/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
#ifndef __Chain__
#  define __Chain__


#include "Memory.hpp"


// *******************Class Chain*******************************

/**
 * Class Chain<C> for implementing double-linked lists
 * @since 03/05/2002 Manchester
 */
template <class C,ClassID ID>
class Chain
{
 public:
  class Link
  {
   private:
    Link* _prev;
    Link* _next;
    C _elem;
   public:
    inline Link (C elem) 
      : _elem (elem) 
      {
      }

    inline Link ()  // for making the header to a chain only
      { 
        _prev = this;
        _next = this; 
      }

    inline Link* prev () const { return _prev; }
    inline void prev ( Link* p ) { _prev = p; }
    inline Link* next () const { return _next; }
    inline void next ( Link* n ) { _next = n; }

    void insertBefore (C elem)
      {
        Link* l = new Link (elem);
        l->_next = this;
        l->_prev = _prev;
        _prev->_next = l;
        _prev = l;
      }

    /**
     * Insert element after this element.
     *
     * @param elem the element to insert.
     */
    void insertAfter (C elem)
      {
        Link* l = new Link (elem);
        l->_prev = this;
        l->_next = _next;
        _next->_prev = l;
        _next = l;
      }

    void del ()
      {
        TRACER( "Chain::del" );
        ASS(_next != this && _prev != this);

        _prev->_next = _next;
        _next->_prev = _prev;
        delete this;
      }

    const C& content () const { return _elem; }
  }; // Class Chain::Link

  // I had to make this public since VC++ ignores my declaration of DelIterator as friend
  inline const Link* header () const { return &_header; }

  // Chain public members
  inline void addLast (C elem)   
    { _header.insertBefore (elem); }

  inline void addFirst (C elem)   
    { _header.insertAfter (elem); }

  inline void delLast ()    
    { _header.prev()->del(); }

  void delFirst ()    
    { _header.next()->del(); }

  inline bool isEmpty () const
    { return _header.prev() == &_header; }

  Link* first () const
    { ASS(! isEmpty());
      return _header.next(); }

  C last () const
    { ASS(! isEmpty());
      return _header.prev(); }


  /**
   * Return the length of the chain.
   *
   * @since 16/07/2003 more() changed to hasNext()
   */
  int length () const
    { int len = 0;
      Iterator links (*this);
      while (links.hasNext()) {
        links.next ();
        len ++;
      }
      return len;
    } // length ()
      

  class Iterator {
   public: 
    inline explicit 
    Iterator (Chain& c) 
      : _header (c.header()),
        _current (_header->next())
        {}
    Iterator (const Chain& c) 
      : _header (c.header()),
        _current (_header->next())
        {}
    
    inline 
    bool hasNext () 
      { return _header != _current; }

    inline 
    C next () 
      { ASS(hasNext());

        C result = _current->content();
        _current = _current->next();

        return result;
      }
    
   private:
    const Link* _header;
    Link* _current;
  };

  class DelIterator {
   public:
    
    inline 
    DelIterator ( Chain& c ) : 
      _header ( c.header() ),
      _current ( _header->next() )
      {}  
    
    inline 
    bool hasNext () 
      { return _header != _current; }

    inline 
    C next () 
      { 
        ASS(_current != _header);
        
        C result = _current->content();
        _current = _current->next();
        return result;
      }
    
    inline
    void del () 
      {
        ASS( _current->prev() != _header ); 

        _current->prev()->del();
      }

    // 07/05/2002 Manchester
    inline
    void replace (C elem) 
      {
        ASS( _current->prev() != _header ); 

        _current->prev()->del();
        _current->insertBefore (elem);
      }

   private:
    const Link* _header;
    Link* _current;
  };

 private:
  Link _header;

 friend class Iterator;
};  // class Chain


#endif


