/* ---------------------------------------------------------------
-- Satnik, a SAT solver                                         --
-- Copyright (c) 2003 Niklas Srensson                          --
-- nik@cs.chalmers.se                                           --
--                                                              --
-- This file is part of Satnik.                                 --
--                                                              --
-- Satnik is free software; you can redistribute it and/or      --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Satnik is distributed in the hope that it will be useful,    --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Satnik; if not, write to the Free         --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
--------------------------------------------------------------- */

#ifndef __CLAUSE__
#define __CLAUSE__

//#include "Satnik.h"
#include <algorithm>

namespace Satnik {

  // The base class for clauses. Contains functionality for activity measures.
  class Clause : public Constraint {
  protected:
    float activity;
    virtual void operator<<(ostream& o) const = 0;
  public:
    Clause(void) : activity(1.0) { }
    void  activate(void) { activity += 1.0; }
    void  decay(float d) { activity *= d; }
    float get_activity(void) const { return activity; }
    bool operator<(const Clause& c) const { return activity > c.activity; }

    friend ostream& operator<<(ostream& o, const Clause& c) { c << o; return o; }

    virtual bool locked(Solver& s) = 0;
    virtual void copy(vector<Lit>& clause) = 0;
  };

  // The implementation of basic Clauses. It is instantiable by the
  // internal literal type, allowing for more compact representation.
  template<typename T>
  class PackedClause : public Clause {
    size_t size;
    T      lits[0];

    PackedClause(const vector<Lit>& clause) {
      for(size_t i = 0; i < clause.size(); i++) lits[i] = index(clause[i]);
      size = clause.size();
    }
  public:
    static PackedClause* create(const vector<Lit>& clause) {
      PackedClause* c = (PackedClause*)malloc(clause.size() * sizeof(T) + sizeof(PackedClause));
      new (c) PackedClause(clause);
      return c;
    }

    void operator<<(ostream& o) const{
      for(size_t i = 0; i < size; i++) o << encode(lits[i]) << ' ';
      o << "0\n"; }

    void attach(Solver& s) { 
      s.add_watch(neg(encode(lits[0])), this); 
      s.add_watch(neg(encode(lits[1])), this); }
    void detach(Solver& s) { 
      s.del_watch(neg(encode(lits[0])), this); 
      s.del_watch(neg(encode(lits[1])), this); }

    bool locked(Solver& s) { 
      return s.values(encode(lits[0])) && s.reason(var(encode(lits[0]))) == this; }

    bool propagate (Lit l, Solver& s);
    void reason    (Lit l, vector<Lit>& r); 
    bool satisfied (Solver& s);
    void copy(vector<Lit>& clause);

  };

  template<typename T>
  void PackedClause<T>::reason   (Lit l, vector<Lit>& r) {
    //activate();
    for(size_t i = 0; i < size; i++)
      if(lits[i] != (T)index(l)) r.push_back(neg(encode(lits[i]))); 
  }

  template<typename T>
  bool PackedClause<T>::satisfied (Solver& s) {
    for(size_t i = 0; i < size; i++)
      if(s.values(encode(lits[i]))) return true;
    return false;
  }

  template<typename T>
  void PackedClause<T>::copy(vector<Lit>& clause) {
    for(size_t i = 0; i < size; i++) clause.push_back(encode(lits[i])); }
    

  template<typename T>
  bool PackedClause<T>::propagate(Lit l, Solver& s) { 
    if (lits[0] == (T)index(neg(l))) swap(lits[0],lits[1]);
    assert(lits[1] == (T)index(neg(l)));
    if(s.values(encode(lits[0]))) { s.add_watch(l, this); return true; }
    
    for (size_t i = 2; i < size; i++)
      if (!s.values(neg(encode(lits[i])))){ 
	swap(lits[1],lits[i]); 
	s.add_watch(neg(encode(lits[1])), this);  
	return true; }
    
    activate();
    s.add_watch(l, this);
    return s.assign(encode(lits[0]), this);
  }

  // A special representation of binary clauses. This does not seem to
  // have a consistently positive effect.
  class BinaryClause : public Clause {
    Lit x,y;
    BinaryClause(Lit a, Lit b) : x(a), y(b) { }
  public:
    static BinaryClause* create(Lit a, Lit b) {
      BinaryClause* c = (BinaryClause*)malloc(sizeof(BinaryClause));
      new (c) BinaryClause(a,b);
      return c;
    }
    void operator<<(ostream& o) const{ o << x << y << " 0\n"; }
    void attach(Solver& s) { s.add_watch(neg(x), this); s.add_watch(neg(y), this); }
    void detach(Solver& s) { s.del_watch(neg(x), this); s.del_watch(neg(y), this); }

    bool locked(Solver& s) { return s.values(x) && s.reason(var(x)) == this ||
			            s.values(y) && s.reason(var(y)) == this; }

    bool propagate (Lit l, Solver& s) {
      assert(l == neg(x) || l == neg(y));
      activate();
      s.add_watch(l, this); return  s.assign(l == neg(x) ? y : x,this); }

    void reason    (Lit l, vector<Lit>& r) { 
      assert(l == x || l == y);
      //activate();
      r.push_back(neg(l == x ? y : x));
    }

    bool satisfied (Solver& s) { return s.values(x) || s.values(y); }

    void copy(vector<Lit>& clause) { 
      clause.push_back(x); clause.push_back(y); }

  };

};

#endif
