/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/**
 * File Clause.hpp.
 * Defines class Clause.
 */


#ifndef __Clause__
#define __Clause__


#include "Literal.hpp"


class Clause 
{
 private:
  class Data;

 public:
  // constructor
  Clause ();
  explicit Clause (const LiteralList& lits);
  Clause (const Clause& t);
  Clause (const VampireKernel::Clause*, const VampireKernel& kernel);
  ~Clause ();
  void operator= (const Clause& rhs);

  // structure access
  const LiteralList& literals () const;

  //miscellaneous
  static bool lessThan (const Clause& c1, const Clause& c2);   
  Compare compare (Clause l) const; // comparison, used for normalisation
  void normalize (Clause& result) const;
  // properties
  bool isEqualityAxiom () const;  // is one of the equality axioms
  bool isEmpty () const;
  bool isFunctionDefinition (Term& lhs, Term& rhs) const; 
  bool isRenamingOf (Clause c) const;
 private:
  // structure
  Data* _data;

  // equality axiom check
  bool isReflexivityAxiom () const;
  bool isFunctionReflexivityAxiom () const;
  bool isPredicateReflexivityAxiom () const;
  bool isSymmetryAxiom () const;
  bool isTransitivityAxiom () const;
}; // class Clause


class Clause::Data
#   if DEBUG_PREPRO
    : public Memory <CID_CLAUSE>
#   endif
{
 public:
  Data ();
  explicit Data (const LiteralList& literals);

  // structure access
  const LiteralList& literals () const;

  void ref ();
  void deref ();

 private:
  // structure
  int _counter;
  LiteralList _literals;
}; // class Clause::Data


class ClauseList 
: public Lst<Clause>
{
 public:
  // constructors
  ClauseList ();
  ClauseList (const ClauseList&);
  explicit ClauseList (const Clause& t); // one-element list
  ClauseList (const Clause& head, const ClauseList& tail);
  explicit ClauseList (LstData<Clause>*);

  // structure access
  const LiteralList& literals ();

  // inherited functions
  const ClauseList& tail () const
    { return static_cast<const ClauseList&>(Lst<Clause>::tail()); }
}; // class ClauseList


ostream& operator << (ostream&, const Clause& c);


// ******************* class Clause, implementation *********************


inline
Clause::Clause () 
  : 
  _data (0) 
{
} // Clause::Clause


inline
Clause::~Clause () 
{
  if (_data) {
    _data->deref ();
  }
} // Clause::~Clause


inline
Clause::Clause (const Clause& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} // Clause::Clause


inline
Clause::Clause (const LiteralList& literals)
  :
  _data (new Data(literals))
{
} // Clause::Clause (const LiteralList&)


inline
const LiteralList& Clause::literals () const
{
  return _data->literals();
} // Clause::literals ()


inline
bool Clause::isEmpty () const
{ 
  return literals().isEmpty(); 
} // Clause::isEmpty ()


// true is this clause is a renaming of c
// 03/10/2002 Manchester
inline
bool Clause::isRenamingOf (Clause c) const
{
  return literals().isRenamingOf (c.literals());
} // Clause::isRenamingOf


// **************** class Clause::Data implementation ******************


inline
Clause::Data::Data (const LiteralList& literals) 
  : 
  _counter (1),
  _literals (literals)
{
} // Clause::Data::Data


inline
const LiteralList& Clause::Data::literals () const
{
  return _literals;
} // Clause::Data::literals ()


inline
void Clause::Data::ref () 
{ 
  ASS (this);

  _counter++;
} // Clause::Data::ref ()


inline
void Clause::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;

  if (_counter == 0) {
    delete this;
  }
} // Clause::Data::deref ()


// ******************* ClauseList definitions ************************

inline
ClauseList::ClauseList () 
  : 
  Lst<Clause> ()
{
} // ClauseList::ClauseList


inline
ClauseList::ClauseList (const ClauseList& ts)
  :
  Lst<Clause> (ts)
{
} // ClauseList::ClauseList


inline
ClauseList::ClauseList (LstData<Clause>* d)
  :
  Lst<Clause> (d)
{
} // ClauseList::ClauseList


inline
ClauseList::ClauseList (const Clause &hd, const ClauseList& tl)
  :
  Lst<Clause> (hd,tl)
{
} // ClauseList::ClauseList


inline
ClauseList::ClauseList (const Clause &hd)
  :
  Lst<Clause> (hd)
{
} // ClauseList::ClauseList



#endif // __Clause__
