/*				FINDER.c
*
*  This is the top file for FINDER, except for Fmain.c where the
*  function main() lives. When FINDER is used as an external utility
*  by other programs, this file will be linked but Fmain.c will not.
*/


#define MAINFILE
#include "FINDER.h"


#ifdef SOFTIE_SCOTT
#include "SOFTIE_SCOTT.h"
#endif


/*
*  Found first tests the given clause in the current model.
*  If the clause is true in the model and add_if_true is set, then it
*  is added to the current theory if add_if_true is true.
*  If the clause is not true in the current model and seek_model is
*  set, then a model of the current theory plus the given clause is
*  sought. If one is found then it becomes the current model and the
*  given clause is added to the current theory.
*
*  The value returned is in every case the truth value of the clause
*  in the (possibly new) current model.
*
*  Note that the timer is reset on every call. This allows a time limit
*  to be set for each individual search.
*/	

int found( const dynstring *s, int add_if_true, int seek_model )
{
  int i, oldgood, rc;
  boolean update = false;
  functor f;
  void short_timesup();

  if (no_semantics) return 0;

  signal( SIGALRM, short_timesup );

  if (seek_model) add_if_true = 1;         /* Sanity condition */

  /*
   * Now set things up so that s is the "latest clause" and the last
   * good model is ready for testing, and test the clause.
   */

  normal_input = false;
  finder_input(s);

  for ( i = 0; sort[i].sname; i++ )
    sort[i].siz = last_sizes[current_theory][i];
  for ( f = FN; f; f = f->nextfunct ) {
    f->fcc = f->old_fcc[current_theory];
    set_fats(f);
  }

  rc = test(last_model[current_theory], 0);

  /*
   * If necessary, try to improve the model to accommodate the
   * new clause. This may require other actions such as re-testing
   * clauses (e.g. re-testing the set of support). Note that if the
   * "latest clause" is too long, no new model will be sought.
   */

  if (clause_is_long() && latest_clause.softweight) {
    seek_model = 0;
    add_if_true = 0;
  }

  if ( seek_model && !rc ) {
    tr_parameters.done = 0;
    tr_parameters.reps[NULL_BACKTRACKS].i_report = 0;
    tr_parameters.reps[SURJ_BACKTRACKS].i_report = 0;
    tr_parameters.reps[SOLUTIONS].i_report = 0;
    tr_parameters.reps[BAD_TESTS].i_report = 0;
    tr_parameters.reps[INVALUES].i_report = 0;
    oldgood = updates[current_theory];
    tr_parameters.max_penalty = theBadness + latest_clause.badness;

    alarm( setting.maxtime );
    begin_time = timestamp();
    go(0);
    end_time = timestamp();
    alarm(0);

    if ( updates[current_theory] > oldgood ) {
      rc = found(s,0,0);
      update = true;
      model_clauses = size[current_theory];
    }
  }

  if ( (rc || latest_clause.softweight) && add_if_true )
    add_to_theory();
  normal_input = true;

  return (rc);
}



/*
 * At present the condition for the clause to be "long" is rather
 * crude. Later it can be made more refined.
 */

boolean clause_is_long()
{
#ifdef SOFTIE_SCOTT
  return (latest_clause.weight > SHORT_CLAUSE_CUTOFF);
#else
  return (latest_clause.weight + latest_clause.varcount > 7);
#endif
}


/*
*	Init_finder must be called before much else happens if FINDER
*	is to be used as a supertester. It reads the input file and 
*	returns a boolean (integer) value recording whether it found 
*	any model of the input theory. If the return value is false 
*	then there will be no "current theory", and adding clauses may 
*	be a bad idea.
*/

boolean init_finder(char s[])
{
  FILE *if_fil;
  int i;

  void short_timesup();

  CLOCK_START(FINDER_TIME);
  
  signal( SIGALRM, short_timesup );
  initialize();
  if ( !(if_fil = fopen(s,"r")) ) inerror("Unable to open file",s);
  read_job( if_fil );
  fclose( if_fil );
  tr_parameters.done = 0;
  for (i=0; i<MAXISET_ARRAY; i++) updates[i] = 0;
  single_model = true;
  Integer->piped = true;
  if ( setting.pm.verbosity & pjob ) print_job();
  current_theory = 0;
  begin_time = timestamp();
  alarm( setting.maxtime * 5 );
  go(0);
  end_time = timestamp();
  alarm(0);

  CLOCK_STOP(FINDER_TIME);
  return (updates[0] > 0);
}



void init_init_finder()
{
  no_semantics = false;
  init_buckets();
  updates[0] = 0;
}





/*
*	Go is the main work loop of FINDER. It calls transref to search
*	every search space, defined by setting the cardinalities of the
*	sorts, within the range stipulated. Go is called from the above 
*	functions, and from itself, and also from main().
*/


static int old_updates;

void go(int x)
{
  if (bug_out) return;

  if ( negative && !clauses_fail ) return;
  if (!x) old_updates = updates[current_theory];
  if ( !sort[x].sname ) {
    if (set_up_cc()) {
      firstmodel = true; // perhaps need to change this?
      transref(&tr_parameters);
    }
    if (bug_out) return;
  }
  else if ( sort[x].piped && !tr_parameters.done ) go(x+1);
  else for ( sort[x].siz = sort[x].sizmin;
             sort[x].siz <= sort[x].sizmax && !tr_parameters.done;
             sort[x].siz++ ) {
    if (updates[current_theory]==old_updates || !setting.sortcut)
      go(x+1);
  }
}



/*
*	FINDER can be used to return models falsifying a badguy.
*/

void refute_current()
{
  cc_badguy = true;
}




/*
*	It is sometimes useful for a calling program to know how many 
*	clauses there are in a theory.
*/

int theory_size(int x)
{
  int i = 0;
  clause c;

  for ( c = TH[x]; c; c = c->nextclause ) i++;
  return i;
}


/*
*	The final trivial function in this section allows a calling
*	program to add FINDER input one line at a time.
*/

void finder_input( const dynstring *s )
{
  cmd_line_input = true;
  input_point = 0;
  dynstring_strcpy( &command_line, s );
  dynstring_strcatc( &command_line, " end.");
  read_job(0);
}



void short_timesup()
{
  job_done( &tr_parameters, "time limit" );
}


boolean initial_model_found()
{
  return (updates[0] > 0);
}


void set_no_semantics()
{
  no_semantics = true;
}
