------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Flags
  ( Flags(..)
  , Print(..)
  , Priority(..)
  , getFlags
  , getTimeLeft
  , getTimeSpent
  , getTimeSpentS
  )
 where

import System.Environment
  ( getArgs
  )

import System

import List
  ( groupBy
  , intersperse
  , (\\)
  )

import Char

import CPUTime

-------------------------------------------------------------------------
-- flags

data Flags
  = Flags
  { files     :: [FilePath]
  , roots     :: [FilePath]
  , mfile     :: Maybe FilePath
  , cfile     :: Maybe FilePath
  , sizes     :: [Int]
  , weakening :: Int
  , prints    :: [Print]
  , dimacs    :: Maybe FilePath
  , start     :: Integer
  , total     :: Maybe Int
  , pretend   :: Bool
  , priority  :: Priority
  , kanin     :: Bool
  , contr     :: Bool
  }
 deriving (Eq, Show)

data Print
  = Original
  | Flattened
  | Instantiated
  | Types
  | Progress
  | Model
 deriving (Eq, Show, Read, Enum)

data Priority
  = No
  | PredConFun
  | ConPredFun
  | ConFunPred
 deriving (Eq, Show, Read, Enum)

secret :: [Print]
secret = []
-- secret = [ Flattened, Instantiated ]

initFlags :: Flags
initFlags =
  Flags
  { files     = []
  , roots     = []
  , mfile     = Nothing
  , cfile     = Nothing
  , sizes     = []
  , weakening = 0
  , prints    = []
  , dimacs    = Nothing
  , start     = error "starting time not properly initialized"
  , total     = Nothing
  , pretend   = False
  , priority  = No
  , kanin     = True
  , contr     = False
  }

-------------------------------------------------------------------------
-- getFlags

getFlags :: IO Flags
getFlags =
  do as    <- getArgs
     picoT <- getCPUTime
     case parseFlags initFlags as of
       Left [] ->
         do putStr (unlines helpMessage)
            exitWith ExitSuccess

       Left err ->
         do putStrLn "Error in arguments:"
            putStr (unlines err)
            putStrLn "Try --help."
            exitWith (ExitFailure (-1))
       
       Right f ->
         do return f{ start = unPico picoT }

getTimeLeft :: Flags -> IO Int
getTimeLeft flags =
  do t' <- getTimeSpent flags
     return (t - t')
 where
  t = maybe 300 id (total flags)

getTimeSpent :: Flags -> IO Int
getTimeSpent flags =
  do picoT <- getCPUTime
     return (fromInteger (unPico picoT - start flags))

getTimeSpentS :: Flags -> IO String
getTimeSpentS flags =
  do picoT <- getCPUTime
     let t = fromInteger (unPico' picoT - start flags)
         m = t `div` 600
         s = (t `div` 10) `mod` 60
         d = t `mod` 10
     return ( show m
           ++ ":"
           ++ let x = show s in replicate (2 - length x) '0' ++ x
           ++ "."
           ++ show d
            )

unPico :: Integer {- picoseconds -} -> Integer {- seconds -}
unPico = let c = 10^12 in (`div` c)

unPico' :: Integer {- picoseconds -} -> Integer {- 0.1 seconds -}
unPico' = let c = 10^11 in (`div` c)

-------------------------------------------------------------------------
-- options

data Option a
  = Option
  { long    :: String
  , meaning :: Arg (a -> a)
  , help    :: [String]
  }

options :: [Option Flags]
options =
  [ Option
    { long    = "sizes"
    , meaning = addSizes <$> argNums
    , help    = [ "A list of domain sizes to check. The format of the"
                , "list consists of numbers n, or intervals n..m,"
                , "separated by commas."
                , "Example: --sizes 2,4,8..12"
                , "Default: --sizes 1..n  (where n is 'reasonable')"
                ]
    }

{-
  , Option
    { long    = "time"
    , meaning = updTime <$> argNum
    , help    = [ "Maximum running time in seconds."
                , "Affects the internal behavior. Is a soft limit."
                , "Example: --time 300"
                , "Default: (off)"
                ]
    }
-}

  , Option
    { long    = "print"
    , meaning = addPrints <$> argList (map show ([Original .. Model] \\ secret))
    , help    = [ "Specify information that should be printed."
                , "Example: --print Original,Model"
                , "Default: --print ''"
                ]
    }

  , Option
    { long    = "model"
    , meaning = addMfile <$> argFile
    , help    = [ "Produce a file containing the model that was found."
                , "Example: --model solution.mod"
                , "Default: (off)"
                ]
    }

  , Option
    { long    = "check"
    , meaning = addCfile <$> argFile
    , help    = [ "Produce a file containing the original problem plus"
                , "extra clauses representing the model that was found."
                , "Example: --check solution.p"
                , "Default: (off)"
                ]
    }

  , Option
    { long    = "dimacs"
    , meaning = addDimacs <$> argFile
    , help    = [ "Produce a CNF file in DIMACS format for each"
                , "domain size. A '%d' in the file name will be"
                , "replaced by the size of the model; a '%f' in the"
                , "filename will be replaced by the name of the problem."
                , "Example: --dimacs /tmp/%f-%d.cnf"
                , "Default: (off)"
                ]
    }

  , Option
    { long    = "pretend"
    , meaning = unit setPretend
    , help    = [ "Pretend that each size has no model. This is"
                , "useful when benchmarking the clause generation"
                , "or when generating DIMACS files."
                , "Default: (off)"
                ]
    }

  , Option
    { long    = "priority"
    , meaning = setPrio <$> argOption (map show ([No .. ConFunPred]))
    , help    = [ "Influences the variable ordering in the SAT solver,"
                , "for Constants, Functions, and Predicates, in that order."
                , "Default: --priority No"
                ]
    }

{-
  , Option
    { long    = "kanin"
    , meaning = unit setKanin
    , help    = [ "(Experimental) Removes the big symmetry-triangle from"
                , "the problem."
                , "Default: (off)"
                ]
    }
-}

  , Option
    { long    = "contr"
    , meaning = unit setContr
    , help    = [ "Check if each instance is contradictory before assumption."
                , "Default: (off)"
                ]
    }

  , Option
    { long    = "root"
    , meaning = addRoot <$> argFile
    , help    = [ "A directory in which included problems will be sought."
                , "Example: --root TPTP-v2.5.0"
                , "Default: --root ."
                ]
    }

  , Option
    { long    = "help"
    , meaning = unit id
    , help    = [ "Displays this help message."
                ]
    }
  ]

-------------------------------------------------------------------------
-- arg

data Arg a = MkArg [String] ([String] -> Either [String] (a, [String]))

instance Functor (Either a) where
  fmap f (Left x)  = Left x
  fmap f (Right y) = Right (f y)

unit :: a -> Arg a
unit x = MkArg [] (\s -> Right (x,s))

(<*>) :: Arg (a -> b) -> Arg a -> Arg b
MkArg fs f <*> MkArg xs x =
  MkArg (fs++xs) (\s ->
    case f s of
      Left err     -> Left err
      Right (h,s') -> case x s' of
                        Left err      -> Left err
                        Right (a,s'') -> Right (h a,s'')
  )

(<$>) :: (a -> b) -> Arg a -> Arg b
f <$> x = unit f <*> x

args :: Arg a -> [String]
args (MkArg as _) = as

-------------------------------------------------------------------------
-- parsers

argNum :: (Read a, Num a) => Arg a
argNum = MkArg ["<num>"] $ \xs ->
  case xs of
    x:xs       | all isDigit x -> Right (read x, xs)
    ('-':x):xs | all isDigit x -> Right (-read x, xs)
    _                          -> Left ["expected a number"]
      
argFile :: Arg FilePath
argFile = MkArg ["<file>"] $ \xs ->
  case xs of
    x:xs -> Right (x, xs)
    _    -> Left ["expected a file"]
      
argNums :: Arg [Int]
argNums = MkArg ["<nums>"] $ \xs ->
  case xs of
    []   -> Left ["expected a number list"]
    x:xs -> ((\a -> (a,xs)) `fmap`) . nums . groupBy (\x y -> isDigit x == isDigit y) $ x ++ ","
     where
      nums []                = Right []
      nums (n:",":ns)        = (read n :) `fmap` nums ns
      nums (n:"..":m:",":ns) = ([read n .. read m] ++) `fmap` nums ns
      nums _                 = Left ["number list garbled"]

argOption :: [String] -> Arg String
argOption as = MkArg ["<" ++ concat (intersperse " | " as) ++ ">"] $ \xs ->
  case xs of
    []   -> Left ["expected an argument"]
    x:xs -> ((\a -> (a,xs)) `fmap`) . elts $ x
     where
      elts x | x `elem` as = Right x
             | otherwise   = Left ["argument garbled"]

argList :: [String] -> Arg [String]
argList as = MkArg ["<" ++ concat (intersperse " | " as) ++ ">*"] $ \xs ->
  case xs of
    []   -> Left ["expected a list"]
    x:xs -> ((\a -> (a,xs)) `fmap`) . elts $ x ++ ","
     where
      elts []              = Right []
      elts s | w `elem` as = (w:) `fmap` elts r
       where
        w = takeWhile (/= ',') s
        r = tail (dropWhile (/= ',') s)
    
      elts _ = Left ["argument list garbled"]

parseFlags :: Flags -> [String] -> Either [String] Flags
parseFlags f []               = Right f
parseFlags f ("--help":xs)    = Left []
parseFlags f (('-':'-':x):xs) =
  case [ opt | opt <- options, x == long opt ] of
    opt:_  -> case h xs of
                Left err      -> Left err
                Right (g,xs') -> parseFlags (g f) xs'
     where
      MkArg _ h = meaning opt
    []     -> Left ["Unrecognized option: '--" ++ x ++ "'"]

parseFlags f (x:xs)           = parseFlags (f{files = files f ++ [x]}) xs

-------------------------------------------------------------------------
-- helpers

addSizes :: [Int] -> Flags -> Flags
addSizes ns f = f{ sizes = sizes f ++ ns }

updTime :: Int -> Flags -> Flags
updTime n f = f{ total = Just n }

setPretend :: Flags -> Flags
setPretend f = f{ pretend = True }

updWeaken :: Int -> Flags -> Flags
updWeaken n f = f{ weakening = n }

addPrints :: [String] -> Flags -> Flags
addPrints ps f = f{ prints = prints f ++ map read ps }

setPrio :: String -> Flags -> Flags
setPrio a f = f{ priority = read a }

setKanin :: Flags -> Flags
setKanin f = f{ kanin = False }

setContr :: Flags -> Flags
setContr f = f{ contr = True }

addRoot :: FilePath -> Flags -> Flags
addRoot r f = f{ roots = roots f ++ [r] }

addDimacs :: FilePath -> Flags -> Flags
addDimacs s f = f{ dimacs = Just s }

addMfile :: FilePath -> Flags -> Flags
addMfile s f = f{ mfile = Just s }

addCfile :: FilePath -> Flags -> Flags
addCfile s f = f{ cfile = Just s }

-------------------------------------------------------------------------
-- help message

helpMessage :: [String]
helpMessage =
  [ "Usage: <option>* <file>*"
  , ""
  , "<file> should be in any of the following formats:"
  , "  - TPTP format; 'include' is understood."
  , "  - Otter format; 'set(prolog_style_variables)' is required."
  , ""
  , "<option> can be any of the following:"
  ] ++
  concat
  [ [ ""
    , "  --" ++ long opt ++ " " ++ unwords (args (meaning opt))
    ] ++ map ("    "++) (help opt)
  | opt <- options
  ]

-------------------------------------------------------------------------
-- the end.





