;======================================================================
;
;                          resprop.scm
;
;                   Gandalf version c-2.6 r1
;
;  Propositional prover & finite satisfiability checker.
;
;                  Copyright (c) Tanel Tammet
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 1, or (at your option)
; any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; The author can be reached at tammet@staff.ttu.ee, tanel@cs.ioc.ee or
; tammet@cs.chalmers.se
;
;
;=======================================================================
;
;
;
; Tasks:
; 1. Satisfiability proving on finite domains: find a finite domain where all sat.
;    Funs with fixed args are fully interpreted to selected constants.
;
; 2. Unsatisfiability proving by instantiating on one domain (initially
;    function-free) and finding a contradiction for that instantiation.
;
; 3. Unsatisfiability proving by instantiating on one domain with functions
;    where functions are interpreted either to new constants or according
;    to the equality units in the formula and finding a contradiction
;    for that instantiation.
;
; 4. Highly incomplete unsatisfiability proving by instantiating by giving all vars
;    one concrete constant value and finding a contradiction
;    for that instantiation.
;
;
;
; Algorithm:
;
; run over all existing clauses, starting with shorter and smaller, for each clause CL do:
;    for each var X in CL do:
;      for all the values V in *prop-constants* do:
;         give X value V
;         for each fully instantiated clause CLI do:
;             compute the values of funs by using mapping *prop-fun-mappings*,
;               generate clause CLIV
;             run over all the literals and find their value using mapping
;               *prop-lit-mappings* and checking:
;                       if value is #t, discard full clause
;                       if value is #f, discard this particular literal
;                          if all lits discarded, contradiction found
;                       else insert value nr either pos or neg
;
; generate a list of vectors of numbers, each having
; a first nr indicating an id, sec len, rest lits.
; generate a vector of lits, each indicating a value of a lit #t, #f or null.
; start binary ordered resolution, preferring smaller clauses. If horn, use unit res.
; If a unit lit is generated, store the value in the value vector.
; generate a subsumption tree
;
;
; Finite domain construction:
;
;  Cycle over domain sizes 1, 2, 3, etc:
;     For each constant symb and fun symb build all mappings for selected domain
;       Select one mapping for each const and fun, run over all such selections:
;           check satisfiability of the mapping and domain.
;           if satisfibale: we are finished, sat proved
;           if not satisfiable: continue.


;----------------------------------------------------------------------
;
;                          compiler defs
;
;----------------------------------------------------------------------

;(define compile-inline
;  '(litinf-left-para? litinf-right-para? litinf-parainto?
;    pos-eq-pred? pos-eq-fun? litinf-binrespara?
;    eq-lit-arg1 eq-lit-arg2 eq-lit-type may-be-demodulated?
;    litinf-use make-binres-history))

;----------------------------------------------------------------------
;
;                 global data
;
;----------------------------------------------------------------------


(define *partial-model-check* #f)

(define *prop-clauses-generation-limit* #f)
(define *prop-clauses-generation-limited-flag* #f)

;Original pred-level data lists:

(define *prop-orig-constants* '()) ; list of constants
(define *prop-orig-funs* '())  ; alist with funs and arities: ((f . 2)(g . 1))
(define *prop-orig-preds* '()) ; alist with preds and arities: ((p . 2)(r . 0))

(define *prop-flat-eq-symb* '()) ; enumeration num corresponding to equality

;Prop-level main lists:

(define *prop-constants* '()) ; selection from prop-orig-constants

(define *prop-constants-mapping* '()) ; alist mapping from orig constants to new ones

(define *prop-fun-mappings* '()) ; for all funs (f,g) and arg trees
				 ; consisting of prop-constants:
	;   ( (f (1 (1 . val)
	;           (2 . val)
	;           (3 . val) )
	;        (2 (1 . val)
	;           (2 . val)
	;           (3 . val) )
	;        (3 (1 . val)
	;           (2 . val)
	;           (3 . val) ))
	;     (g  ...) )

(define *prop-lit-mappings* '()) ; analogous to *prop-fun-mappings*,
				 ; val may be nr, #t or #f

(define *prop-atom-maxnr* 10000) ; enumerates all prop atoms

(define *prop-interp-value-pairs* '()) ; a list of pairs where each cdr should
				       ; replaced with a value from domain

(define *prop-interp-termlist* '()) ; a list of terms which will be interpreted

(define *tmp-varlist* '())

(define *prop-orig-constants-enum* '())
(define *prop-orig-funs-enum* '())
(define *prop-orig-preds-enum* '())
(define *prop-last-enum-value* 0)

(define *propclauses* '())
(define *propclause-tree* '())
(define *prop-time-ran-out* #f)

;----------------------------------------------------------------------
;
;       top level of satisfiability checking with model building
;
;----------------------------------------------------------------------

; prop-build-simple-model returns #t iff satisfiable, #f if building
; failed for any reason

(define (prop-build-simple-model predclauses)
  (let ((stopflag #f)
        (domainsize 0)
        (tmpnr 0)
        (res #f) )
    (set! *prop-clauses-generation-limit* #f)
    (set! *prop-clauses-generation-limited-flag* #f)
    (set! *prop-time-ran-out* #f)
    (set! *given-checkpoint-keptnrs* '())
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)
    (set! predclauses (prop-uncurry-clauses predclauses))
    (extract-pred-data-to-prop predclauses)

    ;(for-each (lambda(x)(newline)(display x)) predclauses)
    ;(display "prop-orig-funs: ")(display *prop-orig-funs*)(newline)

    (cond
      ((null? *prop-orig-funs*)
      ; no function symbols: interpret all constants to themselves
            (set! domainsize (length *prop-orig-constants*))
            (if (zero? domainsize) (set! domainsize 1))
            (if (> *display-info-level* 30)
               (begin ; (newline)
	              (display "no function symbols, starting search for domain size ")
	              (display domainsize)
		      (newline) ))
                ; last arg #t iff const are interp
           (set! *prop-interp-value-pairs* '())
           (make-prop-interpretation-lists domainsize #t #f)
           ;(display *prop-interp-value-pairs* )
           (create-prop-var-vectors *prop-atom-maxnr*)
           ;(display *prop-interp-value-pairs* )
           ;(abort)
           ; now try to solve
           (set! res (check-prop-one-interpretation predclauses domainsize))
           (if *prop-time-ran-out*
    	       (begin
	               (if (> *display-info-level* 38)
           	           (begin (newline)
	                          (display "allocated model search time has run out")
		                  (newline) ))
	                #f)
               res))
      ; function symbols present: full domain size iteration
     (else
    	(do ((domainsize 1 (+ 1 domainsize)))
            ((or stopflag
                 res
                 (> (- (get-check-time 1) *slow-start-time*)
	            *one-pass-time*))
	        ;(if (and (not stopflag)(not res))
	        ;    (store-time-checkpoint! 1) )
	        (if *prop-time-ran-out*
    	            (begin
	               (if (> *display-info-level* 38)
           	           (begin (newline)
	                          (display "allocated model search time has run out")
		                  (newline) ))
	                #f)
                    res))
           (if (> *display-info-level* 28)
               (begin (newline)
	              (display "starting search for domain size ")
	              (display domainsize)
		      (newline) ))

           (make-prop-interpretation-lists domainsize #t #f) ;
           (create-prop-var-vectors *prop-atom-maxnr*)

           (if *prop-debug*
            (begin
             (display "*prop-interp-value-pairs*: ")(newline)
             (display *prop-interp-value-pairs*)(newline)
             (display "*prop-constants-mapping*: ")(newline)
             (display *prop-constants-mapping*)(newline)
             (display "*prop-fun-mappings*: ") (newline)
             (display *prop-fun-mappings*)(newline)
             (display "*prop-lit-mappings*: ") (newline)
             (display *prop-lit-mappings*)(newline)
             (display "*prop-atom-maxnr*: ")(newline)
             (display *prop-atom-maxnr*)(newline) ))

           (set! res (check-prop-interpretations predclauses domainsize 0)) )))))

            ; (if (> domainsize 2) (set! stopflag #t)) ))))

(define *max-var-stat-pre* 0)
(define *max-var-stat-post* 0)

(define (prop-build-flatten-model predclauses)
  (let ((stopflag #f)
        (domainsize 0)
        (flatclauses '())
        (tmpnr 0)
        (res #f) )

    (newline)
    (display "mace-style model building with nonground splitting starts.")
    (newline)

    (set! *prop-clauses-generation-limit* #f)
    (set! *prop-clauses-generation-limited-flag* #f)
    ;(prop-build-incremental-model  predclauses)
    ;(prop-build-simple-model  predclauses)
    ;(quit)

    (set! *prop-conversion-to-file* #t)

    ; first create unique file names

    (set! *prop-conversion-file-name*
          (string-append *prop-conversion-file-name-start* *run-id*) )
    (set! *prop-conversion-temp-file-name*
          (string-append *prop-conversion-temp-file-name-start* *run-id*) )
    (set! *prop-conversion-header-file-name*
           (string-append *prop-conversion-header-file-name-start* *run-id*) )
    (set! *prop-conversion-result-file-name*
           (string-append *prop-conversion-result-file-name-start* *run-id*) )

    (set! *max-var-stat-pre* 0)
    (set! *max-var-stat-post* 0)

    ; (set! *prop-atom-maxnr* *prop-last-enum-value*)
    (set! *prop-time-ran-out* #f)
    (set! *given-checkpoint-keptnrs* '())
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)

    ;(newline)
    ;(display "orig clauses before extracting pred data to prop: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) predclauses)
    ;(newline)

    (set! predclauses (prop-uncurry-clauses predclauses))

    ;(newline)
    ;(display "orig clauses before renaming ground: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) predclauses)
    ;(newline)

    (set! predclauses (prop-groundprocess-clauses predclauses))

    ;(newline)
    ;(display "orig clauses after renaming ground: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) predclauses)
    ;(newline)

    ;(newline)
    ;(display "*prop-orig-preds*:")(newline)
    ;(display *prop-orig-preds*)(newline)

    (extract-pred-data-to-prop predclauses)

    ;(newline)
    ;(display "orig clauses before flattening: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) predclauses)
    ;(newline)

    (set! flatclauses (flatten-clause-set predclauses))

    ;(newline)
    ;(display "*prop-orig-preds*:")(newline)
    ;(display *prop-orig-preds*)(newline)

    (set! *flat-split-literal-nr* 10000) ; new predicates from splitting

    ;(newline)
    ;(display "orig clauses before splitting: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) flatclauses)
    ;(newline)

    ;(set! *max-var-stat-pre* (compute-max-vars-in-clauses flatclauses))

    (set! flatclauses (split-flat-clause-set flatclauses))

    ;(set! *max-var-stat-post* (compute-max-vars-in-clauses flatclauses))

    ;(newline)
    ;(display "*prop-orig-preds*:")(newline)
    ;(display *prop-orig-preds*)(newline)

    ;(display "orig clauses after splitting: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) flatclauses)
    ;(newline)

    (set! *prop-clauses-generation-limit* #f)
    (set! *prop-clauses-generation-limited-flag* #f)

        ; full domain size iteration
    	(do ((domainsize 1 (+ 1 domainsize)))
            ((or stopflag
                 res
                 (> (- (get-check-time 1) *slow-start-time*)
	            *one-pass-time*))
	        ;(if (and (not stopflag)(not res))
	        ;    (store-time-checkpoint! 1) )

	        (set! *prop-conversion-to-file* #f)
	        ; remove temporary files
	        (system (string-append "rm " *prop-conversion-file-name*))
	        (system (string-append "rm " *prop-conversion-temp-file-name*))
	        (system (string-append "rm " *prop-conversion-header-file-name*))
	        (system (string-append "rm " *prop-conversion-result-file-name*))

	        ; check res
	        (if *prop-time-ran-out*
    	            (begin
	               (if (> *display-info-level* 38)
           	           (begin (newline)
	                          (display "allocated model search time has run out")
		                  (newline) ))
	                #f)
                    res))
           (if (> *display-info-level* 28)
               (begin (newline)
	              (display "starting search for domain size ")
	              (display domainsize)
		      (newline) ))

           (make-prop-interpretation-lists domainsize #f #t) ;

           (create-prop-var-vectors *prop-atom-maxnr*)

           (if *prop-debug*
            (begin
             (display "*prop-interp-value-pairs*: ")(newline)
             ;(display *prop-interp-value-pairs*)(newline)
             (for-each (lambda(x)(display x)(newline)) *prop-interp-value-pairs*)
             (display "*prop-constants-mapping*: ")(newline)
             (display *prop-constants-mapping*)(newline)
             (display "*prop-fun-mappings*: ") (newline)
             ;(display *prop-fun-mappings*)(newline)
             (for-each (lambda(x)(display x)(newline)) *prop-fun-mappings*)
             (display "*prop-lit-mappings*: ") (newline)
             ;(display *prop-lit-mappings*)(newline)
             (for-each (lambda(x)(display x)(newline)) *prop-lit-mappings*)
             (display "*prop-atom-maxnr*: ")(newline)
             (display *prop-atom-maxnr*)(newline) ))

           (set! res (check-prop-flatten-interpretations
                         predclauses flatclauses domainsize #t)) )))

            ; (if (> domainsize 2) (set! stopflag #t)) ))))

(define *prop-orig-constants-to-flat-domainels* '())

(define (prop-build-nofuns-model predclauses)
  (let ((stopflag #f)
        (domainsize 0)
        (flatclauses '())
        (tmpnr 0)
        (tmp '())
        (checkres '())
        (eqaxioms '())
        (res #f) )

    (set! *sos-exhausted-flag* #f)

    ;(set! tmp (prop-build-flatten-model predclauses))
    ;(newline)
    ;(display "RES: ")(display tmp)(newline)
    ;(quit)

    ;(newline)
    ;(display "Satisfiability checking for problems without ")
    ;(newline)
    ;(display "nonconstant function symbols is not supported ")
    ;(newline)
    ;(display "yet in this version of Gandalf.")
    ;(newline)
    ;(quit)

    ;(newline)
    ;(display "prop-build-nofuns-model called")
    ;(newline)

    (set! *prop-conversion-to-file* #t)

    ; first create unique file names

    (set! *prop-conversion-file-name*
          (string-append *prop-conversion-file-name-start* *run-id*) )
    (set! *prop-conversion-temp-file-name*
          (string-append *prop-conversion-temp-file-name-start* *run-id*) )
    (set! *prop-conversion-header-file-name*
           (string-append *prop-conversion-header-file-name-start* *run-id*) )
    (set! *prop-conversion-result-file-name*
           (string-append *prop-conversion-result-file-name-start* *run-id*) )

    (set! *max-var-stat-pre* 0)
    (set! *max-var-stat-post* 0)

    ; (set! *prop-atom-maxnr* *prop-last-enum-value*)
    (set! *prop-time-ran-out* #f)
    (set! *given-checkpoint-keptnrs* '())
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)

    ;(newline)
    ;(display "orig clauses before extracting pred data to prop: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) predclauses)
    ;(newline)

    (set! predclauses (prop-uncurry-clauses predclauses))

    (extract-pred-data-to-prop predclauses)
    (set! domainsize (length *prop-orig-constants*))
    (if (<= domainsize 0) (set! domainsize 1))
    (newline)
    ;(set! domainsize 1)
    (display "domain size: ")(display domainsize)(newline)
    (if (>= domainsize 200)
        '()
        (begin
         (set! tmp (- 100000 1))
         (set! *prop-orig-constants-to-flat-domainels*
              (map (lambda(x)
                     (set! tmp (+ 1 tmp))
                     (cons x tmp) )
                    *prop-orig-constants*))

    ;(newline)
    ;(display "orig clauses before flattening: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) predclauses)
    ;(newline)

        (set! *prop-orig-constants* '())
        (set! flatclauses (pseudo-flatten-clause-set predclauses))

    ;(newline)
    ;(display "pseudo-flattened clause set:")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) flatclauses)
    ;(newline)

        (if *input-contains-eq*
           (set! eqaxioms (make-nofuns-equality-axioms))
           (set! eqaxioms '()) )


    ;(newline)
    ;(display "eq axioms created after flattening: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) eqaxioms)
    ;(newline)

        (set! flatclauses (append eqaxioms flatclauses))

    ;(newline)
    ;(display "pseudo-flattened clause set 2:")
    ;(newline)
    ;(display flatclauses)
    ;(newline)


        (set! *flat-split-literal-nr* 10000) ; new predicates from splitting

    ;(newline)
    ;(display "orig clauses before splitting: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) flatclauses)
    ;(newline)

    ;(set! *max-var-stat-pre* (compute-max-vars-in-clauses flatclauses))

       (set! flatclauses (split-flat-clause-set flatclauses))

    ;(set! *max-var-stat-post* (compute-max-vars-in-clauses flatclauses))

    ;(display "orig clauses after splitting: ")
    ;(newline)
    ;(for-each (lambda(x)(display x)(newline)) flatclauses)
    ;(newline)

         (make-prop-interpretation-lists domainsize #f #t)

         ;(newline)
         ;(display "interpr lists created")
         ;(newline)

         (create-prop-var-vectors *prop-atom-maxnr*)

           (if #f
            (begin
             (display "*prop-interp-value-pairs*: ")(newline)
             (display *prop-interp-value-pairs*)(newline)
             (display "*prop-constants-mapping*: ")(newline)
             (display *prop-constants-mapping*)(newline)
             (display "*prop-fun-mappings*: ") (newline)
             (display *prop-fun-mappings*)(newline)
             (display "*prop-lit-mappings*: ") (newline)
             (display *prop-lit-mappings*)(newline)
             (display "*prop-atom-maxnr*: ")(newline)
             (display *prop-atom-maxnr*)(newline) ))

        (set! *prop-pre-evalres* '())
        (set! *prop-clauses-generation-limit* 1000000)
        (set! *prop-clauses-generation-limited-flag* #f)
        (set! res (check-prop-flatten-interpretations
                          predclauses flatclauses domainsize #f))
        (set! *prop-conversion-to-file* #f)


    ;(newline)(display "res1: ")(display res)(newline)
    ;(newline)(display "*prop-pre-evalres*: ")(display *prop-pre-evalres*)(newline)

        (set! checkres '())
        (if (eq? *prop-pre-evalres* #t)
            (if (and (not *prop-clauses-generation-limited-flag*)
                     (check-prop-prover-outfile
                       *prop-conversion-result-file-name*
                       "Instance satisfiable") )
                (set! checkres #t)
                (if  (check-prop-prover-outfile
                        *prop-conversion-result-file-name*
                        "Instance unsatisfiable")
                     (set! checkres #f) ))
            (set! checkres #f) )

    ;(newline)(display "checkres: ")(display checkres)(newline)

    ;(newline)(display "sat check: ")
    ;(display (check-prop-prover-outfile
    ;            *prop-conversion-result-file-name*
    ;            "Instance satisfiable"))
    ;(newline)(display "unsat check: ")
    ;(display (check-prop-prover-outfile
    ;            *prop-conversion-result-file-name*
    ;            "Instance unsatisfiable"))

    ; remove temporary files

        (system (string-append "rm " *prop-conversion-file-name*))
        (system (string-append "rm " *prop-conversion-temp-file-name*))
        (system (string-append "rm " *prop-conversion-header-file-name*))
        (system (string-append "rm " *prop-conversion-result-file-name*))

        checkres))))


(define (compute-max-vars-in-clauses clauses)
  (let ((mx 0)
        (nr 0)
        (vars '()) )
     (for-each
       (lambda(cl)
         (set! vars (get-fullclause-varslist cl))
         (set! nr (length vars))
         (if (> nr mx) (set! mx nr)) )
       clauses)
     mx))

; prop-build-random-model returns #t iff satisfiable, #f if building
; failed for any reason

(define (prop-build-random-model predclauses)
  (let ((stopflag #f)
        (domainsize 0)
        (randomsize 20)
        (tmpnr 0)
        (res #f) )
    (set! *prop-time-ran-out* #f)
    (set! *given-checkpoint-keptnrs* '())
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)

    (set! predclauses (prop-uncurry-clauses predclauses))
    (extract-pred-data-to-prop predclauses)

    (do ((domainsize 1 (+ 1 domainsize)))
        ((or stopflag
             (> domainsize 30)
             res
                 (> (- (get-check-time 1) *slow-start-time*)
	            *one-pass-time*))
	        ;(if (and (not stopflag)(not res))
	        ;    (store-time-checkpoint! 1) )
	        (if *prop-time-ran-out*
    	            (begin
	               (if (> *display-info-level* 38)
           	           (begin (newline)
	                          (display "allocated model search time has run out")
		                  (newline) ))
	                #f)
                    res))
           (if (> *display-info-level* 38)
               (begin (newline)
	              (display "starting search for domain size ")
	              (display domainsize)
		      (newline) ))

           (make-prop-interpretation-lists domainsize #t #f) ;
           (create-prop-var-vectors *prop-atom-maxnr*)

           (if *prop-debug*
            (begin
             (display "*prop-interp-value-pairs*: ")(newline)
             (display *prop-interp-value-pairs*)(newline)
             (display "*prop-constants-mapping*: ")(newline)
             (display *prop-constants-mapping*)(newline)
             (display "*prop-fun-mappings*: ") (newline)
             (display *prop-fun-mappings*)(newline)
             (display "*prop-lit-mappings*: ") (newline)
             (display *prop-lit-mappings*)(newline)
             (display "*prop-atom-maxnr*: ")(newline)
             (display *prop-atom-maxnr*)(newline) ))

           (set! res (check-prop-interpretations predclauses domainsize randomsize)) )))

            ; (if (> domainsize 2) (set! stopflag #t)) ))))


(define *incremental-model-flag* #f)


(define (prop-build-incremental-model predclauses)
  (let ((stopflag #f)
        (domainsize 0)
        (tmpnr 0)
        (res #f) )

    (set! *prop-time-ran-out* #f)
    (set! *given-checkpoint-keptnrs* '())
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)

    (set! predclauses (prop-uncurry-clauses predclauses))
    (extract-pred-data-to-prop predclauses)

    ;(newline)
    ;(display "prop-build-incremental-model called")(newline)
    ;(display "predclauses: ")(newline)
    ;(for-each (lambda(x)(newline)(display x)) predclauses)(newline)
    ;(display "prop-orig-funs: ")(display *prop-orig-funs*)(newline)

    (cond
      ((null? *prop-orig-funs*)
      ; no function symbols: interpret all constants to themselves
            (set! domainsize (length *prop-orig-constants*))
            (if (zero? domainsize) (set! domainsize 1))
            (if (> *display-info-level* 38)
               (begin (newline)
	              (display "no function symbols, starting search for domain size ")
	              (display domainsize)
		      (newline) ))
                ; last arg #t iff const are interp
           (set! *prop-interp-value-pairs* '())
           (make-prop-interpretation-lists domainsize #t #f)
           ;(display *prop-interp-value-pairs* )
           (create-prop-var-vectors *prop-atom-maxnr*)
           ;(display *prop-interp-value-pairs* )
           ;(abort)
           ; now try to solve
           (set! res (check-prop-one-interpretation predclauses domainsize))
           (if *prop-time-ran-out*
    	       (begin
	               (if (> *display-info-level* 38)
           	           (begin (newline)
	                          (display "allocated model search time has run out")
		                  (newline) ))
	                #f)
               res))
      ; function symbols present: full domain size iteration
     (else
    	(do ((domainsize 1 (+ 1 domainsize)))
            ((or stopflag
                 res
                 (> (- (get-check-time 1) *slow-start-time*)
	            *one-pass-time*))
	        ;(if (and (not stopflag)(not res))
	        ;    (store-time-checkpoint! 1) )
	        (if *prop-time-ran-out*
    	            (begin
	               (if (> *display-info-level* 38)
           	           (begin (newline)
	                          (display "allocated model search time has run out")
		                  (newline) ))
	                #f)
                    res))
           (if #t ; (> *display-info-level* 38)
               (begin (newline)(display "-------------------------------------")(newline)
	              (display "starting search for domain size ")(display domainsize)
	              (newline)(display "-------------------------------------")(newline)
		      (newline) ))

           (make-prop-interpretation-lists domainsize #t #f) ;
           (create-prop-var-vectors *prop-atom-maxnr*)

           (if *prop-debug*
            (begin
             (display "*prop-interp-value-pairs*: ")(newline)
             (display *prop-interp-value-pairs*)(newline)
             (display "*prop-constants-mapping*: ")(newline)
             (display *prop-constants-mapping*)(newline)
             (display "*prop-fun-mappings*: ") (newline)
             (display *prop-fun-mappings*)(newline)
             (display "*prop-lit-mappings*: ") (newline)
             (display *prop-lit-mappings*)(newline)
             (display "*prop-atom-maxnr*: ")(newline)
             (display *prop-atom-maxnr*)(newline) ))

           (set! res (check-prop-interpretations-incremental predclauses domainsize 0)) )))))

            ; (if (> domainsize 2) (set! stopflag #t)) ))))

;----------------------------------------------------------------------
;
;                 pred-level data extraction, initialisation
;
;----------------------------------------------------------------------

; takes a full fully-built clause list and extracts all constant, fun and pred symbs
; assigns:
; *prop-orig-constants*
; *prop-orig-funs*
; *prop-orig-preds*

(define (extract-pred-data-to-prop predclauselist)
  (let ()
     (set! *prop-orig-constants* '())
     (set! *prop-orig-funs* '())
     (set! *prop-orig-preds* '())
     (set! *prop-last-enum-value* 0)

     ; build lists of original constants, funs and preds

     (for-each
        (lambda(fullcl)

          ;(newline)(display "fullcl: ")(display fullcl)(newline)

          (if (and (pair? fullcl)(pair? (cdr fullcl)))
              (for-each
                (lambda(lit)
                  (extract-pred-data-to-prop-aux-lit lit) )
                (cdr fullcl) )))
        predclauselist)

     (set! *prop-orig-constants-enum*
           (map (lambda(x)
                  (set! *prop-last-enum-value* (+ 1 *prop-last-enum-value*))
                  (cons x *prop-last-enum-value*) )
                *prop-orig-constants*))
     (set! *prop-orig-funs-enum*
           (map (lambda(x)
                  (set! *prop-last-enum-value* (+ 1 *prop-last-enum-value*))
                  (cons (car x) *prop-last-enum-value*) )
                *prop-orig-funs*))
     (set! *prop-orig-preds-enum*
           (map (lambda(x)
                  (set! *prop-last-enum-value* (+ 1 *prop-last-enum-value*))
                  (if (pos-eq-pred? (car x))
                      (set! *prop-flat-eq-symb* *prop-last-enum-value*) )
                  (cons (car x) *prop-last-enum-value*) )
                *prop-orig-preds*))

     (if *prop-debug*
          (begin
     		(display *prop-orig-constants*)
     		(display *prop-orig-funs*)
     		(display *prop-orig-preds*)

     		(display *prop-orig-constants-enum*)
     		(display *prop-orig-funs-enum*)
     		(display *prop-orig-preds-enum*) ))))




(define (extract-pred-data-to-prop-aux-lit lit)
  ;(newline)
  ;(display "lit: ")(display lit)
  ;(newline)
  (let ((symb '())
        (argpart '()) )
     (set! symb (prop-spec-pure-unsigned-fun (car lit)))

     (if (not (assq symb *prop-orig-preds*))
         (begin

           ;(newline)
          ;(display "symb not found: ")(display symb)
          ;(newline)

           (set! *prop-orig-preds*
                (cons (cons symb (length (cdr lit))) *prop-orig-preds*) ))
         (begin
           ;(newline)
           ;(display "symb found: ")(display symb)
           ;(newline)
            ))
     (do ((argpart (cdr lit) (cdr argpart)))
         ((null? argpart))
        (extract-pred-data-to-prop-aux-fun (car argpart)) )))


(define (extract-pred-data-to-prop-aux-fun term)
  (let ((symb '())
        (argpart '()) )
     (cond ((pair? term)
     	      (set! symb (prop-spec-pure-fun (car term)))
     	      (if (not (assq symb *prop-orig-funs*))
                  (set! *prop-orig-funs*
                        (cons (cons symb (length (cdr term))) *prop-orig-funs*) ))
     	      (do ((argpart (cdr term) (cdr argpart)))
         	  ((null? argpart))
        	(extract-pred-data-to-prop-aux-fun (car argpart)) ))
           ((const? term)
              (set! symb term)
              (if (not (memq symb *prop-orig-constants*))
                  (set! *prop-orig-constants* (cons symb *prop-orig-constants*)) )))))



;----------------------------------------------------------------------
;
;                 interpretation lists construction
;
;----------------------------------------------------------------------
;
;
;  const a, b, c to 1,2:
;   a=1
;   b=1
;   c=2
;
;  fun f with arity 2 to 1,2:
;   f(1,1)=1
;   f(1,2)=1
;   f(2,1)=2
;   f(2,2)=2

; make-prop-interpretation-lists creates interpretation lists to be
; assigned and backtraced over later. If interpconstants is #f, constant
; symbols will not be put there.


(define (make-prop-interpretation-lists domainsize interpconstants flattenflag)
  (let ((tmp '())
        (modified '()) )
     (set! *prop-interp-value-pairs* '())
     (set! *prop-interp-termlist* '())
     (set! *prop-constants-mapping* '())
     (set! *prop-fun-mappings* '())
     (set! *prop-lit-mappings* '())
     (set! *prop-atom-maxnr* 0)

     ;(newline)(display "preds: ")
     ;(display *prop-orig-preds-enum*)(newline)
     ;(display *prop-orig-preds*)
     ;(newline)(display "funs: ")
     ;(display *prop-orig-funs-enum*)(newline)
     ;(display *prop-orig-funs*)
     ;(newline)(display "consts: ")
     ;(display *prop-orig-constants-enum*)(newline)
     ;(display *prop-orig-constants*)(newline)

     (if flattenflag
       ; interpret both preds, funs and constants as preds (plus one arg ...)
       (begin
         (set! modified
                 (append
       		    (map (lambda(x)
       		           (set! tmp (assq (car x) *prop-orig-preds-enum*))
       		           (cons (cdr tmp) (cdr x)) )
       		         *prop-orig-preds*)
       		   (append
       		        (map (lambda(x)
       		               (set! tmp (assq (car x) *prop-orig-funs-enum*))
       		               (cons (cdr tmp)(+ 1 (cdr x))))
       		             *prop-orig-funs*)
       		        (map (lambda(x)
       		               (set! tmp (assq x *prop-orig-constants-enum*))
       		               (cons (cdr tmp) 1))
       		             *prop-orig-constants*))))

          ;(newline)(display "modified prop mapping input: ")
          ;(newline)(display modified)

         (set! *prop-lit-mappings*
           (map (lambda(funinfo)
                  (cons (car funinfo)
                        (make-prop-interpretation-lists-aux
                           (cdr funinfo) (list (car funinfo))
                           domainsize #f flattenflag) ))
       		modified)))
       ; normal interpretationmark
       (begin
       	(set! *prop-lit-mappings*
           (map (lambda(funinfo)
                  (cons (car funinfo)
                        (make-prop-interpretation-lists-aux
                           (cdr funinfo) (list (car funinfo))
                           domainsize #f flattenflag) ))
       		*prop-orig-preds*))
        (set! *prop-fun-mappings*
           (map (lambda(funinfo)
                  (cons (car funinfo)
                        (make-prop-interpretation-lists-aux
                          (cdr funinfo) (list (car funinfo))
                          domainsize #t flattenflag) ))
       		*prop-orig-funs*))
        (set! *prop-constants-mapping*
     	   (map (lambda(const)
     	          (set! tmp (cons const '()))
     	          (if interpconstants
     	              (begin
     	                 (set! *prop-interp-termlist*
     	          	      (cons (cons const '()) *prop-interp-termlist*) )
     	          	 (set! *prop-interp-value-pairs*
     	          	      (cons tmp *prop-interp-value-pairs*) )))
     	          tmp)
     	        *prop-orig-constants*))))))


(define (make-prop-interpretation-lists-aux
             argsleft term domainsize addinterpflag flattenflag)

  ; (newline)(display "make-prop-interpretation-lists-aux called: ")
  ; (newline)(display argsleft)(newline)
  ; (display term)(newline)


  (let ((nr 0)
        (tmp '())
        (res '()) )
    (do ((nr 0 (+ 1 nr)))
        ((>= nr domainsize) res)
      (if (<= argsleft 1)
          (if addinterpflag
              (begin
                (set! tmp (cons nr '()))
                (set! *prop-interp-value-pairs*
     	              (cons tmp *prop-interp-value-pairs*) )
     	        (set! *prop-interp-termlist*
     	              (cons (reverse (cons nr term)) *prop-interp-termlist*) )
     	        (set! res (cons tmp res)) )
     	      (begin
     	        (set! *prop-atom-maxnr* (+ 1 *prop-atom-maxnr*))
     	        (set! tmp (cons nr *prop-atom-maxnr*))
     	        (set! res (cons tmp res)) ))
          (begin (set! tmp (cons nr (make-prop-interpretation-lists-aux
                                        (- argsleft 1)
                                        (cons nr term)
                                        domainsize addinterpflag flattenflag)))
                 (set! res (cons tmp res)) )))))



;----------------------------------------------------------------------
;
;       pred-level formula saturation with domain values,
;       prop form construction
;
;----------------------------------------------------------------------

; check-prop-one-interpretations  returns #t iff it finds a model for
; a given domain size and given interpretation
; ie iff the predclauses are found satisfiable
; a value #f means a failure to find model
;
; it DOES NOT generate new interpretations of fun symbs


(define (check-prop-one-interpretation predclauses domainsize)
  (let ((valuations '())
        (selectedvals '())
        (pre-evalres #t)
        (evalres #t)
        (allchecked #f)
        (tmpnr -1)
        (maxval (- domainsize 1)) )

       (newline)(display " check-prop-one-interpretation called on domainsize: ")
       (display domainsize)(newline)

       (set! *propclause-tree* '())
       ; create a simple valuation for all the constants
       (set! valuations (map (lambda(x)
                                  (set! tmpnr (+ 1 tmpnr))
                                  tmpnr)
                             *prop-interp-value-pairs*))
       ; insert the newly created valuations
       (insert-valuations! *prop-interp-value-pairs* valuations)

       (if *prop-debug*
          (begin
            (newline)
            (display "valuations: ")(display valuations) ))

       (if (> *display-info-level* 50)
         (begin
           (for-each (lambda(x y)
                       (display x) (display " = ") (display y) (newline) )
                     *prop-interp-termlist*
                     valuations)
           (newline) ))


       (if (> *display-info-level* 70)
           (begin (newline)
	          (display "starting to check satisfiability for symbol interp:")
	          (newline)
	          (display *prop-interp-value-pairs*)
		  (newline) ))

       ; clear the vector used for storing values during clause conversion to prop

       (vector-fill! *propvar-enqued-value* '())

       ; make-predclauses-to-propclauses gives #t if might be sat,
        ; #f iff certainly not satisfiable

       (set! pre-evalres (make-predclauses-to-propclauses predclauses domainsize #f))

       (if (> *display-info-level* 70)
           (begin (newline)
	          (if pre-evalres
	             (display "conversion to prop did not find unsatisfiability")
	             (display "conversion to prop found unsatisfiability") )
		  (newline) ))

       (if *prop-debug*
          (begin
             (newline)
             (display "pre-evalres: ")(display pre-evalres) ))

       ; solve-prop-clauses gives #t if satisfiable, #f if not satisfiable

       (if pre-evalres
           (set! evalres (solve-prop-clauses *propclauses* *prop-atom-maxnr*)) )

       (if (> *display-info-level* 70)
           (begin (newline)
	          (if evalres
	             (if (not *prop-clauses-generation-limited-flag*)
	                 (display "full prop evaluation found prop satisfiable") )
	             (display "full prop evaluation found prop unsatisfiable") )
		  (newline) ))

       (if *prop-debug*
          (begin (display "funsymb interp vals: ")(display valuations)(newline)) )

      (if *prop-debug*
          (begin (newline)(display "evalres: ")(display evalres)(newline)) )

      evalres))





; check-prop-flatten-interpretations  returns #t iff it finds a model for
; a given domain size
; ie iff the predclauses are found satisfiable
; a value #f means a failure to find model
;
; NB! no function or constant symbols are present!

(define *prop-pre-evalres*  '())

(define (check-prop-flatten-interpretations predclauses flatclauses domainsize extraflag)
  ; (newline)(display "check-prop-flatten-interpretations called")(newline)
  (let ((valuations '())
        (selectedvals '())
        (pre-evalres #f)
        (evalres #f)
        (extraclauses '())
        (allchecked #f)
        (tmpnr -1)
        (maxval (- domainsize 1)) )
       (set! *propclause-tree* '())

       (if (and *global-prop-prover* *prop-conversion-to-file*
                *sat-strategy-flag*)
           (begin (open-prop-temp-output-file
                        *prop-conversion-temp-file-name*) )
           (begin (set! *prop-conversion-to-file* #f) ))

       ; clear the vector used for storing values during clause conversion to prop

       (vector-fill! *propvar-enqued-value* '())

       (set! *prop-pred-maxval-pairs*
             (make-prop-pred-maxval-pairs flatclauses domainsize) )

       (if *prop-debug*
           (begin
            (newline)
            (display "*prop-pred-maxval-pairs*:")(newline)
            (display  *prop-pred-maxval-pairs*) ))

       ; make-predclauses-to-propclauses gives #t if might be sat,
       ; #f iff certainly not satisfiable

       (if extraflag
           (begin
              (set! extraclauses (make-extra-flatten-propclauses domainsize))
              (if #f
                  (begin
                   (newline)
                   (for-each (lambda(x)(display x)(newline)) extraclauses)
                   (newline) ))
              (set! flatclauses
                   (append extraclauses flatclauses) )))


       (set! pre-evalres (make-predclauses-to-propclauses flatclauses domainsize #t))


       (if #t ; (> *display-info-level* 70)
           (begin (newline)
	          (if pre-evalres
	             (display "conversion to prop did not find unsatisfiability")
	             (display "conversion to prop found unsatisfiability") )
               (newline) ))

       (if *prop-debug*
          (begin
             (newline)
             (display "pre-evalres: ")(display pre-evalres) ))

       ; solve-prop-clauses gives #t if satisfiable, #f if not satisfiable

       ;(if pre-evalres
       ;    (set! pre-evalres
       ;          (make-prop-infile
       ;            filename propclauses prop-atom-maxnr one-pass-time)))

       (if pre-evalres
           (begin
            (set! *prop-pre-evalres* #t)
            (if *global-prop-prover*
               (set! evalres
                  (external-solve-prop-clauses
                      *propclauses* *prop-atom-maxnr* *one-pass-time*))
               (set! evalres (solve-prop-clauses *propclauses* *prop-atom-maxnr*)) ))
           (begin
             (set! *prop-pre-evalres* #f) ))

       (if (> *display-info-level* 70)
           (begin (newline)
	          (if evalres
	             (if (not *prop-clauses-generation-limited-flag*)
	                 (display "full prop evaluation found prop satisfiable") )
	             (display "full prop evaluation found prop unsatisfiable") )
		  (newline) ))

      (if *prop-debug*
          (begin (newline)(display "evalres: ")(display evalres)(newline)) )

      evalres))



; check-prop-interpretations  returns #t iff it finds a model for
; a given domain size
; ie iff the predclauses are found satisfiable
; a value #f means a failure to find model
;
; it generates and tests all interpretations of fun symbs


(define (check-prop-interpretations predclauses domainsize randomsize)
  (let ((valuations '())
        (selectedvals '())
        (pre-evalres #f)
        (evalres #f)
        (count 0)
        (allchecked #f)
        (maxval (- domainsize 1)) )
     (set! *propclause-tree* '())
     (set! valuations (map (lambda(x) 0) *prop-interp-value-pairs*))

     ;(newline)
     ;(display "check-prop-interpretations called on domainsize: ")
     ;(display domainsize)(newline)
     ;(display "valuations: ")(display valuations)(newline)

     (do ((count 0 (+ 1 count)))
         ((or evalres
              allchecked
              (and (> randomsize 0) (> count randomsize))
              (> (- (get-check-time 1) *slow-start-time*)
	         *one-pass-time*))
	    ;(if (and (not evalres)(not allchecked))
	    ;    (store-time-checkpoint! 1) )

            (if (and (not evalres)(not allchecked))
                (set! *prop-time-ran-out* #t) )

            (if *prop-debug*
               (begin (newline)(display "evalres: ")(display evalres)(newline)))

            evalres)
       (insert-valuations! *prop-interp-value-pairs* valuations)

       (if *prop-debug*
          (begin
            (newline)
            (display "valuations: ")(display valuations) ))

       (if (> *display-info-level* 50)
         (begin
           (display " *prop-interp-termlist* and valuations: ")(newline)
           (for-each (lambda(x y)
                       (display x) (display " = ") (display y) (newline) )
                     *prop-interp-termlist*
                     valuations)
           (newline) ))


       (if (> *display-info-level* 70)
           (begin (newline)
	          (display "starting to check satisfiability for symbol interp:")
	          (newline)
	          (display *prop-interp-value-pairs*)
		  (newline) ))

       ; clear the vector used for storing values during clause conversion to prop

       (vector-fill! *propvar-enqued-value* '())

       ; make-predclauses-to-propclauses gives #t if might be sat,
       ; #f iff certainly not satisfiable

       (set! pre-evalres (make-predclauses-to-propclauses predclauses domainsize #f))

       (if (> *display-info-level* 70)
           (begin (newline)
	          (if pre-evalres
	             (display "conversion to prop did not find unsatisfiability")
	             (display "conversion to prop found unsatisfiability") )
		  (newline) ))

       (if *prop-debug*
          (begin
             (newline)
             (display "pre-evalres: ")(display pre-evalres) ))

       ; solve-prop-clauses gives #t if satisfiable, #f if not satisfiable

       (if pre-evalres
           (set! evalres (solve-prop-clauses *propclauses* *prop-atom-maxnr*)) )

       (if (> *display-info-level* 70)
           (begin (newline)
	          (if evalres
	             (display "full prop evaluation found prop satisfiable")
	             (display "full prop evaluation found prop unsatisfiable") )
		  (newline) ))

       (if *prop-debug*
          (begin (display "funsymb interp vals: ")(display valuations)(newline)) )

       (if (> randomsize 0)
           (valuation-list-nextrandomvalues valuations valuations domainsize)
           (set! allchecked
                (or (null? valuations)
                    (valuation-list-nextvalues valuations valuations domainsize) ))))))




(define (check-prop-interpretations-incremental predclauses domainsize randomsize)
  (let ((valuations '())
        (selectedvals '())
        (pre-evalres #f)
        (evalres #f)
        (count 0)
        (allchecked #f)
        (maxval (- domainsize 1)) )

  ; (newline)
  ; (display "check-prop-interpretations-incremental called: ")(newline)
  ;(display "predclauses: ")(newline)
  ;(for-each (lambda(x)(display x)(newline)) predclauses)
  ;(display "domainsize: ")(display domainsize)(newline)
  ;(display "*prop-interp-value-pairs*: ")(newline)
  ;(for-each (lambda(x)(display x)(newline)) *prop-interp-value-pairs*)(newline)


     (set! *incremental-model-flag* #f)
     (set! *propclause-tree* '())
     (set! valuations (map (lambda(x) #f) *prop-interp-value-pairs*))

     (set! evalres
       (check-prop-interpretations-incremental-aux
           '() valuations  predclauses domainsize randomsize))

     ;(newline)
     ;(display "check-prop-interpretations-incremental returns: ")
     ;(display evalres)
     evalres))


(define (check-prop-interpretations-incremental-aux
           filledvaluations unfilledvaluations
           predclauses domainsize randomsize)
  ;(newline)
  ;(display "check-prop-interpretations-incremental-aux called: ")(newline)
  ;(display "    filled: ")(display filledvaluations)(newline)
  ;(display "  unfilled: ")(display unfilledvaluations)(newline)
  ;(display "domainsize: ")(display domainsize)(newline)
  (let ((new-filledvaluations '())
        (res #f)
        (pre-evalres #f)
        (evalres #f)
        (partialokflag #f) )
     (cond
       ((null? unfilledvaluations)

          ;(newline)(display "unfilledvaluations is null")(newline)

           (set! *propclause-tree* '())
           (insert-valuations! *prop-interp-value-pairs* filledvaluations)
           (vector-fill! *propvar-enqued-value* '())

           (if (> *display-info-level* 50)
              (begin
                (display " *prop-interp-termlist* and valuations: ")(newline)
                 (for-each (lambda(x y)
                              (display x) (display " = ") (display y) (newline) )
                           *prop-interp-termlist*
                           filledvaluations)
                 (newline) ))

           (set! *incremental-model-flag* #f)
           (set! evalres #f)
           (set! *propclauses* '())
           (set! pre-evalres
                (make-predclauses-to-propclauses predclauses domainsize #f))

           ;(newline)(display "pre-evalres: ")(display pre-evalres)
           ;(newline)(display "propclauses created: ")(newline)
           ;(for-each (lambda(x)(display x)(newline)) *propclauses*)
           ;(newline)

           (if pre-evalres
                (set! evalres (solve-prop-clauses *propclauses* *prop-atom-maxnr*)) )
           evalres)
       ((> (- (get-check-time 1) *slow-start-time*)
	         *one-pass-time*)
	   #f)
       (else
          (set! res #f)
          (do ((i 0 (+ 1 i)))
              ((or res
                   (>= i domainsize))
                  res)

            ;(newline)(display "domain elem loop i in check-...incremental-aux: ")
            ;(display i)(newline)

            (set! new-filledvaluations
                  (cons i
                        filledvaluations))
            ;(display "unfilledvaluations: ")(display unfilledvaluations)(newline)
            ;(display "new-filledvaluations: ")(display new-filledvaluations)(newline)

            (set! partialokflag #f)
            (if (null? (cdr unfilledvaluations))
                (set! partialokflag #t)
                (begin

                   (set! *propclause-tree* '())
                   (set! *incremental-model-flag* #t)
                   (insert-valuations!
                       *prop-interp-value-pairs*
                       (append (cdr unfilledvaluations) new-filledvaluations))
           	   (vector-fill! *propvar-enqued-value* '())
                   (if (> *display-info-level* 50)
                      (begin
                         (display " *prop-interp-termlist* and valuations: ")(newline)
                         (for-each (lambda(x y)
                                     (display x) (display " = ") (display y) (newline) )
                                   *prop-interp-termlist*
                                   (append (cdr unfilledvaluations) new-filledvaluations) )
                         (newline) ))

                   ;(clear-valuations! *prop-interp-value-pairs* (cdr unfilledvaluations))
                   (set! evalres #f)
                   (set! *propclauses* '())
                   (set! pre-evalres
                     (make-predclauses-to-propclauses
                          predclauses domainsize #f))

                   ;(newline)
                   ;(display "partial pre-evalres: ")(display pre-evalres)
                   ;(newline)(display "partial propclauses created: ")(newline)
                   ;(for-each (lambda(x)(display x)(newline)) *propclauses*)
                   ;(newline)

                   (if pre-evalres
                       (display "+")
                       (display "-") )


                   (if pre-evalres
                      (set! evalres
                           (solve-prop-clauses *propclauses* *prop-atom-maxnr*)) )

                   ;(newline)
                   ;(display "partial evalres: ")(display evalres)
                   ;(newline)

                   (if pre-evalres
                       (if evalres
                          (display "*")
                          (display "%") ))

                   (set! *incremental-model-flag* #f)
                   (set! partialokflag evalres) ))
            (if partialokflag
                (set! res (check-prop-interpretations-incremental-aux
                            new-filledvaluations (cdr unfilledvaluations)
                            predclauses domainsize randomsize))) )))))




(define (valuation-list-nextvalues valuationlist origvaluations domainsize)

    ;(begin (display "in nextvalues: ")
    ;       (display valuationlist)(display " ")
    ;       (display origvaluations)(display " ")
    ;       (display domainsize)(newline) )

    (if (and (pair? valuationlist)
             (or (and (pair? origvaluations)
                      (> (car valuationlist)
                            (max-of-two (max-of-list (cdr valuationlist))
                                        (max-of-corr-fun-args
                                          valuationlist origvaluations) )))
                 (>= (car valuationlist) (- domainsize 1)) ))
        (begin
              (if (null? (cdr valuationlist))
                   #t
                   (begin
                     (set-car! valuationlist 0)
                     (valuation-list-nextvalues
                         (cdr valuationlist) origvaluations domainsize) )))
        (begin (set-car! valuationlist (+ 1 (car valuationlist)))
               #f)))


(define (max-of-two one two)
  (if (> one two) one two) )

(define (max-of-list lst)
  (let ((res 0))
     (do ((part lst (cdr part)))
         ((null? part) res)
       (if (> (car lst) res) (set! res (car lst))) )))

(define (max-of-corr-fun-args valuationlist origvaluationlist)
   (let ((term (get-valuationlistpos-term valuationlist origvaluationlist)))
      ;(display term)
      (if (pair? term)
          (max-of-list (cdr term))
          term)))

(define (get-valuationlistpos-term valuationlist origvaluationlist)
    ;(display valuationlist)(display origvaluationlist)(display *prop-interp-termlist*)
    (let ((res '()))
       (do ((part1 origvaluationlist (cdr part1))
            (part2 *prop-interp-termlist* (cdr part2)) )
           ((or (null? part1) (eq? part1 valuationlist))
               (if (null? part1)
                   999999
                   (car part2) )))))


(define (insert-valuations! value-pairs valuations)
  (for-each (lambda(x y)(set-cdr! x y))
             value-pairs
             valuations))



(define (valuation-list-nextrandomvalues valuationlist origvaluations domainsize)

    ;(begin (display "in nextrandomvalues: ")
    ;       (display valuationlist)(display " ")(display domainsize)(newline))
  (let ((tmp 0))
    (do ((part valuationlist (cdr part)))
        ((null? part))
       (set! tmp (random-random domainsize))
       (if (or (not (integer? tmp)) (< tmp 0) (>= tmp domainsize))
           (begin
              (display "bad random: ")(display tmp)(newline)
              (set! tmp 0) ))
       (set-car! part tmp) )))



; sets the global value *propclauses*

; make-predclauses-to-propclauses returns #f iff found to be unsatisfiable,
; else #t

(define *prop-clauses-generated* 0)

(define (make-predclauses-to-propclauses predclauses domainsize flatflag)
  (let ((tmp '())
  	(unsatisfiable #f)
        (predclausespart '()) )
    (set! *prop-clauses-generated* 0)
    (set! *propclauses* '())
    (set! *propclause-tree* '())
    (if flatflag
        (set! predclauses (sort-predclauses-for-propconversion predclauses)) )

    ;(newline)
    ;(display "*prop-flat-eq-symb*: ")(display *prop-flat-eq-symb*)(newline)
    ;(display "clauses to be converted to prop: ")(newline)
    ;(for-each (lambda(x)(display x)(newline)) predclauses)
    ;(newline)(newline)

    (if *prop-debug*
          (begin
     		(display "predclauses before conversion: ")(newline)
    		(newline)(for-each (lambda(x)(newline)(display x)) predclauses)(newline)
    		(display "now convert: ")(newline) ))

    (do ((predclausespart predclauses (cdr predclausespart)))
        ((or (null? predclausespart) unsatisfiable)

           (if *prop-debug*
          	(begin
           	 (newline)(display "unsatisfiable:")(display unsatisfiable)(newline)))

           (not unsatisfiable))

       (if *prop-debug*
          	(begin (display (car predclausespart))(newline)) )

      (if (and (pair? (car predclausespart))
               (pair? (cdar predclausespart)))
          (begin
            (set! tmp
               (make-predclause-to-propclauses
                   (car predclausespart) domainsize flatflag))

            (if *prop-debug*
                (begin (newline)(display "tmp: ")(display tmp)(newline)) )

            (if tmp (set! unsatisfiable #t)) )))))



(define (sort-predclauses-for-propconversion predclauses)
  (let ((cl '())
        (res0 '())
        (res1 '())
        (res2 '())
        (res3 '())
        (res4 '())
        (res5 '())
        (res '()) )
    (do ((part predclauses (cdr part)))
        ((null? part)
          (append res0
           (append res1
            (append res2
              (append res3
                 (append res4 res5) ))) ))
      (set! cl (car part))
      (cond
         ((not (pair? cl)))
         ((not (pair? (cdr cl))))
         ((null? (cddr cl))
            (if  (fun-may-cont-vars? (car (cadr cl)))
                 (set! res1 (cons cl res1))
                 (set! res0 (cons cl res0)) ))
         ((null? (cdddr cl))
            (set! res2 (cons cl res2)) )
         ((null? (cdddr (cdr cl)))
            (set! res3 (cons cl res3)) )
         ((null? (cdddr (cddr cl)))
            (set! res4 (cons cl res4)) )
         (else
            (set! res5 (cons cl res5)) )))))



;(define (sort-clause clause pair-flag)
;;
;(define (sort-clause-aux cl len pair-flag)
;  (cond ((< len 2) cl)
;	((eq? len 2)
;	   (if (if pair-flag
;                   (>-fun-ordering? (cadar (cdr cl)) (cadar cl))
;                   (>-fun-ordering? (caadr cl) (caar cl)) )
;	       (list (cadr cl) (car cl))
;	       cl))
;	(else
;	  (let* ((l1 (quotient len 2))
;		 (l2 (- len l1))
;		 (s (split-list cl l1)) )
;	    (sort-merge-clauses
;	       (sort-clause-aux (car s) l1 pair-flag)
;	       (sort-clause-aux (cdr s) l2 pair-flag)
;               pair-flag)))))
;
;(define (sort-merge-clauses cl1 cl2 pair-flag)
;  (cond ((null? cl1) cl2)
;	((null? cl2) cl1)
;        ((and pair-flag (>-fun-ordering? (cadar cl1) (cadar cl2)))
;           (cons (car cl1)
;                 (sort-merge-clauses (cdr cl1) cl2 #t) ))
;	((and (not pair-flag) (>-fun-ordering? (caar cl1) (caar cl2)))
;	   (cons (car cl1)
;		 (sort-merge-clauses (cdr cl1) cl2 #f) ))
;	(else
;	   (cons (car cl2)
;		 (sort-merge-clauses cl1 (cdr cl2) pair-flag) ))))


(define (make-predclause-to-propclauses fullpredclause domainsize flatflag)
  (let* ((info (car fullpredclause))
         (limitlocalclausesnr #f)
         (clausesgenerated 0)
         (propclausepos '())
         (propclauseneg '())
         (curlit '())
         (litvalue '())
         (curproplit '())
         (poscurproplit '())
         (varslist (if flatflag
                       (get-fullclause-varslist fullpredclause)
                       (get-fullclause-varslist fullpredclause) ))
         (varvaluations '())
         (maxvar (- domainsize 1))
         (allchecked #f)
         (unsatisfiable #f)
         (tmp '())
         (tautology #f)
         (clausepart '())
         (varlimits '()) )

    (if (or *prop-clauses-generation-limited-flag*
            (and *prop-clauses-generation-limit*
                 (>= *prop-clauses-generated* *prop-clauses-generation-limit*) ))
        (begin
            (set! limitlocalclausesnr 1000)
            (set! *prop-clauses-generation-limited-flag* #t) )
        (set! limitlocalclausesnr #f) )


    ;(newline)
    ;(display "make-predclause-to-propclauses called")(newline)
    ;(display "predclause raw: ")(display fullpredclause)
    ; (display "predclause: ")(print-pure-clause (cdr fullpredclause))
    ;(newline)
    ;(display "varslist: ")(display varslist)
    ;(newline)

    ; if vars exist, create a valuation list for vars
    (if (not (null? varslist))
        (begin
          (set! varvaluations
               (map (lambda(x) 0) varslist) )))

    ; loop over all the valuations for vars
    (do ()
        ((or unsatisfiable
             allchecked
             (and limitlocalclausesnr
                  (> clausesgenerated limitlocalclausesnr) )
             (> (- (get-check-time 1) *slow-start-time*)
	         *one-pass-time*))
	   (if (and limitlocalclausesnr
	            (> clausesgenerated limitlocalclausesnr) )
	       (set! *prop-clauses-generation-limited-flag* #t) )
	   (if (and (not unsatisfiable)(not allchecked))
               (set! *prop-time-ran-out* #t) )
           unsatisfiable)
      (set! propclausepos '())
      (set! propclauseneg '())


      (set! varlimits '())
      (set! varlimits
        (create-prop-flat-varlimits
           varslist fullpredclause domainsize) )

      (if *prop-debug*
          (begin
           (newline)
           (display "varslist:")(newline)
           (display varslist)
           (newline)
           (display "varvaluations:")(newline)
           (display varvaluations)
           (newline)
           (display "varlimits:")(newline)
           (display varlimits)(newline) ))


      (if (if (null? varlimits)
              (begin
                (set-pred-var-values varslist varvaluations)
                #t)
              (set-pred-var-values-limited
                    varslist varvaluations varlimits) )
          (begin

            ;(display "varvaluations: ")(display varvaluations)

           (set! tautology #f)
           ; loop over literals in the clause
           (do ((clausepart (cdr fullpredclause) (cdr clausepart)))
      	       ((or (null? clausepart) tautology unsatisfiable)
      	          ; if no tautology and an empty clause generated, unsatisfiability is found
      	          (if (and (not tautology) (null? propclauseneg) (null? propclausepos))
      	              (set! unsatisfiable #t) )
      	          ; if unsatisfiability was not found directly, try to add the clause
      	          (if (not unsatisfiable)
      	              (begin
      	                (if (not tautology)
      	                    (begin
      	                       (set! clausesgenerated (+ 1 clausesgenerated))
      	                       (set! *prop-clauses-generated* (+ 1 *prop-clauses-generated*))
      	                       (set! tmp (add-new-prop-clause
      	                                   (cons propclauseneg propclausepos) ))))
      	                (if (not tmp) (set! unsatisfiable #t)) )))

             ;(newline)(display "about to call make-...-pred-to-proplit with curlit: ")
             ;(display (car clausepart))(newline)

      	     (set! curlit (car clausepart))

      	     ;(newline)
      	     ;(display "inside make-predclause-to-propclauses we have ")
      	     ;(display "*incremental-model-flag*: ")(display *incremental-model-flag*)
      	     ;(newline)

      	     (cond
      	       (*incremental-model-flag*
      	         (set! curproplit (make-pred-to-proplit-incremental curlit)) )
      	       (flatflag
      	         (set! curproplit (make-flat-pred-to-proplit curlit)) )
      	       (else
      	         (set! curproplit (make-pred-to-proplit curlit)) ))

            (if  *prop-debug*
                (begin (newline)(display "curproplit generated: ")(display curproplit)
                       (newline)(display "curproplit pos value so far: ")
                       (if (and curproplit (integer? curproplit))
                         (display
                           (vector-ref *propvar-enqued-value*
                                    (if (< curproplit 0)
      	                                (- 0 curproplit)
      	                                curproplit))))))

      	     ; check found curproplit: equality, subs, cut, tautology, neg, pos?
      	     (cond
      	       ((eq? curproplit #t) ; a clause is a tautology
      	          (set! tautology #t) )
      	       ((eq? curproplit #f)) ; do not add a literal to a clause!
      	       ((begin (if (< curproplit 0)
      	                   (set! poscurproplit (- 0 curproplit))
      	                   (set! poscurproplit curproplit) )
      	               (set! litvalue (vector-ref *propvar-enqued-value* poscurproplit))
      	                ; (set! litvalue '())
      	               (not (null? litvalue)) )
      	           ; subsumed clause or cut literal
      	           (if (eq? #t litvalue)
      	                ; if same polarity, mark subsumed, otherwise do not add lit (cut)
      	               (if (>= curproplit 0)
      	                   (set! tautology #t) )
                       (if (< curproplit 0)
      	                   (set! tautology #t) )))
      	       ((< curproplit 0)
      	             (set! tmp poscurproplit)
      	             (if (memq tmp propclausepos)
                         (set! tautology #t)
      	                 (if (not (memq tmp propclauseneg))
      	                     (set! propclauseneg (cons tmp propclauseneg)) )))
      	                          ; (add-prop-lit-to-clause tmp propclauseneg)
      	       (else
      	             (if (memq curproplit propclauseneg)
      	                 (set! tautology #t)
      	                 (if (not (memq curproplit propclausepos))
      	                     (set! propclausepos (cons curproplit propclausepos)) ))) ))))
      	                          ; (add-prop-lit-to-clause curproplit propclausepos)
           ; take a next valuation

      (if *prop-debug*
          (begin (newline)(display "maxvar: ")(display maxvar)(newline)
      		 (display "varvaluations: ")(display varvaluations)(newline) ))

      (set! allchecked
          (or (null? varvaluations)
              (valuation-list-nextvalues varvaluations #t domainsize) )))))


(define *prop-pred-maxval-pairs* '())

(define (make-prop-pred-maxval-pairs flatclauses domainsize)
   (let ((res '()))

     (if  *prop-debug*
       (begin
        (newline)
        (display "make-prop-pred-maxval-pairs called:")
        (newline)
        (display "flatclauses:")(newline)
        (display flatclauses)(newline)
        (display "domainsize:")(newline)
        (display domainsize)(newline)
        (display "*prop-orig-constants-enum*:")(newline)
        (display *prop-orig-constants-enum*)(newline)
        (display "*prop-orig-funs-enum*:")(newline)
        (display *prop-orig-funs-enum*)(newline) ))


     (do ((cnstpart *prop-orig-constants-enum* (cdr cnstpart))
          (maxel 0 (+ 1 maxel)) )
         ((or (>= maxel domainsize)
              (null? cnstpart) ))
       (set! res (cons (cons (cdar cnstpart) maxel) res)) )
     res))


(define (create-prop-flat-varlimits
               varslist fullpredclause domainsize)
   (let ((res '())
         (predabs 0)
         (tmp #f)
         (foundoneflag #f)
         (maxel (- domainsize 1))
         (foundflag #f) )

     (if *prop-debug*
         (begin
          (newline)
          (display "create-prop-flat-varlimits called:")
          (newline)
          (display "varslist:")(newline)
          (display varslist)(newline)
          (display "fullpredclause:")(newline)
          (display fullpredclause)(newline)
          (display "domainsize:")(newline)
          (display domainsize)(newline) ))

     ;(display "*prop-interp-value-pairs*:")(newline)
     ;(display *prop-interp-value-pairs*)(newline)
     ;(display "*prop-interp-termlist*:")(newline)
     ;(display *prop-interp-termlist*)(newline)
     ;(display "*prop-constants-mapping*:")(newline)
     ;(display *prop-constants-mapping*)(newline)
     ;(display "*prop-fun-mappings*:")(newline)
     ;(display *prop-fun-mappings*)(newline)
     ;(display "*prop-lit-mappings*:")(newline)
     ;(display *prop-lit-mappings*)(newline)
     ;(display "*prop-atom-maxnr*:")(newline)
     ;(display *prop-atom-maxnr*)(newline)
     ;(newline)(display "*prop-orig-constants-enum*:")
     ;(newline)(display *prop-orig-constants-enum*)
     ;(newline)(display "*prop-orig-funs-enum*:")
     ;(newline)(display *prop-orig-funs-enum*)

     ; *prop-lit-mappings*
     ; ((2 (1 (1 . 4) (0 . 3)) (0 (1 . 2) (0 . 1))) (1 (1 (1 . 8) (0 . 7)) (0 (1 . 6) (0 . 5))))
     ; *prop-orig-preds*
     ; ((2097152 . 2))
     ; *prop-orig-funs*
     ; ((2097152 . 1))
     ; *prop-orig-constants*
     ; ()

     (do ((varspart varslist (cdr varspart)))
         ((null? varspart))
       (set! foundoneflag #f)
       (do ((clpart (cdr fullpredclause) (cdr clpart)))
           ((or (null? clpart)
                foundoneflag)
             (if (not foundoneflag)
                 (set! res (cons maxel res)) ))
         (if (eq? (car varspart) (cadar clpart))
             (begin
               (if (< (caar clpart) 0)
                   (set! predabs (- 0 (caar clpart)))
                   (set! predabs (caar clpart)) )
               (set! tmp (assq predabs *prop-pred-maxval-pairs*))
               (if (and tmp (< (cdr tmp) maxel))
                   (begin
                     (set! foundoneflag #t)
                     (set! foundflag #t)
                     (set! res (cons (cdr tmp) res)) ))))))
     (if foundflag
         (set! res (reverse res))
         (set! res '()) )
     res))



(define *prop-funinfolist* '())

(define (make-extra-flatten-propclauses domainsize)
   (let ((tmp '())
         (res '())
         (wholeresult '())
         (closedaxiom '())
         (funinfo '())
         (var 0)
         (args '())
         (maxel '())
         (funinfolist '()) )
     ; create the input: funs and constants with their increased arities
     (set! funinfolist
         (append
       	    (map (lambda(x)
       		   (set! tmp (assq (car x) *prop-orig-funs-enum*))
       		   (cons (cdr tmp)(+ 1 (cdr x))))
       	         *prop-orig-funs*)
            (map (lambda(x)
       		   (set! tmp (assq x *prop-orig-constants-enum*))
       	           (cons (cdr tmp) 1))
       		 *prop-orig-constants*)))
     (set! *prop-funinfolist* funinfolist)
     ; loop over all funs and constants in the infolist
     (do ((lstpart funinfolist (cdr lstpart)))
         ((null? lstpart)
             (if *prop-debug*
                 (begin
                   (newline)
                   (display "extraclauses:")
                   (newline)
                   (for-each (lambda(x)(display x)(newline)) wholeresult)
                   (newline) ))
             wholeresult)
       (set! funinfo (car lstpart))
       (set! args '())
       ; create arglist main part for a concrete fun, one shorter than actual
       (do ((x 1 (+ 1 x)))
           ((>= x (cdr funinfo)))
         (set! args (cons x args)) )
       ; create closed and total axiom, where constants have been added 100000
       (set! tmp (assq (car funinfo) *prop-pred-maxval-pairs*))
       (if (and tmp (< (cdr tmp) (- domainsize 1)))
           (set! maxel (cdr tmp))
           (set! maxel (- domainsize 1)) )
       (set! closedaxiom '())
       (do ((y 0 (+ 1 y)))
           ((> y maxel))
         (set! tmp (cons (car funinfo)
                         (cons (+ y 100000) args) ))
         (set! closedaxiom (cons tmp closedaxiom)) )
       (set! closedaxiom (cons '() closedaxiom))
       (if (not (null? closedaxiom))
           (set! wholeresult (cons closedaxiom wholeresult)) )
       ; create the well-definedness axioms

       ; (newline)(display "welldefinedness for funinfo:") (display (car funinfo))

       (do ((u 0 (+ 1 u)))
           ((>= u domainsize))

         ; (newline)(display u)(display  ": ")

         (set! res '())
         (do ((v 0 (+ 1 v)))
             ((>= v u))

           ; (display v)(display " ")

           (set! res
             (list '()
                   (cons (- 0 (car funinfo)) (cons (+ v 100000) args))
         	   (cons (- 0 (car funinfo)) (cons (+ u 100000) args)) ))

            (if (not (null? res))
                (set! wholeresult (cons res wholeresult)) ))))))



(define (make-nofuns-equality-axioms)
   (let ((tmp '())
         (res '())
         (wholeresult '())
         (closedaxiom '())
         (funinfo '())
         (var 0)
         (args '())
         (maxarg 0)
         (funinfolist '()) )
     ; create the input: predicates and constants with their arities
     (set! funinfolist
       	    (map (lambda(x)
       		   (set! tmp (assq (car x) *prop-orig-preds-enum*))
       		   (cons (cdr tmp)(cdr x)))
       	         *prop-orig-preds*))

     ; make the standard equality axioms
     (set! wholeresult
           (cons (list '() (list *prop-flat-eq-symb* 1 1))
                 wholeresult))
     (set! wholeresult
           (cons (list '()
                       (list (- 0 *prop-flat-eq-symb*) 1 2)
                       (list *prop-flat-eq-symb* 2 1) )
                 wholeresult))
     (set! wholeresult
           (cons (list '()
                       (list (- 0 *prop-flat-eq-symb*) 1 2)
                       (list (- 0 *prop-flat-eq-symb*) 2 3)
                       (list *prop-flat-eq-symb* 1 3) )
                 wholeresult))
     ; loop over all predicates in the infolist

     (do ((lstpart funinfolist (cdr lstpart)))
         ((null? lstpart) wholeresult)
       (set! funinfo (car lstpart))
       (if (not (eq? *prop-flat-eq-symb* (car funinfo)))
           (begin
             (set! args '())
             ; create arglist for a predicate list
             (do ((x 1 (+ 1 x)))
                 ((> x (cdr funinfo))
                   (set! maxarg x) )
              (set! args (cons x args)) )
             ; loop over arglist, creating axioms for each pos
             (do ((x 1 (+ 1 x)))
                 ((> x (cdr funinfo)))
               (set! wholeresult
                    (cons (list '()
                                (list (- 0 *prop-flat-eq-symb*) maxarg (+ 1 maxarg))
                                (cons (- 0 (car funinfo))
                                      (replace-one-eq x maxarg args) )
                                 (cons (car funinfo)
                                      (replace-one-eq x (+ 1 maxarg) args) ))
                          wholeresult))))))
     ; return the result
     wholeresult))





(define (add-prop-lit-to-clause lit clause)
    (cond ((null? clause) (cons lit '()))
          ((<= lit (car clause)) (cons lit clause))
          (else (cons (car clause)
                      (add-prop-lit-to-clause lit (cdr clause)) ))))



(define (add-new-prop-clause propclause)
  (let ((tmp #t))
    (if  *prop-debug*
         (begin (newline)(display "add-new-prop-cl: ")(display propclause)) )
   ; if unit clause, assign the literal value
   (cond
     ((and (null? (car propclause))
           (null? (cddr propclause)) )
         (vector-set! *propvar-enqued-value* (cadr propclause) #t) )
     ((and (null? (cdr propclause))
           (null? (cdar propclause)) )
         (vector-set! *propvar-enqued-value* (caar propclause) #f) ))
        ; add clause to the discrimination tree
        ;(set! tmp (add-discr-tree-propclause (car propclause) (cdr propclause)))
        ;(set! tmp (check-exist-propclause (car propclause) (cdr propclause)))
   ; add clause if not subsumed
   (if (and *prop-conversion-to-file* *sat-strategy-flag*)
       (begin
         (output-prop-clause propclause *prop-conversion-temp-file-port*) )
       (set! *propclauses* (cons propclause *propclauses*)) )

   #t))


(define (check-exist-propclause neglits poslits)
  (let ((found #f))
   (do ((part *propclauses* (cdr part)))
       ((or found (null? part))
         ;(display found)
         (not found))

    ;(newline)
    ;(display "car part: ")(display (car part))
    ;(display " neg: ")(display neglits)(newline)
    ;(display " pos: ")(display poslits)(newline)

    (if (and (sub-ordered-list? (caar part) neglits)
             (sub-ordered-list? (cdar part) poslits)  )
        (set! found #t) ))))


(define (sub-ordered-list? slist glist)
  (or (null? slist)
      (and (not (null? glist))
           (or (and (eq? (car slist) (car glist))
           	    (sub-ordered-list? (cdr slist) (cdr glist)) )
	       (and (> (car slist) (car glist))
	            (sub-ordered-list? slist (cdr glist)) )))))



; ((1 (-2 #t))
;
;
;


;(define (add-discr-tree-propclause tree neglits poslits)
;   (let* ((treepart '())
;          (tmp '())
;          (nneq neglits)
;          (npos poslits)
;          (nlit '())
;     (if (null? neglits)
;         (if (null? poslits)
;             '()
;             (begin (set! npos (cdr poslits))
;                    (set! nlit (car poslits)) ))
;         (begin (set! nneg (cdr neglits))
;                (set! nlit (- 0 (car neglits))) ))
;     (set! treepart (assq nlit (cdr tree)))
;     (if treepart
;        (set-cdr! tree
;                (cons
;                   (add-discr-tree-propclause tree nneg npos)
;                   (cdr tree)))))))


(define (get-fullclause-varslist fullpredclause)
   (set! *tmp-varlist* '())
   (for-each (lambda(lit)(get-fullclause-varslist-aux! lit))
             (cdr fullpredclause) )
   *tmp-varlist*)


(define (get-fullclause-varslist-aux! term)

    (if *prop-debug*
          (begin (newline)(display "term: ")(display term)
    		 (newline) (display "fun-may-cont-vars: ")
    		 (display (fun-may-cont-vars? (car term)))(newline) ))

    (do ((part (cdr term) (cdr part)))
	((null? part))
      (set! term (car part))

    (if *prop-debug*
          (begin
    		(newline)(display "inner term: ")(display term)
    		(newline) (display "fun-may-cont-vars: ")
    		(if (pair? term)(display (fun-may-cont-vars? (car term))))(newline) ))


      (if (non-composite? term)
	  (if (non-composite:-var? term)
	       (if (not (memq (var-nr term) *tmp-varlist*))
	           (set! *tmp-varlist* (cons (var-nr term) *tmp-varlist*) )))
	  (if #t ; (fun-may-cont-vars? (car term))
	      (get-fullclause-varslist-aux! term) ))))


(define (set-pred-var-values varslist varvaluations)

   ;(if *prop-debug*
   ;       (begin
   ; 		(newline)(display "varslist: ")(display varslist)
   ;		(newline)(display "varvaluations: ")(display varvaluations)
   ;		(newline) ))

   (for-each (lambda(var val)(set-var-ref! var val))
              varslist
              varvaluations))

(define (set-pred-var-values-limited varslist varvaluations varlimits)

   ;(if *prop-debug*
   ;       (begin
   ; 		(newline)(display "varslit: ")(display varslit)
   ;		(newline)(display "varvaluations: ")(display varvaluations)
   ;		(newline) ))
  (do ((part1 varslist (cdr part1))
       (part2 varvaluations (cdr part2))
       (part3 varlimits (cdr part3)) )
      ((or (null? part2)
           (> (car part2) (car part3)) )
         (if (null? part2)
             #t
             #f))
    (set-var-ref! (car part1) (car part2)) ))


(define (make-pred-to-proplit curlit)
  (let* ((negflag (neg-fun? (car curlit)))
         (purefun (prop-spec-pure-unsigned-fun (car curlit)))
         (arg1 '())
         (arg2 '())
         (treepart '())
         (res '()) )

       (if *prop-debug*
          (begin
       	   (display "curlit: ")(display curlit)(newline)
       	   (display "*prop-lit-mappings*: ")(display *prop-lit-mappings*)(newline) ))

      (if (pos-eq-pred? purefun)
        ; equality case
        (begin
          (if (eq? (get-pred-term-interpvalue (cadr curlit))
                   (get-pred-term-interpvalue (caddr curlit)) )
              (if negflag (set! res #f) (set! res #t))
              (if negflag (set! res #t) (set! res #f)) )
          #f)
        ; standard lit case
        (begin
         (set! treepart (assq purefun *prop-lit-mappings*))
         (if *prop-debug*
             (begin (display "treepart: ")(display treepart)(newline)) )

         (if (not treepart)
             (begin (display "error in pred-to-prop pred evaluation")(quit)) )
         (set! treepart (make-pred-to-proplit-aux (cdr curlit) (cdr treepart)))
         (if negflag
             (set! res (- 0 treepart))
             (set! res treepart) )))
     res))


(define (make-pred-to-proplit-incremental curlit)
  (let* ((negflag (neg-fun? (car curlit)))
         (purefun (prop-spec-pure-unsigned-fun (car curlit)))
         (arg1 '())
         (arg2 '())
         (treepart '())
         (res '()) )

       (if *prop-debug*
          (begin
           (newline)
       	   (display "curlit: ")(display curlit)
       	   ;(newline)(display "*prop-lit-mappings*: ")(display *prop-lit-mappings*)
       	   (newline) ))

      (if (pos-eq-pred? purefun)
        ; equality case
        (begin
          (set! arg1 (get-pred-term-interpvalue-incremental (cadr curlit)))
          (set! arg2 (get-pred-term-interpvalue-incremental (caddr curlit)))

          ;(newline)
          ;(display "arg1: ")(display arg1)
          ;(display "arg2: ")(display arg2)
          ;(newline)

          (cond
            ((not (integer? arg1)) (set! res #t))
            ((not (integer? arg2)) (set! res #t))
            ((eq? arg1 arg2)
              (if negflag (set! res #f) (set! res #t)) )
            (else
              (if negflag (set! res #t) (set! res #f)) ))
          #f)
        ; standard lit case
        (begin

         ;(newline)
         ;(display "was standardlit")
         ;(newline)

         (set! treepart (assq purefun *prop-lit-mappings*))
         (if *prop-debug*
             (begin (display "treepart: ")(display treepart)(newline)) )

         (if (not treepart)
             (begin (display "error in pred-to-prop pred evaluation")(quit)) )
         (set! treepart (make-pred-to-proplit-incremental-aux (cdr curlit) (cdr treepart)))

         ;(newline)
         ;(display "treepart: ")(display treepart)
         ;(newline)

         (cond
           ((eq? #t treepart)
             (set! res #t) )
           (negflag
             (set! res (- 0 treepart)) )
           (else
             (set! res treepart) ))))
     res))



(define (make-flat-pred-to-proplit curlit)
  (let* ((negflag (< (car curlit) 0))
         (purefun (if negflag (- 0 (car curlit)) (car curlit)))
         (arg1 '())
         (arg2 '())
         (treepart '())
         (res '()) )

       (if *prop-debug*
          (begin
           (newline)
           (display "make-flat-pred-to-proplit called on: ")(newline)
       	   (display "curlit: ")(display curlit)(newline)
       	   (display "*prop-flat-eq-symb*: ")(display *prop-flat-eq-symb*)(newline)
       	   ;(display "*prop-lit-mappings*: ")(display *prop-lit-mappings*)(newline)
           ))

      (if (and (eq? *prop-flat-eq-symb* purefun)
               (not *mace-nofuns-model-strat*) )
         ; equality case
         ; here different values in eq must be really different
        (if (eq? (get-pred-term-interpvalue (cadr curlit))
                   (get-pred-term-interpvalue (caddr curlit)) )
            (if negflag (set! res #f) (set! res #t))
            (if negflag (set! res #t) (set! res #f)) )
        ; standard lit case
        (begin
         (set! treepart (assq purefun *prop-lit-mappings*))
         (if *prop-debug*
             (begin (display "treepart: ")(display treepart)(newline)) )

         (if (not treepart)
             (begin (display "error in pred-to-prop pred evaluation")(quit)) )
         (set! treepart (make-flat-pred-to-proplit-aux (cdr curlit) (cdr treepart)))
         (if negflag
             (set! res (- 0 treepart))
             (set! res treepart) )))
     res))


; check arglist of a fun

(define (make-pred-to-proplit-aux funargs treepart)
  (let ((tmp '())
        (argval '()) )
    ; cycle over all funargs
    (do ((argspart funargs (cdr argspart)))
       ((null? argspart) treepart)
     ; evaluate interpreted funarg
     (set! argval (get-pred-term-interpvalue (car argspart)))

     ; go further in the tree

     (if *prop-debug*
          (begin
     		(display "funargs: ")(display funargs)(newline)
     		(display "argval: ")(display argval)(newline)
     		;(display "treepart: ")(display treepart)(newline)
        ))

     (set! tmp (assq argval treepart))


     (if (not tmp)
         (begin (display "error in pred-to-prop fun evaluation")(abort)) )
     (set! treepart (cdr tmp)) )))


(define (make-pred-to-proplit-incremental-aux funargs treepart)
  (let ((tmp '())
        (stopflag #f)
        (argval '()) )

    ;(newline)
    ;(display "make-pred-to-proplit-incremental-aux called with: ")
    ;(newline)(display funargs)(newline)
    ;(display treepart)
    ;(newline)

    ; cycle over all funargs
    (do ((argspart funargs (cdr argspart)))
       ((or (null? argspart) stopflag) treepart)
     ; evaluate interpreted funarg
     (set! argval (get-pred-term-interpvalue-incremental (car argspart)))

     ; go further in the tree

     (if *prop-debug*
          (begin
     		(display "funargs: ")(display funargs)(newline)
     		(display "argval: ")(display argval)(newline)
     		;(display "treepart: ")(display treepart)(newline)
        ))

     (cond
       ((boolean? argval)
          (set! stopflag #t)
          (set! treepart #t) )
       (else
         (set! tmp (assq argval treepart))
         (cond
         ((not tmp)
            (set! stopflag #t)
            (set! treepart #t) )
         ((eq? tmp #t)
            (set! stopflag #t)
            (set! treepart #t) )
         (else
            (set! treepart (cdr tmp)) )))))))

(define (make-flat-pred-to-proplit-aux funargs treepart)
  (let ((tmp '())
        (arg '())
        (argval '()) )

    ; cycle over all funargs
    (do ((argspart funargs (cdr argspart)))
       ((null? argspart) treepart)
     (set! arg (car argspart))
     ; evaluate interpreted funarg
     (cond
       ((<= 100000 arg)
          (set! argval (- arg 100000)) ) ; constant in special new axioms
       ((<= *prop-const-min* arg)
          (set! argval (- arg *prop-const-min*)) ) ; splitting constant
       (else
         (set! argval (var-ref (var-nr arg))) )) ; variable

     ; go further in the tree

     (if  *prop-debug*
          (begin
     		(display "argspart: ")(display argspart)(newline)
     		(display "argval: ")(display argval)(newline)
     		(display "treepart: ")(display treepart)(newline) ))

     (set! tmp (assq argval treepart))

     (if (not tmp)
         (begin (display "error in pred-to-prop fun evaluation")(abort)) )
     (set! treepart (cdr tmp)) )))

; check a full term

(define (get-pred-term-interpvalue term)
  (let ((tmp '())
        (treepart '()) )
    (if (non-composite? term)
        (if (non-composite:-var? term)
            (var-ref (var-nr term))
            (cond
              ((<= 100000 term) (- term 100000))
              (else
                 (set! tmp (assq term *prop-constants-mapping*))
                 (if (not tmp)
                     (begin (display "error in pred-to-prop fun evaluation")(quit)) )
                 (cdr tmp) )))
        (begin
          (set! treepart (assq (prop-spec-pure-fun (car term)) *prop-fun-mappings*))
          (if (not treepart)
              (begin (display "error in pred-to-prop fun evaluation")(quit)) )
          (make-pred-to-proplit-aux (cdr term) (cdr treepart)) ))))


(define (get-pred-term-interpvalue-incremental term)

   ;(newline)
   ;(display "get-pred-term-interpvalue-incremental called with: ")
   ;(newline)(display term)(newline)
   ;(newline)

  (let ((tmp '())
        (treepart '()) )
    (if (non-composite? term)
        (if (non-composite:-var? term)
            (var-ref (var-nr term))
            (begin
              (set! tmp (assq term *prop-constants-mapping*))
              (if (not tmp)
                  #t
                  (cdr tmp) )))
        (begin
          (set! treepart (assq (prop-spec-pure-fun (car term)) *prop-fun-mappings*))
          (if (not treepart)
              #t
              (make-pred-to-proplit-incremental-aux (cdr term) (cdr treepart)) )))))

;----------------------------------------------------------------------
;
;       flattening a first-order formula
;
;----------------------------------------------------------------------
;
; each first-order clause is first converted to a simple form with
; numbers representing predicate and function symbols, plus constants
; represented as 0-argument function applications, plus vars represented
; as integers.
;
; afterward n clauses are built from the converted clause
;
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define *prop-flatten-lastvar* 0)
(define *prop-flatten-newlits* '())
(define *prop-flatten-varslist* '())
(define *prop-flatten-clause-processed-terms* '())

(define (flatten-clause-set clauses)
 (let ((tmp '())
       (clause '())
       (res '()))
   (do ((clpart clauses (cdr clpart)))
       ((null? clpart)
          res)
     (set! clause (car clpart))
     (if (and (pair? clause)(pair? (cdr clause)))
         (begin

            ;(newline)(display "about to flatten clause: ")
            ;(print-clause clause)(newline)


           (set! tmp (flatten-clause clause))

            ;(display "flattening gives: ")(newline)
            ;(for-each (lambda(x)(display x)(newline)) tmp)

           (set! res (append tmp res)) )))))


; flatten-clause takes a predclause and returns the set of fully flattened
; clauses, having at the head the list of all vars in the new clause

(define (flatten-clause clause)

   ; (display "starting to flatten clause: ")(display clause)(newline)

   (let ((preprocessed '()))
     (set! *prop-flatten-lastvar* 100)
     (set! preprocessed (preprocess-flatten-clause (cdr clause)))
     (if (>= *prop-flatten-lastvar* 150)
         (begin (display "too large clause for flattening satisfiability check")(quit)) )

     ;(newline)(display "flatten-clause gave preprocessed set: ")(newline)
     ;(for-each (lambda(x)(display x)(newline)) preprocessed)


     (map (lambda(cl)(flatten-clause-aux cl))
          preprocessed) ))

; preprocess-flatten-clause creates a list of clauses from a single clause, splitting
; all the positive equalities

(define (preprocess-flatten-clause pureclause)
  (let ((lit '())
        (fun '())
        (purefun '())
        (negfun '())
        (negflag #f)
        (eqflag #f)
        (var '())
        (l1 '())
        (l2 '())
        (tmp '()))

    ;(newline)(display "preprocess-flatten-clause called with: ")(newline)
    ;(display pureclause)(newline)

    (if (null? pureclause)
        (list '())
        (begin
          (set! lit (car pureclause))
          (set! fun (car lit))
          (set! purefun (prop-spec-pure-unsigned-fun fun))
          (set! negflag (neg-fun? fun))
          (set! eqflag (pos-eq-pred? purefun))
          (if (and eqflag
                   (not negflag)
                   (not (and (non-composite? (cadr lit)) ; not var?
                             (non-composite:-var? (cadr lit)) ))
                   (not (and (non-composite? (caddr lit)) ; not var?
                             (non-composite:-var? (caddr lit)) )))
              ; both args are non-var
              (begin

                 ; (newline)(display "both args non-var")(newline)

                 (set! var (make-prop-flatten-newvar))
                 (set! l1 (list purefun (cadr lit) var))
                 (set! l2 (list purefun (caddr lit) var))
                 (set! negfun (logior *sgn-mask* purefun))
                 (set! tmp (preprocess-flatten-clause (cdr pureclause)))

                 ; (newline)
                 ; (display "var: ")(display var)(newline)
                 ; (display "l1: ")(display l1)(newline)
                 ; (display "l2: ")(display l2)(newline)
                 ; (display "negfun: ")(display negfun)(newline)
                 ; (newline)


                 (append (map (lambda(cl)(cons l1 (cons (cons negfun (cdr l2)) cl)))
                              tmp)
                         (map (lambda(cl)(cons (cons negfun (cdr l1)) (cons l2 cl)))
                              tmp)))

              ; either not equality or at least one of the vars was var
              (begin

                 ; (newline)(display "not both args non-var")(newline)

                 (set! tmp (preprocess-flatten-clause (cdr pureclause)))
                 (map (lambda(cl)(cons lit cl)) tmp) ))))))



; flatten-clause-aux takes a full predclause WITHOUT SPLITTABLE EQUALITY
; and returns the fully flattened
; clause, having at the head the list of all vars in the new clause.


(define (flatten-clause-aux clause)
 (let ((lit '())
       (fun '())
       (purefun '())
       (negflag #f)
       (eqflag #f)
       (varslist (get-fullclause-varslist clause))
       (res '())
       (processed-lit '()) )

   ; (newline)(display "original varslist: ")(display varslist)(newline)

   (if (> (length varslist) 40)
       (begin
         (display "too many vars (over 40) in the clause for the sat checker")
         (quit) ))

   (set! *prop-flatten-varslist* varslist)
   (set! *prop-flatten-lastvar* 150)
   (set! *prop-flatten-newlits* '())
   (set! *prop-flatten-clause-processed-terms* '())
   (do ((clpart clause (cdr clpart)))
       ((null? clpart)
         (if (> *prop-flatten-lastvar* 250)
             (begin
              (display "too many vars (over 250) in the clause for the sat checker")
              (quit) ))
           (cons '() (append (reverse *prop-flatten-newlits*) (reverse res))) )
       ;  (cons
       ;    *prop-flatten-varslist*
       ;     (append (reverse *prop-flatten-newlits*) (reverse res)) ))
     (set! processed-lit (flatten-composite-term (car clpart) #t))
     (set! res (cons processed-lit res)) )))


(define (flatten-composite-term term litflag)
 (let ((funsymb '())
       (fun '())
       (negflag #f)
       (eqflag #f)
       (purefun '())
       (tmp '())
       (newargs '())
       (newvar '()) )

   ; (newline)(display "about to flatten comp term: ")(display term)

  ; first process the leading symbol of the term
  (if (not litflag)
    ; process leading function symbol
    (begin
        (set! funsymb
             (cdr (assq (prop-spec-pure-fun (car term)) *prop-orig-funs-enum*)) )
        (set! term (cons funsymb (cdr term))) )
    ; process predicate symbol
    (begin
        (set! fun (car term))
        (set! purefun (prop-spec-pure-unsigned-fun fun))
        (set! negflag (neg-fun? fun))
        (set! eqflag (pos-eq-pred? purefun))
        (set! funsymb (cdr (assq purefun *prop-orig-preds-enum*)))

        ; (newline)
        ; (display "purefun: ")(display purefun)(newline)
        ; (display "eqflag: ")(display eqflag)(newline)
        ; (display "negflag: ")(display negflag)(newline)
        ; (newline)


        (cond
          ; not an equality literal
          ((not eqflag)
             (if negflag (set! funsymb (- 0 funsymb)))
             (set! term (cons funsymb (cdr term))) )
          ; eq, both args vars, handle normally as eq
          ((and eqflag
                (non-composite? (cadr term))
                (non-composite:-var? (cadr term))
                (non-composite? (caddr term))
                (non-composite:-var? (caddr term)) )
             (if negflag (set! funsymb (- 0 funsymb)))
             (set! term (cons funsymb (cdr term))) )
             ; (display "equality has two var args in flatten-composite-term")(abort)
          ; eq, first arg var?
          ((and eqflag
                (non-composite? (cadr term))
                (non-composite:-var? (cadr term)) )
            (if (non-composite? (caddr term))
                (begin
                  (set! tmp (assq (caddr term) *prop-orig-constants-enum*))
                  (set! term
                        (cons (if negflag (- 0 (cdr tmp)) (cdr tmp))
                              (cons (cadr term) '())) ))
                (begin
                  (set! tmp (assq (prop-spec-pure-fun (car (caddr term))) *prop-orig-funs-enum*))
                  (set! term
                        (cons (if negflag (- 0 (cdr tmp)) (cdr tmp))
                              (cons (cadr term) (cdr (caddr term)))) ))))
          ; eq, second arg var?
          ((and eqflag
                (non-composite? (caddr term))
                (non-composite:-var? (caddr term)) )

            ; (display "second eq arg var")(newline)

            (if (non-composite? (cadr term))
                (begin
                  (set! tmp (assq (cadr term) *prop-orig-constants-enum*))
                  (set! term
                        (cons (if negflag (- 0 (cdr tmp)) (cdr tmp))
                              (cons (caddr term) '())) ))
                (begin
                  (set! tmp (assq (prop-spec-pure-fun (car (cadr term))) *prop-orig-funs-enum*))
                  (set! term
                        (cons (if negflag (- 0 (cdr tmp)) (cdr tmp))
                              (cons (caddr term) (cdr (cadr term)))) ))))
          ; negative eq, none of the args var
          ((and eqflag negflag)
             (set! term (cons (- 0 funsymb) (cdr term))) )
          ; positive eq, none of the args var: ERROR!
          (else
             (display "equality has two non-var args in flatten-composite-term")
             (abort) ))))


   ; (newline)(display "term after funsymb-processing: ")(display term)


  (set! newargs
   (map
    (lambda(curterm)
     (if (non-composite? curterm)
         (if (non-composite:-var? curterm)
            ; (var-ref (var-nr curterm))
            (set! tmp (var-nr curterm))
            (begin
              ; test if an equal curterm already processed in the clause
              (set! tmp (assoc curterm *prop-flatten-clause-processed-terms*))
              (if tmp
                  ; an equal term already processed, use assigned var
                  (set! tmp (cdr tmp))
                  ; not processed yet, create a new var
                  (begin
              	    (set! tmp (assq curterm *prop-orig-constants-enum*))
                    (if (not tmp)
                        (begin (display "error in pred-to-prop fun evaluation")(quit)) )
                    (set! newvar (make-prop-flatten-newvar))
                    (make-prop-flatten-addpair curterm (list (cdr tmp)) newvar)
                    (set! tmp newvar) ))))
        (begin
          ; test if an equal term already processed in the clause
          (set! tmp (assoc curterm *prop-flatten-clause-processed-terms*))
          (if tmp
              ; an equal term already processed, use assigned var
              (set! tmp (cdr tmp))
              ; not processed yet, create a new var
              (begin
                (set! tmp (flatten-composite-term curterm #f))
                (set! newvar (make-prop-flatten-newvar))
                (make-prop-flatten-addpair curterm tmp newvar)
                (set! tmp newvar) ))))
      tmp)
    (cdr term) ))

   ; (newline)(display "newargs found for the term: ")(display newargs)

   ; compose the whole
   (cons (car term) newargs) ))


(define (make-prop-flatten-newvar)
  (set! *prop-flatten-lastvar* (+ 1 *prop-flatten-lastvar*))
  (set! *prop-flatten-varslist* (cons *prop-flatten-lastvar* *prop-flatten-varslist*))
  *prop-flatten-lastvar*)


(define (make-prop-flatten-addpair origterm newterm newvar)
  (let ((newlit '()))
     (set! *prop-flatten-clause-processed-terms*
           (cons (cons origterm newvar) *prop-flatten-clause-processed-terms*) )
     (set! newlit (cons (- 0 (car newterm))
                        (cons newvar
                              (cdr newterm) )))
     (set! *prop-flatten-newlits*
          (cons newlit *prop-flatten-newlits*) )))



(define (pseudo-flatten-clause-set clauses)
 (let ((tmp '())
       (clause '())
       (res '()))
   (do ((clpart clauses (cdr clpart)))
       ((null? clpart)
          res)
     (set! clause (car clpart))
     (if (and (pair? clause)(pair? (cdr clause)))
         (begin
           (set! tmp (pseudo-flatten-clause clause))
           (set! res (append tmp res)) )))))


; pseudo-flatten-clause takes a predclause and returns the set of fully flattened
; clauses, having at the head the list of all vars in the new clause

(define (pseudo-flatten-clause clause)

   ; (display "starting to flatten clause: ")(display clause)(newline)

   (let ((preprocessed '()))
     (set! *prop-flatten-lastvar* 100)
     (set! preprocessed (pseudo-preprocess-flatten-clause (cdr clause)))
     ;(set! preprocessed (list clause))
     (if (>= *prop-flatten-lastvar* 150)
         (begin (display "too large clause for flattening satisfiability check")(quit)) )

     ;(newline)(display "flatten-clause gave preprocessed set: ")(newline)
     ;(for-each (lambda(x)(display x)(newline)) preprocessed)

     (map (lambda(cl)(pseudo-flatten-clause-aux cl))
          preprocessed) ))



(define (pseudo-preprocess-flatten-clause pureclause)
  (let ((lit '())
        (fun '())
        (purefun '())
        (negfun '())
        (negflag #f)
        (eqflag #f)
        (var '())
        (l1 '())
        (l2 '())
        (tmp '()))

    ;(newline)(display "preprocess-flatten-clause called with: ")(newline)
    ;(display pureclause)(newline)

    (if (null? pureclause)
        (list '())
        (begin
          (set! lit (car pureclause))
          (set! fun (car lit))
          (set! purefun (prop-spec-pure-unsigned-fun fun))
          (set! negflag (neg-fun? fun))
          (set! eqflag (pos-eq-pred? purefun))
          (set! tmp (pseudo-preprocess-flatten-clause (cdr pureclause)))
          (map (lambda(cl)(cons lit cl)) tmp) ))))

(define (pseudo-flatten-clause-aux clause)
 (let ((lit '())
       (fun '())
       (purefun '())
       (negflag #f)
       (eqflag #f)
       (varslist (get-fullclause-varslist clause))
       (res '())
       (processed-lit '()) )

   ; (newline)(display "original varslist: ")(display varslist)(newline)

   (if (> (length varslist) 40)
       (begin
         (display "too many vars (over 40) in the clause for the sat checker")
         (quit) ))
   (set! *prop-flatten-varslist* varslist)
   (set! *prop-flatten-lastvar* 150)
   (set! *prop-flatten-newlits* '())
   (set! *prop-flatten-clause-processed-terms* '())
   (do ((clpart clause (cdr clpart)))
       ((null? clpart)
         (if (> *prop-flatten-lastvar* 250)
             (begin
              (display "too many vars (over 250) in the clause for the sat checker")
              (quit) ))
           (cons '() (append (reverse *prop-flatten-newlits*) (reverse res))) )
       ;  (cons
       ;    *prop-flatten-varslist*
       ;     (append (reverse *prop-flatten-newlits*) (reverse res)) ))
     (set! processed-lit (pseudo-flatten-composite-term (car clpart) #t))
     (set! res (cons processed-lit res)) )))


(define (pseudo-flatten-composite-term term litflag)
 (let ((funsymb '())
       (fun '())
       (negflag #f)
       (eqflag #f)
       (purefun '())
       (tmp '())
       (newargs '())
       (newvar '()) )

  ;(newline)(display "about to flatten comp term: ")(display term)

  (set! fun (car term))
  (set! purefun (prop-spec-pure-unsigned-fun fun))
  (set! negflag (neg-fun? fun))
  (set! eqflag (pos-eq-pred? purefun))
  (set! funsymb (cdr (assq purefun *prop-orig-preds-enum*)))

        ; (newline)
        ; (display "purefun: ")(display purefun)(newline)
        ; (display "eqflag: ")(display eqflag)(newline)
        ; (display "negflag: ")(display negflag)(newline)
        ; (newline)

  (if negflag (set! funsymb (- 0 funsymb)))
  (set! term (cons funsymb (cdr term)) )


   ; (newline)(display "term after funsymb-processing: ")(display term)


  (set! newargs
   (map
    (lambda(curterm)
     (if (non-composite? curterm)
         (if (non-composite:-var? curterm)
            ; (var-ref (var-nr curterm))
            (set! tmp (var-nr curterm))
            (begin
              (set! tmp (assq curterm *prop-orig-constants-to-flat-domainels*))
              (set! tmp (cdr tmp)) ))
        (begin
          (newline)
          (display "ERROR: function symbol detected")(newline)
          (quit) ))
      tmp)
    (cdr term) ))

   ; (newline)(display "newargs found for the term: ")(display newargs)

   ; compose the whole
   (cons (car term) newargs) ))



;----------------------------------------------------------------------
;
;       prop formula solving: unsatisfiable or not?
;
;----------------------------------------------------------------------
;
;  variable:
;   an integer, pointing to four vectors:
;   *propvar-value*                   vector containing (), #t or #f
;   *propvar-negative-occ-clauses*    vector pointing to lists of clauses
;   *propvar-positive-occ-clauses*    vector pointing to lists of clauses
;   *propvar-enqued-value*            vector
;
;  clause:
;
;   propclause-nr-of-negative-active
;   propclause-nr-of-positive-active
;   negative-literals
;   positive-literals
;   subsumer
;
;
;  clause list:
;
;  queue:
;   *prop-var-assignment-queue*
;
;  backtrace data:
;   *prop-assigned-varvalpairs*
;   *prop-unsat-found-occlistpart*
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define *prop-debug* #f)

(define *propvar-value*  '())
(define *propvar-negative-occ-clauses* '())
(define *propvar-positive-occ-clauses* '())
(define *propvar-enqued-value* '())
(define *prop-var-assignment-queue* '())
(define *prop-assigned-varvalpairs* '())
(define *prop-unsat-found-occlistpart* '())

; create-prop-var-vectors creates necessary propvar vectors

(define (create-prop-var-vectors maxvar)
  (set! *propvar-value* (make-vector (+ 1 maxvar) '()))
  (set! *propvar-positive-occ-clauses* (make-vector (+ 1 maxvar) '()))
  (set! *propvar-negative-occ-clauses* (make-vector (+ 1 maxvar) '()))
  (set! *propvar-enqued-value* (make-vector (+ 1 maxvar) '())) )

; initialize-prop-var-values sets the values of vars to initial null status

(define (initialize-prop-var-values maxvar)
  (vector-fill! *propvar-value* '())
  (vector-fill! *propvar-positive-occ-clauses* '())
  (vector-fill! *propvar-negative-occ-clauses* '())
  (vector-fill! *propvar-enqued-value* '()) )

 ;(do ((var 0 (+ 1 var)))
 ;     ((> var maxvar))
 ;  (vector-set! *propvar-value* var '())
 ;  (vector-set! *propvar-positive-occ-clauses* var '())
 ;  (vector-set! *propvar-negative-occ-clauses* var '())
 ;  (vector-set! *propvar-enqued-value* var '()) ))

; build-prop-internal-structure creates an ldpp-usable clause list
; and sets the variable vectors to point to right values
;
; input clauses are assumed to be a list of clauses, where each clause
; is a pair of two lists: positive occ var nrs and negative occ var nrs.


(define (build-prop-internal-structure clauses)
  (let ((resclauses '())
        (tmp '())
        (var '())
        (repeatflag #f)
        (resclause '())
        (neglits '())
        (poslits '()) )
    (do ((part clauses (cdr part)))
        ((null? part) resclauses)
      (set! poslits (cdar part))
      (set! neglits (caar part))

      (if *prop-debug*
          (begin (display poslits)(newline)(display neglits)(newline)) )

      ; create a new clause
      (set! resclause
           (list (length neglits) (length poslits) neglits poslits '()) )
      (set! resclauses (cons resclause resclauses))
      ; loop over poslits, set variable values
      (do ((part poslits (cdr part)))
          ((null? part))
        (set! var (car part))
        (vector-set! *propvar-positive-occ-clauses* var
           (cons resclause (vector-ref *propvar-positive-occ-clauses* var)) ))
      ; loop over neglits, set variable values
      (do ((part neglits (cdr part)))
          ((null? part))
        (set! var (car part))
        (vector-set! *propvar-negative-occ-clauses* var
           (cons resclause (vector-ref *propvar-negative-occ-clauses* var)) )))))


; solve-prop-clauses gives #t if satisfiable, #f if not,
;
; input clauses are assumed to be a list of clauses, where each clause
; is a pair of two lists: positive occ var nrs and negative occ var nrs.
;
; maxvar should be at least as big as the biggest var nr. OK to be bigger.
;
; variable value vectors must be created before.

(define (solve-prop-clauses clauses maxvar)
  (let ((splittableclauses '())
        (newsplitvar '())
        (internalclauses '())
        (res '()) )
    (initialize-prop-var-values maxvar)

    (if *prop-debug*  (begin
      (newline)(display "-- prop values initialised")(newline)
      (show-prop-internal-data '() maxvar) ))

    (set! internalclauses (build-prop-internal-structure clauses))
    (set! *prop-internal-clauses* internalclauses)
    (set! *prop-maxvar* maxvar)

    (if *prop-debug*
     (begin
      (newline)(display "-- prop internal structure built")(newline)
      (show-prop-internal-data internalclauses maxvar) ))

    (if #f ;(or *mace-model-strat*
           ;   (null? *prop-orig-funs*)
           ;   (and *prop-unsat-strat* *simple-model-strat*) )
        (begin
          (newline)
          (display (length  clauses))
          (display " propositional clauses containing ")
          (display maxvar)
          (display " variables were built.")
          (newline) ))



    (if (> *display-info-level* 90)
           (begin (newline)
                  (display "clauses for full prop evaluation: ")
                  (newline)
	          (for-each (lambda(cl)(display cl)(newline)) *prop-internal-clauses*)
		  (newline) ))

    (if #f
     (begin
       (newline)(newline)(newline)
       (for-each
         (lambda(cl)
           (for-each (lambda(n)
                      (display "-")(display "p")(display n)(display " | ") )
                     (caddr cl) )
           (for-each (lambda(p)
                      (display "p")(display p)(display " | ") )
                     (cadddr cl) )
          (display ".")
          (newline) )
         *prop-internal-clauses*)
       (newline)(newline)(newline) ))


    (set! *prop-var-assignment-queue* '())
    (set! *prop-assigned-varvalpairs* '())
    (set! *prop-unsat-found-occlistpart* '())
    ; try simple propagation of all units first
    (set! res (initial-prop-unit-propagate internalclauses)) ; if returns #t, found unsat

    (if *prop-debug*
     (begin
      (newline)(display "-- initial-prop-unit-propagate finished:")(newline)
      (display "  satisfiable?: ")(display (not res))(newline)
      (show-prop-internal-data internalclauses maxvar) ))

    ; (prop-try-subsume internalclauses)

    (if res
        #f ; unsatisfiable
        (begin
          ; compute splittable
          (set! splittableclauses (prop-find-splittable internalclauses))
          ; (set! splittableclauses (reverse splittableclauses))
          ; (set! splittableclauses (prop-sort-splittable splittableclauses))

          ;(display "splittable: ")
          ;(newline)
          ;(for-each (lambda(x)
          ;            (display x)
          ;            (for-each (lambda(x)(display " ")
          ;                                (display
          ;                                 (length
          ;                                  (vector-ref *propvar-negative-occ-clauses*
          ;                                              x)))
          ;                                (display ":")
          ;                                (display
          ;                                 (length
          ;                                  (vector-ref *propvar-positive-occ-clauses*
          ;                                              x))) )
          ;                      (caddr x) )
          ;              (for-each (lambda(x)(display " ")
          ;                                (display
          ;                                 (length
          ;                                  (vector-ref *propvar-negative-occ-clauses*
          ;                                              x)))
          ;                                (display ":")
          ;                                (display
          ;                                 (length
          ;                                  (vector-ref *propvar-positive-occ-clauses*
          ;                                              x))) )
          ;                      (cadddr x) )
          ;
          ;            (newline))
          ;           splittableclauses)

          (set! newsplitvar (prop-give-next-split splittableclauses))

          (if *prop-debug*  (begin
            (newline)(display "-- newsplitvar: ")(display newsplitvar)(newline) ))

          (if (null? newsplitvar)
              #t ; satisfiable, since could not split
              (or (solve-prop-clauses-branch
                       newsplitvar #t splittableclauses)
                  (solve-prop-clauses-branch
                       newsplitvar #f splittableclauses)))))))

; initial-prop-unit-propagate finds and propagates all the
; initially unit clauses in the input set of clauses
; if contradiction found, it returns #t

(define (initial-prop-unit-propagate clauses)
  (let ((tmp '())
        (tmp2 '())
        (res #f) )
    (do ((part clauses (cdr part)))
        ((or (null? part) res)
            (if res
                res ; contradiction found while pushing to the queue
                ; propagate all units
                (loop-prop-unit-queue) ))
      (set! tmp (car part))

      (if *prop-debug*  (begin (display tmp)(newline)))

      (if (eq? 1 (+ (car tmp) (cadr tmp))) ; sum of active lits is 1
          (begin
            (set! tmp2 (cddr tmp))
            (if (null? (car tmp2))
                ; negative empty, hence a positive unit
                (set! res (prop-enque-unit (caadr tmp2) #t))
                ; negative not empty, hence a negative unit
        	(set! res (prop-enque-unit (caar tmp2) #f)) ))))))



; solve-prop-clauses-branch returns
; #t if satisfiable,
; #f if unsatisfiable
; assignments etc are not cleared, if found to be satisfiable

(define (solve-prop-clauses-branch splitvar splitvalue splittableclauses)

 (if *prop-debug*  (begin
   (newline)(display "-- solve-prop-clauses-branch called: ")(newline)
   (display "  splitvar, splitvalue: ")
   (display splitvar)(display " ")(display splitvalue)
   (newline) ))

  (let ((res '())
        (newsplitvar '())
        (local-assignment-queue '())
        (local-assigned-varvalpairs '()) )
    (set! *prop-time-ran-out* #f)
    (set! *prop-var-assignment-queue* '())
    (set! *prop-assigned-varvalpairs* '())
    (set! *prop-unsat-found-occlistpart* '())
    (set! *prop-var-assignment-queue*
          (cons (cons splitvar splitvalue) '()) )

    ;;(set! test-list
    ;;  (map (lambda(cl)(cons cl (cons (car cl) (cadr cl))))
    ;;       *prop-internal-clauses*))

    (set! res (loop-prop-unit-queue)) ; if returns #t, found unsat
    (if res
        ; found to be unsat by simple propagation
        (begin
           ; undo assignment and propagation results
           (prop-undo-assignments *prop-assigned-varvalpairs*)
           (prop-undo-queue-assignments *prop-var-assignment-queue*)

           ;;(test-prop-undo-assignments *prop-assigned-varvalpairs* )
           ;;(prop-undo-queue-assignments *prop-var-assignment-queue*)
           ;;(for-each (lambda(back)
           ;;            (set-car! (car back) (cadr back))
           ;;            (set-car! (cdar back) (cddr back)) )
           ;;          test-list)

           #f) ; return unsatisfiable
        ; not found to be unsat by simple propagation
        (if (> (- (get-check-time 1) *slow-start-time*)
	        *one-pass-time*)
          ; time ran out
          (begin
            (set! *prop-time-ran-out* #t)
            #f)
          ; time left
          (begin
           ; store backtracing lists locally
           (set! local-assigned-varvalpairs *prop-assigned-varvalpairs*)
           (set! *prop-assigned-varvalpairs* '())
           (set! local-assignment-queue *prop-var-assignment-queue*)
           (set! *prop-var-assignment-queue* '())
           ; propagation did not find contradiction, now try to split
           (set! newsplitvar (prop-give-next-split splittableclauses))

           (if *prop-debug*  (begin
            (newline)(display "-- newsplitvar: ")(display newsplitvar)(newline) ))

           (if (null? newsplitvar)
               #t ; splitting not possible, now satisfiable.
               ; try to use splitting: if one branch gives #t, satisfiable
               (if (if
                    (> (length (vector-ref *propvar-negative-occ-clauses* newsplitvar))
                       (length (vector-ref *propvar-positive-occ-clauses* newsplitvar)) )

                   (or (solve-prop-clauses-branch
                         newsplitvar #t splittableclauses)
                       (solve-prop-clauses-branch
                         newsplitvar #f splittableclauses))

                   (or (solve-prop-clauses-branch
                         newsplitvar #f splittableclauses)
                       (solve-prop-clauses-branch
                         newsplitvar #t splittableclauses)) )

                    ; one branch satisfiable, hence branch satisfiable
                    #t
                    ; branch unsatisfiable, undo assignments
                    ; in simple propagation at the beginning of branch
                    (begin
                      (set! *prop-unsat-found-occlistpart* #f)
                      (set! *prop-assigned-varvalpairs* local-assigned-varvalpairs)
                      (prop-undo-assignments local-assigned-varvalpairs)
                      (prop-undo-queue-assignments local-assignment-queue)

                      ;;(test-prop-undo-assignments local-assigned-varvalpairs)
                      ;;(prop-undo-queue-assignments local-assignment-queue)
           	      ;;(for-each (lambda(back)
                      ;; 			(set-car! (car back) (cadr back))
                      ;; 			(set-car! (cdar back) (cddr back)) )
                      ;;         test-list)

                      #f))))))))


; prop-find-splittable selects all clauses with more than two positive
; literals active

(define (prop-find-splittable clauses)
  (let ((res '()))
    (do ((part clauses (cdr part)))
        ((null? part) res)
      (if (> (cadar part) 1) (set! res (cons (car part) res))) )))



; prop-give-next-split returns a var, if possible,
; if no splittable vars, it returns ()


(define (prop-give-next-split-tst splittableclauses)

 (if *prop-debug*
  (begin
   (newline)(display "-- prop-give-next-split-called, splittable: ")(newline)
   (for-each (lambda(cl)(display "  ")(display cl)(newline)) splittableclauses)
   (newline) ))

 (let ((bestclause '())
       (bestlen 0)
       (foundflag #f)
       (foundlit #f)
       (tmp '()) )
  (do ((part splittableclauses (cdr part)))
      ((or (null? part) foundflag)
           (if (null? bestclause)
               bestclause
               ; give the fist unassigned positive lit of the found clause
               (begin
                ; (newline)(display "splittable taken: ")(display bestclause)(newline)
                (do ((litpart (car (cdddr bestclause)) (cdr litpart)))
                    ((or (null? litpart) foundlit) foundlit)
                  (if (null? (vector-ref *propvar-value* (car litpart)))
                      (set! foundlit (car litpart)) )))))
     ; consider only fully positive nonunit, not subsumed clauses,
     ; find shortest
     (set! tmp 0)
     (if (and (eq? 0 (caar part)) ; fully positive
              (> (cadar part) 1) ; nonunit
              (null? (car (cddddr (car part)))) ; not subsumed
              (begin
                     (for-each
                      (lambda(x)
                       (set! tmp
                         (+ tmp
                            (+ (length (vector-ref *propvar-negative-occ-clauses* x))
                               (length (vector-ref *propvar-positive-occ-clauses* x)) ))))
                        (car (cdddar part)) )
                      #t)
              (> tmp bestlen) )
         ; might be a good candidate, check subsumption further
         (if (not (prop-clause-subsumed-by-var? (car part)))
             (begin
                (set! bestclause (car part))
                (set! bestlen tmp) ))))))




(define (prop-give-next-split splittableclauses)

 (if *prop-debug*
  (begin
   (newline)(display "-- prop-give-next-split-called, splittable: ")(newline)
   (for-each (lambda(cl)(display "  ")(display cl)(newline)) splittableclauses)
   (newline) ))

 (let ((bestclause '())
       (bestlen 100000)
       (foundflag #f)
       (foundlit #f)
       (tmp '()) )
  (do ((part splittableclauses (cdr part)))
      ((or (null? part) foundflag)
           (if (null? bestclause)
               bestclause
               ; give the fist unassigned positive lit of the found clause
               (begin
                ; (newline)(display "splittable taken: ")(display bestclause)(newline)
                (do ((litpart (car (cdddr bestclause)) (cdr litpart)))
                    ((or (null? litpart) foundlit) foundlit)
                  (if (null? (vector-ref *propvar-value* (car litpart)))
                      (set! foundlit (car litpart)) )))))
     ; consider only fully positive nonunit, not subsumed clauses,
     ; find shortest
     (if (and (eq? 0 (caar part)) ; fully positive
              (begin (set! tmp (cadar part)) (> tmp 1)) ; nonunit
              (null? (car (cddddr (car part)))) ; not subsumed
              (or (eq? 2 tmp) (< tmp bestlen)) )
         ; might be a good candidate, check subsumption further
         (if (not (prop-clause-subsumed-by-var? (car part)))
             (begin
                (set! bestclause (car part))
                (set! bestlen tmp)
                (if (eq? 2 tmp) (set! foundflag #t)) ))))))



(define (prop-clause-subsumed-by-var? clause)
  ;(if (prop-subsumed-clause? clause *prop-internal-clauses*)
  ;    (begin (display "splittable subsumed: ")(display clause)(newline)
  ;           #t)
  (let ((subsflag #f))
    ; check negative literals
    (do ((part (car (cddr clause)) (cdr part)))
        ((or (null? part) subsflag) subsflag)
      (if (eq? (vector-ref *propvar-value* (car part)) #f)
          (set! subsflag #t) ))
    ; check positive literals
    (if subsflag
        subsflag
        (do ((part (car (cdddr clause)) (cdr part)))
            ((or (null? part) subsflag) subsflag)
          (if (eq? (vector-ref *propvar-value* (car part)) #t)
              (set! subsflag #t) )))))


; loop-prop-unit-queue loops over the queue, assigns values,
; propagates, etc until either the queue is empty or contradiction
; is found
; iff contradiction is found, it returns #t, else #f


(define (loop-prop-unit-queue)

  (if *prop-debug*  (begin
    (newline)(display "-- loop-prop-unit-queue called")(newline)
    (show-prop-internal-data '() 4)  ))

  (let ((unsat-flag #f)
        (tmp '()) )
    (do ()
        ((or (null? *prop-var-assignment-queue*) unsat-flag) unsat-flag)
      (set! tmp (car *prop-var-assignment-queue*))
      (set! *prop-var-assignment-queue* (cdr *prop-var-assignment-queue*))
      (set! unsat-flag (assign-prop-var tmp)) )))


; assign-prop var returns #t iff unsatisfiable else #f
; it calls prop-resolve-clause when propagating the value

(define (assign-prop-var varvalpair)

  (if *prop-debug*  (begin
    (newline)(display "++ assign-prop-var called: ")
    (display varvalpair)(newline) ))

  (let* ((varnr (car varvalpair))
         (varval (cdr varvalpair))
         (tmpres #f)
         (unsat-flag #f) )
    ; set var value
    (vector-set! *propvar-value* varnr varval)
    ; keep the setting for backtracing
    (set! *prop-assigned-varvalpairs*
          (cons varvalpair *prop-assigned-varvalpairs*) )
    ; loop over correctly signed occurrences of var in all clauses
    (if varval
        ; if var set to #t, loop over all negative occurrences
        (do ((occlist (vector-ref *propvar-negative-occ-clauses* varnr)
                      (cdr occlist) ))
            ;((or (null? occlist) unsat-flag) unsat-flag)
            ((null? occlist) unsat-flag)

          (if *prop-debug*  (begin (display (car occlist))(newline)))

          (set! tmpres (prop-resolve-clause occlist varval))
          (if tmpres (set! unsat-flag #t)) )
        ; if var set to #f, loop over all positive occurrences
        (do ((occlist (vector-ref *propvar-positive-occ-clauses* varnr)
                      (cdr occlist) ))
            ;((or (null? occlist) unsat-flag) unsat-flag)
            ((null? occlist) unsat-flag)

          (if *prop-debug*  (begin (display (car occlist))(newline)))

          (set! tmpres (prop-resolve-clause occlist varval))
          (if tmpres (set! unsat-flag #t)) ))))

; prop-resolve-clause decreases the clause active propvar count and if necessary,
; puts a new element to the assignment queue
; or finds a contradiction:
; iff contradiction found, #t is returned

(define (prop-resolve-clause propclauselst varval)

  (if *prop-debug*  (begin
    (display " ++ prop-resolve-clause called: ")(newline)
    (display propclauselst)(newline)
    (display varval)(newline) ))

  (let* ((propclause (car propclauselst))
         (clcdr (cdr propclause))
         (clsum (+ (car propclause) (car clcdr)))
         (unsat-flag #f) )
    (if varval
        (set-car! propclause (- (car propclause) 1))
        (set-car! clcdr (- (car clcdr) 1)) )
    (if (> clsum 2)
        ; nothing interesting
        #f
        ; smth special happens
        (if (<= clsum 1)
            ; contradiction found
            #t
            ;;(begin
            ;;  (set! *prop-unsat-found-occlistpart* propclauselst)
            ;;  #t)
            ; unit generated: locate which unit, enque
            (begin
               (set! varval #f) ; used as a flag for termination
               ; check negative literals
               (do ((propclause (cadr clcdr) (cdr propclause)))
                   ((or (null? propclause) varval))

                 (if *prop-debug*  (begin
                   (display  (car propclause))
                   (display " ")
                   (display (vector-ref *propvar-value* (car propclause)))
                   (newline) ))

                 (if (null? (vector-ref *propvar-value* (car propclause)))
                     (begin
                       (set! varval #t)
                       (set! unsat-flag
                            (prop-enque-unit (car propclause) #f) ))))
                       ;;(if unsat-flag
                       ;;    (set! *prop-unsat-found-occlistpart*
                       ;;          propclauselst))
               ; check positive literals
               (if unsat-flag
                   unsat-flag
                  (if varval
                      #f
                      (do ((propclause (caddr clcdr) (cdr propclause)))
                          ((or (null? propclause) varval) unsat-flag)

                        (if *prop-debug*  (begin
                          (display  (car propclause))
                          (display " ")
                          (display (vector-ref *propvar-value* (car propclause)))
                          (newline) ))

                        (if (null? (vector-ref *propvar-value* (car propclause)))
                            (begin
                              (set! varval #t)
                              (set! unsat-flag
                                   (prop-enque-unit
                                     (car propclause) #t)) ))))))))))

                              ;;(if unsat-flag
                              ;;   (set! *prop-unsat-found-occlistpart*
                              ;;         propclauselst))


(define (prop-enque-unit varnr varval)
  (let* ((queue-val (vector-ref *propvar-enqued-value* varnr)))
    ;; (set! queue-val '())
    (if (null? queue-val)
        ; enque normally
        (begin
          (set! *prop-var-assignment-queue*
                (cons (cons varnr varval) *prop-var-assignment-queue*) )
          (vector-set! *propvar-enqued-value* varnr varval)
          #f)
        ; either already enqued with same val or contradiction found
        (if (eq? queue-val varval)
            #f   ; do nothing
            #t))))  ; contradiction found


(define (prop-undo-assignments assignmentlist)
  (let ((tmp '())
        (stopflag #f)
        (varnr '()) )
   (do ((part assignmentlist (cdr part)))
       ((null? part))
     (set! tmp (car part))
     (set! varnr (car tmp))
     (vector-set! *propvar-value* varnr '())
     (vector-set! *propvar-enqued-value* varnr '())
     (if (cdr tmp)
        ; if var set to #t, loop over all negative occurrences
        (do ((occlist (vector-ref *propvar-negative-occ-clauses* varnr)
                      (cdr occlist) ))
            ((or (null? occlist) stopflag))
          (set-car! (car occlist) (+ (caar occlist) 1))
          ; the same place where we stopped cause of contradiction?
          (if #f ; (eq? *prop-unsat-found-occlistpart* occlist)
              (begin
                (set! stopflag #t) )))
        ; if var set to #f, loop over all positive occurrences
        (do ((occlist (vector-ref *propvar-positive-occ-clauses* varnr)
                      (cdr occlist) ))
            ((or (null? occlist) stopflag))
          (set-car! (cdar occlist) (+ (cadar occlist) 1))
          ; the same place where we stopped cause of contradiction?
          (if #f ; (eq? *prop-unsat-found-occlistpart* occlist)
              (begin
                (set! stopflag #t) )))))))


(define (prop-undo-queue-assignments queue)
   (do ((part queue (cdr part)))
       ((null? part))
     (vector-set! *propvar-enqued-value* (caar part) '()) ))


(define (test-prop-undo-assignments assignmentlist)
  (let ((tmp '())
        (stopflag #f)
        (varnr '()) )
   ; clear set vars
   (do ((part assignmentlist (cdr part)))
       ((null? part))
     (set! tmp (car part))
     (set! varnr (car tmp))
     (vector-set! *propvar-value* varnr '())
     (vector-set! *propvar-enqued-value* varnr '()) )))


(define (prop-try-subsume clauses)
 (newline)(display "prop-subsumption started: ")(newline)
 (let ((nr 0))
  (do ((part clauses (cdr part)))
      ((null? part)
         (newline)(display "prop-subsumption ended, subsumed: ")(display nr)(newline) )
      (if (prop-subsumed-clause? (car part) clauses)
          (begin (set! nr (+ 1 nr))
                 (display ".") )))))

(define (prop-subsumed-clause? cl clauses)
  (let ((foundflag #f))
    (if (prop-unit-subsumed? cl)
        (set! foundflag #t) )
    (do ((part clauses (cdr part)))
        ((null? part)
             foundflag)
        (if (and (prop-subsumes? (car part) cl)
                 (not (eq? (car part) cl)) )
            (set! foundflag #t) ))))

(define (prop-unit-subsumed? cl)
  (or (some? (lambda(x)(eq? #f (vector-ref *propvar-value* x)))
             (car (cddr cl)) )
      (some? (lambda(x)(eq? #t (vector-ref *propvar-value* x)))
             (car (cdddr cl)) )))

(define (prop-subsumes? gencl subscl)
  (let ()
    (and (>= (car gencl) (car subscl))
         (>= (cadr gencl) (cadr subscl))
         (prop-subclause-polarity? (cadddr gencl) (cadddr subscl) #t)
         (prop-subclause-polarity? (caddr gencl) (caddr subscl) #f)  )))

(define (prop-subclause-polarity? glist slist polarity)
 (let ((foundflag #t)
       (lit '())
       (val '())
       (npol (not polarity)) )
  (do ((gpart glist (cdr gpart)))
      ((or (null? gpart) (not foundflag))
         foundflag)
     (set! foundflag #f)
     (set! lit (car gpart))
     (set! val (vector-ref *propvar-value* lit))
     (if (eq? val npol)
         (set! foundflag #t)
         (do ((spart slist (cdr spart)))
             ((null? spart))
          (if (eq? lit (car spart))
            (set! foundflag #t) ))))))


(define (prop-sort-splittable clauselst)
  (let ((tmp (map (lambda(x)(cons (prop-clause-occ-sum x) x)) clauselst)))
   (map (lambda(x)(cdr x))
        (prop-sort-splittable-aux tmp) )))

(define (prop-clause-occ-sum cl)
  (let ((sum 0))
     (for-each
       (lambda(x)
        (set! sum
              (+ sum
                  (+ (length (vector-ref *propvar-negative-occ-clauses* x))
                     (length (vector-ref *propvar-positive-occ-clauses* x)) ))))
       (car (cdddr cl)) )
      sum))

(define (prop-sort-splittable-aux clauselst)
 (let ((part1 '())
       (part2 '())
       (tmp '())
       (len (length clauselst)) )
  (cond ((null? clauselst) clauselst)
        ((null? (cdr clauselst)) clauselst)
        (else
           (set! tmp (prop-sort-splittable-split clauselst (quotient len 2)))
           (prop-sort-splittable-merge
              (prop-sort-splittable-aux (car tmp))
              (prop-sort-splittable-aux (cdr tmp)) )))))


(define (prop-sort-splittable-merge lst1 lst2)
 (let ((m1 0)
       (m2 0) )
  (cond ((null? lst1) lst2)
        ((null? lst2) lst1)
        ((>= (caar lst1) (caar lst2))
            (cons (car lst1) (prop-sort-splittable-merge (cdr lst1) lst2)) )
        (else
            (cons (car lst2) (prop-sort-splittable-merge lst1 (cdr lst2))) ))))

(define (prop-sort-splittable-split clause halfnr)
  (let ((p1 '()) )
   (do ((part clause (cdr part))
        (nr 0 (+ 1 nr)) )
       ((or (null? part) (>= nr halfnr))
          (cons p1 part) )
       (set! p1 (cons (car part) p1)) )))




;-----------------------------------------------------------------
;
; external prop
;
;------------------------------------------------------------------

(define *global-prop-prover* 'zchaff)
(define *prop-conversion-to-file* #f)
; (define *global-prop-prover* 'chaffscript)

(define *prop-conversion-file-port* #f)
(define *prop-conversion-temp-file-port* #f)
(define *prop-conversion-file-name-start* "/tmp/ganprglin")
(define *prop-conversion-temp-file-name-start* "/tmp/ganprgltmp")
(define *prop-conversion-header-file-name-start* "/tmp/ganprhdrtmp")
(define *prop-conversion-result-file-name-start* "/tmp/ganprrslt")
(define *prop-conversion-file-name* "")
(define *prop-conversion-temp-file-name* "")
(define *prop-conversion-header-file-name* "")
(define *prop-conversion-result-file-name* "")
(define *prop-conversion-output-clause-nr* 0)



(define (external-solve-prop-clauses propclauses prop-atom-maxnr one-pass-time)
  (let (;(time (number->string (quotient (quotient one-pass-time 4)
        ; 					*internal-time-units-per-second*)))
	;(time (quotient (quotient one-pass-time 4)
	;				*internal-time-units-per-second*))

        (time (quotient (quotient one-pass-time 2)
		        *internal-time-units-per-second*))
        (callstr "")
        (infile '())
        (mainport '())
        (tempport '())
        (headerport '())
        (res '()) )
    ;(newline)(display "external-solve-prop-clauses called with:")(newline)
    ;(display propclauses)(newline)
    ;(display prop-atom-maxnr)(newline)
    ;(display one-pass-time)(newline)
    (set! infile *prop-conversion-file-name*)
    (if (and *prop-conversion-to-file* *sat-strategy-flag*)
        (begin
          (if *prop-conversion-temp-file-port*
              (close-output-port *prop-conversion-temp-file-port*) )
          (set! headerport
		(open-output-file *prop-conversion-header-file-name*))
          (display (string-append "p cnf " (number->string prop-atom-maxnr) " "
			(number->string *prop-conversion-output-clause-nr*)
			"\n") headerport)
          (close-output-port headerport)
          (set! callstr (string-append "cat "
				       *prop-conversion-header-file-name* " "
				       *prop-conversion-temp-file-name* " > "
				       *prop-conversion-file-name*))

          (display "external: ")
          (display callstr)
          (newline)
          (system callstr))
        (make-prop-infile infile propclauses prop-atom-maxnr one-pass-time) )

    (set! res (call-prop-prover *global-prop-prover* infile time))
    ;(newline)
    ;(display "external prop prover returns: ")(display res)
    ;(newline)
    (cond
      ((eq? res #f) (set! *prop-time-ran-out* #f) #f)
      ((eq? res #t) (set! *prop-time-ran-out* #f) #t)
      ((eq? res '()) (set! *prop-time-ran-out* #t) #t)
      (else (set! *prop-time-ran-out* #t) #f) )))


; Call an external propositional prover returning the result of the proof
; search as
;   0   Contradiction found
;   1   No contradiction found, propositional prover ran out of time
;   2   No contradiction found
;
; call-prop-prover :: Symbol -> String -> Int
;
; Note: The system command return value might not be portable.

(define (call-prop-prover prover file time)
  (let ((callstring "")
        (tmp '())
        (tmptime 0)
	; (timelimit (number->string time))
	(timelimit "3600") )

    ;(newline)(display "	call-prop-prover time: ")
    ;(display time)(newline)

    ;(display "get-check-time: ")(display (get-check-time 1))(newline)
    ;(display "*slow-start-time*: ")(display *slow-start-time*)(newline)
    ;(display "*one-pass-time*: ")(display *one-pass-time*)(newline)

    ; (> (- (get-check-time 1) *slow-start-time*)
    ;	     *one-pass-time*))

    (set! tmptime
         (- *one-pass-time*
            (- (get-check-time 1) *slow-start-time*) ))

    (if (< tmptime (* 2  *internal-time-units-per-second*))
        (set! tmptime 0) )

    ;(display "tmptime: ")(display tmptime)

    (set! timelimit (quotient tmptime *internal-time-units-per-second*))
    (set! timelimit (number->string timelimit))

    (set! callstring
       (string-append
         (string-append
            (if *intepreter-mode*
                (symbol->string prover)
                (string-append
                     (path-string-startpart (car *gandalf-program-arguments*))
                     (symbol->string prover) ))
            (string-append " "
              (string-append file
                  (string-append " -t " timelimit) )))
         (string-append " > " *prop-conversion-result-file-name*) ))
    (display "external: ")
    (display  callstring)
    (newline)
    (set! tmp (system callstring))
    ;(newline)
    ;(display "system call returns: ")
    ;(display tmp)
    ;(newline)

    (cond
       ((check-prop-prover-outfile
            *prop-conversion-result-file-name*
            "Instance satisfiable")
          #t)
       ((check-prop-prover-outfile
            *prop-conversion-result-file-name*
            "Instance unsatisfiable")
          #f)
       (else
          '() ))))


(define (check-prop-prover-outfile filename checkstring)
  (let ((port (open-input-file filename))
        (chr '())
        (lst '())
        (searchstr "")
        (searchlst '())
        (line "") )
    ;(set! searchstr "Instance satisfiable")
    (set! searchstr checkstring)
    (set! searchlst (string->list searchstr))
    (if (or (not port)(null? port))
        (begin
          (newline)
          (display "ERROR: cannot read prop prover outputfile")
          (newline)
          (quit) )
        (begin
          (do ((chr (read-char port)  (read-char port)))
              ((eof-object? chr)
                 (close-input-port port)
                 (set! lst (reverse lst))

                 ;(newline)
                 ;(display searchlst)
                 ;(newline)
                 ;(display lst)

                 (if (sublist? searchlst lst)
                     (begin
                        ;(newline)(display "SAT!!!")(newline)
                        #t)
                     (begin
                        ;(newline)(display "NOT SAT!!!")(newline)
                        #f) ))
            (set! lst (cons chr lst)) )))))


(define (sublist? small big)
  (cond
    ((null? small) #t)
    ((null? big) #f)
    (else
      (or (sublist-aux? small big)
          (sublist? small (cdr big)) ))))


(define (sublist-aux? small big)
  (cond
    ((null? small) #t)
    ((null? big) #f)
    (else
       (and (eq? (car small) (car big))
            (sublist-aux? (cdr small) (cdr big)) ))))

(define (make-prop-infile filename propclauses prop-atom-maxnr one-pass-time)
  (let ((port '()))
    (set! port (open-output-file filename))
    ;(display "c " port)
    ;(display (quotient (quotient one-pass-time 2)
    ;  		       *internal-time-units-per-second*)
    ;	     port)
    ;(newline port)
    (display "p cnf " port)
    (display prop-atom-maxnr port)
    (display " " port)
    (display (length propclauses) port)
    (newline port)
    (if #f
	(begin
	  (display "propclauses: ")(newline)
	  (for-each
	   (lambda(cl)
	     (for-each
	      (lambda(lit)
		(display (- 0 lit))
		(display " ") )
	      (car cl) )
	     (for-each
	      (lambda(lit)
		(display lit)
		(display " ") )
	      (cdr cl) )
	     (newline) )
	   propclauses)))
    (for-each
     (lambda(cl)
       ;; (display "prop: " port)(display cl port)(newline port)
       (output-prop-clause cl port) )
     propclauses)
    (close-output-port port) ))

(define (output-prop-clause cl port)
   ; possible display to screen
   (if #f
       (begin
          (for-each
            (lambda(lit)
             (display (- 0 lit))
             (display " ") )
            (car cl) )
          (for-each
            (lambda(lit)
             (display lit)
             (display " ") )
           (cdr cl) )
          (display 0)
          (newline) ))
   ; actual output to port
   (for-each
       (lambda(lit)
          (display (- 0 lit) port)
          (display " " port) )
       (car cl) )
   (for-each
       (lambda(lit)
          (display lit port)
          (display " " port) )
      (cdr cl) )
   (set! *prop-conversion-output-clause-nr*
         (+ 1 *prop-conversion-output-clause-nr*) )
   (display 0 port)
   (newline port) )

(define (open-prop-temp-output-file filename)
  (let ((port '()))
   (if *prop-conversion-temp-file-port*
       (close-output-port *prop-conversion-temp-file-port*))
   (set! port (open-output-file filename))
   (set! *prop-conversion-output-clause-nr* 0)
   (set! *prop-conversion-temp-file-port* port) ))



;----------------------------------------------------------------------
;
;         test stuff
;
;----------------------------------------------------------------------

(define *prop-in*
;'(
;((1 3 4) 2)
;(() 1)
;((1) 3 2 5)
;((1) 4 2 5)
;((2)) ))

;'(
;((1 2))
;((1) 2)
;(() 1 2)
;((2) 1) ))

'(
((1 2 3))
((1  2) 3)
((1 3) 2)
((2 3) 1)
((1) 2 3)
((2) 1 3)
((3) 1 2)
(() 1 2 3) ))


(define *prop-internal-clauses* '())
(define *prop-maxvar* 0)

(define (tstprop)
  (let ((maxvar 5)
        (internalclauses '())
        (res '()) )

    (create-prop-var-vectors maxvar)

    ;(initialize-prop-var-values maxvar)
    ;(newline)(display "-- prop values initialised")(newline)
    ;(show-prop-internal-data '() maxvar)
    ;(set! internalclauses (build-prop-internal-structure *prop-in*))
    ;(newline)(display "-- prop internal structure built")(newline)
    ;(show-prop-internal-data internalclauses maxvar)

    (set! res (solve-prop-clauses *prop-in* maxvar))

    (newline)(display "-- res is: ")(display res)(newline)
    (show-prop-internal-data internalclauses maxvar) ))


(define (show-prop-internal-data clauses maxvar)
 (let ()
  (set! maxvar *prop-atom-maxnr*)
  (newline)
  (display "showing vars 1 to ")(display maxvar)
  (display " : val, queue, pos, neg")(newline)
  (do ((var 0 (+ 1 var)))
      ((> var maxvar))
   (display var)(display ": " )
   (display (vector-ref *propvar-value* var))(display ", ")
   (display (vector-ref *propvar-enqued-value* var))(display ", ")
   (newline)(display "  pos: ")
   (display (vector-ref *propvar-positive-occ-clauses* var))
   (newline)(display "  neg: ")
   (display (vector-ref *propvar-negative-occ-clauses* var))
   (newline) )
  (newline)
  (display "showing clauses: ")(newline)
  (for-each (lambda(cl)(display cl)(newline)) clauses)
  (newline)
  (display "*prop-var-assignment-queue* : ")(newline)
  (display "  ")(display *prop-var-assignment-queue*)(newline)
  (display "*prop-assigned-varvalpairs* : ")(newline)
  (display "  ")(display *prop-assigned-varvalpairs*)(newline)
  (display "*prop-unsat-found-occlistpart* : ")(newline)
  (display "  ")(display *prop-unsat-found-occlistpart*)(newline) ))


(define (tst x)
 (set! *prop-orig-constants* '(a b))
 (set! *prop-orig-funs* '((f . 2)(g . 1)) )
 (set! *prop-orig-preds* '((p . 1)(r . 0)) )
 ; (make-prop-interpretation-lists x #t)
 (pretty-print *prop-interp-value-pairs*)(newline)
 (pretty-print *prop-constants-mapping*)(newline)
 (pretty-print *prop-fun-mappings*)(newline) )
 ; (check-prop-interpretations x) )

;----------------------------------------------------------------------
;
;       random generation functions
;
;----------------------------------------------------------------------


(define random-tap 24)
(define random-size 55)

(define (random-size-int l)
  (let ((trial (string->number (make-string l #\f) 16)))
  (if (and (exact? trial) (>= most-positive-fixnum trial))
      l
      (random-size-int (- l 1)))))
(define random-chunk-size (* 4 (random-size-int 8)))

(define random-MASK
  (string->number (make-string (quotient random-chunk-size 4) #\f) 16))

(define *random-state*
  '#(
 "d909ef3e" "fd330ab3" "e33f7843" "76783fbd" "f3675fb3"
 "b54ef879" "0be45590" "a6794679" "0bcd56d3" "fabcdef8"
 "9cbd3efd" "3fd3efcd" "e064ef27" "dddecc08" "34444292"
 "85444454" "4c519210" "c0366273" "54734567" "70abcddc"
 "1bbdac53" "616c5a86" "a982efa9" "105996a0" "5f0cccba"
 "1ea055e1" "fe2acd8d" "1891c1d4" "e6690270" "6912bccc"
 "2678e141" "61222224" "907abcbb" "4ad6829b" "9cdd1404"
 "57798841" "5b892496" "871c9cd1" "d1e67bda" "8b0a3233"
 "578ef23f" "28274ef6" "823ef5ef" "845678c5" "e67890a5"
 "5890abcb" "851fa9ab" "13efa13a" "b12278d6" "daf805ab"
 "a0befc36" "0068a7b5" "e024fd90" "a7b690e2" "27f3571a"
 0))

(let ((random-strings *random-state*))
  (set! *random-state* (make-vector (+ random-size 1) 0))
  (let ((nibbles (quotient random-chunk-size 4)))
    (do ((i 0 (+ i 1)))
        ((= i random-size))
      (vector-set!
       *random-state* i
       (string->number (substring (vector-ref random-strings i)
                                  0 nibbles)
                       16)))))

;;; random-chunk returns an integer in the range of
;;; 0 to (- (expt 2 random-chunk-size) 1)
(define (random-chunk v)
  (let* ((p (vector-ref v random-size))
         (ans (logxor
               (vector-ref v (modulo (- p random-tap) random-size))
               (vector-ref v p))))
    (vector-set! v p ans)
    (vector-set! v random-size (modulo (- p 1) random-size))
    ans))

(define (random-random modu . args)
  (let ((state (if (null? args) *random-state* (car args))))
    (if #t
        (do ((ilen 0 (+ 1 ilen))
             (s random-MASK
                (+ random-MASK (* (+ 1 random-MASK) s))))
            ((>= s (- modu 1))
             (let ((slop (modulo (+ s (- 1 modu)) modu)))
               (let loop ((n ilen)
                          (r (random-chunk state)))
                 (cond ((not (zero? n))
                        (loop (+ -1 n)
                              (+ (* r (+ 1 random-MASK))
                                 (random-chunk state))))
                       ((>= r slop) (modulo r modu))
                       (else (loop ilen (random-chunk state))))))))

        1)))

(define (random-make-random-state . args)
  (let ((state (if (null? args) *random-state* (car args))))
    (list->vector (vector->list state))))


;----------------------------------------------------------
;
;                uncurrying
;
;----------------------------------------------------------


(define (prop-uncurry-clauses clauses)
  (let ((res '()))
    (set! res (map (lambda(x)
                     (cons (car x)
                           (prop-uncurry-clause (cdr x)) ))
                   clauses))

    ;(newline)
    ;(display "uncurried clauses: ")(newline)
    ;(for-each (lambda(x)(display x)(newline)) res)
    ;(newline)

    res))

(define (prop-uncurry-clause clause)
  (let ((res '()))
    (set! res (map (lambda(x)(prop-uncurry-term x)) clause))
    res))

(define (prop-uncurry-term term)
  (let ((res '())
        (fun '())
        (args '()) )
   (cond
     ((pair? term)
        (set! fun (car term))
        (set! args (cdr term))
        (cond
          ((and (pair? args)
                (curry-const? (car args)) )
             (set! fun
               (logior (- (car args) *mincurryconst*)
                       (pure-fun fun) ))
             (set! args (cdr args)) )
          (else
             (set! fun (pure-fun fun)) ))
        (set! args (map (lambda(x)(prop-uncurry-term x)) args))
        (cons fun args) )
     (else
        term) )))


(define (prop-spec-pure-fun x)
  x)

(define (prop-spec-pure-unsigned-fun x)
  (if (neg-fun? x)
      (logxor x *sgn-mask*)
      x))



;=======================================================================
;
;       for sat only: converting clauses with ground terms
;
;======================================================================

(define *prop-groundprocess-const-nr* 100000)
(define *prop-groundprocess-term-const-pairs* '())

(define (prop-groundprocess-clauses readylst)

    ;(newline)
    ;(pretty-print readylst)
    ;(newline)
    ;(display "readylst: ")(newline)
    ;(pretty-print readylst)

  (let ((res '())
        (tmp '())
        (sosaddedflag #f) )
    (set! *prop-groundprocess-const-nr* 100000)
    (set! *prop-groundprocess-term-const-pairs* '())

    (set! res (prop-groundprocess-clause-list readylst))

    (if (null? *prop-groundprocess-term-const-pairs*)
        (begin
          (set! res readylst) )
        (begin
          (set! res (append (prop-groundprocess-make-equalities) res))

          (newline)
          (display "Some ground terms were converted.")
          ;(for-each (lambda(x)(print-clause x)(newline))
          ;          (prop-groundprocess-make-equalities) )
          (newline) ))

    res))

(define (prop-groundprocess-make-equalities)
   (let ((tmp '())
         (varsymb '())
         (eqfun '()) )
     (set! eqfun *eq-pred-nr*)
     (set! tmp
           (map (lambda(x)
                  (list '() (list eqfun (car x) (cdr x))) )
                 *prop-groundprocess-term-const-pairs*))
     (set! varsymb 1)
     (set! tmp
          (cons (list '() (list eqfun varsymb varsymb)) tmp))
     tmp))

(define (prop-groundprocess-clause-list clauselst)
  (map (lambda(x)
         (cons (car x) (prop-groundprocess-clause (cdr x))) )
       clauselst))

(define (prop-groundprocess-clause clause)
  (map (lambda(x)(prop-groundprocess-lit x))
       clause))


(define (prop-groundprocess-lit lit)
  (let ((tmp '()))
    (cond
      ((not (pair? lit))
         lit)
      (else
         (set! tmp (map (lambda(x)(prop-groundprocess-term x))
                        (cdr lit) ))
         (cons (car lit) tmp) ))))


(define (prop-groundprocess-term term)
  (let ((tmp '())
        (newc '()) )
    (cond
      ((not (pair? term))
         term)
      (else
         (set! tmp (map (lambda(x)(prop-groundprocess-term x))
                        (cdr term) ))
         (if (every? (lambda(y)(const? y))
                     tmp)
             (begin

               ;(newline)
               ;(print-term term)
               ;(newline)

               (set! newc (assoc (cons (car term) tmp)
                                 *prop-groundprocess-term-const-pairs*))
               (if newc
                   (cdr newc)
                   (begin
                     (set! newc (prop-groundprocess-term-newconstant))
                     (set! *prop-groundprocess-term-const-pairs*
                           (cons (cons (cons (car term) tmp) newc)
                                 *prop-groundprocess-term-const-pairs*))
                     newc)))

             (cons (car term) tmp) )))))

(define (prop-groundprocess-term-newconstant)
  (set! *prop-groundprocess-const-nr* (+ 1 *prop-groundprocess-const-nr*))
  *prop-groundprocess-const-nr*)




;======================= EOF resprop.scm ====================================
