/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/**
 * File Literal.cpp.
 * Implements class Literal.
 */

#include "Literal.hpp"
#include "Substitution.hpp"


// build literal from a kernel literal
// 28/09/2002 Manchester
Literal::Literal (const VampireKernel::Literal* literal, 
		  const VampireKernel& kernel)
{
  TermList ts (literal->firstArgument (), kernel);
  ulong sym = literal->predicate();
  Atom a (sig->createPred (kernel.symbolPrintName (sym),
			   kernel.symbolArity(sym)),
	  ts);

  _data = new Data (literal->isPositive(), a);
} // Literal::Literal (const VampireKernel::Literal* lit, ...)


// build literal list from a kernel literal list
// 28/09/2002 Manchester
LiteralList::LiteralList (const VampireKernel::Literal* literal, 
		  const VampireKernel& kernel)
{
  if (! literal) {
    _data = 0;
    return;
  }

  _data = new LstData<Literal> ( Literal (literal, kernel),
		     LiteralList (literal->next(), kernel) );
} // Literal::Literal (const VampireKernel::Literal* lit, ...)


// assignment operator
// 25/08/2002 Torrevieja
void Literal::operator = (const Literal& t)
{
  if (t._data) {
    t._data->ref ();
  }

  if (_data) {
    _data->deref ();
  }

  _data = t._data;
} // Literal::operator =


/**
 * Comparison of two literals, needed for normalization. All negative
 * literals are smaller than positive ones. All other literals are
 * compared by comparing their atoms.
 *
 * @since 26/06/2002 Manchester
 * @since 17/07/2003 Manchester, slightly changed
 * @return result of comparison of the literal with l
 */
Compare Literal::compare (const Literal& l) const  
{
  // negative literals are less than positive
  if ( negative() ) {
    if ( l.positive() )
      return LESS;
  }
  else { // positive
    if ( l.negative() )
      return GREATER;
  }

  // polarities are equal
  return atom().compare(l.atom());
} // Literal::compare


/*
 * Normalization.
 *
 * @since 10/09/2002, Manchester
 * @since 17/07/2003 Manchester changed to return the result in the argument
 * @param result the literal to write the result in
 */
void Literal::normalize (Literal& result) const
{ 
  Atom a;
  atom().normalize(a);
  result = a == atom() ?
    *this :
    Literal (sign(),a);
} // Literal::normalize ()


// the list is a renaming of lst (up to a permutation)
// 03/10/2002 Manchester
bool LiteralList::isRenamingOf (LiteralList lst) const
{
  // easy check first
  if (length() != lst.length()) {
    return false;
  }

  Substitution sbst;
  LiteralList empty;

  return isRenamingOf (empty, lst, sbst);
} // LiteralList::isRenamingOf


// the list is a renaming of (tried U remaining) up to a permutation
// but the first literal should match a literal in remaining
// sbst is a previously computed substitution
// 03/10/2002 Manchester
bool LiteralList::isRenamingOf (LiteralList tried, 
				LiteralList remaining, 
				Substitution sbst) const
{
  if (isEmpty()) {
    return tried.isEmpty() &&
           remaining.isEmpty();
  }

  while (remaining.isNonEmpty()) {
    // try to match the first literal
    // copy sbst, or else it may be spoiled by the first match
    Substitution s (sbst);
    if ( head().isRenamingOf(remaining.head(), s) ) {
      LiteralList all (tried);
      all.append (remaining.tail());
      LiteralList empty;
      if ( tail().isRenamingOf(empty, all, s) ) {
	return true;
      }
    }
    // the match of the first literal did not help, try the rest
    tried.push (remaining.head());
    remaining.pop();
  }

  // remaining is empty
  return false;
} // LiteralList::isRenamingOf


// the literal is a renaming of l
// sbst is a previously computed substitution
// 03/10/2002 Manchester
bool Literal::isRenamingOf (Literal l, Substitution& sbst) const
{
  return sign() == l.sign() &&
         atom().isRenamingOf(l.atom(), sbst);
} // Literal::isRenamingOf


/*
// replace the literal
//   f(X,y,Z) != f(X,u,Z)
// by 
//   i(f(X,y,Z),X,Z) = y
void Literal::introduceInverse ( int argNo, FSymbol* inv )
{
  // make literal into positive
  _sign = true;

  atom()->introduceInverse (argNo, inv);
} // Literal::introduceInverse


void Literal::splitLargeArity (const PSymbol* p, FSymbol::List* newFs )
{
  atom()->splitLargeArity (p, newFs);
} // Literal::splitLargeArity


void Literal::splitLargeArity (const FSymbol* f, FSymbol::List* newFs )
{
  atom()->splitLargeArity (f, newFs);
} // Literal::splitLargeArity


Literal* Literal::negativeEqualitySplit ( int thresholdWeight )
{
  if ( positive() ) // only negative equalities can be split
    return 0;

  Atom* a = atom()->negativeEqualitySplit (thresholdWeight);

  return a ? new Literal ( true, a ) : 0 ; 
} // Literal::negativeEqualitySplit
*/
