/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/**
 * File Literal.hpp.
 * Defines class Literal.
 */


#ifndef __Literal__
#define __Literal__


#include "Atom.hpp"


// ******************* class Literal, definition *********************


class Literal {
 private:
  class Data;

 public:
  // constructor
  Literal ();
  Literal (bool sign, const Atom& atom);
  Literal (const Literal& t);
  Literal (const VampireKernel::Literal* lit, const VampireKernel& kernel);
  ~Literal ();
  void operator= (const Literal& rhs);

  // query/change the structure
  bool positive () const;
  bool negative () const;
  bool sign () const;
  const Atom& atom () const;
  bool exists () const;

  /**
   * Comparison operator, required for using Sort<...>.
   *
   * @param l1 first literal
   * @param l2 second literal
   * @return true if l1 is less than l2
   * @since 26/06/2002 Manchester
   * @since 17/07/2003 Manchester, slightly changed
   */
  inline
  static bool lessThan (const Literal& l1, const Literal& l2) {
    return l1.compare(l2) == LESS;
  }

  // various
  // comparison, used for normalisation
  Compare compare (const Literal& l) const; // comparison, used for normalisation
  void normalize (Literal& result) const;
  bool isRenamingOf (Literal l, Substitution& sbst) const;
 private:
  // structure
  Data* _data;
}; // class Literal


class Literal::Data 
#   if DEBUG_PREPRO
    : public Memory <CID_LITERAL>
#   endif
{
 public:
  Data ();
  Data (bool sign, const Atom& atom);

  // structure access
  bool positive () const;
  bool negative () const;
  bool sign () const;
  const Atom& atom () const;

  void ref ();
  void deref ();

 protected:
  // structure
  int _counter;
  bool _sign;
  Atom _atom;
}; // class Literal::Data


class LiteralList 
: public Lst<Literal>
{
 public:
  // constructors
  LiteralList ();
  LiteralList (const LiteralList&);
  explicit LiteralList (const Literal& t); // one-element list
  LiteralList (const Literal& head, const LiteralList& tail);
  explicit LiteralList (LstData<Literal>*);
  LiteralList (const VampireKernel::Literal* lit, const VampireKernel& kernel);

  // inherited functions
  const LiteralList& tail () const
    { return static_cast<const LiteralList&>(Lst<Literal>::tail()); }

  // various
  bool isRenamingOf (LiteralList lst) const;
 private:
  bool isRenamingOf (LiteralList tried, LiteralList remaining, Substitution sbst) const;
}; // class LiteralList


// ******************* class Literal, implementation *********************


inline
Literal::Literal () 
  : 
  _data (0) 
{
} // Literal::Literal


inline
Literal::~Literal () 
{
  if (_data) {
    _data->deref ();
  }
} // Literal::~Literal


inline
Literal::Literal (const Literal& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} // Literal::Literal


inline
Literal::Literal (bool sign, const Atom& atom)
  :
  _data (new Data(sign,atom))
{
} // Literal::Literal (bool sign, const Atom& atom)


inline
bool Literal::exists () const 
{ 
  return _data != 0; 
} // Literal::exists ()


inline
bool Literal::positive () const
{
  return _data->sign();
} // Literal::positive ()


inline
bool Literal::negative () const
{
  return ! _data->sign();
} // Literal::negative ()


inline
bool Literal::sign () const
{
  return _data->sign();
} // Literal::sign ()


inline
const Atom& Literal::atom () const
{
  return _data->atom();
} // Literal::Atom ()


// **************** class Literal::Data implementation ******************


inline
Literal::Data::Data (bool sign, const Atom& atom) 
  : 
  _counter (1),
  _sign (sign), 
  _atom (atom)
{
} // Literal::Data::Data


inline
bool Literal::Data::sign () const 
{ 
  return _sign; 
} // Literal::Data::sign ()


inline
const Atom& Literal::Data::atom () const 
{ 
  return _atom; 
} // Literal::Data::atom ()


inline
void Literal::Data::ref () 
{ 
  ASS (this);

  _counter++;
} // Literal::Data::ref ()


inline
void Literal::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;

  if (_counter == 0) {
    delete this;
  }
} // Literal::Data::deref ()


// ******************* LiteralList definitions ************************

inline
LiteralList::LiteralList () 
  : 
  Lst<Literal> ()
{
} // LiteralList::LiteralList


inline
LiteralList::LiteralList (const LiteralList& ts)
  :
  Lst<Literal> (ts)
{
} // LiteralList::LiteralList


inline
LiteralList::LiteralList (LstData<Literal>* d)
  :
  Lst<Literal> (d)
{
} // LiteralList::LiteralList


inline
LiteralList::LiteralList (const Literal &hd, const LiteralList& tl)
  :
  Lst<Literal> (hd,tl)
{
} // LiteralList::LiteralList


inline
LiteralList::LiteralList (const Literal &hd)
  :
  Lst<Literal> (hd)
{
} // LiteralList::LiteralList


#endif // Literal
