/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
//  file Output.hpp
//  defines various output classes
//

#ifndef __Output__
#define __Output__


#include "Formula.hpp"


// class ostream;
class Literal;
class LiteralList;
class Clause;
class Unit;
class UnitList;
class Refutation;



// ****************** class Output, definition *************************

  /**
   * Output syntax of Vampire. The following kinds are available:
   * <ul>
   * <li><b>tptp:</b> the TPTP syntax, 
   *                  <font color=red>not implemented</font>.</li>
   * <li><b>kif</b> the KIF XML-style syntax.</li>
   * <li><b>native</b> the native syntax,
   *                  <font color=red>not implemented</font>.</li>
   * <li><b>xml</b> the XML syntax.</li>
   * <li><b>latex</b> the LaTeX syntax, can be used to generate LaTeX files.</li>
   * </ul>
   * They can be set using the <font color=red>output_syntax</font>
   * option in the command line.
   */
enum OutputSyntax {
  TPTP = 0,       // the numbers should not be changed
  NATIVE = 1,     // they are used in some analogues of
  KIF = 2,        // virtual tables
  XML = 3,            // XML
  LATEX = 4       // must be the last one
};


/**
 * Base class for various output classes.
 */ 
class Output {
 public:
  // create an output
  static Output* create (OutputSyntax, ostream&);

  // constructors/destructors
  Output (OutputSyntax syntax, ostream& str);

  virtual void pretty (const Refutation&, int answerNumber) = 0;
  virtual void formula (const Formula&);
  virtual void clause (const Clause&) = 0;

 protected:
  class Map; // defined in Output.cpp
  virtual void symbol (const Signature::Symbol*);
  void refutation (const Unit& goal, int indent, Map& map);
  void refutations (const UnitList& goal, int indent, Map& map);
  virtual void inference (const Unit& conclusion, int indent) = 0;
  virtual void connective (Formula::Connective);
  virtual void var (Var v);
  virtual void term (const Term&);
  virtual void terms (const TermList&);
  virtual void atom (const Atom&);
  virtual void literal (const Literal&);
  virtual void clause (const Clause&, int indent) = 0;
  virtual void formula (const Formula&, int indent) = 0;
  void subformula (const Formula&, int indent, Formula::Connective outer);
  void subformula (const Formula&, Formula::Connective outer);
  void content (const Unit& unit); // print clause or formula from unit
  virtual void content (const Unit& unit, int indent);
  bool fitsInOneLine (const Formula&) const;
  static bool fitsInOneLine (const Formula&, int& limit);
  static bool parenthesesRequired (Formula::Connective inner, 
				   Formula::Connective outer);
  void vars (const VarList&);

  // structure
  OutputSyntax _syntax;
  ostream& _stream;

  // printing various data types
  virtual void spaces (int times);
  void repeat (char character, int times);

  // virtual tables
  static int _oneLineLengthLimit [LATEX+1];
}; // class Output


class NativeOutput
  : public Output
{
 public:
  NativeOutput (ostream&);

  // output functions
  void pretty (const Refutation&, int answerNumber);

 public:
  void inference (const Unit& conclusion, int indent);
  void clause (const Clause&);
  void clause (const Clause&, int indent);
  void formula (const Formula&, int indent);
  void formula (const Formula&);
  virtual void connective (Formula::Connective);
};  // class NativeOutput


class TPTPOutput
  : public Output
{
 public:
  TPTPOutput (ostream&);

  // output functions
  void pretty (const Refutation&, int answerNumber);

 private:
  void inference (const Unit& conclusion, int indent);
  void clause (const Clause&);
  void clause (const Clause&, int indent);
  void formula (const Formula&);
  void formula (const Formula&, int indent);
};  // class TPTPOutput


class KIFOutput
  : public Output
{
 public:
  KIFOutput (ostream&);

  // output functions
  void pretty (const Refutation&, int answerNumber);

 private:
  void inference (const Unit& conclusion, int indent);
  void content (const Unit& unit, int indent);
  void clause (const Clause&);
  void clause (const Clause&, int indent);
  void formula (const Formula&);
  void formula (const Formula&, int indent);
  void literal (const Literal&);
  void atom (const Atom&);
  void connective (Formula::Connective);
  void var (Var v);
  void terms (const TermList&);
  void term (const Term&);

  static int _connectivePrintLength[];
};  // class KIFOutput


/**
 * Implements XML output.
 */
class XMLOutput
  : public Output
{
 public:
  XMLOutput (ostream&);

  // output functions
  void pretty (const Refutation&, int answerNumber);

 private:
  void inference (const Unit& conclusion, int indent);
  void content (const Unit& unit, int indent);
  void clause (const Clause&);
  void clause (const Clause&, int indent);
  void formula (const Formula&);
  void formula (const Formula&, int indent);
  void literal (const Literal&);
  void atom (const Atom&);
  void connective (Formula::Connective);
  void var (Var v);
  void terms (const TermList&);
  void term (const Term&);

  static int _connectivePrintLength[];
};  // class XMLOutput


class LaTeXOutput
  : public Output
{
 public:
  LaTeXOutput (ostream&);

  // output functions
  void pretty (const Refutation&, int answerNumber);

 private:
  void symbol (const Signature::Symbol*);
  void inference (const Unit& conclusion, int indent);
  void clause (const Clause&);
  void literals (const LiteralList& lits);
  void literals (const LiteralList& lits, int indent);
  void clause (const Clause&, int indent);
  void formula (const Formula&, int indent);
  void formula (const Formula&);
  void connective (Formula::Connective);
  void var (Var v);
  void spaces (int times);
};  // class LaTeXOutput


// ***************** class Output, implementation ***************


// for debugging only
ostream& operator<< (ostream&, const Formula&);

#endif // __Output__

