/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/**
 * File Property.cpp (syntactic properties of problems)
 *
 * @since 06/06/2001 Manchester
 * @since 17/07/2003 Manchester, changed to new representation
 */

#include "Property.hpp"
#include "Problem.hpp"


Property* Property::global = 0;


/** 
 * Initialize Property. Must be applied to the preprocessed problem.
 *
 * @param p the problem
 * @since 29/06/2002, Manchester
 */
Property::Property (Problem& p)
  :
  _goalClauses (0),
  _axiomClauses (0),
  _equalityAtoms (0),
  _atoms (0),
  _goalFormulas (0),
  _axiomFormulas (0),
  _subformulas (0),
  _terms (0),
  _unitGoals (0),
  _unitAxioms (0),
  _hornGoals (0),
  _hornAxioms (0),
  _equationalClauses (0),
  _pureEquationalClauses (0),
  _groundUnitAxioms (0),
  _positiveAxioms (0),
  _groundPositiveAxioms (0),
  _groundGoals (0),
  _maxFunArity (0), 
  _maxPredArity (0)
{
  read (p);

  // determine the category
  if (formulas() > 0) { // FOF, either FEQ or FNE
    if (_equalityAtoms == 0) {
      _category = FNE;
    }
    else {
      _category = FEQ;
    }
  }
  // no formulas in the input, one of NEQ, HEQ, PEQ, HNE, NNE, EPR, UEQ
  else 
  if (_maxFunArity == 0) { // all function symbols are constants
    _category = EPR;
  }
  // one of NEQ, HEQ, PEQ, HNE, NNE, UEQ
  else if (_equationalClauses == 0 ) { // HNE, NNE
    if (clauses() == hornClauses()) { // all clauses are horn
      _category = HNE;
    }
    else {
      _category = NNE;
    }
  }
  // one of NEQ, HEQ, PEQ, UEQ
  else if (_pureEquationalClauses == clauses()) { // only equations, UEQ or PEQ
    if (clauses() == unitClauses()) { // all clauses are unit
      _category = UEQ;
    }
    else {
      _category = PEQ;
    }
  }
  // one of NEQ, HEQ
  else if (clauses() == hornClauses()) { // all clauses are horn
    _category = HEQ;
  }
  else {
    _category = NEQ;
  }
} // Property::Property


/** 
 * Read property from a problem
 * @since 29/06/2002 Manchester
 */
void Property::read( Problem& p)
{
  TRACER("Property::read (Problem& p)");

  UnitChain::Iterator us (p.units());
  while (us.hasNext()) {
    read (us.next());
  }
} // Property::read (Problem&)


/**
 * Read property from a unit.
 *
 * @since 29/06/2002 Manchester
 * @since 17/07/2003 Manchester, changed to non-pointer types, 
 *        formulas reading added
 */
void Property::read (const Unit& u)
{
  TRACER("Property::read(const Unit&)");

  switch (u.unitType()) 
    {
    case CLAUSE:
      read (u.clause(), u.inputType() == AXIOM);
      return;

    case FORMULA:
      if (u.inputType() == AXIOM) {
	_axiomFormulas ++;
      }
      else {
	_goalFormulas++;
      }
      read(u.formula());
      return;

#if DEBUG_PREPRO
    default:
      ASS(false);
#endif
    }

} // Property::read (const Unit&)


/**
 * Read property from a clause.
 *
 * @param isAxiom true if the clause is an axiom
 * @since 29/06/2002 Manchester
 * @since 17/07/2003 Manchester, changed to non-pointer types
 */
void Property::read (const Clause& clause, bool isAxiom)
{
  TRACER("void Property::read (const Clause& clause, bool isAxiom)");

  int positiveLiterals = 0;
  int negativeLiterals = 0;
  int equationalLiterals = 0;
  int groundLiterals = 0;
  Iterator<Literal> lits (clause.literals());
  while (lits.more()) {
    Literal literal (lits.next());
    if (literal.positive()) {
      positiveLiterals ++;
    }
    else {
      negativeLiterals ++;
    }

    Atom atom (literal.atom());
    if (atom.isEquality()) {
      equationalLiterals++;
    }

    bool isGround = true;
    read (atom, isGround);

    if (isGround) {
      groundLiterals++;
    }
  }
  int literals = positiveLiterals + negativeLiterals;
  _atoms += literals;

  if ( equationalLiterals > 0 ) {
    _equationalClauses ++;
    _equalityAtoms += equationalLiterals;
  }
  if ( literals == equationalLiterals ) {
    _pureEquationalClauses ++;
  }

  if (isAxiom) {
    _axiomClauses ++;
    if ( literals == 1) {
      _unitAxioms ++;
      if ( groundLiterals == 1) {
	_groundUnitAxioms ++;
      }
    }
    if (positiveLiterals <= 1) {
      _hornAxioms ++;
    }
    if (negativeLiterals == 0) {
      _positiveAxioms ++;
      if (literals == groundLiterals) {
	_groundPositiveAxioms ++;
      }
    }
  }
  else {
    _goalClauses ++;
    if ( literals == 1) {
      _unitGoals ++;
    }
    if (positiveLiterals <= 1) {
      _hornGoals ++;
    }
    if (literals == groundLiterals) {
      _groundGoals ++;
    }
  }
} // Property::read (const Clause* clause, bool isAxiom)


/**
 * Read property from a formula.
 *
 * @since 17/07/2003 Manchester
 */
void Property::read (const Formula& formula)
{
  TRACER("void Property::read (const Formula&)");

  _subformulas++;
  switch (formula.connective()) 
    {
    case Formula::ATOM:
      {
	_atoms++;
	Atom a (formula.atom());
	if (a.isEquality()) {
	  _equalityAtoms++;
	}
	bool dummy = false;
	read(a,dummy);
	return;
      }
      
    case Formula::AND:
    case Formula::OR:
      {
	Iterator<Formula> args (formula.args());
	while (args.more()) {
	  read(args.next());
	}
	return;
      }

    case Formula::IMP:
    case Formula::IFF:
    case Formula::XOR:
      read(formula.left());
      read(formula.right());
      return;

    case Formula::NOT:
      read(formula.uarg());
      return;

    case Formula::FORALL:
    case Formula::EXISTS:
      read(formula.qarg());
      return;

#if DEBUG_PREPRO
    default:
      ASS (false);
#endif
    }
} // Property::read (const Formula&)


/**
 * Read property from an atom.
 *
 * @param isGround will be set to false if the atom contains a variable
 * @since 29/06/2002 Manchester
 * @since 17/07/2003 Manchester, changed to non-pointer types
 */
void Property::read (const Atom& atom, bool& isGround)
{
  TRACER("void Property::read (const Atom& atom, bool& isGround)");

  if (! atom.isEquality()) {
    int arity = atom.functor()->arity();
    if (arity > _maxPredArity) {
      _maxPredArity = arity;
    }
  }

  Iterator<Term> args (atom.args());
  while ( args.more() ) {
    read (args.next(), isGround);
  }
} // Property::read (const Atom& term, bool& isGround)


/**
 * Read property from a term.
 *
 * @param isGround will be set to false if the term contains a variable
 * @since 29/06/2002 Manchester
 * @since 17/07/2003 Manchester, changed to non-pointer types, 
 *        also NUMERIC case added
 */
void Property::read (const Term& term, bool& isGround)
{
  TRACER("void Property::read (const Term& term, bool& isGround)");

  _terms ++;

  switch (term.tag()) 
    {
    case Term::VAR:
      isGround = false;
      return;

    case Term::NUMERIC:
      return;

    case Term::COMPOUND: 
      {
	int arity = term.functor()->arity();
	if (arity > _maxFunArity) {
	  _maxFunArity = arity;
	}

	Iterator<Term> args (term.args());
	while ( args.more() ) {
	  read (args.next(), isGround);
	}
	return;
      }

#if DEBUG_PREPRO
    default:
      ASS(false);
#endif
    }
} // Property::read (const Term& term, bool& isGround)


/**
 * Return the string representation of the CASC category.
 */ 
string Property::categoryString() const
{
  TRACER("string Property::categoryString() const");

  switch (_category)
    {
    case NEQ:
      return "NEQ";
    case HEQ:
      return "HEQ";
    case PEQ:
      return "PEQ";
    case HNE:
      return "HNE";
    case NNE:
      return "NNE";
    case FEQ:
      return "FEQ";
    case FNE: 
      return "FNE";
    case EPR:
      return "EPR";
    case UEQ:
      return "UEQ";
#if DEBUG_PREPRO
    default:
      ASS(false);
      return "";
#endif
    }
} // categoryString


// from the previous version
// char Property::axiomTypes () const
// {
//   if (_unitAxioms == _axiomClauses) { // all axioms are unit
//     return 'u';
//   }
//   if (_hornAxioms == _axiomClauses) { // all axioms are horn
//     return 'h';
//   }
//   // general axiom types
//   return 'g';
// } // Property::axiomTypes


// char Property::goalTypes () const
// {
//   if (_unitAxioms == _goalClauses) { // all goals are unit
//     return 'u';
//   }
//   if (_hornAxioms == _goalClauses) { // all goals are horn
//     return 'h';
//   }
//   // general axiom types
//   return 'g';
// } // Property::goalTypes


// char Property::equalityContent () const
// {
//   if (_equationalClauses == 0) { // no equality
//     return 'n';
//   }
//   if (_pureEquationalClauses == clauses()) { // all clauses are pure equational
//     return 'p';
//   }
//   // some equational literals but not all
//   return 's';
// } // Property::equalityContent


// // This function has been generated automatically
// // on Sun Jun 30 11:44:28 2002
// char Property::nonGroundUnitContent () const
// {
//   float quantity = _unitAxioms ?
//     ((float)(_unitAxioms - _groundUnitAxioms)) / (float)_unitAxioms :
//     0.5;
 
//   switch (_category) {
//   case EPR:
//     if (quantity <= 0.105263) { return 'f'; } // few
//     if (quantity <= 0.5) { return 's'; } // some
//     return 'm'; // many
//   case HEQ:
//     if (quantity <= 0.5) { return 'f'; } // few
//     if (quantity <= 0.99999) { return 's'; } // some
//     return 'm'; // many
//   case UEQ:
//     if (quantity <= 0.666667) { return 'f'; } // few
//     if (quantity <= 0.99999) { return 's'; } // some
//     return 'm'; // many
//   case PEQ:
//     if (quantity <= 0.933333) { return 'f'; } // few
//     if (quantity <= 0.99999) { return 's'; } // some
//     return 'm'; // many
//   case NNE:
//     if (quantity <= 1e-05) { return 'f'; } // few
//     if (quantity <= 0.5) { return 's'; } // some
//     return 'm'; // many
//   case NEQ:
//     if (quantity <= 0.7) { return 'f'; } // few
//     if (quantity <= 0.807692) { return 's'; } // some
//     return 'm'; // many
//   case HNE:
//     if (quantity <= 0.5) { return 'f'; } // few
//     if (quantity <= 0.99999) { return 's'; } // some
//     return 'm'; // many
//   default:
//     ASSERT(debug_property, false);
//   }
// } // Property::nonGroundUnitContent ()


// // This function has been generated automatically
// // on Sun Jun 30 11:36:41 2002
// char Property::groundPositiveContent () const
// {
//   float quantity = _positiveAxioms ?
//     ((float)(_groundPositiveAxioms)/ (float)(_positiveAxioms)) :
//     0.5;
 
//   switch (_category) {
//   case EPR:
//     if (quantity <= 0.105263) { return 'f'; } // few
//     if (quantity <= 0.5) { return 's'; } // some
//     return 'm'; // many
//   case HEQ:
//     if (quantity <= 0.5) { return 'f'; } // few
//     if (quantity <= 0.99999) { return 's'; } // some
//     return 'm'; // many
//   case UEQ:
//     if (quantity <= 0.666667) { return 'f'; } // few
//     if (quantity <= 0.99999) { return 's'; } // some
//     return 'm'; // many
//   case PEQ:
//     if (quantity <= 0.933333) { return 'f'; } // few
//     if (quantity <= 0.99999) { return 's'; } // some
//     return 'm'; // many
//   case NNE:
//     if (quantity <= 1e-05) { return 'f'; } // few
//     if (quantity <= 0.5) { return 's'; } // some
//     return 'm'; // many
//   case NEQ:
//     if (quantity <= 0.7) { return 'f'; } // few
//     if (quantity <= 0.807692) { return 's'; } // some
//     return 'm'; // many
//   case HNE:
//     if (quantity <= 0.5) { return 'f'; } // few
//     if (quantity <= 0.99999) { return 's'; } // some
//     return 'm'; // many
//   default:
//     ASSERT(debug_property, false);
//   }
// } // Property::groundPositiveContent ()


// char Property::goalsAreGround () const
// {
//   if (_goalClauses == _groundGoals) { // all goals are ground
//     return 'g';
//   }
//   // general axiom types
//   return 'n';
// } // Property::axiomTypes


// // This function has been generated automatically
// // on Sun Jun 30 12:53:21 2002
// char Property::setClauseSize () const
// {
//   int quantity = clauses ();
 
//   switch (_category) {
//   case EPR:
//     if (quantity <= 51) { return 'f'; } // few
//     if (quantity <= 340) { return 's'; } // some
//     return 'm'; // many
//   case HEQ:
//     if (quantity <= 11) { return 'f'; } // few
//     if (quantity <= 18) { return 's'; } // some
//     return 'm'; // many
//   case UEQ:
//     if (quantity <= 5) { return 'f'; } // few
//     if (quantity <= 16) { return 's'; } // some
//     return 'm'; // many
//   case PEQ:
//     if (quantity <= 4) { return 'f'; } // few
//     if (quantity <= 8) { return 's'; } // some
//     return 'm'; // many
//   case NNE:
//     if (quantity <= 30) { return 'f'; } // few
//     if (quantity <= 38) { return 's'; } // some
//     return 'm'; // many
//   case NEQ:
//     if (quantity <= 95) { return 'f'; } // few
//     if (quantity <= 162) { return 's'; } // some
//     return 'm'; // many
//   case HNE:
//     if (quantity <= 5) { return 'f'; } // few
//     if (quantity <= 9) { return 's'; } // some
//     return 'm'; // many
//   default:
//     ASSERT(debug_property, false);
//   }
// } // Property::setClauseSize ()


// // This function has been generated automatically
// // on Sun Jun 30 13:21:57 2002
// char Property::setLiteralSize () const
// {
//   int quantity = _literals;
 
//   switch (_category) {
//   case EPR:
//     if (quantity <= 98) { return 'f'; } // few
//     if (quantity <= 1060) { return 's'; } // some
//     return 'm'; // many
//   case HEQ:
//     if (quantity <= 14) { return 'f'; } // few
//     if (quantity <= 29) { return 's'; } // some
//     return 'm'; // many
//   case UEQ:
//     if (quantity <= 5) { return 'f'; } // few
//     if (quantity <= 16) { return 's'; } // some
//     return 'm'; // many
//   case PEQ:
//     if (quantity <= 6) { return 'f'; } // few
//     if (quantity <= 12) { return 's'; } // some
//     return 'm'; // many
//   case NNE:
//     if (quantity <= 77) { return 'f'; } // few
//     if (quantity <= 93) { return 's'; } // some
//     return 'm'; // many
//   case NEQ:
//     if (quantity <= 219) { return 'f'; } // few
//     if (quantity <= 351) { return 's'; } // some
//     return 'm'; // many
//   case HNE:
//     if (quantity <= 7) { return 'f'; } // few
//     if (quantity <= 14) { return 's'; } // some
//     return 'm'; // many
//   default:
//     ASSERT(debug_property, false);
//   }
// } // Property::setLiteralSize ()


// // This function has been generated automatically
// // on Sun Jun 30 13:40:24 2002
// char Property::setTermSize () const
// {
//   int quantity = _terms;
 
//   switch (_category) {
//   case EPR:
//     if (quantity <= 142) { return 'f'; } // few
//     if (quantity <= 2026) { return 's'; } // some
//     return 'm'; // many
//   case HEQ:
//     if (quantity <= 70) { return 'f'; } // few
//     if (quantity <= 120) { return 's'; } // some
//     return 'm'; // many
//   case UEQ:
//     if (quantity <= 47) { return 'f'; } // few
//     if (quantity <= 127) { return 's'; } // some
//     return 'm'; // many
//   case PEQ:
//     if (quantity <= 51) { return 'f'; } // few
//     if (quantity <= 72) { return 's'; } // some
//     return 'm'; // many
//   case NNE:
//     if (quantity <= 184) { return 'f'; } // few
//     if (quantity <= 218) { return 's'; } // some
//     return 'm'; // many
//   case NEQ:
//     if (quantity <= 834) { return 'f'; } // few
//     if (quantity <= 1282) { return 's'; } // some
//     return 'm'; // many
//   case HNE:
//     if (quantity <= 38) { return 'f'; } // few
//     if (quantity <= 78) { return 's'; } // some
//     return 'm'; // many
//   default:
//     ASSERT(debug_property, false);
//   }
// } // Property::setTermSize ()


// char Property::maxFunArity () const
// {
//   switch (_maxFunArity) {
//   case 0:
//     return '0';
//   case 1:
//     return '1';
//   case 2:
//     return '2';
//   default:
//     return '3';
//   }
// } // Property::maxFunArity


// char Property::maxPredArity () const
// {
//   switch (_maxPredArity) {
//   case 0:
//     return '0';
//   case 1:
//     return '1';
//   case 2:
//     return '2';
//   default:
//     return '3';
//   }
// } // Property::maxPredArity


// void Property::toString (char* string) const
// {
//   *string++ = categoryToChar ();
//   *string++ = axiomTypes ();
//   *string++ = goalTypes ();
//   *string++ = equalityContent ();
//   *string++ = nonGroundUnitContent ();
//   *string++ = groundPositiveContent ();
//   *string++ = goalsAreGround ();
//   *string++ = setClauseSize ();
//   *string++ = setLiteralSize ();
//   *string++ = setTermSize ();
//   *string++ = maxPredArity ();
//   *string++ = maxFunArity ();

//   // mark end of string
//   *string = 0;
// } // Property::toString


// const char* Property::categoryToString () const
// {
//   switch ( _category ) {
//   case NEQ:
//     return "NEQ";
//   case HEQ:
//     return "HEQ";
//   case PEQ:
//     return "PEQ";
//   case HNE:
//     return "HNE";
//   case NNE:
//     return "NNE";
// //   case FEQ:       
// //     return "FEQ";
// //   case FNE:
// //     return "FNE";
//   case EPR:  
//     return "EPR";
//   case UEQ:
//     return "UEQ";
//   default:
//     ASSERT (debug_property, false);
//   } 
// } // Property::categoryToString


// char Property::categoryToChar () const
// {
//    if ( _hasFormulas ) { // FOF, either FEQ or FNE
//      return _equationalClauses == 0 ? 'f' : 'F';
//   }

//   switch ( _category ) {
//   case NEQ:
//     return 'N';
//   case HEQ:
//     return 'H';
//   case PEQ:
//     return 'P';
//   case HNE:
//     return 'h';
//   case NNE:
//     return 'n';
// //   case FEQ:       
// //     return 'F';
// //   case FNE:
// //     return 'f';
//   case EPR:  
//     return 'E';
//   case UEQ:
//     return 'U';
//   default:
//     ASSERT (debug_property, false);
//   } 
// } // Property::categoryToString
