/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/**
 * File Property.hpp (syntactic properties of problems).
 *
 * @since 06/06/2001 Manchester
 * @since 17/07/2003 Manchester, changed to new representation
 */


#ifndef __Property__
#define __Property__


#include <string>


class Problem;
class Unit;
class Clause;
class Literal;
class Atom;
class Term;
class Formula;
class ostream;            // for debugging


/**
 * Represents syntactic properties of problems.
 */
class Property 
{
 public:

  /**
   * CASC category
   */
  enum Category {
    NEQ,  
    HEQ,
    PEQ,
    HNE,
    NNE,
    FEQ,       // fof division 
    FNE, 
    EPR,       // epr division
    UEQ        // ueq division
  };

  // constructor
  explicit Property (Problem&);

  /**
   * Return the CASC category of the problem
   */
  Category category() const {
    return _category;
  }

  /**
   * Return the string representation of the category, for example, "HEQ".
   */
  string categoryString() const;

  /**
   * Global property, can be set by preprocessing and accessed by 
   * option analysers.
   */
  static Property* global;

  // output
  //void toString (char* string) const;
  //const char* categoryToString () const;
  //char categoryToChar () const;

 private:
  // reading in properties of problems
  void read (Problem&);
  void read (const Unit&);
  void read (const Formula&);
  void read (const Clause&, bool isAxiom);
  void read (const Atom&, bool& isGround);
  void read (const Term&, bool& isGround);

  char axiomTypes () const;
  char goalTypes () const;
  char equalityContent () const;
  char nonGroundUnitContent () const;
  char groundPositiveContent () const;
  char goalsAreGround () const;
  char setClauseSize () const;
  char setLiteralSize () const;
  char setTermSize () const;
  char maxFunArity () const;
  char maxPredArity () const;

 public:
  /**
   * Total number of clauses in the problem.
   */ 
  inline int clauses () const { 
    return _goalClauses + _axiomClauses; 
  }

  /**
   * Total number of formulas in the problem.
   */ 
  inline int formulas () const { 
    return _goalFormulas + _axiomFormulas; 
  }

  /**
   * Total number of unit clauses in the problem.
   */ 
  inline int unitClauses () const { 
    return _unitGoals + _unitAxioms; 
  }

  /**
   * Total number of Horn clauses in the problem.
   */ 
  inline int hornClauses () const { 
    return _hornGoals + _hornAxioms; 
  }

  // structure
  int _goalClauses;
  int _axiomClauses;

  int _equalityAtoms;
  int _atoms;

  int _goalFormulas;
  int _axiomFormulas;
  int _subformulas;

  int _terms;
  int _unitGoals;
  int _unitAxioms;
  int _hornGoals;
  int _hornAxioms;
  int _equationalClauses;
  int _pureEquationalClauses;
  int _groundUnitAxioms;
  int _positiveAxioms;
  int _groundPositiveAxioms;
  int _groundGoals;
  int _maxFunArity;
  int _maxPredArity;

  // derived
  Category _category;
}; // class Property


// for debugging
ostream& operator << ( ostream&, Property& );


#endif // __Property__
