/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
//  file Refutation.cpp
//  
//  implements class Refutation
//


#include "Refutation.hpp"
#include "Problem.hpp"
#include "IntNameTable.hpp"
#include "Unit.hpp"
#include "Map.hpp"


class Refutation::Map 
  : public ::Map <int, Unit>
{
}; // class Refutation::Map


Refutation::Refutation (const VampireKernel& kernel, const Problem& problem)
  : _kernel (kernel),
    _problem(problem)
{
  TRACER ("Refutation::Refutation");
 
  Map map;

  _goal = collect(kernel.refutation(), map);
} // Refutation::Refutation


// build the refutation rooting at the returned unit
// 29/09/2002 Manchester
// 04/10/2002 Manchester changed and simplified
Unit Refutation::collect (const VampireKernel::Clause* clause, Map& map)
{
  TRACER ("Refutation::collect (VampireKernel::Clause ...)");

  int number = clause->number();

  Unit unit;
  if ( map.find(number, unit) ) { // clause number is already in the map
    return unit;
  }

  // the number of the clause was NOT in the map
  if (clause->isInputClause()) { // the clause is an input clause 
    // the clause should not be added to the refutation but his parent unit should
    Unit parent (clause->inputClauseOrigin());
    return parent;
  }

  // the clause is not an input clause
  // collect all parents of the clause
  // and create the list of parent units
  UnitList parents = collect (clause->firstAncestor(), map);

  // create the unit itself
  Clause cls (clause, _kernel);
  Unit result (cls, parents);
  map.insert (number, result);

  return result;
} //  Refutation::collect (VampireKernel::Clause* ...)


// build a list of units representing a list of refutations refutations
// 04/10/2002 Manchester
UnitList Refutation::collect (const VampireKernel::Clause::Ancestor* parent, Map& map)
{
  TRACER ("Refutation::collect (VampireKernel::Clause::Ancestor* ...)");

  if (! parent) {
    return UnitList ();
  }

  // the list of parents is non-empty
  const VampireKernel::Clause* clause = parent->clause();
  const VampireKernel::Clause::Ancestor* rest = parent->next ();
  if (clause->isBuiltInTheoryFact()) {
    return collect (rest, map);
  }

  // not built-in theory fact
  Unit u (collect (clause, map));
  return UnitList (u, collect (rest, map));
} //  Refutation::collect (VampireKernel::Clause* ...)


// condense the refutation by removing all single-parent inferences
// 04/10/2002 Manchester
Unit Refutation::condense () const
{
  Map condensed;

  return condense (_goal, condensed);
} // Refutation::condense ()


// condense the refutation by removing all single-parent inferences
// 04/10/2002 Manchester
Unit Refutation::condense (Unit root, Map& prev) const
{
  if ( prev.find (root.number(), root) ) { // already condensed
    return root;
  }

  // not condensed yet
  UnitList condensedParents = condense (root.parents(), prev);
  if ( condensedParents.isNonEmpty() && 
       condensedParents.tail().isEmpty() ) { // one parent, case for condensing
    Unit parent ( condensedParents.head() ); // the single parent
    if ( parent.unitType() == FORMULA &&
         root.unitType() == CLAUSE ) { // clause obtained from formula
      // don't condense any more
      if ( condensedParents == root.parents() ) { // the parent was not in fact condensed
	prev.insert (root.number(), root);
	return root;
      }
      // the parent was condensed
      Unit result (root.clause(), condensedParents);
      prev.insert (root.number(), result);
      return result;      
    }
    // both the parent and the root have the same unit type, unconditionally condense
    // root to parent
    prev.insert (root.number(), parent);
    return parent;
  }

  // several parents, do not condense
  if ( condensedParents == root.parents() ) { // no change at all
    prev.insert (root.number(), root);
    return root;
  }

  // parents have change, so return a new unit
  if (root.unitType() == CLAUSE) {
    Unit result (root.clause(), condensedParents);
    prev.insert (root.number(), result);
    return result;      
  }
  // unit contains a formula
  Unit result (root.formula(), condensedParents);
  prev.insert (root.number(), result);
  return result;      
} // Refutation::condense (... Unit ...)


// condense a list of refutations by removing all single-parent inferences
// 04/10/2002 Manchester
UnitList Refutation::condense (UnitList units, Map& prev) const
{
  if (units.isEmpty()) {
    return units;
  }

  Unit first (units.head());
  UnitList rest (units.tail());
  Unit f ( condense (first, prev) );
  UnitList r ( condense (rest, prev) );

  if (f == first && r == rest) { // nothing has changed
    return units;
  }

  return UnitList (f, r);
} // Refutation::condense (... UnitList units ...)


/*
// BlodKorv      15,361
// Gematogen     10,559
// VampireKernel 72,429
// preprocessor  23,729
// --------------------
//             

*/
