;;;*************************************************************************
;; ---------------------
;; UNIFICATION PROCEDURE
;; ---------------------
(module unification
	(include "dctp.sch")
	(export (unbind-vars) (unbind-vars-with-count) (ref term) (unify-lists k l) (unify-terms s t) (unbind-vars-and-undo-var-counter)
		(occurs-list var termlist) (occurs-term var term) (copy-trailed-bindings) (untrail-subgoals) (match-termlists k l) (match-terms s t))
	(import (display "display.scm") (globals "globals.scm") (free-index-variable index "index.scm") (index-variable-binding index "index.scm")))

(define (ref term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
  (if (index-variable term) 
      (if (free-index-variable term) 
	  term
	  (ref (index-variable-binding term)))
      (if (or (non-variable term) (free-variable term) (domain-element term))
	  term
	  (ref (car term)))))

(define (unify-lists k l)
;; unifies termlists k and l.
;; If unification succeeds, #t is returned, else #f.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     #t
     (if (unify-terms (ref (car k)) (ref (car l)))
	 (unify-lists (cdr k) (cdr l))
	 #f)))

(define (unify-terms s t)          
;; unifies terms s and t. 
;; If unification succeeds, #t is returned, else #f.
;; Side effect: terms and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
         (if (variable t) 
             (bind s t)
	     (if (occurs-list s (cdr t))
                 (unbind-vars)
  	         (bind s t)))
         (if (variable t)
             (if (occurs-list t (cdr s))
	         (unbind-vars)
                 (bind t s))
             (if (eq? (car s) (car t))
	         (unify-lists (cdr s) (cdr t))
                 (unbind-vars))))))

(define (occurs-list var termlist)              
;; returns #t if var occurs in termlist, and #f otherwise
 (if (null? termlist)
     #f
     (if (occurs-term var (ref (car termlist)))
         #t
         (occurs-list var (cdr termlist)))))

(define (occurs-term var term)       
;; returns #t if var occurs in term, and #f otherwise
 (if (eq? var term) 
     #t
     (and (non-variable term) (occurs-list var (cdr term)))))


;;;*************************************************************************
;; -----------------------------
;; TRAIL MODIFICATION PROCEDURES

(define (unbind-vars)
;; makes free all variables back to the previous trailstop
 (if (trail-content)
     (begin 
       ;;(display "unbinding ") 
       ;;(if (variable (caar trail)) (display "variable "))
       ;;(if (caaar trail)
	   ;;(display-ref (ref (caar trail)))
	   ;;(display (caaar trail)))
       (set-car! (caar trail) '()) 
       ;;(display " to ") 
       ;;(display-ref (caar trail))
       ;;(newline)
       (trail-previous) (unbind-vars))
     (begin 
       (trail-previous) 
       #f)))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Trail functions for \forall-unification
;; ------------------------------------------------------------

;; this works without destructive manipulation of the trail, so all the (trail-... macros
;; connot be used
(define (copy-trailed-bindings)
  (copy-bindings trail))

(define (copy-bindings trail-pointer)
  (if (caar trail-pointer)
      (begin
	(set!  nr-copied-terms (+ nr-copied-terms 1))
	(if (not (and local-vars (local-variable (caar trail-pointer))))
	    (bind-var (caar trail-pointer) (copy-binding-term (ref (caaar trail-pointer)))))
	(copy-bindings (cdar trail-pointer))
	(if (and local-vars (local-variable (caar trail-pointer)))
	    (unbind-var (caar trail-pointer))))
      #t))
	
(define (copy-binding-term term)
  (if (free-variable term)
      (let ((new-var (new-variable)))
	(set-variable-name new-var (gensym `new-var))
	new-var)
      (if (constant term)
	  term
	  (cons (car term) (copy-binding-terms (cdr term))))))

(define (copy-binding-terms terms)
  (if (pair? terms)
      (cons (copy-binding-term (ref (car terms))) (copy-binding-terms (cdr terms)))
      '()))

(define (unbind-vars-and-undo-var-counter)
;; makes free all variables back to the previous trailstop, decrementing var-counter for each freed variable
 (if (trail-content)
     (begin 
       (set-car! (caar trail) '()) 
       (set! var-counter (- var-counter 1)) 
       (trail-previous) 
       (unbind-vars-and-undo-var-counter))
     (begin (trail-previous) #f)))

(define (unbind-vars-with-count)
;; makes free all variables back to the previous trailstop, returning the number of freed vars
 (if (trail-content)
     (let ((binding (ref (caar trail))))
       (set-car! (caar trail) '()) (trail-previous) 
       (if (variable binding)
	   (unbind-vars-with-count)
	   (+ 1 (unbind-vars-with-count))))
     (begin (trail-previous) 0)))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Trail functions for unit propagation
;; ------------------------------------------------------------

(define (untrail-subgoals)
;; undoes the subgoal references back to the previous trailstop
 (if (trail-content)
     (begin 
       (set-car! (car trail) #f) 
       (trail-previous) (untrail-subgoals))
     (begin 
       (trail-previous) 
       #f)))



;;;*************************************************************************
;; ----------------------------------------------------------------
;; Term matching functions (depend on glob-var instantiatable-vars)
;; ----------------------------------------------------------------

(define (match-termlists k l)
  ;; tries to match termlists k and l.
  ;; If matching succeeds, #t is returned, else #f.
  ;; Side effect: termlists and trail destructively modified.
  ;; Before every matching, (trail-stop) must be called in order
  ;; to mark the trail position for the stored substitution
  ;; Also, instantiatable-vars must be set to the list of variables occurring in k
   (if (null? k)
     #t
     (if (match-terms (ref (car k)) (ref (car l)))
	 (match-termlists (cdr k) (cdr l))
	 #f)))

(define (match-terms s t)
  ;; tries to match terms s and t.
  ;; If matching succeeds, #t is returned, else #f.
  ;; Side effect: terms and trail destructively modified.
  ;; Before every unification, (trail-stop) must be called in order
  ;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
	 (if (memq s instantiatable-vars)
	     (bind s t)
	     (unbind-vars))
	 (if (variable t)
	     (unbind-vars)
	     (if (eq? (car s) (car t))
	         (match-termlists (cdr s) (cdr t))
                 (unbind-vars))))))
