(module constraints
	(include "setheo.sch")
	(export (undo-v-trail) (propagate-substitution)
		(extract-disequation-constraint-relevant)
		(make-disequation-constraint l1 l2))
	(import (globals "globals.scm")
		(display-ref display "display.scm")
		(display-var-name display "display.scm")
		(unification "unification.scm")))

; General information.
; --------------------
; This is the code for integrating disequation constraints into Scheme-Setheo.
; General changes to the prover are the following.
;
; 1) Generation of structural constraints.
; After every extension step, structural constraints (regularity) is generated
; and pushed on the constraint stack, and the constraint lists of the
; respective variables are modified. The basic procedure for normalizing and
; integrating a constraint is make-disequation-constraint. It uses the
; recursive procedure extract-constraint as subroutine.
;
; 2) Propagation of substitutions to constraints.
; After every successful inference step, the last computed substitution is
; propagated to the constraints. The main procedure for doing this is
; propagate-substitution. Again the procedure extract-constraint is used as
; subroutine. 
; 
; 3) Backtracking.
; The changes performed on the constraint data are stored in the value trail,
; which is organized similar to the trail (for variable bindings). 
; Before every sequence of constraint modifications associated with an 
; inference step, a v-trail-stop is set. When backtracking an inference step,
; the changes documented on the value trail have to be undone (from the
; current top of trail (the global variable v-trail) back to the previous
; v-trail-stop. The respective procedure is undo-v-trail. 
;
; Important note: if a constraint violation occurs during the generation or
; propagation of constraints, then the unwinding of the value stack is not
; done implicitly in those procedures, but most be done explicitly afterwards.
; Furthermore, in order to permit reuse of the constraint stack, the global
; variable c-stack has to be set back to the old top of the constraint stack.
; Both operations are done in the procedures extension and reduction-steps.
;
; Relevant data structures:
; -------------------------
; VARIABLE is ( VALUE CONSTRAINTS . NAME )
;
; CONSTRAINT-STACK is list of SUBCONSTRAINTS or HEADERS
; global variable c-stack-root is entire CONSTRAINT-STACK
; global variable c-stack is actual top of CONSTRAINT-STACK
; SUBCONSTRAINT is #( BINDING HEADER NEXT-SUBCONSTRAINT )
; HEADER is #( LENGTH #t/#f DUMMY )
;
; VALUE-TRAIL is list of VALUE-TRAILPOINTS
; VALUE-TRAILPOINT is #( STRUCT POS OLD-VALUE PREVIOUS-VALUE-TRAILPOINT ) 
; POS is car, cdr (for cons cells), or a natural number (the vector index)
; global variable v-trail-root is entire VALUE-TRAIL
; global variable v-trail is actual top of VALUE-TRAIL

;;;*************************************************************************
;; -------------------------------
;; GENERATE DISEQUATION CONSTRAINT
;; -------------------------------

(define (make-disequation-constraint l1 l2)
; integrates the normalization of the disequation constraint l1 =/= l2
; output: #f if the constraint is already violated, otherwise #t.
; Side effects on the constraints are stored on the value trail.
; Note: the procedure misuses the variable trail, which is restored afterwards
 (if (unify-lists-stop l1 l2) ; protocol unification performed
     (if (trail-content) ; constraint has length > 0
	 (begin (constraint-stack-next) ; new header is made
		(vector-set! (car c-stack) 0 0) ; length in header set to 0
		(vector-set! (car c-stack) 1 #f) ; constraint is not true
		(extract-constraint 1 (car c-stack) #f)
		(trail-previous) #t)
	 (begin (trail-previous) #f)) ; constraint is immediately violated
     #t)) ; term lists not unifiable: no constraint generated

(define (extract-disequation-constraint-relevant)
; extracts relevant disequation constraints from last binding on trail:
; all variables whose (var-name var) are = successful-inferences are
; irrelevant
  (if (trail-content) ; constraint has length > 0
      (if (not (eq? (var-name (caar trail)) actual-copy)) ; variable relevant
	  (begin (constraint-stack-next) ; new header is made
		 (vector-set! (car c-stack) 0 0) ; length in header set to 0
		 (vector-set! (car c-stack) 1 #f) ; constraint is not true
		 (extract-constraint 1 (car c-stack) #t)
		 (trail-previous) #t)
	  (begin (set-car! (caar trail) '()) ; variable binding undone
		 (trail-previous)
		 ; (display-var-name (caar trail)) (print " irrelevant")
		 (extract-disequation-constraint-relevant)))
      (begin (trail-previous) #f))) ; constraint is immediately violated

(define (extract-constraint length c-header rel-tag)
;; makes free all variables back to the previous trailstop
;; and pushes the bindings as subconstraints on the constraint stack;
;; relevant changes are documented on the value trail;
; (caar trail) is current variable, (car c-stack) is new subconstraint
 
; (if debug (begin (display "SUBCONSTRAINT: ") (display-var-name (caar trail))
; 		 (display "/") ;(display-ref (var-binding (caar trail)))
; 		 (print "BLA")
; 		 (newline)))
 (v-trail-next) ; old value of variable constraint list is stored:
 (vector-set! (car v-trail) 0 (cdr (caar trail))) ; affected cons cell  
 (vector-set! (car v-trail) 1 'car) ; memorizes that car was changed
 (vector-set! (car v-trail) 2 (var-constraints (caar trail))) ; old value

 (constraint-stack-next) ; new subconstraint generated and added:
 (vector-set! (car c-stack) 0 (ref (var-binding (caar trail)))) ; binding
 (vector-set! (car c-stack) 1 c-header) ; pointer to constraint header
 (vector-set! (car c-stack) 2 (var-constraints (caar trail))) ; old constraints
 (set-car! (cdr (caar trail)) (car c-stack)) ; new variable constraint list 
	    
 (set-car! (caar trail) '()) ; var. binding from protocol unification undone
 (trail-previous) ; moves to previous trail-point 
 (if rel-tag (remove-irrelevant-from-trail))
 (if (trail-content) ; next binding to be processed
     (extract-constraint (+ length 1) c-header rel-tag)
     ; below: all bindings processed; now constraint length inserted in header
     (vector-set! c-header 0 (+ (vector-ref c-header 0) length))
     ))

(define (remove-irrelevant-from-trail)
  (if (trail-content)
      (if (eq? (var-name (caar trail)) actual-copy) ; variable irrelevant
	  (begin ; (display-var-name (caar trail)) (print " irrelevant")
		 (set-car! (caar trail) '()) (trail-previous)
		 (remove-irrelevant-from-trail)))))

;;;*************************************************************************
;; -------------------------------------------------
;; PROPAGATE SUBSTITUTION TO DISEQUATION CONSTRAINTS
;; -------------------------------------------------

(define (propagate-substitution)
; applies current substitution, which is documented on the variable trail,
; to the constraints in the constraint stack;
; output: #t/#f if constraints are satisfiable/unsatisfiable.
; Side effects on the constraints are stored on the value trail.
; Note: the procedure misuses the variable trail, which is restored afterwards
 (if (trail-content) ; substitution is nonempty
     (propagate-bindings trail)
     #t)) ; no propagation to be performed

(define (propagate-bindings b-trail)
; (caar b-trail) is current variable
; b-trail is the trail position of the respective binding; the global variable
; trail cannot be used for this purpose, since the trail is misused for
; protocol unifications in subroutines of this procedure
 (if (propagate-binding (var-binding (caar b-trail)) ; current binding does not
			(var-constraints (caar b-trail))) ; violate constraints
     (begin (set! b-trail (trail-previous-loc b-trail)) ; process next binding 
	    (if (trail-content-loc b-trail) ; remaining substitution nonempty
		(propagate-bindings b-trail)
		#t)) ; all bindings propagated and no constraint violated
     #f)) ; some constraint violated

(define (propagate-binding binding subconstraint)
; subconstraint is the top subconstraint of the respective variable, i.e.,
; either #f or otherwise a vector of length 3 with its last element pointing
; to the next subconstraint
 (if subconstraint
     (if (or (is-passiv-constraint? subconstraint)
	     (expand-subconstraint binding subconstraint))
	 (propagate-binding binding (vector-ref subconstraint 2))
	 #f) ; current constraint is violated
     #t)) ; binding has been propagated to all subconstraints of this variable

(define (expand-subconstraint binding subconstraint)
 (if (unify-terms-stop ; protocol unification performed
		       (ref binding)
		       (ref (vector-ref subconstraint 0))
     )		       
     (if (trail-content) ; new constraint has length > 0
	 (begin (decrement-and-trail-header (vector-ref subconstraint 1))
		(extract-constraint 1 (vector-ref subconstraint 1) #f)
		(trail-previous) #t)
	 (begin (trail-previous) ; subconstraint is violated
		(if (= 1 (vector-ref (vector-ref subconstraint 1) 0)) ; entire
		    #f                                ; constraint is violated
		    (begin (decrement-and-trail-header
			    (vector-ref subconstraint 1))
			   #t))))
     ; below: unification failed, hence constraint deactivated
     (begin (deactivate-and-trail-header (vector-ref subconstraint 1))
	    #t))) ; term lists not unifiable: no new constraint generated

(define (decrement-and-trail-header c-header)
 (v-trail-next) ; change is trailed on the value trail
 (vector-set! (car v-trail) 0 c-header) ; affected vector
 (vector-set! (car v-trail) 1 0) ; memorizes that index 0 was changed
 (vector-set! (car v-trail) 2 (vector-ref c-header 0)) ; old value stored;
 (vector-set! c-header 0 (- (vector-ref c-header 0) 1))) ; header length - 1

(define (deactivate-and-trail-header c-header)
 (v-trail-next) ; change is trailed on the value trail
 (vector-set! (car v-trail) 0 c-header) ; affected vector
 (vector-set! (car v-trail) 1 1) ; memorizes that index 1 was changed
 (vector-set! (car v-trail) 2 (vector-ref c-header 1)) ; old value stored
 (vector-set! c-header 1 #t)) ; constraint deactivated

;;;*************************************************************************
;; ------------------------
;; VALUE TRAIL MODIFICATION
;; ------------------------

(define (undo-v-trail)
; undos all modifications stored on the value trail back to the previous
; trail-stop, then moves one step back to the previous v-trail position
 (if (v-trail-content) ; some modification has to be undone
     (begin (cond ((vector? (vector-ref (car v-trail) 0)) ; vector 
		   (vector-set! (vector-ref (car v-trail) 0)
				(vector-ref (car v-trail) 1) ; index
				(vector-ref (car v-trail) 2))) ; old value
		  ((eq? 'car (vector-ref (car v-trail) 1)) ; car of cell
		   (set-car! (vector-ref (car v-trail) 0) 
			     (vector-ref (car v-trail) 2))) ; old value
		  (else (set-cdr! (vector-ref (car v-trail) 0) ; cdr of cell
			     (vector-ref (car v-trail) 2)))) ; old value
	    (v-trail-previous)
	    (undo-v-trail))
     (v-trail-previous)))