;;;*************************************************************************
;; -------------------------
;; STRUCTURE COPY PROCEDURES
;; -------------------------
(module copy
	(include "setheo.sch")
	(export (copy-termlist termlist) (copy-term term)
		(copy-constraints constraints) 
		(ref-copy-termlist termlist)
		(make-subgoal-copy subgoal clause-copy) (make-subgoals-copy subgoals generic length cp) 
		(make-clause-copy generic length variables clause constraints))
	(import (globals "globals.scm")
		(successful-inferences globals "globals.scm")
		(current-copy globals "globals.scm")
		(ref unification "unification.scm")
		(unbind-vars unification "unification.scm")))

(define (copy-termlist termlist)
;; computes a copy of termlist with the variables consistently renamed.
;; Since the trail is used for keeping the variable consistency, 
;; - before each call to copy-termlist, one has to execute (trail-stop)
;; - and afterwards one has to execute (unbind-vars)
;; NOTE: termlist has to be free of ref-cells
 (if (null? termlist)
     '()
     (cons (copy-term (car termlist)) (copy-termlist (cdr termlist)))))

(define (copy-term term)
;; computes a copy of term with the variables consistently renamed.
;; Since the trail is used for keeping the variable consistency, 
;; - before each call to copy-term, one has to execute (trail-stop)
;; - and afterwards one has to execute (unbind-vars)
;; NOTE: term has to be free of ref-cells
 (if (non-variable term)
     (cons (car term) (copy-termlist (cdr term)))
     (if (free-variable term)
         (begin ; trick for identification of in which clause variables are:
	   ; each variable in a new clause gets as name the current clause copy
	   ; this way one can identify in the antilemma generation whether the
	   ; variable is relevant
	   (bind term (cons '() (cons #f current-copy)))
;(cons (copy-constraints (var-constraints term)))
	   (car term))
         (car term))))

; Note: for using static constraints, this function has to be implemented
; and integrated into the copy-term function
(define (copy-constraints constraints)
 (if constraints
     (display "copy-constraints not yet implemented")
     constraints))

(define (make-subgoal-copy subgoal clause-copy)
;; computes a variable-renamed copy of a subgoal.
;; Since the trail is used for keeping the variable consistency, 
;; - before each call to make-subgoal-copy, one has to execute (trail-stop)
;; - and afterwards one has to execute (unbind-vars)
 (let ((sg (make-vector 10)))
  (set-subgoal-sign sg (subgoal-sign subgoal))
  (set-subgoal-pred sg (subgoal-pred subgoal))
  (set-subgoal-terms sg (copy-termlist (subgoal-terms subgoal)))
  (set-subgoal-links sg (subgoal-links subgoal))
  (set-subgoal-copy sg clause-copy)
  (set-subgoal-selected sg #f)
  (set-subgoal-next sg #f)
  (set-subgoal-number sg (subgoal-number subgoal))
  (set-subgoal-time-stamp sg 0)
  sg))

;(define (make-eq-subgoal s t)
;;; makes an equal-subgoal with termlist (s t)
; (let ((sg (make-vector 7)))
;  (set-subgoal-sign sg '-)
;  (set-subgoal-pred sg 'equal)
;  (set-subgoal-terms sg (list s t))
;  (set-subgoal-links sg equality-links)
;  (set-subgoal-copy sg clause-copy)

(define (make-subgoals-copy subgoals generic length cp)
;; constructs a variable-renamed copy of the <generic> subgoals of a clause.
;; subgoals is the vector of length k+1 (with k subgoals in the clause)
;; which is destructively modified in the procedure. 
;; The procedure has no output.
;; Since the trail is used for keeping the variable consistency, 
;; - before each call to make-subgoal-copy, one has to execute (trail-stop)
;; - and afterwards one has to execute (unbind-vars)
 (if (>= length 0)
     (begin (vector-set! subgoals length
                         (make-subgoal-copy (vector-ref generic length) cp))
            (make-subgoals-copy subgoals generic (- length 1) cp))))

(define (make-clause-copy generic length variables clause constraints)
;; computes a new variable-renamed <copy> of a clause.
;; The trail is used during the procedure, but restored afterwards 
;; Except for possible trail extensions, the procedure is nondestructive 
 (let ((cp (make-vector 9)) (subgoals (make-vector (+ length 1) #f)))
  (trail-stop)
;  (set-copy-variables cp '())
  (set! current-copy cp)
  (make-subgoals-copy subgoals generic (- length 1) cp)
  (set-copy-subgoals cp subgoals)
  (set-copy-clause cp clause)
  (set-copy-constraints cp constraints)
  (unbind-vars)
  cp))

;(define (make-eq-subgoals copy s t index)
; (if (>= index 0)
;     (begin
;       ())))

(define (ref-copy-termlist termlist)
;; computes a copy of termlist with the variables consistently renamed.
;; Since the list var-assoc is used for keeping the variable consistency, 
;; - before each call to ref-copy-termlist, one has to empty var-assoc
 (if (null? termlist)
     '()
     (cons (ref-copy-term (ref (car termlist)))
	   (ref-copy-termlist (cdr termlist)))))

(define (ref-copy-term term)
;; computes a copy of term with the variables consistently renamed.
 (if (non-variable term)
     (cons (car term) (ref-copy-termlist (cdr term)))
     (if (variable term)
         (let ((head-assoc (memq term var-assoc)))
	   (if head-assoc
	       (cadr head-assoc)
	       (begin
		 (set! var-assoc
		       (cons term (cons (cons '() (cons #f 'V)) var-assoc)))
	         (cadr var-assoc)))))))











