(module index
	(include "setheo.sch")
	(export (find-negmatching-lemma subgoal)
		(in-lemma-index-lemmas lemmas)
		(print-predicate-lemma-indices predicates)
		(comp-in-lemma-index-subgoal subgoal)
		(is-in-infindex-subgoal? subgoal)
		(in-infindex-subgoal subgoal)
		(in-lemma-index-subgoal subgoal))
	(import (globals "globals.scm") (unification "unification.scm")
		(display-subgoal display "display.scm")
		(display-comp-subgoal display "display.scm")
		(ref-int solve "solve.scm")
		(insert-lemma solve "solve.scm")
		(ref-equal pruning "pruning.scm")
		))

;; INDEX 
;; -----
;; Description.
;; An index stores a set of terms in such a way that each term occurs in only
;; one variant. The checking for the existence of a variant and the insertion
;; of a new variant can be done very efficiently.
;;
;; Index data structures.
;; INDEX_VARIABLE ::= POSINT 
;; index variables are identical if their POSINTs are identical
;; INDEX_FUNC ::= #( ATOM ARITY NUMBER )
;; INDEX_HEAD ::= INDEX_VARIABLE | INDEX_FUNC
;; INDEX_END ::= ( INDEX_HEAD )
;; by simultaneous induction:
;; OR_NODE ::= ( AND_NODE . OR_NODES )
;; OR_NODES ::= () | ( OR_NODE . OR_NODES )
;; AND_NODE ::= ( INDEX_HEAD . OR_NODES )
;;;end********************************************************************
;;
;; Definition (WELL-FORMED_OR_NODE).
;; The OR_TREE of an OR_NODE is the labelled tree representation of the
;; list structure of the OR_NODE where INDEX_HEAD and INDEX_END
;; expressions are taken as atomic. This means that the OR_TREE of an 
;; OR_NODE has the following structure: all inner nodes are unlabelled,
;; the leftmost successor node of every inner node is a leaf node labelled 
;; with an INDEX_HEAD expression, and all other leaf nodes are labelled 
;; with INDEX_END expressions.
;; An OR_NODE is well-formed if every INDEX_END i has the following 
;; property: the number of predecessor nodes of i equals the sum of the 
;; ARITIES in the INDEX_FUNC expressions which are at the leftmost brother 
;; nodes of i and its predecessor nodes.

;; Clearly, every INDEX_END of a WELL-FORMED_OR_NODE represents a 
;; well-formed logical term.
;; 
;; INDEX ::= () | ( WELL-FORMED_OR_NODE . INDEX )
;; INDEX_ROOT ::= ( INDEX . NATNUM )
;; NATNUM is the maximal number of distinct variables in an index term 
;;
;; Note that a propositional index is simply #t or #f, depending on whether
;; a propositional literal occurs in the index or not

;; ---------------------------------------------------------
;; PROCEDURES FOR CHECKING WHETHER A TERMLIST IS IN AN INDEX
;; ---------------------------------------------------------

(define (is-in-index-root? termlist term index-root-in)
;(print "is-in-index-root?: ") 
; checks whether a nonempty termlist with first car term is in index;
; the procedure uses the standard binding and trailing mechanisms for
; associating term variables with index variables
 (if (pair? (car index-root-in)) ; index is nonempty
     (begin (set! index (car index-root-in)) (set! nth-variable 0)
	    (set! max-variables (cdr index-root-in)) (trail-stop) 
	    (if (is-in-or-node? termlist term)
		(begin (unbind-vars) #t)
		(begin (unbind-vars) #f)))
     #f))

(define (is-in-or-node? termlist term)
;(print "is-in-or-node?: ") 
 (if (or (integer? term)         ; term is a variable
	 (and (free-variable term) 
              (begin (set! nth-variable (+ 1 nth-variable))
		     (and (<= nth-variable max-variables)
                          (begin (bind term nth-variable)
                                 (set! term nth-variable))))))
     (and (number? (caar index)) ; first entry in index is a variable
	  (is-in-nodes-var? term (caar index)) ; note: term is variable nr 
          (or (eq? '() (cdr termlist)) 
	      (is-in-or-node? (cdr termlist) (ref-int (cadr termlist)))))
     (and (is-in-nodes-func? (car term))
          (or (eq? '() (cdr term))
              (is-in-or-node? (cdr term) (ref-int (cadr term))))
          (or (eq? '() (cdr termlist))
              (is-in-or-node? (cdr termlist) (ref-int (cadr termlist)))))))

(define (is-in-nodes-var? var-number next-var-number)
;(print "is-in-nodes-var?: ") 
 (if (= var-number next-var-number) 
     (set! index (cdar index))
     (and (< var-number next-var-number)
	  (pair? (cdr index)) 
          (number? (caadr index))
	  (begin (set! index (cdr index)) 
                 (is-in-nodes-var? var-number (caar index))))))

(define (is-in-nodes-func? funcsymb)
(display "is-in-nodes-func?: ") 
(print index)
 (if (number? (caar index)) ; top of first OR_NODE is a variable
     (and (pair? (cdr index))
          (begin (set! index (cdr index)) (is-in-nodes-func? funcsymb)))
     (is-in-nodes-func1? (functor-number funcsymb) ; no variable
                         (functor-number (caar index)))))

(define (is-in-nodes-func1? number next-number) 
; auxiliary to is-in-nodes-func1?
(print "is-in-nodes-func1?: ") 
 (if (= number next-number) 
     (set! index (cdar index)) ; funcsymb in index
     (and (< number next-number)
          (pair? (cdr index)) 
	  (begin (set! index (cdr index))
                 (is-in-nodes-func1? number (functor-number (caar index)))))))

;; --------------------------------------------------------------------
;; PROCEDURES FOR CHECKING WHETHER AN INDEX TERMLIST MATCHES A TERMLIST
;; --------------------------------------------------------------------

(define (match-index-root? termlist term index-root-in)
; checks whether a nonempty termlist with car term has a match from
; a termlist in the index;
; the procedures use the trailing mechanisms for keeping track of the
; parts of the input termlist that have been delayed
 (set! in-index? #f)
 (if (pair? (car index-root-in)) ; index is nonempty
     (begin (trail-stop)
	    (match-or-node? (car index-root-in) termlist term)
	    (trail-previous)))
 in-index?)

(define (match-or-node? or-node termlist term)
;(newline) (display "V: ") (display-ref-list termlist)
;(display " vs ") (display (caar or-node))
 (let (tl undo-binding)
   (if (integer? (caar or-node)) ; first entry in index is a variable
       ; index-bindings is vector of bindings of index variables,
       ; initially #(#f ... #f)
       ; undo-binding is position of current binding in index-bindings
       (if (or (and (not (vector-ref index-bindings (caar or-node))) ; free
		    (begin (vector-set! index-bindings (caar or-node) term)
			   (set! undo-binding (caar or-node)) #t)) ; bound    
	       (ref-equal (vector-ref index-bindings (caar or-node)) term));***
	   (begin (cond ((pair? (cdr termlist))
			 (match-or-node? (cdar or-node) (cdr termlist) 
					 (ref (cadr termlist))))
			((trail-content) ; remaining termlist(s) to be checked
			 (set! tl (trail-content)) (trail-previous)
			 (match-or-node? (cdar or-node) tl (ref (car tl)))
			 (push-trail tl)
;			 (newline) (display "VRT: ") 
;			 (display-ref-list (trail-content))
			 ); rearrange trail-pointer
			(else (set! in-index? #t))) ; matching term found
		  (if (integer? undo-binding) 
		      (vector-set! index-bindings undo-binding #f))
		  (if (and (not in-index?) (pair? (cdr or-node))) ; deep back-
		      (match-or-node? (cdr or-node) termlist term))) ; tracking
	   (if (pair? (cdr or-node)) ; shallow backtracking
	       (match-or-node? (cdr or-node) termlist term)))
       (if (non-variable term) ; term must be no variable, too
	   (match-or-node-func? or-node termlist term
				(functor-number (car term))
				(functor-number (caar or-node)))))))

(define (match-or-node-func? or-node termlist term funcnr-term funcnr-next)
;(newline) (display "F: ") (display-ref-list termlist)
;(display " vs ") (display (caar or-node))
(cond ((< funcnr-term funcnr-next)) ; no matching term exists in or-node
      ((> funcnr-term funcnr-next) ; try next function symbol in index
       (if (pair? (cdr or-node))
	   (match-or-node-func? (cdr or-node) termlist term funcnr-term
				(functor-number (caadr or-node)))))
      ; below: top function symbols are equal 
      ((pair? (cdr term)) ; term is complex
       (if (pair? (cdr termlist))
	   (begin (push-trail (cdr termlist))
;	       (newline) (display "IT: ") (display-ref-list (trail-content))
	      (match-or-node? (cdar or-node) (cdr term) (ref (cadr term)))
	      (trail-previous)) ; rearrange trail-pointer
	   (match-or-node? (cdar or-node) (cdr term) (ref (cadr term)))))
      ((pair? (cdr termlist)) 
       (match-or-node? (cdar or-node) (cdr termlist) (ref (cadr termlist))))
      ((trail-content) ; remaining termlist(s) to be checked
       (set! termlist (trail-content)) (trail-previous)
       (match-or-node? (cdar or-node) termlist (ref (car termlist)))
       (push-trail termlist) ; rearrange trail-pointer
;	(newline) (display "FRT: ") (display-ref-list (trail-content))
         )
      (else (set! in-index? #t))))

;; -------------------------------------------------------------------------
;; PROCEDURES FOR CHECKING WHETHER AN INDEX TERMLIST UNIFIES WITH A TERMLIST
;; -------------------------------------------------------------------------

;(define (unify-index-root? termlist term index-root-in)
;; checks whether a nonempty termlist with car term unifies with
;; a termlist in the index;
;; the procedures use the trailing mechanisms for keeping track of the
;; parts of the input termlist that have been delayed and the unif-trail
;; for unifications
; (set! in-index? #f)
; (if (pair? (car index-root-in)) ; index is nonempty
;     (begin (trail-stop)
;	    (unify-or-node? (car index-root-in) termlist term)
;	    (trail-previous)))
; in-index?)

;(define (unify-or-node? or-node termlist term) ; term is referenced 
;(newline) (display "V: ") (display-ref-list termlist)
;(display " vs ") (display (caar or-node))
; (let (tl undo-binding)
;   (if (integer? (caar or-node)) ; first entry in index is a variable
;       ; index-bindings is vector of bindings of index variables,
;       ; initially #(#f ... #f)
;       ; undo-binding is position of current binding in index-bindings
;       (if (or (and (not (vector-ref index-bindings (caar or-node))) ; free
;		    (begin (vector-set! index-bindings (caar or-node) term)
;			   (set! undo-binding (caar or-node)) #t)) ; bound    
;	       (ref-equal (vector-ref index-bindings (caar or-node)) term));***
;	   (begin (unify-or-node-rec or-node termlist term tl)
;		  (if (integer? undo-binding) 
;		      (vector-set! index-bindings undo-binding #f))
;		  (if (and (not in-index?) (pair? (cdr or-node))) ; deep back-
;		      (unify-or-node? (cdr or-node) termlist term))) ; tracking
;	   (if (pair? (cdr or-node)) ; shallow backtracking
;	       (unify-or-node? (cdr or-node) termlist term)))
;       (if (non-variable term) ; term is no variable, too
;	   (unify-or-node-func? or-node termlist term
;				(functor-number (car term))
;				(functor-number (caar or-node)))
;           ; below: term is a variable 
;	   (if (constant (caar or-node))
;               (begin (set-car! term (list (caar or-node))) ; bind term
;		      (unif-trail-next) 
;		      (set-car! (car unif-trail) term)
;		      (unify-or-node-rec or-node termlist term tl)
;		      (unif-trail-previous)
;		      (set-car! term '())) ; unbind term on backtracking
;               (begin (set-car! term (cons (caar or-node)
;				   (make-list (functor-arity (caar or-node)))))
;	     )))))

;(define (unify-or-node-rec or-node termlist term tl)
; (cond ((pair? (cdr termlist))
;	(unify-or-node? (cdar or-node) (cdr termlist) 
;			(ref (cadr termlist))))
;       ((trail-content) ; remaining termlist(s) to be checked
;	(set! tl (trail-content)) (trail-previous)
;	(unify-or-node? (cdar or-node) tl (ref (car tl)))
;	(push-trail tl)
;;	 (newline) (display "VRT: ") 
;;	 (display-ref-list (trail-content))
;	); rearrange trail-pointer
;       (else (set! in-index? #t))) ; unifying term found
;)

;(define (unify-or-node-func? or-node termlist term funcnr-term funcnr-next)
;;(newline) (display "F: ") (display-ref-list termlist)
;;(display " vs ") (display (caar or-node))
;(cond ((< funcnr-term funcnr-next)) ; no unifying term exists in or-node
;      ((> funcnr-term funcnr-next) ; try next function symbol in index
;       (if (pair? (cdr or-node))
;	   (unify-or-node-func? (cdr or-node) termlist term funcnr-term
;				(functor-number (caadr or-node)))))
;      ; below: top function symbols are equal 
;      ((pair? (cdr term)) ; term is complex
;       (if (pair? (cdr termlist))
;	   (begin (push-trail (cdr termlist))
;;	       (newline) (display "IT: ") (display-ref-list (trail-content))
;	      (unify-or-node? (cdar or-node) (cdr term) (ref (cadr term)))
;	      (trail-previous)) ; rearrange trail-pointer
;	   (unify-or-node? (cdar or-node) (cdr term) (ref (cadr term)))))
;      ((pair? (cdr termlist)) 
;       (unify-or-node? (cdar or-node) (cdr termlist) (ref (cadr termlist))))
;      ((trail-content) ; remaining termlist(s) to be checked
;       (set! termlist (trail-content)) (trail-previous)
;       (unify-or-node? (cdar or-node) termlist (ref (car termlist)))
;       (push-trail termlist) ; rearrange trail-pointer
;;	(newline) (display "FRT: ") (display-ref-list (trail-content))
;      )
;      (else (set! in-index? #t))))

;; -------------------------------------------------
;; PROCEDURES FOR INSERTING A TERMLIST INTO AN INDEX
;; -------------------------------------------------

(define (in-index-root termlist term index-root-in) 
; top procedure for inserting a nonempty TERMLIST with car TERM into an
; INDEX-ROOT
 (set! index-root index-root-in) (set! max-variables (cdr index-root))
 (set! nth-variable 0) (set! var-assoc-nr? #f)
 (if (pair? (car index-root)) ; index is nonempty
     (begin (set! in-index? #t) (set! index (car index-root))
            (in-or-node termlist term))
     (begin (set! in-index? #f) ; index-root is empty
        (if (variable term) ; first term is a variable
            (begin (set-car! index-root (list (list (var-assoc-nr term))))
		   (set! index (caar index-root)))
            (begin (set-car! index-root (list (list (car term))))
		   (set! index (caar index-root))
		   (if (pair? (cdr term)) 
		       (in-and-node (cdr term) (ref (cadr term))))))
	(if (pair? (cdr termlist)) 
	    (in-and-node (cdr termlist) (ref (cadr termlist))))))
 (set-cdr! index-root max-variables)
 (if (> max-variables (vector-length index-bindings))
     (set! index-bindings (make-vector (+ 1 max-variables) #f)))
 in-index?)

(define (in-or-node termlist term)
; main procedure for inserting a nonempty TERMLIST with car TERM into an
; INDEX-ROOT
; global variables:
; nth-variable: NATNUM, the number of the last scanned variable;
; max-variables: NATNUM; if a term with more distinct variables as in the
; index is processed, max-variables is set to the new value;
; var-assoc-nr?: #f or list of already seen input variables in termlist; 
; whenever a new variable is encountered in termlist, the list ( variable )
; is destructively attached at the end of var-assoc-nr?;
; var-assoc-nr? must be set to #f before the inital call to in-index; 
; in-index?: #t or #f depending on whether the TERM is in INDEX or not; 
; must be set to #t before the inital call to in-index
; index: is always set to the current OR_NODE or the last inserted AND_NODE;
; SIDE EFFECTS: 
; 1. in-index?, nth-variable, max-variables, var-assoc-nr?, index;
; 2. index-root is possibly modified, by integrating TERMLIST 
 (if (variable term) ; first term is a variable: finds or inserts symbol at
     (begin (var-assoc-nr term) ; the right position in the current index
            (if (number? (caar index)) ; first entry in index is a variable
                (in-nodes-var nth-variable (caar index))
                (begin (set! in-index? #f) ; since variables come first
                       (set-cdr! index (cons (car index) (cdr index))) 
                       (set-car! index (list nth-variable))
                       (set! index (car index)))))
     (begin ; term is no variable: finds or inserts the top function symbol 
            (in-nodes-func (car term))
            (if (pair? (cdr term))
                (if in-index? 
		    (in-or-node (cdr term) (ref (cadr term)));index is: OR_NODE
		    (in-and-node (cdr term) (ref (cadr term))))))) ; : AND_NODE
 (if (pair? (cdr termlist))
     (if in-index? 
         (in-or-node (cdr termlist) (ref (cadr termlist))) ; index is: OR_NODE 
         (in-and-node (cdr termlist) (ref (cadr termlist)))))) ;   : AND_NODE

(define (in-nodes-var var-number next-var-number)
;; checks whether the var-number-th universal variable in a respective
;; term position is in index; 
;; if not, a new OR_NODE is inserted at the right index position
 (cond ((= var-number next-var-number)
	(set! loc-index (car index)) (set! index (cdar index)))
       ((< var-number next-var-number) (set! in-index? #f)
        (set-cdr! index (cons (car index) (cdr index))) 
        (set-car! index (list var-number))
        (set! index (car index)))
       ((and (pair? (cdr index)) 
                    (number? (caadr index))) ; next OR_NODE is a variable
        (set! loc-index index)
	(set! index (cdr index)) (in-nodes-var var-number (caar index)))
       (else (set! in-index? #f) 
             (set-cdr! index (cons (list var-number) (cdr index)))
             (set! index (cadr index)))))

(define (in-nodes-func funcsymb)
; checks whether a function symbol in a respective position is in index;
; at the beginning, the list of variables in the index is skipped;
; depending on whether the function symbol is in the index, a new
; OR_NODE adn AND_NODE is inserted into the index at the right position
; (ie function symbols are inserted in increasing order of their numbers);
; when entering the procedure, the global variable index is an OR_NODE;
; at the end of the computation, index is set to 
; * the next actual OR_NODE       if the function symbol is in the index
; * the last inserted AND_NODE    otherwise
; (newline) (display "INDEX: ") (display index)
; (newline) (display "FUNCSYM: ") (display funcsymb)
 (if (number? (caar index)) ; top of first OR_NODE is a variable
     (if (pair? (cdr index))
         (begin (set! loc-index index) (set! index (cdr index))
		(in-nodes-func funcsymb))
         (begin (set! in-index? #f) (set-cdr! index (list (list funcsymb)))
                (set! index (cadr index))))
     (in-nodes-func1 funcsymb (functor-number funcsymb) ; no variable 
		     (functor-number (caar index)))))
     
(define (in-nodes-func1 funcsymb number next-number) 
; auxiliary to in-nodes-func
 (cond ((= number next-number) (set! loc-index (car index))
	(set! index (cdar index))) ; funcsymb in index
       ((< number next-number) (set! in-index? #f)
        (set-cdr! index (cons (car index) (cdr index))) 
        (set-car! index (list funcsymb))
        (set! index (car index))) ; below: (> number next-number)
       ((pair? (cdr index)) (set! loc-index index)
	(set! index (cdr index)) ; next OR_NODE 
        (in-nodes-func1 funcsymb number (functor-number (caar index))))
       (else (set! in-index? #f) ; last OR_NODE reached
        (set-cdr! index (list (list funcsymb)))
        (set! index (cadr index)))))

(define (in-and-node termlist term)
; (if (= nr-sub;goals 17)-in-nodes-var? var-number 
;     (begin (newline) (display "IN-AND-NODE ENTERED: ")
;	    (display-ref-list termlist)
;	    (display index)))
; inserts nonempty termlist with first term into the current and-node
; when entering the procedure, index must be set to an AND_NODE
 (if (variable term) ; first term is a variable
     (begin (set-cdr! index (list (list (var-assoc-nr term))))
	    (set! index (cadr index))) ; index set to next AND_NODE
     (begin (set-cdr! index (list (list (car term))))
	    (set! index (cadr index)) ; index set to next AND_NODE
            (if (pair? (cdr term)) 
		(in-and-node (cdr term) (ref (cadr term))))))
 (if (pair? (cdr termlist))
     (in-and-node (cdr termlist) (ref (cadr termlist)))))

(define (var-assoc-nr variable)
;; performs association of term variables with index variables
(if var-assoc-nr?
    (begin (set! nth-variable (nth-occ-insert variable var-assoc-nr? 1))
           (if (> nth-variable max-variables) 
	       (set! max-variables nth-variable))
           nth-variable)
    (begin (set! var-assoc-nr? (list variable))
           (if (= max-variables 0) (set! max-variables 1))
           (set! nth-variable 1) 1)))

(define (nth-occ-insert expr list1 position) 
;; INPUT: list1: list of length m > 0
;;        position: NATNUM, initially 1
;; OUTPUT: if expr eq-occurs in list1, the first such position k is given,
;;         else m+1 is outputted and
;;         SIDE EFFECT: expr is destructively attached to the end of list1
 (cond ((eq? expr (car list1)) position)
       ((pair? (cdr list1)) (nth-occ-insert expr (cdr list1) (+ 1 position))) 
       (else (set-cdr! list1 (list expr)) (+ 1 position))))

(define (print-or-node prefix or-node)
 (if (pair? or-node)
     (begin (print-and-node prefix (car or-node))
            (print-or-node prefix (cdr or-node)))))

(define (print-and-node prefix and-node) 
 (set! prefix (cons (car and-node) prefix))
 (if (pair? (cdr and-node))
     (print-or-node prefix (cdr and-node))
     (let ((terms (reverse prefix)))
           (print-tptp-clause terms))))

(define (print-tptp-clause terms)
 (newline) (set! variable-names '())
 (display "input_clause(lemma,unknown,[")
 (display (car terms))
 (print-term (cdr terms) '()))
 
(define (print-term terms aux-list)
 (if (pair? terms)
     (begin
       (if (vector? (car terms))
	   (display (functor-symbol (car terms)))
	   (begin (display "X") (display (car terms))))
       (if (and (vector? (car terms)) (> (functor-arity (car terms)) 0))
           (begin (display "(")
		  (print-term (cdr terms)
			      (cons (functor-arity (car terms)) aux-list)))
	   (begin (set! aux-list (close-brackets aux-list))
		  (if (pair? aux-list)
		      (begin (display ",")
			     (set-car! aux-list (- (car aux-list) 1))
			     (print-term (cdr terms) aux-list))
		      (print-term (cdr terms) aux-list)))))
     (display "]).")))

(define (close-brackets aux-list)
 (if (and (pair? aux-list) (= 1 (car aux-list)))
     (begin (display ")") (close-brackets (cdr aux-list)))
     aux-list))

;(define (print-or-node prefix or-node)
; (if (pair? or-node)
;     (begin (print-and-node prefix (car or-node))
;            (print-or-node prefix (cdr or-node)))))

;(define (print-and-node prefix and-node) 
; (if (vector? (car and-node))
;     (set! prefix (cons (functor-symbol (car and-node)) prefix))
;     (set! prefix 
;           (cons (string-append "X" (number->string (car and-node))) prefix)))
; (if (pair? (cdr and-node))
;     (print-or-node prefix (cdr and-node))
;     (begin (newline) (display (print-list (reverse prefix))))))

(define (print-list list1)
 (if (pair? list1)
     (begin (display (car list1)) (display " ") (print-list (cdr list1)))
     '#\ ))

; Examples for testing

;(define index-root-in (cons '() 0))
 
;(set! filename "test1.p") (read-tptp filename)
;(begin (sort-clauses (formula-clauses formula)) 'done)
;(make-connection-graph (formula-clauses formula) (formula-clauses formula))
;(define generic1 (clause-generic (car (formula-clauses formula))))
;(define s0 (vector-ref generic1 0)) (define t0 (subgoal-terms s0))
;(define s1 (vector-ref generic1 1)) (define t1 (subgoal-terms s1))
;(define s2 (vector-ref generic1 2)) (define t2 (subgoal-terms s2))
;(define s3 (vector-ref generic1 3)) (define t3 (subgoal-terms s3))
;(define s4 (vector-ref generic1 4)) (define t4 (subgoal-terms s4))
;(define s5 (vector-ref generic1 5)) (define t5 (subgoal-terms s5))
;(define s6 (vector-ref generic1 6)) (define t6 (subgoal-terms s6))
;(define s7 (vector-ref generic1 7)) (define t7 (subgoal-terms s7))
;(define s8 (vector-ref generic1 8)) (define t8 (subgoal-terms s8))

;(define (inindex termlist)
; (in-index-root termlist (ref (car termlist)) index-root-in))

;(define (inindex? termlist)
; (is-in-index-root? termlist (ref (car termlist)) index-root-in))

;(define (matchindex? termlist)
; (match-index-root? termlist (ref (car termlist)) index-root-in))

;(define (unifyindex? termlist)
; (unify-index-root? termlist (ref (car termlist)) index-root-in))
;
;(define (printindex) (print-or-node '() (car index-root-in)))

(define (in-lemma-index-lemmas lemmas)
 (if (pair? lemmas)
     (begin 
       (if (= 1 (clause-length (car lemmas)))
	   (in-lemma-index-subgoal
	             (vector-ref (clause-generic (car lemmas)) 0)))
       (in-lemma-index-lemmas (cdr lemmas)))))

(define (is-in-infindex-subgoal? subgoal)
; checks whether subgoal is in infindex
 (if (> (predsym-arity (subgoal-pred subgoal)) 0)
     (if (eq? '+ (subgoal-sign subgoal))
         (is-in-index-root? (subgoal-terms subgoal)
			(ref (car (subgoal-terms subgoal)))
			(predsym-pos-infindex (subgoal-pred subgoal)))
	 (is-in-index-root? (subgoal-terms subgoal)
			(ref (car (subgoal-terms subgoal)))
			(predsym-neg-infindex (subgoal-pred subgoal))))
     #f))

(define (in-infindex-subgoal subgoal)
; insert termlist of subgoal into respective infindex
 (if (> (predsym-arity (subgoal-pred subgoal)) 0)
     (begin
       (if (eq? '+ (subgoal-sign subgoal))
	   (in-index-root (subgoal-terms subgoal)
			  (ref (car (subgoal-terms subgoal)))
			  (predsym-pos-infindex (subgoal-pred subgoal)))
	   (in-index-root (subgoal-terms subgoal)
			  (ref (car (subgoal-terms subgoal)))
			  (predsym-neg-infindex (subgoal-pred subgoal))))
;       (display " index-end: ")
       (if (not (pair? index)) (set! index loc-index))
;       (print index)
;       (display "in-index?: ") (print in-index?)
       )
     #f))

(define (in-lemma-index-subgoal subgoal)
 (if (> (predsym-arity (subgoal-pred subgoal)) 0)
     (if (eq? '+ (subgoal-sign subgoal))
	 (in-index-root (subgoal-terms subgoal)
			(ref (car (subgoal-terms subgoal)))
			(predsym-pos-index (subgoal-pred subgoal)))
	 (in-index-root (subgoal-terms subgoal)
			(ref (car (subgoal-terms subgoal)))
			(predsym-neg-index (subgoal-pred subgoal))))
     (if (eq? '+ (subgoal-sign subgoal))
	 (begin (if (predsym-pos-index (subgoal-pred subgoal))
		    (set! in-index? #t) (set! in-index? #f))
		(set-predsym-pos-index (subgoal-pred subgoal) #t))
	 (begin (if (predsym-neg-index (subgoal-pred subgoal))
		    (set! in-index? #t) (set! in-index? #f))
		(set-predsym-neg-index (subgoal-pred subgoal) #t))))
 (if (not in-index?)
     (begin (set! nr-lemmas (+ 1 nr-lemmas))
	    (if lemmas (insert-lemma subgoal #t))
	    (if printindex 
		(begin (newline)
	;	       (display "Type of lemma ") (display nr-lemmas)
	;	       (display " is: ") (display lemmatype) (newline)
		       (display "input_clause(l")
		       (display nr-lemmas)
		       (if (= 0 lemmatype)
			   (display ",conjecture,[")
			   (display ",axiom,["))
		       (display-subgoal subgoal)
		       (display "]).")
		       (newline)
		       )))))

(define (comp-in-lemma-index-subgoal subgoal)
 (if (> (predsym-arity (subgoal-pred subgoal)) 0)
     (if (eq? '- (subgoal-sign subgoal))
	 (in-index-root (subgoal-terms subgoal)
			(ref (car (subgoal-terms subgoal)))
			(predsym-pos-index (subgoal-pred subgoal)))
	 (in-index-root (subgoal-terms subgoal)
			(ref (car (subgoal-terms subgoal)))
			(predsym-neg-index (subgoal-pred subgoal))))
     (if (eq? '- (subgoal-sign subgoal))
	 (begin (if (predsym-pos-index (subgoal-pred subgoal))
		    (set! in-index? #t) (set! in-index? #f))
		(set-predsym-pos-index (subgoal-pred subgoal) #t))
	 (begin (if (predsym-neg-index (subgoal-pred subgoal))
		    (set! in-index? #t) (set! in-index? #f))
		(set-predsym-neg-index (subgoal-pred subgoal) #t))))
 (if (not in-index?)
     (begin (set! nr-lemmas (+ 1 nr-lemmas))
	    (if lemmas (insert-lemma subgoal #f))
	    (if printindex
		(begin (newline)
	;	       (display "Type of lemma ") (display nr-lemmas)
	;	       (display " is: ") (display lemmatype) (newline)
		       (display "input_clause(c")
		       (display nr-lemmas)
		       (display ",axiom,[")
		       (display-comp-subgoal subgoal)
		       (display "]).")
		       (newline)
		       )))))

(define (print-predicate-lemma-indices predicates)
 (if (pair? predicates)
     (begin
;           (newline) (newline) 
;	    (display "% + ") (display (predsym-symbol (car predicates)))
;	    (display ": ---------------------------")
	    (print-or-node (list (car predicates) '++)
			   (car (predsym-pos-index (car predicates))))
;	    (newline) (newline) 
;	    (display "% - ") (display (predsym-symbol (car predicates)))
;	    (display ": ---------------------------")
	    (set! variable-names '())
	    (print-or-node (list (car predicates) '--)
			   (car (predsym-neg-index (car predicates))))
	    (print-predicate-lemma-indices (cdr predicates)))))

; (define (find-matching-lemma subgoal)
;  (let (old-sign result)
;    (set! old-sign (subgoal-sign subgoal))
;    (if (eq? '- old-sign)
;        (set-subgoal-sign subgoal '+)
;        (set-subgoal-sign subgoal '-))
;    (set! result (find-negmatching-lemma subgoal))
;    (set-subgoal-sign subgoal old-sign)
;    result))

(define (find-negmatching-lemma subgoal)
 (if (> (predsym-arity (subgoal-pred subgoal)) 0)
     (if (eq? '- (subgoal-sign subgoal))
	 (match-index-root? (subgoal-terms subgoal)
			    (ref (car (subgoal-terms subgoal)))
			    (predsym-pos-index (subgoal-pred subgoal)))
	 (match-index-root? (subgoal-terms subgoal)
			    (ref (car (subgoal-terms subgoal)))
			    (predsym-neg-index (subgoal-pred subgoal))))
     (if (eq? '- (subgoal-sign subgoal))
	 (predsym-pos-index (subgoal-pred subgoal))
	 (predsym-neg-index (subgoal-pred subgoal)))))


(define (make-list length)
 (if (<= length 0)
     '()
     (cons #unspecified (make-list (- length 1)))))


