;;;*************************************************************************
;; ---------------------
;; UNIFICATION PROCEDURE
;; ---------------------
(module unification
	(include "setheo.sch")
	(export (unbind-vars) (ref term) (unify-lists k l) (unify-terms s t)
		(d-unify-lists k l d-pairs) 
		(d-unify-terms s t d-pairs) 
		(d-unify-eq-terms s t d-pairs) 
		(subgoals->eq-subgoals subgoals d-pairs clause-copy sgs)
		(extension-unify-lists k l) 
		(unbind-collect-vars variables))
	(import (globals "globals.scm")
		(display-subgoal display "display.scm")
		(ref-equal pruning "pruning.scm")))

(define (ref term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
 (if (or (non-variable term) (free-variable term))
     term
     (ref (car term))))

(define (unify-lists k l)
;; unifies termlists k and l.
;; If unification succeeds, #t is returned, else #f.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     #t
     (if (unify-terms (ref (car k)) (ref (car l)))
	 (unify-lists (cdr k) (cdr l))
	 #f)))

(define (unify-terms s t)          
;; unifies terms s and t. 
;; If unification succeeds, #t is returned, else #f.
;; Side effect: terms and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
         (if (variable t) 
             (bind s t)
	     (if (occurs-list s (cdr t))
                 (unbind s t)
  	         (bind s t)))
         (if (variable t)
             (if (occurs-list t (cdr s))
	         (unbind t s)
                 (bind t s))
             (if (eq? (car s) (car t))
	         (unify-lists (cdr s) (cdr t))
                 (unbind s t))))))

(define (occurs-list var termlist)              
;; returns #t if var occurs in termlist, and #f otherwise
 (if (null? termlist)
     #f
     (if (occurs-term var (ref (car termlist)))
         #t
         (occurs-list var (cdr termlist)))))

(define (occurs-term var term)       
;; returns #t if var occurs in term, and #f otherwise
 (if (eq? var term) 
     #t
     (and (non-variable term) (occurs-list var (cdr term)))))

;;;END *********************************************************************

;;;*************************************************************************
;; -------------------------------
;; OPTIMIZED UNIFICATION PROCEDURE
;; -------------------------------

;; For extension steps it is known that initially the two termlists are
;; variable disjoint. Under these conditions the unification can be 
;; optimized such that certain occurs-checks can be omitted, as follows.
;; We assume that for bindings always the first argument position (the k 
;; or s-position, i.e. the entry literal) is preferred; also, we suppose 
;; that no ref-cells are in the terms at the first argument position. 
;; 
;; Initially, the fully optimized procedures extension-unify-* are called.
;; As long as no binding is performed, we remain in this routines.
;;     THE REST IS TO BE CORRECTED ...
;; If the first binding is performed, say, for variable x, then
;; 1. if x is on the second argument position (l or t-position, i.e. from 
;;    the calling subgoal), then we move to standard unification for the 
;;    rest of the unification;
;; 2. otherwise (x must be a free variable) we move to the semi-optimized
;;    procedures semi-extension-* which work as follows:
;;    only free variables at the first argument positions may be bound 
;;    without performing occurs-checks; whenever a variable at the second 
;;    argument position is bound, we move to standard unification.

(define (extension-unify-lists k l)
;; like unify-lists, but fully optimized for extension steps;
;; it is assumed that no binding has been performed yet
 (if 
  (null? k)
  #t
  (if (extension-unify-terms (car k) (ref (car l)))
      (cond ((eq? optimized-unification 0) 
             (extension-unify-lists (cdr k) (cdr l)))
            ((eq? optimized-unification 1) 
             (semi-extension-unify-lists (cdr k) (cdr l)))
	    (else (unify-lists (cdr k) (cdr l))))
      #f)))

(define (extension-unify-terms s t)          
;; like unify-terms, but fully optimized for extension steps;
;; it is assumed that no binding has been performed yet
 (if (variable s)
     (begin (set! optimized-unification 1) (bind s t))
     (if (variable t)
         (begin (set! optimized-unification #f) (bind t s))
         (if (eq? (car s) (car t))
	     (extension-unify-lists (cdr s) (cdr t))
             (unbind s t)))))

(define (semi-extension-unify-lists k l)
;; like unify-lists, but semi-optimized for extension steps;
;; it is assumed that only bindings for variables at the first argument
;; positions have been performed yet
 (cond ((null? k) #t)
       ((free-variable (car k)) (bind (car k) (ref (car l)))
        (semi-extension-unify-lists (cdr k) (cdr l)))
       ((variable (car k)) 
        (if (unify-terms (ref (car k)) (ref (car l)))
            (semi-extension-unify-lists (cdr k) (cdr l))
            #f))
       ((semi-extension-unify-terms (car k) (ref (car l)))
        (if (eq? 1 optimized-unification)
            (semi-extension-unify-lists (cdr k) (cdr l))
            (unify-lists (cdr k) (cdr l))))
       (else #f)))

(define (semi-extension-unify-terms s t)          
;; like unify-terms, but semi-optimized for extension steps;
;; it is assumed that only bindings for variables at the first argument
;; positions have been performed yet
 (if (variable t)
     (if (occurs-list t (cdr s))
	 (unbind t s)
         (begin (set! optimized-unification #f) (bind t s)))
     (if (eq? (car s) (car t))
	 (semi-extension-unify-lists (cdr s) (cdr t))
         (unbind s t))))

;;;END**********************************************************************

;;;*************************************************************************
;; -----------------------------------
;; DISAGREEMENT UNIFICATION PROCEDURES
;; -----------------------------------

(define (d-unify-lists k l d-pairs)
;; Disagreement unifies termlists k and l.
;; unification always succeeds, list of disagreement pairs is returned
;; Side effect: termlists and trail destructively modified.
;; Returns the list of disagreement pairs.
;; Eager variable elimination for variable/variable pairs
;; and elimination of trivial pairs is included.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     d-pairs
     (begin (set! d-pairs (append! (d-unify-terms (ref (car k)) (ref (car l)) '())
				   d-pairs))
	    (d-unify-lists (cdr k) (cdr l) d-pairs))))

(define (d-unify-terms s t d-pairs)          
;; Disagreement unifies variable-disjoint terms s and t. 
;; Side effect: terms and trail destructively modified.
;; Returns the singleton list containing the new disagreement pair 
;; or otherwise '()
 (if (not (ref-equal s t))
     (if (and (variable s)
	      (or (variable t)
		  (and eager-vel (not (occurs-list s (cdr t))))))
	 (bind s t) ; variable elimination
         (if (and eager-vel (variable t) (not (occurs-list t (cdr s))))
	     (bind t s) ; variable elimination
             (set! d-pairs (cons (list t s) d-pairs)))))
 d-pairs)

(define (d-unify-eq-terms s t d-pairs)   
;; Disagreement unifies variable-disjoint terms s and t with s from
;; a ~eq-subgoal
;; Side effect: terms and trail destructively modified.
;; Returns the list of disagreement pairs or otherwise #f when
;; s and t are not root-unifiable
;; Eager variable elimination and elimination of trivial pairs is included.
 (if (variable s) ; occurs check not needed, since first occurrence of s
     (bind s t) ; forced substitution
     (if (variable t) ; occurs check not needed, since first occurrence of t
	 (bind t s)
         (if (eq? (car s) (car t))
	     (set! d-pairs (d-unify-lists (cdr s) (cdr t) '()))
	     (set! d-pairs #f))))
 d-pairs)

; (define (root-unifiable s t)
; ;; Checks whether two variable-disjoint terms are root-unifiable
;  (or (variable s) (variable t) (eq? (car s) (car t))))

(define (subgoals->eq-subgoals subgoals d-pairs clause-copy sgs)
 (set! sgs (get-subgoals (+ (vector-length subgoals) (length d-pairs)) #f))
 (fill-vector subgoals sgs (- (vector-length subgoals) 2))
 (insert-neq-subgoals d-pairs
		      (- (vector-length subgoals) 1) sgs clause-copy #f)
 sgs)

(define (insert-neq-subgoals d-pairs index sgs clause-copy sg)
;; Transfers d-pairs into the standard subgoals format 
 (set! sg (get-eq-subgoal #f))
 (set-subgoal-terms sg (car d-pairs))
 (set-subgoal-copy sg clause-copy)
 (set-subgoal-next sg #f)
 (set-subgoal-number sg index)
; (set-subgoal-foldup-predecessor sg 'fu)
 (set-subgoal-time-stamp sg 0)
 (vector-set! sgs index sg)
; (display "New ~eq-Subgoal: ")  (display-subgoal sg) (newline)
 (set! d-pairs (cdr d-pairs))
 (if (pair? d-pairs)
     (insert-neq-subgoals d-pairs (+ index 1) sgs clause-copy #f)))

(define (fill-vector vec1 vec2 index)
;; fills vec2 with contents of vec1 until index position
 (if (>= index 0)
     (begin (vector-set! vec2 index (vector-ref vec1 index))
	    (fill-vector vec1 vec2 (- index 1)))))

(define (get-subgoals length sgs)
;; Gets a new list of subgoals in standard format; 
;; length should be the number of subgoals in the clause plus 1;
;; sgs may be anything, since it is only used as a local variable
 (if (eq? '() (vector-ref subgoals-stack length))
     (set! sgs (make-vector length #f))
     (begin (set! sgs (car (vector-ref subgoals-stack length)))
	    (vector-set! subgoals-stack length
			 (cdr (vector-ref subgoals-stack length)))))
 sgs)

(define (insert-vectors vec length index)
 (if (>= index 0)
     (begin (vector-set! vec index (make-vector length))
	    (insert-vectors vec length (- index 1)))))

(define (free-subgoals length subgoals)
;; When a list of subgoals is no more needed on backtracking, it may be freed
;; by putting it on the subgoals-stack at the respective length position
 (vector-set! subgoals-stack length
	      (cons subgoals (vector-ref subgoals-stack length))))

(define (get-eq-subgoal sg)
;; Gets a new eq-subgoal in standard format for subgoals; 
;; sg may be anything, since it is only used as a local variable
 (if (eq? '() eq-subgoal-stack)
     (begin (set! sg (make-vector 10 #f))
	    (set-subgoal-sign sg '-)
	    (set-subgoal-pred sg eq-predicate)
	    (set-subgoal-links sg eq-links))
     (begin (set! sg (car eq-subgoal-stack))
	    (set! eq-subgoal-stack (cdr eq-subgoal-stack))))
 sg)

(define (free-eq-subgoal sg)
;; When an eq-subgoal is no more needed on backtracking, it may be freed
;; by putting it on the eq-subgoal-stack
  (set! eq-subgoal-stack (cons sg eq-subgoal-stack)))
	    
;;;END**********************************************************************

;;;*************************************************************************
;; -----------------------------
;; TRAIL MODIFICATION PROCEDURES

(define (unbind-vars)
;; makes free all variables back to the previous trailstop
 (if (trail-content)
     (begin (set-car! (caar trail) '()) (trail-previous) (unbind-vars))
     (begin (trail-previous) #f)))

(define (unbind-collect-vars variables)
;; makes free and collects all variables back to the previous trailstop
 (if (trail-content)
     (begin (set-car! (caar trail) '()) 
	    (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (unbind-collect-vars variables))
     (begin (trail-previous) variables)))

(define (last-binding-variables)
 (let ((old-trail trail) variables)
      (set! variables (last-binding-vars '()))
      (set! trail old-trail)
      variables))

(define (last-binding-vars variables)
 (if (trail-content)
     (begin (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (last-binding-vars variables))
     variables))

; (define (make-eq-clause)
;  (let ((var1 (cons '() (cons #f 'x))) (var2 (cons '() (cons #f 'y))) sg)
;  (set! eq-clause (make-vector 12))
;  (set! sg (make-vector 10) #f)
;  (set-subgoal-sign sg '+)
;  (set-subgoal-pred sg eq-predicate)
;  (set-subgoal-terms sg (list var1 var2))
;  (set-subgoal-links sg '())
;  (set-subgoal-copy sg '())
; ; (set-subgoal-selected sg #f)
; ; (set-subgoal-next sg #f)   
;  (set-subgoal-number sg 0)
; ; (set-subgoal-foldup-predecessor sg #f)
; ; (set-subgoal-time-stamp sg 0)
;  (set! clause-generic eq-clause (vector sg #f))
;  (set-clause-copies eq-clause '())
;  (set-clause-generic eq-clause '())
;  (set-clause-length eq-clause 1)
;  (set-clause-nr-variables eq-clause 2)
;  (set-clause-variables eq-clause '(list var1 var2))
;  (set-clause-constraints eq-clause '())
;  (set-clause-type eq-clause 'axiom)
; ; (set-clause-number eq-clause #f)
;  (set-clause-name eq-clause 'eq-clause)
;  (set-clause-all-copies eq-clause '())
;  (set-clause-complexity eq-clause 0)
;  (set-clause-signature eq-clause '())
; ))
 



