;;;*************************************************************************
;;;module name: semprop.sch
;;;author     : Reinhold Letz, TU Munich
;;;date       : 10.08.00
;;;description: propositional decision procedure using the semantic tree 
;;;		calculus (DPLL-kernel)
;;;remarks    :
;;;compiling  : yes
;;;*************************************************************************

;; -------------------------------------------------------------------------
;; internal data structures:
;; 
;; clause: vector of four/six elements
;; 1st element: atom which activates the clause [#f or atom]
;; 2nd element: current clause length [positive integer]
;; 3rd element: list of positive atoms
;; 4th element: list of negative atoms
;; 5th element: context relevance vector for lemmas,
;;              5th element not existing for input clauses
;; 6th element: context relevance vector depth/30 for lemmas,
;;              6th element not existing for input clauses
;;
;; atom: vector of ... elements
;; 1st element: clauses with positive occurrence of atom
;; 2nd element: clauses with negative occurrence of atom
;; 3rd element: last depth at which atom was evaluated
;; 4th element: positive atom frequency in formula
;; 5th element: negative atom frequency in formula
;; 6th element: atom name
;; 7th element: atom sign
;; 8th element: list of positive neighbors
;; 9th element: list of negative neighbors
;;10th element: list of context unit clauses
;;11th element: last depth mod 30 at which atom was evaluated
;;12th element: list of positive context non-unit clauses
;;13th element: list of negative context non-unit clauses
;;14th element: #f if atom is evaluated and on right branch
;;15th element: list of context lemmas to be deleted when backtracking atom
;;16th element: #f if initial freq of pos atom is less or equal than neg atom
;; -------------------------------------------------------------------------

;; macros for accessing and modifying basic data structures
;;
(define-macro (c-activating-atom clause) `(vector-ref ,clause 0))
(define-macro (c-length clause)          `(vector-ref ,clause 1))
(define-macro (c-pos-literals clause)    `(vector-ref ,clause 2))
(define-macro (c-neg-literals clause)    `(vector-ref ,clause 3))
(define-macro (c-rel-vector clause)      `(vector-ref ,clause 4))
(define-macro (c-rel-depth30 clause)     `(vector-ref ,clause 5))
(define-macro (c-usages clause)          `(vector-ref ,clause 6))

;(define-macro (c-number clause)          `(vector-ref ,clause 4))
;(define-macro (c-unit-atom-index clause) `(vector-ref ,clause 5))

(define-macro (set-c-activating-atom clause val) `(vector-set! ,clause 0 ,val))
(define-macro (set-c-length clause val)          `(vector-set! ,clause 1 ,val))
(define-macro (set-c-pos-literals clause val)    `(vector-set! ,clause 2 ,val))
(define-macro (set-c-neg-literals clause val)    `(vector-set! ,clause 3 ,val))
(define-macro (set-c-rel-vector clause val)      `(vector-set! ,clause 4 ,val))
(define-macro (set-c-rel-depth30 clause val)     `(vector-set! ,clause 5 ,val))
(define-macro (set-c-usages clause val)          `(vector-set! ,clause 6 ,val))

;(define-macro (set-c-number clause val)          `(vector-set! ,clause 4 ,val))
;(define-macro (set-c-unit-atom-index clause val) `(vector-set! ,clause 5 ,val))

(define-macro (a-pos-clauses atom)       `(vector-ref ,atom 0))
(define-macro (a-neg-clauses atom)       `(vector-ref ,atom 1))
(define-macro (a-last-depth atom)        `(vector-ref ,atom 2))
(define-macro (a-pos-frequency atom)     `(vector-ref ,atom 3))
(define-macro (a-neg-frequency atom)     `(vector-ref ,atom 4))
(define-macro (a-name atom)              `(vector-ref ,atom 5))
(define-macro (a-sign atom)              `(vector-ref ,atom 6))
(define-macro (a-pos-neighbors atom)     `(vector-ref ,atom 7))
(define-macro (a-neg-neighbors atom)     `(vector-ref ,atom 8))
(define-macro (a-unit-lemmas atom)       `(vector-ref ,atom 9))
(define-macro (a-last-depth-mod30exp atom)  `(vector-ref ,atom 10))
(define-macro (a-pos-lemmas atom)        `(vector-ref ,atom 11))
(define-macro (a-neg-lemmas atom)        `(vector-ref ,atom 12))
(define-macro (a-left-branch atom)       `(vector-ref ,atom 13))
(define-macro (a-lemmas-out atom)        `(vector-ref ,atom 14))
(define-macro (a-neg-less-pos atom)      `(vector-ref ,atom 15))
(define-macro (a-pos-occ atom)           `(vector-ref ,atom 16))
(define-macro (a-neg-occ atom)           `(vector-ref ,atom 17))

(define-macro (a-last-depth30 atom) `(vector-ref vector30 (a-last-depth ,atom)))

(define-macro (a-realname atom)
`(if general-input
     (vector-ref hash-table (a-name ,atom))
     (a-name ,atom)))

(define-macro (set-a-pos-clauses atom val)      `(vector-set! ,atom 0 ,val))
(define-macro (set-a-neg-clauses atom val)      `(vector-set! ,atom 1 ,val))
(define-macro (set-a-last-depth atom val)       `(vector-set! ,atom 2 ,val))
(define-macro (set-a-pos-frequency atom val)    `(vector-set! ,atom 3 ,val))
(define-macro (set-a-neg-frequency atom val)    `(vector-set! ,atom 4 ,val))
(define-macro (set-a-name atom val)             `(vector-set! ,atom 5 ,val))
(define-macro (set-a-sign atom val)             `(vector-set! ,atom 6 ,val))
(define-macro (set-a-pos-neighbors atom val)    `(vector-set! ,atom 7 ,val))
(define-macro (set-a-neg-neighbors atom val)    `(vector-set! ,atom 8 ,val))
(define-macro (set-a-unit-lemmas atom val)      `(vector-set! ,atom 9 ,val))
(define-macro (set-a-last-depth-mod30exp atom val)
                                                `(vector-set! ,atom 10 ,val))
(define-macro (set-a-pos-lemmas atom val)       `(vector-set! ,atom 11 ,val))
(define-macro (set-a-neg-lemmas atom val)       `(vector-set! ,atom 12 ,val))
(define-macro (set-a-left-branch atom val)      `(vector-set! ,atom 13 ,val))
(define-macro (set-a-lemmas-out atom val)       `(vector-set! ,atom 14 ,val))
(define-macro (set-a-neg-less-pos atom val)     `(vector-set! ,atom 15 ,val))
(define-macro (set-a-pos-occ atom val)          `(vector-set! ,atom 16 ,val))
(define-macro (set-a-neg-occ atom val)          `(vector-set! ,atom 17 ,val))

(define-macro (context-lemma? clause)
`(and (> (vector-length ,clause) 4) (c-rel-vector ,clause)))

(define-macro (input-clause? clause) `(= (vector-length ,clause) 4))

(define-macro (lemma? clause) `(not (input-clause? ,clause)))

(define-macro (is-pure? atom) `(or (= 0 (a-pos-occ ,atom)) (= 0 (a-neg-occ ,atom))))

(define-macro (make-atom atom-name)
  `(vector '() '() 0 0 0 ,atom-name #f '() '() '() 0 '() '() #f '() #t 0 0))

(define-macro (select-sign atom branch)
`(if (and (> random-branch 0) 
          (or (< (quotient (a-pos-frequency ,atom)
			   (+ (a-neg-frequency ,atom) 1))
		 freq-divisor)
              (< (quotient (a-neg-frequency ,atom)
			   (+ (a-pos-frequency ,atom) 1))
		 freq-divisor))
          (> random-percentage
             (modulo (modulo random-branch (+ inference 1)) 100)))
     (if (= 0 (modulo (modulo random-branch (+ inference 1)) 2))
         (set! ,branch #f)
         (set! ,branch #t))
     (if neg-first
	 (set! ,branch #f)
	 (if pos-first
	     (set! ,branch #t)
	     (if init-sign
		 (set! ,branch (a-neg-less-pos ,atom))
		 (if occ
		     (if (<= (a-pos-occ ,atom) (a-neg-occ ,atom))
			 (set! ,branch #f)
			 (set! ,branch #t))
		     (if (<= (a-pos-frequency ,atom) (a-neg-frequency ,atom))
			 (set! ,branch #f)
			 (set! ,branch #t))))))))
 
(define-macro (hash-insert! obj) `(hash-insert-key! (hash-key ,obj) ,obj))

(define-macro (hash-member? obj) `(hash-member-key? (hash-key ,obj) ,obj))

(define-macro (hashify-clause term) 
`(hashify-poslits (car ,term) (cadr ,term)))

(define-macro (max-atom atoms)
`(let (m-pos m-neg)
  (set! m-pos (a-pos-frequency (car ,atoms))) 
  (set! m-neg (a-neg-frequency (car ,atoms)))
  (if max-frequency
      (if (< m-pos m-neg) 
	  (max-atom-atoms (cdr ,atoms) ,atoms m-neg m-pos)
	  (max-atom-atoms (cdr ,atoms) ,atoms m-pos m-neg))
      (if (< m-pos m-neg)
	  (max-atom-atoms (cdr ,atoms) ,atoms (+ m-pos m-neg) m-neg)
	  (max-atom-atoms (cdr ,atoms) ,atoms (+ m-pos m-neg) m-pos)))))

(define-macro (atom-irrelevant? rel-vector atom)
`(= 0 (bit-and (vector-ref ,rel-vector (a-last-depth30 ,atom))
	       (a-last-depth-mod30exp ,atom))))

(define-macro (extend-plits plits atom)
`(set! ,plits (cons ,atom ,plits)))

(define-macro (extend-nlits nlits atom)
`(set! ,nlits (cons ,atom ,nlits)))

