(module semprop
	(include "semprop.sch")
	(import
		(globals "globals.scm")
	        (evaluate "evaluate.scm")
		(delete-lemmas lemmas "lemmas.scm")
		)
	(export
		(atom->= m n)
		(display-atom atom)
		(display-atom-names atoms)
		(display-clause clause)
		(display-clauses clauses)
		(display-path-index index)
		(model)
		(order-atoms)
	        (semprop n)
		))
 
;;;*************************************************************************
;;;module name: semprop.scm
;;;author     : Reinhold Letz, TU Munich
;;;date       : 25.01.2001
;;;description: propositional decision procedure using the semantic tree 
;;;		calculus (DPLL) plus
;;;		*) context lemmata and
;;;		*) dependency-directed backtracking
;;;remarks    : input is a sequence of clauses;
;;;             there are three input formats:
;;;             1) an input clause must be a list of the form 
;;;                ( positive_literals negative_literals ) 
;;;                where positive_literals and negative_literals are lists
;;;                of arbitrary scheme symbols or strings
;;;             2) or, a clause is a list of integers different from 0,
;;;                negation is expressed with arithmetic negation
;;;             3) or, clauses must be in dimacs format
;;;             the input formats must not be mixed
;;;bugs       : clause length is limited to 1000
;;;*************************************************************************

(define (semprop args)
 (input-formula args)
 (if lemmas-on 
     (begin (initial-lemmas2clauses atom-list)
            (if (= nr-lemmas 0) (set! nr-lemmas 50)) ;(quotient clause-nr 10)
	    (if (= 0 max-lemma-length) (set! max-lemma-length 20))
	    (set! foldup-stack (make-vector nr-atoms))
	    ))
 (set! path-vector (make-vector nr-atoms #f))
 (set! nr-atoms30 (quotient nr-atoms 30))
 (set! vector30 (make-vector (+ 1 nr-atoms)))
 (set! vector-mod30exp (make-vector (+ 1 nr-atoms)))
 (vector30in vector30 vector-mod30exp nr-atoms 30)
 (set! inference 0) (set! lemma-diff 0)
 (compute-frequencies atom-list) 
 (compute-occurrences atom-list) ; NOTE: must come before purity check
 (if purity (extract-pure atom-list))
 (set! nr-open-clauses clause-nr)
 (if statistics
     (begin (display "Atoms") (if purity (display " (pure)"))
	    (display ": ") (display nr-atoms)
	    (if purity
		(begin (display " (") (display (+ 1 purity-index)) (display ")")))
	    (display ", clauses (units): ") (display clause-nr)
	    (display " (") (display (+ 1 unit-index)) (display ")")
	    (display ", av. length: ")
	    (display (quotient nr-literals clause-nr)) (display ".")
	    (display (modulo (quotient (* nr-literals 100) clause-nr) 100))
	    (print " ")))
 (if compile-only (begin (print "formula compiled") (exit 0)))
 (set! atom-list-init atom-list)
 (if (not order-unit)
     (order-atoms)
     (set! atom-vector (list->vector atom-list)))
 (set! output (evaluate 0 0 0 0))
 (if (> inference-limit 0)
     (if (= auto 0) (random-evaluate) (auto-evaluate)))
 (if split 
     (begin (display "Nr of splitted clauses: ")
	    (display splitted-clauses) (newline)))
 (set! inference (+ inference total-inferences))
 (set! branchings (+ branchings total-branchings))
 (if lemmas-on
     (begin (display "Context lemmas: ") (display nr-gen-lemmas)
	    (display ", max in memory: ") (display max-nr-kept-lemmas)
	    (display ", root: ") (display nr-kept-lemmas) (newline)
	    (display "Lemmas used/usages: context: ") (display used-lemmas)
	    (display "/") (display lemma-usages)
	    (set! used-lemmas 0) (set! lemma-usages 0)
	    (delete-lemmas root-lemmas)
	    (display ", used/usages root: ") (display used-lemmas)
	    (display "/") (display lemma-usages)
	    (print " ")))
 (if (and output (= nr-models 0))
     (begin (display "unsatisfiable: ") (display inference))
     (begin (display "satisfiable: ") 
	    (if allmodels
		(begin
		  (if (or (= verbose 0) (and (= verbose 1) (> nr-models 1)))
		      (begin (display "nr of models: ") (display nr-models)
			     (display ", ")))
		  (display inference))
		(display inference))))
 (display"/") (display branchings) (display " branches")
 (if (> attempts 1) (begin (display " (strategy ") (display attempts) (display ")")))
 (newline)
; (display-atoms atom-list)
)

(define (auto-evaluate)
 (if (and (not output) (= 0 nr-models))
     (begin ;(display "(") (display output)
	    ;(display " ") (display nr-models)
	    ;(display " ") (display inference) (print ")")
	    (set! auto (+ 1 auto)) (auto-mode auto)
	    (set! unit-index -1) (insert-units-in-vector unit-list)
	    (set! purity-index -1) (if purity (extract-pure atom-list))
	    (set! current-atom #f)
	    (set! total-inferences (+ inference total-inferences))
	    (set! total-branchings (+ branchings total-branchings))
	    (set! inference 0) (set! lemma-diff 0) (set! branchings 0)
	    (set! attempts (+ attempts 1))
	    (set! output (evaluate 0 0 0 0))
	    (auto-evaluate))))

(define (random-evaluate)
 (if (and (not output) (= 0 nr-models))
     (begin (set! random-branch (- random-branch 1))
            (set! attempts (+ attempts 1))
;            (display random-branch) (newline)
            (set! inference-limit (+ inference-limit
                (quotient (* inference-limit-increase inference-limit) 100)))
            (set! unit-index -1) (insert-units-in-vector unit-list)
            (set! inference 0) (set! lemma-diff 0) (set! branchings 0)
	    (set! output (evaluate 0 0 0 0))
            (random-evaluate))))

(define (input-formula args)
; (display "Argument list: ") (display args) (newline)
 (update-defaults args)
; (display "DEFAULTS UPDATED")
 (set! clauses-vector (make-vector (+ max-cl-length 1) '()))
 (set! clauses-weight-vector (make-vector (+ max-cl-length 1)))
 (if (not (file-exists? (car (last-pair args))))
     (begin (display "Error: ") (display "given input file does not exist!")
	    (newline) (exit 1)))
 (set! infile (open-input-file (car (last-pair args))))
 (if (and split)
     (set! input-size (+ hash-size max-def-number))
     (set! input-size hash-size))
; (display "input-size: ") (display input-size) (newline)
 (load-formula-initial infile)
 (compile)
)

(define (load-formula-initial port)
 (set! interm (read port))
 (if (eof-object? interm)
     (begin (display "Error: ") (display "given input file is empty!")
	    (newline) (exit 1))
     (if (pair? interm)
	 (if (and (list? (car interm))
		  (pair? (cdr interm))
		  (list? (cadr interm))
;		  (or (and (pair? (car interm))
;			(or (string? (caar interm)) (symbol? (caar interm))))
;		      (and (pair? (cadr interm))
;		        (or (string? (cadr interm)) (symbol? (cadr interm)))))
		  )
	     (begin (set! hash-table (make-vector hash-size #f))
	            (set! interm (hashify-clause interm))
		    (if convert-format (print interm))
		    (set! clause-nr (+ 1 clause-nr))
		    (clause-insert)
		    (load-formula-general port))
	     (if (integer? (car interm))
		 (begin (set! clause-nr (+ 1 clause-nr))
			(clause-insert)
			(load-formula port))
		 (begin (display "Error: ") (display "error in input!")
			(newline) (exit 1))))
	 (if (and (equal? interm 'p)
		  (equal? (read port) 'cnf))
	     (begin (read port) (read port)
		    (set! clause-nr (+ 1 clause-nr))
		    (load-formula-dimacs port))
	     (begin (display "Error: ") (display "error in input!")
		    (newline) (exit 1)))
)))

(define (load-formula-dimacs port)
 (set! interm (read port))
 (if (eof-object? interm)
     (begin (if (not (eq? in-clause '()))
		(begin (set! interm (reverse! in-clause))
		       (set! clause-nr (+ 1 clause-nr))
		       (clause-insert)))
	    (close-input-port infile))
     (begin (if (= interm 0) 
		(begin (set! interm (reverse! in-clause))
		       (set! clause-nr (+ 1 clause-nr))
		       (clause-insert)
		       (set! in-clause '()))
		(set! in-clause (cons interm in-clause)))
	    (load-formula-dimacs port))))

(define (load-formula port)
 (set! interm (read port))
 (if (eof-object? interm)
     (close-input-port infile)
     (begin (set! clause-nr (+ 1 clause-nr))
	    (clause-insert)
            (load-formula port))))

(define (load-formula-general port)
 (set! interm (read port))
 (if (eof-object? interm)
     (close-input-port infile)
     (begin (set! interm (hashify-clause interm))
	    (if convert-format (print interm))
            (set! clause-nr (+ 1 clause-nr))
	    (clause-insert)
            (load-formula-general port))))

(define (next-neighbor neighbors)
 (if (pair? neighbors)
     (if (integer? (car neighbors))
	 (if (>= (car neighbors) neighbor-selection)
	 ; ie (car neighbors) occurs >= neighbor-selection times as neighbor
	     (if (a-sign (cadr neighbors))
		 (next-neighbor (cddr neighbors))
		 (cadr neighbors))
	     #f)
	 (if (a-sign (car neighbors))
	     (next-neighbor (cdr neighbors))
	     (car neighbors)))
     #f))

(define (next-neighbor-global neighbors)
 (if (pair? neighbors)
     (if (a-sign (car neighbors))
	 (next-neighbor-global (cdr neighbors))
	 (car neighbors))
     #f))

(define (model)
;d; (newline) (display "MODEL FOUND AT INFERENCE: ") (display inference)
; outputs always #f
 (set! nr-models (+ 1 nr-models))
 (if (or (= verbose 2) (and (= 1 nr-models) (= verbose 1)))
     (begin (if (not also-neg-in-model) 
		(display "positive part of "))
	    (display "model ")
	    (display nr-models)
	    (display ":")
	    (newline)
	    (print-model line atom-list (- line 1))
	    (newline)))
 #f)

(define (print-model nr atom-list line)
 (if (pair? atom-list)
     (begin 
       (if (< (a-name (car atom-list)) 65536)
	   (if (= 1 (a-sign (car atom-list)))
	       (begin (if (= 0 nr) 
			  (begin (newline) (set! nr line))
			  (set! nr (- nr 1)))
		      (display " ") (display (a-realname (car atom-list))))
	       (if also-neg-in-model 
		   (begin (if (= 0 nr) 
			      (begin (newline) (set! nr line))
			      (set! nr (- nr 1)))
			  (display " -")
			  (display (a-realname (car atom-list)))))))
       (print-model nr (cdr atom-list) line))))

(define (^ m n) (if (> n 0) (* m (^ m (- n 1))) 1))

(define (compute-frequencies atoms)
  (if (pair? atoms)
      (begin
	(if (not (a-sign (car atoms)))
	    (begin
	      (set-a-pos-frequency (car atoms) (freq-clauses 0 (a-pos-clauses (car atoms))))
	      (set-a-neg-frequency (car atoms) (freq-clauses 0 (a-neg-clauses (car atoms))))))
	(compute-frequencies (cdr atoms)))))

(define (compute-occurrences atoms)
  (if (pair? atoms)
      (begin (set-a-pos-occ (car atoms) (length (a-pos-clauses (car atoms))))
	     (set-a-neg-occ (car atoms) (length (a-neg-clauses (car atoms))))
	     (compute-occurrences (cdr atoms)))))

(define (freq-clauses freq clauses)
 (if (pair? clauses) 
     (begin
       (if (and (not (c-activating-atom (car clauses))) (input-clause? (car clauses)))
	   (set! freq (+ freq (vector-ref clauses-weight-vector (c-length (car clauses))))))
       (freq-clauses freq (cdr clauses)))
     freq))
	 
(define (auto-mode arg)
 (cond ((= arg 1) ; (display "1:: ")
	(set! lemmas-on #t) (initial-lemmas2clauses atom-list)
	(set! lemma-adaptive #t) (initial-lemmas2clauses atom-list)
	(set! foldup-stack (make-vector nr-atoms))
	(set! tableau 0) (set! tableau-right 0) (set! leaf-factor 0)
	(set! clause-divisor -1) (set! base-divisor 20)
	(set! inference-limit 20000))
       ((= arg 2) ; (print inference) (display "2:: ")
	(set! clause-divisor 0) (set! base-divisor 0)
	(set! dynamic-steps 10) (set! dynamic-depth 15)
	(set! depth-nontab 15) (set! tableau-steps 0)
	(set! inference-limit 30000))
       ((= arg 3) ; (print inference) (display "3:: ")
	(set! clause-divisor -1) (set! base-divisor 20)
	(set! dynamic-steps 10) (set! dynamic-depth 15)
	(set! depth-nontab 0) (set! tableau-steps 5)
	(set! inference-limit 0))))

(define (update-freq clauses)
 (if (pair? clauses)
     (begin (update-freq-pos (c-pos-literals (car clauses)))
            (update-freq-neg (c-neg-literals (car clauses)))
	    (update-freq (cdr clauses)))))

(define (update-freq-pos atoms)
 (if (pair? atoms)
     (begin (set-a-pos-frequency (car atoms) (+ (a-pos-frequency (car atoms)) weight))
	    (update-freq-pos (cdr atoms)))))

(define (update-freq-neg atoms)
 (if (pair? atoms)
     (begin (set-a-neg-frequency (car atoms) (+ (a-neg-frequency (car atoms)) weight))
	    (update-freq-neg (cdr atoms)))))
 
(define (extract-pure atoms)
 (if (pair? atoms)
     (begin (if (or (= 0 (a-pos-occ (car atoms)))
		    (= 0 (a-neg-occ (car atoms))))
		(begin (set! purity-index (+ purity-index 1))
		       (vector-set! purity-vector purity-index (car atoms))))
	    (extract-pure (cdr atoms)))))

;;;*************************************************************************
;;;module name: compiler.sch
;;;author     : Reinhold Letz, TU Munich
;;;date       : 29.10.98
;;;description: compiler for semprop
;;;remarks    : literals are integers /= 0 and may occur multiply in a clause
;;;*************************************************************************

(define (compile) 
 (set! input-vector (make-vector (+ input-size 1) #f))
 (set! cl-length 1)
 (if (> clause-divisor 0)
     (set! end-divisor (max 1 (quotient clause-nr (^ 2 clause-divisor)))))
 (if (> base-divisor 0) (set! end-divisor clause-nr))
 (set! clause-nr 0) (compile-to-semprop #t)
 (set! unit-vector-length clause-nr)
 (set! unit-vector (make-vector unit-vector-length #f))
 (if purity (set! purity-vector (make-vector nr-atoms #f)))
 (insert-units-in-vector unit-list)
; (newline) (display-atom-names atom-list)
; (if purity (delete-pure atom-list))
 (if reverse-atomlist (set! atom-list (reverse! atom-list)))
 (set! atom-list-init atom-list)
; (if (not order-unit) (order-atoms))
; (order-atoms)
; (set! atom-vector (list->vector atom-list))
; (newline) (display-atom-names atom-list)
 (if (or neighbor-selection neighbor-global)
	 (insert-atom-neighbors-atoms atom-list))
 (if reverse-clauses (reverse-clauses-in-atoms atom-list))
 (set! clauses-weight-diff-vector (make-vector (+ max-cl-length 1)))
 (make-clauses-weight-diff-vector max-cl-length)
 (if init-sign (insert-initial-freq-pos-neg atom-list))
)

(define (compile-to-semprop init) 
 (if (<= cl-length max-cl-length)
     (begin 
       (cond ((> base-divisor 0)
	      (cond
	       ((= cl-length 2) (set! weight
				      (quotient greatest-int end-divisor))
		(vector-set! clauses-weight-vector 2 weight))
	       ((> cl-length 2)
		(set! weight (vector-ref clauses-weight-vector
					 (- cl-length 1)))
		(set! weight (max 1 (quotient (* weight 10) base-divisor))))))
             ((= clause-divisor 0) (set! divisor 1))
             ((= clause-divisor 1) (set! divisor cl-length))
             ((= clause-divisor 2) (set! divisor (^ cl-length 2)))
             ((= clause-divisor 3) 
              (if (> cl-length 812) ; 812^3 <= greatest-int < 813^3
		  (set! divisor greatest-int)
		  (set! divisor (^ cl-length 3))))
             ((= clause-divisor 4)
	      (if (> cl-length 152) ; 152^4 <= greatest-int < 153^4  
		  (set! divisor greatest-int)
		  (set! divisor (^ cl-length 4))))
             ((= clause-divisor 5) 
	      (if (> cl-length 55) ; 55^5 <= greatest-int < 56^5
		  (set! divisor greatest-int)
		  (set! divisor (^ cl-length 5))))
             ((= clause-divisor 6) 
	      (if (> cl-length 28) ; 28^6 <= greatest-int < 29^6
		  (set! divisor greatest-int)
		  (set! divisor (^ cl-length 6))))
             ((= clause-divisor 7) 
	      (if (> cl-length 17) ; 17^7 <= greatest-int < 18^7
		  (set! divisor greatest-int)
		  (set! divisor (^ cl-length 7))))
             ((= clause-divisor 8) 
	      (if (> cl-length 12) ; 12^8 <= greatest-int < 13^8
		  (set! divisor greatest-int)
		  (set! divisor (^ cl-length 8))))
             ((= clause-divisor 9) 
	      (if (> cl-length 9) ; 12^9 <= greatest-int < 13^9
		  (set! divisor greatest-int)
		  (set! divisor (^ cl-length 9))))
             (else (newline) (display "Error: wrong clause divisor!") 
                   (exit 1)))
       (if (or (= cl-length 1) (= clause-divisor 0))
	   (set! weight 1)
	   (if (= base-divisor 0)
	       (set! weight (max 1
		    (quotient (quotient greatest-int divisor) end-divisor)))))
       (vector-set! clauses-weight-vector cl-length weight)
       (if init
	   (compile-to-semprop-list (vector-ref clauses-vector cl-length))
	   (vector-set! clauses-weight-vector cl-length weight))
       (set! cl-length (+ cl-length 1))
       (compile-to-semprop init))))

(define (compile-to-semprop-list input-clauses) 
 (if (pair? input-clauses)
     (begin (set! clause-nr (+ 1 clause-nr))
            (compile-input-clause (car input-clauses) #f)
            (compile-to-semprop-list (cdr input-clauses)))))

(define (prune-clause)
 (if (element (- (car interm)) (cdr interm))
     (set! tautology #t)
     (if (element (car interm) (cdr interm))
         (begin (set! interm (cdr interm)) (prune-clause))
         (prune-clause1 interm (cdr interm)))))

(define (prune-clause1 precell literals)
 (if (pair? literals)
     (if (element (- (car literals)) (cdr literals))
         (set! tautology #t)
         (if (element (car literals) (cdr literals))
             (begin (set-cdr! precell (cdr literals))
                    (prune-clause1 precell (cdr literals)))
             (begin (set! cl-length (+ cl-length 1))
                    (prune-clause1 literals (cdr literals)))))))

(define (clause-insert)
 (set! tautology #f) (set! cl-length 1) (prune-clause)
   (if (not tautology)
     (begin 
       (if (and split (> cl-length 3)
                (< def-number input-size))
	   (begin (set! splitted-clauses (+ 1 splitted-clauses))
	     (split-clause-insert (car interm) (cadr interm)
				  (cddr interm) cl-length))
           (begin (set! nr-literals (+ nr-literals cl-length))
	     (set! cl-length (min cl-length max-cl-length))
             (vector-set! clauses-vector cl-length
                   (cons interm (vector-ref clauses-vector cl-length))))))))

(define (split-clause-insert lit1 lit2 lits length)
 (set! def-number (+ def-number 1))
; (newline) (display "def-number: ") (display def-number)
 (vector-set! clauses-vector 3 (cons (list lit1 lit2 (- def-number))
                                     (vector-ref clauses-vector 3)))
 (if strongsplit
     (vector-set! clauses-vector 2 (cons (list (- lit1) def-number)
				       (cons (list (- lit2) def-number)
					     (vector-ref clauses-vector 2)))))
 (set! length (- length 1))
 (if (and (> length 3) (< def-number input-size))
     (split-clause-insert def-number (car lits) (cdr lits) length)
     (vector-set! clauses-vector length (cons (cons def-number lits)
                                      (vector-ref clauses-vector length)))))

(define (element elem list)
 (if (pair? list)
     (if (eq? elem (car list))
         list
         (element elem (cdr list)))
     #f))

(define (nosubset list1 list2)
 (if (pair? list1)
     (if (element (car list1) list2)
	 (nosubset (cdr list1) list2)
	 (car list1))
     #f))

(define (consistent-disjunction-branch disjunction branch)
 (if (pair? disjunction)
     (if (element (- (car disjunction)) branch)
	 (consistent-disjunction-branch (cdr disjunction) branch)
	 (car disjunction))
     #f))

(define (compile-input-clause input-clause clause) 
 (set! clause (vector #f cl-length '() '()))
 (set! clause-list (cons clause clause-list))
 (compile-input-literals input-clause clause #f)
 (if (= 1 cl-length) (set! unit-list (cons clause unit-list))))

(define (compile-input-literals literals clause atom)
 (if (pair? literals)
     (begin 
       (if (> (car literals) 0)
           (begin (set! atom (vector-ref input-vector (car literals)))
              (if (not atom) 
                  (begin (set! atom (make-atom (car literals)))
;			 (if (= (a-realname atom) 505) (set! atom505 atom))
                         (if (> (car literals) hash-size)
                             (set! def-list (cons atom def-list))
                             (set! atom-list (cons atom atom-list)))
                         (set! nr-atoms (+ 1 nr-atoms))
                         (vector-set! input-vector (car literals) atom))) 
              (set-a-pos-clauses atom (cons clause (a-pos-clauses atom)))
;              (set-a-pos-frequency atom (+ weight (a-pos-frequency atom)))
              (set-c-pos-literals clause (cons atom (c-pos-literals clause))))
           (begin (set! atom (vector-ref input-vector (- (car literals))))
              (if (not atom) 
                  (begin (set! atom (make-atom (- (car literals))))
;                         (if (= (a-realname atom) 505) (set! atom505 atom))
			 (if (> (- (car literals)) hash-size)
                             (set! def-list (cons atom def-list))
                             (set! atom-list (cons atom atom-list)))
                         (set! nr-atoms (+ 1 nr-atoms))
                         (vector-set! input-vector (- (car literals)) atom)))
              (set-a-neg-clauses atom (cons clause (a-neg-clauses atom)))
;              (set-a-neg-frequency atom (+ weight (a-neg-frequency atom)))
              (set-c-neg-literals clause (cons atom (c-neg-literals clause)))))
       (compile-input-literals (cdr literals) clause #f))))

(define (insert-units-in-vector unit-list)
 (if (pair? unit-list)
     (begin (set! unit-index (+ unit-index 1))
            (vector-set! unit-vector unit-index (car unit-list))
            (insert-units-in-vector (cdr unit-list)))))

(define (insert-initial-freq-pos-neg atomlist)
 (if (pair? atomlist)
     (begin (if (<= (a-pos-frequency (car atomlist))
		    (a-neg-frequency (car atomlist)))
		(set-a-neg-less-pos (car atomlist) #f))
	    (insert-initial-freq-pos-neg (cdr atomlist)))))
	    
(define (order-atoms)
; (if display-atomlist (begin (display-atom-names atom-list) (newline)))
 (if atom-ordering (set! atom-list (merge-sort atom-list)))
 (if (and split (pair? def-list))
     (begin (if atom-ordering (set! def-list (merge-sort def-list)))
            (if deflast
		(set-cdr! (last-pair atom-list) def-list)
		(begin (set-cdr! (last-pair def-list) atom-list)
		       (set! atom-list def-list)))))
 (if display-atomlist (begin (display-atoms atom-list) (newline)))
 (set! atom-vector (list->vector atom-list)))

(define (merge-sort unsorted-list)
; outputs a sorted list of input:
;   procedure is externally side-effect free;
;   elements are listed in decreasing order according
;   to the function atom->=
 (if (pair? unsorted-list)
     (merge-sort-divide unsorted-list (length unsorted-list))
     unsorted-list))

(define (merge-sort-divide unsorted up2l)
; up2l must be <= (length unsorted) and unsorted must be non-empty
 (if (> up2l 2)
     (begin
       (let* ((l2 (quotient up2l 2)) (l1 (- up2l l2)))
	 (merge-sort-merge (merge-sort-divide unsorted l1)
			   (merge-sort-divide (list-tail unsorted l1) l2))))
     (if (= up2l 2)
	 (if (atom->= (car unsorted) (cadr unsorted))
	     (list (car unsorted) (cadr unsorted))
	     (list (cadr unsorted) (car unsorted)))
	 (list (car unsorted)))))

(define (merge-sort-merge sorted1 sorted2)
; sorted1 and sorted2 must be non-empty
 (if (atom->= (car sorted1) (car sorted2))
     (begin (merge-sort-merge1 (cdr sorted1) sorted2 sorted1)
            sorted1)
     (begin (merge-sort-merge1 (cdr sorted2) sorted1 sorted2)
            sorted2)))

(define (merge-sort-merge1 sorted1 sorted2 sorted1pred)
; sorted2 must be non-empty and sorted1 = (list-tail sorted1pred 1)
 (if (pair? sorted1)
     (if (atom->= (car sorted1) (car sorted2))
         (merge-sort-merge1 (cdr sorted1) sorted2 sorted1)
         (begin (set-cdr! sorted1pred sorted2)
		(merge-sort-merge1 (cdr sorted2) sorted1 sorted2)))
     (set-cdr! sorted1pred sorted2)))
         
(define (atom->= m n) 
  (let ((m_pos (a-pos-frequency m)) (m_neg (a-neg-frequency m)) 
        (n_pos (a-pos-frequency n)) (n_neg (a-neg-frequency n)))
;     (if (= 0 m_pos)
; 	(set! m_neg 0)
; 	(if (= 0 m_neg) (set! m_pos 0)))
;     (if (= 0 n_pos)
; 	(set! n_neg 0)
; 	(if (= 0 n_neg) (set! n_pos 0)))
    (cond
      (sum-frequency
	(or (> (+ m_pos m_neg) (+ n_pos n_neg))
	    (and (= (+ m_pos m_neg) (+ n_pos n_neg))
		 (>= (max m_pos m_neg) (max n_pos n_neg)))))
      (max-frequency
	(cond ((> m_pos m_neg)
	       (cond ((> n_pos n_neg)
		      (cond ((> m_pos n_pos) #t)
			    ((< m_pos n_pos) #f)
			    (else (>= m_neg n_neg))))
		     ((< n_pos n_neg)
		      (cond ((> m_pos n_neg) #t)
			    ((< m_pos n_neg) #f)
			    (else (>= m_neg n_pos))))
                     (else (> m_pos n_pos))))
	      ((< m_pos m_neg)
	       (cond ((> n_pos n_neg)
		      (cond ((> m_neg n_pos) #t)
			    ((< m_neg n_pos) #f)
			    (else (>= m_pos n_neg))))
		     ((< n_pos n_neg)
		      (cond ((> m_neg n_neg) #t)
			    ((< m_neg n_neg) #f)
			    (else (>= m_pos n_pos))))
                     (else (> m_neg n_pos))))
	      ((> n_pos n_neg) (>= m_pos n_pos))
	      ((< n_pos n_neg) (>= m_pos n_neg))
	      (else #t)))
      (min-frequency
        (cond ((< m_pos m_neg)
	       (cond ((< n_pos n_neg)
		      (cond ((> m_pos n_pos) #t)
			    ((< m_pos n_pos) #f)
			    (else (>= m_neg n_neg))))
		     ((> n_pos n_neg)
		      (cond ((> m_pos n_neg) #t)
			    ((< m_pos n_neg) #f)
			    (else (>= m_neg n_pos))))
		     (else (> m_pos n_pos))))
	      ((> m_pos m_neg)
	       (cond ((< n_pos n_neg)
		      (cond ((> m_neg n_pos) #t)
			    ((< m_neg n_pos) #f)
			    (else (>= m_pos n_neg))))
		     ((> n_pos n_neg)
		      (cond ((> m_neg n_neg) #t)
			    ((< m_neg n_neg) #f)
			    (else (>= m_pos n_pos))))
                     (else (> m_neg n_pos))))
	      ((< n_pos n_neg) (> m_pos n_pos))
	      ((> n_pos n_neg) (> m_pos n_neg))
	      (else #t)))
      (prod-frequency
        (cond ((> m_pos n_pos)
	       (>= (quotient (min greatest-int (* m_pos 10)) (+ n_pos 1))
		   (quotient (min greatest-int (* n_neg 10)) (+ m_neg 1))))
              ((= m_pos n_pos) (>= m_neg n_neg))
	      (else (>= (quotient (min greatest-int (* m_neg 10)) (+ n_neg 1))
		     (quotient (min greatest-int (* n_pos 10)) (+ m_pos 1))))))
      (else ; sum-max-frequency
        (>= (+ (* sum-max1 (+ m_pos m_neg)) (* sum-max2 (max m_pos m_neg)))
            (+ (* sum-max1 (+ n_pos n_neg)) (* sum-max2 (max n_pos n_neg)))))
      )))

(define (hash-reset) (vector-fill! hash-table #f))

(define (hash-insert-key! index obj)
 (if (or (eq? index 0) (eq? index hash-size))
     (set! index 1))
 (if (and (vector-ref hash-table index) 
	  (not (equal?  (vector-ref hash-table index) obj)))
     (hash-insert-key! (+ hash-step index) obj)
     (begin (vector-set! hash-table index obj)
	    index)))

(define (hash-member-key? index obj)
 (if (or (eq? index 0) (eq? index hash-size))
     (set! index 1))
 (if (not (vector-ref hash-table index))
     #f
     (if (equal? (vector-ref hash-table index) obj)
	 index
	 (hash-member-key? (hash-step  index) obj))))

(define (hashify-poslits pos-lits neg-lits)
 (if (null? pos-lits)
     (hashify-neglits neg-lits)   
     (cons (hash-insert! (car pos-lits)) 
	   (hashify-poslits (cdr pos-lits) neg-lits))))

(define (hashify-neglits neg-lits)
 (if (null? neg-lits)
     '()
     (cons (- (hash-insert! (car neg-lits))) 
	   (hashify-neglits (cdr neg-lits)))))

(define (insert-atom-neighbors-atoms atoms)
 (if (pair? atoms)
     (begin (insert-atom-neighbors (car atoms))
	    (insert-atom-neighbors-atoms (cdr atoms)))))

(define (insert-atom-neighbors atom)
 (set! current-atom atom)
; (display "------ ")
; (display (a-realname current-atom))
; (display " ------") (newline)
 (set! neighbor-list '())
 (atom-neighbors (a-pos-clauses atom))
; (display "+1: ") (print-neighbors neighbor-list)
 (set-a-pos-neighbors atom (extract-sort-even-int neighbor-list))
; (display "+2: ") (print-neighbors (a-pos-neighbors atom))
 (set! neighbor-list '())
 (atom-neighbors (a-neg-clauses atom))
; (display "-1: ") (print-neighbors neighbor-list)
 (set-a-neg-neighbors atom (extract-sort-even-int neighbor-list))
; (display "-2: ") (print-neighbors (a-neg-neighbors atom))
 (if neighbor-global
     (begin
       (set! neighbor-list (a-pos-neighbors atom))
       (set! neighbor-list (delete-ints neighbor-list))
;       (display "NG pos: ") (print-neighbors neighbor-list)
       (set-a-pos-neighbors atom (merge-sort neighbor-list))
;       (display "NG pos: ") (print-neighbors (a-pos-neighbors atom))
       (set! neighbor-list (a-neg-neighbors atom))
       (set! neighbor-list (delete-ints neighbor-list))
;       (display "NG neg: ") (print-neighbors neighbor-list)
       (set-a-neg-neighbors atom (merge-sort neighbor-list))
;       (display "NG neg: ") (print-neighbors (a-neg-neighbors atom))
       )))


(define (print-neighbors list)
 (if (pair? list)
     (begin (if (vector? (car list))
		(begin (display " * ")
		       (display (a-realname (car list))))
		(display (car list)))
	    (display " ")
	    (print-neighbors (cdr list)))
     (newline)))
	    
(define (atom-neighbors clauses)
 (if (pair? clauses)
     (begin (if (> (c-length (car clauses)) 2)
		(begin (atom-neighbors1 (c-pos-literals (car clauses)))
		       (atom-neighbors1 (c-neg-literals (car clauses)))))
	    (atom-neighbors (cdr clauses)))))

(define (atom-neighbors1 atoms)
 (if (pair? atoms)
     (begin 
       (if (not (eq? (car atoms) current-atom))
	   (if (pair? neighbor-list)
	       (incr-neighbor-freq (car atoms) (a-name (car atoms))
				   (a-name (car neighbor-list)) neighbor-list)
	       (set! neighbor-list (list (car atoms) 1))))
       (atom-neighbors1 (cdr atoms)))))

(define (incr-neighbor-freq atom atom-index neighbor-index neighbors)
 (cond ((> atom-index neighbor-index)	    
	(if (pair? (cddr neighbors))
	    (incr-neighbor-freq atom atom-index (a-name (caddr neighbor-list))
				(cddr neighbors))
	    (set-cdr! (cdr neighbors) (list atom 1))))
       ((= atom-index neighbor-index)
	(set-car! (cdr neighbors) (+ 1 (cadr neighbors))))
       (else (set-cdr! (cdr neighbors) (cons (car neighbors)
				     (cons (cadr neighbors) (cddr neighbors))))
	       (set-car! neighbors atom)
	       (set-car! (cdr neighbors) 1))))

(define (extract-sort-even-int list)
 (let ((freq (make-vector (+ 1 (max-int-even 0 list)) '())))
   (in-vector-int-even freq list)
   (vector-int-list freq 1 (vector-length freq) '())))

(define (max-int-even max list)
 (if (pair? list)
     (if (< max (cadr list))
	 (max-int-even (cadr list) (cddr list))
	 (max-int-even max (cddr list)))
     max))

(define (in-vector-int-even freq list)
 (if (pair? list)
     (begin (vector-set! freq (cadr list) 
			 (cons (car list) (vector-ref freq (cadr list))))
	    (in-vector-int-even freq (cddr list)))))

(define (vector-int-list freq index length list)
(if (< index length)
    (begin (if (pair? (vector-ref freq index))
	       (set! list (cons index (append! (vector-ref freq index) list))))
	   (vector-int-list freq (+ index 1) length list))
    list))

(define (clause-literals clause)
 (clause-literals-pos (c-pos-literals clause)
		      (c-neg-literals clause)))

(define (clause-literals-pos pos neg)
 (if (pair? pos)
     (cons (a-name (car pos)) (clause-literals-pos (cdr pos) neg))
     (clause-literals-neg neg)))

(define (clause-literals-neg neg)
 (if (pair? neg)
     (cons (- (a-name (car neg))) (clause-literals-neg (cdr neg)))
     '()))

(define (path-literals index)
 (if (>= index 0)
     (if (a-sign (vector-ref path-vector index))
	 (if (= 0 (a-sign (vector-ref path-vector index)))
	     (cons (- (a-name (vector-ref path-vector index)))
		   (path-literals (- index 1)))
	     (cons (a-name (vector-ref path-vector index))
		   (path-literals (- index 1))))
	 (path-literals (- index 1)))
     '()))

(define (interpretation-literals atoms)
 (if (pair? atoms)
     (if (a-sign (car atoms))
	 (if (= 0 (a-sign (car atoms)))
	     (cons (- (a-name (car atoms)))
		   (interpretation-literals (cdr atoms)))
	     (cons (a-name (car atoms))
		   (interpretation-literals (cdr atoms))))
	 (interpretation-literals (cdr atoms)))
     '()))

;(define (nconc list1 list2)
; (if (pair? list1)
;     (begin (set-cdr! (last-pair list1) list2) list1)
;     list2))
(define (delete-ints list)
 (if (pair? list)
     (if (not (integer? (car list)))
	 (cons (car list) (delete-ints (cdr list)))
	 (delete-ints (cdr list)))
     '()))

(define (vector30in vec1 vec2 index divisor)
 (if (>= index 0)
     (begin (vector-set! vec1 index (quotient index divisor))
	    (vector-set! vec2 index (^ 2 (modulo index divisor)))
	    (vector30in vec1 vec2 (- index 1) divisor))))

(define (reverse-clauses-in-atoms atomlist)
 (if (pair? atomlist)
     (begin (set-a-pos-clauses (car atomlist)
			       (reverse! (a-pos-clauses (car atomlist))))
            (set-a-neg-clauses (car atomlist)
			       (reverse! (a-neg-clauses (car atomlist))))
	    (reverse-clauses-in-atoms (cdr atomlist)))))

(define (delete-pure atoms)
; pure clauses are not really deleted, but the literal weights are not counted;
; this achieves that pure literals are put to the end of the atom list
 (if (pair? atoms)
     (begin (set! pure '())
;	    (display " ") (display (a-realname (car atoms)))
;	    (display ":")
       (if (= 0 (a-pos-frequency (car atoms)))
	   (if (> (a-neg-frequency (car atoms)) 0)
	       (delete-pure-clauses (a-neg-clauses (car atoms))))
	   (if (= 0 (a-neg-frequency (car atoms)))
	       (delete-pure-clauses (a-pos-clauses (car atoms)))))
       (set! atoms (append! pure (cdr atoms)))
       (delete-pure atoms))))
;     (reactivate-pure pure-clauses)
 
(define (delete-pure-clauses clauses)
 (if (pair? clauses)
     (begin
       (if (not (c-activating-atom (car clauses)))
	   (begin
;	     (display "|")
	     (set-c-activating-atom (car clauses) 0)
	     (set! pure-clauses (cons (car clauses) pure-clauses))
	     (delete-pure-pos (c-pos-literals (car clauses))
	       (vector-ref clauses-weight-vector (c-length (car clauses))))
	     (delete-pure-neg (c-neg-literals (car clauses))
	       (vector-ref clauses-weight-vector (c-length (car clauses))))))
       (delete-pure-clauses (cdr clauses)))))

(define (delete-pure-pos lits weight)
 (if (pair? lits)
     (begin (set-a-pos-frequency (car lits)
				 (- (a-pos-frequency (car lits)) weight))
	    (if (= 0 (a-pos-frequency (car lits)))
		(set! pure (cons (car lits) pure)))
	    (delete-pure-pos (cdr lits) weight))))

(define (delete-pure-neg lits weight)
 (if (pair? lits)
     (begin (set-a-neg-frequency (car lits)
				 (- (a-neg-frequency (car lits)) weight))
	    (if (= 0 (a-neg-frequency (car lits)))
		(set! pure (cons (car lits) pure)))
	    (delete-pure-neg (cdr lits) weight))))

(define (reactivate-pure clauses)
 (if (pair? clauses)
     (begin (set-c-activating-atom (car clauses) #f)
	    (reactivate-pure (cdr clauses)))))

(define (make-clauses-weight-diff-vector index)
 (if (> index 2)
     (begin (vector-set! clauses-weight-diff-vector index
			 (- (vector-ref clauses-weight-vector (- index 1))
			    (vector-ref clauses-weight-vector index)))
	    (make-clauses-weight-diff-vector (- index 1)))))

(define (display-clauses clauses)
 (if (pair? clauses)
     (begin
       (if (or proof (not (c-activating-atom (car clauses))))
	   (display-clause (car clauses)))
       (display-clauses (cdr clauses)))))

(define (display-clause clause)
 (display "( ")
 (display-pos-literals (c-pos-literals clause))
 (display-neg-literals (c-neg-literals clause))
 (display ")")
 (if (and proof (> (vector-length clause) 4))
     (begin (display "length: ") (display (c-length clause))
            (display ", rel-vector: ") (display (c-rel-vector clause))
	    ))
)

(define (display-all-clauses clauses)
 (if (pair? clauses)
     (begin (display-all-clause (car clauses))
            (display-all-clauses (cdr clauses)))))

(define (display-all-clause clause)
 (display " ") (display (c-length clause))
 (display "( ")
 (display-pos-literals (c-pos-literals clause))
 (display-neg-literals (c-neg-literals clause))
 (display ")") 
 (if (c-activating-atom clause) 
     (display (a-realname (c-activating-atom clause))))
 (newline))

(define (display-pos-literals atoms)
 (if (pair? atoms)
     (begin (display (a-realname (car atoms))) (display " ")
            (display-pos-literals (cdr atoms)))))

(define (display-neg-literals atoms)
 (if (pair? atoms)
     (begin (display "-") (display (a-realname (car atoms))) (display " ")
            (display-neg-literals (cdr atoms)))))

(define (display-atom-names atoms)
 (if (pair? atoms)
     (begin (display (a-realname (car atoms))) (display " ")
            (display-atom-names (cdr atoms)))))

(define (display-atoms atoms)
 (if (pair? atoms)
     (begin
       (if (not (a-sign (car atoms))) (display-atom (car atoms)))
       (display-atoms (cdr atoms)))))

(define (display-atom atom)
 (if (a-sign atom) (if (= 0 (a-sign atom)) (display "-") (display "+")))
 (display (a-realname atom)) (display ": freq ")
 (display (a-pos-frequency atom)) (display " ")
 (display (a-neg-frequency atom))
 (newline) (display "pos: ") 
 (display-clauses (a-pos-clauses atom))
 (newline) (display "neg: ") 
 (display-clauses (a-neg-clauses atom))
; (display " neg-less-pos: ") (display (a-neg-less-pos atom))
 (newline))

;(define (display-nr-clauses clauses)
; (if (pair? clauses)
;     (begin (display (c-number (car clauses))) (display " ")
;            (display-nr-clauses (cdr clauses)))))

;(define (display-trail-content trail)
; (display (caar trail)) (display " ")
; (if (pair? (cdr trail)) (display-trail-content (cdr trail))))

;(define (display-inference-lengths atoms)
; (if (pair? atoms)
;     (begin (display (a-realname (car atoms))) (display ":")
;	    (display (length (a-used-inferences (car atoms)))) (display " ")
;	    (display-inference-lengths (cdr atoms)))
;     (newline)))

(define (display-interpretation atoms)
 (if (pair? atoms)
     (begin (if (a-sign (car atoms))
		(begin (if (= 0 (a-sign (car atoms))) (display "-"))
		       (display (a-realname (car atoms)))
		       (display " ")))
	    (display-interpretation (cdr atoms)))))
		    
(define (display-path-index index)
 (if (>= index 0)
     (begin 
       (display-path-index (- index 1))
;       (display "(") (display index) (display ")")
       (if (a-sign (vector-ref path-vector index))
	   (begin
	     (if (= 0 (a-sign (vector-ref path-vector index)))
		 (display " -")
		 (display "  "))
	     (display (a-realname (vector-ref path-vector index))))
;          (begin (display "[")
;		  (display (a-name (vector-ref path-vector index)))
;		  (display "] "))
       ))))

; (define (display-dimacs-format index)
;  (if (<= index max-cl-length)
;      (begin (display-dimacs-format-clauses (vector-ref clauses-vector index))
; 	    (display-dimacs-format (+ index 1)))))

; (define (display-dimacs-format-clauses clauselist)
;  (if (pair? clauselist)
;      (begin (newline)
; 	    (display-pos-ints (c-pos-literals (car clauselist)))
; 	    (display-neg-ints (c-neg-literals (car clauselist)))
; 	    (display " 0")
; 	    (display-dimacs-format-clauses (cdr clauselist)))))

; (define (display-pos-ints naturals)
;  (if (pair? naturals)
;      (begin (display (car naturals))
; 	    (display " ")
; 	    (display-pos-ints (cdr naturals)))))

; (define (display-neg-ints naturals)
;  (if (pair? naturals)
;      (begin (display "-")
; 	    (display (car naturals))
; 	    (display " ")
; 	    (display-neg-ints (cdr naturals)))))


(define (initial-lemmas2clauses atoms)
; globals: atom-vector, path-atom
 (if (pair? atoms)
     (begin (set! path-atom (car atoms))
       (if (pair? (a-pos-clauses path-atom))
	   (set-a-pos-lemmas path-atom (last-pair (a-pos-clauses path-atom))))
       (if (pair? (a-neg-clauses path-atom))
	   (set-a-neg-lemmas path-atom (last-pair (a-neg-clauses path-atom))))
       (initial-lemmas2clauses (cdr atoms)))))

;;;*************************************************************************
;; ---------------------------
;; UPDATE OF DEFAULTS SETTINGS
;; ---------------------------

(define (update-defaults args)
 (if (pair? (cdr args))
     (cond ((equal? (car args) "-negmodel")
            (set! also-neg-in-model #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-allmodels")
            (set! allmodels #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-atomlist")
	    (set! display-atomlist #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-reverse-atomlist")
	    (set! reverse-atomlist #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-stat")
            (set! statistics #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-sum-max")
            (set! max-frequency #f) (set! prod-frequency #f) (set! min-frequency #f)
	    (set! sum-frequency #f) (set! sum-max1 (string->integer (cadr args))) 
	    (set! sum-max2 (string->integer (caddr args))) 
	    (update-defaults (cdddr args)))
	   ((equal? (car args) "-max")
            (set! sum-frequency #f) (set! prod-frequency #f) (set! min-frequency #f)
	    (set! max-frequency #t) 
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-min")
            (set! sum-frequency #f) (set! max-frequency #f) (set! prod-frequency #f)
	    (set! min-frequency #t) 
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-prod")
            (set! sum-frequency #f) (set! max-frequency #f) (set! min-frequency #f) 
            (set! prod-frequency #t) 
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-neg")
            (set! neg-first #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-pos")
            (set! pos-first #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-reverse")
            (set! reverse-clauses #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-noswitch")
            (set! current-switch #f)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-tableau")
            (set! tableau (string->integer (cadr args)))
            (set! tableau-right (string->integer (caddr args)))
            (set! leaf-factor (string->integer (cadddr args)))
	    (update-defaults (cddddr args)))
	   ((equal? (car args) "-tablim")
            (set! tableau-selection (string->integer (cadr args)))
	    (update-defaults (cddr args)))
	   ((equal? (car args) "-neighbor")
            (set! neighbor-selection (string->integer (cadr args)))
	    (set! neighbor-global #f)
	    (update-defaults (cddr args)))
	   ((equal? (car args) "-neighbor-global")
            (set! neighbor-selection #f)
	    (set! neighbor-global #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-lemmas")
            (set! lemmas-on #t)
            (set! nr-lemmas (string->integer (cadr args)))
            (if (= 0 nr-lemmas)
		(begin (set! lemma-adaptive #t) (update-defaults (cdr args)))
		(update-defaults (cddr args))))
	   ((equal? (car args) "-lemma-length")
	    (set! lemmas-on #t)
	    (set! max-lemma-length (string->integer (cadr args)))
	    (update-defaults (cddr args)))
	   ((equal? (car args) "-lemma-filter")
	    (set! lemma-filter (string->integer (cadr args)))
	    (update-defaults (cddr args)))
	   ((equal? (car args) "-lemma-filo")
	    (set! lemma-filo #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-lemma-diff")
	    (set! lemma-min (string->integer (cadr args)))
	    (update-defaults (cddr args)))
	   ((equal? (car args) "-auto")
	    (set! purity #t) (set! order-unit #t) (set! auto 1) (auto-mode 1))
	   ((equal? (car args) "-lemma-rel")
	    (set! lemma-relevance (string->integer (cadr args)))
	    (update-defaults (cddr args)))
	   ((equal? (car args) "-lemma-dyn")
	    (set! lemma-lengthrel (string->integer (cadr args)))
	    (update-defaults (cddr args)))
	   ((equal? (car args) "-lemma-modulo")
	    (set! lemma-modulo (string->integer (cadr args)))
	    (update-defaults (cddr args)))
	   ((equal? (car args) "-compare-selected")
	    (set! compare-selected (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-verbose")
            (set! verbose (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-noverbose")
            (set! verbose 0)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-noorder")
            (set! atom-ordering #f)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-order-unit")
            (set! order-unit #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-size")
	    (update-defaults (cddr args)))
           ((equal? (car args) "-limit")
            (set! inference-limit (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-rand")
            (set! random-branch (string->integer (cadr args)))
            (if (= 0 random-branch) (set! random-branch random-branch-default))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-randp")
            (set! random-percentage (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-freq-div")
            (set! freq-divisor (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-limit+")
            (set! inference-limit-increase (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-div")
            (set! clause-divisor (string->integer (cadr args)))
	    (set! base-divisor 0)
	    (update-defaults (cddr args)))
           ((equal? (car args) "-base-div")
            (set! base-divisor (string->integer (cadr args)))
	    (set! clause-divisor -1)
	    (update-defaults (cddr args)))
           ((equal? (car args) "-occ")
            (set! occ (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-dynamic-depth")
            (set! dynamic-depth (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-dynamic-steps")
            (set! dynamic-steps (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-depth-nontab")
            (set! depth-nontab (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-tableau-steps")
            (set! tableau-steps (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-backjump")
            (set! backjump (string->integer (cadr args)))
            (set! inference-limit backjump)
	    (update-defaults (cddr args)))
           ((equal? (car args) "-init-sign")
            (set! init-sign #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-purity")
            (set! purity #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-proof")
            (set! proof #t)
            (update-defaults (cdr args)))
           ((equal? (car args) "-first-unit")
            (set! first-pure #f)
            (update-defaults (cdr args)))
           ((equal? (car args) "-split")
            (set! split #t)
            (set! max-def-number (min max-def-number 
                                      (string->integer (cadr args))))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-weaksplit")
            (set! split #t) (set! strongsplit #f)
            (set! max-def-number (min max-def-number 
                                      (string->integer (cadr args))))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-deflast")
            (set! deflast #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-compile-only")
            (set! compile-only #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-convert-format")
            (set! convert-format #t)
	    (update-defaults (cdr args)))
	   (else (display "Abort: error in parameter list") 
	         (newline) (exit 1))
)))






