
/******************************************************************
*                                                                 *
*   This file contains functions for converting OTTER's clause    *
*   data structure into SFINDER input string format. The only     *
*   function to be called from outside is                         *
*   char *convert_to_sfinder_clause(), all                        *
*   other functions are merely helpers for this one.              *
*                                                                 *
*   That's not true any more. There are also functions for        *
*   manipulating the bucket-sorted set of support.                *
*                                                                 *
******************************************************************/


#include "SOFTIE_SCOTT.h"

// declared in Fproto.h, but including that causes problems
boolean initial_model_found();

/***********************************
* Converts the OTTER clause struct to a string containing the clause
* in SFINDER input format and writes it to str.
************************************/
void convert_to_sfinder_clause(struct clause *cl,
			       int softweight,
			       dynstring *str) {

  dynstring fcl;      // hold the sfinder formatted clause
  struct literal *lit;
  struct term *positives[MAX_NUMBER_LITERALS];  // list of positive literals
  struct term *negatives[MAX_NUMBER_LITERALS];  // list of negative literals
  int pos_count = 0;
  int neg_count = 0;                            // current index counters
  int i;
  dynstring tmpStr;

  dynstring_initialise(&fcl, "");
  dynstring_initialise(&tmpStr, "");

  dynstring_sprintf(&fcl, "clause (%d) { ",softweight);
  for (lit = cl->first_lit; lit; lit = lit->next_lit) {
    if (lit->sign == POS) {
      positives[pos_count] = lit->atom;
      pos_count++;
    }
    else {
      negatives[neg_count] = lit->atom;
      neg_count++;
    }
  }

  for (i = 0; i < neg_count; i++) {
    convert_to_sfinder_term(negatives[i], PREDICATE, &tmpStr);
    dynstring_strcat(&fcl, &tmpStr);
    if (i != neg_count-1) dynstring_strcatc(&fcl, ", ");
  }
  dynstring_strcatc(&fcl, " -> ");
  for (i = 0; i < pos_count; i++) {
    convert_to_sfinder_term(positives[i], PREDICATE, &tmpStr);
    dynstring_strcat(&fcl, &tmpStr);
    if (i != pos_count-1) dynstring_strcatc(&fcl, "; ");
  }

  dynstring_strcatc(&fcl, ". }");  // close clause declaration
  dynstring_strcpy(str, &fcl);  // copy result into argument string

  dynstring_free(&fcl);
  dynstring_free(&tmpStr);
  return;
} /**************** end convert_to_sfinder_clause(...) ****************/



/***********************************
* Converts the OTTER term struct to a string containing the term
* in SFINDER input format and and writes it to str.
* complex tells the function if, should a complext be encountered,
* it is to be treated as a predicate or a function, i.e. whether
* this is the first or a recursive call.
* Recursively calls convert_to_sfinder_complex for predicates and functions.
************************************/
void convert_to_sfinder_term(struct term *t,
			     unsigned char complex,
			     dynstring *str) {

  dynstring ft;  // the sfinder formatted term
  dynstring tmpStr;   // temporary string

  dynstring_initialise(&ft, "");
  dynstring_initialise(&tmpStr, "");

  switch (t->type) {
  case NAME:          // constant
    dynstring_sprintf(&ft, "c_%s", sn_to_str(t->sym_num));
    break;
  case VARIABLE:      // variable
    dynstring_sprintf(&ft, "v%d", t->varnum);
    break;
  case COMPLEX:
    convert_to_sfinder_complex(t, complex, &tmpStr);
    dynstring_strcpy(&ft, &tmpStr);
    break;
  }
  dynstring_strcpy(str, &ft);  // copy result into argument string

  dynstring_free(&ft);
  dynstring_free(&tmpStr);
  
  return;
} /**************** end convert_to_sfinder_term(...) ****************/



/***********************************
* Converts the complex OTTER term to a SFINDER input formatted
* predicate or function and writes it to str.
* Recursively calls convert_to_sfiner_term for arguments.
************************************/
void convert_to_sfinder_complex(struct term *t,
				unsigned char complex,
				dynstring *str) {

  dynstring ft;      // the sfinder formatted complex
  dynstring tmpStr;       // temporary string
  struct rel *current_arg = t->farg;  // pointer to current argument

  dynstring_initialise(&ft, "");
  dynstring_initialise(&tmpStr, "");

  switch (complex) {
  case PREDICATE:
    // reserved characters
    if (!strcmp(sn_to_str(t->sym_num), "=") ||
	!strcmp(sn_to_str(t->sym_num), "equal")) {
      HASEQ = TRUE;
      dynstring_sprintf(&ft, "p_%s(", "eq");
    }
    else if (!strcmp(sn_to_str(t->sym_num), "!="))
      dynstring_sprintf(&ft, "p_%s(", "neq");
    else if (!strcmp(sn_to_str(t->sym_num), "<"))
      dynstring_sprintf(&ft, "p_%s(", "less");
    else if (!strcmp(sn_to_str(t->sym_num), ">"))
      dynstring_sprintf(&ft, "p_%s(", "greater");
    else if (!strcmp(sn_to_str(t->sym_num), "<="))
      dynstring_sprintf(&ft, "p_%s(", "less_eq");
    else if (!strcmp(sn_to_str(t->sym_num), ">="))
      dynstring_sprintf(&ft, "p_%s(", "greater_eq");
    else if (!strcmp(sn_to_str(t->sym_num), "=="))
      dynstring_sprintf(&ft, "p_%s(", "eq_eq");
    else if (!strcmp(sn_to_str(t->sym_num), "=/="))
      dynstring_sprintf(&ft, "p_%s(", "eq_sl_eq");

    // commonly used but conflicting with SFINDER special characters
    else if (!strcmp(sn_to_str(t->sym_num), "->"))
      dynstring_sprintf(&ft, "p_%s(", "implies");
    else if (!strcmp(sn_to_str(t->sym_num), "<-"))
      dynstring_sprintf(&ft, "p_%s(", "implies_reverse");
    else if (!strcmp(sn_to_str(t->sym_num), "-->"))
      dynstring_sprintf(&ft, "p_%s(", "impl");
    else if (!strcmp(sn_to_str(t->sym_num), "<--"))
      dynstring_sprintf(&ft, "p_%s(", "impl_reverse");
    else if (!strcmp(sn_to_str(t->sym_num), "<-->"))
      dynstring_sprintf(&ft, "p_%s(", "iff");
    else     // if no reserved character
      dynstring_sprintf(&ft, "p_%s(", sn_to_str(t->sym_num));
    break;
  case FUNCTION:
    //reserved characters
    if (!strcmp(sn_to_str(t->sym_num), "+"))
      dynstring_sprintf(&ft, "f_%s(", "plus");
    else if (  !strcmp(sn_to_str(t->sym_num), "-")  )
      dynstring_sprintf(&ft, "f_%s(", "minus");

    // OTTER predefined
    else if (!strcmp(sn_to_str(t->sym_num), "="))
      dynstring_sprintf(&ft, "f_%s(", "eq");
    else if (!strcmp(sn_to_str(t->sym_num), "!="))
      dynstring_sprintf(&ft, "f_%s(", "neq");
    else if (!strcmp(sn_to_str(t->sym_num), "<"))
      dynstring_sprintf(&ft, "f_%s(", "less");
    else if (!strcmp(sn_to_str(t->sym_num), ">"))
      dynstring_sprintf(&ft, "f_%s(", "greater");
    else if (!strcmp(sn_to_str(t->sym_num), "<="))
      dynstring_sprintf(&ft, "f_%s(", "less_eq");
    else if (!strcmp(sn_to_str(t->sym_num), ">="))
      dynstring_sprintf(&ft, "f_%s(", "greater_eq");
    else if (!strcmp(sn_to_str(t->sym_num), "=="))
      dynstring_sprintf(&ft, "f_%s(", "eq_eq");
    else if (!strcmp(sn_to_str(t->sym_num), "=/="))
      dynstring_sprintf(&ft, "f_%s(", "eq_sl_eq");

    // commonly used but conflicting with SFINDER special characters
    else if (!strcmp(sn_to_str(t->sym_num), "->"))
      dynstring_sprintf(&ft, "f_%s(", "implies");
    else if (!strcmp(sn_to_str(t->sym_num), "<-"))
      dynstring_sprintf(&ft, "f_%s(", "implies_reverse");
    else if (!strcmp(sn_to_str(t->sym_num), "-->"))
      dynstring_sprintf(&ft, "f_%s(", "impl");
    else if (!strcmp(sn_to_str(t->sym_num), "<--"))
      dynstring_sprintf(&ft, "f_%s(", "impl_reverse");
    else if (!strcmp(sn_to_str(t->sym_num), "<-->"))
      dynstring_sprintf(&ft, "f_%s(", "iff");
    
    else  // if no reserved character
      dynstring_sprintf(&ft, "f_%s(", sn_to_str(t->sym_num));
    break;
  }

  /*** loop thru argument list and construct sfinder string ***/
  while (current_arg != NULL) {
    convert_to_sfinder_term(current_arg->argval, FUNCTION, &tmpStr);
    dynstring_strcat(&ft, &tmpStr);
    if (current_arg->narg != NULL) dynstring_strcatc(&ft, ",");
    current_arg = current_arg->narg;  // go on to next argument
  }
  dynstring_strcatc(&ft, ")");  //  close argument list
  dynstring_strcpy(str, &ft);  // copy result into argument string

  dynstring_free(&ft);
  dynstring_free(&tmpStr);
  
  return;
} /**************** end convert_to_sfinder_complex(...) ****************/



/*
 * First mark all the buckets as empty.
 */

void init_buckets()
{
  int i;
  int val;

  for (i=0; i<BUCKETS; i++) 
    for (val=0; val<SEMANTIC_VALUES; val++) {
      bucket[i][val].first_cl = bucket[i][val].last_cl = 0;
      bucket[i][val].name[0] = 0;
    }
  for (val=0; val<SEMANTIC_VALUES; val++)
    FNE[val] = BUCKETS;
}



void put_in_bucket(struct clause *cl)
{
  dynstring clstr;
  int val;
  int b;
  boolean got_model;

  dynstring_initialise(&clstr, "");

  got_model = initial_model_found();
  b = cl->pick_weight;
  if (b >= BUCKETS) {
    b = BUCKETS-1;
    val = cl->sval = 0;
  }
  else if (b && got_model) {
    convert_to_sfinder_clause(cl,0,&clstr);
    CLOCK_START(FINDER_TIME);
    val = cl->sval = found(&clstr,0,0);
    CLOCK_STOP(FINDER_TIME);
  }
  else val = cl->sval = 0;
  if (bucket[b][val].first_cl) {
    bucket[b][val].last_cl->bucket_next_cl = cl;
    cl->bucket_prev_cl = bucket[b][val].last_cl;
  }
  else {
    bucket[b][val].first_cl = cl;
    cl->bucket_prev_cl = 0;
  }
  cl->bucket_next_cl = 0;
  bucket[b][val].last_cl = cl;
  cl->bucket_container = b;

  if (b && b < FNE[val])
    FNE[val] = b;

  dynstring_free(&clstr);
}


void remove_from_bucket(struct clause *cl)
{
  int b;
  int val;
  struct clause *clnxt;
  struct clause *clprev;

  if (!cl) return;

  b = cl->bucket_container;
  val = cl->sval;
  clnxt = cl->bucket_next_cl;
  clprev = cl->bucket_prev_cl;

  if (clnxt)
    clnxt->bucket_prev_cl = clprev;
  else
    bucket[b][val].last_cl = clprev;

  if (clprev)
    clprev->bucket_next_cl = clnxt;
  else
    bucket[b][val].first_cl = clnxt;

  if (!bucket[b][val].first_cl && FNE[val] == b) {
    while (FNE[val] < BUCKETS && !bucket[FNE[val]][val].first_cl)
      FNE[val]++;
  }
}



/*
 * Newly added clauses may be in bucket zero because their weight had not
 * been calculated at the time they were included. They should be put into
 * their correct buckets before a given clause is selected.
 */

void empty_bucket_zero()
{
  struct clause *cl;

  while ((cl = bucket[0][0].first_cl)) {
    remove_from_bucket(cl);
    put_in_bucket(cl);
  }
}



/*
 * After the model has been updated, lists of clauses (usually just Sos)
 * should be re-tested to get them into the right buckets
 */

void retest_list(struct list *lp)
{
  struct clause *cl;

  for (cl = lp->first_cl; cl; cl = cl->next_cl) {
    remove_from_bucket(cl);
    put_in_bucket(cl);
  }
}



void print_buckets()
{
  int i;
  struct clause *c;

  for (i=0; i<BUCKETS; i++) {
    if (bucket[i][FALSE_IN_MODEL].first_cl) {
      printf("%3d.FALSE: ",i);
      for (c = bucket[i][FALSE_IN_MODEL].first_cl; c; c = c->bucket_next_cl)
	printf("%d ",c->id);
      printf("\n");
    }
    if (bucket[i][TRUE_IN_MODEL].first_cl) {
      printf("%3d.TRUE: ",i);
      for (c = bucket[i][TRUE_IN_MODEL].first_cl; c; c = c->bucket_next_cl)
	printf("%d ",c->id);
      printf("\n");
    }
  }
}
