
/******************************************************************
*                                                                 *
* This file contains the SFINDER theory initiation function       *
* for Softie.                                                     *
*                                                                 *
******************************************************************/


#include "SOFTIE_SCOTT.h"


/**** keep track of constants/functions/predicates already written ****/
unsigned short constants[MAX_CONSTANTS];
unsigned short functions[MAX_FUNCTIONS];
unsigned short predicates[MAX_PREDICATES];

/**** counters to keep track of how full the arrays are ****/
int const_count = 0;
int fct_count = 0;
int pred_count = 0;

/**** SFINDER input file to be created ****/
FILE *sfinder_bool_complex_file;
FILE *sfinder_bool_constant_file;
FILE *sfinder_thing_complex_file;
FILE *sfinder_thing_constant_file;
FILE *sfinder_init_file;

/*** keep track of how many constant, predicate, function symbols ***/
static int const_num = 0;
static int pred_num = 0;
static int func_num = 0;

void transfer_input_file();


/***********************************
* Initializes SFINDER theory.
* Looks through all OTTER lists to find constants, functions, and predicates.
* Sets up model for initial UL.
************************************/
void init_softie_scott() {

  /********** read OTTER lists for predicate and function names **********/

  struct list *lists[6];  // hold OTTER's clause lists
  struct clause *current_cl;  // hold currently processed clause
  struct literal *current_lit;  // hold currently processed literal
  struct term *current_term;  // currently processed term
  int i;  // counter for the for loop
  dynstring tmpStr;  // SFINDER clause input string

  HASEQ = FALSE;
  dynstring_initialise(&tmpStr, "");

  /**** open file to write SFINDER formatted initializing input to ****/

  if ((sfinder_bool_complex_file = fopen("sfind_bcx.tmp", "w")) == NULL)
    fprintf(stderr,
	    "Error encountered while attempting to create temporary file.");
  if ((sfinder_bool_constant_file = fopen("sfind_bct.tmp", "w")) == NULL)
    fprintf(stderr,
	    "Error encountered while attempting to create temporary file.");
  if ((sfinder_thing_complex_file = fopen("sfind_tcx.tmp", "w")) == NULL)
    fprintf(stderr,
	    "Error encountered while attempting to create temporary file.");
  if ((sfinder_thing_constant_file = fopen("sfind_tct.tmp", "w")) == NULL)
    fprintf(stderr,
	    "Error encountered while attempting to create temporary file.");

  if ((sfinder_init_file = fopen("sfind_init.tmp", "w")) == NULL)
    fprintf(stderr,
	    "Error encountered while attempting to create temporary file.");

  /**** the OTTER clause lists to be checked for symbols ****/
  lists[0] = Usable;
  lists[1] = Sos;
  lists[2] = Demodulators;
  lists[3] = Passive;
  lists[4] = Hot;
  lists[5] = Hints;

  /**** write sort to file; cardinality is specified in SOFTIE_SCOTT.h ****/
  fprintf(sfinder_init_file,
	  "sort thing {cardinality = %d}\n\n",
	  SFINDER_SORT_CARDINALITY);

  for (i = 0; i <= 5; i++) {  // loop thru all lists
    for (current_cl = lists[i]->first_cl;
	 current_cl;
	 current_cl = current_cl->next_cl) {
      for (current_lit = current_cl->first_lit;
	   current_lit;
	   current_lit = current_lit->next_lit) {
	current_term = current_lit->atom;
	switch (current_lit->atom->type) {
	case NAME:
	  write_sfinder_constant(current_lit->atom, BOOL);
	  break;
	case VARIABLE:
	  break;  // do nothing for variables
	case COMPLEX:
	  write_sfinder_complex(current_lit->atom, PREDICATE);
	  break;
	}
      }
    }
  }
  transfer_input_file(sfinder_bool_complex_file,"sfind_bcx.tmp");
  transfer_input_file(sfinder_bool_constant_file,"sfind_bct.tmp");
  transfer_input_file(sfinder_thing_complex_file,"sfind_tcx.tmp");
  transfer_input_file(sfinder_thing_constant_file,"sfind_tct.tmp");

  /**** read UL and convert all clauses into SFINDER input format ****/

  if (HASEQ) {
    fprintf(sfinder_init_file, "\nfunction p_eq: thing,thing -> bool {}");
    fprintf(sfinder_init_file, "\nclause {(x p_eq y) = (x = y).}");
  }
  fprintf(sfinder_init_file, "\n");
  for (current_cl = Usable->first_cl;
       current_cl;
       current_cl = current_cl->next_cl) {
    convert_to_sfinder_clause(current_cl, 0, &tmpStr);
    fprintf(sfinder_init_file, "%s\n", tmpStr.c_str);
    current_cl->sval = TRUE_IN_MODEL;
  }

  /**** write settings to file ****/
  fprintf(sfinder_init_file, "\nsetting {\n");
  fprintf(sfinder_init_file, "	verbosity models none");
  fprintf(sfinder_init_file,
       "  time_limit: %d\n", SFINDER_TIME_LIMIT);
  /*
   *  fprintf(sfinder_init_file, "  backtracks 50000\n");
   */
  /*
   * fprintf(sfinder_init_file, "  insertions 100000\n");
   */
  fprintf(sfinder_init_file, "  random %ld\n", Stats[CL_GENERATED]);
  fprintf(sfinder_init_file, "  time_stamp\n");
  fprintf(sfinder_init_file, "  silent_breaks\n");
  fprintf(sfinder_init_file, "}\n");

  /**** write end to file ****/
  fprintf(sfinder_init_file, "\nend");

  fclose(sfinder_init_file);

  if (!init_finder("sfind_init.tmp")) {
      /*
        printf("\n\nFailed to model usable list: aborting!\n\n");
        exit(100);
      */
      printf("\n\nFailed to model usable list: disabling FINDER\n\n");
      set_no_semantics();
  }
  remove("sfind_init.tmp");

  init_dump_file();  // initialize runtime data dump

  dynstring_free(&tmpStr);
} /**************** end function init_softie_scott() ****************/


/*
 * Concatenate the contents (if any) of file f with name s onto the
 * end of the sfinder input file. Close and delete f
 */

void transfer_input_file(FILE *f, char *s)
{
  char line[10000];

  fclose(f);
  f = fopen(s,"r");
  while (fgets(line,10000,f))
    fputs(line,sfinder_init_file);
  fclose(f);
  remove(s);
}


/***********************************
* Writes SFINDER formated constant to given file.
* Passed term is assumed to be a constant. -> not checked for
* The argument type specifies whether this is a bool or thing constant.
************************************/
void write_sfinder_constant(struct term *constant, unsigned char type) {

  if (!is_used(constant->sym_num, CONSTANT)) {  // if not yet written
    const_num++;  // keep count of symbols
    switch (type) {
    case BOOL:
      fprintf(sfinder_bool_constant_file,
	      "constant c_%s bool.\n",
	      sn_to_str(constant->sym_num));
      break;
    case THING:
      fprintf(sfinder_thing_constant_file,
	      "constant c_%s thing.\n",
	      sn_to_str(constant->sym_num));
      break;
    }

    constants[const_count] = constant->sym_num;
    const_count++;
  }

} /**************** end function write_sfinder_constant(...) ****************/



/***********************************
* Writes SFINDER formated predicate/function to given file.
* Passed term is assumed to be a predicate/function. -> not checked for
* Argument type specifies whether complex is PREDICATE or FUNCTION.
************************************/
void write_sfinder_complex(struct term *compl, unsigned char type) {

  struct rel *args = compl->farg;  // the argument list
  dynstring tmpStr;  // temporary predicate declaration
  dynstring declaration;  // declaration for file

  /*** keep track of whether this symbol has been written yet ***/
  /*** until stack unwinds back to here when actual file entry written ***/
  BOOLEAN already_written = FALSE;

  unsigned short *comp_array = NULL;
  int *comp_count = NULL;

  dynstring_initialise(&tmpStr, "");
  dynstring_initialise(&declaration, "");

  switch (type) {
  case PREDICATE:
    comp_array = predicates;
    comp_count = &pred_count;

    // OTTER predefined
    if (!strcmp(sn_to_str(compl->sym_num), "=") ||
	!strcmp(sn_to_str(compl->sym_num), "equal")) {
      HASEQ = TRUE;
      dynstring_sprintf(&tmpStr, "function p_%s: ", "eq");
    }
    else if (!strcmp(sn_to_str(compl->sym_num), "!="))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "neq");
    else if (!strcmp(sn_to_str(compl->sym_num), "<"))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "less");
    else if (!strcmp(sn_to_str(compl->sym_num), ">"))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "greater");
    else if (!strcmp(sn_to_str(compl->sym_num), "<="))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "less_eq");
    else if (!strcmp(sn_to_str(compl->sym_num), ">="))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "greater_eq");
    else if (!strcmp(sn_to_str(compl->sym_num), "=="))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "eq_eq");
    else if (!strcmp(sn_to_str(compl->sym_num), "=/="))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "eq_sl_eq");

    // commonly used but conflicting with SFINDER format
    else if (!strcmp(sn_to_str(compl->sym_num), "->"))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "implies");
    else if (!strcmp(sn_to_str(compl->sym_num), "<-"))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "implies_reverse");
    else if (!strcmp(sn_to_str(compl->sym_num), "-->"))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "impl");
    else if (!strcmp(sn_to_str(compl->sym_num), "<--"))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "impl_reverse");
    else if (!strcmp(sn_to_str(compl->sym_num), "<-->"))
      dynstring_sprintf(&tmpStr, "function p_%s: ", "iff");

    else
      dynstring_sprintf(&tmpStr, "function p_%s: ", sn_to_str(compl->sym_num));
    break;

  case FUNCTION:
    comp_array = functions;
    comp_count = &fct_count;

    if (!strcmp(sn_to_str(compl->sym_num), "+"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "plus");
    else if (!strcmp(sn_to_str(compl->sym_num), "-"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "minus");
      
    // OTTER predefined
    else if (!strcmp(sn_to_str(compl->sym_num), "="))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "eq");
    else if (!strcmp(sn_to_str(compl->sym_num), "!="))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "neq");
    else if (!strcmp(sn_to_str(compl->sym_num), "<"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "less");
    else if (!strcmp(sn_to_str(compl->sym_num), ">"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "greater");
    else if (!strcmp(sn_to_str(compl->sym_num), "<="))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "less_eq");
    else if (!strcmp(sn_to_str(compl->sym_num), ">="))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "greater_eq");
    else if (!strcmp(sn_to_str(compl->sym_num), "=="))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "eq_eq");
    else if (!strcmp(sn_to_str(compl->sym_num), "=/="))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "eq_sl_eq");
      
    // commonly used but conflicting with SFINDER format
    else if (!strcmp(sn_to_str(compl->sym_num), "->"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "implies");
    else if (!strcmp(sn_to_str(compl->sym_num), "<-"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "implies_reverse");
    else if (!strcmp(sn_to_str(compl->sym_num), "-->"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "impl");
    else if (!strcmp(sn_to_str(compl->sym_num), "<--"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "impl_reverse");
    else if (!strcmp(sn_to_str(compl->sym_num), "<-->"))
      dynstring_sprintf(&tmpStr, "function f_%s: ", "iff");

    else   // if no reserved character
      dynstring_sprintf(&tmpStr, "function f_%s: ", sn_to_str(compl->sym_num));
    break;
  }

  dynstring_strcat(&declaration, &tmpStr);

  if (is_used(compl->sym_num, type))
    already_written = TRUE;
  else {
    comp_array[*comp_count] = compl->sym_num;
    (*comp_count)++;
  }

  while (args != NULL) {  /**** while not at end of argument list ****/
    switch (args->argval->type) {
    case NAME:
      write_sfinder_constant(args->argval, THING);
      break;
    case VARIABLE:
      break;  // do nothing for variables
    case COMPLEX:
      write_sfinder_complex(args->argval, FUNCTION);
      break;
    }

    // write argument type to file
    dynstring_strcatc(&declaration, "thing");

    if (args->narg != NULL) {
      // comma if more to come
      dynstring_strcatc(&declaration, ", ");
    }

    args = args->narg;
  } /**** end while loop thru arg list ****/

  switch (type) {
  case PREDICATE:
    dynstring_sprintf(&tmpStr, " -> bool {no_priority}");
    break;
  case FUNCTION:
    dynstring_sprintf(&tmpStr, " -> thing {no_priority}");
    break;
  }
  dynstring_strcat(&declaration, &tmpStr);

  /*** only write to file if not yet written ***/
  if (!already_written) {
    if (type == PREDICATE) {
      pred_num++;  // keep count of symbols
      fprintf(sfinder_bool_complex_file, "%s\n", declaration.c_str);
    }
    if (type == FUNCTION) {
      func_num++;  // keep count of symbols
      fprintf(sfinder_thing_complex_file, "%s\n", declaration.c_str);
    }
  }

  dynstring_free(&tmpStr);
  dynstring_free(&declaration);

} /**************** end function write_sfinder_complex(...) ****************/



/***********************************
* Checks if this symbol (type = CONSTANT, FUNCTION, or PREDICATE) has
* been written before (according to the respective list)
************************************/
BOOLEAN is_used(unsigned char symbol, char type) {

  int i;
  unsigned short *array = NULL;  // respective distinct symbol array
  int counter = 0;  // respective array counter

  switch(type) {
  case CONSTANT:
    array = constants; counter = const_count;
    break;
  case FUNCTION:
    array = functions; counter = fct_count;
    break;
  case PREDICATE: array = predicates; counter = pred_count;
    break;
  }

  for (i = 0; i < counter; i++)
    if (symbol == array[i])
      return TRUE; 

  return FALSE;

} /**************** end function is_used(...) ****************/



/***********************************
* Returns number of distinct symbols of symbol type specified.
* Symbol can be CONSTANT, PREDICATE, or FUNCTION.
************************************/
int get_count(unsigned char type) {
  switch (type) {
  case CONSTANT:
    return const_num;
  case PREDICATE:
    return pred_num;
  case FUNCTION:
    return func_num;
  }
  return -1;  // error in symbol type	
} /**************** end function get_count(...) ****************/

