
/*******************************************************************
*											 *
* This file contains functions for statistics tracking and output. *
*											 *
*******************************************************************/


#include <math.h>  /* for some square rooting */
#include "SOFTIE_SCOTT.h"


/***********************************
* Print statistics about the initial clause sets.
* To be called before starting OTTER's main loop.
************************************/
void print_initial_softie_stats() {

	/*** compute stats ***/

  int pred_count;
  int func_count;
  int const_count;
  int total_count;
  int percentage_symbols_in_UL;
  int usable_count;
  int total_clause_count;
  int percent_clauses_in_UL;
  int percent_short_clauses;

  pred_count = get_count(PREDICATE);
  func_count = get_count(FUNCTION);
  const_count = get_count(CONSTANT);
  total_count = pred_count + func_count + const_count;
  get_UL_symbol_counts();
  percentage_symbols_in_UL = get_percentage();
  usable_count = get_clause_count(Usable);
  total_clause_count =
    usable_count + 
    get_clause_count(Sos) + 
    get_clause_count(Demodulators) +
    get_clause_count(Passive) +
    get_clause_count(Hot) +
    get_clause_count(Hints);
  percent_clauses_in_UL =
    (int)
    (100 *
     (
      ((double) usable_count) /
      ((double) total_clause_count)
      )
     );
  percent_short_clauses =
    (int)
    (100 *
     (
      (
       ((double) get_number_short_clauses(Usable)) + 
       ((double) get_number_short_clauses(Sos)) +
       ((double) get_number_short_clauses(Demodulators)) +
       ((double) get_number_short_clauses(Passive)) +
       ((double) get_number_short_clauses(Hot)) +
       ((double) get_number_short_clauses(Hints))
       ) /
      ((double) total_clause_count)
      )
     );

  /*** print stats ***/

  printf("\n\n-------------- Softie stats --------------\n\n");

  printf("UPDATE_STOP: %d\n", UPDATE_STOP);
  printf("SFINDER_TIME_LIMIT: %d\n", SFINDER_TIME_LIMIT);
  printf("SHORT_CLAUSE_CUTOFF: %d\n", SHORT_CLAUSE_CUTOFF);

  printf("number of clauses in intial UL: %d\n",
	 usable_count);
  printf("number of clauses initially in problem: %d\n",
	 total_clause_count);
  printf("percentage of clauses intially in UL: %d\n",
	 percent_clauses_in_UL);
  printf("percentage of distinct symbols occuring in initial UL: %d\n",
	 percentage_symbols_in_UL);
  printf("percent of all initial clauses that are short: %d\n",
	 percent_short_clauses);
  printf("absolute distinct symbol count: %d\n",
	 total_count);
  printf("   distinct predicate count: %d\n",
	 pred_count);
  printf("   distinct function count: %d\n",
	 func_count);
  printf("   distinct constant count: %d\n\n",
	 const_count);

  printf("---------- no more Softie stats ----------\n\n\n");

} /************ end function print_initial_softie_stats(...) ************/



/***********************************
* Returns number of clauses in given list.
************************************/
int get_clause_count(struct list *clause_list) {

  int counter= 0;

  struct clause *current_cl = clause_list->first_cl;
  while (current_cl != NULL) {
    counter++;
    current_cl = current_cl->next_cl;
  }
  return counter;

} /**************** end function get_clause_count(...) ****************/




int sym_count;

/***********************************
* Returns number of short clauses in all lists.
************************************/
int get_number_short_clauses(struct list *clause_list) {

  struct literal *current_lit;
  struct clause *current_cl = clause_list->first_cl;
  struct term *current_term;

  int counter = 0;
  short_tested_syms_index = 0;

  while (current_cl != NULL) {
    // reset
    current_lit = current_cl->first_lit;	
    sym_count = 0;
    while (current_lit != NULL) {
      current_term = current_lit->atom;
      if (current_term->type == COMPLEX) count_shorts(current_term);
      current_lit = current_lit->next_lit;
    }
    if (sym_count <= SHORT_CLAUSE_CUTOFF) counter++;
    current_cl = current_cl->next_cl;  // go on to next clause in list
  }
  return counter;

} /**************** end function get_number_short_clauses(...) ****************/



void count_shorts(struct term *current_term) {

  struct rel *args = current_term->farg;  // the argument list

  if (!symbol_short_counted(current_term->sym_num)) sym_count++;
  while (args != NULL) {
    if (args->argval->type == COMPLEX) count_shorts(args->argval);
    args = args->narg;  // next argument
  }

} /**************** end function count_shorts(...) ****************/



boolean symbol_short_counted(unsigned short sym_num) {

  int i;

  for (i = 0; i < MAX_SYMBOLS; i++) {
    if (sym_num == short_tested_symbols[i]) return true;
  }
  short_tested_syms_index++;
  short_tested_symbols[short_tested_syms_index] = sym_num;
  return false;

} /**************** end function symbol_short_counted(...) ****************/



/***********************************
* Returns true if model should still be updated,
* false otherwise. Currently, false is returned after
* a fixed number of given clauses have been selected,
* but the criteria for stopping to generate new models
* can be changed by rewriting this function.
************************************/
boolean still_update_model() {
  static int num_sel_giv_cls = 0;
  static int stopped_generating = 0;

  if (stopped_generating)
      return false;

  num_sel_giv_cls++;

  if ((Parms[MAX_SECONDS].val <= 0 && num_sel_giv_cls >= UPDATE_STOP) ||
      (Parms[MAX_SECONDS].val > 0 && clock_val(FINDER_TIME) > Parms[MAX_SECONDS].val * 500)) {
      printf("\nModelling stopped after %d given clauses and %0.2f seconds\n\n",
             num_sel_giv_cls, clock_val(FINDER_TIME) / 1000.0);
      stopped_generating = 1;
      return false;
  }
  return true;
} /**************** end function still_update_model(...) ****************/



// returns percentage of UL symbols of total problem symbol count
int get_percentage() {
  return
    (int)
    (100 *
     (
      ((double) symbol_count) /
      ((double)
       (
	get_count(CONSTANT) +
	get_count(PREDICATE) +
	get_count(FUNCTION)
	)
       )
      )
     );
} /**************** end function still_update_model() ****************/



/***********************************
* Determines number of different symbol types currently in UL
* and stores the number in the appropriate variables.
************************************/
int get_UL_symbol_counts() {

  struct clause *current_cl;
  struct literal *current_lit;
  struct term *current_term;

  sym_index = 0;
  symbol_count = 0;

  current_cl = Usable->first_cl;
  while (current_cl != NULL) {
    current_lit = current_cl->first_lit;
    while (current_lit != NULL) {
      current_term = current_lit->atom;
      switch (current_lit->atom->type) {
      case NAME: count_constant(current_term); break;
      case COMPLEX: count_complex(current_term); break;
      }
      current_lit = current_lit->next_lit;
    }
    current_cl = current_cl->next_cl;
  }
	return symbol_count;

} /************* end function get_UL_symbol_counts() *************/



/***********************************
* Includes a constant in count if not yet counted.
************************************/
void count_constant(struct term *current_term) {
  int i;

  if (!is_counted(current_term->sym_num)) {
    symbol_count++;
    symbols[sym_index].symbol = current_term->sym_num;
    symbols[sym_index].count = 1;
    sym_index++;
  }
  else {
    // find right place
    for (i = 0; symbols[i].symbol != current_term->sym_num; i++) {}
    symbols[i].count++;
  }
} /**************** end function count_constant(...) ****************/


/***********************************
* Includes a complex in count if not yet counted.
************************************/
void count_complex(struct term *current_term) {
  int i;

  struct rel *args = current_term->farg;  // the argument list

  if (!is_counted(current_term->sym_num)) {
    symbol_count++;
    symbols[sym_index].symbol = current_term->sym_num;
    symbols[sym_index].count = 1;
    sym_index++;
  }
  else {
    // find right place
    for (i = 0; symbols[i].symbol != current_term->sym_num; i++) {}
    symbols[i].count++;
  }

  while (args != NULL) {
    switch (args->argval->type) {
    case NAME: count_constant(args->argval); break;
    case COMPLEX: count_complex(args->argval); break;
    }
    args = args->narg;
  }
} /**************** end function count_complex(...) ****************/


/***********************************
* Returns truth value of whether the specified symbols has already been
* counted or not.
************************************/
boolean is_counted(int sym_num) {
  int i;
  for (i = 0; i <= sym_index; i++) {
    if (sym_num == symbols[i].symbol) return true;
  }
  return false;
} /**************** end function is_counted(...) ****************/






/***
*-*-*-*-*-*-*-* Everything from here is for the runtime
                data dumping for anim illustration.     *-*-*-*-*-*-*-*
***/


FILE *dump_file = NULL;
int true_given = 0;           // number of clauses true in current model
int sos_clauses = 0;          // number of clauses in set of support
int subsumed_clauses = 0;     // total number of deleted clauses

// defined in SOFITE_SCOTT.h:
// model_clauses              // no. clauses on which current model based
// false_given;               // number of given clauses false in model


/***********************************
* Initializes dump file by creating it.
************************************/
void init_dump_file() {
/*----
  ----No dump for CASC
  if ((dump_file = fopen("runtime_data.dump", "w")) == NULL)
    fprintf(stderr,
	    "Error encountered while attempting to create runtime ");
    fprintf(stderr,
	    "data dump file.\n");
----*/
  false_given = 0;
  model_clauses = 0;
} /**************** end init_dump_file() ****************/



/***********************************
* Ends and closes dump file.
************************************/
void close_dump_file() {
  if (dump_file) {
    fprintf(dump_file, "-1\n.");  // write closing statement
    fprintf(dump_file, "\n1. Given clauses\n");
    fprintf(dump_file, "2. False given clauses\n");
    fprintf(dump_file, "3. Semantically affected choices\n");
    fprintf(dump_file, "4. SOS size\n");
    fprintf(dump_file, "5. Clauses subsumed\n");
    fprintf(dump_file, "6. Clauses generated\n");
    fprintf(dump_file, "7. Clauses generating latest model\n");
    fclose(dump_file);
  }
} /**************** end close_dump_file() ****************/



/***********************************
* Writes dump stats to respective variables.
************************************/
void update_dump_stats() {
  // updated in find_given_clause function in clause.c:   false_given
  // updated in found function in FINDER.c:               model_clauses

  true_given = Stats[CL_GIVEN] - false_given;
  sos_clauses = Stats[SOS_SIZE];
  subsumed_clauses = Stats[CL_FOR_SUB] + Stats[CL_BACK_SUB];
} /**************** end update_dump_stats() ****************/


/***********************************
* Writes the runtime stats to the dump file.
************************************/
void dump_runtime_data() {

  update_dump_stats();  // update stats variables

  if (dump_file) {
  fprintf(dump_file, "%ld %d %ld %d %d %ld %d\n",
	  Stats[CL_GIVEN],
	  false_given,
	  Stats[GIVEN_AFFECTED],
	  sos_clauses,
	  subsumed_clauses,
          Stats[CL_GENERATED],
	  model_clauses);
  }
  return;

} /**************** end dump_runtime_data() ****************/


void output_soft_stats(FILE *fp)
{
  int i, j, k;
  int sostrue=0;
  int sosfalse=0;
  struct clause *cl;

  fprintf(fp,"\n----------- soft-scott stats ----------\n\n");
  fprintf(fp,"true clauses given %11ld      (%0.1f%c)\n",
	  Stats[CL_GIVEN]-false_given,
	  (100.0 * (Stats[CL_GIVEN]-false_given))/Stats[CL_GIVEN],
	  '%');
  fprintf(fp,"false clauses given %10d\n", false_given);
  fprintf(fp,"\n      FALSE     TRUE\n");
  for (i=1; i<BUCKETS; i++)
    if (bucket[i][0].first_cl || bucket[i][1].first_cl) {
      for (j=0,cl=bucket[i][0].first_cl; cl; cl = cl->bucket_next_cl) j++;
      for (k=0,cl=bucket[i][1].first_cl; cl; cl = cl->bucket_next_cl) k++;
      fprintf(fp,"%4d  %-9d %d\n",i,j,k);
      sosfalse += j;
      sostrue += k;
    }
  fprintf(fp,"tot:  %-9d %d      (%0.1f%c true)\n\n",
	  sosfalse,sostrue,(100.0*sostrue)/(sostrue+sosfalse),'%');
}
