------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Sat
  ( SolverM       -- :: * -> *; Monad
  , try           -- :: SolverM a -> SolverM (Maybe a)
  , lift          -- :: IO a -> SolverM a
  , run           -- :: SolverM a -> IO (Maybe a)
  , contradiction -- :: SolverM a

  , Var           -- :: *; Eq, Ord
  , Lit           -- :: *; Eq
  , var           -- :: Lit -> Var
  , lit           -- :: Var -> Lit
  , neg           -- :: Lit -> Lit
  , sign          -- :: Lit -> Bool
		  
  , addVar        -- :: SolverM Var
  , addVarP       -- :: Int -> SolverM Var
  , addVars       -- :: Int -> SolverM [Var]
  , addVarsP      -- :: Int -> Int -> SolverM [Var]
  , addClause     -- :: [Lit] -> SolverM ()
  , simplify      -- :: SolverM ()
  , solve         -- :: [Lit] -> SolverM Bool
  , dimacs        -- :: String -> [Lit] -> SolverM ()
  , value         -- :: Var   -> SolverM Bool
  , values        -- :: [Var] -> SolverM [Bool]
  
  , getNClauses   -- :: SolverM Int
  , getNVars      -- :: SolverM Int
  , getNConflicts -- :: SolverM Int

  , setVerbose    -- :: Bool -> SolverM ()

  , identify      -- :: IO ()
  )
 where

import Foreign
import Foreign.C.String
import CTypes
import IO ( hFlush, stdout )

-------------------------------------------------------------------------
-- types

type Var
  = Int -- CUInt

newtype Lit
  = MkLit CUInt
 deriving Eq

newtype Solver
  = MkSolver (Ptr ())

-------------------------------------------------------------------------
-- solver monad

newtype SolverM a
  = MkSolverM (forall b . Solver -> IO b -> (a -> IO b) -> IO b)

instance Monad SolverM where
  return x =
    MkSolverM (\_s _fail ok -> ok x)
    
  MkSolverM m >>= k =
    MkSolverM (\s fail ok ->
      m s fail (\y ->
        let MkSolverM m' = k y in
          m' s fail ok
      )
    )

-------------------------------------------------------------------------
-- primitive operations

try :: SolverM a -> SolverM (Maybe a)
try (MkSolverM m) =
  MkSolverM (\s _fail ok ->
    m s (ok Nothing) (ok . Just)
  )

solver :: SolverM Solver
solver = MkSolverM (\s _fail ok -> ok s)

lift :: IO a -> SolverM a
lift io = MkSolverM (\_s _fail ok -> io >>= ok)

withSolver :: (Solver -> IO a) -> SolverM a
withSolver f = MkSolverM (\s _fail ok -> f s >>= ok)

contradiction :: SolverM a
contradiction = MkSolverM (\_s fail _ok -> fail)

run :: SolverM a -> IO (Maybe a)
run (MkSolverM m) =
  do s <- solver_new
     let fail =
           do solver_free s
              return Nothing
         
         ok a =
           do solver_free s
              return (Just a)
     
      in m s fail ok

-------------------------------------------------------------------------
-- derived operations

addClause :: [Lit] -> SolverM ()
addClause lits =
  do r <- withSolver $ \s ->
       do solver_clause_begin s
          mapM_ (solver_clause_add_lit s) lits
          solver_clause_commit s
     if r == 1
       then return ()
       else contradiction

addVar :: SolverM Var
addVar = withSolver $ \s -> solver_add_var s

addVarP :: Int -> SolverM Var
addVarP i = withSolver $ \s -> solver_add_varp s (fromIntegral i)

addVars :: Int -> SolverM [Var]
addVars n = sequence (replicate n addVar)

addVarsP :: Int -> Int -> SolverM [Var]
addVarsP i n = sequence (replicate n (addVarP i))

simplify :: SolverM ()
simplify =
  do r <- withSolver solver_simplify
     if r == 0
       then contradiction
       else return ()

solve :: [Lit] -> SolverM Bool
solve lits =
  withSolver $ \s -> 
    do solver_prefix_clear s
       mapM_ (solver_prefix_add_lit s) lits
       r <- solver_solve s
       return (r == 1)

dimacs :: String -> [Lit] -> SolverM ()
dimacs file lits =
  withSolver $ \s -> 
    do solver_prefix_clear s
       mapM_ (solver_prefix_add_lit s) lits
       withCString file (solver_dimacs s)

value :: Var -> SolverM Bool
value v = 
  withSolver $ \s -> 
    do r <- solver_model_value s (lit v)
       return (r == 1)

values :: [Var] -> SolverM [Bool]
values vs = 
  withSolver $ \s -> 
    do let val v =
             do r <- solver_model_value s (lit v)
                return (r == 1)
       mapM val vs

getNClauses, getNVars, getNConflicts :: SolverM Int
getNClauses   = withSolver $ \s -> fromIntegral `fmap` solver_get_nof_clauses s
getNVars      = withSolver $ \s -> fromIntegral `fmap` solver_get_nof_vars s
getNConflicts = withSolver $ \s -> fromIntegral `fmap` solver_get_nof_conflicts s

setVerbose :: Bool -> SolverM ()
setVerbose b = withSolver $ \s -> solver_set_verbose s (if b then 1 else 0)

identify :: IO String
identify =
  do cs <- prim_identify
     peekCString cs

--------------------------------------------------------------------------------
-- foreign imports

foreign import ccall "_lit"  lit              :: Var -> Lit
foreign import ccall "_var"  var              :: Lit -> Var
foreign import ccall "_neg"  neg              :: Lit -> Lit
foreign import ccall "_sign" sign             :: Lit -> Bool
					      
foreign import ccall solver_new               :: IO Solver
foreign import ccall solver_free              :: Solver -> IO ()
					      
foreign import ccall solver_add_var           :: Solver -> IO Var
foreign import ccall solver_add_varp          :: Solver -> CInt -> IO Var
					      
foreign import ccall solver_clause_begin      :: Solver -> IO ()
foreign import ccall solver_clause_add_lit    :: Solver -> Lit -> IO ()
foreign import ccall solver_clause_commit     :: Solver -> IO CUInt
					      
foreign import ccall solver_prefix_clear      :: Solver -> IO ()
foreign import ccall solver_prefix_add_lit    :: Solver -> Lit -> IO ()
					      
foreign import ccall solver_solve             :: Solver -> IO CUInt
foreign import ccall solver_dimacs            :: Solver -> CString -> IO ()
foreign import ccall solver_simplify          :: Solver -> IO CUInt
foreign import ccall solver_model_value       :: Solver -> Lit -> IO CUInt

foreign import ccall solver_get_nof_clauses   :: Solver -> IO CUInt
foreign import ccall solver_get_nof_vars      :: Solver -> IO CUInt
foreign import ccall solver_get_nof_conflicts :: Solver -> IO CUInt
foreign import ccall solver_set_verbose       :: Solver -> CUInt -> IO ()

foreign import ccall "identify" prim_identify :: IO CString

--------------------------------------------------------------------------------
-- the end.
