/* ---------------------------------------------------------------
-- Satnik, a SAT solver                                         --
-- Copyright (c) 2003 Niklas Srensson                          --
-- nik@cs.chalmers.se                                           --
--                                                              --
-- This file is part of Satnik.                                 --
--                                                              --
-- Satnik is free software; you can redistribute it and/or      --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Satnik is distributed in the hope that it will be useful,    --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Satnik; if not, write to the Free         --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
--------------------------------------------------------------- */

#include <stdlib.h>
#include <math.h>

#include "Satnik.h"

namespace Satnik {

  /*******************************************************************
   ** Assumes the literal <l> in the current context. Does nothing if
   ** the literal already is true. Otherwise it stores the assigment
   ** on the <assignments> vector and sets the value, updates the
   ** relevant information (reason/level/value) if the assignment was
   ** not contradictory. Returns wether it gave a conflict or not. A
   ** constradiction should be managed directly, most methods relies
   ** on the fact that the solver is in a non contradictory state.
   ******************************************************************/
  bool Solver::assign(Lit l, Constraint* from) {
    //cout << "Assign " << l << " @ " << current_level() << '\n';
    if(values(l))
      return true;  // an old assignment 
    else {
      assigns.push_back(l);
      values(l)      = true;

      if(values(neg(l)))
	return false;

      num_assign++;
      level (var(l)) = current_level();
      reason(var(l)) = from;

      queue.push_back(l);

      return true;
    }
  }

  //void apa(vector<Constraint*>& w, size_t i) { w.erase(w.begin(), w.begin() + i); }
  /*******************************************************************
   **  Propagates all literals in the <queue>. The method return true
   **  if no conflict is found, false otherwise. <conflict> is set to
   **  the conflicting consraint, and the conflicting literal is
   **  the last in the <assignment> vector. The <queue> is always
   **  empty afterwards.
   ******************************************************************/
#define NEWPROP
  bool Solver::propagate(void) {
#ifdef NEWPROP
    bool                result = true;
    while(queue.size() > 0) {
      Lit l  = queue.front(); queue.pop_front(); 
      size_t limit = watches(l).size();
      size_t i;
      //cout << "propagate : " << l << '\n';
      for(i = 0; i < limit; i++) {
	//cout << "propagate: " << l << " in " << *((Clause*)watches(l)[i]) << '\n';
	
	if(!(watches(l)[i]->propagate(l,*this))) { 
	  //cout << "conflict!\n";
	  queue.clear(); 
	  conflict = watches(l)[i];
	  result = false;
	  i++; break;
	}
      }
      //cout << "result ? " << result << '\n';
      //cout << "wsize: " << watches(l).size() << " i = " << i << '\n';
      //for(size_t j = 0; j < watches(l).size() - i; j++) watches(l)[j] = watches(l)[j+i];
      //watches(l).resize(watches(l).size() - i);
      watches(l).erase(watches(l).begin(), watches(l).begin() + i);
      //apa(watches(l),i);
    }
    //cout << "done!\n";
    return result;
#else
    vector<Constraint*> cs;
    while(queue.size() > 0) {
      Lit l  = queue.front(); queue.pop_front(); 
      cs = watches(l);
      //cs.clear(); for(size_t i = 0; i < watches(l).size(); i++) cs.push_back(watches(l)[i]);

      watches(l).clear();
      for(size_t i = 0; i < cs.size(); i++)
	if(!(cs[i]->propagate(l,*this))) { 
	  //if(!((Clause*)cs[i])->propagate_(l,*this)) {  // non-virtual call!
	  for(size_t j = i + 1; j < cs.size(); j++) add_watch(l, cs[j]);
	  queue.clear(); conflict = cs[i];
	  return false; 
	}
    }
    return true;
#endif
  }

  /*******************************************************************
   ** Push the current point in the search tree on the undo
   ** stack. 
   ******************************************************************/
  void Solver::push(void) {
    assert(queue.size() == 0);
    frames.push_back(make_pair(assigns.size(), undos.size()));
  }

  /*******************************************************************
   ** Pop the previous point in the search tree from the undo
   ** stack. 
   ******************************************************************/
  void Solver::pop(void) {
    assert(queue.size() == 0);
    size_t prev_assigns_frame  = frames.back().first;
    size_t prev_undo_frame     = frames.back().second;

    //    if(current_level() - root_level < 10) {
//      Var   decision          = var(assigns[prev_assigns_frame]);
//      float diff              =  (float)assigns.size() - prev_assigns_frame - 1;
//      float exp               = -(float)current_level() - root_level;
//      float activate          =  diff * pow(2,exp);
//					   
//      var_activity[decision] += activate;
//      cout << 
//	"activate = " << activate << 
//	" exp  = "    << exp      <<
//	" diff = "    << diff     <<
//	'\n';
      //}

    frames.pop_back();


    for(size_t i = prev_assigns_frame; i < assigns.size(); i++) { 
      Lit l         = assigns[i]; 
      next_decision = min(next_decision,var_order_index[var(l)]);
      values(l)     = false; 
    }
    assigns.resize(prev_assigns_frame); 
    
    for(size_t i = prev_undo_frame; i < undos.size(); i++)  undos[i]->pop();

    undos.resize(prev_undo_frame);
  }
  
  /*******************************************************************
   ** Create completely fresh variables.
   ******************************************************************/
  Var Solver::add_vars(int num) {
    Var v = next_var; next_var = next_var+num;
    var_reason     .resize(var_reason     .size() + num  , NULL);
    var_level      .resize(var_level      .size() + num  , 0);
    var_activity   .resize(var_activity   .size() + num  , 0.0);
    var_polarity   .resize(var_polarity   .size() + num  , 0.0);
    var_order_index.resize(var_order_index.size() + num  , 0);
    lit_values     .resize(lit_values     .size() + num*2, false);
    lit_watches    .resize(lit_watches    .size() + num*2);
    lit_visited    .resize(lit_visited    .size() + num*2, false);
    return v;
  }

  /*******************************************************************
   ** Check if the literal <l> is visited. If not, mark is as visited
   ** and add it to <conflict_clause> if its level is less than the
   ** current.
   ******************************************************************/
  void Solver::analyze_lit(Lit l, vector<Lit>&  conflict_clause) {
    if(!values(l)) { cout << "Hmm? " << l << '\n'; }
    assert(values(l)); 
    //printf("Checking %c%d ", sign(l) ? '+' : '-', var(l));
    if(!visited(l)) { 
      //printf("not visited.\n");
      if(level(var(l)) < current_level()) { 
	conflict_clause.push_back(neg(l));  
      } else { 
        assert(level(var(l)) == current_level()); 
        num_marked++; 
      } 
      visited(l) = true;
    }
    //else printf("visited.\n");

  }

  /*******************************************************************
   ** Compute the reason for <conflict_set> using the standard first
   ** UIP algorithm. The method returns the level where the conflict
   ** occured, which can be used for non chronological backtracking.
   ** 
   ** The first two literals of <conflict_clause> are of level larger
   ** than all the others, which makes them suitable for being
   ** watched in that clause.
   ******************************************************************/
  int Solver::analyze(const vector<Lit>& conflict_set, vector<Lit>& conflict_clause) {
    assert(num_marked == 0);
    for(size_t j = 0; j < conflict_set.size(); j++) 
      analyze_lit(conflict_set[j], conflict_clause);
    vector<Lit> r;
    size_t      i;
    for(i = assigns.size() - 2; i >= frames.back().first; i--) {
      //printf("*** i = %d, num_marked = %d\n", i, num_marked);
      Lit l = assigns[i];
      if(visited(l)) {
	visited(l) = false;
	//printf("--- i = %d, num_marked = %d\n", i, num_marked);
	if(--num_marked == 0) {
	  //if(i != frames.back().first) 
	  //printf("Got here %d!\n", i - frames.back().first);
	  for(size_t j = 0; j < conflict_clause.size(); j++) {
	    visited(neg(conflict_clause[j])) = false;
	  }
	  conflict_clause.push_back(neg(assigns[i]));
	  swap(conflict_clause.front(), conflict_clause.back());
	  break; // UIP
	}
	assert(reason(var(l)) != NULL);
	r.clear(); reason(var(l))->reason(l,r);
	for(size_t j = 0; j < r.size(); j++) analyze_lit(r[j], conflict_clause);
      }
    }
    //printf("num_marked = %d\n", num_marked);
    assert(num_marked == 0);


    //visited(assigns[i]) = false;
    

    if(conflict_clause.size() > 1) {
      size_t maxi = 1;
      for(size_t j = 2; j < conflict_clause.size(); j++)
	if(level(var(conflict_clause[j])) > level(var(conflict_clause[maxi]))) 
	  maxi = j;
      swap(conflict_clause[1], conflict_clause[maxi]);
      return max(root_level, level(var(conflict_clause[1])));
    } else return root_level;
    UNREACHABLE(0);
  }

  /*******************************************************************************
   ** Contains machine dependent code!
   *******************************************************************************/
  Clause* Solver::create_clause(const vector<Lit>& clause) {
    assert(clause.size() > 1);
    //if(clause.size() == 2)
    //return (Clause*) BinaryClause::create(clause[0],clause[1]);
    Lit max = *max_element(clause.begin(), clause.end(), IndexLitComparator());
    if(index(max) < 1<<8) 
      return (Clause*)PackedClause<unsigned char>::create(clause);
    else if(index(max) < 1<<16)
      return (Clause*)PackedClause<unsigned short>::create(clause);
    else
      return (Clause*)PackedClause<unsigned int>::create(clause);
  }

  /*******************************************************************
   ** Add a clause to the set of learned clauses. Adjust activities
   ** for the variables contained in the clause, polarity is included
   ** if <variable_polarity_dynamic> is true..
   ******************************************************************/
  Clause* Solver::learn_clause(const vector<Lit>& clause) {
    Clause* c = create_clause(clause); c->attach(*this); learned_clauses.push_back(c);
    for(size_t i = 0; i < clause.size(); i++)
      var_activity[var(clause[i])] += 1;

    if(variable_polarity_dynamic)
      for(size_t i = 0; i < clause.size(); i++) 
	var_polarity[var(clause[i])] += sign(clause[i]) ? 1 : -1;

    return c;
  }

  struct ClauseCompare {
    inline bool operator()(Clause* x, Clause* y) const { return *x < *y; }
  };
  
  /*******************************************************************
   ** Removes learned clauses according to the following scheme:
   **
   ** 1) No clauses that are locked are ever removed.
   ** 2) Otherwise, a clause is removed if its activity is lower than the 
   **    configurable parameter <clause_useless>.
   ** 3) If the parameter <clause_max_learned> is greater than 0, the 
   **    clause with the lowest activity is removed until either the 
   **    number of clauses is equal to <clause_max_learned> or all 
   **    remaining clauses are locked.
   ******************************************************************/
  void    Solver::delete_learned_clauses(void) {
    // find locked clauses
    int num_locked = 0;
    for(int i = learned_clauses.size() - 1; i >= num_locked; )
      if(learned_clauses[i]->locked(*this)) {
	swap(learned_clauses[i],learned_clauses[num_locked]); num_locked++;
      } else i--;

    int deleted = 0;
    //int total   = learned_clauses.size();

    for(int i = num_locked; (size_t)i < learned_clauses.size(); )
      if(learned_clauses[i]->get_activity() < clause_useless) {
	swap(learned_clauses[i],learned_clauses.back());
	Clause* c = learned_clauses.back(); learned_clauses.pop_back();
	c->detach(*this); free(c);
	deleted++;
      } else i++;

    if(clause_max_learned > 0) {
      sort(learned_clauses.begin() + num_locked, learned_clauses.end(), ClauseCompare());
      while(learned_clauses.size() > (size_t)max(clause_max_learned,num_locked)) {
	Clause* c = learned_clauses.back(); learned_clauses.pop_back();
	c->detach(*this); free(c);
	deleted++;
      }
    }

    //cout << "Limit = " << num << ", Deleted = " << deleted
    //<< ", Locked = " << num_locked << ", Total = " << total << '\n';
  }

  /*******************************************************************
   ** Simplifies the set of learned clauses, under the assumption that
   ** the solver is at the top level.
   ******************************************************************/
  void    Solver::simplify_clauses(void) {
    assert(current_level() == 0);
    int deleted = 0;
    for(size_t i = 0; i < learned_clauses.size(); i++) 
      if(learned_clauses[i]->satisfied(*this)) {
	swap(learned_clauses[i],learned_clauses.back());
	Clause* c = learned_clauses.back(); learned_clauses.pop_back();
	c->detach(*this); free(c);
	deleted++;
      } else i++;
    //if(verbose) cout << "deleted " << deleted << " learned clauses";
    deleted = 0;
    for(size_t i = 0; i < problem_constraints.size(); i++) 
      if(problem_constraints[i]->satisfied(*this)) {
	swap(problem_constraints[i],problem_constraints.back());
	Constraint* c = problem_constraints.back(); problem_constraints.pop_back();
	c->detach(*this); free(c);
	deleted++;
      } else i++;
    //if (verbose) cout << ", deleted " << deleted << " problem clauses\n";
  }
  

  /*******************************************************************
   ** Decay the activities of all clauses in <learned_clauses>.
   ******************************************************************/
  void Solver::decay_clause_activity(void) {
    for(size_t i = 0; i < learned_clauses.size(); i++) 
      if(!learned_clauses[i]->locked(*this))
	learned_clauses[i]->decay(clause_decay);
  }

  /*******************************************************************
   ** Delete all learned clauses and set variable activities to zero.
   ******************************************************************/
  void Solver::clear_all(void) {
    for(Var v = first_var(); v < end_var(); v++) {
      var_activity[v] = 0.0;
      var_polarity[v] = 0.0;
    }

    while(learned_clauses.size() > 0) {
      Clause* c = learned_clauses.back();
      learned_clauses.pop_back();
      c->detach(*this); free(c);
    }
  }

  /*******************************************************************
   ** Add a clause to the set of problem clauses. The clause is
   ** simplified. If simplification gives an empty clause, false is
   ** returned, true is returned otherwise. Preferred polarity for the
   ** variables in the simplified clause is updated if
   ** <variable_polarity_problem_based> is true. 
   ******************************************************************/
  bool Solver::add_clause(const vector<Lit>& clause) {
    //cout << "Pre Adding clause: " << clause << " 0\n";

    assert(current_level() == 0);
    vector<Lit> cs;
    // cout << "Pre Adding clause: " << clause << " 0\n";

    for(size_t i = 0; i < clause.size(); i++)
      if(values(clause[i]))            return true;
      else if(!values(neg(clause[i]))) cs.push_back(clause[i]);

    //cout << "Mid Adding clause: " << cs << " 0\n";

    sort(cs.begin(), cs.end(), IndexLitComparator());
    vector<Lit>::iterator new_end = unique(cs.begin(), cs.end());
    cs.erase(new_end,cs.end());
    for(size_t i = 1; i < cs.size(); i++)
      if(cs[i] == neg(cs[i-1]))
	return true;

    // cout << "Post Adding clause: " << cs << " 0\n";

    if(variable_polarity_problem_based)
      for(size_t i = 0; i < cs.size(); i++) 
	var_polarity[var(cs[i])] += sign(cs[i]) ? 1 : -1;

    switch(cs.size()) {
    case 0:  return false;
    case 1:  return assign(cs[0]);
    default: 
      Clause *c = create_clause(cs); c->attach(*this); 
      problem_constraints.push_back((Constraint*)c);
      return true;
    }
    UNREACHABLE(false);
  }

#define c2(c,x,y)   { c.clear(); c.push_back(x); c.push_back(y); }
#define c3(c,x,y,z) { c.clear(); c.push_back(x); c.push_back(y); c.push_back(z); }

  bool Solver::define_or  (Lit x, Lit y, Lit z) {
    vector<Lit> cl;
    c3(cl,neg(x),y,z); if(!add_clause(cl)) return false;
    c2(cl,neg(y),x);   if(!add_clause(cl)) return false;
    c2(cl,neg(z),x);   if(!add_clause(cl)) return false;
    return true;
  }

  bool Solver::define_and (Lit x, Lit y, Lit z) {
    vector<Lit> cl;
    c2(cl,neg(x),y);        if(!add_clause(cl)) return false;
    c2(cl,neg(x),z);        if(!add_clause(cl)) return false;
    c3(cl,neg(y),neg(z),x); if(!add_clause(cl)) return false;
    return true;
  }

  bool Solver::define_equ (Lit x, Lit y, Lit z) {
    vector<Lit> cl;
    c3(cl,neg(x),neg(y),z); if(!add_clause(cl)) return false;
    c3(cl,neg(x),y,neg(z)); if(!add_clause(cl)) return false;
    c3(cl,x,neg(y),neg(z)); if(!add_clause(cl)) return false;
    c3(cl,x,y,z);           if(!add_clause(cl)) return false;
    return true;
  }

  /*******************************************************************
   ** Fill <order> with all currently free variables.
   ******************************************************************/
  void Solver::initialize_order(void) {
//    vector<int> lit_occurrences;
//    lit_occurrences.resize(lit_values.size(), 0);
//    for(size_t i = 0; i < problem_constraints.size(); i++)
//      ((Clause*)problem_constraints[i])->occurrences(lit_occurrences);
    //int useless = 0;

    order.clear();
    for(Var v = first_var(); v < end_var(); v++) 
      if(free_var(v)) { 
//	if(lit_occurrences[index(lit(v))] == 0 && 
//	   lit_occurrences[index(neg(lit(v)))] == 0) useless++;
	order.push_back(v);
      }

    //cout << "Useless variables: " << useless << '\n';
  }


  // Assumes that all problem constraints are clauses.
  void Solver::static_clausebased_order(int steps) {
    vector<float> x,y;
    x.resize(lit_values.size(), 1.0);

    for(int i = 0; i < steps; i++) {
      static_clausebased_order_step(x,y);
      x = y; }

    float max_act = 0.0;
    for(Var v = first_var(); v < end_var(); v++) {
      var_activity[v] += (y[index(lit(v))] + y[index(neg(lit(v)))]) / 2;
      var_polarity[v] +=  y[index(lit(v))] - y[index(neg(lit(v)))] / 2;
      max_act = max(max_act, var_activity[v]);
      cout << "activity: " << var_activity[v] << '\n';
    }
    cout << "Max activity " << max_act << '\n';
  }

  void Solver::static_clausebased_order_step(const vector<float>& activities_in,
					     vector<float>& activities_out) {
    vector<Lit>   clause;
    activities_out.clear(); activities_out.resize(activities_in.size(), 0.0);

    for(size_t i = 0; i < problem_constraints.size(); i++) {
      clause.clear(); ((Clause*)problem_constraints[i])->copy(clause);
      float sum = 0.0;

      for(size_t j = 0; j < clause.size(); j++)
	sum += activities_in[index(clause[j])];

      for(size_t j = 0; j < clause.size(); j++) 
	activities_out[index(neg(clause[j]))] += 
	  (sum - activities_in[index(clause[j])]) 
	  / ((clause.size() - 1) * clause.size());
      /// (clause.size() - 1);
    }
  }

  /*******************************************************************
   ** Decay the activities of all variables with <d>. 
   ** If <variable_polarity_dynamic> is true, the polarities are also
   ** decayed.
   ******************************************************************/
  void Solver::decay_var_activity(void) {
    for(size_t i = 0; i < order.size(); i++)
      var_activity[order[i]] *= variable_decay;

    if(variable_polarity_dynamic)
      for(size_t i = 0; i < order.size(); i++)
	var_polarity[order[i]] *= variable_decay;
  }


  /*******************************************************************
   ** Choose randomly a free variable in order. Returns end_var() if
   ** all variables are bound.
   ******************************************************************/
  Var  Solver::get_random_decision(void) { 
    size_t next = rand() / (RAND_MAX / order.size());
    for(size_t i = 0; i < order.size(); i++, next++)
      if(free_var(order[next % order.size()])) return order[next % order.size()];
    return end_var();
  }

  /*******************************************************************
   ** Find the first free variable in order. Returns end_var() if all
   ** variables are bound.
   ******************************************************************/
  Var  Solver::get_decision(void) { 
    for(size_t i = 0; i < next_decision; i++)
      assert(!free_var(order[i]));

//    cout << "Order: ";
//    for(size_t j = 0; j < order.size(); j++)
//      cout << order[j] << ' ';
//    cout << '\n';

    for(; next_decision < order.size(); next_decision++)
      if(free_var(order[next_decision])) {
	return order[next_decision]; }


//    for(size_t i = 0; i < order.size(); i++)
//      if(free_var(order[i])) return order[i];

    return end_var();
  }

  /*******************************************************************
   ** The top level solving method. <prefix> is a set of literals that
   ** is assumed to be true in addition to the clauses.
   ******************************************************************/
  bool Solver::solve(const vector<Lit>& prefix) { 
    InspectModel m; return solve(activity_var_comp, m, prefix); }


  /*******************************************************************
   ** Export the internal clauses of the solver to a file. If
   ** <learned> is true, then the learned clauses are included.  
   ** Note: this currently only works if all constraints are clauses!
   ******************************************************************/
  void Solver::export_dimacs(ostream& o, const vector<Lit>& prefix, bool learned) {
    int total = problem_constraints.size() +
                assigns.size() +
                prefix.size() +
                (learned ? learned_clauses.size() : 0);
    o << "p cnf " << end_var()-1 << ' ' << total << '\n';

//    cout << "Problem clauses : " << problem_constraints.size() << '\n';
//    cout << "Learned clauses : " << learned_clauses.size()     << '\n';
//    cout << "Assigns         : " << assigns.size()             << '\n';
//    cout << "Total           : " << total                      << '\n';

    for(size_t i = 0; i < problem_constraints.size(); i++)
      o << *(Clause*)problem_constraints[i];

    if(learned)
      for(size_t i = 0; i < learned_clauses.size(); i++)
	o << *learned_clauses[i];

    for(size_t i = 0; i < assigns.size(); i++) o << assigns[i] << " 0\n";
    for(size_t i = 0; i < prefix.size(); i++)  o << prefix[i]  << " 0\n";
  }


  /*******************************************************************
   ** relise on the fact that order contains the real variables 
   ** (compared to 1 .. next_var -1)
   ******************************************************************/
  double Solver::progress(void) {
    double progress = 0.0;
    
    int old = 0, i = 0, next;
    double step;
    for(; i < current_level(); i++) {
      step      = pow(2.0, (double) (- i));
      next      = frames[i].first;
      progress += (next - old) * step;
      old       = next;
    }
    step      = pow(2.0, (double) (- i));
    progress += (assigns.size() - old) * step;
    return progress / (double)(order.size());
  }

};
