/* ---------------------------------------------------------------
-- Satnik, a SAT solver                                         --
-- Copyright (c) 2003 Niklas Srensson                          --
-- nik@cs.chalmers.se                                           --
--                                                              --
-- This file is part of Satnik.                                 --
--                                                              --
-- Satnik is free software; you can redistribute it and/or      --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Satnik is distributed in the hope that it will be useful,    --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Satnik; if not, write to the Free         --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
--------------------------------------------------------------- */

#ifndef __SATNIK__
#define __SATNIK__

#include <vector>
#include <deque>
#include <algorithm>
#include <iostream>
#include <math.h>

#define UNREACHABLE(x) assert(false); return (x);


namespace Satnik {
  using namespace std;

  typedef int Var;

  // an abstraction for literals, implemented as integers
  class Lit {
    int     x;
    
    Lit(Var var, bool s) : x((var+var) + s) { }

    Var var(void) { return x >> 1; }
  public:
    Lit(void)            /* undefined value allowed */  { }
    bool operator==(Lit l) const { return x == l.x; }

    friend Lit  lit   (Var v, bool s = true) { return Lit(v,s); }
    friend Lit  neg   (Lit l)                { Lit lit; lit.x = l.x ^ 1; return lit; }
    friend bool sign  (Lit l)                { return l.x & 1; }
    friend Var  var   (Lit l)                { return l.var(); }
    friend int  index (Lit l)                { return l.x; }
    friend Lit  encode(int x)                { Lit lit; lit.x = x; return lit; }
    friend ostream& operator<<(ostream& o, Lit l) {
      if(!sign(l)) o << '-'; o << l.var(); return o; }
  };

  // print a vector of literals
  inline ostream& operator<<(ostream& o, const vector<Lit>& literals) {
    for(size_t i = 0; i < literals.size(); i++) o << literals[i] << ' ';
    return o; }

  class Solver;

  class Undo {
  public:
    virtual void pop (void) = 0;
  };


  // An interface for arbitrary constraints over boolean variables
  // TODO: explain!!
  class Constraint : public Undo {
  public:

    virtual void attach    (Solver& s) = 0;
    virtual void detach    (Solver& s) = 0;

    virtual void pop       (void) {};
    virtual void reason    (Lit l, vector<Lit>& r) = 0;
    virtual bool propagate (Lit l, Solver& s) = 0;

    virtual bool satisfied (Solver& s) = 0;
  };

  typedef enum { Unsatisfiable, Satisfiable, Indeterminate }  Result;

  class Clause;

  // The core solver class
  class Solver {
    Var next_var;

    // problem + learned clauses
    vector<Constraint*>          problem_constraints;
    vector<Clause*>              learned_clauses;

    // search dependent information
    deque<Lit>                   queue;
    vector<Lit>                  assigns;
    vector<Undo*>                undos;
    vector<pair<size_t,size_t> > frames;  // (assign,undo) frames
    vector<Var>                  order;

    vector<Lit>                  unit_conflicts;

    size_t                       next_decision;

    int                          num_marked;
    int                          root_level;
    int                          simplified_frame;
    Constraint*                  conflict;

    // counters
    int       num_prop;
    int       num_conf;
    int       num_deci;
    int       num_restart;
    long long num_assign;

    // information for variables
    vector<Constraint*>          var_reason;
    vector<int>                  var_level;
    vector<float>                var_activity;
    vector<float>                var_polarity;
    vector<size_t>               var_order_index;

    // information for literals
    vector<int>                  lit_values;
    vector<vector<Constraint*> > lit_watches;
    vector<int>                  lit_visited;

    bool  verbose;

    // parameters affecting the behaviour of search
    int   variable_reorder_interval;
    int   variable_randomness;
    float variable_decay;

    bool  variable_polarity_problem_based;
    bool  variable_polarity_dynamic;
    bool  variable_polarity_sign;

    int   clause_compaction_interval;
    int   clause_max_learned;
    float clause_decay;
    float clause_useless;

    int   restart_initial_interval;
    float restart_increase;


  public:
    void  set_variable_parameters(int reorder_interval, float decay, int randomness) {
      variable_reorder_interval = reorder_interval;
      variable_decay            = pow(decay,reorder_interval);
      variable_randomness       = randomness;
    }

    void  set_polarity_parameters(bool problem_based, bool dynamic, bool sign) {
      variable_polarity_problem_based = problem_based;
      variable_polarity_dynamic       = dynamic;
      variable_polarity_sign          = sign;
    }

    void  set_learning_parameters(int compaction_interval, int max_learned, float decay, float useless) {
      clause_compaction_interval = compaction_interval;
      clause_max_learned         = max_learned;
      clause_decay               = pow(decay,compaction_interval);
      clause_useless             = useless;
    }

    void  set_restart_parameters(int initial_interval, float increase) {
      restart_initial_interval = initial_interval;
      restart_increase         = increase;
    }

    void  set_verbose(bool v) { verbose = v; }

    // convenient access
    int&                 values  (Lit l) { return lit_values [index(l)]; }
    vector<Constraint*>& watches (Lit l) { return lit_watches[index(l)]; }
    int&                 visited (Lit l) { return lit_visited[index(l)]; }

    Constraint*&         reason  (Var v) { return var_reason [v]; }
    int&                 level   (Var v) { return var_level  [v]; }

    int                  current_level(void) { return frames.size(); }

    // watcher references
    void add_watch(Lit l, Constraint* c) { lit_watches[index(l)].push_back(c); }
    void del_watch(Lit l, Constraint* c) {
      vector<Constraint*>& cs = lit_watches[index(l)];
      for(size_t i = 0; i < cs.size(); i++)
	if(cs[i] == c) {  cs[i] = cs.back(); cs.pop_back();  break; }
    }

    // propagation + search related basic tools
    bool propagate(void);
    bool assign   (Lit l, Constraint* from = NULL);
    void push     (void);
    void pop      (void);

    // variable order related methods
    void initialize_order(void);
    void static_clausebased_order(int steps);
    void static_clausebased_order_step(const vector<float>& activities_in,
				       vector<float>& activities_out);
    void decay_var_activity(void);

    /*******************************************************************
     ** Reorder the <order> vector with respect to the ordering
     ** predicate <c>. The pointer to the next_decision is no longer
     ** valid and must be reset.
     ******************************************************************/
    template<class C> 
    void reorder(C c) {
      sort(order.begin(), order.end(), c); 
      //stable_sort(order.begin(), order.end(), c);
      for(size_t i = 0; i < order.size(); i++)
	var_order_index[order[i]] = i;

      next_decision = 0;
    }

    Var  get_decision(void);
    Var  get_random_decision(void);

    // conflict analysis + clause learning
    void    analyze_lit (Lit l, vector<Lit>&  conflict_clause);
    int     analyze     (const vector<Lit>& conflict_set, vector<Lit>& conflict_clause);
    Clause* learn_clause(const vector<Lit>& clause);
    void    decay_clause_activity(void);
    void    delete_learned_clauses(void);
    void    simplify_clauses(void);

    // literal and variable comparators
    struct IndexLitComparator {
      bool operator()(Lit x, Lit y) const { return index(x) < index(y); }
    };

    class ActivityVarComparator {
      vector<float>& var_activity;
    public:
      ActivityVarComparator(vector<float>& a) : var_activity(a) {}
      bool operator()(Var x,Var y) const { return var_activity[x] > var_activity[y]; }
    };

    template<class A, class B>
    class ComposedVarComparator {
      A a; B b;
    public:
      ComposedVarComparator(A _a, B _b) : a(_a), b(_b) {};
      bool operator()(Var x,Var y) const { return a(x,y) || (!a(y,x) && b(x,y)); }
    };

    const ActivityVarComparator activity_var_comp;


    // standard model inspection (that does nothing)
    struct InspectModel { void operator()(const Solver& s) {} };

    // the core searching method
    template <class C> Result search(C &c, int conflicts); 

    Clause* create_clause(const vector<Lit>& clause);

    // Misc
    double progress(void);


  public:
    Solver(void) : 
      next_var(0), next_decision(0),
      num_marked(0), num_prop(0), num_conf(0), num_deci(0), num_restart(0), num_assign(0),
      activity_var_comp(var_activity) { 
      add_vars(); 

      set_variable_parameters(60,.99,100);
      set_polarity_parameters(true,false,true);
      set_learning_parameters(1000,0,.999,.0001);
      set_restart_parameters (1000,1.4);

      set_verbose(false);
    };

    template<class A>
    inline ComposedVarComparator<A,ActivityVarComparator>
    combine(A a) {
      return ComposedVarComparator<A,ActivityVarComparator>(a,activity_var_comp);
    }


    // variable creation + iteration over all variables
    Var  add_vars  (int num = 1);
    Var  first_var (void)   { return 1; }
    Var  end_var   (void)   { return next_var; }

    bool free_var(Var v) { return !values(lit(v)) && !values(neg(lit(v))); }

    // controlling the behaviour of search
    void  clear_all(void);
    // set_xxx should be here ...

    // statistics
    int  get_nof_vars(void)        { return next_var - 1; }
    int  get_nof_constraints(void) { return problem_constraints.size(); }
    int  get_nof_conflicts(void)   { return num_conf; }


    // problem specification
    bool add_clause (const vector<Lit>& clause);
    bool define_or  (Lit x, Lit y, Lit z);
    bool define_and (Lit x, Lit y, Lit z);
    bool define_equ (Lit x, Lit y, Lit z);

    // the toplevel solving method + variants
    template <class C, class M>
    bool solve(C c, M& m, const vector<Lit>& prefix = vector<Lit>());

    bool solve(const vector<Lit>& prefix = vector<Lit>());

    template <class C>
    bool solve(C c, const vector<Lit>& prefix = vector<Lit>()) 
      { InspectModel m; return solve(c, m, prefix); }

    template <class M>
    bool solve_(M& m, const vector<Lit>& prefix = vector<Lit>()) 
      { return solve(activity_var_comp, m, prefix); }

    void copy_model(vector<int>& m) { m = lit_values; }

    void print_stat(void) {
      cout << "Restarts:     " << num_restart  << '\n';
      cout << "Conflicts:    " << num_conf     << '\n';
      cout << "Decisions:    " << num_deci     << '\n';
      cout << "Propagations: " << num_prop     << '\n';
      cout << "Assignments:  " << num_assign   << '\n';
    }

    // temporary debug
    const vector<Var>& inspect_order(void) { return order; }

    void export_dimacs(ostream& o, const vector<Lit>& prefix = vector<Lit>(), bool learned = false);
				

  };

};

#include "Clause.h"
#include "Search.h"



#endif
