/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/**
 *  File Schedule.cpp for running scheduled tasks.
 */

#include "VampireKernel.hpp"
#include "GlobAlloc.hpp"

#include "Property.hpp"
#include "Problem.hpp"
#include "Options.hpp"
#include "Schedule.hpp"
#include "Int.hpp"
#include "Statistics.hpp"
#include "Refutation.hpp"
#include "vampire.hpp"


/**
 * Initialize schedule.
 *
 * @param p problem properties
 * @param cmdLineOptions options read from the command line only
 *        time limit is used
 * @since 21/07/2003 Manchester, switch on CASCs added
 */
Schedule::Schedule (const Property& p, const Options& cmdLineOptions)
  :
  _commandLineOptions (cmdLineOptions)
{
  int timeLimit = cmdLineOptions.timeLimit ();

  switch (options->mode()) 
    {
    case Options::CASC_18:
      casc18(p,timeLimit);
      return;
    case Options::CASC_19:
      casc19(p,timeLimit);
      return;
#if DEBUG_PREPRO
    default:
      ASS(false);
#endif
    }
} // Schedule::Schedule


/**
 * Initialise schedule to the CASC-19 mode
 *
 * @param p problem properties
 * @param timeLimit time limit
 */

void Schedule::casc19(const Property& p, int timeLimit) 
{
  switch (p.category()) 
    {
    case Property::EPR:
      {
	Options* o1 = Options::lrs (Options::P_4);
	o1->casc19();
	o1->ageWeightRatio (1);
	o1->timeLimit(470);
	o1->backwardSubsumption (Options::SUBSUMPTION_OFF);

	Options* o2 = Options::lrs (Options::P_1002);
	o2->casc19();
	o2->timeLimit(320);
	o2->backwardSubsumption (Options::SUBSUMPTION_OFF);

	Options* o3 = Options::lrs (Options::N_1005, Options::KINKY);
	o3->casc19();
	o3->timeLimit(520);
	o3->backwardSubsumption (Options::SUBSUMPTION_OFF);
	o3->elimDef (1);

	Options* o4 = Options::discount (Options::N_7, Options::PREDICATE);
	o4->casc19();
	o4->timeLimit(30);

	Options* o5 = Options::lrs (Options::P_7, Options::PREDICATE);
	o5->casc19();
	o5->timeLimit(570);
	o5->ageWeightRatio (4);
	o5->elimDef (1);

	Options* general = Options::lrs (Options::P_1005);
	general->casc19();
	general->timeLimit (0);

	_schedule = new List (o1, new List (o2, new List (o3, 
                    new List (o4, new List (o5, new List (general))))));
	return;
      }

    case Property::UEQ:
      {
	Options* general = Options::lrs (Options::P_1);
	general->casc19();
	general->ageWeightRatio (6);
	general->backwardSubsumption (Options::SUBSUMPTION_OFF);
	general->simplificationOrdering (Options::KBO);
	general->timeLimit(0);

	_schedule = new List (general);
	return;
      }

    case Property::HEQ: 
      { 
	Options* general = Options::lrs (Options::N_1005);
	general->casc19();
	general->ageWeightRatio(1);
	general->nongoalWeightCoefficient(1.5);
	general->elimDef(1);
	general->timeLimit (0);
	_schedule = new List (general);

	Options* o1 = Options::lrs(Options::P_6);
	o1->casc19();
	o1->ageWeightRatio(1);
	o1->elimDef(1);
	o1->timeLimit(620);
	List::push(o1,_schedule);

	Options* o2 = Options::lrs(Options::P_1002);
	o2->casc19();
	o2->ageWeightRatio(2);
	o2->elimDef(1);
	o2->timeLimit(520);
	List::push(o2,_schedule);


	Options* o3 = Options::lrs(Options::P_4);
	o3->ageWeightRatio(1);
	o3->casc19();
	o3->elimDef(1);
	o3->timeLimit(220);
	List::push(o3,_schedule);

	Options* o4 = Options::discount(Options::P_7);
	o4->ageWeightRatio(9);
	o4->casc19();
	o4->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o4->elimDef(1);
	o4->timeLimit(420);
	List::push(o4,_schedule);

	Options* o5 = Options::discount(Options::N_14);
	o5->ageWeightRatio(4);
	o5->casc19();
	o5->timeLimit(70);
	List::push(o5,_schedule);

	return;
      }

    case Property::PEQ:
      { 
	Options* general = Options::lrs (Options::P_4);
	general->ageWeightRatio(1);
	general->casc19();
	general->timeLimit (0);
	_schedule = new List (general);

	Options* o1 = Options::lrs(Options::N_2);
	o1->ageWeightRatio(6);
	o1->casc19();
	o1->timeLimit(670);
	List::push(o1,_schedule);

	Options* o2 = Options::lrs(Options::P_15);
	o2->ageWeightRatio(1);
	o2->casc19();
	o2->nongoalWeightCoefficient(1.5);
	o2->fullSplitting();
	o2->timeLimit(220);
	List::push(o2,_schedule);
	return;
      }

    case Property::HNE:
      { 
	Options* general = Options::lrs (Options::P_4);
	general->ageWeightRatio(2);
	general->casc19();
	general->timeLimit (0);
	_schedule = new List (general);

	Options* o1 = Options::lrs(Options::P_1003);
	o1->ageWeightRatio(1);
	o1->casc19();
	o1->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o1->timeLimit(timeLimit / 9);
	List::push(o1,_schedule);

	Options* o2 = Options::discount(Options::P_11);
	o2->ageWeightRatio(12);
	o2->casc19();
	o2->timeLimit(timeLimit/9);
	List::push(o2,_schedule);

	Options* o3 = Options::lrs(Options::N_4);
	o3->ageWeightRatio(1);
	o3->casc19();
	o1->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o3->timeLimit(timeLimit / 9);
	List::push(o3,_schedule);

	Options* o4 = Options::discount(Options::P_6);
	o4->ageWeightRatio(2);
	o4->casc19();
	o4->elimDef(1);
	o4->timeLimit(timeLimit / 9);
	List::push(o4,_schedule);

	return;
      }

    case Property::FEQ:
      { 
	Options* general = Options::lrs (Options::P_4);
	general->ageWeightRatio(1);
	general->casc19();
	general->nongoalWeightCoefficient(1.5);
	general->timeLimit (0);
	_schedule = new List (general);

	Options* o1 = Options::lrs(Options::P_1012);
	o1->ageWeightRatio(2);
	o1->casc19();
	o1->timeLimit(timeLimit / 9);
	List::push(o1,_schedule);

	Options* o2 = Options::discount(Options::P_6);
	o2->ageWeightRatio(16);
	o2->casc19();
	o2->timeLimit(timeLimit/9);
	List::push(o2,_schedule);

	Options* o3 = Options::lrs(Options::P_13);
	o3->ageWeightRatio(1);
	o3->casc19();
	o3->timeLimit(timeLimit / 9);
	List::push(o3,_schedule);

	Options* o4 = Options::lrs(Options::N_14);
	o4->ageWeightRatio(1);
	o4->casc19();
	o4->timeLimit(timeLimit / 9);
	List::push(o4,_schedule);
	return;
      }

    case Property::NNE:
      { 
	Options* general = Options::lrs (Options::P_4);
	general->ageWeightRatio(1);
	general->casc19();
	general->nongoalWeightCoefficient(1.5);
	general->timeLimit (0);
	_schedule = new List (general);

	Options* o1 = Options::lrs(Options::P_1012);
	o1->ageWeightRatio(4);
	o1->casc19();
	o1->nongoalWeightCoefficient(1.5);
	o1->timeLimit(timeLimit / 7);
	List::push(o1,_schedule);

	Options* o2 = Options::lrs(Options::P_13);
	o2->ageWeightRatio(1);
	o2->casc19();
	o2->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o2->timeLimit(timeLimit/7);
	List::push(o2,_schedule);

	return;
      }

    case Property::FNE:
    default:
      { 
	Options* general = Options::lrs (Options::P_4);
	general->ageWeightRatio(1);
	general->casc19();
	general->timeLimit (0);

	_schedule = new List (general);
	break;
      }

    case Property::NEQ:
      { 
	Options* general = Options::lrs (Options::P_4);
	general->ageWeightRatio(1);
	general->casc19();
	general->nongoalWeightCoefficient(2.0);
	general->timeLimit (0);
	_schedule = new List (general);

	Options* o1 = Options::lrs(Options::P_15);
	o1->ageWeightRatio(1);
	o1->casc19();
	o1->nongoalWeightCoefficient(1.5);
	o1->timeLimit(620);
	List::push(o1,_schedule);

	Options* o2 = Options::lrs(Options::P_5);
	o2->ageWeightRatio(1);
	o2->casc19();
	o2->timeLimit(270);
	List::push(o2,_schedule);

	Options* o3 = Options::lrs(Options::P_4);
	o3->ageWeightRatio(1);
	o3->casc19();
	o3->timeLimit(520);
	List::push(o3,_schedule);

	Options* o4 = Options::lrs(Options::N_1005,Options::KINKY);
	o4->ageWeightRatio(5);
	o4->casc19();
	o4->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o4->elimDef(1);
	o4->timeLimit(520);
	List::push(o4,_schedule);

	Options* o5 = Options::discount(Options::P_13);
	o5->ageWeightRatio(1);
	o5->casc19();
	o5->timeLimit(620);
	List::push(o5,_schedule);

	Options* o6 = Options::lrs(Options::P_1012);
	o6->ageWeightRatio(1);
	o6->casc19();
	general->nongoalWeightCoefficient(1.5);
	o6->timeLimit(520);
	List::push(o6,_schedule);

	return;
      }

#if DEBUG_PREPRO
    default:
      ASSERT( debug_schedule, false);
#endif
    }

} // Schedule::casc19




/**
 * Initialise schedule to the CASC-18 mode
 *
 * @param p problem properties
 * @param timeLimit time limit
 */
void Schedule::casc18(const Property& p, int timeLimit) 
{
  switch (p.category()) 
    {
    case Property::FNE: 
      { 
	Options* general = Options::lrs (Options::P_4);
	general->casc18();
	general->timeLimit (0);

	_schedule = new List (general);
	return;
      }

    case Property::NNE: 
      { 
	Options* o1 = Options::lrs (Options::P_7);
	o1->casc18();
	o1->ageWeightRatio (1);
	o1->timeLimit(620);

	Options* o2 = Options::lrs (Options::P_1003, Options::KINKY);
	o2->casc18();
	o2->ageWeightRatio (16);
	o2->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o2->timeLimit(620);
    
	Options* o3 = Options::lrs (Options::P_1005, Options::KINKY);
	o3->casc18();
	o3->ageWeightRatio (13);
	o3->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o3->timeLimit(420);

	Options* o4 = Options::lrs (Options::P_6, Options::KINKY);
	o4->casc18();
	o4->fullSplitting ();
	o4->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o4->timeLimit(70);

	Options* o5 = Options::lrs (Options::P_1);
	o5->casc18();
	o5->backwardSubsumption(Options::SUBSUMPTION_OFF);
	o5->timeLimit(170);

	Options* o6 = Options::lrs (Options::N_1005);
	o6->casc18();
	o6->literalComparisonMode (Options::KINKY);
	o6->symbolPrecedence (Options::BY_REVERSE_ARITY);
	o6->ageWeightRatio (11);
	o6->backwardSubsumption (Options::SUBSUMPTION_OFF);
	o6->timeLimit(270);

	Options* o7 = Options::discount (Options::P_7);
	o7->casc18();
	o7->symbolPrecedence (Options::BY_REVERSE_ARITY);
	o7->ageWeightRatio (172);
	o7->timeLimit(320);

	Options* general = Options::lrs (Options::P_1005);
	general->casc18();
	general->timeLimit (0);

	_schedule = new List (o1, new List (o2, new List (o3, 
		    new List (o4, new List (o5, new List (o6, 
	  	    new List (o7, new List (general))))))));
	return;
      }

    case Property::HNE:
      {
	Options* o1 = Options::lrs (Options::P_4);
	o1->casc18();
	o1->ageWeightRatio (3);
	o1->timeLimit(520);
	o1->backwardSubsumption (Options::SUBSUMPTION_OFF);
	o1->elimDef (1);
	
	Options* o2 = Options::discount (Options::N_1005);
	o2->casc18();
	o2->ageWeightRatio (2);
	o2->timeLimit(70);
	o2->backwardSubsumption (Options::SUBSUMPTION_OFF);

	Options* o3 = Options::discount (Options::N_2);
	o3->casc18();
	o3->timeLimit(70);
	o3->backwardSubsumption (Options::SUBSUMPTION_OFF);

	Options* general = Options::lrs (Options::P_1, Options::KINKY);
	general->casc18();
	general->ageWeightRatio (54);
	general->backwardSubsumption (Options::SUBSUMPTION_OFF);
	general->timeLimit(0);

	_schedule = new List (o1, new List (o2, new List (o3, 
                    new List (general))));
	return;
      }

    case Property::HEQ:
      {
	Options* o1 = Options::lrs (Options::P_4);
	o1->casc18();
	o1->ageWeightRatio (1);
	o1->elimDef (1);
	o1->timeLimit(520);

	Options* o2 = Options::discount (Options::P_7);
	o2->casc18();
	o2->ageWeightRatio (9);
	o2->elimDef (1);
	o2->backwardSubsumption (Options::SUBSUMPTION_OFF);    
	o2->timeLimit(570);

	Options* o3 = Options::discount (Options::N_1009);
	o3->casc18();
	o3->ageWeightRatio (8);
	o3->symbolPrecedence (Options::BY_REVERSE_ARITY);
	o3->timeLimit(70);

	Options* o4 = Options::lrs (Options::N_1005, Options::KINKY);
	o4->casc18();
	o4->ageWeightRatio (2);
	o4->elimDef (1);
	o4->timeLimit(620);

	Options* o5 = Options::lrs (Options::P_1005);
	o5->casc18();
	o5->ageWeightRatio (4);
	o5->timeLimit(520);
	o5->backwardSubsumption (Options::SUBSUMPTION_OFF);    
	o5->forwardDemodulation (Options::DEMODULATION_PREORDERED);
	o5->maxWeight (16);

	Options* general = Options::lrs (Options::P_1003, Options::KINKY);
	general->casc18();
	general->timeLimit(0);

	_schedule = new List (o1, new List (o2, new List (o3, 
                   new List (o4, new List (o5, new List (general))))));
	return;
      }

    case Property::PEQ:
      {
	Options* o1 = Options::lrs (Options::N_2);
	o1->casc18();
	o1->ageWeightRatio (6);
	o1->timeLimit(670);

	Options* o2 = Options::discount (Options::P_1);
	o2->casc18();
	o2->ageWeightRatio (3);
	o2->timeLimit(120);

	Options* o3 = Options::otter (Options::P_1007);
	o3->casc18();
	o3->ageWeightRatio (100);
	o3->backwardSubsumption (Options::SUBSUMPTION_OFF);
	o3->timeLimit(120);

	Options* o4 = Options::discount (Options::N_1005);
	o4->casc18();
	o4->ageWeightRatio (3);
	o4->timeLimit(770);

	Options* general = Options::lrs (Options::P_4);
	general->casc18();
	general->timeLimit (0);

	_schedule = new List (o1, new List (o2, new List (o3, 
                    new List (o4, new List (general)))));
	return;
      }

    case Property::FEQ:
      {
	Options* o1 = Options::lrs (Options::P_4);
	o1->casc18();
	o1->ageWeightRatio (1);
	o1->timeLimit(620);
	o1->elimDef (1);

	Options* o2 = Options::lrs (Options::P_6);
	o2->casc18();
	o2->ageWeightRatio (9);
	o2->timeLimit(320);
	o2->backwardSubsumption (Options::SUBSUMPTION_OFF);

	Options* o3 = Options::discount (Options::P_1003);
	o3->casc18();
	o3->ageWeightRatio (9);
	o3->timeLimit(320);

	Options* o4 = Options::lrs (Options::P_1003, Options::KINKY);
	o4->casc18();
	o4->fullSplitting ();
	o4->timeLimit(420);
	o4->ageWeightRatio (7);
	o4->backwardSubsumption (Options::SUBSUMPTION_OFF);
	o4->elimDef (1);

	Options* general = Options::lrs (Options::P_1005);
	general->casc18();
	general->timeLimit (0);

	_schedule = new List (o1, new List (o2, new List (o3, 
                    new List (o4, new List (general)))));
	return;
      }

    case Property::EPR:
    case Property::NEQ:
      {
	Options* o1 = Options::lrs (Options::P_4);
	o1->casc18();
	o1->ageWeightRatio (1);
	o1->timeLimit(470);
	o1->backwardSubsumption (Options::SUBSUMPTION_OFF);

	Options* o2 = Options::lrs (Options::P_1002);
	o2->casc18();
	o2->timeLimit(320);
	o2->backwardSubsumption (Options::SUBSUMPTION_OFF);

	Options* o3 = Options::lrs (Options::N_1005, Options::KINKY);
	o3->casc18();
	o3->timeLimit(520);
	o3->backwardSubsumption (Options::SUBSUMPTION_OFF);
	o3->elimDef (1);

	Options* o4 = Options::discount (Options::N_7, Options::PREDICATE);
	o4->casc18();
	o4->timeLimit(30);

	Options* o5 = Options::lrs (Options::P_7, Options::PREDICATE);
	o5->casc18();
	o5->timeLimit(570);
	o5->ageWeightRatio (4);
	o5->elimDef (1);

	Options* general = Options::lrs (Options::P_1005);
	general->casc18();
	general->timeLimit (0);

	_schedule = new List (o1, new List (o2, new List (o3, 
                    new List (o4, new List (o5, new List (general))))));
	return;
      }

    case Property::UEQ:
      {
	Options* general = Options::lrs (Options::P_1);
	general->casc18();
	general->ageWeightRatio (6);
	general->backwardSubsumption (Options::SUBSUMPTION_OFF);
	general->simplificationOrdering (Options::KBO);
	general->timeLimit(0);

	_schedule = new List (general);
	return;
      }

#if DEBUG_PREPRO
    default:
      ASSERT( debug_schedule, false);
#endif
    }
} // Schedule::casc18


Schedule::~Schedule ()
{
  List::Iterator opts (_schedule);
  while (opts.more()) {
    delete opts.next ();
  }
  _schedule->destroy ();
} // Schedule::~Schedule


void Schedule::run (Problem& problem, BK::GlobalClock::Timer& timer)
{
  // create job id
  const char* jobId = "dummy";

  while (_schedule->isNonEmpty()) {
    // create Vampire kernel
    VampireKernel kernel;
    // create options for the kernel
    VampireKernel::Options kernelOptions;

    Options* opts = List::pop (_schedule);
    // recalculate the time limit
    opts->timeLimit (opts->timeLimit());

    kernel.closeSession(); // safe even if there is no session
    BK::GlobAlloc::reset();

    // pass the values of options from vampire to the kernel
    passOptions (*opts, kernelOptions, jobId);

    int remainingTime = _commandLineOptions.timeLimit () -
                        timer.elapsedDeciseconds ();
    int timeLimit = _schedule ->isEmpty () ?
                    remainingTime :
                    Int::min (opts->timeLimit(), remainingTime);

    switch (::run (kernel, timeLimit, kernelOptions, problem)) 
      {
      case '+':
	cout << "--------------------- PROVED ----------------------\n"
	     << flush;

	if ( options->proof() != Options::PROOF_OFF ) {
	  Refutation ref (kernel, problem);
	  Output* output = Output::create (opts->outputSyntax(), cout);
	  output->pretty (ref, 1); // 1 - proof number
	  delete output;
	}

	//kernel.tabulate ();
	kernel.closeSession(); // not really necessary here
	return;

      case '-':
	cout << "--------------------- UNPROVABLE ----------------------\n"
	     << flush;
	//kernel.tabulate ();
	kernel.closeSession(); // not really necessary here
	return;

      case '0':
	delete opts;
	break;
      }
  }

  cout << "--------------------- CANNOT PROVE ----------------------\n"
       << flush;
} // Schedule::run


/*

void Schedule::run (Problem& problem, BK::GlobalClock::Timer& timer)
{
  cout << "<?xml version=\"1.0\"?>\n<vampire>\n";
  options->outputSyntax(XML);

  // create job id
  const char* jobId = "dummy";

  while (_schedule->isNonEmpty()) {
    // create Vampire kernel
    VampireKernel kernel;
    // create options for the kernel
    VampireKernel::Options kernelOptions;

    Options* opts = List::pop (_schedule);
    opts->outputSyntax(XML);

    BK::GlobAlloc::reset();

    // pass the values of options from vampire to the kernel
    passOptions (*opts, kernelOptions, jobId);

    int remainingTime = _commandLineOptions.timeLimit () -
                        timer.elapsedDeciseconds ();
    int timeLimit = _schedule ->isEmpty () ?
                    remainingTime :
                    Int::min (opts->timeLimit(), remainingTime);

    char result = ::run (kernel, timeLimit, kernelOptions, problem);
    switch (result) 
      {
      case '0':
	if (_schedule->isNonEmpty()) {
	  kernel.closeSession();
	  break;
	}
	// else tabulate the last attempt

      case '+':
      case '-':
	{
	  Statistics statistics (kernel);
	  statistics.result(result);
	  cout << statistics;
	  kernel.closeSession(); // not really necessary here
	  if (opts->outputSyntax() == XML) {
	    cout << "</vampire>\n";
	  }
	  return;
	}
      }
  }
} // Schedule::run


*/
