/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/**
 * Class Sort<C,n> for sorting
 * @since 06/05/2002 
 * @since 16/07/2003 flight Moscow-London, changed from a
 *   previous one-type version.
 */


#ifndef __Sort__
#  define __Sort__


#include "Memory.hpp"


// ******************* Class CList *******************************


/**
 * A template class to sort data. The type ToCompare is the
 * type of elements to sort. The type Comp is the class that
 * contains a static function bool lessThan(ToCompare, ToCompare).
 * A typical work with this class is as follows: (1) an element
 * of the class Sort is created; (2) elements of the class ToCompare
 * are added one by one; (3) the resulting sorted collection is read.
 *
 * @since 16/07/2003 flight Moscow-London, changed from a
 * previous one-type version.
 */
template <typename ToCompare, class Comp>
class Sort
{
 public:
  /**
   * Initialize the sort element.
   *
   * @param length the number of elements to sort
   */
  inline
  Sort (int length) 
    :
    _elems ( new ToCompare[length] ),
    _size (length),
    _length (0)
    {
      if (! _elems) {
        NO_MEMORY;
      }
    }

  ~Sort () 
    {
      delete [] _elems;
    }

  /**
   * Return the length (the number of elements to sort)
   */
  inline int length () const 
    { return _length; }

  /** 
   * Return the nth element, used to retrieve sorted elements
   */
  inline ToCompare operator [] (int n) const
    { 
      ASS(n < _length);
      return _elems[n]; 
    } // Sort::operator []

  /**
   * Add an element to sort.
   * 
   * @param elem the element.
   */ 
  inline void add (ToCompare elem)
    {
      ASS(_length < _size);

      _elems[_length++] = elem;
    } // Sort::add

  /**
   * Remove duplicates from sorted list
   */
  void removeDuplicates ()
    {
      int cur = 0;

      for (int next = 1; next < _length; next++) {
        if ( _elems[cur] != _elems[next] ) {
          cur++;
        }
      }

      // reset length
      _length = cur + 1;
    } // Sort::removeDuplicates

  /**
   * Sort elements.
   */
  inline void sort ()
    { sort (0,_length-1); }

  /**
   * Check membership in the sorted list. Membership is checked using
   * == for equality.
   *
   * @return true if elem is the member of Sort
   * @since 17/07/2003 Manchester, changed to the new two-argument Sort class
   */
   inline bool member (const ToCompare elem) const { 
     return member (elem,0,_length-1); 
   }

 protected:  // structure

  ToCompare* _elems;
  int _size;
  int _length;

  /** 
   * Quicksort, copied from Cormen et.al's "Introduction to Algorithms"
   */
  void sort ( int p,int r ) 
    {
      ASS(r < _length);

      if (p < r) {
        int q = partition(p,r);
        sort(p,q);
        sort(q+1,r);    
      }
    } // Sort::sort

  /** 
   * Quicksort partition, copied from Cormen et.al's 
   * "Introduction to Algorithms"
   */
  int partition( int p,int r ) 
  {
    ToCompare x = _elems[p];
    int i = p-1;
    int j = r+1;

    for (;;) {
      do --j;
      while (Comp::lessThan(x,_elems[j]));
      do ++i;
      while (Comp::lessThan(_elems[i],x));
      if (i < j) {
        // swap [i] and [j]
        ToCompare tmp = _elems[i];
        _elems[i] = _elems[j];
        _elems[j] = tmp;
      }
      else
        return j;
    }
  } // Sort::partitition

  /**
   * Check membership in the sorted list between the indexes fst
   * and snd. Membership is checked using == for equality. The function
   * works correctly only if lessThan is a total order.
   *
   * @return true if elem is the member of Sort[fst] ... Sort[snd]
   * @since 17/07/2003 Manchester, changed to the new two-argument Sort class
   */
  bool member (const ToCompare elem, int fst, int lst) const { 
    for (;;) {
      if (fst > lst) {
	return false;
      }

      int mid = (fst + lst) / 2;
      
      if (_elems[mid] == elem) {
	return true;
      }

      if (Comp::lessThan(_elems[mid],elem)) {
	lst = mid-1;
      }
      else { // _elems[mid] > c
	fst = mid+1;
      }
    }
  } // Sort::member
};  // class Sort


#endif


