/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
//  file Statistics.cpp 
//  implements class Statistics
//


#include <iostream>

#ifdef _MSC_VER
#  include <winsock2.h>
#  include <process.h>
#else 
#  include <unistd.h>
#endif

#include "Statistics.hpp"
#include "Options.hpp"


bool Statistics::anyGeneratingInference () const
{
  return _resolutions ||
         _forwardSuperpositions ||
         _backwardSuperpositions ||
         _equalityResolutions ||
        _equalityFactorings;
} // Statistics::anyGeneratingInference ()


bool Statistics::anySimplifyingInference () const
{
  return _forwardSubsumptionResolutions ||
         _forwardDemodulated ||
         _backwardDemodulated ||
         _theoryInferences;
} // Statistics::anySimplifyingInference ()


bool Statistics::anyRedundantClause () const
{
  return _forwardSubsumed ||
         _backwardSubsumed ||
         _discardedPropositionalTautologies ||
         _discardedEquationalTautologies ||
         _murderedPassiveOrphans ||
         _murderedActiveOrphans;
} // Statistics::anyRedundantClause ()


inline
bool Statistics::anySplitting () const
{
  return _splittingInferences > 0;
} // Statistics::anySplitting ()


inline
bool Statistics::anyDiscarded () const
{
  return _discardedWithTooHeavyLiterals ||
         _discardedTooDeep ||
         _discardedTooHeavy ||
         _discardedWithTooManyVariables ||
         _newClausesDiscardedDueToLackOfMemory ||
         _recycledDueToLackOfResources;
} // Statistics::anyDiscarded ()


/**
 * Build statistics from the kernel.
 * 
 * @param kernel the Vampire kernel
 * @since 01/01/2003 Manchester
 * @since 28/06/2003 Manchester, result initialization added
 */
Statistics::Statistics (const VampireKernel& kernel)
  :
  // general
  _versionNumber (VampireKernel::versionNumber()),
  _versionDescription (VampireKernel::versionDescription()),
  _terminationReason (kernel.terminationReason()),
  _time (kernel.statTime()),
  _memory (kernel.statMemory()),
  // number of various categories of clauses
  _generated (kernel.statNumOfGeneratedCl()),
  _retained (kernel.statNumOfRetainedCl()),
  _used (kernel.statNumOfSelectedCl()),
  _currentlyActive (kernel.statNumOfCurrentlyActiveCl()),
  _currentlyPassive (kernel.statNumOfCurrentlyPassiveCl()),
  // generating inferences
  _resolutions (kernel.statNumOfClGeneratedByResolution()),
  _forwardSuperpositions (kernel.statNumOfClGeneratedByForwSup()),
  _backwardSuperpositions (kernel.statNumOfClGeneratedByBackSup()),
  _equalityResolutions (kernel.statNumOfClGeneratedByEqRes() +
                        kernel.statNumOfClSimplifiedByEqualityResol()),
  _equalityFactorings (kernel.statNumOfClGeneratedByEqFact()),
  _preprocessed (kernel.statNumOfClGeneratedByPrepro()),
  _reanimated (kernel.statNumOfClGeneratedByReanimation()),
  // simplifying inferences
  _forwardSubsumptionResolutions (kernel.statNumOfClSimplifiedByForwSubsResol()),
  _forwardDemodulated (kernel.statNumOfClSimplifiedByForwDemod()),
  _backwardDemodulated (kernel.statNumOfClSimplifiedByBackDemod()),
  _theoryInferences (kernel.statNumOfClSimplifiedByBuiltInTheories()),
  // redundant
  _forwardSubsumed (kernel.statNumOfForwSubsumedCl()),
  _backwardSubsumed (kernel.statNumOfBackSubsumedCl()),
  _discardedPropositionalTautologies (kernel.statNumOfDiscardedPropositionalTautologies()),
  _discardedEquationalTautologies (kernel.statNumOfDiscardedEquationalTautologies()),
  _murderedPassiveOrphans (kernel.statNumOfMurederedPassiveOrphans()),
  _murderedActiveOrphans (kernel.statNumOfMurederedActiveOrphans()),
  // splitting
  _splittingInferences (kernel.statNumOfSplitCl()),
  _differentSplitComponents (kernel.statNumOfDifferentSplitComponents()),
  _averageLiteralsPerComponent(kernel.statAverageNumOfSplitComponentsPerCl()),
  _forwardDemodulatedOnBranches (kernel.statNumOfClSimplifiedByForwDemodOnSplitBranches()),
  _backwardDemodulatedOnBranches (kernel.statNumOfClSimplifiedByBackDemodOnSplitBranches()),
  // discarded
  _discardedWithTooHeavyLiterals (kernel.statNumOfDiscardedClWithTooBigLiterals()),
  _discardedTooDeep (kernel.statNumOfDiscardedTooDeepCl()),
  _discardedTooHeavy (kernel.statNumOfDiscardedTooBigCl()),
  _discardedWithTooManyVariables (kernel.statNumOfDiscardedClWithTooManyVariables()),
  _newClausesDiscardedDueToLackOfMemory (kernel.statNumOfNewClDiscardedDueToLackOfMemory()),
  _recycledDueToLackOfResources (kernel.statNumOfClRecycledDueToLackOfResources()),
  _result("unknown")
{
} // Statistics::Statistics


// output statistics to a astream
// 02/01/2003 Manchester
/**
 * Output statistics to a stream.
 * 
 * @param str the output stream.
 * @since 02/01/2003 Manchester
 * @since 28/06/2003 Manchester, changed to output also in the XML format
 */
void Statistics::output (ostream& str) const
{
  Options::Statistics level = options->statistics();

  if (level == Options::STATISTICS_OFF) {
    return;
  }
  if (options->outputSyntax() == XML) {
    toXML(str);
    return;
  }

  str << "=========== Statistics ==========\n";

  if (level == Options::STATISTICS_DETAILED) {
    str << "version: " << _versionNumber << ' '
	<< _versionDescription << '\n'
	<< "=== General:\n";
  }

  str << "time: " << _time << "s\n"
      << "memory: " << ((float)(_memory/100000))/10 << "Mb\n"
      << "termination reason: ";

  outputTerminationReason(str);
  str << '\n';

  if (level == Options::STATISTICS_DETAILED) {
    // Statistics on clauses
    str << "=== Clauses:\n" 
	<< "generated: " << _generated << '\n'
        << "retained: " << _retained << '\n'
        << "total active: " << _used << '\n'
        << "currently active: " << _currentlyActive << '\n'
        << "currently passive: " << _currentlyPassive << '\n';

    // Generating inferences
    if ( anyGeneratingInference() ) {
      str << "=== Generating inferences:\n";
      if ( _resolutions) {
	str << "resolution: " << _resolutions << '\n';
      }
      if ( _forwardSuperpositions) {
	str << "forward superposition: " << _forwardSuperpositions << '\n';
      }
      if ( _backwardSuperpositions) {
	str << "backward superposition: " << _backwardSuperpositions << '\n';
      }
      if ( _equalityResolutions) {
	str << "equality resolution: " << _equalityResolutions << '\n';
      }
      if ( _equalityFactorings) {
	str << "equality factoring: " << _equalityFactorings << '\n';
      }
    }

    // Simplifying inferences
    if ( anySimplifyingInference() ) {
      str << "=== Simplifying inferences:\n";
      if (_forwardSubsumptionResolutions) {
	str << "forward subsumption resolution: " << _forwardSubsumptionResolutions << '\n';
      }
      if (_forwardDemodulated) {
	str << "forward demodulation: " << _forwardDemodulated << '\n';
      }
      if (_backwardDemodulated) {
	str << "backward demodulation: " << _backwardDemodulated << '\n';
      }
      if (_theoryInferences) {
	str << "theory inferences: " << _theoryInferences << '\n';
      }
    }

    if ( anyRedundantClause() ) {
      str << "=== Redundant clauses:\n";
      if ( _forwardSubsumed ) {
	str << "forward subsumed: " << _forwardSubsumed << '\n';
      }
      if ( _backwardSubsumed) {
	str << "backward subsumed: " << _backwardSubsumed << '\n';
      }
      if (_discardedPropositionalTautologies) {
	str << "propositional tautologies: " << _discardedPropositionalTautologies << '\n';
      }
      if (_discardedEquationalTautologies) {
	str << "equational tautologies: " << _discardedEquationalTautologies << '\n';
      }
      if (_murderedActiveOrphans) {
	str << "active orphans murdered: " << _murderedActiveOrphans << '\n';
      }
      if (_murderedPassiveOrphans) {
	str << "passive orphans murdered: " <<_murderedPassiveOrphans << '\n';
      }
    }

    // Splitting (if any)
    if ( anySplitting() ) {
      str << "=== Splitting:\n"
	  << "splitting inferences: " <<_splittingInferences << '\n'
	  << "distinct components: " << _differentSplitComponents << '\n'
	  << "average number of literals in a component: " << _averageLiteralsPerComponent << '\n';
      if (_forwardDemodulatedOnBranches) {
	str << "forward demodulation on branches: " << _forwardDemodulatedOnBranches << '\n';
      }
      if (_backwardDemodulatedOnBranches) {
	str << "backward demodulation on branches: " << _backwardDemodulatedOnBranches << '\n';
      }
    }

    if ( anyDiscarded() ) {
      str << "=== Discarded clauses:\n";
      if (_discardedTooHeavy) {
	str << "too heavy: " << _discardedTooHeavy << '\n';
      }
      if (_discardedTooDeep) {
	str << "too deep: " << _discardedTooDeep << '\n';
      }
      if (_discardedWithTooHeavyLiterals) {
	str << "having too heavy literals: " << _discardedWithTooHeavyLiterals << '\n';
      }
      if (_discardedWithTooManyVariables) {
	str << "having too many variables: " << _discardedWithTooManyVariables << '\n';
      }
      if (_newClausesDiscardedDueToLackOfMemory) {
	str << "due to lack of memory: " << _newClausesDiscardedDueToLackOfMemory << '\n';
      }
      if (_recycledDueToLackOfResources) {
	str << "recycled: " << _recycledDueToLackOfResources << '\n';
      }
    }
  }

  str << "======= End of statistics =======\n";
} // Statistics::output


/**
 * Output statistics to a stream in the XML format.
 * 
 * @param str the output stream.
 * @since 28/06/2003 Manchester
 */
void Statistics::toXML (ostream& str) const
{
  bool detailed = options->statistics() == Options::STATISTICS_DETAILED;

  str << "<statistics test_id='" 
      << options->testId()
      << "' test_number='" 
      << options->testNumber()
      << "'>\n";

  toXML(str, "problem_name", options->problemName());
  char hostname[256];
  gethostname(hostname,256);
  toXML(str, "hostname", hostname);

  if (detailed) {
    toXML(str, "version_date", __DATE__);
    toXML(str, "version_time", __TIME__);
    toXML(str, "version_number", VampireKernel::versionNumber());
    toXML(str, "version_description", VampireKernel::versionDescription());
  }

  toXML(str, "time", _time);
  toXML(str, "memory", (float)(_memory/100000)/10);

  // termination reason
  startXMLElement(str, "termination_reason");
  outputTerminationReason(str);
  endEmptyXMLElement(str);

  if (detailed) {
    toXML(str, "generated_clauses", _generated);
    toXML(str, "retained_clauses", _retained);
    toXML(str, "total_active", _used);
    toXML(str, "currently_active", _currentlyActive);
    toXML(str, "currently_passive", _currentlyPassive);

    // generating inferences
    if ( _resolutions) {
      toXML(str, "resolution", _resolutions);
    }
    if ( _forwardSuperpositions) {
      toXML(str, "forward_superposition", _forwardSuperpositions);
    }
    if ( _backwardSuperpositions) {
      toXML(str, "backward_superposition", _backwardSuperpositions);
    }
    if ( _equalityResolutions) {
      toXML(str, "equality_resolution", _equalityResolutions);
    }
    if ( _equalityFactorings) {
      toXML(str, "equality_factoring", _equalityFactorings);
    }

    // simplifying inferences
    if (_forwardSubsumptionResolutions) {
      toXML(str, "forward_subsumption_resolution", _forwardSubsumptionResolutions);
    }
    if (_forwardDemodulated) {
      toXML(str, "forward_demodulation", _forwardDemodulated);
    }
    if (_backwardDemodulated) {
      toXML(str, "backward_demodulation", _backwardDemodulated);
    }
    if (_theoryInferences) {
      toXML(str, "theory_inferences", _theoryInferences);
    }
    // discarded clauses
    if ( _forwardSubsumed ) {
      toXML(str, "forward_subsumed", _forwardSubsumed);
    }
    if ( _backwardSubsumed) {
      toXML(str, "backward_subsumed", _backwardSubsumed);
    }
    if (_discardedPropositionalTautologies) {
      toXML(str, "propositional_tautologies", _discardedPropositionalTautologies);
    }
    if (_discardedEquationalTautologies) {
      toXML(str, "equational_tautologies", _discardedEquationalTautologies);
    }
    if (_murderedActiveOrphans) {
      toXML(str, "murdered_active_orphans", _murderedActiveOrphans);
    }
    if (_murderedPassiveOrphans) {
      toXML(str, "murdered_passive_orphans", _murderedPassiveOrphans);
    }

    // splitting
    if ( anySplitting() ) {
      toXML(str, "splitting_inferences", _splittingInferences);
      toXML(str, "distinct_components", _differentSplitComponents);
      toXML(str, "average_number_of_literals_per_component", _averageLiteralsPerComponent);
      if (_forwardDemodulatedOnBranches) {
	toXML(str, "forward_demodulation_on_branches", _forwardDemodulatedOnBranches);
      }
      if (_backwardDemodulatedOnBranches) {
	toXML(str, "backward_demodulation_on_branches", 
	      _backwardDemodulatedOnBranches);
      }
    }
    
    // discarded
    if (_discardedTooHeavy) {
      toXML(str, "too_heavy_clauses", _discardedTooHeavy);
    }
    if (_discardedTooDeep) {
      toXML(str, "too_deep_clauses", _discardedTooDeep);
    }
    if (_discardedWithTooHeavyLiterals) {
      toXML(str, "having_too_heavy_literals", _discardedWithTooHeavyLiterals);
    }
    if (_discardedWithTooManyVariables) {
      toXML(str, "having_too_many_variables", _discardedWithTooManyVariables);
    }
    if (_newClausesDiscardedDueToLackOfMemory) {
      toXML(str, "new_clausesdiscarded_due_to_lack_of_memory", 
	    _newClausesDiscardedDueToLackOfMemory);
    }
    if (_recycledDueToLackOfResources) {
      toXML(str, "recycled", _recycledDueToLackOfResources);
    }
  }

  toXML(str, "result", _result);

  str << "</statistics>\n";
} // Statistics::toXML


/**
 * Tabulate statistics to a stream in the Prolog format.
 *
 * @since 27/05/2003 Manchester
 * @param str the output stream
 */
void Statistics::tabulate (ostream& str) const
{
  tabulate(str, "time", _time);
  tabulate(str, "memory", (float)(_memory/100000)/10);

  // termination reason
  startTab(str, "termination_reason");
  str << '\'';
  outputTerminationReason(str);
  str << '\'';
  endTab(str);

  tabulate(str, "generated_clauses", _generated);
  tabulate(str, "retained_clauses", _retained);
  tabulate(str, "total_active", _used);
  tabulate(str, "currently_active", _currentlyActive);
  tabulate(str, "currently_passive", _currentlyPassive);

  // generating inferences
  if ( _resolutions) {
    tabulate(str, "resolution", _resolutions);
  }
  if ( _forwardSuperpositions) {
    tabulate(str, "forward_superposition", _forwardSuperpositions);
  }
  if ( _backwardSuperpositions) {
    tabulate(str, "backward_superposition", _backwardSuperpositions);
  }
  if ( _equalityResolutions) {
    tabulate(str, "equality_resolution", _equalityResolutions);
  }
  if ( _equalityFactorings) {
    tabulate(str, "equality_factoring", _equalityFactorings);
  }

  // simplifying inferences
  if (_forwardSubsumptionResolutions) {
    tabulate(str, "forward_subsumption_resolution", _forwardSubsumptionResolutions);
  }
  if (_forwardDemodulated) {
    tabulate(str, "forward_demodulation", _forwardDemodulated);
  }
  if (_backwardDemodulated) {
    tabulate(str, "backward_demodulation", _backwardDemodulated);
  }
  if (_theoryInferences) {
    tabulate(str, "theory_inferences", _theoryInferences);
  }
  // discarded clauses
  if ( _forwardSubsumed ) {
    tabulate(str, "forward_subsumed", _forwardSubsumed);
  }
  if ( _backwardSubsumed) {
    tabulate(str, "backward_subsumed", _backwardSubsumed);
  }
  if (_discardedPropositionalTautologies) {
    tabulate(str, "propositional_tautologies", _discardedPropositionalTautologies);
  }
  if (_discardedEquationalTautologies) {
    tabulate(str, "equational_tautologies", _discardedEquationalTautologies);
  }
  if (_murderedActiveOrphans) {
    tabulate(str, "murdered_active_orphans", _murderedActiveOrphans);
  }
  if (_murderedPassiveOrphans) {
    tabulate(str, "murdered_passive_orphans", _murderedPassiveOrphans);
  }

  // splitting
  if ( anySplitting() ) {
    tabulate(str, "splitting_inferences", _splittingInferences);
    tabulate(str, "distinct_components", _differentSplitComponents);
    tabulate(str, "average_number_of_literals_per_component", _averageLiteralsPerComponent);
    if (_forwardDemodulatedOnBranches) {
      tabulate(str, "forward_demodulation_on_branches", _forwardDemodulatedOnBranches);
    }
    if (_backwardDemodulatedOnBranches) {
      tabulate(str, "backward_demodulation_on_branches", 
	       _backwardDemodulatedOnBranches);
    }
  }

  // discarded
  if (_discardedTooHeavy) {
    tabulate(str, "too_heavy_clauses", _discardedTooHeavy);
  }
  if (_discardedTooDeep) {
    tabulate(str, "too_deep_clauses", _discardedTooDeep);
  }
  if (_discardedWithTooHeavyLiterals) {
    tabulate(str, "having_too_heavy_literals", _discardedWithTooHeavyLiterals);
  }
  if (_discardedWithTooManyVariables) {
    tabulate(str, "having_too_many_variables", _discardedWithTooManyVariables);
  }
  if (_newClausesDiscardedDueToLackOfMemory) {
    tabulate(str, "new_clausesdiscarded_due_to_lack_of_memory", 
	     _newClausesDiscardedDueToLackOfMemory);
  }
  if (_recycledDueToLackOfResources) {
    tabulate(str, "recycled", _recycledDueToLackOfResources);
  }
} // Statistics::tabulate


/**
 * Start tab record for a statistics field in the Prolog format.
 *
 * @since 27/05/2003 Manchester
 * @param str the output stream
 * @param field the field name
 */
void Statistics::startTab (ostream& str, const char* field)
{
  str << "stat(" << field << ',';
} // Statistics::startTab


/**
 * Start tab record for a statistics field in the XML format.
 *
 * @since 28/06/2003 Manchester
 * @param str the output stream
 * @param field the field name
 */
void Statistics::startXMLElement (ostream& str, const char* field)
{
  str << " <stat name='" 
      << field 
      << "' value='";
} // Statistics::startXML


/**
 * End tab record for a statistics field in the Prolog format.
 *
 * @since 27/05/2003 Manchester
 * @param str the output stream
 */
void Statistics::endTab (ostream& str)
{
  str << ").\n";
} // Statistics::endTab


/**
 * End tab record for a statistics field in the XML format.
 *
 * @since 28/06/2003 Manchester
 * @param str the output stream
 */
void Statistics::endXMLElement (ostream& str)
{
  str << "'>\n";
} // Statistics::endXML


/**
 * End tab record for a statistics field in the XML format.
 *
 * @since 04/07/2003 Eindhoven
 * @param str the output stream
 */
void Statistics::endEmptyXMLElement (ostream& str)
{
  str << "'/>\n";
} // Statistics::endXML


/**
 * Tabulate a statistics record to a stream in the Prolog format.
 * 
 * @since 27/05/2003 Manchester
 * @param str the output stream
 * @param field the record name
 * @param value the record value
 */
void Statistics::tabulate (ostream& str, const char* field, const char* value)
{
  startTab(str, field);
  str << ",\'" << value;
  endTab(str);
  str << '\'';
} // Statistics::tabulate


/**
 * Write a statistics record to a stream in the XML format.
 * 
 * @since 27/05/2003 Manchester
 * @param str the output stream
 * @param field the record name
 * @param value the record value
 */
void Statistics::toXML (ostream& str, const char* field, const char* value)
{
  startXMLElement(str, field);
  str << value;
  endEmptyXMLElement(str);
} // Statistics::toXML


/**
 * Tabulate a statistics record to a stream in the Prolog format.
 * 
 * @since 27/05/2003 Manchester
 * @param str the output stream
 * @param field the record name
 * @param value the record value
 */
void Statistics::tabulate (ostream& str, const char* field, float value)
{
  startTab(str, field);
  str << value;
  endTab(str);
} // Statistics::tabulate


/**
 * Write a statistics record to a stream in the XML format.
 * 
 * @since 27/05/2003 Manchester
 * @param str the output stream
 * @param field the record name
 * @param value the record value
 */
void Statistics::toXML (ostream& str, const char* field, float value)
{
  startXMLElement(str, field);
  str << value;
  endEmptyXMLElement(str);
} // Statistics::toXML


/**
 * Tabulate a statistics record to a stream in the Prolog format.
 * 
 * @since 27/05/2003 Manchester
 * @param str the output stream
 * @param field the record name
 * @param value the record value
 */
void Statistics::tabulate (ostream& str, const char* field, long value)
{
  startTab(str, field);
  str << value;
  endTab(str);
} // Statistics::tabulate


/**
 * Write a statistics record to a stream in the XML format.
 * 
 * @since 27/05/2003 Manchester
 * @param str the output stream
 * @param field the record name
 * @param value the record value
 */
void Statistics::toXML (ostream& str, const char* field, long value)
{
  startXMLElement(str, field);
  str << value;
  endEmptyXMLElement(str);
} // Statistics::toXML


// output statistics to a astream
// 02/01/2003 Manchester
void Statistics::outputTerminationReason (ostream& str) const
{
  switch (_terminationReason) 
    {
    case VampireKernel::UnknownTerminationReason:
      str << "unknown";
      break;
    case VampireKernel::RefutationFound:
      str << "refutation found";
      break;
    case VampireKernel::Satisfiable:
      str << "satisfiability detected";
      break;
    case VampireKernel::EmptyPassive:
      str << "no passive clauses left";
      break;
    case VampireKernel::MaxPassive:
      str << "max_passive exceeded";
      break;
    case VampireKernel::MaxActive:
      str << "max_active exceeded";
      break;
    case VampireKernel::TimeLimitExpired:
      str << "time limit expired";
      break;
    case VampireKernel::AbortedOnAllocationFailure:
      str << "allocation failure";
      break;
    case VampireKernel::AbortedOnExternalInterrupt:
      str << "external interrupt";
      break;
    case VampireKernel::AbortedOnErrorSignal:
      str << "error signal";
      break;
    case VampireKernel::AbortedOnError:
      str << "error";
      break;
    }
} // Statistics::outputTerminationReason 


/**
 * Set the result field.
 * @param res character representing the result
 */
void Statistics::result (char res) 
{ 
  switch (res) 
    {
    case '0':
      _result = "0";
      return;
    case '+':
      _result = "+";
      return;
    case '-':
      _result = "-";
      return;
    default:
      ASS(false);
    } 
} // Statistics::result
