/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         Chaining.cpp
// Description:  Ordered chaining for transitive predicates.  
// Created:      Jun 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Jun 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Computation of main premise residual size was optimised.
//==================================================================
#include "Chaining.hpp"
#include "Clause.hpp"
#include "NewClause.hpp"
#include "RuntimeError.hpp"
#include "prefix.hpp"
#include "variables.hpp"
#include "Term.hpp"
using namespace BK;
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CHAINING
#  define DEBUG_NAMESPACE "Chaining"
#endif
#include "debugMacros.hpp"
//===================================================================


Chaining::Chaining(NewClause* newClauseBuffer,OR_INDEX* unifIndex) :
  _transitivity(0),
  _newClauseBuffer(newClauseBuffer),
  _unifIndex(unifIndex),
  _queryPremise(newClauseBuffer),
  _storedPremise(newClauseBuffer),
  _queryVarIndex(0UL),
  _storedVarIndex(1UL),
  _compiledConstraint(0),
  _storedConstraintCheck(0),
  _pseudoHeader((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity),
  _statBlockedChainingsThroughVariables(0UL),  
  _statSubsumedMainPremiseInstances(0UL),
  _statSubsumedSidePremiseInstances(0UL),
  _mainSelectedLitDemodulation(false),
  _sideSelectedLitDemodulation(false),   
   _statDemodulatedMainSelectedLits(0UL),
   _statDemodulatedSideSelectedLits(0UL)
{
  _residualQuerySizeFormula.SetIndex(_queryVarIndex);
  _unificationMachine.SetIndex(_storedVarIndex);
}; // Chaining::Chaining(NewClause* newClauseBuffer)

Chaining::~Chaining()
{  
  if (DestructionMode::isThorough())
    { 
      if (_compiledConstraint)
	{
	  SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
	};
      if (_storedConstraintCheck)
	{
	  SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
	};
    };
}; // Chaining::~Chaining()

void Chaining::init(NewClause* newClauseBuffer,OR_INDEX* unifIndex) 
{
  _transitivity = 0;
  _newClauseBuffer = newClauseBuffer;
  _unifIndex = unifIndex;
  _queryPremise.init(newClauseBuffer);
  _storedPremise.init(newClauseBuffer);
  _queryVarIndex = 0UL;
  _storedVarIndex = 1UL;
  _residualQuerySizeFormula.init();
  _compiledConstraint = 0;
  _storedConstraintCheck = 0;
  _pseudoHeader.init((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity);
  _unificationMachine.init();
  _statBlockedChainingsThroughVariables = 0UL; 
  _statSubsumedMainPremiseInstances = 0UL;
  _statSubsumedSidePremiseInstances = 0UL; 
  _mainSelectedLitDemodulation = false;
  _sideSelectedLitDemodulation = false;  
  _statDemodulatedMainSelectedLits = 0UL;
  _statDemodulatedSideSelectedLits = 0UL;

  _residualQuerySizeFormula.SetIndex(_queryVarIndex);
  _unificationMachine.SetIndex(_storedVarIndex);
}; // Chaining::init(NewClause* newClauseBuffer)

void Chaining::destroy()
{
  CALL("destroy()");
  if (DestructionMode::isThorough())
    { 
      if (_compiledConstraint)
	{
	  SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
	};
      if (_storedConstraintCheck)
	{
	  SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
	};
    };
  _unificationMachine.destroy();
  _residualQuerySizeFormula.destroy();
  _storedPremise.destroy();
  _queryPremise.destroy(); 
}; // void Chaining::destroy()

bool Chaining::loadQuery(Clause* cl)
{
  CALL("loadQuery(Clause* cl)");
  _query = cl;
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const TERM* lit = iter.currentLiteral();
      const Clause* transLaw = 
	_transitivity->transitivityLaw(lit->functor());
      if (transLaw)
	{
	  // at least one eligible literal
	  _outerContinuation = LoadForwardPositive;
	  // collect residual query size formula
	  _residualQuerySizeFormula.Reset();
	  ulong remainingSelected = cl->LitList()->numOfSelectedLits();
	  for (LiteralList::Iterator iter1(cl->LitList()); 
	       iter1.notEnd(); 
	       iter1.next())
	    {
	      if (remainingSelected)
		{
		  if (!_transitivity->transitivityLaw(iter1.currentLiteral()->functor()))
		    {
		      // residual literal
		      _residualQuerySizeFormula.Compile(iter1.currentLiteral());
		    };
		  --remainingSelected;
		}
	      else //  nonselected literal
		if (!iter1.currentLiteral()->isAnswerLiteral())
		  {
		    // residual literal
		    _residualQuerySizeFormula.Compile(iter1.currentLiteral());
		  };
	    };
	  
	  
	  _queryPremise.Load(_queryVarIndex,_query);

	  return true; 
	}; // if (transLaw) 
      --numOfSelectedLits;
    };
  return false;
}; // bool Chaining::loadQuery(Clause* cl)

bool Chaining::nextInference()
{
  CALL("nextInference()");
  switch (_outerContinuation)
    {
    case LoadForwardPositive:      
      _innerContinuation = FirstLit;
      // no break here

    case NextForwardPositive:
      if (nextForwardPositive())
	{
	  _outerContinuation = NextForwardPositive;
	  return true;
	};         
      _innerContinuation = FirstLit;
      // no break here

    case NextBackwardPositive:
      if (nextBackwardPositive())
	{
	  _outerContinuation = NextBackwardPositive;
	  return true;
	};
      _innerContinuation = FirstLit;
      // no break here

    case NextForwardNegative1:
      if (nextForwardNegative1())
	{
	  _outerContinuation = NextForwardNegative1;
	  return true;
	};         
      _innerContinuation = FirstLit;
      // no break here

    case NextBackwardNegative1:
      if (nextBackwardNegative1())
	{
	  _outerContinuation = NextBackwardNegative1;
	  return true;
	};         
      _innerContinuation = FirstLit;
      // no break here

    case NextForwardNegative2:
      if (nextForwardNegative2()) 
	{
	  _outerContinuation = NextForwardNegative2;
	  return true;
	};         
      _innerContinuation = FirstLit;
      // no break here
      

    case NextBackwardNegative2:
      if (nextBackwardNegative2())
	{
	  _outerContinuation = NextBackwardNegative2;
	  return true;
	};
      return false;

#ifdef DEBUG_NAMESPACE
    default:
      ICP("ICP0");
      return false;
#else
#  ifdef _SUPPRESS_WARNINGS_
      RuntimeError::report("Wrong _outerContinuation value in Chaining::nextInference()");
      return false;
#  endif
#endif
    };

#ifdef DEBUG_NAMESPACE
  ICP("ICP1");
  return false;
#else
#  ifdef _SUPPRESS_WARNINGS_
      RuntimeError::report("Illegal control point in Chaining::nextInference()");
      return false;
#  endif
#endif

}; // bool Chaining::nextInference()


bool Chaining::nextForwardPositive()
{
  CALL("nextForwardPositive()");
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.LastMax()) return false;
  lit = _queryPremise.nthMaxLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Negative())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg2();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	  
      if (!_chainingThroughVariables)
	{
	  ++_statBlockedChainingsThroughVariables;
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg1();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      if (!_compiledConstraint)
	_compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  ASSERT(_newLitHeader.Positive());
  _unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead);
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos1ChainingTree(_newLitHeader.functor()))))
    {
      _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize >_newClauseBuffer->MaxWeight())
	{
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_leaf;
	}; 

      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  goto try_next_leaf;
	};
	        
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->MinResWeight() 
	  + _query->MinResWeight() 
	  > _newClauseBuffer->MaxWeight())
        {
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  if (!_storedConstraintCheck)
	    _storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
	  ASSERT(_storedConstraintCheck);
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    goto try_next_cand;
	};
      
      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      _storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate);

      // first inference without factoring
      if (!buildResolvent(_queryTail,_currentCandidate->chainingTail()))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      ASSERT(_storedConstraintCheck);
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		goto next_factor;;
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_queryTail,_currentCandidate->chainingTail())) goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextForwardPositive()



bool Chaining::nextBackwardPositive()
{
  CALL("nextBackwardPositive()");
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.LastMax()) return false;
  lit = _queryPremise.nthMaxLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Negative())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg1();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	 
      if (!_chainingThroughVariables)
	{
	  ++_statBlockedChainingsThroughVariables;
	  goto next_lit;
	};  
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg2();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      if (!_compiledConstraint)
	_compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  ASSERT(_newLitHeader.Positive());
  _unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead);
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos2ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize >_newClauseBuffer->MaxWeight())
	{
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_leaf;
	}; 


      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;        
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->MinResWeight() 
	  + _query->MinResWeight() 
	  > _newClauseBuffer->MaxWeight())
        {
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  if (!_storedConstraintCheck)
	    _storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
	  ASSERT(_storedConstraintCheck);
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    goto try_next_cand;
	};


      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      _storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate);

      // first inference without factoring
      if (!buildResolvent(_currentCandidate->chainingTail(),_queryTail))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      ASSERT(_storedConstraintCheck);
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		goto next_factor;;
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),_queryTail)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextBackwardPositive()


bool Chaining::nextForwardNegative1()
{
  CALL("nextForwardNegative1()");
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.LastMax()) return false;
  lit = _queryPremise.nthMaxLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Negative())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg1();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      // non-singleton variable restriction
      // does not apply to this instance of the rule 
      if (!_chainingThroughVariables)
	{
	  ++_statBlockedChainingsThroughVariables;
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg2();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      if (!_compiledConstraint)
	_compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  _newLitHeader.Inversepolarity();
  ASSERT(_newLitHeader.Negative());
  _unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead);
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->neg1ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize >_newClauseBuffer->MaxWeight())
	{
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_leaf;
	}; 


      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  goto try_next_leaf;
	};

      if (subsumeMainPremise()) goto try_next_leaf;  
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->MinResWeight() 
	  + _query->MinResWeight() 
	  > _newClauseBuffer->MaxWeight())
        {
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  if (!_storedConstraintCheck)
	    _storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
	  ASSERT(_storedConstraintCheck);
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    goto try_next_cand;
	};

      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      _storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate);

      // first inference without factoring
      if (!buildResolvent(_queryTail,_currentCandidate->chainingTail()))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      ASSERT(_storedConstraintCheck);
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		goto next_factor;;
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_queryTail,_currentCandidate->chainingTail())) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextForwardNegative1()


bool Chaining::nextBackwardNegative1()
{
  CALL("nextBackwardNegative1()");
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.LastMax()) return false;
  lit = _queryPremise.nthMaxLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Positive())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg1();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	   
      if (!_chainingThroughVariables)
	{
	  ++_statBlockedChainingsThroughVariables;
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg2();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      if (!_compiledConstraint)
	_compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  ASSERT(_newLitHeader.Negative());
  _unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead);
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos1ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize >_newClauseBuffer->MaxWeight())
	{
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_leaf;
	}; 

      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  goto try_next_leaf;
	};
	        
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->MinResWeight() 
	  + _query->MinResWeight() 
	  > _newClauseBuffer->MaxWeight())
        {
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  if (!_storedConstraintCheck)
	    _storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
	  ASSERT(_storedConstraintCheck);
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    goto try_next_cand;
	};
      
      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      _storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate);

      // first inference without factoring
      if (!buildResolvent(_currentCandidate->chainingTail(),_queryTail))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      ASSERT(_storedConstraintCheck);
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		goto next_factor;;
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),_queryTail)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextBackwardNegative1()



bool Chaining::nextForwardNegative2()
{
  CALL("nextForwardNegative2()");
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.LastMax()) return false;
  lit = _queryPremise.nthMaxLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Negative())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg2();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      // non-singleton variable restriction
      // does not apply to this instance of the rule 
      if (!_chainingThroughVariables)
	{
	  ++_statBlockedChainingsThroughVariables;
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg1();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if ((_headTailComparison == Less) || 
      (_headTailComparison == Equal))
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if (_headTailComparison != Greater)
    {
      ASSERT(_headTailComparison == Incomparable);
      if (!_compiledConstraint)
	_compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  _newLitHeader.Inversepolarity();
  ASSERT(_newLitHeader.Negative());
  _unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead);
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->neg2ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize >_newClauseBuffer->MaxWeight())
	{
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_leaf;
	}; 

      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  goto try_next_leaf;
	};
	        
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->MinResWeight() 
	  + _query->MinResWeight() 
	  > _newClauseBuffer->MaxWeight())
        {
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  if (!_storedConstraintCheck)
	    _storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
	  ASSERT(_storedConstraintCheck);
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    goto try_next_cand;
	};

      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      _storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate);

      // first inference without factoring
      if (!buildResolvent(_currentCandidate->chainingTail(),_queryTail))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      ASSERT(_storedConstraintCheck);
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		goto next_factor;;
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_currentCandidate->chainingTail(),_queryTail)) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextForwardNegative2()


bool Chaining::nextBackwardNegative2()
{
  CALL("nextBackwardNegative2()");
  PrefixSym* lit;
  ulong predNum;
  Unifier::Variable* varInHead;
  
  switch (_innerContinuation)
    {
    case FirstLit:
      _currentQueryLit = 0UL;
      goto check_lit;
    case FirstFactoring:
      goto first_factor;
    case NextFactoring:
      goto next_factor;
    };

 next_lit:
  ++_currentQueryLit;

 check_lit:
  if (_currentQueryLit > _queryPremise.LastMax()) return false;
  lit = _queryPremise.nthMaxLit(_currentQueryLit);
  ASSERT(lit);
  // check if the literal is eligible 
  predNum = lit->Head().Func().functor();
  _currentTransitivityLaw = _transitivity->transitivityLaw(predNum);
  if ((!_currentTransitivityLaw) ||
      lit->Head().Func().Positive())
    goto next_lit;
  
  // positive transitive literal
  _queryHead = lit->Arg2();

  varInHead = 0;
  if (_queryHead->Head().isVariable())
    {
      varInHead = _queryHead->Head().var();
      ulong varNum =  varInHead->relativeNum();
      if (_query->singletonVar(varNum))
	goto next_lit;	   
      if (!_chainingThroughVariables)
	{
	  ++_statBlockedChainingsThroughVariables;
	  goto next_lit;
	};
    };

  // check maximality of the head
  // and if a constraint is needed  

  _queryTail = lit->Arg1();

  _headTailComparison = 
    SimplificationOrdering::current()->compareOnCurrentUnifier(_queryHead,_queryTail);
  
  if (_headTailComparison == Less)
    {
      goto next_lit;
    };	

  // compile ordering constraint, if necessary
  _constraintOnQuery = 0; 
  if ((_headTailComparison != Greater) &&
      (_headTailComparison != Equal))
    {
      ASSERT(_headTailComparison == Incomparable);
      if (!_compiledConstraint)
	_compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
      _compiledConstraint->load(_queryTail,_queryHead);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreater();
	  _constraintOnQuery = _compiledConstraint;
	};	
    };

  // prepare the unification machine and _queryPremise

  _newLitHeader = lit->Head().Func();
  ASSERT(_newLitHeader.Negative());
  _unificationMachine.CompilePseudoLiteral(_pseudoHeader,_queryHead);
  _queryPremise.Cut(_currentQueryLit);
  
  Unifier::current()->UnbindAll();
  
  if (_unificationMachine.FindFirstLeaf(*(_unifIndex->pos2ChainingTree(_newLitHeader.functor()))))
    {
      
    _mainSelectedLitDemodulationAttempted = false;
    process_leaf:  
    
      _residualQuerySize = _residualQuerySizeFormula.Compute();
      if (_residualQuerySize >_newClauseBuffer->MaxWeight())
	{
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_leaf;
	}; 

      // check the constraint associated with the query
      if (_constraintOnQuery &&
	  _constraintOnQuery->holdsOnCurrentUnifier())
	{
	  // bad substitution
	  goto try_next_leaf;
	};
	        
      if (subsumeMainPremise()) goto try_next_leaf;
      _unificationMachine.ResetCandidatesFromLeaf();

    try_next_cand:
      _currentCandidate  = _unificationMachine.NextCandidate();      
      if (!_currentCandidate) goto try_next_leaf;           
      if (_currentCandidate->TheClause()->MinResWeight() 
	  + _query->MinResWeight() 
	  > _newClauseBuffer->MaxWeight())
        {
	  _newClauseBuffer->AnotherTooHeavyClause();
	  goto try_next_cand;
        }; 
      
      // check the constraint associated with the candidate
      
      _storedConstraint = _currentCandidate->orderingConstraint();
      if (_storedConstraint)
	{	  
	  if (!_storedConstraintCheck)
	    _storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
	  ASSERT(_storedConstraintCheck);
	  if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
	    goto try_next_cand;
	};

      if (demodulateSelectedLits()) goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      _storedPremise.LoadOptimized(_storedVarIndex,_currentCandidate);

      // first inference without factoring
      if (!buildResolvent(_queryTail,_currentCandidate->chainingTail()))
	goto first_factor;
      _innerContinuation = FirstFactoring;
      return true;

    first_factor:
      _queryPremise.ResetFactoring();
      _storedPremise.ResetFactoring();

    next_factor:
      if (nextFactoring())
	{
	  if (_constraintOnQuery &&
	      _constraintOnQuery->holdsOnCurrentUnifier())
	    {
	      // bad substitution
	      goto next_factor;
	    };

	  if (_storedConstraint)
	    {	  
	      ASSERT(_storedConstraintCheck);
	      if (_storedConstraintCheck->holds(_storedConstraint,_storedVarIndex))
		goto next_factor;;
	    };
	  
	  if (subsumeSidePremise()) goto next_factor;
	  if (!buildResolvent(_queryTail,_currentCandidate->chainingTail())) 
	    goto next_factor;	           
           _innerContinuation = NextFactoring;
           return true;          
	};
      
      // no more inferences with this candidate
      _storedPremise.Unload();
      goto try_next_cand;

      
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (_unificationMachine.FindNextLeaf()) 
	{
	  _mainSelectedLitDemodulationAttempted = false;
	  goto process_leaf;
	};
    }; // if (_unificationMachine.FindFirstLeaf())

  // no more inferences for this literal
  _queryPremise.UnmarkCut();
  goto next_lit;
}; // bool Chaining::nextBackwardNegative2()



bool Chaining::posChainingLitSelectedIn(Clause* cl)
{
  CALL("posChainingLitSelectedIn(Clause* cl)");
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const TERM* lit = iter.currentLiteral();
      if (_transitivity->transitivityLaw(lit->functor()))
	{
	  if (lit->Positive()) return true;
	};
      --numOfSelectedLits;
    };
  // no appropriate literal
  return false;
}; // bool Chaining::posChainingLitSelectedIn(Clause* cl)

bool Chaining::negChainingLitSelectedIn(Clause* cl)
{
  CALL("negChainingLitSelectedIn(Clause* cl)");
  ulong numOfSelectedLits = cl->LitList()->numOfSelectedLits();
  for (LiteralList::Iterator iter(cl->LitList()); 
       numOfSelectedLits; 
       iter.next())
    {
      const TERM* lit = iter.currentLiteral();
      if (_transitivity->transitivityLaw(lit->functor()))
	{
	  if (lit->Negative()) return true;
	};
      --numOfSelectedLits;
    };
  // no appropriate literal
  return false;
}; // bool Chaining::negChainingLitSelectedIn(Clause* cl)



bool Chaining::buildResolvent(PrefixSym* newLitArg1,const TERM& newLitArg2)
{
  CALL("buildResolvent(PrefixSym* newLitArg1,const TERM& newLitArg2)");
  ASSERT(newLitArg2.isVariable() || newLitArg2.IsReference()); 

  if ((!_query->Active())
      || (!_storedPremise.TheClause()->Active()))
    return false;

  if (_residualQuerySize + 
      _storedPremise.TheClause()->MinResWeight() > 
      _newClauseBuffer->MaxWeight())
    {
      _newClauseBuffer->AnotherTooHeavyClause();
      return false;
    }; 

  _newClauseBuffer->openGenerated(ClauseBackground::BinRes);
  if (!_queryPremise.CollectNonmarkedLiterals())
    return false;
  if (!_storedPremise.CollectNonmarkedLiterals())
    return false;

  if ((!_newClauseBuffer->OpenNonpropLiteral(_newLitHeader,false)) ||
      (!_newClauseBuffer->PushPrefixTerm(newLitArg1)) ||
      (!_newClauseBuffer->PushStandardTerm(_storedVarIndex,newLitArg2)) ||
      (!_newClauseBuffer->CloseLiteral()))
    return false;
    

  // resolvent collected in _newClauseBuffer  
  _newClauseBuffer->UsedRule(ClauseBackground::BinRes);
  _newClauseBuffer->PushAncestor(_query);
  _newClauseBuffer->PushAncestor(_storedPremise.TheClause()); 
  _newClauseBuffer->PushAncestor(const_cast<Clause*>(_currentTransitivityLaw)); 
  return true;
}; // bool Chaining::buildResolvent(PrefixSym* newLitArg1,const TERM& newLitArg2)



bool Chaining::buildResolvent(const TERM& newLitArg1,PrefixSym* newLitArg2)
{
  CALL("buildResolvent(const TERM& newLitArg1,PrefixSym* newLitArg)");
  
  if ((!_query->Active())
      || (!_storedPremise.TheClause()->Active()))
    return false;

  if (_residualQuerySizeFormula.Compute() + 
      _storedPremise.TheClause()->MinResWeight() > 
      _newClauseBuffer->MaxWeight())
    {
      _newClauseBuffer->AnotherTooHeavyClause();
      return false;
    }; 

  _newClauseBuffer->openGenerated(ClauseBackground::BinRes);
  if (!_queryPremise.CollectNonmarkedLiterals())
    return false;
  if (!_storedPremise.CollectNonmarkedLiterals())
    return false;

  if ((!_newClauseBuffer->OpenNonpropLiteral(_newLitHeader,false)) ||
      (!_newClauseBuffer->PushStandardTerm(_storedVarIndex,newLitArg1)) ||
      (!_newClauseBuffer->PushPrefixTerm(newLitArg2)) ||
      (!_newClauseBuffer->CloseLiteral()))
    return false;
    

  // resolvent collected in _newClauseBuffer  
  _newClauseBuffer->UsedRule(ClauseBackground::BinRes);
  _newClauseBuffer->PushAncestor(_query);
  _newClauseBuffer->PushAncestor(_storedPremise.TheClause()); 
  _newClauseBuffer->PushAncestor(const_cast<Clause*>(_currentTransitivityLaw)); 
  return true;
}; // bool Chaining::buildResolvent(const TERM& newLitArg1,PrefixSym* newLitArg2)



bool Chaining::nextFactoring()
{
  CALL("nextFactoring()");
  
  if (_queryPremise.WasFactored())
   {
    if (_queryPremise.GetNextFactor())
     {
      return true;
     }
    else 
     {
      _queryPremise.ResetFactoring();
      if (_storedPremise.WasFactored())
       {
        return _storedPremise.GetNextFactor();
       }
      else
       {
        return _storedPremise.GetFirstFactor(_storedPremise.CutLit()+1);
       };
     };
   }
  else 
   {
    if (_queryPremise.GetFirstFactor(0))
     {
      return true;
     }
    else
     {
      if (_storedPremise.WasFactored())
       {
        return _storedPremise.GetNextFactor();
       }
      else
       {
        return _storedPremise.GetFirstFactor(_storedPremise.CutLit()+1);
       };
     };
   };
}; // bool Chaining::nextFactoring()

bool Chaining::subsumeMainPremise()
{
  CALL("subsumeMainPremise()"); 
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_query->hasSubsumedPremiseInstance()) return false;
  _query->blockForSubsumption();
  ulong maxWeight = _newClauseBuffer->MaxWeight();
  _newClauseBuffer->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _newClauseBuffer->openGenerated(ClauseBackground::Unspecified);
  if (collectMainPremiseInstance())   
    {
      _newClauseBuffer->endOfClause();
      subsumingClause = _newClauseBuffer->FSubsume();
      ASSERT(subsumingClause != _query);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_newClauseBuffer->WasTautology()) goto discarded;
      goto return_false;
    };
  
 return_false:
  _query->unblockForSubsumption();
  _newClauseBuffer->SetMaxWeight(maxWeight);
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++_statSubsumedMainPremiseInstances;
 discarded:
  _query->unblockForSubsumption();
  _newClauseBuffer->SetMaxWeight(maxWeight);
  return true;
}; // bool Chaining::subsumeMainPremise()

bool Chaining::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");
  for(LiteralList::Iterator iter(_query->LitList());
      iter.notEnd();
      iter.next())     
    if (!(_newClauseBuffer->PushStandardLit(_queryVarIndex,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) 
      return false; 
  return true;
}; // bool Chaining::collectMainPremiseInstance()



bool Chaining::subsumeSidePremise()
{
  CALL("subsumeSidePremise()"); 
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (_currentCandidate->TheClause()->hasSubsumedPremiseInstance()) return false;
  _currentCandidate->TheClause()->blockForSubsumption();
  ulong maxWeight = _newClauseBuffer->MaxWeight();
  _newClauseBuffer->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _newClauseBuffer->openGenerated(ClauseBackground::Unspecified);
  if (collectSidePremiseInstance())   
    {
      _newClauseBuffer->endOfClause();
      subsumingClause = _newClauseBuffer->FSubsume();
      ASSERT(subsumingClause != _currentCandidate->TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_newClauseBuffer->WasTautology()) goto discarded;
      goto return_false;
    };
  
 return_false:
  _currentCandidate->TheClause()->unblockForSubsumption();
  _newClauseBuffer->SetMaxWeight(maxWeight);
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++_statSubsumedSidePremiseInstances;
 discarded:
  _currentCandidate->TheClause()->unblockForSubsumption();
  _newClauseBuffer->SetMaxWeight(maxWeight);
  return true;
}; // bool Chaining::subsumeSidePremise()

bool Chaining::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");
  for(LiteralList::Iterator iter(_currentCandidate->TheClause()->LitList());
      iter.notEnd();
      iter.next())     
    if (!(_newClauseBuffer->PushStandardLit(_storedVarIndex,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) 
      return false; 
  return true;
}; // bool Chaining::collectSidePremiseInstance()




bool VK::Chaining::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");

  Clause* simplifyingClause;
  if (_mainSelectedLitDemodulation)
    {   
      if (_mainSelectedLitDemodulationAttempted)
	{
	  if (_mainSelectedLitDemodulated)
	    {
	      ++_statDemodulatedMainSelectedLits;
	      return true;
	    };
	  // proceed with the side premise
	}
      else
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  ulong maxWeight = _newClauseBuffer->MaxWeight();
	  _newClauseBuffer->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
	  _newClauseBuffer->openGenerated(ClauseBackground::Unspecified);
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto())
	    {	   
	      _newClauseBuffer->endOfClause();   	
      
	      simplifyingClause = _newClauseBuffer->canBeDemodulated();
	      ASSERT(simplifyingClause != _query);
	      ASSERT(simplifyingClause != _currentCandidate->TheClause());
	      _newClauseBuffer->SetMaxWeight(maxWeight);
	      
	      if (simplifyingClause)
		{	
		  _query->protectParamodulationsInto();
		  _currentCandidate->TheClause()->protectParamodulationsInto();
		  ++_statDemodulatedMainSelectedLits;
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	      // proceed with the side premise
	    }
	  else // analyse the result
	    {
	      _newClauseBuffer->SetMaxWeight(maxWeight);
	      return _newClauseBuffer->WasTautology();
	    };
	};
    };

  if (_sideSelectedLitDemodulation)
    {
      ulong maxWeight = _newClauseBuffer->MaxWeight();
      _newClauseBuffer->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
      _newClauseBuffer->openGenerated(ClauseBackground::Unspecified);
      if (collectSidePremiseSubtermsAvailableForParamodulationInto())
	{	   
	  _newClauseBuffer->endOfClause();   	
      
	  simplifyingClause = _newClauseBuffer->canBeDemodulated();
	  ASSERT(simplifyingClause != _query);
	  ASSERT(simplifyingClause != _currentCandidate->TheClause());
	  _newClauseBuffer->SetMaxWeight(maxWeight);
	      
	  if (simplifyingClause)
	    {	
	      _query->protectParamodulationsInto();
	      _currentCandidate->TheClause()->protectParamodulationsInto();
	      ++_statDemodulatedSideSelectedLits;
	      return true;
	    }
	  else
	    return false;
	}
      else // analyse the result
	{
	  _newClauseBuffer->SetMaxWeight(maxWeight);
	  return _newClauseBuffer->WasTautology();
	};
    }
  else
    return false;
}; // bool VK::Chaining::demodulateSelectedLits()



bool 
VK::Chaining::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()");
  ulong numOfSelectedLits = _query->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_query->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_newClauseBuffer->pushPseudoLit(_queryVarIndex,
					     TERM::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_newClauseBuffer->PushStandardLit(_queryVarIndex,lit,false))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_newClauseBuffer->PushStandardLit(_queryVarIndex,lit,false))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; // bool VK::Chaining::collectMainPremiseSubtermsAvailableForParamodulationInto()



bool 
VK::Chaining::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  ulong numOfSelectedLits = _currentCandidate->TheClause()->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_currentCandidate->TheClause()->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_newClauseBuffer->pushPseudoLit(_storedVarIndex,
					     TERM::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_newClauseBuffer->PushStandardLit(_storedVarIndex,lit,false))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_newClauseBuffer->PushStandardLit(_storedVarIndex,lit,false))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; // bool VK::Chaining::collectSidePremiseSubtermsAvailableForParamodulationInto()




//=================================================================

