/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         ForwardMatchingSubstitutionCache.h
// Description:  Fast mapping of clause variables
//               into their instances from a substitution
//               collected by forward matching.
// Created:      Apr 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//===================================================================
#ifndef FORWARD_MATCHING_SUBSTITUTION_CACHE_H
#define FORWARD_MATCHING_SUBSTITUTION_CACHE_H
//===================================================================
#ifndef NO_DEBUG_VIS
#include <iostream.h>
#endif
#include "jargon.h"
#include "VampireKernelDebugFlags.h"
#include "VampireKernelConst.h"
#include "Flatterm.h"
#include "Term.h"

#ifndef NO_DEBUG
#include "LinearCombination.h"
#endif
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_SUBSTITUTION_CACHE
 #define DEBUG_NAMESPACE "ForwardMatchingSubstitutionCache"
#endif
#include "debugMacros.h"
//===================================================================

namespace VK
{
  using namespace BK;
  using namespace Gematogen;

class ForwardMatchingSubstitutionCache
{
 public: 
  class Instance;
 public:
  ForwardMatchingSubstitutionCache() {};
  ~ForwardMatchingSubstitutionCache() {};
  void reset() 
    {
      DOP(_debugVariables.reset());
    };
  void associate(ulong var,const Flatterm* term)
    {
      CALL("associate(ulong var,const Flatterm* term)");
      ASSERT(!_debugVariables.coefficient(var));
      _subst[var] = term;
      DOP(_debugVariables.add1(var));
    };
  void endOfCollection() {};
  const Flatterm* map(ulong var) const 
    {
      CALL("map(ulong var) const");
      ASSERT(_debugVariables.coefficient(var));
      return _subst[var]; 
    };

#ifndef NO_DEBUG_VIS
  ostream& output(ostream& str) const
    {
      for (const ulong* v = _debugVariables.begin(); v < _debugVariables.end(); ++v)
	str << "X" << *v << " ---> " << map(*v) << "\n";
      return str;
    };
#endif

 private:
  const Flatterm* _subst[VampireKernelConst::MaxNumOfVariables];
#if (defined DEBUG_NAMESPACE) || (!(defined NO_DEBUG_VIS))
  LinearCombination<long,VampireKernelConst::MaxNumOfVariables> _debugVariables;
#endif
}; // class ForwardMatchingSubstitutionCache

}; // namespace VK

#ifndef NO_DEBUG_VIS
inline
ostream& operator<<(ostream& str,const VK::ForwardMatchingSubstitutionCache* subst)
{
  return subst->output(str);
};
#endif

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_SUBSTITUTION_CACHE
 #define DEBUG_NAMESPACE "ForwardMatchingSubstitutionCache::Instance"
#endif
#include "debugMacros.h"
//===================================================================

namespace VK
{
  using namespace BK;
  using namespace Gematogen;

class ForwardMatchingSubstitutionCache::Instance
{
 public:
  Instance() : 
    _dummyTerm((TERM::Arity)0,(TERM::Functor)0,TERM::PositivePolarity)  
    {
    };
  ~Instance() {};
  void reset(const TERM& t,const ForwardMatchingSubstitutionCache* subst);
  void reset(const TERM* t,const ForwardMatchingSubstitutionCache* subst);
  const TERM& symbol() const;
  bool next();
  bool after();
 private:
  const ForwardMatchingSubstitutionCache* _subst;
  TERM::Iterator _topLevel;
  Flatterm::Iterator _substIter;
  bool _onTop;
  TERM _dummyTerm;
}; // class ForwardMatchingSubstitutionCache::Instance

}; // namespace VK
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_MATCHING_SUBSTITUTION_CACHE
 #define DEBUG_NAMESPACE "ForwardMatchingSubstitutionCache::Instance"
#endif
#include "debugMacros.h"
//===================================================================

namespace VK
{
  using namespace BK;
  using namespace Gematogen;

inline
void 
ForwardMatchingSubstitutionCache::Instance::reset(const TERM& t,
						  const ForwardMatchingSubstitutionCache* subst)
{
  CALL("reset(const TERM& t,const ForwardMatchingSubstitutionCache* subst)");
  if (t.isVariable())
    {
      _subst = subst;
      _substIter.Reset(subst->map(t.var()));
      _topLevel.Reset(&_dummyTerm);
      NEVER(_topLevel.Next());
      _onTop = false;
    }
  else
    {
      ASSERT(t.IsReference());
      reset(t.First(),subst);
    };
}; // void ForwardMatchingSubstitutionCache::Instance::reset(const TERM& t,const ForwardMatchingSubstitutionCache* subst)

inline
void 
ForwardMatchingSubstitutionCache::Instance::reset(const TERM* t,
						  const ForwardMatchingSubstitutionCache* subst)
{
  CALL("reset(const TERM* t,const ForwardMatchingSubstitutionCache* subst)");
  ASSERT(t->isComplex());
  _subst = subst;
  _topLevel.Reset(t);
  _onTop = true;
}; // void ForwardMatchingSubstitutionCache::Instance::reset(const TERM* t,const ForwardMatchingSubstitutionCache* subst)


inline const TERM& ForwardMatchingSubstitutionCache::Instance::symbol() const
{
  CALL("symbol() const");
  if (_onTop) return _topLevel.symbolRef();
  ASSERT(!_substIter.EndOfTerm());
  return _substIter.CurrPos()->symbolRef();
}; // const TERM& ForwardMatchingSubstitutionCache::Instance::symbol() const

inline bool ForwardMatchingSubstitutionCache::Instance::next()
{
  CALL("next()");
  if (_onTop)
    {
    next_on_top:
      if (!_topLevel.Next()) return false;
      if (_topLevel.CurrentSymIsVar())
	{
	  _onTop = false;
	  _substIter.Reset(_subst->map(_topLevel.symbolRef().var()));
	};
      return true;
    }
  else // inside a substitution term
    {
      _substIter.NextSym();
      if (!_substIter.EndOfTerm()) return true;
      _onTop = true;
      goto next_on_top;
    };
}; // bool ForwardMatchingSubstitutionCache::Instance::next()



inline bool ForwardMatchingSubstitutionCache::Instance::after()
{
  CALL("after()");
  if (_onTop)
    {
    next_on_top:
      if (!_topLevel.SkipTerm()) return false;
      if (_topLevel.CurrentSymIsVar())
	{
	  _onTop = false;
	  _substIter.Reset(_subst->map(_topLevel.symbolRef().var()));
	};
      return true;
    }
  else // inside a substitution term
    {
      _substIter.after();
      if (!_substIter.EndOfTerm()) return true;
      _onTop = true;
      goto next_on_top;
    };
}; // bool ForwardMatchingSubstitutionCache::Instance::after()


}; // namespace VK



//===================================================================
#endif

