/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         Splitting.h
// Description:  Core of splitting without backtracking.
//               Additionally implements negative equality splitting.
// Created:      Dec 12, 2001 
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Revised:      Dec 17, 2001.
//               Array ulong _litNumMap[HellConst::MaxNumOfLiterals]
//               is replace by 
//               TMP_LITERAL* _numLitMap[HellConst::MaxNumOfLiterals] 
//============================================================================
#ifndef SPLITTING_H
//============================================================================
#define SPLITTING_H 
#include "jargon.h"
#include "RuntimeError.h"
#include "VampireKernelDebugFlags.h"
#include "VampireKernelConst.h"
#include "Array.h"
#include "ExpandingMultiset.h"
#include "GlobAlloc.h"
#include "DivIntoComp.h"
#include "ClauseQueue.h"
#include "TmpLitList.h"
#include "OpenNewClause.h"
#include "DestructionMode.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SPLITTING
 #define DEBUG_NAMESPACE "Splitting"
#endif
#include "debugMacros.h"
//============================================================================

namespace VK
{
class ComponentName;
class ForwardSubsumptionFrontEnd;
class Clause;
class PagedPassiveClauseStorage;
class ClauseAssembler;

class Splitting
{
 public:
  class Statistics
  {
   public:
    Statistics() { reset(); };
    ~Statistics() {};
    void reset() 
    {
     forwardSubsumed = 0;
     simplifiedByFSR = 0;
     differentComponents = 0;     
     split = 0;
     components = 0;
    };
   
    ulong forwardSubsumed;
    ulong simplifiedByFSR;
    ulong differentComponents;
    ulong split;
    ulong components;
  }; // class Statistics

 public:
  Splitting(OpenClauseBackground* ancestors);

  ~Splitting() 
    { 
      CALL("destructor ~Splitting()"); 
      if (DestructionMode::isThorough())
	{
	  for (ulong i = 0; i < _components.size(); ++i)
	    _components[i].reset();
	};
    };
  
  void setStyle(long num) { _splittingStyle = num; };
  void setSplittingWithNamingFlag(bool fl)
  {
   _splittingWithNamingFlag = fl;
  };  
  void setForwardSubsumptionFlag(bool fl)
  {
   _forwardSubsumptionFlag = fl;
  };
  void setForwardSubsumptionResolutionFlag(bool fl)
  {
   _forwardSubsumptionResolutionFlag = fl;
  };
  void setNegEqSplittingMinEligibleTermSize(ulong sz)
  {
   _negEqSplittingMinEligibleTermSize = sz;
  };
  void setForwardSubsumption(ForwardSubsumptionFrontEnd* fs)
  {
   _forwardSubsumption = fs;
  };
  void setComponentNamingService(ComponentName* compNaming)
  {
   _compName = compNaming;
  };   
  void setClauseAssembler(ClauseAssembler* clauseAssembler)
  {
   _clauseAssembler = clauseAssembler;
  };

  void adjustSizes(ulong newNumOfLits);


  // getting statistics 

  ulong statForwardSubsumed() const { return _statistics.forwardSubsumed; };
  ulong statSimplifiedByFSR() const { return _statistics.simplifiedByFSR; };
  ulong statSplit() const { return _statistics.split; };
  ulong statDifferentComponents() const { return _statistics.differentComponents; };
  ulong statComponents() const { return _statistics.components; };
  
  // keeping statistics
  void registerAnotherForwardSubsumed() { _statistics.forwardSubsumed++; };  
  void registerAnotherSimplifiedByFSR() { _statistics.simplifiedByFSR++; };
  void registerAnotherSplit(ulong numOfComponents) 
  { 
   _statistics.split++; 
   _statistics.components += numOfComponents;
  };
  void registerAnotherComponent() { _statistics.differentComponents++; };

  bool Split(TmpLitList& lits,ClauseQueue& newClauses);
  bool Split3(TmpLitList& lits,
              Clause*& answerClause,
              PagedPassiveClauseStorage& storage,
              ulong& numOfNewClauses,
              ulong& numOfRejectedNewClauses,
	      Clause::NewNumber* clauseNumberGenerator,
              bool showNewClauses);

  void splitNegativeEqualities(TmpLitList& lits,ClauseQueue& newClauses);

 private:

  bool NonblockingSplitting(TmpLitList& lits,ClauseQueue& newClauses);
  bool NonblockingSplitting3(TmpLitList& lits,  
                             Clause*& answerClause, 
                             PagedPassiveClauseStorage& storage,
                             ulong& numOfNewClauses,
                             ulong& numOfRejectedNewClauses,
                             Clause::NewNumber* clauseNumberGenerator,
                             bool showNewClauses);
   


  bool BlockingSplitting(TmpLitList& lits,ClauseQueue& newClauses);
  bool BlockingSplitting3(TmpLitList& lits,                            
                          Clause*& answerClause,
                          PagedPassiveClauseStorage& storage,
                          ulong& numOfNewClauses,
                          ulong& numOfRejectedNewClauses,
                          Clause::NewNumber* clauseNumberGenerator,
                          bool showNewClauses);

  void collectRegisteredComponents(TmpLitList& lits)
    {
      CALL("collectRegisteredComponents(TmpLitList& lits)");
      for (ulong n = 0; n < _registeredCompNums.numOfDiffElements(); n++) 
	{
	  lits.collectFrom(_components[_registeredCompNums.nth(n)]);
	};
    };


 private:
  long _splittingStyle;
  bool _splittingWithNamingFlag;
  bool _forwardSubsumptionFlag;
  bool _forwardSubsumptionResolutionFlag;
  long _negEqSplittingMinEligibleTermSize;
  OpenClauseBackground* _ancestors;
  Statistics _statistics;  
  ulong _maxNumOfLitsSoFar;
  DivIntoComp<VampireKernelConst::MaxNumOfVariables> _divisionIntoComponents;


  Array<GlobAlloc,TMP_LITERAL*,32UL,Splitting> _numLitMap;

  OpenNewClause _basicComponent;
  ComponentName* _compName;
  ClauseAssembler* _clauseAssembler;
  ForwardSubsumptionFrontEnd* _forwardSubsumption;

  Array<GlobAlloc,TmpLitList,32UL,Splitting> _components;


  ExpandingMultiset<GlobAlloc,ulong,32UL,Splitting> _registeredCompNums;


  Array<GlobAlloc,ulong,32UL,Splitting> _componentName;    

 
  Array<GlobAlloc,bool,32UL,Splitting> _newCompName;  

}; // class Splitting

}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SPLITTING
 #define DEBUG_NAMESPACE "Splitting"
#endif
#include "debugMacros.h"
//============================================================================


namespace VK
{
inline void Splitting::adjustSizes(ulong newNumOfLits)
{
 CALL("adjustSizes(ulong newNumOfLits)");
 if (newNumOfLits > _maxNumOfLitsSoFar)
  {
   _divisionIntoComponents.adjustSizes(newNumOfLits);
   
   DOP(_numLitMap.unfreeze());
   _numLitMap.expand(newNumOfLits);
   DOP(_numLitMap.freeze());

   ulong prevSize = _components.size();  
   DOP(_components.unfreeze());  
   _components.expand(newNumOfLits);
   DOP(_components.freeze());
   for (ulong i = prevSize; i < _components.size(); i++)
     _components[i].init();


   DOP(_registeredCompNums.unfreeze());
   _registeredCompNums.expand(newNumOfLits);
   DOP(_registeredCompNums.freeze());


   DOP(_componentName.unfreeze());
   _componentName.expand(newNumOfLits);
   DOP(_componentName.freeze());   


   DOP(_newCompName.unfreeze());
   _newCompName.expand(newNumOfLits);
   DOP(_newCompName.freeze());

   _maxNumOfLitsSoFar = newNumOfLits;
  }; 
}; // void Splitting::adjustSizes(ulong newNumOfLits) 

inline bool Splitting::Split(TmpLitList& lits,ClauseQueue& newClauses)
{
 CALL("Split(TmpLitList& lits,ClauseQueue& newClauses)");

 adjustSizes(lits.numOfAllLiterals()); 

 switch (_splittingStyle)
  {
   case 1L: return NonblockingSplitting(lits,newClauses);
   case 2L: return BlockingSplitting(lits,newClauses);
   default: 
    RuntimeError::report("Wrong splitting style number."); 
    return false; 
  };
}; //  bool Splitting::Split(TmpLitList& lits,ClauseQueue& newClauses)

inline bool Splitting::Split3(TmpLitList& lits,                            
                              Clause*& answerClause,
                              PagedPassiveClauseStorage& storage,
                              ulong& numOfNewClauses,
                              ulong& numOfRejectedNewClauses,
                              Clause::NewNumber* clauseNumberGenerator,
                              bool showNewClauses)
{
  CALL("Split3(TmpLitList& lits,...)"); 

  adjustSizes(lits.numOfAllLiterals()); 

  switch (_splittingStyle)
  {
   case 1L: return NonblockingSplitting3(lits,
                                         answerClause,
                                         storage,
                                         numOfNewClauses,
                                         numOfRejectedNewClauses,
                                         clauseNumberGenerator,
                                         showNewClauses);
   case 2L: return BlockingSplitting3(lits,
                                      answerClause,
                                      storage,
                                      numOfNewClauses,
                                      numOfRejectedNewClauses,
                                      clauseNumberGenerator,
                                      showNewClauses);
   default: 
    RuntimeError::report("Wrong splitting style number."); 
    return false; 
  };

}; // bool Splitting::Split3(TmpLitList& lits,PagedPassiveClauseStorage& storage,...



}; // namespace VK
//======================================================================
#endif
