/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         VampireKernel.h
// Description:  Vampire kernel interface. 
// Created:      May 19, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//======================================================================
#ifndef VAMPIRE_KERNEL_H
#define VAMPIRE_KERNEL_H
//======================================================================
#include <iostream.h>
#include <fstream.h>
#include "jargon.h"
#include "Float.h"
namespace VK
{
class Clause;
}; // namespace VK
//======================================================================

class VampireKernel
{
 public:
  typedef long TermWeightType;
  class Options;
  class Emergency;
  enum TerminationReason
  {
    UnknownTerminationReason,
    RefutationFound,
    Satisfiable,
    EmptyPassive,
    MaxPassive,
    MaxActive,
    TimeLimitExpired,
    AbortedOnAllocationFailure,
    AbortedOnExternalInterrupt,
    AbortedOnErrorSignal,
    AbortedOnError
  };
 public:
  // Initialisation and destruction
  VampireKernel();
  ~VampireKernel();
  void openSession(const Options& opt);
  void closeSession();
  void destroy(); // for debugging, mainly for detecting memory leaks

  // Version information
  
  static float versionNumber();
  static const char* versionDescription();

  // Receiving signature information

  void beginInputSignature();  
  void endInputSignature();
  void registerInputFunction(const char* name,
			     ulong arity,
			     ulong& id); 
  void registerInputPredicate(const char* name,
			      ulong arity,
			      bool isAnswer,
			      ulong& id); 
  
  void setSymbolWeightConstPart(ulong symId,const TermWeightType& weight);
  void setSymbolPrecedence(ulong symId,long prec); // prec can not be 0

  // Receiving input clauses
  
  void beginInputClauses();
  void endInputClauses();
  
  void beginInputClause(ulong clauseNumber,void* origin,bool goal);
  void endInputClause();
  
  void beginLiteral(bool positive,ulong predicateId);
  void endLiteral();
  
  void receiveFunction(ulong symId);
  void receiveVariable(ulong varNum);
  
  // Searching for a proof, includes preprocessing
  
  void saturate(ulong nextVacantClauseId,
		long timeLimit); // deciseconds 
  // throws Emergency and intercepts SIGVTALRM
  static VampireKernel* currentSession();
  void suspend();
  void resume();
  void tabulate();
  void reportInterrupt(int signalNumber);
  void reportErrorSignal(int signalNumber);
  void reportError();
  // Communicating search results

  TerminationReason terminationReason();
  const VK::Clause* refutation() const;
  
 private:
  class _Implementation; 
  
 private:
  _Implementation* _implementation;
}; // class VampireKernel

//======================================================================

class VampireKernel::Options
{
 public:
  enum SimplificationOrdering
  {
    NonrecursiveKBOrdering,
    StandardKBOrdering
  };
  enum TermWeightingScheme
  {
    UniformTermWeighting,
    NonuniformConstantTermWeighting
  };
 public:
  Options() { setDefaultValues(); };
  ~Options() {};
  void setDefaultValues();
  ostream& output(ostream& str);
 public:
  
  // session attributes
  const char* job_id;
  

  // main algorithm
    
  long main_alg;

  // inference rules
  bool no_inferences_between_rules;
  bool static_splitting;
  bool dynamic_splitting;
  long splitting_style;
  bool splitting_with_naming;
  long static_neg_eq_splitting;
  bool paramodulation;
  long sim_back_sup;
  long selection;
  long literal_comparison_mode;
  bool inherited_negative_selection;
     
  // strategy 
  bool forward_subsumption;
  bool fs_set_mode;
  bool fs_optimised_multiset_mode;
  bool fs_use_sig_filters;
  bool fs_old_style;
  long simplify_by_fs;
  long forward_demodulation;
  bool fd_on_splitting_branches;
  bool normalize_by_commutativity;
  long backward_demodulation;
  bool bd_on_splitting_branches;
  bool simplify_by_eq_res;
  bool backward_subsumption;
  bool bs_set_mode;
  bool bs_optimised_multiset_mode;
  long orphan_murder;
  long lrs;
  long elim_def;
  SimplificationOrdering simplification_ordering;
  long symbol_precedence_by_arity;
  bool header_precedence_kinky;
  TermWeightingScheme term_weighting_scheme;
  long symbol_weight_by_arity;

  // numeric options 
  long memory_limit; // kilobytes
  long allocation_buffer_size;
  double allocation_buffer_price;
  long max_skolem; 
  long max_active;
  long max_passive;
  long max_weight;
  long age_weight_ratio;
  float selected_pos_eq_penalty_coefficient;    
  float nongoal_penalty_coefficient;
  long first_time_check;   
  long first_mem_check;   
    
  // interface options    
  bool show_opt;
  bool show_input;
  bool show_prepro;
  bool show_gen;
  bool show_passive;
  bool show_active; 
  bool proof;
  bool show_profile;
  bool show_weight_limit_changes;  
  ofstream* tabulation;
}; // class VampireKernel::Options

inline 
ostream& operator<<(ostream& str,
		    const VampireKernel::Options::SimplificationOrdering& so)
{
  switch (so)
    {
    case VampireKernel::Options::NonrecursiveKBOrdering: 
      str << "nonrecursive_KB";
      break;
    case VampireKernel::Options::StandardKBOrdering: 
      str << "standard_KB";
      break;
    };
  return str;
};

inline 
ostream& operator<<(ostream& str,
		    const VampireKernel::Options::TermWeightingScheme& tws)
{
  switch (tws)
    {
    case VampireKernel::Options::UniformTermWeighting:
      str << "uniform";
      break;
    case VampireKernel::Options::NonuniformConstantTermWeighting:
      str << "nonuniform_constant";
      break;
    };
  return str;
};


//======================================================================

class VampireKernel::Emergency
{
 public:
  enum Type
  {
    AllocationFailure,
    TimeLimit
  };
 public:
  Emergency(const Type& t) : type_(t) {}; 
  ~Emergency() {};
  const Type& type() const { return type_; };
 private:
  Emergency() {};
 private:
  Type type_;
};

//======================================================================
#endif



