/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         backward_superposition.cpp
// Description:  Classes dealing with backward superposition.
// Created:      May 06, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::CompiledConstraintForUnifiers
//               replaced ORD_CONSTR
// Revised:      May 16, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bug fix in RecollectMaxWeightForm().
// Revised:      Jul 15, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Major refurbishment.
// Revised:      Jul 16, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Computation of main premise residual size was optimised.
//==================================================================
#include "backward_superposition.hpp"
#include "DestructionMode.hpp"
#include "VampireKernelInterceptMacros.hpp"
using namespace BK;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "BACKWARD_SUPERPOSITION"
#endif
#include "debugMacros.hpp"
//================================================


VK::BACKWARD_SUPERPOSITION::BACKWARD_SUPERPOSITION(NewClause* res,OR_INDEX* unifIndex) : 
  // main parameters
  _resolvent(res),
  _unifIndex(unifIndex),
  _simultaneousness(0L),
  _noInfBetweenRulesFlag(false),
  _paramodulateFromVariables(true),
  _mainPremiseSubsumption(false),
  _sidePremiseSubsumption(false),
  _mainSelectedLitDemodulation(false),
  _sideSelectedLitDemodulation(false),

  // statistics
  _statSubsumedMainPremiseInstances(0UL),
  _statSubsumedSidePremiseInstances(0UL),
  _statDemodulatedMainSelectedLits(0UL),
  _statDemodulatedSideSelectedLits(0UL),
  _statBlockedParamodulationsFromVariables(0UL),
  
  // main premise data
  _mainPremiseVarIndex(0UL),

  // current selected literal in the main premise
  _compiledConstraint(0), 

  // side premise data
  _sidePremiseVarIndex(1UL), 
  _storedConstraintCheck(0)
{
  _mainPremiseResidualSizeFormula.SetIndex(0UL);
  _mainPremiseSelectedLitResidualSizeFormula.SetIndex(0UL);

  _currentMainPosEqArg1Mem[0].Head().MkFunc(TERM((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity));
  _currentMainPosEqArg2Mem[0].Head().MkFunc(TERM((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity));
  
  _retrievalOfUnifiable.SetIndex(_sidePremiseVarIndex);
}; // BACKWARD_SUPERPOSITION::BACKWARD_SUPERPOSITION(NewClause* res,OR_INDEX* unifIndex)

VK::BACKWARD_SUPERPOSITION::~BACKWARD_SUPERPOSITION() 
{ 
  CALL("destructor ~BACKWARD_SUPERPOSITION()"); 
    
  if (DestructionMode::isThorough())
    {
      if (_storedConstraintCheck)
	{
	  SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
	};
      if (_compiledConstraint)
	{
	  SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
	};
    };
};  // BACKWARD_SUPERPOSITION::~BACKWARD_SUPERPOSITION() 
  
void VK::BACKWARD_SUPERPOSITION::init(NewClause* res,OR_INDEX* unifIndex)
{
  CALL("init()");
  // main parameters
  _resolvent = res;
  _unifIndex = unifIndex;
  _simultaneousness = 0L;
  _noInfBetweenRulesFlag = false;
  _paramodulateFromVariables = true;
  _mainPremiseSubsumption = false;
  _sidePremiseSubsumption = false;
  _mainSelectedLitDemodulation = false;
  _sideSelectedLitDemodulation = false;
  
  // statistics
  _statSubsumedMainPremiseInstances = 0UL;
  _statSubsumedSidePremiseInstances = 0UL;
  _statDemodulatedMainSelectedLits = 0UL;
  _statDemodulatedSideSelectedLits = 0UL;
  _statBlockedParamodulationsFromVariables = 0UL;

  // main premise data
  _mainPremiseVarIndex = 0UL;
  _mainSelectedLitIter.init();
  _mainPremiseResidualSizeFormula.init();
  _mainPremiseSelectedLitResidualSizeFormula.init();

 

  // current selected literal in the main premise
  _compiledConstraint = 0; 
  
  

  // side premise data
  _sidePremiseVarIndex = 1UL; 
  _storedConstraintCheck = 0; 



  // retrieval
  _retrievalOfUnifiable.init();
  _candidates.init();


  _mainPremiseResidualSizeFormula.SetIndex(0UL);
  _mainPremiseSelectedLitResidualSizeFormula.SetIndex(0UL);
    
  _currentMainPosEqArg1Mem[0].Head().MkFunc(TERM((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity));
  _currentMainPosEqArg2Mem[0].Head().MkFunc(TERM((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity));
  _storedConstraintCheck = 0; 
    
  _retrievalOfUnifiable.SetIndex(_sidePremiseVarIndex);
}; // void VK::BACKWARD_SUPERPOSITION::init(NewClause* res,OR_INDEX* unifIndex)


void VK::BACKWARD_SUPERPOSITION::destroy()
{
  CALL("destroy()");
  
  

  if (DestructionMode::isThorough())
    {
      if (_storedConstraintCheck)
	{
	  SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
	};
      if (_compiledConstraint)
	{
	  SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
	};
    };


  _candidates.destroy();


  // retrieval
  _retrievalOfUnifiable.destroy();

  
  // main premise data
  _mainPremiseSelectedLitResidualSizeFormula.destroy();
  _mainPremiseResidualSizeFormula.destroy();
  _mainSelectedLitIter.destroy();
}; // void VK::BACKWARD_SUPERPOSITION::destroy()



bool VK::BACKWARD_SUPERPOSITION::SetQuery(Clause* cl) 
{ 
  _noRulesAsSidePremises = _noInfBetweenRulesFlag && cl->IsRule();

  loadMainPremise(cl);
  _currentMainPosEqLit = findNextMainSelectedPosEq();
  if (!_currentMainPosEqLit) return false;
  LoadEq(_currentMainPosEqLit,_mainPremiseVarIndex);
  return true; 
}; // bool VK::BACKWARD_SUPERPOSITION::SetQuery(Clause* cl) 

bool VK::BACKWARD_SUPERPOSITION::CollectResult()
{
  CALL("CollectResult()");
  if ((!_mainPremise->Active()) 
      || (!_sidePremise->Active()))
    return false;
      
  if (_mainPremiseResidualSize
      + _sidePremise->MinResWeight()
      > _resolvent->MaxWeight())
    {
      _resolvent->AnotherTooHeavyClause();
      return false;
    };            
     
  _resolvent->openGenerated(ClauseBackground::BackSup);
  if (collectSidePremiseLiterals() && collectMainPremiseLiterals())
    {
      _resolvent->UsedRule(ClauseBackground::BackSup);
      _resolvent->PushAncestor(_mainPremise);
      _resolvent->PushAncestor(_sidePremise);
      return true;
    }
  else return false;
}; // bool VK::BACKWARD_SUPERPOSITION::CollectResult()

bool VK::BACKWARD_SUPERPOSITION::NextInference()
{
 process_literal:
  while (Paramodulate())
    { 
      if (CollectResult()) return true;
    };
  // try next literal
  _currentMainPosEqLit = findNextMainSelectedPosEq();       
  if (!_currentMainPosEqLit) return false;
  LoadEq(_currentMainPosEqLit,_mainPremiseVarIndex);
  goto process_literal;
}; // bool VK::BACKWARD_SUPERPOSITION::NextInference()


inline void VK::BACKWARD_SUPERPOSITION::compileMainPremiseResidualSizeFormula()
{
  CALL("compileMainPremiseResidualSizeFormula()");
  _mainPremiseResidualSizeFormula.Reset();

  for (LiteralList::Iterator iter(_allMainPremiseLiterals,
				  _firstNonselectedMainLit);
       iter.notEnd();
       iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      _mainPremiseResidualSizeFormula.Compile(iter.currentLiteral());
}; // void VK::BACKWARD_SUPERPOSITION::compileMainPremiseResidualSizeFormula()


inline void VK::BACKWARD_SUPERPOSITION::recompileMainPremiseSelectedLitResidualSizeFormula()
{
  CALL("recompileMainPremiseSelectedLitResidualSizeFormula()");
  _mainPremiseSelectedLitResidualSizeFormula.Reset();

  for (LiteralList::Iterator iter(_allMainPremiseLiterals); 
       iter.currentElement() != _firstNonselectedMainLit;
       iter.next())
    {
      ASSERT(!iter.currentLiteral()->isAnswerLiteral());
      if (iter.currentElement() == _currentSelectedLitListElement)
	{              
	  ASSERT(iter.currentElement()->isEquality());
	  ASSERT(iter.currentElement()->positive());
	  if (iter.currentElement()->isOrderedEq())
	    _mainPremiseSelectedLitResidualSizeFormula.Compile(*(iter.currentElement()->arg2()));
	}
      else 
	_mainPremiseSelectedLitResidualSizeFormula.Compile(iter.currentElement()->literal());
    }; 
}; // void VK::BACKWARD_SUPERPOSITION::recompileMainPremiseSelectedLitResidualSizeFormula()


inline void VK::BACKWARD_SUPERPOSITION::loadMainPremise(Clause* cl) 
{
  CALL("loadMainPremise(Clause* cl)");
  _mainPremise = cl;
  _allMainPremiseLiterals = cl->LitList();
  _firstNonselectedMainLit = _allMainPremiseLiterals->nth(_allMainPremiseLiterals->numOfSelectedLits());
  _mainSelectedLitIter.reset(_allMainPremiseLiterals);
  compileMainPremiseResidualSizeFormula();
}; // void VK::BACKWARD_SUPERPOSITION::loadMainPremise(Clause* cl) 


inline const VK::TERM* VK::BACKWARD_SUPERPOSITION::findNextMainSelectedPosEq()
{
  CALL("findNextMainSelectedPosEq()");
  const TERM* lit;
 check_lit:
  if (_mainSelectedLitIter.currentElement() == _firstNonselectedMainLit) return 0;  
  _currentSelectedLitListElement = _mainSelectedLitIter.currentElement();
  lit = _mainSelectedLitIter.currentLiteral();
  _mainSelectedLitIter.next();
  if (lit->Positive() && lit->IsEquality()) 
    {
      recompileMainPremiseSelectedLitResidualSizeFormula();
      return lit;
    }; 
  goto check_lit;
}; //  const TERM* VK::BACKWARD_SUPERPOSITION::findNextMainSelectedPosEq()


inline bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseLiterals() 
{
  CALL("collectMainPremiseLiterals()");
  // collect max. literals except the used equality

  //cout << "REWR RULE " << _mainSelectedLitIter.currentLiteral() << "\n";

  for (LiteralList::Iterator iter(_allMainPremiseLiterals); iter.currentElement() != _firstNonselectedMainLit;iter.next())
    if (iter.currentElement()->next() != _mainSelectedLitIter.currentElement())
      {      

	//cout << "PUSH MAX " << iter.currentLiteral() << "\n";

	if (!_resolvent->PushStandardLit(_mainPremiseVarIndex,iter.currentLiteral(),iter.currentElement()->isInherentlySelected())) return false; 

	//_resolvent->output(cout << "Success, RES3 = ") << "\n";

      };      

  // collect nonmax. literals
  for (LiteralList::Iterator iter1(_allMainPremiseLiterals,_firstNonselectedMainLit);iter1.notEnd();iter1.next())
    {
      ASSERT(!(iter1.currentElement()->isInherentlySelected()));
      if (!_resolvent->PushStandardLit(_mainPremiseVarIndex,iter1.currentLiteral(),false)) return false;   
    };  
  return true;
}; // bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseLiterals() 




inline void VK::BACKWARD_SUPERPOSITION::LoadEq(const TERM* eq_lit,ulong eq_ind)
{
  CALL("LoadEq(const TERM* eq_lit,ulong eq_ind)");
  ASSERT(eq_lit->IsEquality());
  ASSERT(eq_lit->Positive());
  _mainPremiseVarIndex = eq_ind; 
  const TERM* arg_ref = eq_lit->Args();
  PrefixSym* mem = _currentMainPosEqArg1Mem + 1;
  StandardTermToPrefix(_mainPremiseVarIndex,*arg_ref,mem);
  _currentMainPosEqArg1Mem[0].SetBracket(mem);
     
  // another argument
  arg_ref = arg_ref->Next();
  mem = _currentMainPosEqArg2Mem + 1;
  StandardTermToPrefix(_mainPremiseVarIndex,*arg_ref,mem);
  _currentMainPosEqArg2Mem[0].SetBracket(mem);
     
  LoadLHS(_currentMainPosEqArg1Mem);
  _currentMainPosEqRHS = _currentMainPosEqArg2Mem + 1;
  _currentMainPosEqIsUnordered = eq_lit->IsUnorderedEq();
  _swapArguments = _currentMainPosEqIsUnordered;

  if (_currentMainPosEqIsUnordered)
    {
      if (!_compiledConstraint)
	_compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
      ASSERT(_compiledConstraint);
      _compiledConstraint->load(_currentMainPosEqArg2Mem,_currentMainPosEqArg1Mem);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _currentCompiledConstraint = _compiledConstraint;
	}
      else
	{
	  _currentCompiledConstraint = 0;
	};
    }
  else
    {
      _currentCompiledConstraint = 0;
    };
     
}; //  void VK::BACKWARD_SUPERPOSITION::LoadEq(const TERM* eq_lit,ulong eq_ind)



inline bool VK::BACKWARD_SUPERPOSITION::Paramodulate()
{
  CALL("Paramodulate()");
 try_next_cand:
  if (NextCandidate(_currentCompiledConstraint,
		    _sidePremise,
		    _candLitRedexPair)) 
    {
      if (_noRulesAsSidePremises && _sidePremise->IsRule()) 
	goto try_next_cand;

      if (demodulateSelectedLits())
	goto try_next_cand;

      if (subsumeSidePremise()) goto try_next_cand;
      return true;
    };
  if (_swapArguments)
    {
      _swapArguments = false;
      LoadLHS(_currentMainPosEqArg2Mem);
      _currentMainPosEqRHS = _currentMainPosEqArg1Mem + 1;


      if (!_compiledConstraint)
	_compiledConstraint = 
	  SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
      ASSERT(_compiledConstraint);
      _compiledConstraint->load(_currentMainPosEqArg1Mem,_currentMainPosEqArg2Mem);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _currentCompiledConstraint = _compiledConstraint;
	}
      else
	{
	  _currentCompiledConstraint = 0;
	};

      goto try_next_cand;          
    };
  return false; 
}; // bool VK::BACKWARD_SUPERPOSITION::Paramodulate()

inline bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseLiterals() 
{
  CALL("collectSidePremiseLiterals()");

  const SimplificationOrdering::StoredConstraint* constr = 
    _candLitRedexPair->constraint();
  if (constr) 
    {
      if (!_storedConstraintCheck)
	_storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
      ASSERT(_storedConstraintCheck);
      if (_storedConstraintCheck->holds(constr,_sidePremiseVarIndex))
	return false;
    };
    
  ulong rewritten_lit_num = _candLitRedexPair->litNum();

  LiteralList::Element* rewritten_lit = _sidePremise->LitList()->nth(rewritten_lit_num);
  for (LiteralList::Iterator iter(_sidePremise->LitList());iter.notEnd();iter.next())
    {
      if (iter.currentElement() == rewritten_lit)
	{        
	  if (!_resolvent->PushRewrittenLit(_sidePremiseVarIndex,
					    iter.currentLiteral(),
					    _candLitRedexPair->redex(),
					    _currentMainPosEqRHS,
					    iter.currentElement()->isInherentlySelected()))
	    return false;
	}
      else
	{
	  if ((_simultaneousness > 0) && (iter.currentElement()->arity()))
	    {
	      if (!_resolvent->PushRewrittenLit(_sidePremiseVarIndex,
						iter.currentLiteral(),
						_candLitRedexPair->redex(),
						_currentMainPosEqRHS,
						iter.currentElement()->isInherentlySelected()))
		return false;   
	    }
	  else
	    if (!_resolvent->PushStandardLit(_sidePremiseVarIndex,iter.currentElement()->literal(),iter.currentElement()->isInherentlySelected())) return false;  
	};  
    };     

  return true;
};  // bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseLiterals()

bool VK::BACKWARD_SUPERPOSITION::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");

  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (_sidePremise->hasSubsumedPremiseInstance()) return false;
  _sidePremise->blockForSubsumption();
  ulong maxWeight = _resolvent->MaxWeight();
  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _resolvent->openGenerated(ClauseBackground::Unspecified);
  if (collectSidePremiseInstance())   
    {
      _resolvent->endOfClause();
      subsumingClause = _resolvent->FSubsume();
      ASSERT(subsumingClause != _sidePremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_resolvent->WasTautology()) goto discarded;
      goto return_false;
    };
  
 return_false:
  _sidePremise->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++_statSubsumedSidePremiseInstances;
 discarded:
  _sidePremise->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return true;

}; // bool VK::BACKWARD_SUPERPOSITION::subsumeSidePremise()

bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");

  for(LiteralList::Iterator iter(_sidePremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!(_resolvent->PushStandardLit(_sidePremiseVarIndex,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) 
      return false; 
  return true;

}; // bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseInstance()



void VK::BACKWARD_SUPERPOSITION::LoadLHS(PrefixSym* t)
{
  CALL("LoadLHS(PrefixSym* t)");
  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_QUERY(t->Arg1());
  _retrievalOfUnifiable.CompileComplexTerm(t);     
  _newRetrievalAttempt = true;
  _isFromVariable = t->Arg1()->Head().isVariable();
}; // void VK::BACKWARD_SUPERPOSITION::LoadLHS(PrefixSym* t)
            
            
inline 
bool 
VK::BACKWARD_SUPERPOSITION::NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr,
			  Clause*& candCl,
			  BackParaCandidates::LitRedexPair*& candLitRedexPair)
{
  CALL("NextCandidate(..)"); 

  if (_newRetrievalAttempt)
    {
      _newRetrievalAttempt = false;
      if ((!_paramodulateFromVariables) && _isFromVariable)
	{
	  ++_statBlockedParamodulationsFromVariables;
	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_TOTAL;
	  return false;
	};

      if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->BParaTree())))
	{	       
	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_RESULT(_retrievalOfUnifiable.CurrentLeaf()->Term().First());

	  computeMainPremiseResidualWeight();
	  if (_mainPremiseResidualSize > _resolvent->MaxWeight())
	    {
	      _resolvent->AnotherTooHeavyClause();
	      goto try_next_leaf;
	    };

	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) goto try_next_leaf;
	  if (subsumeMainPremise()) goto try_next_leaf;
	  _candidates.reset(*(_retrievalOfUnifiable.BPCandidatesFromLeaf()));            
	  ALWAYS(_candidates.next(candCl,candLitRedexPair));
	  return true;
	}
      else 
	{
	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_TOTAL;
	  return false;
	};
    }
  else // !_newRetrievalAttempt
    {
      if (_candidates.next(candCl,candLitRedexPair)) 
	{
	  return true;
	};
    try_next_leaf: 

      if (_retrievalOfUnifiable.FindNextLeaf())
	{	        

	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_RESULT(_retrievalOfUnifiable.CurrentLeaf()->Term().First());

	  computeMainPremiseResidualWeight();
	  if (_mainPremiseResidualSize > _resolvent->MaxWeight())
	    {
	      _resolvent->AnotherTooHeavyClause();
	      goto try_next_leaf;
	    };

	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) goto try_next_leaf;
	  if (subsumeMainPremise()) goto try_next_leaf;
	  _candidates.reset(*(_retrievalOfUnifiable.BPCandidatesFromLeaf()));            
	  ALWAYS(_candidates.next(candCl,candLitRedexPair));
	  return true;
	}
      else 
	{
	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_TOTAL;
	  return false;  
	};
    };     
}; // bool VK::BACKWARD_SUPERPOSITION::NextCandidate(..) 

bool VK::BACKWARD_SUPERPOSITION::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");

  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_mainPremise->hasSubsumedPremiseInstance()) return false;
  _mainPremise->blockForSubsumption();
  ulong maxWeight = _resolvent->MaxWeight();
  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _resolvent->openGenerated(ClauseBackground::Unspecified);
  if (collectMainPremiseInstance())   
    {
      _resolvent->endOfClause();
      subsumingClause = _resolvent->FSubsume();
      ASSERT(subsumingClause != _mainPremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_resolvent->WasTautology()) goto discarded;
      goto return_false;
    };
  
 return_false:
  _mainPremise->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++_statSubsumedMainPremiseInstances;
 discarded:
  _mainPremise->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return true;

}; // bool VK::BACKWARD_SUPERPOSITION::subsumeMainPremise()

bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");

  for(LiteralList::Iterator iter(_mainPremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!(_resolvent->PushStandardLit(_mainPremiseVarIndex,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) 
      return false; 
  return true;

}; // bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseInstance()




bool VK::BACKWARD_SUPERPOSITION::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");
  if (_sidePremise->paramodulationsIntoProtected() &&
      _mainPremise->Unit())
    return false; // paramodulations into this clause are protected

  Clause* simplifyingClause;
  if (_mainSelectedLitDemodulation)
    {   
      if (_mainSelectedLitDemodulationAttempted)
	{
	  if (_mainSelectedLitDemodulated)
	    {
	      ++_statDemodulatedMainSelectedLits;
	      return true;
	    };
	  // proceed with the side premise
	}
      else
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  _mainPremise->blockForDemodulationFrom();
	  _sidePremise->blockForDemodulationFrom();
	  ulong maxWeight = _resolvent->MaxWeight();
	  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
	  _resolvent->openGenerated(ClauseBackground::Unspecified);
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto())
	    {	   
	      _resolvent->endOfClause();   	
      
	      simplifyingClause = _resolvent->canBeDemodulated();
	      ASSERT(simplifyingClause != _mainPremise);
	      ASSERT(simplifyingClause != _sidePremise);
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      _resolvent->SetMaxWeight(maxWeight);
	      
	      if (simplifyingClause)
		{	
		  _mainPremise->protectParamodulationsInto();
		  _sidePremise->protectParamodulationsInto();
		  ++_statDemodulatedMainSelectedLits;
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	      // proceed with the side premise
	    }
	  else // analyse the result
	    {
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      _resolvent->SetMaxWeight(maxWeight);
	      return _resolvent->WasTautology();
	    };
	};
    };

  if (_sideSelectedLitDemodulation)
    {
      _mainPremise->blockForDemodulationFrom();
      _sidePremise->blockForDemodulationFrom();
      ulong maxWeight = _resolvent->MaxWeight();
      _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
      _resolvent->openGenerated(ClauseBackground::Unspecified);
      if (collectSidePremiseSubtermsAvailableForParamodulationInto())
	{	   
	  _resolvent->endOfClause();   	
      
	  simplifyingClause = _resolvent->canBeDemodulated();
	  ASSERT(simplifyingClause != _mainPremise);
	  ASSERT(simplifyingClause != _sidePremise);
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  _resolvent->SetMaxWeight(maxWeight);
	      
	  if (simplifyingClause)
	    {	
	      _mainPremise->protectParamodulationsInto();
	      _sidePremise->protectParamodulationsInto();
	      ++_statDemodulatedSideSelectedLits;
	      return true;
	    }
	  else
	    return false;
	}
      else // analyse the result
	{
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  _resolvent->SetMaxWeight(maxWeight);
	  return _resolvent->WasTautology();
	};
    }
  else
    return false;
}; // bool VK::BACKWARD_SUPERPOSITION::demodulateSelectedLits()




bool 
VK::BACKWARD_SUPERPOSITION::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()");  
  bool atLeastOneLiteralCollected = false;
  for(LiteralList::Iterator iter(_allMainPremiseLiterals);
      iter.currentElement() != _firstNonselectedMainLit;
      iter.next())
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_resolvent->pushPseudoLit(_mainPremiseVarIndex,
					     TERM::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_resolvent->PushStandardLit(_mainPremiseVarIndex,lit,false))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_resolvent->PushStandardLit(_mainPremiseVarIndex,lit,false))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
    };
  return atLeastOneLiteralCollected;
}; // bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseSubtermsAvailableForParamodulationInto()



bool 
VK::BACKWARD_SUPERPOSITION::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  ulong numOfSelectedLits = _sidePremise->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_sidePremise->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_resolvent->pushPseudoLit(_sidePremiseVarIndex,
					     TERM::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_resolvent->PushStandardLit(_sidePremiseVarIndex,lit,false))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_resolvent->PushStandardLit(_sidePremiseVarIndex,lit,false))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; // bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseSubtermsAvailableForParamodulationInto()





//==================================================================
