 /*===================================================\
 |               FILE: control.c:                     |
 |                                                    |
 | SCOTT functions that control the size and          |
 | composition of the clauses on the Set of Support.  |
 \===================================================*/


/*
  Requires the following struct...

  struct Distn {
       int min,max;
       int freq[MAX_DIST];
       int freq_over_comp[MAXISET_ARRAY][MAX_DIST];
     
       int max_outlier;
       int num_outliers_up;
       int outliers_up_over_comp[MAXISET_ARRAY];
     
       int min_outlier;
       int num_outliers_down;
       int outliers_down_over_comp[MAXISET_ARRAY];
     
       float average, median, variance;
       int interval;
       char *name;
       int (*property)();
       int *limit;
       int *cull_over;
       int *limit_at;
  };

  and the following parameters...

  tParms[MEMORY_CONTROL_POINT].name = "memory_control_point";
  tParms[MEMORY_CONTROL_POINT].val = 0; // off by default
  tParms[MEMORY_CONTROL_POINT].min = 0;
  tParms[MEMORY_CONTROL_POINT].max = MAX_INT;

  tParms[LIMIT_WEIGHT_TO].name = "limit_weight_to";
  tParms[LIMIT_WEIGHT_TO].val = 10; // percentile
  tParms[LIMIT_WEIGHT_TO].min = 0;
  tParms[LIMIT_WEIGHT_TO].max = 100;
  
  tParms[CULL_WEIGHTS_OVER].name = "cull_weights_over";
  tParms[CULL_WEIGHTS_OVER].val = 90;  // percentile
  tParms[CULL_WEIGHTS_OVER].min = 0;
  tParms[CULL_WEIGHTS_OVER].max = 100;

  tParms[LIMIT_VARS_TO].name = "limit_vars_to";
  tParms[LIMIT_VARS_TO].val = 50; // percentile
  tParms[LIMIT_VARS_TO].min = 0;
  tParms[LIMIT_VARS_TO].max = 100;
  
  tParms[CULL_VARS_OVER].name = "cull_vars_over";
  tParms[CULL_VARS_OVER].val = 90;  // percentile
  tParms[CULL_VARS_OVER].min = 0;
  tParms[CULL_VARS_OVER].max = 100;

 */

#include "scott.h"

static struct Distn pick_weight;
static struct Distn max_vars;

void scott_control_memory()
 /*--------------------------------------------------------------.
 |              Called by infer_and_process()                    |
 |                                                               |
 | Checks memory allocation once every 20 given clause, and      |
 | resets weight and variable limits if appropriate -- memory is |
 | greter than 1/3rd the of the memory_limit parameter and a     |
 | sensible new limit can be set.  This implements a *soft*      |
 | limit on memory since exceeding the control_memory_limit does |
 | not cause the program to exit.  Also back culls a proportion  |
 | of the existing clauses so memory consumption will often be   |
 | less after a control point.                                   |
 `--------------------------------------------------------------*/
{
  static int control_point = 0;

  /* should we control? */
  if (tParms[MEMORY_CONTROL_POINT].val == 0 ) return;
  if (total_mem() < tParms[MEMORY_CONTROL_POINT].val) return;

  if (control_point && control_point != Stats[CL_GIVEN]) return;
  control_point = Stats[CL_GIVEN] + 20;
  control_distn(&pick_weight);
  control_distn(&max_vars);
}/*--scott_control_memory()--*/


void control_distn(struct Distn *distn)
 /*-------------------------------------------------------------.
 |              Called by scott_control_memory().               |
 |                                                              |
 | Examinies distribution and (possibly) sets a new value for   |
 | the associated limit. depending on the distribution we are   |
 | controlling the new limit is set at the percentile given by  |
 | either CONTROL_WT_AT or CONTROL_VARS_AT and existing SOS     |
 | clauses are be discarded if they exceed the percentile given |
 | by CULL_WT_OVER or CULL_VARS_OVER.                           |
 `-------------------------------------------------------------*/
{
  int new_limit = MAX_INT;
  int back_cull = MAX_INT;

  set_distn(distn);

  // is the distn sufficietly fine grained to control?
  if ((distn->max - distn->min)*2 < MAX_DIST) return;

  new_limit = percentile(distn,*(distn->limit_at));

  if ((new_limit < *(distn->limit))
      || (*(distn->limit) == 0) ) {
    
    // set new value
    *(distn->limit) = new_limit;

    message(3,"");
    message(3,"Memory usage at control point %d Kbytes",total_mem());
    message(3,"");
    if (tParms[VERBOSITY].val > 2) 
      output_distn(distn);

    back_cull = percentile(distn,*(distn->cull_over));
    
    message(2,"Resetting %s limit to %d and culling over %d"
	    "\n    (search is now incomplete)",
	    distn->name, new_limit, back_cull);
    message(2,"");
    
    cull_clauses_over(back_cull,distn->property);
  }
  
}/*--control_distn()--*/


void cull_clauses_over(int max_value, int (*property)())
 /*--------------------------------------------------------------.
 |       Called by control_distn(), make_finder_clause(), and    |
 | possibly execute_strategy().                                  |
 |                                                               |
 | Scan SOS discarding clause that are heavier then the new      |
 | max_weight. Must be called after the scan is complete so that |
 | partition and scan data structures are correct.               |
 `--------------------------------------------------------------*/
{  
  struct clause *cl, *next_cl;
  
  cl = Sos->first_cl;
  while (cl) {
    next_cl = cl->next_cl;
    if ((*property)(cl) > max_value){	
      CLOCK_START(UN_INDEX_TIME);
      un_index_lits_all(cl);
      CLOCK_STOP(UN_INDEX_TIME);
      rem_from_list(cl);
      hide_clause(cl);
    }
    cl = next_cl;
  }
  
}/*--cull_clauses_over()--*/


int weight_percentile(int percent)
 /*--------------------------------------------------------------.
 |                    Called by start_scan().                    |
 |                                                               |
 | Returns the maximum value of the given percentile in the      |
 | "pick_weight" distn.  If there are less than 100 clause to on |
 | the SOS returns MAX_INT because a sensible decison cannot be  |
 | made with so few clauses.                                     |
 `--------------------------------------------------------------*/
{
  static int last_num_givens = 0;

  // too few clauses to make a sensible judgement
  if (Stats[SOS_SIZE] <= 100) return MAX_INT;

  // don't reset more than once between given clauses
  if (Stats[CL_GIVEN] > last_num_givens)
    set_distn(&pick_weight);
  last_num_givens = Stats[CL_GIVEN];

  if ((pick_weight.max - pick_weight.min)*2 < MAX_DIST)
    return MAX_INT;

  return percentile(&pick_weight, percent);
  
}/*--weight_percentile()--*/

int variable_percentile(int percent)
 /*-------------------------------------------------------------.
 |                   Called by start_scan().                    |
 |                                                              |
 | Returns the maximum value of the given percentile in the     |
 | "distinct_vars" distn.  If there are less than 100 clause to |
 | on the SOS returns MAX_INT because a sensible decison cannot |
 | be made with so few clauses.                                 |
 `-------------------------------------------------------------*/
{
  static int last_num_givens = 0;
  
  // too few clauses to make a sensible judgement
  if (Stats[SOS_SIZE] <= 100) return MAX_INT;

  // don't reset more than once between given clauses
  if (Stats[CL_GIVEN] > last_num_givens)
    set_distn(&max_vars);
  last_num_givens = Stats[CL_GIVEN];
  
  if ((max_vars.max - max_vars.min)*2 < MAX_DIST)
    return MAX_INT;
  
  return percentile(&max_vars, percent);
  
}/*--variable_percentile()--*/


int clause_pick_weight(struct clause *cl)
 /*--------------------------------------------------------------.
 | Called by pick_weight->property(). Returns the pick weight of |
 | the clause -- so that we can pass this property to a          |
 | function.                                                     |
 `--------------------------------------------------------------*/
{
  return cl->pick_weight;

}/*clause_pick_weight()--*/

void init_distributions()
 /*----------------------------------------------------------.
 | Called by init().  Sets up the distributions that we will |
 | monitor.                                                  |
 `----------------------------------------------------------*/
{
  init_a_distribution(&pick_weight);
  init_a_distribution(&max_vars);

  pick_weight.name = "weight";
  max_vars.name =    "variable";

  pick_weight.property = clause_pick_weight;
  max_vars.property = distinct_vars;

  pick_weight.limit = &(Parms[MAX_WEIGHT].val);
  max_vars.limit = &(Parms[MAX_DISTINCT_VARS].val);

  pick_weight.cull_over = &(tParms[CULL_WEIGHTS_OVER].val);
  max_vars.cull_over = &(tParms[CULL_VARS_OVER].val);

  pick_weight.limit_at = &(tParms[LIMIT_WEIGHT_TO].val);
  max_vars.limit_at = &(tParms[LIMIT_VARS_TO].val);
  
  pick_weight.interval = max_vars.interval = -1;

}/*--init_distributions()--*/

void init_a_distribution(struct Distn *distn)
 /*--------------------------------------------------------------.
 | Called by init_distributions() Initialise fields in the given |
 | Distn structure.                                              |
 `--------------------------------------------------------------*/
{
 int class, maxiset;

  // re-init distributions
  for (class = 0; class < MAX_DIST; class++) {
    distn->freq[class] = 0;
#ifdef SCOTT
    if (get_semantic_guidance_flag()) {
      for (maxiset = 0; maxiset <= last_set_index; maxiset++)
	distn->freq_over_comp[maxiset][class] = 0;
    }
#endif
  }

}/*--init_a_distribution()--*/

void get_distn(struct Distn *distn)
 /*--------------------------------------------------------------.
 |                       Called by set_distn().                  |
 |                                                               |
 | Determines frequencies of clauses on the SOS with             |
 | distn->property within the bounds sets by ditsn->max and      |
 | distn->min which we assume have been set appropriately (ie so |
 | as to skip coinsideration of any outliers).  Also assumes:    |
 | MAX_DIST > 0 distn->max != distn->min|                        |
 `--------------------------------------------------------------*/
{
  struct clause *cl;
  int class, maxiset, diff, sum;
  float median_mark, average;
  int interval, max, min;

  init_a_distribution(distn); // reset frequency values

  max = distn->max;
  min = distn->min;
  average = distn->average;

  // select an appropriate interval length
  diff = max - min;
  if (diff % MAX_DIST) interval = (diff/MAX_DIST) + 1;
  else interval = diff/MAX_DIST;
  
  distn->interval = interval;

  // scan for frequencies in corresponding ranges
  sum = 0;
  for (cl = Sos->first_cl; cl; cl = cl->next_cl) {
    diff = distn->property(cl) - average;
    sum += diff * diff;
    class = ((distn->property(cl) - min) / interval);

    // skip outliers
    if (class < 0 || class >= MAX_DIST) continue;

    distn->freq[class]++;
#ifdef SCOTT
    if (get_semantic_guidance_flag()) {
      for (maxiset = 1; maxiset <= last_set_index; maxiset++){
	if (active[maxiset] && !cl->in_set[maxiset]) 
	  distn->freq_over_comp[maxiset][class]++;
      }
      // the zero set contains clauses that belong to all maxisets
      if (num_maxisets == cl->num_sets)
	distn->freq_over_comp[0][class]++;
    }
#endif
  }
  distn->variance = sum/(float)Stats[SOS_SIZE];
  
  // estimate the median
  median_mark = (Stats[SOS_SIZE]+1)/2;
  class = 0;
  while (class < MAX_DIST) {
    if (median_mark < distn->freq[class]) break;
    median_mark -= distn->freq[class];
    class++;
  }
  if (!distn->freq[class]) {
    distn->median = (class*interval)+min;
  }
  else {
    distn->median = (class*interval)+min+
      (interval*(median_mark/distn->freq[class]));
  }
}/*--get_distn()--*/


void set_distn(struct Distn *distn)
 /*--------------------------------------------------------------.
 |                  Called by percentile().                      |
 |                                                               |
 | Determine frequencies on the SOS so that outliers are treated |
 | sensibly.                                                     |
 `--------------------------------------------------------------*/
{
  struct clause *cl;
  int class, max_freq, maxiset;
  int sos_wt = 0;

  if (Stats[SOS_SIZE] == 0) return;

  distn->min = MAX_INT;
  distn->max = 0;
  distn->interval = 1;

#ifdef SCOTT
  if (get_semantic_guidance_flag()) {
    for (maxiset = 0; maxiset <= last_set_index; maxiset++)
      distn->outliers_up_over_comp[maxiset] =
	distn->outliers_down_over_comp[maxiset] = 0;
  }
#endif

  if (Stats[SOS_SIZE] == 0){
    message(0,"WARNING: set_distn() called with an empty SOS");
    return;
  }
  // scan for min and max value of the distribution
  for (cl = Sos->first_cl; cl; cl = cl->next_cl) {
    if (distn->property(cl) > distn->max) distn->max =  distn->property(cl);
    if (distn->property(cl) < distn->min) distn->min =  distn->property(cl);
    sos_wt += cl->pick_weight;
  }

  if (distn->max == distn->min) {
    distn->average = distn->median = distn->max;
    distn->freq[0] = Stats[SOS_SIZE];
    return;
  }
  distn->average = (sos_wt/(float)Stats[SOS_SIZE]);  
  
  // distn->max != distn->min 
  get_distn(distn);

  // check for outliers, that is, any frequencies in the distribution that
  // is less than 100th of the maximum frequency
  
  for (class = 0, max_freq = 0; class < MAX_DIST; class++)
    if (distn->freq[class] > max_freq) max_freq = distn->freq[class];

  // first upwards outliers
  distn->max_outlier = distn->max;
  distn->num_outliers_up = 0;
  for (class = MAX_DIST-1; class >= 0; class--){
    if (distn->freq[class] > max_freq/(float)(100)) {
      distn->max = ((class+1)*(distn->interval))+(distn->min)-1;
      break;
    }
    distn->num_outliers_up += distn->freq[class];
#ifdef SCOTT
    if (get_semantic_guidance_flag()) {
      for (maxiset = 0; maxiset <= last_set_index; maxiset++){
	distn->outliers_up_over_comp[maxiset]
	  += distn->freq_over_comp[maxiset][class];
      }
    }
#endif
  }

  // next downwards outliers
  distn->min_outlier = distn->min;
  distn->num_outliers_down = 0;
  for (class = 0; class < MAX_DIST; class++){
    if (distn->freq[class] > max_freq/(float)(100)) {
      distn->min = (class*(distn->interval))+(distn->min);
      break;
    }
    distn->num_outliers_down += distn->freq[class];
#ifdef SCOTT
    if (get_semantic_guidance_flag()) {
      for (maxiset = 0; maxiset <= last_set_index; maxiset++){
	distn->outliers_down_over_comp[maxiset] 
	  += distn->freq_over_comp[maxiset][class];
      }
    }
#endif
  }
  
  // if we have any outliers recompute freqencies
  if (distn->num_outliers_up || distn->num_outliers_down) {
    if (distn->max == distn->min) {
      distn->average = distn->median = distn->max;
      distn->freq[0] = Stats[SOS_SIZE];
      return;
    }
    get_distn(distn);
  }
}/*--set_distn()--*/


void output_distn(struct Distn *distn)
 /*-----------------------------------------------------------.
 | Called by OTTER fcn main().                                |
 |                                                            |
 | Outputs formated freqency counts (and stats) for "distn".  |
 | Assumes distn has been set.                                |
 `-----------------------------------------------------------*/
{
  int class, maxiset;

  if (Stats[SOS_SIZE] == 0) return;

  printf("\n-------------- Begin %s distibution ---------------\n", distn->name);
  printf("\nmin %d, max %d, average %.1f, median %.1f, variance %.1f"
	 "\n\n", 
	 (distn->min_outlier? distn->min_outlier : distn->min), 
	 (distn->max_outlier? distn->max_outlier : distn->max),
	 distn->average, 
	 distn->median, 
	 distn->variance);
  
  if (distn->interval > 1) 
    printf("  interval    :  freq");
  else  
    printf("     value    :  freq");
#ifdef SCOTT
  if (get_semantic_guidance_flag()) {
    for (maxiset = 0; maxiset <=last_set_index; maxiset++)
      if (active[maxiset]) printf(" %5d", maxiset);
  }
#endif
  printf("\n---------------------");
#ifdef SCOTT
  if (get_semantic_guidance_flag()) {
    for (maxiset = 0; maxiset <=last_set_index; maxiset++)
      if (active[maxiset]) printf("------");
  }
#endif
  printf("\n");
  if (distn->num_outliers_down) {
    if ((distn->min - distn->min_outlier) > 1) {
      printf("%5d - %-5d : %5d",
	     distn->min_outlier, distn->min - 1, 
	     distn->num_outliers_down);
    }
    else {
      printf("%5s   %-5d : %5d", "",
	     distn->min_outlier,
	     distn->num_outliers_down);
    }
#ifdef SCOTT
    if (get_semantic_guidance_flag()) {
      for (maxiset = 0; maxiset <= last_set_index; maxiset++)
	if (active[maxiset]) 
	  printf(" %5d",distn->outliers_down_over_comp[maxiset]);
    }
#endif
    printf("\n");
  }
  // print the frequencies
  for (class = 0; class < MAX_DIST; class++) {
    if ((class*distn->interval)+distn->min > distn->max) break;

    if (distn->interval > 1) {
      printf("%5d - %-5d : %5d",
	     (class*distn->interval)+distn->min,
	     ((class+1)*distn->interval)+distn->min-1,
	     distn->freq[class]);
    }
    else  {
      printf("%5s   %-5d : %5d", "",
	     class+distn->min,
	     distn->freq[class]);
    }
#ifdef SCOTT
    if (get_semantic_guidance_flag()) {
      for (maxiset = 0; maxiset <= last_set_index; maxiset++)
	if (active[maxiset]) 
	  printf(" %5d",distn->freq_over_comp[maxiset][class]);
    }
#endif
    printf("\n");
  }
  if (distn->num_outliers_up) {
    printf("\n");
    if ((distn->max_outlier - distn->max) > 1) {
      printf("%5d - %-5d : %5d",
	     distn->max+1,
	     distn->max_outlier, 
	     distn->num_outliers_up);
    }
    else {
      printf("%5s   %-5d : %5d", "",
	     distn->max_outlier,
	     distn->num_outliers_up);
    }
#ifdef SCOTT
    if (get_semantic_guidance_flag()) {
      for (maxiset = 0; maxiset <= last_set_index; maxiset++)
	if (active[maxiset]) 
	  printf(" %5d",distn->outliers_up_over_comp[maxiset]);
    }
#endif
    printf("\n");
  }
  printf("\n-------------- End %s distribution -----------------\n",distn->name);
  fflush(stdout); 
  
}/*--output_distn()--*/


void output_weight_distn()
 /*--------------------------------------------------------.
 | Called by main() and output_finder_stats().             |
 |                                                         |
 | Prints a leaf plot of the current pick weight frequency |
 | distribution.                                           |
 `--------------------------------------------------------*/
{
  set_distn(&pick_weight);
  message(2,"");
  output_distn(&pick_weight);
  message(2,"");

}/*--output_weight_distn()--*/

void output_variable_distn()
 /*--------------------------------------------------------.
 | Called by output_finder_stats().                        |
 |                                                         |
 | Prints a leaf plot of the current pick weight frequency |
 | distribution.                                           |
 `--------------------------------------------------------*/
{
  set_distn(&max_vars);
  message(2,"");
  output_distn(&max_vars);
  message(2,"");

}/*--output_variable_distn()--*/


int percentile(struct Distn *distn, int percent)
 /*-----------------------------------------------------------.
 |         Called by start_scan() and control_memory().       |
 |                                                            |
 | Returns the maximum value of the given percentile in the   |
 | given "distn". Assumes the frequency distribution has been |
 | set.  Takes account of outliers and maxiset complements.   |
 `-----------------------------------------------------------*/
{
  int class, class_in_set, maxiset;
  float percentile_mark, interpolant;
  int value;
  char str[10];
  
  // estimate the top value in the percentile 
  percentile_mark = (Stats[SOS_SIZE]*percent)/(float)100;
  class = 0;
  percentile_mark -= distn->num_outliers_down;
  if (percentile_mark >= distn->freq[class]) {
    while (class < MAX_DIST) {
      if (percentile_mark < distn->freq[class]) break;
      percentile_mark -= distn->freq[class];
      class++;
    }
  }

  if (distn->freq[class]) 
    interpolant = 
      distn->interval
      *percentile_mark
      /(float)distn->freq[class];
  else 
    interpolant = 0;

#ifdef SCOTT
  if (get_semantic_guidance_flag()) {
    // check percentile across maxiset complements
    for (maxiset = 1; maxiset <= last_set_index; maxiset++) {
      if (!active[maxiset]) continue;
      percentile_mark = (get_comp_on_sos(maxiset)*percent)/(float)100;

      // skip empty theories
      if (percentile_mark == 0) continue;

      class_in_set = 0;
      percentile_mark -= distn->num_outliers_down;
      if (percentile_mark >= distn->freq[class_in_set]) {
	while (class_in_set < MAX_DIST) {
	  if (percentile_mark 
	      < distn->freq_over_comp[maxiset][class_in_set]) break;
	  percentile_mark -= distn->freq_over_comp[maxiset][class_in_set];
	  class_in_set++;
	} 
      }
      if (class_in_set > class) {//chose the higher value
	class = class_in_set;
	if (distn->freq_over_comp[maxiset][class])
	  interpolant = 
	    distn->interval
	    *percentile_mark
	    /(float)distn->freq_over_comp[maxiset][class];
	else
	  interpolant = 0;
      }
    }
  }
#endif
 
  // kludge rounding rather than use a math library
  sprintf(str,"%.0f", (class*distn->interval)+distn->min+interpolant);
  sscanf(str,"%d",&value);
  
  return value;
  
}/*--percentile()--*/







