#include "dynstring.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

// note, global strings don't need to be initialised, they will have
// memory allocated for them when they are first used

#define min(a, b) ((a) < (b) ? (a) : (b))
#define max(a, b) ((a) > (b) ? (a) : (b))

#define DEFAULT_INITIAL_ALLOCATION 64 
void dynstring_initialise(dynstring *d, const char *initial)
{
    d->length = strlen(initial);
    d->allocated = max(d->length + 1, DEFAULT_INITIAL_ALLOCATION);
    d->c_str = (char *) malloc(d->allocated);
    memcpy(d->c_str, initial, d->length + 1);
}

void dynstring_free(dynstring *d)
{
    if (d->c_str != NULL) {
        free(d->c_str);
        d->allocated = 0;
        d->length = 0;
        d->c_str = NULL;
    }
}

static void ensure_initialised(dynstring *d)
{
    if (d->c_str == NULL) {
        d->allocated = DEFAULT_INITIAL_ALLOCATION;
        d->length = 0;
        d->c_str = (char *) malloc(d->allocated);
    }
    
}

void dynstring_ensure_space(dynstring *d, int index)
{
    if (index >= d->allocated) {
        d->allocated = max(d->allocated * 2, index + 1);
        d->c_str = realloc(d->c_str, d->allocated);
        if (d->c_str == NULL) {
            perror("Error allocating memory for string");
            exit(1);
        }
    }
}

void dynstring_strcat(dynstring *d, const dynstring *s)
{
    ensure_initialised(d);
    dynstring_ensure_space(d, d->length + s->length);
    memcpy(d->c_str + d->length, s->c_str, s->length + 1);
    d->length += s->length;
}

void dynstring_strcatc(dynstring *d, const char *s)
{
    int s_length = strlen(s);

    ensure_initialised(d);

    dynstring_ensure_space(d, d->length + s_length);
    memcpy(d->c_str + d->length, s, s_length + 1);
    d->length += s_length;
}

void dynstring_sprintf(dynstring *d, const char *fmt, ...)
{
    va_list ap;
    int rv;

    ensure_initialised(d);

    va_start(ap, fmt);
    rv = vsnprintf(d->c_str, d->allocated, fmt, ap);
    /* handle pre C99 standard */
    while (rv == -1) {
        dynstring_ensure_space(d, d->allocated * 2);
        
        va_end(ap);
        va_start(ap, fmt);
        rv = vsnprintf(d->c_str, d->allocated, fmt, ap);
    }
    /* handle post C99 standard */
    if (rv + 1 > d->allocated) {
        dynstring_ensure_space(d, rv + 1);

        va_end(ap);
        va_start(ap, fmt);
        rv = vsnprintf(d->c_str, d->allocated, fmt, ap);
    }
    d->length = rv;
    va_end(ap);
}

void dynstring_strcpy(dynstring *d, const dynstring *s)
{
    ensure_initialised(d);

    dynstring_ensure_space(d, s->length);
    d->length = s->length;
    strcpy(d->c_str, s->c_str);
}

void dynstring_strcpyc(dynstring *d, const char *s)
{
    int s_length = strlen(s);
    ensure_initialised(d);
    
    dynstring_ensure_space(d, s_length);
    d->length = s_length;
    strcpy(d->c_str, s);
}

void dynstring_clear(dynstring *d)
{
    ensure_initialised(d);
 
    d->length = 0;
    d->c_str[0] = 0;
}

char *dynstring_fgets(dynstring *d, FILE *f)
{
    ensure_initialised(d);
    if (fgets(d->c_str, d->allocated, f) == NULL) {
        dynstring_clear(d);
        return NULL;
    }
    d->length = strlen(d->c_str);
    while (d->c_str[d->length - 1] != '\n') {
        // read hit the end of the buffer, reallocate, then read some more

        int extra;

        dynstring_ensure_space(d, d->allocated * 2);
        fgets(d->c_str + d->length, d->allocated - d->length, f);

        extra = strlen(d->c_str + d->length);
        if (extra == 0)
            break;
        d->length += extra;
    }
    return d->c_str;
}

void dynstring_append_char(dynstring *d, char ch)
{
    ensure_initialised(d);
    dynstring_ensure_space(d, d->length + 1);
    d->c_str[d->length] = ch;
    d->length++;
    d->c_str[d->length] = '\0';
}
