/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
//  file formula.h 
//  defines class Formula
//

#ifndef __formula__
#define __formula__


#include "typedefs.h"
#include "term.h"
#include "clause.h"
#include "clist.h"
#include "memory.h"
#include "tracer.h"


class Atom;
class Substitution;


class Formula 
#   if debug_memory
    : public Memory <CID_FORMULA>
#   endif
{
 public:

  typedef CList <Formula*,CID_FORMULA_LIST> List;
  
  enum Connective {
    ATOM = 0,     // WARNING: the order is important for
    AND = 1,      // normalize() to work
    OR = 2,
    IMP = 3,
    IFF = 4,
    XOR = 5,
    NOT = 6,
    FORALL = 7,
    EXISTS = 8
  }; // enum Formula::Connective

  // constructors
  Formula ( Connective c, Formula* l, Formula* r) : // for binary formulas
    _con (c),
    _left (l),
    _right (r),
    _count (0)
    { ASSERT(debug_formula, c == IFF || c == XOR || c == IMP);
      l->inc ();
      r->inc (); }
  Formula ( Connective c, Formula* a ) : // for unary formulas
    _con ( c ),
    _arg ( a ),
    _count (0)
    { ASSERT ( debug_formula, c == NOT );
      a->inc (); }
  Formula ( Connective c, List* fs ) : // for and/or formulas
    _con ( c ),
    _args ( fs ),
    _count (0)
    { ASSERT ( debug_formula, c == AND || c == OR );
      if (fs->isNonEmpty()) 
        fs->inc (); } 
  Formula (Connective c, VarList* vs, Formula* a) : // for quantified formulas
    _con (c),
    _vars (vs),
    _arg (a),
    _count (0)
    { ASSERT ( debug_formula, c == FORALL || c == EXISTS );
      a->inc (); } 
  explicit Formula ( Atom* a ) : // for atomic formulas
    _con (ATOM),
    _atom (a),
    _count (0)
    { a->inc(); } 

  // similar to a binary constructor
  static Formula* junction (Connective, Formula* l, Formula* r); 
  // the same but only used by the parser
  static Formula* parseJunction (Connective, Formula* l, Formula* r); 

  void inc () { _count++; }
  void dec () 
    { ASSERT (debug_formula, _count > 0);
      TRACER( "Formula::dec" );

      _count--; 
      if (_count == 0) {
        delete this;
      }
    }

  // query the structure
  Connective connective () const { return _con; } 

  Atom* atom () const
    { ASSERT ( debug_formula, _con == ATOM );
      return _atom; } 
  Formula* left () const
    { ASSERT ( debug_formula, _con == IFF || _con == XOR || _con == IMP );
      return _left; } 
  Formula* right () const
    { ASSERT ( debug_formula, _con == IFF || _con == XOR || _con == IMP );
      return _right; } 
  Formula* arg () const
    { ASSERT ( debug_formula, _con == FORALL || _con == EXISTS || _con == NOT );
      return _arg; } 
  List* args () const
    { ASSERT ( debug_formula, _con == AND || _con == OR );
      return _args; }
  VarList* vars () const
    { ASSERT ( debug_formula, _con == FORALL || _con == EXISTS );
      return _vars; } 

  // statistics
  bool isEqualityAxiom () const;  // is one of the equality axioms
  bool occurs ( const Signature::Pred* p ) const; // predicate symbol f occurs in the formula

  // miscellaneous
  void prettyPrint ( int indent, ostream& );
  Atom* isPredicateDefinition ( Formula* & rhs );   // return lhs of the definition or 0
  Term* isFunctionDefinition ( Term* & rhs ) const; // return lhs of the definition or 0
  bool hasFreeVars () const; 
  List* conjunctionsOut (bool polarity);
  Formula* ennf (bool polarity);
  Formula* removeIff (bool polarity);
  Formula* miniscope ();
  Formula* flatten ();  // flatten the formula
  Formula* normalize ();
  bool less (const Formula*) const; // used to compare formulas, needed for normalisation
  Compare compare ( const Formula* ) const; // comparison of normalized formulas
  static Compare compare ( const List*, const List* ); // comparison of normalized formulas
  Formula* rectify (); // renames variables in the order of their occurrence
  Formula* skolemizeNNF () const;

/*
  int size () const; // the number of connective, quantifiers and atoms
  int weight () const; // the number of symbols
  void polarities ( const PSymbol* p, bool sign, int& pol ) const;
  void makeHalfDefinition ( bool positive,const Formula* rhs );
*/

 private: 
  // structure
  Connective _con;

  union {
    Formula* _left;
    Atom* _atom;
    VarList* _vars;
  };
  union {
    Formula* _right;
    Formula* _arg;
    List* _args;
  };
  long _count;

  // destructor
  ~Formula ();

  // structure change
  void connective ( Connective c ) { _con = c; }

  void args ( List* as )
    { ASSERT ( debug_formula, _con == AND || _con == OR );
      as->inc ();
      _args->dec ();
      _args = as; }

  // auxiliary functions
  static char* printName ( Connective );
  void print ( ostream& );
  void print ( Connective outer, ostream& );
  void prettyPrint ( Connective outer, int indent, ostream& );
  bool printOneLine ( int& limit ) const;  // 
  bool printOneLine () const;
  bool hasFreeVars (VarListList*) const; 
  static List* conjunctionsOut (List*, bool polarity);
  static Formula* ennf (Connective andOr, List*, bool polarity);
  static List* removeIff (List*, bool polarity);
  bool occurs (Var v) const; // variable v occurs free in the formula
  static List* flatten ( List*, Connective );  
  Formula* rectify ( Substitution&, Var& last ); 
  static List* rectify ( List*, Substitution&, Var& last ); 
  static VarList* rectifyVars ( VarList*, Substitution&, Var& last ); 
  Formula* skolemizeNNF (Substitution& subst, const VarList*) const; 
  static List* skolemizeNNF (const List*, Substitution& subst, const VarList*);  

  // miniscoping
  class Miniscope;
  static List* miniscope (Connective, List*);
  Formula* topMiniscope (Connective, VarList*);
  static List* topMiniscope (Connective, VarList*, List*);
  VarList* occurring (VarList*);
  static List* pushVarsDown ( List*, Connective quant, int j, VarList*);

/*
  bool isLinearAndFlat (); 
  static List* copy( List* );
  void copyFrom ( const Formula* );
  void removeIff ();  // remove IFF and XOR
  void removeIff ( Connective c );  // remove IFF and XOR, c is the external connective (AND or OR)
  void normalize1 ();
  Formula* copy ();
  static List* sort ( List* );
  static List* sort ( List* toSort, List* appendTo );
  static void split ( Formula*, List*, // for sorting
                      List*& less, List*& equal, List*& greater );
*/
  // equality axiom check
  bool isReflexivityAxiom () const;
  bool isFunctionReflexivityAxiom () const;
  bool isPredicateReflexivityAxiom () const;
  bool isSymmetryAxiom () const;
  bool isTransitivityAxiom () const;
/*
  void miniscope (bool sign);
  void miniscope (bool sign, Connective quantifier, VarList*);
  // void moveDown (bool sign,VarList*); // auxiliary for miniscope
  VarList* occur ( const VarList* vs ) const; // sublist of those variables in vs occuring in the formula
  void nnf (bool sign); 

  typedef Alist <Var,Term*> Subst;
                          Subst& sbst, Clause::List*& result );
*/
 friend ostream& operator << ( ostream&, Formula* );
}; // class Formula


ostream& operator << ( ostream&, Formula* );

#endif // __formula__
