/*--------------------------------------------------------------------------*/
/*Routines for testing clauses against existing models.*/
/*--------------------------------------------------------------------------*/

#include "FINDER.h"

#include <stdlib.h>

static int thisvector[V_LENGTH];  /* Copy of the test vector   */


/*--------------------------------------------------------------------------*/
int test( int info[], TRS *T )
     
     /*Called by model() and find_theory(), purely as a tester -- just
      *ends up calling is_good() -- and by transref for the serious search.

      *Decodes the vector info[] (the current model) into testable data 
      *structures. 
      *Tests any unpreprocessed clauses against this model.
      *Uses the external function Ref to record a refutation if one is found
      *If the model passes all the tests, sends the it to the printup routine.
      * 
      *Returns TRUE if all tests are passed
      *and FALSE if some refutation occurs.
      *
      *If the global flag negative flag is set then we want to keep 
      *structures that come through the whole generate-and-test procedure 
      *without ever passing the tests. 
      *So instead of calling the print function, model simply records that 
      *the clause was found to fail.  
      *This flag is set in the fcn main but nowhere else. */

{
  clause c;
  boolean cc_is_good = TRUE;
  
  vect_into_model(info);
  
  /* search for a bad guy successful? */
  if ( negative && !clauses_fail ) return FALSE;
  
  /*Initial call sets T=0; this calls is_good and sets up the modeling
    structure T such that initial latest_clause will be satisfied*/

  if ( !T ) { /* just perform a straight test */
    if (clause_is_long()) {
      cc_is_good = is_good(&latest_clause);
      latest_clause.badness = (cc_is_good? 0: -1);
    }
    else {
      latest_clause.badness = find_badness(&latest_clause);
      cc_is_good = (latest_clause.badness == 0);
    }

    /* if searching for a bad guy this reverses the sense of the search */
    if ( negative ) clauses_fail = !cc_is_good;
    return (cc_badguy != cc_is_good); /* skip out with the result */
  }
  
  /* Get to here only if called by transref and supplied with a T */

  /* Do the following iff SCOTT is operating */
  if ( latest_clause.prem || latest_clause.conc ) {
    /*Perhaps test each clause in the existing theory first*/
    for ( c = TH[current_theory]; c; c = c->nextclause )
      if (!c->preprocessed && is_after( &latest_clause, c )) {
	if ( !is_good(c) ) {
	  record_refutation( c, T );
	  lift_clause(c);
	  return FALSE;           /* refutation found */
	}
      }

    /*then test the new clause */
    cc_is_good = is_good(&latest_clause);
    if ( !cc_is_good && !cc_badguy ) {
      record_refutation( &latest_clause, T );
      return FALSE;               /* refutation found */
    }

    /*Perhaps test each clause in the existing theory after instead*/
    for ( c = TH[current_theory]; c; c = c->nextclause )
      if (!c->preprocessed && is_after( c, &latest_clause )) {
	if ( !is_good(c) ) {
	  record_refutation( c, T );
	  lift_clause(c);
	  return FALSE;           /* refutation found */
	}
      }
  } /*if latest clause is undefined do old FINDER */
  else {
    for ( c = TH[current_theory]; c; c = c->nextclause ) {
      if ( !c->preprocessed ) {
	if ( !is_good(c) ) {
	  record_refutation( c, T );
	  lift_clause(c);
	  return FALSE;             /* refutation found */
	}
      }
    }
  }

  /*Skip out if doing a negative search; else record some stats */  
  if ( negative ) clauses_fail = 0;
  else if ( cc_badguy && cc_is_good ) return TRUE;
  else {
    if ( T->firstchange < 0 || T->firstchange >= tr_parameters.vlength )
      first_change = tr_parameters.vlength;
    else first_change = T->inorder[T->firstchange];

    set_best(T);
    theBadness = T->best_penalty;
    theWork = T->job->reps[INVALUES].i_report;
    theTime = timestamp();
    set_last_model();
    if (!theBadness) {
      updates[current_theory]++;
      print_model();
      if ((updates[current_theory] && 
	   updates[current_theory]==setting.max_models) ||
	  single_model)
	job_done( &tr_parameters, "limit on number of solutions" );
    }
  }

  return TRUE;
} /* test() */



void print_best_model(int info[], trs T)
{
  vect_into_model(info);
  updates[current_theory]++;
  workNow = T->job->reps[INVALUES].i_report;
  print_model();
}


/*--------------------------------------------------------------------------*/
void vect_into_model( int vector[] )

     /*Called by test().
      *Translates vector[] (contains the values assigned to functions) 
      *into testable data structures.

      *The array thisvector[] is filled with a copy of info[] for purposes 
      *of deciding what should be printed, and of recording the current 
      *model where FINDER is used as a TESTER.
      *(Really just another copy of last_model[])

      *Note that piped functions do not feature in the vector, though they 
      *are in the list of testable functions. */

{
  functor f;
  int i;
  int argument[ARGMAX];

  /* Store a copy of vector[] in case we need to record a new good model*/
  for ( i = 0; i < tr_parameters.vlength; i++ )
    thisvector[i] = vector[i];

  /* Translate to testable data structure */
  for ( f = FN; f; f = f->nextfunct ) 
    if ( !f->piped ) {
      for (i=0; i<f->adicity; i++) argument[i] = 0;
      do
	*(f->table+getoffset(f,argument)) = vector[vec_pos(f,argument)];
      while (another_argument(f,argument));
    }
} /* vect_into_model() */



/*--------------------------------------------------------------------------*/
void set_last_model()
     
     /*Called by test().
      *If all the tests are passed, stores the most recent good model of 
      *the current theory in last_model[] and last_sizes[], and old_fcc[] 
      *fields of the function symbols. 
      *These are later used by model() for testing new clauses*/

{
  int i;
  functor f;
  
  for ( i = 0; i < tr_parameters.vlength; i++ )
    last_model[current_theory][i] = thisvector[i];

  for ( i = 0; sort[i].sname; i++ ) {
    last_sizes[current_theory][i] = sort[i].siz;
    assignset(last_sizmasks[current_theory][i],sort[i].sizmask);
  }

  for ( f = FN; f; f = f->nextfunct ) 
    f->old_fcc[current_theory] = f->fcc;

  last_vlength[current_theory] = tr_parameters.vlength;

} /*  set_last_model() */



/*--------------------------------------------------------------------------*/
boolean is_good( clause cl )

     /*Called by test().
      *Returns TRUE iff the clause is satisfied by the current model.*/
     
{
  wff_stack wf;

  /* if too many variables, search for a refutation instead of testing*/
  if ( cl->varcount > VCMAX ) return tisgood(cl);

  for ( wf = cl->varlist; wf; wf = wf->nxt ){
    /* set the first assignment to variables */
    /* note variables should never have a val_type->siz=0 */
    wf->w->val = (setting.invert_change? 0: wf->w->sym->val_type->siz-1 );
  }
  do
    if (clause_fails(cl) )
      return false;
  while (next_assignment(cl));

  return true;

} /* is_good() */



/*--------------------------------------------------------------------------*/
int find_badness( clause cl )

     /*Called by test().
      *Returns TRUE iff the clause is satisfied by the current model.*/

{
  wff_stack wf;
  int clause_badness=0;

  /* if too many variables, search for a refutation instead of testing*/
  if ( cl->varcount > VCMAX ) {
    if (tisgood(cl)) return 0;
    else return -1;
  }

  for ( wf = cl->varlist; wf; wf = wf->nxt )
    wf->w->val = (setting.invert_change? 0:
		  wf->w->sym->val_type->siz-1 );

  do
    if (clause_fails(cl) )
      clause_badness++;
  while (next_assignment(cl));

  return clause_badness;

} /* is_good() */



/*--------------------------------------------------------------------------*/
void sanity_check(wff w)

     /*Called by clause_fails() .
      *Check that formula is well formed */
{
  int i;
  boolean sane=true;

  if (w->matrix != w->sym->table) {
    report_insanity("Incorrect table pointer",w,-1);
    sane = false;
  }
  if (w->val > w->sym->val_type->siz ||  /* value assigned is greater than 
					    the number of possible values*/
      w->val < 0 ||                      /* value assigned is negative */

      ((w->val == w->sym->val_type->siz) /* value is too big unless the
                                            function is partial */
       && w->sym->total)) {
      report_insanity("Value out of range",w,-1);
      sane = false;

    }

  for (i=0; i < w->sym->adicity ; i++)
    if (w->sbval[i] != &(w->sb[i]->val)) {
      if (!sane) {
	report_insanity("Mismatch of subvalues",0,i);
      }
      else {
	report_insanity("Mismatch of subvalues",w,i);
      }
      sane = false;
    }
  if (!sane) exit(3);
} /* sanity_check() */



/*--------------------------------------------------------------------------*/
void report_insanity(char error_message[], wff w, int code)

     /* Called by  sanity_check() */

{
  if (w) {
    fprintf(WARNING_STREAM,"Insane formula ");
    outwff(WARNING_STREAM,w);
    fprintf(WARNING_STREAM,"\n");
  }
  fprintf(WARNING_STREAM,"%s",error_message);
  if (code>=0) fprintf(WARNING_STREAM," (%d)",code);
  fprintf(WARNING_STREAM,"\n");
} /* insane() */


/*--------------------------------------------------------------------------*/
boolean clause_fails( clause c )

     /*Called by is_good().
      *This is the actual test. 
      *The value of each subformula is calculated and assigned to its 
      *_val field. Then the premises and conclusions of the clause are 
      *checked to see whether it has been refuted.
      *Returns TRUE iff the clause is true in the current model.*/
     
{
  wff_stack wf;
  int i, arity;
  int **sbval;
  int subvalues[ARGMAX];

  for ( wf = c->testlist; wf; wf = wf->nxt ) {
    arity =  wf->w->sym->adicity;
    sbval = wf->w->sbval;
    for (i=0; i<arity; i++) subvalues[i] = *(sbval[i]);
    wf->w->val = *(wf->w->matrix + getoffset(wf->w->sym,subvalues));
  }

  for ( wf = c->conc; wf; wf = wf->nxt ) if ( wf->w->val ) return false;
  for ( wf = c->prem; wf; wf = wf->nxt ) if ( !wf->w->val ) return false;
  return true;
  
} /* clause_fails() */



/*--------------------------------------------------------------------------*/
boolean next_assignment( clause c )
     
     /*Called by is_good().
      *Gets the next assignment of values to the variables in the clause c.
      *Returns FALSE iff all possible values have been exhausted.*/

{
  wff_stack wf;
  boolean not_invert_change;
  int pretest_limit;

  // rather than calculate inside the for loop
  not_invert_change = !setting.invert_change; 
  pretest_limit = setting.pretest_limit;

  /* foreach variable ... */
  for ( wf = c->varlist; wf; wf = wf->nxt )
    if ( c->weight <= pretest_limit || not_invert_change ) {
      if ( !wf->w->val ) wf->w->val = wf->w->sym->val_type->siz-1;
      else {
	wf->w->val--;
	return true;
      }
    }
    else {
      if ( wf->w->val == wf->w->sym->val_type->siz-1 ) wf->w->val = 0;
      else {
	wf->w->val++;
	return true;
      }
    }
  return false;
} /* next_assignment() */



/*--------------------------------------------------------------------------*/
void record_refutation( clause c, TRS *T )

     /*Called by test().
      *Describes the new refutaion to transref.*/

{
  wff_stack wf;
  int argument[ARGMAX];
  int i;

  for ( wf = c->testlist; wf; wf = wf->nxt )
    if ( is_open(wf->w->sym) ) {
      for (i=0; i<wf->w->sym->adicity; i++)
	argument[i] = *(wf->w->sbval[i]);
      Ref( vec_pos(wf->w->sym,argument), T );
    }
} /* record_refutation() */


/*------------------------------------------------------------------------*/
void lift_clause(clause c)

     /*Called by test().
      *Lifts clauses that fail up the order of testing, 
      *so that the clauses doing the work congregate at the top.*/
{
  clause d,e;

  if (c==TH[current_theory]) return;
  for (d=TH[current_theory]; d->nextclause != c; d = d->nextclause);
  if (d==TH[current_theory]) {
    d->nextclause = c->nextclause;
    c->nextclause = d;
    TH[current_theory] = c;
  }
  else {
    for (e=TH[current_theory]; e->nextclause != d; e = e->nextclause);
    d->nextclause = c->nextclause;
    c->nextclause = d;
    e->nextclause = c;
  }
} /* lift_clause() */

/*=============================== END ====================================*/
