/*----------------------------------------------------------------*/
/*      hash.c
/*
/*
/*      functions:
/*
/*      These two functions check whether a clause is in the
/*      clause hash table or whether a new clause can be added
/*
/*      is_clause_in_ht().............................#1
/*      add_clause_to_ht()............................#2
/*
/*      The next four functions deal with the management of the unit
/*      clause hash table.
/*
/*      contradiction()...............................#3
/*      is_unit_in_ht()...............................#4
/*      is_comp_in_ht()...............................#5
/*
/*      The following five functions deal with calculating hashcodes
/*      of literals and clauses.
/*
/*      calc_lit_hc().................................#6
/*      normalize_literal()...........................#7
/*      count_vars()..................................#8
/*      replace_vars()................................#9
/*      calc_alt_hashcodes()..........................#10
/*      test_alt_hashcodes()..........................#11
/*      is_test_in_ht()...............................#12
/*      calc_alt_hash()...............................#13
/*      calc_rev_hash()...............................#13a
/*      calc_clause_hashcode()  ......................#14
/*      sort_hash_array().............................#15
/*      quick_sort()..................................#16
/*      insertion_sort()..............................#17
/*      count_constants().............................#18
/*      
/*----------------------------------------------------------------*/
 
#include <stdio.h>
#include <math.h>
 
#include "const.h"
#include "clause.h"

/*-----------------------------------------------------------------*/
/* externally defined functions                                    */ 
/*-----------------------------------------------------------------*/

extern void fatal();
extern void clear_var_substitution_tble();
extern void print_continuation();

/*-----------------------------------------------------------------*/
/* externally defined constants                                    */
/*-----------------------------------------------------------------*/

extern int *var_substitution_tbl[];
extern clause *clist[];
extern hnode *clause_hash_table;   /* ptrs to clause hash tables   */
extern int array_sizem1;
extern int array_power;
 
extern rnode *rand_array;           /* arrary for random numbers   */
 
extern int s_nod, s_nod0, iter_depth, ply, nextc;
extern int o_hash;
extern int o_maxterm, o_maxvar;
extern int number_fxns, max_lit;
extern int number_const_hash;
extern int number_constants;
extern int extra_pass, phase;
extern int in_elim;
extern int var_in_lit;
extern int o_step, o_print;

extern int s_unit_sub;
extern int s_unit_res;
extern int s_htelim;
 
extern int numb_rows, rows_less_nfxns;
 
extern int s_htentry, s_hthit, s_htfail;
extern int s_ltentry, s_lthit, s_ltfail;
 
/*-----------------------------------------------------------------*/
/* internally defined functions                                    */
/*-----------------------------------------------------------------*/ 

int is_clause_in_ht();
int add_clause_to_ht();
int contradiction();
int is_unit_in_ht();
int is_comp_in_ht();
int calc_lit_hc();
void normalize_literal();
void count_vars();
void replace_vars();
int calc_alt_hashcodes();
void calc_alt_hash();
void calc_rev_hash();
void calc_clause_hashcode();
int test_alt_hashcodes();
int is_test_in_ht();
void sort_hash_array();
void quick_sort();
void insertion_sort();
int count_constants();

/*-----------------------------------------------------------------*/
/* internally defined constants                                    */
/*-----------------------------------------------------------------*/ 

int contradiction_found;
int var_array[64];
int reverse[400];
int *l_reverse, l_reverse_HX, l_reverse_HY;
int var_array_ptr;
 
int magic_hashx,  magic_hashy;  /* hash code for contradictory lit */
int magic_hashz;
int alt_hx,  alt_hy;          /* alternate hash codes for variants */
int v1_count, v2_count;
int lxx;
int llxx;

/*----------------------------------------------------------*/
/* displacements from probe location into hash_table        */
/*----------------------------------------------------------*/
 
static int probe[] = 
{0,3,198991,31,-100853,1727,-218159,2423,-3995821,4529,
		  -5223,50724,-60625,80326,193,0};
     
/*-----------------------------------------------------------------*/
/*      is_clause_in_ht(c)      .............................#1
/*
/*      see if clause c is in the hash table
/*      if so, return 1, otherwise, return 0.
/*
/*      called by:  install_clause()         search.c
/*
/*      calls:      nobody
/*
/*      returns:    0 if clause not in TT,  1 if it is.
/*-----------------------------------------------------------------*/
 
int is_clause_in_ht(c)
clause *c;
{
hnode *hn;
int *p, ht;

p = probe;                                  /* ht probe pointer   */

do {
 
/*-------------------------*/
/* get bits to decide the  */
/* table location to probe */
/*-------------------------*/   
 
   ht = (c->c_X + *p) & array_sizem1;
   hn = &clause_hash_table[ht];
 
/*-----------------------------------------*/
/* return a -1 if hash table does not have */
/* any entry at this hash table location   */
/*-----------------------------------------*/
 
   if(!hn->h_X) {
   return(0);}
 
/*-----------------------------------------*/
/* if same hash code found, check to see   */
/* if clause was a base clause; if so      */
/* return a hit.  also if clause was       */
/* previously found on same iteration      */
/* and no deeper in tree, return a hit.    */
/*-----------------------------------------*/
 
   if(hn->h_X == c->c_X) {
      if(hn->h_Y == c->c_Y) {
         if(!hn->h_in) {
            s_htelim++;
            return(1);}
         if(hn->h_in == iter_depth) {
            if((int) ( hn->h_ply & PLY_MASK) <= ply) {
               s_htelim++;
 
/*-----------------------------*/
/* return 0 on a hit           */
/*-----------------------------*/
 
               return(1);}}}}   
   } while(*(++p));
 
/*-----------------------------*/
/* clause not in hash table    */
/*-----------------------------*/
 
return(0);
}
 
/*-----------------------------------------------------------------*/
/*    add_clause_to_ht(c)           ........................#2
/*
/*    called by:    hash_all_base_clauses()              theo.c
/*                  prove_by_iterative_deepening()       theo.c
/*                  install_clause()                     search.c
/*                  add_link_resolvents()                elim.c
/*                  replace_clause()                     elim.c
/*
/*    calls:        contradiction()                      hash.c
/*
/*    returns:      nil
/*-----------------------------------------------------------------*/
 
int add_clause_to_ht(hx,hy,lits)
int hx, hy, lits;
{
int *p, ht, best_yet;
hnode *hn, *hm;

if(!o_hash) {return(0);}

/*-------------------------------------------------*/
/* if clause has only one literal, see if          */
/* complement has been previously hashed.          */
/*-------------------------------------------------*/

if(lits == 1) {
   contradiction_found = contradiction(hx,hy);
   if(contradiction_found) {return(2);}}

/*--------------------------------------------------*/
/* now set p to point to probe displacement array   */
/*--------------------------------------------------*/

best_yet = 0;
p = probe;

do {
   
/*------------------------------------*/
/* get bits to decide hash table      */
/* location to probe                  */
/*------------------------------------*/   
   
   ht = (hx + *p) & array_sizem1;
   hn = &clause_hash_table[ht];
 
/*--------------------------------------------------------------*/
/* enter new clauses if hn->h_X is zero (ie, no entry)          */
/*                                                              */
/* each entry is four words                                     */
/*                                                              */
/* hash code (2 words)                                          */
/* iteration (1 integer)                                        */
/* ply (1 integer) + 256 * number of literals in clause         */
/*--------------------------------------------------------------*/ 
	
   if(!hn->h_X) { 
      hn->h_X     = hx;
      hn->h_Y     = hy;
      hn->h_in    = iter_depth;
      hn->h_ply   = ply + 256 * lits;  
      if(lits == 1) {s_ltentry++;}
      else {s_htentry++;}
      return(0);}

/*------------------------------*/
/* update an existing entry     */
/*------------------------------*/
        
   if(hn->h_X == hx && hn->h_Y == hy) { 
      if(lits == 1) {s_lthit++;}
      else {s_hthit++;}
      if((int) (hn->h_ply & PLY_MASK) > 1) {
         if(hn->h_in == iter_depth) {	
            if((int) (hn->h_ply & PLY_MASK) >= ply) {
               hn->h_ply = ply + 256 * lits;}
            return(1);}
         else {               /* if different iteration */ 
            hn->h_ply =  ply + 256 * lits;
            hn->h_in  =  iter_depth;
            return(1);}}
      else hn->h_in = iter_depth; /* entry is at ply 1	*/
      return(1);}
 
/*---------------------------------------------*/
/* if another entry is already here, update    */
/* best_yet (best_yet is the best previous     */
/* entry to replace.  that entry with the      */
/* most literals is chosen first, and then     */
/* the one deepest in the tree is chosen next. */
/*---------------------------------------------*/
 
   else {
      if(hn->h_ply > best_yet) {
         best_yet = hn->h_ply;
         hm = hn;}}
   } while(*(++p));
 
/*--------------------------------------------------------------*/
/* if no location found for new entry, (after trying 13 probes) */
/* overwrite an existing entry -- the one with most literals    */
/* and deepest in tree -- if it is a worse clause than the      */
/* current one.                                                 */
/*--------------------------------------------------------------*/

if((best_yet & 0xffffff00)>>8 > lits) { /* replace with new clause if lits equal */
   hm->h_X 	= hx;
   hm->h_Y 	= hy; 
   hm->h_ply  = ply + 256 * lits;
   hm->h_in   = iter_depth;}

if(lits == 1) {s_ltfail++;}
else {s_htfail++;}
return(0);                        /* overwrite entry return    */
}
 
/*--------------------------------------------------------------*/
/*      contradiction(l)          ......................#3
/*
/*      test unit to see whether its complement is in the
/*      unit clause hash table.
/*
/*      called by:  is_unit_in_hash()        hash.c
/*
/*      calls:      nobody
/*
/*      returns:    0  if no contradiction found
/*                  1  if found (negative hashcode found
/*---------------------------------------------------------------*/
 
int contradiction(lx,ly)
int lx,ly;
{
int *p, ht;
hnode *l_try;
 
p = probe;                               /* probe pointer       */
    
do {   
   ht = (-lx + *p) & array_sizem1;
   l_try = &clause_hash_table[ht];
        
   if(!l_try->h_X) {return(0);}
   else 
      if(l_try->h_X == -lx && l_try->h_Y == -ly) {
	       magic_hashx = lx;
	       magic_hashy = ly;
               s_unit_res++;
	       s_ltentry++;
	       return(1);}                                    
      } while(*(++p));
 
return(0);                                 /* no contradiction */
}
 
/*---------------------------------------------------------------*/
/*      is_unit_in_ht(c)          .................#4
/*
/*      examines clause c to see if one of the literals
/*      in c is in hash table and
/*          (1) was placed in the hash table on the
/*              current iteraltion
/*          (2) the hash table unit was at a ply
/*              less than or equal to the current ply
/*
/*      called by:  calc_inferences()          search.c
/*
/*      calls:      nobody
/*
/*      returns:    1 if literal found
/*                  0 if literal not found
/*---------------------------------------------------------------*/
 
int is_unit_in_ht(c)
clause *c;
{
int  *p, ht;
literal *l;
hnode *l_try;
int index;
 
/*----------------------------------*/
/* test for the same literal        */
/*----------------------------------*/
    
index = 0;


for(l = c->c_lit ; l != 0; l = l->l_next) {    /* scan literals     */
   index++;
   p = probe;                                  /* probe pointer     */
   if(!(l->l_id & ELIMLIT))     {     /* if literal not masked out  */
      do {
         ht = (l->l_X + *p) & array_sizem1;
         l_try = &clause_hash_table[ht];
	
         if(!l_try->h_X) break;      /* literal not in hash table  */
            if(l_try->h_X == l->l_X && l_try->h_Y == l->l_Y) {
 
/*----------------------------------*/
/* unit found in hash table         */
/*----------------------------------*/
	       
               if(l_try->h_in == iter_depth) {
                  if((int) (l_try->h_ply & PLY_MASK) <= ply &&
                           ply >= iter_depth){
                     s_unit_sub++;
                     return(index);}}
               else if(l_try->h_in < 2 && 
                    (int) (l_try->h_ply & PLY_MASK) < 2) {
                  s_unit_sub++;
		  return(index);}}            /* lit in hash_table */
      } while(*(++p));}}
return(0);                               /* wasn't found in table */
}
 
/*----------------------------------------------------------------*/
/*      is_comp_in_ht(c)       .............................#5
/*
/*      examines clause c to see if the complement of
/*      a unit in c is in the hash table;
/*
/*      returns: the index of the first such literal found.
/*
/*      called by:  install_clause()        search.c
/*
/*      calls:      nobody
/*
/*      returns:    0   if literal not found in clause  c  or
/*                  lit_index  (the index of the literal found)
/*----------------------------------------------------------------*/
 
int is_comp_in_ht(c)
clause *c;
{
int  *p, ht;
int lit_index; /* records the index of the complem'y lit */ 
literal *l;
hnode *l_try;
 
lit_index = 0;
 
/*--------------------------------------------------*/
/* scan literals of c for a complementary match     */
/*--------------------------------------------------*/
 
for(l = c->c_lit; l != 0; l = l->l_next) {
   p = probe;                                   /* probe pointer     */
   lit_index++;
   if(!(l->l_id & ELIMLIT)) {
      do {
         ht = (-l->l_X + *p) & array_sizem1;
         l_try = &clause_hash_table[ht];
         if(!l_try->h_X) {break;}
         if(l_try->h_X == -l->l_X) {
            if(l_try->h_Y == -l->l_Y) {
               s_unit_res++;
               return(lit_index);}}} while (*(++p));}}
 
return(0);             /* complementary literal not in table   */
}
 
/*---------------------------------------------------------------*/
/*      calc_lit_hc(l,c)       ..........................#6
/*
/*      calc_lit_hc will compute the hashcode of a lit and
/*      store the result in the l_xsum field
/*
/*      called by  hash_all_base_clauses()       theo.c
/*                 generate_literal()            gen.c
/*
/*      calls:     nobody
/*
/*      returns:   lt_terms = number of terms in lit
/*---------------------------------------------------------------*/
 
int calc_lit_hc(l,c)
literal *l;
clause *c;
{
int a, s, s1, t, *p, q;    
rnode *r1;
int lt_terms;            /* l_terms = # of terms in lit  */
 
/*-------------------------------------------------------*/
/* p..........a pointer to the first term of the literal */
/* q..........used as an index into the rand_array       */
/* s,t........two words of 64-bit hash code for lit l    */
/* lt_terms...count of the number of terms in lit l      */
/*-------------------------------------------------------*/ 

p = &l->l_id;        
s = t = 0;     
q = numb_rows;
lt_terms = 0;

for(a = *p++ & ARGMASK; a > 0; a--) {

/*--------------------------------------------------------*/
/* get next term; determine the number of arguments       */
/* increment the count of the number of terms in lit;     */
/* return during Phase 1,2 if too many terms in literal   */
/* the variable "a" keeps track of how many terms remain  */
/* unevalauted.  When a = 0, hashcode has been calculated.*/
/*--------------------------------------------------------*/

   if(lt_terms++ > o_maxterm) {
      if(in_elim) {return(lt_terms);}}
   if(*p & FCNMASK) {           
 
/*---------------------------------------------------------*/
/* function found; set pointer into rand_array             */
/*---------------------------------------------------------*/
 
      r1 = 
        &rand_array[((*p&IDMASK)>>ARGSHIFT)+rows_less_nfxns+q];
      a += *p++ & ARGMASK;}
   else {                     
 
/*---------------------------------------------------------*/
/* variable found; set pointer into rand_array             */
/*---------------------------------------------------------*/

      var_in_lit++; 
      r1 = &rand_array[(*p & MAX_VARS_M1) + q];
      p++;}
	
/*---------------------------------------------------------*/
/* accumulate hashcode in s and t                          */
/*---------------------------------------------------------*/
 
   s = s ^ r1->r_hash1;        
   t = t ^ r1->r_hash2;
	
/*---------------------------------------------------------*/
/* set q to index of top of next row in rand_array         */
/*---------------------------------------------------------*/
 
   q = q + numb_rows;}

/*---------------------------------------------------------*/
/* add in hash code for predicate                          */
/*---------------------------------------------------------*/

s1 = (&rand_array[((l->l_id & IDMASK) >> ARGSHIFT) + 
		MAX_VARS_M1])->r_hash1;
s = s ^ s1;
s1 = (&rand_array[((l->l_id & IDMASK) >> ARGSHIFT) + 
		MAX_VARS_M1])->r_hash2;
t = t ^ s1;

/*---------------------------------------------------------*/
/* If complemented literal, negate hashcode                */
/* save value in literal header elements, l->l_X,l->l_Y    */
/*---------------------------------------------------------*/
 
if(l->l_id & NEGMASK) { 
   l->l_X = -s;
   l->l_Y = -t;}        
else {
   l->l_X = s;
   l->l_Y = t;}

return(lt_terms);
}
 
/*------------------------------------------------------------------*/
/*      normalize_literal(l,c)         ....................#7
/*
/*      called by:      install_clause()                search.c
/*			prove_by_iterative_deepening()  theo.c
/*                      hash_all_base_clauses()         theo.c
/*                      elim_base_clause_hc()           pass2.c
/*                      elim_same_cont_hc()             pass2.c
/*                      save_line()                     pass2.c
/*
/*      calls:          clear_var_substitution_tble()   infer.c
/*                      count_vars()                    hash.c
/*
/*      returns:        nil
/*------------------------------------------------------------------*/
 
void normalize_literal(l,c)
literal *l;
clause *c;
{
int a, *p, k; /* l_terms = # of terms in lit  */
literal *lit0, *lit;
int var_cnt;

lit0 = l;


p = &lit0->l_id;        
var_array_ptr = 0;     

for(k = 0 ; k < o_maxvar ; k++) {var_array[k] = 0;}
	
for(a = *p++ & ARGMASK; a > 0; a--) {
   if(*p & FCNMASK) {a += *p++ & ARGMASK;}
   else { 
 
/*-----------------------------------------------------------*/
/* see if var_array has det a new var name for this variable */
/*-----------------------------------------------------------*/
 
      if(!var_array[*p & MAX_VARS_M1]) {
 
/*---------------------------------------------------------------*/
/* if not, enter new var name                                    */
/*---------------------------------------------------------------*/
 
	     var_array[*p & MAX_VARS_M1] = var_array_ptr++ + 1;}
      *p = var_array[*p & MAX_VARS_M1] - 1;
      if(c->c_bits & VID) *p += MAX_VARS;
      p++;}}

for(lit = c->c_lit; lit != 0; lit = lit->l_next) {

   if((!(lit->l_id & ELIMLIT)) && lit != l) {
      p = &lit->l_id;        
      for(a = *p++ & ARGMASK; a > 0; a--) {
         if(*p & FCNMASK) {a += *p++ & ARGMASK;}
         else {
            if(!var_array[*p & MAX_VARS_M1]) {
	        var_array[*p & MAX_VARS_M1] = var_array_ptr++ + 1;}
            *p = var_array[*p & MAX_VARS_M1] - 1;
            if(c->c_bits & VID) *p += MAX_VARS;
            p++;}}}}
c->c_nv = var_array_ptr;

return;
}

/*--------------------------------------------------------------*/
/* count_vars()         .................... #8
/*
/* count variables in clause after normalized
/*
/* called by: normalize_literal()          hash.c
/*
/* calls:     nobody
/*
/* return:    nil
/*--------------------------------------------------------------*/

void count_vars(l,c) 
literal *l;
clause *c;
{
int *p, a;

p = &l->l_id;

for(a = *p++ & ARGMASK; a> 0; a--) {
   if(*p & FCNMASK) {
      a += *p++ & ARGMASK;}

   else {
       if(!var_substitution_tbl[*p & MAX_VARS_M1]) {
	     var_array[*p & MAX_VARS_M1] = var_array_ptr++ + 1;}
/*	  var_substitution_tbl[*p & MAX_VARS_M1] = (int *)++c->c_nv;}*/
      p++;}}
return;
}

/*--------------------------------------------------------------*/
/* replace_vars()        ............................... #9
/*
/* count variables in clause after normalized
/*
/* called by: normalize_literal()          hash.c
/*
/* calls:     nobody
/*
/* return:    nil
/*--------------------------------------------------------------*/

void replace_vars(c,change) 
clause *c;
int change;
{
int *p, a;
int v1, v2, c1, c2;
literal *l;

v1 = change & 255;
c1 = (change>>8) & 255;
v2 = (change>>16) & 255;
if(v2 == 0) v2 = -1;
c2 = (change>>24) & 255;
for(l = c->c_lit; l != 0; l = l->l_next) {
   p = &l->l_id;

   for(a = *p++ & ARGMASK; a> 0; a--) {
     if(*p & FCNMASK) {
        a += *p++ & ARGMASK;}

     else {
       *p = *p & ~VID;
       if(*p == v1 && v1_count == 1) {
	  *p = 0x80000000 | (c1<<5);}
       if(*p == v2 && v2_count == 1) {
	  *p = 0x80000000 | (c2<<5);}
       p++;}}}
c->c_change = change;
return;
}

/*--------------------------------------------------------------*/
/*      calc_alt_hashcodes(l,c)   ................#10
/*
/*      calculates var in the formula, and for each constant
/*      in the list of constants for theorem, calculate a hash
/*      code and add it to the clause hash table
/*
/*      called by:      install_clause()               search.c
/*
/*      calls:          calc_alt_hash()                hash.c
/*                      add_clause_to_ht()             hash.c
/*
/*      returns:        0;2 if contradiction found
/*---------------------------------------------------------------*/
 
int calc_alt_hashcodes(l,c) 
literal *l;
clause *c;
{
int const1, const2, m, varbl, varbl2;

/*-----------------------------------------------*/
/* if there is one variable, replace it with     */
/* each constant and calculate the resulting     */
/* hash code.  add the hash code to the clause   */
/* hash table.                                   */
/*-----------------------------------------------*/

if(c->c_nv == 1) {
   for(const1 = 1 ; const1 <= number_constants ; const1++) {
      calc_alt_hash(l,0,-1,const1,0);    
      if(add_clause_to_ht(alt_hx,alt_hy,c->c_nlit) == 2) {
	     return(2);}}

/*-----------------------------------------------*/
/* calc hashcode with single variable replaced   */
/* with the variable y.  most likely, but not    */
/* for sure, this will produce a new hashcode    */
/*-----------------------------------------------*/
 
   for(m = 0; m <= 3; m++) {
      calc_alt_hash(l,0,-1,1 - rows_less_nfxns + m,0);
      if(add_clause_to_ht(alt_hx,alt_hy,c->c_nlit) == 2) {
         return(2);}}}
 
/*-----------------------------------------------*/
/* if there are two variables, replace each one  */
/* with a constant and calculate the hashcode    */
/*-----------------------------------------------*/

if(c->c_nv == 2) {
   for(varbl = 0; varbl < 2 ; varbl++) {
     for(const1 = 1; const1 <= number_constants ; const1++) {
        calc_alt_hash(l,var_array[varbl] - 1,-1,const1,0);
        if(add_clause_to_ht(alt_hx,alt_hy,c->c_nlit) == 2) {
            return(2);}}}}

/*------------------------------------------------*/
/* replace every pair of variables with every     */
/* pair of constants, calculate the hashcode      */
/* and add the hashcode to the literal hash table */
/*------------------------------------------------*/ 

if(c->c_nv >= 2 && c->c_nv < 4) {
   for(varbl = 0; varbl < c->c_nv - 1; varbl++) {
      for(varbl2 = varbl + 1 ; varbl2 < c->c_nv; varbl2++) {
         for(const1=1; const1 <= number_const_hash; const1++) {
	    for(const2=1;const2<=number_const_hash; const2++) {
	       calc_alt_hash(l,varbl,varbl2,const1,const2);
	       if(add_clause_to_ht(alt_hx,alt_hy,c->c_nlit) == 2) {
	          return(2);}}}}}}

return(0);
}

/*--------------------------------------------------------------*/
/*      test_alt_hashcodes(l,c)   ................#11
/*
/*      test to see if a variant of, say P(x) is in hashtable
/*
/*      called by:      install_clause()               search.c
/*
/*      calls:          calc_alt_hash()                hash.c
/*                      is_test_in_ht()                hash.c
/*
/*      returns:        0;2 if contradiction found
/*---------------------------------------------------------------*/
 
int test_alt_hashcodes(l,c) 
literal *l;
clause *c;
{
int const1, const2, varbl, varbl2;

/*-----------------------------------------------*/
/* if there is one variable, replace it with     */
/* each constant and calculate the resulting     */
/* hash code.  add the hash code to the clause   */
/* hash table.                                   */
/*-----------------------------------------------*/
 
if(c->c_nv == 1) {
   for(const1 = 1; const1 <= number_constants; const1++) {
      calc_alt_hash(l,0,-1,const1,0);    
      if(is_test_in_ht(-alt_hx,-alt_hy)) {
         return(const1*256 + 0);}}}

/*-----------------------------------------------*/
/* if there are two variables, replace each one  */
/* with a constant and calculate the hashcode    */
/*-----------------------------------------------*/

if(c->c_nv == 2) {
   for(varbl = 0; varbl < 2; varbl++) {
     for(const1 = 1; const1 <= number_constants ; const1++) {
        calc_alt_hash(l,varbl,-1,const1,0);
           if(is_test_in_ht(-alt_hx,-alt_hy)) {
	        return(const1*256 + varbl);}}}}
			
/*------------------------------------------------*/
/* replace every pair of variables with every     */
/* pair of constants, calculate the hashcode      */
/* and add the hashcode to the literal hash table */
/*------------------------------------------------*/ 

if(c->c_nv >= 2  && c->c_nv < 4) {
   for(varbl = 0; varbl < c->c_nv - 1; varbl++) {
      for(varbl2 = varbl + 1; varbl2 < c->c_nv; varbl2++) {
         for(const1=1; const1 <= number_const_hash; const1++) {
	    for(const2=1;const2<=number_const_hash; const2++) {
	       calc_alt_hash(l,varbl,
			       varbl2,const1,const2);
	       if(is_test_in_ht(-alt_hx,-alt_hy)) {
	          return(const1*256 + varbl + 
			      const2*16777216 +
			      varbl2*65536);}}}}}}

return(0);
}
 
/*-----------------------------------------------------------------*/
/*      is_test_in_ht(hx,hy)      .............................#12
/*
/*      see if alternate hash code is in the hash table
/*      if so, return 1, otherwise, return 0.
/*
/*      called by:  test_alt_hashcodes()     hash.c
/*
/*      calls:      nobody
/*
/*      returns:    0 if clause not in TT, depth+1 of node if found
/*-----------------------------------------------------------------*/
 
int is_test_in_ht(hx,hy)
int hx, hy;
{
hnode *hn;
int *p, ht;
	     
p = probe;                                  /* ht probe pointer   */

do {
 
/*------------------------------------------------*/
/* get bits to decide the table location to probe */
/*------------------------------------------------*/   
 
   ht = (hx + *p) & array_sizem1;
   hn = &clause_hash_table[ht];
 
/*-----------------------------------------*/
/* return a -1 if hash table does not have */
/* any entry at this hash table location   */
/*-----------------------------------------*/
 
   if(!hn->h_X) {return(0);}
 
/*-----------------------------------------*/
/* if same hash code found, check to see   */
/* if clause was a base clause; if so      */
/* return a hit.  also if clause was       */
/* previously found on same iteration      */
/* and no deeper in tree, return a hit.    */
/*-----------------------------------------*/
 
   if(hn->h_X == hx) {
      if(hn->h_Y == hy) {
 
/*-----------------------------*/
/* return 1 on a hit           */
/*-----------------------------*/
 
               return(1);}}   
   } while(*(++p));
 
/*-----------------------------*/
/* clause not in hash table    */
/*-----------------------------*/
 
return(0);
}


/*----------------------------------------------------------------*/
/*      calc_alt_hash(l,v1,v2,sub1,sub2)    ...............#13
/*
/*      calculate the hash code for lit l with varible v1 replaced
/*      by constant sub1 and variable v2 replaced by constant sub2.
/*
/*      called by:    calc_alt_hashcodes()         hash.c
/*                    elim_base_clause_hc()        pass2.c
/*                    elim_same_cont_hc()          pass2.c
/*                    alt_on_hash_code_list()      pass2.c
/*
/*      calls:        nobody
/*
/*      returns:      nil
/*----------------------------------------------------------------*/
 
void calc_alt_hash(l,v1,v2,sub1,sub2)
literal *l;
int v1, v2, sub1, sub2;
{
int a, *p, q, s1;
rnode *r1;
int third_var;

p = &l->l_id;        
alt_hx = alt_hy = v1_count = v2_count = 0;     
q = numb_rows;
 
for(a = *p++ & ARGMASK; a > 0; a--) {
   if(*p & FCNMASK) {
 
/*---------------------------------------------------------------*/
/* function found, create pointer to element                     */
/* ((*p & IDMASK)>>ARGSHIFT) + rows_less_nfxns + q               */
/* in rand_array.  Advance pointer p to next term.               */
/*---------------------------------------------------------------*/
 
      r1 = &rand_array[
         ((*p & IDMASK)>>ARGSHIFT) + rows_less_nfxns + q];
      a += *p++ & ARGMASK;}
      
   else if((*p & MAX_VARS_M1) == v1) {
 
/*---------------------------------------------------------------*/
/* variable v1 found, replace with constant sub1, create ptr to  */
/* element    sub1 + rows_less_nfxns + q                         */
/* in rand_array.  Then advance pointer p to next term           */
/*---------------------------------------------------------------*/
 
      v1_count = 1;
      r1 = &rand_array[sub1 + rows_less_nfxns + q];
      p++;}
      
   else if((*p & MAX_VARS_M1) == v2) {
 
/*---------------------------------------------------------------*/
/* variable v2 found, replace with constant sub2, create ptr to  */
/* element    sub1 + rows_less_nfxns + q                         */
/* in rand_array.  Then advance pointer p to next term           */
/*---------------------------------------------------------------*/
 
      v2_count = 1;
      r1 = &rand_array[sub2 + rows_less_nfxns + q];
      p++;}
      
   else  {
 
/*---------------------------------------------------------------*/
/* another variable found -- do not substitute -- create ptr to  */
/* element    (*p & MAX_VARS_M1) + q                             */
/* in rand_array.  Then advance pointer p to next term           */
/*---------------------------------------------------------------*/

      third_var = *p & MAX_VARS_M1; 
      if(third_var > v1 && v1 >= 0) third_var--;
      if(third_var > v2 && v2 >= 0) third_var--;
      r1 = &rand_array[third_var + q];
      p++;}
 
/*---------------------------------------------------------------*/
/* exclusive-or random numbers found in rand_array with current  */
/* partially-calculated hash-code                                */
/*---------------------------------------------------------------*/
 
   alt_hx = alt_hx ^ r1->r_hash1;
   alt_hy = alt_hy ^ r1->r_hash2; 
   q = q + numb_rows;}

/*---------------------------------------------------------------*/
/* Add in hash code for predicate................................*/
/*---------------------------------------------------------------*/

s1 = (&rand_array[((l->l_id & IDMASK) >> ARGSHIFT) +
		MAX_VARS_M1])->r_hash1;
alt_hx = alt_hx ^ s1;
s1 = (&rand_array[((l->l_id & IDMASK) >> ARGSHIFT) +
		MAX_VARS_M1])->r_hash2;
alt_hy = alt_hy ^ s1;
	
/*---------------------------------------------------------------*/
/* If complemented lit, negate hashcode                          */
/*---------------------------------------------------------------*/
 
if(l->l_id & NEGMASK) {  /* If compl lit, negate hc */
   alt_hx = -alt_hx;
   alt_hy = -alt_hy;}     

return;
}

/*----------------------------------------------------------------*/
/*      calc_rev_hash(l,c)    ...............#13a
/*
/*      calls:        nobody
/*
/*      returns:      nil
/*----------------------------------------------------------------*/
 
void calc_rev_hash(l,c)
literal *l;
clause *c;
{
int a, *p, *pp, *p0, i, *end, *end_ptr, j, m; 

p0 = p = &l->l_id;        
i = 200;
j = 200;

for(a = *p++ & ARGMASK; a > 0; a--) {
   if(a == 1 && j == 200) {j--;pp = p;}
   if(j == 200) reverse[i++] = *p;
   if(*p & FCNMASK) {
      a += *p++ & ARGMASK;}
   else {p++;}}

end = p;
end--;
//printf("Append end j=%d end=%d pp=%d\n",j,end,pp);
for(end_ptr = end; end_ptr >= pp; end_ptr--) {
   reverse[j--] = *end_ptr;}

reverse[j] = *p0;
l_reverse = &reverse[j-3];

calc_lit_hc(&reverse[j-3],c);

//printf("\nHC calculated %d %d\n",reverse[j-2],reverse[j-1]);
//for(m = j-3; m < i; m++) {
//   printf("m=%d %x %x\n",m,&reverse[m],reverse[m]);}
//printf("\nBased on clause:");print_clause(c);
//printf("\n");getchar();}
return;
}
 
/*-------------------------------------------------------------*/
/*      calc_clause_hashcode(c)        ...............#14
/*
/*      The hashcode of a cls is calculated. The hash codes for
/*      each literal are placed on a list and then ordered,
/*      then summed and multiplied to yield a hopefully random
/*      number.
/*
/*      called by:  read_in_clause()        theo.c
/*                  generate_clause()       gen.c
/*                  install_proof_line()    pass2.c
/*                  install_clause()        search.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*---------------------------------------------------------------*/
 
void calc_clause_hashcode(c)
clause *c;
{
literal *l;
int temp_lit_total, j, c_hash1, c_hash2;
int l_hash_tempx, l_hash_tempy;  /* used when reordering */
int l_hashx[MAX_LIT_IN_CLAUSE], l_hashy[MAX_LIT_IN_CLAUSE];
int longest_lit, lit_length;
 
/*-------------------------------------------------*/
/* if c = null clause, return special hash code    */
/*-------------------------------------------------*/

if(!c->c_nlit) {
   c->c_X = 362879; c->c_Y = 4021; return;}
		
/*-------------------------------------------------*/
/* scan clause for literal not represented by its  */
/* hash code and put its hash value on the top     */
/* of the list of literal hash codes               */
/*-------------------------------------------------*/

for(l = c->c_lit; l != 0 ; l = l->l_next) {                           
   if(!(l->l_id & ELIMLIT)) {break;}}

if(l->l_next) {
   longest_lit = (int) l->l_next - (int) l;}
else {longest_lit = (int) c->c_end - (int) l;}

l_hashx[0]     = l->l_X ;              
l_hashy[0]     = l->l_Y ;              
l              = l->l_next;                        
temp_lit_total  = 0;

/*-------------------------------------------------*/           
/* list hash values for literals in                */
/* descending order                                */
/*-------------------------------------------------*/

while(l) {
   if(!(l->l_id & ELIMLIT)) {

      if(!l->l_next) {
	 lit_length = (int) c->c_end - (int) l;}
      else {lit_length = (int) l->l_next - (int) l;}
      if(lit_length > longest_lit) {
	 longest_lit = lit_length;}
      l_hashx[temp_lit_total + 1] = l->l_X;
      l_hashy[temp_lit_total + 1] = l->l_Y;
      temp_lit_total++;
	
      for(j = temp_lit_total; j > 0 ; j--) {
	     if(l_hashx[j] < l_hashx[j - 1])  {break;}
	     else if(l_hashx[j]  == l_hashx[j - 1]) {
	        if(l_hashy[j] < l_hashy[j - 1])  {break;}}
	     l_hash_tempx = l_hashx[j - 1];
	     l_hash_tempy = l_hashy[j - 1];
	     l_hashx[j - 1] = l_hashx[j];
	     l_hashy[j - 1] = l_hashy[j];
	     l_hashx[j] = l_hash_tempx;
	     l_hashy[j] = l_hash_tempy;}}
   l = l->l_next;}

c_hash1 = l_hashx[0];
c_hash2 = l_hashy[0];
 
/*-------------------------------------------------*/
/* now form c_hash1, and c_hash2 by adding and     */
/* multiplying the values on the list              */
/*-------------------------------------------------*/
 
for(j = 1 ; j <= temp_lit_total ; j++) {

   if(l_hashx[j] > 0) {
      c_hash1  *= l_hashx[j] * (l_hashy[j] + 5039);
      c_hash2  += l_hashx[j] +  l_hashy[j] * 71;}
   else {       
      c_hash1  *= l_hashx[j] *  (l_hashy[j] * 97);
      c_hash2  += l_hashx[j] +  (l_hashy[j] + 20723);}}


if(c->c_nlit != 1) {
   c_hash1 = (c_hash2 * 20723) ^ c_hash1;}        
c->c_X = c_hash1;
c->c_Y = c_hash2;
return;
}

/*---------------------------------------------------------*/
/* sort_hash_array() .............................# 15
/*
/* called from: main()
/*
/* calls: nobody
/*
/* returns: nil
/*---------------------------------------------------------*/

void sort_hash_array()
{
hnode *hm, *hn, *hp;
int j, k, i, htemp, htemp1, htemp2;

printf("Sort_hash_codes\n");
j = 0;
k = array_sizem1;

test00:
hm = &clause_hash_table[j];
if(hm->h_X == 0 && hm->h_Y == 0) {
   goto test01;}

j++;
goto test00;
test01:
if(j >= k) goto test03;
hn = &clause_hash_table[k];
if(hn->h_X != 0 || hn->h_X != 0) goto test02;
k--;
goto test01;

test02:
htemp = hm->h_X << (32 - array_power);
htemp = htemp + (hm->h_X >> array_power);
hm->h_X = hn->h_X << (32 - array_power);
hm->h_X = hm->h_X + (hn->h_X >> array_power);
hn->h_X = htemp;

htemp = hm->h_Y << (32 - array_power);
htemp = htemp + (hm->h_Y >> array_power);
hm->h_Y = hn->h_Y << (32 - array_power);
hm->h_Y = hm->h_Y + (hn->h_Y >> array_power);
hn->h_Y = htemp;
j++,k--;
if(j < k) goto test00;

test03:

/*
for(i = 0; i < 30; i++) {
if((i%3) == 0) printf("\n");
hn = &clause_hash_table[i];
printf("%3d H%8x %8x ",i,hn->h_X,hn->h_Y);}
printf("END");getchar();


for(i = 0; i < j; i++) {
if((i%3) == 0) printf("\n");
hn = &clause_hash_table[i];
printf("%3d H%10x %10x ",i,hn->h_X,hn->h_Y);}
printf("END");getchar();
*/
quick_sort(0,j - 1);

/*
for(i = 0; i < 30; i++) {
if((i%3) == 0) printf("\n");
hn = &clause_hash_table[i];
printf("%3d H%10x %10x ",i,hn->h_X,hn->h_Y);}
printf("END");getchar();
*/
return;
}

/*-----------------------------------------------------------*/
/* quick_sort() ..............................#16
/*
/* called by: 
/*
/* calls:     Nobody
/*
/* return:    Nil
/*
/*-----------------------------------------------------------*/

void quick_sort(first,last)
int first, last;
{
int i, j, qtemp1, qtemp2, middle, pivot;
hnode *itemp, *jtemp, *ptemp, *ktemp, *qtemp;

if(last - first > 40) {
   middle = (first + last)/2;

   itemp = &clause_hash_table[first];
   jtemp = &clause_hash_table[middle];
   ktemp = &clause_hash_table[last];

   if((unsigned int) itemp->h_X < (unsigned int) jtemp->h_X ||
      ((itemp->h_X == jtemp->h_X) && 
	     ((unsigned int) itemp->h_Y < (unsigned int) jtemp->h_Y))) {
	 qtemp1 = itemp->h_X;
	 qtemp2 = itemp->h_Y;
	 itemp->h_X = jtemp->h_X;
	 itemp->h_Y = jtemp->h_Y;
	 jtemp->h_X = qtemp1;
	 jtemp->h_Y = qtemp2;}
   if((unsigned int) itemp->h_X < (unsigned int) ktemp->h_X ||
      ((itemp->h_X == ktemp->h_X) 
	 && ((unsigned int) itemp->h_Y < (unsigned int) ktemp->h_Y))) {
	 qtemp1 = itemp->h_X;
	 qtemp2 = itemp->h_Y;
	 itemp->h_X = ktemp->h_X;
	 itemp->h_Y = ktemp->h_Y;
	 ktemp->h_X = qtemp1;
	 ktemp->h_Y = qtemp2;}
   if((unsigned int) jtemp->h_X < (unsigned int) ktemp->h_X ||
      ((jtemp->h_X == ktemp->h_X) && 
	  ((unsigned int) jtemp->h_Y < (unsigned int) ktemp->h_Y))) {
	 qtemp1 = jtemp->h_X;
	 qtemp2 = jtemp->h_Y;
	 jtemp->h_X = ktemp->h_X;
	 jtemp->h_Y = ktemp->h_Y;
	 ktemp->h_X = qtemp1;
	 ktemp->h_Y = qtemp2;}
   qtemp = &clause_hash_table[last - 1];
   qtemp1 = qtemp->h_X;
   qtemp2 = qtemp->h_Y;
   qtemp->h_X = jtemp->h_X;
   qtemp->h_Y = jtemp->h_Y;
   jtemp->h_X = qtemp1;
   jtemp->h_Y = qtemp2;

   pivot = last - 1;
   ptemp =&clause_hash_table[pivot];
   i = first;
   j = last - 2;
   for(;;) {
      move_right:
      itemp = &clause_hash_table[i];
      if((unsigned int) itemp->h_X > (unsigned int) ptemp->h_X || 
	   ((itemp->h_X == ptemp->h_X) &&
   	    ((unsigned int) itemp->h_Y > (unsigned int) ptemp->h_Y))) {
	    i++;goto move_right;}
      move_left:
      jtemp = &clause_hash_table[j];
      if((unsigned int) jtemp->h_X < (unsigned int) ptemp->h_X || 
	   ((jtemp->h_X == ptemp->h_X) &&
	   ((unsigned int) jtemp->h_Y < (unsigned int) ptemp->h_Y))) {
	   j--;goto move_left;}
      if( i < j) {
	 qtemp1 = itemp->h_X;
	 qtemp2 = itemp->h_Y;
	 itemp->h_X = jtemp->h_X;
	 itemp->h_Y = jtemp->h_Y;
	 jtemp->h_X = qtemp1;
	 jtemp->h_Y = qtemp2;
	 i++;j--;}
      else {
	 break;}}

   qtemp1 = itemp->h_X;          /* restore pivot */
   qtemp2 = itemp->h_Y;
   ktemp = &clause_hash_table[last - 1];
   itemp->h_X = ktemp->h_X;
   itemp->h_Y = ktemp->h_Y;
   ktemp->h_X = qtemp1;
   ktemp->h_Y = qtemp2;
/*   printf(". ");*/
   quick_sort(first,i - 1);
   quick_sort(i + 1,last);}
else {
   insertion_sort(first,last);}
}
      
/*-----------------------------------------------------------*/
/* insertion_sort() ...............................# 17
/*
/* called by:
/*
/* calls: Nobody
/*
/* returns: Nil
/*
/*-----------------------------------------------------------*/

void insertion_sort(first,last)
int first, last;
{
hnode *hm, *hn, *hp;
int j, k, i, htemp, htemp1, htemp2;
int k1;

/*
printf("\nI FIRST%d-LAST%d\n",first,last);getchar();
for(i = 0; i <= qtotal; i++) {
if((i%3) == 0) printf("\n");
hm = &clause_hash_table[i];
if(i >= first && i <= last) {printf("%3d H%8x %8x ",i,hm->h_X,hm->h_Y);}
else {printf("%3d Z%8x %8x ",i,hm->h_X,hm->h_Y);}}
printf("END");getchar();
*/

k1 = 0 ;
for(i = first + 1; i <= last; i++) { 
   k = i - 1;
   hm = &clause_hash_table[i];
   hn = &clause_hash_table[k];
   if((unsigned int) hm->h_X > (unsigned int) hn->h_X || 
     ((unsigned int) hm->h_X == (unsigned int) hn->h_X && 
      ((unsigned int) hm->h_Y > (unsigned int) hn->h_Y))) {
      htemp1 = hm->h_X;
      htemp2 = hm->h_Y;
      hm->h_X = hn->h_X;
      hm->h_Y = hn->h_Y;
      next_element:
      if(k == first) {
	 hn->h_X = htemp1;
	 hn->h_Y = htemp2;}
      else {
         hp = &clause_hash_table[--k];
         if((unsigned int) hp->h_X < (unsigned int) htemp1 || 
	   ((unsigned int) hp->h_X == (unsigned int) htemp1 && (
	    (unsigned int) hp->h_Y < (unsigned int) htemp2))) {
	    k1++;
	    hn->h_X = hp->h_X;
	    hn->h_Y = hp->h_Y;
	    hn = hp;
	    goto next_element;}
	 else {
	    hn->h_X = htemp1;
	    hn->h_Y = htemp2;}}}}
return;
}

/*---------------------------------------------------------------*/
/*      count_constants()       ..........................#18
/*
/* called by:
/*
/* calls:   Nobody
/*
/* returns: 0 if constant is in literal
/*          1 if no constant is in literal
/*
/*---------------------------------------------------------------*/
 
int count_constants(c)
clause *c;
{
int a, *p;    
literal *l;
 
/*-------------------------------------------------------*/
/* p..........a pointer to the first term of the literal */
/* q..........used as an index into the rand_array       */
/* s,t........two words of 64-bit hash code for lit l    */
/* lt_terms...count of the number of terms in lit l      */
/*-------------------------------------------------------*/ 

l = c->c_lit;
p = &l->l_id;        

for(a = *p++ & ARGMASK; a > 0; a--) {
   if(*p & FCNMASK) {           
      if((*p & ARGMASK) == 0) {return(0);}
      a += *p++ & ARGMASK;}
   else {                     
      p++;}}

return(1);
}
