#include "../ladr/header.h"

#include "../ladr/ioutil.h"
#include "../ladr/commands.h"
#include "../ladr/mindex.h"
#include "../ladr/demod.h"
#include "../ladr/fastparse.h"

#define PROGRAM_NAME    "idfilter"
#define PROGRAM_VERSION "2003-C"
#define PROGRAM_DATE    "Nov 2003"

static char Help_string[] =
"\nThis program takes a file of demodulators (arg 1) and a\n"
"stream of equations (stdin).  Each equation is rewritten.\n"
"If it rewrites to an instance of x=x, it is discarded;\n"
"otherwise it (in the unrewritten form) is send to stdout.\n"
"For example,\n\n"
"   idfilter demods < terms.in > terms.out\n\n"
"The file of demodulators contains optional commands\n"
"then a list of demodulators.  The commands can be used to\n"
"declare infix operations and associativity/commutativity.\n"
"Example file of demodulators:\n\n"
"    op(400, infix, ^).\n"
"    op(400, infix, v).\n"
"    assoc_comm(^).\n"
"    assoc_comm(v).\n"
"    clauses(theory).\n"
"    x ^ x = x.\n"
"    x ^ (x v y) = x.\n"
"    x v x = x.\n"
"    x v (x ^ y) = x.\n"
"    end_of_list.\n\n";


/* Flag IDs */

#define VERBOSE        0

static BOOL member_args(int argc, char **argv, char *str)
{
int i;
 for (i = 1; i < argc; i++)
   if (str_ident(argv[i], str))
     return TRUE;
 return FALSE;
}  /* member_args */

int main(int argc, char **argv)
{
  FILE *head_fp;
  Ilist just;
  Clist demodulators;
  Clist_pos p;
  Mindex idx;
  Term t;
  int tested = 0;
  int passed = 0;
  Uniftype unification_type;
  BOOL backward = member_args(argc, argv, "x");
  BOOL fast_parse = member_args(argc, argv, "fast");

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "-help") ||
      argc < 2) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }

  declare_standard_parse_types();
  init_options();
  init_flag(VERBOSE, "verbose", FALSE);

  if (fast_parse) {
    /* Declare the symbols for fastparse. */
    fast_set_defaults();
  }

  head_fp = fopen(argv[1], "r");
  if (head_fp == NULL)
    fatal_error("demodulator file can't be opened for reading");

  t = read_commands(head_fp, stderr, FALSE, KILL_UNKNOWN);

  if (!is_term(t, "clauses", 1))
    fatal_error("clauses(demodulators) not found");

  if (assoc_comm_symbols() || comm_symbols())
    unification_type = BACKTRACK_UNIF;
  else
    unification_type = ORDINARY_UNIF;

  /* Read list of demodulators. */

  demodulators = read_clause_list(head_fp, stderr, "demodulators", TRUE);

  fclose(head_fp);

  /* AC-canonicalize and index the demodulators. */

  idx = mindex_init(DISCRIM, unification_type, 0);
  for (p = demodulators->first; p != NULL; p = p->next) {
    /* assume positive equality unit */
    Literal lit = p->c->literals;
    Term alpha = lit->atom->args[0];
    mark_oriented_eq(lit->atom);  /* don not check for termination */
    if (assoc_comm_symbols())
      ac_canonical(lit->atom, -1);
    mindex_update(idx, alpha, INSERT);
  }

  if (flag(VERBOSE))
    fwrite_clause_list(stdout, demodulators, TRUE, FALSE);

  /* Read and demodulate terms. */

  t = term_reader(fast_parse);

  while (t != NULL) {
    Term tcopy = copy_term(t);
    BOOL ident;
    tested++;
    if (flag(VERBOSE)) {
      fprintf(stdout, "\nBefore:   "); fwrite_term_nl(stdout, t);
    }

    if (assoc_comm_symbols())
      ac_canonical(tcopy, -1);
    just = NULL;
    tcopy = demodulate(tcopy, idx, &just);

    ident = (is_term(t, "=", 2) && term_ident(ARG(tcopy,0),ARG(tcopy,1)));
    if ((ident && !backward) ||
	(!ident && backward)) {
      passed++;
      term_writer(t, fast_parse);
      fflush(stdout);
    }

    zap_ilist(just);
    zap_term(t);
    zap_term(tcopy);
    t = term_reader(fast_parse);
  }

  printf("%% %s %s %s: tested %d, passed %d in %.2f seconds.\n",
	 PROGRAM_NAME,
	 argv[1], backward ? "x" : "",
	 tested, passed,
	 user_time() / 1000.0);
    
  exit(0);

}  /* main */

