#include "../ladr/header.h"

#include "../ladr/clock.h"
#include "../ladr/ioutil.h"
#include "../ladr/interp.h"
#include "../ladr/commands.h"
#include "../ladr/fastparse.h"

#define PROGRAM_NAME    "modfilter"
#define PROGRAM_VERSION "2003-C"
#define PROGRAM_DATE    "Nov 2003"

static char Help_string[] =
"\nThis program takes a file of interpretations (arg 1) and a stream of\n"
"clauses (stdin).  Clauses that pass the given test (arg 2) are sent\n"
"to stdout. The tests are true_in_all, true_in_some, false_in_all,\n"
"false_in_some.  An optional third argument 'fast' says to read and\n"
"write the clauses in fastparse form (e.g., =mxxx.).  The file of\n"
"interpretations can have operator declarations at the beginning.\n"
"For example,\n\n"
"   modfilter interps true_in_all fast < clauses.in > clauses.out\n\n";

static
BOOL member_args(int argc, char **argv, char *str)
{
  int i;
  for (i = 1; i < argc; i++)
    if (str_ident(argv[i], str))
      return TRUE;
  return FALSE;
}  /* member_args */

static Ordertype interp_compare(Interp a, Interp b)
{
  if (interp_size(a) < interp_size(b))
    return LESS_THAN;
  else if (interp_size(a) > interp_size(b))
    return GREATER_THAN;
  else
    return SAME_AS;
}  /* interp_compare

/*************
 *
 *   sort_interps()
 *
 *************/

static
void sort_interps(Plist interps)
{
  int n = plist_count(interps);
  void **a = malloc(n * sizeof(void *));
  Plist p;
  int i = 0;

  for (p = interps; p; p = p->next)
    a[i++] = p->v;
  merge_sort(a, n, (Ordertype (*)(void*,void*)) interp_compare);
  i = 0;
  for (p = interps; p; p = p->next)
    p->v = a[i++];
}  /* sort_interps */

/*************
 *
 *   find_interp(clause, list-of-interpretations, models-flag)
 *
 *   Look for an interpretation that models (or doesn't model) a given clause.
 *   That is, return the first interpretation in which the clause is true
 *   (not true); return NULL if the clause is false (true) in all the
 *   interpretations.
 *
 *************/

static Interp find_interp(Clause c, Plist interps, BOOL models)
{
  Plist p = interps;
  BOOL value;

  while (p != NULL) {
    value = eval_clause(c, p->v);
    if ((models && value) || (!models && !value))
      return p->v;
    else
      p = p->next;
  }
  return NULL;
}  /* find_interp */

#define TRUE_IN_ALL   0
#define TRUE_IN_SOME  1
#define FALSE_IN_ALL  2
#define FALSE_IN_SOME 3

int main(int argc, char **argv)
{
  FILE *interp_fp;
  Clause c;
  Term t;
  Interp interp;
  Plist interps = NULL;
  int operation = -1;
  BOOL fast;
  unsigned long int checked = 0;
  unsigned long int passed = 0;

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "-help") ||
      argc < 3) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }

  if (str_ident(argv[2], "true_in_all"))
    operation = TRUE_IN_ALL;
  else if (str_ident(argv[2], "true_in_some"))
    operation = TRUE_IN_SOME;
  else if (str_ident(argv[2], "false_in_all"))
    operation = FALSE_IN_ALL;
  else if (str_ident(argv[2], "false_in_some"))
    operation = FALSE_IN_SOME;
  else
    fatal_error("modfilter: operation should be {true,false}_in_{all,some}");

  interp_fp = fopen(argv[1], "r");
  if (interp_fp == NULL)
    fatal_error("modfilter: interpretation file cannot be opened for reading");

  fast = (argc == 4 && str_ident(argv[3], "fast"));

  declare_standard_parse_types();
  init_options();

  t = read_commands(interp_fp, stderr, FALSE, KILL_UNKNOWN);

  /* terms(interpretations) and end_of_list are optional. */
  
  if (is_term(t, "terms", 1))
    t = read_term(interp_fp, stderr);  /* get first interpretation */

  while (t != NULL && !end_of_list_term(t)) {
    interp = compile_interp(t);
    interps = plist_append(interps, interp);
    zap_term(t);
    t = read_term(interp_fp, stderr);
  }
  fclose(interp_fp);

  sort_interps(interps);  /* puts smallest ones first */

  if (fast) {
    /* Declare the symbols for fastparse. */
    fast_set_defaults();
  }
  
  /* Evaluate each clause on stdin. */

  c = clause_reader(fast);

  while (c != NULL) {

    checked++;

    interp = find_interp(c, interps,
			 operation == TRUE_IN_SOME ||
			 operation == FALSE_IN_ALL);

    if ((interp && (operation==TRUE_IN_SOME || operation==FALSE_IN_SOME)) ||
	(!interp && (operation==TRUE_IN_ALL || operation==FALSE_IN_ALL))) {
    
      passed++;
      clause_writer(c, fast);
    }
    zap_clause(c);
    c = clause_reader(fast);
  }
  
  printf("%% modfilter %s %s: checked %lu, passed %lu, %.2f seconds.\n",
	 argv[1], argv[2], checked, passed,
	 user_time() / 1000.0);

  exit(passed > 0 ? 0 : 1);
}  /* main */

