#include "formula.h"

/* Private definitions and types */

/*
 * memory management (mostly copied from term.c)
 */

/*
 * Note that this is unusual, because there is an array of avail lists
 * for formulas of different arities. Formulas with arities >= MAX_FORM_AVAIL-1
 * all go into the same list.
 */

#define MAX_FORM_AVAIL       100
#define FSTACK_ARRAY_SIZE    100  /* see use below */

static Formula formula_avail[MAX_FORM_AVAIL];
static unsigned formula_gets, formula_frees, formula_avails, formula_mem;

/* end of counters for each type */

/*************
 *
 *   Formula get_formula(arity)
 *
 *************/

static
Formula get_formula(int arity)
{
  Formula p;
  
  formula_gets++;
  
  if (arity == 0) {
    if (formula_avail[0] != NULL) {
      formula_avails--;
      p = formula_avail[0];
      formula_avail[0] = (Formula) p->kids;
    }
    else {
      p = tp_alloc(sizeof(struct formula));
      p->arity = 0;
      formula_mem += sizeof(struct formula);
    }
    p->kids = NULL;
  }
  else if (arity < MAX_FORM_AVAIL-1) {
    if (formula_avail[arity] != NULL) {
      formula_avails--;
      p = formula_avail[arity];
      formula_avail[arity] = p->kids[0];
    }
    else {
      p = tp_alloc(sizeof(struct formula));
      p->kids = tp_alloc(sizeof(Formula) * arity);
      p->arity = arity;
      formula_mem += sizeof(struct formula) + sizeof(Formula) * arity;
    }
  }
  else {  /* must use multiple-arity list */
    Formula t2;
    for (p = formula_avail[MAX_FORM_AVAIL-1], t2 = NULL;
	 p && p->arity != arity;
	 t2 = p, p = p->kids[0]);
    if (p != NULL) {
      formula_avails--;
      if (t2 != NULL)
	t2->kids[0] = p->kids[0];
      else
	formula_avail[MAX_FORM_AVAIL-1] = p->kids[0];
    }
    else {
      p = tp_alloc(sizeof(struct formula));
      p->kids = tp_alloc(sizeof(Formula) * arity);
      p->arity = arity;
      formula_mem += sizeof(struct formula) + sizeof(Formula) * arity;
    }
  }
  
  p->type = ATOM_FORM;
  p->qvar = "";
  p->atom = NULL;
  return p;
}  /* get_formula */

/*************
 *
 *    free_formula()
 *
 *************/

/* DOCUMENTATION
This routine frees a formula node only.  If you wish to recursively
free all of the subformulas as well, call zap_formula(t) instead.
*/

/* PUBLIC */
void free_formula(Formula p)
{
  formula_frees++;
  formula_avails++;
  if (p->arity == 0) {
    p->kids = (struct formula **) formula_avail[0];
    formula_avail[0] = p;
  }
  else if (p->arity < MAX_FORM_AVAIL-1) {
    p->kids[0] = formula_avail[p->arity];
    formula_avail[p->arity] = p;
  }
  else {
    p->kids[0] = formula_avail[MAX_FORM_AVAIL-1];
    formula_avail[MAX_FORM_AVAIL-1] = p;
  }
}  /* free_formula */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_formula_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for Formulas.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_formula_mem(FILE *fp, BOOL heading)
{
  int n;
  
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");
  
  n = sizeof(struct formula);
  fprintf(fp, "formula (%4d)      %11u%11u%11u%11u%9.1f K\n",
	  n, formula_gets, formula_frees, formula_gets - formula_frees, formula_avails,
	  formula_mem / 1024.);
  
  /* end of printing for each type */
  
}  /* fprint_formula_mem */

/*************
 *
 *   p_formula_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints memory usage statistics for Formulas to stdout.
*/

/* PUBLIC */
void p_formula_mem(void)
{
  fprint_formula_mem(stdout, 1);
}  /* p__mem */

/*
 *  end of memory management
 */

/*************
 *
 *   formula_get()
 *
 *************/

static
Formula formula_get(int arity, Ftype type)
{
  Formula f = get_formula(arity);
  f->type = type;
  return f;
}  /* formula_get */

/*************
 *
 *   quant_form()
 *
 *************/

static
int quant_form(Formula f)
{
  return (f->type == ALL_FORM || f->type == EXISTS_FORM);
}  /* quant_form */

/*************
 *
 *   zap_formula()
 *
 *************/

/* DOCUMENTATION
Free a formula, including all of its subformulas, including its atoms.
*/

/* PUBLIC */
void zap_formula(Formula f)
{
  if (f->type == ATOM_FORM)
    zap_term(f->atom);
  else {
    int i;
    for (i = 0; i < f->arity; i++)
      zap_formula(f->kids[i]);
  }
  free_formula(f);
}  /* zap_formula */

/*************
 *
 *   flatten_top()
 *
 *************/

static
Formula flatten_top(Formula f)
{
  if (f->type != AND_FORM && f->type != OR_FORM)
    return f;
  else {
    BOOL operate = FALSE;
    int n = 0;  /* count new arity */
    int i;
    for (i = 0; i < f->arity; i++) {
      if (f->type != f->kids[i]->type)
	n++;
      else {
	n += (f->kids[i]->arity);
	operate = TRUE;
      }
    }
    if (!operate)
      return f;
    else {
      Formula g = formula_get(n, f->type);
      int i, j;
      j = 0;
      for (i = 0; i < f->arity; i++) {
	if (f->kids[i]->type != f->type)
	  g->kids[j++] = f->kids[i];
	else {
	  int k;
	  for (k = 0; k < f->kids[i]->arity; k++)
	    g->kids[j++] = f->kids[i]->kids[k];
	  free_formula(f->kids[i]);
	}
      }
      free_formula(f);
      /* If the new formula has just one argument, return that argument. */
      if (g->arity == 1) {
	Formula h = g->kids[0];
	free_formula(g);
	return h;
      }
      else
	return g;
    }
  }
}  /* flatten_top */

/*************
 *
 *   formula_flatten()
 *
 *************/

/* DOCUMENTATION
This routine (recursively) flattens all AND and OR formulas.
*/

/* PUBLIC */
Formula formula_flatten(Formula f)
{
  int i;
  for (i = 0; i < f->arity; i++)
    f->kids[i] = formula_flatten(f->kids[i]);
  return flatten_top(f);
}  /* flatten */
  
/*************
 *
 *   term_to_formula()
 *
 *************/

/* DOCUMENTATION
<P>
Assume that no subterm (of t) representing a formula is a
term of type VARIABLE.
*/

/* PUBLIC */
Formula term_to_formula(Term t)
{
  Formula f;
  Ftype type;
  char *str = sn_to_str(SYMNUM(t));

  if (ARITY(t) == 1 && str_ident(str, QUANT_SYM)) {
    /* example: [all,x,exists,y,all,z,form] */
    Term l = t->args[0];
    int n = listterm_length(l);
    if (n % 2 != 1) {
      fatal_error("term_to_formula, bad quantified formula.");
    }
    f = term_to_formula(listterm_i(l, n));
    for (n = n-2; n > 0; n = n - 2) {
      Term quant = listterm_i(l, n);
      Term var   = listterm_i(l, n+1);
      Formula g;
      Ftype qtype;
      qtype = (is_symbol(SYMNUM(quant),ALL_SYM,0) ? ALL_FORM : EXISTS_FORM);
      g = formula_get(1, qtype);
      g->kids[0] = f;
      g->qvar = sn_to_str(SYMNUM(var));
      f = g;
    }
  }
  else {
    if (ARITY(t) == 0 && str_ident(str, TRUE_SYM))
      type = AND_FORM;
    else if (ARITY(t) == 0 && str_ident(str, FALSE_SYM))
      type = OR_FORM;
    else if (ARITY(t) == 1 && str_ident(str, NOT_SYM))
      type = NOT_FORM;
    else if (ARITY(t) == 2 && str_ident(str, AND_SYM))
      type = AND_FORM;
    else if (ARITY(t) == 2 && str_ident(str, OR_SYM))
      type = OR_FORM;
    else if (ARITY(t) == 2 && str_ident(str, IFF_SYM))
      type = IFF_FORM;
    else if (ARITY(t) == 2 && str_ident(str, IMP_SYM))
      type = IMP_FORM;
    else if (ARITY(t) == 2 && str_ident(str, IMPBY_SYM))
      type = IMPBY_FORM;
    else
      type = ATOM_FORM;

    if (type == ATOM_FORM) {
      f = formula_get(0, ATOM_FORM);
      f->atom = copy_term(t);
    }
    else if (type == NOT_FORM) {
      f = formula_get(1, NOT_FORM);
      f->kids[0] = term_to_formula(t->args[0]);
    }
    else if (ARITY(t) == 0) {
      f = formula_get(0, type);
    }
    else {
      f = formula_get(2, type);
      f->kids[0] = term_to_formula(t->args[0]);
      f->kids[1] = term_to_formula(t->args[1]);
    }
  }
  return flatten_top(f);
}  /* term_to_formula */

/*************
 *
 *   formula_to_term()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Term formula_to_term(Formula f)
{
  Term t = NULL;

  switch (f->type) {
  case ATOM_FORM:
    t = copy_term(f->atom);
    break;
  case NOT_FORM:
    t = get_rigid_term(NOT_SYM, 1);
    t->args[0] = formula_to_term(f->kids[0]);
    break;
  case IFF_FORM:
    t = get_rigid_term(IFF_SYM, 2);
    t->args[0] = formula_to_term(f->kids[0]);
    t->args[1] = formula_to_term(f->kids[1]);
    break;
  case IMP_FORM:
    t = get_rigid_term(IMP_SYM, 2);
    t->args[0] = formula_to_term(f->kids[0]);
    t->args[1] = formula_to_term(f->kids[1]);
    break;
  case IMPBY_FORM:
    t = get_rigid_term(IMPBY_SYM, 2);
    t->args[0] = formula_to_term(f->kids[0]);
    t->args[1] = formula_to_term(f->kids[1]);
    break;
  case AND_FORM:
  case OR_FORM:
    if (f->arity == 0)
      t = get_rigid_term(f->type == AND_FORM ? TRUE_SYM : FALSE_SYM, 0);
    else {
      int i = f->arity-1;
      t = formula_to_term(f->kids[i]);
      for (i--; i >= 0; i--) {
	Term t1 = get_rigid_term(f->type == AND_FORM ? AND_SYM : OR_SYM, 2);
	t1->args[0] = formula_to_term(f->kids[i]);
	t1->args[1] = t;
	t = t1;
      }
    }
    break;
  case ALL_FORM:
  case EXISTS_FORM:
    {
      Term l = get_nil_term();
      Formula g;
      for (g = f; quant_form(g); g = g->kids[0]) {
	Term quant, var;
	quant = get_rigid_term(g->type == ALL_FORM ? ALL_SYM : EXISTS_SYM, 0);
	var = get_rigid_term(g->qvar, 0);
	l = listterm_append(l, quant);
	l = listterm_append(l, var);
      }
      l = listterm_append(l, formula_to_term(g));
      t = get_rigid_term(QUANT_SYM, 1);
      t->args[0] = l;
    }
    break;
  }
  return t;
}  /* formula_to_term */

/*************
 *
 *   fprint_formula()
 *
 *************/

/* DOCUMENTATION
This routine prints a formula to a file.
If you wish to have a formula printed without extra parentheses,
you can call fprint_formula_term() instead.
*/

/* PUBLIC */
void fprint_formula(FILE *fp, Formula f)
{
  if (f->type == ATOM_FORM) {
    fprintf(fp, "(");
    fprint_term(fp, f->atom);
    fprintf(fp, ")");
  }
  else if (f->type == NOT_FORM) {
    fprintf(fp, "(%s ", NOT_SYM);
    fprint_formula(fp, f->kids[0]);
    fprintf(fp, ")");
  }
  else if (f->type == IFF_FORM) {
    fprintf(fp, "(");
    fprint_formula(fp, f->kids[0]);
    fprintf(fp, " %s ", IFF_SYM);
    fprint_formula(fp, f->kids[1]);
    fprintf(fp, ")");
  }
  else if (f->type == IMP_FORM) {
    fprintf(fp, "(");
    fprint_formula(fp, f->kids[0]);
    fprintf(fp, " %s ", IMP_SYM);
    fprint_formula(fp, f->kids[1]);
    fprintf(fp, ")");
  }
  else if (f->type == IMPBY_FORM) {
    fprintf(fp, "(");
    fprint_formula(fp, f->kids[0]);
    fprintf(fp, " %s ", IMPBY_SYM);
    fprint_formula(fp, f->kids[1]);
    fprintf(fp, ")");
  }
  else if (quant_form(f)) {
    fprintf(fp, "(%s %s ", f->type==ALL_FORM ? ALL_SYM : EXISTS_SYM, f->qvar);
    fprint_formula(fp, f->kids[0]);
    fprintf(fp, ")");
  }
  else if (f->type == AND_FORM || f->type == OR_FORM) {
    if (f->arity == 0)
      fprintf(fp, "%s", f->type == AND_FORM ? TRUE_SYM : FALSE_SYM);
    else {
      int i;
      fprintf(fp, "(");
      for (i = 0; i < f->arity; i++) {
	fprint_formula(fp, f->kids[i]);
	if (i < f->arity-1)
	  fprintf(fp, " %s ", f->type == AND_FORM ? AND_SYM : OR_SYM);
      }
      fprintf(fp, ")");
    }
  }
}  /* fprint_formula */

/*************
 *
 *   p_formula()
 *
 *************/

/* DOCUMENTATION
This routine prints a formula, followed by ".\n" and fflush, to stdout.
If you wish to have a formula printed without extra parentheses,
you can call p_formula_term() instead.
If you don't want the newline, use fprint_formula() instead.
*/

/* PUBLIC */
void p_formula(Formula c)
{
  fprint_formula(stdout, c);
  printf(".\n");
  fflush(stdout);
}  /* p_formula */

/*************
 *
 *   formula_ident()
 *
 *************/

/* DOCUMENTATION
This Boolean function checks if two formulas are identical.
The routine term_ident() checks identity of atoms.
<P>
The test is for strict identity---it does not consider
renamability of bound variables, permutability of AND or OR,
or symmetry of IFF or equality.
*/

/* PUBLIC */
int formula_ident(Formula f, Formula g)
{
  if (f->type != g->type || f->arity != g->arity)
    return 0;
  else if (f->type == ATOM_FORM)
    return term_ident(f->atom, g->atom);
  else if (quant_form(f))
    return (str_ident(f->qvar,g->qvar) &&
	    formula_ident(f->kids[0],g->kids[0]));
  else {
    int i;
    int ok = 0;
    for (i = 0; i < f->arity && ok; i++)
      ok = formula_ident(f->kids[i], g->kids[i]);
    return ok;
  }
}  /* formula_ident */

/*************
 *
 *   formula_copy()
 *
 *************/

/* DOCUMENTATION
This function returns a copy of the given formula.
All subformulas, including the atoms, are copied.
*/

/* PUBLIC */
Formula formula_copy(Formula f)
{
  Formula g = formula_get(f->arity, f->type);

  if (f->type == ATOM_FORM)
    g->atom = copy_term(f->atom);
  else {
    int i;
    if (quant_form(f))
      g->qvar = f->qvar;
    for (i = 0; i < f->arity; i++)
      g->kids[i] = formula_copy(f->kids[i]);
  }
  return g;
}  /* formula_copy */


/*************
 *
 *   dual()
 *
 *************/

static
int dual(int op)
{
  switch (op) {
  case AND_FORM: return OR_FORM;
  case OR_FORM: return AND_FORM;
  case ALL_FORM: return EXISTS_FORM;
  case EXISTS_FORM: return ALL_FORM;
  default: return op;
  }
}  /* dual */

/*************
 *
 *   and()
 *
 *************/

static
Formula and(Formula a, Formula b)
{
  Formula f = formula_get(2, AND_FORM);
  f->kids[0] = a;
  f->kids[1] = b;
  return f;
}  /* and */

/*************
 *
 *   or()
 *
 *************/

static
Formula or(Formula a, Formula b)
{
  Formula f = formula_get(2, OR_FORM);
  f->kids[0] = a;
  f->kids[1] = b;
  return f;
}  /* or */

/*************
 *
 *   imp()
 *
 *************/

static
Formula imp(Formula a, Formula b)
{
  Formula f = formula_get(2, IMP_FORM);
  f->kids[0] = a;
  f->kids[1] = b;
  return f;
}  /* imp */

/*************
 *
 *   impby()
 *
 *************/

static
Formula impby(Formula a, Formula b)
{
  Formula f = formula_get(2, IMPBY_FORM);
  f->kids[0] = a;
  f->kids[1] = b;
  return f;
}  /* impby */

/*************
 *
 *   not()
 *
 *************/

static
Formula not(Formula a)
{
  Formula f = formula_get(1, NOT_FORM);
  f->kids[0] = a;
  return f;
}  /* not */

/*************
 *
 *   nnf()
 *
 *************/

/* DOCUMENTATION
Transform a formula into negation normal form (NNF).  (NNF means
that all propositional connectives have been rewritten in terms of
AND, OR and NOT, and all negation signs ar up against atomic formulas).
<P>
The argument "pref" should be either CONJUNCTION or DISJUNCTION,
and it specifies the preferred form to use when translating IFFs.
<P>
Do not refer to the given formula after the call; a good way to call
this routine is <TT>f = nnf(f, CONJUNCTION)</TT>.
*/

/* PUBLIC */
Formula nnf(Formula f, Fpref pref)
{
  if (f->type == ATOM_FORM)
    return f;
  else if (quant_form(f)) {
    f->kids[0] = nnf(f->kids[0], pref);
    return f;
  }
  else if (f->type == AND_FORM || f->type == OR_FORM) {
    int i;
    for (i = 0; i < f->arity; i++)
      f->kids[i] = nnf(f->kids[i], pref);
    return f;
  }
  else if (f->type == IMP_FORM) {
    Formula g = nnf(or(not(f->kids[0]), f->kids[1]), pref);
    free_formula(f);
    return g;
  }
  else if (f->type == IMPBY_FORM) {
    Formula g = nnf(or(f->kids[0], not(f->kids[1])), pref);
    free_formula(f);
    return g;
  }
  else if (f->type == IFF_FORM) {
    Formula g;
    Formula a = f->kids[0];
    Formula b = f->kids[1];
    Formula ac = formula_copy(a);
    Formula bc = formula_copy(b);

    if (pref == CONJUNCTION)
      g = nnf(and(imp(a,b), impby(ac,bc)), pref);
    else
      g = nnf(or(and(a,b),and(not(ac),not(bc))), pref);
	     
    free_formula(f);
    return g;
  }

  /* NOT */

  else if (f->type == NOT_FORM) {
    Formula h = f->kids[0];
    if (h->type == ATOM_FORM)
      return f;
    else if (h->type == NOT_FORM) {
      Formula g = nnf(h->kids[0], pref);
      free_formula(h);
      free_formula(f);
      return g;
    }
    else if (quant_form(h)) {
      Formula g = formula_get(1, dual(h->type));
      g->qvar = h->qvar;
      g->kids[0] = nnf(not(h->kids[0]), pref);
      free_formula(h);
      free_formula(f);
      return g;
    }
    else if (h->type == AND_FORM || h->type == OR_FORM) {
      Formula g = formula_get(h->arity, dual(h->type));
      int i;
      for (i = 0; i < h->arity; i++)
	g->kids[i] = nnf(not(h->kids[i]), pref);
      free_formula(h);
      free_formula(f);
      return g;
    }
    else if (h->type == IMP_FORM) {
      Formula g = nnf(and(h->kids[0], not(h->kids[1])), pref);
      free_formula(h);
      free_formula(f);
      return g;
    }
    else if (h->type == IMPBY_FORM) {
      Formula g = nnf(and(not(h->kids[0]), h->kids[1]), pref);
      free_formula(h);
      free_formula(f);
      return g;
    }
    else if (h->type == IFF_FORM) {
      Formula g;
      Formula a = h->kids[0];
      Formula b = h->kids[1];
      Formula ac = formula_copy(a);
      Formula bc = formula_copy(b);

      if (pref == CONJUNCTION)
	g = nnf(and(or(a,b),or(not(ac),not(bc))), pref);
      else
	g = nnf(or(and(a,not(b)),and(not(ac),bc)), pref);

      free_formula(h);
      free_formula(f);
      return g;
    }
    else
      return f;
  }  /* NOT */
  else
    return f;
}  /* nnf */

/*************
 *
 *   idempotence()
 *
 *************/

static
Formula idempotence(Formula f)
{
  if (f->type != AND_FORM && f->type != OR_FORM)
    return f;
  else {
    int new_arity = f->arity;
    int i;
    for (i = 0; i < f->arity-1; i++) {
      if (f->kids[i] != NULL) {
	int j;
	for (j = i+1; j < f->arity; j++) {
	  if (f->kids[j] != NULL && formula_ident(f->kids[i], f->kids[j])) {
	    zap_formula(f->kids[j]);
	    f->kids[j] = NULL;
	    new_arity--;
	  }
	}
      }
    }
    if (new_arity == f->arity)
      return f;
    else {
      int j = 0;  /* index for g */
      Formula g = formula_get(new_arity, f->type);
      for (i = 0; i < f->arity; i++) {
	if (f->kids[i] != NULL) {
	  g->kids[j] = f->kids[i];
	  j++;
	}
      }
      free_formula(f);
      if (g->arity != 1)
	return g;
      else {
	Formula h = g->kids[0];
	free_formula(g);
	return h;
      }
    }
  }
}  /* idempotence */

/*************
 *
 *   contains_complements()
 *
 *************/

static
int contains_complements(Formula f)
{
  int i, complements;
  for (i = 0, complements = 0; i < f->arity-1 && !complements; i++) {
    int j;
    for (j = i+1; j < f->arity && !complements ;j++) {
      Formula a = f->kids[i];
      Formula b = f->kids[j];
      if (a->type == NOT_FORM && formula_ident(a->kids[0], b))
	complements = 1;
      else if (b->type == NOT_FORM && formula_ident(a, b->kids[0]))
	complements = 1;
    }
  }
  return complements;
}  /* contains_complements */

/*************
 *
 *   contains_empty_child()
 *
 *************/

static
int contains_empty_child(Formula f, int type)
{
  int i, empty_child;
  for (i = 0, empty_child = 0; i < f->arity && !empty_child; i++)
    empty_child = (f->kids[i]->type == type && f->kids[i]->arity == 0);
  return empty_child;
}  /* contains_empty_child */

/*************
 *
 *   simplify_top()
 *
 *************/

static
Formula simplify_top(Formula f)
{
  if (f->arity == 0)
    return f;
  else if (f->type == AND_FORM || f->type == OR_FORM) {
    f = flatten_top(f);
    /* Since flattened, so no child of AND_FORM is "true" (empty AND_FORM),
     * and dually for OR_FORM.
     */
    if (contains_empty_child(f, dual(f->type)) || contains_complements(f)) {
      Formula g = formula_get(0, dual(f->type));  /* "false" or "true" */
      zap_formula(f);
      return g;
    }
    else {
      return idempotence(f);
    }
  }  /* AND/OR */
  else if (f->type == NOT_FORM && f->kids[0]->type == NOT_FORM) {
    Formula g = f->kids[0]->kids[0];
    free_formula(f->kids[0]);
    free_formula(f);
    return g;
  }
  else
    return f;
}  /* simplify_top */

/*************
 *
 *   formula_simplify()
 *
 *************/

/* DOCUMENTATION
This routine applies basic propositional simplification to a formula.
And/or flattening, idempotence, not-not, and true-false, simplification
are applied, but distributivity is <I>not</I> applied.
The given formula is assumed to be in NNF (negation normal form).
<P>
Do not refer to the given formula after the call; a good way to call
this routine is <TT>f = simplify(f)</TT>.
*/

/* PUBLIC */
Formula formula_simplify(Formula f)
{
  if (f->type == ATOM_FORM)
    return f;
  else {
    int i;
    for (i = 0; i < f->arity; i++) {
      f->kids[i] = formula_simplify(f->kids[i]);
    }
    return simplify_top(f);
  }
}  /* formula_simplify */

/*************
 *
 *   bump_odometer()
 *
 *************/

static
int bump_odometer(int *odo, int *limit, int n)
{
  int i = n-1;
  while (i > 0 && odo[i] == limit[i]-1) {
    odo[i] = 0;
    i--;
  }
  if (i < 0)
    return 0;
  else {
    odo[i]++;
    return 1;
  }
}  /* bump_odometer */

/*************
 *
 *   distribute()
 *
 *************/

static
Formula distribute(Formula f, BOOL simp_flag)
{
  int i, j, new_arity;
  Formula g;
  int arity = f->arity;

  /* Some trickery: If we need more room than is reserved on the
   * stack, malloc some from the heap (and return it when done).
   */

  int a1[FSTACK_ARRAY_SIZE], a2[FSTACK_ARRAY_SIZE], *odo, *limit;
  
  if (arity > FSTACK_ARRAY_SIZE) {
    odo   = malloc((arity * sizeof(int)));
    limit = malloc((arity * sizeof(int)));
  }
  else {
    odo = a1;
    limit = a2;
  }

  new_arity = 1;
  for (i = 0; i < arity; i++) {
    int tmp;
    odo[i] = 0;
    limit[i] = (f->kids[i]->type == dual(f->type) ? f->kids[i]->arity : 1);
    tmp = new_arity * limit[i];
    /* check for overflow */
    if (tmp < new_arity)
      fatal_error("distribute: formula too big");
    else
      new_arity = tmp;
  }
  g = formula_get(new_arity, dual(f->type));
  for (j = 0; j < new_arity; j++) {
    Formula h = formula_get(arity, f->type);
    for (i = 0; i < arity; i++) {
      if (limit[i] == 1)
	h->kids[i] = formula_copy(f->kids[i]);
      else
	h->kids[i] = formula_copy(f->kids[i]->kids[odo[i]]);
    }
    if (simp_flag)
      g->kids[j] = simplify_top(h);
    else
      g->kids[j] = flatten_top(h);
    bump_odometer(odo, limit, arity);
  }

  if (arity > FSTACK_ARRAY_SIZE) {
    free(odo);
    free(limit);
  }
  zap_formula(f);
  if (simp_flag)
    return simplify_top(g);
  else
    return g;
}  /* distribute */

/*************
 *
 *   cnf_dnf()
 *
 *************/

static
Formula cnf_dnf(Formula f, Ftype target, BOOL simp_flag)
{
  if (f->type == ATOM_FORM || f->type == NOT_FORM)
    return f;
  else if (quant_form(f)) {
    f->kids[0] = cnf_dnf(f->kids[0], target, simp_flag);
    return f;
  }
  else {
    /* AND_FORM || OR_FORM */
    int i;
    BOOL target_kid = FALSE;
    for (i = 0; i < f->arity; i++) {
      f->kids[i] = cnf_dnf(f->kids[i], target, simp_flag);
      if (f->kids[i]->type == target)
	target_kid = TRUE;
    }
    f = flatten_top(f);
    if (f->type == target || !target_kid)
      return f;
    else
      return distribute(f, simp_flag);
  }
}  /* cnf_dnf */

/*************
 *
 *   cnf()
 *
 *************/

/* DOCUMENTATION
This routine transforms a formula into CNF (conjunctive normal form).
The argument simplify_flag tells whether to perform basic propositional
simplification (not-not, idempotence, $T, $F) along the way.
<P>
It is assumed that the given formula is in NNF (negation normal form)
(but it need not be flattened).
<P>
The result is flattened (w.r.t AND and OR), including all subformulas.
<P>
If the given formula is contradictory and simplify_flag is 1,
the result should be the empty disjunction, that is, false.
<P>
Do not refer to the given formula after the call; a good way to call
this routine is <TT>f = cnf(f, 1)</TT>.
*/

/* PUBLIC */
Formula cnf(Formula f, int simplify_flag)
{
  return cnf_dnf(f, AND_FORM, simplify_flag);
}  /* cnf */

/*************
 *
 *   dnf()
 *
 *************/

/* DOCUMENTATION
This routine transforms a formula into DNF (disjunctive normal form).
The argument simplify_flag tells whether to perform basic propositional
simplification (not-not, idempotence, $T, $F) along the way.
<P>
It is assumed that the given formula is in NNF (negation normal form)
(but it need not be flattened).
<P>
The result is flattened (w.r.t AND and OR), including all subformulas.
<P>
If the given formula is valid and simplify_flag is 1, the result
should be the empty conjunction, that is, true.
<P>
Do not refer to the given formula after the call; a good way to call
this routine is <TT>f = dnf(f, 1)</TT>.
*/

/* PUBLIC */
Formula dnf(Formula f, int simplify_flag)
{
  return cnf_dnf(f, OR_FORM, simplify_flag);
}  /* dnf */

/*************
 *
 *   subst_free_var()
 *
 *************/

static
void subst_free_var(Formula f, Term target, Term replacement)
{
  if (f->type == ATOM_FORM)
    f->atom = subst_term(f->atom, target, replacement);
  else if (quant_form(f) && str_ident(sn_to_str(SYMNUM(target)), f->qvar)) {
    ; /* Do nothing, because we have a quantified variable of the same name. */
  }
  else {
    int i;
    for (i = 0; i < f->arity; i++)
      subst_free_var(f->kids[i], target, replacement);
  }
}  /* subst_free_var */

/*************
 *
 *   skolem()
 *
 *************/

static
Formula skolem(Formula f, Term uvars)
{
  if (f->type == ATOM_FORM || f->type == NOT_FORM)
    return f;
  else if (f->type == ALL_FORM) {
    Term var = get_rigid_term(f->qvar, 0);
    Term lst;
    if (listterm_member(var, uvars)) {
      /* We are in the scope of another variable with this name, so
       * rename this variable.
       */
      int sn = fresh_symbol("x", 0);
      Term newvar = get_rigid_term(sn_to_str(sn), 0);
      subst_free_var(f->kids[0], var, newvar);
      f->qvar = sn_to_str(sn);
      free_term(var);
      var = newvar;
    }

    lst = listterm_cons(var, uvars);
    f->kids[0] = skolem(f->kids[0], lst);
    free_term(var);
    free_term(lst);
    return f;
  }
  else if (f->type == EXISTS_FORM) {
    Formula g;
    int n = listterm_length(uvars);
    int sn = next_skolem_symbol(n);
    Term sk = get_rigid_term(sn_to_str(sn), n);
    Term evar = get_rigid_term(f->qvar, 0);
    int i;

    for (i = 0; i < n; i++)
      sk->args[i] = copy_term(listterm_i(uvars, n-i)); /* uvars is backward */

    subst_free_var(f->kids[0], evar, sk);

    zap_term(sk);
    zap_term(evar);

    g = skolem(f->kids[0], uvars);
    free_formula(f);
    return g;
  }
  else if (f->type == AND_FORM || f->type == OR_FORM) {
    int i;
    for (i = 0; i < f->arity; i++) {
      f->kids[i] = skolem(f->kids[i], uvars);
    }
    return f;
  }
  else {
    /* Not in NNF!  Let the caller beware! */
    return f;
  }
}  /* skolem */

/*************
 *
 *   skolemize()
 *
 *************/

/* DOCUMENTATION
This routine Skolemizes an NNF formula.
The quantified variables need not be named in any particular way.
If there are universally quantified variables with the same name,
one in the scope of another, the inner variable will be renamed.
(Existential nodes are removed.)
*/

/* PUBLIC */
Formula skolemize(Formula f)
{
  Term uvars = get_nil_term();
  f = skolem(f, uvars);
  free_term(uvars);
  return f;
}  /* skolemize */

/*************
 *
 *   unique_qvars()
 *
 *************/

static
Term unique_qvars(Formula f, Term vars)
{
  if (f->type == ATOM_FORM)
    return vars;
  else if (quant_form(f)) {
    Term var = get_rigid_term(f->qvar, 0);
    Term lst;
    if (listterm_member(var, vars)) {
      /* Rename this variable. */
      int sn = fresh_symbol("x", 0);
      Term newvar = get_rigid_term(sn_to_str(sn), 0);
      subst_free_var(f->kids[0], var, newvar);
      f->qvar = sn_to_str(sn);
      free_term(var);
      var = newvar;
    }
    lst = listterm_cons(var, vars);
    return unique_qvars(f->kids[0], lst);
  }
  else {
    int i;
    for (i = 0; i < f->arity; i++) {
      vars = unique_qvars(f->kids[i], vars);
    }
    return vars;
  }
}  /* unique_qvars */

/*************
 *
 *   unique_quantified_vars()
 *
 *************/

/* DOCUMENTATION
Rename quantified variables, if necessary, so that each is unique.
This works for any formula.
<P>
If you wish to rename a quantified variable only if it occurs in
the scope of of a quantified variable with the same name, you
can use the routine eliminate_rebinding() instead.
<P>
(This could be a void routine, because
none of the formula nodes is changed; I made it return the Formula
so that it is consistent with its friends.)
*/

/* PUBLIC */
Formula unique_quantified_vars(Formula f)
{
  Term uvars = unique_qvars(f, get_nil_term());
  zap_term(uvars);
  return f;
}  /* unique_quantified_vars */

/*************
 *
 *   mark_free_vars_formula()
 *
 *************/

/* Replace all free occurrences of CONSTANT *varname with
 * a VARIABLE of index varnum.
 */

static
void mark_free_vars_formula(Formula f, char *varname, int varnum)
{
  if (f->type == ATOM_FORM)
    f->atom = subst_var_term(f->atom, str_to_sn(varname, 0), varnum);
  else if (quant_form(f) && str_ident(f->qvar, varname))
    return;
  else {
    int i;
    for (i = 0; i < f->arity; i++)
      mark_free_vars_formula(f->kids[i], varname, varnum);
  }
}  /* mark_free_vars_formula */

/*************
 *
 *   remove_uni_quant()
 *
 *************/

static
Formula remove_uni_quant(Formula f, int *varnum_ptr)
{
  if (f->type == AND_FORM || f->type == OR_FORM) {
    int i;
    for (i = 0; i < f->arity; i++)
      f->kids[i] = remove_uni_quant(f->kids[i], varnum_ptr);
    return f;
  }
  else if (f->type == ALL_FORM) {
    Formula g = f->kids[0];
    mark_free_vars_formula(g, f->qvar, *varnum_ptr);
    *varnum_ptr += 1;
    free_formula(f);
    return remove_uni_quant(g, varnum_ptr);
  }
  else {
    /* If not ATOM_FORM, something's probably wrong,
     * but let the caller beware!
     */
    return f;
  }
}  /* remove_uni_quant */

/*************
 *
 *   remove_universal_quantifiers()
 *
 *************/

/* DOCUMENTATION
For each universally quantified variable in the given formula,

*/

/* PUBLIC */
Formula remove_universal_quantifiers(Formula f)
{
  int varnum = 0;
  return remove_uni_quant(f, &varnum);
}  /* remove_universal_quantifiers */

/*************
 *
 *   clausify_prepare()
 *
 *************/

/* DOCUMENTATION
This routine gets a formula all ready for translation into clauses.
The sequence of transformations is
<UL>
<LI> change to negation normal form;
<LI> propositional simplification;
<LI> skolemize (nothing fancy here);
<LI> make the universally quantified variables unique;
<LI> remove universal quantifiers, changing the
     constants-which-represent-variables into genuine variables;
<LI> change to conjunctive normal form
     (with basic propositional simplification).
</UL>
The caller should not refer to the given formula f after the call;
A good way to call is <TT>f = clausify_prepare(f)</TT>
*/

/* PUBLIC */
Formula clausify_prepare(Formula f)
{
  f = nnf(f, CONJUNCTION);
  f = formula_simplify(f);
  f = skolemize(f);
  f = unique_quantified_vars(f);
  f = remove_universal_quantifiers(f);
  f = cnf(f, 1);
  return f;
}  /* clausify_prepare */

/*************
 *
 *   greatest_qvar()
 *
 *************/

/* DOCUMENTATION
Return the greatest SYMNUM of a quantified variable in Formula f.
<P>
Recall that in Formulas, a quantified variable is represented
as a constant (which is bound by the quantifier).
If the formula has no quantified variables, return -1.
*/

/* PUBLIC */
int greatest_qvar(Formula f)
{
  if (quant_form(f)) {
    int sn = str_to_sn(f->qvar, 0);
    int max_sub = greatest_qvar(f->kids[0]);
    return (sn > max_sub ? sn : max_sub);
  }
  else {
    int max = -1;
    int i;
    for (i = 0; i < f->arity; i++) {
      int max_sub = greatest_qvar(f->kids[i]);
      max = (max_sub > max ? max_sub : max);
    }
    return max;
  }
}  /* greatest_qvar */

/*************
 *
 *   greatest_symnum_in_formula()
 *
 *************/

/* DOCUMENTATION
Return the greatest SYMNUM of a any subterm.  This includes quantifed
variables that don't occur in any term.
<P>
This routine is intended to be used if you need malloc an array
for indexing by SYMNUM.
*/

/* PUBLIC */
int greatest_symnum_in_formula(Formula f)
{
  if (f->type == ATOM_FORM) {
    return greatest_symnum_in_term(f->atom);
  }
  if (quant_form(f)) {
    int sn = str_to_sn(f->qvar, 0);
    int max_sub = greatest_symnum_in_formula(f->kids[0]);
    return (sn > max_sub ? sn : max_sub);
  }
  else {
    int max = -1;
    int i;
    for (i = 0; i < f->arity; i++) {
      int max_sub = greatest_symnum_in_formula(f->kids[i]);
      max = (max_sub > max ? max_sub : max);
    }
    return max;
  }
}  /* greatest_symnum_in_formula */

/*************
 *
 *   elim_rebind()
 *
 *************/

static
Formula elim_rebind(Formula f, Term uvars)
{
  if (quant_form(f)) {
    Term var = get_rigid_term(f->qvar, 0);
    Term lst;
  
    if (listterm_member(var, uvars)) {
      /* We are in the scope of another variable with this name, so
       * rename this variable.
       */
      int sn = fresh_symbol("y", 0);
      Term newvar = get_rigid_term(sn_to_str(sn), 0);
      subst_free_var(f->kids[0], var, newvar);
      f->qvar = sn_to_str(sn);
      free_term(var);
      var = newvar;
    }

    lst = listterm_cons(var, uvars);
    f->kids[0] = elim_rebind(f->kids[0], lst);
    free_term(var);
    free_term(lst);
    return f;
  }
  else {
    int i;
    for (i = 0; i < f->arity; i++) {
      f->kids[i] = elim_rebind(f->kids[i], uvars);
    }
    return f;
  }
}  /* elim_rebind */

/*************
 *
 *   eliminate_rebinding()
 *
 *************/

/* DOCUMENTATION
This routine renames quantified variables so that
no quantified variable occurs in the scope of a quantified variable
with the same name.
<P>
If you wish to rename variables so that each
quantifer has a unique variable,
you can use the routine unique_quantified_vars() instead.
<P>
(This could be a void routine, because
none of the formula nodes is changed; I made it return the Formula
so that it is consistent with its friends.)
*/

/* PUBLIC */
Formula eliminate_rebinding(Formula f)
{
  Term uvars = get_nil_term();
  f = elim_rebind(f, uvars);
  free_term(uvars);
  return f;
}  /* eliminate_rebinding */
