#include "memory.h"

static int Malloc_megs=DEFAULT_MALLOC_MEGS; /* megs allocated by each malloc */
static int Max_megs=DEFAULT_MAX_MEGS;       /* max number of megs to malloc */

static char *Alloc_block=NULL;  /* location returned by most recent malloc */
static char *Alloc_pos;         /* current position in block */

static int Malloc_calls=0;      /* number of calls to malloc */

/*************
 *
 *    void *tp_alloc(n)
 *
 *    Allocate n contiguous bytes, aligned on pointer boundry.
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a block of at least n bytes of memory,
aligned so that it can be used for any kind of data.  (The size of the block
returned is a multiple of the size of a pointer.)  The way it works is
that a large block is obtained from malloc(), and this routine gives
out pieces until it has to get more from malloc(), and so on.
<P>
If the operating system runs out of memory, or the amount requested
is too great, or Max_megs is exceeded, the whole process just dies.
*/

/* PUBLIC */
void *tp_alloc(size_t n)
{
  if (n == 0)
    return NULL;
  else {

    char *return_block;
    int malloc_bytes = Malloc_megs*1024*1024;
    int psize;


    /* If n is not a multiple of sizeof(void *), round up so that it is */
  
    psize = sizeof(void *);
    if (n % psize != 0)
      n = n + (psize - (n % psize));

    /* If first call or not enough in the current block, get a new block. */
  
    if (Alloc_block==NULL || Alloc_block + malloc_bytes - Alloc_pos < n) {
      if (n > malloc_bytes)
	fatal_error("tp_alloc, request too big; see set_malloc_megs()");
      else if ((Malloc_calls+1)*Malloc_megs > Max_megs)
	fatal_error("tp_alloc, Max_megs parameter exceeded");
      else {
	Alloc_pos = Alloc_block = malloc((size_t) malloc_bytes);
	Malloc_calls++;
	if (!Alloc_pos)
	  fatal_error("tp_alloc, operating system is out of memory");
      }
    }
    return_block = Alloc_pos;
    Alloc_pos += n;
    return(return_block);
  }
}  /* tp_alloc */

/*************
 *
 *    void *tp_calloc(n)
 *
 *************/

/* DOCUMENTATION
This is just like tp_alloc(), except the memory is initialized to 0.
*/

/* PUBLIC */
void *tp_calloc(size_t n)
{
  int i;
  char *p = tp_alloc(n);

  for (i = 0; i < n; i++)
    p[i] = 0;
  return p;
}  /* tp_calloc */

/*************
 *
 *    int mbytes_malloced() -- How many K have been dynamically allocated?
 *
 *************/

/* DOCUMENTATION
This routine returns the number of kilobytes that tp_alloc()
has obtained from the operating system by malloc();
*/

/* PUBLIC */
int megs_malloced(void)
{
  return Malloc_calls * Malloc_megs;
}  /* mbytes_malloced */

/*************
 *
 *   set_max_megs()
 *
 *************/

/* DOCUMENTATION
This routine changes the limit on the amount of memory obtained
from malloc() by tp_alloc().  The argument is in megabytes.
The default value is DEFAULT_MAX_MEGS.
*/

/* PUBLIC */
void set_max_megs(int megs)
{
  Max_megs = megs;
}  /* set_max_megs */

/*************
 *
 *   set_malloc_megs()
 *
 *************/

/* DOCUMENTATION
This routine changes the size of the blocks obtained from malloc() by
tp_alloc().  The argument is in megabytes.  The default value is
DEFAULT_MALLOC_MEGS.
*/

/* PUBLIC */
void set_malloc_megs(int megs)
{
  Malloc_megs = megs;
}  /* set_malloc_megs */
