#include "clausify.h"

/* Private definitions and types */

/*************
 *
 *   formula_to_literal()
 *
 *************/

/* DOCUMENTATION
This routine takes a Formula f and returns a Literal representation.
If the formula is not an ATOM_FORM or the negation of an ATOM_FORM,
NULL is returned.
<P>
The returned literal is an entirely new copy,
and the given formula is not changed.
*/

/* PUBLIC */
Literal formula_to_literal(Formula f)
{
  if (f->type == NOT_FORM && f->kids[0]->type == ATOM_FORM)
    return new_literal(0, copy_term(f->kids[0]->atom));
  else if (f->type == ATOM_FORM)
    return new_literal(1, copy_term(f->atom));
  else
    return NULL;
}  /* formula_to_literal */

/*************
 *
 *   formula_to_clause()
 *
 *************/

/* DOCUMENTATION
This routine takes a Formula f and returns a Clause representation.
If f is not an atom, literal, or disjunction of literals, the
returned clause will be NULL or not well formed.
<P>
The returned clause is an entirely new copy, and the given formula
is not changed.
*/

/* PUBLIC */
Clause formula_to_clause(Formula f)
{
  Clause c = get_clause();
  if (f->type == ATOM_FORM || f->type == NOT_FORM)
    append_literal(c, formula_to_literal(f));
  else if (f->type == OR_FORM) {
    int i;
    for (i = 0; i < f->arity; i++)
      append_literal(c, formula_to_literal(f->kids[i]));
  }
  return c;
}  /* formula_to_clause */

/*************
 *
 *   formula_to_clauses()
 *
 *************/

/* DOCUMENTATION
This routine takes a Formula f and returns a Clist of Clauses
representation.
If f is not an atom, literal, or disjunction of literals, or a
conjunction of those things, the clauses in the returned list may be NULL
or not well formed.
<P>
The returned clauses are entirely new copies, and the given formula
is not changed.
*/

/* PUBLIC */
Clist formula_to_clauses(Formula f)
{
  Clist lst = clist_init(NULL);
  if (f->type == AND_FORM) {
    int i;
    for (i = 0; i < f->arity; i++)
      clist_append(formula_to_clause(f->kids[i]), lst);
  }
  else if (f->type == OR_FORM || f->type == NOT_FORM || f->type == ATOM_FORM)
    clist_append(formula_to_clause(f), lst);
  return lst;
}  /* formula_to_clauses */

/*************
 *
 *   clausify()
 *
 *************/

/* DOCUMENTATION
This routine translates a Formula f into a Clist of clauses.
The translation includes Skolemization, so the result should
be unsatisfiable iff f is unsatisfiable.  The variables in
each clause are renumbered.  If there are more than MAX_VARS
variables in a clause, a fatal error occurs.
<P>
Formula f is not changed
*/

/* PUBLIC */
Clist clausify(Formula f)
{
  Formula g;
  Clist lst;
  Clist_pos p;

  g = clausify_prepare(formula_copy(f));
  lst = formula_to_clauses(g);

  for (p = lst->first; p != NULL; p = p->next)
    renumber_variables(p->c, MAX_VARS);
  zap_formula(g);
  return lst;
}  /* clausify */

/*************
 *
 *   clausify_list()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Clist clausify_list(Plist g)
{
  Clist lst = clist_init(NULL);
  Plist p;
  for (p = g; p != NULL; p = p->next) {
    Clist l2 = clausify(p->v);
    clist_append_all(lst, l2);
  }
  return lst;
}  /* clausify_list */

