#include "clock.h"

/* Private definitions and types */

struct clock {
  char       *name;                 /* name of clock */
  unsigned   accum_msec;            /* accumulated time */
  unsigned   curr_msec;             /* time since clock has been turned on */
  int        level;                 /* STARTs - STOPs */
};

static struct clock *Clocks = NULL;  /* array of clocks */
static int Clocks_avail = 0;         /* current size of Clocks */
static int Clocks_in_use = 0;

static BOOL Clocks_enabled = TRUE;   /* getrusage can be slow */
static unsigned Clock_starts = 0;    /* Keep a count */

#define CLOCK_ARRAY_INCREMENT 100    /* in case Clocks is too small */

static int Wall_start;               /* for measuring wall-clock time */


#define CPU_TIME(msec) { struct rusage r; getrusage(RUSAGE_SELF, &r);\
       msec = r.ru_utime.tv_sec * 1000 + r.ru_utime.tv_usec / 1000; }

/*************
 *
 *   expand_clock_array()
 *
 *************/

static
void expand_clock_array()
{
  int new_size = Clocks_avail + CLOCK_ARRAY_INCREMENT;
  struct clock *tmp;
  int i;

  /* Replace the old array with a new and bigger one. */

  tmp = malloc(new_size * sizeof(struct clock));
  for (i = 0; i < Clocks_avail; i++) {
    tmp[i].name = Clocks[i].name;
    tmp[i].accum_msec = Clocks[i].accum_msec;
    tmp[i].curr_msec = Clocks[i].curr_msec;
    tmp[i].level = Clocks[i].level;
  }
  if (Clocks != NULL)
    free(Clocks);
  Clocks = tmp;
  Clocks_avail = new_size;
}  /* expand_clock_array */

/*************
 *
 *   clock_init()
 *
 *************/

/* DOCUMENTATION
This routine initializes a clock.  You give it a string (any length,
which is copied), representing the name of the new clock, and it
returns a nonnegative integer ID which is to be used for all of the
clock operations.  There is no limit on the number of clocks you can
have.  If you already have a clock with the given name, -1 is
returned.
<P>
The clock operations are clock_start(), clock_stop(), clock_value(),
and clock_reset().
*/

/* PUBLIC */
int clock_init(char *str)
{
  int i;
  BOOL already_here;

  for (i = 0, already_here = FALSE; i < Clocks_in_use && !already_here; i++) {
    if (strcmp(str, Clocks[i].name) == 0)
      already_here = TRUE;
  }
  if (already_here)
    return -1;
  else {
    int n;
    
    if (Clocks_in_use == Clocks_avail)
      expand_clock_array();

    n = Clocks_in_use;
    Clocks_in_use++;
    Clocks[n].name = new_str_copy(str);;
    Clocks[n].level = 0;
    Clocks[n].accum_msec = 0;
    return n;
  }
}  /* clock_init */

/*************
 *
 *   clock_start()
 *
 *************/

/* DOCUMENTATION
This routine starts clock n.  It is okay if the clock is already going.
*/

/* PUBLIC */
void clock_start(int n)
{
  if (Clocks_enabled) {
    struct clock *cp = &Clocks[n];
    cp->level++;
    if (cp->level == 1) {
      CPU_TIME(cp->curr_msec);
      Clock_starts++;
    }
  }
}  /* clock_start */

/*************
 *
 *   clock_stop()
 *
 *************/

/* DOCUMENTATION
This routine stops clock n and adds the time to the accumulated total,
<I>unless there have been too many starts and not enough stops</I>.
See the introduction.
*/

/* PUBLIC */
void clock_stop(int n)
{
  if (Clocks_enabled) {
    struct clock *cp = &Clocks[n];
    if (cp->level <= 0)
      fprintf(stderr, "WARNING, clock_stop: clock %d not running.\n", n); 
    else {
      cp->level--;
      if (cp->level == 0) {
	unsigned msec;
	CPU_TIME(msec);
	cp->accum_msec += msec - cp->curr_msec;
      }
    }
  }
}  /* clock_stop */

/*************
 *
 *   clock_value()
 *
 *************/

/* DOCUMENTATION
This routine returns the current value of a clock.  The value
is in milliseconds.  The clock need not be stopped.
*/

/* PUBLIC */
unsigned clock_value(int n)
{
  struct clock *cp = &Clocks[n];
  int i = cp->accum_msec;

  if (cp->level == 0)
    return i;
  else {
    int msec;
    CPU_TIME(msec);
    return i + (msec - cp->curr_msec);
  }
}  /* clock_value */

/*************
 *
 *   clock_running()
 *
 *************/

/* DOCUMENTATION
This routine tells you whether or not a clock is running.
*/

/* PUBLIC */
BOOL clock_running(int n)
{
  struct clock *cp = &Clocks[n];
  return cp->level > 0;
}  /* clock_running */

/*************
 *
 *   clock_reset()
 *
 *************/

/* DOCUMENTATION
This routine resets a clock, as if it had just been initialized.
(You shouldn't need this routine under normal circumstances.)
*/

/* PUBLIC */
void clock_reset(int n)
{
  Clocks[n].level = 0;
  Clocks[n].accum_msec = 0;
}  /* clock_reset */

/*************
 *
 *   fprint_clock()
 *
 *************/

/* DOCUMENTATION
This routine
*/

/* PUBLIC */
void fprint_clock(FILE *fp, int n)
{
  if (n < 0 || n >= Clocks_in_use)
    fprintf(fp, "Clock %d is out of range.\n", n);
  else
    fprintf(fp, "clock %-15s: %.2f seconds.\n",
	    Clocks[n].name, clock_value(n) / 1000.0);
}  /* fprint_clock */

/*************
 *
 *   fprint_all_clocks()
 *
 *************/

/* DOCUMENTATION
This routine
*/

/* PUBLIC */
void fprint_all_clocks(FILE *fp)
{
  int i;
  fprintf(fp, "All clocks (%u clock_starts):\n", Clock_starts);
  for (i = 0; i < Clocks_in_use; i++)
    fprint_clock(fp, i);
  fflush(fp);
}  /* fprint_all_clocks */

/*************
 *
 *   get_date()
 *
 *************/

/* DOCUMENTATION
This routine returns a string representation of the current date and time.
*/

/* PUBLIC */
char * get_date(void)
{
  time_t i = time((time_t *) NULL);
  return asctime(localtime(&i));
}  /* get_date */

/*************
 *
 *   user_time()
 *
 *************/

/* DOCUMENTATION
This routine returns the user CPU time, in milliseconds, that the
current process has used so far.
*/

/* PUBLIC */
unsigned user_time()
{
  struct rusage r;
  unsigned sec, usec;

  getrusage(RUSAGE_SELF, &r);
  sec = r.ru_utime.tv_sec;
  usec = r.ru_utime.tv_usec;

  return((sec * 1000) + (usec / 1000));
}  /* user_time */

/*************
 *
 *   system_time()
 *
 *************/

/* DOCUMENTATION
This routine returns the system CPU time, in milliseconds, that has
been spent on the current process.
(System time measures low-level operations such
as system calls, paging, and I/O that the operating systems does
on behalf of the process.)
*/

/* PUBLIC */
unsigned system_time()
{
  struct rusage r;
  unsigned sec, usec;

  getrusage(RUSAGE_SELF, &r);
  sec = r.ru_stime.tv_sec;
  usec = r.ru_stime.tv_usec;

  return((sec * 1000) + (usec / 1000));
}  /* system_time */

/*************
 *
 *   wall_init()
 *
 *************/

/* DOCUMENTATION
This routine initializes the wall-clock timer.
*/

/* PUBLIC */
void wall_init()
{
  Wall_start = time((time_t *) NULL);
}  /* wall_init */

/*************
 *
 *   wall_time()
 *
 *************/

/* DOCUMENTATION
This routine returns the number of wall-clock seconds since wall_init()
was called.
*/

/* PUBLIC */
int wall_time()
{
  time_t i;
  i = time((time_t *) NULL);
  return (int) (i - Wall_start);
}  /* wall_time */

/*************
 *
 *   disable_clocks()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void disable_clocks(void)
{
  Clocks_enabled = FALSE;
}  /* disable_clocks */

/*************
 *
 *   enable_clocks()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void enable_clocks(void)
{
  Clocks_enabled = TRUE;
}  /* enable_clocks */

