#ifndef TP_DISCRIM_H
#define TP_DISCRIM_H

#include "glist.h"
#include "unify.h"
#include "index.h"

/* INTRODUCTION
This package implements two kinds of discrimination indexing
for first-order terms.  Both kinds support GENERALIZATION retrieval
only (e.g., for forward demodulation and forward subsumption).
<P>
The "wild" kind is an imperfect filter, and it does not bind variables.
The caller must also call a routine, say match(), to check if
the answers are really more general than the query term and to
construct the substitution.
Wild indexing supports associative-commutative (AC) and
commutative/symmetric (C) symbols.
Indexing terms with AC symbols works by considering the number
of arguments and the number of nonvariable arguments of 
AC terms that do not occur in other AC terms.
(The term "wild" is used because all variables in the discrimination
tree are treated as the the wildcard symbol *).
<P>
With the "tame" kind, every answer is more general than the query term,
and the matching substitution is constructed during the retrieval.
Wild indexing supports commutative/summetric (C) symbols,
but it <I>does not support</I> associative-commutative (AC) symbols.
Retrieval with C symbols can produce duplicate answers.
(The term "tame" is used because the indexing is not wild.)
<P>
Note that some of the routines in this package (the ones without
"wild" or "tame" in their names) are used with either wild or tame indexing.
<P>
There is probably a higher-level package (mindex ?) which
provides a uniform interface to these and other indexing methods.
*/
/* Public definitions */

typedef struct discrim * Discrim;
typedef struct discrim_pos * Discrim_pos;

/* End of public definitions */

/* Public function prototypes from discrim.c */

void fprint_discrim_mem(FILE *fp, BOOL heading);

void p_discrim_mem(void);

Discrim discrim_init(void);

void discrim_dealloc(Discrim d);

void fprint_discrim_tame_index(FILE *fp, Discrim d);

void p_discrim_tame_index(Discrim d);

void discrim_tame_update(Term t, Discrim root, void *object, Indexop op);

void *discrim_tame_retrieve_first(Term t, Discrim root,
			     Context subst, Discrim_pos *ppos);

void *discrim_tame_retrieve_next(Discrim_pos pos);

void discrim_tame_cancel(Discrim_pos pos);

void fprint_discrim_wild_index(FILE *fp, Discrim d);

void p_discrim_wild_index(Discrim d);

void discrim_wild_update(Term t, Discrim root, void *object, Indexop op);

void *discrim_wild_retrieve_first(Term t, Discrim root,
				  Discrim_pos *ppos);

void *discrim_wild_retrieve_next(Discrim_pos pos);

void discrim_wild_cancel(Discrim_pos pos);

void destroy_discrim_tree(Discrim d);

BOOL discrim_empty(Discrim d);

#endif  /* conditional compilation of whole file */
