#ifndef TP_FPA_H
#define TP_FPA_H

#include "glist.h"
#include "unify.h"
#include "index.h"

/*
We use a trick to save memory.  We need to be able to order terms, and
the trick is that we use their addresses.  This is risky, because,
according to the C definition, two pointers can be compared with < and
> only if they are members of the same array.

I suspect it's ok to do this on all 32-bit architectures/compilers,
but you never know.  (On some old C compilers in DOS, you can have
two pointers A and B, where none of the following holds: A<B, A>B, A==B,
if A and B are in different memory segments.)

If you wish to be safe (or to check if this trick is the problem when
something goes wrong) you can go to term.h and "#define TERM_ID_FIELD".
This will add an ID field to terms, and that field will be given a
unique id.  This FPA package will then use those fields instead of the
addresses.  The results should be exactly the same, because the
two methods give the same order on terms.
*/

#ifdef TERM_ID_FIELD
#  define FPA_ID_TYPE unsigned
#  define FPA_ID_MAX UINT_MAX
#  define FPA_ID(t) (((Term) t)->id)
#else
#  define FPA_ID_TYPE Term
#  define FPA_ID_MAX ((Term) ULONG_MAX) /* a pointer bigger than all others? */
#  define FPA_ID(t) ((Term) t)
#endif

/* INTRODUCTION
This package implements FPA/Path indexing for first-order terms.
It is used to retrieive terms that are likely to unify (in various
ways) with a query term.  The answers are not guaranteed to unify,
and the caller must call some kind of unification routine to
check and to construct a substitution.
<P>
The unification types are
FPA_UNIFY, FPA_INSTANCE, FPA_GENERALIZATION, FPA_VARIANT, and FPA_IDENTICAL.
Performance is very good for
FPA_INSTANCE, FPA_VARIANT, and FPA_IDENTICAL,
fair for FPA_UNIFY, and
poor for FPA_GENERALIZATION.
(Use Discrimination indexing for FPA_GENERALIZATION, e.g.,
forward demodulation and forward subsumption.)
<P>
Associative-commutative (AC) function symbols are handled
by simply ignoring all subterms of AC symbols (which can give
bad performance).
Commutative/symmetric (C) operations are handled.
*/

/* Public definitions */

/* #define FPA_DEBUG */

typedef struct fpa_index * Fpa_index;
typedef struct fpa_state * Fpa_state;

/* End of public definitions */

/* Public function prototypes from fpa.c */

void fprint_fpa_mem(FILE *fp, BOOL heading);

void p_fpa_mem(void);

void fprint_path(FILE *fp, Ilist p);

void p_path(Ilist p);

void fprint_fpa_index(FILE *fp, Fpa_index idx);

void p_fpa_index(Fpa_index idx);

Fpa_index fpa_init_index(int depth);

void fpa_update(Term t, Fpa_index idx, Indexop op);

void fprint_fpa_state(FILE *fp, Fpa_state q, int depth);

void p_fpa_state(Fpa_state q);

void p_fpa_query(Term t, Querytype query_type, Fpa_index idx);

Term fpa_next_answer(Fpa_state q);

Term fpa_first_answer(Term t, Context c, Querytype query_type,
		      Fpa_index idx, Fpa_state *ppos);

void fpa_cancel(Fpa_state q);

void zap_fpa_index(Fpa_index idx);

BOOL fpa_empty(Fpa_index idx);

#endif  /* conditional compilation of whole file */
