#ifndef FSOLO
#    include "header.h"
#else /* stand-alone debug mode */
#    include <stdio.h>
#    include <stdlib.h>
#    include <time.h>
#    include "fpa2.h"
#    include "fpa2.h"
#endif

/*
 * memory management
 */

static Fnode fnode_avail;
static unsigned fnode_gets, fnode_frees, fnode_avails;

/* end of counters for each type */

/*************
 *
 *    Fnode get_fnode()
 *
 *************/

static
Fnode get_fnode(void)
{
  Fnode p;

  fnode_gets++;
  if (fnode_avail == NULL)
    p = (void *) tp_alloc(sizeof(struct fnode));
  else {
    fnode_avails--;
    p = fnode_avail;
    fnode_avail = fnode_avail->next;
  }
  /* initialization */	
  if (0) {
    int i;
    for (i = 0; i < FMAX; i++)
      p->d[i] = FDEFAULT;
  }
  p->next = NULL;
  /* end of initialization */	
  return(p);
}  /* get_fnode */

/*************
 *
 *    free_fnode()
 *
 *************/

static
void free_fnode(Fnode p)
{
  fnode_frees++;
  fnode_avails++;
  p->next = fnode_avail;
  fnode_avail = p;
}  /* free_fnode */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_flist_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for ??
and associated structures.  The argument heading tells whether
to print a heading on the table.
*/

/* PUBLIC */
void fprint_flist_mem(FILE *fp, BOOLEAN heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct fnode);
  fprintf(fp, "fnode (%4d)        %11u%11u%11u%11u%9.1f K\n",
	  n, fnode_gets, fnode_frees,
	  fnode_gets - fnode_frees, fnode_avails,
	  (((fnode_gets - fnode_frees) + fnode_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_flist_mem */

/*************
 *
 *   p_flist_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for String_bufs
and associated structures.
*/

/* PUBLIC */
void p_flist_mem()
{
  fprint_flist_mem(stdout, 1);
}  /* p_flist_mem */

/*
 *  end of memory management
 */

/*************************************************************************/

/*
  Now we get to the interesging stuff.  (Most of the preceding
  could be moved to av.c.

  The design is determined by the following properties of the
  application:  (1) items will nearly always be inserted in
  increasing order, (2) the lists will be traversed, and the
  items must be kept in decreasing order, and (3) deletions
  will be arbitrary and occasionally extensive.

/* First and last items in chunk.  Items are right-justified. */

#define FLAST(f) (f)->d[FMAX-1]
#define FFIRST(f) (f)->d[FMAX-((f)->n)]

/*************
 *
 *   flist_insert()
 *
 *   If the item is greater than any in the list, insertion should
 *   be constant time.
 *
 *************/

Fnode flist_insert(Fnode f, FTYPE x)
{
  if (f == NULL) {
    f = get_fnode();
    FLAST(f) = x;
    f->n = 1;
  }
  else if (f->n == FMAX) {
    if (FLT(x,FLAST(f)))
      f->next = flist_insert(f->next, x);
    else if (FEQ(x,FLAST(f)))
      fprintf(stderr, "WARNING: flist_insert, item %d already here (1)!\n", x);
    else if (FGT(x,FFIRST(f))) {
      /* 
	 This special case isn't necessary.  It is to improve performance.
	 The application for which I'm writing this inserts items in
	 increasing order (most of the time), and this prevents a lot of
	 half-empty nodes in that case.
      */
      Fnode f2 = flist_insert(NULL, x);
      f2->next = f;
      f = f2;
    }
    else {
      /* split this node in half */
      Fnode f2 = get_fnode();
      int move = FMAX / 2;
      int i, j;
      for (i = 0, j = FMAX-move; i < move; i++, j++) {
	f2->d[j] = f->d[i];
	f->d[i] = FDEFAULT;
      }
      f2->n = move;
      f->n = FMAX - move;
      f2->next = f;
      f = flist_insert(f2, x);
    }
  }
  else {
    if (f->next && FLE(x,FFIRST(f->next)))
      f->next = flist_insert(f->next, x);
    else {
      /* insert into this node */
      int n = f->n;
      int i = FMAX - n;
      while (i < FMAX && FLT(x,f->d[i]))
	i++;
      if (i < FMAX && FEQ(x,f->d[i]))
	fprintf(stderr, "WARNING: flist_insert, item %d already here (2)!\n", x);
      else if (i == FMAX - n) {
	f->d[i-1] = x;
	f->n = n+1;
      }
      else {
	/* insert at i-1, shifting the rest */
	int j;
	for (j = FMAX-n; j < i; j++)
	  f->d[j-1] = f->d[j];
	f->d[i-1] = x;
	f->n = n+1;
      }
    }
  }
  return f;
}  /* flist_insert */

/*************
 *
 *   consolidate() - try to join f and f->next; not recursive
 *
 *************/

static
Fnode consolidate(Fnode f)
{
  if (f->next && f->n + f->next->n <= FMAX) {
    Fnode f2 = f->next;
    int i;
    for (i = 0; i < f->n; i++)
      f2->d[FMAX - (f2->n + i + 1)] = f->d[FMAX - (i+1)];
    f2->n = f->n + f2->n;
    free_fnode(f);
    return f2;
  }
  else
    return f;
}  /* consolidate */

/*************
 *
 *   flist_delete()
 *
 *************/

Fnode flist_delete(Fnode f, FTYPE x)
{
  if (f == NULL)
    fprintf(stderr, "WARNING: flist_delete, item %d not found (1)!\n", x);
  else if (FLT(x,FLAST(f)))
    f->next = flist_delete(f->next, x);
  else {
    int n = f->n;
    int i = FMAX - n;
    while (i < FMAX && FLT(x,f->d[i]))
      i++;
    if (FNE(x,f->d[i]))
      fprintf(stderr, "WARNING: flist_delete, item %d not found (2)!\n", x);
    else {
      /* delete and close the hole */
      int j;
      for (j = i; j > FMAX-n; j--)
	f->d[j] = f->d[j-1];
      f->d[j] = FDEFAULT;
      f->n = n-1;
      if (f->n == 0) {
	/* delete this node */
	Fnode next = f->next;
	free_fnode(f);
	f = next;
      }
      else {
	/* try to join this node with the next */
	f = consolidate(f);
      }
    }
  }
  return f;
}  /* flist_delete */

/*************
 *
 *   first_fpos()
 *
 *************/

struct fposition first_fpos(Fnode f)
{
  if (f == NULL)
    return (struct fposition) {NULL, 0};
  else
    return (struct fposition) {f, FMAX - f->n};
}  /* first_fpos */

/*************
 *
 *   next_fpos()
 *
 *************/

struct fposition next_fpos(struct fposition p)
{
  int i = p.i+1;
  if (i < FMAX)
    return (struct fposition) {p.f, i};
  else
    return first_fpos((p.f)->next);
}  /* next_fpos */

/*****************************************************************************/
/* The rest of the file is for stand-alone debugging mode only               */
/*****************************************************************************/

#ifdef FSOLO

void p_flist(Fnode f)
{
  Fnode p;
  for (p = f; p != NULL; p = p->next) {
    int i;
    printf("fnode %p, n=%d:", p, p->n);
    for (i = 0; i < FMAX; i++) {
      printf("\t%d", p->d[i]);
    }
    printf("\n");
  }
}  /* p_flist */

int flist_population(Fnode f)
{
  if (f == NULL)
    return 0;
  else
    return f->n + flist_population(f->next);
}  /* flist_population */

int user_time()
{
  return (clock() / (CLOCKS_PER_SEC / 1000));   /* return milliseconds */
}  /* user_time */

void *tp_alloc(size_t size)
{
  return malloc(size);
}  /* tp_alloc */

void random_permutation(int *a, int size)
{
  int n, i, x;

  for (i = 0; i < size; i++)
    a[i] = -1;

  n = 0;
  while (n < size) {
    x = rand() % size;
    if (a[x] == -1)
      a[x] = n++;
    else {
      i = x+1;
      while (i < size && a[i] != -1)
        i++;
      if (i < size)
        a[i] = n++;
      else {
        i = x-1;
        while (i >= 0 && a[i] != -1)
          i--;
        a[i] = n++;
      }
    }
  }
}  /* random_permutation */

/*************
 *
 *   main()
 *
 *************/

#define N 1000  /* objects to insert */
#define R (N*10)   /* range of objects is [1 .. R] */

FTYPE Record[N];   /* record the objects inserted */

int main(int argc, char **argv)
{
  int perm[N];
  Fnode f = NULL;
  int i = 0;
  float t1, t2;

  if (argc == 2) {
    int seed = atoi(argv[1]);
    srand(seed);
  }

  while (i < N) {
    // FTYPE r = (FTYPE) (1 + (rand() % R));       // random
    FTYPE r = (FTYPE) i;                      // increasing
    // FTYPE r = (FTYPE) ((N-1) - i);              // decreasing

    f = flist_insert(f, r);
    // printf("\ninserted %d\n", r);
    // p_flist(f);
    Record[i] = r;
    i++;
  }

  p_flist(f);
  t1 = user_time() / 1000.0;
  printf("FMAX=%d, insert time: %.2f, chunks=%d\n", FMAX, t1, fnode_gets-fnode_frees);

  // printf("density: %.2f\n",
  //       flist_population(f) / (float) ((fnode_gets-fnode_frees) * FMAX));

  random_permutation(perm, N);

  for (i = 0; i < N; i++) {
    int j = perm[i];       // random
    // int j = (N-1) - i;     // reverse order from inserted
    // int j = i;             // same order as inserted

    FTYPE r = Record[j];

    f = flist_delete(f, r);
    // printf("\ndeleted %d\n", r);
    // p_flist(f);
  }

  t2 = user_time() / 1000.0;
  printf("FMAX=%d, delete time: %.2f\n", FMAX, t2 - t1);

  p_flist_mem();
  exit(0);
}  /* main */

#endif  /* FSOLO */
