#include "fpalist.h"

/* Private definitions and types */

/*
 * memory management
 */

/* end of counters for each type */

static Fpa_chunk fpa_chunk_avail;
static unsigned fpa_chunk_gets, fpa_chunk_frees, fpa_chunk_avails;

/*************
 *
 *    Fpa_chunk get_fpa_chunk()
 *
 *************/

static
Fpa_chunk get_fpa_chunk(void)
{
  Fpa_chunk p;

  fpa_chunk_gets++;
  if (fpa_chunk_avail == NULL)
    p = (void *) tp_alloc(sizeof(struct fpa_chunk));
  else {
    fpa_chunk_avails--;
    p = fpa_chunk_avail;
    fpa_chunk_avail = fpa_chunk_avail->next;
  }
  /* initialization */	
  if (0) {  /* this is not necessary */
    int i;
    for (i = 0; i < FMAX; i++)
      p->d[i] = FDEFAULT;
  }
  p->next = NULL;
  /* end of initialization */	
  return(p);
}  /* get_fpa_chunk */

/*************
 *
 *    free_fpa_chunk()
 *
 *************/

static
void free_fpa_chunk(Fpa_chunk p)
{
  fpa_chunk_frees++;
  fpa_chunk_avails++;
  p->next = fpa_chunk_avail;
  fpa_chunk_avail = p;
}  /* free_fpa_chunk */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_fpalist_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with ??.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_fpalist_mem(FILE *fp, int heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct fpa_chunk);
  fprintf(fp, "fpa_chunk (%4d)     %11u%11u%11u%11u%9.1f K\n",
	  n, fpa_chunk_gets, fpa_chunk_frees, fpa_chunk_gets - fpa_chunk_frees,
	  fpa_chunk_avails,
	  (((fpa_chunk_gets - fpa_chunk_frees) + fpa_chunk_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_fpalist_mem */

/*************
 *
 *   p_fpalist_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with ??.
*/

/* PUBLIC */
void p_fpalist_mem()
{
  fprint_fpalist_mem(stdout, 1);
}  /* p_fpalist_mem */

/*
 *  end of memory management
 */

/*
  This if the new code (April 2004) for inserting/deleting/traversing
   FPA lists.  It should function exactly the same as the old code,
   except that deletions should be much faster, and memory usage will
   be somewhat different (not much).

   Instead of a (singly-linked) list of pointers to terms, we have a
   (singly-linked) list of *arrays* of pointers to terms.  As before,
   the terms are kept in decreasing order.  Recall that in practice,
   terms being inserted will usually be greater than anything already
   in the list.
*/

/*
  The design is determined by the following properties of the
  application:  (1) items will nearly always be inserted in
  increasing order, (2) the lists will be traversed, and the
  items must be kept in decreasing order, and (3) deletions
  will be arbitrary and occasionally extensive.
*/

/* First and last items in chunk.  Items are right-justified. */

#define FLAST(f) (f)->d[FMAX-1]
#define FFIRST(f) (f)->d[FMAX-((f)->n)]

/*************
 *
 *   flist_insert()
 *
 *   If the item is greater than any in the list, insertion should
 *   be constant time.
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Fpa_chunk flist_insert(Fpa_chunk f, FTYPE x)
{
  if (f == NULL) {
    f = get_fpa_chunk();
    FLAST(f) = x;
    f->n = 1;
  }
  else if (f->n == FMAX) {
    if (FLT(x,FLAST(f)))
      f->next = flist_insert(f->next, x);
    else if (FEQ(x,FLAST(f)))
      fprintf(stderr, "WARNING: flist_insert, item %d already here (1)!\n", x);
    else if (FGT(x,FFIRST(f))) {
      /* 
	 This special case isn't necessary.  It is to improve performance.
	 The application for which I'm writing this inserts items in
	 increasing order (most of the time), and this prevents a lot of
	 half-empty chunks in that case.
      */
      Fpa_chunk f2 = flist_insert(NULL, x);
      f2->next = f;
      f = f2;
    }
    else {
      /* split this chunk in half */
      Fpa_chunk f2 = get_fpa_chunk();
      int move = FMAX / 2;
      int i, j;
      for (i = 0, j = FMAX-move; i < move; i++, j++) {
	f2->d[j] = f->d[i];
	f->d[i] = FDEFAULT;
      }
      f2->n = move;
      f->n = FMAX - move;
      f2->next = f;
      f = flist_insert(f2, x);
    }
  }
  else {
    if (f->next && FLE(x,FFIRST(f->next)))
      f->next = flist_insert(f->next, x);
    else {
      /* insert into this pa_chunk */
      int n = f->n;
      int i = FMAX - n;
      while (i < FMAX && FLT(x,f->d[i]))
	i++;
      if (i < FMAX && FEQ(x,f->d[i]))
	fprintf(stderr, "WARNING: flist_insert, item %d already here (2)!\n", x);
      else if (i == FMAX - n) {
	f->d[i-1] = x;
	f->n = n+1;
      }
      else {
	/* insert at i-1, shifting the rest */
	int j;
	for (j = FMAX-n; j < i; j++)
	  f->d[j-1] = f->d[j];
	f->d[i-1] = x;
	f->n = n+1;
      }
    }
  }
  return f;
}  /* flist_insert */

/*************
 *
 *   consolidate() - try to join f and f->next; not recursive
 *
 *************/

static
Fpa_chunk consolidate(Fpa_chunk f)
{
  if (f->next && f->n + f->next->n <= FMAX) {
    Fpa_chunk f2 = f->next;
    int i;
    for (i = 0; i < f->n; i++)
      f2->d[FMAX - (f2->n + i + 1)] = f->d[FMAX - (i+1)];
    f2->n = f->n + f2->n;
    free_fpa_chunk(f);
    return f2;
  }
  else
    return f;
}  /* consolidate */

/*************
 *
 *   flist_delete()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Fpa_chunk flist_delete(Fpa_chunk f, FTYPE x)
{
  if (f == NULL)
    fprintf(stderr, "WARNING: flist_delete, item %d not found (1)!\n", x);
  else if (FLT(x,FLAST(f)))
    f->next = flist_delete(f->next, x);
  else {
    int n = f->n;
    int i = FMAX - n;
    while (i < FMAX && FLT(x,f->d[i]))
      i++;
    if (FNE(x,f->d[i]))
      fprintf(stderr, "WARNING: flist_delete, item %d not found (2)!\n", x);
    else {
      /* delete and close the hole */
      int j;
      for (j = i; j > FMAX-n; j--)
	f->d[j] = f->d[j-1];
      f->d[j] = FDEFAULT;
      f->n = n-1;
      if (f->n == 0) {
	/* delete this chunk */
	Fpa_chunk next = f->next;
	free_fpa_chunk(f);
	f = next;
      }
      else {
	/* try to join this chunk with the next */
	f = consolidate(f);
      }
    }
  }
  return f;
}  /* flist_delete */

/*************
 *
 *   first_fpos()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
struct fposition first_fpos(Fpa_chunk f)
{
  if (f == NULL)
    return (struct fposition) {NULL, 0};
  else
    return (struct fposition) {f, FMAX - f->n};
}  /* first_fpos */

/*************
 *
 *   next_fpos()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
struct fposition next_fpos(struct fposition p)
{
  int i = p.i+1;
  if (i < FMAX)
    return (struct fposition) {p.f, i};
  else
    return first_fpos((p.f)->next);
}  /* next_fpos */

/*************
 *
 *   zap_fpa_chunks()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void zap_fpa_chunks(Fpa_chunk p)
{
  if (p != NULL) {
    zap_fpa_chunks(p->next);
    free_fpa_chunk(p);
  }
}  /* zap_fpa_chunks */
