#ifndef TP_FPALIST_H
#define TP_FPALIST_H

#include "term.h"

/* INTRODUCTION
*/

/*
We use a trick to save memory.  We need to be able to order terms, and
the trick is that we use their addresses.  This is risky, because,
according to the C definition, two pointers can be compared with < and
> only if they are members of the same array.

I suspect it's ok to do this on all 32-bit architectures/compilers,
but you never know.  (On some old C compilers in DOS, you can have
two pointers A and B, where none of the following holds: A<B, A>B, A==B,
if A and B are in different memory segments.)

If you wish to be safe (or to check if this trick is the problem when
something goes wrong) you can go to term.h and "#define TERM_ID_FIELD".
This will add an ID field to terms, and that field will be given a
unique id.  This FPA package will then use those fields instead of the
addresses.  The results should be exactly the same, because the
two methods give the same order on terms.
*/

#ifdef TERM_ID_FIELD
#  define FPA_ID_TYPE unsigned
#  define FPA_ID_MAX UINT_MAX
#  define FPA_ID(t) (((Term) t)->id)
#else
#  define FPA_ID_TYPE Term
#  define FPA_ID_MAX ((Term) ULONG_MAX) /* a pointer bigger than all others? */
#  define FPA_ID(t) ((Term) t)
#endif

/* Public definitions */

/* The objects in FPA lists are pointers to terms, and the lists
   are ordered (decreasing) by fpa_id field in terms.

   The code was originally written and debugged with integers instead;
   hence, the following, which makes it easy to change back to ints.
*/

#define FTYPE Term        /* type of items in chunk */
#define FDEFAULT NULL     /* default value of FTYPE */

#define FLT(x,y) (FPA_ID(x) <  FPA_ID(y))
#define FGT(x,y) (FPA_ID(x) >  FPA_ID(y))
#define FLE(x,y) (FPA_ID(x) <= FPA_ID(y))
#define FGE(x,y) (FPA_ID(x) >= FPA_ID(y))
#define FEQ(x,y) ((x) == (y))
#define FNE(x,y) ((x) != (y))

#define FTERM(p) ((p).f == NULL ? NULL : (p).f->d[(p).i])

/* A chunk of an FPA list */

#define FMAX 400  /* maximum number of items per chunk */

typedef struct fpa_chunk *Fpa_chunk;

struct fpa_chunk {
  FTYPE d[FMAX];    /* array for chunk */
  int n;            /* current size of chunk (right justified in array) */
  Fpa_chunk next;       /* list of chunks is singly-linked */
};

/* to maintain a position in an FPA list while traversing for set operations */

struct fposition {
  Fpa_chunk f;
  int i;
};

/* End of public definitions */

/* Public function prototypes from fpalist.c */

void fprint_fpalist_mem(FILE *fp, int heading);

void p_fpalist_mem();

Fpa_chunk flist_insert(Fpa_chunk f, FTYPE x);

Fpa_chunk flist_delete(Fpa_chunk f, FTYPE x);

struct fposition first_fpos(Fpa_chunk f);

struct fposition next_fpos(struct fposition p);

void zap_fpa_chunks(Fpa_chunk p);

#endif  /* conditional compilation of whole file */
