#include "hash.h"

/* Private definitions and types */

struct hashtab {
  int      size;
  Hashnode *table;
  Hashtab  next;  /* for avail list only */
};

struct hashnode {
  int      id;
  void     *v;
  Hashnode next;
};

/*
 * memory management
 */

static Hashtab hashtab_avail;
static unsigned hashtab_gets, hashtab_frees, hashtab_avails;

static Hashnode hashnode_avail;
static unsigned hashnode_gets, hashnode_frees, hashnode_avails;

/* end of counters for each type */

/*************
 *
 *    Hashtab get_hashtab()
 *
 *************/

static
Hashtab get_hashtab(void)
{
  Hashtab p;

  hashtab_gets++;
  if (hashtab_avail == NULL)
    p = tp_alloc(sizeof(struct hashtab));
  else {
    hashtab_avails--;
    p = hashtab_avail;
    hashtab_avail = hashtab_avail->next;
  }
  /* initialization */	
  	
  p->next = NULL;

  /* end of initialization */	
  return(p);
}  /* get_hashtab */

/*************
 *
 *    free_hashtab()
 *
 *************/

static
void free_hashtab(Hashtab p)
{
  hashtab_frees++;
  hashtab_avails++;
  p->next = hashtab_avail;
  hashtab_avail = p;
}  /* free_hashtab */

/*************
 *
 *    Hashnode get_hashnode()
 *
 *************/

static
Hashnode get_hashnode(void)
{
  Hashnode p;

  hashnode_gets++;
  if (hashnode_avail == NULL)
    p = tp_alloc(sizeof(struct hashnode));
  else {
    hashnode_avails--;
    p = hashnode_avail;
    hashnode_avail = hashnode_avail->next;
  }
  /* initialization */	
  	
  p->next = NULL;

  /* end of initialization */	
  return(p);
}  /* get_hashnode */

/*************
 *
 *    free_hashnode()
 *
 *************/

static
void free_hashnode(Hashnode p)
{
  hashnode_frees++;
  hashnode_avails++;
  p->next = hashnode_avail;
  hashnode_avail = p;
}  /* free_hashnode */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_hash_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with hash.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_hash_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct hashtab);
  fprintf(fp, "hashtab (%4d)      %11u%11u%11u%11u%9.1f K\n",
	  n, hashtab_gets, hashtab_frees,
	  hashtab_gets - hashtab_frees, hashtab_avails,
	  (((hashtab_gets - hashtab_frees) + hashtab_avails) * n) / 1024.);

  n = sizeof(struct hashnode);
  fprintf(fp, "hashnode (%4d)     %11u%11u%11u%11u%9.1f K\n",
	  n, hashnode_gets, hashnode_frees,
	  hashnode_gets - hashnode_frees, hashnode_avails,
	  (((hashnode_gets - hashnode_frees) + hashnode_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_hash_mem */

/*************
 *
 *   p_hash_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated .
*/

/* PUBLIC */
void p_hash_mem()
{
  fprint_hash_mem(stdout, 1);
}  /* p_hash_mem */

/*
 *  end of memory management
 */

/*************
 *
 *   hash_init()
 *
 *************/

/* DOCUMENTATION
Allocate and initialize a hash table of the given size.
*/

/* PUBLIC */
Hashtab hash_init(int size)
{
  Hashtab p = get_hashtab();
  int i;

  p->size = size;
  p->table = malloc(size * sizeof(Hashnode));
  for (i = 0; i < size; i++) {
    p->table[i] = NULL;
  }
  return p;
}  /* hash_init */

/*************
 *
 *   hash_insert()
 *
 *************/

/* DOCUMENTATION
Insert a pair (integer, void *) into a hash table.
If the integer is already there, a fatal error occurs.
*/

/* PUBLIC */
void hash_insert(int id, void *v, Hashtab h)
{
  Hashnode p, prev, new;
  int i = id % h->size;

  p = h->table[i];
  prev = NULL;
  while (p != NULL && p->id < id) {
    prev = p;
    p = p->next;
  }
  if (p != NULL && p->id == id)
    fatal_error("hash_insert, id already there.");
  else {
    new = get_hashnode();
    new->id = id;
    new->v = v;
    new->next = p;
    if (prev != NULL)
      prev->next = new;
    else
      h->table[i] = new;
  }
}  /* hash_insert */

/*************
 *
 *   hash_lookup()
 *
 *************/

/* DOCUMENTATION
Find and return the pointer associated with the given integer ID.
If the ID is not in the table, NULL is returned.
*/

/* PUBLIC */
void * hash_lookup(int id, Hashtab h)
{
  Hashnode p = h->table[id % h->size];
  while (p != NULL && p->id < id)
    p = p->next;
  if (p != NULL && p->id == id)
    return p->v;
  else
    return NULL;
}  /* hash_lookup */

/*************
 *
 *   hash_delete()
 *
 *************/

/* DOCUMENTATION
Delete the pair (ID, void *) for the given integer ID.
If the ID is not in the table, nothing happens.
*/

/* PUBLIC */
void hash_delete(int id, Hashtab h)
{
  Hashnode p, prev;
  int i = id % h->size;

  p = h->table[i];
  prev = NULL;
  while (p != NULL && p->id < id) {
    prev = p;
    p = p->next;
  }

  if (p != NULL && p->id == id) {
    if (prev == NULL)
      h->table[i] = p->next;
    else
      prev->next = p->next;
    free_hashnode(p);
  }
}  /* hash_delete */

/*************
 *
 *   hash_destroy()
 *
 *************/

/* DOCUMENTATION
Free all of the memory used by the given hash table.  Do not
refer to the table after calling this routine.  The hash table
need not be empty;
*/

/* PUBLIC */
void hash_destroy(Hashtab h)
{
  int i;
  for (i = 0; i < h->size; i++) {
    Hashnode p = h->table[i];
    while (p != NULL) {
      Hashnode p2 = p;
      p = p->next;
      free_hashnode(p2);
    }
  }
  free(h->table);
  free_hashtab(h);
}  /* hash_destroy */
