#include "ioutil.h"

/* Private definitions and types */

/*************
 *
 *   fwrite_formula()
 *
 *************/

/* DOCUMENTATION
This routine prints a formula, followed by ".\n" to a file.
This version does not print extra parentheses
(it first translates the formula to a term, then
prints the term, then frees the term).
To print the formula directly, with extra parentheses,
call fprint_formula() instead.
*/

/* PUBLIC */
void fwrite_formula(FILE *fp, Formula f)
{
  Term t = formula_to_term(f);
  if (t == NULL)
    fatal_error("fwrite_formula, formula_to_term returns NULL.");
  fwrite_term_nl(fp, t);
  fflush(fp);
  zap_term(t);
}  /* fwrite_formula */

/*************
 *
 *   read_clause()
 *
 *************/

/* DOCUMENTATION
This routine reads a clause from FILE *fin.
If there are no more clauses in *fin, NULL is returned.
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
<P>
Variables are "set", upward links ar made from all subterms to the
clause, and ac_canonical() is called on each atom.
*/

/* PUBLIC */
Clause read_clause(FILE *fin, FILE *fout)
{
  Term t;

  t = read_term(fin, fout);
  if (t == NULL)
    return NULL;
  else {
    Clause c = term_to_clause(t);
    Literal lit;
    clause_set_variables(c, MAX_VARS);  /* fatal if too many vars */
    zap_term(t);
    for (lit = c->literals; lit != NULL; lit = lit->next)
      ac_canonical(lit->atom, -1);
    upward_clause_links(c);
    return c;
  }
}  /* read_clause */

/*************
 *
 *   end_of_list_clause()
 *
 *************/

static
BOOL end_of_list_clause(Clause c)
{
  if (c == NULL)
    return FALSE;
  else if (number_of_literals(c) != 1)
    return FALSE;
  else {
    Term a = c->literals->atom;
    if (!CONSTANT(a))
      return FALSE;
    else
      return is_symbol(SYMNUM(a), "end_of_list", 0);
  }
}  /* end_of_list_clause */

/*************
 *
 *   read_clause_list()
 *
 *************/

/* DOCUMENTATION
This routine reads a list of clauses from FILE *fin.
If you with the list to have a name, send a string;
othersize send NULL.  (You can name the list later with name_clist().)
If there are no more clauses in *fin, an empty Clist is returned.
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
*/

/* PUBLIC */
Clist read_clause_list(FILE *fin, FILE *fout, char *name, BOOL assign_id)
{
  Clist lst = clist_init(name);
  Clause c;

  c = read_clause(fin, fout);
  while (c != NULL && !end_of_list_clause(c)) {
    if (assign_id)
      assign_clause_id(c);
    clist_append(c, lst);
    c = read_clause(fin, fout);
  }
  if (c != NULL)
    zap_clause(c);  /* end_of_list_clause */
  return lst;
}  /* read_clause_list */

/*************
 *
 *   fwrite_clause()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a clause in mixfix form,
followed by ".\n".
*/

/* PUBLIC */
void fwrite_clause(FILE *fp, Clause c, BOOL print_id, BOOL justification_last)
{
  Term t = clause_to_term(c);

  if (t == NULL)
    fatal_error("fwrite_clause, clause_to_term returns NULL.");
  
  if (!print_id)
    fwrite_term_nl(fp, t);
  else if (!justification_last) {
    fprintf(fp, "%d ", c->id);
    fprint_just(fp, c->justification);
    fprintf(fp, " ");
    fwrite_term_nl(fp, t);
  }
  else {
    fprintf(fp, "%d ", c->id);
    fwrite_term(fp, t);
    fprintf(fp, ".  ");
    fprint_just(fp, c->justification);
    fprintf(fp, "\n");
  }
  fflush(fp);
  zap_term(t);
}  /* fwrite_clause */

/*************
 *
 *   f_clause()
 *
 *************/

/* DOCUMENTATION
Write a clause to stdout, with id, with justification last.
*/

/* PUBLIC */
void f_clause(Clause c)
{
  fwrite_clause(stdout, c, TRUE, TRUE);
}  /* f_clause */

/*************
 *
 *   fwrite_clause_list()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a list of clauses in mixfix form.
Example:
<PRE>
list(sos).
a = b.
end_of_list.
</PRE>
If the name of the list is "", it is written as list(anonymous).
*/

/* PUBLIC */
void fwrite_clause_list(FILE *fp, Clist lst, BOOL print_ids, BOOL justifications_last)
{
  Clist_pos p;
    
  if (lst->name == NULL || str_ident(lst->name, ""))
    fprintf(fp, "\nclauses(anonymous).\n");
  else
    fprintf(fp, "\nclauses(%s).\n", lst->name);
	  
  for (p = lst->first; p != NULL; p = p->next)
    fwrite_clause(fp, p->c, print_ids, justifications_last);
  fprintf(fp, "end_of_list.\n");
  fflush(fp);
}  /* fwrite_clause_list */

/*************
 *
 *   read_formula()
 *
 *************/

/* DOCUMENTATION
This routine reads a formula from FILE *fin.
If there are no more formulas in *fin, NULL is returned.
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
*/

/* PUBLIC */
Formula read_formula(FILE *fin, FILE *fout)
{
  Term t;

  t = read_term(fin, fout);
  if (t == NULL)
    return NULL;
  else {
    Formula c = term_to_formula(t);
    zap_term(t);
    return c;
  }
}  /* read_formula */

/*************
 *
 *   end_of_list_formula()
 *
 *************/

static
BOOL end_of_list_formula(Formula f)
{
  if (f == NULL)
    return FALSE;
  else if (f->type != ATOM_FORM)
    return FALSE;
  else {
    Term a = f->atom;
    if (!CONSTANT(a))
      return FALSE;
    else
      return is_symbol(SYMNUM(a), "end_of_list", 0);
  }
}  /* end_of_list_formula */

/*************
 *
 *   read_formula_list()
 *
 *************/

/* DOCUMENTATION
This routine reads a list of formulas from FILE *fin.
If there are no more formulas in *fin, NULL is returned.
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
*/

/* PUBLIC */
Plist read_formula_list(FILE *fin, FILE *fout)
{
  Plist p = NULL;
  Formula f;

  f = read_formula(fin, fout);
  while (f != NULL && !end_of_list_formula(f)) {
    p = plist_append(p, f);
    f = read_formula(fin, fout);
  }
  if (f != NULL)
    zap_formula(f);
  return p;
}  /* read_formula_list */

/*************
 *
 *   fwrite_formula_list()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a list of formulas in mixfix form.
Example:
<PRE>
list(sos).
a = b.
end_of_list.
</PRE>
If the name of the list is "", it is written as list(anonymous).
*/

/* PUBLIC */
void fwrite_formula_list(FILE *fp, Plist lst, char *name)
{
  Plist p;

  if (name == NULL || str_ident(name, ""))
    fprintf(fp, "\nformulas(anonymous).\n");
  else
    fprintf(fp, "\nformulas(%s).\n", name);

  for (p = lst; p != NULL; p = p->next)
    fwrite_formula(fp, p->v);
  fprintf(fp, "end_of_list.\n");
  fflush(fp);
}  /* fwrite_formula_list */

/*************
 *
 *   zap_formula_list()
 *
 *************/

/* DOCUMENTATION
Free a Plist of formulas.
*/

/* PUBLIC */
void zap_formula_list(Plist lst)
{
  Plist p = lst;
  while (p != NULL) {
    Plist p2 = p;
    p = p->next;
    zap_formula(p2->v);
    free_plist(p2);
  }
}  /* zap_formula_list */

/*************
 *
 *   end_of_list_term()
 *
 *************/

/* DOCUMENTATION
Check if a term is the constant "end_of_list".
*/

/* PUBLIC */
BOOL end_of_list_term(Term t)
{
  if (t == NULL)
    return FALSE;
  else if (!CONSTANT(t))
    return FALSE;
  else
    return is_symbol(SYMNUM(t), "end_of_list", 0);
}  /* end_of_list_term */

/*************
 *
 *   end_of_commands_term()
 *
 *************/

/* DOCUMENTATION
Check if a term is the constant "end_of_commands".
*/

/* PUBLIC */
BOOL end_of_commands_term(Term t)
{
  if (t == NULL)
    return FALSE;
  else if (!CONSTANT(t))
    return FALSE;
  else
    return is_symbol(SYMNUM(t), "end_of_commands", 0);
}  /* end_of_commands_term */

/*************
 *
 *   read_term_list()
 *
 *************/

/* DOCUMENTATION
This routine reads a list of terms from FILE *fin.
If there are no more terms in *fin, NULL is returned.
<P>
None of the returned terms or their subterms will have type VARIABLE,
even if it looks like a variable (e.g., x).
<P>
If any error occurs, a message is sent to FILE *fout and a fatal_error occurs.
*/

/* PUBLIC */
Plist read_term_list(FILE *fin, FILE *fout)
{
  Plist p = NULL;
  Term t;

  t = read_term(fin, fout);
  while (t != NULL && !end_of_list_term(t)) {
    p = plist_append(p, t);
    t = read_term(fin, fout);
  }
  if (t != NULL)
    zap_term(t);
  return p;
}  /* read_term_list */

/*************
 *
 *   fwrite_term_list()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a list of terms in mixfix form.
Example:
<PRE>
list(sos).
a = b.
end_of_list.
</PRE>
If the name of the list is "" or NULL, it is written as list(anonymous).
*/

/* PUBLIC */
void fwrite_term_list(FILE *fp, Plist lst, char *name)
{
  Plist p;

  if (name == NULL || str_ident(name, ""))
    fprintf(fp, "\nterms(anonymous).\n");
  else
    fprintf(fp, "\nterms(%s).\n", name);

  for (p = lst; p != NULL; p = p->next) {
    fwrite_term(fp, p->v);
    fprintf(fp, ".\n");
  }
  fprintf(fp, "end_of_list.\n");
  fflush(fp);
}  /* fwrite_term_list */

/*************
 *
 *   zap_term_list()
 *
 *************/

/* DOCUMENTATION
Free a Plist of terms.
*/

/* PUBLIC */
void zap_term_list(Plist lst)
{
  Plist p = lst;
  while (p != NULL) {
    Plist p2 = p;
    p = p->next;
    zap_term(p2->v);
    free_plist(p2);
  }
}  /* zap_term_list */

/*************
 *
 *   term_reader()
 *
 *************/

/* DOCUMENTATION
If your program has optional fast parsing/writing, you
can use this routine to save a few lines of code.
The flag "fast" says whether or not to use fast parse form.
A term is read from stdin (NULL if there are none).  
Errors are fatal.
*/

/* PUBLIC */
Term term_reader(BOOL fast)
{
  if (fast)
    return fast_read_term(stdin, stderr);
  else
    return read_term(stdin, stderr);
}  /* term_reader */

/*************
 *
 *   term_writer()
 *
 *************/

/* DOCUMENTATION
If your program has optional fast parsing/writing, you
can use this routine to save a few lines of code.
The flag "fast" says whether or not to use fast parse form.
The term is written to stdout, with a period and newline.
*/

/* PUBLIC */
void term_writer(Term t, BOOL fast)
{
  if (fast)
    fast_fwrite_term_nl(stdout, t);
  else
    fwrite_term_nl(stdout, t);
  fflush(stdout);
}  /* term_writer */

/*************
 *
 *   clause_reader()
 *
 *************/

/* DOCUMENTATION
If your program has optional fast parsing/writing, you
can use this routine to save a few lines of code.
The flag "fast" says whether or not to use fast parse form.
A clause is read from stdin (NULL if there are none).  
Errors are fatal.
*/

/* PUBLIC */
Clause clause_reader(BOOL fast)
{
  if (fast)
    return fast_read_clause(stdin, stderr);
  else
    return read_clause(stdin, stderr);
}  /* clause_reader */

/* DOCUMENTATION
If your program has optional fast parsing/writing, you
can use this routine to save a few lines of code.
The flag "fast" says whether or not to use fast parse form.
The clause is written to stdout, with a period and newline.
*/

/* PUBLIC */
void clause_writer(Clause c, BOOL fast)
{
  if (fast)
    fast_fwrite_clause(stdout, c);
  else
    fwrite_clause(stdout, c, FALSE, FALSE);
  fflush(stdout);
}  /* clause_writer */

/*************
 *
 *   bad_constants_in_clause()
 *
 *************/

/* DOCUMENTATION
Given a clause, check if it contains any constants that
satisfy variable_name().  If so, print a message for
each and return TRUE.
*/

/* PUBLIC */
BOOL bad_constants_in_clause(Clause c)
{
  Plist constants = constants_in_clause(c);
  Plist p;
  BOOL bad = FALSE;
  for (p = constants; p; p = p->next) {
    Term t = p->v;
    char *s = sn_to_str(SYMNUM(t));
    if (variable_name(s)) {
      printf("\nThe following clause contains a constant that might be\n"
	     "misinterpreted by the user as a variable. It probably came\n"
	     "from a formula with a missing quantifier. The constant is %s.\n",
	     s);
      fwrite_clause(stdout, c, TRUE, TRUE);
      bad = TRUE;
    }
  }
  zap_plist(constants);
  return bad;
}  /* bad_constants_in_clause */

/*************
 *
 *   check_constants_in_clist()
 *
 *************/

/* DOCUMENTATION
Given a list of clauses, check if it contains any constants that
satisfy variable_name().  If so, print messages and die
by calling fatal_error().
*/

/* PUBLIC */
void check_constants_in_clist(Clist l)
{
  BOOL die = FALSE;
  Clist_pos p;
  for (p = l->first; p; p = p->next) {
    if (bad_constants_in_clause(p->c))
      die = TRUE;
  }
  if (die)
    fatal_error("check_constants_in_clist (see stdout)");
}  /* check_constants_in_clist */

