#include "just.h"

/* Private definitions and types */

typedef struct parajust * Parajust;

struct parajust {
  int from_id;
  int into_id;
  Ilist from_pos;
  Ilist into_pos;
};

struct just {
  Just_type type;
  Just next;
  union {
    int id;
    Ilist lst;
    Parajust para;
  } u;
};

/*
 * memory management
 */

static Just just_avail;
static unsigned just_gets, just_frees, just_avails;

static Parajust parajust_avail;
static unsigned parajust_gets, parajust_frees, parajust_avails;

/* end of counters for each type */

/*************
 *
 *    Just get_just()
 *
 *************/

static
Just get_just(void)
{
  Just p;
    
  just_gets++;
  if (just_avail == NULL)
    p = tp_alloc(sizeof(struct just));
  else {
    just_avails--;
    p = just_avail;
    just_avail = just_avail->next;
  }
  p->type = 0;
  p->u.id = 0;
  p->next = NULL;
  return(p);
}  /* get_just */

/*************
 *
 *    free_just()
 *
 *************/

static
void free_just(Just p)
{
  just_frees++;
  just_avails++;
  p->next = just_avail;
  just_avail = p;
}  /* free_just */

/*************
 *
 *    Parajust get_parajust()
 *
 *************/

static
Parajust get_parajust(void)
{
  Parajust p;
    
  parajust_gets++;
  if (parajust_avail == NULL)
    p = tp_alloc(sizeof(struct parajust));
  else {
    parajust_avails--;
    p = parajust_avail;
    parajust_avail = (Parajust) parajust_avail->into_pos;
  }
  p->from_id = 0;
  p->into_id = 0;
  p->from_pos = NULL;
  p->into_pos = NULL;
  return(p);
}  /* get_parajust */

/*************
 *
 *    free_parajust()
 *
 *************/

static
void free_parajust(Parajust p)
{
  parajust_frees++;
  parajust_avails++;
  p->into_pos = (Ilist) parajust_avail;
  parajust_avail = p;
}  /* free_parajust */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_just_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with ??.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_just_mem(FILE *fp, int heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct just);
  fprintf(fp, "just (%4d)         %11u%11u%11u%11u%9.1f K\n",
	  n, just_gets, just_frees, just_gets - just_frees,
	  just_avails,
	  (((just_gets - just_frees) + just_avails) * n) / 1024.);

  n = sizeof(struct parajust);
  fprintf(fp, "parajust (%4d)     %11u%11u%11u%11u%9.1f K\n",
	  n, parajust_gets, parajust_frees, parajust_gets - parajust_frees,
	  parajust_avails,
	  (((parajust_gets - parajust_frees) + parajust_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_just_mem */

/*************
 *
 *   p_just_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with ??.
*/

/* PUBLIC */
void p_just_mem()
{
  fprint_just_mem(stdout, 1);
}  /* p_just_mem */

/*
 *  end of memory management
 */

/*************
 *
 *   fprint_just()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) a clause justification.
No whitespace is printed before or after.
*/

/* PUBLIC */
void fprint_just(FILE *fp, Just just)
{
  Just g = just;
  fprintf(fp, "[");
  while (g != NULL) {
    Just_type rule = g->type;
    if (rule==BINARY_RES_JUST || rule==HYPER_RES_JUST || rule==UR_RES_JUST) {
      Ilist p;
      char *s;
      switch (rule) {
      case BINARY_RES_JUST: s = "res"; break;
      case HYPER_RES_JUST:  s = "hyper"; break;
      case UR_RES_JUST:     s = "ur"; break;
      default:              s = "?";
      }
      fprintf(fp, "%s (", s);
      for (p = g->u.lst; p; p = p->next) {
	fprintf(fp, "%d", p->i);
	fprintf(fp, "%s", p->next == NULL ? ")" : " ");
      }
    }
    else if (rule == PARA_JUST || rule == PARA_FX_JUST ||
	     rule == PARA_IX_JUST || rule == PARA_FX_IX_JUST) {
      char *s;
      Parajust p = g->u.para;

      switch (rule) {
      case PARA_JUST:       s = "para";       break;
      case PARA_FX_JUST:    s = "para_fx";    break;
      case PARA_IX_JUST:    s = "para_ix";    break;
      case PARA_FX_IX_JUST: s = "para_fx_ix"; break;
      default: s = "?";
      }

      fprintf(fp, "%s (%d ", s, p->from_id);
      fprint_ilist(fp, p->from_pos);
      fprintf(fp, " %d ", p->into_id);
      fprint_ilist(fp, p->into_pos);
      fprintf(fp, ")");
    }
    else if (rule == DEMOD_JUST) {
      Ilist p;
      fprintf(fp, " demod (");
      for (p = g->u.lst; p; p = p->next) {
	fprintf(fp, "%d", p->i);
	fprintf(fp, "%s", p->next == NULL ? "" : " ");
      }
      fprintf(fp, ")");
    }
    else if (rule == UNIT_DEL_JUST) {
      Ilist p = g->u.lst;
      fprintf(fp, " unit_del (%d %d)", p->i, p->next->i);
    }
    else if (rule == FACTOR_JUST) {
      Ilist p = g->u.lst;
      fprintf(fp, "factor (%d %d %d)", p->i, p->next->i, p->next->next->i); 
    }
    else if (rule == CD_JUST) {
      Ilist p = g->u.lst;
      fprintf(fp, "cd (%d %d)", p->i, p->next->i); 
    }
    else if (rule == FLIP_JUST)
      fprintf(fp, " flip %d", g->u.id); 
    else if (rule == XX_JUST)
      fprintf(fp, " xx %d", g->u.id); 
    else if (rule == MERGE_JUST)
      fprintf(fp, " merge %d", g->u.id); 
    else if (rule == BACK_DEMOD_JUST)
      fprintf(fp, "back_demod %d", g->u.id); 
    else {
      printf("\nunknown rule: %d\n", rule);
      fatal_error("fprint_just: unknown rule");
    }
    g = g->next;
  }
  fprintf(fp, "]");
}  /* fprint_just */

/*************
 *
 *   zap_parajust()
 *
 *************/

static
void zap_parajust(Parajust p)
{
  zap_ilist(p->from_pos);
  zap_ilist(p->into_pos);
  free_parajust(p);
}  /* zap_parajust */

/*************
 *
 *   zap_just()
 *
 *************/

/* DOCUMENTATION
This routine frees a justification list, including any sublists.
*/

/* PUBLIC */
void zap_just(Just just)
{
  if (just != NULL) {
    zap_just(just->next);
    
    switch (just->type) {
    case INPUT:
    case BACK_DEMOD_JUST:
    case FLIP_JUST:
    case XX_JUST:
    case MERGE_JUST:
      break;  /* nothing to do */
    case CD_JUST:
    case BINARY_RES_JUST:
    case HYPER_RES_JUST:
    case UR_RES_JUST:
    case DEMOD_JUST:
    case UNIT_DEL_JUST:
    case FACTOR_JUST:
      zap_ilist(just->u.lst); break;
    case PARA_JUST:
    case PARA_FX_JUST:
    case PARA_IX_JUST:
    case PARA_FX_IX_JUST:
      zap_parajust(just->u.para); break;
    default: fatal_error("zap_just: unknown type");
    }
    free_just(just);
  }
}  /* zap_just */

/*************
 *
 *   get_clanc()
 *
 *************/

static
Plist get_clanc(int id, Plist anc)
{
  Clause c = find_clause_by_id(id);
  if (c == NULL) {
    printf("get_clanc, clause with id=%d not found.\n", id);
    fatal_error("get_clanc, clause not found.");
  }

  if (!plist_member(anc, c)) {
    Just g = c->justification;
    anc = insert_clause_into_plist(anc, c, TRUE);

    while (g != NULL) {
      Just_type rule = g->type;
      if (rule==BINARY_RES_JUST || rule==HYPER_RES_JUST || rule==UR_RES_JUST) {
	/* [rule (nucid nuclit sat1id sat1lit nuclit2 sat2id sat2lit ...)] */
	Ilist p = g->u.lst;
	int nuc_id = p->i;
	anc = get_clanc(nuc_id, anc);
	p = p->next;
	while (p != NULL) {
	  int sat_id = p->next->i;
	  anc = get_clanc(sat_id, anc);
	  p = p->next->next->next;
	}
      }
      else if (rule == PARA_JUST || rule == PARA_FX_JUST ||
	       rule == PARA_IX_JUST || rule == PARA_FX_IX_JUST) {
	Parajust p   = g->u.para;
	anc = get_clanc(p->from_id, anc);
	anc = get_clanc(p->into_id, anc);
      }
      else if (rule == FACTOR_JUST) {
	int parent_id = g->u.lst->i;
	anc = get_clanc(parent_id, anc);
      }
      else if (rule == UNIT_DEL_JUST) {
	int parent_id = g->u.lst->i;
	anc = get_clanc(parent_id, anc);
      }
      else if (rule == CD_JUST) {
	Ilist p = g->u.lst;
	int major_id = p->i;
	int minor_id = p->next->i;
	anc = get_clanc(major_id, anc);
	anc = get_clanc(minor_id, anc);
      }
      else if (rule == BACK_DEMOD_JUST) {
	int parent_id = g->u.id;
	anc = get_clanc(parent_id, anc);
      }
      else if (rule == DEMOD_JUST) {
	Ilist p;
	for (p = g->u.lst; p; p = p->next)
	  anc = get_clanc(p->i, anc);
      }
      else if (rule == FLIP_JUST)
	;  /* do nothing */
      else if (rule == XX_JUST)
	;  /* do nothing */
      else if (rule == MERGE_JUST)
	;  /* do nothing */
      else {
	fatal_error("get_clanc, unknown rule.");
      }
      g = g->next;
    }
  }
  return anc;
}  /* get_clanc */

/*************
 *
 *   get_clause_ancestors()
 *
 *************/

/* DOCUMENTATION
This routine returns the Plist of clauses that are ancestors of Clause c,
including clause c.  The result is sorted (increasing) by ID.
*/

/* PUBLIC */
Plist get_clause_ancestors(Clause c)
{
  Plist ancestors = get_clanc(c->id, NULL);
  return ancestors;
}  /* get_clause_ancestors */

/*************
 *
 *   proof_length()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int proof_length(Plist proof)
{
  Plist p;
  int n = 0;
  for (p = proof; p; p = p->next) {
    Clause c = p->v;
    if (c->justification != NULL)
      n++;
  }
  return n;
}  /* proof_length */

/*************
 *
 *   cd_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for a factorization.
*/

/* PUBLIC */
Just cd_just(Clause major, Clause minor)
{
  /* (CD_JUST (major_id minor_id)) */
  Just j = get_just();
  j->type = CD_JUST;
  j->u.lst = ilist_append(ilist_append(NULL, major->id), minor->id);
  return j;
}  /* cd_just */

/*************
 *
 *   unit_del_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for a factorization.
*/

/* PUBLIC */
Just unit_del_just(Clause deleter, int literal_num)
{
  /* UNIT_DEL (clause-id literal-num) */
  Just j = get_just();
  j->type = UNIT_DEL_JUST;
  j->u.lst = ilist_append(ilist_append(NULL, deleter->id), literal_num);
  return j;
}  /* cd_just */

/*************
 *
 *   back_demod_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for back_demod.
*/

/* PUBLIC */
Just back_demod_just(Clause c)
{
  /* (BACK_DEMOD_JUST parent_id) */
  Just j = get_just();
  j->type = BACK_DEMOD_JUST;
  j->u.id = c->id;
  return j;
}  /* back_demod_just */

/*************
 *
 *   flip_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification equality flipping.
*/

/* PUBLIC */
Just flip_just(int n)
{
  /* (FLIP_JUST literal-number) */
  Just j = get_just();
  j->type = FLIP_JUST;
  j->u.id = n;
  return j;
}  /* flip_just */

/*************
 *
 *   xx_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification for the XX rule,
which removes literals that are instances of x!=x.
*/

/* PUBLIC */
Just xx_just(int n)
{
  /* (FLIP_JUST literal-number) */
  Just j = get_just();
  j->type = XX_JUST;
  j->u.id = n;
  return j;
}  /* xx_just */

/*************
 *
 *   merge_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification for the merging
a literal.  The n-th literal has been removed because it is
identical to another literal.
*/

/* PUBLIC */
Just merge_just(int n)
{
  /* (merge literal-number) */
  Just j = get_just();
  j->type = MERGE_JUST;
  j->u.id = n;
  return j;
}  /* merge_just */

/*************
 *
 *   binary_res_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for binary resolution.
(Binary res justifications may also be constructed in resolve(), along
with hyper and UR.)
*/

/* PUBLIC */
Just binary_res_just(Clause c1, int n1, Clause c2, int n2)
{
  /* (BINARY_RES_JUST (id1 lit1 id2 lit2) */
  Just j = get_just();
  j->type = BINARY_RES_JUST;
  j->u.lst = ilist_append(
                  ilist_append(
                       ilist_append(
                            ilist_append(NULL,c1->id),n1),c2->id),n2);

  return j;
}  /* binary_res_just */

/*************
 *
 *   para_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for a paramodulation
inference.  The position vectors are copied and reversed.
*/

/* PUBLIC */
Just para_just(Just_type rule,
		Clause from, Ilist from_vec,
		Clause into, Ilist into_vec)
{
  Just j = get_just();
  j->type = rule;
  j->u.para = get_parajust();

  j->u.para->from_id = from->id;
  j->u.para->into_id = into->id;
  j->u.para->from_pos = reverse_ilist(copy_ilist(from_vec));
  j->u.para->into_pos = reverse_ilist(copy_ilist(into_vec));

  return j;
}  /* para_just */

/*************
 *
 *   demod_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for demodulation.
*/

/* PUBLIC */
Just demod_just(Ilist g)
{
  Just j = get_just();
  j->type = DEMOD_JUST;
  j->u.lst = g;
  return j;
}  /* demod_just */

/*************
 *
 *   resolve_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification resolution rules.
*/

/* PUBLIC */
Just resolve_just(Ilist g, Just_type type)
{
  Just j = get_just();
  j->type = type;
  j->u.lst = g;
  return j;
}  /* resolve_just */

/*************
 *
 *   factor_just()
 *
 *************/

/* DOCUMENTATION
This routine builds and returns a justification list for a factorization.
*/

/* PUBLIC */
Just factor_just(Clause c, int lit1, int lit2)
{
  /* (FACTOR_JUST (clause_id lit1 lit2)) */
  Just j = get_just();
  j->type = FACTOR_JUST;
  j->u.lst = ilist_append(ilist_append(ilist_append(NULL,c->id),lit1),lit2);
  return j;
}  /* factor_just */

/*************
 *
 *   append_just()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Just append_just(Just j1, Just j2)
{
  if (j1 == NULL)
    return j2;
  else {
    j1->next = append_just(j1->next, j2);
    return j1;
  }
}  /* append_just */

