#include "listterm.h"

/* Private definitions and types */

/*************
 *
 *   get_nil_term()
 *
 *************/

/* DOCUMENTATION
Allocate and return an empty listterm.
*/

/* PUBLIC */
Term get_nil_term()
{
  return get_rigid_term("$nil", 0);
}  /* get_nil_term */

/*************
 *
 *   listterm_cons()
 *
 *************/

/* DOCUMENTATION
This routine returns the "cons" of two terms.
The two terms are not copied.
*/

/* PUBLIC */
Term listterm_cons(Term t1, Term t2)
{
  Term c = get_rigid_term("$cons", 2);
  c->args[0] = t1;
  c->args[1] = t2;
  return c;
}  /* listterm_cons */

/*************
 *
 *   cons_term()
 *
 *************/

/* DOCUMENTATION
This function checks if a term is a "cons", that is,
arity 2 with the official "cons" symbol.   
*/

/* PUBLIC */
BOOL cons_term(Term t)
{
  return is_symbol(SYMNUM(t), "$cons", 2);
}  /* cons_term */

/*************
 *
 *   nil_term()
 *
 *************/

/* DOCUMENTATION
This function checks if a term is a "nil", that is,
arity 0 with the official "nil" symbol.   
*/

/* PUBLIC */
BOOL nil_term(Term t)
{
  return is_symbol(SYMNUM(t), "$nil", 0);
}  /* nil_term */

/*************
 *
 *   proper_listterm()
 *
 *************/

/* DOCUMENTATION
This function checks if a term is a proper listterm,
that is, a nil_term(), or a cons_term() whose tail
is a proper_listterm().
*/

/* PUBLIC */
BOOL proper_listterm(Term t)
{
  if (nil_term(t))
    return TRUE;
  else if (cons_term(t))
    return proper_listterm(t->args[1]);
  else
    return FALSE;
}  /* proper_listterm */

/*************
 *
 *   listterm_append()
 *
 *************/

/* DOCUMENTATION
This routine appends an element to a listterm.  The resulting
listterm is returned.  Neither the list nor the element is copied.
You should not refer to the argument "list" after calling this
routine---a good way to call it is like this:
<PRE>list = listterm_append(list, element)</PRE>
<P>
If "list" is not a proper_listterm(), the result will be
well-formed, but it might not be what you expect.
*/

/* PUBLIC */
Term listterm_append(Term list, Term element)
{
  if (!cons_term(list))
    return listterm_cons(element, list);
  else {
    list->args[1] = listterm_append(list->args[1], element);
    return list;
  }
}  /* listterm_append */

/*************
 *
 *   listterm_length()
 *
 *************/

/* DOCUMENTATION
This function returns the length of a listterm.
*/

/* PUBLIC */
int listterm_length(Term t)
{
  if (!cons_term(t))
    return 0;
  else {
    return 1 + listterm_length(t->args[1]);
  }
}  /* listterm_length */

/*************
 *
 *   listterm_i()
 *
 *************/

/* DOCUMENTATION
Return the i-th member, counting from 1, of a listterm.
If there are less than i members, return NULL.
*/

/* PUBLIC */
Term listterm_i(Term lst, int i)
{
  if (!cons_term(lst))
    return NULL;
  else if (i == 1)
    return lst->args[0];
  else
    return listterm_i(lst->args[1], i-1);
}  /* listterm_i */

/*************
 *
 *   listterm_member()
 *
 *************/

/* DOCUMENTATION
This function checks if Term t is a member of a listterm (Term lst).
*/

/* PUBLIC */
BOOL listterm_member(Term t, Term lst)
{
  if (cons_term(lst)) {
    if (term_ident(t, lst->args[0]))
      return TRUE;
    else
      return listterm_member(t, lst->args[1]);
  }
  else
    return FALSE;
}  /* listterm_member */

