#include "options.h"

/* Private definitions and types */

struct flag {  /* Flags are boolean valued options */
    char *name;
    int val;
    };

struct parm {  /* Parms are integer valued options */
    char *name;
    int val;
    int min, max;  /* minimum and maximum permissible values */
    };

static struct flag Flags[MAX_FLAGS];
static struct parm Parms[MAX_PARMS];

/*
 * memory management
 */

/*
 *  end of memory management
 */

/*************
 *
 *    init_options()
 *
 *************/

/* DOCUMENTATION
This routine initializes all of the flags and parameters.  It should
be called once at the beginning of the process, before doing any
options operations.
*/

/* PUBLIC */
void init_options(void)
{
  int i;
  for (i = 0; i < MAX_FLAGS; i++)
    Flags[i].name = "";
  for (i = 0; i < MAX_PARMS; i++)
    Parms[i].name = "";
}  /* init_options */

/*************
 *
 *    next_available_flag_id()
 *
 *************/

/* DOCUMENTATION
If you're not sure what flag IDs have already been claimed,
for example if you're adding options to an existing program,
you can use this function to find an available one.
Return -1 if none is available.
*/

/* PUBLIC */
int next_available_flag_id(void)
{
  int i = 0;
  while (i < MAX_FLAGS && !str_ident(Flags[i].name, ""))
    i++;
  return (i < MAX_FLAGS ? i : -1);
}  /* next_available_flag_id */

/*************
 *
 *    next_available_parm_id()
 *
 *************/

/* DOCUMENTATION
If you're not sure what parm IDs have already been claimed,
for example if you're adding options to an existing program,
you can use this function to find an available one.
Return -1 if none is available.
*/

/* PUBLIC */
int next_available_parm_id(void)
{
  int i = 0;
  while (i < MAX_PARMS && !str_ident(Parms[i].name, ""))
    i++;
  return (i < MAX_PARMS ? i : -1);
}  /* next_available_parm_id */

/*************
 *
 *    init_flag()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_flag(int flag_id,
	       char *flag_name,
	       BOOL default_value)
{
  if (flag_id < 0 || flag_id >= MAX_FLAGS)
    fatal_error("init_flag: flag ID out of range");
  else if (!str_ident(Flags[flag_id].name, ""))
    fatal_error("init_flag: flag ID already in use");
  else if (str_to_flag_id(flag_name) != -1)
    fatal_error("init_flag: flag name already in use");
  else {
    Flags[flag_id].name = flag_name;
    Flags[flag_id].val = default_value;
  }
}  /* init_flag */

/*************
 *
 *    init_parm()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_parm(int parm_id,
	       char *parm_name,
	       int default_value,
	       int min_value,
	       int max_value)
{
  if (parm_id < 0 || parm_id >= MAX_PARMS)
    fatal_error("init_parm: parm ID out of range");
  else if (!str_ident(Parms[parm_id].name, ""))
    fatal_error("init_parm: parm ID already in use");
  else if (str_to_parm_id(parm_name) != -1)
    fatal_error("init_parm: parm name already in use");
  else if (default_value < min_value || default_value > max_value)
    fatal_error("init_parm: default parm value out of range");
  else {
    Parms[parm_id].name = parm_name;
    Parms[parm_id].val = default_value;
    Parms[parm_id].min = min_value;
    Parms[parm_id].max = max_value;
  }
}  /* init_parm */

/*************
 *
 *    fprint_options(fp)
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) the current values of all of the
the flags and parameters.
*/

/* PUBLIC */
void fprint_options(FILE *fp)
{
  int i, j;

  fprintf(fp, "\n--------------- options ---------------\n");

  j = 0;
  for (i = 0; i < MAX_FLAGS; i++)  /* print set flags */
    if (Flags[i].name[0] != '\0') {
      fprintf(fp, "%s", Flags[i].val ? "set(" : "clear(");
      fprintf(fp, "%s). ", Flags[i].name);
      j++;
      if (j % 3 == 0)
	fprintf(fp, "\n");
    }
  fprintf(fp, "\n\n");

  j = 0;
  for (i = 0; i < MAX_PARMS; i++)  /* print parms */
    if (Parms[i].name[0] != '\0') {
      fprintf(fp, "assign(");
      fprintf(fp, "%s, %d). ", Parms[i].name, Parms[i].val);
      j++;
      if (j % 3 == 0)
	fprintf(fp, "\n");
    }
  fprintf(fp, "\n");
  fflush(fp);
}  /* fprint_options */

/*************
 *
 *    p_options()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) the current values of the
the flags and parameters.
*/

/* PUBLIC */
void p_options(void)
{
  fprint_options(stdout);
}  /* p_options */

/*************
 *
 *   flag()
 *
 *************/

/* DOCUMENTATION
This Boolean routine returns the value of a flag.  If the Flag
index is out of range, bad things can happen.
*/

/* PUBLIC */
int flag(int flag_id)
{
  return Flags[flag_id].val;
}  /* flag */

/*************
 *
 *   parm()
 *
 *************/

/* DOCUMENTATION
This integer routine returns the value of a parameter.  If the parm
index is out of range, bad things can happen.
*/

/* PUBLIC */
int parm(int parm_id)
{
  return Parms[parm_id].val;
}  /* parm */

/*************
 *
 *   set_flag()
 *
 *************/

/* DOCUMENTATION
This routine sets a flag.  The flag is identified by its
integer ID (which is available from str_to_flag_id()).
If the ID is not valid, a fatal error occurs.
*/

/* PUBLIC */
void set_flag(int id)
{
  if (id < 0 || id >= MAX_FLAGS) {
    fprintf(stdout, "set_flag: flag id %d, is out of range.\n", id);
    fprintf(stderr, "set_flag: flag id %d, is out of range.\n", id);
    fatal_error("set_flag");
  }
  else
    Flags[id].val = TRUE;
}  /* set_flag */

/*************
 *
 *   clear_flag()
 *
 *************/

/* DOCUMENTATION
This routine clears a flag.  The flag is identified by its
integer ID (which is available from str_to_flag_id()).
If the ID is not valid, a fatal error occurs.
*/

/* PUBLIC */
void clear_flag(int id)
{
  if (id < 0 || id >= MAX_FLAGS) {
    fprintf(stdout, "clear_flag: flag id %d, is out of range.\n", id);
    fprintf(stderr, "clear_flag: flag id %d, is out of range.\n", id);
    fatal_error("clear_flag");
  }
  else
    Flags[id].val = FALSE;
}  /* clear_flag */

/*************
 *
 *   assign_parm()
 *
 *************/

/* DOCUMENTATION
This routine assigns a value to a parm.  The parm is identified by
its integer ID (which is available from str_to_parm_id()).
If the ID is not valid, or if the value is out of range, a
fatal error occurs.
*/

/* PUBLIC */
void assign_parm(int id, int val)
{
  if (id < 0 || id >= MAX_PARMS) {
    fprintf(stdout, "assign_parm: parm id %d, is out of range.\n", id);
    fprintf(stderr, "assign_parm: parm id %d, is out of range.\n", id);
    fatal_error("assign_parm");
  }
  else if (val < Parms[id].min || val > Parms[id].max) {
    fprintf(stdout, "assign_parm: parm %s, value %d out of range [%d..%d].\n",
	    Parms[id].name, val, Parms[id].min, Parms[id].max);
    fprintf(stderr, "assign_parm: parm %s, value %d out of range [%d..%d].\n",
	    Parms[id].name, val, Parms[id].min, Parms[id].max);
    fatal_error("assign_parm");
  }
  else
    Parms[id].val = val;
}  /* assign_parm */

/*************
 *
 *   str_to_flag_id()
 *
 *************/

/* DOCUMENTATION
This routine converts the string name of a flag to its integer ID.
If the string name is not valid, -1 is returned.
*/

/* PUBLIC */
int str_to_flag_id(char *flag_name)
{
  int id = 0;
  BOOL already_here = FALSE;

  while (id < MAX_FLAGS && !already_here)
    if (str_ident(flag_name, Flags[id].name))
      already_here = TRUE;
    else
      id++;
  if (already_here)
    return id;
  else
    return -1;
}  /* str_to_flag_id */

/*************
 *
 *   str_to_parm_id()
 *
 *************/

/* DOCUMENTATION
This routine converts the string name of a parm to its integer ID.
If the string name is not valid, -1 is returned.
*/

/* PUBLIC */
int str_to_parm_id(char *parm_name)
{
  int id = 0;
  BOOL already_here = FALSE;

  while (id < MAX_PARMS && !already_here)
    if (str_ident(parm_name, Parms[id].name))
      already_here = TRUE;
    else
      id++;
  if (already_here)
    return id;
  else
    return -1;
}  /* str_to_parm_id */

/*************
 *
 *   flag_id_to_str()
 *
 *************/

/* DOCUMENTATION
Given a flag ID, return the corresponding name of the flag.
*/

/* PUBLIC */
char *flag_id_to_str(int id)
{
  if (id < 0 || id > MAX_FLAGS)
    fatal_error("flag_id_to_str, bad id");
  return Flags[id].name;
}  /* flag_id_to_str */

/*************
 *
 *   parm_id_to_str()
 *
 *************/

/* DOCUMENTATION
Given a parm ID, return the corresponding name of the parm.
*/

/* PUBLIC */
char *parm_id_to_str(int id)
{
  if (id < 0 || id > MAX_PARMS)
    fatal_error("parm_id_to_str, bad id");
  return Parms[id].name;
}  /* parm_id_to_str */

