#include "parautil.h"

/* Private definitions and types */

int Oriented_flag       = -1;  /* termflag to mark oriented equality atoms */
int Renamable_flip_flag = -1;  /* termflag to mark renamable-flip eq atoms */

/*************
 *
 *   eq_atom()
 *
 *************/

/* DOCUMENTATION
This function checks if an atom is an equality atom (positive or negative)
for the purposes of paramodulation and demodulation.
*/

/* PUBLIC */
BOOL eq_atom(Term a)
{
  return is_symbol(SYMNUM(a), "=", 2);
}  /* eq_atom */

/*************
 *
 *   pos_eq_unit()
 *
 *************/

/* DOCUMENTATION
This function checks if a clause is a positive equality unit
for the purposes of paramodulation and demodulation.
*/

/* PUBLIC */
BOOL pos_eq_unit(Clause c)
{
  return (number_of_literals(c) == 1 &&
	  c->literals->sign &&
	  eq_atom(c->literals->atom));
}  /* pos_eq_unit */

/*************
 *
 *   mark_renamable_flip()
 *
 *************/

/* DOCUMENTATION
This routine marks an atom as "renamable_flip".
*/

/* PUBLIC */
void mark_renamable_flip(Term atom)
{
  if (Renamable_flip_flag == -1)
    Renamable_flip_flag = claim_term_flag();
  term_flag_set(atom, Renamable_flip_flag);
}  /* mark_renamable_flip */

/*************
 *
 *   renamable_flip_eq()
 *
 *************/

/* DOCUMENTATION
This function checks if an atom is a renamable_flip equality atom.
(The terms are not actually compared.  Only the mark is checked.)
*/

/* PUBLIC */
BOOL renamable_flip_eq(Term atom)
{
  if (Renamable_flip_flag == -1)
    /* Nothing has ever been marked renamable_flip. */
    return FALSE;
  else
    return term_flag(atom, Renamable_flip_flag);
}  /* renamable_flip_eq */

/*************
 *
 *   mark_oriented_eq()
 *
 *************/

/* DOCUMENTATION
This routine marks an atom as an oriented equality.
*/

/* PUBLIC */
void mark_oriented_eq(Term atom)
{
  if (Oriented_flag == -1)
    Oriented_flag = claim_term_flag();
  term_flag_set(atom, Oriented_flag);
}  /* mark_oriented_eq */

/*************
 *
 *   oriented_eq()
 *
 *************/

/* DOCUMENTATION
This function checks if an atom is an oriented equality atom.
(The terms are not actually compared.  Only the mark is checked.)
*/

/* PUBLIC */
BOOL oriented_eq(Term atom)
{
  if (Oriented_flag == -1)
    /* Nothing has ever been oriented. */
    return FALSE;
  else
    return term_flag(atom, Oriented_flag);
}  /* oriented_eq */

/*************
 *
 *   term_compare_primary()
 *
 *************/

/* DOCUMENTATION
Compare two terms with LRPO.
Return GREATER_THAN, LESS_THAN, or NOT_COMPARABLE.
*/

/* PUBLIC */
Ordertype term_compare_primary(Term alpha, Term beta)
{
  if (lrpo(alpha, beta))
    return GREATER_THAN;
  else if (lrpo(beta, alpha))
    return LESS_THAN;
  else
    return NOT_COMPARABLE;
}  /* term_compare_primary */

/*************
 *
 *   same_term_structure()
 *
 *************/

/* DOCUMENTATION
Do terms t1 and t2 have the same structure?  That is, if
we rename all variables to x, are t1 and t2 identical?
*/

/* PUBLIC */
BOOL same_term_structure(Term t1, Term t2)
{
  if (VARIABLE(t1) || VARIABLE(t2))
    return VARIABLE(t1) && VARIABLE(t2);
  else if (SYMNUM(t1) != SYMNUM(t2))
    return FALSE;
  else {
    int i;
    for (i = 0; i < ARITY(t1); i++)
      if (!same_term_structure(ARG(t1,i), ARG(t2,i)))
	return FALSE;
    return TRUE;
  }
}  /* same_term_structure */

/*************
 *
 *   flip_eq()
 *
 *   Flip an equality atom and update the just. of the containing clause.
 *
 *************/

static
void flip_eq(Term atom, int n)
{
  Clause c = atom->container;
  Term t = ARG(atom,0);
  ARG(atom,0) = ARG(atom,1);
  ARG(atom,1) = t;
  c->justification = append_just(c->justification, flip_just(n));
}  /* flip_eq */

/*************
 *
 *   orient_equalities()
 *
 *************/

/* DOCUMENTATION
For each equality literal (positive or negative) of Clause c,
compare the arguments; if the left is greater, mark the
atom as oriented, and if the the right is greater, flip the
arguments (add a note about flipping to the justification),
and mark the atom as oriented.
<P>
(<I>What ordering method will be used???</I>
*/

/* PUBLIC */
void orient_equalities(Clause c)
{
  Literal lit;
  int i;
  for (lit = c->literals, i = 1; lit != NULL; lit = lit->next, i++) {
    Term atom = lit->atom;
    if (eq_atom(atom)) {
      Term alpha = ARG(atom,0);
      Term beta  = ARG(atom,1);
      if (!term_ident(alpha, beta)) {
	/* Check in the primary ordering. */
	Ordertype rc = term_compare_primary(alpha, beta);
	if (rc == GREATER_THAN)
	  mark_oriented_eq(atom);
	else if (rc == LESS_THAN) {
	  flip_eq(atom, i);
	  mark_oriented_eq(atom);
	}
	else {
	  /* Not orientable by the primary ordering. */
	  Term a1 = copy_term(atom);
	  Term a2 = copy_term(atom);
	  Term tmp = ARG(a2,0);
	  ARG(a2,0) = ARG(a2,1);
	  ARG(a2,1) = tmp;
	  term_renumber_variables(a1, MAX_VARS);
	  term_renumber_variables(a2, MAX_VARS);
	  if (term_ident(a1,a2))
	    /* The renumbered flip is identical to the renumbered original. */
	    mark_renamable_flip(atom);  /* We won't para from both sides. */
	  else {
	    /* Flip if the right side is greater in the secondary ordering. */
	    rc = term_compare_vcp(a1, a2);
	    if (rc == LESS_THAN)
	      flip_eq(atom, i);
	  }
	  zap_term(a1);
	  zap_term(a2);
	}
      }  /* not identical */
    }  /* eq_atom */
  }  /* foreach literal */
}  /* orient_equalities */

/*************
 *
 *   eq_tautology()
 *
 *************/

/* DOCUMENTATION
This routine returns TRUE if the clause has any literals of the form t=t.
*/

/* PUBLIC */
BOOL eq_tautology(Clause c)
{
  Literal l1;
  for (l1 = c->literals; l1; l1 = l1->next) {
    Term a = l1->atom;
    if (l1->sign && eq_atom(a) && term_ident(ARG(a,0), ARG(a,1)))
      return TRUE;
  }
  return FALSE;
}  /* eq_tautology */

