/*****************************************************************************
 *                                                                           *
 *   Prover5                                                                 *
 *                                                                           *
 *****************************************************************************/

/*  TO DO

 + Term ordering for AC?

 + Symmetric unification for equality?

 + Nonoriented eq back subsumes oriented eq (demodulator).
   (empty back demodulation inference)

 + unit conflict before weight test??  optional??

 + AC/C inference

 + pair algorithm

 + dependent options?

 + merge identical answer attributes?

 + factor_simp, back_unit_deletion

 */

#define PROVER_VERSION "5D"
#define VERSION_DATE "15 April 2003"

#include "header.h"
	  
#include "clock.h"
#include "options.h"
#include "resolve.h"
#include "paramod.h"
#include "ioutil.h"
#include "subsume.h"
#include "lindex.h"
#include "clause2.h"
#include "backdemod.h"
#include "detach.h"
#include "commands.h"
#include "hints.h"
#include "sos.h"
#include "nonport.h"
#include "weight.h"
#include "hash.h"
#include "symbolcheck.h"
#include "forward_subsume.h"
#include "bsub_conflict.h"
#include "control_mem.h"
#include "maximal.h"
#include "demodulate.h"

// Global Variables (all are for this file only, i.e., static)

// Clause Lists

static Clist Sos;
static Clist Usable;
static Clist Demodulators;
static Clist Limbo;          // Kept, but not yet in Sos
static Clist Disabled;       // Back subsumed, back demodulated
static Clist Hints;

// Term Lists

static Plist Weights;

// Indexes (most indexes are hidden in packages)

static Lindex Clashable_idx;        // resolution inference rules, given alg

// Clause Statistics

static int Given             = 0;
static int Generated         = 0;
static int Kept              = 0;
static int Subsumed          = 0;
static int Weight_deleted    = 0;
static int Back_subsumed     = 0;
static int New_demodulators  = 0;
static int Back_demodulated  = 0;

// Clocks

static int Infer_clock;
static int Preprocess_clock;
static int Demod_clock;
static int Hints_clock;
static int Subsume_clock;
static int Back_subsume_clock;
static int Back_demod_clock;
static int Disable_clock;
static int Index_clock;
static int Pick_given_clock;
static int Conflict_clock;
static int Unit_del_clock;
static int Control_sos_clock;

// Flags

enum {
  BINARY_RESOLUTION,
  HYPER_RESOLUTION,
  UR_RESOLUTION,
  FACTOR,
  PARAMODULATION,
  BACK_DEMOD,
  PRINT_KEPT,
  PRINT_GEN,
  CHECK_FOR_LEAKS,
  BACK_SUBSUME,
  COND_DETACH,
  NO_FAPL,
  UNINDEX_BACK_DEMOD,
  PROCESS_INITIAL_SOS,
  PROLOG_STYLE_VARIABLES,
  DEGRADE_HINTS,
  JUSTIFICATION_LAST,
  UNIT_DELETION,
  ORDERED_RESOLUTION,
  DISABLE_CLOCKS,
  PARA_UNITS_ONLY,
  BASIC_PARAMODULATION
};

// Parms

enum {
  MAX_KEPT,
  MAX_GIVEN,
  MAX_WEIGHT,
  MAX_MEGS,
  MAX_SECONDS,
  PICK_GIVEN_RATIO,
  BSUB_HINT_ADD_WT,
  BSUB_HINT_WT,
  LRS_SECONDS,
  CONSTANT_WEIGHT,
  VARIABLE_WEIGHT
};

// Clause attribute IDs

static int Label_att;
static int Answer_att;
static int Bsub_hint_add_wt_att;
static int Bsub_hint_wt_att;

// Internal Flags

static Uniftype Unification_type = ORDINARY_UNIF;
static BOOL     Use_clash_idx    = FALSE;
static int      Max_weight       = INT_MAX;  // starts as parm(MAX_WEIGHT)

/*************
 *
 *   symbol_sanity()
 *
 *   1. Make sure that symbols are being used in reasonable ways.
 *   2. If the user hasn't specified (or completely specified) a
 *      symbol ordering, do so now.
 *
 *************/

static
void symbol_sanity(void)
{
  Ilist function_syms, relation_syms, p;
  // Get lists of function and relation symbols from the clauses.
  Plist all_clauses = NULL;
  all_clauses = plist_prepend(all_clauses, Usable);
  all_clauses = plist_prepend(all_clauses, Sos);
  all_clauses = plist_prepend(all_clauses, Demodulators);
  function_syms = function_symbols_in_clists(all_clauses, NULL);
  relation_syms = relation_symbols_in_clists(all_clauses, NULL);
  zap_plist(all_clauses);

  // Check if any symbol is used as both a relation and function symbol.
  function_relation_check(function_syms, relation_syms, TRUE);

  // Check if any symbol is used with multiple arities.
  arity_check(function_syms, relation_syms, TRUE);

  // At this point, the symbol package doesn't know functions from relations.
  // We declare them here so that they can be ordered correctly.
  declare_functions_and_relations(function_syms, relation_syms);

  // This orders all symbols, taking into account the (maybe incomplete)
  // order already given in a user's lex command.
  lex_order_2(function_syms, relation_syms);

  function_syms = sort_by_lex_val(function_syms);
  printf("\nFunction symbol order:  ");
  for (p = function_syms; p; p = p->next)
    printf("%s/%d%s",
	   sn_to_str(p->i), sn_to_arity(p->i), p->next ? "  <  " : ".\n");
  
  relation_syms = sort_by_lex_val(relation_syms);
  printf("\nRelation symbol order:  ");
  for (p = relation_syms; p; p = p->next)
    printf("%s/%d%s",
	   sn_to_str(p->i), sn_to_arity(p->i), p->next ? "  <  " : ".\n");
  
  zap_ilist(function_syms);
  zap_ilist(relation_syms);
}  /* symbol_sanity */

/*************
 *
 *   index_clashable()
 *
 *************/

static
void index_clashable(Clause c, Indexop operation)
{
  if (Use_clash_idx) {
    clock_start(Index_clock);
    lindex_update_maximal(Clashable_idx, c, operation);
    clock_stop(Index_clock);
  }
}  /* index_clashable */

/*************
 *
 *   index_back_demod()
 *
 *************/

static
void index_back_demod(Clause c, Indexop operation)
{
  if (flag(BACK_DEMOD)) {
    clock_start(Index_clock);
    if (operation == INSERT || flag(UNINDEX_BACK_DEMOD))
      back_demod_index(c, operation);
    clock_stop(Index_clock);
  }
}  /* index_back_demod */

/*************
 *
 *   index_bsub_conflict()
 *
 *************/

static
void index_bsub_conflict(Clause c, Indexop operation)
{
  if (!flag(NO_FAPL) || !positive_clause(c)) {
    clock_start(Index_clock);
    bsub_index(c, operation);
    clock_stop(Index_clock);
  }
}  /* index_bsub_conflict */

/*************
 *
 *   index_fsub()
 *
 *************/

static
void index_fsub(Clause c, Indexop operation)
{
  clock_start(Index_clock);
  fsub_index(c, operation);
  clock_stop(Index_clock);
}  /* index_fsub */

/*************
 *
 *   index_demod()
 *
 *************/

static
void index_demod(Clause c, Indexop operation)
{
  clock_start(Index_clock);
  index_demodulator(c, operation);
  clock_stop(Index_clock);
}  /* index_demod */

/*************
 *
 *   read_lists() -- read and echo lists of objects, up to EOF
 *
 *   Supported clause lists: Usable, Sos, Demodulators, Hints
 *   Supported term lists: Weight.
 *
 *************/

static
void read_lists(FILE *fin, FILE *fout)
{
  Term t;
  while ((t = read_term(fin, fout))) {
    if (!is_term(t, "list", 1))
      fatal_error("read_lists: bad list identifier");
    else {
      char *listname = term_symbol(ARG(t,0));
      if (str_ident(listname, "usable")) {
	Clist l = read_clause_list(fin, fout, listname, TRUE);
	fwrite_clause_list(fout, l, TRUE, flag(JUSTIFICATION_LAST));
	clist_append_all(Usable, l);
      }
      else if (str_ident(listname, "sos")) {
	// Don't assign IDs if we're going to process the clauses.
	Clist l = read_clause_list(fin, fout, listname,
				   flag(PROCESS_INITIAL_SOS) ? FALSE : TRUE);
	fwrite_clause_list(fout, l, TRUE, flag(JUSTIFICATION_LAST));
	clist_append_all(Sos, l);
      }
      else if (str_ident(listname, "demodulators")) {
	Clist l = read_clause_list(fin, fout, listname, TRUE);
	fwrite_clause_list(fout, l, TRUE, flag(JUSTIFICATION_LAST));
	clist_append_all(Demodulators, l);
      }
      else if (str_ident(listname, "hints")) {
	Clist l = read_clause_list(fin, fout, listname, TRUE);
	fwrite_clause_list(fout, l, TRUE, flag(JUSTIFICATION_LAST));
	clist_append_all(Hints, l);
      }
      else if (str_ident(listname, "weights")) {
	// This is a list of terms.
	Plist g = read_term_list(fin, fout);
	fwrite_term_list(fout, g, "weights");
	Weights = plist_cat(Weights, g);
      }
      zap_term(t);
    }
  }

  // Now that we've read all the clauses, we have all the symbols.

  symbol_sanity();

}  // read_lists

/*************
 *
 *   p_memory_stats()
 *
 *************/

void p_memory_stats(void)
{
  fprintf(stdout, "\nMegs malloced by tp_alloc(): %d.\n", megs_malloced());
  
  fprint_strbuf_mem(stdout, TRUE);
  fprint_symbols_mem(stdout, FALSE);
  fprint_parse_mem(stdout, FALSE);
  fprint_hash_mem(stdout, FALSE);
  fprint_term_mem(stdout, FALSE);
  fprint_attrib_mem(stdout, FALSE);
  fprint_glist_mem(stdout, FALSE);

  fprint_just_mem(stdout, FALSE);
  fprint_formula_mem(stdout, FALSE);
  fprint_clause_mem(stdout, FALSE);
  fprint_clist_mem(stdout, FALSE);

  fprint_unify_mem(stdout, FALSE);
  fprint_btm_mem(stdout, FALSE);
  fprint_btu_mem(stdout, FALSE);

  fprint_fpa_mem(stdout, FALSE);
  fprint_discrim_mem(stdout, FALSE);
  fprint_mindex_mem(stdout, FALSE);
  fprint_lindex_mem(stdout, FALSE);
  fprint_clash_mem(stdout, FALSE);

}  // p_memory_stats

/*************
 *
 *   inferences_to_make()
 *
 *************/

static
BOOL inferences_to_make(void)
{
  return Sos->first != NULL;
}  // inferences_to_make

/*************
 *
 *   disable_clause()
 *
 *************/

static
void disable_clause(Clause c)
{
  // Assume c is in Usable, Sos, or Disabled.
  // Also, c may be in Demodulators.
  //
  // Unindex c according to which lists it is on and
  // the flags that are set, remove c from the lists,
  // and append c to Disabled.  Make sure you don't
  // have a Clist_pos for c during the call, because
  // it will be freed during the call.

  clock_start(Disable_clock);

  if (clist_member(c, Demodulators)) {
    index_demod(c, DELETE);
    clist_remove(c, Demodulators);
  }
  
  if (clist_member(c, Usable)) {
    index_fsub(c, DELETE);
    index_bsub_conflict(c, DELETE);
    index_back_demod(c, DELETE);
    index_clashable(c, DELETE);
    clist_remove(c, Usable);
  }
  else if (clist_member(c, Sos)) {
    index_fsub(c, DELETE);
    index_bsub_conflict(c, DELETE);
    index_back_demod(c, DELETE);
    remove_from_sos_lightest(c, Sos);  // remove from sos and update "lightest"
  }
  else if (clist_member(c, Limbo)) {
    index_fsub(c, DELETE);
    index_bsub_conflict(c, DELETE);
    clist_remove(c, Limbo);
  }
  else
    fatal_error("disable_clause: where IS the clause?");

  clist_append(c, Disabled);
  clock_stop(Disable_clock);
}  // disable_clause

/*************
 *
 *   free_everything()
 *
 *   This frees all memory so that we can check for memory leaks.
 *
 *************/

static
void free_everything(void)
{
  // Demodulators

  while (Demodulators->first) {
    Clause c = Demodulators->first->c;
    index_demod(c, DELETE);
    clist_remove(c, Demodulators);
    if (c->containers == NULL)
      delete_clause(c);
  }
  clist_free(Demodulators);
  
  destroy_demodulation_index();

  // Lightest

  zap_sos_lightest();

  // Usable, Sos, Limbo

  while (Usable->first)
    disable_clause(Usable->first->c);
  clist_free(Usable);

  while (Sos->first)
    disable_clause(Sos->first->c);
  clist_free(Sos);

  while (Limbo->first)
    disable_clause(Limbo->first->c);
  clist_free(Limbo);

  fsub_destroy_index();
  bsub_destroy_index();
  destroy_back_demod_index();
  lindex_destroy(Clashable_idx);

  delete_clist(Disabled);

  if (Hints->first) {
    Clist_pos p;
    for(p = Hints->first; p; p = p->next)
      unindex_hint(p->c);
    done_with_hints();
  }
  delete_clist(Hints);

  empty_sym_tab();
  zap_variables();
}  // free_everything

/*************
 *
 *   p_stats()
 *
 *************/

static
void p_stats()
{
  p_memory_stats();

  printf("\nGiven=%d, Generated=%d, Kept=%d.\n", Given, Generated, Kept);
  printf("Weight_deleted=%d, Subsumed=%d, Back_subsumed=%d.\n",
	 Weight_deleted, Subsumed, Back_subsumed);
  printf("New_demodulators=%d, Back_demodulated=%d.\n",
	 New_demodulators, Back_demodulated);

  printf("\nDemod_attempts=%d, Demod_rewrites=%d.\n",
	 demod_attempts(), demod_rewrites());

  printf("\nuser_CPU=%.2f, system_CPU=%.2f.\n\n",
	 user_time() / 1000.,
	 system_time() / 1000.);

  fprint_all_clocks(stdout);

  printf("\nnonunit_subsumption_tests=%d.\n\n",
	 nonunit_subsumption_tests());

  fflush(stdout);

}  // p_stats

/*************
 *
 *   exit_with_message()
 *
 *************/

static
void exit_with_message(char *message)
{
  if (message)
    fprintf(stderr, "\n------ %s exit ------\n", message);
  bell(stderr);

  if (flag(CHECK_FOR_LEAKS))
    free_everything();

  p_stats();
  printf("\nJob %d finished (%s) %s", my_process_id(), message, get_date());
  exit(1);
}  /* exit_with_message */

/*************
 *
 *   check_limits()
 *
 *************/

static
char *check_limits()
{
  if (Kept > parm(MAX_KEPT))
    return "max_kept";
  else if (Given > parm(MAX_GIVEN))
    return "max_given";
  else if (megs_malloced() > parm(MAX_MEGS))
    return "max_megs";
  else if (user_time() / 1000 > parm(MAX_SECONDS))
    return "max_seconds";
  else
    return NULL;
}  // check_limits

/*************
 *
 *   exit_if_over_limit()
 *
 *************/

static
void exit_if_over_limit(void)
{
  char *s = check_limits();
  if (s != NULL)
    exit_with_message(s);
}  // exit_if_over_limit

/*************
 *
 *   print_proof_and_exit()
 *
 *************/

static
void print_proof_and_exit(Clause empty_clause)
{
  Plist ancestors = get_clause_ancestors(empty_clause);
  printf("\n-------- PROOF -------\n");
  fprintf(stderr, "-------- PROOF -------\n"); bell(stderr);
  printf("Length of proof is %d.\n", proof_length(ancestors));
  while (ancestors != NULL) {
    fwrite_clause(stdout, ancestors->v, TRUE, flag(JUSTIFICATION_LAST));
    ancestors = plist_pop(ancestors);
  }
  printf("\n-------- end of proof  -------\n");
  exit_with_message("proof");
}  // print_proof_and_exit

/*************
 *
 *   pre_process()
 *
 *   Process a newly inferred (or input) clause.  If the empty clause is
 *   detected, print proof and exit.
 *
 *   It is likely that a pointer to this routine was passed to
 *   an inference rule, and that inference rule called this routine
 *   with a new clause.
 *
 *   If this routine decides to keep the clause, it is appended
 *   to the Limbo list rather than to Sos.  Clauses in Limbo have been
 *   kept, but operations that can delete clauses (back subsumption,
 *   back demoulation) have not yet been applied.  The Limbo list
 *   is processed after the inference rule is finished.
 *
 *   Why use the Limbo list?  Because we're not allowed to delete
 *   clauses (back subsumption, back demodulation) while inferring
 *   clauses.
 *
 *   Why not infer the whole batch of clauses, and then process them?
 *   Because there can be too many.  We have to do demodulation and
 *   subsumption right away, and get kept clauses indexed for
 *   forward demodulation and forward subsumption right away,
 *   so they can be used on the next inferred clause.
 *
 *************/

void pre_process(Clause c)
{
  // If the Infer_clock is running, stop it and restart it when done.

  BOOL infer_clock_stopped = FALSE;
  exit_if_over_limit();  /* exit if over some limit */

  if (clock_running(Infer_clock)) {
    clock_stop(Infer_clock);
    infer_clock_stopped = TRUE;
  }
  clock_start(Preprocess_clock);

  Generated++;
  if (flag(PRINT_GEN))
    fwrite_clause(stdout, c, TRUE, flag(JUSTIFICATION_LAST));

  if (!clist_empty(Demodulators)) {
    clock_start(Demod_clock);
    demodulate_clause(c);
    clock_stop(Demod_clock);
  }

  simplify_literals(c);  // Remove any instances of x!=x.

  orient_equalities(c);

  merge_literals(c);

  if (flag(UNIT_DELETION)) {
    clock_start(Unit_del_clock);
    unit_deletion(c);
    clock_stop(Unit_del_clock);
  }

  if (tautology(c) || eq_tautology(c)) {
    Subsumed++;
    delete_clause(c);
  }
  else if (number_of_literals(c) == 0) {
    assign_clause_id(c);
    print_proof_and_exit(c);
  }
  else {
    c->weight = clause_weight(c);
    if (!clist_empty(Hints)) {
      // This can be expensive.  Should we do subsumption first??
      clock_start(Hints_clock);
      renumber_variables(c, MAX_VARS);
      adjust_weight_with_hints(c,
			       Bsub_hint_add_wt_att,
			       Bsub_hint_wt_att,
			       Label_att,
			       flag(DEGRADE_HINTS));
      clock_stop(Hints_clock);
    }
    if (c->weight > Max_weight) {
      Weight_deleted++;
      delete_clause(c);
    }
    else {
      Clause subsumer;
      clock_start(Subsume_clock);
      subsumer = forward_subsumption(c);
      clock_stop(Subsume_clock);
      if (subsumer != NULL) {
	Subsumed++;
	delete_clause(c);
      }
      else {
	Clause empty_clause;
	Kept++;
	if (clist_empty(Hints))
	  renumber_variables(c, MAX_VARS);
	assign_clause_id(c);
	if (flag(ORDERED_RESOLUTION))
	  mark_non_maximal_literals(c);
	new_atom_nodes(c);  /* this makes FPA indexing behave well */
	if (flag(PRINT_KEPT)) {
	  printf("KEPT: ");
	  // p_clause(c);
	  fwrite_clause(stdout, c, TRUE, flag(JUSTIFICATION_LAST));
	}

	clock_start(Conflict_clock);
	empty_clause = unit_conflict(c);
	clock_stop(Conflict_clock);
	if (empty_clause) {
	  assign_clause_id(empty_clause);
	  print_proof_and_exit(empty_clause);
	}
	else {
	  if (flag(BACK_DEMOD) && pos_eq_unit(c) &&
	      oriented_eq(c->literals->atom)) {
	    if (flag(PRINT_KEPT))
	      printf("    new demodulator: %d.\n", c->id);
	    clist_append(c, Demodulators);
	    index_demod(c, INSERT);
	    New_demodulators++;
	  }
	  index_fsub(c, INSERT);
	  // bsub index is used for unit conflict and for back subsumption.
	  // We insert c into the bsub index now so that it can be
	  // available for unit conflict while it's in limbo.  It will also
	  // be available for back subsumption, but it should not be
	  // back subsumed while in limbo.  See fatal error in limbo_process.
	  index_bsub_conflict(c, INSERT);
	  clist_append(c, Limbo);
	}  // no unit conflict
      }  // not subsumed
    }  // not deleted by weight
  }  // not empty clause
  clock_stop(Preprocess_clock);
  if (infer_clock_stopped)
    clock_start(Infer_clock);
}  // pre_process

/*************
 *
 *   back_demod()
 *
 *   For each clause that can be back demodulated, make a
 *   copy, disable the original, pre_process the copy.
 *
 *************/

static
void back_demod(Clause demod)
{
  Plist results, p, prev;

  results = back_demodulatable(demod);
  p = results;
  while(p != NULL) {
    Clause old = p->v;
    if (!clist_member(old, Disabled)) {
      Clause new;
      if (flag(BASIC_PARAMODULATION))
	new = copy_clause_with_flag(old, nonbasic_flag());
      else
	new = copy_clause(old);
      upward_clause_links(new);
      Back_demodulated++;
      if (flag(PRINT_KEPT))
	printf("        %d back demodulating %d.\n", demod->id, old->id);
      new->justification = back_demod_just(old);
      new->attributes = copy_inheritable_attributes(old->attributes);
      disable_clause(old);
      pre_process(new);
    }
    prev = p;
    p = p->next;
    free_plist(prev);
  }
}  // back_demod

/*************
 *
 *   limbo_process()
 *
 *   Apply back subsumption and back demodulation to the Limbo
 *   list.  Since back demodulated clauses are pre_processed,
 *   the Limbo list can grow while it is being processed.
 *
 *   If this prover had a hot-list, or any other feature that
 *   generates clauses from newly kept clauses, it probably would
 *   be done here.
 *
 *   The Limbo list operates as a queue.  An important property
 *   of the Limbo list is that if A is ahead of B, then A does
 *   not simplify or subsume B.  However, B can simplify or subsume A.
 *
 *************/

static
void limbo_process(void)
{
  while (Limbo->first) {
    Clause c = Limbo->first->c;

    // back subsumption

    if (flag(BACK_SUBSUME)) {
      Plist subsumees;
      clock_start(Back_subsume_clock);
      subsumees = back_subsumption(c);
      while (subsumees != NULL) {
	Clause d = subsumees->v;
	if (clist_member(d, Limbo))  // See comment in pre_process.
	  fatal_error("back subsume limbo clause");
	Back_subsumed++;
	if (flag(PRINT_KEPT))
	  printf("    %d back subsumes %d.\n", c->id, d->id);
	disable_clause(d);
	subsumees = plist_pop(subsumees);
      }
      clock_stop(Back_subsume_clock);
    }

    // back demodulation

    if (clist_member(c, Demodulators)) {
      clock_start(Back_demod_clock);
      if (flag(PRINT_KEPT))
	printf("    starting back demodulation with %d.\n", c->id);
      back_demod(c);  // This calls pre_process on rewritable clauses.
      clock_stop(Back_demod_clock);
    }

    // move to Sos and index for back demodulation

    clist_remove(c, Limbo);
    insert_into_sos_lightest(c, Sos);
    index_back_demod(c, INSERT);
  }
}  // limbo_process

/*************
 *
 *   given_infer()
 *
 *   Make given_clause inferences according to the flags that are set.
 *   Inferred clauses are sent to pre_process().
 *
 *   We could process the Limbo list after each inference rule,
 *   and this might improve performance in some cases.  But it seems
 *   conceptually simplier if we process the Limbo clauses after
 *   all of the inferences have been made.
 *
 *************/

static
void given_infer(Clause given)
{
  clock_start(Infer_clock);

  if (flag(BINARY_RESOLUTION))
    binary_resolution(given, Clashable_idx, pre_process);

  if (flag(FACTOR))
    binary_factors(given, Unification_type, pre_process);

  if (flag(HYPER_RESOLUTION))
    hyper_resolution(given, Clashable_idx, pre_process);

  if (flag(UR_RESOLUTION))
    ur_resolution(given, Clashable_idx, pre_process);

  if (flag(COND_DETACH)) {
    // The built-in condensed detachment does not use indexing.
    Clist_pos p;
#if 1
    // This way generates clasues in the same order as Otter.
    for (p = Usable->last; p; p = p->prev)
      cond_detach(p->c, given, pre_process);
    for (p = Usable->last; p; p = p->prev)
      cond_detach(given, p->c, pre_process);
#else
    // This way scans Usable only once.
    for (p = Usable->first; p; p = p->next) {
      cond_detach(given, p->c, pre_process);
      cond_detach(p->c, given, pre_process);
    }
#endif
  }

  if (flag(PARAMODULATION) &&
      (!flag(PARA_UNITS_ONLY) || unit_clause(given))) {
    // This paramodulation does not use indexing.
    Clist_pos p;
    for (p = Usable->first; p != NULL; p = p->next) {
      if (unit_clause(p->c)) {
	para_from_into(given, p->c, PARA_ALL,            pre_process);
	para_from_into(p->c, given, PARA_ALL_EXCEPT_TOP, pre_process);
      }
    }
  }
  clock_stop(Infer_clock);
}  // given_infer

/*************
 *
 *   get_given_clause()
 *
 *************/

static
Clause get_given_clause(void)
{
  Clause c;
  int ratio = parm(PICK_GIVEN_RATIO);
  if (ratio != -1 && Given % (ratio + 1) == 0)
    c = extract_first_clause(Sos);
  else
    c = extract_lightest_clause(Sos);
  return c;
}  // get_given_clause

/*************
 *
 *   make_inferences()
 *
 *   Assume that there are inferences to make.
 *
 *   If we had the option of using the pair algorithm instead of
 *   the given algorithm, we could make that decision here.
 *
 *************/

void make_inferences(void)
{
  Clause given_clause;

  clock_start(Pick_given_clock);
  given_clause = get_given_clause();
  clock_stop(Pick_given_clock);

  if (given_clause != NULL) {
    Given++;
    // Apply limited-resource strategy every 10 given clasues.
    if (parm(LRS_SECONDS) > 0 && Given % 10  == 0) {
      clock_start(Control_sos_clock);
      Max_weight = control_sos_2(Sos,
				 Given,
				 parm(LRS_SECONDS) * 1000,
				 Max_weight,
				 parm(PICK_GIVEN_RATIO),
				 disable_clause);
      clock_stop(Control_sos_clock);
    }
    printf("\ngiven #%d (wt=%d): ", Given, given_clause->weight);
    fwrite_clause(stdout, given_clause, TRUE, flag(JUSTIFICATION_LAST));
    clist_append(given_clause, Usable);
    index_clashable(given_clause, INSERT);
    given_infer(given_clause);
  }
}  // make_inferences

/*************
 *
 *    init_prover_options()
 *
 *************/

static
void init_prover_options(void)
{
  //        ID                       name                    default

  init_flag(BINARY_RESOLUTION,      "binary_resolution",      FALSE);
  init_flag(HYPER_RESOLUTION,       "hyper_resolution",       FALSE);
  init_flag(UR_RESOLUTION,          "ur_resolution",          FALSE);
  init_flag(PARAMODULATION,         "paramodulation",         FALSE);
  init_flag(COND_DETACH,            "cond_detach",            FALSE);
  init_flag(ORDERED_RESOLUTION,     "ordered_resolution",     TRUE);
  init_flag(PARA_UNITS_ONLY,        "para_units_only",        FALSE);
  init_flag(BASIC_PARAMODULATION,   "basic_paramodulation",   FALSE);

  init_flag(PRINT_GEN,              "print_gen",              FALSE);
  init_flag(PRINT_KEPT,             "print_kept",             FALSE);
  init_flag(UNIT_DELETION,          "unit_deletion",          FALSE);
  init_flag(FACTOR,                 "factor",                 FALSE);
  init_flag(BACK_SUBSUME,           "back_subsume",           TRUE);
  init_flag(BACK_DEMOD,             "back_demod",             FALSE);

  init_flag(PROCESS_INITIAL_SOS,    "process_initial_sos",    TRUE);
  init_flag(NO_FAPL,                "no_fapl",                FALSE);
  init_flag(UNINDEX_BACK_DEMOD,     "unindex_back_demod",     FALSE);

  init_flag(CHECK_FOR_LEAKS,        "check_for_leaks",        FALSE);
  init_flag(PROLOG_STYLE_VARIABLES, "prolog_style_variables", FALSE);
  init_flag(DEGRADE_HINTS,          "degrade_hints",          TRUE);
  init_flag(JUSTIFICATION_LAST,     "justification_last",     TRUE);
  init_flag(DISABLE_CLOCKS,         "disable_clocks",         FALSE);

  //        ID                 name               default       min       max

  init_parm(MAX_MEGS,         "max_megs",              96,        0,  INT_MAX);
  init_parm(MAX_SECONDS,      "max_seconds",      INT_MAX,        0,  INT_MAX);
  init_parm(MAX_KEPT,         "max_kept",         INT_MAX,        0,  INT_MAX);
  init_parm(MAX_GIVEN,        "max_given",        INT_MAX,        0,  INT_MAX);
  init_parm(MAX_WEIGHT,       "max_weight",       INT_MAX,  INT_MIN,  INT_MAX);
  init_parm(PICK_GIVEN_RATIO, "pick_given_ratio",      -1,       -1,  INT_MAX);
  init_parm(BSUB_HINT_ADD_WT, "bsub_hint_add_wt",   -1000,  INT_MIN,  INT_MAX);
  init_parm(BSUB_HINT_WT,     "bsub_hint_wt",     INT_MAX,  INT_MIN,  INT_MAX);
  init_parm(LRS_SECONDS,      "lrs_seconds",           -1,       -1,  INT_MAX);
  init_parm(CONSTANT_WEIGHT,  "constant_weight",        1,  INT_MIN,  INT_MAX);
  init_parm(VARIABLE_WEIGHT,  "variable_weight",        1,  INT_MIN,  INT_MAX);
}  // init_prover_options

/*************
 *
 *   init_prover_clocks()
 *
 *************/

static
void init_prover_clocks(void)
{
  Pick_given_clock    = clock_init("pick_given");
  Infer_clock         = clock_init("infer");
  Preprocess_clock    = clock_init("preprocess");
  Demod_clock         = clock_init("demod");
  Unit_del_clock      = clock_init("unit_deletion");
  Conflict_clock      = clock_init("conflict");
  Hints_clock         = clock_init("hints");
  Subsume_clock       = clock_init("subsume");
  Back_subsume_clock  = clock_init("back_subsume");
  Back_demod_clock    = clock_init("back_demod");
  Index_clock         = clock_init("index");
  Disable_clock       = clock_init("disable");
  Control_sos_clock   = clock_init("control_sos");

  wall_init();  // initialize wall-clock timer
}  // init_prover_clocks

/*************
 *
 *   init_prover_attributes()
 *
 *************/

static
void init_prover_attributes(void)
{
  Label_att            = register_attribute("label",   STRING_ATTRIBUTE);
  Bsub_hint_add_wt_att = register_attribute("bsub_hint_add_wt", INT_ATTRIBUTE);
  Bsub_hint_wt_att     = register_attribute("bsub_hint_wt", INT_ATTRIBUTE);
  Answer_att           = register_attribute("answer",  TERM_ATTRIBUTE);
  declare_term_attribute_inheritable(Answer_att);
}  // Init_prover_attributes

/*************
 *
 *   index_and_process_input_clauses()
 *
 *************/

static
void index_and_process_input_clauses(void)
{
  Clist_pos p;

  // Allocate and initialize indexes (even if they won't be used).

  init_fsub_index(DISCRIM_BIND, Unification_type, 0);
  init_bsub_index(FPA, Unification_type, 100);
  init_demodulator_index(DISCRIM_BIND, Unification_type, 0);
  init_back_demod_index(FPA, Unification_type, 100);

  Clashable_idx = lindex_init(FPA, Unification_type, 100,
			      FPA, Unification_type, 100);

  // Usable: index the clauses

  for (p = Usable->first; p != NULL; p = p->next) {
    Clause c = p->c;
    orient_equalities(c);
    if (flag(ORDERED_RESOLUTION))
      mark_non_maximal_literals(c);
    index_fsub(c, INSERT);
    index_bsub_conflict(c, INSERT);
    index_back_demod(c, INSERT);
    index_clashable(c, INSERT);
  }

  // Demodulators: index the clauses

  for (p = Demodulators->first; p != NULL; p = p->next) {
    Clause c = p->c;
    if (!pos_eq_unit(c))
      fatal_error("input demodulator is not equation");
    else {
      Term atom = c->literals->atom;
      Term alpha = ARG(atom,0);
      Term beta  = ARG(atom,1);
      orient_equalities(c);
      if (ARG(atom,0) != alpha) {
	printf("NOTE: input demodulator %d flipped.\n", c->id);
	fprintf(stderr, "NOTE: input demodulator %d flipped.\n", c->id);
	bell(stderr);
      }
      if (!oriented_eq(atom)) {
	if (term_ident(alpha, beta))
	  fatal_error("input demodulator is instance of x=x");
	if (!variables_subset(alpha, beta) && !variables_subset(beta, alpha))
	  fatal_error("input demoulator does not have variable-subset property");
	
	printf("NOTE: input demodulator %d is not orientable.\n", c->id);
	fprintf(stderr, "NOTE: input demodulator %d is not orientable.\n", c->id);
	bell(stderr);
      }
      index_demod(c, INSERT);
    }
  }

  // Hints (we have to do this before processing Sos).

  if (Hints->first) {
    init_hints();  // Initialize hints package (hints index).
    for (p = Hints->first; p != NULL; p = p->next) {
      Clause h = p->c;
      orient_equalities(h);
      // If the hint already has a weight attribute, it overrides the
      // following, because it is earlier in the list of attributes.
      if (parm(BSUB_HINT_ADD_WT) != 0)
	h->attributes = set_int_attribute(h->attributes,
					  Bsub_hint_add_wt_att,
					  parm(BSUB_HINT_ADD_WT));
      if (parm(BSUB_HINT_WT) != INT_MAX)
	h->attributes = set_int_attribute(h->attributes,
					  Bsub_hint_wt_att,
					  parm(BSUB_HINT_WT));
      index_hint(h);  // This is handled in the hints package.
    }
  }

  // Sos: either pre_process or index the clauses

  if (flag(PROCESS_INITIAL_SOS)) {
    // Treat them EXACTLY as if they had been inferred.
    while (Sos->first) {
      Clause c = Sos->first->c;
      clist_remove(c, Sos);
      pre_process(c);
    }
    // Sos is now empty.  This will put processed clauses back into Sos.
    limbo_process();  // back subsumption and back demodulation.

    printf("\nClauses after processing initial sos:\n");
    fwrite_clause_list(stdout, Usable,       TRUE, flag(JUSTIFICATION_LAST));
    fwrite_clause_list(stdout, Sos,          TRUE, flag(JUSTIFICATION_LAST));
    fwrite_clause_list(stdout, Demodulators, TRUE, flag(JUSTIFICATION_LAST));
  }
  else {
    // Move Sos to a temporary list, then traverse that temporary list,
    // putting the clauses back into Sos in the "correct" way.
    // We do this round-about operation because insert_into_sos_lightest
    // wants to put the clauses into Sos.  Also index the clauses.
    Clist temp = Sos;          // move Sos to a temporary list
    name_clist(temp, "temp");  // not really necessary
    Sos = clist_init("sos");   // get a new Sos list
    while (temp->first) {
      Clause c = temp->first->c;
      clist_remove(c, temp);

      orient_equalities(c);
      if (flag(ORDERED_RESOLUTION))
	mark_non_maximal_literals(c);
      c->weight = clause_weight(c);
      insert_into_sos_lightest(c, Sos);
      
      index_fsub(c, INSERT);
      index_bsub_conflict(c, INSERT);
      index_back_demod(c, INSERT);
    }
    clist_zap(temp);  // free the empty list
  }
}  // index_and_process_input_clauses

/*************
 *
 *   process_options_1()
 *
 *   If we have to do anything related to options, based on the
 *   commands input by the user, we can do it here.  Because some
 *   of this can affect how clauses are read (e.g., variable style)
 *   we assume that that only the commands have been read.
 *
 *************/

static
void process_options_1(void)
{
  Max_weight = parm(MAX_WEIGHT);

  // Index Usable clauses if hyper, UR, or binary-res are set.
  Use_clash_idx = (flag(BINARY_RESOLUTION) ||
		   flag(HYPER_RESOLUTION) ||
		   flag(UR_RESOLUTION));

  // If any AC or C symbols were declared, use backtrack-unification.
  if (assoc_comm_symbols() || comm_symbols())
    Unification_type = BACKTRACK_UNIF;

  // Variable style.
  if (flag(PROLOG_STYLE_VARIABLES))
    set_variable_style(PROLOG_STYLE);

  // Tell the clock package if the user doesn't want timing.
  if (flag(DISABLE_CLOCKS))
    disable_clocks();

  // Tell the basic paramodulation package if the flag is set.
  if (flag(BASIC_PARAMODULATION))
    set_basic_paramod();

  // Tell the memory package when to quit in case we don't catch it in time.
  set_max_megs(parm(MAX_MEGS));

}  // process_options_1

/*************
 *
 *   process_options_2()
 *
 *   This is meant to be called after all input has been read.
 *
 *************/

static
void process_options_2(void)
{
  if (parm(CONSTANT_WEIGHT) != 1)
    set_weights_by_arity(0, parm(CONSTANT_WEIGHT));
  if (parm(VARIABLE_WEIGHT) != 1)
    set_variable_weight(parm(VARIABLE_WEIGHT));
}  // process_options_2

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

static
void print_banner(int argc, char **argv)
{
  int i;
  printf("----- Prover %s, %s -----\n", PROVER_VERSION, VERSION_DATE);
  printf("Process %d was started by %s on %s,\n%s",
	 my_process_id(), username(), hostname(), get_date());
	 
  printf("The command was \"");
  for(i = 0; i < argc; i++)
    printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
  printf("\".\n\n");
}  // print_banner

/*************
 *
 *    main -- a theorem prover (given clause algorithm)
 *
 *************/

int main(int argc, char **argv)
{
  Term t;

  print_banner(argc, argv);

  declare_standard_parse_types();   // LADR routine
  translate_neg_equalities(TRUE);   // LADR routine
  init_options();                   // LADR routine

  init_prover_options();            // local routine
  init_prover_clocks();             // local routine
  init_prover_attributes();         // local routine

  t = read_commands(stdin, stdout, TRUE, WARN_UNKNOWN);
  if (!end_of_commands_term(t))
    fatal_error("main: commands must end with end_of_commands");
  if (t != NULL)
    zap_term(t);

  process_options_1();               // local routine

  // Allocate main clause lists (ones that can be input).

  Usable       = clist_init("usable");
  Sos          = clist_init("sos");
  Demodulators = clist_init("demodulators");
  Hints        = clist_init("hints");

  // Read and echo the rest of the input (Usable, Sos, etc).

  read_lists(stdin, stdout);

  // Initialize the weighting package (ok if Weights NULL).

  init_weight_scheme(Weights);
  zap_term_list(Weights);
  Weights = NULL;

  // Allocate auxiliary clause lists.

  Limbo        = clist_init("limbo");
  Disabled     = clist_init("disabled");

  process_options_2();               // local routine

  index_and_process_input_clauses();

  printf("\nStarting search.\n");

  // ****************************** Main Loop ******************************

  while (inferences_to_make()) {

    // make_inferences: each inferred clause is pre_processed, which
    // does forward demodulation and subsumption; if the clause is kept
    // it is put on the Limbo list, and it is indexed so that it can be
    // used immediately with subsequent newly inferred clauses.
	 
    make_inferences();

    // limbo_process: this applies back subsumption and back demodulation,
    // and moves the clauses to the Sos list.  The reason for the Limbo
    // list is that we cannot disable any clauses while inferences are being
    // made.

    limbo_process();

  }  // ************************ end of main loop ************************

  exit_with_message("sos_empty");

  exit(1);
}  // main
