#include "resolve.h"

/* Private definitions and types */

/*************
 *
 *   unit_check()
 *
 *************/

static
BOOL unit_check(Literal lit)
{
  return unit_clause(lit->atom->container);
}  /* unit_check */

/*************
 *
 *   positive_check()
 *
 *************/

static
BOOL positive_check(Literal lit)
{
  return positive_clause(lit->atom->container) && maximal_literal(lit);
}  /* positive_check */

/*************
 *
 *   first_negative_maximal_literal()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Literal first_negative_maximal_literal(Clause c)
{
  Literal lit = c->literals;
  while (lit && (lit->sign || !maximal_literal(lit)))
    lit = lit->next;
  return lit;
}  /* first_negative_maximal_literal */

/*************
 *
 *   binary_check()
 *
 *************/

static
BOOL binary_check(Literal lit)
{
#if 0
  return maximal_literal(lit);
#else
  Clause c = lit->atom->container;
  if (positive_clause(c))
    return maximal_literal(lit);
  else {
    return first_negative_maximal_literal(c) == lit;
  }
#endif
}  /* binary_check */

/*************
 *
 *   hyper_satellite()
 *
 *************/

static
void hyper_satellite(Clause c, Lindex idx,
		      void (*proc_proc) (Clause))
{
  /* Assume c is positive. */
  Context sat_subst = get_context();
  Context nuc_subst = get_context();
  Literal slit;
  for (slit = c->literals; slit != NULL; slit = slit->next) {
    if (maximal_literal(slit)) {
      /* for each acceptable nucleus */
      Term fnd_atom;
      Mindex_pos mate_pos;
      fnd_atom = mindex_retrieve_first(slit->atom, idx->neg, UNIFY,
				       sat_subst, nuc_subst,
				       FALSE, &mate_pos);
      while (fnd_atom != NULL) {
	Clause nuc = fnd_atom->container;
	Literal nuc_lit = atom_to_literal(fnd_atom);
	if (!positive_clause(nuc)) {
	  /* We have a nucleus.  Set up the clash list, then clash. */
	  Clash p = NULL;
	  Clash first = NULL;
	  Literal nlit;
	  for (nlit = nuc->literals; nlit != NULL; nlit = nlit->next) {
	    p = append_clash(p);
	    if (first == NULL)
	      first = p;
	    p->nuc_lit = nlit;
	    p->nuc_subst = nuc_subst;
	    if (nlit == nuc_lit) {
	      p->sat_lit = slit;
	      p->sat_subst = sat_subst;
	      p->clashable = TRUE;
	      p->clashed = TRUE;
	    }
	    else {
	      p->clashable = !nlit->sign;
	      if (p->clashable) {
		p->mate_index = idx->pos;
		p->sat_subst = get_context();
	      }
	    }
	  }  /* for each literal of nucleus */
	  clash(first, positive_check, HYPER_RES_JUST, proc_proc);
	  zap_clash(first);
	}  /* if we have a nuc */
	fnd_atom = mindex_retrieve_next(mate_pos);
      }  /* for each mate */
    }  /* if sat is ok */
  }  /* for each literal of satellite */
  free_context(sat_subst);
  free_context(nuc_subst);
}  /* hyper_satellite */

/*************
 *
 *   hyper_nucleus()
 *
 *************/

static
void hyper_nucleus(Clause c, Lindex idx,
		    void (*proc_proc) (Clause))
{
  /* Assume c is not positive. */
  Clash p = NULL;
  Clash first = NULL;
  Literal lit;
  Context nuc_subst = get_context();
  for (lit = c->literals; lit != NULL; lit = lit->next) {
    p = append_clash(p);
    if (first == NULL)
      first = p;
    p->clashable = !lit->sign;
    p->nuc_lit = lit;
    p->nuc_subst = nuc_subst;
    if (p->clashable) {
      p->mate_index = idx->pos;
      p->sat_subst = get_context();
    }
  }
  clash(first, positive_check, HYPER_RES_JUST, proc_proc);
  free_context(nuc_subst);
  zap_clash(first);  /* This also frees satellite contexts. */
}  /* hyper_nucleus */

/*************
 *
 *   hyper_resolution()
 *
 *************/

/* DOCUMENTATION
Hyperresolution.
*/

/* PUBLIC */
void hyper_resolution(Clause c, Lindex idx,
		      void (*proc_proc) (Clause))
{
  if (positive_clause(c))
    hyper_satellite(c, idx, proc_proc);
  else
    hyper_nucleus(c, idx, proc_proc);
}  /* hyper_resolution */

/*************
 *
 *   ur_satellite()
 *
 *************/

static
void ur_satellite(Clause c, Lindex idx,
		  void (*proc_proc) (Clause))

{
  /* Assume C is a unit. */
  Context sat_subst = get_context();
  Context nuc_subst = get_context();
  Term fnd_atom;
  Mindex_pos mate_pos;
  Literal slit = c->literals;

  fnd_atom = mindex_retrieve_first(slit->atom, (slit->sign?idx->neg:idx->pos),
				   UNIFY, sat_subst, nuc_subst,
				   FALSE, &mate_pos);
  while (fnd_atom != NULL) {
    Clause nuc = fnd_atom->container;
    Literal fnd_lit = atom_to_literal(fnd_atom);
    if (!unit_clause(nuc)) {
      Literal target;
      for (target=nuc->literals; target!=NULL; target=target->next) {
	if (target != fnd_lit) {
	  Clash p = NULL;
	  Clash first = NULL;
	  Literal nlit;
	  for (nlit = nuc->literals; nlit != NULL; nlit = nlit->next) {
	    p = append_clash(p);
	    if (first == NULL)
	      first = p;
	    p->nuc_lit = nlit;
	    p->nuc_subst = nuc_subst;
	    if (nlit == fnd_lit) {
	      p->sat_lit = slit;
	      p->sat_subst = sat_subst;
	      p->clashable = TRUE;
	      p->clashed = TRUE;
	    }
	    else {
	      p->clashable = (nlit != target);
	      if (p->clashable) {
		p->mate_index = (nlit->sign ? idx->neg : idx->pos);
		p->sat_subst = get_context();
	      }
	    }
	  }  /* for each literal of nucleus */
	  clash(first, unit_check, UR_RES_JUST, proc_proc);
	  zap_clash(first);
	}  
      }  /* for each target */
    }  /* if we have a nuc */
    fnd_atom = mindex_retrieve_next(mate_pos);
  }  /* for each mate */
  free_context(sat_subst);
  free_context(nuc_subst);
}  /* ur_satellite */

/*************
 *
 *   ur_nucleus()
 *
 *************/

static
void ur_nucleus(Clause c, Lindex idx,
		 void (*proc_proc) (Clause))
{
  Literal target;
  for (target = c->literals; target != NULL; target = target->next) {
    Clash p = NULL;
    Clash first = NULL;
    Literal lit;
    Context nuc_subst = get_context();
    for (lit = c->literals; lit != NULL; lit = lit->next) {
      p = append_clash(p);
      if (first == NULL)
	first = p;
      p->clashable = (lit != target);
      p->nuc_lit = lit;
      p->nuc_subst = nuc_subst;
      if (p->clashable) {
	p->mate_index = (lit->sign ? idx->neg : idx->pos);
	p->sat_subst = get_context();
      }
    }
    clash(first, unit_check, UR_RES_JUST, proc_proc);
    free_context(nuc_subst);
    zap_clash(first);  /* This also frees satellite contexts. */
  }
}  /* ur_nucleus */

/*************
 *
 *   ur_resolution()
 *
 *************/

/* DOCUMENTATION
Unit-resulting resolution.
*/

/* PUBLIC */
void ur_resolution(Clause c, Lindex idx,
		   void (*proc_proc) (Clause))
{
  if (unit_clause(c))
    ur_satellite(c, idx, proc_proc);
  else
    ur_nucleus(c, idx, proc_proc);
}  /* ur_resolution */

/*************
 *
 *   binary_resolution()
 *
 *************/

/* DOCUMENTATION
Binary resolution.
*/

/* PUBLIC */
void binary_resolution(Clause c, Lindex idx,
		       void (*proc_proc) (Clause))
{
  Literal lit;
  for (lit = c->literals; lit != NULL; lit = lit->next) {
    if (binary_check(lit)) {
      Clash p = NULL;
      Clash first = NULL;
      Literal l2;
      Context nuc_subst = get_context();
      for (l2 = c->literals; l2 != NULL; l2 = l2->next) {
	p = append_clash(p);
	if (first == NULL)
	  first = p;
	p->clashable = (l2 == lit);
	p->nuc_lit = l2;
	p->nuc_subst = nuc_subst;
	if (p->clashable) {
	  p->mate_index = (lit->sign ? idx->neg : idx->pos);
	  p->sat_subst = get_context();
	}
      }
      clash(first, binary_check, BINARY_RES_JUST, proc_proc);
      free_context(nuc_subst);
      zap_clash(first);  /* This also frees satellite contexts. */
    }
  }
}  /* binary_resolution */

/*************
 *
 *   binary_factors()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void binary_factors(Clause c, Uniftype utype,
		    void (*proc_proc) (Clause))
{
  Literal l1;
  int i = 1;
  Context subst = get_context();
  for (l1 = c->literals; l1 != NULL; l1 = l1->next, i++) {
    Literal l2;
    int j = i+1;
    for (l2 = l1->next; l2 != NULL; l2 = l2->next, j++) {

      if (utype == ORDINARY_UNIF) {
	Trail tr = NULL;
	if (l1->sign == l2->sign && unify(l1->atom,subst,l2->atom,subst,&tr)) {
	  Clause f = get_clause();
	  Literal l3;
	  f->justification = factor_just(c, i, j);
	  for (l3 = c->literals; l3 != NULL; l3 = l3->next) {
	    if (l3 != l2)
	      append_literal(f, apply_lit(l3, subst));
	  }
	  undo_subst(tr);
	  upward_clause_links(f);
	  f->attributes = transfer_inheritable_attributes(c->attributes,
							  subst,f->attributes);
	  (*proc_proc)(f);
	}
      }  /* ordinary unification */
      else {
	/* backtrack unification */
	Btu_state bt = unify_bt_first(l1->atom, subst, l2->atom, subst);
	while (bt != NULL) {
	  Clause f = get_clause();
	  Literal l3;
	  f->justification = factor_just(c, i, j);
	  for (l3 = c->literals; l3 != NULL; l3 = l3->next) {
	    if (l3 != l2)
	      append_literal(f, apply_lit(l3, subst));
	  }
	  upward_clause_links(f);
	  f->attributes = transfer_inheritable_attributes(c->attributes,
							  subst,f->attributes);
	  (*proc_proc)(f);
	  bt = unify_bt_next(bt);
	}
      }  /* backtrack unification */
    }
  }
  free_context(subst);
}  /* binary_factors */

/*************
 *
 *   merge_literals()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void merge_literals(Clause c)
{
  Literal l1;
  int n1;
  BOOL null_literals = FALSE;
  for (l1 = c->literals, n1 = 1; l1; l1 = l1->next, n1++) {
    if (l1->atom) {
      Literal l2;
      int n2;
      for (l2 = l1->next, n2 = n1+1; l2; l2 = l2->next, n2++) {
	if (l2->atom) {
	  if (l1->sign == l2->sign && term_ident(l1->atom, l2->atom)) {
	    zap_term(l2->atom);
	    c->justification = append_just(c->justification, merge_just(n2));
	    l2->atom = NULL;  /* remove it below */
	    null_literals = TRUE;
	  }
	}
      }
    }
  }
  if (null_literals)
    c->literals = remove_null_literals(c->literals);
}  /* merge_literals */
