#include "strbuf.h"

/*
 * private definitions and types
 */

struct string_buf {
  int size;
  struct chunk *first;
  struct string_buf *link;
};

#define CHUNK 500

struct chunk {
  char s[CHUNK];
  struct chunk *next;
};

typedef struct chunk *Chunk;

/*
 * memory management
 */

static Chunk chunk_avail;
static unsigned chunk_gets, chunk_frees, chunk_avails;

static String_buf string_buf_avail;
static unsigned string_buf_gets, string_buf_frees, string_buf_avails;

/* end of counters for each type */

/*************
 *
 *    Chunk get_chunk()
 *
 *************/

static
Chunk get_chunk(void)
{
  Chunk p;

  chunk_gets++;
  if (chunk_avail == NULL)
    p = tp_alloc(sizeof(struct chunk));
  else {
    chunk_avails--;
    p = chunk_avail;
    chunk_avail = chunk_avail->next;
  }
  /* initialization */	
  if (0) {
    int i;
    for (i = 0; i < CHUNK; i++)
      p->s[i] = '\0';
  }
  	
  p->next = NULL;

  /* end of initialization */	
  return(p);
}  /* get_chunk */

/*************
 *
 *    free_chunk()
 *
 *************/

static
void free_chunk(Chunk p)
{
  chunk_frees++;
  chunk_avails++;
  p->next = chunk_avail;
  chunk_avail = p;
}  /* free_chunk */

/*************
 *
 *    String_buf get_string_buf()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns an empty String_buf.  Don't
forget to call zap_string_buf(sb) when finished with it.
Also see init_string_buf().
*/

/* PUBLIC */
String_buf get_string_buf(void)
{
  String_buf p;

  string_buf_gets++;
  if (string_buf_avail == NULL)
    p = tp_alloc(sizeof(struct string_buf));
  else {
    string_buf_avails--;
    p = string_buf_avail;
    string_buf_avail = string_buf_avail->link;
  }
  /* initialization */	
  p->first = NULL;
  p->size = 0;

  p->link = NULL;

  /* end of initialization */	
  return(p);
}  /* get_string_buf */

/*************
 *
 *    free_string_buf()
 *
 *************/

static
void free_string_buf(String_buf p)
{
  string_buf_frees++;
  string_buf_avails++;
  p->link = string_buf_avail;
  string_buf_avail = p;
}  /* free_string_buf */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_strbuf_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for String_bufs
and associated structures.  The argument heading tells whether
to print a heading on the table.
*/

/* PUBLIC */
void fprint_strbuf_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct chunk);
  fprintf(fp, "chunk (%4d)        %11u%11u%11u%11u%9.1f K\n",
	  n, chunk_gets, chunk_frees,
	  chunk_gets - chunk_frees, chunk_avails,
	  (((chunk_gets - chunk_frees) + chunk_avails) * n) / 1024.);

  n = sizeof(struct string_buf);
  fprintf(fp, "string_buf (%4d)   %11u%11u%11u%11u%9.1f K\n",
	  n, string_buf_gets, string_buf_frees,
	  string_buf_gets - string_buf_frees, string_buf_avails,
	  (((string_buf_gets - string_buf_frees) + string_buf_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_strbuf_mem */

/*************
 *
 *   p_strbuf_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for String_bufs
and associated structures.
*/

/* PUBLIC */
void p_strbuf_mem()
{
  fprint_strbuf_mem(stdout, 1);
}  /* p_strbuf_mem */

/*
 *  end of memory management
 */

/*************
 *
 *    String_buf init_string_buf()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a String_buf, initialized
to string s.  Don't forget to call zap_string_buf(sb) when
finished with it.
Also see get_string_buf().
*/

/* PUBLIC */
String_buf init_string_buf(char *s)
{
  String_buf p = get_string_buf();
  sb_append(p, s);
  return(p);
}  /* init_string_buf */

/*************
 *
 *   fprint_sb()
 *
 *************/

/* DOCUMENTATION
This routine prints String_buf sb to FILE *fp.
*/

/* PUBLIC */
void fprint_sb(FILE *fp, String_buf sb)
{
  Chunk h = sb->first;
  int i = 0;

  while (i < sb->size) {
    fprintf(fp, "%c", h->s[i % CHUNK]);
    i++;
    if (i % CHUNK == 0)
      h = h->next;
  }
}  /* fprint_sb */

/*************
 *
 *   p_sb()
 *
 *************/

/* DOCUMENTATION
This routine prints String_buf sb, followed by '\n' and fflush, to stdout.
If you don't want the newline, use fprint_sb() instead.
*/

/* PUBLIC */
void p_sb(String_buf sb)
{
  fprint_sb(stdout, sb);
  printf("\n");
  fflush(stdout);
}  /* p_sb */

/*************
 *
 *   sb_append()
 *
 *************/

/* DOCUMENTATION
This routine appends string s to String_buf sb.
The NULL character that marks the end of s does not go into
the String_buf.
*/

/* PUBLIC */
void sb_append(String_buf sb, char *s)
{
  int i;
  int n = sb->size;
  Chunk last = sb->first;

  while (last != NULL && last->next != NULL)
    last = last->next;

  for (i = 0; s[i] != '\0'; i++) {
    if (n % CHUNK == 0) {
      Chunk new = get_chunk();
      if (last != NULL)
	last->next = new;
      else
	sb->first = new;
      last = new;
    }
    last->s[n % CHUNK] = s[i];
    n++;
  }
  sb->size = n;
}  /* sb_append */

/*************
 *
 *   sb_append_char()
 *
 *************/

/* DOCUMENTATION
This routine appends character c to String_buf sb.
*/

/* PUBLIC */
void sb_append_char(String_buf sb, char c)
{
  char s[2];
  s[0] = c;
  s[1] = '\0';
  sb_append(sb, s);
}  /* sb_append_char */

/*************
 *
 *   sb_char()
 *
 *************/

/* DOCUMENTATION
This routine returns the n-th character (counting from 0) of String_buf sb.
If index n is out of range, the NULL character '\0' is returned.
*/

/* PUBLIC */
char sb_char(String_buf sb, int n)
{
  if (n < 0 || n >= sb->size)
    return '\0';
  else {
    Chunk h = sb->first;
    int i;
    for (i = 0; i < n / CHUNK; i++)
      h = h->next;
    return h->s[n % CHUNK];
  }
}  /* sb_char */

/*************
 *
 *   sb_cat_copy()
 *
 *************/

/* DOCUMENTATION
This routine appends a copy of sb2 to sb1.
String_buf sb2 is not changed.
You can use sb_cat() instead if you won't be needing sb2.
*/

/* PUBLIC */
void sb_cat_copy(String_buf sb1, String_buf sb2)
{
  /* Note that this is inefficient if there are many chunks in either sb. */
  int i;
  char c;

  for (i = 0; (c = sb_char(sb2, i)); i++)
    sb_append_char(sb1, c);
}  /* sb_cat_copy */

/*************
 *
 *   sb_cat()
 *
 *************/

/* DOCUMENTATION
This routine appends a copy of sb2 to sb1, then deallocates sb2.
Do not refer to sb2 after calling this rouine because it won't exist.
You can use sb_cat_copy() instead if you need to save sb2.
*/

/* PUBLIC */
void sb_cat(String_buf sb1, String_buf sb2)
{
  sb_cat_copy(sb1, sb2);
  zap_string_buf(sb2);
}  /* sb_cat */

/*************
 *
 *   zap_string_buf()
 *
 *************/

/* DOCUMENTATION
This routine deallocates a String_buf and frees all memory
associated with it.
*/

/* PUBLIC */
void zap_string_buf(String_buf sb)
{
  Chunk curr, prev;
  curr = sb->first;
  while (curr != NULL) {
    prev = curr;
    curr = curr->next;
    free_chunk(prev);
  }
  free_string_buf(sb);
}  /* zap_string_buf */

/*************
 *
 *   sb_to_new_string()
 *
 *************/

/* DOCUMENTATION
This routine returns a new, ordinary C string corresponding to the
String_buf argument sb.  WARNING: the new string, say s, is
dynamically allocated (malloced), so don't forget to call the system
routine free(s) when you are finished with the string.  (This routine
is not intended for printing String_bufs; use fprint_sb() instead.)
<P>
String_bufs do not have a NULL character marking the end;
instead, they keep a count of the number of characters.
*/

/* PUBLIC */
char *sb_to_new_string(String_buf sb)
{
  char *s = malloc(sb->size + 1);

  if (s == NULL)
    return NULL;
  else {
    int i;
    for (i = 0; i < sb->size; i++)
      s[i] = sb_char(sb, i);
    s[sb->size] = '\0';
    return s;
  }
}  /* sb_to_new_string */

/*************
 *
 *   sb_size()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int sb_size(String_buf sb)
{
  return sb->size;
}  /* sb_size */
