#include "subsume.h"

/* Private definitions and types */

static int Nonunit_subsumption_tests = 0;
static int Unit_deletion_failures = 0;

/* PUBLIC */
int nonunit_subsumption_tests(void)
{
  return Nonunit_subsumption_tests;
}

/* PUBLIC */
int unit_deletion_failures(void)
{
  return Unit_deletion_failures;
}

/*************
 *
 *   subsume_literals()
 *
 *   This is a recursive routine that tries to map a list of literals
 *   into a clause d.  If successful, the trail is updated; if failure,
 *   the trail is unchanged.
 *
 *   This version uses ordinary unification.
 *
 *************/

static
BOOL subsume_literals(Literal clit, Context subst, Clause d, Trail *trp)
{
  BOOL subsumed = FALSE;
  Literal dlit;

  if (clit == NULL)
    return TRUE;
  else {
    for (dlit = d->literals; !subsumed && dlit != NULL; dlit = dlit->next) {
      /* printf("testing "); fprint_term(stdout, clit->atom);
      printf("  "); fprint_term(stdout, dlit->atom); printf("\n"); */
      if (clit->sign == dlit->sign) {
	Trail mark = *trp;
	if (match(clit->atom, subst, dlit->atom, trp)) {
	  /* printf("                    Match\n"); */
	  if (subsume_literals(clit->next, subst, d, trp))
	    subsumed = TRUE;
	  else {
	    /* undo bindings from this match only */
	    /* printf("                    UNDO\n"); */
	    undo_subst_2(*trp, mark);
	    *trp = mark;
	  }
	}
      }
    }
    return subsumed;
  }
}  /* subsume_literals */

/*************
 *
 *   subsume_bt_literals()
 *
 *   This version uses backtrack unification.
 *
 *************/

static
BOOL subsume_bt_literals(Literal clit, Context subst,
			 Clause d, Plist *gp)
{
  if (clit == NULL)
    return TRUE;
  else {
    BOOL subsumed = FALSE;
    Literal dlit;
    *gp = plist_prepend(*gp, NULL);

    for (dlit=d->literals; !subsumed && dlit!=NULL; dlit=dlit->next) {
      /* printf("testing "); fprint_term(stdout, clit->atom);
      printf("  "); fprint_term(stdout, dlit->atom); printf("\n"); */
      if (clit->sign == dlit->sign) {
	Btm_state bt = match_bt_first(clit->atom, subst, dlit->atom, FALSE);
	while (bt != NULL && !subsumed) {
	  (*gp)->v = bt;
	  if (subsume_bt_literals(clit->next, subst, d, gp))
	    subsumed = TRUE;
	  else {
	    bt = match_bt_next(bt);
	  }
	}
      }
    }
    if (subsumed)
      return TRUE;
    else {
      *gp = plist_pop(*gp);
      return FALSE;
    }
  }
}  /* Subsume_bt_literals */

/*************
 *
 *   subsumes()
 *
 *************/

/* DOCUMENTATION
This routine checks if Clause c subsumes Clause d.
Ordinary unification is used; in particular, symmetry of
equality is not built-in.
<P>
*/

/* PUBLIC */
BOOL subsumes(Clause c, Clause d)
{
  Context subst = get_context();
  Trail tr = NULL;
  BOOL subsumed = subsume_literals(c->literals, subst, d, &tr);
  if (subsumed)
    undo_subst(tr);
  free_context(subst);
  Nonunit_subsumption_tests++;
  return subsumed;
}  /* subsumes */

/*************
 *
 *   subsumes_bt()
 *
 *************/

/* DOCUMENTATION
This routine checks if Clause c subsumes Clause d.
Backtrack unification is used; in particular, AC and
commutative/symmetric matching are applied where appropriate.
*/

/* PUBLIC */
BOOL subsumes_bt(Clause c, Clause d)
{
  Context subst = get_context();
  Plist g = NULL;
  int rc = subsume_bt_literals(c->literals, subst, d, &g);
  if (rc) {
    /* Cancel the list (stack) of btm_states */
    while (g != NULL) {
      Btm_state bt = g->v;
      match_bt_cancel(bt);
      g = plist_pop(g);
    }
  }
  free_context(subst);
  return rc;
}  /* subsumes_bt */

/*************
 *
 *   forward_subsume()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Clause forward_subsume(Clause d, Lindex idx)
{
  Literal dlit;
  Clause subsumer = NULL;
  Context subst = get_context();
  int nd = number_of_literals(d);

  /* We have to consider all literals of d, because when d is
     subsumed by c, not all literals of d have to match with
     a literal in c.  c is indexed on the first literal only.
   */

  for (dlit=d->literals; dlit!=NULL && subsumer==NULL; dlit=dlit->next) {
    Mindex mdx = dlit->sign ? idx->pos : idx->neg;
    Mindex_pos pos;
    Term catom = mindex_retrieve_first(dlit->atom, mdx, GENERALIZATION,
				       NULL, subst, FALSE, &pos);
    BOOL backtrack = lindex_backtrack(idx);
    while (catom != NULL && subsumer == NULL) {
      Clause c = catom->container;
      if (atom_number(c, catom) == 1) {
	int nc = number_of_literals(c);
	/* If c is a unit then we already know it subsumes d; otherwise,
	 * do a full subsumption check on the clauses.  (We don't let
	 * a clause subsume a shorter one, because that would cause
	 * factors to be deleted.)
	 */
	if (nc == 1 || (nc <= nd && (backtrack
				     ? subsumes_bt(c,d)
				     : subsumes(c,d)))) {
	  subsumer = c;
	  mindex_retrieve_cancel(pos);
	}
      }
      if (subsumer == NULL)
	catom = mindex_retrieve_next(pos);
    }
  }
  free_context(subst);
  return subsumer;
}  /* forward_subsume */

/*************
 *
 *   back_subsume()
 *
 *************/

/* DOCUMENTATION
Look in the index and return the list of clauses subsumed by c.
*/

/* PUBLIC */
Plist back_subsume(Clause c, Lindex idx)
{
  int nc = number_of_literals(c);

  if (nc == 0)
    return NULL;
  else {
    Plist subsumees = NULL;
    Context subst = get_context();
    Literal clit = c->literals;

    /* We only have to consider the first literal of c, because when
       c subsumes a clause d, all literals of c have to map into d.
       All literals of d are indexed.
     */

    Mindex mdx = clit->sign ? idx->pos : idx->neg;
    Mindex_pos pos;

    Term datom = mindex_retrieve_first(clit->atom, mdx, INSTANCE,
				       subst, NULL, FALSE, &pos);
    BOOL backtrack = lindex_backtrack(idx);
    while (datom != NULL) {
      Clause d = datom->container;
      if (d != c) {  /* in case c is already in idx */
	int nd = number_of_literals(d);
	/* If c is a unit the we already know it subsumes d; otherwise,
	 * do a full subsumption check on the clauses.  (We don't let
	 * a clause subsume a shorter one.)
	 */
	if (nc == 1 || (nc <= nd && (backtrack
				     ? subsumes_bt(c, d)
				     : subsumes(c, d))))
	  subsumees = insert_clause_into_plist(subsumees, d, FALSE);
      }
      datom = mindex_retrieve_next(pos);
    }
    free_context(subst);
    return subsumees;
  }
}  /* back_subsume */

/*************
 *
 *   back_subsume_one()
 *
 *************/

/* DOCUMENTATION
Look in the index for a clause subsumed by c.
The first one found is returned.  (It is not
necessarily the first of the subsumees that
was inserted into the index.)
*/

/* PUBLIC */
Clause back_subsume_one(Clause c, Lindex idx)
{
  int nc = number_of_literals(c);

  if (nc == 0)
    return NULL;
  else {
    Context subst = get_context();
    Literal clit = c->literals;

    Mindex mdx = clit->sign ? idx->pos : idx->neg;
    Mindex_pos pos;

    Term datom = mindex_retrieve_first(clit->atom, mdx, INSTANCE,
				       subst, NULL, FALSE, &pos);
    BOOL backtrack = lindex_backtrack(idx);
    BOOL found = FALSE;
    Clause d = NULL;

    while (datom != NULL && !found) {
      d = datom->container;
      if (d != c) {  /* in case c is already in idx */
	int nd = number_of_literals(d);
	/* If c is a unit the we already know it subsumes d; otherwise,
	 * do a full subsumption check on the clauses.  (We don't let
	 * a clause subsume a shorter one.)
	 */
	if (nc == 1 || (nc <= nd && (backtrack
				     ? subsumes_bt(c, d)
				     : subsumes(c, d)))) {
	  found = TRUE;
	  mindex_retrieve_cancel(pos);
	}
	else
	  datom = mindex_retrieve_next(pos);
      }
    }
    free_context(subst);
    return found ? d : NULL;
  }
}  /* back_subsume_one */

/*************
 *
 *   any_subsumes()
 *
 *************/

/* DOCUMENTATION
Return the first clause in Plist g that subsumes Clause d.
(We don't consider clasues in g that are longer than Clause d.)
*/

/* PUBLIC */
Clause any_subsumes(Plist g, Clause d, Uniftype utype)
{
  int dn = number_of_literals(d);
  Plist p = g;
  Clause subsumer = NULL;

  while (subsumer == NULL && p != NULL) {
    Clause c = p->v;
    if (number_of_literals(c) <= dn && (utype == BACKTRACK_UNIF
					? subsumes_bt(c,d)
					: subsumes_bt(c,d)))
      subsumer = c;
    else
      p = p->next;
  }
  return subsumer;
}  /* any_subsumes */

/*************
 *
 *   unit_conflict_by_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Clause unit_conflict_by_index(Clause c, Lindex idx)
{
  if (number_of_literals(c) != 1)
    return NULL;
  else {
    Clause conflictor = NULL;
    Clause empty = NULL;
    Context subst1 = get_context();
    Context subst2 = get_context();
    Mindex mdx = c->literals->sign ? idx->neg : idx->pos;
    Mindex_pos pos;
    Term atom = mindex_retrieve_first(c->literals->atom, mdx,
				      UNIFY, subst1, subst2, FALSE, &pos);

    while (atom != NULL && conflictor == NULL) {
      Clause d = atom->container;
      if (number_of_literals(d) == 1) {
	conflictor = d;
	empty = get_clause();
	empty->justification = binary_res_just(c, 1, conflictor, 1);
	inherit_attributes(c, subst1, conflictor, subst2, empty);
	mindex_retrieve_cancel(pos);
      }
      else 
	atom = mindex_retrieve_next(pos);
    }
    free_context(subst1);
    free_context(subst2);
    return empty;  /* May be NULL. */
  }
}  /* unit_conflict_by_index */

/*************
 *
 *   simplify_literals()
 *
 *   Remove any literals t!=t.
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void simplify_literals(Clause c)
{
  Literal l;
  int i;
  BOOL null_literals = FALSE;

  for (l = c->literals, i = 1; l; l = l->next, i++) {
    Term a = l->atom;
    BOOL sign = l->sign;
    if ((!sign && eq_atom(a) && term_ident(ARG(a,0), ARG(a,1))) ||
	(!sign && true_term(a)) ||
	(sign && false_term(a))) {

      c->justification = append_just(c->justification, xx_just(i));
      zap_term(l->atom);
      l->atom = NULL;
      null_literals = TRUE;
    }
  }
  if (null_literals)
    c->literals = remove_null_literals(c->literals);
}  /* simplify_literals */

/*************
 *
 *   unit_delete()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void unit_delete(Clause c, Lindex idx)
{
  Context subst = get_context();
  Literal l;
  int i;
  BOOL null_literals = FALSE;

  for (l = c->literals, i = 1; l; l = l->next, i++) {
    Mindex mdx = l->sign ? idx->neg : idx->pos;
    Mindex_pos pos;
    Term datom = mindex_retrieve_first(l->atom, mdx, GENERALIZATION,
				       NULL, subst, FALSE, &pos);
    BOOL ok = FALSE;
    while (datom && !ok) {
      Clause d = datom->container;
      if (unit_clause(d)) {
	ok = TRUE;
	mindex_retrieve_cancel(pos);
	c->justification = append_just(c->justification,
				       unit_del_just(d, i));
	zap_term(l->atom);
	l->atom = NULL;  /* remove it below */
	null_literals = TRUE;
      }
      else {
	Unit_deletion_failures++;
	datom = mindex_retrieve_next(pos);
      }
    }
  }
  if (null_literals)
    c->literals = remove_null_literals(c->literals);
  free_context(subst);
}  /* unit_delete */

/*************
 *
 *   forward_subsumex()
 *
 *************/

static
Plist insert_x(Plist p, Clause c)
{
  int nc = number_of_literals(c);
  Plist prev, curr, new;
  prev = NULL;
  curr = p;
  while (curr != NULL &&
	 number_of_literals(curr->v) <= nc &&
	 ((Clause) curr->v)->id != c->id) {
    prev = curr;
    curr = curr->next;
  }
  if (curr && ((Clause) curr->v)->id == c->id)
    ;  
  else if (curr == NULL || ((Clause) curr->v)->id != c->id) {
    new = get_plist();
    new->v = c;
    new->next = curr;
    if (prev != NULL)
      prev->next = new;
    else
      p = new;
  }
  return p;
}  /* insert_x */

/* DOCUMENTATION
*/

/* PUBLIC */
Clause forward_subsumex(Clause d, Lindex idx)
{
  if (number_of_literals(d) == 1)
    return forward_subsume(d, idx);
  else {
    Plist candidates = NULL;
    int nd = number_of_literals(d);

    Context subst = get_context();
    Literal dlit;

    for (dlit=d->literals; dlit!=NULL; dlit=dlit->next) {
      Mindex mdx = dlit->sign ? idx->pos : idx->neg;
      Mindex_pos pos;
      Term catom = mindex_retrieve_first(dlit->atom, mdx, GENERALIZATION,
					 NULL, subst, FALSE, &pos);
      while (catom != NULL) {
	Clause c = catom->container;
	if (number_of_literals(c) <= nd)
	  candidates = insert_x(candidates, c);
	catom = mindex_retrieve_next(pos);
      }
    }
    free_context(subst);

     printf("candidates=%d\n", plist_count(candidates)); fflush(stdout);

    {
      Plist p;
      Clause subsumer = NULL;
      for (p = candidates; p && !subsumer; p = p->next) {
	if (subsumes(p->v, d))
	  subsumer = p->v;
      }
      zap_plist(candidates);
      return subsumer;
    }
  }
}  /* forward_subsumex */

/*************
 *
 *   forward_subsumey()
 *
 *************/

/* DOCUMENTATION
like ordinary forward_subsume, except that we keep a list of
"tried-and-failed" candidates, and don't repeat the test on them.
*/

/* PUBLIC */
Clause forward_subsumey(Clause d, Lindex idx)
{
  Literal dlit;
  Clause subsumer = NULL;
  Context subst = get_context();
  int nd = number_of_literals(d);
  Plist tried = NULL;

  for (dlit=d->literals; dlit!=NULL && subsumer==NULL; dlit=dlit->next) {
    Mindex mdx = dlit->sign ? idx->pos : idx->neg;
    Mindex_pos pos;
    Term catom = mindex_retrieve_first(dlit->atom, mdx, GENERALIZATION,
				       NULL, subst, FALSE, &pos);
    while (catom != NULL && subsumer == NULL) {
      Clause c = catom->container;
      if (atom_number(c, catom) == 1) {
	int nc = number_of_literals(c);
	if (nc == 1)
	  subsumer = c;
	else if (nc > nd)
	  ;  /* do nothing */
	else if (plist_member(tried, c))
	  ;  /* do nothing */
	else if (subsumes(c,d))
	  subsumer = c;
	else
	  tried = insert_clause_into_plist(tried, c, FALSE);
      }
      if (subsumer == NULL)
	catom = mindex_retrieve_next(pos);
      else
	mindex_retrieve_cancel(pos);
    }
  }
  free_context(subst);
  zap_plist(tried);
  return subsumer;
}  /* forward_subsumey */

