#include "symbols.h"

/* Private definitions and types*/

struct sym_ent {
  int         sym_num;          /* unique identifier */
  int         arity;            /* 0 for constants */
  char        *name;            /* the print symbol */
  Symbol_type type;             /* function, relation, unspecified */
  Parsetype   parse_type;
  int         parse_prec;
  Unif_theory unif_theory;      /* e.g., associative-commutative */
  int         lex_val;
  Lrpo_status lrpo_status;
  Sym_ent     next;
};

#define SYM_TAB_SIZE  500

struct symbol_table {
  struct sym_ent *table[SYM_TAB_SIZE];
};

/* Private variables */

static struct symbol_table *Sym_tab;  /* pointer to symbol table */
static unsigned Sym_ent_count;
static Variable_style Var_style = STANDARD_STYLE;

static int Next_skolem_constant = 1;
static int Next_skolem_function = 1;

static BOOL Assoc_comm_symbols = FALSE;
static BOOL Comm_symbols = FALSE;

/*
 * memory management
 */

static Sym_ent sym_ent_avail;
static unsigned sym_ent_gets, sym_ent_frees, sym_ent_avails;

/* end of counters for each type */

/*************
 *
 *    Sym_ent get_sym_ent()
 *
 *************/

static
Sym_ent get_sym_ent(void)
{
  Sym_ent p;

  sym_ent_gets++;
  if (sym_ent_avail == NULL)
    p = tp_alloc(sizeof(struct sym_ent));
  else {
    sym_ent_avails--;
    p = sym_ent_avail;
    sym_ent_avail = sym_ent_avail->next;
  }
  
  p->name = "";
  p->sym_num = 0;
  p->parse_type = NOTHING_SPECIAL;
  p->parse_prec = 0;
  p->arity = -1;
  p->unif_theory = EMPTY_THEORY;
  p->lex_val = INT_MAX;
  p->lrpo_status = LRPO_LR_STATUS;
  p->type = UNSPECIFIED_SYMBOL;

  p->next = NULL;
  return(p);
}  /* get_sym_ent */

/*************
 *
 *    free_sym_ent()
 *
 *************/

static
void free_sym_ent(Sym_ent p)
{
  sym_ent_frees++;
  sym_ent_avails++;
  p->next = sym_ent_avail;
  sym_ent_avail = p;
}  /* free_sym_ent */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_symbols_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for Sym_ents.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_symbols_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");
  
  n = sizeof(struct sym_ent);
  fprintf(fp, "sym_ent (%4d)      %11u%11u%11u%11u%9.1f K\n",
	  n, sym_ent_gets, sym_ent_frees, sym_ent_gets - sym_ent_frees,
	  sym_ent_avails,
	  (((sym_ent_gets - sym_ent_frees) + sym_ent_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_symbols_mem */

/*************
 *
 *    p_symbols_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for Sym_ents.
*/

/* PUBLIC */
void p_symbols_mem(void)
{
    fprint_symbols_mem(stdout, 1);
}  /* p_symbols_mem */

/*
 *  end of memory management
 */

/*************
 *
 *    init_symbol_table()
 *
 *************/

static
void init_symbol_table(void)
{
  int i;

  Sym_tab = (struct symbol_table *) tp_alloc(sizeof(struct symbol_table));
  for (i = 0; i < SYM_TAB_SIZE; i++)
    Sym_tab->table[i] = NULL;
  Sym_ent_count = 0;
}  /* init_symbol_table */

/*************
 *
 *    empty_sym_tab() -- free all symbols in the symbol table
 *
 *************/

/* DOCUMENTATION
This routine deletes all entries in the symbol table.
*/

/* PUBLIC */
void empty_sym_tab(void)
{
  Sym_ent p1, p2;
  int i;
  
  for (i = 0; i < SYM_TAB_SIZE; i++) {
    p1 = Sym_tab->table[i];
    while (p1) {
      p2 = p1;
      p1 = p1->next;
      free_sym_ent(p2);
    }
    Sym_tab->table[i] = NULL;
  }
}  /* empty_sym_tab */

/*************
 *
 *    int new_sym_num()
 *
 *    Return the next available symbol number.  It is always POSITIVE.
 *
 *************/

static
int new_sym_num(void)
{
  Sym_ent_count++;
  return(Sym_ent_count);
}  /* new_sym_num */

/*************
 *
 *    hash_sym(sym_num)
 *
 *************/

static
int hash_sym(int sym_num)
{
  return(abs(sym_num) % SYM_TAB_SIZE);
}  /* hash_sym */

/*************
 *
 *    Sym_ent insert_sym(string, arity)
 *
 *    Insert string/arity into the symbol table and return the symbol
 *    table node.  Do not check if string/arity is already there.
 *
 *************/

static
Sym_ent insert_sym(char *s, int arity)
{
  Sym_ent p;
  int i;
  
  p = get_sym_ent();

  p->name = new_str_copy(s);
  p->arity = arity;
  p->sym_num = new_sym_num();
  i = hash_sym(p->sym_num);
  p->next = Sym_tab->table[i];
  Sym_tab->table[i] = p;
  return(p);
}  /* insert_sym */

/*************
 *
 *    int str_to_sn()
 *
 *************/

/* DOCUMENTATION
This routine takes a string and an arity, and returns an integer
identifier for the pair.  If the pair is not already in the symbol
table, a new entry is inserted into the table.  A pair, say ("f",2),
is sometimes written as f/2, which is a different symbol from f/3.
There is no limit on the length of the string.
*/

/* PUBLIC */
int str_to_sn(char *str, int arity)
{
  Sym_ent p;
  int i;

  if (Sym_tab == NULL)
    init_symbol_table();
  
  for (i = 0; i < SYM_TAB_SIZE; i++) {
    p = Sym_tab->table[i];
    while (p != NULL) {
      if (!str_ident(str, p->name))
	p = p->next;
      else if (p->arity != arity)
	p = p->next;
      else
	return(p->sym_num);
    }
  }
  
  /* String/arity not in table, so create an entry. */
  
  p = insert_sym(str, arity);
  
  return(p->sym_num);
  
}  /* str_to_sn */

/*************
 *
 *    fprint_syms(file_ptr) -- Display the symbol list.
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) the symbol table, including all
of the attributes of each symbol.
*/

/* PUBLIC */
void fprint_syms(FILE *fp)
{
  Sym_ent p;
  int i;
  
  for (i = 0; i < SYM_TAB_SIZE; i++) {
    for (p = Sym_tab->table[i]; p; p = p->next)
      fprintf(fp, "%d  %s/%d %s\n",
	      p->sym_num, p->name, p->arity,
	      p->type == FUNCTION_SYMBOL ? "function" : p->type == RELATION_SYMBOL ? "relation" : "unknown type");
  }
}  /* fprint_syms */

/*************
 *
 *    p_syms()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) the symbol table, including all
of the attributes of each symbol.
*/

/* PUBLIC */
void p_syms(void)
{
  fprint_syms(stdout);
}  /* p_syms */

/*************
 *
 *    sn_to_node(sym_num)
 *
 *    Given a symbol number, return the symbol table node.
 *
 *************/

static
Sym_ent sn_to_node(int sym_num)
{
  if (sym_num <= 0)
    return NULL;
  else {
    Sym_ent p = Sym_tab->table[hash_sym(sym_num)];
    while (p != NULL && p->sym_num != sym_num)
      p = p->next;
    return p;  /* possibly NULL */
  }
}  /* sn_to_node */

/*************
 *
 *    str_to_node()
 *
 *    Given a string/arity, return the symbol table node if it exists.
 *
 *************/

static
Sym_ent str_to_node(char *str, int arity)
{
  Sym_ent p;
  int i;

  if (Sym_tab == NULL)
    return NULL;
  
  for (i = 0; i < SYM_TAB_SIZE; i++) {
    p = Sym_tab->table[i];
    while (p != NULL) {
      if (!str_ident(str, p->name))
	p = p->next;
      else if (p->arity != arity)
	p = p->next;
      else
	return p;
    }
  }
  return NULL;
}  /* str_to_node */

/*************
 *
 *    char *sn_to_str(sym_num)  --  given a symbol number, return the name
 *
 *************/

/* DOCUMENTATION
This routine returns the string assocated with a symbol ID.
*/

/* PUBLIC */
char *sn_to_str(int sym_num)
{
  Sym_ent p = sn_to_node(sym_num);
  
  if (p == NULL)
    return("");
  else
    return(p->name);
}  /* sn_to_str */

/*************
 *
 *    int sn_to_arity(sym_num)  --  given a symbol number, return the arity
 *
 *************/

/* DOCUMENTATION
This routine returns the arity associated with a symbol ID.
*/

/* PUBLIC */
int sn_to_arity(int sym_num)
{
  Sym_ent p = sn_to_node(sym_num);
  if (p == NULL)
    return(-1);
  else
    return(p->arity);
}  /* sn_to_arity */

/*************
 *
 *    int is_symbol(symbol, str, arity)
 *
 *************/

/* DOCUMENTATION
This Boolean routine checks if a given symbol ID matches a given
(string,arity) pair.
*/

/* PUBLIC */
BOOL is_symbol(int symnum, char *str, int arity)
{
  Sym_ent n = sn_to_node(symnum);
  if (n == NULL)
    return FALSE;
  else
    return (n->arity == arity && str_ident(n->name, str));
}  /* is_symbol */

/*************
 *
 *    int is_eq_symbol(symbol)
 *
 *************/

/* DOCUMENTATION
This Boolean routine checks if a given symbol ID is for =/2.
One could use is_symbol(symnum, "=", 2) instead, but this
should be a bit faster.
*/

/* PUBLIC */
BOOL is_eq_symbol(int symnum)
{
  static int eq_symnum = 0;
  if (eq_symnum == 0) {
    eq_symnum = str_to_sn("=", 2);
  }
  return (symnum == eq_symnum ? TRUE : FALSE);
}  /* is_eq_symbol */

/*************
 *
 *    set_lex_val()
 *
 *************/

/* DOCUMENTATION
This routine is used to assign a lexical value to a symbol.
The value can be retrieved later with sn_to_lex_val();
*/

/* PUBLIC */
void set_lex_val(int symnum, int lex_val)
{
  Sym_ent p = sn_to_node(symnum);
  p->lex_val = lex_val;
}  /* set_lex_val */

/*************
 *
 *    set_assoc_comm()
 *
 *************/

/* DOCUMENTATION
This routine declares a string to be a (binary) symbol with the
logical property "associative-commutative".  This property is used for
AC unification/matching/identity.
(If you wish to print AC expressions without parentheses, see
set_parse_type().)

*/

/* PUBLIC */
void set_assoc_comm(char *str)
{
  int sn = str_to_sn(str, 2);
  Sym_ent p = sn_to_node(sn);
  p->unif_theory = ASSOC_COMMUTE;
  Assoc_comm_symbols = TRUE;
}  /* set_assoc_comm */

/*************
 *
 *    set_commutative()
 *
 *************/

/* DOCUMENTATION
This routine declares a string to be a (binary) symbol with the
logical property "commutative".  This includes symmetric relations as
well as commutative functions.  This property is used for commutative
unification/matching/identity.
*/

/* PUBLIC */
void set_commutative(char *str)
{
  int sn = str_to_sn(str, 2);
  Sym_ent p = sn_to_node(sn);
  p->unif_theory = COMMUTE;
  Comm_symbols = TRUE;
}  /* set_commutative */

/*************
 *
 *    assoc_comm_symbols()
 *
 *************/

/* DOCUMENTATION
This function tells you if any symbol has been declared to be
associative-commutative;
*/

/* PUBLIC */
BOOL assoc_comm_symbols(void)
{
  return Assoc_comm_symbols;
}  /* assoc_comm_symbols */

/*************
 *
 *    comm_symbols()
 *
 *************/

/* DOCUMENTATION
This function tells you if any symbol has been declared to be
commutative/symmetric;
*/

/* PUBLIC */
BOOL comm_symbols(void)
{
  return Comm_symbols;
}  /* comm_symbols */

/*************
 *
 *    is_assoc_comm
 *
 *    NOTE: may wish to avoid hash if no AC symbols exist.
 *
 *************/

/* DOCUMENTATION
This function checks if a symbol ID has the
associative-commutative property.  Note that set_assoc_comm() takes a
string, but this routine takes a symbol ID.  Recall that str_to_sn()
and sn_to_str() translate between the two forms.
*/

/* PUBLIC */
BOOL is_assoc_comm(int sn)
{
  Sym_ent p = sn_to_node(sn);
  return (p == NULL ? 0 : p->unif_theory == ASSOC_COMMUTE);
}  /* is_assoc_comm */

/*************
 *
 *    is_commutative
 *
 *    NOTE: may wish to avoid hash if no C symbols exist.
 *
 *************/

/* DOCUMENTATION
This function checks if a symbol ID has the commutative property.
Note that set_commutative() takes a string, but this routine
takes a symbol ID.  Recall that str_to_sn() and sn_to_str()
translate between the two forms.
*/

/* PUBLIC */
BOOL is_commutative(int sn)
{
  Sym_ent p = sn_to_node(sn);
  return (p == NULL ? 0 : p->unif_theory == COMMUTE);
}  /* is_commutative */

/*************
 *
 *    fprint_sym()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) the string associated with a symbol ID.
A newline is NOT printed.
*/

/* PUBLIC */
void fprint_sym(FILE *fp, int symnum)
{
  fprintf(fp, "%s", sn_to_str(symnum));
}  /* fprint_sym */

/*************
 *
 *    sprint_sym()
 *
 *************/

/* DOCUMENTATION
This routine appends, to String_buf sb, the string associated with
a symbol ID.  A newline is NOT printed.
*/

/* PUBLIC */
void sprint_sym(String_buf sb, int symnum)
{
  sb_append(sb, sn_to_str(symnum));
}  /* fprint_sym */

/*************
 *
 *    p_sym()
 *
 *************/

/* DOCUMENTATION
This routine prints (stdout) the string associated with a symbol ID.
A newline is NOT printed.
*/

/* PUBLIC */
void p_sym(int symnum)
{
  fprint_sym(stdout, symnum);
}  /* p_sym */

/*************
 *
 *    sn_to_lex_val
 *
 *************/

/* DOCUMENTATION
This routine returns the lexical value associated with a symbol ID.
The default value is INT_MAX.  If the symbol ID is not valid, INT_MIN
is returned.
*/

/* PUBLIC */
int sn_to_lex_val(int sn)
{
  Sym_ent p = sn_to_node(sn);
  return (p == NULL ? INT_MIN : p->lex_val);
}  /* sn_to_lex_val */

/*************
 *
 *    sn_to_lrpo_status
 *
 *************/

/* DOCUMENTATION
This routine returns the LRPO status associated with a symbol ID.
The default value is LRPO_LR_STATUS.  See order.h for the possible
values.  If the symbol ID is not valid, 0 is returned.
*/

/* PUBLIC */
Lrpo_status sn_to_lrpo_status(int sn)
{
  Sym_ent p = sn_to_node(sn);
  return (p == NULL ? 0 : p->lrpo_status);
}  /* sn_to_lrpo_status */

/*************
 *
 *   set_parse_type()
 *
 *************/

/* DOCUMENTATION
This routine sets the parse/print properties of a binary or unary symbol.
The types for binary infix symbols are
INFIX_LEFT,
INFIX_RIGHT,
INFIX.
The types for prefix unary symbols are
PREFIX,
PREFIX_PAREN.
The types for postfix unary symbols are
POSTFIX,
POSTFIX_PAREN.
<P>
If the precedence is out of range [MIN_PRECEDENCE ... MAX_PRECEDENCE],
nothing happens.
*/

/* PUBLIC */
void set_parse_type(char *str, int precedence, Parsetype type)
{
  if (precedence < MIN_PRECEDENCE || precedence > MAX_PRECEDENCE)
    return;
  else {
    Sym_ent p = NULL;
    switch (type) {
    case INFIX_LEFT:
    case INFIX_RIGHT:
    case INFIX:
      p = sn_to_node(str_to_sn(str, 2));
      p->parse_type = type;
      p->parse_prec = precedence;
      break;
    case PREFIX:
    case PREFIX_PAREN:
    case POSTFIX:
    case POSTFIX_PAREN:
      p = sn_to_node(str_to_sn(str, 1));
      p->parse_type = type;
      p->parse_prec = precedence;
      break;
    case NOTHING_SPECIAL:
      break;
    }
  }
}  /* set_parse_type */

/*************
 *
 *   binary_parse_type()
 *
 *************/

/* DOCUMENTATION
This routine gets the parse/print properties for a binary symbol.
If *str is a binary symbol, TRUE is returned and the properties are filled in.
If *str is a not a binary symbol, FALSE is returned.
*/

/* PUBLIC */
BOOL binary_parse_type(char *str, int *precedence, Parsetype *type)
{
  Sym_ent p = str_to_node(str, 2);
  if (p == NULL || p->parse_type == NOTHING_SPECIAL)
    return FALSE;
  else {
    *precedence = p->parse_prec;
    *type = p->parse_type;
    return TRUE;
  }
}  /* binary_parse_type */

/*************
 *
 *   unary_parse_type()
 *
 *************/

/* DOCUMENTATION
This routine gets the parse/print properties for a unary symbol.
If *str is a unary symbol, TRUE is returned and the properties are filled in.
If *str is a not a unary symbol, FALSE is returned.
*/

/* PUBLIC */
BOOL unary_parse_type(char *str, int *precedence, Parsetype *type)
{
  Sym_ent p = str_to_node(str, 1);
  if (p == NULL || p->parse_type == NOTHING_SPECIAL)
    return FALSE;
  else {
    *precedence = p->parse_prec;
    *type = p->parse_type;
    return TRUE;
  }
}  /* unary_parse_type */

/*************
 *
 *   variable_name()
 *
 *************/

/* DOCUMENTATION
Is the given name a variable?  In some cases (in particular clauses)
variables are not explicitly quantified, so we have a rule to
distinguish variables from constants.  This is it.
*/

/* PUBLIC */
BOOL variable_name(char *s)
{
  if (variable_style() == PROLOG_STYLE)
    return ((*s >= 'A' && *s <= 'Z') || *s == '_');
  else if (variable_style() == INTEGER_STYLE)
    return (*s >= '0' && *s <= '9');
  else
    return (*s >= 'u' && *s <= 'z');
}  /* variable_name */

/*************
 *
 *   str_exists()
 *
 *************/

/* DOCUMENTATION
This function checks if the given string occurs in the
symbol table (with any arity).
*/

/* PUBLIC */
BOOL str_exists(char *str)
{

  if (Sym_tab == NULL)
    return 0;
  else {
    Sym_ent p;
    int i;
    int found = 0;
  
    for (i = 0; i < SYM_TAB_SIZE && !found; i++) {
      for (p = Sym_tab->table[i]; p != NULL && !found; p = p->next)
	found = str_ident(str, p->name);
    }
    return found;
  }
}  /* str_exists */

/*************
 *
 *   next_skolem_symbol()
 *
 *************/

/* DOCUMENTATION
This routine returns a fresh symbol ID, which is intended to be
used as a Skolem symbol.  
The symbols are $c1, $c2, $c3, ... for arity 0 (constants) and
$f1, $f2, $f3, ... for arity != 0.
If some of those symbols already exist
in the symbol table (with any arity), they will be skipped.
*/

/* PUBLIC */
int next_skolem_symbol(int arity)
{
  char name[20];

  do {
    if (arity == 0) {
      sprintf(name, "$c%d", Next_skolem_constant);
      Next_skolem_constant++;
    }
    else {
      sprintf(name, "$f%d", Next_skolem_function);
      Next_skolem_function++;
    }
  } while (str_exists(name));

  return str_to_sn(name, arity);
}  /* next_skolem_symbol */

/*************
 *
 *   fresh_symbol()
 *
 *************/

/* DOCUMENTATION
This routine returns a symbol ID for a new symbol with the given
arity.  The symbol is made up of the given prefix followed by the
smallest natural number that results in a new symbol (regardless of
arity).  The prefix must be less than MAX_NAME characters.
*/

/* PUBLIC */
int fresh_symbol(char *prefix, int arity)
{
  char name[MAX_NAME+20];
  int i = 0;

  if (strlen(prefix) > MAX_NAME) {
    fatal_error("fresh_symbol, prefix is too big.");
  }

  do {
    sprintf(name, "%s%d", prefix, i);
    i++;
  } while (str_exists(name));

  return str_to_sn(name, arity);
}  /* fresh_symbol */

/*************
 *
 *   str_to_int()
 *
 *************/

/* DOCUMENTATION
This routine tries to convert a string into an integer (using strtol()).
If successful, TRUE is returned and *ip is set to the integer.
If failure, FALSE is returned.
*/

/* PUBLIC */
BOOL str_to_int(char *str, int *ip)
{
  char *end;
  *ip = strtol(str, &end, 10);
  if (*end != '\0')
    return FALSE;
  else
    return TRUE;
}  /* str_to_int */

/*************
 *
 *   reverse_chars()
 *
 *************/

/* DOCUMENTATION
This routine reverses an array of characters.
You must give the starting and ending positions.
*/

/* PUBLIC */
void reverse_chars(char *s, int start, int end)
{
  if (start < end) {
    char c = s[start]; s[start] = s[end]; s[end] = c;
    reverse_chars(s, start+1, end-1);
  }
}  /* reverse_chars */

/*************
 *
 *   itoa()
 *
 *************/

/* DOCUMENTATION
This routine converts an integer to a string (in decimal form).
The character array s must be large enough to hold the string.
*/

/* PUBLIC */
void itoa(int n, char *s)
{
  int i = 0;
  BOOL sign = (n >= 0);
  n = abs(n);
  if (n == 0)
    s[i++] = '0';
  else {
    while (n > 0) {
      s[i++] = '0' + (n % 10);
      n = n / 10;
    }
    if (!sign)
      s[i++] = '-';
  }
  reverse_chars(s, 0, i-1);
  s[i] = '\0';
}  /* itoa */

/*************
 *
 *   greatest_symnum()
 *
 *************/

/* DOCUMENTATION
This function returns the greatest symnum (symbol ID) currently in use.
This can be used if you need to dynamnically allocate an array
of objects to be indexed by symnum.
*/

/* PUBLIC */
int greatest_symnum(void)
{
  return Sym_ent_count;
}  /* greatest_symnum */

/*************
 *
 *    int sym_precedence(symnum_1, symnum_2)
 *
 *************/

/* DOCUMENTATION
This routine compares two symbol IDs by looking at their lex_val
in the symbol table.  The range of return values is<BR>
{SAME_AS, GREATER_THAN, LESS_THAN, NOT_COMPARABLE}.
*/

/* PUBLIC */
Ordertype sym_precedence(int symnum_1, int symnum_2)
{
  int p1, p2;

  if (symnum_1 == symnum_2)
    return SAME_AS;
  else {
    p1 = sn_to_lex_val(symnum_1);
    p2 = sn_to_lex_val(symnum_2);
	
    if (p1 == INT_MAX || p2 == INT_MAX)
      return NOT_COMPARABLE;
    else if (p1 > p2)
      return GREATER_THAN;
    else if (p1 < p2)
      return LESS_THAN;
    else
      return SAME_AS;
  }
}  /* sym_precedence */

/*************
 *
 *   set_variable_style()
 *
 *************/

/* DOCUMENTATION
This routine determines how variables are parsed and printed.
*/

/* PUBLIC */
void set_variable_style(Variable_style style)
{
  Var_style = style;
}  /* set_variable_style */

/*************
 *
 *   variable_style()
 *
 *************/

/* DOCUMENTATION
This routine gives the current variable style.
*/

/* PUBLIC */
Variable_style variable_style(void)
{
  return Var_style;
}  /* set_variable_style */

/*************
 *
 *   no_symbols_have_lex_val()
 *
 *************/

/* DOCUMENTATION
Return TRUE if no symbol has been given a lex_val;
*/

/* PUBLIC */
BOOL no_symbols_have_lex_val(void)
{
  Sym_ent p;
  int i;
  
  for (i = 0; i < SYM_TAB_SIZE; i++) {
    for (p = Sym_tab->table[i]; p; p = p->next) {
      if (p->lex_val != INT_MAX)
	return FALSE;
    }
  }
  return TRUE;
}  /* no_symbols_have_lex_val */

/*************
 *
 *   syms_with_lex_val()
 *
 *************/

/* DOCUMENTATION
Return an Ilist containing symnums of symbols to which lex_vals have
been assigned.
*/

/* PUBLIC */
Ilist syms_with_lex_val(void)
{
  Sym_ent p;
  Ilist g = NULL;
  int i;
  
  for (i = 0; i < SYM_TAB_SIZE; i++) {
    for (p = Sym_tab->table[i]; p; p = p->next) {
      if (p->lex_val != INT_MAX)
	g = ilist_prepend(g, i);
    }
  }
  return g;
}  /* syms_with_lex_val */

/*************
 *
 *   set_symbol_type()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void set_symbol_type(int sym_num, Symbol_type type)
{
  Sym_ent p = sn_to_node(sym_num);
  if (p == NULL)
    fatal_error("set_symbol_type: bad sym_num");
  p->type = type;
}  /* set_symbol_type */

/*************
 *
 *   get_symbol_type()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Symbol_type get_symbol_type(int sym_num)
{
  Sym_ent p = sn_to_node(sym_num);
  if (p == NULL)
    fatal_error("get_symbol_type: bad sym_num");
  return p->type;
}  /* get_symbol_type */

/*************
 *
 *   function_symbol()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
BOOL function_symbol(int sym_num)
{
  return get_symbol_type(sym_num) == FUNCTION_SYMBOL;
}  /* function_symbol */

/*************
 *
 *   relation_symbol()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
BOOL relation_symbol(int sym_num)
{
  return get_symbol_type(sym_num) == RELATION_SYMBOL;
}  /* relation_symbol */

/*************
 *
 *   lex_compare_1()
 *
 *   1. constants < arity-2 < arity-3 < ... < arity-infinity < arity-1
 *   2. within arity, use strcmp.
 *
 *************/

static
Ordertype lex_compare_1(Sym_ent s1, Sym_ent s2)
{
  if (s1 == NULL)
    return LESS_THAN;
  else if (s2 == NULL)
    return GREATER_THAN;
  else if (s1->arity == s2->arity) {
    int i = strcmp(s1->name, s2->name);
    if (i < 0)
      return LESS_THAN;
    else if (i > 0)
      return GREATER_THAN;
    else
      return SAME_AS;
  }
  else if (s1->arity == 1)
    return GREATER_THAN;
  else if (s2->arity == 1)
    return LESS_THAN;
  else if (s1->arity < s2->arity)
    return LESS_THAN;
  else
    return GREATER_THAN;
}  /* lex_compare_1 */

/*************
 *
 *   lex_order_1()
 *
 *************/

/* DOCUMENTATION
This routine imposes a default total ordering on symbols.
<BR>
The primary order is
<PRE>
constants < arity-2 < arity-3 < ... < arity-infinity < arity-1
<PRE>
Within arity, the order is determined by the C library function
strcmp, which is (usually) lexicographic ASCII order.
*/

/* PUBLIC */
void lex_order_1(void)
{
  int n = greatest_symnum() + 1;
  Sym_ent *a = malloc(n * sizeof(void *));
  int i;
  for (i = 0; i < n; i++)
    a[i] = sn_to_node(i);

  merge_sort((void **) a, n,
	     (Ordertype (*)(void*, void*)) lex_compare_1);

  for (i = 0; i < n; i++) {
    if (a[i]) {
      a[i]->lex_val = i;
#if 0
      printf("symbol %s/%d, lex_val=%d.\n",
	     a[i]->name, a[i]->arity, a[i]->lex_val);
#endif
    }
  }
  free(a);
}  /* lex_order_1 */

/*************
 *
 *   lex_compare_2()
 *
 *   1. functions < relations < unspecified
 *   2. existing lex_val  (default is INT_MAX)
 *   3. function: constants < arity-2 < ... < arity-infinity < arity-1
 *   4. relations & unspecified: constants < arity-1 < ... < arity-infinity < =
 *   5. = is greatest
 *   6. strcmp
 *
 *************/

static
Ordertype lex_compare_2(Sym_ent s1, Sym_ent s2)
{
  if (s1 == s2)
    return SAME_AS;
  else if (s1 == NULL)
    return LESS_THAN;
  else if (s2 == NULL)
    return GREATER_THAN;

  else if (s1->type == FUNCTION_SYMBOL && s2->type != FUNCTION_SYMBOL)
    return LESS_THAN;
  else if (s1->type != FUNCTION_SYMBOL && s2->type == FUNCTION_SYMBOL)
    return GREATER_THAN;

  else if (s1->type == RELATION_SYMBOL && s2->type != RELATION_SYMBOL)
    return LESS_THAN;
  else if (s1->type != RELATION_SYMBOL && s2->type == RELATION_SYMBOL)
    return GREATER_THAN;

  /* now they have the same type */

  else if (s1->lex_val < s2->lex_val)
    return LESS_THAN;
  else if (s1->lex_val > s2->lex_val)
    return GREATER_THAN;

  else if (s1->type == RELATION_SYMBOL && is_eq_symbol(s2->sym_num))
    return LESS_THAN;
  else if (s1->type == RELATION_SYMBOL && is_eq_symbol(s1->sym_num))
    return GREATER_THAN;

  else if (s1->arity == s2->arity) {
    if (!is_eq_symbol(s1->sym_num) && is_eq_symbol(s2->sym_num))
      return LESS_THAN;
    else if (is_eq_symbol(s1->sym_num) && !is_eq_symbol(s2->sym_num))
      return GREATER_THAN;
    else {
      int i = strcmp(s1->name, s2->name);
      if (i < 0)
	return LESS_THAN;
      else if (i > 0)
	return GREATER_THAN;
      else
	return SAME_AS;
    }
  }
  else if (s1->type == FUNCTION_SYMBOL && s1->arity == 1)
    return GREATER_THAN;
  else if (s1->type == FUNCTION_SYMBOL && s2->arity == 1)
    return LESS_THAN;
  else if (s1->arity < s2->arity)
    return LESS_THAN;
  else
    return GREATER_THAN;
}  /* lex_compare_2 */

/*************
 *
 *   lex_order_2()
 *
 *************/

/* DOCUMENTATION
Assign a total order on lex_vals of (fsyms U rsyms).
If any of the symbols already have lex_vals, they
are considered, but overwritten by this rouine.
For the rules, see lex_compare_2.
*/

/* PUBLIC */
void lex_order_2(Ilist fsyms, Ilist rsyms)
{
  int n = ilist_count(fsyms) + ilist_count(rsyms);
  Sym_ent *a = malloc(n * sizeof(void *));
  Ilist p;
  int i = 0;
  for (p = fsyms; p; p = p->next)
    a[i++] = sn_to_node(p->i);
  for (p = rsyms; p; p = p->next)
    a[i++] = sn_to_node(p->i);
  
  merge_sort((void **) a, n,
	     (Ordertype (*)(void*, void*)) lex_compare_2);
  
  for (i = 0; i < n; i++)
    a[i]->lex_val = i;
  free(a);
}  /* lex_order_2 */

/*************
 *
 *   insert_by_lex_val()
 *
 *************/

static
Ilist insert_by_lex_val(Ilist head, Ilist tail)
{
  if (tail == NULL) {
    head->next = NULL;
    return head;
  }
  else if (sn_to_lex_val(head->i) < sn_to_lex_val(tail->i)) {
    head->next = tail;
    return head;
  }
  else {
    tail->next = insert_by_lex_val(head, tail->next);
    return tail;
  }
}  /* insert_by_lex_val */

/*************
 *
 *   sort_by_lex_val()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Ilist sort_by_lex_val(Ilist p)
{
  if (p == NULL)
    return NULL;
  else {
    return insert_by_lex_val(p, sort_by_lex_val(p->next));
  }
}  /* sort_by_lex_val */

/*************
 *
 *   declare_functions_and_relations()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void declare_functions_and_relations(Ilist fsyms, Ilist rsyms)
{
  Ilist p;
  for (p = fsyms; p; p = p->next)
    set_symbol_type(p->i, FUNCTION_SYMBOL);
  for (p = rsyms; p; p = p->next)
    set_symbol_type(p->i, RELATION_SYMBOL);
}  /* declare_functions_and_relations */
