#ifndef TP_SYMBOLS_H
#define TP_SYMBOLS_H

#include "string.h"
#include "strbuf.h"
#include "order.h"
#include "glist.h"

/* INTRODUCTION
This collection of routines manages the global symbol table.
Each symbol is a pair (string,arity) and has a unique ID number.
For example, ("f",2), sometimes written f/2, is a different symbol
from f/3.
<P>
These symbols are used mostly as constant, function, and predicate
symbols, but they can be used for variables and other things as
well.
*/

/* Public definitions */

/* maximum number of chars in string part of symbol (includes '\0') */

#define MAX_NAME      51

/* parse/print properties of symbols */

typedef enum {NOTHING_SPECIAL,
	      INFIX,         /* xfx */
	      INFIX_LEFT ,   /* yfx */
	      INFIX_RIGHT,   /* xfy */
	      PREFIX_PAREN,  /* fx  */
	      PREFIX,        /* fy  */
	      POSTFIX_PAREN, /* xf  */
	      POSTFIX        /* yf  */
             } Parsetype;

#define MIN_PRECEDENCE      1
#define MAX_PRECEDENCE    999

/* Function/relation properties of symbols */

typedef enum { FUNCTION_SYMBOL,
	       RELATION_SYMBOL,
	       UNSPECIFIED_SYMBOL
             } Symbol_type;

/* Unification properties of symbols */

typedef enum { EMPTY_THEORY,
	       COMMUTE,
	       ASSOC_COMMUTE
             } Unif_theory;

/* LRPO status */

typedef enum { LRPO_LR_STATUS,
	       LRPO_MULTISET_STATUS
             } Lrpo_status;

/* Variable style */

typedef enum { STANDARD_STYLE,      /* x,y,z,... */
	       PROLOG_STYLE,        /* A,B,C,... */
	       INTEGER_STYLE        /* 0,1,2,... */
             } Variable_style;

/* Logic symbols when in Term form */

#define TRUE_SYM    "$T"
#define FALSE_SYM   "$F"
#define AND_SYM     "&"
#define OR_SYM      "|"
#define NOT_SYM     "~"
#define IFF_SYM     "<->"
#define IMP_SYM     "->"
#define IMPBY_SYM   "<-"
#define ALL_SYM     "all"
#define EXISTS_SYM  "exists"
#define QUANT_SYM   "$quantified"

/* Other symbols when in Term form */

#define ATTRIB_SYM  "#"  /* operator for attaching attributes */

typedef struct sym_ent * Sym_ent;

/* End of public definitions */

/* Public function prototypes from symbols.c */

void fprint_symbols_mem(FILE *fp, BOOL heading);

void p_symbols_mem(void);

void empty_sym_tab(void);

int str_to_sn(char *str, int arity);

void fprint_syms(FILE *fp);

void p_syms(void);

char *sn_to_str(int sym_num);

int sn_to_arity(int sym_num);

BOOL is_symbol(int symnum, char *str, int arity);

BOOL is_eq_symbol(int symnum);

void set_lex_val(int symnum, int lex_val);

void set_assoc_comm(char *str);

void set_commutative(char *str);

BOOL assoc_comm_symbols(void);

BOOL comm_symbols(void);

BOOL is_assoc_comm(int sn);

BOOL is_commutative(int sn);

void fprint_sym(FILE *fp, int symnum);

void sprint_sym(String_buf sb, int symnum);

void p_sym(int symnum);

int sn_to_lex_val(int sn);

Lrpo_status sn_to_lrpo_status(int sn);

void set_parse_type(char *str, int precedence, Parsetype type);

BOOL binary_parse_type(char *str, int *precedence, Parsetype *type);

BOOL unary_parse_type(char *str, int *precedence, Parsetype *type);

BOOL variable_name(char *s);

BOOL str_exists(char *str);

int next_skolem_symbol(int arity);

int fresh_symbol(char *prefix, int arity);

BOOL str_to_int(char *str, int *ip);

void reverse_chars(char *s, int start, int end);

void itoa(int n, char *s);

int greatest_symnum(void);

Ordertype sym_precedence(int symnum_1, int symnum_2);

void set_variable_style(Variable_style style);

Variable_style variable_style(void);

BOOL no_symbols_have_lex_val(void);

Ilist syms_with_lex_val(void);

void set_symbol_type(int sym_num, Symbol_type type);

Symbol_type get_symbol_type(int sym_num);

BOOL function_symbol(int sym_num);

BOOL relation_symbol(int sym_num);

void lex_order_1(void);

void lex_order_2(Ilist fsyms, Ilist rsyms);

Ilist sort_by_lex_val(Ilist p);

void declare_functions_and_relations(Ilist fsyms, Ilist rsyms);

#endif  /* conditional compilation of whole file */
