#include "termorder.h"

/* Private definitions and types */

/*************
 *
 *    int term_compare_ncv(t1, t2)
 *
 *************/

/* DOCUMENTATION
This routine compares two terms.  The ordering is
CONSTANT < COMPLEX < VARIABLE; within type, the order is by VARNUM
and lexigocgaphic by SYMNUM.  The range of return values is<BR>
{SAME_AS, GREATER_THAN, LESS_THAN}.
*/

/* PUBLIC */
Ordertype term_compare_ncv(Term t1, Term t2)
{
  Ordertype  rc;

  if ((COMPLEX(t1) && COMPLEX(t2)) || (CONSTANT(t1) && CONSTANT(t2))) {
    if (SYMNUM(t1) == SYMNUM(t2)) {
      int i;
      for (rc = SAME_AS, i = 0; rc == SAME_AS && i < t1->arity; i++)
	rc = term_compare_ncv(t1->args[i], t2->args[i]);
    }
    else if (SYMNUM(t1) > SYMNUM(t2))
      rc = GREATER_THAN;
    else
      rc = LESS_THAN;
  }
  else if (VARIABLE(t1) && VARIABLE(t2)) {
    if (VARNUM(t1) == VARNUM(t2))
      rc = SAME_AS;
    else
      rc = (VARNUM(t1) > VARNUM(t2) ? GREATER_THAN : LESS_THAN);
  }
  /* Now we know they are different types. */
  else if (VARIABLE(t1))
    rc = GREATER_THAN;
  else if (VARIABLE(t2))
    rc = LESS_THAN;
  else if (COMPLEX(t1))
    rc = GREATER_THAN;
  else 
    rc = LESS_THAN;  /* CONSTANT(t1) && COMPLEX(t2) */

  return rc;
}  /* term_compare_ncv */

/*************
 *
 *    int term_compare_vcp(t1, t2)
 *
 *************/

/* DOCUMENTATION
This routine compares two terms.  The ordering is
VARIABLE < CONSTANT < COMPLEX; within type, the order is by VARNUM
and lexigocgaphic by SYMNUM.  The range of return values is<BR>
{SAME_AS, GREATER_THAN, LESS_THAN}.
*/

/* PUBLIC */
Ordertype term_compare_vcp(Term t1, Term t2)
{
  Ordertype  rc;

  if ((COMPLEX(t1) && COMPLEX(t2)) || (CONSTANT(t1) && CONSTANT(t2))) {
    if (SYMNUM(t1) == SYMNUM(t2)) {
      int i;
      for (rc = SAME_AS, i = 0; rc == SAME_AS && i < t1->arity; i++)
	rc = term_compare_vcp(t1->args[i], t2->args[i]);
    }
    else if (SYMNUM(t1) > SYMNUM(t2))
      rc = GREATER_THAN;
    else
      rc = LESS_THAN;
  }
  else if (VARIABLE(t1) && VARIABLE(t2)) {
    if (VARNUM(t1) == VARNUM(t2))
      rc = SAME_AS;
    else
      rc = (VARNUM(t1) > VARNUM(t2) ? GREATER_THAN : LESS_THAN);
  }
  /* Now we know they are different types. */
  else if (VARIABLE(t1))
    rc = LESS_THAN;
  else if (VARIABLE(t2))
    rc = GREATER_THAN;
  else if (COMPLEX(t1))
    rc = GREATER_THAN;
  else 
    rc = LESS_THAN;  /* CONSTANT(t1) && COMPLEX(t2) */

  return rc;
}  /* term_compare_vcp */

/*************
 *
 *    lrpo_lex(t1, t2) -- Is t1 > t2 ?
 *
 *    t1 and t2 have same symbol and the symbol has lr status.
 *
 *************/

static
BOOL lrpo_lex(Term t1, Term t2)
{
  int i;
  BOOL ok;
  int arity = t1->arity;

  /* First skip over any identical arguments. */

  for (i = 0; i < arity && term_ident(t1->args[i],t2->args[i]); i++);

  if (i == arity)
    return 0;  /* t1 and t2 identical */
  else if (lrpo(t1->args[i], t2->args[i])) {
    /* return (t1 > each remaining arg of t2) */
    for (ok = TRUE, i++; ok && i < arity; i++)
      ok = lrpo(t1, t2->args[i]);
    return ok;
  }
  else {
    /* return (there is a remaining arg of t1 s.t. arg == t2 or arg > t2) */
    for (ok = FALSE, i++; !ok && i < arity; i++)
      ok = (term_ident(t1->args[i], t2) || lrpo(t1->args[i], t2));
    return ok;
  }
}  /* lrpo_lex */

/*************
 *
 *   num_occurrences()
 *
 *************/

static
int num_occurrences(Term t, Term a[], int n)
{
  int i, count;
  for (i = 0, count = 0; i < n; i++)
    if (term_ident(a[i], t))
      count++;
  return count;
}  /* num_occurrences */

/*************
 *
 *   set_of_more_occurrences() 
 *
 *   Return the list of elements with more occurrences in a1.
 *
 *************/

static
Plist set_of_more_occurrences(Term a1[], int n1, Term a2[], int n2)
{
  Plist answer = NULL;
  int i, j;
    
  for (i = 0; i < n1; i++) {
    Term e1 = a1[i];

    /* Check if this is the first occurrence of e1. */

    for (j = 0; j != i && !term_ident(e1, a1[j]); j++);

    if (i == j && num_occurrences(e1,a1,n1) > num_occurrences(e1,a2,n2))
      answer = plist_prepend(answer, e1);
  }
  return answer;
}  /* set_of_more_occurrences */

/*************
 *
 *   greater_multiset()
 *
 *   a1 >> a2 iff 
 *      (1) there is an element with more occurrences in a1; and
 *      (2) for each element with more occurrences in a2,
 *          there is a greater element with more occurrences in a1.
 *
 *************/

/* DOCUMENTATION
Given two arrays of terms, check if the first is greater
in the multiset extension of lrpo.
*/

/* PUBLIC */
BOOL greater_multiset(Term a1[], int n1, Term a2[], int n2)
{
  Plist s1, s2, p1, p2;
  BOOL ok;
  s1 = set_of_more_occurrences(a1, n1, a2, n2);  /* more occurrences in t1 */
  s2 = set_of_more_occurrences(a2, n2, a1, n1);  /* more occurrences in t2 */
  /*
   * return (s1 not empty and foreach e2 in s2
   * there is an e1 in s1 such that lrpo(e1, e2)).
   */
  if (s1 == NULL)
    ok = FALSE;
  else {
    for (p2 = s2, ok = TRUE; p2 && ok; p2 = p2->next)
      for (p1 = s1, ok = FALSE; p1 && !ok; p1 = p1->next)
	ok = lrpo(p1->v, p2->v);
  }
  zap_plist(s1);
  zap_plist(s2);
  return ok;
}  /* greater_multiset */

/*************
 *
 *   int lrpo_multiset(t1, t2) -- Is t1 > t2 in the lrpo multiset ordering?
 *
 *************/

/* DOCUMENTATION
This routine
*/

/* PUBLIC */
BOOL lrpo_multiset(Term t1, Term t2)
{
  return greater_multiset(t1->args, t1->arity, t2->args, t2->arity);
}  /* lrpo_multiset */

/*************
 *
 *    lrpo()
 *                      
 *************/

/* DOCUMENTATION
This Boolean routine checks if Term t1 > Term t2 in the
Lexicographic Recursive Path Ordering.  Function symbols can have
either multiset or left-to-right status.  See symbols.h to
set those properties.
*/

/* PUBLIC */
BOOL lrpo(Term t1, Term t2)
{
  if (VARIABLE(t1))
    /* A variable is never greater than anything. */
    return 0;  

  else if (VARIABLE(t2))
    /* t1 > variable iff t1 properly contains that variable. */
    return occurs_in(t2, t1);

  else if (SYMNUM(t1) == SYMNUM(t2) &&
	   sn_to_lrpo_status(SYMNUM(t1)) == LRPO_LR_STATUS)
    /* Both have the same "left-to-right" symbol. */
    return lrpo_lex(t1, t2);

  else {
    int i;
    BOOL ok;
    Ordertype p = sym_precedence(SYMNUM(t1), SYMNUM(t2));

    if (p == SAME_AS)
      return lrpo_multiset(t1, t2);

    else if (p == GREATER_THAN) {
      /* return (t1 > each arg of t2) */
      for (ok = TRUE, i = 0; ok && i < t2->arity; i++)
	ok = lrpo(t1, t2->args[i]);
      return ok;
    }

    else {  /* LESS_THEN or NOT_COMPARABLE */
      /* return (there is an arg of t1 s.t. arg == t2 or arg > t2) */
      for (ok = FALSE, i = 0; !ok && i < t1->arity; i++)
	ok = term_ident(t1->args[i], t2) || lrpo(t1->args[i], t2);

      return ok;
    }
  }
}  /* lrpo */

