#include "mace4.h"

#define DEBUGno

Symbol_data Symbols;  /* List of symbols */

/* This maps OPS symbol IDs to MACE symbol IDs, which start with 0. */

int Sn_map_size;
Symbol_data *Sn_to_mace_sn;

/* Cell table is indexed by eterm IDs. */

int Number_of_cells;
struct cell *Cells;           /* the table of cells (dynamically allocated) */
struct cell **Ordered_cells;  /* (pointers to) permutation of Cells */

/* Misc */

int Max_domain_element_in_input;  /* For Least Number Heuristic */

int Domain_size;
Term *Domain;    /* array of terms representing (shared) domain elements  */

Plist Ground_clauses;

int Relation_flag;  /* term flag */
int Negation_flag;  /* term flag */

int Eq_sn;
int Or_sn;
int Not_sn;

/* ---------------------- Statistics ------------------------- */

/* stats for entire run */

unsigned All_models;

/* stats for the current domain size */

int Mace4_clock;

unsigned Current_models;
unsigned Selections;
unsigned Assignments;
unsigned Propagations;
unsigned Cross_offs;
unsigned Rewrite_terms;
unsigned Rewrite_bools;
unsigned Indexes;
unsigned Ground_clauses_seen;
unsigned Ground_clauses_kept;
unsigned Rules_from_neg;

unsigned Neg_elim_attempts;
unsigned Neg_elim;
unsigned Neg_elim_agone;
unsigned Neg_elim_egone;

unsigned Neg_assign_attempts;
unsigned Neg_assign;
unsigned Neg_assign_agone;
unsigned Neg_assign_egone;

unsigned Neg_near_assign_attempts;
unsigned Neg_near_assign;
unsigned Neg_near_assign_agone;
unsigned Neg_near_assign_egone;

unsigned Neg_near_elim_attempts;
unsigned Neg_near_elim;
unsigned Neg_near_elim_agone;
unsigned Neg_near_elim_egone;

/* Ground terms.  MACE4 operates on ground clauses, which are
   represented by the structure mclause.  Ground terms (and
   atoms) are represented with ordinary OPS terms.  There are
   a few tricks:

   (1) We use upward pointers from terms to superterms
   and from atoms to clauses.
 
   (2) We need to mark atoms with a termflag, so that we know
   when to stop when following the upward pointers.  Also,
   a termflag is used to indicate that an atom is negated.

   (3) Domain elements are represented by variables (sorry,
   but it is very convenient to do so).  Also, there is only
   one actual copy of each domain element (structure sharing
   of domain elements).  Global array *Domain contains them.

   IDs.  If all of the arguments of a term (including atoms) are
   domain elements, that term is called an eterm.  For example,
   f(3,4), a, P(0), and Q.

   Each eterm has a unique ID which is used as an index into
   the cell table, for example, when a new eterm is obtained by
   evaluating a subterm to a domain element, we have to quickly
   check if this new eterm can be evaluated.  We do this by
   calculating its ID and looking up in Cells[ID].value.
   And when we have a new assignment, say f(3,4)=2, we find the
   list of occurrences of f(3,4) by looking in Cells[ID].occurrences.
*/

/*************
 *
 *   initialize_for_search()
 *
 *   This is the initialization that has to be done only once
 *   for a given set of clauses.  It is independent of the
 *   domain size.
 *
 *************/

static
void initialize_for_search(Clist clauses)
{
  Clist_pos p;
  int max, i;
  Symbol_data s;

  /* In ground clauses, VARIABLEs represent domain elements,
     so from here on, print variables as integers. */

  set_variable_style(INTEGER_STYLE);

  /* These flags are for ground clause (mclause) literals. */

  Relation_flag = claim_term_flag();
  Negation_flag = claim_term_flag();

  /* Cache some symbol numbers. */

  Eq_sn  = str_to_sn("=", 2);
  Or_sn  = str_to_sn("|", 2);
  Not_sn = str_to_sn("~", 1);

  /* Set up Symbols list. */

  init_built_in_symbols();  /* =/2 (and maybe others) */

  /* Collect data for each symbol. */

  Max_domain_element_in_input = -1;
  for (p = clauses->first; p != NULL; p = p->next) {
    Literal lit;
    for (lit = p->c->literals; lit != NULL; lit = lit->next) {
      int x = collect_mace_syms(lit->atom, RELATION);
      Max_domain_element_in_input = MAX(Max_domain_element_in_input, x);
    }
  }

  if (Max_domain_element_in_input == -1)
    printf("\n%% There are no domain elements in the input.\n");
  else
    printf("\n%% The maximum domain element in the input is %d.\n",
	   Max_domain_element_in_input);

  /* Set up map from ordinary symnums to mace symnums. */

  max = 0;
  i = 0;

  for (s = Symbols; s != NULL; s = s->next) {
    s->mace_sn = i++;
    max = (s->sn > max ? s->sn : max);
  }

  Sn_map_size = max+1;

  Sn_to_mace_sn = malloc(Sn_map_size * sizeof(void *));

  for (i = 0; i < Sn_map_size; i++)
    Sn_to_mace_sn[i] = NULL;

  for (s = Symbols; s != NULL; s = s->next) {
    Sn_to_mace_sn[s->sn] = s;
  }
}  /* initialize_for_search */

/*************
 *
 *   init_for_domain_size()
 *
 *   Given the list of (general) clauses, set up the various data
 *   structures that will be needed for a given domain size.
 *
 *************/

static
void init_for_domain_size(Clist clauses)
{
  int i, j, nextbase, id;
  Symbol_data s;

  /* Give each symbol its "base" value, which is used to index cells.  */

  nextbase = 0;
  for (s = Symbols; s != NULL; s = s->next) {
    s->base = nextbase;
    nextbase += int_power(Domain_size, s->arity);
  }

  /* Set up the array of domain terms.  All ground terms refer to these. */

  Domain = malloc(Domain_size * sizeof(void *));
  for (i = 0; i < Domain_size; i++)
    Domain[i] = get_variable_term(i);
  
  /* Set up the table of cells. */

  Number_of_cells = nextbase;
  Cells           = malloc(Number_of_cells * sizeof(struct cell));
  Ordered_cells   = malloc(Number_of_cells * sizeof(void *));

  for (id = 0; id < Number_of_cells; id++) {
    struct cell *c = Cells + id;
    int n;
    c->id = id;
    c->occurrences = NULL;
    c->value = NULL;
    c->symbol = find_symbol_node(id);
    c->eterm = decode_eterm_id(id);
    c->max_index = max_index(id, c->symbol);
    n = id_to_domain_size(id);
    c->possible = malloc(n * sizeof(void *));
    for (j = 0; j < n; j++)
      c->possible[j] = Domain[j];  /* really just a flag */
  }

  order_cells();

  if (flag(NEGPROP))
    init_negprop_index();

} /* init_for_domain_size */

/*************
 *
 *   built_in_assignments()
 *
 *************/

static
void built_in_assignments(void)
{
  Symbol_data s;
  for (s = Symbols; s != NULL; s = s->next) {
    if (s->attribute == EQUALITY_SYMBOL) {
      int i, j;
      for (i = 0; i < Domain_size; i++)
	for (j = 0; j < Domain_size; j++)
          Cells[X2(s->base,i,j)].value = (Domain[i==j ? 1 : 0]);
    }
  }
}  /* built_in_assignments */

/*************
 *
 *   special_assignments()
 *
 *************/

static
void special_assignments(void)
{
  if (flag(INTEGER_RING)) {
    /* Fix [+,-,*] as the ring of integers mod domain_size. */
    /* If any of those operations doesn't exist, then ignore it.*/
    Symbol_data s;
    for (s = Symbols; s != NULL; s = s->next) {
      int i, j;
      if (is_symbol(s->sn, "+", 2)) {
	for (i = 0; i < Domain_size; i++)
	  for (j = 0; j < Domain_size; j++)
	    Cells[X2(s->base,i,j)].value = Domain[(i + j) % Domain_size];
      }
      else if (is_symbol(s->sn, "*", 2)) {
	for (i = 0; i < Domain_size; i++)
	  for (j = 0; j < Domain_size; j++)
	    Cells[X2(s->base,i,j)].value = Domain[(i * j) % Domain_size];
      }
      else if (is_symbol(s->sn, "-", 1)) {
	for (i = 0; i < Domain_size; i++)
	  Cells[X1(s->base,i)].value = Domain[(Domain_size - i) % Domain_size];
      }
    }
    if (flag(LNH)) {
      printf("\nClearing flag lnh, because the least number heuristic\n"
	     "is not compatible with flag integer_ring.\n");
      clear_flag(LNH);
    }
  }
}  /* special_assignments */

/*************
 *
 *   check_that_ground_clauses_are_true()
 *
 *************/

static
void check_that_ground_clauses_are_true(void)
{
  Plist g;
  BOOL ok = TRUE;
  for (g = Ground_clauses; g != NULL; g = g->next) {
    Mclause c = g->v;
    if (!c->subsumed) {
      fprintf(stderr, "ERROR, model reported, but clause not true!\n");
      fprintf(stdout, "ERROR, model reported, but clause not true! ");
      p_mclause(c);
      ok = FALSE;
    }
  }
  if (!ok)
    fatal_error("check_that_ground_clauses_are_true, bad model found");
}  /* check_that_ground_clauses_are_true */

/*************
 *
 *   we_have_a_model()
 *
 *************/

static
void we_have_a_model(void)
{
  static int next_message = 1;
  All_models++;
  Current_models++;
  check_that_ground_clauses_are_true();  /* in case of a soundness bug */
  if (flag(PRINT_MODELS_PORTABLE))
    print_model_portable(stdout, TRUE);
  else if (flag(PRINT_MODELS))
    p_model(TRUE);
  else if (next_message == All_models) {
    printf("\nModel %d has been found.\n", All_models);
    next_message *= 10;
  }
  fflush(stdout);
  if (parm(MAX_MODELS) != -1 && All_models >= parm(MAX_MODELS))
    mace4_exit(MAX_MODELS_EXIT);
}  /* we_have_a_model */

/*************
 *
 *   check_time()
 *
 *************/

static
void check_time(void)
{
  if (parm(MAX_SECONDS) != -1 && user_time() > parm(MAX_SECONDS) * 1000)
    mace4_exit(All_models > 1 ? MAX_SEC_YES_EXIT : MAX_SEC_NO_EXIT);
}  /* check_time */

/*************
 *
 *   p_possible_values()
 *
 *************/

void p_possible_values(void)
{
  int i;
  for (i = 0; i < Number_of_cells; i++) {
    if (Cells[i].symbol->attribute == ORDINARY_SYMBOL) {
      int j;
      printf("Cell %d: ", i);
      for (j = 0; j < id_to_domain_size(i); j++) {
	if (Cells[i].possible[j] != NULL)
	  printf(" %d", j);
      }
      printf("\n");
    }
  }
}  /* p_possible_values */

/*************
 *
 *   search()
 *
 *   Max_constrained is the maximum constrained domain element
 *   (or -1 is none is constrained).  Greater domain elements
 *   can all be considered symmetric.  An element can become
 *   constrained in two ways:  (1) it is an index of some selected
 *   cell, or (2) it is the value assigned to some selected cell.
 *   (Propagation does not constrain elements.  This might need
 *   careful justification.)
 *
 *   To apply the least number heuristic, we consider values
 *   0 ... MIN(max_constrained+1, Domain_size-1).
 *
 *   To make this effective, we should keep max_constrained as low as
 *   possible by selecting cells with maximum index <= max_constrained.
 *
 *************/

static
void search(int max_constrained, int depth)
{
  int id;
  check_time();
  id = select_cell(max_constrained);
  if (id == -1)
    we_have_a_model();
  else {
    int i, last;  /* we'll do 0 .. last */
    int x = Cells[id].max_index;
    max_constrained = MAX(max_constrained, x);
    Selections++;

    if (flag(TRACE)) {
      printf("select: ");
      p_model(FALSE);
      /* p_possible_values(); */
    }
	
    if (Cells[id].symbol->type == RELATION)
      last = 1;
    else if (flag(LNH))
      last = MIN(max_constrained+1, Domain_size-1);
    else
      last = Domain_size-1;

    for (i = 0; i <= last; i++) {
      Estack stk;
      Assignments++;

      if (flag(TRACE)) {
	printf("assign: ");
	fwrite_term(stdout, Cells[id].eterm);
	printf("=%d (%d) depth=%d\n", i, last, depth);
      }
	
      stk = assign_and_propagate(id, Domain[i]);

      if (stk != NULL) {
	search(MAX(max_constrained, i), depth+1);
	restore_from_stack(stk);
      }
    }
  }
}  /* search */

/*************
 *
 *   mace4() -- look for a model of a specific size
 *
 *************/

static
void mace4(Clist clauses, int order)
{
  Clist_pos p;
  Plist g;
  int i;
  Mstate initial_state = get_mstate();

  if (Max_domain_element_in_input >= order)
    fatal_error("mace4: domain_element too big");

  Domain_size = order;

  init_for_domain_size(clauses);

  built_in_assignments();  /* Fill out equality table (and maybe others). */

  special_assignments();  /* assignments determined by options */

  /* Instantiate clauses over the domain.  This also 
     (1) makes any domain element constants into real domain elements,
     (2) applies OR, NOT, and EQ simplification, and
     (3) does unit propagation (which pushes events onto initial_state->stack).
     Do the units first, then the 2-clauses, then the rest. */

  for (p = clauses->first; initial_state->ok && p != NULL; p = p->next)
    if (number_of_literals(p->c) < 2)
      generate_ground_clauses(p->c, initial_state);

  for (p = clauses->first; initial_state->ok && p != NULL; p = p->next)
    if (number_of_literals(p->c) == 2)
      generate_ground_clauses(p->c, initial_state);

  for (p = clauses->first; initial_state->ok && p != NULL; p = p->next)
    if (number_of_literals(p->c) > 2)
      generate_ground_clauses(p->c, initial_state);

  /* The preceding calls push propagation events onto initial_state->stack.
     We won't have to undo those initial events during the search,
     but we can undo them after the search.
  */

  if (flag(VERBOSE)) {
    printf("\nInitial partial model:\n");
    p_model(FALSE);
    fflush(stdout);
  }

  /* Here we go! */

  if (initial_state->ok)
    search(Max_domain_element_in_input, 0);

  /* OK, we're done.  (We may have exited from within search.) */

  /* Free all of the memory associated with the current domain size. */

  restore_from_stack(initial_state->stack);
  free_mstate(initial_state);

  if (flag(NEGPROP))
    free_negprop_index();

  free(Ordered_cells);
  Ordered_cells = NULL;

  for (i = 0; i < Number_of_cells; i++) {
    zap_mterm(Cells[i].eterm);
    free(Cells[i].possible);
  }
  free(Cells);
  Cells = NULL;

  for (i = 0; i < Domain_size; i++)
    zap_term(Domain[i]);
  free(Domain);
  Domain = NULL;

  for (g = Ground_clauses; g != NULL; g = g->next)
    zap_mclause(g->v);
  zap_plist(Ground_clauses);
  Ground_clauses = NULL;

}  /* mace4 */

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

static
void print_banner(int argc, char **argv)
{
  int i;
  int pid = 0;
    
  printf("----- %s %s, %s -----\n",
	 PROGRAM_NAME, PROGRAM_VERSION, PROGRAM_DATE);
  printf("Built with library %s %s, %s.\n",
	 LIBRARY_NAME, LIBRARY_VERSION, LIBRARY_DATE);
  printf("The process was started by %s on %s,\n%s",
	 username(), hostname(), get_date());
  pid = my_process_id();

  printf("The command was \"");
  for(i = 0; i < argc; i++)
    printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
  printf("\".");
  if (pid != 0)
    printf("  The process ID is %d.\n\n", pid);
  else
    printf("\n\n");
    
}  /* print_banner */

/*************
 *
 *   mace4_exit()
 *
 *************/

void mace4_exit(int exit_code)
{
  char *m = (All_models == 1 ? "model" : "models");

  if (flag(VERBOSE)) {
    p_mem();

    if (exit_code == MAX_MODELS_EXIT ||
	exit_code == MAX_SEC_YES_EXIT ||
	exit_code == MAX_SEC_NO_EXIT)
      p_stats();
  }

  switch(exit_code) {
  case MAX_MODELS_EXIT:
    fprintf(stdout, "\n----- MAX_MODELS EXIT (%u %s) -----\n", All_models, m);
    fprintf(stderr, "\n----- MAX_MODELS EXIT (%u %s) -----\n", All_models, m);
    break;
  case ALL_MODELS_EXIT:
    fprintf(stdout, "\n----- ALL_MODELS EXIT (%u %s) -----\n", All_models, m);
    fprintf(stderr, "\n----- ALL_MODELS EXIT (%u %s) -----\n", All_models, m);
    break;
  case EXHAUSTED_EXIT:
    fprintf(stdout, "\n----- exhausted exit -----\n");
    fprintf(stderr, "\n----- exhausted exit -----\n");
    break;
  case MAX_SEC_YES_EXIT:
    fprintf(stdout, "\n----- max_seconds exit (with models) -----\n");
    fprintf(stderr, "\n----- max_seconds exit (with models) -----\n");
    break;
  case MAX_SEC_NO_EXIT:
    fprintf(stdout, "\n----- max_seconds exit (no models) -----\n");
    fprintf(stderr, "\n----- max_seconds exit (no models) -----\n");
    break;
  default:
    fprintf(stdout, "\n----- unknown exit code: %d -----\n", exit_code);
    fprintf(stderr, "\n----- unknown exit code: %d -----\n", exit_code);
    break;
  }
  if (!flag(VERBOSE))
    printf("CPU time: %.2f seconds.\n", user_time() / 1000.0);
  printf("The process finished %s", get_date());
  exit(exit_code);
}  /* mace4_exit */

/*************
 *
 *   read_clauses()
 *
 *   In case the input file designed for a theorem prover, we allow
 *   (and ignore) unrecognized lists.
 *
 *************/

static
Clist read_clauses(Term listid, FILE *fin, FILE *fout, BOOL allow_unknown_lists)
{
  Term t = listid;
  Clist clauses = clist_init("input_clauses");

  while (t != NULL) {
    if (is_term(t, "clauses", 1)) {
      char *listname = term_symbol(ARG(t,0));
      Clist l = read_clause_list(fin, fout, listname, FALSE);
      fwrite_clause_list(fout, l, FALSE, FALSE);
      clist_append_all(clauses, l);
    }
    else if (is_term(t, "formulas", 1)) {
      char *listname = term_symbol(ARG(t,0));
      Plist forms = read_formula_list(fin, fout);
      Clist l = clausify_list(forms);
      fwrite_formula_list(fout, forms, listname);
      check_constants_in_clist(l);
      name_clist(l, "clausified_formulas");
      fprintf(fout, "\n%% The preceding list of formulas produces the "
	      "following list of clauses.\n");
      fwrite_clause_list(fout, l, FALSE, FALSE);
      clist_append_all(clauses, l);
    }
    else if (allow_unknown_lists &&
	     (str_ident(term_symbol(t),"clauses") ||
	      str_ident(term_symbol(t),"formulas") ||
	      str_ident(term_symbol(t),"terms"))) {
      Plist g = read_term_list(fin, fout);
      zap_term_list(g);
      fprintf(fout, "\nNOTE: The list \"");
      fwrite_term(fout, t);
      fprintf(fout, "\" was read and ignored.\n");
    }
    else {
      fwrite_term_nl(stdout, t);
      fwrite_term_nl(stderr, t);
      fatal_error("bad list identifier");
    }
    zap_term(t);
    t = read_term(fin, fout);
  }
  return clauses;
}  /* read_clauses */

/*************
 *
 *   overall_init()
 *
 *************/

void overall_init(void)
{
  declare_standard_parse_types();  /* we always have to do this */
  translate_neg_equalities(TRUE);  /* we can use a != b instead of ~(a = b) */

  init_options();

  init_parm(DOMAIN_SIZE,      "domain_size",      2,       2, 200);
  init_parm(ITERATE_UP_TO,    "iterate_up_to",    0,      -1, 200);
  init_parm(MAX_MODELS,       "max_models",       1,      -1, INT_MAX);
  init_parm(MAX_SECONDS,      "max_seconds",     -1,      -1, INT_MAX);
  init_parm(SELECTION_ORDER,  "selection_order",  2, 0, 2); /* see select.c */
  init_parm(SELECTION_MEASURE,"selection_measure",4, 0, 4); /* see select.c */
  init_parm(MAX_MEGS,         "max_megs",         192,     0, INT_MAX);
  init_parm(INCREMENT,        "increment",        1,       1, INT_MAX);
         
  init_flag(PRINT_MODELS,            "print_models",            TRUE);
  init_flag(PRINT_MODELS_PORTABLE,   "print_models_portable",   FALSE);
  init_flag(LNH,                     "lnh",                     TRUE);
  init_flag(TRACE,                   "trace",                   FALSE);
  init_flag(NEGPROP,                 "negprop",                 TRUE);
  init_flag(NEG_ASSIGN,              "neg_assign",              TRUE);
  init_flag(NEG_ASSIGN_NEAR,         "neg_assign_near",         TRUE);
  init_flag(NEG_ELIM,                "neg_elim",                TRUE);
  init_flag(NEG_ELIM_NEAR,           "neg_elim_near",           TRUE);
  init_flag(PROLOG_STYLE_VARIABLES,  "prolog_style_variables",  FALSE);
  init_flag(VERBOSE,                 "verbose",                 FALSE);
  init_flag(INTEGER_RING,            "integer_ring",            FALSE);
  init_flag(ITERATE_PRIMES,          "iterate_primes",          FALSE);

  Mace4_clock = clock_init("Mace4");
}  /* overall_init */

/*************
 *
 *   read_the_input()
 *
 *************/

Clist read_the_input(int argc, char **argv, BOOL allow_unknown_things)
{
  /* First, read any commands in the input.  Second, get options from
     the command line (these can override the commands in the input).
     Third, read the clauses in the input (this has to be after we
     have all the options, because they can affect the reading
     of clauses).
  */

  Term t = read_commands(stdin, stdout, TRUE,
			 allow_unknown_things ? NOTE_UNKNOWN : KILL_UNKNOWN);
  /* t is the first unregonized term, and it should be a list identifer. */

  process_command_line_args(argc, argv);

  if (flag(PROLOG_STYLE_VARIABLES))
    set_variable_style(PROLOG_STYLE);
  set_max_megs(parm(MAX_MEGS));

  return read_clauses(t, stdin, stdout, allow_unknown_things);
}  /* read_the_input */

/*************
 *
 *   next_domain_size()
 *
 *************/

int next_domain_size(n)
{
  int top = parm(ITERATE_UP_TO);

  if (top == 0)
    top = parm(DOMAIN_SIZE);
  else if (top == -1)
    top = INT_MAX;
      
  if (flag(ITERATE_PRIMES)) {
    n++;
    while (!prime(n))
      n++;
  }
  else
    n += parm(INCREMENT);

  return (n > top ? -1 : n);
}  /* next_domain_size */

/*************
 *
 *   main()
 *
 *************/

int main(int argc, char **argv)
{
  Clist clauses;
  int n;
  BOOL prover_compatability_mode;  /* ignore unregognized set/clear/assigns? */

  overall_init();  /* We must do this before calling usage_message. */

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "-help")) {
    usage_message(stderr);
    exit(1);
  }

  print_banner(argc, argv);

  prover_compatability_mode = member_args(argc, argv, "-c");

  clauses = read_the_input(argc, argv, prover_compatability_mode);

  initialize_for_search(clauses);

  n = parm(DOMAIN_SIZE);

  while (n >= 2) { 
    fprintf(stdout,"\n====== Starting search for domain size %d. ======\n",n);
    fprintf(stderr,"\n====== Starting search for domain size %d. ======\n",n);
    fflush(stdout);
    clock_start(Mace4_clock);
    mace4(clauses, n);
    clock_stop(Mace4_clock);
    if (flag(VERBOSE))
      p_stats();
    reset_current_stats();
    clock_reset(Mace4_clock);
    n = next_domain_size(n);
  }

  clist_zap(clauses);

  /* Note that max_models and max_seconds exits occur within mace4(). */

  mace4_exit(All_models == 0 ? EXHAUSTED_EXIT : ALL_MODELS_EXIT);

  exit(0);  /* This won't happen, but the compiler wants it. */

}  /* main */

