#include "mace4.h"

/*
 * memory management
 */

static Mstate mstate_avail;
static unsigned mstate_gets, mstate_frees, mstate_avails;

static Jnode jnode_avail;
static unsigned jnode_gets, jnode_frees, jnode_avails;

/* end of counters for each type */

/*************
 *
 *    Mstate get_mstate()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a Mstate node.
*/

/* PUBLIC */
Mstate get_mstate(void)
{
  Mstate p;
    
  mstate_gets++;
  if (mstate_avail == NULL)
    p = tp_alloc(sizeof(struct mstate));
  else {
    mstate_avails--;
    p = mstate_avail;
    mstate_avail = mstate_avail->next;
  }
  p->first_job = NULL;
  p->last_job = NULL;
  p->ok = TRUE;
  p->stack = NULL;
  p->next = NULL;
  return(p);
}  /* get_mstate */

/*************
 *
 *    free_mstate()
 *
 *************/

/* DOCUMENTATION
This routine frees a Mstate node.  If you wish to free a whole
Mstate list, call zap_mstate();
*/

/* PUBLIC */
void free_mstate(Mstate p)
{
  mstate_frees++;
  mstate_avails++;
  p->next = mstate_avail;
  mstate_avail = p;
}  /* free_mstate */

/*************
 *
 *    Jnode get_jnode()
 *
 *************/

/* DOCUMENTATION
This routine allocates and returns a Jnode node.
*/

/* PUBLIC */
Jnode get_jnode(void)
{
  Jnode p;
    
  jnode_gets++;
  if (jnode_avail == NULL)
    p = tp_alloc(sizeof(struct jnode));
  else {
    jnode_avails--;
    p = jnode_avail;
    jnode_avail = jnode_avail->next;
  }
  return(p);
}  /* get_jnode */

/*************
 *
 *    free_jnode()
 *
 *************/

void free_jnode(Jnode p)
{
  jnode_frees++;
  jnode_avails++;
  p->next = jnode_avail;
  jnode_avail = p;
}  /* free_jnode */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_mstate_mem()
 *
 *************/

void fprint_mstate_mem(FILE *fp, BOOL heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");
  
  n = sizeof(struct mstate);
  fprintf(fp, "mstate (%4d)       %11u%11u%11u%11u%9.1f K\n",
	  n, mstate_gets, mstate_frees, mstate_gets - mstate_frees,
	  mstate_avails,
	  (((mstate_gets - mstate_frees) + mstate_avails) * n) / 1024.);

  n = sizeof(struct jnode);
  fprintf(fp, "jnode (%4d)        %11u%11u%11u%11u%9.1f K\n",
	  n, jnode_gets, jnode_frees, jnode_gets - jnode_frees,
	  jnode_avails,
	  (((jnode_gets - jnode_frees) + jnode_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_mstate_mem */

/*************
 *
 *   p_mstate_mem()
 *
 *************/

void p_mstate_mem(void)
{
  fprint_mstate_mem(stdout, 1);
}  /* p_mstate_mem */

/*
 *  end of memory management
 */

/*************
 *
 *   job_append()
 *
 *************/

void job_append(Mstate s, int type, int id, Term alpha, Term beta, int pos)
{
  Jnode j = get_jnode();
  j->type = type;
  j->id = id;
  j->alpha = alpha;
  j->beta = beta;
  j->pos = pos;

  if (s->first_job == NULL) {
    j->prev = NULL;
    j->next = NULL;
    s->first_job = j;
    s->last_job = j;
  }
  else {
    j->prev = s->last_job;
    j->next = NULL;
    s->last_job->next = j;
    s->last_job = j;
  }
}  /* job_append */

/*************
 *
 *   job_prepend()
 *
 *************/

void job_prepend(Mstate s, int type, int id, Term alpha, Term beta, int pos)
{
  Jnode j = get_jnode();
  j->type = type;
  j->id = id;
  j->alpha = alpha;
  j->beta = beta;
  j->pos = pos;

  if (s->first_job == NULL) {
    j->prev = NULL;
    j->next = NULL;
    s->first_job = j;
    s->last_job = j;
  }
  else {
    j->next = s->first_job;
    j->prev = NULL;
    s->first_job->prev = j;
    s->first_job = j;
  }
}  /* job_prepend */

/*************
 *
 *   job_pop()
 *
 *************/

void job_pop(Mstate s)
{
  if (s->first_job == NULL)
    fatal_error("job_pop: empty list");
  else {
    Jnode p = s->first_job;
    s->first_job = p->next;
    if (s->first_job == NULL)
      s->last_job = NULL;
    else
      s->first_job->prev = NULL;
    free_jnode(p);
  }
}  /* job_pop */

/*************
 *
 *   zap_jobs()
 *
 *************/

void zap_jobs(Mstate ms)
{
  int n;
  Jnode j;
  for (j = ms->first_job, n = 0; j != NULL; j = j->next, n++);
  jnode_frees += n;
  jnode_avails += n;
  if (ms->first_job != NULL) {
    ms->last_job->next = jnode_avail;
    jnode_avail = ms->first_job;
  }
}  /* zap_jobs */
