#include "mace4.h"

extern Symbol_data Symbols;
extern int Domain_size;

extern struct cell *Cells;

/*************
 *
 *    Symbol_data get_symbol_data()
 *
 *************/

Symbol_data get_symbol_data(void)
{
  Symbol_data p = tp_alloc(sizeof(struct symbol_data));

  p->sn = 0;
  p->mace_sn = 0;
  p->arity = 0;
  p->type = 0;
  p->base = 0;
  p->attribute = ORDINARY_SYMBOL;
  p->next = NULL;
  return(p);
}  /* get_symbol_data */

/*************
 *
 *    Symbol_data find_symbol_data()
 *
 *************/

Symbol_data find_symbol_data(int sn)
{
  Symbol_data p = Symbols;
  while (p != NULL && p->sn != sn)
    p = p->next;
  return p;
}  /* find_symbol_data */

/*************
 *
 *   init_built_in_symbols()
 *
 *************/

void init_built_in_symbols(void)
{
  Symbol_data s;

  s = get_symbol_data();
  s->sn = str_to_sn("=", 2);
  s->arity = 2;
  s->type = RELATION;
  s->attribute = EQUALITY_SYMBOL;
  s->next = Symbols;
  Symbols = s;

}  /* init_built_in_symbols */

/*************
 *
 *   compare_sn()
 *
 *   1. lex_val  (lex command)
 *   2. functions < relations
 *   3. arity n < arity n+1
 *   4. symbol id (sn)
 *
 *************/

static
Ordertype compare_sn(Symbol_data s1, Symbol_data s2)
{
  int lv1 = sn_to_lex_val(s1->sn);
  int lv2 = sn_to_lex_val(s2->sn);
  if (lv1 < lv2)
    return LESS_THAN;
  else if (lv1 > lv2)
    return GREATER_THAN;
  else {
    if (s1->type == FUNCTION && s2->type == RELATION)
      return LESS_THAN;
    else if (s1->type == RELATION && s2->type == FUNCTION)
      return GREATER_THAN;
    else {
      int a1 = sn_to_arity(s1->sn);
      int a2 = sn_to_arity(s2->sn);
      if (a1 < a2)
	return LESS_THAN;
      else if (a1 > a2)
	return GREATER_THAN;
      else {
	if (s1->sn < s2->sn)
	  return LESS_THAN;
	else if (s1->sn > s2->sn)
	  return GREATER_THAN;
	else
	  return SAME_AS;
      }
    }
  }
}  /* compare_sn */

/*************
 *
 *   insert_mace_sym()
 *
 *************/

static
Symbol_data insert_mace_sym(Symbol_data head, Symbol_data new)
{
  /* order by ordinary symnum */
  if (head == NULL)
    return new;
  else if (compare_sn(new, head) == LESS_THAN) {
    new->next = head;
    return new;
  }
  else {
    head->next = insert_mace_sym(head->next, new);
    return head;
  }
}  /* insert_mace_sym */

/*************
 *
 *   collect_mace_syms()
 *
 *   Return the largest domain element seen (-1 if none).
 *
 *************/

int collect_mace_syms(Term t, int type)
{
  if (VARIABLE(t))
    return -1;
  else {
    int i;
    int max = -1;
    Symbol_data s = find_symbol_data(SYMNUM(t));
    if (s != NULL) {
      /* we've already seen this symbol */
      if (type != s->type) {
	printf("\nThe overloaded symbol is: %s\n", sn_to_str(SYMNUM(t)));
	fatal_error("collect_mace_syms: symbol is both relation and function");
      }
    }
    else {
      max = natural_constant(t);
      if (max == -1) {
	s = get_symbol_data();
	s->sn = SYMNUM(t);
	s->arity = sn_to_arity(SYMNUM(t));
	s->type = type;
	Symbols = insert_mace_sym(Symbols, s);
      }
      else if (type == RELATION) {
	printf("\nThe bad symbol is: %s\n", sn_to_str(SYMNUM(t)));
	fatal_error("collect_mace_syms: relation symbol is domain element");
      }
    }
    for (i = 0; i < ARITY(t); i++) {
      int x = collect_mace_syms(ARG(t,i), FUNCTION);
      max = MAX(max, x);
    }
    return max;
  }
}  /* collect_mace_syms */

/*************
 *
 *   p_symbol_data()
 *
 *************/

void p_symbol_data(Symbol_data s)
{
  if (s != NULL)
    printf("%s/%d, sn=%d, mace_sn=%d, base=%d, type=%d, attribute=%d.\n",
	   sn_to_str(s->sn), s->arity,
	   s->sn, s->mace_sn, s->base, s->type, s->attribute);
}  /* p_symbol_data */

/*************
 *
 *   find_symbol_node()
 *
 *************/

Symbol_data find_symbol_node(int id)
{
  /* Assume bases are increasing and that the id is in range.
     Return node with the largest base <= the id. */
  Symbol_data prev = NULL;
  Symbol_data curr = Symbols;
  while (curr != NULL && curr->base <= id) {
    prev = curr;
    curr = curr->next;
  }
  return prev;
}  /* find_symbol_node */

/*************
 *
 *   id_to_domain_size()
 *
 *************/

int id_to_domain_size(int id)
{
  /* Assume the id is in range. */
  Symbol_data s = Cells[id].symbol;
  return (s->type == RELATION ? 2 : Domain_size);
}  /* id_to_domain_size */

/*************
 *
 *   max_index()
 *
 *   If the cell given by ID represents f(i,j,k,...), return the 
 *   maximum of (i,j,k,...).  If it is a constant, return -1.
 *
 *************/

int max_index(int id, Symbol_data s)
{
  int max = -1;
  int n = Domain_size;
  int x = id - s->base;
  int i;
  for (i = s->arity - 1; i >= 0; i--) {
    int p = int_power(n, i);
    int e = x / p;
    max = MAX(e,max);
    x = x % p;
  }
  return max;
}  /* max_index */

