/**************************************************************************************************

VarOrder.h -- (C) Niklas Een, Niklas Srensson, 2004

ADT for maintaining the variable ordering. It will keep a list of all decision
variables sorted on the current activity.

**************************************************************************************************/

#ifndef VarOrder_h
#define VarOrder_h

#include "SolverTypes.h"


//=================================================================================================


class VarOrder {
    const vec<char>&    assigns;    // var->val. Pointer to external assignment table.
    const vec<double>&  activity;   // var->act. Pointer to external activity table.
    vec<int>            varpos;     // var->pos.
    vec<Var>            order;      // pos->var.  

    double  random_seed;            // For the internal random number generator

    int     last_var;

public:
    VarOrder(const vec<char>& ass, const vec<double>& act) :
        assigns(ass), activity(act), random_seed(91648253), last_var(0) { }


    inline void newVar(bool decision_var = true);
    inline void update(Var x);                  // Called when variable increased in activity.
           void updateAll(void);                // Called when all variables have been assigned new activity.
    inline void undo(Var x);                    // Called when variable is unassigned and may be selected again.
    inline Var  select(double random_freq =.0); // Selects a new, unassigned variable (or 'var_Undef' if none exists).

    inline bool decisionVar(Var x) { return varpos[x] != -1; }
};


void VarOrder::newVar(bool decision_var)
{
    varpos.push(decision_var ? order.size() : -1);
    if (decision_var) order.push(varpos.size()-1);
}


void VarOrder::update(Var x)
{
    if (decisionVar(x)) {
        for (int i = varpos[x]; i > 0 && activity[order[i-1]] < activity[x]; i--){
            assert(order[i] == x);
            varpos[order[i-1]]++; varpos[order[i]]--;
            order[i-1] ^= order[i]; order[i] ^= order[i-1]; order[i-1] ^= order[i];
        }
    }
}


void VarOrder::undo(Var x)
{
    if (decisionVar(x) && varpos[x] < last_var)
        last_var = varpos[x];
}


Var VarOrder::select(double random_var_freq)
{
    // Random decision:
    if (drand(random_seed) < random_var_freq){
        Var next = order[irand(random_seed, order.size())];
        assert(next >= 0 && next < assigns.size());
        if (toLbool(assigns[next]) == l_Undef)
            return next;
    }

    // Activity based decision:
    for (int i = last_var; i < order.size(); i++){
        if (toLbool(assigns[order[i]]) == l_Undef){
            last_var = i;
            return order[i]; }
    }

    return var_Undef;
}


//=================================================================================================
#endif
