{-# OPTIONS -fglasgow-exts #-} -- options: -auto-all
------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Flatten where

import Data
import Set
import List ( (\\), sortBy, minimumBy, partition )
import AnalysisTypes
import Maybe
  ( fromJust
  )

import Monad
  ( mplus
  )

import List
  ( groupBy
  , sort
  )

import Trace

-------------------------------------------------------------------------
-- macify

macify :: [Clause] -> ([(Type,Int)], [Clause], [QClause])
macify cs = ([(t,length ts)|(t,ts) <- cliques], flattenedCs, functionCs)
 where
  flattenedCs =
    [ c'
    | (c, App _ (Elt i) _) <- tabulate assignCons
    , let d i = Elt i `app` []
    , c' <- unit (Pos (Equal `prd` [c, d i]))
          : [ unit (Neg (Equal `prd` [c, d j]))
            | j <- [1..i-1]
            ]
    ] ++
    splitAll (concatMap (flatten assignCons) (defCs ++ coreCs))
  
  (ds, defCs, coreCs) =
    definitions cs
    -- ([], [], cs)
    
  functionCs =
    [ Uniq y
      ( unit (Pos (t =. y))
      )
    | (f,n) <- elements fs ++ ds
    , let FType tps tp = typing f
          xs = [ Var (varNameX % i ::- VType tp) | (i,tp) <- [1..n] `zip` tps ]
          y  = varNameY ::- VType tp
          t  = f `app` xs
    , t `notMember` assignCons
    ]

  constants =
    map (\g@((t,_):_) -> (t,map snd g)) $
    groupBy (\x y -> fst x == fst y) $
    sort $
    [ (typ t,t)
    | (c,0) <- elements fs
    , let t = c `app` []
    ]

  units =
    [ lit
    | [lit] <- map elements cs
    , size (free lit) <= 2
    ]

  cliques =
    [ (t,findClique cs units)
    | (t,cs) <- constants
    ]

  assignCons =
    table [ (t, Elt i `app` [])
          | (_,ts) <- cliques
          , (t,i)  <- ts `zip` [1..]
          ]

  (_,fs,_) = symbols cs

(=.) :: Term -> Symbol -> Atom
t =. v = Equal `prd` [t, Var v]

findClique :: [Term] -> [Lit] -> [Term]
findClique [] units = []
findClique ts units = elements (largestClique empty graph)
 where
  cons = set ts
  
  (posInfo, negInfo) = foldr gather (empty, empty) units
   where
    gather (Pos p) (pos, neg) = (pos `union` inst p, neg)
    gather (Neg p) (pos, neg) = (pos, neg `union` inst p)
  
    inst p = set [ norm (subst sub p) | sub <- subs (elements (free p)) ]
     where
      norm (Pred _ Equal [a,b]) | b < a = Equal `prd` [b,a]
      norm p                            = p
    
    subs []     = [empty]
    subs (x:xs) = [ extend x t sub | sub <- subs xs, t <- ts ]
  
  edges = 
    [ (a,b)
    | (a,b) <- pairs (elements cons)
    , a `notEqual` b
    ]
   where
    a `notEqual` b = a `eqNotEqual` b || a `repNotEqual` b

    a `eqNotEqual` b =
      (Equal `prd` [a,b]) `member` negInfo

    a `repNotEqual` b =
      size (rep posInfo `intersect` rep negInfo) > 0
     where
      rep xs = set [ p `prd` map (replace a b) ys
                   | Pred _ p ys <- elements xs
                   ]

      replace a b a' | a == a' = b
      replace a b (App _ f xs) = f `app` map (replace a b) xs
  
  graph = 
    accumulate union $
      concat [ [(a, unit b), (b, unit a)]
             | (a,b) <- edges
             ]
        ++ [ (a, empty)
           | a <- ts
           ]
  
  largestClique cl gr =
    case pick gr of
      Nothing -> cl
      Just ((a,bs),rest)
        | size cl >= 1 + size bs
           || size cl >= size cl' -> largestClique cl gr'
        | otherwise               -> largestClique cl' gr'
       where
        gr'   = mapElt (a `delete`) rest
        cl'   = a `insert` largestClique empty subgr
        subgr = table [ (x, xs `intersect` bs)
                      | x <- elements bs
                      , let xs = fromJust (look x gr')
                      ]

  pairs []     = []
  pairs (x:xs) = [ (x,y) | y <- xs ] ++ pairs xs

-------------------------------------------------------------------------
-- definitions

definitions :: [Clause] -> ([(Symbol,Int)], [Clause], [Clause])
definitions cs = ([ (f,length xs) | (_, App _ f xs) <- list ], defCs, coreCs)
 where
  deepTerms =
    [ t'
    | t <- elements (subterms cs)
    , let (_,t') = normalize t
    , isOkTerm (t' `member` ts) t'
    ]
   where
    ts = topterms cs
   
  list =
    [ (t, (defName % i ::- FType (map typ xs) (typ t)) `app` xs)
    | (t,i) <- deepTerms `zip` [1..]
    , let xs = map Var (elements (free t))
    ]

  defCs =
    [ unit (Pos (Equal `prd` [deft, f `app` map replace xs]))
    | (App _ f xs, deft) <- list
    ]

  coreCs =
    [ set (map replaceLit (elements c))
    | c <- cs
    ]

  tab =
    table list

  normalize t =
    norm t empty [varNameX % i | i <- [1..]] (\t' sub _ -> (sub,t'))
   where
    norm (Var v)      sub (w:ws) k = k (Var v) (extend (w ::- typing v) (Var v) sub) ws
    norm (App _ f xs) sub ws     k =
      norms xs sub ws $ \xs' sub' ws' ->
        k (f `app` xs') sub' ws'
    
    norms []     sub ws k = k [] sub ws
    norms (t:ts) sub ws k =
      norm t sub ws $ \t' sub' ws' ->
        norms ts sub' ws' $ \ts' sub'' ws'' ->
          k (t':ts') sub'' ws''
  
  isOkTerm top (Var _)      = False
  isOkTerm top (App _ _ xs) = any (not . isVar) xs
                          -- && size (free xs) <= 1
                           && any (isAlreadyConnected (free xs)) xs
                          -- && (not top || all isGroundTerm xs)
   where
    isVar (Var _) = True
    isVar _       = False
    
    isAlreadyConnected vs (Var v)      = vs `subset` set [v]
    isAlreadyConnected vs (App _ _ xs) = vs `subset` set [ v | Var v <- xs ]
                                      || any (isAlreadyConnected vs) xs

  topterms []     = empty
  topterms (c:cs) = tops (elements c) `union` topterms cs
   where
    tops ls =
      set [ t
          | l <- ls
          , t <- case l of
                   Pos (Pred _ _ ts) -> ts
                   Neg (Pred _ _ ts) -> ts
          ]

  replaceLit (Pos a) = Pos (replaceAtom a)
  replaceLit (Neg a) = Neg (replaceAtom a)

  replaceAtom (Pred _ p xs) = p `prd` map replace xs
  
  replace (Var v)        = Var v
  replace t@(App _ f xs) =
    case look t' tab of
      Nothing  -> f `app` map replace xs
      Just t'' -> subst sub t''
   where
    (sub,t') = normalize t

-------------------------------------------------------------------------
-- flatten

flatten :: (Term :=> Term) -> Clause -> [Clause]
flatten assignCons ls = simplify (defs `union` flatLs)
 where
  fls = free ls
  
  vs = [ v
       | i <- [1..]
       , let v = varNameX % i
       ]
  
  tab =
    table
      [ (t,v)
      | (t,fv) <- elements (subterms ls) `zip` vs
      , let v = case t of
                  Var v -> v
                  _     -> fv ::- VType (typ t)
      ]
  
  var t =
    case look t assignCons of
      Just ci                       -> ci
      Nothing 
        | tdomain (typ t) == Just 1 -> Elt 1 `app` []
        | otherwise                 -> Var (fromJust (look t tab))

  defs =
    set
      [ Neg ((f `app` map var ts) =. v)
      | (t@(App _ f ts), v) <- tabulate tab
      , t `notMember` assignCons
      ]
  
  flatLs =
    set
      [ flat `fmap` l
      | l <- elements ls
      ]
  
  flat (Pred _ p ts) =
    p `prd` map var ts

-------------------------------------------------------------------------
-- simplify

-- rules:
-- P | -P               | C         ==>   { }                          (truth)
-- X  = X               | C         ==>   { }                          (truth)
-- X != X               | C         ==>   { C }                        (falsity)
-- X != Y               | C[X,Y]    ==>   { C[X,X] }                   (subst)
-- f(X1..Xk) != Y | Z=Y | C[X,Z]    ==>   { f(X1..Xk) = Z | C[X,Z] }   (subst)

simplify :: Clause -> [Clause]
simplify = simp
 where
  simp c0 =
    [ c4
    | c1 <- trivial c0
    , c2 <- identEq c1
    , c3 <- substVarEq c2
    , c4 <- substFunEq c3
    ]
  
  trivial ls
    | any ((`member` ls) . neg) (elements ls) = []
    | otherwise                               = [ls]
  
  identEq ls =
    case [ ()
         | Pos (Pred _ Equal [s, t]) <- elements ls
         , s == t
         ] of
      [] -> [ls]
      _  -> []
  
  substVarEq ls = substVar [] (elements ls)
   where
    substVar ls' (Neg (Pred _ Equal [Var v, Var w]) : ls) =
      simp (set (subst (pair v (Var w)) (ls' ++ ls)))
    
    substVar ls' (Neg (Pred _ Equal [Var v, t@(App _ (Elt _) [])]) : ls) =
      simp (set (subst (pair v t) (ls' ++ ls)))
    
    substVar ls' (Neg (Pred _ Equal [t@(App _ (Elt _) []), Var v]) : ls) =
      simp (set (subst (pair v t) (ls' ++ ls)))
    
    substVar ls' (Neg (Pred _ Equal [t1@(App _ (Elt _) []), t2@(App _ (Elt _) [])]) : ls)
      | t1 /= t2 = []
    
    substVar ls' (l:ls) =
      substVar (l:ls') ls
    
    substVar ls' [] =
      [set ls']
  
  substFunEq ls = substFun [] (elements ls)
   where
    substFun ls' (l@(Neg (Pred _ Equal [t, Var v])) : ls)
      | v `notMember` free t =
      substTerm v t [] (ls' ++ ls) (substFun (l:ls') ls)
    
    substFun ls' (l:ls) =
      substFun (l:ls') ls
    
    substFun ls' [] =
      [set ls']

    substTerm v t ls' (Pos (Pred _ Equal [t1,t2]):ls) k
      | leftSubst  = substTerm v t (Pos (Equal `prd` [t,t2]):ls') ls k
      | rightSubst = substTerm v t (Pos (Equal `prd` [t,t1]):ls') ls k
     where
      leftSubst  = t1 == Var v && isSmall t2
      rightSubst = t2 == Var v && isSmall t1
      
      isSmall (Var _)            = True
      isSmall (App _ (Elt _) []) = True
      isSmall _                  = False

    substTerm v t ls' (l:ls) k
      | v `member` free l = k
      | otherwise         = substTerm v t (l:ls') ls k
    
    substTerm v t ls' [] k =
      simp (set ls')

-------------------------------------------------------------------------
-- split

splitAll :: [Clause] -> [Clause]
splitAll cs = splitting 1 cs (\_ -> [])
 where
  splitting i []     k = k i
  splitting i (c:cs) k = hyper i connections ls (\i -> splitting i cs k)
   where
    ls  = elements c
    n   = size (free ls)
    lvs = map free ls

    connections =
      [ (v,ws)
      | (v,ws) <- 
          tabulate $
            accumulate union
              [ (v,set ws)
              | vs     <- lvs
              , (v,ws) <- select (elements vs)
              ]
      , size ws < n-1
      ]

    hyper i []   ls k = break i ls k
    hyper i cons ls k = break (i+1) (Neg p : left)
                        (\i -> hyper i cons' (Pos p : right) k)
     where
      (v,_) = minimumBy siz cons
      rest  = [ x | x@(w,_) <- cons, v /= w ]
      (v1,s1) `siz` (v2,s2) =
        (size s1,tpsize v2,inter s2) `compare` (size s2,tpsize v1,inter s1)

      tpsize v = case tdomain (typ (Var v)) of
                   Nothing -> maxBound
                   Just d  -> d

      inter s =
        sum [ size (s `intersect` vs) | (v,vs) <- cons, v `member` s  ]
       where
        xs = elements s

      cons' =
        [ (w,ws')
        | (w,ws) <- rest
        , w `member` freeRight
        , let ws' = (ws `union` extra) `intersect` freeRight
              extra | w `member` vs = vs
                    | otherwise     = empty
        , size ws' < size freeRight-1
        ]

      vs           = free left `intersect` freeRight
      p            = (splitName % i ::- PType (map typ xs)) `prd` xs
      xs           = map Var (elements vs)
      (left,right) = partition ((v `member`) . free) ls
      freeRight    = free right

    break i ls k =
      case [ ls'
           
           -- only try when number of variables is at least 3
           | size (free ls) >= 3
           
           -- for all "non-recursive" definitions in ls
           , def@(Neg (Pred _ Equal [t@(App _ _ _), Var y])) <- ls
           , y `notMember` free t
           
           -- gather literals which contain y
           , let (lsWithY, lsWithoutY) =
                   partition ((y `member`) . free) (ls \\ [def])
                 
           -- there should be at least 2 such literals ...
           , lWithY:(lsWithY'@(_:_)) <- [lsWithY]
           
           -- now, partition the literals containing y
           -- into two separate groups, not containing
           -- any overlapping variables except the variables in def
           , let connToY = y `delete` free lsWithY
                 ok      = free def
           
                 (lsLeft, lsRight) =
                   part (free lWithY) [lWithY] [] lsWithY'
                  where
                   part vs left right [] = (left, right)
                   part vs left right (l:ls)
                     | size ((ws `intersect` vs) `difference` ok) == 0 =
                         part vs left (l:right) ls
                     
                     | otherwise =
                         part (vs `union` ws) (l:left) right ls
                    where
                     ws = free l
           
           -- they should not all end up on one side
           , not (null lsRight)
           
           -- construct the new clause with the extra literal
           , let y'  = varNameCopy % i ::- typing y
                 ls' = [ def
                       , Neg (Equal `prd` [t, Var y'])
                       ]
                    ++ subst (pair y (Var y')) lsLeft
                    ++ lsRight
                    ++ lsWithoutY
           ] of
        -- _       -> set ls : k i
        []      -> set ls : k i
        (ls':_) -> splitting (i+1) [set ls'] k
    
select :: [a] -> [(a,[a])]
select []     = []
select (x:xs) = (x,xs) : [ (y,x:ys) | (y,ys) <- select xs ]

-------------------------------------------------------------------------
-- the end.
