#include "SatnikInstantiateClause.h"

#include <ostream.h>

/**************************************************************************************************/
/*** Local helpers ********************************************************************************/
/**************************************************************************************************/

static inline int max(int x, int y) { return x < y ? y : x; }

static inline bool operator<(const vector<int>& d1, const vector<int>& d2) {
  assert(d1.size() == d2.size());
  for(size_t i = 0; i < d1.size(); i++) if(d2[i] <= d1[i]) return false;
  return true;
}

static inline bool operator==(const vector<int>& d1, const vector<int>& d2) {
  assert(d1.size() == d2.size());
  for(size_t i = 0; i < d1.size(); i++) if(d2[i] != d1[i]) return false;
  return true;
}

// least upper bound, of two equal-sized vectors
static inline vector<int> lub(const vector<int>& d1, const vector<int>& d2) {
  assert(d1.size() == d2.size());
  vector<int> result; 
  for(size_t i = 0; i < d1.size(); i++) 
    result.push_back(max(d1[i],d2[i]));
  return result;
}

// dot product
static inline int operator*(const vector<int>& d1, const vector<int>& d2) {
  assert(d1.size() == d2.size());
  int sum = 0; for(size_t i = 0; i < d1.size(); i++) sum += d1[i] * d2[i]; 
  return sum; }


// debug
static ostream& operator<<(ostream& o, const vector<int>& i) {
  o << "[";
  if(i.size() == 0) { o << "]"; return o; }
  for(size_t j = 0; j < i.size()-1; j++) o << i[j] << ", ";
  o << i[i.size()-1] << "]";
  return o;
}

static inline Var& get(vector<Var>& table, const vector<int>& sizes, const vector<int>& i) {
  return table[sizes * i]; }

/**************************************************************************************************/
/*** Loc implementaion ****************************************************************************/
/**************************************************************************************************/

Var Loc::get(Solver& s, const vector<Arg>& args, const vector<int>& bindings) { 
  assert(args.size() == (size_t)arity);

  //cout << "get [1] " << arity << args << bindings << endl;

  DomElem                argv[args.size()];
  Var                    result;

  for(int i = 0; i < arity; i++) 
    argv[i] = isVar(args[i]) ? bindings[getArg(args[i])] : getArg(args[i]);

  varMap::const_iterator i = vmap.find(argv);

  if(i == vmap.end()) {
    DomElem* argp = mem.alloc();
    for(int i = 0; i < arity; i++) argp[i] = argv[i];
    result     = s.add_vars();
    vmap[argp] = result;
  } else result = (*i).second;

  return result;
}

size_t nvars(const Literal& l, const vector<bool>& vset) {
  vector<Var> vs; 
  l.vars(vs); 

  size_t r = 0; 
  for(size_t i = 0; i < vs.size(); i++) if(!vset[vs[i]]) r++;
  return r;
}

void schedule(vector<Literal>& lits, vector<int>& variables, vector<int>& chunksize)
{
  assert(variables.size() == chunksize.size());
  vector<bool> varset; varset.resize(variables.size(),false);
  size_t i,j,k;

  // order literals
  for(i = 0; i < lits.size(); i++) {
    size_t minindex = i;
    size_t minvars  = nvars(lits[i],varset);
    for(j = i; j < lits.size(); j++) {
      size_t nv = nvars(lits[j],varset);
      if(nv < minvars) { minindex = j; minvars = nv; }
    }
    swap(lits[i],lits[minindex]);
    vector<int> vs; lits[i].vars(vs);
    for(j = 0; j < vs.size(); j++) varset[vs[j]] = true;
  }

  // calculate tables
  k = 0;
  for(i = 0; i < varset.size(); i++) varset[i] = false;
  for(i = 0; i < lits.size(); i++) {
    vector<int> vs; lits[i].vars(vs);
    for(j = 0; j < vs.size(); j++) 
      if(!varset[vs[j]]) { variables[k++] = vs[j]; varset[vs[j]] = true; }
    if(k > 0) chunksize[k-1]++;
  }

}


// print a vector of literals
inline ostream& operator<<(ostream& o, const vector<Literal>& literals) {
  for(size_t i = 0; i < literals.size(); i++) o << literals[i] << ' ';
  return o; }

int Loc::n = 0;

/**************************************************************************************************/
/*** The implementation of First-order Clauses ****************************************************/
/**************************************************************************************************/

// precondition: fresh >= max(sizes)-1
bool FOClause::instantiate(Solver& s, int fresh) {

  vector<int> bindings;  bindings .resize(sizes.size(), -1);
  vector<int> variables; variables.resize(sizes.size());
  vector<int> chunksize; chunksize.resize(sizes.size());

  int maxe = 0;  // the largest domain-size of a variable
  int maxi = -1; // the first largest variable
  int i,j;

  for(i = 0; (size_t)i < sizes.size(); i++) if(maxe < sizes[i]) { maxe = sizes[i]; maxi = i; } 
  if(maxe > 0) maxe--;

  int count = s.get_nof_constraints();

  if(fresh > maxe) { 
    //cout << "Done! " << s.get_nof_constraints() - count << endl; 
    return true; 
  }
  assert(fresh == 0 || fresh == maxe);

  //cout << "(" << s.get_nof_constraints() << ") clause:    " << lits << endl;
  //  cout << "clause:    " << lits << endl;
//  cout << "  lits:    " << lits.size() << endl;
//  cout << "  vars:    " << sizes.size() << endl;
//  cout << "  fresh:   " << fresh  << endl;
//  cout << "  max:     " << maxe  << endl;
//  cout << "  sizes:   " << sizes << endl;

  schedule(lits,variables,chunksize);

  //cout << "reordered: " << lits << endl;
//  cout << "variables: " << variables << endl;
//  cout << "chunksize: " << chunksize << endl;

  vector<Lit> cls;
  //  LitSet      lset;
  
  // handle ground prefix
  size_t lind; { vector<bool> varset; varset.resize(variables.size(),false);
  for(lind = 0; lind < lits.size(); lind++) {
    Literal& l = lits[lind];
    //cout << "lit " << l << endl;
    if(nvars(l,varset) == 0) {
      Lit g = lit(l.get(s,bindings),l.sign);
      if(s.values(g) || lset.member(neg(g))) {
	//cout << "Done! " << s.get_nof_constraints() - count << endl; return true; }
	//cout << "Done! " << s.get_nof_constraints()  << endl; 
	// remove literals in set !
	for(i = 0; (size_t)i < cls.size(); i++) lset.del(cls[i]);
	return true; }
      else if(!s.values(neg(g)) && !lset.member(g)) {
	cls.push_back(g);
	lset.add(g);
      }
    }
    else break;
  } }
  
  // if whole clause is ground, add it and return
  if(variables.size() == 0) {
    //cout << bindings << " : " << cls << endl;
    if(cls.size() == 0) return false;
    bool ret = s.unsafe_add_clause(cls);
    //cout << "adding: " << cls << endl;
    assert(ret);
    for(i = 0; (size_t)i < cls.size(); i++) lset.del(cls[i]);
    //cout << "Done! " << s.get_nof_constraints() - count << endl; return ret;
    //cout << "Done! " << s.get_nof_constraints() << endl; 
    return ret;
  }
  
  // find last fresh variable
  int fvar = 0;
  if(fresh > 0) 
    for(i = variables.size()-1; i >= 0; i--) {
      //cout << "var" << variables[i] << ':' << sizes[variables[i]] << endl;
      if(sizes[variables[i]]-1 == fresh) { fvar = i; break; }
    }
  //cout << "fresh: " << fresh << endl;
  //cout << "fvar:  " << fvar << endl;

  vector<int> currsize; currsize.resize(chunksize.size(),0);

  // enumerate variable substitutions
  int top = 0; // number of variables with the value "fresh"
  for(i = 0; i >= 0;) {

    // full substitution constructed?
    if((size_t)i == sizes.size()) {
      // check if clause is empty
      if(cls.size() == 0) return false;
      
      //cout << bindings << " : " << cls << endl;
//      {
//	LitSet      ls;
//	for(size_t k = 0; k < cls.size(); k++) {
//	  Lit l = cls[k];
//	  assert(!s.values(l));
//	  assert(!s.values(neg(l)));
//	  assert(!ls.member(l));
//	  assert(!ls.member(neg(l)));
//	  ls.add(cls[k]);
//	}
//	bool fr = false;
//	for(size_t l = 0; l < bindings.size(); l++)
//	  fr = fr | bindings[l] == maxe;
//	if(fresh > 0) assert(fr);
//      }
      //vector<Lit> cs = cls;
      //sort(cs.begin(), cs.end(), Solver::IndexLitComparator());
#ifndef NDEBUG
      int count2 = s.get_nof_constraints();
#endif
      bool ret = s.unsafe_add_clause(cls);
      //bool ret = s.unsafe_add_clause(cs);
      //cout << "adding: " << cls << endl;
      assert(ret);
      assert(cls.size() == 1 || s.get_nof_constraints() == count2 + 1);
      i--;
      while(currsize[i] > 0) { Lit g = cls.back(); cls.pop_back(); lset.del(g); currsize[i]--; }
      lind -= chunksize[i];
      continue;
    }

    // not done
    int var = variables[i];
    assert(cls.size() <= lind);
    if(bindings[var] == sizes[var]-1) {
      // backtrack
      if(bindings[var] == fresh) top--;
      bindings[var] = -1; i--;

      if(i >= 0) {
//	cout << "unset x" << var << endl;
//	cout << "backtrack: " << currsize[i] << endl;
//	cout << "    chunk: " << chunksize[i] << endl;
//	cout << "     lind: " << lind << endl;
	while(currsize[i] > 0) { Lit g = cls.back(); cls.pop_back(); lset.del(g); currsize[i]--; }
	lind -= chunksize[i];
      }
    } else {
      // advance
      if(fresh > 0 && top == 0 && i == fvar) 
	bindings[var] = fresh;
      else
	bindings[var]++;

      if(bindings[var] == fresh) top++;
      //cout << "set x" << var << " = " << bindings[var] << endl;
      //cout << "top: " << top << endl;
      //cout << "advance: " << chunksize[i] << endl;

      currsize[i] = 0;
      int oldind = lind;
      for(j = 0; j < chunksize[i]; j++) {
	//cout << "lind: " << lind << endl;
	Lit g = lit(lits[lind].get(s,bindings),lits[lind].sign);
	//cout << "lit: " << g << endl;
	lind++;
	if(s.values(g) || lset.member(neg(g))) {
	  while(currsize[i] > 0) { Lit g = cls.back(); cls.pop_back(); lset.del(g); currsize[i]--; }
	  lind = oldind;
	  goto apa;
	}
	else if(!s.values(neg(g)) && !lset.member(g)) {
	  cls.push_back(g);
	  lset.add(g);
	  currsize[i]++;
	}
      }
      //cout << "curr clause: " << cls << endl;
	
      i++;
    }
  apa:;
    //cout << "Var: " << var << ", Binding: " << bindings[var] << ", Top: " << top << endl;
  }
  //cout << "Done! " << s.get_nof_constraints() - count << endl;
  //cout << "Done! " << s.get_nof_constraints() << endl;
  for(i = 0; (size_t)i < cls.size(); i++) lset.del(cls[i]);
  return true;
}


#if 0

/**************************************************************************************************/
/*** Main, for testing purposes *******************************************************************/
/**************************************************************************************************/

int main() {
  Solver      s;
  Loc         pr(2);
  FOClause    c;

  // transitivity of p

  #if 1
  c.addLit(&pr, false);
  c.lastLit().addArg(varArg(0));
  c.lastLit().addArg(varArg(1));
  c.addLit(&pr, false);
  c.lastLit().addArg(varArg(1));
  c.lastLit().addArg(varArg(2));
  c.addLit(&pr, true);
  c.lastLit().addArg(varArg(0));
  c.lastLit().addArg(varArg(2));

  cout << "Instantiating all for d = 2:" << endl;
  c.clearSize();
  c.addSize(2);
  c.addSize(2);
  c.addSize(2);
  c.instantiate(s,0);
  s.export_dimacs(cout);


  cout << "Instantiating new for d = 3:" << endl;
  c.clearSize();
  c.addSize(3);
  c.addSize(3);
  c.addSize(3);
  c.instantiate(s,2);
  s.export_dimacs(cout);

  vector<Arg> args; args.push_back(varArg(0)); args.push_back(varArg(1)); 
  vector<int> bind; bind.resize(2);

  for(int i = 0; i < 3; i++)
    for(int j = 0; j < 3; j++) {
      bind[0] = i; bind[1] = j;
      cout << "p(" << i << "," << j << ") = " << pr.get(s,args,bind) << endl;
    }
  #endif

  //exit(-1);
  

  c.clearLits();
  
  c.addLit(&pr, true);
  c.lastLit().addArg(varArg(0));
  c.lastLit().addArg(varArg(1));
  c.addLit(&pr, true);
  c.lastLit().addArg(varArg(2));
  c.lastLit().addArg(varArg(3));
  c.addLit(&pr, true);
  c.lastLit().addArg(varArg(4));
  c.lastLit().addArg(varArg(5));

  Loc         qr(6);
  c.addLit(&qr, true);
  c.lastLit().addArg(varArg(0));
  c.lastLit().addArg(varArg(1));
  c.lastLit().addArg(varArg(2));
  c.lastLit().addArg(varArg(3));
  c.lastLit().addArg(varArg(4));
  c.lastLit().addArg(varArg(5));

  c.addLit(&qr, false);
  c.lastLit().addArg(conArg(0));
  c.lastLit().addArg(conArg(0));
  c.lastLit().addArg(conArg(0));
  c.lastLit().addArg(conArg(0));
  c.lastLit().addArg(conArg(0));
  c.lastLit().addArg(conArg(0));

  for(int siz = 1; siz < 11; siz++) {
    cout << "Instantiating a lot! (" << siz << " ^ 6):" << endl;
    c.clearSize();
    c.addSize(siz);
    c.addSize(siz);
    c.addSize(siz);
    c.addSize(siz);
    c.addSize(siz);
    c.addSize(siz);

    c.instantiate(s,siz-1);
    cout << "Number of clauses: " << s.get_nof_constraints() << endl;
    cout << "Number of vars:    " << s.get_nof_vars() << endl;

    //exit(-1);
  }

  return 0;
}

#endif
