/* ---------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
--------------------------------------------------------------- */

#include "Satnik.h"
#include "SatnikInstantiateClause.h"

#include <stdio.h>
#include <stdlib.h>

#if (__GNUC__ > 2)
#include <ext/hash_map>
using namespace __gnu_cxx;
#else
#include <hash_map>
#endif

using namespace std;
using namespace Satnik;

typedef unsigned int uint;
typedef void         ExtSolver;

struct ModelInspect {
  vector<int>& model;
  ModelInspect(vector<int>& m) : model(m) {};
  void operator()(Solver& s) { s.copy_model(model); }
};

struct SolverWrapper {
  Solver            s;
  vector<Lit>       prefix;
  vector<int>       model;
  FOClause          clause;
  Loc*              loc;
  uint              sig;
  vector<int>       args;
  vector<int>       dummy;
};

extern "C" {
  SolverWrapper* solver_new(void) { 
    SolverWrapper* s = (SolverWrapper*)malloc(sizeof(SolverWrapper));
    new (s) SolverWrapper();
    
    s->s.set_variable_parameters(70,.998,0);
    s->s.set_polarity_parameters(true,false,true);
    s->s.set_learning_parameters(1000, 0 * s->s.get_nof_constraints(),.9985,.0001);
    s->s.set_restart_parameters (100,1.3);
    s->s.set_verbose(false);
    
    return s;
  }

  void     solver_free(SolverWrapper* w)                    { w->~SolverWrapper(); free(w); }

  Loc*     loc_new    (uint a) { Loc* l = (Loc*)malloc(sizeof(Loc)); new (l) Loc(a); return l; }
  void     loc_free   (Loc* l) { free(l); }
  uint     loc_arity  (Loc* l) { return l->getArity(); }

  // build a clause
  void solver_clause_begin       (SolverWrapper* w)           { w->clause.clear();}
  void solver_clause_add_lit     (SolverWrapper* w, Loc* l, uint s) { w->clause.addLit(l,(bool)s); }
  void solver_clause_add_lit_var (SolverWrapper* w, int v)          {
    w->clause.lastLit().addArg(varArg(v-1));}
  void solver_clause_add_lit_con (SolverWrapper* w, int c)          { 
    w->clause.lastLit().addArg(conArg(c-1));}
  void solver_clause_add_size    (SolverWrapper* w, int s)          { w->clause.addSize(s); }
  uint solver_clause_commit      (SolverWrapper* w, int fresh)      { 
    return w->clause.instantiate(w->s, fresh) ? 1 : 0; }

  // setting a literal to be read or assumed during solving
  void solver_lit_begin   (SolverWrapper* w, Loc* l, uint sign)  { 
    w->args.clear(); w->loc = l; w->sig = sign; }
  void solver_lit_add_con (SolverWrapper* w, int c)              { w->args.push_back(conArg(c-1)); }

  // reading the value of a predicate (location)
  uint solver_read        (SolverWrapper* w) { 
    // creates variables if they don't exists!! fix this !! 
    Var v = w->loc->get(w->s, w->args, w->dummy); 
    return w->model[index(lit(v,w->sig))] ? 1 : 0; }

  uint solver_solve(SolverWrapper* w) {
    ModelInspect m(w->model);
    vector<Lit>  prefix;
    if(w->loc != NULL) {
      Var v = w->loc->get(w->s, w->args, w->dummy); 
      prefix.push_back(lit(v,w->sig));
    }
    return w->s.solve_(m,prefix) ? 1 : 0; 
  }

  void solver_dimacs(SolverWrapper* w, char* name) { 
    ofstream out;
    out.open(name, ofstream::out);
    if(!out.is_open()) { cout << "Could not open file " << name << "!\n"; exit(-1); }
    vector<Lit>  prefix;
    if(w->loc != NULL) {
      Var v = w->loc->get(w->s, w->args, w->dummy); 
      prefix.push_back(lit(v,w->sig));
    }
    w->s.export_dimacs(out,w->prefix);
    out.close();
  }

  uint solver_simplify(SolverWrapper* w) {
    if ( w->s.propagate() )
    {
      w->s.simplify_clauses();
      return 1;
    }
    return 0;
  }


  // statistics
  uint     solver_get_nof_clauses   (SolverWrapper* w)       { return w->s.get_nof_constraints(); }
  uint     solver_get_nof_vars      (SolverWrapper* w)       { return w->s.get_nof_vars(); }
  uint     solver_get_nof_conflicts (SolverWrapper* w)       { return w->s.get_nof_conflicts(); }

  void     solver_set_verbose       (SolverWrapper* w, uint b) { w->s.set_verbose((bool)b); }

  char    *identify (void) { return "Satnik"; }
}
