/* --------------------------------------------------------------------------------------------------------*/
/* -                                                                                                       */
/* - TermpairRepresentation.c                                                                              */
/* - ************************                                                                              */
/* -                                                                                                       */
/* - Repraesentation von kritischen Paaren im Speicher.                                                    */
/* -                                                                                                       */
/* - 14.03.96 Arnim. Angelegt.                                                                             */
/* -                                                                                                       */
/* -                                                                                                       */
/* --------------------------------------------------------------------------------------------------------*/

#include "Ausgaben.h"
#include "compiler-extra.h"
#include "general.h"
#include "InfoStrings.h"
#include "RUndEVerwaltung.h"
#include "SpeicherVerwaltung.h"
#include "Stringterms.h"
#include "SymbolOperationen.h"
#include "TermpairRepresentation.h"
#include "TermOperationen.h"
#include "Unifikation1.h"
#include "pcl.h"

typedef enum {tpr_packed, tpr_implicit, tpr_regular} TPR;

static TPR tpr = tpr_packed;

/* --------------------------------------------------------------------------------------------------------*/
/* Packen:

Dokumentation:

Funktionalitaet etc.:

1. Terme sollen in Byte-Ketten abgelegt bzw. aus solchen restauriert werden koennen.

2. Terme als Byteketten sollen als Grundlage der KP-Menge herhalten koennen. Dazu ist eine
   entsprechende Speicherverwaltung zu etablieren. Diese soll:
   - fuer alle Byteketten-Laengen in Vierer-Schritten Freispeicherlisten fuehren,
     (-> Array von Speichermanagern)
   - fuer Byteketten oberhalb einer bestimmten Laenge einzeln den Platz allokieren

3. Da Terme in der KP-Menge nur paarweise auftreten, werden Termpaare hier gemeinsam eingepackt.
   Dadurch entsteht nur ein- statt zweimal Platzverschwendung durch Alignment.

4. Jedes gepackte Termpaar steht in einem Byte-String im Speicher. Dieser String enthaelt als erstes Byte
   die Nummer des entsprechenden Speichermanagers. Ist der Platz einzeln allokiert worden, so findet sich dort
   die Zahl 0xff == 255. Dadurch ergibt sich folgende Restriktion:
   es duerfen max. 254 Speichermanager angelegt werden!

*/

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------------- Speichermanagement fuer eingepackte Terme/ Globale Variablen ----------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define Alignment (sizeof(unsigned long int))
#define AnzahlManager 50        /* AnzahlManager muss kleiner als PseudoManagerNummer sein ! */
#define PseudoManagerNummer 255 /* Manager-Nummer, die bei einzeln allokierten PTermpaaren eingetragen wird. */
#define EinzelnAnlegenSchwelle (Alignment * AnzahlManager)

static SV_SpeicherManagerT SpeicherManagerArray[AnzahlManager];

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------------- Anlegen und Freigeben von gepackten Termpaaren ------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static PTermpaarT TermpaarAllokieren(unsigned long int groesse)
/* Liefert den Speicherbereich fuer ein Termpaar der Laenge groesse. 
 * Beruecksichtig den Platz fuer PCL und Manager-Nummer */
{
   PTermpaarT pp;

   groesse = groesse + 1 + (PCL_EsWirdProtokolliert()? sizeof(PCL_ObjektNummerT) : 0);
   
   if (groesse  >= EinzelnAnlegenSchwelle){
      pp = (PTermpaarT) our_alloc(groesse);
      *pp = (unsigned char) PseudoManagerNummer;
   }
   else{
      unsigned int index = groesse / Alignment; 
      SV_Alloc(pp,SpeicherManagerArray[index],PTermpaarT);
      *pp = (unsigned char) index;
   }
   return pp;
}

static void TermpaarFreigeben(PTermpaarT pp)
/* Termpaar je nach ManagerID freigeben.*/
{
  unsigned char ManagerNummer = (unsigned char) *pp;
  if (ManagerNummer == PseudoManagerNummer)
    our_dealloc(pp);
  else
    SV_Dealloc(SpeicherManagerArray[ManagerNummer], pp);
}

static PTermpaarT pclIDEinpacken(PTermpaarT tp, PCL_ObjektNummerT pclID)
{
  if (PCL_EsWirdProtokolliert()){
    PCL_ObjektNummerT pclid = pclID;
    memcpy(tp,&pclid,sizeof(PCL_ObjektNummerT));
    tp = tp + sizeof(PCL_ObjektNummerT);
  }
  return tp;
}

static PTermpaarT pclIDAuspacken(PTermpaarT tp, PCL_ObjektNummerT *pclPtr)
{
  if (PCL_EsWirdProtokolliert()){
    memcpy(pclPtr,tp,sizeof(PCL_ObjektNummerT));
    tp = tp + sizeof(PCL_ObjektNummerT);
  }
  return tp;
}

static PTermpaarT TermpaarEinpacken(TermT lhs, TermT rhs, PCL_ObjektNummerT pclID)
/* Termpaar lhs=rhs einpacken und als gepacktes Termpaar zurueckgeben.
   Dieses steht dann bereits als eigenstaendiges Objekt zur Verfuegung. */
{
   PTermpaarT Termpaar;

   Termpaar = TermpaarAllokieren(STT_TermpaarPlatzbedarf(lhs,rhs));
   STT_TermpaarEinpackenFragil(lhs, rhs, pclIDEinpacken(Termpaar+1,pclID));
   return (TermpairT) Termpaar;
}

static void TermpaarAuspacken(PTermpaarT tp, TermT *lhsPtr, TermT *rhsPtr, PCL_ObjektNummerT *pclPtr)
{
  STT_TermpaarAuspacken(pclIDAuspacken(tp+1,pclPtr),lhsPtr,rhsPtr);
}

static void DruckeTermpaar(PTermpaarT tp, const char *trenner)
{
  STT_DruckePTermpaar(tp + 1 + (PCL_EsWirdProtokolliert() ? sizeof(PCL_ObjektNummerT) : 0), trenner);
}

/* -------------------------------------------------------------------------------------------------------------
 - Implizite Darstellung von Paaren von flatterms durch die Stelle ihrer Ueberlappung.                   
 Doku:

 Paare von Flachtermen werden durch die Stelle der Ueberlappung repraesentiert, durch die sie entstanden sind.
 Zum Restaurieren sind folglich die Eltern sowie die Information, in welchen der Elternteile hineinueberlappt
 wurde, noetig. Ferner sind Gleichungen aus der Spezifikation sowie Interreduktions-Opfer zu beruecksichtigen.

 Zu diesem Zwecke werden die Termpaare wie folgt repraesentiert:
 
 - IR-Opfer, Gleichungen aus der Spezifikation: als Flachterm-Paare (1 Pointer)
 - Unifikations-Produkte: als Nummer der Stelle, an der in die aeussere Regel/Gleichung hineinueberlappt wurde.
 
 Bemerkungen:

 - IR-Opfer und Spez-Gleichungen sind leicht daran erkennbar, dass sie den Halbgott als ELternteile haben.
 - Die Information, in welchen der beiden Elternteile letztendlich hineinueberlappt wurde, wird als Vorzeichen
   der Stelle gespeichert:

   tp<0: In OtherParent hineinueberlappt  (CPWithFather)
   tp>0: In ActualParent hineinueberlappt (CPWithMother)


 Voraussetzungen:
 
 - Terme werden nicht laenger als 2^30-1 Symbole.

   --------------------------------------------------------------------------------------------------------*/

/* --------------------------------------------------------------------------------------------------------*/
/* ------- New Flats generated ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void TPR_NewFlatsGenerated(void)
{
  PCL_StartChain(tpr != tpr_implicit); /* == feychain bei tpr_implicit*/
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- FLATTERMS to TERMPAIR --------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


TermpairT TPR_Flat2TP (TermT lhs, TermT rhs, RegelOderGleichungsT ActualParent, 
                       RegelOderGleichungsT OtherParent, UTermT actualPosition, 
                       PCL_ObjektNummerT pclID, BOOLEAN isCPWithFather, TermT lhsReplica)
{
   TermpairT tp;
   if ((tpr == tpr_packed) || (OtherParent == NULL)){
      tp = (TermpairT) TermpaarEinpacken(lhs, rhs, pclID);
      PCL_SaveChain();
   }
   else if (tpr == tpr_regular){
     TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
      PCL_SaveChain();
     return (TermpairT) lhs;
   }
   else {                 /* Ueberlappungsstelle berechnen */
     PCL_DeleteChain();
     if (isCPWithFather){ /* actualPosition ist Teil von OtherParent ==> negatives Vorzeichen */
       tp = (TermpairT) (-TO_StelleInTerm(actualPosition, TP_LinkeSeite(OtherParent)));
     }
     else{                /* actualPosition ist Teil von ActualParent ==> positives Vorzeichen */
       tp = (TermpairT) TO_StelleInTerm(actualPosition, lhsReplica);
     }
   }
   TO_TermeLoeschen(lhs, rhs);
   return tp;
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- TERMPAIR to PARALLEL ---------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static void TP2Par (TermpairT tp, TermT *lhsPtr, TermT *rhsPtr, RegelOderGleichungsT ActualParent, 
                         RegelOderGleichungsT OtherParent, PCL_ObjektNummerT *pclIDPtr, BOOLEAN realChain)
{
   if ((tpr == tpr_packed) || (OtherParent == NULL)){
     TermpaarAuspacken((PTermpaarT) tp, lhsPtr, rhsPtr, pclIDPtr);
     PCL_StartChain(TRUE);
   }
   else if (tpr == tpr_regular){
     *lhsPtr = (TermT) tp;
     *rhsPtr = TO_Schwanz(TO_TermEnde((TermT) tp));
     TO_NachfBesetzen(TO_TermEnde((TermT) tp), NULL);
     PCL_StartChain(TRUE);
   }
   else{
      long int Stelle = ptr2ul(tp);
      PCL_StartChain(realChain); 
      if (Stelle < 0){
         /* Stelle in OtherParent */
         U1_KPRekonstruieren(OtherParent, -Stelle, ActualParent, lhsPtr, rhsPtr);
         PCL_ReUnifikation(pclIDPtr, *lhsPtr, *rhsPtr, OtherParent, -Stelle, ActualParent);
      }
      else{
         /* Stelle in ActualParent */
         U1_KPRekonstruieren(ActualParent, Stelle, OtherParent, lhsPtr, rhsPtr);
         PCL_ReUnifikation(pclIDPtr, *lhsPtr, *rhsPtr, ActualParent, Stelle, OtherParent);
      }
   }
}

void TPR_TP2ParIntermed (TermpairT tp, TermT *lhsPtr, TermT *rhsPtr, RegelOderGleichungsT ActualParent, 
                         RegelOderGleichungsT OtherParent, PCL_ObjektNummerT *pclIDPtr)
{
  TP2Par (tp, lhsPtr, rhsPtr, ActualParent, OtherParent, pclIDPtr, FALSE /* == feychain */);
}


void TPR_TP2ParFinal (TermpairT tp, TermT *lhsPtr, TermT *rhsPtr, RegelOderGleichungsT ActualParent, 
                      RegelOderGleichungsT OtherParent, PCL_ObjektNummerT *pclIDPtr)
{
  TP2Par (tp, lhsPtr, rhsPtr, ActualParent, OtherParent, pclIDPtr, TRUE /* != feychain */);
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to FLATTERMS --------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void TPR_Par2Flat (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent)
/* Destroys the termpair-representation irrecoverable */
{
  if ((tpr == tpr_packed) || (OtherParent == NULL)) {
    TermpaarFreigeben((PTermpaarT) tp);
  }
  PCL_SaveChain(); 
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to TERMPAIR ---------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

TermpairT TPR_Par2TP (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT ActualParent, 
                      RegelOderGleichungsT OtherParent, BOOLEAN changed,
                      PCL_ObjektNummerT pclID)
{
   if ((tpr == tpr_packed) || (OtherParent == NULL)){
     if (changed) {
       TermpaarFreigeben((PTermpaarT) tp);
       tp = (TermpairT) TermpaarEinpacken(lhs, rhs, pclID);
     }
     PCL_SaveChain();
   }
   else if (tpr == tpr_regular){
     PCL_SaveChain();
     TO_NachfBesetzen(TO_TermEnde(lhs), rhs);
     return (TermpairT) lhs;
   }
   else {
     /* tp bleibt das alte */
     PCL_DeleteChain();
   }
   TO_TermeLoeschen(lhs, rhs);
   return tp;
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- DESTROY ----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


void TPR_DestroyTP  (TermpairT tp, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent)
/* Termpaar zerstoeren */
{
   if ((tpr == tpr_packed) || (OtherParent == NULL)){
     TermpaarFreigeben((PTermpaarT) tp);
   }
   else if (tpr == tpr_regular){
     TO_TermzellenlisteLoeschen((TermT) tp, TO_TermEnde(TO_Schwanz(TO_TermEnde((TermT) tp))));
   }
   else{
      /* nix tun */
   }
}

void TPR_DestroyPar (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent)
/* Termpaar und Flachterme zerstoeren */
{
   if ((tpr == tpr_packed) || (OtherParent == NULL)){
     TermpaarFreigeben((PTermpaarT) tp);
   }
   PCL_DeleteChain();
   TO_TermeLoeschen(lhs, rhs);
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Output -----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void TPR_PrintTermpair (TermpairT tp, RegelOderGleichungsT ActualParent, RegelOderGleichungsT OtherParent, const char *trenner)
{
  if ((tpr == tpr_packed) || (OtherParent == NULL)){
    DruckeTermpaar((PTermpaarT) tp, trenner);
  }
  else if (tpr == tpr_regular){
    IO_DruckeFlex("%t %s %t", tp, TO_NaechsterTeilterm((TermT)tp), trenner);
  }
  else {                 /* Ueberlappungsstelle berechnen */
    TermT lhs, rhs;
    PCL_ObjektNummerT bla;

    /* Hier stellt sich die Frage, ob Terme berechnet werden sollten oder nicht!!! */    
    TPR_TP2ParIntermed(tp, &lhs, &rhs, ActualParent, OtherParent, &bla);
    IO_DruckeFlex("%t %s %t", lhs, trenner, rhs);
    TPR_Par2TP (tp, lhs, rhs, ActualParent, OtherParent, FALSE, bla);
  }
}

/* --------------------------------------------------------------------------------------------------------*/
/* -------- Initialisierung -------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

/* XXX Vermutlich ueberfluessig */
static BOOLEAN TPR_SchonInitialisiert = FALSE;

BOOLEAN TPR_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
   if (!TPR_SchonInitialisiert){    /* Etablieren der Speichermanager */
     unsigned int i;
     for (i=0; i < AnzahlManager; i++) {
       SpeicherManagerArray[i].block_size = 0; /* XXX ??? */                    /* Initialisierung erzwingen */
       SV_BeliebigenManagerInitialisieren(&(SpeicherManagerArray[i]), (i+1)*Alignment, 
                                          TPR_BlocksizeInBytes/((i+1)*Alignment), FALSE, "Gepackte Termpaare");
     }
     TPR_SchonInitialisiert = TRUE;
   }

   switch (IS_FindOption(DEF, USR, 3, "packed", "implicit", "regular")){
   case 1:
     tpr = tpr_implicit;
     break;
   case 2:
     tpr = tpr_regular;
     if (PCL_EsWirdProtokolliert()){
       our_error("TPR regular unvertraeglich mit PCL!");
     }
     break;
   case 0:
   default:
     tpr = tpr_packed;
   }
#if !COMP_POWER
   fprintf(IO_Strom, "\n* TERMPAIR-REPRESENTATION\n\n  %s", (tpr == tpr_packed ? "Packed. " : tpr == tpr_implicit ? "Implicit. " : "Regular. "));
#endif
   return TRUE;
}

void TPR_ParamInfo(ParamInfoArtT what MAYBE_UNUSEDPARAM, InfoStringT DEF MAYBE_UNUSEDPARAM, InfoStringT USR MAYBE_UNUSEDPARAM)
{
   fprintf(IO_Strom, "\n* Termpair-Representation\n\n  ");
   fprintf(IO_Strom, "TPR_ParamInfo not yet implemented\n\n");
}
