/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Puffer.c                                                                                                   */
/* - ********                                                                                                   */
/* -                                                                                                            */
/* - Realisierung von Puffern fuer Objekte, die kurzfristig zwischengespeichert werden muessen, da ihre weitere */
/* - Behandlung noch nicht vorgenommen werden kann. Der Aufwand an Platz und Zeit soll minimal sein.            */
/* -                                                                                                            */
/* - REPuffer : Hier koennen Regeln und Gleichungen zwischengelagert werden, die der Interreduktion zum Opfer   */
/* -            gefallen sind.                                                                                  */
/* -            Zur Verwendung kommt ein Array, da andernfalls zu viele Pointer umgebogen werden muesten:       */
/* -            - Termendezeiger der linken Seite auf rechte Seite umbiegen                                     */
/* -            - Nachfolgerzeiger der letzten Zelle der linken Seite auf Objekt selbst umbiegen                */
/* -            - Nachfolgerzeiger der letzten Zelle der rechten Seite auf Rest des Puffers umbiegen            */
/* -            - ... plus drei Ruecksetzungen beim Auslesen aus dem Puffer                                     */
/* -            Platzbedarf beim Array: lediglich Pointergroesse, da nur ganzes Objekt referenziert werden muss.*/
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* - TPPuffer : Hier werden waehrend der Modi confluence, convergence, termination als nicht richtbar erkannten */
/* -            Ausgangsgleichungen bzw. nicht zusammenfuehrbaren KPs abgelegt. Hierzu wird Platz vorgesehen    */
/* -            fuer Eltern-Nummern sowie die Nummern der Ausgangsgleichungen, nicht aber fuer PCL-IDs.         */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* - SGPuffer : Subsumptionspuffer fuer Subgleichungen. Wennn eine neue Gleichung aufgenommen werden soll, wer- */
/* -            den zunaechst ihre Subgleichungen darauf ueberprueft, ob sie durch die Gleichungsmenge bereits  */
/* -            subsummiert werden. Spaeter werden diese Subgleichungen aber wieder beneotigt, wennn ueber-     */
/* -            prueft werden soll, ob eine neue Gleichung aeltere subsummiert. Waehrend der erstgenannten      */
/* -            Ueberpruefung werden deshalb die entdeckten Subgleichungen im SG-Puffer gespeichert, um sie     */
/* -            anschliessend ggf. in eine Struktur zu uebernehmen, die die Durchfuehrung der letztgenannten    */
/* -            Ueberpruefung erleichtert.                                                                      */
/* -                                                                                                            */
/* -                                                                                                            */
/* - 16.06. Arnim. Angelegt.                                                                                    */
/* - 14.07. Arnim. SG-Puffer dazugefuegt.                                                                       */
/* - 02.05. Arnim. KP-Puffer geaendert und zu TP-Puffer gemacht.                                                */
/* -------------------------------------------------------------------------------------------------------------*/

#include "general.h"
#include "Parameter.h"
#include "SpeicherParameter.h"
#include "TermOperationen.h"
#include "RUndEVerwaltung.h"
#include "pcl.h"
#include "Puffer.h"

/*----------------------- Typen und Konstanten -----------------------------------------------------------------*/

typedef struct TPPuffTS{
   TermT     linkeSeite;
   TermT     rechteSeite;
   unsigned long Nummer;
} TPPuffZellenT;

static RegelOderGleichungsT *REPuffer;	                /* Array von Gleichungen bzw. Regeln (als void *) */
static unsigned long hoechsteREPufferPos, REPufferPos, REMerker;

static TPPuffZellenT *TPPuffer;	        /* Array von (oben def.) Puffer-Zellen */
static unsigned long hoechsteTPPufferPos, TPPufferPos, TPLesePos;

/* ----------------------- Initialisierung ---------------------------------------------------------------------*/
static BOOLEAN PU_SchonInitialisiert = FALSE;

void PU_InitAposteriori(void)
/* In den Modi completion und proof werden beide Puffer eingerichtet, in den Modi convergence und confluence
   nur der TP-Puffer, in den uebrigen Modi keiner der beiden Puffer.
*/
{
  if (!PU_SchonInitialisiert){
    switch (PA_ExtModus){
    case completion:
    case proof:                  /* RE-Puffer einrichten, fuer TP-Puffer weiter 'durchfallen' */
      hoechsteREPufferPos = PU_REPufferGroesse;
      REPuffer = (RegelOderGleichungsT *) our_alloc(hoechsteREPufferPos * sizeof(RegelOderGleichungsT));
    case convergence:
    case confluence:             /* TP-Puffer einrichten*/
    case termination:
      hoechsteTPPufferPos = PU_TPPufferGroesse;
      TPPuffer = (TPPuffZellenT *) our_alloc(hoechsteTPPufferPos * sizeof(TPPuffZellenT));
      break;
    default:                     /* reduction:keine Puffer noetig.*/
      break;
    }
  }
  PU_SchonInitialisiert = TRUE;
}
   
   
/* ------------------------------------------ Funktionen RE-Puffer ---------------------------------------------*/


void PU_REPufferNeu(void)
/* Bereitet Puffer auf neue Eingaben vor (Initialisierung).*/
{
   REPufferPos = 0;
   REMerker = 0;
   
}

void PU_REObjektEinfuegen(RegelOderGleichungsT RoderE)
/* Fuegt ein neues Objekt in den Puffer ein.*/
{
   if (REPufferPos == hoechsteREPufferPos){
      hoechsteREPufferPos += PU_REPufferGroesse;
      REPuffer = (RegelOderGleichungsT *) our_realloc(REPuffer, 
                 hoechsteREPufferPos * sizeof(RegelOderGleichungsT),"Erweitern des Regel/Gleichungsspuffers");
   }
   REPuffer[REPufferPos] = RoderE;
   REPufferPos += 1;
}

void PU_REMerkerSetzen(void)
/* Merker auf aktuelle Puffer-Position setzen, um spaeter alle folgenden Eintraege behandeln zu koennen.*/
{
  REMerker = REPufferPos;
}

void PU_REPufferDurchlauf(REPufferAnwendungsprozT AP)
/* Wendet AP auf alle Eintraege des Puffers an.*/
{
  unsigned long pos = 0;

  while (pos < REPufferPos)
    (*AP)(REPuffer[pos++]);
  REPufferPos = 0;
}

void PU_REPufferDurchlauf2(REPufferAnwendungsprozT AP)
/* Wendet AP auf alle Eintraege des Puffers an.*/
{
  unsigned long pos = 0;

  while (pos < REPufferPos)
    (*AP)(REPuffer[pos++]);
}

void PU_REPufferDurchlaufAbMerker(REPufferAnwendungsprozT AP)
/* Wendet AP auf alle Eintraege des Puffers an, die nach dem Setzen des Merkers eingefuegt wurden. */
{
  unsigned long pos = REMerker;

  while (pos < REPufferPos)
    (*AP)(REPuffer[pos++]);
}

BOOLEAN PU_ObjektNochNichtInREPuffer(RegelOderGleichungsT RoderE)
/* Liefert TRUE, wenn das angegebene Objekt noch nicht im REPuffer gespeichert ist. Test auf ==.*/
{
  unsigned long pos = 0;
  while (pos<REPufferPos)
    if (REPuffer[pos++] == RoderE)
      return FALSE;
  return TRUE;
}
        


/* ------------------------------------------ Funktionen TP-Puffer ---------------------------------------------*/

static BOOLEAN TPPufferBenutzt = FALSE;

void PU_TPPufferNeu(void)
/* Bereitet Puffer auf neue Eingaben vor (Initialisierung).*/
{
   TPPufferBenutzt = TRUE;
   TPPufferPos = TPLesePos = 0;
}


void PU_TPEinfuegen(TermT links, TermT rechts, unsigned long int Nummer)
/* Fuegt ein Termpaar (aus der Unifikation entstanden) in den Puffer ein, wobei ein Elternteil als Vermerk mit-
   gefuehrt werden muss (das zweite Elternteil ist diejenige Regel oder Gleichung, zu der gerade TPs gebildet werden).*/
{
  if (TPPufferPos == hoechsteTPPufferPos){
    hoechsteTPPufferPos += PU_TPPufferGroesse;
    TPPuffer = (TPPuffZellenT *) our_realloc(TPPuffer, 
                                             hoechsteTPPufferPos * sizeof(TPPuffZellenT), "Erweitern des TP-Puffers");
  }
  TPPuffer[TPPufferPos].linkeSeite = links;
  TPPuffer[TPPufferPos].rechteSeite = rechts;
  TPPuffer[TPPufferPos++].Nummer = Nummer;
}


void PU_TPAuslesen(TermT *links, TermT *rechts, unsigned long int  *nr)
/* Liefert ein TP, das im Puffer steht. Vor.: Puffer nicht leer! Die Leseposition wird eins weitergesetzt.*/
{
  *links = (TPPuffer[TPLesePos].linkeSeite);
  *rechts = (TPPuffer[TPLesePos].rechteSeite);
  *nr = (TPPuffer[TPLesePos++].Nummer);
}


BOOLEAN PU_TPPufferNichtLeer(void)
/* Liefert TRUE, wenn der TP-Puffer noch nicht leer ist.*/
{
   if (TPLesePos < TPPufferPos)
      return TRUE;
   else{
      TPLesePos = 0;
      return FALSE;
   }
}

void PU_TPPufferLoeschen(void)
/* Loescht alle Terme, die sich noch im TP-Puffer befinden. */
{
   if (TPPufferBenutzt){
      TPLesePos = 0;
      while (TPLesePos < TPPufferPos){
         TO_TermeLoeschen(TPPuffer[TPLesePos].linkeSeite,TPPuffer[TPLesePos].rechteSeite);
         TPLesePos++;
      }
   }
}
