(* $Id: admissible.ml,v 1.13 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(* makes all remainder literals
   - sharing (universal or parameteric) variables parametric
   - containing parameters and sharing variables non-remainder literals parametric
   - reverses all universal -> parametric bindings of the remaining unprocessed
     remainder literals
   - makes the remainder literals still containing parameters parametric
   - leaves the remaining remainder literals universal
*)


(*** types ***)

type var = Var.var
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type raw_context_unifier = Context_unifier.raw_context_unifier

(* just pair the variables of a remainder literal
   and its producing input literal for convenience *)
type remainder_vars = {
  rv_remainder_input_vars: Subst.var list;
  rv_remainder_literal_vars: Subst.var list;
}






(* get the variables for each input literal producing a remainder literal *)
let vars_of_remainder_input_literals
  (raw_context_unifier: raw_context_unifier)
  (remainder_states: bool array) :
  (Subst.var list) list =

  Array.fold_left
    (fun acc input_partner ->
       if remainder_states.(input_partner.Context_unifier.ip_index) then
	 input_partner.Context_unifier.ip_vars :: acc
       else
	 acc
    )
    []
    raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_input_partners


(* adds the parameters of context literals not producing a remainder literal *)
let add_pars_of_non_remainder_context_literals
  (raw_context_unifier: raw_context_unifier)
  (remainder_states: bool array)
  (old_pars: Subst.var list)
  : Subst.var list =

  Array.fold_left
    (fun acc input_partner ->
       let index =
	 input_partner.Context_unifier.ip_index
       in
	 if remainder_states.(index) then
	   acc
	 else
	   let context_partner =
	     raw_context_unifier.Context_unifier.rcu_context_partners.(index)
	   in
	   let context_offset =
	     Subst.context_literal_offset index
	   in
	     List.fold_left
	       (fun pars par ->
		  Subst.make_var par context_offset :: pars
	       )
	       acc
	       context_partner.Context_unifier.cp_element.Context.el_pars
    )
    old_pars
    raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_input_partners





(* binds all unbound vars to fresh parameters, thus producing a v-free literal

   returns the adapted substitution and the previously plus the newly
   created fresh parameters

   fresh parameters use their own offset, fresh_par_offset,
   their id is just determined by counting the number of
   currently created fresh paramters. *)
let rec make_v_free
  (subst: subst)
  (vars: Subst.var list)
  (fresh_pars: Subst.var list)
  : subst * Subst.var list =

  match vars with
    | [] ->
	(subst, fresh_pars)

    | var :: variables_tail ->
	(* only bind variables, not parameters *)
	if Var.is_universal var.Subst.sv_var then begin

	  match Subst.get subst var with
	    | None ->
		let fresh_par =
		  Var.create_parametric (List.length fresh_pars)
		in
		let fresh_term = 
		  Term.request_var fresh_par
		in

		let new_subst = 
		  Subst.set_ ~p_preserving:true ~i_preserving:false
		    subst var.Subst.sv_var var.Subst.sv_offset fresh_term Subst.fresh_par_offset
		in
		  make_v_free
		    new_subst
		    variables_tail
		    ((Subst.make_var fresh_par Subst.fresh_par_offset) :: fresh_pars)

	    | Some _ ->
		(* already bound (to a fresh var?),
		   probably by make_v_free on another remainder literal *)
		make_v_free subst variables_tail fresh_pars
	end

	else
	  make_v_free subst variables_tail fresh_pars



let is_pure (vars: Subst.var list) : bool =
  List.for_all
    (fun var ->
       Var.is_universal var.Subst.sv_var
    )
    vars




(* - makes all remainder literals which share variables (universal and parameteric)
     (universal) variable free and removes them
   - removes all parameter free literals (as pure)
   - makes all remaining remainder literals
     which share variables with non-remainder literals
     (universal) variable free and removes them

   returns
   - the adapted subst
   - the yet unprocessed remainder literals
   - the created fresh parameters *)
let process_shared_remainder_literals
  (_subst: subst)
  (_remainder_vars: remainder_vars list)
  (_shared_remainder_literal_vars: Subst.var list)
  (_shared_input_literal_vars: Subst.var list)
  : subst * remainder_vars list * Subst.var list =

  let rec do_process_shared_remainder_literals
    (subst: subst)
    (remainder_vars: remainder_vars list)
    (unprocessed_remainder_vars: remainder_vars list)
    (fresh_pars: Subst.var list)
    : subst * remainder_vars list * Subst.var list =
    
    match remainder_vars with
      | [] ->
	  subst, unprocessed_remainder_vars, fresh_pars
	    
      | remainder_vars_head :: remainder_vars_tail ->
	  (* variables shared between remainders? *)
	  if 
	    Tools.do_lists_intersect
	      Subst.var_equal
	      _shared_remainder_literal_vars
	      remainder_vars_head.rv_remainder_literal_vars
	  then begin
	    let (new_subst, new_fresh_pars) =
	      make_v_free subst remainder_vars_head.rv_remainder_literal_vars fresh_pars
	    in
	      do_process_shared_remainder_literals
		new_subst
		remainder_vars_tail
		unprocessed_remainder_vars
		new_fresh_pars
	  end

          (* pure remainder literals? *)
	  else if is_pure remainder_vars_head.rv_remainder_literal_vars then begin
	    do_process_shared_remainder_literals
	      subst
	      remainder_vars_tail
	      unprocessed_remainder_vars
	      fresh_pars
	  end
	    
	  (* remainder literals containing clause variables shared with other input literals? *)
	  else if 
	    Tools.do_lists_intersect
	      Subst.var_equal
	      _shared_input_literal_vars
	      remainder_vars_head.rv_remainder_literal_vars
	  then begin
	    let (new_subst, new_fresh_pars) =
	      make_v_free subst remainder_vars_head.rv_remainder_literal_vars fresh_pars
	    in
	      do_process_shared_remainder_literals
		new_subst
		remainder_vars_tail
		unprocessed_remainder_vars
		new_fresh_pars
	  end

	  else begin
	    do_process_shared_remainder_literals
	      subst
	      remainder_vars_tail
	      (remainder_vars_head :: unprocessed_remainder_vars)
	      fresh_pars
	end

  in
    do_process_shared_remainder_literals
      _subst
      _remainder_vars
      []
      []




(* reverse all var -> par bindings of the remaining remainder literals.
   don't reverse parameters of non-remainder context literals *)
let reverse_subst
  (subst: subst)
  (remainder_vars: remainder_vars list)
  (keep_pars: Subst.var list)
  : subst =

  List.fold_left
    (fun acc remainder_vars ->
       Subst.reverse_var_to_par
       acc
       remainder_vars.rv_remainder_input_vars
       keep_pars
    )
    subst
    remainder_vars







let rec parameterize_mixed_literals
  (subst: subst)
  (remainder_vars: remainder_vars list)
  (fresh_pars: Subst.var list)
  : subst =
  
  match remainder_vars with
    | [] ->
	subst
	
    | remainder_vars_head :: remainder_vars_tail ->
	let reversed_vars =
	  Subst.get_bound_vars2 subst remainder_vars_head.rv_remainder_input_vars
	in
	  if is_pure reversed_vars then begin
	    parameterize_mixed_literals subst remainder_vars_tail fresh_pars
	  end

	  else begin
	    let universal_vars_in_remainder_literal =
	      List.find_all
		(fun var ->
		   Var.is_universal var.Subst.sv_var)
		reversed_vars
	    in

	    let (new_subst, new_fresh_pars) =
	      make_v_free subst universal_vars_in_remainder_literal fresh_pars
	    in
	      parameterize_mixed_literals new_subst remainder_vars_tail new_fresh_pars
	  end



      








(* transforms subst so that all remainder literals are parameter or variable free *)
let make_admissible
  (raw_context_unifier: raw_context_unifier)
  (subst: subst)
  (remainder_states: bool array)
  : subst =

  (* get the variables for each input literal producing a remainder literal *)
  let (remainder_input_vars: Subst.var list list) =
    vars_of_remainder_input_literals raw_context_unifier remainder_states
  in

  (* get the variables of the remainder literals.
     as the remainder literals are not build yet,
     get the variables in the terms bound to the input literals instead. *)
  let (remainder_literal_vars: Subst.var list list) =
    Subst.get_bound_vars3 subst remainder_input_vars
  in

  (* pair the input and remainder variables for convenience *)
  let (remainder_vars: remainder_vars list) =
    List.map2
      (fun remainder_input_vars remainder_literal_vars ->
	 {
	   rv_remainder_input_vars = remainder_input_vars;
	   rv_remainder_literal_vars = remainder_literal_vars;
	 }
      )
      remainder_input_vars
      remainder_literal_vars
  in

  (* get the variables occuring in more than one remainder literal *)
  let (shared_remainder_literal_vars: Subst.var list) =
    Tools.lists_shared Subst.var_equal remainder_literal_vars
  in

  (* process and remove the literals sharing variables *)
  let (unshared_subst, unshared_remainder_vars, fresh_pars) =
    process_shared_remainder_literals
      subst
      remainder_vars
      shared_remainder_literal_vars
      raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_shared_vars
  in
    match unshared_remainder_vars with
      | [] ->
	  unshared_subst
	    
      | _ ->
	  (* reverse all var -> par bindings of the remaining remainder literals
	     which do not introduce new remainder variables
	     this excludes:
	     - fresh parameters
	     - parameters occuring in several remainder literals
	     - non-remainder parameters  *)

	  let shared_fixed_pars =
	    List.fold_left
	      (fun acc var ->
		 if Var.is_parametric var.Subst.sv_var then
		   var :: acc
		 else
		   acc
	      )
	      fresh_pars
	      shared_remainder_literal_vars
	  in

	  let all_fixed_pars =
	    add_pars_of_non_remainder_context_literals 
	      raw_context_unifier
	      remainder_states
	      shared_fixed_pars
	  in

  	  let reversed_subst =
	    reverse_subst
	      unshared_subst
	      unshared_remainder_vars
	      all_fixed_pars
	  in
	    (* make all mixed literal variable free *)
	    parameterize_mixed_literals
	      reversed_subst
	      unshared_remainder_vars
	      fresh_pars


	      
