(* $Id: config.mli,v 1.29 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** global configuration

  As part of the module initialization the command line arguments are evaluated
  and the options set accordingly.
*)




(** {6 Types} *)

(** the derivation is aborted (because of a limitation of the implementation)
  although no proof has been found. *)
exception NO_SOLUTION




(** {6 Functions} *)

(** {2 Options - Mechanism } *)

(** use the {e Resolve} rule *)
val get_use_resolve: unit -> bool
val set_use_resolve: bool -> unit

(** use the {e Subsume} rule *)
val get_use_subsume: unit -> bool
val set_use_subsume: bool -> unit

(** use the {e Compact} rule *)
val get_use_compact: unit -> bool
val set_use_compact: bool -> unit

(** don't use (by the implementation detectable) unproductive remainders *)
val get_use_productivity: unit -> bool
val set_use_productivity: bool -> unit

(** use {e +v} (instead of {e -v}) as the default context literal.

  may not be altered during the derivation,
  otherwise the context
  a change doesn't work as this variable is read only once
  by the term database *)
val get_use_plus_v: unit -> bool
val set_use_plus_v: bool -> unit

(** use backjumping instead of dynamic backtracking.
  backjumping has less overhead, dynamic backtracking might (hopefully) need less Splits *)
val get_use_chronological_backtracking: unit -> bool
val set_use_chronological_backtracking: bool -> unit

(** check if the problem is Horn.
    Otherwise {!Config.set_horn} is never used. *)
val get_use_horn_detection: unit -> bool
val set_use_horn_detection: bool -> unit

(** use term indexing context checks. *)
val get_use_term_indexing: unit -> bool
val set_use_term_indexing: bool -> unit


(** {2 Options - Verboseness } *)

(** show the options settings after the derivation *)
val get_show_options: unit -> bool

(** show the statistic after the derivation *)
val get_show_statistic: unit -> bool

(** show the model after the derivation (if the problem is satisfiable) *)
val get_show_model: unit -> bool

(** print the model to a file (if the problem is satisfiable)
  (returns the filename) *)
val get_print_model_to_file: unit -> string
val set_print_model_to_file: string -> unit

(** show the application of the derivation rules
  assert/split/close and backtracking. *)
val get_show_derivation: unit -> bool
val set_show_derivation: bool -> unit

(** print the derivation/proof in the graphviz dot format to a file,
  see {!Derivation_to_dot}.
  (returns the filename) *)
val get_derivation_to_dot_file_name: unit -> string
val set_derivation_to_dot_file_name: string -> unit
val get_derivation_to_dot_file: unit -> bool

(** show the context unifiers used in the derivation *)
val get_show_context_unifier: unit -> bool
val set_show_context_unifier: bool -> unit

(** show the context after every change to it (i.e. Assert, Split, Backtrack, Restart) *)
val get_show_derivation_context: unit -> bool
val set_show_derivation_context: bool -> unit

(** show every computed assert candidate *)
val get_show_computed_assert_candidates: unit -> bool
val set_show_computed_assert_candidates: bool -> unit

(** show every computed split candidate *)
val get_show_computed_split_candidates: unit -> bool
val set_show_computed_split_candidates: bool -> unit


(** is the problem a Horn problem? (this enables some optimizations) *)
val get_horn: unit -> bool

(** mark the problem as a Horn problem.
  implicitely also activates {!Config.get_use_chronological_backtracking}
  as backtracking never occurs for Horn. *)
val set_horn: unit -> unit



(** {2 Problem File } *)

(** the problem file name given on startup *)
val get_problem_file_name: unit -> string

(** change the problem file name. E.g. add a default suffix *)
val set_problem_file_name: string -> unit


(** {2 Depth Bound } *)

(** (Iterative Deepening over the depth bound).
  Split and Assert candidates exceeding a certain term depth bound
  are ignored during the derivation. Thus if a model has been found
  but some candidates have been dropped, the derivation has to be restarted
  with a higher depth bound. *)

(** set the initial depth bound for a derivation *)
val set_initial_depth_bound: int -> unit

(** get current depth bound *)
val get_current_depth_bound: unit -> int

(** get previous depth bound. -1 if there has been no restart yet *)
val get_prev_depth_bound: unit -> int

(** [check_depth_bound depth] checks if [depth] is not above the current depth bound.
    Failures are remembered for later restarts and depth bound increment. *)
val check_depth_bound: int -> bool

(** [true], if some candidates have been dropped because of {!Config.check_depth_bound}.
    Then the current and previous depth bound are adjusted accordingly *)
val update_depth_bound: unit -> bool

(** reset the depth bound to the inital depth bound *)
val reset_depth_bound: unit -> unit


(** {2 Representation} *)

val to_string: unit -> string

(** {2 Misc} *)

(** compile the (slow) debug version with additional consistency checks *)
val debug: bool
