(* $Id: context.ml,v 1.32 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(*** types ***)


type symbol = Symbol.symbol
type var = Var.var
type term = Term.term
type literal = Term.literal
type subst =  Subst.subst
type state = State.state

type element = {
  el_literal: literal;
  el_state: state;
  el_pars: var list;
  el_generation: int;
  el_mandatory_at: state option;
  mutable el_compacted: state list;
}



(*** string representation ***)

let element_to_string (element: element) : string =
  "state:    " ^ State.state_to_string element.el_state ^ "\n"
  ^ "literal: " ^ Term.literal_to_string element.el_literal ^ "\n"


module STIndex =
  Substitution_tree.Make (
    struct
      type t = element

      let is_equal (first: t) (second: t) : bool =
	first == second

      let to_string (element: t) : string =
	element_to_string element
    end
  )


module DTIndex =
  Discrimination_tree.Make (
    struct
      type t = element

      let is_equal (first: t) (second: t) : bool =
	first == second

      let to_string (element: t) : string =
	element_to_string element
    end
  )

module Indexes =
  Hashtbl.Make (
    struct
      type t = bool * term

      let equal (sign1, term1) (sign2, term2) : bool =
	(sign1 = sign2)
	&&
	match term1, term2 with
	  | Term.Const symbol1, Term.Const symbol2
	  | Term.Func (symbol1, _, _), Term.Func (symbol2, _, _) ->
	      Symbol.equal symbol1 symbol2

	  | _ ->
	      false

      let hash (_, term) : int =
	match term with
	  | Term.Const symbol
	  | Term.Func (symbol, _, _) ->
	      Symbol.id_of_symbol symbol

	  | _ ->
	      failwith "Context.Indexes.hash"
    end
  )

type index_type =
  | STI of STIndex.t
  | DTI of DTIndex.t

type indexing_method =
  | NONE (** non term indexing *)
  | ST   (** substitution trees *)
  | DT   (** discrimination trees *)


(* the context is stored in
- a list for easy backtracking
- an index for fast term checks *)
type context = {
  (* the context elements *)
  mutable context: element list;

  (* the used index type *)
  mutable indexing_method: indexing_method;

  (* term indexing over the context. an index for each sign/predicate pair *)
  index: index_type Indexes.t;
}




(*** constants ***)

let null_element : element =
  {
    el_literal = Term.null_literal;
    el_pars = [];
    el_generation = 0;
    el_state = State.invalid_state;
    el_compacted = [];
    el_mandatory_at = None;
  }



(*** globals ***)

(* the context *)
let (global: context) = {
  context = [];
  indexing_method = ST;
  index = Indexes.create 16;
}






(*** init ***)

let init (indexing_method: indexing_method) =
  global.context <- [];
  global.indexing_method <- indexing_method;
  Indexes.clear global.index




(*** string representation ***)
let context_to_string ?(show_compacted: bool = false) () : string =
  State.current_state_to_string ()
  ^ "Context: "
  ^ "["
  ^ String.concat "; "
    (List.map 
       (fun element ->
	  let show =
	    not show_compacted
	    ||
	    match element.el_compacted with
	      | [] ->
		  true

	      | _ ->
		  false
	  in
	    if show then
	      Term.literal_to_string element.el_literal
	    else
	      ""
       )
       global.context)
  ^ "]"




let iter (func: element -> unit) : unit =
  List.iter func global.context

let fold (func: 'a -> element -> 'a) (acc: 'a) : 'a =
  List.fold_left func acc global.context





(*** term indexing wrapper ***)

(* unfortunately, although both Discrimination_tree and Substitution_tree
   are instances of Term_indexing, because of OCamls type safety handling
   we have to have thse wrappers in order to be able to make a decision
   at runtime on which indexing module we want to use *)

let index_create () : index_type =
  match global.indexing_method with
    | ST ->
	STI (STIndex.create ())

    | DT ->
	DTI (DTIndex.create ())

    | NONE ->
	failwith "index_create: NONE"


let index_add (index: index_type) (term: term) (element: element) : unit =
  match index with
    | STI index ->
	STIndex.add index term element

    | DTI index ->
	DTIndex.add index term element

let index_remove (index: index_type) (term: term) (element: element option) : bool =
  match index with
    | STI index ->
	STIndex.remove index term element

    | DTI index ->
	DTIndex.remove index term element

let index_iter (func: term -> element -> unit) (index: index_type) : unit =
  match index with
    | STI index ->
	STIndex.iter func index

    | DTI index ->
	DTIndex.iter func index


let index_find_all_instances ~(p_preserving: bool) (index: index_type) (term: term) : element list =
  match index with
    | STI index ->
	STIndex.find_all_instances ~p_preserving:p_preserving index term

    | DTI index ->
	DTIndex.find_all_instances ~p_preserving:p_preserving index term


let index_find_generalization ~(p_preserving: bool) (index: index_type) (term: term) : element option =
  match index with
    | STI index ->
	STIndex.find_generalization ~p_preserving:p_preserving index term

    | DTI index ->
	DTIndex.find_generalization ~p_preserving:p_preserving index term

let index_find_unifiable ~(p_preserving: bool) (index: index_type) (term: term) : element option =
  match index with
    | STI index ->
	STIndex.find_unifiable ~p_preserving:p_preserving index term

    | DTI index ->
	DTIndex.find_unifiable ~p_preserving:p_preserving index term

let index_find_non_productive (index: index_type) (producer: term) (produced: term) : element option =
  match index with
    | STI index ->
	STIndex.find_non_productive index producer produced

    | DTI index ->
	DTIndex.find_non_productive index producer produced

let index_find_non_msg (index: index_type) (producer: term) (produced: term) : element option =
  match index with
    | STI index ->
	STIndex.find_non_msg index producer produced

    | DTI index ->
	DTIndex.find_non_msg index producer produced

let index_to_string (index: index_type) : string =
  match index with
    | STI index ->
	STIndex.to_string index

    | DTI index ->
	DTIndex.to_string index




(*** access ***)

let update_max_size () : unit =
  Statistic.set_max_context_size 
    (max
       (Statistic.get_max_context_size ())
       (List.length global.context)
    )


let parameters_of_literal (literal: literal) : var list =
  List.find_all
    Var.is_parametric
    (Term.vars_of_literal literal)


(* compact the context with literal *)
let compact (index: index_type) (literal: literal) : unit =
  (* compact context with the new literal *)
  if not (Config.get_use_compact ()) then begin
    ()
  end

  else begin
    (* find all compacted elements *)
    let compacted_elements =
      if Config.get_use_term_indexing () then
	index_find_all_instances ~p_preserving:true index literal.Term.term
      else
	List.find_all
	  (fun element ->
	     try
	       ignore (Unification.match_literals ~p_preserving:true
			 literal 0
			 element.el_literal 1);
	       true
	     with
	       | Unification.UNIFICATION_FAIL ->
		   false
	  )
	  global.context
    in
      (* mark all uncompacted elements compacted *)
      List.iter
	(fun compacted_element ->
	   if not
	     (List.exists
		(fun compacted_at ->
		   State.depends_on (State.current_state ()) compacted_at
		)
		compacted_element.el_compacted
	     )
	   then begin
	     Statistic.inc_compact ();
	     compacted_element.el_compacted <- (State.current_state ()) :: compacted_element.el_compacted;
	     
	     (* remove compacted literals from the term index *)
	     if Config.get_use_term_indexing () then begin
	       if not (index_remove index compacted_element.el_literal.Term.term (Some compacted_element)) then
		 failwith "Context.compact"
	     end;
	   end
	)
	compacted_elements
  end


let add (literal: literal) (mandatory_at: state option)
  (generation: int): element =

  let new_context_element = {
    (* the index of an element is always its index in global.context *)
    el_literal = literal;
    el_state = State.current_state ();
    el_pars = parameters_of_literal literal;
    el_generation = generation;
    el_mandatory_at = mandatory_at;
    el_compacted = [];
  }
  in

  (* get the index where the literal must be inserted *)
  let index =
    try
      Indexes.find global.index
	(new_context_element.el_literal.Term.sign,
	 new_context_element.el_literal.Term.term)
    with
      | Not_found ->
	  let new_index =
	    index_create ()
	  in
	    Indexes.add
	      global.index
	      (new_context_element.el_literal.Term.sign,
	       new_context_element.el_literal.Term.term)
	      new_index;
	    new_index
  in
    compact index literal;

    (* add new literal to context and index *)
    global.context <- new_context_element :: global.context;

    if Config.get_use_term_indexing () then begin
      index_add index new_context_element.el_literal.Term.term new_context_element;
    end;

    if Config.debug then begin
      Indexes.iter
	(fun _ index ->
	   index_iter
	   (fun term element ->
	      if not (Term.term_equal term element.el_literal.Term.term) then begin
		print_endline ("\n\n\nFAIL=====\n");
		print_endline (index_to_string index);
		print_endline ("ADD: " ^ Term.literal_to_string new_context_element.el_literal);
		print_endline ("Term             : " ^ Term.term_to_string term);
		print_endline ("But node contains: " ^ element_to_string element);
		failwith "Context.add"
	      end
	   )
	   index
	)
	global.index;
    end;
    

(*    if (List.length global.context) mod 100 = 0 then begin
      print_endline (Statistic.to_string ());
    end;*)

(*    if (List.length global.context) = 150 then begin
      exit 0;
    end;*)

(*    Term.clean ();*)
(*
    print_endline ("GLOBAL LENGTH: " ^ string_of_int (List.length global.context));
    if (List.length global.context) = 10000 then begin
    if (List.length global.context) mod 100 = 0 then begin
(*      Indexes.iter
	(fun literal index ->
	   print_endline ("INDEX FOR : " ^ Term.literal_to_string literal);
	   print_endline (Index.to_string index);
	)
	global.index;
      print_endline (Statistic.to_string ());
      exit 0;*)
      Term.dump ();
      exit 0;
    end;
*)

    (* verboseness *)
    if Config.get_show_derivation_context () then begin
      print_endline (context_to_string ());
    end;
    

    new_context_element




let to_file (file_name: string) : unit =
  let out_channel =
    open_out file_name
  in

    List.iter
      (fun element ->
	 match element.el_compacted with
	   | [] ->
	       output_string out_channel (Term.literal_to_string element.el_literal);
	       output_string out_channel "\n";

	   | _ ->
	       ()
      )
      (* print older elements first *)
      (List.rev global.context);

    close_out out_channel



let from_file ?(indexing_method:indexing_method = ST) (file_name: string) : unit =
  (* drop the old context *)
  init (indexing_method);

  let in_channel =
    open_in file_name
  in

  let rec read_literals () =
    try
      let line =
	input_line in_channel
      in
      let literal =
	Read_darwin.to_literal line
      in
	ignore (add literal None 0);
	read_literals ()
    with
      | End_of_file ->
	  ()
  in
    read_literals ();
    close_in in_channel






(*** checks ***)


let is_element_mandatory_for_state (element: element) (candidate_state : state) : bool =
  (* the candidate's state is dependend on the conflicting element *)
  State.depends_on candidate_state element.el_state
  ||
  (* the element is mandatory in a state on which the candidate's state is dependend *)
  match element.el_mandatory_at with
    | None ->
	false

    | Some mandatory_state ->
	State.depends_on candidate_state mandatory_state




(*** contradictory ***)

let check_contradictory_no_indexing (literal: literal) : element option =

  let rec do_check (elements: element list) : element option =
    match elements with
      | [] ->
	  None

      | element :: tail ->
	  if literal.Term.sign = element.el_literal.Term.sign then
	    do_check tail

	  else begin
	    try
	      ignore (Unification.unify_terms ~p_preserving:true
			element.el_literal.Term.term 0
			literal.Term.term 1);
	      Some element;
	    with
	      | Unification.UNIFICATION_FAIL ->
		  do_check tail
	  end
  in
    do_check global.context


let check_contradictory_indexing (literal: literal) : element option =
  try
    let index =
      Indexes.find global.index (not literal.Term.sign, literal.Term.term)
    in
      index_find_unifiable ~p_preserving:true index literal.Term.term
	
  with
    | Not_found ->
	None


let check_contradictory (literal: literal) : element option =
  if Config.get_use_term_indexing () then
    check_contradictory_indexing literal
  else
    check_contradictory_no_indexing literal





(*** subsumed ***)


let check_subsumed_no_indexing (literal: literal) : element option =

  let rec do_check (elements: element list) : element option =
    match elements with
      | [] ->
	  None

      | element :: tail ->
	  if literal.Term.sign <> element.el_literal.Term.sign then
	    do_check tail

	  else begin
	    try
	      ignore (Unification.match_terms ~p_preserving:true
			element.el_literal.Term.term 0
			literal.Term.term 1);
	      Some element;
	    with
	      | Unification.UNIFICATION_FAIL ->
		  do_check tail
	  end
  in
    do_check global.context


let check_subsumed_indexing (literal: literal) : element option =
  try
    let index =
      Indexes.find global.index (literal.Term.sign, literal.Term.term)
    in
      index_find_generalization ~p_preserving:true index literal.Term.term
	      
  with
    | Not_found ->
	None


let check_subsumed (literal: literal) : element option =
  if Config.get_use_term_indexing () then
    check_subsumed_indexing literal
  else
    check_subsumed_no_indexing literal







(*** productive ***)


let check_productive_no_indexing (producer: literal) (produced: literal) : element option =

  let neg_producer =
    Term.request_negated_literal producer
  and neg_produced =
    Term.request_negated_literal produced
  in

  let rec do_check (elements: element list) : element option =
    match elements with
      | [] ->
	  None

      | element :: tail ->
	  (* producer must be ignored *)
	  if Unification.are_literals_variants producer element.el_literal then
	    do_check tail
	      
	  (* produced (must) be ignored *)
	  else if Unification.are_literals_variants produced element.el_literal then
	    do_check tail

	  else begin
	    (* msg *)
	    try
	      ignore (Unification.match_literals producer 0 element.el_literal 1);
	      ignore (Unification.match_literals element.el_literal 0 neg_produced 1);
	      Some element;
	    with
	      | Unification.UNIFICATION_FAIL ->
		  begin
		    (* instance *)
		    try
		      (* produced might be a mixed non-remainder literal - but that's of no importance for match *)
		      ignore (Unification.match_literals element.el_literal 0 produced 1);
		
		      begin
			match element.el_pars with
			  | [] ->
			      (* non-parametric, so must be unifiable *)
			      ignore (Unification.unify_literals neg_producer 0 element.el_literal 1);
			      Some element;
			      
			  | _ ->
			      (* parametric, so must be instance *)
			      ignore (Unification.match_literals neg_producer 0 element.el_literal 1);
			      Some element;
		      end
		    with
		      | Unification.UNIFICATION_FAIL ->
			  do_check tail
		  end
	  end
  in

    (* if they are identical one literal always produces the other  *)
    if Term.literal_equal neg_producer produced then
      None
    else
      do_check global.context



let check_productive_indexing (producer: literal) (produced: literal) : element option =

  (* check second part of productivity *)
  let do_check_productive () =
    try
      let index_produced =
	Indexes.find global.index (produced.Term.sign, produced.Term.term)
      in
	index_find_non_productive index_produced producer.Term.term produced.Term.term
    with
      | Not_found ->
	  None
  in

  (* check first part of productivity, i.e. msg *)
  try
    let index_producer =
      Indexes.find global.index (producer.Term.sign, producer.Term.term)
    in
      begin
	match index_find_non_msg index_producer producer.Term.term produced.Term.term with
	  | None ->
	      do_check_productive ()
	      
	  | Some element as conflict ->
	      conflict
      end
	  
  with
    | Not_found ->
	do_check_productive ()


let check_productive (producer: literal) (produced: literal) : element option =
  if Config.get_use_productivity () then begin
    let check =
      if Config.get_use_term_indexing () then
	check_productive_indexing producer produced
      else
	check_productive_no_indexing producer produced
    in
      begin
	match check with
	  | None ->
	      ()
	      
	  | Some _ ->
	      Statistic.inc_filtered_by_productivity ();
      end;
      check
  end
    
  else
    None




(*** Backtracking ***)

(* remove all context literal added in invalid states *)
let backtrack_elements () : unit =

  let remove_from_index (element: element) : unit =
    if Config.get_use_term_indexing () then begin
      let index =
	try
	  Indexes.find
	    global.index
	    (element.el_literal.Term.sign, element.el_literal.Term.term)
	with
	  | Not_found ->
	      failwith "Context.backtrack index I"
      in
      let removed =
	index_remove
	  index
	  element.el_literal.Term.term
	  (Some element)
      in
	if Config.debug then begin
	  Indexes.iter
	    (fun _ index ->
	       index_iter
	       (fun term element ->
		  if not (Term.term_equal term element.el_literal.Term.term) then begin
		    print_endline ("\n\n\nFAIL=====\n");
		    print_endline (index_to_string index);
		    print_endline ("REMOVE: " ^ Term.literal_to_string element.el_literal);
		    print_endline ("Term             : " ^ Term.term_to_string term);
		    print_endline ("But node contains: " ^ element_to_string element);
		    failwith "Context.backtrack"
		end
	       )
	       index
	    )
	    global.index;
	end;

      if Config.debug && not removed then begin
	failwith "Context.backtrack index II"
      end

    end;
  in

    if Config.get_use_chronological_backtracking () then begin
      (* all invalidated states are the at the beginning of the context *)
      let rec backtrack_chronological () =
	match global.context with
	  | [] ->
	      ()
	      
	  | element :: tail ->
	      if State.is_state_invalid element.el_state then begin
		global.context <- tail;
		remove_from_index element;
		backtrack_chronological ();
	      end
      in
	backtrack_chronological ()
    end
      
    else begin
      global.context <- 
      List.find_all
	(fun element ->
	   if State.is_state_valid element.el_state then begin
	     true
	   end
	     
	   else begin
	     remove_from_index element;
	     false
	   end
	)
	global.context
    end


(* uncompact all context literal uncompacted in invalid states *)
let backtrack_compact () : unit =
  List.iter
    (fun context_element ->
       match context_element.el_compacted with
	 | [] ->
	     ()

	 | _ ->
	     context_element.el_compacted <-
	     List.find_all State.is_state_valid context_element.el_compacted;

	     (* readd not longer compacted literals to the term index *)
	     if Config.get_use_term_indexing () then begin
	       match context_element.el_compacted with
		 | [] ->
		     (* get the index where the literal must be inserted *)
		     let index =
		       try
			 Indexes.find global.index
			   (context_element.el_literal.Term.sign,
			    context_element.el_literal.Term.term)
		       with
			 | Not_found ->
			     let new_index =
			       index_create ()
			     in
			       Indexes.add
				 global.index
				 (context_element.el_literal.Term.sign,
				  context_element.el_literal.Term.term)
				 new_index;
			       new_index
		     in
		       index_add index context_element.el_literal.Term.term context_element
			 
		 | _ ->
		     ()
	     end

    )
    global.context


let backtrack () : unit =
  backtrack_elements ();
  backtrack_compact ()
