(* $Id: context.mli,v 1.16 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** context with operations

  Applies the {e Compact} rule of the model evolution calculus.
*)

(** {6 Types} *)

type var = Var.var
type term = Term.term
type literal = Term.literal
type state = State.state

(** a context literal *)
type element = {
  el_literal: literal;
  (** the context literal *)

  el_state: state;
  (** the state in which the context literal was added *)

  el_pars: var list;
  (** the parameters of the context literal *)

  el_generation: int;
  (** depth of the tree of context literals
    needed to computed this context literal *)

  el_mandatory_at: state option;
  (** this context element must be valid in this state
     (and thus all states depending on this state).
     I.e. in these states this literal was computed
     as an assert literal *)

  mutable el_compacted: state list;
  (** the states in which literals were added to the context
    which compact this literal *)
}


(** the different indexing methods usable by the context *)
type indexing_method =
  | NONE (** non term indexing *)
  | ST   (** substitution trees *)
  | DT   (** discrimination trees *)



(** {6 Functions} *)




(** {2 Initialization} *)

(** empties the context and fixes the indexing method based. *)
val init: indexing_method -> unit




(** {2 Constants} *)

(** for initialization use *)
val null_element: element





(** {2 Access} *)

(** update the maximum context size statistic *)
val update_max_size: unit -> unit


(** [add literal mandatory_at generation]
  adds a new literal to the context *)
val add: literal -> state option -> int -> element


(** {2 Checks} *)

(** [check_contradictory literal]
  checks if [literal] is contradictory with the context.
  Returns the contradictory context element. *)
val check_contradictory: literal -> element option

(** [check_subsumed literal]
  checks if [literal] is subsumed by the context.
  Returns a subsuming context element. *)
val check_subsumed: literal -> element option

(** [check_productive producer produced]
  checks if [producer] is producing [produced] in the context.
  Returns a conflicting context element. *)
val check_productive: literal -> literal -> element option

val iter: (element -> unit) -> unit

val fold: ('a -> element -> 'a) -> 'a -> 'a



(** {2 Properties} *)

(** is the context literal mandatory for [state]?
  I.e. the literal must be part of the context in [state]
  or there must be a candidate which asserts literal to the context in [state].
*)
val is_element_mandatory_for_state: element -> state -> bool



(** {2 Representation} *)

val element_to_string: element -> string

(** represents the context as a list of literals.
    compacted literals are only shown when [show_compacted] is true
    (by default it is false) *)
val context_to_string : ?show_compacted:bool -> unit -> string


(** [to_file file_name]
  writes the (not compacted) context literals to a file, one literal per line.

  raises [Sys_error] on file access failure. *)
val to_file: string -> unit


(** [from_file file_name]
  Reads the context from a file, replacing the current context.
  The file should contain one literal per line,
  readable by {!Read_tme}

  The indexing method may be given, default is ST.

  raises [Sys_error] on file access failure.

  raises [Failure] if file_name contains invalid input. *)
val from_file: ?indexing_method:indexing_method -> string -> unit



(** {2 Backtracking} *)

(** removes all context elements added in invalid states,
  and backtracks compacting of the context. *)
val backtrack : unit -> unit
