(* $Id: context_unifier.ml,v 1.30 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type var = Var.var
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type state = State.state
type explanation = State.explanation
type packed_subst = Unification.packed_subst


type context_partner = {
  cp_element: Context.element;
  mutable cp_subst: packed_subst;
}

type context_partners =
    context_partner array

type input_partner = {
  ip_index: int;
  ip_literal: literal;
  ip_vars: Subst.var list;
  mutable ip_context_partners: context_partner list;
  mutable ip_resolved: context_partner list;
}

type 'a _context_unifier_space = {
  cus_clause: clause;
  cus_shared_vars: Subst.var list;
  cus_input_partners: input_partner array;
  mutable cus_subsumed: State.state list;
  cus_process_assert_candidate: ('a -> unit);
  cus_process_split_candidate: ('a -> unit);
}


type raw_context_unifier = {
  rcu_space: raw_context_unifier _context_unifier_space;
  rcu_context_partners: context_partners;
}

type context_unifier_space = raw_context_unifier _context_unifier_space





exception CLOSED of explanation





(*** constants ***)


let assert_partner = {
  cp_element = {
    Context.el_literal = Term.assert_literal;
    Context.el_state = State.valid_state;
    Context.el_pars = [];
    Context.el_generation = 0;
    Context.el_mandatory_at = None;
    Context.el_compacted = [];
  };
  cp_subst = [| |];
}

let null_partner = {
  cp_element = {
    Context.el_literal = Term.null_literal;
    Context.el_state = State.invalid_state;
    Context.el_pars = [];
    Context.el_generation = 0;
    Context.el_mandatory_at = None;
    Context.el_compacted = [];
  };
  cp_subst = [| |];
}


























(*** misc ***)


(* as long as the states of all its context literals are valid,
   the context unifier is valid *)
let is_raw_context_unifier_invalid (raw_context_unifier: raw_context_unifier) : bool =
  Tools.array_exists
    (fun context_partner ->
       State.is_state_invalid context_partner.cp_element.Context.el_state
    )
    raw_context_unifier.rcu_context_partners


let recompute_unifier (raw_context_unifier: raw_context_unifier) : subst =

  let rec recompute_at (i: int) (acc: subst) : subst =
    if i >= Array.length raw_context_unifier.rcu_space.cus_input_partners then
      acc
    else
      let input_index =
	raw_context_unifier.rcu_space.cus_input_partners.(i).ip_index
      in
      let context_partner =
	raw_context_unifier.rcu_context_partners.(input_index)
      in
      let merged_subst =
	try
	  Unification.unify_packed_substs acc context_partner.cp_subst
	with
	  | Unification.UNIFICATION_FAIL ->
	      print_endline ("CONTEXT_UNIFIER.RECOMPUTE_UNIFIER");
	      print_endline (Subst.subst_to_string acc);
	      print_endline (Subst.subst_to_string (Array.to_list context_partner.cp_subst));
	      failwith "Context_unifier.recompute_unifier"
      in
	recompute_at (i + 1) merged_subst
	  
  in
    recompute_at 0 Subst.empty




(* is this pair producing a remainder literal?
   yes, if one of its parameters is bound to a non-parameter *)
let is_remainder (subst: subst) (pars: var list) (offset: int) =

  let par_offset =
    Subst.context_literal_offset offset
  in

    List.exists
      (fun par ->
	 match Subst.get_ subst par par_offset with
	   | None -> 
	       false
	       
	   | Some term ->
	       begin
		 match term.Subst.st_term with
		   | Term.Var var ->
		       Var.is_universal var

		   | _ -> 
		       true
	       end
      )
      pars





let compute_remainder_states (raw_context_unifier: raw_context_unifier) (subst: subst) : bool array =
  (* no map function as the arrays have a different order *)
  let remainder_states =
    Array.make (Array.length raw_context_unifier.rcu_context_partners) false
  in

    for i = 0 to Array.length raw_context_unifier.rcu_space.cus_input_partners - 1 do
      let input_index =
	raw_context_unifier.rcu_space.cus_input_partners.(i).ip_index
      in
      let context_partner =
	raw_context_unifier.rcu_context_partners.(input_index)
      in
	remainder_states.(input_index) <-
	  is_remainder
	  subst
	  context_partner.cp_element.Context.el_pars
	  input_index;
    done;

    remainder_states




(* a generation of a candidate is
   the max generation of its context_partners + 1 *)
let generation_of_candidate (context_partners: context_partners) : int =
  let max_generation =
    Array.fold_left
      (fun acc context_partner ->
	 max acc context_partner.cp_element.Context.el_generation
      )
      0
    context_partners
  in
    max_generation + 1


let explanation_of_context_partners (context_partners: context_partners) : explanation =

  let rec do_fold (i: int) (acc: literal list) =
    if i >= Array.length context_partners then
      acc
    else begin
      let context_partner =
	context_partners.(i)
      in
	if context_partner == assert_partner then
	  (* skip an assert gap *)
	  do_fold (i + 1) acc
	else
	  let term =
	    context_partner.cp_element.Context.el_literal
	  in
	    do_fold (i + 1) (term :: acc)
    end
  in

  let terms =
    do_fold 0 []
  in

    State.get_explanation terms

let get_creation_state (raw_context_unifier: raw_context_unifier) : state =
  let youngest_state =
    Array.fold_left
      (fun acc context_partner ->
	 match acc with
	   | None ->
	       Some context_partner.cp_element.Context.el_state

	   | Some best_state ->
	       if State.compare_age best_state context_partner.cp_element.Context.el_state < 0 then
		 Some context_partner.cp_element.Context.el_state
	       else
		 acc
      )
      None
      raw_context_unifier.rcu_context_partners
  in
    match youngest_state with
      | None ->
	  failwith "get_creation_state"

      | Some state ->
	  state



let get_input_clause (context_unifier_space: context_unifier_space) : clause =
  context_unifier_space.cus_clause


let get_context_clause (context_partners: context_partners) : clause =
  (* create a clause where the i.th element is the context literal
     paired to the i.th literal in the input clause *)
  Array.fold_right
    (fun context_literal acc ->
       context_literal.cp_element.Context.el_literal :: acc
    )
    context_partners
    []










(*** representation ***)


let context_unifier_to_string
  (context_unifier_space: context_unifier_space) (context_partners: context_partners)
  : string =
  "Clause : "
  ^ Term.clause_to_string (get_input_clause context_unifier_space)
  ^ "\n"
  ^ "Context: "
  ^ Term.clause_to_string (get_context_clause context_partners)
  ^ "\n"
(*  ^ "Explanation: "
  ^ State.explanation_to_string (explanation_of_context_partners context_partners)
  ^ "\n"*)
(*  ^ "Subst: "
  ^ Subst.subst_to_string subst
  ^ "\n"*)


let raw_context_unifier_to_string (raw_context_unifier: raw_context_unifier) : string =
  context_unifier_to_string
    raw_context_unifier.rcu_space
    raw_context_unifier.rcu_context_partners


let unit_clause_context_unifier_to_string (literal: literal) : string =
  "Clause : "
  ^ Term.clause_to_string [literal]
  ^ "\n"
  ^ "Context: "
  ^ Term.clause_to_string [Term.assert_literal]
  ^ "\n"
