(* $Id: context_unifier.mli,v 1.21 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** basic types for and operations on context unifiers

  contains also the types for the search for context unifiers,
  but this is actually done in {!Context_unifier_space}.
*)


(** {6 Types} *)

type var = Var.var
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type state = State.state
type explanation = State.explanation
type packed_subst = Unification.packed_subst


(** a context literal paired to an input clause literal *)
type context_partner = {
  cp_element: Context.element;
  (** the context literal *)
  mutable cp_subst: packed_subst;
  (** the unifier of this context literal and the corresponding input literal *)
}

(** the context literals paired with an input clause.
  A {!Context_unifier.input_partner} is paired with the context literal
  at [context_partners.(input_partner.ip_index)].
  This ensures that [context_partners] remains valid
  even if the [input_partners] of a [context_unifier_space] are reordered. *)
type context_partners =
    context_partner array


(** a literal of an input clause, with some properties
  and the results of its unification with all context literals. *)
type input_partner = {
  ip_index: int;
  (** the literal is the i.th's literal of the clause
    (the actual order in the datastructure is dynamic during the computation) *)

  ip_literal: literal;
  (** the input literal *)

  ip_vars: Subst.var list;
  (** the variables of the literal *)

  mutable ip_context_partners: context_partner list;
  (** all current context literals (negated) unifiable with this input literal *)

  mutable ip_resolved: context_partner list;
  (** all current (from each other independent) context literals
    resolving this input literal.
    I.e. at most one in chronological backtracking.*)
}

(** represents an input clause and its relationship with the context,
i.e. unifiable literal pairs, ...

  See {!Context_unifier.context_unifier_space}. *)
type 'a _context_unifier_space = {
  cus_clause: clause;
  (** the input clause *)

  cus_shared_vars: Subst.var list;
  (** the variables occuring in several clause literals *)

  cus_input_partners: input_partner array;
  (** the clause literals - may be reordered *)

  mutable cus_subsumed: state list;
  (** subsumed by context literals from these states *)

  cus_process_assert_candidate: ('a -> unit);
  (* to be called on each found assert candidate
     ({!Context_unifier.check_assert} *)

  cus_process_split_candidate: ('a -> unit);
  (* to be called on each found split candidate
     ({!Context_unifier.check_split} *)
}


(** the minimum information needed to recompute a found context unifier.

  See {!Context_unifier.context_unifier_space}. *)
type raw_context_unifier = {
  rcu_space: raw_context_unifier _context_unifier_space;
  (** the clause of the context unifier *)

  rcu_context_partners: context_partners;
  (** the context literals of the context unifier *)
}

(** {!Context_unifier.context_unifier_space} and
  {!Context_unifier.raw_context_unifier}
  are actually mutually recursive types.

  As these are not possible in OCaml
  {!Context_unifier._context_unifier_space}
  is introduced as a trick to break the immediate dependency *)
type context_unifier_space = raw_context_unifier _context_unifier_space



(** raised if {e Close} is triggered.
  Contains the explanation for the closing context unifier. *)
exception CLOSED of explanation



(** {6 Functions} *)


(** {2 Constants} *)

(** used for initialization *)
val null_partner: context_partner

(** used as the gap marker in an 'assert context unifier' *)
val assert_partner: context_partner




  




  

(** {2 Misc} *)

(** is the raw_context_unifier invalid in the current state? *)
val is_raw_context_unifier_invalid: raw_context_unifier -> bool


(** recomputes the substitution of the [raw_context_unifier] *)
val recompute_unifier: raw_context_unifier -> subst

(** [is_remainder context_unifier context_literal_parameters input_literal_index].
  does the context literal
  which contains the parameters [context_literal_parameters]
  is paired with the [input_literal_index].th input literal
  in the [context_unifier]
  produce a remainder literal?
  I.e. one of its parameters is bound to a non-parameter *)
val is_remainder: subst -> var list -> int -> bool

(** computes which literals produced by the context unifier
  are in the remainder.
  The entry at i corresponds to the literal produced with
  the input literal with [ip_index] = i. *)
val compute_remainder_states: raw_context_unifier -> subst -> bool array

(** depth of the tree of the context partners
  needed to computed this context unifier *)
val generation_of_candidate: context_partners -> int

(** the explanation for the used context literals *)
val explanation_of_context_partners: context_partners -> explanation

(** the state in which this unifier has been computed *)
val get_creation_state: raw_context_unifier -> state




(** {2 Representation} *)

val raw_context_unifier_to_string: raw_context_unifier -> string

val context_unifier_to_string: context_unifier_space -> context_partners -> string

val unit_clause_context_unifier_to_string: literal -> string
