(* $Id: context_unifier_check.ml,v 1.2 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type var = Var.var
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type state = State.state
type explanation = State.explanation
type packed_subst = Unification.packed_subst
type context_unifier_space = Context_unifier.context_unifier_space
type context_partners = Context_unifier.context_partners
type raw_context_unifier = Context_unifier.raw_context_unifier



(*** checks ***)


let check_close
  (subst: subst)
  (context_unifier_space: context_unifier_space)
  (context_partners: context_partners)
  : unit =

  let rec empty_remainder (i: int) : bool =
    if i >= Array.length context_partners then
      true
    else begin
      let input_index =
	context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_index
      in
      let context_partner =
	context_partners.(input_index)
      in
	if
	  Context_unifier.is_remainder 
	    subst 
	    context_partner.Context_unifier.cp_element.Context.el_pars
	    input_index
	then
	  false
	else
	  empty_remainder (i + 1)
    end
  in

    if empty_remainder 0 then begin
      Context.update_max_size ();

      let explanation =
	Context_unifier.explanation_of_context_partners context_partners
      in
	if Config.get_show_derivation () then begin
	  print_string (
	    State.current_state_to_string ()
	    ^ "Close:\n"
	    ^
	    if Config.get_show_context_unifier () then
	      Context_unifier.context_unifier_to_string context_unifier_space context_partners
	    else
	      ""
	  )
	end;
	
	if Config.get_derivation_to_dot_file () then begin
	  if Config.get_show_context_unifier () then
	    Derivation_to_dot.print_close_detailed
	      (Context_unifier.context_unifier_to_string context_unifier_space context_partners)
	  else
	    Derivation_to_dot.print_close ()
	end;
	
	Statistic.inc_close ();
	raise (Context_unifier.CLOSED explanation)
    end



let check_assert
  (subst: subst)
  (context_unifier_space: context_unifier_space)
  (context_partners: context_partners) :
  raw_context_unifier option =

  (* the assert literal must be parameter free *)
  let is_assert_valid_at (i : int) : bool =
    let assert_vars =
      Subst.get_bound_vars2
	subst
	context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_vars
    in
      List.for_all 
	(fun var ->
	   Var.is_universal var.Subst.sv_var
	)
	assert_vars
  in

  (* has the unifier an empty remainder
     and leaves the assert literal parameter free? *)
  let rec is_valid_at (i: int) (assert_index: int) : bool * int =
    if i >= Array.length context_partners then
      true, assert_index
    else
      let input_index =
	context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_index
      in
      let context_partner =
	context_partners.(input_index)
      in
	if context_partner == Context_unifier.assert_partner then begin
	  (* no parameter in the assert literal? *)
	  if is_assert_valid_at i then
	    is_valid_at (i + 1) i
	  else
	    false, assert_index
	end

	else begin
	  (* empty remainder? *)
	  if
	    Context_unifier.is_remainder 
	      subst 
	      context_partner.Context_unifier.cp_element.Context.el_pars
	      input_index
	  then
	    false, assert_index
	  else
	    is_valid_at (i + 1) assert_index
	end
  in

    match is_valid_at 0 (-1) with
      | false, _ ->
	  None

      | true, assert_index ->
	  let depth =
	    Term_quality.depth_of_literal_subst
	      subst
	      context_unifier_space.Context_unifier.cus_input_partners.(assert_index).Context_unifier.ip_literal
	      Subst.input_literal_offset
	  in
	    (* drop candidates at depth bound *)
	    if Config.check_depth_bound depth then begin
	      let raw_context_unifier = {
		Context_unifier.rcu_space = context_unifier_space;
		Context_unifier.rcu_context_partners = Array.copy context_partners;
	      }
	      in
		Some raw_context_unifier
	    end

	    else
	      None


let check_split
  (subst: subst)
  (context_unifier_space: context_unifier_space)
  (context_partners: context_partners)
  : raw_context_unifier option =


  let rec depth_at (i: int) (acc: int option) : int option =
    if i >= Array.length context_partners then
      acc
    else
      let input_index =
	context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_index
      in
      let context_partner =
	context_partners.(input_index)
      in
      let remainder =
	Context_unifier.is_remainder 
	  subst
	  context_partner.Context_unifier.cp_element.Context.el_pars
	  input_index
      in
	if remainder then begin
	  let new_depth =
	    Term_quality.depth_of_literal_subst
	      subst
	      context_unifier_space.Context_unifier.cus_input_partners.(i).Context_unifier.ip_literal
	      Subst.input_literal_offset
	  in
	    match acc with
	      | None ->
		  depth_at (i + 1) (Some new_depth)
		    
	      | Some old_depth ->
		  if old_depth > new_depth  then
		    depth_at (i + 1) (Some new_depth)
		  else
		    depth_at (i + 1) acc
	end

	else
	  depth_at (i + 1) acc
  in

  let min_depth =
    match depth_at 0 None with
      | None ->
	  failwith "Context_unifier.check_split"
	  
      | Some depth ->
	  depth
  in

    (* drop candidates at depth bound *)
    if Config.check_depth_bound min_depth then
      let raw_context_unifier = {
	Context_unifier.rcu_space = context_unifier_space;
	(* the old array is reused and altered, so copy it *)
	Context_unifier.rcu_context_partners = Array.copy context_partners;
      }
      in
	Some raw_context_unifier

    else
      None


