(* $Id: context_unifier_space.ml,v 1.34 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type var =  Var.var
type literal =  Term.literal
type clause =  Term.clause
type subst =  Subst.subst
type explanation = State.explanation
type context_unifier_space = Context_unifier.context_unifier_space
type context_partners = Context_unifier.context_partners
type context_partner = Context_unifier.context_partner
type input_partner = Context_unifier.input_partner
type raw_context_unifier = Context_unifier.raw_context_unifier



exception CLOSED = Context_unifier.CLOSED




(*** process completed context unifiers ***)

    


(* a context unifier is completed, now check it for Close, Assert, Split.
   writes the result to candidates *)
let process_context_unifier
  (context_unifier_space: context_unifier_space)
  (context_partners: context_partners)
  (subst: subst)
  (assert_unifier: bool)
  (non_empty_remainder: bool)
  : unit =

  if assert_unifier then begin
    (* an assert context unifier *)
    match Context_unifier_check.check_assert subst context_unifier_space context_partners with
      | None ->
	  ()
	    
      | Some candidate ->
	  Statistic.inc_computed_assert_candidates ();
	  context_unifier_space.Context_unifier.cus_process_assert_candidate candidate;
  end

  else begin
    (* a split or close context unifier *)

    (* first check for close *)
    begin
      (* if the remainder is known to be not empty, close is not possible *)
      if not non_empty_remainder then
	Context_unifier_check.check_close subst context_unifier_space context_partners
    end;

    begin
      match Context_unifier_check.check_split subst context_unifier_space context_partners with
	| None ->
	    (* dropped at depth bound *)
	    ()
	      
	| Some candidate ->
	    Statistic.inc_computed_split_candidates ();
	    context_unifier_space.Context_unifier.cus_process_split_candidate candidate;
    end;
  end







(*** search for context unifiers ***)


(* to avoid extending partial context unifiers with non-empty remainders
   when only assert and close unifiers are targeted,
   the remainder state of every extending context literal is checked for.

   if it is producing a remainder literal,
   the context unifier has a non-empty remainder.
   if it is not producing a remainder literal,
   nothing is known, as another later used context literal might change this.

   but despite of this imperfectness
   even this partial pruning of context unifiers pays off *)
let update_non_empty_remainder (non_empty_remainder: bool)
  (subst: subst) (context_partner: context_partner) (input_partner: input_partner)
  : bool =

  (* Horn never produces non-empty remainders *)
  not (Config.get_horn ())
  &&
  (
    non_empty_remainder
    ||
    Context_unifier.is_remainder 
     subst
     context_partner.Context_unifier.cp_element.Context.el_pars
     input_partner.Context_unifier.ip_index
  )



(* is the potential context_partner compacted when seen from the new context partner? *)
let is_compacted (context_partner: context_partner) : bool =
  if Config.get_use_chronological_backtracking () then begin
    match context_partner.Context_unifier.cp_element.Context.el_compacted with
      | [] ->
	  false
	    
      | _ ->
	  true
  end
  else
    List.exists
      (fun compacted_at ->
	 State.depends_on
	   (State.current_state ())
           compacted_at
      )
      context_partner.Context_unifier.cp_element.Context.el_compacted






(* build all possible unifiers where one input literal is fixed (fixed_index)
   to the new context literal.

   the literals are considered from 0 to n.
   - current_index: try to pair this input literal with a context literal at this step
   - fixed_index: ignored, the initially paired input literal
   - context_partners: the current pairings of context literal to clause literals
   - subst: the partial context unifier
   - candidates: the already computed split and assert candidates, new ones are added
   - assert_unifier: this is an assert unifier, i.e. one literal is not paired with a context literal
   - non_empty_remainder: the partial context unifier produces at least one remainder literal
     (false does not imply that the partial context unifier does _not_ contain remainder literals
 *)
let rec assemble_context_unifier
  (context_unifier_space: context_unifier_space)
  (current_index: int) (fixed_index: int)
  (context_partners: context_partners)
  (subst: subst)
  (assert_unifier: bool) (non_empty_remainder: bool)
  : unit =

  (* the context unifier is complete *)
  if current_index >= Array.length context_unifier_space.Context_unifier.cus_input_partners then begin
    process_context_unifier
      context_unifier_space
      context_partners
      subst
      assert_unifier
      non_empty_remainder
  end

  (* ignore the fixed literal - it has been matched in the initialization *)
  else if current_index = fixed_index then begin
    assemble_context_unifier
      context_unifier_space
      (current_index + 1)
      fixed_index
      context_partners
      subst
      assert_unifier
      non_empty_remainder
  end

  else begin
    (* extend from here to an assert context unifier *)
    assemble_assert
      context_unifier_space
      current_index
      fixed_index
      context_partners
      subst
      assert_unifier
      non_empty_remainder;
    
    (* and also extend from here to a split context unifier *)
    assemble_split
      context_unifier_space
      current_index
      fixed_index
      context_partners
      subst
      assert_unifier
      non_empty_remainder
  end


and assemble_assert
  (context_unifier_space: context_unifier_space)
  (current_index: int)
  (fixed_index: int)
  (context_partners: context_partners)
  (subst: subst)
  (assert_unifier: bool) (non_empty_remainder: bool)
  : unit =

  if not (
    (* this is already an assert unifier *)
    assert_unifier
    ||
    (* no point in trying to build an assert unifier with a non empty remainder *)
    non_empty_remainder
  )
  then begin
    let input_partner =
      context_unifier_space.Context_unifier.cus_input_partners.(current_index)
    in
      (* mark this literal as the assert candidate *)
      context_partners.(input_partner.Context_unifier.ip_index) <-
        Context_unifier.assert_partner;
      
      assemble_context_unifier
	context_unifier_space
	(current_index + 1)
	fixed_index
	context_partners
	subst
	true
	non_empty_remainder
  end

(* checks if the next literal to choose is subsumed.
   raiss Not_found otherwise *)
and assemble_resolved_split
  (context_unifier_space: context_unifier_space)
  (current_index: int) (fixed_index: int)
  (context_partners: context_partners)
  (subst: subst)
  (assert_unifier: bool) (non_empty_remainder: bool)
  : unit =
  
  let input_partner =
    context_unifier_space.Context_unifier.cus_input_partners.(current_index)
  in

  let resolving_partner =
    (* is this input literal resolved? *)
    if Config.get_use_chronological_backtracking () then begin
      match input_partner.Context_unifier.ip_resolved with
	| [] ->
	    raise Not_found
	    
	| resolving_partner :: _ ->
	    resolving_partner
    end

    else begin
      List.find
	(fun resolve_partner ->
	   State.depends_on
	   (State.current_state ())
	   resolve_partner.Context_unifier.cp_element.Context.el_state
	)
	input_partner.Context_unifier.ip_resolved
    end
  in
    (* yes resolved, so use the resolving literal *)

    (* this implies in chronological backtracking that all following literals
       are also resolve.
       if context_partners is already set to this resolved literal,
       we already passed here when completing another context unifier.
       And thus all remaining literals are also already set in context_partners. *)
    if Config.get_use_chronological_backtracking ()
      &&
      context_partners.(input_partner.Context_unifier.ip_index) == resolving_partner
    then begin
      (* check completed context unifier *)
      process_context_unifier
	context_unifier_space
	context_partners
	subst
	assert_unifier
	non_empty_remainder
    end

    else begin
      context_partners.(input_partner.Context_unifier.ip_index) <- resolving_partner;
      
      (* as this substitution is only used to check for close
	 and is recomputed and made admissible for productivity, subsumption, ...
	 the resolving substitution can be ignored *)
      assemble_context_unifier
	context_unifier_space
	(current_index + 1)
	fixed_index
	context_partners
	subst
	assert_unifier
	non_empty_remainder
    end

(* try all context literals of the selected queue 
   in order to extend the current context unifier *)
and assemble_split
  (context_unifier_space: context_unifier_space)
  (current_index: int) (fixed_index: int)
  (context_partners: context_partners)
  (subst: subst)
  (assert_unifier: bool) (non_empty_remainder: bool)
  : unit =
  
  try
    assemble_resolved_split
      context_unifier_space
      current_index
      fixed_index
      context_partners
      subst
      assert_unifier
      non_empty_remainder
  with
    | Not_found ->
	let input_partner =
	  context_unifier_space.Context_unifier.cus_input_partners.(current_index)
	in
	  List.iter
	    (fun (context_partner: context_partner) ->

	       (* ignore compacted context literals *)
	       if is_compacted context_partner then
		 ()

	       else begin
		 (* try to extend the unifier *)
		 try
		   let extended_subst =
		     Unification.unify_packed_substs subst context_partner.Context_unifier.cp_subst
		   in

		   let new_non_empty_remainder =
		     update_non_empty_remainder
		       non_empty_remainder
		       extended_subst
		       context_partner
		       input_partner
		   in
		     
		     if
		       (* a non empty remainder is not interesting for assert *)
		       assert_unifier
		       &&
		       new_non_empty_remainder
		     then
		       ()

		     else begin
		       context_partners.(input_partner.Context_unifier.ip_index) <- context_partner;
		       
		       assemble_context_unifier
			 context_unifier_space
			 (current_index + 1)
			 fixed_index
			 context_partners
			 extended_subst
			 assert_unifier
			 new_non_empty_remainder
		     end;
		     
		 with
		   | Unification.UNIFICATION_FAIL ->
		       (* unusable context literal *)
		       ()
	       end;
	    )
	    input_partner.Context_unifier.ip_context_partners



(* searchs all context unifiers and adds the new candidates to candidates.
   one input_partner may be exempted from the search by setting fixed_index to its index
   and initializing partial_context_unifier and subst accordingly *)
let search_context_unifiers
  (context_unifier_space: context_unifier_space)
  (fixed_index: int)
  (partial_context_unifier: context_partners)
  (subst: subst)
  (non_empty_remainder: bool)
  : unit =

  assemble_context_unifier
    context_unifier_space
    0
    fixed_index
    partial_context_unifier
    subst
    false
    non_empty_remainder







(*** resolve ***)


(* is this input literal already resolved
   with respect to the new context element? *)
let is_resolved
  (input_partner: input_partner) (context_element: Context.element) : bool =
  if Config.get_use_chronological_backtracking () then begin
    match input_partner.Context_unifier.ip_resolved with
      | [] ->
	  false
	    
      | _ ->
	  true
  end
  else begin
    List.exists
      (fun resolve_partner ->
	 State.depends_on
	   context_element.Context.el_state
	   resolve_partner.Context_unifier.cp_element.Context.el_state
      )
      input_partner.Context_unifier.ip_resolved
  end


(*** subsume ***)

(* yes, this special treatment for chronological backtracking is significantly faster *)
let is_already_subsumed (context_unifier_space: context_unifier_space) (context_element: Context.element) : bool =
  Config.get_use_subsume ()
  &&
  (
    if Config.get_use_chronological_backtracking () then begin
      match context_unifier_space.Context_unifier.cus_subsumed with
	| [] ->
	    false
	      
	| _ ->
	    true
    end
    else
      List.exists
	(State.depends_on context_element.Context.el_state)
	context_unifier_space.Context_unifier.cus_subsumed
  )


let mark_subsumed context_unifier_space state =
  match context_unifier_space.Context_unifier.cus_subsumed with
    | top_state :: _
	when State.state_equal top_state state ->
	()

    | _ ->
	context_unifier_space.Context_unifier.cus_subsumed <-
          state :: context_unifier_space.Context_unifier.cus_subsumed



(*** reorder ***)

(* sort input literals by increasing number of context partners.
   keep resolved literals at the end.
   thus less partial context unifier extending steps need to be done,
   as the first considered literals have fewer context partners
   and thus branching is lower in the top part of the search tree.

   this information is used assemble_resolved_split
   when literals are check for beeing resolved. *)


(* shift the resolved literal to the end *)
let move_resolved_input_literal (context_unifier_space: context_unifier_space) (index: int) : unit =
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in
  let resolved =
    input_partners.(index)
  in
    for i = index + 1 to Array.length input_partners - 1 do
      input_partners.(i - 1) <- input_partners.(i);
    done;
    input_partners.(Array.length input_partners - 1) <- resolved



(* shift the literal with the increased queue
   behind unresolved literals with a shorter queue *)
let move_increased_input_literal (context_unifier_space: context_unifier_space) (index: int) : unit =
  let input_partners =
    context_unifier_space.Context_unifier.cus_input_partners
  in

  let rec swap_with (i : int) : unit =
    (* end of queue *)
    if i >= Array.length input_partners then
      ()

    (* only resolved literal ahead *)
    else if
      match input_partners.(i).Context_unifier.ip_resolved with
	| [] ->
	    (* unresolved, so continue *)
	    false

	| _ ->
	    (* resolved, so stop here *)
	    true
    then
      ()

    (* the next literals queue is smaller, so swap *)
    else if
      List.length input_partners.(i).Context_unifier.ip_context_partners
      <
      List.length input_partners.(i - 1).Context_unifier.ip_context_partners
    then begin
      let swap =
	input_partners.(i - 1)
      in
	input_partners.(i - 1) <- input_partners.(i);
	input_partners.(i) <- swap;

	swap_with (i + 1)
    end
  in
    swap_with (index + 1)


let move_updated_input_literal
  (context_unifier_space: context_unifier_space) (index: int)
  (input_partner: input_partner) (context_partner: context_partner)
  : unit =
  if
    Config.get_use_chronological_backtracking ()
    &&
    is_resolved input_partner context_partner.Context_unifier.cp_element
  then
    move_resolved_input_literal context_unifier_space index
  else
    move_increased_input_literal context_unifier_space index









(*** add a literal ***)


(* pair the new context literal with the clause literals one by one
   and compute all new context unifiers. *)
let add_to_clause
  (context_unifier_space: context_unifier_space)
  (input_partner: input_partner)
  (context_partner: context_partner)
  (subst: subst)
  : unit =

  if is_already_subsumed context_unifier_space context_partner.Context_unifier.cp_element then begin
    ()
  end

  else begin
    (* create the inital partial context unifier *)
    let partial_context_unifier =
      Array.make
	(Array.length context_unifier_space.Context_unifier.cus_input_partners)
	Context_unifier.null_partner
    in
      
      partial_context_unifier.(input_partner.Context_unifier.ip_index) <-
      context_partner;
      
      let rec find_i i =
	if i > Array.length context_unifier_space.Context_unifier.cus_input_partners then
	  failwith "find_i"
	else if input_partner == context_unifier_space.Context_unifier.cus_input_partners.(i) then
	  i
	else
	  find_i (i + 1)
      in
      let i =
	find_i 0
      in
		    
	(* compute new candidates *)
	search_context_unifiers
	  context_unifier_space
	  i
	  partial_context_unifier
	  subst
	  (update_non_empty_remainder false subst context_partner input_partner);
		  
	move_updated_input_literal context_unifier_space i input_partner context_partner;
  end



(* checks if the new context literal is contradictory with this unit clause *)
let add_to_unit_clause
  (context_unifier_space: context_unifier_space)
  (context_element: Context.element)
  : unit =

  (* already subsumed? *)
  if is_already_subsumed context_unifier_space context_element then
    ()

  (* now subsumed? *)
  else if
    Config.get_use_subsume ()
    &&
    begin
      try
	ignore (Unification.match_literals
		  ~p_preserving:true
		  context_element.Context.el_literal
		  0
		  context_unifier_space.Context_unifier.cus_input_partners.(0).Context_unifier.ip_literal
		  1);
	mark_subsumed context_unifier_space (State.current_state ());
	Statistic.inc_subsume ();
	true
      with
	| Unification.UNIFICATION_FAIL -> 
	    false
    end
  then
    ()

  (* contradictory? *)
  else begin
    let context_literal =
      context_element.Context.el_literal
    in
    let input_literal =
      context_unifier_space.Context_unifier.cus_input_partners.(0)
    in
    let clause_literal =
      input_literal.Context_unifier.ip_literal
    in
      if context_literal.Term.sign = clause_literal.Term.sign then begin
	()
      end

      else begin
	try
	  (* are the clause and the context literal contradictory? *)
	  let subst =
	    Unification.unify_terms
	      ~p_preserving:true
	      clause_literal.Term.term
	      0
	      context_literal.Term.term
	      1
	  in
	    
	    (* yes *)
	    Context.update_max_size ();
	    
	    let explanation =
	      State.get_explanation [context_element.Context.el_literal]
	    in
	      
	      if Config.get_show_derivation () then begin
		print_string (
		  State.current_state_to_string ()
		  ^ "Close:\n"
		  ^
		  if Config.get_show_context_unifier () then
		    let context_partner = {
		      Context_unifier.cp_element = context_element;
		      Context_unifier.cp_subst = Array.of_list subst;
		    }
		    in
		      Context_unifier.context_unifier_to_string
			context_unifier_space
			[| context_partner |]
		  else
		    ""
		)
	      end;
	      
	      if Config.get_derivation_to_dot_file () then begin
		if Config.get_show_context_unifier () then
		  let context_partner = {
		    Context_unifier.cp_element = context_element;
		    Context_unifier.cp_subst = Array.of_list subst;
		  }
		  in
		    Derivation_to_dot.print_close_detailed
		      (Context_unifier.context_unifier_to_string
			 context_unifier_space
			 [| context_partner |])
		else
		  Derivation_to_dot.print_close ()
	      end;
	      
	      Statistic.inc_close ();
	      raise (CLOSED explanation);

	with
	  | Unification.UNIFICATION_FAIL ->
	      ()
      end
  end

      







(*** creation ***)


(* convert the clause into queues. *)
let clause_to_input_partners (clause: clause) : input_partner array =
  let input_partners =
    Array.mapi
      (fun index literal ->
	 Problem_literals.register_literal literal index
      )
      (Array.of_list clause)
  in

    Array.sort
      (fun queue1 queue2 ->
	 compare
	 (List.length queue1.Context_unifier.ip_context_partners)
	 (List.length queue2.Context_unifier.ip_context_partners)
      ) 
      input_partners;

    input_partners



let create_clause
  (process_assert_candidate: raw_context_unifier -> unit)
  (process_split_candidate: raw_context_unifier -> unit)
  (clause: clause)
  : (context_unifier_space * literal option) =

  let input_partners =
    clause_to_input_partners clause
  in

  let input_vars =
    Array.map
      (fun partner -> partner.Context_unifier.ip_vars)
      input_partners
  in

  let shared_subst_vars =
    Tools.lists_shared
      Subst.var_equal
      (Array.to_list input_vars)
  in

  let context_unifier_space = {
    Context_unifier.cus_clause = clause;
    Context_unifier.cus_shared_vars = shared_subst_vars;
    Context_unifier.cus_input_partners = input_partners;
    Context_unifier.cus_subsumed = [];
    Context_unifier.cus_process_assert_candidate = process_assert_candidate;
    Context_unifier.cus_process_split_candidate = process_split_candidate;
  }
  in

  let do_add_to_clause =
    add_to_clause context_unifier_space
  and do_mark_subsumed =
    mark_subsumed context_unifier_space
  in
    Array.iter
      (fun input_partner ->
	 Problem_literals.register_clause_literal
	 input_partner.Context_unifier.ip_literal
	 input_partner.Context_unifier.ip_index
	 do_add_to_clause
	 do_mark_subsumed
      )
      input_partners;

  let context_partners =
    Array.make
      (Array.length context_unifier_space.Context_unifier.cus_input_partners)
      Context_unifier.null_partner
  in
    search_context_unifiers
      context_unifier_space
      (-1)
      context_partners
      Subst.empty
      false;
    
    (context_unifier_space, None)








(* a unit clause must be asserted and checked for contradictory against every context literal.
   but no context unifiers need to be computed, thus no queue is needed *)
let create_unit_clause
  (process_assert_candidate: raw_context_unifier -> unit)
  (process_split_candidate: raw_context_unifier -> unit)
  (literal: literal)
  : (context_unifier_space * literal option) =

  let vars =
    List.map
      (fun var ->
	 Subst.make_var var Subst.input_literal_offset
      )
      (Term.vars_of_literal literal)
  in

  let input_partner = {
      Context_unifier.ip_literal = literal;
      Context_unifier.ip_vars = vars;
      Context_unifier.ip_index = 0;
      Context_unifier.ip_context_partners = [];
      Context_unifier.ip_resolved = [];
    };
  in

  let context_unifier_space =
    {
      Context_unifier.cus_clause = [literal];
      Context_unifier.cus_shared_vars = [];
      Context_unifier.cus_input_partners = [| input_partner |];
      Context_unifier.cus_subsumed = [];
      Context_unifier.cus_process_assert_candidate = process_assert_candidate;
      Context_unifier.cus_process_split_candidate = process_split_candidate;
    }
  in
    Problem_literals.register_unit_literal
      (add_to_unit_clause context_unifier_space);

    Statistic.inc_computed_assert_candidates ();
    
    (context_unifier_space, Some literal)



(* e.g. false :- true. in TPTP problem SWC128-1 *)
let create_empty_clause
  (process_assert_candidate: raw_context_unifier -> unit)
  (process_split_candidate: raw_context_unifier -> unit)
  : (context_unifier_space * literal option) =
  {
    Context_unifier.cus_clause = [];
    Context_unifier.cus_shared_vars = [];
    Context_unifier.cus_input_partners = [| |];
    Context_unifier.cus_subsumed = [];
    Context_unifier.cus_process_assert_candidate = process_assert_candidate;
    Context_unifier.cus_process_split_candidate = process_split_candidate;
  },
  None


(* creates a context_unifier_space from a clause and returns the initial candidates,
   i.e. a split context unifier computed by using -v *)    
let create
  (clause: clause)
  (process_assert_candidate: raw_context_unifier -> unit)
  (process_split_candidate: raw_context_unifier -> unit)
  : (context_unifier_space * literal option) =
  match clause with
    | [] ->
	(* e.g. false :- true. in TPTP problem SWC128-1 *)
	create_empty_clause process_assert_candidate process_split_candidate

    | literal :: [] ->
	create_unit_clause process_assert_candidate process_split_candidate literal

    | _ ->
	create_clause process_assert_candidate process_split_candidate clause








(*** backtracking ***)

let backtrack_subsumed (context_unifier_space: context_unifier_space) : unit =
  context_unifier_space.Context_unifier.cus_subsumed <-
    List.find_all State.is_state_valid context_unifier_space.Context_unifier.cus_subsumed


(* sort the newly arranged queues by length. small queues first, resolved queues at the end *)
let backtrack_sort (context_unifier_space: context_unifier_space) : unit =
  if Config.get_use_chronological_backtracking () then begin
    Array.sort
      (fun queue1 queue2 ->
	 match queue1.Context_unifier.ip_resolved, queue2.Context_unifier.ip_resolved with
	   | [], [] ->
	       compare
	       (List.length queue1.Context_unifier.ip_context_partners)
	       (List.length queue2.Context_unifier.ip_context_partners)
	       
	   | [], _ ->
	       -1

	   | _, [] ->
	       1

	   | _ ->
	       0
      )
      context_unifier_space.Context_unifier.cus_input_partners
  end
  else begin
    Array.sort
      (fun queue1 queue2 ->
	 compare
	 (List.length queue1.Context_unifier.ip_context_partners)
	 (List.length queue2.Context_unifier.ip_context_partners)
      )
      context_unifier_space.Context_unifier.cus_input_partners
  end


let backtrack (context_unifier_space: context_unifier_space) : unit =
  backtrack_subsumed context_unifier_space;
  backtrack_sort context_unifier_space
