(* $Id: darwin.ml,v 1.47 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(* Sketch of the algorithm:
   i) add candidate to context
   ii) if this triggers a Close:
     - backtrack to a left split state and right split on it
     - remove all now invalidated information
   iii) compute new assert and split candidates
        based on the new context literal
   iv) select a new candidate, asserts preferred to splits
   v) if it's a split candidate then create a new state
   vi) continue with i)
*)


(*** types ***)

type state = State.state
type literal = Term.literal
type clause = Term.clause
type explanation = State.explanation
type selection = Selection.selection
type selected_split = Selection_types.selected_split


(* the information for the right split on a split_decision *)
type right_split = {
  rs_left_closing_explanation: explanation;
  (* the left split was closed with this explanation *)

  rs_skolemized: literal;
  (* the skolemization of the split literal *)

  rs_literal: literal;
  (* the negation of the split literal *)
}


(* a split - first left, the possibly right *)
type split_decision = {
  sd_selected: selected_split;
  (* the split literal splitted on *)

  mutable sd_right_split: right_split option;
  (* None -> this is a left split
     Some _ -> this is a right split *)
}


exception CLOSED = Context_unifier.CLOSED
exception SATISFIABLE = Selection.SATISFIABLE
exception NO_SOLUTION = Config.NO_SOLUTION




(*** globals ***)


let (split_decisions: split_decision list ref) =
  ref []





(*** representation ***)

let split_decision_to_string (split_decision: split_decision) : string =
  match split_decision.sd_right_split with
    | None ->
	"Left Split at: " ^ State.state_to_string !(split_decision.sd_selected.Selection_types.ss_valid)
	^ "\n" 
	^ "Literal: " ^ Term.literal_to_string split_decision.sd_selected.Selection_types.ss_literal
	^ "\n" 

    | Some right_split ->
	"Right Split at: " ^ State.state_to_string !(split_decision.sd_selected.Selection_types.ss_valid)
	^ "\n" 
	^ "Literal: " ^ Term.literal_to_string right_split.rs_literal
	^ "\n" 
	^ "Left Explanation: " ^ State.explanation_to_string right_split.rs_left_closing_explanation
	^ "\n" 


let split_decisions_to_string () : string =
  String.concat
    ("--------------\n")
    (List.map split_decision_to_string !split_decisions)






(*** add ***)


let add (selection: selection) (literal: literal)
  (mandatory_at: state option) (generation: int) : unit =

  let context_element =
    Context.add literal mandatory_at generation
  in
    Selection.add selection context_element





(*** backtrack ***)

(* backtrack over the split decisions *)
let backtrack_decisions (_explanation: explanation) : split_decision * explanation =
  
  let rec do_backtrack_decisions
      remaining_decisions processed_decisions explanation =

    match remaining_decisions with
      | [] ->
	  split_decisions := List.rev processed_decisions;
	  raise (CLOSED explanation)

      | decision :: tail ->
	  begin
	    let (new_explanation, contained) =
	      State.prune_up_to_state explanation !(decision.sd_selected.Selection_types.ss_valid)
	    in
	      match decision.sd_right_split with
		| None ->
		    (* a left split *)
		    if contained then begin
		      (* do a right split *)
		      State.invalidate_state !(decision.sd_selected.Selection_types.ss_valid);
		      split_decisions := decision :: List.rev processed_decisions @ tail;

		      decision, new_explanation
		    end

		    else begin
		      if Config.debug && State.is_state_invalid !(decision.sd_selected.Selection_types.ss_valid) then begin
			failwith "Darwin.backtrack_states.left_split";
		      end;

		      do_backtrack_decisions tail (decision :: processed_decisions) new_explanation
		    end
		      
		| Some right_split ->
		    (* a right split *)

		    (* add the left closing explanation of the split *)
		    let merged_closing_explanation =
		      State.merge_explanations
			right_split.rs_left_closing_explanation
			new_explanation
		    in

		    (* add the explanation of the split context unifier *)
		    let merged_explanation =
		      State.merge_explanations
			decision.sd_selected.Selection_types.ss_explanation
			merged_closing_explanation
		    in
		      if contained then begin
			(* close this split and all splits depending on it *)
			State.invalidate_state !(decision.sd_selected.Selection_types.ss_valid);
		      end;

		      do_backtrack_decisions tail (decision :: processed_decisions) merged_explanation
	  end
  in
    do_backtrack_decisions !split_decisions [] _explanation




(*** inference rules ***)

(* assert a literal *)
let assert_rule (selection: selection) (selected: Selection_types.selected_assert)
  (is_unit_split: bool) : unit =

  (* add to the current state *)
  State.set_assert
    selected.Selection_types.sa_literal
    selected.Selection_types.sa_explanation;

  let to_string_extended raw_context_unifier_option =
    match raw_context_unifier_option with
      | None ->
	  Context_unifier.unit_clause_context_unifier_to_string selected.Selection_types.sa_literal
	    
      | Some raw_context_unifier ->
	  Context_unifier.raw_context_unifier_to_string raw_context_unifier
	  ^
	  if is_unit_split then
	    Selection_split.remainder_literals_to_string raw_context_unifier
	  else
	    ""
  in
    if Config.get_show_derivation () then begin
      print_endline (
	State.state_to_string (State.current_state ())
	^ 
	if is_unit_split then
	  "Unit Split: " ^ Term.literal_to_string selected.Selection_types.sa_literal
	else
	  "Assert: " ^ Term.literal_to_string selected.Selection_types.sa_literal
      );
      
      if Config.get_show_context_unifier () then begin
	print_string (to_string_extended selected.Selection_types.sa_raw_context_unifier);
      end;
    end;
  
    if Config.get_derivation_to_dot_file () then begin
      if is_unit_split then begin
	if Config.get_show_context_unifier () then
	  Derivation_to_dot.print_unit_split_detailed
	    selected.Selection_types.sa_literal
	    (to_string_extended selected.Selection_types.sa_raw_context_unifier)
	else
	  Derivation_to_dot.print_unit_split
	    selected.Selection_types.sa_literal
      end

      else begin
	if Config.get_show_context_unifier () then
	  Derivation_to_dot.print_assert_detailed
	    selected.Selection_types.sa_literal
	    (to_string_extended selected.Selection_types.sa_raw_context_unifier)
	else
	  Derivation_to_dot.print_assert
	    selected.Selection_types.sa_literal
      end
    end;


  (* add to the context *)
  add
    selection
    selected.Selection_types.sa_literal
    selected.Selection_types.sa_mandatory_at
    selected.Selection_types.sa_generation



(* left split on a split literal *)
let split_left (selection: selection) (selected: Selection_types.selected_split) : unit =
  (* create a new state *)
  let state =
    State.create_state
      selected.Selection_types.ss_literal
      selected.Selection_types.ss_explanation
  in

    selected.Selection_types.ss_valid := state;

  let split_state = {
    sd_selected = selected;
    sd_right_split = None;
  }
  in
    split_decisions := split_state :: !split_decisions;

    (* show the derivation only after the new state has been created *)

    if Config.get_show_derivation () then begin
      print_endline (
	State.state_to_string (State.current_state ())
	^ "Split Left: " ^ Term.literal_to_string selected.Selection_types.ss_literal
      );
      
      if Config.get_show_context_unifier () then begin
	print_string (Context_unifier.raw_context_unifier_to_string selected.Selection_types.ss_raw_context_unifier);
	print_string (Selection_split.remainder_literals_to_string selected.Selection_types.ss_raw_context_unifier);
      end;
    end;
    
    if Config.get_derivation_to_dot_file () then begin
      if Config.get_show_context_unifier () then
      Derivation_to_dot.print_split_left_detailed
	selected.Selection_types.ss_literal
	(Context_unifier.raw_context_unifier_to_string selected.Selection_types.ss_raw_context_unifier
	 ^ Selection_split.remainder_literals_to_string selected.Selection_types.ss_raw_context_unifier)
      else
	Derivation_to_dot.print_split_left
	  selected.Selection_types.ss_literal
    end;

    (* add to the context *)
    add
      selection
      selected.Selection_types.ss_literal
      None
      selected.Selection_types.ss_generation


(* a CLOSED has been triggered -
   so find a left splitted literal do undo and right split on *)
let rec split_right (selection: selection) (explanation: explanation) =
  (* backtrack to a left split responsible for the CLOSED,
     i.e. a left split in explanation *)
  let backtrack_decision, left_explanation =
    backtrack_decisions explanation
  in

    (* except for the decision backtracked to remove all now invalid states *)
    split_decisions :=
    List.find_all
      (fun split_state ->
	 (backtrack_decision == split_state)
	 ||
	 (State.is_state_valid !(split_state.sd_selected.Selection_types.ss_valid))
      )
      !split_decisions;


    (* update the decision to right split *)
    let left_skolemized =
      Term.request_skolemized_literal backtrack_decision.sd_selected.Selection_types.ss_literal
    in
      
    let right_split = {
      rs_left_closing_explanation = left_explanation;
      rs_skolemized = left_skolemized;
      rs_literal = Term.request_negated_literal left_skolemized;
    }
    in
      backtrack_decision.sd_right_split <- Some right_split;
    
      (* first invalidate all backtracked states *)
      State.backtrack ();

      (* create a new state and attach the right split to it -
         otherwise the split candidate would be removed as invalid
         by Selection.backtrack *)
      backtrack_decision.sd_selected.Selection_types.ss_valid :=
        State.create_state
	  right_split.rs_literal
	  backtrack_decision.sd_selected.Selection_types.ss_explanation;
      
      (* backtrack all modules *)
      Context.backtrack ();
      Selection.backtrack selection explanation;


      (* show the derivation only after the new state has been created *)
      if Config.get_show_derivation () then begin
	print_endline (
	  State.current_state_to_string ()
	  ^ "Split Right: " ^ Term.literal_to_string right_split.rs_literal
	);
      end;
	
      if Config.get_derivation_to_dot_file () then begin
	if Config.get_show_context_unifier () then
	  Derivation_to_dot.print_split_right_detailed
	    right_split.rs_literal
	else
	  Derivation_to_dot.print_split_right
	    right_split.rs_literal
      end;	

      try
	(* add to the context *)
	add
	  selection
	  right_split.rs_literal
	  None
	  backtrack_decision.sd_selected.Selection_types.ss_generation
      with
	| CLOSED explanation ->
	    (* the right split immediately closed, try the next right split *)
	    split_right selection explanation





(*** selection ***)

(* endless loop:
   - select a candidate
   - add it to the context
   - backtrack on CLOSED and right split

   loop is exited when
   - CLOSED can not be caught,
     i.e. there is no left split left to switch to a right split
   - SATISFIABLE is raised *)
let rec select (selection: selection) : exn =
  begin
    try
      match Selection.select selection with
	| Selection.Assert selected ->
	    Statistic.inc_assert ();
	    assert_rule selection selected false
	      
	| Selection.Assert_split selected ->
	    Statistic.inc_split ();
	    assert_rule selection selected true
	      
	| Selection.Split selected ->
	    Statistic.inc_split ();
	    split_left selection selected
    with
      | CLOSED explanation ->
	  split_right selection explanation
  end;

  select selection




	  



(*** main loop ***)

(* initialization before every start of the proof procedure,
   i.e. every start and restart *)
let restart (clauses: clause list) : unit =
  (* empty proof tree *)
  split_decisions := [];

  (* just the statistics, no dependencies *)
  Statistic.restart ();

  (* must be called first,
     as other modules use global terms (default_v) when initializing *)
  Term.restart (clauses);

  Selection_lookahead.init ();

  (* uses Term.default_v *)
  State.restart ();

  (* no dependencies *)
  let index_type =
    if Config.get_use_term_indexing () then begin
    
      if Config.get_horn () then
	Context.DT
      else
	Context.ST
    end
    else
      Context.NONE
  in
    Context.init (index_type);

  (* no dependencies - uses Term.default_v, but only after the initialization *)
  Problem_literals.init ()

  

(* start select and check its termination:
   - CLOSED: problem is unsatisfiable
   - SATISFIABLE: problem is satisfiable,
     if no candidates have been dropped because of the depth bound.
     then a restart is necessary *)
let rec start (clauses: clause list) : unit =
  restart (clauses);

  if Config.get_derivation_to_dot_file () then begin
    Derivation_to_dot.init (Config.get_derivation_to_dot_file_name ());
  end;

  if Config.get_show_derivation () then begin
    print_endline (
      State.state_to_string (State.current_state ())
      ^ "Depth bound: " ^ string_of_int (Config.get_current_depth_bound ())
    );
  end;


  try
    ignore (select (Selection.create clauses));
  with
    | SATISFIABLE when
	Config.update_depth_bound () ->
	(* candidates have been dropped because of the depth bound *)
	
	if Config.get_derivation_to_dot_file () then begin
	  Derivation_to_dot.finish ();
	end;
	  
	  start clauses

    | exn ->
	if Config.get_derivation_to_dot_file () then
	  Derivation_to_dot.finish ();

	raise exn








(*** input processing **)


(* add .tme as default extension if the problem file has no extension *)
let check_extension (file_name: string) : string =
  try
    ignore (Filename.chop_extension (Filename.basename file_name));
    file_name
  with
    | Invalid_argument _ ->
	(* if no extension is given assume .tme *)
	file_name ^ ".tme"


(* is the claus Horn? *)
let rec is_horn (clause: literal list) (pos_count: int) : bool =
  match clause with
    | [] ->
	true
	  
    | literal :: tail ->
	if literal.Term.sign then begin
	  if pos_count = 0 then
	    is_horn tail 1
	  else
	    false
	end
	else
	  is_horn tail pos_count

(* are the clauses horn? *)
let are_horn (clauses: clause list) : bool =
  List.for_all
    (fun clause ->
       is_horn clause 0
    )
    clauses

	   

(* read the problem file, check if it's Horn,
   and start the derivation *)

let prove (file_name: string) : unit =
  (* set initial depth bound *)
  Config.reset_depth_bound ();


(*  let file_name =
    check_extension file_name
  in*)
    
  let clauses =
    try
      if Filename.check_suffix file_name ".darwin" then
	Read_darwin.to_clauses_from_file file_name
      else (*if Filename.check_suffix file_name ".tme" then*)
	Read_tme.to_clauses_from_file file_name
(*      else begin
	print_endline ("unsupported input format: " ^ file_name);
	[]
      end*)
    with
      | Sys_error error_message ->
	  print_endline error_message;
	  []
  in
    begin
      (*print_endline ("Proving " ^ file_name ^ " ...");*)
      
      match clauses with
	| [] ->
	    ()

	| _ ->
	    if ((Config.get_use_horn_detection ())
		&&
		(are_horn clauses))
	    then
	      Config.set_horn ();
	    
	    
	    if Config.get_show_derivation () then
	      print_endline ("START OF DERIVATION");
	    
	    begin
	      try
		start clauses;
	      with
		| SATISFIABLE ->
		    if Config.get_show_derivation () then
		      print_endline ("END OF DERIVATION\n");
		    
		    print_endline ("SATISFIABLE");
		    
		    if Config.get_show_model () then begin
		      print_endline ("\nSTART OF CONTEXT ('MODEL'):");
		      Context.iter (fun context_element -> print_endline (Term.literal_to_string context_element.Context.el_literal));
		      print_endline ("END OF CONTEXT");
		    end;
		    
		    if Config.get_print_model_to_file () <> "" then begin
		      try
			Context.to_file (Config.get_print_model_to_file ());
		      with 
			| exn ->
			    print_endline ("Couldn't print context to file "
					   ^ Config.get_print_model_to_file ()
					 ^ ": " ^ Printexc.to_string exn);
		    end;
		    
		    
		| CLOSED _ ->
		    if Config.get_show_derivation () then
		      print_endline ("END OF DERIVATION\n");
		    
		    print_endline ("REFUTATION FOUND");
		    
		| Counter.OVERFLOW
		| NO_SOLUTION ->
		    (* Note: this case never happened in current tests,
		       but a candidate overflow is bound to happen for longer run times. *)
		    if Config.get_show_derivation () then
		      print_endline ("END OF DERIVATIONx\n");
		    
		    print_endline ("NO SOLUTION");
	    end;
	    

	    if Config.get_show_options () then begin
	      print_newline ();
	      print_string (Config.to_string ());
	    end;
	    
	    if Config.get_show_statistic () then begin
	      print_newline ();
	      print_string (Statistic.to_string ());
	    end
    end;
    (* trying several proofs in a row is broken,
       probably due to kept/recreated symbols/literals. *)
    exit 0





let _ =
  if Config.get_problem_file_name () = "" then begin
    (* to make debugging/interactive toplevel possible
       where no file is given *)
    print_endline "No input file specified. try darwin -help.";
    ()
  end
  else begin
    let file_name =
      Config.get_problem_file_name ()
    in
      prove file_name;
  end      
