(* $Id: derivation_to_dot.ml,v 1.2 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type state = State.state
type literal = Term.literal





(*** globals ***)

(* the channel to write the derivation to.
   has to be opened/closed with init/finish *)
let (out_channel: out_channel option ref) =
  ref None





(*** node ids ***)

(* the graph nodes are just a fixed name plus the id counter as a suffix. *)

let node_id =
  Counter.create_with 0

let node_name (id: int) =
  "node_" ^ string_of_int id

let current_node_name () =
  node_name (Counter.value node_id)




(*** split state -> node id mapping ***)

(* maps the state left by a split to the if of the node created by the split *)

let state_to_id =
  Hashtbl.create 64

let store_state_id () : unit =
  Hashtbl.add
    state_to_id
    (State.parent_state (State.current_state ()))
    (Counter.value node_id)

let get_state_id (state: state) : int =
  try
    Hashtbl.find state_to_id state
  with
    | Not_found ->
	failwith "get_state_id"
  







(*** color definitions ***)

let yellow = "\"#ffffa0\""
let green = "\"#b0ffb0\""
let blue = "\"#b0e0ff\""
let red = "\"#ffb0b0\""





(*** printing ***)

let print_line (line: string) : unit =
  match !out_channel with
    | None ->
	()

    | Some out_channel ->
	output_string out_channel line;
	output_string out_channel "\n"



let print_header () =
  (* start graph *)
  print_line "digraph proof {";

  (* node defaults *)
  print_line "node [fontsize=\"10\", fontname=\"Arial\"]";
  print_line "node [shape=box, style = filled]";
  print_line "node [label=\"\", fillcolor = black]";

  (* edge defaults *)
  print_line "edge [fontsize=\"10\", fontname=\"Arial\"]"


let print_footer () =
  (* end graph *)
  print_line "}"





(*** print node ***)

(* sets the attributes of the current node *)
let print_node (label: string) (color: string) =
  print_line (
    current_node_name ()
    ^ " [label = \"" ^ String.escaped label ^ "\", fillcolor = "
    ^ color ^ "]")

let print_assert_node (label: string) =
  print_node label blue

let print_unit_split_node (label: string) =
  print_node label yellow

let print_split_node (label: string) =
  print_node label green

let print_close_node (label: string) =
  print_node label red







(*** print edge ***)


(* prints an edge to a (implicitely created) new node *)
let print_edge (from_node: string) (label: string) =
  Counter.inc node_id;

  print_line (
    from_node
    ^ " -> "
    ^ (current_node_name ())
    ^ " [label = \"" ^ String.escaped label ^ "\"]"
  )


let print_edge_from_current_node (label: string) : unit =
  print_edge
    (current_node_name ())
    label







(*** compact rule printing ***)

let print_assert (literal: literal) : unit =
  print_assert_node (Term.literal_to_string literal);
  print_edge_from_current_node ""
    
let print_unit_split (literal: literal) : unit =
  print_unit_split_node (Term.literal_to_string literal);
  print_edge_from_current_node ""

let print_split_left (literal: literal)  : unit =
  (* if the first action is a split we need a special dummy root node *)
  if Counter.value node_id = 0 then begin
    print_edge_from_current_node ""
  end;

  store_state_id ();
  print_split_node (Term.literal_to_string literal);
  print_edge_from_current_node ""

let print_split_right (literal: literal) : unit =
  (* the left split node has been associated with the parent state.
     make the right split node a child of the left split node's parent *)
  let parent_id =
    get_state_id (State.parent_state (State.current_state ()))
  in
    print_edge
      (node_name (parent_id - 1))
      "";

  print_split_node (Term.literal_to_string literal);
  print_edge_from_current_node ""

let print_close () : unit =
  print_close_node "";
  Counter.inc node_id






(*** detailed rule printing ***)

let print_assert_detailed (literal: literal) (node_label: string) : unit =
  print_assert_node node_label;
  print_edge_from_current_node (Term.literal_to_string literal)

let print_unit_split_detailed (literal: literal) (node_label: string) : unit =
  print_unit_split_node node_label;
  print_edge_from_current_node (Term.literal_to_string literal)

let print_split_left_detailed (literal: literal) (node_label: string) : unit =
  store_state_id ();
  print_split_node node_label;
  print_edge_from_current_node (Term.literal_to_string literal)

let print_split_right_detailed (literal: literal) : unit =
  (* the (left/right) split node has been associated with the parent state.
     attach the right split edge to the split node. *)
  let state_id =
    get_state_id (State.parent_state (State.current_state ()))
  in
    print_edge
      (node_name state_id)
      (Term.literal_to_string literal)


let print_close_detailed (node_label: string) : unit =
  print_close_node node_label;
  Counter.inc node_id







(*** initialization/finalization ***)

(* create and open the out file, write a header *)
let init (file_name: string) =
  try
    let channel =
      open_out file_name
    in
      out_channel := Some channel;

      print_header ()
  with
    | Sys_error error_message ->
	print_endline error_message


(* close the out file, write a footer *)
let finish () =
  print_footer ();

  match !out_channel with
    | None ->
	()

    | Some channel ->
	try
	  close_out channel
	with
	  | Sys_error error_message ->
	      print_endline error_message
