(* $Id: discrimination_tree.ml,v 1.5 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(* a basic implementation of a discrimination tree term indexing tree.
   Reference: Peter Graf, Term Indexing, Springer-Verlag, New York, NY, 1995

   Copyright (C) 2004 John Wheeler 
   
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(* This implementation is based on code found in the file
   dtIndex.ml contained in the source code for KRHyper, copyright
   2003, by Universitaet Koblenz-Landau.  The file dtIndex.ml
   included the following copyright notice:

   * Copyright (C) 2003 Universitaet Koblenz-Landau
   * 
   * This file is part of KRHyper.
   * 
   * KRHyper is free software; you can redistribute it and/or modify it under the
   * terms of the GNU General Public License as published by the Free Software
   * Foundation; either version 2 of the License, or (at your option) any later
   * version.
   * 
   * KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
   * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
   * details.
   * 
   * You should have received a copy of the GNU General Public License along with
   * KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
   * Place, Suite 330, Boston, MA 02111-1307 USA
   * 
   * Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(* The current author (John Wheer) has adapted the basic implementation of discrimination tree
   term indexing found in the dtIndex.ml file for use in the darwin program and 
   extended the implementation to handle both universal and paremetric variables.  
   This adaptation and extension retained the basic outlines of implementing functions, 
   but rewrote both the call interfaces to match the darwin term indexing 
   interface (see term_index.ml* for details) and the code that handles the term and
   data items that are stored in the darwin term index.  Approximately 70% 
   of the code in the file dtIndex.ml has been modified or rewritten and approximately 
   300 lines of code of additional code were created to implement extensions to 
   address universal and parametric variables.
*)

   

(* hkey *)

type hkey =
    Variable 
  | Parameter 
  | Symbol of Term.symbol 

      (* hkey_equal, modulo renaming *)

let hkey_equal (k1: hkey) (k2: hkey) : bool =
  match k1, k2 with
      Variable, Variable ->
	true
    | Parameter, Parameter ->
	true
    | Symbol s1, Symbol s2 ->
	Symbol.equal s1 s2
    | _, _ ->
	false
	  
let term_hkey (term: Term.term ): hkey =
  match term with
      Term.Var var ->
	begin
	  if Var.is_universal var then
	    Variable
	  else
	    Parameter
	end
    | Term.Const const ->
	Symbol const
    | Term.Func (symbol, _, _) ->
	Symbol symbol

let hkey_arity (hkey: hkey) : int =
  match hkey with
      Symbol s ->
	Symbol.arity_of_symbol s
    | _ -> 0

module HKeyHashtbl =
  Hashtbl.Make(
    struct
      type t = hkey
      let equal = hkey_equal
      let hash = Hashtbl.hash
    end)

module type Table =
sig
  type key = hkey
  type 'a t
  val make_table: unit -> 'a t
  val add: 'a t -> key -> 'a -> unit
  val remove: 'a t -> key -> unit
  val find: 'a t -> key -> 'a
  val iter: (key -> 'a -> unit) -> 'a t -> unit
  val iter_and_throw_away: (key -> 'a -> unit) -> 'a t -> unit
  val fold: (key -> 'a -> 'c -> 'c) -> 'a t -> 'c -> 'c
end

module Table =
struct
  type key = hkey
  type 'a t = 'a HKeyHashtbl.t
  let make_table () = HKeyHashtbl.create 8
  let add = HKeyHashtbl.add
  let remove = HKeyHashtbl.remove
  let find = HKeyHashtbl.find
  let iter = HKeyHashtbl.iter
  let iter_and_throw_away f t = 
    (* Hashtables from the standard library do not to allow removal
       of entries while iterating. *)
    HKeyHashtbl.iter f t;
    HKeyHashtbl.clear t
  let fold = HKeyHashtbl.fold 
end

module Make (DataType : Term_indexing.DataType) :
  Term_indexing.Index
with type data = DataType.t =
struct

  (*** types ***)

  type data = DataType.t
  type term = Term.term
  type subst = Subst.subst

  (* used by the find-first queries to return on success *)
  exception FOUND of data
    
  type entry = {
    term: term;
    data: data
  }
      
  let make_entry (t: term ) (d: data) :entry =
    {term=t ; data=d }

  let get_term (e: entry) : term =
    e.term

  let get_data (e: entry) : data =
    e.data
      
  type entries = {mutable list: entry list}

  type node =
      DtInnerNode of table
    | DtLeafNode of entries
	
  and table =
      node Table.t
	
  type discrimination_tree = {
    mutable root:node;
    mutable size:int
  }
      
  let index_root index = index.root

  type t = discrimination_tree

  (*** consts ***)

  let data_is_equal =
    DataType.is_equal
      
  let data_to_string =
    DataType.to_string

  (* flatten term *)
      
  let rec flatten_term term =
    flatten_term_1 term []
      
  and flatten_term_1 term rest =
    let hkey = term_hkey term in
      match term with
	  Term.Func ( k, termarry, _ ) ->
	    hkey :: flatten_term_1_args
	    rest
	    termarry
	    (hkey_arity hkey)
	    0
	| _ -> hkey :: rest
	    
  and flatten_term_1_args rest args arity i =
    if i >= arity then
      rest
    else
      flatten_term_1
	args.(i)
	(flatten_term_1_args
	   rest
	   args
	   arity
	   (i + 1))

  (* skip some terms in fterm. *)
  (* count is the number of terms to be skipped. *)

  let rec skip_terms_fterm fterm count =
    if count = 0 then
      fterm
    else
      match fterm with 
	  hkey::fterm1 ->
	    skip_terms_fterm fterm1 (count + (hkey_arity hkey) - 1)
	| _ ->
	    raise (Failure "discrimination_tree:skip_terms_fterm: Bad flattened term structure.")

  (*** instance, genralization, and unification predicates ***)
    (* thanks to Alex Fuchs for pointing out how to code these *)
    (* in terms of Unificaition.* funcitons. *)

  let is_instance ?(p_preserving: bool = false) query_term candidate_term =
    (* returns true if candidate_term is an instance of query_term *)
    try
      ignore (Unification.match_terms ~p_preserving:p_preserving 
                query_term 0 candidate_term 1);
      true
    with
      | Unification.UNIFICATION_FAIL ->
          false

  and is_generalization ?(p_preserving: bool = false) query_term candidate_term = 
    (* returns true if candidate_term is a generalization of query_term *)
    try
      ignore(Unification.match_terms ~p_preserving:p_preserving
               candidate_term 0 query_term 1);
      true
    with
      | Unification.UNIFICATION_FAIL ->
          false

  and is_unifiable  ?(p_preserving: bool = false) query_term candidate_term  = 
    (* returns true if candidate_term is unifiable with query_term *)
    try
      ignore (Subst.subst_to_string(Unification.unify_terms ~p_preserving:p_preserving
              query_term 0 candidate_term 1));
      true
    with
      | Unification.UNIFICATION_FAIL ->
	  false 

  (*** creation ***)

  let make_index () = 
    {root=DtInnerNode (Table.make_table ()); size=0}

  let clear index =
    index.root <- DtInnerNode (Table.make_table ());
    index.size <- 0

  let size index =
    index.size 
      
  let create () : discrimination_tree =
    make_index () 

  (*** add ***)

  let rec insert index term entry =
    match index_root index with
	DtInnerNode table ->
	  ignore (insert_1 entry term table 1);
	  index.size <- index.size + 1
      | _ ->
	  raise (Failure "discrimination_tree:insert: Bad index toplevel type.")
	    
  and insert_1 entry term table n =
    (* n is a counter of the terms to be processed. It is used to
       find out when we are at the last subterm in the preorder. *)
    let hkey = term_hkey term in
    let hkey_arity = (hkey_arity hkey) in
    let n1 = n + hkey_arity - 1 in
    let node =
      try 
	Table.find table hkey
      with Not_found ->
	let node =
	  if n1 = 0 then (* At the last subterm in the preorder. *)
	    DtLeafNode {list=[]}
	  else
	    DtInnerNode (Table.make_table ())
	in
	  Table.add table hkey node; (* We could also use replace here. *)
	  node
    in
      match n1, node with
	  0, DtLeafNode entries ->
	    entries.list <- entry :: entries.list;
	    table, n1 (* Return the table just to stay well typed. *)
	| _, DtInnerNode table1 ->
	    ( match term with
		  Term.Func ( _, termarry, _ ) ->
		    insert_1_args
		    entry
		    termarry
		    hkey_arity
		    0
		    table1
		    n1
		| _ -> table1, n1 )
	| _ , _ ->
	    raise (Failure "discrimination_tree:insert_1: Bad index structure.")
	      
  and insert_1_args entry args arity i table n =
    if i >= arity then
      table, n
    else
      let table1, n1 = insert_1 entry args.(i) table n in
	insert_1_args entry args arity (i+1) table1 n1
	  
  and  add (tree: discrimination_tree) (term: term) (data: data) : unit =
    insert tree term (make_entry term data)

  (*** iter ***)

  let rec iter_all_1 (func: term -> data -> unit) node =
    match node with
	DtInnerNode table ->
	  Table.iter (fun k v -> iter_all_1 func v) table
      | DtLeafNode entries ->
	  List.iter (fun entry -> func entry.term entry.data ) entries.list

  and iter_all (func: term -> data -> unit) (tree: discrimination_tree) : unit =
    if tree.size > 0 then
      iter_all_1 func (index_root tree)
	
  and iter (func: term -> data -> unit) (tree: discrimination_tree) : unit =
    iter_all func tree

  (*** remove ***)
      
  (*** remove a variant ***)
      
  exception FoundVar
    
  let is_empty_table table =
    try
      Table.iter 
	(fun k v -> raise FoundVar) table;
      true
    with FoundVar ->
      false
      
  let rec remove_variant index term (data: data option) =
    match (rm_variant term 
	     (flatten_term term)
	     (index_root index) data  ) with
	(empty, foundp) ->
	  ignore (if foundp then 
		    index.size <- index.size - 1);
	  foundp
	    
  and rm_variant term fterm node (data: data option) =
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    try
	      let node1 = Table.find table hkey in
		match (rm_variant term fterm1 node1 data) with
		    (empty , foundp) ->
		      if empty then
			begin
			  Table.remove table hkey;
			  ((is_empty_table table), foundp)
			end
		      else
			(false, foundp)
	    with Not_found ->
	      (false, false)
	  end
      | [], DtLeafNode entries ->
	  begin
	    match list_remove_variant term entries.list data with
		(entlist, foundp) ->
		  begin
		    entries.list <- entlist ;
		    (entries.list = [], foundp)
		  end
	  end
      | _, _ ->
	  raise (Failure "discrimination_tree:rm_variant: Bad index structure.")
		  
  and list_remove_variant term entries (data: data option)=
    match entries with
	[] ->
	  ([] , false)
      | entry :: entries1 ->
	  if Unification.are_terms_variants (get_term entry) term  then 
	    match data with
		Some termdata -> 
		  if data_is_equal (get_data entry) termdata then
(*		  if (get_data entry) = termdata then*)
		    (entries1, true)
		  else (
		    match list_remove_variant term entries1 data with 
			(rest, foundp) -> (entry :: rest , foundp) 
		  )
	      | None ->
		    (entries1, true)
	  else
	    match list_remove_variant term entries1 data with
		(rest, foundp) ->
		  (entry :: rest , foundp) 

  let remove (tree: discrimination_tree) (term: term) (data: data option) : bool =
    (* remove - removes p-preserving u_preserving variants only, per alex 040416/1240-0500 . *)
    remove_variant tree term data

  (*** variants ***)

  (* graf notation for candidate set for variants : *)
  (*   variants(N,t) := var(N,[t]) *)
  (*   var(empty,[X]) := emptyset *)
  (*   var(N, []) := {N} *)
  (* if p_preserving, i.e., U to U and P to P, only*)
  (*     var(N, [x, X]) := var(next(N,*U),[X]) U var(next(N,*P),[X])  *)
  (*     var(N, [p, X]) := var(next(N,*P),[X])  *)
  (* else not p_preserving, i.e., [U|P] to [U|P] *)
  (*     var(N, [[x|p], X]) := var(next(N,*U),[X]) U var(next(N,*P),[X])  *)
  (* end if *)
  (*   var(N, [a, X]) := var(next(N,a),[X]) *)
  (*   var(N, [f(t1, ... tn),X]) := var(next(N,f),[t1,...,tn,X]) *)

  let rec map_variants ?(p_preserving: bool = true) f index term : data list=
    if index.size > 0 then
      map_var ~p_preserving:p_preserving f (flatten_term term) (index_root index)
    else
      []
	
  and map_var ?(p_preserving: bool = true)  f fterm node : data list=
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    try
	      begin 
		if p_preserving then
		  begin 
		    match hkey with
			Variable ->
			  List.append 
			    (try
			       let node1 = Table.find table Variable in
				 map_var ~p_preserving:p_preserving f fterm1 node1
			     with Not_found ->
			       []
			    )
			    (try
			       let node1 = Table.find table Parameter in
				 map_var ~p_preserving:p_preserving f fterm1 node1
			     with Not_found ->
			       []
			    )
		      | _ ->
			  let node1 = Table.find table hkey in
			    map_var ~p_preserving:p_preserving f fterm1 node1
		  end
		else
		  begin 
		    match hkey with
			Variable
		      | Parameter ->
			  List.append 
			    (try
			       let node1 = Table.find table Variable in
				 map_var ~p_preserving:p_preserving f fterm1 node1
			     with Not_found ->
			       []
			    )
			    (try
			       let node1 = Table.find table Parameter in
				 map_var ~p_preserving:p_preserving f fterm1 node1
			     with Not_found ->
			       []
			    )
		      | _ ->
			  let node1 = Table.find table hkey in
			    map_var ~p_preserving:p_preserving f fterm1 node1
		  end
	      end
	    with Not_found ->
	      []
	  end
      | [], DtLeafNode entries ->
(* 	  List.map f entries.list *)
	  List.fold_right f entries.list []
      | _, _ ->
          raise (Failure "discrimination_tree:map_var: Bad index structure.")

  let find_all_variants (tree: discrimination_tree) (term: term) : data list =
    map_variants 
      (fun entry dlist -> 
	 if Unification.are_terms_variants (get_term entry) term  then
	   get_data entry :: dlist
	 else
	   dlist)
      tree term

  (*** generalization ***)

  (* graf notation for candidate set for generalization : *)
  (*   general(N,t) := gen(N,[t]) *)
  (*   gen(empty,[X]) := emptyset *)
  (*   gen(N, []) := {N} *)
  (* if p_preserving, P to P and U to U, only *)
  (*     gen(N, [x, X]) := gen(next(N,*U),[X]) *)
  (*     gen(N, [p, X]) := gen(next(N,*U),[X]) U gen(next(N,*P),[X])  *)
  (*   only a Universal variable may generalize a constant or a function *)
  (*     gen(N, [a, X]) := gen(next(N,*U),[X]) U gen(next(N,a),[X]) *)
  (*     gen(N, [f(t1, ... tn),X]) := gen(next(N,*U),[X]) U *)
  (*                                 gen(next(N,f),[t1,...,tn,X]) *)
  (* else not p_preserving, i.e., [U|P] to [U|P] *)
  (*     gen(N, [[x|p], X]) := gen(next(N,*U),[X]) U gen(next(N,*P),[X]) *)
  (*     gen(N, [a, X]) := gen(next(N,*U),[X]) U gen(next(N,*P),[X]) U gen(next(N,a),[X]) *)
  (*     gen(N, [f(t1, ... tn),X]) := gen(next(N,*U),[X]) U gen(next(N,*P),[X]) U *)
  (*                                 gen(next(N,f),[t1,...,tn,X]) *)
  (* end if *)

  let rec map_generalization_candidates ?(p_preserving: bool = true) ?(find_all: bool = false) f index term : data list=
    if index.size > 0 then
      map_gen ~p_preserving:p_preserving ~find_all:find_all f (flatten_term term) (index_root index)
    else
      []
	
  and map_gen ?(p_preserving: bool = true) ?(find_all: bool = false) f fterm node : data list=
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    if p_preserving then
	      match hkey with
		  Parameter ->
                     List.append 
                      (try
			 let node1 = Table.find table Variable in
			   map_gen ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
		       with Not_found ->
			 []
		      )
                      (try
			 let node1 = Table.find table Parameter in
			   map_gen ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
		       with Not_found ->
			 []
		      )
		|  Variable ->
		    begin 
		      try
			let node1 = Table.find table hkey in
			  map_gen ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
		      with Not_found ->
			[]
		    end
                |  _ ->
		    List.append 
		      (try
			  let node1 = Table.find table Variable in
			    map_gen ~p_preserving:p_preserving ~find_all:find_all 
			      f
			      (skip_terms_fterm fterm 1)
			      node1
			with Not_found ->
			  []
		      )
		      (match hkey with
			   Variable ->
			     []
			 | _ ->
			     try 
			       let node1 = Table.find table hkey in
				 map_gen ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
			     with Not_found ->
			       []
		      )
	    else (* not p_preserving *)
	      List.append
		(List.append 
		   (try
		      let node1 = Table.find table Variable in
			map_gen ~p_preserving:p_preserving ~find_all:find_all 
			  f
			  (skip_terms_fterm fterm 1)
			  node1
		    with Not_found ->
		      []
		   )
		   (try
		      let node1 = Table.find table Parameter in
			map_gen ~p_preserving:p_preserving ~find_all:find_all 
			  f
			  (skip_terms_fterm fterm 1)
			  node1
		    with Not_found ->
		      []
		   )
		)
		( match hkey with
		      Variable 
		    | Parameter ->
			[]
		    | _ ->
			try 
			  let node1 = Table.find table hkey in
			    map_gen ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
			with Not_found ->
			  []
		)
	  end
      | [], DtLeafNode entries ->
	  List.fold_right f entries.list []
      | _, _ ->
	  raise (Failure "discrimination_tree:map_gen: Bad index structure.")
	    
  let find_all_generalizations ~(p_preserving: bool) (tree: discrimination_tree) (term: term) : data list =
    map_generalization_candidates ~p_preserving:p_preserving ~find_all:true 
      (fun entry dlist -> 
         if is_generalization ~p_preserving:p_preserving term (get_term entry) then
           (get_data entry) :: dlist
         else
           dlist)
      tree term

  let find_generalization ~(p_preserving: bool) (tree: discrimination_tree) (term: term) : data option =  
    try
      ignore(
	map_generalization_candidates ~p_preserving:p_preserving ~find_all:false 
	       (fun entry dlist -> 
		  if is_generalization ~p_preserving:p_preserving term (get_term entry) then
		    raise (FOUND (get_data entry))
		  else
		    dlist) 
	       tree term );
      None
    with FOUND data ->
      Some data
      

  (*** unifiable ***)

  (* graf notation for skip set : *)
  (*   skip(N) := U over N' sonsof (N) skip'(N', arity(label(N,N')))  *)
  (*   skip'(N,0) := {N} *)
  (*   skip'(N,l) := U over N' sonsof (N) skip'(N',l - 1 +arity(label(N,N'))) *)

  (* graf notation for candidate set for unifiable terms : *)
  (*   unifiable(N,t):= uni(N,[t])  *)
  (*   uni(empty,[X]) := emptyset  *)
  (*   uni(N, []) := {N}  *)
  (* if p_preserving, i.e., P to P only, but either [U to U | U to term(on both U and P) ] allowed *)
  (*     uni(N, [x, X]) := U over M \in skip(N) uni(M,[X])  *)
  (*     uni(N, [p, X]) := uni(next(N,*V),[X]) U uni(next(N,*P),[X])  *)
  (*   only a Universal variable may generalize a constant or a function *)
  (*     uni(N, [a, X]) := uni(next(N,*U),[X]) U uni(next(N,a),[X]) *)
  (*     uni(N,[f(t1, ... tn),X]) := uni(next(N,*U),[X]) U *)
  (*                                uni(next(N,f),[t1,...,tn,X]) *)
  (* else not p_preserving, i.e., [([U|P] to [U|P]) | ([U|P] to term)] allowed *)
  (*     uni(N, [[x|p] , X]) := U over M \in skip(N) uni(M,[X])  *)
  (*     uni(N, [a, X]) := uni(next(N,*U),[X]) U uni(next(N,*P),[X]) U uni(next(N,a),[X]) *)
  (*     uni(N,[f(t1, ... tn),X]) := uni(next(N,*U),[X]) U uni(next(N,*p),[X]) U *)
  (*                                 uni(next(N,f),[t1,...,tn,X]) *)
  (* end if  *)

  let rec map_unification_candidates ?(p_preserving: bool = true) ?(find_all: bool = false) f index term : data list =
    if index.size > 0 then
      map_unif ~p_preserving:p_preserving ~find_all:find_all f (flatten_term term) (index_root index)
    else
      []
	
  and map_unif ?(p_preserving: bool = true) ?(find_all: bool = false) f fterm node : data list =
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    if p_preserving then
	      match hkey with
		  Variable ->
		    map_unif_skip ~p_preserving:p_preserving ~find_all:find_all f fterm1 node 1
                | Parameter ->
                     List.append 
                      (try 
		         let node1 = Table.find table Variable in
			   map_unif ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
		       with Not_found ->
		         []
		      )
                      (try 
		         let node1 = Table.find table Parameter in
			   map_unif ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
		       with Not_found ->
		         []
		      )
	        | _ ->
 		    List.append
		      (try 
			 let node1 = Table.find table Variable in
			   map_unif ~p_preserving:p_preserving ~find_all:find_all 
			     f
			     (skip_terms_fterm fterm 1)
			     node1
		       with Not_found ->
			 []
		      )
                      (try 
			 let node1 = Table.find table hkey in
			   map_unif ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
		       with Not_found ->
			 []
		      )
	    else
	      match hkey with
		  Variable 
		| Parameter ->
		    map_unif_skip ~p_preserving:p_preserving ~find_all:find_all f fterm1 node 1
		| _ ->
		    List.append
		      (List.append
			 (try 
			    let node1 = Table.find table Variable in
			      map_unif ~p_preserving:p_preserving ~find_all:find_all 
				f
				(skip_terms_fterm fterm 1)
				node1
			  with Not_found ->
			    []
			 )
			 (try 
			    let node1 = Table.find table Parameter in
			      map_unif ~p_preserving:p_preserving ~find_all:find_all 
				f
				(skip_terms_fterm fterm 1)
				node1
			  with Not_found ->
			    []
			 )
		      )
		      (try 
			  let node1 = Table.find table hkey in
			    map_unif ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
			with Not_found ->
			  []
		      )

	  end
      | [], DtLeafNode entries ->
	  List.fold_right f entries.list []
      | _, _ ->
	  raise (Failure "discrimination_tree:map_unif: Bad index structure.")	
	    
  and map_unif_skip ?(p_preserving: bool = true) ?(find_all: bool = false) f fterm node count : data list =
    if count = 0 then
      map_unif ~p_preserving:p_preserving ~find_all:find_all f fterm node
    else
      match node with
	  DtInnerNode table ->
	    (*  val fold : ('a -> 'b -> 'c -> 'c) -> ('a, 'b) t -> 'c -> 'c *)
	    Table.fold
	    (fun hkey node1 dlist ->
	       List.append 
	       (map_unif_skip ~p_preserving:p_preserving ~find_all:find_all f fterm node1 (count + (hkey_arity hkey) - 1))
	       dlist
	    )
	    table
	    []
	| _ -> raise (Failure "discrimination_tree:map_unif_skip: Bad index structure.")
	    
  let find_all_unifiable ~(p_preserving: bool) (tree: discrimination_tree) (term: term) : data list = 
    map_unification_candidates ~p_preserving:p_preserving ~find_all:true 
      (fun entry dlist -> 
         if is_unifiable ~p_preserving:p_preserving term (get_term entry) then
           (get_data entry) :: dlist
         else
           dlist)
      tree term 

  let find_unifiable ~(p_preserving: bool) (tree: discrimination_tree) (term: term) : data option = 
    try
      ignore(
	map_unification_candidates  ~p_preserving:p_preserving ~find_all:false 
	       (fun entry dlist -> 
		  if is_unifiable ~p_preserving:p_preserving term (get_term entry) then
		    raise (FOUND (get_data entry))
		  else
		    dlist)
	       tree term);
      None
    with FOUND data ->
      Some data


  (*** instance ***)

  (* graf notation for candidate set for instances : *)
  (*   instances(N,t) := inst(N,[t]) *)
  (*   inst(empty,[X]) := emptyset *)
  (*   inst(N, []) := {N} *)
  (* if p_preserving, i.e., P to P only, but either [U to U | U to term(on both U and P) ] allowed *)
  (*     inst(N, [x, X]) := U over M \in skip(N) inst(M,[X])  *)
  (*     inst(N, [p, X]) := inst(next(N,*P),[X])  *)
  (* else not p_preserving, i.e., [([U|P] to [U|P]) | ([U|P] to term)] allowed *)
  (*     inst(N, [[x|p] , X]) := U over M \in skip(N) inst(M,[X])  *)
  (* end if *)
  (*   inst(N, [a, X]) := inst(next(N,a),[X]) *)
  (*   inst(N,[f(t1, ... tn),X]) := inst(next(N,f),[t1,...,tn,X]) *)

  let rec map_instance_candidates ?(p_preserving: bool = true) ?(find_all: bool = false) f index term : data list =
    if index.size > 0 then
      map_inst ~p_preserving:p_preserving ~find_all:find_all f (flatten_term term) (index_root index)
    else
      []

  and map_inst ?(p_preserving: bool = true) ?(find_all: bool = false) f fterm node : data list=
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    if p_preserving then
	      match hkey with
		  Variable ->
		    map_inst_skip ~p_preserving:p_preserving ~find_all:find_all f fterm1 node 1
		| _ ->
		    try
		      let node1 = Table.find table hkey in
			map_inst ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
		    with Not_found ->
		      []
	    else
	      match hkey with
		  Variable 
		| Parameter ->
		    map_inst_skip ~p_preserving:p_preserving ~find_all:find_all f fterm1 node 1
		| _ ->
		    try
		      let node1 = Table.find table hkey in
			map_inst ~p_preserving:p_preserving ~find_all:find_all f fterm1 node1
		    with Not_found ->
		      []
	  end
      | [], DtLeafNode entries ->
(* 	    ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b *)
 	    List.fold_right f entries.list [] 

      | _, _ ->
	  raise (Failure "discrimination_tree:map_inst: Bad index structure.")

  and map_inst_skip ?(p_preserving: bool = true) ?(find_all: bool = false) f fterm node count : data list =
    if count = 0 then
      map_inst ~p_preserving:p_preserving ~find_all:find_all f fterm node
    else
      match node with
	  DtInnerNode table ->
	    (*  val fold : ('a -> 'b -> 'c -> 'c) -> ('a, 'b) t -> 'c -> 'c *)
	    Table.fold
	    (fun hkey node1 dlist ->
	       List.append
	       (map_inst_skip ~p_preserving:p_preserving ~find_all:find_all f fterm node1 (count + (hkey_arity hkey) - 1))
	       dlist
	    )
	    table
	    []
	| _ -> raise (Failure "discrimination_tree:map_inst_skip: Bad index structure.")

  let find_all_instances ~(p_preserving: bool) (tree: discrimination_tree) (term: term) : data list =
    map_instance_candidates ~p_preserving:p_preserving ~find_all:true 
      (fun entry dlist -> 
         if is_instance ~p_preserving:p_preserving term (get_term entry) then
           (get_data entry) :: dlist
         else
           dlist)
      tree term 

  let find_instance ~(p_preserving: bool) (tree: discrimination_tree) (term: term) : data option = 
    try 
      ignore(
	map_instance_candidates ~p_preserving:p_preserving ~find_all:false 
	       (fun entry dlist -> 
		  if is_instance ~p_preserving:p_preserving term (get_term entry) then
		    raise (FOUND (get_data entry))
		  else
		    dlist) 
	       tree term );
      None
    with FOUND data ->
      Some data

  (*** productivity ***)

  (* find the first or all elements according to the specified function *)
  let find_non_productive (tree: discrimination_tree) (producer: term) (produced: term) : data option =
    None

  (*** msg ***)

  let find_non_msg (tree: discrimination_tree) (producer: term) (produced: term) : data option =
    None

  (*** to_string ***)

  let hkey_to_string hkey : string =
    match hkey with
	Symbol sym ->
	  (Symbol.name_of_symbol sym) ^ "/" ^ (string_of_int (Symbol.arity_of_symbol sym))
      | Variable ->
	  "*U/0"
      | Parameter ->
	  "*P/0"

  let rec indent d : string =
      if d <= 0 then
	""
      else
	"  " ^ indent (d - 1)

  let rec index_to_string index : string =
    let node = index_root index in
      node_to_string node 0

  and node_to_string node d = 
    match node with
	DtInnerNode table ->
	  Table.fold
	  (fun k v str ->
	     (indent d) ^ 
	     (hkey_to_string k) ^ "\n" ^ 
	     (node_to_string v (d + 1)) ^ 
	     str 
	  )
	  table
	  ""
      | DtLeafNode entries ->
	  (indent d) ^ "[ " ^ entrylist_to_string entries.list ^ "]\n"

  and entrylist_to_string (entrylist: entry list) : string =
    match entrylist with
	entry :: entrylist1 ->
	  let data = get_data entry in
	    data_to_string data ^ ", " ^ entrylist_to_string entrylist1
      | _ ->
	  ""

  let to_string (tree : discrimination_tree) : string =
    index_to_string tree

end



