(* $Id: heap.mli,v 1.7 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** imperative heap 

  The heap data structure.
  Adding elements and removing the minimum element
  have logarithmic complexity.

  The heap can take up to Sys.max_array_length elements
*)


module type OrderedType = 
  sig
    type t

    val compare : t -> t -> int
  end
(** Input signature of the functor {!Heap.Make}.

  It is acceptable if different [t]s are considered to be equal by compare.
  (This is not the case for e.g. the [Set] data structure of the OCaml library)
*)


module Make (Ord: OrderedType) : sig
  type data = Ord.t

  (** the heap *)
  type t

  exception OVERFLOW

  (** creates an empty heap *)
  val create: unit -> t


  (** adds an element.
    raises OVERFLOW if the heap is full. *)
  val add: t -> data -> unit


  (** returns the minimum element.
     raises [Not_found] if the heap is empty *)
  val min: t -> data

  (** returns the 'last' element in the heap.
    this is in general not the maximum element,
    but usually a relative big one.
    raises [Not_found] if the heap is empty *)
  val last: t -> data


  (** removes the minimum element and returns it.
     raises [Not_found] if the heap is empty *)
  val remove_min: t -> data

  (** removes the 'last' element in the heap and returns it.
    this is in general not the maximum element,
    but usually a relative big one.
    raises [Not_found] if the heap is empty *)
  val remove_last: t -> data

  (** iterates over all elements *)
  val iter: (data -> unit) -> t -> unit

    
  (** no elements? *)
  val empty: t -> bool

  (** number of elements *)
  val size: t -> int
end
(** Functor building an implementation of the heap structure
   given a ordered type. *)
