(* $Id: problem_literals.ml,v 1.12 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type literal = Term.literal
type binding = Subst.binding
type subst = Subst.subst
type packed_subst = Subst.packed_subst
type context_partner = Context_unifier.context_partner
type input_partner = Context_unifier.input_partner


(*** callback function types ***)
type on_add_unit = Context.element -> unit
type on_add = input_partner -> context_partner -> subst -> unit
type on_subsume =  State.state -> unit




(* binding database *)
module Bindings =
  Weak.Make (
    struct
      type t = binding

      let equal (binding1: binding) (binding2: binding) : bool =
	(Subst.var_equal binding1.Subst.sb_var binding2.Subst.sb_var)
	&&
	(Subst.term_equal binding1.Subst.sb_term binding2.Subst.sb_term)

      let hash (binding: binding) : int =
	let var_hash =
	  Var.id_of_var binding.Subst.sb_var.Subst.sv_var
	and term_hash =
	  match binding.Subst.sb_term.Subst.st_term with
	    | Term.Var var ->
		Var.id_of_var var
		
	    | Term.Const symbol ->
		Symbol.id_of_symbol symbol
		
	    | Term.Func (_, _, hash_value) ->
		hash_value
	in
	  (var_hash + term_hash) * 131

    end
  )



(* problem literals accessed by literal *)
module LiteralTable =
  Hashtbl.Make (
    struct
      type t = literal

      let equal (literal1: literal) (literal2: literal) : bool =
	Term.literal_equal literal1 literal2

      let hash (literal: literal) : int =
	match literal.Term.term with
	  | Term.Var var ->
	      failwith ("Problem_literals.");
	      Var.id_of_var var
	      
	  | Term.Const symbol ->
	      Symbol.id_of_symbol symbol
	      
	  | Term.Func (_, _, hash_value) ->
	      hash_value

    end
  )

(* problem literals accessed by sign and predicate *)
module Index =
  Hashtbl.Make (
    struct
      type t = literal

      let equal (literal1: literal) (literal2: literal) : bool =
	(literal1.Term.sign = literal2.Term.sign)
	&&
	match literal1.Term.term, literal2.Term.term with
	  | Term.Const symbol1, Term.Const symbol2
	  | Term.Func (symbol1, _, _), Term.Func (symbol2, _, _) ->
	      Symbol.equal symbol1 symbol2

	  | _ ->
	      false

      let hash (literal: literal) : int =
	match literal.Term.term with
	  | Term.Const symbol
	  | Term.Func (symbol, _, _) ->
	      Symbol.id_of_symbol symbol

	  | _ ->
	      failwith "Problem_literals.Indexes.hash"
    end
  )


(* input literals with offsets -
   stored in a problem literal which ignores any offset *)
module Elements =
  Map.Make (
    struct
      type t = int

      let compare (first: int) (second: int) : int =
	compare first second
    end
  )



(* data per problem literal with a specific offset *)
type element = {
  (* the problem literal is kept as an input_partner *)
  input_partner: input_partner;

  (* registered subsume call backs (for all clauses containing this literal) *)
  mutable callbacks_on_add: on_add list;
}

(* data per problem literal without offset *)
type problem_literal = {
  (* variants of the problem literal, i.e. pairing with different offsets *)
  mutable elements: element Elements.t;

  (* registered add call backs (for all clauses containing this literal) *)
  mutable callbacks_on_subsume: on_subsume list;

  (* which states subsume the literal? *)
  mutable subsumed_at: State.state list;
}







type global = {
  (* access to problem literals by literal - used during creation *)
  mutable problem_literals: problem_literal LiteralTable.t;

  (* access to problem literals by sign and predicate -
     used when checked against new context literals
     to quickly ignore literals with wrong sign and predicate *)
  index: ((literal * problem_literal) list) Index.t;

  (* all registered call backs for unit clauses *)
  mutable callbacks_on_add_unit: on_add_unit list;

  (* database for the bindings used in the substitution
     between the problem literals and the context partners (cp_subst).
     this can significantly decrease the memory usage, esp. for horn problems *)
  bindings: Bindings.t;
}


let global = {
  problem_literals = LiteralTable.create 0;
  index = Index.create 8;
  callbacks_on_add_unit = [];
  
  bindings = Bindings.create 1024;
}







(*** initialization ***)



let init () =
  (* use the number of literals to initialize the Hashtbl.
     as of OCaml 3.07 this has a noticable impact on the performance
     of the collision handling,
     even if the Hashtbl size is increased afterwards *)
  global.problem_literals <- LiteralTable.create 16;
  global.callbacks_on_add_unit <- [];
  Index.clear global.index;
  
  Bindings.clear global.bindings




(*** binding database ***)  

let register_binding (binding: binding) : binding =
  try
    Bindings.find global.bindings binding
  with
    | Not_found ->
	Bindings.add global.bindings binding;
	binding


let register_subst (subst: subst) : packed_subst =
  Array.map register_binding (Array.of_list subst)







(*** registration ***)




(* dummy values for -v/+v:
   just provide a parameter which is bound to a non-parameter term.
   
   as the remainder literals are built from the variables of the input clause
   and not from the parameters of the context literals,
   this binding is never applied.
   it is only checked in order to determine its remainder status,
   so a binding to any non-parameter is sufficient *)
let minus_v_par = 
  Var.create_parametric 0

and minus_v_const = 
  Symbol.create_constant ()
    
let minus_v_cp_element () = {
  Context.el_pars = [minus_v_par];
  Context.el_state = State.valid_state;
  Context.el_literal = Term.get_default_v ();
  Context.el_generation = 0;
  Context.el_mandatory_at = None;
  Context.el_compacted = [];
}
      

(* creates the initial queue of context partners for a literal.
   i.e. either no initial partner or -v for a positive literal. *)
let create_initial_context_partners (literal: literal) (index: int) : context_partner list =
  if 
    (* in the horn case split never happens,
       so only assert or close context unifiers are of interest.
       thus -v which produces non-empty remainders is only doing bad *)
    
    Config.get_horn ()
    ||
    (* -v or +v may be used - and only applyable if the literal has the opposite sign  *)
    literal.Term.sign = Config.get_use_plus_v ();
  then
    []
  else begin
    (* just ensuring that this context partner is not p-preserving *)
    let subst =
      Subst.set_  ~p_preserving:false ~i_preserving:false
	Subst.empty
	minus_v_par
	(Subst.context_literal_offset index)
	(* ensure that the minus_v term is in the database, even after a restart *)
	(Term.request_const minus_v_const) 
	(Subst.context_literal_offset index);
    in
    let context_partner = {
      Context_unifier.cp_element = minus_v_cp_element ();
      Context_unifier.cp_subst = register_subst subst;
    }
    in
      [context_partner]
  end



let register_to_index (literal: literal) (problem_literal: problem_literal) : unit =
  let problem_literals =
    try
      Index.find global.index literal
    with
      | Not_found ->
	  Index.add global.index literal [];
	  []
  in
    Index.replace global.index literal ((literal, problem_literal) :: problem_literals)


let register_literal (literal: literal) (index: int) : input_partner =
  let problem_literal =
    try
      LiteralTable.find global.problem_literals literal
    with
      | Not_found ->
	  let new_problem_literal = {
	    elements = Elements.empty;
	    callbacks_on_subsume = [];
	    subsumed_at = [];
	  }
	  in
	    LiteralTable.add global.problem_literals literal new_problem_literal;
	    register_to_index literal new_problem_literal;
	    new_problem_literal;
  in

  let element =
    try
      Elements.find index problem_literal.elements
    with
      | Not_found ->
	  (* first occurance of this literal/index: create it *)
	  let vars =
	    List.map
	      (fun var ->
		 Subst.make_var var Subst.input_literal_offset
	      )
	      (Term.vars_of_literal literal)
	  in
	    
	  let new_input_partner = {
	    Context_unifier.ip_index = index;
	    Context_unifier.ip_literal = literal;
	    Context_unifier.ip_vars = vars;
	    Context_unifier.ip_context_partners = create_initial_context_partners literal index;
	    Context_unifier.ip_resolved = [];
	  }
	  in

	  let new_element = {
	    input_partner = new_input_partner;
	    callbacks_on_add = [];
	  }
	  in

	    problem_literal.elements <- Elements.add index new_element problem_literal.elements;
	    new_element
  in
    element.input_partner




let register_clause_literal (literal: literal) (index: int)
  (on_add: on_add) (on_subsume: on_subsume) : unit  =

  let problem_literal =
    try
      LiteralTable.find global.problem_literals literal
    with
      | Not_found ->
	  failwith "register_clause_literal: problem_literal"
  in

  let element =
    try
      Elements.find index problem_literal.elements
    with
      | Not_found ->
	  failwith "register_clause_literal: element"
  in

    problem_literal.callbacks_on_subsume <- on_subsume :: problem_literal.callbacks_on_subsume;

    element.callbacks_on_add <- on_add :: element.callbacks_on_add

    
      
let register_unit_literal (on_add_unit: on_add_unit) : unit =
  global.callbacks_on_add_unit <- on_add_unit :: global.callbacks_on_add_unit







(*** resolve ***)


(* is the problem literal already resolved with respect to
   the dependencies on the context literal? *)
let is_already_resolved (input_partner: input_partner) (context_element: Context.element) : bool =
  if Config.get_use_chronological_backtracking () then begin
    match input_partner.Context_unifier.ip_resolved with
      | [] ->
	  false
	    
      | _ ->
	  true
  end
  else begin
    List.exists
      (fun resolve_partner ->
	 State.depends_on
	   context_element.Context.el_state
	   resolve_partner.Context_unifier.cp_element.Context.el_state
      )
      input_partner.Context_unifier.ip_resolved
  end




(* mark the problem literal as resolved by this context literal *)
let mark_resolved (element: input_partner) (context_partner: context_partner): unit =
  Statistic.inc_resolve ();

  element.Context_unifier.ip_resolved <-
    context_partner :: element.Context_unifier.ip_resolved;

  (* no backtracking in horn - so this stays resolved forever
     and some entries can be removed for garbage collection *)
  if Config.get_horn () then begin
    element.Context_unifier.ip_context_partners <- context_partner :: [];
  end


(* checks the resolve inference rule against
   the problem literal and the context literal *)
let resolve (input_partner: input_partner) (context_partner: context_partner) : bool =
  if not (Config.get_use_resolve ()) then
    false

  else if
    context_partner.Context_unifier.cp_element.Context.el_literal.Term.sign
    =
    input_partner.Context_unifier.ip_literal.Term.sign
  then
    false

  else begin
    let context_term =
      context_partner.Context_unifier.cp_element.Context.el_literal.Term.term
    and problem_term =
      input_partner.Context_unifier.ip_literal.Term.term
    in
      (* variants? *)
      if Unification.are_terms_variants context_term problem_term then begin
	mark_resolved input_partner context_partner;
	(* this is only a variable renaming, so the substitution can be removed.
	   This can vastly improve the speed of merging substitutions. *)
	context_partner.Context_unifier.cp_subst <- [| |];
	true
      end
	
      (* subsumed? this implies Resolve,
	 as the effect on the rest clause can only be a variable remaining *)
      else begin
	try
	  let subst =
	    Unification.match_terms
	      ~p_preserving:true
	      context_term
	      (Subst.context_literal_offset input_partner.Context_unifier.ip_index)
	      problem_term
	      Subst.input_literal_offset
	  in
	    mark_resolved input_partner context_partner;
	    (* use this substitution instead of the originally computed -
	       it is subsuming,
	       thus only variables of the context literal are bound,
	       thus merging of substitutions might be more efficient *)
	    context_partner.Context_unifier.cp_subst <- register_subst subst;
	    true
	    
	with
	  | Unification.UNIFICATION_FAIL ->
	      false
      end
  end



(*** subsume ***)

let is_already_subsumed (problem_literal: problem_literal) (context_element: Context.element) : bool =
  if Config.get_use_chronological_backtracking () then begin
    match problem_literal.subsumed_at with
      | [] ->
	  false
	  
      | _ ->
	  true
  end
  else
    List.exists
      (State.depends_on context_element.Context.el_state)
      problem_literal.subsumed_at


(* try subsume with the context literal on the problem literal *)
let is_literal_subsumed (literal: literal) (context_element: Context.element) : bool =
  try
    ignore (Unification.match_literals
	      ~p_preserving:true
	      context_element.Context.el_literal
	      0
	      literal
	      1);
    true
  with
    | Unification.UNIFICATION_FAIL -> 
	false



(* try subsume with the context literal on all problem literals *)
let subsume (problem_literal: problem_literal)
  (literal: literal)
  (context_element: Context.element) : unit =
  if not (Config.get_use_subsume ()) then
    ()
      
  else if is_already_subsumed problem_literal context_element then
    ()
      
  else if not (is_literal_subsumed literal context_element) then
    ()
      
  else begin
    Statistic.inc_subsume ();
    
    (* process callbacks *)
    List.iter
      (fun on_subsume ->
	 on_subsume (State.current_state ())
      )
      problem_literal.callbacks_on_subsume;
    
    (* mark as subsumed *)
    problem_literal.subsumed_at <- State.current_state () :: problem_literal.subsumed_at;
    
    (* no backtracking in horn - so this stays subsumed forever
       and all context partner may be forgotten *)
    if Config.get_horn () then begin
      Elements.iter
	(fun _ element ->
	   element.input_partner.Context_unifier.ip_context_partners <- [];
	)
	problem_literal.elements
    end;
  end


(*** add ***)



let process_callbacks_on_add_unit (context_element: Context.element) : unit =
  List.iter
    (fun callback ->
       callback context_element
    )
    global.callbacks_on_add_unit



let add_to_element
  (problem_literal: problem_literal)
  (neg_new_context_literal: literal)
  (context_element: Context.element)
  : unit =
  (* if the first literal * offset fails,
     the literal's other offset combination need not to be checked.
     otherwise the information can be reused *)
  try
    let _ =
      Elements.fold
	(fun index element cache ->
	   
	   match cache with
	       (* first time that this literal is processed with (any index) *)
	     | None ->
		 (* already resolved? *)
		 if is_already_resolved element.input_partner context_element then begin
		   raise Exit
		 end
		   
		 (* unify *)
		 else begin
		   try
		     let subst =
		       Unification.unify_literals
			 element.input_partner.Context_unifier.ip_literal
			 Subst.input_literal_offset
			 neg_new_context_literal
			 (Subst.context_literal_offset index)
		     in

		     (* create the initial context partner *)
		     let context_partner = {
		       Context_unifier.cp_element = context_element;
		       Context_unifier.cp_subst = register_subst subst;
		     }
		     in
		       
		       (* and add it *)
		       element.input_partner.Context_unifier.ip_context_partners <-
		       context_partner :: element.input_partner.Context_unifier.ip_context_partners;
		       
		       (* resolving? *)
		       let resolved =
			 resolve element.input_partner context_partner
		       in
			 
			 (* process all callbacks *)
			 List.iter
			   (fun callback ->
			      callback element.input_partner context_partner subst
			   )
			   element.callbacks_on_add;
			 
			 Some (element, context_partner, resolved)
		   with
		     | Unification.UNIFICATION_FAIL ->
			 raise Exit
		 end

	     (* literal has already been processed with another index *)
	     | Some (cached_element, cached_context_partner, cached_resolved) ->
		 (* create the initial context partner *)
		 let new_subst =
		   Subst.replace_offset
		     (Array.to_list cached_context_partner.Context_unifier.cp_subst)
		     (Subst.context_literal_offset cached_element.input_partner.Context_unifier.ip_index)
		     (Subst.context_literal_offset index)
		 in
		 let new_context_partner = {
		   Context_unifier.cp_element = context_element;
		   Context_unifier.cp_subst = register_subst new_subst;
		 }
		 in

		   (* and add it *)
		   element.input_partner.Context_unifier.ip_context_partners <-
		   new_context_partner :: element.input_partner.Context_unifier.ip_context_partners;

		   if cached_resolved then begin
		     mark_resolved element.input_partner new_context_partner;
		   end;

		   (* process all callbacks *)
		   List.iter
		     (fun callback ->
			callback element.input_partner new_context_partner new_subst
		     )
		     element.callbacks_on_add;

		   cache
	)
	problem_literal.elements
	None
    in
      ()
  with
    | Exit ->
	()



let add (context_element: Context.element) : unit =

  process_callbacks_on_add_unit context_element;

  (* subsume *)
  let subsume_candidates =
    try
      Index.find global.index context_element.Context.el_literal
    with
      | Not_found ->
	  (* a unit literal and thus not registered *)
	  []
  in
    List.iter
      (fun (literal, problem_literal) ->
	 subsume problem_literal literal context_element
      )
      subsume_candidates;

  (* unify *)
  let neg_new_context_literal =
    Term.request_negated_literal context_element.Context.el_literal
  in
  let unifying_candidates =
    try
      Index.find global.index neg_new_context_literal
    with
      | Not_found ->
	  (* a unit literal and thus not registered *)
	  []
  in
    List.iter
      (fun (literal, problem_literal) ->
	 add_to_element
	 problem_literal
	 neg_new_context_literal
	 context_element
      )
      unifying_candidates





(*** backtrack ***)


let backtrack () =
  LiteralTable.iter
    (fun _ problem_literal ->

       (* subsumed *)
       problem_literal.subsumed_at <-
         List.find_all State.is_state_valid problem_literal.subsumed_at;

       Elements.iter
         (fun _ element ->
	    (* context_partners *)
	    if Config.get_use_chronological_backtracking () then begin
	      let rec remove_invalid_context_partner context_partners =
		match context_partners with
		  | [] ->
		      element.input_partner.Context_unifier.ip_context_partners <- []

		  | context_partner :: tail ->
		      if State.is_state_valid context_partner.Context_unifier.cp_element.Context.el_state then
			element.input_partner.Context_unifier.ip_context_partners <- context_partners
		       
		      else begin
			remove_invalid_context_partner tail
		      end
	      in
		remove_invalid_context_partner element.input_partner.Context_unifier.ip_context_partners
	    end

	    else begin
	      element.input_partner.Context_unifier.ip_context_partners <-
	      List.find_all
		(fun context_partner ->
		   State.is_state_valid context_partner.Context_unifier.cp_element.Context.el_state
		)
		element.input_partner.Context_unifier.ip_context_partners
	    end;
	  
	    (* resolved *)
	    element.input_partner.Context_unifier.ip_resolved <-
	      List.find_all
                (fun context_partner ->
		   State.is_state_valid context_partner.Context_unifier.cp_element.Context.el_state
		)
                element.input_partner.Context_unifier.ip_resolved;

       )
       problem_literal.elements;

    )
    global.problem_literals
