(* $Id: problem_literals.mli,v 1.7 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** checks all problem literals against a new context literal

  Contains all literals of the problem set.
  
  As literals with different offsets are to be considered as variants
  a literal is stored paired with its index in a clause.
  Thus the same literal might be stored several times
  but with different offsets.

  Checks for every new context literal its unifiability
  with all problem literals, stores the information
  (substitution, resolved, subsumed, ...),
  and calls the registered call back functions.

  Applies the {e Resolve} and {e Subsume} rule
  of the model evolution calculus.
  As this module contains no information about clauses,
  {e Resolve} is only applied in a restricted - clause independent - way. *)


(** {6 Types} *)



type literal = Term.literal
type subst = Subst.subst
type context_partner = Context_unifier.context_partner
type input_partner = Context_unifier.input_partner

(** should be used by unit clauses to register
  and be notified whenever a context literal is added.
  The intention is to raise CLOSED on a contradiction *)
type on_add_unit = Context.element -> unit

(** should be used by non-unit clauses to register literals
  and be notified whenever a unifiable context literal is added. *)
type on_add = input_partner -> context_partner -> subst -> unit

(** should be used by non-unit clauses to register literals
  and be notified whenever a subsuming context literal is added. *)
type on_subsume = State.state -> unit


(** {6 Functions} *)


(** {2 Initialization} *)

(** needs to be called before the first usage of any other function,
  i.e. also after each restart. *)
val init: unit -> unit


(** {2 Registration} *)

(** [register_literal literal index]
  retrieves the input_partner for the problem [literal]
  used as the [index].th literal in a clause. *)
val register_literal: literal -> int -> input_partner

(** [register_clause_literal literal index on_add on_subsume]
  registers {!Problem_literals.on_add}
  and {!Problem_literals.on_subsume}
  for the problem [literal]
  used as the [index].th literal in a clause.

  [literal] must have been previously registered
  with {!Problem_literals.register_literal} *)
val register_clause_literal: literal -> int -> on_add -> on_subsume -> unit

(** [register_unit_literal on_add_unit]
  registers {!Problem_literals.on_add_unit} *)
val register_unit_literal: on_add_unit -> unit


(** {2 Adding} *)

(** checks the new context literal against all stored problem literals,
  stores the results,
  and calls the appropriate registered call back functions. *)
val add: Context.element -> unit


(** {2 Backtracking} *)

(** removes information based on invalid context literals. *)
val backtrack: unit -> unit
