(* $Id: selection.ml,v 1.27 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(*** types ***)

type state = State.state
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type explanation = State.explanation
type context_unifier_space = Context_unifier.context_unifier_space
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected_assert = Selection_types.selected_assert
type selected_split = Selection_types.selected_split

type selection = {
  sel_context_unifier_spaces: context_unifier_space array;
  (* the context_unifier_spaces build on the input clauses *)

  mutable sel_unit_asserts: literal list;
  (* unit assert candidates *)

  mutable sel_assert: Selection_assert.candidates;
  (* assert candidates *)

  mutable sel_split: Selection_split.candidates;
  (* split candidates *)
}


type selected =
  | Assert of selected_assert
  | Assert_split of selected_assert
  | Split of selected_split




exception SATISFIABLE
exception CLOSED = Context_unifier.CLOSED




(*** creation ***)

let create (clauses: clause list) : selection =
  let selection_assert =
    Selection_assert.create ()
  and selection_split =
    Selection_split.create ()
  in

  let process_assert_candidate (candidate: raw_context_unifier) : unit =
    Selection_assert.add selection_assert candidate
  and process_split_candidate (candidate: raw_context_unifier) : unit =
    Selection_split.add selection_split candidate
  in

  (* create a context_unifier_space for every clause,
     gather all candidates computed on creation *)
  let context_unifier_spaces, unit_asserts =
    List.fold_left
      (fun (context_unifier_spaces, unit_asserts) clause ->
	 let new_context_unifier_space, new_unit_assert =
	   Context_unifier_space.create
	     clause
	     process_assert_candidate
	     process_split_candidate
	 in

	 let new_unit_asserts =
	   match new_unit_assert with
	     | None ->
		 unit_asserts
		 
	     | Some assert_literal ->
		 assert_literal :: unit_asserts
	 in
	   (new_context_unifier_space :: context_unifier_spaces),
	   new_unit_asserts
      )
      ([], [])
      clauses
  in

  (* sort the assert literals of unit clauses by term_weight *)
  let sorted_unit_asserts =
    List.sort
      (fun (x: literal) (y: literal) ->
	 compare
	 (Term_quality.weight_of_literal x)
	 (Term_quality.weight_of_literal y)
      )
      unit_asserts
  in    

  let context_unifier_spaces_array =
    Array.of_list context_unifier_spaces
  in

  let selection =
    {
      sel_context_unifier_spaces = context_unifier_spaces_array;
      sel_unit_asserts = sorted_unit_asserts;
      sel_assert = selection_assert;
      sel_split = selection_split;
    }
  in

    selection








(*** add ***)

(* add the new context literal to all unsubsumed clauses
   and get the new candidates *)
let add (selection: selection) (context_element: Context.element) : unit =
  Problem_literals.add context_element





(*** selection ***)


let rec select_unit_assert (selection: selection) (candidates: literal list): selected_assert option =
  match candidates with
    | [] ->
	selection.sel_unit_asserts <- [];
	None
	
    | candidate :: tail ->
	begin
	  match Context.check_subsumed candidate with
	    | Some _ ->
		select_unit_assert selection tail
		
	    | None ->
		selection.sel_unit_asserts <- tail;
		
		Some {
		  Selection_types.sa_literal = candidate;
		  Selection_types.sa_mandatory_at = Some (State.current_state ());
		  Selection_types.sa_explanation = State.empty_explanation;
		  Selection_types.sa_generation = 0;
		  Selection_types.sa_raw_context_unifier = None;
		}
	end


let select (selection: selection) : selected =
  (* first try to assert a unit input clause *)
  match select_unit_assert selection selection.sel_unit_asserts with
    | Some selected ->
	Assert selected

    | None ->
	begin
	  (* then try to find an assert literal *)
	  match Selection_assert.select selection.sel_assert with
	    | Some selected ->
		Assert selected
		
	    | None ->
		(* finally try to find a split literal *)
		begin
		  match Selection_split.select selection.sel_split with
		    | Some (Selection_split.Assert selected) ->
			Assert_split selected;
			
		    | Some (Selection_split.Split selected) ->
			Split selected;
			
		    | None ->
			raise SATISFIABLE
		end
	end
	







(*** backtrack ***)
let backtrack (selection: selection) (explanation: explanation) : unit =
  (* context partners *)
  Problem_literals.backtrack ();

  (* partial context unifiers - dependend on Problem_literals.backtrack *)
  Array.iter Context_unifier_space.backtrack selection.sel_context_unifier_spaces;

  (* candidates *)
  Selection_lookahead.backtrack ();
  Selection_assert.backtrack selection.sel_assert;
  Selection_split.backtrack selection.sel_split
