(* $Id: selection_assert.ml,v 1.47.2.3 2004/06/14 05:28:15 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type state = State.state
type literal = Term.literal
type explanation = State.explanation
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected_assert = Selection_types.selected_assert

type quality = {
  cq_term_weight: int;
  cq_generation: int;
}

type candidate = {
  ca_raw_context_unifier: raw_context_unifier;

  (* the assert literal *)
  ca_literal: literal;
  
  (* comparison properties for the selection heuristic *)
  ca_quality: quality;
}



(*** comparison ***)

let compare_qualities (first: quality) (second: quality) : int =
  let term_weight =
    compare first.cq_term_weight second.cq_term_weight
  in
    if term_weight <> 0 then
      term_weight
    else
      compare first.cq_generation second.cq_generation

let compare_candidates (first: candidate) (second: candidate) : int =
  compare_qualities first.ca_quality second.ca_quality


module Active =
  Heap.Make ( 
    struct
      type t = candidate

      let compare (first: candidate) (second: candidate) : int =
	compare_candidates first second
    end
  )

module Inactive =
  Heap.Make ( 
    struct
      type t = raw_context_unifier * quality

      let compare (_, first) (_, second) : int =
	compare_qualities first second
    end
  )

type candidates = {
  (* freshly computed context unifiers
     which have to be transformed into candidates
     at the beginning of the next candidate selection. *)
  mutable cd_new: raw_context_unifier list;

  (* the assert candidates currently valid for adding to the context
     (at least valid when last checked). *)
  mutable cd_active: Active.t;

  (* to save memory only a limited amount of candidates is stored
     with full informatio nin cd_active,
     the others are kept here *)
  cd_inactive: Inactive.t;

  (* invalid candidates and the state of the responsible context literal *)
  mutable cd_invalid: (raw_context_unifier * state) list;

  (* candidates added to the context (at state), to consider in backtracking *)
  mutable cd_added: (candidate * state) list;
}


(* how many active candidates may be stored?
   other are temporarily kept inactive. *)
let active_threshhold =
  1024 * 8




(*** representation ***)


let candidate_to_string (title: string) (candidate: candidate) : string =
  State.current_state_to_string ()
  ^ title ^ ": " ^ Term.literal_to_string candidate.ca_literal ^ "\n"
  ^
  if Config.get_show_context_unifier () then
    Context_unifier.raw_context_unifier_to_string candidate.ca_raw_context_unifier
  else
    ""





			      
(*** creation ***)

let create () : candidates =
  {
    cd_new = [];
    cd_active = Active.create ();
    cd_inactive = Inactive.create ();
    cd_invalid = [];
    cd_added = [];
  }

  
let add_active_candidate (candidates: candidates) (candidate: candidate) : unit =
  Selection_lookahead.add candidate.ca_literal candidate.ca_raw_context_unifier;
  Active.add candidates.cd_active candidate



(* add a valid candidate to the active/inactive queue *)
let add_valid_candidate (candidates: candidates) (candidate: candidate) : candidate option =
  if Selection_lookahead.check candidate.ca_literal candidate.ca_raw_context_unifier then
    Some candidate
  else
    try
      if Active.size candidates.cd_active < active_threshhold then begin
	(* active queue is not full, so add here *)
	add_active_candidate candidates candidate
      end
      else begin
	
	if compare_candidates (Active.last candidates.cd_active) candidate > 0 then begin
	  (* candidate is better than an element in active queue,
	     so switch them *)
	  add_active_candidate candidates candidate;
	  
	  let last_heap_element =
	    Active.remove_last candidates.cd_active
	  in
	    (* silently drop invalid candidates *)
	    if not (Context_unifier.is_raw_context_unifier_invalid last_heap_element.ca_raw_context_unifier) then begin
	      Selection_lookahead.remove last_heap_element.ca_literal last_heap_element.ca_raw_context_unifier;
	      
	      Inactive.add candidates.cd_inactive (last_heap_element.ca_raw_context_unifier, last_heap_element.ca_quality);
	    end	  
	end
	  
	else begin
	  (* candidate is worse than the elements in active queue,
	     so make it inactive *)
	  Inactive.add candidates.cd_inactive (candidate.ca_raw_context_unifier, candidate.ca_quality)
	end
      end;

      None
    with
      | Inactive.OVERFLOW
      | Active.OVERFLOW ->
	  raise Config.NO_SOLUTION
	  

(* build an assert candidate from a context unifier. *)
let build_candidate
  (candidates: candidates) (raw_context_unifier: raw_context_unifier)
  (old_quality: quality option) (* rebuild an inactive candidate and reuse its quality *)
  : candidate option =
  let subst =
    Context_unifier.recompute_unifier raw_context_unifier
  in

  let input_partners =
    raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_input_partners
  in

  (* find the assert literal input partner *)
  let assert_literal_partner =
    let rec check_at (index: int) : literal =
      if index >= Array.length input_partners then
	failwith "Selection_assert.build_candidate"
      else begin
	let input_partner =
	  input_partners.(index)
	in
	  if raw_context_unifier.Context_unifier.rcu_context_partners.(input_partner.Context_unifier.ip_index)
	    ==
	    Context_unifier.assert_partner
	  then
	    input_partner.Context_unifier.ip_literal
	  else
	    check_at (index + 1)
      end
    in
      check_at 0
  in

  let assert_literal =
    Subst.apply_to_literal
      subst
      assert_literal_partner
      Subst.input_literal_offset
  in

    match Context.check_subsumed assert_literal with
      | Some element ->
	  if not (
	    Context.is_element_mandatory_for_state
		    element
		    (Context_unifier.get_creation_state raw_context_unifier)
	  ) then begin
	    candidates.cd_invalid <-
	    (raw_context_unifier, element.Context.el_state) :: candidates.cd_invalid
	  end;
	  
	  None
	    
      | _ ->
	  let quality =
	    match old_quality with
	      | Some quality ->
		  quality
		    
	      | None ->
		  {
		    cq_term_weight = Term_quality.weight_of_literal assert_literal;
		    cq_generation = Context_unifier.generation_of_candidate
				      raw_context_unifier.Context_unifier.rcu_context_partners;
		  }
	  in
	  let candidate = {
	    ca_raw_context_unifier = raw_context_unifier;
	    ca_literal = assert_literal;
	    ca_quality = quality;
	  }
	  in
	    if Config.get_show_computed_assert_candidates () then begin
	      print_string (candidate_to_string "New Assert Candidate" candidate);
	    end;

	    add_valid_candidate candidates candidate




let rec remove_invalid_active_candidates (candidates: candidates) : unit =

  if Active.empty candidates.cd_active then
    ()

  else begin
    let candidate =
      Active.min candidates.cd_active
    in
      if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
	let removed_candidate =
	  Active.remove_min candidates.cd_active
	in
	  remove_invalid_active_candidates candidates;
      end
	
      else begin
	match Context.check_subsumed candidate.ca_literal with
	  | Some element ->
	      if
		Context.is_element_mandatory_for_state
		  element
		  (Context_unifier.get_creation_state candidate.ca_raw_context_unifier)
	      then begin
		()
	      end
		
	      else begin
		candidates.cd_invalid <- (candidate.ca_raw_context_unifier, element.Context.el_state) :: candidates.cd_invalid;
	      end;
	    
	      let removed_candidate =
		Active.remove_min candidates.cd_active
	      in
		Selection_lookahead.remove removed_candidate.ca_literal removed_candidate.ca_raw_context_unifier;
		remove_invalid_active_candidates candidates;
	      
	  | None ->
	      ()
      end
  end




let rec build_inactive_candidates (candidates: candidates) : candidate option =
  if Inactive.empty candidates.cd_inactive then
    None
      
  else if Active.size candidates.cd_active >= active_threshhold then
    None
      
  else
    let inactive_candidate, quality =
      Inactive.remove_min candidates.cd_inactive
    in
      begin
	if Context_unifier.is_raw_context_unifier_invalid inactive_candidate then
	  build_inactive_candidates candidates
	else
	  match build_candidate candidates inactive_candidate (Some quality) with
	    | None ->
		build_inactive_candidates candidates
		
	    | Some _ as conflicting ->
		conflicting
      end


let rec build_new_candidates (candidates: candidates) : candidate option =
  match candidates.cd_new with
    | [] ->
	None

    | candidate :: tail ->
	begin
	  match build_candidate candidates candidate None with
	    | None ->
		candidates.cd_new <- tail;
		build_new_candidates candidates
		  
	    | Some _ as conflicting ->
		candidates.cd_new <- tail;
		conflicting
	end


(* builds all candidates from the raw context unifiers *)
let build_candidates (candidates: candidates) : candidate option =
  remove_invalid_active_candidates candidates;

  match build_inactive_candidates candidates with
    | None ->
	build_new_candidates candidates

    | Some _ as conflicting ->
	conflicting



(*** add ***)


let add (candidates: candidates) (candidate: raw_context_unifier) : unit =
  candidates.cd_new <- candidate :: candidates.cd_new









(*** select ***)

	


let select (candidates: candidates) : selected_assert option =

  (* transform all unchecked candidates *)
  let candidate =
    match build_candidates candidates with
      | Some _ as conflicting ->
	  conflicting
	  
      | None ->
	  if Active.empty candidates.cd_active then
	    None
	  else begin
	    let candidate =
	      Active.remove_min candidates.cd_active
	    in
	      Selection_lookahead.remove candidate.ca_literal candidate.ca_raw_context_unifier;
	      Some candidate
	  end
  in
    match candidate with
      | None ->
	  None

      | Some candidate ->
	  (* move the chosen candidate to the used candidates *)
	  if not (Config.get_horn ()) then begin
	    candidates.cd_added <- (candidate, State.current_state ()) :: candidates.cd_added;
	  end;
	  
	  let explanation =
	    Context_unifier.explanation_of_context_partners
	      candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners
	  in
	  let creation_state =
	    Context_unifier.get_creation_state candidate.ca_raw_context_unifier
	  in
	    Some {
	      Selection_types.sa_literal = candidate.ca_literal;
	      Selection_types.sa_mandatory_at = Some creation_state;
	      Selection_types.sa_explanation = explanation;
	      Selection_types.sa_generation = candidate.ca_quality.cq_generation;
	      Selection_types.sa_raw_context_unifier = Some candidate.ca_raw_context_unifier;
	    }




(*** backtrack ***)

(* all unbuild candidates must be add after the last split decision -
   and as a close always invalidates the most recent split decision
   (it uses a context literal added in the splits state)
   it also invalidates all these candidates *)
let backtrack_new (candidates: candidates) : unit =
  candidates.cd_new <- []


let backtrack_invalid (candidates: candidates) : unit =
  let old_invalid =
    candidates.cd_invalid
  in
    candidates.cd_invalid <- [];

    (* this may implicitly readd candidates which are not contradictory with conflicting_state
       but another context literal to the invalid candidates. *)
    let still_invalid =
      List.find_all
	(fun (raw_context_unifier, conflicting_state) ->
	   if Context_unifier.is_raw_context_unifier_invalid raw_context_unifier then begin
	     false
	   end
	     
	   else if State.is_state_invalid conflicting_state then begin
	     add candidates raw_context_unifier;
	     false
	   end
	     
	   else
	     true
	)
	old_invalid
    in
      candidates.cd_invalid <- candidates.cd_invalid @ still_invalid

(* garbage collection *)
let backtrack_added (candidates: candidates) : unit =
  candidates.cd_added <-
  List.find_all
    (fun (candidate, add_state) ->
       if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
	 false
       end

       else if State.is_state_invalid add_state then begin
	 if Config.debug && Selection_lookahead.check candidate.ca_literal candidate.ca_raw_context_unifier then
	   failwith "Selection_assert.backtrack_added";

	 add_active_candidate candidates candidate;
	 false
       end

       else
	 true
    )
    candidates.cd_added



let backtrack (candidates: candidates) : unit =
  backtrack_new candidates;

  backtrack_invalid candidates;

  backtrack_added candidates
