(* $Id: selection_lookahead.ml,v 1.3.2.2 2004/06/14 05:28:15 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)



type term = Term.term
type literal = Term.literal
type raw_context_unifier = Context_unifier.raw_context_unifier

(* discrimination trees are faster than substitution trees
   for lots of index maintenance operations *)
module Index =
  Discrimination_tree.Make ( 
    struct
      type t = raw_context_unifier

      let is_equal (first: t) (second: t) : bool =
	first == second

      let to_string (element: t) : string =
	Context_unifier.raw_context_unifier_to_string element
    end
  )

module Indexes =
  Hashtbl.Make (
    struct
      type t = bool * term

      let equal (sign1, term1) (sign2, term2) : bool =
	(sign1 = sign2)
	&&
	match term1, term2 with
	  | Term.Const symbol1, Term.Const symbol2
	  | Term.Func (symbol1, _, _), Term.Func (symbol2, _, _) ->
	      Symbol.equal symbol1 symbol2

	  | _ ->
	      false

      let hash (_, term) : int =
	match term with
	  | Term.Const symbol
	  | Term.Func (symbol, _, _) ->
	      Symbol.id_of_symbol symbol

	  | _ ->
	      failwith "Context.Indexes.hash"
    end
  )




(*** globals ***)

(* store the assert/unit split candidates in one index
   for each predicate symbol *)
let unit_candidates : Index.t Indexes.t =
  Indexes.create 16





(*** functions ***)


let init () : unit =
  Indexes.clear unit_candidates



let backtrack () : unit =
  (* for each index... *)
  Indexes.iter
    (fun literal index ->
       let remove_entries =
	 ref []
       in
	 (* ...find all invalid candidates... *)
	 Index.iter
	   (fun term raw_context_unifier ->
	      if Context_unifier.is_raw_context_unifier_invalid raw_context_unifier then
		remove_entries := (term, raw_context_unifier) :: !remove_entries
	   )
	   index;

	 (* ...and remove them *)
	 List.iter
	   (fun (term, raw_context_unifier) ->
	      if not (Index.remove index term (Some raw_context_unifier)) then begin
		print_endline (Term.term_to_string term);
		print_endline (Context_unifier.raw_context_unifier_to_string raw_context_unifier);
		failwith "Selection_lookahead.backtrack"
	      end
	   )
	   !remove_entries
    )
    unit_candidates




let add (literal: literal) (raw_context_unifier: raw_context_unifier) : unit =
  let index =
    try
      Indexes.find unit_candidates
	(literal.Term.sign, literal.Term.term)
    with
      | Not_found ->
	  let new_index =
	    Index.create ()
	  in
	    Indexes.add
	      unit_candidates
	      (literal.Term.sign, literal.Term.term)
	      new_index;
	    new_index
  in
    Index.add index literal.Term.term raw_context_unifier



let remove (literal: literal) (raw_context_unifier: raw_context_unifier) : unit =
  let index =
    try
      Indexes.find unit_candidates
	(literal.Term.sign, literal.Term.term)
    with
      | Not_found ->
	  failwith "Selection_lookahead.remove I";
  in
    if not (Index.remove index literal.Term.term (Some raw_context_unifier)) then begin
      failwith "Selection_lookahead.remove II"
    end





let check (literal: literal) (raw_context_unifier: raw_context_unifier) : bool =
  try
    let index =
      Indexes.find unit_candidates
	(not literal.Term.sign, literal.Term.term)
    in
      begin
	match Index.find_unifiable ~p_preserving:true index literal.Term.term with
	  | None ->
	      false
		
	  | Some conflicting_raw_context_unifier ->
	      Statistic.inc_debug ();
	      true
      end
	
  with
    | Not_found ->
	false
  
