(* $Id: selection_split.ml,v 1.51.2.3 2004/06/14 05:28:15 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type state = State.state
type literal = Term.literal
type subst = Subst.subst
type explanation = State.explanation
type context_unifier_space = Context_unifier.context_unifier_space
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected_assert = Selection_types.selected_assert
type selected_split = Selection_types.selected_split

type quality = {
  (* properties of the remainder *)
  rq_generation: int;
  rq_size: int;

  (* properties of the best remainder literal *)
  rq_is_parametric: bool;
  rq_term_weight: int;
}

(* actually only a remainder literal if rl_is_remainder_literal is true.
   otherwise just a non-remainder literal computed by the context unifier *)
type remainder_literal = {
  (* the context literal used to computed the remainder literal *)
  rl_context_literal: Context.element;

  (* the remainder literal *)
  rl_remainder_literal: literal;

  (* really a remainder literal? *)
  rl_is_remainder_literal: bool;
}


(* actually all literals of a context unifier, not only the remainder *)
type remainder = {
  (* the computed context unifier *)
  rm_raw_context_unifier: raw_context_unifier;

  (* the 'best' remainder literal *)
  rm_best_remainder_literal: literal;

  rm_quality: quality;
}



(* compare two remainders *)
let compare_qualities (first: quality) (second: quality) : int =
  let pureness =
    Term_quality.cmp_pureness first.rq_is_parametric second.rq_is_parametric
  in
    if pureness <> 0 then
      pureness
    else
      let size =
        compare first.rq_size second.rq_size
      in
        if size <> 0 then
          size
        else
 	  let term_weight =
	    compare first.rq_term_weight second.rq_term_weight
	  in
	    if term_weight <> 0 then
	      term_weight
	    else
	      compare first.rq_generation second.rq_generation

let compare_candidates (first: remainder) (second: remainder) : int =
  compare_qualities first.rm_quality second.rm_quality



module Active =
  Heap.Make (
    struct
      type t = remainder

      let compare x y = compare_candidates x y
    end
  )

module Inactive =
  Heap.Make ( 
    struct
      type t = raw_context_unifier * quality

      let compare (_, first) (_, second) : int =
	compare_qualities first second
    end
  )

(* remainders whose best split literal are of this depth *)
type candidates = {
  (* candidates not considered yet *)
  mutable cd_new: raw_context_unifier list;

  (* the remainders currently valid for adding to the context
     (at least valid when last checked). *)
  cd_active: Active.t;

  (* to save memory only a limited amount of candidates is stored in cd_active.
     the others are kept here with their explanation *)
  cd_inactive: Inactive.t;

  (* to save memory only a limited amount of candidates is stored
     with full informatio nin cd_active,
     the others are kept here *)
  mutable cd_invalid: (raw_context_unifier * state) list;

  (* candidates added to the context (at step) *)
  mutable cd_added: (remainder * state ref) list;
}


type selected =
  | Assert of Selection_types.selected_assert
  | Split of Selection_types.selected_split



(* how many active candidates may be stored?
   other are temporarily kept inactive. *)
let active_threshhold =
  1024 * 8








(*** comparison ***)


    


        
(* finds the best remainder literal *)
let best_remainder_literal (literals: remainder_literal list) :
  (remainder_literal * bool * int) =
  let best =
    List.fold_left
      (fun acc remainder_literal ->
	 if remainder_literal.rl_is_remainder_literal then begin
	   match acc with
	     | None ->
		 let depth =
		   Term_quality.depth_of_literal remainder_literal.rl_remainder_literal
		 in
		   (* ignore literals exceeding the depth bound *)
		   if depth > Config.get_current_depth_bound () then
		     None
		   else
		     Some (
		       remainder_literal,
		       Term_quality.is_parametric remainder_literal.rl_remainder_literal,
		       Term_quality.weight_of_literal remainder_literal.rl_remainder_literal
		     )
		   
		 
	     | Some (best_remainder_literal, best_pureness, best_weight) ->
		 (* ignore literals exceeding the depth bound *)
		 let depth =
		   Term_quality.depth_of_literal remainder_literal.rl_remainder_literal
		 in
		   if depth > Config.get_current_depth_bound () then
		     acc
		   else
		     let pureness =
		       Term_quality.is_parametric remainder_literal.rl_remainder_literal
		     in
		     let cmp_pureness =
		       Term_quality.cmp_pureness best_pureness pureness
		     in
		       if cmp_pureness < 0 then
			 acc
		       else if cmp_pureness > 0 then
			 Some (
			   remainder_literal,
			   pureness,
			   Term_quality.weight_of_literal remainder_literal.rl_remainder_literal
			 )
		       else
			 let weight =
			   Term_quality.weight_of_literal remainder_literal.rl_remainder_literal
			 in
			 let cmp_weight =
			   compare best_weight weight
			 in
			   if cmp_weight < 0 then
			     acc
			   else if cmp_weight > 0 then
			     Some (
			       remainder_literal,
			       pureness,
			       weight
			     )
			   else
			     acc

	 end
	   
	 else
	   acc
      )
      None
      literals
  in

    match best with
      | None ->
	  failwith "best_remainder_literal"
	    
      | Some best ->
	  best



(*** creation ***)


let create () : candidates =
  {
    cd_new = [];
    cd_active = Active.create ();
    cd_inactive = Inactive.create ();
    cd_invalid = [];
    cd_added = []
  }




(*** build ***)



let build_remainder_literals (raw_context_unifier: raw_context_unifier) : remainder_literal list =
  let subst =
    Context_unifier.recompute_unifier raw_context_unifier
  in
    
  let remainder_states =
    Context_unifier.compute_remainder_states raw_context_unifier subst
  in

  let admissible =
    Admissible.make_admissible raw_context_unifier subst remainder_states
  in

    if Config.debug then begin
      let remainder_states_admissible =
	Context_unifier.compute_remainder_states raw_context_unifier subst
      in
	if not (remainder_states = remainder_states_admissible) then begin
	  failwith "Selection_split.remainder_states";
	end;
    end;

    let input_partners =
      raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_input_partners
    in

    let rec build_at (i : int) (acc: remainder_literal list) : remainder_literal list =
      if i >= Array.length input_partners then
	acc

      else begin
	let input_partner =
	  input_partners.(i)
	in
	  
	let index =
	  input_partner.Context_unifier.ip_index
	in
	  
	let context_literal =
	  raw_context_unifier.Context_unifier.rcu_context_partners.(index).Context_unifier.cp_element;
	in
	  
	let split_literal =
	  Subst.apply_to_literal 
	    admissible 
	    input_partner.Context_unifier.ip_literal
	    Subst.input_literal_offset 
	in
	  
	let remainder_literal = {
	  rl_context_literal = context_literal;
	  rl_remainder_literal = split_literal;
	  rl_is_remainder_literal = remainder_states.(index);
	}
	in
	  build_at (i + 1) (remainder_literal :: acc)
      end
    in
      build_at 0 []








(*** representation ***)

let remainder_literals_to_string_ (raw_context_unifier: raw_context_unifier)
  (remainder_literals: remainder_literal list) : string =
  let _, remainder_literals_repr =
    Array.fold_right
      (fun context_partner (remainder_literals, remainder_repr) ->
	 try
	   (* we need to remove the remainder literal as it is identified by the context literal
	      and a clause might contain the same context literal more than once *)
	   let remainder_literal =
	     List.find
	       (fun remainder_literal ->
		  context_partner.Context_unifier.cp_element == remainder_literal.rl_context_literal
	       )
	       remainder_literals
	   in
	   let remaining_remainder_literals =
	     List.find_all
	       (fun remaining_remainder_literal ->
		  remainder_literal != remaining_remainder_literal
	       )
	       remainder_literals
	   in

	   let repr =
	     if remainder_literal.rl_is_remainder_literal then
	       Term.literal_to_string remainder_literal.rl_remainder_literal
	     else
	       "___"
	   in
	     (remaining_remainder_literals, repr :: remainder_repr)
	 with
	   | Not_found ->
	       failwith "remainder_literals_to_string"
      )
      raw_context_unifier.Context_unifier.rcu_context_partners
      (remainder_literals, [])
  in
    "Remain.: " ^ "[" ^ String.concat ", " remainder_literals_repr ^ "]" ^ "\n"


let remainder_literals_to_string (raw_context_unifier: raw_context_unifier) : string =

  let remainder_literals =
    build_remainder_literals raw_context_unifier
  in
    remainder_literals_to_string_ raw_context_unifier remainder_literals




let remainder_to_string_ (title: string)
  (remainder: remainder) (remainder_literals: remainder_literal list) : string =
    State.current_state_to_string ()
    ^ title ^ ": " ^ Term.literal_to_string remainder.rm_best_remainder_literal ^ "\n"
    ^
    if Config.get_show_context_unifier () then
      Context_unifier.raw_context_unifier_to_string remainder.rm_raw_context_unifier
      ^ remainder_literals_to_string_ remainder.rm_raw_context_unifier remainder_literals
    else
      ""

let remainder_to_string (title: string) (remainder: remainder) : string =
  let remainder_literals =
    build_remainder_literals remainder.rm_raw_context_unifier
  in
    remainder_to_string_ title remainder remainder_literals





(*** context checks ***)

let rec check_context
  (func: remainder_literal -> Context.element option)
  (remainder_literals: remainder_literal list)
  (remainder_state: state)
  (acc: Context.element option) :
  Context.element option =

  match remainder_literals with
    | [] ->
	acc
	
    | remainder_literal :: tail ->
	let check_status =
	  func remainder_literal
	in
	  begin
	    match check_status with
	      | None ->
		  check_context func tail remainder_state acc
	    
	      | Some conflicting_element ->
		  check_status
	  end


let check_subsumed_func (remainder_literal: remainder_literal) : Context.element option =
  if remainder_literal.rl_is_remainder_literal then
    Context.check_subsumed remainder_literal.rl_remainder_literal
  else
    None


let check_productive_func (remainder_literal: remainder_literal) : Context.element option =
  if Term_quality.is_parametric remainder_literal.rl_context_literal.Context.el_literal then begin
    Context.check_productive
      remainder_literal.rl_context_literal.Context.el_literal
      remainder_literal.rl_remainder_literal
  end

  else
    None


let check_remainder_literals
  (remainder_literals: remainder_literal list) (best_remainder_literal: literal) (remainder_state: state) :
  Context.element option =

  let subsumed_state =
    check_context check_subsumed_func remainder_literals remainder_state None
  in
    match subsumed_state with
      | Some _ ->
	  subsumed_state
	    
      | None ->
	  let productive_state =
	    check_context check_productive_func remainder_literals remainder_state subsumed_state
	  in
	    match productive_state with
	      | Some _ ->
		  productive_state
		  
	      | None ->
		  Context.check_contradictory best_remainder_literal




let check_remainder (remainder: remainder) : Context.element option =
  let remainder_literals =
    build_remainder_literals remainder.rm_raw_context_unifier
  in
    check_remainder_literals
      remainder_literals
      remainder.rm_best_remainder_literal
      (Context_unifier.get_creation_state remainder.rm_raw_context_unifier)





(*** build remainers ***)



let add_active_candidate (candidates: candidates) (candidate: remainder) : unit =
  if candidate.rm_quality.rq_size = 1 then
    Selection_lookahead.add candidate.rm_best_remainder_literal candidate.rm_raw_context_unifier;

  Active.add candidates.cd_active candidate



(* add a valid candidate to the active/inactive queue *)
let add_valid_candidate (candidates: candidates) (candidate: remainder) : remainder option =
  if
    (candidate.rm_quality.rq_size = 1)
    &&
    (Selection_lookahead.check candidate.rm_best_remainder_literal candidate.rm_raw_context_unifier)
  then
    Some candidate
  else
    try
      if Active.size candidates.cd_active < active_threshhold then begin
	(* active queue is not full, so add here *)
	add_active_candidate candidates candidate
      end
      else begin
	
	if compare_candidates (Active.last candidates.cd_active) candidate > 0 then begin
	  add_active_candidate candidates candidate;
	  
	  (* candidate is better than an element in active queue,
	     so switch them *)
	  let last_heap_element =
	    Active.remove_last candidates.cd_active
	  in
	    (* silently drop invalid candidates *)
	    if not (Context_unifier.is_raw_context_unifier_invalid last_heap_element.rm_raw_context_unifier) then begin
	      if last_heap_element.rm_quality.rq_size = 1 then
		Selection_lookahead.remove
		  last_heap_element.rm_best_remainder_literal
		  last_heap_element.rm_raw_context_unifier;
	      
	      Inactive.add candidates.cd_inactive (last_heap_element.rm_raw_context_unifier, last_heap_element.rm_quality)
	    end
	end
	  
	else begin
	  (* candidate is worse than the elements in active queue,
	     so make it inactive *)
	  Inactive.add candidates.cd_inactive (candidate.rm_raw_context_unifier, candidate.rm_quality)
	end
      end;

      None
    with
      | Inactive.OVERFLOW
      | Active.OVERFLOW ->
	  raise Config.NO_SOLUTION






let build_candidate
  (candidates: candidates) (raw_context_unifier: raw_context_unifier)
  (old_quality: quality option)
  : remainder option =

  let remainder_literals =
    build_remainder_literals raw_context_unifier
  in
    
  let best_literal, pureness, weight =
    best_remainder_literal remainder_literals
  in
	    
  let best_remainder_literal =
    best_literal.rl_remainder_literal;
  in

  let creation_state =
    Context_unifier.get_creation_state raw_context_unifier
  in
    match check_remainder_literals remainder_literals best_remainder_literal creation_state with
      | Some conflicting_element ->
	  if not (Context.is_element_mandatory_for_state conflicting_element creation_state) then begin
	    candidates.cd_invalid <-
	    (raw_context_unifier, conflicting_element.Context.el_state) :: candidates.cd_invalid
	  end;
	  
	  None
	    
      | _ ->
	  let size =
	    List.fold_left
	      (fun acc remainder_literal ->
		 if remainder_literal.rl_is_remainder_literal then
		   acc + 1
		 else
		   acc
	      )
	      0
	      remainder_literals
	  in
	  let quality = {
	    rq_generation = Context_unifier.generation_of_candidate
			      raw_context_unifier.Context_unifier.rcu_context_partners;
	    rq_size = size;
	    rq_is_parametric = pureness;
	    rq_term_weight = weight;
	  }
	  in
	  let candidate = {
	    rm_raw_context_unifier = raw_context_unifier;
	    rm_best_remainder_literal = best_remainder_literal;
	    rm_quality = quality;
	  }
	  in
	    if Config.get_show_computed_split_candidates () then begin
	      print_string (remainder_to_string_ "New Split Candidate" candidate remainder_literals);
	    end;
	    
	    add_valid_candidate candidates candidate
	



    

(*** fill the active candidates buffer ***)


let rec remove_invalid_active_candidates (candidates: candidates) : unit =

  if Active.empty candidates.cd_active then
    ()

  else begin
    let candidate =
      Active.min candidates.cd_active
    in
      if Context_unifier.is_raw_context_unifier_invalid candidate.rm_raw_context_unifier then begin
	let removed_candidate =
	  Active.remove_min candidates.cd_active
	in
	  remove_invalid_active_candidates candidates;
      end
	
      else begin
	match check_remainder candidate with
	  | Some element ->
	      begin
		if Context.is_element_mandatory_for_state
		  element
		  (Context_unifier.get_creation_state candidate.rm_raw_context_unifier)
		then
		  ()
		else
		  candidates.cd_invalid <-
		  (candidate.rm_raw_context_unifier, element.Context.el_state) :: candidates.cd_invalid
	      end;

	      let removed_candidate =
		Active.remove_min candidates.cd_active
	      in
		if removed_candidate.rm_quality.rq_size = 1 then
		  Selection_lookahead.remove removed_candidate.rm_best_remainder_literal removed_candidate.rm_raw_context_unifier;
		remove_invalid_active_candidates candidates;
		
	  | None ->
	      ()
      end
  end




let rec build_inactive_candidates (candidates: candidates) : remainder option =
  if Inactive.empty candidates.cd_inactive then
    None
      
  else if Active.size candidates.cd_active >= active_threshhold then
    None
      
  else
    let inactive_candidate, quality =
      Inactive.remove_min candidates.cd_inactive
    in
      begin
	if Context_unifier.is_raw_context_unifier_invalid inactive_candidate then
	  build_inactive_candidates candidates
	else
	  match build_candidate candidates inactive_candidate (Some quality) with
	    | None ->
		build_inactive_candidates candidates
		
	    | Some _ as conflicting ->
		conflicting
      end


let rec build_new_candidates (candidates: candidates) : remainder option =
  match candidates.cd_new with
    | [] ->
	None

    | candidate :: tail ->
	begin
	  match build_candidate candidates candidate None with
	    | None ->
		candidates.cd_new <- tail;
		build_new_candidates candidates
		  
	    | Some _ as conflicting ->
		candidates.cd_new <- tail;
		conflicting
	end


(* builds all candidates from the raw context unifiers *)
let build_candidates (candidates: candidates) : remainder option =
  remove_invalid_active_candidates candidates;

  match build_inactive_candidates candidates with
    | None ->
	build_new_candidates candidates

    | Some _ as conflicting ->
	conflicting














(*** add ***)
let add (candidates: candidates) (candidate: raw_context_unifier) : unit =
  candidates.cd_new <- candidate :: candidates.cd_new








(*** select ***)




let select (candidates: candidates) : selected option =

  (* transform all unchecked candidates *)
  let remainder =
    match build_candidates candidates with
      | Some _ as conflicting ->
	  conflicting
	  
      | None ->
	  if Active.empty candidates.cd_active then
	    None
	  else begin
	    let remainder =
	      Active.remove_min candidates.cd_active
	    in
	      if remainder.rm_quality.rq_size = 1 then
		Selection_lookahead.remove remainder.rm_best_remainder_literal remainder.rm_raw_context_unifier;

	      Some remainder
	  end
  in
    match remainder with
      | None ->
	  None

      | Some remainder ->
	  let explanation =
	    Context_unifier.explanation_of_context_partners
	      remainder.rm_raw_context_unifier.Context_unifier.rcu_context_partners
	  in
	  let valid =
	    ref (State.current_state ())
	  in
	    candidates.cd_added <- (remainder, valid) :: candidates.cd_added;
	    
	    if (*false &&*) remainder.rm_quality.rq_size <= 1 then begin
	      (* unit split - valid until state is backtracked *)
	      Some (
		Assert {
		  Selection_types.sa_literal = remainder.rm_best_remainder_literal;
		  Selection_types.sa_mandatory_at = None;
		  Selection_types.sa_explanation = explanation;
		  Selection_types.sa_generation = remainder.rm_quality.rq_generation;
		  Selection_types.sa_raw_context_unifier = Some remainder.rm_raw_context_unifier;
		}
	      )
	    end
	      
	    else begin
	      (* real split - valid until left and right split are invalid *)
	      Some (
		Split {
		  Selection_types.ss_literal = remainder.rm_best_remainder_literal;
		  Selection_types.ss_explanation = explanation;
		  Selection_types.ss_valid = valid;
		  Selection_types.ss_generation = remainder.rm_quality.rq_generation;
		  Selection_types.ss_raw_context_unifier = remainder.rm_raw_context_unifier;
		}
	      )
	    end








(*** backtrack ***)

(* when using Selection_lookahead a candidate may be chosen
   before all candidates are built,
   thus some still valid candidates might be left *)
let backtrack_new (candidates: candidates) : unit =
  candidates.cd_new <-
  List.find_all
    (fun raw_context_unifier ->
       not (Context_unifier.is_raw_context_unifier_invalid raw_context_unifier)
    )
    candidates.cd_new


let rec backtrack_invalid (candidates: candidates) : unit =
  let old_invalid =
    candidates.cd_invalid
  in
    candidates.cd_invalid <- [];

    (* this may implicitly readd candidates which are not contradictory with conflicting_state
       but another context literal to the invalid candidates. *)
    let still_invalid =
      List.find_all
	(fun (raw_context_unifier, conflicting_state) ->
	   if Context_unifier.is_raw_context_unifier_invalid raw_context_unifier then
	     false
	       
	   else if State.is_state_invalid conflicting_state then begin
	     add candidates raw_context_unifier;
	     false
	   end
	     
	   else
	     true
	)
	old_invalid
    in
      candidates.cd_invalid <- candidates.cd_invalid @ still_invalid



let backtrack_added (candidates: candidates) : unit =
  candidates.cd_added <-
  List.find_all
    (fun (candidate, is_state_valid) ->
       if Context_unifier.is_raw_context_unifier_invalid candidate.rm_raw_context_unifier then begin
	 false
       end

       else if (State.is_state_invalid !is_state_valid) then begin
	 try
	   if
	     Config.debug
	     &&
	     candidate.rm_quality.rq_size = 1
	     &&
	     Selection_lookahead.check candidate.rm_best_remainder_literal candidate.rm_raw_context_unifier then
	       failwith "Selection_split.backtrack_added";

	   add_active_candidate candidates candidate;
	   false
	 with
	   | Inactive.OVERFLOW
	   | Active.OVERFLOW ->
	       raise Config.NO_SOLUTION
       end

       else
	 true
    )
    candidates.cd_added



let backtrack (candidates: candidates) : unit =
  backtrack_new candidates;

  backtrack_invalid candidates;

  backtrack_added candidates

