(* $Id: state.ml,v 1.17 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(*** types ***)


type literal = Term.literal

(* explanations are alway kept decreasingly sorted,
   i.e. youngest states first *)
type explanation = state list

and state = {
  (* states are enumerated by age *)
  mutable st_id: int;
  (* is this state valid? *)
  mutable st_valid: bool;


  (* the following two are only needed for non-chronological backtracking *)

  (* the transitive closure of all states this state depends on -
     for dependency testing. *)
  mutable st_dependend_on: explanation;
  (* all states which immediately depend on this state -
     for invalidating states.  *)
  mutable st_base_for: state list;
}

(* basically copied from Term_database -
   but here literal_equal can be used *)
module LiteralTable =
  Hashtbl.Make (
    struct
      type t = literal

      let equal (literal1: literal) (literal2: literal) : bool =
	Term.literal_equal literal1 literal2

      let hash (literal: literal) : int =
	match literal.Term.term with
	  | Term.Var var ->
	      failwith ("State.LiteralTable.hash");
	      Var.id_of_var var
	      
	  | Term.Const symbol ->
	      Symbol.id_of_symbol symbol
	      
	  | Term.Func (_, _, hash_value) ->
	      hash_value

    end
  )


(*** constants ***)

let empty_explanation : explanation =
  []

let valid_state : state = {
  st_id = -1;
  st_valid = true;
  st_dependend_on = empty_explanation;
  st_base_for = [];
}

let invalid_state : state = {
  st_id = -1;
  st_valid = false;
  st_dependend_on = empty_explanation;
  st_base_for = [];
}




(*** globals ***)

(* the current states *)
let states =
  let first_state =
    {
      st_id = 0;
      st_valid = true;
      st_dependend_on = empty_explanation;
      st_base_for = [];
    }
  in
    ref [first_state]


(* the active state, i.e. the youngest state *)
let current_state () : state =
  match !states with
    | [] ->
	failwith "State.current_state"

    | head :: _ ->
	head


(* the explanations of context literals *)
let (literal_explanations: explanation LiteralTable.t) =
  let literal_explanations =
    LiteralTable.create 1024
  in
    (* v has no explanation *)
    LiteralTable.add literal_explanations (Term.get_default_v ()) [current_state ()];
    literal_explanations













(*** representation ***)

let state_to_string (state: state) : string =
  "[" ^ string_of_int state.st_id ^ "] "

let current_state_to_string () : string =
  "[" ^ string_of_int (current_state ()).st_id ^ "] "

let explanation_to_string (explanation: explanation) : string =
  "[" ^ String.concat ", " (List.map state_to_string explanation) ^ "]"

let state_to_string_extended (state: state) : string =
  "[" ^ string_of_int state.st_id ^ "] :"
  ^ "\nValid: " ^ string_of_bool state.st_valid
  ^ "\nDependend On: " ^ explanation_to_string state.st_dependend_on
  ^ "\nBase For: " ^ explanation_to_string state.st_base_for

let current_states_to_string () : string =
  List.fold_left
    (fun acc state ->
       state_to_string_extended state ^ "\n" ^ acc
    )
    ""
    !states







(*** properties ***)

let parent_state (state: state) : state =

  let rec do_find states =
    match states with
      | [] ->
	  failwith "parent_state.unknown state"

      | head :: parent :: _ when
	  head == state ->
	  parent

      | head :: [] when
	  head == state ->
	  raise Not_found

      | head :: tail ->
	  do_find tail

  in
    do_find !states

let is_state_valid (state: state) : bool =
  state.st_valid

let is_state_invalid (state: state) : bool =
  not state.st_valid

let is_explanation_invalid (explanation: explanation) : bool =
  List.exists is_state_invalid explanation

let is_closing_explanation (explanation: explanation) : bool =
  match explanation with
    | [] ->
	true

    | _ ->
	false

let contains_explanation_state (explanation: explanation) (state: state) : bool =
  List.memq state explanation

(* List.memq actually works quite fine -
   more sophisticated approaches like
   - building and traversing a tree over the state ids
   - or assigning a prime number to each state,
     computing the factor of the primes of all states it depends on,
     and then checking if base divides the factor
   proved to be worse *)
let depends_on (dependend: state) (base: state) : bool =
  if Config.get_use_chronological_backtracking () then
    base.st_id <= dependend.st_id
  else
    List.memq base dependend.st_dependend_on


let compare_age (first: state) (second: state) : int =
  compare first.st_id second.st_id




(*** Comparison ***)

let state_equal (first: state) (second: state) : bool =
  first == second


let compare_states (first: state) (second: state) : int =
  compare first.st_id second.st_id

(* explanations are decreasingly sorted.
   older explanations are preferred. *)
let compare_explanations (_first: explanation) (_second: explanation) : int =

  let rec do_compare_explanations (first: explanation) (second: explanation) =
    match first, second with
      | [], [] ->
	  0

      | [], _ ->
	  -1
	  
      | _, [] ->
	  1
	  
      | first_head :: tail1, second_head :: tail2 ->
	  if first_head.st_id < second_head.st_id then
	    -1
	  else if first_head.st_id > second_head.st_id then
	    1
	  else
	    do_compare_explanations tail1 tail2
  in
    do_compare_explanations _first _second







(*** explanations ***)

(* explanations are decreasingly sorted *)
let rec merge_explanations (first: explanation) (second: explanation) : explanation =
  match first, second with
    | [], _ ->
	second
	  
    | _, [] ->
	first
	  
    | head1 :: tail1, head2 :: tail2 ->
	begin
	  match compare head1.st_id head2.st_id with
	    | -1 ->
		head2 :: merge_explanations first tail2
		  
	    | 0 ->
		head1 :: merge_explanations tail1 tail2
		  
	    | _ ->
		head1 :: merge_explanations tail1 second
	end


let get_explanation (literals: literal list) : explanation =
  List.fold_left
    (fun acc literal ->
       try
	 let new_explanation =
	   LiteralTable.find literal_explanations literal
	 in
	   merge_explanations acc new_explanation
       with
	 | Not_found ->
	     failwith ("Explanation.get_explanation: " ^ Term.literal_to_string literal)
    )
    []
    literals


let rec prune_up_to_state (explanation: explanation) (state: state) : explanation * bool =
  match explanation with
    | [] ->
	[], false

    | head :: tail ->
	(* head is older than state, so state is not part of the explanation *)
	if state.st_id > head.st_id then
	  explanation, false

	(* found it *)
	else if state_equal state head then
	  tail, true

	else
	  prune_up_to_state tail state










(*** state access ***)

(* register a state by all states in its direct explanation *)
let register_dependend_state (dependend: state) (explanation: explanation) : unit =
  List.iter
    (fun state ->
       match state.st_base_for with
	 | [] ->
	     state.st_base_for <- dependend :: state.st_base_for
	       
	 | head :: _ ->
	     if not (state_equal dependend head) then
	       state.st_base_for <- dependend :: state.st_base_for
    )
    explanation




(* the reasons for an assert are the reasons for its creating literals *)
let set_assert (literal: literal) (explanation: explanation) =
  if Config.debug && LiteralTable.mem literal_explanations literal then begin
    failwith ("State.set_assert: " ^ Term.literal_to_string literal)
  end;

  if Config.debug && is_explanation_invalid explanation then begin
    failwith "State.set_assert.invalid";
  end;

  let state =
    current_state ()
  in

(* the current state must be part of the explanation.
   this is not automatically the case for non-chronologicall backtracking.
   an assert literal might on creation be subsumed by an older context literal,
   but later on this other literal becomes retracted
   and now the assert literal becomes asserted *)
  let literal_explanation =
    match explanation with
      | [] ->
	  [state]

      | head :: _ ->
	  if state_equal head state then
	    explanation
	  else
	    state :: explanation
  in
    LiteralTable.add literal_explanations literal literal_explanation;

    if not (Config.get_use_chronological_backtracking ()) then begin
      register_dependend_state state explanation;

      state.st_dependend_on <- 
        merge_explanations state.st_dependend_on explanation
    end


(* a split *)
let create_state (literal: literal) (explanation: explanation) : state =
  let new_state = {
    st_id = List.length !states;
    st_valid = true;
    st_dependend_on = [];
    st_base_for = [];
  }
  in
    LiteralTable.add literal_explanations literal [new_state];

    if not (Config.get_use_chronological_backtracking ()) then begin
      register_dependend_state new_state explanation;

      new_state.st_dependend_on <-  new_state :: explanation;
    end;

    states := new_state :: !states;
    new_state







(*** backtracking ***)


let rec invalidate_state (state: state) : unit =
  if state.st_valid then begin
    state.st_valid <- false;

    if Config.get_use_chronological_backtracking () then begin
      (* only the youngest in a sequence *)
      List.iter
	(fun invalidate_state ->
	   if invalidate_state.st_id > state.st_id then
	     invalidate_state.st_valid <- false;
	)
	!states;
    end
    else begin
      (* recurvisely *)
      List.iter invalidate_state state.st_base_for;
    end
  end




let backtrack_states () : unit =

  let rec do_backtrack_chronologically (states: state list) : state list =
    match states with
      | [] ->
	  []

      | state :: tail ->
	  if is_state_valid state then begin
	    if Config.get_show_derivation () then begin
	      print_endline (current_state_to_string () ^ "Backtrack To " ^ state_to_string state);
	    end;
	    
	    states
	  end
	  else begin
	    do_backtrack_chronologically tail
	  end
  in

  let rec do_backtrack_non_chronologically (counter: int) (states: state list) (acc: state list) : state list =
    match states with
      | [] ->
	  acc

      | state :: tail ->
	  if is_state_invalid state then begin
	    if Config.get_show_derivation () then begin
	      print_endline (current_state_to_string () ^ "Invalidating " ^ state_to_string state);
	    end;
	    
	    do_backtrack_non_chronologically counter tail acc
	  end
	  else begin
	    if Config.get_show_derivation ()
	      &&
	      state.st_id <> counter
	    then begin
	      print_endline (current_state_to_string () ^ "Renaming [" ^ string_of_int state.st_id ^ "] to [" ^ string_of_int counter ^ "]");
	    end;
	    
	    state.st_id <- List.length acc;
	    state.st_dependend_on <- List.find_all is_state_valid state.st_dependend_on;
	    state.st_base_for <- List.find_all is_state_valid state.st_base_for;

	    do_backtrack_non_chronologically (counter + 1) tail (state :: acc)
	  end
  in
    states :=
      if Config.get_use_chronological_backtracking () then
	do_backtrack_chronologically !states
      else
	do_backtrack_non_chronologically 0 (List.rev !states) []




let backtrack_literal_explanations () : unit =
  let invalid_keys =
    LiteralTable.fold
      (fun key explanation acc ->
	 if is_explanation_invalid explanation then begin
	   key :: acc
	 end
	 else
	   acc
      )
      literal_explanations
      []
  in
    List.iter
      (LiteralTable.remove literal_explanations)
      invalid_keys


let backtrack () =
  backtrack_states ();
  backtrack_literal_explanations ()






(*** restart ***)

let restart () : unit =
  List.iter
    (fun invalidate_state ->
       invalidate_state.st_valid <- false;
    )
    !states;

  let first_state = {
    st_id = 0;
    st_valid = true;
    st_dependend_on = empty_explanation;
    st_base_for = empty_explanation;
  }
  in
    states := [first_state];
    
    LiteralTable.clear literal_explanations;
    LiteralTable.add literal_explanations (Term.get_default_v ())  [first_state]
