(* $Id: state.mli,v 1.11 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** a Split decision plus subsequent Asserts
  
  A {e state} represents a Split decision (plus subsequent Asserts)
  and the dependencies ({e explanation}) of this decision on other states.

  The {e explanation} for a context literal is the set of the states
  which are responsible for the context literals which were used to compute the remainder
  and the state in which the literal was added to the context.

  But: For backtracking only the state which were immediately responsible
  for the closing context unifier are of interest.
  Thus for splitted context literals the explanation is only its state.

  An {e explanation} for a state is the set of the explanations
  for the context literals added in this state
  (splitted literals treated the same way as asserted).

  I.e.:
  - state s is created by a Split on literal l of the context unifier r
  - the context literals of r were added to the context in states m_0, ..., m_i
  -> the explanation for l is s and the set of states m_0, ..., m_i ,
     which is the set of states s, m_0, ..., m_i
  -> the current explanation for s is the explanation for l

  - now (and still in state s) literal k is asserted because of the context unifier s
  - the context literals of s were added to the context in states n_0, ..., n_j
  -> the explanation for k is the set of states s, n_0, ..., n_i
  -> the current explanation for s is the union of the explanations for l and k,
     which is the set of states s, m_0, ..., m_i, n_0, ..., n_j

  A state d is dependend on another state s if
  s is in the transitive closure of the explanations of d.

  States are enumerated by 'age', i.e. in the order of creation.

  If a states become invalid (backtracking)
  all states, context literal and remainders dependend on this state
  become also invalid (transitive).
  If in chronological backtracking a state becomes invalid
  all younger states became invalid, too.
  Thus they are implicitely dependend on every older state.

  In non-chronological backtracking this is not the case,
  only the explanation dependencies exist.

  In Backjumping, a chronological backtracking method,
  states which are not part of a closing explanation
  (i.e. the explanation of a Close context unifier)
  are not tried to be repaired by toggling the split decision.
  These states are just skipped on backtracking
  and thus automically invalidated
  when the split decision represented by a younger state,
  (which is part of the explanation)
  is toggled.
*)



(** {6 Types} *)

type literal = Term.literal
type state
type explanation


(** {6 Functions} *)


(** {2 Constants } *)

(** a valid state *)
val valid_state: state

(** an invalid state *)
val invalid_state: state

(** an empty explanation -
  i.e. the only dependency is the input formula set *)
val empty_explanation: explanation



(** {2 State Access } *)

(** returns the currently active state *)
val current_state: unit -> state

(** returns the parent state of a state.
  raises Not_found on the root state *)
val parent_state: state -> state

(** [create_state literal explanation]
  creates a new state initiated by a split on a new context literal [literal]
  with its [explanation]. *)
val create_state: literal -> explanation -> state

(** [set_assert literal explanation]
  adds a [literal] and its [explanation] to the responsibility of the current state *)
val set_assert: literal -> explanation -> unit




(** {2 Properties } *)

(** is this state valid? *)
val is_state_valid: state -> bool

(** is this state invalid? *)
val is_state_invalid: state -> bool

(** is this explanation (i.e. on of its states) invalid? *)
val is_explanation_invalid: explanation -> bool

(** is this a closing explanation,
  i.e. an empty explanation which closes the whole derivation? *)
val is_closing_explanation: explanation -> bool

(** is [state] part of [explanation]? *)
val contains_explanation_state: explanation -> state -> bool

(** [depends_on dependend base]
  does [dependend] depend on [base]?
*)
val depends_on: state -> state -> bool

(** compares the 'age' of two states,
  i.e. if a state has been created before another state.
  returns -1 if the first state is older,
  1 if the second is older,
  0 if the states are identical *)
val compare_age: state -> state -> int


(** {2 Comparison} *)

val state_equal: state -> state -> bool

(** older states are considered to be better (farther backjumping) *)
val compare_states: state -> state -> int

(** explanations using older states are considered to be better
  (farther backjumping) *)
val compare_explanations: explanation -> explanation -> int





(** {2 Explanations } *)


(** [get_explanation context_literals]
  returns the explanation for the [context_literals] *)
val get_explanation: literal list -> explanation

(** merges explanations into one explanation *)
val merge_explanations: explanation -> explanation -> explanation

(** [prune_up_to_state explanation state]
  returns
  - the [explanation] pruned to contain only states older than [state].
  - true if [state] was in [explanation] *)
val prune_up_to_state: explanation -> state -> explanation * bool





(** {2 Backtracking} *)

(** invalidates a state and all states based on it (transitively). *)
val invalidate_state: state -> unit

(** removes invalid states and context literal explanations. *)
val backtrack: unit -> unit





(** {2 Representation} *)

val state_to_string: state -> string
val current_state_to_string: unit -> string
val current_states_to_string: unit -> string
val explanation_to_string: explanation -> string




(** {2 Restart} *)

(** dependend on {!Term.get_default_v} *)
val restart: unit -> unit
