(* $Id: subst.ml,v 1.15 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(*** types ***)

type symbol = Symbol.symbol
type literal = Term.literal

type var = {
  sv_var: Var.var;
  sv_offset: int;
}

type term = {
  st_term: Term.term;
  st_offset: int;
}

type binding = {
  sb_var: var;
  sb_term: term;
}
    
type subst = 
    binding list

type packed_subst =
    binding array





exception SET_FAIL




(*** offsets used in a context unifier substitution *)

(* literals of the input clause *)
let input_literal_offset  =
  0

(* fresh parameters introduced when making the substitution admissible,
   i.e. var/par unmixing *)
let fresh_par_offset =
  1

(* offset of the context literal paired to the index.th input clause literal *)
let context_literal_offset (index: int) =
  2 + index









(*** creation ***)

let empty =
  []


let make_var (var: Var.var) (offset: int) : var =
  {
    sv_var = var;
    sv_offset = offset;
  }

let make_term (term: Term.term) (offset: int) : term =
  {
    st_term = term;
    st_offset = offset;
  }



(*** string representation ***)

let var_to_string (var: var) : string =
  string_of_int var.sv_offset ^ ": " ^ Var.to_string var.sv_var
	  
let term_to_string (term: term) : string = 
  string_of_int term.st_offset ^ ": " ^ Term.term_to_string term.st_term

let binding_to_string (binding: binding) : string =
  "[ "
  ^ var_to_string binding.sb_var
  ^ " -> "
  ^ term_to_string binding.sb_term
  ^ " ]"

let subst_to_string (subst: subst) : string =
  " { \n"
  ^ String.concat "\n"
    (List.map binding_to_string subst)
  ^ " }" 




(*** equality ***)

let var_equal (var1: var) (var2: var) : bool =
  (var1.sv_offset = var2.sv_offset)
  &&
  (Var.equal var1.sv_var var2.sv_var)
	

let term_equal (term1: term) (term2: term) : bool =
  (term1.st_offset = term2.st_offset)
  &&
  (Term.term_equal term1.st_term term2.st_term)


(* for testing only *)
let subst_equal (subst1: subst) (subst2: subst) : bool =
  subst1 == subst2
  ||
  Tools.lists_unordered_equal
    (fun binding1 binding2 ->
       (var_equal binding1.sb_var binding2.sb_var)
       &&
       (term_equal binding1.sb_term binding2.sb_term)
    )
    subst1
    subst2







(*** access ***)

let rec get_ (subst: subst) (var: Var.var) (offset: int) : term option =
  match subst with
    | [] ->
	None

    | binding :: tail ->
	if 
	  (offset = binding.sb_var.sv_offset)
	  &&
	  (Var.equal var binding.sb_var.sv_var)
	then
	  Some binding.sb_term
	else
	  get_ tail var offset

let rec get (subst: subst) (var: var) : term option =
  get_ subst var.sv_var var.sv_offset




(* caching var -> bound_vars information
   between calls from get_bound_vars2/3 didn't pay off *)
let rec get_bound_vars (subst: subst) (vars: var list) (acc: var list) : var list =
  List.fold_left
    (fun acc var ->
       match get subst var with
	 | None ->
	     (* unbound, so return the variable itself *)
	     if List.exists (var_equal var) acc then
	       acc
	     else
	       var :: acc
		   
	 | Some term ->
	     (* unbbound var, so return the variables the bound term *)
	     let term_vars =
	       List.map
		 (fun term_var -> 
		    make_var term_var term.st_offset
		 )
		 (Term.vars_of_term term.st_term)
	     in
	       get_bound_vars
		 subst
		 term_vars
		 acc
    )
    acc
    vars


let get_bound_vars1 (subst: subst) (var: var) : var list =
  get_bound_vars subst [var] []

let get_bound_vars2 (subst: subst) (vars: var list) : var list =
  get_bound_vars subst vars []

let get_bound_vars3 (subst: subst) (vars: var list list) : var list list =
  List.fold_left
    (fun acc vars ->
       get_bound_vars2 subst vars :: acc
    )
    []
    vars












let dereference (subst: subst) (term: term) : term =
  match term.st_term with
    | Term.Var var ->
	begin
	  match get_ subst var term.st_offset with
	    | None ->
		term
		  
	    | Some bound_term ->
		bound_term
	end

    | _ ->
	term



(* raises SET_FAIL on a cyclic variable dependency *)
let rec occur_check (subst: subst) (var: var) (term: Term.term) (term_offset: int) : unit =
  match term with
    | Term.Var term_var ->
	if ((var.sv_offset = term_offset) 
	    && 
	    (Var.equal var.sv_var term_var))
	then begin
	  raise SET_FAIL
	end
	else begin
	  match get_ subst term_var term_offset with
	    | None -> 
		()

	    | Some term -> 
		occur_check subst var term.st_term term.st_offset
	end

    | Term.Const _ ->
	()

    | Term.Func (_, terms, _) ->
	Array.iter
	  (fun term ->
	     occur_check subst var term term_offset
	  )
	  terms



(* returns true if some bindings must be compressed.

   raises SET_FAIL if the substitution can not be extended
   in a p-preserving (i-preserving) way with var -> term *)
let check_bindings
  (p_preserving: bool) (i_preserving: bool) 
  (subst: subst) (var: var) (term: term) : bool =

  (* ensure p-preserving on var -> term *)
  let is_p_renaming =
    if p_preserving && Var.is_parametric var.sv_var then begin
      match term.st_term with
	| Term.Var bound_var when
	    Var.is_parametric bound_var
	    &&
	    var.sv_offset <> term.st_offset
	    ->
	      true
		    
	| _ ->
	    raise SET_FAIL
    end

    else
      false
  in
    
  (* ensure i-preserving on var -> term *)
  let is_i_renaming =
    if i_preserving && Var.is_indicator var.sv_var then begin

      match term.st_term with
	| Term.Var bound_var when
	    Var.is_indicator bound_var
	    &&
	    var.sv_offset <> term.st_offset
	    &&
	    (Var.is_universal bound_var = Var.is_universal var.sv_var)
	    ->
	      true

	| _ ->
	    raise SET_FAIL
    end

    else
      false
  in

    (* is_p_renaming (is_i_renaming) marks that var -> term
       binds a parameter to a parameter (indicater to indicator).
       thus the bindings to var and term have to be checked
       if they are still p-preserving (i-preserving) after
       the new binding is added.
       I.e. if var is the only parameter (indicator) bound to term.

       Other bindings can not be made non-p-preserving (non-i-preserving)
       by rebinding a parameter to a non-parameter,
       as this would imply that var is a parameter and term a non-parameter
       and would have been caught above. *)
    List.fold_left
      (fun acc binding ->
	 (* var already bound *)
	 if var_equal var binding.sb_var then
	   raise SET_FAIL
	     
	 (* variables bound to new term to bind *)
	 else if term_equal term binding.sb_term then begin
	   if is_p_renaming && Var.is_parametric binding.sb_var.sv_var then
	     raise SET_FAIL
	   else if is_i_renaming && Var.is_indicator binding.sb_var.sv_var then
	     raise SET_FAIL
	   else
	     acc
	 end
	   
	 (* variables bound to new variable to bind *)
	 else if (not acc) || is_p_renaming || is_i_renaming then begin
	   
	   match binding.sb_term.st_term with 
	     | Term.Var bound_var
		 when
		   ((var.sv_offset = binding.sb_term.st_offset)
		    &&
		    (Var.equal var.sv_var bound_var)) ->
		 
		 if is_p_renaming && Var.is_parametric binding.sb_var.sv_var then
		   raise SET_FAIL
		 else if is_i_renaming && Var.is_indicator binding.sb_var.sv_var then
		   raise SET_FAIL
		 else
		   true
		     
	     | _ ->
		 acc
	 end
	 else
	   acc
      )
      false
      subst


(* var is bound to term -
   so all variables already bound to var are rebound to term. *)
let compress_old_bindings
  (subst: subst) (var: var) (term: term) : subst =
  List.map
    (fun binding ->
       match binding.sb_term.st_term with 
	 | Term.Var bound_var
	       when
		 ((var.sv_offset = binding.sb_term.st_offset)
		  &&
		  (Var.equal var.sv_var bound_var)) ->
	     {
	       sb_var = binding.sb_var; 
		 sb_term = term;
	     }
	     
	 | _ ->
	     binding
      )
    subst


(* binds a variable to a term,
   does occur check and path compression. *)
let set ?(p_preserving:bool = false) ?(i_preserving:bool = false)
  (subst: subst) (var: var) (term: term) : subst =

  if Config.debug &&
    List.exists
      (fun binding ->
	 var_equal var binding.sb_var
      )
      subst
  then begin
    failwith "ohaohaoha";
    raise SET_FAIL;
  end;

  (* check for cyclic bindings *)
  occur_check subst var term.st_term term.st_offset;

  (* path compression: if term is a variable, bind to its binding instead *)
  let new_bound_term =
    dereference subst term
  in

  (* path compression: replace bindings to var by bindings to its new binding *)
  let compressed_subst =
    if check_bindings p_preserving i_preserving subst var new_bound_term then
      compress_old_bindings subst var new_bound_term
    else
      subst
  in

  let new_binding = { 
    sb_var = var;
    sb_term = new_bound_term;
  }
  in
   
    new_binding :: compressed_subst


let set_ ?(p_preserving:bool = false) ?(i_preserving:bool = false)
  (subst: subst) (var: Var.var) (var_offset: int) (term: Term.term) (term_offset: int) : subst =
  set ~p_preserving:p_preserving ~i_preserving:i_preserving
    subst (make_var var var_offset) (make_term term term_offset)







let rec reverse_var_to_par (subst: subst) (vars_to_reverse: var list) (pars_to_keep: var list): subst =
  match vars_to_reverse with
    | [] ->
	subst
	
    | var_to_reverse :: tail ->
	begin
	  try
	    let binding_to_reverse =
	      List.find
		(fun binding ->
		   var_equal var_to_reverse binding.sb_var
		)
		subst
	    in
	      begin
		match binding_to_reverse.sb_term.st_term with
		  | Term.Var bound_par ->
		      if
			(* the var is not bound to a par but to a var, so ignore it *)
			(Var.is_universal bound_par)
			||
			(* this parameter may not be reversed *)
			(List.exists 
			   (var_equal (make_var bound_par binding_to_reverse.sb_term.st_offset))
			   pars_to_keep)
		      then
			reverse_var_to_par subst tail pars_to_keep
		      else begin
			if 
			  List.exists
			    (fun binding ->
			       (* ignore the binding to reverse *)
			       (not (binding == binding_to_reverse))
			       &&
			       (* does this binding bind to the parameter to reverse? *)
			       term_equal binding.sb_term binding_to_reverse.sb_term
			       &&
			       (* and is this binding a parameter which may not be reversed? *)
			       List.exists (var_equal binding.sb_var) pars_to_keep
			    )
			    subst
			then
			  (* reversing this var -> par would also reverse a par which may not be reversed *)
			  reverse_var_to_par subst tail pars_to_keep
			else begin
			  (* ok, now the binding can be reversed *)
			  let new_bound_term =
			    make_term (Term.request_var var_to_reverse.sv_var) var_to_reverse.sv_offset
			  in

			  let reversed_subst =
			    List.map
			      (fun binding ->
				 (* replace the old binding by its reversed binding *)
				 if binding == binding_to_reverse then
				   {
				     sb_var = make_var bound_par binding_to_reverse.sb_term.st_offset;
				     sb_term = new_bound_term;
				   }

				 (* replace a binding to the reversed var *)
				 else if term_equal binding.sb_term binding_to_reverse.sb_term then
				   { binding with
				       sb_term = new_bound_term;
				   }

				 else
				   binding
			      )
			      subst
			  in
			    reverse_var_to_par reversed_subst tail pars_to_keep
			end
		      end

		  | _ ->
		      (* not bound to a variable *)
		      reverse_var_to_par subst tail pars_to_keep
	      end
	  with
	    | Not_found ->
		(* the var to reverse is not bound, so ignore it *)
		reverse_var_to_par subst tail pars_to_keep
	end














(*** apply ***)

(* applies a substitution to a term and normalizes it.

   var_mapping maps old variables to new ones,
   thus normalizing the term by numbering the new variables from 0 on *)
let rec do_apply_to_term (subst: subst) (term: Term.term) (offset: int)
  (var_mapping: (var * Var.var) list) : Term.term * ((var * Var.var) list) =
  
  match term with
    | Term.Var var ->
	begin
	  match get_ subst var offset with
	    | None ->
		let var_to_normalize =
		  (make_var var offset)
		in

		let new_var, var_mapping =
		  try 
		    List.assoc var_to_normalize var_mapping, var_mapping
		  with
		    | Not_found ->
			let new_var =
			  (* the next id can be deduced from the number of remapped vars *)
			  Var.clone_renumbered var (List.length var_mapping)
			in
			  new_var, (var_to_normalize, new_var) :: var_mapping
		in
		  (Term.request_var new_var), var_mapping

	    | Some term -> 
		do_apply_to_term subst term.st_term term.st_offset var_mapping
	end

    | Term.Const _ -> 
	term, var_mapping

    | Term.Func (symbol, terms, _) ->
	let new_terms =
	  Array.copy terms
	in

	let rec do_at (i : int) mapping =
	  if i >= Array.length new_terms then
	    mapping
	  else
	    let new_term, altered_var_mapping =
	      do_apply_to_term subst terms.(i) offset mapping
	    in
	      new_terms.(i) <- new_term;
	      do_at (i + 1) altered_var_mapping
	in
	  
	(* side effect on new_terms *)
	let new_var_mapping =
	  do_at 0 var_mapping
	in
	  Term.request_func (symbol, new_terms), new_var_mapping


let apply_to_term (subst: subst) (term: Term.term) (offset: int) : Term.term =
  fst (do_apply_to_term subst term offset [])

let apply_to_literal (subst: subst) (literal: Term.literal) (offset: int) : Term.literal =
  let term, _ =
    do_apply_to_term subst literal.Term.term offset []
  in
    Term.request_literal literal.Term.sign term






let replace_binding_offset (binding: binding) (old_offset: int) (new_offset: int) : binding =
  if
    binding.sb_var.sv_offset = old_offset
    &&
    binding.sb_term.st_offset = old_offset
  then
    { 
      sb_var = {
	sv_var = binding.sb_var.sv_var;
	sv_offset = new_offset;
      };
      sb_term = {
	st_term = binding.sb_term.st_term;
	st_offset = new_offset;
      };
    }
  else if
    binding.sb_var.sv_offset = old_offset
  then
    { binding with
	sb_var = {
	  sv_var = binding.sb_var.sv_var;
	  sv_offset = new_offset;
	};
    }
    
  else if
    binding.sb_term.st_offset = old_offset
  then
    { binding with
	sb_term = {
	  st_term = binding.sb_term.st_term;
	  st_offset = new_offset;
	};
    }
    
  else
    binding

let replace_offset (subst: subst) (old_offset: int) (new_offset: int) : subst =
  List.map
    (fun binding -> replace_binding_offset binding old_offset new_offset)
    subst






let replace_in_bound_terms (subst: subst) (old_term: Term.term) (new_term: Term.term) : subst =
  if Term.term_equal old_term new_term then
    subst
  else
    List.map
      (fun binding ->
	 let new_bound_term =
	   Term.replace_in_term binding.sb_term.st_term old_term new_term
	 in
	   { binding with
	       sb_term = {
		 binding.sb_term with
		   st_term = new_bound_term
	       };
	   }
      )
      subst
