(* $Id: subst.mli,v 1.9 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(** a substitution (immutable)

  A substitution is a mapping from variables to terms.

  In order to be able to efficiently reuse a literal
  when several fresh variants of it are needed during a unification process,
  each variable and term is paired with an {e offset}.
  Thus creating a fresh variant of a literal
  consists just of pairing the literal with fresh offsets.

  Therefore a {!Subst.subst} stores not a mapping from {!Var.var} to {!Term.term},
  but from {!Subst.var} to {!Subst.term}.
  
  A substitution is represented in the triangular form.
  I.e. if variables occuring in bound terms
  are also bound in the substitution,
  they have to be replaced by their binding.

  E.g. instead of [x -> (g(a)); y -> a] 
  the substitution may be [x -> (g(y)), y -> a].

  Path compression is exhaustively done.
  I.e. if a variable x is to be bound to a bound variable y,
  x is bound to y's binding instead.

  E.g. [x -> (g(y)), y -> z] 
  would be represented as [x -> (g(z)), y -> z].

  A substitution is {e parameter preserving (p_preserving)} if
  parameters are only bound to parameters of other terms.

  A substitution is {e indicator preserving (i_preserving)} if
  universal (parametric) indicators are only bound to
  universal (parametric) indicators of other terms.

  Thus a {e p_preserving} ({i p_preserving}) unifier between two terms
  contains a renaming between the parameters (indicator variables)
  of the two terms.

  If {e p_preserving} ({i p_preserving}) is an optional argument of a function
  it always defaults to false,
  thus completely ignoring parameter and indicator variables.
*)



(** {6 Types} *)

(** associates a {!Var.var} with an offset *)
type var = {
  sv_var: Var.var;
  sv_offset: int;
}

(** associates a {!Term.term} with an offset *)
type term = {
  st_term: Term.term;
  st_offset: int;
}

(** a binding of a {!Subst.var} to a {!Subst.term} *)
type binding = {
  sb_var: var;
  sb_term: term;
}
    
(** a substitution is simply a list of bindings *)
type subst = 
    binding list

(** more compact representation for storing a substitution *)
type packed_subst =
    binding array



(** see {!Subst.set} *)
exception SET_FAIL



(** {6 Functions} *)


(** {2 Offsets} *)
  

(** literals of the input clause *)
val input_literal_offset : int

(** [context_literal_offset input_clause_literal_index]
  computes the offset of the context literal 
  based on the index of the input literal it is paired to.
  
  A context unifier is not build at once,
  its literal pairs are unified separately
  and the partial substitutions are then merged afterwards.
  Creating a fresh variant of a literal is
  pairing the literal with an offset.
  Thus in order to be able to correctly merge the partial substitutions
  they must use these different and defined offsets. *)
val context_literal_offset : int -> int

(** to make a context unifier admissible it might be necessary
  to bind universal variables to new parameters.
  these have to use this offset *)
val fresh_par_offset : int



(** {2 Creation} *)

(** empty substitution *)
val empty: subst

(** conveniently creating a Subst.var *)
val make_var: Var.var -> int -> var

(** conveniently creating a Subst.term *)
val make_term: Term.term -> int -> term





(** {2 Access} *)

(** [get subst var] returns the term bound to [var],
  or None if [var] is unbound.
  Because of the triangular substitution (see module description)
  the returned term may contain other variables bound in the substitution.
  To get the real term {!Subst.apply_to_term} must be used on the term. *)
val get: subst -> var -> term option

(** wrapper for {!Subst.get} *)
val get_: subst -> Var.var -> int -> term option

(** [get_bound_vars subst var] returns
  - the subst_vars contained in the term bound to [var],
  - or [var] if it is unbound *) 
val get_bound_vars1 : subst -> var -> var list

(** like {!Subst.get_bound_vars1}, but for a list of vars *)
val get_bound_vars2 : subst -> var list -> var list  

(** like a mapping with {!Subst.get_bound_vars2}. *)
val get_bound_vars3 : subst -> var list list -> var list list


(** [set subst var term].
  extends [subst] by the binding [var -> term].

  raises SET_FAIL if
  - the binding would lead to a cycle (occur check)
  - the variable is already set - a setting can not be undone

  {e p_preserving} ({e i_preserving}) is by default false.
  if true [subst] is expected to be {e p_preserving} ({e i_preserving})
  and is tried to extend with the new binding [var -> term]
  to a {e p_preserving} ({e i_preserving}) substitution.
*)
val set: ?p_preserving:bool -> ?i_preserving:bool -> subst -> var -> term -> subst

(** wrapper for {!Subst.set} *)
val set_: ?p_preserving:bool -> ?i_preserving:bool -> subst -> Var.var -> int -> Term.term -> int -> subst



(** [reverse_var_to_par subst vars_to_reverse pars_to_keep]
  reverses one by one the bindings of the universal variables in [vars_to_reverse], if
  - the variable is bound to a parameter
  - no parameter in [pars_to_keep] is bound to the same parameter as the variable -
    this would also reverse the parameters binding and let it be bound to a universal variable,
    thus loosing p_preserving bindings *)
val reverse_var_to_par : subst -> var list -> var list -> subst






(** {2 Application} *)


(** [apply_to_term subst term offset]
  applies [subst] to [term] with [offset] and returns a new normalized term.
  I.e. the variables' ids in the new term are enumerated from 0 on. *)
val apply_to_term : subst -> Term.term -> int -> Term.term

(** like {!Subst.apply_to_term}. *)
val apply_to_literal : subst -> Term.literal -> int -> Term.literal

(** [replace_offset subst old_offset new_offset]
  replaces all occurences of [old_offset] in [subst]
  by [new_offset] *)
val replace_offset: subst -> int -> int -> subst

(** [replace_offset subst old_term new_term]
  replaces in all terms bound in [subst]
  all occurences of [old_term] by [new_term].

  As no offsets are be specified there may only be {b one}
  offset used in the whole substitution. *)
val replace_in_bound_terms: subst -> Term.term -> Term.term -> subst




(** {2 Comparison} *)

val var_equal : var -> var -> bool
val term_equal: term -> term -> bool
val subst_equal : subst -> subst -> bool



(** {2 String Representation} *)

val var_to_string: var -> string
val term_to_string: term -> string
val binding_to_string: binding -> string

(** returns the triangular form,
  i.e. bound variables are not replaced by their bound term. *)
val subst_to_string: subst -> string
