(* $Id: substitution_tree.ml,v 1.13 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(* a basic implementation of a substitution tree.

   PARAMETERS:
   The only modifications is the usage of parameters.
   thus there are two kinds of indicators: universal and parametric.
   non-indicator variables, i.e. variables used to abstract over subterms,
   are always universal.

   INSERTION:
   Only very basic attempts are undertaken to keep the tree shallow:
   - if the term is an instance of a tree node,
     the search descends into this branch.
     The first such node found is used,
     even if some of its siblings are also valid.
     The only slight attempt to use a node which keeps the tree flat
     is to keep sibling nodes ordered by increasing depth of their branches
     and thus to choose the shallowest of the possible branches.

   - if the term is not an instance of any of the child nodes of a node,
     it is inserted under this node.
     If a non-empty mscg with a child exists it is used as a generalization
     (if several exists, the one introducing less bindings is chosen),
     otherwise the term is inserted as a new child.

   NORMALIZATION:
   Prior to insertion a term has to be normalized.
   I.e. all is variables are replaced by indicator variables.
   To allow easy merging into the tree
   a term is normalized using indicator variables
   not already contained in the tree.
   To do this easily the tree contains a counter for indicator ids.

   OFFSETS:
   The tree has one offset (tree_offset) for all contained variables and terms.
   A query term / insertion term gets a different offset (query_offset)
   for the unification operations.
   For an insertion term the query_offset is replace by tree_offset
   when the place to insert the term has been found.

   
*)


(*module Make (DataType : Term_indexing.DataType) : Term_indexing.Index with type data = DataType.t =*)
module Make (DataType : Term_indexing.DataType) :
  Term_indexing.Index
with type data = DataType.t =
struct


  (*** types ***)

  type counter = Counter.counter
  type var = Var.var
  type term = Term.term
  type binding = Subst.binding
  type subst = Subst.subst
  type data = DataType.t

  (* a leaf - i.e. representing a term with its attached data *)
  type leaf = {
    sl_subst: subst;
    (* subst from parent node to this leaf *)

    sl_data: data;
    (* attached data *)

    mutable sl_indicators: Var.var list;
    (* the indicator variables which are contained
       in the substitution composed from the root up to and including this node *)
  }
      
  (* a generalization over leaves *)
  and inner_node = {
    sn_subst: subst;
    (* subst from parent node to this node *)
    
    sn_open_vars: Subst.var list;
    (* the variables which are open in the substitution composed from the root
       up to and including this node. I.e. these variables have to be bound
       in a path down to a leaf. *)

    sn_closed_vars: Subst.var list;
    (* variables which are bound in the path up to and including the node. *)

    mutable sn_indicators: Var.var list;
    (* the indicator variables which are contained
       in the substitution composed from the root up to and including this node *)

    mutable sn_children: node list;
    (* child nodes *)

    mutable sn_depth: int;
    (* maximum of the depth of the subtrees under the children *)

  }
      
  and node =
    | Node of inner_node
    | Leaf of leaf

	
  type substitution_tree = {
    st_counter: Counter.counter;
    (* enumeration of new tree variables *)

    (* empty tree or top node *)
    mutable st_root_node: node option;
  }

  type t = substitution_tree



  (* used by the find-first queries to return on success *)
  exception FOUND of data








  (*** consts ***)
    
  let tree_offset =
    0
      
  let query_offset =
    1



  let data_is_equal =
    DataType.is_equal

  let data_to_string =
    DataType.to_string




  (*** node accesss ***)

  let get_node_subst (node: node) : subst =
    match node with
      | Leaf leaf ->
	  leaf.sl_subst
	  
      | Node inner_node ->
	  inner_node.sn_subst

  let get_node_indicators (node: node) : Var.var list =
    match node with
      | Leaf leaf ->
	  leaf.sl_indicators
	  
      | Node inner_node ->
	  inner_node.sn_indicators


  (*** tree root access ***)

  let get_root_var (tree: substitution_tree) : Subst.var =
    match tree.st_root_node with
      | None ->
	  failwith "Substitution_tree.get_root_var: empty tree"

      | Some node ->
	  begin
	    match get_node_subst node with
	      | [] ->
		  failwith "Substitution_tree.get_root_var: empty root subst"
		  
	      | binding :: [] ->
		  binding.Subst.sb_var
		  
	      | subst ->
		  print_endline ("Substitution_tree.get_root_var: " ^ Subst.subst_to_string subst);
		  failwith "Substitution_tree.get_root_var: root subst contains more than one binding"
	  end

  let get_root_term (tree: substitution_tree) : term =
    match tree.st_root_node with
      | None ->
	  failwith "Substitution_tree.get_root_term: empty tree"

      | Some node ->
	  begin
	    match get_node_subst node with
	      | [] ->
		  failwith "Substitution_tree.root_term: empty root subst. probably inserted terms with different top symbol."
		  
	      | binding :: [] ->
		  binding.Subst.sb_term.Subst.st_term
		  
	      | subst ->
		  print_endline ("Substitution_tree.get_root_term: " ^ Subst.subst_to_string subst);
		  failwith "Substitution_tree.root_term: root subst contains more than one binding"
	  end





  (*** term normalization ***)

  (* replace indicator vars by non-indicator vars *)
  let clone_term_without_indicator_vars (term: term) =

    let rec do_clone_term_without_indicator_vars (term: term) : term =
      
      match term with
	| Term.Var var ->
	    Term.request_var (Var.clone_as_non_indicator var)
	    
	| Term.Const _ -> 
	    term
	    
	| Term.Func (symbol, terms, _) ->
	    Term.request_func (symbol, Array.map do_clone_term_without_indicator_vars terms)
    in
      do_clone_term_without_indicator_vars term

	
  (* replace variables with fresh indicator variables not contained in the tree

     a term to insert
     - may not contain any variables already used in the tree
     - and must contain only indicator variables
  *)
  let clone_term_renumbered (counter: counter) (term: term) =

    let rec do_clone_term_renumbered (var_mapping: (var * var) list) (term: term) :
      term * ((var * var) list) =
      
      match term with
	| Term.Var var ->
	    let new_var, new_var_mapping =
	      try 
		List.assoc var var_mapping, var_mapping
	      with
		| Not_found ->
		    let new_var =
		      (* the next id can be deduced from the number of remapped vars *)
		      Var.clone_renumbered (Var.clone_as_indicator var) (Counter.next counter)
		    in
		      new_var, (var, new_var) :: var_mapping
	    in
	      Term.request_var new_var, new_var_mapping
		
	| Term.Const _ ->
	    term, var_mapping
	      
	| Term.Func (symbol, terms, _) ->
	    let new_terms =
	      Array.copy terms
	    in

	    let rec do_at (i : int) mapping =
	      if i >= Array.length new_terms then
		mapping
	      else
		let new_term, altered_var_mapping =
		  do_clone_term_renumbered mapping terms.(i)
		in
		  new_terms.(i) <- new_term;
		  do_at (i + 1) altered_var_mapping
	    in
	      
	    (* side effect on new_terms *)
	    let new_var_mapping =
	      do_at 0 var_mapping
	    in
	      Term.request_func (symbol, new_terms), new_var_mapping
    in
    let renumbered_term, _ =
      do_clone_term_renumbered [] term
    in
      renumbered_term







  (*** representation  ***)



  let rec leaf_to_string
    (root_term: term)
    (subst_to_parent_node: subst)
    (depth: int)
    (leaf: leaf)
    : string =

    let subst_to_current_node =
      try
	Unification.unify_substs subst_to_parent_node leaf.sl_subst
      with
	| Unification.UNIFICATION_FAIL ->
	    print_endline ("subst_to_parent: " ^ Subst.subst_to_string subst_to_parent_node);
	    print_endline ("leaf_subst: " ^ Subst.subst_to_string leaf.sl_subst);
	    failwith "leaf_to_string"
    in

    let depth_prefix =
      String.make (2 * depth) ' ' ^ string_of_int depth ^ ": "
    in

    let indicator_vars =
      "Indicator_vars: [ " ^ String.concat ", " (List.map Var.to_string leaf.sl_indicators) ^ " ]"
    in

    let term =
      Subst.apply_to_term
	subst_to_current_node
	  root_term
	tree_offset
    in
    let term_representation =
      Term.term_to_string term
    in

    let data_representation =
      data_to_string leaf.sl_data
    in

      "\n" ^ depth_prefix ^ term_representation ^ ": " ^ data_representation ^ "\n"
      ^ depth_prefix ^ Subst.subst_to_string leaf.sl_subst ^ "\n"
      ^ indicator_vars ^ "\n"


  and node_to_string
    (root_term: term)
    (subst_to_parent_node: subst)
    (depth: int)
    (node: inner_node)
    (leaf_count: int ref)
    (inner_count: int ref)
    : string =

    inner_count := !inner_count + 1;

    let subst_to_current_node =
      try
	Unification.unify_substs subst_to_parent_node node.sn_subst
      with
	| Unification.UNIFICATION_FAIL ->
	    print_endline ("subst_to_parent: " ^ Subst.subst_to_string subst_to_parent_node);
	    print_endline ("node_subst: " ^ Subst.subst_to_string node.sn_subst);
	    failwith "node_to_string"
    in

    let depth_prefix =
      String.make (2 * depth) ' ' ^ string_of_int depth ^ ": "
    in

    let open_vars =
      "Open_vars: [ " ^ String.concat ", " (List.map Subst.var_to_string node.sn_open_vars) ^ " ]"
    in
    let indicator_vars =
      "Indicator_vars: [ " ^ String.concat ", " (List.map Var.to_string node.sn_indicators) ^ " ]"
    in


    let term =
      Subst.apply_to_term
	subst_to_current_node
	  root_term
	tree_offset
    in
    let term_representation =
      Term.term_to_string term
    in

    let children_representation =
      String.concat ""
	(List.map
	   (fun child_node ->
	      match child_node with
		| Leaf leaf ->
		    leaf_count := !leaf_count + 1;
		    leaf_to_string
		      root_term
		      subst_to_current_node
		      (depth + 1)
		      leaf

		| Node node ->
		    node_to_string
		    root_term
		    subst_to_current_node
		    (depth + 1)
		    node
		    leaf_count
		    inner_count
	   )
	   node.sn_children)
    in

      "\n" ^ depth_prefix
(*      ^ "(" ^ string_of_int node.sn_depth ^ ")"*)
      ^ term_representation ^ "\n"
      ^ depth_prefix ^ Subst.subst_to_string node.sn_subst ^ "\n"
      ^ open_vars ^ "\n"
      ^ indicator_vars ^ "\n"
      ^ "\n" ^ children_representation



  let to_string (tree: substitution_tree) : string =
    let leaf_count = ref 0
    in
    let inner_count = ref 0
    in
      match tree.st_root_node with
	| None ->
	    "empty tree"

	| Some (Leaf leaf) ->
	    leaf_to_string
	    (get_root_term tree)
	    Subst.empty
	    0
	    leaf

	| Some (Node node) ->
	    let rep =
	      node_to_string
		(get_root_term tree)
		Subst.empty
		0
		node
		leaf_count
		inner_count
	    in
	      "LEAFS: " ^ string_of_int !leaf_count
	      ^ "  NODES: " ^ string_of_int !inner_count
	      ^ rep







  (*** traversal ***)

  let rec iter_node
    (func: term -> data -> unit) (root_term: term) (subst_to_node: subst) (node: node) : unit =

    match node with
      | Leaf leaf ->
	  let node_term =
	    Subst.apply_to_term
	      subst_to_node
	      root_term
	      tree_offset
	  in
	  let normalized_term =
	    clone_term_without_indicator_vars node_term
	  in
	    func normalized_term leaf.sl_data;

      | Node inner_node ->
	  List.iter
	  (fun child_node ->
	     let child_subst =
	       Unification.match_substs
		 ~p_preserving:true
		 tree_offset subst_to_node (get_node_subst child_node)
	     in
	       iter_node func root_term child_subst child_node
	  )
	  inner_node.sn_children

  let iter (func: term -> data -> unit) (tree: substitution_tree) : unit =
    match tree.st_root_node with
      | None ->
	  ()

      | Some node ->
	  iter_node
	  func
	  (get_root_term tree)
	  Subst.empty
	  node
	  


  (* for debugging: checks that no variable is bound twice in a path *)
  let rec validate_node
    (substs: subst list) (node: node) : unit =

    let node_subst =
      get_node_subst node
    in
      try
	let reused_var =
	  List.find
	    (fun node_binding ->
	       List.exists
	       (fun previous_subst ->
		  List.exists
		  (fun previous_binding ->
		     Subst.var_equal
		     node_binding.Subst.sb_var
		     previous_binding.Subst.sb_var
		  )
		  previous_subst
	       )
	       substs
	    )
	    node_subst
	in
	  failwith ("variable reused: " ^ Subst.binding_to_string reused_var)
      with
	| Not_found ->
	    begin
	    match node with
	      | Leaf _ ->
		  ()

	      | Node inner_node ->
		  let new_substs =
		    node_subst :: substs
		  in
		    List.iter (validate_node new_substs) inner_node.sn_children
	  end

  let validate (tree: substitution_tree) : unit =
    match tree.st_root_node with
      | None ->
	  ()

      | Some node ->
	  validate_node [] node
	  
















  (*** mscg ***)

  (* computation of the most specific generalization
     of two terms/substitutions, e.g.:
     f(a, X, g(a)) and f(a, Y, g(Y)) -> f(a, X, g(Z)); [Z -> a]; [Z -> X]

     I.e. the substitutions applied to the mscg term produce variants
     of the original terms, not the terms themselves.
     
     these functions are specific for the usage in a substitution tree.
     they expect some prerequirements:
     - the substitutions only contain one offset: tree_offset
     - the left term/substitution is alread part of the tree, i.e. a leaf or node,
     the right term/substitution represents the term to insert into the tree.
  *)

  let mscg_of_terms
    (_tree: substitution_tree)

    (* build the mscg of these terms *)
    (_tree_term: term) (_query_term: term)

    (* when called by mscg_of_substs these contain generalizations
       of previous calls to mscg_of_terms for other terms
       of the tree and the query *)
    (_tree_subst: subst) (_query_subst: subst)

    (* the open vars of the parent node. *)
    (_open_vars: Subst.var list)
    (_closed_vars: Subst.var list)
    : term * subst * subst =

    (* create a new variable to generalize over the two incompatible terms *)
    let do_generalize (tree_term: term) (query_term: term) (tree_subst: subst) (query_subst: subst)
      : term * subst * subst =

      let var =
	Var.create_universal (Counter.next _tree.st_counter)
      in
	Term.request_var var,
	Subst.set_ tree_subst var tree_offset tree_term tree_offset,
	Subst.set_ query_subst var tree_offset query_term tree_offset
    in


    (* generalize over two incompatible terms *)
    let generalize (tree_term: term) (query_term: term) (tree_subst: subst) (query_subst: subst)
      : term * subst * subst =

      (* try to reuse previous generalizations over the same terms.
	 only non-open vars may be reused.
	 otherwise open vars might be lost due tue variable reusage and never closed. *)

      (* if this is the top term this binding will be pruned
	 by mscg_of_subst to the open var on top of the two top terms.
	 thuse an open var is involved and reusage is not possible. *)
      if Term.term_equal _tree_term tree_term then begin
	do_generalize tree_term query_term tree_subst query_subst
      end

      (* try to find a previous generalization over the same two terms *)
      else begin
	try
	  let reuse_binding =
	    List.find
	      (fun binding ->
		 (* var is bound to tree_term ... *)
		 Term.term_equal tree_term binding.Subst.sb_term.Subst.st_term
		 &&
		 (* and var is not an open var ... *)
		 (List.for_all
		    (fun open_var ->
		       not (Var.equal binding.Subst.sb_var.Subst.sv_var open_var.Subst.sv_var)
		    )
		    _open_vars
		 )
		 &&
		 (* and var is bound to query_term ... *)
		 (List.exists
		    (fun binding2 ->
		       Var.equal binding.Subst.sb_var.Subst.sv_var binding2.Subst.sb_var.Subst.sv_var
		       &&
		       Term.term_equal query_term binding2.Subst.sb_term.Subst.st_term
		    )
		    query_subst
		 )
	      )
	      tree_subst
	  in
	    Term.request_var reuse_binding.Subst.sb_var.Subst.sv_var, tree_subst, query_subst
	with
	  | Not_found ->
	      do_generalize tree_term query_term tree_subst query_subst
      end
    in


    let rec do_mscg_of_terms (tree_term: term) (query_term: term) (tree_subst: subst) (query_subst: subst)
      : term * subst * subst =
      (* try to use var as the generalization,
	 i.e. avoid doing anything for the left side *)
      match tree_term, query_term with
	| Term.Var var, _ when
	    (* indicator may never be bound in an mscg *)
	    not (Var.is_indicator var)
	    &&
	    (* reusage of open vars is handled in mscg_of_subst
	       as the open var might also be bound in the query subst. *)
	    List.for_all
	      (fun open_var -> not (Var.equal var open_var.Subst.sv_var))
	      _open_vars

	    &&
	    List.for_all
	      (fun closed_var -> not (Var.equal var closed_var.Subst.sv_var))
	      _closed_vars

	    ->

	    begin
	      match Subst.get_ query_subst var tree_offset with
		| None ->
		    tree_term,
		    tree_subst,
		    Subst.set_ query_subst var tree_offset query_term tree_offset
		      
		(* already bound to the same term?
		   so the old binding can be reused by simply reusing the variable *)
		| Some bound_term when
		    Term.term_equal query_term bound_term.Subst.st_term ->
		    tree_term, tree_subst, query_subst
		      
		| _ ->
		    generalize tree_term query_term tree_subst query_subst
	    end


	| Term.Const symbol1, Term.Const symbol2 when
	    Symbol.equal symbol1 symbol2 ->

	    tree_term, tree_subst, query_subst


	| Term.Func (symbol1, terms1, _), Term.Func (symbol2, terms2, _) when
	    Symbol.equal symbol1 symbol2 ->

	    let new_terms =
	      Array.copy terms1
	    in
	      
	    let rec map_terms (index: int) (subst1: subst) (subst2: subst) : subst * subst =
	      if index >= Array.length terms1 then
		subst1, subst2

	      else
		let mscg, subst1, subst2 =
		  do_mscg_of_terms terms1.(index) terms2.(index) subst1 subst2
		in
		  new_terms.(index) <- mscg;
		  map_terms (index + 1) subst1 subst2
	    in
	      
	    let subst1, subst2 =
	      map_terms 0 tree_subst query_subst
	    in
	      
	      Term.request_func (symbol1, new_terms), subst1, subst2;

	      
	| _ ->
	    generalize tree_term query_term tree_subst query_subst
	    
    in
      do_mscg_of_terms _tree_term _query_term _tree_subst _query_subst





  let mscg_of_substs
    (_tree: substitution_tree)

    (* build the mscg of these substitutions *)
    (_tree_subst: subst) (_query_subst: subst)

    (* the open vars of the parent node. *)
    (_open_vars_parent: Subst.var list)
    (* the open vars of the parent node. *)
    (_closed_vars_parent: Subst.var list)

    (* the open vars of the tree node. *)
    (_open_vars_tree: Subst.var list)

    (* the indicator vars used in the parent substitutions *)
    (_parent_indicators: Var.var list)

    (* the indicator vars used in the query substitution *)
    (_query_indicators: Var.var list)
    : subst * subst * subst =

    (* get the binding of open_var in _query_subst *)
    let get_query_only_bound_var (open_var: Subst.var) : binding =
      try
	List.find
	  (fun binding ->
	     Subst.var_equal open_var binding.Subst.sb_var
	  )
	  _query_subst
      with
	| Not_found ->
	    print_endline ("TREE: " ^ to_string _tree);
	    print_endline ("OPEN_VAR: " ^ Subst.var_to_string open_var);
	    print_endline ("TREE_SUBST: " ^ Subst.subst_to_string _tree_subst);
	    print_endline ("QUERY_SUBST: " ^ Subst.subst_to_string _query_subst);
	    failwith "get_query_only_bound_var"
    in

    (* get the bindinds of open_var in _term_subst and _query_subst *)
    let get_binding_pair (open_var: Subst.var) : binding * binding =
      try
	List.find
	  (fun binding ->
	     Subst.var_equal open_var binding.Subst.sb_var
	  )
	  _tree_subst,
	  
	  List.find
	    (fun binding ->
	       Subst.var_equal open_var binding.Subst.sb_var
	    )
	    _query_subst
      with
	| Not_found ->
	    print_endline ("TREE: " ^ to_string _tree);
	    print_endline ("OPEN_VAR: " ^ Subst.var_to_string open_var);
	    print_endline ("TREE_SUBST: " ^ Subst.subst_to_string _tree_subst);
	    print_endline ("QUERY_SUBST: " ^ Subst.subst_to_string _query_subst);
	    failwith "get_binding_pair"
    in




      
    (* the mscg of two terms was mscg_term, which is mscg_var.
       thus the generalization basically failed, as the only generalization is a variable.
       so try to reuse variables from equal previous generalizations *)
    let reuse_previous_generalizations
      (open_var: Subst.var)
      (mscg_term: term) (mscg_var: Var.var)
      (mscg_subst: subst) (new_mscg_tree: subst) (new_mscg_query: subst)
      (* previous to the generalization to mscg_term *)
      (old_mscg_tree: subst) (old_mscg_query: subst)
      : subst * subst * subst =


      (* does a generalization with a variable (x)
	 over the same terms (t1, t2) already exist?
	 if so, reuse it:
	 replace open_var -> mscg_var; mscg_var -> t1; mscg_var -> t2
	 with open_var -> x if x -> t1, x -> t2 does already exist.

	 raises Not_found on failure *)
      let bind_to_previous_generalization
	(tree_binding: binding) (query_binding: binding) : subst * subst * subst =

	(* find such a reusable generalization *)
	let reuse_binding =
	  List.find
	    (fun binding1 ->
	       Subst.term_equal tree_binding.Subst.sb_term binding1.Subst.sb_term
	       &&
	       not (Var.equal mscg_var binding1.Subst.sb_var.Subst.sv_var)
	       &&
	       (
		 List.exists
		  (fun binding2 ->
		     Subst.var_equal binding1.Subst.sb_var binding2.Subst.sb_var
		     &&
		     Subst.term_equal query_binding.Subst.sb_term binding2.Subst.sb_term
		  )
	       )
	       new_mscg_query
	    )
	    new_mscg_tree
	in

	(* so bind open_var to the previously used generalization var
	   and drop the new generalizations *)
	let new_mscg_term =
	  Term.request_var reuse_binding.Subst.sb_var.Subst.sv_var
	in
	let mscg_binding = {
	  Subst.sb_var = open_var;
	  Subst.sb_term = Subst.make_term new_mscg_term tree_offset;
	}
	in
	  mscg_binding :: mscg_subst, old_mscg_tree, old_mscg_query
      in



      (* the mscg is only a variable, thus we have:
	 open_var -> mscg_var; mscg_var -> t1; mscg_var -> t2
	 and can get rid of mscg_var with
	 ; open_var -> t1; open_var -> t2 *)
      let prune_mscg_var
	(tree_binding: binding) (query_binding: binding) : subst * subst * subst =
	
	let prune_subst (subst: subst) : subst =
	  List.map
	    (fun binding ->
	       if Var.equal mscg_var binding.Subst.sb_var.Subst.sv_var then
		 { binding with
		     Subst.sb_var = {
		       binding.Subst.sb_var with
			 Subst.sv_var = open_var.Subst.sv_var
		     }
		 }
	       else
		 binding
	    )
	    subst
	in
	  (*	print_endline ("\nPRUNE_MSCG_VAR: " ^ Subst.var_to_string open_var);
		print_endline ("\nopen_var: " ^ Subst.var_to_string open_var);
		print_endline ("\nmscg_var: " ^ Var.to_string mscg_var);*)
	  mscg_subst, prune_subst new_mscg_tree, prune_subst new_mscg_query
      in



      (* reusage is only done if mscg_var is bound in mscg_tree and mscg_query

	 raises Not_found on failure *)
      let do_reuse_previous_generalizations () : subst * subst * subst =
	(* is mscg_var bound in new_mscg_tree? if not, fail *)
	let tree_binding =
	  List.find
	    (fun binding ->
	       Var.equal mscg_var binding.Subst.sb_var.Subst.sv_var
	    )
	    new_mscg_tree
	in
	  try
	    let query_binding =
	      List.find
		(fun binding ->
		   Var.equal mscg_var binding.Subst.sb_var.Subst.sv_var
		)
		new_mscg_query
	    in
	      begin
		try
		  bind_to_previous_generalization tree_binding query_binding
		with
		  | Not_found ->
		      prune_mscg_var tree_binding query_binding
	      end		       
	  with
	    | Not_found ->
		print_endline ("MSCG_VAR: " ^ Var.to_string mscg_var);
		print_endline ("TREE_MSCG: " ^ Subst.subst_to_string new_mscg_tree);
		print_endline ("QUERY_MSCG: " ^ Subst.subst_to_string new_mscg_query);
		failwith "Substitution_tree.do_reuse_previous_generalizations"
      in


	try
	  do_reuse_previous_generalizations ()
	with
	  | Not_found ->
	      (* no, so mscg_var was a variable from tree_subst reused to generalize.
		 just keep it *)
	      let mscg_binding = {
		Subst.sb_var = open_var;
		Subst.sb_term = Subst.make_term mscg_term tree_offset;
	      }
	      in
		mscg_binding :: mscg_subst,
		new_mscg_tree,
		new_mscg_query
    in


    (* compute the mscg for the two substs
       and do all variable reusing/optimization except for indicator variables *)
    let compute_mscg () : subst * subst * subst =
      List.fold_left
	(fun (mscg_subst, mscg_tree, mscg_query) open_var ->
	   (* if this open_var remains an open var for the tree_subst
	      is is only bound in the query_subst
	      (and must be bound there as it is a leaf)
	      and the binding can just be propagated *)
	   if List.exists (Subst.var_equal open_var) _open_vars_tree then begin
	     mscg_subst, mscg_tree, get_query_only_bound_var open_var :: mscg_query
	   end

	   else begin
	     let first_binding, second_binding =
	       get_binding_pair open_var
	     in
	     let mscg_term, new_mscg_tree, new_mscg_query =
	       mscg_of_terms
		 _tree
		 first_binding.Subst.sb_term.Subst.st_term
		 second_binding.Subst.sb_term.Subst.st_term
		 mscg_tree
		 mscg_query
		 _open_vars_parent
		 _closed_vars_parent
	     in
	       begin
		 match mscg_term with
		   | Term.Var mscg_var ->
 		       (* generalization basically failed, as the only generalization is a variable.
			  so try to reuse variables from equal previous generalizations *)
		       reuse_previous_generalizations
		       open_var
		       mscg_term
		       mscg_var
		       mscg_subst
		       new_mscg_tree
		       new_mscg_query
		       mscg_tree
		       mscg_query

		   | _ ->
		       (* some generalization has been done, add it to the mscg *)
		       let mscg_binding = {
			 Subst.sb_var = open_var;
			 Subst.sb_term = Subst.make_term mscg_term tree_offset;
		       }
		       in
			 mscg_binding :: mscg_subst,
			 new_mscg_tree,
			 new_mscg_query
	       end
	   end
	)
	(Subst.empty, Subst.empty, Subst.empty)
	_open_vars_parent
    in


    (* replace the query indicator variable by the tree indicator variable.
       this permanently changes the indicator variables
       contained in the query leaf *)
    let use_tree_indicator
      (tree_indicator_var: Var.var)
      (query_indicator_var: Var.var)
      (tree_binding: binding) (query_binding: binding)
      (mscg_subst: subst) (mscg_tree: subst) (mscg_query: subst)
      : subst * subst * subst =

      (* if the variable used to generalize the indicator variables is
	 a open var of the parent node it must be kept *)
      let new_mscg =
	if List.exists (Subst.var_equal tree_binding.Subst.sb_var) _open_vars_parent then
	  (* open var, so bind it to the tree indicator *)
	  let mscg_binding = {
	    Subst.sb_var = tree_binding.Subst.sb_var;
	    Subst.sb_term = Subst.make_term (Term.request_var tree_indicator_var) tree_offset;
	  }
	  in
	    mscg_binding :: mscg_subst
	else
	  (* new generalization var, replace it with the tree indicator var *)
	  Subst.replace_in_bound_terms
	    mscg_subst
	    (Term.request_var tree_binding.Subst.sb_var.Subst.sv_var)
	    (Term.request_var tree_indicator_var)
      in

      (* remove the lifted bindings *)
      let pruned_mscg_tree =
	List.find_all
	  (fun binding -> tree_binding != binding)
	  mscg_tree
      in
      let pruned_mscg_query =
	List.find_all
	  (fun binding -> query_binding != binding)
	  mscg_query
      in
	
      (* replace the query indicator with the tree indicator used in the mscg *)
      let mapped_mscg_query =
	if Var.equal query_indicator_var tree_indicator_var then
	  pruned_mscg_query
	else
	  Subst.replace_in_bound_terms
	    pruned_mscg_query
	    (Term.request_var query_indicator_var)
	    (Term.request_var tree_indicator_var)
      in
	new_mscg, pruned_mscg_tree, mapped_mscg_query
    in



    let is_liftable
      (tree_indicator: Var.var) (query_indicator: Var.var)
      (common_indicators: Var.var list)
      : bool =
      Var.equal tree_indicator query_indicator
      ||
      (
	Var.is_indicator query_indicator
	&&
	(Var.is_universal tree_indicator = Var.is_universal query_indicator)
	&&
	(* an indicator which is part of the mscg or parent node can not be touched.
	   otherwise indicators which are equal in the tree subst
	   but different in the query subst
	   might be replaced by the same indicator,
	   thus changing the query subst *)
	(not
	   (List.exists
	      (fun common ->
		 (* f(x, x) and f(x, y) *)
		 Var.equal tree_indicator common
		 ||
		 (* f(x, y) and f(x, x) considering var replacement *)
		 Var.equal query_indicator common
	      )
	      common_indicators)
	)
	&&
	(* similiar as above.
	   but an indicator which is only in the query,
	   neither in the mscg or parent node,
	   may be replaced by another indicator -
	   is this replacing indicator is not already part of the query subst *)
	(not
	   (List.exists
	      (fun blocked ->
		 (* f(x, x) and f(x, y) *)
		 Var.equal tree_indicator blocked
		 (*					    ||
		 (* f(x, y) and f(x, x) considering var replacement *)
							    Var.equal right_indicator blocked*)
	      )
	      _query_indicators)
	)
      )
    in

    (* try to reuse indicator variables from the tree_subst
       by lifting them to the mscg and replacing the corresponding
       indicator variables in query_subst. *)
    let lift_indicators
      (_mscg_subst: subst) (_mscg_tree: subst) (_mscg_query: subst)
      : subst * subst * subst =

      let (mscg_subst, mscg_tree, mscg_query, _) =
	List.fold_left
	  (fun (mscg_subst, mscg_tree, mscg_query, common_indicators as acc) tree_binding ->
	     match tree_binding.Subst.sb_term.Subst.st_term with
	       | Term.Var tree_indicator when
		   Var.is_indicator tree_indicator ->
		   
		   begin
		     try
		       let query_binding =
			 List.find
			   (fun binding ->
			      Subst.var_equal tree_binding.Subst.sb_var binding.Subst.sb_var
			   )
			   mscg_query
		       in
			 begin
			   match query_binding.Subst.sb_term.Subst.st_term with
			     | Term.Var query_indicator ->
				 if is_liftable tree_indicator query_indicator common_indicators then
				   let new_mscg, new_tree_mscg, new_query_mscg =
				     use_tree_indicator
				       tree_indicator query_indicator
				       tree_binding query_binding
				       mscg_subst mscg_tree mscg_query
				   in
				     (* tree_indicator is lifted to the mscg
					and thus now also used in both the tree and the query subst *)
				     new_mscg, new_tree_mscg, new_query_mscg, tree_indicator :: common_indicators
				 else
				   acc
				     
			     | _ ->
				 acc
			 end

		     with
		       | Not_found ->
			   (* generalization variables of the tree must also be in the query *)
			   failwith "lift_indicators";
		   end
		   
	       | _ ->
		   acc
	  )
	  (_mscg_subst, _mscg_tree, _mscg_query, _parent_indicators)
	  _mscg_tree
      in
	mscg_subst, mscg_tree, mscg_query
    in


    let mscg_subst, mscg_tree, mscg_query =
      compute_mscg ()
    in
      if
	Config.debug
	&&
	(List.exists
	   (fun binding ->
	      List.exists
	      (fun var ->
		 not (Var.is_indicator var)
	      )
	      (Term.vars_of_term binding.Subst.sb_term.Subst.st_term)
	   )
	   mscg_query)
      then begin
	print_endline ("TREE_SUBST: " ^ Subst.subst_to_string _tree_subst);
	print_endline ("QUERY_SUBST: " ^ Subst.subst_to_string _query_subst);
	print_endline ("OPEN_VARS_PARENT: " ^ String.concat ", " (List.map Subst.var_to_string _open_vars_parent));
	print_endline ("OPEN_VARS_TREE: " ^ String.concat ", " (List.map Subst.var_to_string _open_vars_tree));
	print_endline ("INDICATORS_PARENT: " ^ String.concat ", " (List.map Var.to_string _parent_indicators));
	print_endline ("INDICATORS_QUERY: " ^ String.concat ", " (List.map Var.to_string _query_indicators));
	print_endline ("MSCG: " ^ Subst.subst_to_string mscg_subst);
	print_endline ("TREE_MSCG: " ^ Subst.subst_to_string mscg_tree);
	print_endline ("QUERY_MSCG: " ^ Subst.subst_to_string mscg_query);
	failwith "non-indicator in leaf mscg"
      end;
      let mscg_subst, mscg_tree, mscg_query =
	lift_indicators mscg_subst mscg_tree mscg_query
      in
      if
	Config.debug
	&&
	(List.exists
	   (fun binding ->
	      List.exists
	      (fun var ->
		 not (Var.is_indicator var)
	      )
	      (Term.vars_of_term binding.Subst.sb_term.Subst.st_term)
	   )
	   mscg_query)
      then begin
	print_endline ("TREE_SUBST: " ^ Subst.subst_to_string _tree_subst);
	print_endline ("QUERY_SUBST: " ^ Subst.subst_to_string _query_subst);
	print_endline ("OPEN_VARS_PARENT: " ^ String.concat ", " (List.map Subst.var_to_string _open_vars_parent));
	print_endline ("OPEN_VARS_TREE: " ^ String.concat ", " (List.map Subst.var_to_string _open_vars_tree));
	print_endline ("INDICATORS_PARENT: " ^ String.concat ", " (List.map Var.to_string _parent_indicators));
	print_endline ("INDICATORS_QUERY: " ^ String.concat ", " (List.map Var.to_string _query_indicators));
	print_endline ("MSCG: " ^ Subst.subst_to_string mscg_subst);
	print_endline ("TREE_MSCG: " ^ Subst.subst_to_string mscg_tree);
	print_endline ("QUERY_MSCG: " ^ Subst.subst_to_string mscg_query);
	failwith "non-indicator in lifted mscg"
      end;
      mscg_subst, mscg_tree, mscg_query








  (*** creation ***)

  let create () : substitution_tree =
    {
      st_counter = Counter.create ();
      st_root_node = None;
    }





  (*** add ***)

  (* sort nodes by increasing depth *)
  let sort_nodes (nodes: node list) : node list =
    List.sort
      (fun first second ->
	 match first, second with
	   | Leaf _, Leaf _ ->
	       0
	       
	   | Leaf _, Node _ ->
	       -1
	       
	   | Node _, Leaf _ ->
	       1
	       
	   | Node first_node, Node second_node ->
	       compare first_node.sn_depth second_node.sn_depth
      )
      nodes


  (* remove indicator renamings *x -> *y by
     dropping *x -> *y
     and replacing all occurences of *y by *x in the substitution

     the substitution must only contain one offset - i.e. tree_offset *)
  let prune_query_indicators (subst: subst) : subst =

    let bindings_to_prune, bindings_to_keep =
      List.partition
	(fun binding ->
	   Var.is_indicator binding.Subst.sb_var.Subst.sv_var
	   &&
	   (
	     match binding.Subst.sb_term.Subst.st_term with
	       | Term.Var var ->
		   Var.is_universal binding.Subst.sb_var.Subst.sv_var
		   =
	      Var.is_universal var
		
	       | _ ->
		   false
	   )
	)
	subst
    in
      List.fold_left
	(fun acc binding ->
	   let old_term =
	     binding.Subst.sb_term.Subst.st_term
	   in
	   let new_term =
	     Term.request_var binding.Subst.sb_var.Subst.sv_var
	   in
	     Subst.replace_in_bound_terms acc old_term new_term
	)
	bindings_to_keep
	bindings_to_prune


  (* returns the indicator variables contained in the terms bound in subst *)
  let get_bound_indicators (subst: subst) (acc: Var.var list) : Var.var list =
    List.fold_left
      (fun acc binding ->
	 let term_vars =
	   List.find_all
	     Var.is_indicator
	     (Term.vars_of_term binding.Subst.sb_term.Subst.st_term)
	 in
	   Tools.lists_merge Var.equal acc term_vars
      )
      acc
      subst



  (* finds a child node more general than the query.
     returns the node with its substitution *)
  let rec find_subsuming_child (children: node list) (subst_to_parent: subst) :
    (inner_node * subst) option =
    match children with
      | [] ->
	  None

      | Leaf _ :: tail ->
	  (* a leaf can not be (truely) more general, as it contains no open vars.
	     but it might be a variant. *)
	  find_subsuming_child tail subst_to_parent

      | Node child_node :: tail ->
	  begin
	    try
	      let subst_to_child =
		Unification.match_substs
		  ~p_preserving:true ~i_preserving:true
		  tree_offset subst_to_parent child_node.sn_subst
	      in
		Some (child_node, subst_to_child)
	    with
	      | Unification.UNIFICATION_FAIL ->
		  find_subsuming_child tail subst_to_parent
	  end


  (* finds the child node with the 'best' mscg with respect to the query.
     returns the node with its mscg, tree, and query substitution.
     
     assumes that all substs contain only tree_offset as an offset *)
  let find_mscg_child
    (_tree: substitution_tree)
    (_parent: inner_node)
    (_parent_to_query_subst: subst)
    : (node * subst * subst * subst) option =

    let query_indicators =
      get_bound_indicators _parent_to_query_subst []
    in

    let rec do_find_mscg_child
      (children: node list) (acc: (node * subst * subst * subst) option)
      : (node * subst * subst * subst) option =

      match children with
	| [] ->
	    acc
	    
	| child_node :: tail ->
	    begin
	      let open_vars_tree =
		match child_node with
		  | Leaf _ ->
		      []
		  | Node node ->
		      node.sn_open_vars
	      in
	      let mscg_subst, mscg_tree, mscg_query =
		mscg_of_substs
		  _tree
		  (get_node_subst child_node)
		  _parent_to_query_subst
		  _parent.sn_open_vars
		  _parent.sn_closed_vars
		  open_vars_tree
		  _parent.sn_indicators
		  query_indicators
	      in
		match mscg_subst with
		  | [] ->
		      (* ignore empty mscgs *)
		      do_find_mscg_child tail acc

		  | _ ->
		      (* choose the best mscg *)
		      begin
			match acc with
			  | Some (_, old_mscg_subst, old_mscg_tree, old_mscg_query) when
			      (* the new mscg is not better than the old one *)
			      (List.length old_mscg_tree + List.length old_mscg_query)
			      <=
			      (List.length mscg_tree + List.length mscg_query)
			      ->
			      do_find_mscg_child tail acc

			  | _ ->
			      let new_acc =
				Some (child_node, mscg_subst, mscg_tree, mscg_query)
			      in
				do_find_mscg_child tail new_acc
		      end
	    end
    in
      do_find_mscg_child _parent.sn_children None







  (* the term to insert is subsumed by node, so
     - find a subsuming child and descend
     - create an mscg with a child
     - or add as a new leaf *)
  let rec add_node
    (tree: substitution_tree)
    (node: inner_node)
    (* substitution from root to node, including the subsumed query*)
    (subst_to_node: subst)
    (data: data) : unit =

    (* try to find a subsuming child *)
    let subsuming_child =
      find_subsuming_child node.sn_children subst_to_node
    in
      begin
	match subsuming_child with
	  | Some (child_node, subst_to_child) ->
	      (* found *)
	      let old_child_depth =
		child_node.sn_depth
	      in
		(* descend *)
		add_node tree child_node subst_to_child data;

		(* reorder children if necessary *)
		if child_node.sn_depth > old_child_depth then begin
		  node.sn_children <- sort_nodes node.sn_children;

		  if child_node.sn_depth >= node.sn_depth then
		    node.sn_depth <- child_node.sn_depth + 1;
		end;
		  
	  | None ->
	      (* failed, so try to find a child with a non-empty mscg *)
	      begin
		(* normalize the substitution: *)
		(* only use tree_offset *)
		let subst_to_node =
		  Subst.replace_offset subst_to_node query_offset tree_offset
		in

		(* and only keep the bindings of open variables of the parent node *)
		let subst_to_node =
		  prune_query_indicators subst_to_node
		in


		(* extract the query subst from the combination
		   of the node and the query subst *)
		let node_to_query_subst =
		  List.find_all
		    (fun binding ->
		       List.exists
		       (fun open_var ->
			  Subst.var_equal binding.Subst.sb_var open_var
		       )
		       node.sn_open_vars
		    )
		    subst_to_node
		in

		let mscg_child =
		  find_mscg_child
		    tree
		    node
		    node_to_query_subst
		in
		  match mscg_child with
		    | None ->
			let child_indicators =
			  get_bound_indicators node_to_query_subst node.sn_indicators
			in

			let new_child_node = {
			  sl_subst = node_to_query_subst;
			  sl_data = data;
			  sl_indicators = child_indicators;
			}
			in
			  node.sn_children <- (Leaf new_child_node) :: node.sn_children;
			  
		    | Some (child_node, mscg_subst, left_subst, right_subst) ->
			let mscg_open_vars =
			  List.find_all
			    (fun open_var ->
			       not (Var.is_indicator open_var.Subst.sv_var)
			    )
			    (Subst.get_bound_vars2 mscg_subst node.sn_open_vars)
			in                      

			let mscg_indicators =
			  get_bound_indicators mscg_subst node.sn_indicators
			in
			let left_indicators =
			  get_bound_indicators right_subst mscg_indicators
			in
			  
			let left_node = Leaf {
			  sl_subst = right_subst;
			  sl_data = data;
			  sl_indicators = left_indicators;
			}
			in
			let right_node, right_node_depth =
			  match child_node with
			    | Node inner_node ->
				Node {
				  inner_node with
				  sn_subst = left_subst;
				},
				inner_node.sn_depth

			    | Leaf leaf ->
				Leaf {
				  leaf with
				  sl_subst = left_subst;
				},
				0
			in
			let mscg_closed_vars =
			  List.fold_left
			    (fun acc binding ->
			       binding.Subst.sb_var :: acc
			    )			    
			    node.sn_closed_vars
			    mscg_subst
			in
			let mscg_node = {
			  sn_subst = mscg_subst;
			  sn_open_vars = mscg_open_vars;
			  sn_closed_vars = mscg_closed_vars;
			  sn_indicators = mscg_indicators;
			  sn_children = [left_node; right_node];
			  sn_depth = right_node_depth + 1;
			}
			in

			(* most often there are only two children,
			   so these operation are not going to be expensive *)
			let new_children =
			  List.map
			    (fun child ->
			       if child == child_node then
				 Node mscg_node
			       else
				 child
			    )
			    node.sn_children
			in
			  node.sn_children <- sort_nodes new_children;
			  if mscg_node.sn_depth >= node.sn_depth then
			    node.sn_depth <- mscg_node.sn_depth + 1
	      end
      end



  (* add term (with attached data) to the tree *)
  let add (tree: substitution_tree) (term: term) (data: data) : unit =
    match tree.st_root_node with
      | None ->
	  let root_var =
	    Var.create_universal (Counter.next tree.st_counter)
	  in
	    
	  let renumbered_term =
	    clone_term_renumbered tree.st_counter term
	  in
	    
	  let subst =
	    Subst.set_ Subst.empty root_var tree_offset renumbered_term tree_offset
	  in
	    
	  let indicators =
	    List.find_all
	      Var.is_indicator
	      (Term.vars_of_term renumbered_term)
	  in

	  let root_node = Leaf {
	    sl_subst = subst;
	    sl_data = data;
	    sl_indicators = indicators;
	  }
	  in
	    tree.st_root_node <- Some root_node;
	    
      | Some (Leaf leaf) ->
	  let renumbered_term =
	    clone_term_renumbered tree.st_counter term
	  in

	  let query_subst =
	    Subst.set Subst.empty
	      (get_root_var tree)
	      (Subst.make_term renumbered_term tree_offset)
	  in

	  let query_indicators =
	    List.find_all
	      Var.is_indicator 
	      (Term.vars_of_term renumbered_term)
	  in

	  let mscg_subst, mscg_tree, mscg_query =
	    mscg_of_substs
	      tree
	      leaf.sl_subst
	      query_subst
	      [(get_root_var tree)]
	      []
	      []
	      []
	      query_indicators
	  in

	  let mscg_open_vars =
	    List.find_all
	      (fun open_var ->
		 not (Var.is_indicator open_var.Subst.sv_var)
	      )
	      (Subst.get_bound_vars2 mscg_subst [(get_root_var tree)])
	  in

	  let mscg_indicators =
	    get_bound_indicators mscg_subst []
	  in
	  let left_indicators =
	    get_bound_indicators mscg_query mscg_indicators
	  in

	  let left_node = Leaf {
	    sl_subst = mscg_query;
	    sl_indicators = left_indicators;
	    sl_data = data;
	  }
	  in
	  let right_node = Leaf {
	    leaf with
	    sl_subst = mscg_tree;
	  }
	  in
	  let mscg_closed_vars =
	    List.map
	      (fun binding ->
		 binding.Subst.sb_var
	      )
	      mscg_subst
	  in
	  let mscg_node = Node {
	    sn_subst = mscg_subst;
	    sn_open_vars = mscg_open_vars;
	    sn_closed_vars = mscg_closed_vars;
	    sn_children = [left_node; right_node];
	    sn_depth = 1;
	    sn_indicators = mscg_indicators;
	  }
	  in
	    tree.st_root_node <- Some mscg_node;
	    
      | Some (Node node) ->
	  begin
	    let renumbered_term =
	      clone_term_renumbered tree.st_counter term
	    in
	      try
		(* instance of root? *)
		let subst_to_root =
		  Unification.match_terms
		    ~p_preserving:true ~i_preserving:true
		    (get_root_term tree) tree_offset
		    renumbered_term query_offset
		in
		  add_node tree node subst_to_root data;

	      with
		| Unification.UNIFICATION_FAIL ->
		    (* no, so create a new root *)

		    let query_subst =
		      Subst.set
			Subst.empty
			(get_root_var tree)
			(Subst.make_term renumbered_term tree_offset)
		    in
		    let query_indicators =
		      List.find_all
			Var.is_indicator 
			(Term.vars_of_term renumbered_term)
		    in

		    let mscg_subst, mscg_tree, mscg_query =
		      mscg_of_substs
			tree
			node.sn_subst
			query_subst
			[(get_root_var tree)]
			[]
			[]
			[]
			query_indicators
		    in

		    let mscg_open_vars =
		      List.find_all
			(fun open_var ->
			   not (Var.is_indicator open_var.Subst.sv_var)
			)
			(Subst.get_bound_vars2 mscg_subst [(get_root_var tree)])
		    in

		    let mscg_indicators =
		      get_bound_indicators mscg_subst []
		    in
		    let left_indicators =
		      get_bound_indicators mscg_query mscg_indicators
		    in

		    let left_node = Leaf {
		      sl_subst = mscg_query;
		      sl_indicators = left_indicators;
		      sl_data = data;
		    }
		    in
		    let right_node = Node {
		      node with
		      sn_subst = mscg_tree;
		    }
		    in
		    let mscg_closed_vars =
		      List.map
			(fun binding ->
			   binding.Subst.sb_var
			)
			mscg_subst
		    in
		    let mscg_node = Node {
		      sn_subst = mscg_subst;
		      sn_closed_vars = mscg_closed_vars;
		      sn_open_vars = mscg_open_vars;
		      sn_children = [left_node; right_node];
		      sn_depth = node.sn_depth + 1;
		      sn_indicators = mscg_indicators;
		    }
		    in
		      tree.st_root_node <- Some mscg_node;
	  end









  (*** remove ***)

  (* merge the substitution of a node and its only child such that
     open vars/indicator vars, ... remains unchanged the child's children *)
  let fold_node_with_child
    (* the open vars of the parent of node *)
    (node_parent_open_vars: Subst.var list)
    (* the mscg and its child to fold *)
    (node_subst: subst) (child_subst: subst)
    : subst =

    (* replace the vars in the child subst
       which are bound to another var by their bindings *)
    let applied_child_subst =
      List.fold_left
	(fun acc binding ->
	   Subst.replace_in_bound_terms
	     acc
	     (Term.request_var binding.Subst.sb_var.Subst.sv_var)
	     binding.Subst.sb_term.Subst.st_term
	)
	child_subst
	child_subst
    in

    (* replace the open vars of the parent subst
       which are bound in the child node
       by their bindings *)
    let applied_parent_subst =
      List.fold_left
	(fun acc binding ->
	   Subst.replace_in_bound_terms
	     acc
	     (Term.request_var binding.Subst.sb_var.Subst.sv_var)
	     binding.Subst.sb_term.Subst.st_term
	)
	node_subst
	applied_child_subst
    in
      (* the open vars of node are closed in applied_subst.
	 the bindings open vars of node's parent which were not bound in node
	 but in child must be propagated. *)
      List.fold_left
	(fun acc binding ->
	   if
	     List.exists
	       (fun open_var ->
		  Subst.var_equal binding.Subst.sb_var open_var
	       )
	       node_parent_open_vars
	   then
	     binding :: acc
	   else
	     acc
	)
	applied_parent_subst
	applied_child_subst
	

  (* descend into node's children *)
  (* returns
     - a node (with only one child) to be replaced by its folded node
     - the data of the removed node  *)
  let rec remove_in_children
    (func: node -> subst -> subst)
    (data: data option)
    (node: inner_node)
    (subst_to_node: subst)
    (children: node list)
    : (inner_node * node) option * bool =
    
    match children with
      | [] ->
	  None, false
	      
      | child_node :: tail ->
	  begin
	    try
	      let subst_to_child =
		func child_node subst_to_node
	      in
		begin
		  match child_node with
		    | Leaf leaf ->
			if
			  match data with
			    | None ->
				true

			    | Some data ->
				data_is_equal leaf.sl_data data
			then begin
			  (* found the leaf to remove *)
			  node.sn_children <-
			  List.find_all
			    (fun child -> child != child_node)
			    node.sn_children;
			  
			  None, true
			end

			else
			  (* didn't find the node under inner_node,
			     so try the next child *)
			  remove_in_children func data node subst_to_node tail
			  
		    | Node inner_node ->
			let replace_node_with, removed =
			  remove_node func data inner_node subst_to_child node.sn_open_vars
			in
			  begin
			    if removed then
			      (* found the node under inner_node, so return *)
			      replace_node_with, removed

			    else
			      (* didn't find the node under inner_node,
				 so try the next child *)
			      remove_in_children func data node subst_to_node tail
			  end
		end
	    with
	      | Unification.UNIFICATION_FAIL ->
		  (* node not suited for descending *)
		  remove_in_children func data node subst_to_node tail
	  end
	  

  and remove_node
    (func: node -> subst -> subst)
    (data: data option)
    (node: inner_node)
    (subst_to_node: subst)
    (parent_open_vars: Subst.var list)
    : (inner_node * node) option * bool =

    let replace_node_with, removed =
      remove_in_children func data node subst_to_node node.sn_children
    in
      (* just removed a leaf of a child of this node
	 and then folded the child and its now only remaining child? *)
      begin
	match replace_node_with with
	  | None ->
	      ()
	      
	  | Some (replace_node, replace_with) ->
	      begin
		let new_children =
		  List.map
		    (fun child ->
		       match child with
			 | Node inner_node when
			     inner_node == replace_node ->
			     replace_with
			       
			 | _ ->
			     child
		    )
		    node.sn_children 
		in
		  node.sn_children <- sort_nodes new_children;
	      end
      end;

      (* removed a node somewhere under node? *)
      begin
	if not removed then begin
	  replace_node_with, removed
	end
	else begin
	  (* yes, so update the depth *)
	  node.sn_depth <- 1 +
	  (List.fold_left
	     (fun acc child ->
		match child with
		  | Leaf _ ->
		      acc
		      
		  | Node inner_node ->
		      max acc inner_node.sn_depth
	     )
	     0
	     node.sn_children
	  );

	  (* is this the node whose child has been removed?
	     if so, it might have only once child left which needs to be folded.
	     every node must have at least two children. *)
	  begin
	    match node.sn_children with
	      | [] ->
		  failwith "Substitution_tree.remove_node: no children"
		  
	      | Leaf only_child :: [] ->
		  let new_subst =
		    fold_node_with_child parent_open_vars node.sn_subst only_child.sl_subst
		  in
		  let replace_with =
		    Leaf {
		      only_child with
			sl_subst = new_subst;
		    }
		  in
		    Some (node, replace_with), removed
		      
	      | Node only_child :: [] ->
		  let new_subst =
		    fold_node_with_child parent_open_vars node.sn_subst only_child.sn_subst
		  in
		  let replace_with =
		    Node {
		      only_child with
			sn_subst = new_subst;
		    }
		  in
		    Some (node, replace_with), removed
		      
	      | _ ->
		  None, removed
	  end
	end
	  
      end



  let remove (tree: substitution_tree) (term: term) (data: data option) : bool =
    match tree.st_root_node with
      | None ->
	  false
	  
      | Some node ->
	  let normalized_term =
	    clone_term_renumbered tree.st_counter term
	  in
	    
	  let query_subst =
	    Subst.set Subst.empty
	      (get_root_var tree)
	      (Subst.make_term normalized_term query_offset)
	  in
	    
	  let func node subst =
	    Unification.match_substs
	      ~p_preserving:true ~i_preserving:true
	      tree_offset subst (get_node_subst node)
	  in

	    try
	      let subst_to_node =
		func node query_subst
	      in
		match node with
		  | Leaf leaf ->
		      tree.st_root_node <- None;
		      true

		  | Node inner_node ->
		      begin
			let replace_with_node, removed =
			  remove_node func data inner_node subst_to_node [(get_root_var tree)];
			in
			  begin
			    match replace_with_node with
			      | None ->
				  ()

			      | Some (replace_node, replace_with) ->
				  if inner_node != replace_node then
				    failwith "Substitution_tree.remove_variant";
				  tree.st_root_node <- Some replace_with;
			  end; 
			  
			  removed
		      end
	    with
	      | Unification.UNIFICATION_FAIL ->
		  false









  (*** tests against the tree ***)

  let rec find_node ?(find_all: bool = false) (func: node -> subst -> subst) (acc: data list)
    (node: node) (subst_to_node: subst) : data list =

    match node with
      | Leaf leaf ->
	  if find_all then
	    leaf.sl_data :: acc
	  else
	    raise (FOUND leaf.sl_data)
	      
      | Node inner_node ->
	  List.fold_left
	  (fun acc child_node ->
	     let subst_to_child, success =
	       try
		 func child_node subst_to_node, true
	       with
		 | Unification.UNIFICATION_FAIL ->
		     Subst.empty, false
	     in
	       if success then
		 find_node ~find_all:find_all func acc child_node subst_to_child
	       else
		 acc
	  )
	  acc
	  inner_node.sn_children


  (* find the first or all elements according to the specified function *)
  let start_find ~(find_all: bool) ~(p_preserving: bool)
    (func_root: term -> substitution_tree -> subst)
    (func: node -> subst -> subst)
    (tree: substitution_tree) (term: term) : data list =
    match tree.st_root_node with
      | None ->
	  []

      | Some node ->
	  begin
	    try
	      let subst_to_node =
		func_root term tree
	      in
		find_node ~find_all:find_all func [] node subst_to_node
	    with
	      | Unification.UNIFICATION_FAIL ->
		  []
	  end


  (* find the first element according to the specified function *)
  let start_find_first ~(p_preserving: bool)
    (func_root: term -> substitution_tree -> subst)
    (func: node -> subst -> subst)
    (tree: substitution_tree) (term: term) :
    data option =

    try
      let found =
	start_find
	  ~find_all:false ~p_preserving:p_preserving
	  func_root
	  func
	  tree term
      in
	match found with
	  | [] ->
	      None
	      
	  | _ ->
	      failwith "Substitution_tree.start_find_first"
    with
      | FOUND data ->
	  Some data





  (*** variants ***)

  let find_all_variants (tree: substitution_tree) (term: term) : data list =
    match tree.st_root_node with
      | None ->
	  []

      | Some node ->
	  let normalized_term =
	    clone_term_renumbered tree.st_counter term
	  in

	  let func node subst =
	    Unification.match_substs
	      ~p_preserving:true ~i_preserving:true
	      tree_offset subst (get_node_subst node)
	  in
	    try
	      let subst_to_node =
		Unification.match_terms
		  ~p_preserving:true ~i_preserving:true
		  (get_root_term tree) tree_offset
		  normalized_term query_offset
	      in
		find_node ~find_all:true func [] node subst_to_node
	    with
	      | Unification.UNIFICATION_FAIL ->
		  []



  (*** generalization ***)

  let func_generalization_root ~(p_preserving: bool) (query_term: term) (tree: substitution_tree) : subst =
    Unification.match_terms
      ~p_preserving:p_preserving
      (get_root_term tree) tree_offset
      query_term query_offset

  let func_generalization ~(p_preserving: bool) (node: node) (subst: subst) : subst =
    Unification.match_substs
      ~p_preserving:p_preserving
      tree_offset subst (get_node_subst node)

  let find_all_generalizations ~(p_preserving: bool) (tree: substitution_tree) (term: term) : data list =
    start_find
      ~find_all:true ~p_preserving:p_preserving
      (func_generalization_root ~p_preserving:p_preserving)
      (func_generalization ~p_preserving:p_preserving)
      tree term

  let find_generalization ~(p_preserving: bool) (tree: substitution_tree) (term: term) : data option =
    start_find_first
      ~p_preserving:p_preserving
      (func_generalization_root ~p_preserving:p_preserving)
      (func_generalization ~p_preserving:p_preserving)
      tree term





  (*** unifiable ***)

  let func_unifiable_root ~(p_preserving: bool) (query_term: term) (tree: substitution_tree) : subst =
    Unification.unify_terms
      ~p_preserving:p_preserving
      (get_root_term tree) tree_offset
      query_term query_offset

  let func_unifiable ~(p_preserving: bool) (node: node) (subst: subst) : subst =
    Unification.unify_substs
      ~p_preserving:p_preserving
      subst (get_node_subst node)

  let find_all_unifiable ~(p_preserving: bool) (tree: substitution_tree) (term: term) : data list =
    start_find
      ~find_all:true ~p_preserving:p_preserving
      (func_unifiable_root ~p_preserving:p_preserving)
      (func_unifiable ~p_preserving:p_preserving)
      tree term

  let find_unifiable ~(p_preserving: bool) (tree: substitution_tree) (term: term) : data option =
    start_find_first
      ~p_preserving:p_preserving
      (func_unifiable_root ~p_preserving:p_preserving)
      (func_unifiable ~p_preserving:p_preserving)
      tree term





  (*** instance ***)

  let func_instance_root ~(p_preserving: bool) (query_term: term) (tree: substitution_tree) : subst =
    Unification.unify_terms
      ~p_preserving:p_preserving ~i_preserving:true
      (get_root_term tree) tree_offset
      query_term query_offset

  let func_instance ~(p_preserving: bool) (node: node) (subst: subst) : subst =
    Unification.unify_substs
      ~p_preserving:p_preserving ~i_preserving:true
      subst (get_node_subst node)

  let find_all_instances ~(p_preserving: bool) (tree: substitution_tree) (term: term) : data list =
    start_find
      ~find_all:true ~p_preserving:p_preserving
      (func_instance_root ~p_preserving:p_preserving)
      (func_instance ~p_preserving:p_preserving)
      tree term

  let find_instance ~(p_preserving: bool) (tree: substitution_tree) (term: term) : data option =
    start_find_first
      ~p_preserving:p_preserving
      (func_instance_root ~p_preserving:p_preserving)
      (func_instance ~p_preserving:p_preserving)
      tree term
      

      

  (*** productivity ***)

  (* assumes that a context literals contains
     either universal or parametric literals.

     As long as it's not known if the current substitution will be completed
     to a universal or parametric term use unification.
     As long as there are no indicator variables
     this doesn't make any difference.

     But as soon as there is a parametric indicator variable
     matching must be used. *)

  let func_producer node subst =
    match get_node_indicators node with
      | indicator :: _ when
	  Var.is_parametric indicator ->

	  (* a parametric node must be an instance of the producer *)
	  func_instance ~p_preserving:false node subst

      | _ ->
	  (* a universal node must be unifying with the producer *)
	  func_unifiable ~p_preserving:false node subst


  let rec find_non_productive_node
    (node: node)
    (func_producer: node -> subst -> subst)
    (subst_to_node_producer: subst)
    (subst_to_node_produced: subst)
    : unit =

    match node with
      | Leaf leaf ->
	  raise (FOUND leaf.sl_data)
	      
      | Node inner_node ->
	  List.iter
	  (fun child_node ->
	     let subst_to_child_producer, subst_to_child_produced, success =
	       try
		 func_producer child_node subst_to_node_producer,
		 func_generalization ~p_preserving:false child_node subst_to_node_produced,
		 true
	       with
		 | Unification.UNIFICATION_FAIL ->
		     Subst.empty, Subst.empty, false
	     in
	       if success then
		 find_non_productive_node
		   child_node
		   func_producer
		   subst_to_child_producer
		   subst_to_child_produced
	  )
	  inner_node.sn_children


  (* find the first or all elements according to the specified function *)
  let find_non_productive (tree: substitution_tree) (producer: term) (produced: term) : data option =

    match tree.st_root_node with
      | None ->
	  None

      | Some node ->
	  begin
	    try
	      let subst_to_node_produced =
		func_generalization_root ~p_preserving:false produced tree
	      in
	      let func_producer, subst_to_node_producer =
		match get_node_indicators node with
		  | [] ->
		      func_producer,
		      func_unifiable_root ~p_preserving:false producer tree

		  | indicator :: _ ->
		      if Var.is_parametric indicator then
			(* a parametric node must be an instance of the producer *)
			func_instance ~p_preserving:false,
			func_instance_root ~p_preserving:false producer tree
		      else
			(* a universal node must be unifying with the producer *)
			func_unifiable ~p_preserving:false,
			func_unifiable_root ~p_preserving:false producer tree

	      in
		find_non_productive_node
		  node
		  func_producer
		  subst_to_node_producer
		  subst_to_node_produced;
		None
	    with
	      | Unification.UNIFICATION_FAIL ->
		  None

	      | FOUND data ->
		  Some data
	  end







  (*** msg ***)
	  
  (* is the subst a renaming from query variables to term indicator variables? *)
  let is_producer (subst: subst) : bool =
    try
      ignore (
	List.fold_left (
	  fun (mapping: (var * var) list) binding ->

	    (* a query var? *)
	    if binding.Subst.sb_var.Subst.sv_offset = query_offset then begin

	      (* bound to itself, so a true match *)
	      if binding.Subst.sb_term.Subst.st_offset = query_offset then begin
		raise Exit;
	      end;
	      
	      match binding.Subst.sb_term.Subst.st_term with
		| Term.Var bound_var when
		    Var.is_universal binding.Subst.sb_var.Subst.sv_var
		    =
		    Var.is_universal bound_var
		    ->
		    (* a bijective mapping? *)
		    Tools.mapping_extend
		      mapping
		      Var.equal
		      Var.equal
		      binding.Subst.sb_var.Subst.sv_var
		      bound_var
		      
		| _ ->
		    (* not bound to a indicator variable *)
		    raise Exit
	    end

	    (* a term non-indicator var *)
	    else
	      mapping
	)
	[]
	subst
      );
      true

    with
      | Exit ->
	  false



  let rec find_non_msg_node
    (node: node)
    (subst_to_node_msg: subst)
    (subst_to_node_produced: subst)
    : unit =

    match node with
      | Leaf leaf ->
	  (* ignore the producer? *)
	  if not (is_producer subst_to_node_msg) then
	    raise (FOUND leaf.sl_data)
	      
      | Node inner_node ->
	  List.iter
	  (fun child_node ->
	     let subst_to_child_msg, subst_to_child_produced, success =
	       try
		 func_instance ~p_preserving:false child_node subst_to_node_msg,
		 func_generalization ~p_preserving:false  child_node subst_to_node_produced,
		 true
	       with
		 | Unification.UNIFICATION_FAIL ->
		     Subst.empty, Subst.empty, false
	     in
	       if success then
		 find_non_msg_node
		   child_node
		   subst_to_child_msg
		   subst_to_child_produced
	  )
	  inner_node.sn_children


  let find_non_msg (tree: substitution_tree) (producer: term) (produced: term) : data option =

    match tree.st_root_node with
      | None ->
	  None

      | Some node ->
	  begin
	    try
	      let subst_to_node_msg =
		func_instance_root ~p_preserving:false producer tree
	      in
	      let subst_to_node_produced =
		func_generalization_root ~p_preserving:false produced tree
	      in
		find_non_msg_node
		  node
		  subst_to_node_msg
		  subst_to_node_produced;
		None
	    with
	      | Unification.UNIFICATION_FAIL ->
		  None

	      | FOUND data ->
		  Some data
	  end





  let clear (tree: substitution_tree) : unit =
    tree.st_root_node <- None;
    Counter.set tree.st_counter 0

end
