(* $Id: symbol.ml,v 1.5 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(* All symbols are stored in a symbol table,
   thus each symbol occurs only once in the system.

   A symbol is represented by a unique id inside the system,
   name and arity are (at least currently) only needed for printing *)

(* a symbol is represented by its id *)
type symbol =
    int

(* the symbol's real data *)
type symbol_data = {
  name: string;
  arity: int;
}








(*** global data ***)


(* mapping from a symbol's id to its name and arity. *)
let (symbol_table: (symbol, symbol_data) Hashtbl.t) = 
  Hashtbl.create 128

(* mapping from a symbol's name to its id.
   needed on creation of a symbol to check
   if it's already registered and with which arity. *)
let (name_table: (string, symbol) Hashtbl.t) = 
  Hashtbl.create 128



let symbol_id_counter =
  Counter.create ()






(*** creation ***)




let create_symbol (name: string) (arity: int) : symbol =
  try
    (* does symbol with this name does already exist? *)
    let id =
      Hashtbl.find name_table name
    in
    let old_entry =
      Hashtbl.find symbol_table id
    in
      (* yes, so check if the old an new arity are equal *)
      if arity <> old_entry.arity then begin
	failwith ("Symbol.create_symbol: tried to reregister symbol " ^ name
		  ^ " with the arity " ^ string_of_int old_entry.arity
		  ^ " with the new arity " ^ string_of_int arity)
(*	Hashtbl.remove symbol_table id;
	raise Not_found;*)
      end
      else
	id
  with
    | Not_found ->
	(* no, so create and register it *)
	let id =
	  try
	    Counter.next symbol_id_counter
	  with
	    | Counter.OVERFLOW ->
		failwith "Symbol.create_symbol: counter_overflow"
	in
	  Hashtbl.add symbol_table id { name = name; arity = arity };
	  Hashtbl.add name_table name id;
	  id


(* create an anonymous constant *)
let create_constant () =
  try
    Counter.next symbol_id_counter
  with
    | Counter.OVERFLOW ->
	failwith "Symbol.create_constant: counter_overflow"





let id_of_symbol (symbol: symbol) : int =
  symbol


(* if symbol is unknown it is interpreted as an anonymous constant.
   slow, but currently only used for printing *)
let name_of_symbol (symbol: symbol) : string =
  try
    (Hashtbl.find symbol_table symbol).name
  with
    | Not_found ->
	"__" ^ string_of_int symbol ^ "__"


(* if symbol is unknown it is interpreted as an anonymous constant.
   slow, but currently only used for printing *)
let arity_of_symbol (symbol: symbol) : int =
  try
    (Hashtbl.find symbol_table symbol).arity
  with
    | Not_found ->
	0




let equal (symbol1: symbol) (symbol2: symbol) = 
  symbol1 == symbol2



let to_string (symbol: symbol) = 
(*  name_of_symbol symbol ^ "/" ^ string_of_int (arity_of_symbol symbol) *)
  name_of_symbol symbol
