(* $Id: term.ml,v 1.19 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(* the term database consists basically of hash tables containing the terms.

   the hash values are for a
   - Var: its var's id
   - Const: its symbol's id
   - Func: its enclosed term's hash_value
   - literal: its enclosed term's hash_value

   terms are simply stored in a weak set.
   as long as a term is used outside the set,
   it is kept alive and subsequent creation requests return the same term.
   as soon as the term is not referenced outside the set
   it may be garbage collected.
*)


(*** types ***)

type var = Var.var
type symbol = Symbol.symbol

type term =
  | Var of var
  | Const of symbol
  | Func of symbol * term array * int

type literal = {
  sign: bool;
  term: term
}

type clause =
    literal list



(* term database *)
module TermTable =
  Weak.Make (
    struct
      type t = term

      let rec are_terms_equal_at (index: int) (terms1: term array) (terms2: term array) : bool =
	if index >= Array.length terms1 then
	  true
	else begin
	  if terms1.(index) == terms2.(index) then
	    are_terms_equal_at (index + 1) terms1 terms2
	  else
	    false
	end


      (* term_equal can not be used, as it uses object identity 
	 which is assured by this module *)
      let equal (term1: term) (term2: term) : bool =
	match term1, term2 with
	  | Var var1, Var var2 ->
	      Var.equal var1 var2

	  | Const const1, Const const2 ->
	      Symbol.equal const1 const2

	  | Func (symbol1, terms1, _), Func (symbol2, terms2, _) ->
	      (Symbol.equal symbol1 symbol2)
	      &&
	      (are_terms_equal_at 0 terms1 terms2)

	  | _ ->
	      false

      let hash (term: term) : int =
	match term with
	  | Var var ->
	      Var.id_of_var var

	  | Const const ->
	      Symbol.id_of_symbol const

	  | Func (_, _, hash_value) ->
	      hash_value
    end
  )


(* literal database *)
module LiteralTable =
  Weak.Make (
    struct
      type t = literal

      let equal (literal1: literal) (literal2: literal) : bool =
	(literal1.sign = literal2.sign)
	&&
	(literal1.term == literal2.term)

      let hash (literal: literal) : int =
	match literal.term with
	  | Var var ->
	      (* invalid literal type *)
	      failwith ("Term.LiteralTable.hash");
	      (* Var.id_of_var var*)
	      
	  | Const symbol ->
	      (* e.g. a special literal like -v *)
	      Symbol.id_of_symbol symbol
	      
	  | Func (_, _, hash_value) ->
	      hash_value

    end
  )



type term_database = {
  (* contans the terms: consts, vars, functions *)
  terms: TermTable.t;
  (* contains the literals *)
  literals: LiteralTable.t;
}






(*** globals ***)

let global : term_database =
  {
    terms = TermTable.create 2048;
    literals = LiteralTable.create 512;
  }






(*** string representation ***)

let rec term_to_string (term: term) : string =
  match term with
    | Var var ->
	Var.to_string var

    | Const const ->
	Symbol.to_string const

    | Func (symbol, terms, _) ->
	Symbol.to_string symbol
	^ "("
	^ String.concat ", " (List.map term_to_string (Array.to_list terms))
	^ ")"



(*** comparison ***)

(* ensure that equal terms have equal pointers *)
let rec debug_term_equal (term1: term) (term2: term) : bool =
  match term1, term2 with
    | Var v1, Var v2 ->
	Var.equal v1 v2

    | Const s1, Const s2 ->
	Symbol.equal s1 s2

    | Func (s1, t1, _), Func (s2, t2, _) ->
	Symbol.equal s1 s2
	&&
	Tools.array_fold2
	  (fun acc term1 term2 ->
	     acc
	     &&
	     debug_term_equal term1 term2
	  )
	  true
	  t1
	  t2

    | _ ->
	false

let rec term_equal (term1: term) (term2: term) : bool =
  (* check if the term database contains all terms *)
  if
    Config.debug
    &&
    not (
      (debug_term_equal term1 term2)
      =
      (term1 == term2)
    )
  then begin
    failwith ("Term.term_equal: " ^ term_to_string term1 ^ " - " ^ term_to_string term2);
  end;

  term1 == term2

let literal_equal (literal1: literal) (literal2: literal) : bool =
  (* check if the term database contains all literals *)
  if
    Config.debug
    &&
    not (
      (
	(literal1.sign == literal2.sign)
	&&
	(debug_term_equal literal1.term literal2.term)
      )
      =
      (literal1 == literal2)
    )
  then begin
    failwith "Term.literal_equal";
  end;

  literal1 == literal2

let clause_equal (clause1: clause) (clause2: clause) : bool =
  Tools.lists_unordered_equal literal_equal clause1 clause2







(*** creation ***)



(* these functions try to find a requested term in the database
   and to return the old instance.

   otherwise a fresh instance is created, stored, and returned. *)


let request_term (term: term) : term =
  try
    TermTable.find global.terms term
  with
    | Not_found ->
	TermTable.add global.terms term;
 	term


let request_var (var: var) : term =
  let new_term =
    Var var
  in
    request_term new_term

let request_const (symbol: symbol) : term =
  let new_term =
    Const symbol
  in
    request_term new_term

let request_func ((symbol, terms): symbol * term array) : term =
  
  (* 131 is known to work quite well by ... someone Cesare Tinelli knows.
     tests seem to support it *)
  let hash (old_value: int) (new_value: int) =
    (old_value + new_value) * 131
  in
    
  let hash_term (term: term) : int =
    match term with
      | Var var ->
	  Var.id_of_var var
	    
      | Const symbol ->
	  Symbol.id_of_symbol symbol
  
      | Func (_, _, hash_value) ->
	  hash_value
  in

  let hash_value =
    Array.fold_left 
      (fun old_hash term ->
	 hash old_hash (hash_term term)
      ) 
      (hash 0 (Symbol.id_of_symbol symbol))
      terms
  in

  let new_term =
    Func (symbol, terms, hash_value)
  in
    request_term new_term




let request_literal (sign: bool) (term: term) : literal =
  let new_literal = {
    sign = sign;
    term = term;
  }
  in
    try
      LiteralTable.find global.literals new_literal
    with
      | Not_found ->
	  LiteralTable.add global.literals new_literal;
 	  new_literal


let request_negated_literal (literal: literal) : literal =
  request_literal (not literal.sign) literal.term


let request_skolemized_term (term: term) : term =
  let (var_mapping: (var * term) list ref) =
    ref []
  in

  let rec do_skolemize (term: term) : term=
    match term with
      | Var var ->
	  if Var.is_universal var then begin
	    try
	      List.assoc var !var_mapping
	    with
	      | Not_found ->
		  let fresh_skolem =
		    Symbol.create_constant ()
		  in
		  let fresh_const =
		    request_const fresh_skolem
		  in
		    var_mapping := (var, fresh_const) :: !var_mapping;
		    fresh_const
	  end
	  else
	    term

      | Const _ ->
	  term

      | Func (symbol, terms, _) ->
	  request_func (symbol, Array.map do_skolemize terms)

  in
    do_skolemize term


let request_skolemized_literal (literal: literal) : literal =
  request_literal literal.sign (request_skolemized_term literal.term)



let rec replace_in_term (term: term) (old_term: term) (new_term: term) : term =
  if term_equal term old_term then
    new_term

  else if term_equal old_term new_term then
    term

  else begin
    match term with
      | Var _
      | Const _ ->
	  term
	  
      | Func (symbol, terms, _) ->
	  let new_terms =
	    Array.map
	      (fun term ->
		 replace_in_term term old_term new_term)
	      terms
	  in
	    request_func (symbol, new_terms)
  end







(*** special literals ***)

let true_literal : literal =
  request_literal true (request_const (Symbol.create_symbol "true" 0))

let false_literal : literal =
  request_literal true (request_const (Symbol.create_symbol "false" 0))

let assert_literal : literal =
  request_literal true (request_const (Symbol.create_constant ()))

let null_literal : literal =
  request_literal true (request_const (Symbol.create_constant ()))

let default_v : literal ref =
  ref (request_literal (Config.get_use_plus_v ()) (request_const (Symbol.create_constant ())))

let get_default_v () : literal =
  !default_v









(*** decomposition ***)

let rec do_vars_of_term (acc: var list) (term: term) : var list =
  match term with
    | Var var ->
	(* store each variable only once *)
	if List.exists
	  (fun acc_var ->
	     Var.equal acc_var var
	  )
	  acc
	then
	  acc
	else
	  var :: acc

    | Const _ ->
	acc

    | Func (_, terms, _) ->
	Array.fold_left do_vars_of_term acc terms


let vars_of_term (term: term) : var list =
  do_vars_of_term [] term


let vars_of_literal (literal: literal) : var list =
  vars_of_term literal.term





(*** string representation ***)




let literal_to_string (literal: literal) : string =
  let prefix =
    if literal.sign then
      "+"
    else
    "-"
  in
  let term =
    if literal_equal !default_v literal then
      "__v__"
    else if literal_equal assert_literal literal then
      "__assert__"
    else if literal_equal null_literal literal then
      "__null__"
    else
      term_to_string literal.term
  in
    prefix ^ term



let clause_to_string (clause: clause) : string =
  "[" ^ String.concat ", " (List.map literal_to_string clause) ^ "]"



(*** restart ***)

(* after a restart the problem and special literals have to be reregistered
   in order to be kept in the new database.
   I.e. the old terms and literals are reinserted into the database *)

let reregister_term (term: term) : unit =

  let rec do_register (term: term) : unit =
    match term with
      | Var _
      | Const _ ->
	  ignore (request_term term)

      | Func (symbol, terms, _) ->
	  Array.iter do_register terms;
	  ignore (request_term term)


  in
    do_register term


let reregister_literal (literal: literal) : unit =
  reregister_term literal.term;

  try
    ignore (LiteralTable.find global.literals literal)
  with
    | Not_found ->
	LiteralTable.add global.literals literal





let restart (clauses: clause list) : unit =
  TermTable.clear global.terms;
  LiteralTable.clear global.literals;

  (* reregister special literals *)
  reregister_literal true_literal;
  reregister_literal false_literal;
  reregister_literal assert_literal;
  reregister_literal null_literal;

  (* the sign plus_v might have changed *)
  default_v :=
  request_literal (Config.get_use_plus_v ()) (request_const (Symbol.create_constant ()));

  (* reregister problem clauses *)
  List.iter
    (fun clause ->
       List.iter reregister_literal clause
    )
    clauses
