(* $Id: term.mli,v 1.16 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** terms, literals, clauses

  All terms are stored in a database
  so that each term exists only once in the system
  (low memory usage by term sharing, fast term comparison by pointer equality).

  Therefore terms and literals must never be explicitly built using the constructors.
  Instead they must be build by using the appropriate {b request} functions.

  Fast access is based on a hash value for every term.
  For variables and constants that's their id,
  for terms it is (for efficiency) stored in the term data structure. *)


(** {6 Types} *)

type var = Var.var
type symbol = Symbol.symbol

(** a term can be a variable, a constant or a term *)
type term =
  | Var of var
  | Const of symbol
  | Func of symbol * term array * int
      (** symbol * arguments * hash_value  *)

(** a literal is a signed term - the term is interpreted as a predicate *)
type literal = {
  sign: bool; 
  term: term; 
}

(** a clause: a list, not a set as in theory. *)
type clause =
    literal list




(** {6 Functions} *)

(** {2 Special Terms} *)

(** the {e true} literal (sign is true) *)
val true_literal: literal

(** the {e false} literal (sign is true) *)
val false_literal: literal

(** used for initialization *)
val null_literal: literal

(** used to mark an assert gap in a context unifier ({!Context_unifier}) *)
val assert_literal: literal

(** the default context literal {e -v} or {e +v}
  based on {!Config.get_use_plus_v}.

  the sign ({!Config.get_use_plus_v} and thus the literal
  might change between proofs. *)
val get_default_v: unit -> literal




(** {2 Request} *)

(** requests a [term.Var] term from the database.

  there might be different instances of the same variable in the system
  (e.g. the universal variable with the id 0).
  As the first request to make_var creates a database entry for a variable,
  and the following requests return this first stored instance,
  this is of no consequence for the term sharing *)
val request_var: var -> term

(** requests a [term.Const] term from the database *)
val request_const: symbol -> term

(** requests a [term.Func] term from the database *)
val request_func: symbol * term array -> term


(** requests a [literal] from the database.
  implicitely registers [literal].

  A literal of a {!Var} term must not be created. *)
val request_literal: bool -> term -> literal

(** requests a negation of a literal from the database.
  implicitely registers the negated [literal]. *)
val request_negated_literal : literal -> literal

(** requests the skolemization of a literal from the database.
  implicitely registers the skolemized [literal]. *)
val request_skolemized_literal: literal -> literal

(** [replace_in_term term old_term new_term]
  builds a new term from [term] by replacing
  every occurrence of [old_term] with [new_term] *)
val replace_in_term: term -> term ->  term -> term




(** {2 Decomposition} *)

(** the variables of the term (each only once) *)
val vars_of_term: term -> var list

(** the variables of the literal (each only once) *)
val vars_of_literal: literal -> var list


(** {2 Comparison} *)

val term_equal : term -> term -> bool

val literal_equal : literal -> literal -> bool

(** currently only used for testing *)
val clause_equal : clause -> clause -> bool




(** {2 String Representation} *)

(** Example: {b p(a, _0)} *)
val term_to_string : term -> string

(** Example: {b -p(a, _0)} *)
val literal_to_string : literal -> string

(** Example: {b \{ p(f(a), _0), -q(b) \}} *)
val clause_to_string : clause -> string





(** {2 Restart} *)

(** the problem clauses have to be given in order to be reregistered
  to the fresh and empty database *)
val restart: clause list -> unit
