(* $Id: term_indexing.mli,v 1.7 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** Term indexing interface

  The signature for a term indexing module.

  A concrete implementation of the term indexing module
  should look like this:

  [module Make (DataType: Term_indexing.DataType):
  Term_indexing.Index with type data = DataType.t].

  For an example see {!Substitution_tree.Make}
*)


(** type of data stored in an indexing module *)
module type DataType = 
  sig
    type t

    val is_equal: t -> t -> bool
    val to_string: t -> string
  end



(** The output signature for an indexing module.

  Mutable.

  Insertion of variants of already inserted terms is allowed
  and keeps the data attached to all variants intact.

  To be used for exactly one predicate/function symbol,
  not for several at once.
  Thus adding f(x) and g(x) is not allowed.

  For an explanation for {e p_preserving} and {e i_preserving} see {!Subst}.

*)
module type Index =
  sig
    type term = Term.term
    type subst = Subst.subst

    (** the index *)
    type t

    (** the contained data *)
    type data

      
    (** {2 Manipulation} *)

    (** creates an empty index *)
    val create: unit -> t

    (** removes all entries *)
    val clear: t -> unit

    (** insert [data] attached to the [term] *)
    val add: t -> term -> data -> unit
      
    (** removes a variant of [term].
      if [data] is given [data] must be attached
      to the removed node is attached.

      returns false if no such entry exists. *)
    val remove: t -> term -> data option -> bool


    (** {2 Traversal} *)

      (** visits each term/data pair stored in the index *)
    val iter: (term -> data -> unit) -> t -> unit
      


    (** {2 Find All} *)

    (** return the data attached to all
      (universal- and parameter-preserving) variant terms of [term]
      (used for Testing only) *)
    val find_all_variants: t -> term -> data list


    (** return the data attached to all terms more general than [term] *)
    val find_all_generalizations: p_preserving:bool -> t -> term -> data list

    (** return the data attached to all terms unifiable with [term] *)
    val find_all_unifiable: p_preserving:bool -> t -> term -> data list

    (** return the data attached to all terms which are instances of [term] *)
    val find_all_instances: p_preserving:bool -> t -> term -> data list


    (** {2 Find First} *)

    (** return the data attached to some term more general than [term] *)
    val find_generalization: p_preserving:bool -> t -> term -> data option

    (** return the data attached to some term unifiable with [term] *)
    val find_unifiable: p_preserving:bool -> t -> term -> data option

    (** return the data attached to some term which is an instance of [term] *)
    val find_instance: p_preserving:bool -> t -> term -> data option

    (** [is_productive index producer produced]
      implements the 2.nd part of the productivity check as
      defined in the model evolution paper (3.5 Productivity),
      where
      - [producer] corresponds to the context literal
      - [produced] to the remainder literal
      - [index] has the same sign as [produced]
      and the opposite sign of [producer].

      there is no term [k] in [index] such that
      there is a [k'] such that
      - [k] p_subsumes [k']
      - [producer] subsumes [k']
      - [k'] subsumes [produced]

      returns the data attached to such a [k].

      {b Note}: if [producer] is in [index] it is falsely taken for
      a suitable element.

      {b Justification}:
      If [producer] is in [index], the context is contradictory,
      as [index] contains the literals with opposite sign of [producer].


      {b Note}: if [produced] is in [index] it is falsely taken for
      a suitable element.

      {b Justification}:
      If [produced] is already in the context,
      then a remainder containing [produced] is already true in the context
      and can be ignored.

      If [producer] is in [index], the context is contradictory,
      as [index] contains the literals with opposite sign of [producer].
      This should be detected by a subsumption test on remainder literals.
      
      {b Note}: [producer] must be parameteric.

      {b Justification}:

      If producer is universal, there can not be such a [k].
      Otherwise the context would be contradictory. *)
    val find_non_productive: t -> term -> term -> data option

    (** [find_non_msg index msg term]
      implements the 1.st part of the productivity check as
      defined in the model evolution paper (3.5 Productivity):

      return None if [msg] is an msg of [term] in [index],
      otherwise the data attached to a term
      preventing [mscg] from being an msg.

      {b Note}: if [term] is in [index] it is falsely taken for
      a suitable element.

      {b Justification}:
      In the current implementation:
      - the msg check is only used in the productivity test
      - [msg] and [term] have opposite signs
      - [term] and [index] have opposite signs
      - thus if [term] is in [index] then it is contradictory
        to the remainder literal [term]
      - thus the remainder is invalidated and not considered for the context
      - this should be detected by a contradiction test
    *)
    val find_non_msg: t -> term -> term -> data option

    (** {2 Representation} *)

    (** string representation of the index *)
    val to_string: t -> string
  end
