(* $Id: term_quality.ml,v 1.12 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)



(*** types ***)

type term = Term.term
type literal = Term.literal
type subst = Subst.subst




(*** pureness ***)




exception Parametric of bool

let is_parametric (literal: literal) : bool =

  let rec do_is_parametric (term: term) : unit =
    match term with
      | Term.Var var ->
	  raise (Parametric (Var.is_parametric var))
	    
      | Term.Const _ ->
	  ()
	  
      | Term.Func (_, terms, _) ->
	  Array.iter do_is_parametric terms
  in

    try
      do_is_parametric literal.Term.term;
      false
    with
      | Parametric is_parametric ->
	  is_parametric


let cmp_pureness (first: bool) (second: bool) : int =
  if first = second then
    0
  else if first then
    1
  else
    -1



(*** depth ***)

(* for performance reasons two functions,
   although depth_of_term is a special case of depth_of_term_subst *)

let rec depth_of_term (term: term) (depth) : int =
  match term with
    | Term.Var _ ->
	depth
	  
    | Term.Const _ ->
	depth
	  
    | Term.Func (_, terms, _) ->
	let current_depth =
	  depth + 1
	in
	  (Array.fold_left 
	     (fun acc term ->
		max acc (depth_of_term term current_depth)
	     )
	     current_depth
	     terms
	  )



let depth_of_literal (literal: literal) =
  depth_of_term literal.Term.term 0



let rec depth_of_term_subst (subst: subst) (term: term) (offset: int) (depth) : int =
  match term with
    | Term.Var var ->
	begin
	  match Subst.get_ subst var offset with
	    | None ->
		depth
		  
	    | Some bound_term ->
		depth_of_term_subst subst bound_term.Subst.st_term bound_term.Subst.st_offset depth
	end
	  
    | Term.Const _ ->
	depth

    | Term.Func (_, terms, _) ->
	let current_depth =
	  depth + 1
	in
	  (Array.fold_left 
	     (fun acc term ->
		max acc (depth_of_term_subst subst term offset current_depth)
	     )
	     current_depth
	     terms
	  )
	    

let depth_of_literal_subst (subst: subst) (literal: literal) (offset: int) =
  depth_of_term_subst subst literal.Term.term offset 0










(*** weight ***)


let weight_of_literal (literal: literal) : int =

  let rec do_weight_of_term (term: term) (weight: int) : int =
    match term with
      | Term.Var var ->
	  if Var.is_universal var then
	    weight
	  else
	    weight + 1

      | Term.Const _ ->
	  weight + 1
	  
      | Term.Func (_, terms, _) ->
	  Array.fold_left
	    (fun acc term ->
	       do_weight_of_term term acc
	    )
	    (weight + 1)
	    terms
  in
    do_weight_of_term literal.Term.term 0
