(* $Id: unification.ml,v 1.11 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(*** Types ***)

type var = Var.var
type term = Term.term
type literal = Term.literal
type binding = Subst.binding
type subst = Subst.subst
type subst_var = Subst.var
type packed_subst = Subst.packed_subst

exception UNIFICATION_FAIL









(*** Variants ***)


let are_terms_variants (_term1: term) (_term2: term) : bool =

  (* acc keeps track of the bijective mapping
     from the variables of _term1 to the variables of _term2 *)
  let rec do_variants (acc: (var * var) list) (term1: term) (term2: term) :
    (var * var) list =
    
    match term1, term2 with
      | Term.Var var1, Term.Var var2 ->
	  if Var.is_universal var1 = Var.is_universal var2 then begin
	    try
	      Tools.mapping_extend acc Var.equal Var.equal var1 var2
	    with
	      | Exit ->
		  raise UNIFICATION_FAIL
	    end
	  else
	    raise UNIFICATION_FAIL
	      
      | Term.Const symbol1, Term.Const symbol2 ->
	  if Symbol.equal symbol1 symbol2 then
	    acc
	  else
	    raise UNIFICATION_FAIL
	      
      | Term.Func (symbol1, terms1, _), Term.Func (symbol2, terms2, _) ->
	  if Symbol.equal symbol1 symbol2 then
	    Tools.array_fold2 do_variants acc terms1 terms2
	  else
	    raise UNIFICATION_FAIL
	      
      | _ ->
	  raise UNIFICATION_FAIL
  in	    
    try
      ignore (do_variants [] _term1 _term2);
      true
    with
      | UNIFICATION_FAIL ->
	  false


let are_literals_variants (literal1: literal) (literal2: literal) : bool =
  (literal1.Term.sign = literal2.Term.sign)
  &&
  (are_terms_variants literal1.Term.term literal2.Term.term)













(*** Matching ***)

let rec do_match_terms ~(p_preserving: bool) ~(i_preserving:bool) (subst: subst)
  (term1: term) (offset1: int) (term2: term) (offset2: int) : subst =
  
  if (offset1 = offset2) && (Term.term_equal term1 term2) then
    subst
  else
    match term1, term2 with
      | Term.Var var, _ ->
	  begin
	    match Subst.get_ subst var offset1 with
	      | Some bound_term ->
		  do_match_terms
		    ~p_preserving:p_preserving
		    ~i_preserving:i_preserving
		    subst
		    bound_term.Subst.st_term
		    bound_term.Subst.st_offset
		    term2
		    offset2
		    
	      | None ->
		  (* term1 is part of the term to match and may thus not be bound. *)
		  if offset1 = offset2 then
		    raise UNIFICATION_FAIL

		  (* var can be bound if p/i preserving don't apply *)
		  else Subst.set_ ~p_preserving:p_preserving ~i_preserving:i_preserving
		    subst var offset1 term2 offset2
	  end
	    
      | Term.Const symbol1, Term.Const symbol2 when 
	  Symbol.equal symbol1 symbol2 ->

	  subst
	      
      | Term.Func (symbol1, terms1, _), Term.Func (symbol2, terms2, _) when
	  Symbol.equal symbol1 symbol2 ->
	    Tools.array_fold2
	      (fun subst term1 term2 ->
		 do_match_terms
		 ~p_preserving:p_preserving ~i_preserving:i_preserving
		 subst term1 offset1 term2 offset2
	      )
	      subst
	      terms1
	      terms2
	      
      | _ -> 
	  raise UNIFICATION_FAIL





let match_terms ?(p_preserving: bool = false) ?(i_preserving:bool = false)
  (term1: term) (offset1: int) (term2: term) (offset2: int) : subst =
  try
    let subst =
      do_match_terms
	~p_preserving:p_preserving ~i_preserving:i_preserving
	Subst.empty term1 offset1 term2 offset2
    in 
      subst
  with
    | Subst.SET_FAIL ->
	raise UNIFICATION_FAIL

let match_literals ?(p_preserving: bool = false) ?(i_preserving:bool = false)
  (literal1: literal) (offset1: int) (literal2: literal) (offset2: int) : subst =

  if (literal1.Term.sign = literal2.Term.sign) then
    match_terms
      ~p_preserving:p_preserving ~i_preserving:i_preserving
      literal1.Term.term offset1 literal2.Term.term offset2
  else
    raise UNIFICATION_FAIL






(*** Unification ***)


let rec do_unify_terms ~(p_preserving: bool) ~(i_preserving: bool) (subst: subst)
  (term1: term) (offset1: int) (term2: term) (offset2: int) : subst =

  if (offset1 = offset2) && (Term.term_equal term1 term2) then 
    subst
  else
    begin
      match term1, term2 with
	| Term.Var var1, Term.Var var2 ->
	    begin
	      (* dereference bound variables *)
	      match Subst.get_ subst var1 offset1, Subst.get_ subst var2 offset2 with
		| Some bound_term1, Some bound_term2 ->
		    do_unify_terms
		      ~p_preserving:p_preserving
		      ~i_preserving:i_preserving
		      subst
		      bound_term1.Subst.st_term
		      bound_term1.Subst.st_offset
		      bound_term2.Subst.st_term
		      bound_term2.Subst.st_offset
		      
		| Some bound_term1, None ->
		    do_unify_terms
		      ~p_preserving:p_preserving
		      ~i_preserving:i_preserving
		      subst
		      bound_term1.Subst.st_term
		      bound_term1.Subst.st_offset
		      term2
		      offset2
		      
		| None, Some bound_term2 ->
		    do_unify_terms
		      ~p_preserving:p_preserving
		      ~i_preserving:i_preserving
		      subst
		      term1
		      offset1
		      bound_term2.Subst.st_term
		      bound_term2.Subst.st_offset
		      
		| None, None ->
		    (* if a variable is an indicator bind the other variable *)
		    if
		      i_preserving
		      &&
		      Var.is_indicator var1
		      &&
		      not (Var.is_indicator var2)
		    then
		      Subst.set_ ~p_preserving:p_preserving ~i_preserving:i_preserving
			subst var2 offset2 term1 offset1

		    (* if a variable is an indicator bind the other variable *)
		    else if
		      i_preserving
		      &&
		      Var.is_indicator var2
		      &&
		      not (Var.is_indicator var1)
		    then
		      Subst.set_ ~p_preserving:p_preserving ~i_preserving:i_preserving
			subst var1 offset1 term2 offset2

		    (* try to bind a universal variable *)
		    else if Var.is_universal var1 then
		      Subst.set_ ~p_preserving:p_preserving ~i_preserving:i_preserving
			subst var1 offset1 term2 offset2
			
		    else (*if Var.is_universal var2 then*)
		      Subst.set_ ~p_preserving:p_preserving ~i_preserving:i_preserving
			subst var2 offset2 term1 offset1
	    end
	      
	| Term.Var var1, _ ->
	    begin
	      (* dereference bound variable *)
	      match Subst.get_ subst var1 offset1 with
		| Some bound_term1 ->
		    do_unify_terms
		      ~p_preserving:p_preserving
		      ~i_preserving:i_preserving
		      subst
		      bound_term1.Subst.st_term
		      bound_term1.Subst.st_offset
		      term2
		      offset2
		      
		| None ->
		    Subst.set_ ~p_preserving:p_preserving ~i_preserving:i_preserving
		    subst var1 offset1 term2 offset2
	    end
	      
	| _, Term.Var var2 ->
	    begin
	      (* dereference bound variable *)
	      match Subst.get_ subst var2 offset2 with
		| Some bound_term2 ->
		    do_unify_terms
		      ~p_preserving:p_preserving
		      ~i_preserving:i_preserving
		      subst
		      term1
		      offset1
		      bound_term2.Subst.st_term
		      bound_term2.Subst.st_offset
		      
		| None ->		    
		    Subst.set_ ~p_preserving:p_preserving ~i_preserving:i_preserving
		    subst var2 offset2 term1 offset1
	    end
	      
	| Term.Const symbol1, Term.Const symbol2 when
	    Symbol.equal symbol1 symbol2 ->

	    subst
		
		
	| Term.Func (symbol1, terms1, _), Term.Func (symbol2, terms2, _) when
	    Symbol.equal symbol1 symbol2 ->

	    Tools.array_fold2
	      (fun subst term1 term2 ->
		 do_unify_terms
		 ~p_preserving:p_preserving
		 ~i_preserving:i_preserving
		 subst
		 term1
		 offset1
		 term2
		 offset2
	      )
	      subst
	      terms1
	      terms2
		
	| _ ->
	    raise UNIFICATION_FAIL
    end
    


let unify_terms ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (term1: term) (offset1: int) (term2: term) (offset2: int) : Subst.subst =
  try  
    do_unify_terms ~p_preserving:p_preserving ~i_preserving:i_preserving
      Subst.empty term1 offset1 term2 offset2
  with
    | Subst.SET_FAIL ->
	raise UNIFICATION_FAIL


let unify_literals ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (literal1: literal) (offset1: int) (literal2: literal) (offset2: int) : Subst.subst =

  if literal1.Term.sign = literal2.Term.sign then
    unify_terms ~p_preserving:p_preserving ~i_preserving:i_preserving
      literal1.Term.term offset1 literal2.Term.term offset2
  else
    raise UNIFICATION_FAIL







(*** Merging ***)

(* does subst contain binding reversed? *)	
let is_included_reversed (subst: subst) (binding: binding) : bool =

  (* does subst contain the binding bound_var -> var? *)	
  let rec contains_reversed_binding (acc: subst) (var: Subst.var) (bound_var: Subst.var) : bool =
    match acc with
      | [] ->
	  false
	  
      | binding :: tail ->
	if Subst.var_equal bound_var binding.Subst.sb_var then begin
	  
	  (binding.Subst.sb_term.Subst.st_offset = var.Subst.sv_offset)
	  &&
	  (match binding.Subst.sb_term.Subst.st_term with
	     | Term.Var binding_bound_var when
		 Var.equal var.Subst.sv_var binding_bound_var ->
		 
		 true
		   
	     | _ ->
		 false
	  )
	end
	else
	  contains_reversed_binding tail var bound_var
  in

    match binding.Subst.sb_term.Subst.st_term with
      | Term.Var term_var when
	  contains_reversed_binding
	    subst
	    binding.Subst.sb_var
	    (Subst.make_var term_var binding.Subst.sb_term.Subst.st_offset)
	  ->
	  true
	    
      | _ ->
	  false



let do_unify_substs ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (acc: subst) (binding: binding) : subst =

  let var, term =
    binding.Subst.sb_var, binding.Subst.sb_term
  in
    match Subst.get acc var with
      | None ->
	  if is_included_reversed acc binding then
	    acc
	  else
	    (* var is unbound till now, so just keep its binding *)
	    Subst.set ~p_preserving:p_preserving ~i_preserving:i_preserving
	      acc var term
	      
      | Some acc_term ->
	  (* var is already bound, so unify the existing bindings *)
	  do_unify_terms
	    ~p_preserving:p_preserving ~i_preserving:i_preserving acc
	    term.Subst.st_term term.Subst.st_offset
	    acc_term.Subst.st_term acc_term.Subst.st_offset


let unify_substs ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (first: subst) (second: subst) : subst =

  match first with
    | [] ->
	second
        
    | _ ->
        begin
          try
	    List.fold_left
	      (fun acc binding ->
		 do_unify_substs
		   ~p_preserving:p_preserving ~i_preserving:i_preserving
		   acc binding
	      )
	      first
	      second
          with
            | Subst.SET_FAIL ->
	        raise UNIFICATION_FAIL
        end


let unify_packed_substs ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (first: subst) (second: packed_subst) : subst =

  match first with
    | [] ->
	Array.to_list second
        
    | _ ->
        begin
          try
	    Array.fold_left
	      (fun acc binding ->
		 do_unify_substs
		   ~p_preserving:p_preserving ~i_preserving:i_preserving
		   acc binding
	      )
	      first
	      second
          with
            | Subst.SET_FAIL ->
	        raise UNIFICATION_FAIL
        end






let do_match_substs ?(p_preserving: bool = false) ?(i_preserving: bool = false)
  (more_general_offset: int) (acc: subst) (subst: subst) : subst =

  List.fold_left
    (fun acc binding ->
       let var, term =
	 binding.Subst.sb_var, binding.Subst.sb_term
       in
	 match Subst.get acc var with
	   | None ->
	       (* var is unbound till now, so just keep its binding *)
	       Subst.set
	         ~p_preserving:p_preserving ~i_preserving:i_preserving
	         acc var term
	       
	   | Some acc_term ->
	       (* var is already bound, so unify the existing bindings *)
		 if term.Subst.st_offset = more_general_offset
		   &&
		   acc_term.Subst.st_offset = more_general_offset
		 then begin
		   (* unify *)
		   do_unify_terms
		     ~p_preserving:p_preserving ~i_preserving:i_preserving acc
		     term.Subst.st_term term.Subst.st_offset
		     acc_term.Subst.st_term acc_term.Subst.st_offset
		 end

		 else if term.Subst.st_offset = more_general_offset then begin
		   (* match term to acc *)
		   do_match_terms
		     ~p_preserving:p_preserving ~i_preserving:i_preserving acc
		     term.Subst.st_term term.Subst.st_offset
		     acc_term.Subst.st_term acc_term.Subst.st_offset
		 end

		 else if acc_term.Subst.st_offset = more_general_offset then begin
		   (* match acc to term *)
		   do_match_terms
		     ~p_preserving:p_preserving ~i_preserving:i_preserving acc
		     acc_term.Subst.st_term acc_term.Subst.st_offset
		     term.Subst.st_term term.Subst.st_offset
		 end

		 else begin
		   (* both terms part of the term to match -
		      no variables can be bound *)
		   if Term.term_equal term.Subst.st_term acc_term.Subst.st_term then
		     acc
		   else
		     raise UNIFICATION_FAIL
		 end

    )
    acc
    subst


let match_substs ?(p_preserving: bool = false)  ?(i_preserving: bool = false)
  (more_general_offset: int) (first: subst) (second: subst) : subst =

  match first with
    | [] ->
	second
        
    | _ ->
        begin
          try
            do_match_substs
	      ~p_preserving:p_preserving ~i_preserving:i_preserving
	      more_general_offset first second
          with
            | Subst.SET_FAIL ->
	        raise UNIFICATION_FAIL
        end
