/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
//  file unit.h 
//
//  11/06/2000 Manchester, moved to from problem.h
//  29/03/2001 Macnchester, Functor changed to Functor*
//  08/04/2001 Manchester, char* changed to String in _name
//  12/04/2001 Manchester, InputUnit changed to Unit
//  09/05/2001 Manchester: occurs ()
//  15/05/2001 Manchester: replaceCancellation()
//  22/05/2001 Manchester: isLinearDefinition()
//  27/05/2001 Manchester: occurs (PSymbol*)
//  30/05/2001 Manchester, splitLargeArity/2
//  31/05/2001 Manchester, normalize/0, negativeEqualitySplit/1
//  03/06/2001 Manchester: isEqualityAxiom/0, isPredicateDefinition/1
//


#ifndef __unit__
#define __unit__


#include "debug.h"
#include "typedefs.h"
#include "string.h"
#include "int.h"
#include "signature.h"
#include "chain.h"
#include "clause.h"


class ReplaceCode;
class Formula;
class Clause;
class Atom;
class Problem;


enum InputType {
  AXIOM,
  CONJECTURE,
  HYPOTHESIS
};


enum UnitType {
  CLAUSE,
  FORMULA
};


enum DefType {
  DT_NONE,
  DT_PRED,
  DT_FUN
};


enum InferenceRule {
  IR_INPUT,         // input formula or clause
  IR_PROP,          // obtained by propositional manipulations, e.g., a -> b replaced by ~a \/ b
  IR_PERMUT,        // obtained by permutations, e.g. f <=> g replaced by g <=> f
  IR_FLATTEN,       // obtained by some form of flattening (quantifiers, conjunctions etc.)
  IR_REORDER_EQ,    // obtained by reordering equality
  IR_HALF_EQUIV,    // obtained by rewriting a positive equivalence into an implication
  IR_AND_ELIM,      // obtain Ai from A1 & ... & An
  IR_ENNF,          // clear
  IR_RM_EQUIV,      // remove equivalences
  IR_MINISCOPE,
  IR_SKOLEMIZE,
  IR_RECTIFY,
  IR_SWAP,          // obtained by using commutativity of connectives, eg &
  IR_CLAUSIFY       // obtain clause from a formula
}; 


// ****************** class Unit, definition *******************


class Unit 
#   if debug_memory
    : public Memory <CID_UNIT>
#   endif
{ 
 public:

  typedef List<Unit*,CID_UNIT_LIST> List;
  typedef Chain<Unit*,CID_UNIT_LINK> Chain;

  Unit ( char* name, InputType untype, Formula* );
  Unit ( char* name, InputType untype, Clause* );
  Unit ( InferenceRule, Formula*, Unit* parent );
  Unit ( InferenceRule, Clause*, Unit* parent );

  ~Unit (); // also destroys the unit

  // query/change structure
  char* name () const { return _name; }
  InferenceRule rule () const { return _rule; }
  InputType inputType () const { return _intype; }
  UnitType unitType () const { return _untype; }
  DefType defType () const { return _defType; }
  void inc () { _count++; }
  void dec ();
  int number () const { return _number; }

  Formula* formula () const 
    { ASSERT ( debug_problem, unitType() == FORMULA );
      return _formula; }
  void formula ( Formula* f ); // note: only resets an already existing formula
  Clause* clause () const
    { ASSERT ( debug_problem, unitType() == CLAUSE );
      return _clause; }

  // properties
  bool occurs ( const Signature::Pred* p ) const;
  bool isEqualityAxiom () const;      // is one of the equality axioms
  bool formulaWithFreeVars () const;  // is a formula with free variables
  Atom* isPredicateDefinition (Formula* & rhs) const;

  // miscellaneous
  Clause::List* skolemize ();
  Unit* makeDefinition (); 
  void normalize ();
  void flatten ();
  bool less (const Unit*) const; // used to compare units, needed for normalisation
/*
  void clause ( Clause* c )
    { ASSERT ( debug_problem, unitType() == CLAUSE );
      _clause = c; } // Unit::clause

  // statistics
  int weight () const;
  int countOccurrences ( const FSymbol* f ) const;
  bool occurs ( const FSymbol* f ) const;
  int countOccurrences ( const PSymbol* p ) const;
  bool hasFunctionSymbols () const;
  bool hasEquality () const;
  bool isHorn () const;
  bool isUnit () const; // contains only one literal
  bool isPropositional () const;
  bool hasNonEquality () const;
  int numberOfLiterals () const;
  int maxLitWeight ();
  bool isGoal () const;

  Unit* resolveOn ( const PSymbol*, Unit*, ostream& report );
  Compare compare ( const Unit* ) const;
  bool removeVarEquality ( ostream& report );
  bool isTautology ( ostream& report ); 
  void replaceRelationByFunction ( PSymbol*, FSymbol*, int argNo, int* vs, ostream& report ); 
  Var* isFunctionalityAxiom ( const PSymbol* pred, FSymbol*& fun, int argNo, Atom* & atom ) const;
  bool isUniquenessAxiom ( PSymbol*& pred, int& argNo ) const; // the unit is P(x,y) & P(x,z) -> y = z
  bool isCancellationAxiom ( FSymbol*& pred, int& argNo ) const; // the unit is f(X,y,Z) = f(X,u,Z) -> y = u
  void apply ( const ReplaceCode&, const Unit* def, ostream& report );
  int maxLitWeight ( const ReplaceCode& ) const;
  Term* isDefinition ( Term* & rhs ) const; 
  Term* isLinearDefinition ( Term* & rhs ) const; 
  static List* sort ( List* );
  void replaceCancellation ( const FSymbol*,int argNo,FSymbol* inv, ostream& report);
  void splitLargeArity ( const PSymbol* p, FSymbol::List* newFs );
  void splitLargeArity ( const FSymbol* p, FSymbol::List* newFs );
  List* negativeEqualitySplit ( int thresholdWeight, ostream& tab );
  void polarities ( const PSymbol* p, int& pol ) const;
  void makeHalfDefinition ( bool positive,const Formula* rhs );
*/

 private:

  // structure
  UnitType _untype;
  InputType _intype;
  InferenceRule _rule;  
  long _number;
  union {
    List* _parents;
    char* _name;
  };
  union {
    Clause* _clause;
    Formula* _formula;
  };
  DefType _defType;
  int _count; // count the number of references for deleted units

  static long _lastNumber;
}; // class Unit


ostream& operator << ( ostream&, InputType );
ostream& operator << ( ostream&, UnitType );
ostream& operator << ( ostream&, Unit::Chain& );



#endif // __unit__
