
/*
 * File CASCMode.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "CASCMode.hpp"
#include "Int.hpp"
#include "Kernel.hpp"
#include "Timer.hpp"
#include "Tracer.hpp"
#include "Options.hpp"
#include "Preprocess.hpp"
#include "Problem.hpp"
#include "Property.hpp"
#include "Run.hpp"
#include "Refutation.hpp"
#include "Signature.hpp"
#include "Statistics.hpp"
#include "TheoryFinder.hpp"
#include "XML.hpp"


#define SPIDER_MODE 0


void CASCMode::run (UnitList& units, Timer& timer)
{
  TRACER ("CASCMode::run");
  Options& options = Options::globalOptions;

  bool logged = options.logFile() != "";
  if (logged) {
    Run::log->addChild(XML::Element::generalProofAttemptInfo());
  }

  int timeLimitInDeciseconds = options.timeLimitInDeciseconds();
  Problem problem(options.problemName(),units);
  Property property(problem);
  property.scanProblem();
  TheoryFinder tf(problem);
  tf.search();
  for (const TheoryFinder::Record* record = tf.records(); 
       record; 
       record = record->next) {
    switch (record->kind) {
    case AX_EXTENSIONALITY:
      property.addProp(PR_HAS_EXTENSIONALITY);
      break;
    case AX_SUBSET:
    case AX_SUBSET1:
      property.addProp(PR_HAS_SUBSET);
      break;
    default:
      break;
    }
  }
  Schedule schedule(property);



  runSchedule(schedule,units,property,timer,timeLimitInDeciseconds,logged ? Run::log : 0,600);
}  



void CASCMode::reportAllocationFailure ()
{
#if SPIDER_MODE
  cout << "? " 
       << Options::globalOptions.problemName() << ' '
       << "0.0 casc\n";
#else
  cout << "--------------------- CANNOT PROVE ----------------------\n";
#endif
} 



void CASCMode::runSchedule (Schedule& schedule,
			    UnitList& units,
			    const Property& property,
			    Timer& timer,
			    int timeLimit,
			    XML::DeepElement* log,
			    int normalSliceTimeLimitInDeciseconds)
{
  Options& globalOptions = Options::globalOptions;
  XML::DeepElement* proofAttempt = 0;
  while (schedule.hasNext()) {
    
    Signature::sig.resetSkolemFunctionNumbers();
    int remainingTime = globalOptions.timeLimitInDeciseconds() -
                        timer.elapsedDeciseconds();
    
    if (remainingTime < 0) {
      break;
    }
    
    Kernel kernel;
    
    VampireKernel::Options kernelOptions;
    Options opts;
    opts.setArityCheck(false);
    string testId = schedule.next();
#if SPIDER_MODE
#else
    cout << "Running slice " << testId << '\n';
#endif
    opts.readFromTestId(testId);
    opts.setMemoryLimit(globalOptions.memoryLimit());
    
    if (opts.mainAlg() == Options::LRS && 
	normalSliceTimeLimitInDeciseconds &&
	opts.timeLimitInDeciseconds()) {
      opts.setLrsSoftness(((float)normalSliceTimeLimitInDeciseconds)/
			  (float)opts.timeLimitInDeciseconds() + 0.001);
    }
    int timeLimit = opts.timeLimitInDeciseconds() ?
                    Int::min (opts.timeLimitInDeciseconds(), remainingTime) :
                    remainingTime;
    opts.setTimeLimitInDeciseconds(timeLimit);


    
    Kernel::passOptions (kernelOptions, opts);

    if (log) {
      proofAttempt = new XML::DeepElement("proof_attempt");
      log->addChild(proofAttempt);
      proofAttempt->addChild(opts.toXML());
    }
    
    UnitList copyOfUnits;
    Stack<Unit> stack(copyOfUnits);
    Iterator<Unit> us(units);
    while (us.hasNext()) {
      stack.push(us.next());
    }
    
    Problem problem(globalOptions.problemName(),copyOfUnits);
    
    Preprocess::preprocess(problem,opts,property);

    BK::DestructionMode::makeThorough();
    RunResult result = kernel.run(timeLimit, kernelOptions, problem, 
				  opts, timer, log);
    switch (result) {
    case REFUTATION:
#if SPIDER_MODE
      cout << "+ " 
	   << globalOptions.problemName() << ' '
	   << timer.elapsedSeconds() << ' ' 
	   << opts.testId() << '\n';
#else
      cout << "--------------------- PROVED ----------------------\n"
	   << flush;
#endif

      if (globalOptions.proof()) {
	Refutation ref (kernel, 
			problem,
			globalOptions);
#if SPIDER_MODE
#else
	ref.prettyPrint(cout,1);
	cout << flush;
#endif
	if (log) {
	  proofAttempt->addChild(ref.toXML());
	}
      }
      break;

    case SATISFIABLE:
#if SPIDER_MODE
      cout << "- " 
	   << globalOptions.problemName() << ' '
	   << timer.elapsedSeconds() << ' ' 
	   << opts.testId() << '\n';
#else
      cout << "--------------------- UNPROVABLE ----------------------\n"
	   << flush;
#endif
      break;
      
    case UNKNOWN:
      break;
    }

    if (log) {
      Statistics statistics (kernel.kernel());
      XML::Element* xmlStat = statistics.toXML();
      if (xmlStat) {
	proofAttempt->addChild(xmlStat);
      }
      string resultString = Run::toString(result);
      proofAttempt->addChild(new XML::TextElement("slice_result",resultString));
      if (result != UNKNOWN || ! schedule.hasNext()) {
	log->addChild(new XML::FloatElement("total_time",
						 (float)timer.elapsedDeciseconds()/10));
	log->addChild(new XML::TextElement("result",resultString));
      }
    }
    if (result != UNKNOWN) {
      return;
    }
    if (schedule.hasNext()) {
#if SPIDER_MODE
#else
      cout << "------------------ SLICE UNSUCCESSFUL -------------------\n"
	   << flush;
#endif
    }
    else {
#if SPIDER_MODE
      cout << "? " 
	   << globalOptions.problemName() << ' '
	   << timer.elapsedSeconds() << ' ' 
	   << opts.testId() << '\n';
#else
      cout << "--------------------- CANNOT PROVE ----------------------\n"
	   << flush;
#endif
      return;
    }
    BK::DestructionMode::makeFast();
    kernel.kernel().closeSession();
  }
#if SPIDER_MODE
  cout << "? " 
       << globalOptions.problemName() << ' '
       << timer.elapsedSeconds() << " casc\n";
#else
  cout << "--------------------- CANNOT PROVE ----------------------\n"
       << flush;
#endif
} 



string CASCMode::Schedule::next ()
{
  unsigned indexOfSpace = schedule.find(' ');
  string nxt;
  if (indexOfSpace == string::npos) {
    nxt = schedule;
    schedule = "";
  }
  else {
    nxt = schedule.substr(0,indexOfSpace);
    schedule = schedule.substr(indexOfSpace+1);
  }

  return nxt;
} 



bool CASCMode::Schedule::hasNext () const
{
  return schedule != "";
} 


CASCMode::Schedule::Schedule (const Property& p)
{
  switch (p.category()) {
  case NEQ:
    switch (p.props()) {
    case 3:
      schedule = 
	"lrs-1012_4:3_bd=off_bs=off_ep=on_lcm=p_sos=on_44 "
	"lrs+1012_3_bs=off_nwc=1.4_18 "
	"dis+1010_1:2_bs=off_ep=on_nwc=2.0_1 "
	"lrs-1012_1_ep=on_lcm=k_nwc=5.0_sos=on_sp=on_52 "
	"dis+10_2_bs=off_ep=on_nwc=1.8_2 "
	"dis+1010_12_bs=off_2 "
	"lrs+13_48_bs=off_nwc=9.0_41 "
	"lrs+10_4:3_bas=on_bs=off_ep=on_fde=all_nwc=1.3_12 "
	"dis-11_25_bas=on_sos=on_16 "
	"dis+1013_32_nwc=1.5_3 "
	"dis+1013_7_bs=off_ep=on_sos=on_26 "
	"lrs+16_2_bs=off_ep=on_fde=all_sp=on_4 "
	"lrs+17_20_ep=on_fde=all_sp=on_12 "
	"dis+13_64_lcm=p_nwc=2.0_45 "
	"dis+1013_5:1_bs=off_ep=on_sos=on_7 "
	"dis+11_256_lcm=p_sos=on_10 "
	"dis+10_15_bas=on_bs=off_ep=on_fde=all_12 "
	"lrs+16_5:4_bs=off_fde=all_nwc=4.0_15 "
	"lrs+16_60_bas=on_bs=off_fde=all_lcm=k_nwc=3.0_sp=on_24 "
	"lrs+10_5:4_bas=on_bs=off_fde=all_ep=on_nwc=1.7_0";
      return;

    case 7:
      schedule =
	"ott+16_1_ep=on_lcm=p_nwc=5.0_1 "
	"lrs+1011_1_nwc=1.5_60 "
	"ott+17_16_bs=off_fde=all_lcm=p_nwc=5.0_1 "
	"ott+16_32_bs=off_lcm=k_nwc=2.5_3 "
	"ott+17_8_bs=off_fde=all_lcm=p_nwc=4.5_1 "
	"ott-1012_4_ep=on_nwc=5.0_1 " 
	"ott-1012_4_bs=off_ep=on_fde=all_lcm=k_nwc=4.0_1 "
	"lrs+10_64_lcm=p_nwc=5.0_4 "
	"ott+10_2_lcm=p_nwc=3.0_1 "
	"ott+1011_5:1_bs=off_ep=on_fde=all_sos=on_1 "
	"ott+13_16_fde=all_nwc=2.0_1 "
	"lrs+17_3:2_bs=off_ep=on_fde=all_lcm=p_nwc=3.0_33 "
	"dis+8_8_ep=on_lcm=p_nwc=3.0_5 "
	"dis+10_7_fde=all_nwc=5.0_12 "
	"lrs+1011_5:1_bs=off_ep=on_sos=on_16 "
	"dis+10_4_ep=on_fde=all_lcm=k_nwc=1.5_2 "
	"ott+17_5:4_bs=off_ep=on_nwc=3.5_3 "
	"lrs+16_3:1_bs=off_fde=all_lcm=k_nwc=3.5_17 "
	"lrs+16_4:1_fde=all_nwc=4.0_18 "
	"lrs+17_2:3_bs=off_ep=on_fde=all_lcm=k_nwc=4.3_5 "
	"lrs+1011_2:1_bs=off_lcm=k_sos=on_6 "
	"lrs+13_3_bs=off_ep=on_lcm=k_nwc=3.0_28 "
	"lrs+16_3_bs=off_lcm=k_nwc=3.0_16 "
	"dis+10_2:1_ep=on_nwc=2.0_13 "
	"lrs+16_1_bs=off_fde=all_lcm=p_nwc=1.3_14 "
	"lrs-11_1_lcm=p_28 "
	"lrs-11_4_bs=off_ep=on_fde=all_lcm=k_nwc=1.8_46 "
	"lrs+10_1_nwc=1.5_0";
      return;

    default:
      schedule = 
	"lrs+16_112_bas=on_bs=off_nwc=5.0_60 "
	"dis+16_32_bs=off_ep=on_fde=all_nwc=2.0_2 "
	"ott+1012_2_bs=off_ep=on_nwc=3.0_2 "
	"dis+1011_36_bs=off_ep=on_lcm=p_nwc=4.0_2 "
	"lrs+10_4:3_bas=on_bs=off_ep=on_fde=all_nwc=1.3_10 "
	"lrs+1011_5:1_bs=off_ep=on_fde=all_sos=on_20 "
	"lrs+17_4:3_bas=on_bs=off_ep=on_nwc=2.3_30 "
	"lrs+1013_3:1_ep=on_fde=all_nwc=2.0_5 "
	"lrs+1011_28_bas=on_bs=off_nwc=5.0_35 "
	"lrs+1010_4_bd=preordered_bs=off_lcm=k_nwc=3.0_22 "
	"lrs+13_1_nwc=2.0_14 "
	"lrs+11_64_bas=on_bs=off_nwc=1.2_14 "
	"lrs+16_60_bas=on_bs=off_fde=all_lcm=k_nwc=3.0_sp=on_27 "
	"lrs+17_5:4_bs=off_ep=on_nwc=3.5_0";
      return;
    }
    return;

  case HEQ: 
    schedule = 
      "dis+13_11_bs=off_fde=all_nwc=9.0_62 "
      "lrs-1012_28_nwc=1.8_55 "
      "dis-1012_2:1_bs=off_fde=all_lcm=k_nwc=4.0_30 "
      "dis+16_6_fde=all_nwc=5.0_27 "
      "lrs+1010_1_lcm=p_30 "
      "lrs+10_1_nwc=1.5_28 "
      "lrs-11_256_nwc=1.3_23 "
      "dis+11_4_lcm=p_5 "
      "ott+1010_16_bd=preordered_nwc=2.0_1 "
      "ott-11_5_bs=off_ep=on_fde=all_sos=on_3 "
      "dis+10_8_bd=preordered_nwc=3.0_5 "
      "dis+10_2:3_fde=all_nwc=2.5_5 "
      "lrs+16_2_bs=off_fsr=off_nwc=1.7_60 "
      "lrs+10_32_nwc=1.2_0 ";
      return;

  case PEQ: 
    schedule = 
      "lrs+1010_1_lcm=p_34 "
      "lrs-11_4_bs=off_ep=on_fde=all_lcm=k_nwc=1.8_20 "
      "ott-11_2_bd=off_bs=off_ep=on_fde=all_nwc=2.0_3 "
      "lrs+11_2_bs=off_lcm=k_14 "
      "lrs-11_1_bs=off_ep=on_lcm=k_nwc=5.0_8 "
      "lrs-11_20_bs=off_ep=on_fde=all_lcm=p_nwc=2.5_46 "
      "lrs-11_256_bs=off_ep=on_nwc=4.0_10 "
      "lrs+17_4_bs=off_nwc=1.4_0";
    return;

  case HNE: 
    schedule = 
      "lrs+10_3:1_nwc=1.3_sp=on_60 "
      "lrs+11_2_bs=off_lcm=k_60 "
      "lrs+10_2_lcm=p_nwc=3.0_60 "
      "lrs+10_4_lcm=k_nwc=3.0_26 "
      "lrs+1011_3_bs=off_14 "
      "ott+10_512_nwc=3.0_sos=on_sp=on_2 "
      "dis+11_256_lcm=p_sos=on_20 "
      "lrs+17_2:1_bs=off_ep=on_fde=all_lcm=p_sos=on_60 "
      "ott-1012_11_bs=off_nwc=4.0_2 "
      "lrs+13_128_0";
    return;

  case NNE: 
    schedule = 
      "lrs+13_3:1_bs=off_ep=on_nwc=2.9_sos=on_60 "
      "dis+1011_12_bs=off_sos=on_56 "
      "lrs+1011_1_bs=off_lcm=p_nwc=5.0_17 "
      "lrs+13_2:1_bs=off_lcm=k_sos=on_60 "
      "dis+10_1_lcm=p_sos=on_1 "
      "dis+1012_1024_bs=off_lcm=p_nwc=4.0_43 "
      "lrs+16_5:4_bs=off_lcm=k_nwc=2.8_20 "
      "lrs+13_1_bs=off_fde=all_lcm=k_nwc=1.5_50 "
      "dis+13_4_lcm=p_sos=on_36 "
      "lrs+1012_32_bs=off_lcm=p_nwc=1.5_50 "
      "lrs+17_4_bs=off_nwc=1.4_0";
    return;

  case FEQ:
    switch (p.props()) {
    case 1:
      schedule = 
	"lrs-1012_3_bd=off_lcm=p_sos=on_55 "
	"lrs+1011_2_bd=off_bs=off_ep=on_nwc=3.0_55 "
	"lrs+1012_2:3_bs=off_ep=on_lcm=k_nwc=1.3_43 "
	"dis-1012_256_ep=on_nwc=2.0_1 "
	"dis+11_3:1_bs=off_ep=on_nwc=1.3_8 "
	"dis+16_32_bs=off_ep=on_nwc=2.0_55 "
	"dis+1013_2_bs=off_ep=on_fde=all_lcm=p_sos=on_60 "
	"dis+1013_8_ep=on_fde=all_lcm=p_sos=on_25 "
	"lrs-1012_11_bs=off_nwc=4.0_13 "
	"lrs+1011_2:1_bs=off_lcm=k_sos=on_28 "
	"dis+1013_8_ep=on_lcm=p_sos=on_21 "
	"lrs+10_3:1_nwc=1.3_sp=on_26 "
	"lrs+16_4:1_fde=all_nwc=4.0_0";
      return;

    default:
      schedule = 
    	"lrs-1012_64_nwc=9_60 "
	"lrs+16_4:1_fde=all_nwc=4.0_48 "
	"lrs+17_2:1_bs=off_ep=on_fde=all_lcm=p_sos=on_12 "
	"lrs-1012_2_bs=off_nwc=2.5_sos=on_sp=on_14 "
	"lrs+1010_4_bd=preordered_bs=off_lcm=k_nwc=3.0_1 "
	"lrs+10_32_lcm=k_nwc=9.0_4 "
	"dis+1010_8_ep=on_lcm=k_sos=on_4 "
	"lrs+1011_28_bas=on_bs=off_nwc=5.0_28 "
	"lrs+13_3:1_bs=off_ep=on_nwc=2.9_sos=on_7 "
	"lrs-11_2_bd=off_bs=off_ep=on_fde=all_nwc=2.0_5 "
	"dis+11_5:1_bs=off_lcm=k_sos=on_5 "
	"lrs+1011_1_bs=off_lcm=p_nwc=5.0_10 "
	"lrs+16_28_bs=off_nwc=2.1_0";
      return;
    }

  case EPR:
  case FNE:
    schedule =
      "lrs+16_2_bs=off_ep=on_fde=all_sp=on_5 "
      "lrs+10_32_nwc=1.2_0";
    return;

  case UEQ:
    if (p.hasProp(PR_HAS_FUNCTION_DEFINITIONS)) {
      schedule = 
	"lrs+8_2:3_fde=all_ls=3.75_80 "
     	"dis+8_2:3_fde=all_ls=1.58_190 "
        "lrs+8_8_bs=off_0";
    }
    else {
      schedule = 
	"lrs+8_2_bd=off_ls=1.82_165 "
	"lrs+8_4_bd=preordered_bs=off_ls=6.39_47 "
	"lrs+8_2:3_bs=off_ls=3.53_85 "
	"lrs+8_3:1_bs=off_0";
    }
    return;
  }
}
