
/*
 * File Compare.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Compare.hpp"
#include "Unit.hpp"
#include "SymCounter.hpp"
#include "Signature.hpp"

SymCounter* Compare::_symbolCounter;


bool Compare::lessThan (const Unit& u1, const Unit& u2)
{
  switch (u1.unitType()) 
    {
    case CLAUSE:
      if (u2.unitType() == FORMULA) {
	return true;
      }
      
      return lessThan(u1.clause(),u2.clause());

    case FORMULA:
      if (u2.unitType() == CLAUSE) {
	return false;
      }
      
      return lessThan(u1.formula(),u2.formula());

#if DEBUG_PREPRO
    default:
      ASS(false);
      return false;
#endif
    }
} 



Comparison Compare::compare (const Term& t1,const Term& t2)
{
  switch (compare (t1.tag(), t2.tag())) 
    {
    case GREATER:
      return GREATER;
    case LESS:
      return LESS;
    case EQUAL:
      break;
    }

  switch (t1.tag()) { 
  case VAR:
    return EQUAL;

  case NUMERIC:
    if (t1.number() > t2.number()) {
      return GREATER;
    }
    if (t1.number() < t2.number()) {
      return LESS;
    }
    return EQUAL;

  case COMPOUND:
    switch (compare(t1.functor(),t2.functor())) {
    case LESS:
      return LESS;
      
    case GREATER:
      return GREATER;

    case EQUAL:
      
      return compare(t1.args(),t2.args());
    }

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



Comparison Compare::compare (const TermList& l1, const TermList& l2)
{
  Iterator<Term> ts1(l1);
  Iterator<Term> ts2(l2);
  while (ts1.hasNext()) {
    if (! ts2.hasNext()) {
      return GREATER;
    }
    Comparison cmp = compare(ts1.next(),ts2.next());
    if (cmp != EQUAL) {
      return cmp;
    }
  }
  if (ts2.hasNext()) {
    return LESS;
  }
  return EQUAL;
} 


Comparison Compare::compare (const Literal& l1, const Literal& l2)
{
  
  if (l1.isNegative()) {
    if (l2.isPositive()) {
      return LESS;
    }
  }
  else { 
    if (l2.isNegative()) {
      return GREATER;
    }
  }

  
  return compare(l1.atom(),l2.atom());
} 



Comparison Compare::compare (const Atom& a1, const Atom& a2)
{ 
  
  if (a1.isEquality()) {
    if (! a2.isEquality()) {
      return LESS;
    }
  }
  else if (a2.isEquality()) {
    return GREATER;
  }

  
  switch (compare(a1.functor(),a2.functor())) {
  case LESS:
    return LESS;

  case EQUAL:
    return compare(a1.args(),a2.args());

  case GREATER:
    return GREATER;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
}  



Comparison Compare::compare (const PredicateSymbol& p1,
			     const PredicateSymbol& p2)
{
  Comparison comp = compare(p1.arity(),p2.arity());
  if (comp != EQUAL) {
    return comp;
  }

  
  SymCounter::Pred* pr1 = _symbolCounter->get(p1);
  SymCounter::Pred* pr2 = _symbolCounter->get(p2);

  comp = compare(pr1->pocc(), pr2->pocc());
  if (comp != EQUAL) {
    return comp;
  }

  comp = compare (pr1->nocc(), pr2->nocc());
  if (comp != EQUAL) {
    return comp;
  }

  return compare (pr1->docc(), pr2->docc());
} 



Comparison Compare::compare (const FunctionSymbol& f1,
			     const FunctionSymbol& f2)
{
  switch (compare(f1.arity(),f2.arity())) {
  case LESS:
    return LESS;
    
  case GREATER:
    return GREATER;

  case EQUAL:
    return compare(_symbolCounter->get(f1)->occ(),
		   _symbolCounter->get(f2)->occ());
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



bool Compare::lessThan (const Clause& cl1, const Clause& cl2)
{
  Iterator<Literal> lits1 (cl1.literals());
  Iterator<Literal> lits2 (cl2.literals()); 

  while (lits1.hasNext()) {
    if (! lits2.hasNext()) {
      return false;
    }

    
    switch (compare(lits1.next(),lits2.next())) {
    case LESS:  
      return true;

    case EQUAL:
      break;

    case GREATER:
      return false;
    }
  }
  return lits2.hasNext();
} 



Comparison Compare::compare (const FormulaList& fs1, const FormulaList& fs2) 
{
  Iterator<Formula> f1 (fs1);
  Iterator<Formula> f2 (fs2);

  while ( f1.hasNext() ) {
    if ( f2.hasNext() ) {
      Comparison comp = compare(f1.next(),f2.next());
      if (comp != EQUAL) {
	return comp;
      }
    }
    else { 
      return GREATER;
    }
  }
  
  return f2.hasNext() ? LESS : EQUAL;
} 



bool Compare::lessThan (const Formula& f1, const Formula& f2)
{
  return compare(f1,f2) == LESS;
}



Comparison Compare::compare (const Formula& f1, const Formula& f2)
{
  Comparison comp = compare ((int)f1.connective(),
			     (int)f2.connective ());
  if (comp != EQUAL) {
    return comp;
  }

  
  switch (f1.connective()) 
    {
    case ATOM:
      return compare(f1.atom(),f2.atom());

    case AND:
    case OR:
      return compare(f1.args(),f2.args());

    case IMP:
    case IFF:
    case XOR:
      
      comp = compare(f1.left(),f2.left());
      if (comp != EQUAL) {
	return comp;
      }
      return compare(f1.right(),f2.right());

    case NOT:
      return compare(f1.uarg(),f2.uarg());

    case FORALL:
    case EXISTS:
      
      
      comp = compare(f1.vars().length(),f2.vars().length());
      if (comp != EQUAL) {
	return comp;
      }
      return compare(f1.qarg(),f2.qarg());

#if DEBUG_PREPRO
    default:
      ASS (false);
#endif
  }
} 



bool Compare::lessThan (const Literal& l1, const Literal& l2)
{
  return compare(l1,l2) == LESS;
}



void Compare::createSymbolCounter (const Problem& problem)
{
  delete _symbolCounter;
  _symbolCounter = new SymCounter(Signature::sig);
  _symbolCounter->count(problem,1);
} 

