
/*
 * File DoubleSubstitution.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Int.hpp"
#include "Literal.hpp"
#include "DoubleSubstitution.hpp"


class DoubleSubstitution::Binding 
{
public:
  
  Binding(const Term& t,VariableBank* b) : term(t), bank(b) {}

  
  Term term;
  
  VariableBank* bank;

  CLASS_NAME("DoubleSubstitution::Binding");
  USE_ALLOCATOR(Binding);
};



bool DoubleSubstitution::resolve (const Literal& l1,const Literal& l2)
{
  return l1.sign() != l2.sign() &&
         unify(l1.atom(),l2.atom());
} 



bool DoubleSubstitution::unify (const Atom& a1,const Atom& a2)
{
  if (a1.functor() != a2.functor()) {
    return false;
  }

  return unify(a1.args(),a2.args(),_bank,_bank+1);
} 



bool DoubleSubstitution::unify (const TermList& ts1,
				const TermList& ts2,
				VariableBank* vb1,
				VariableBank* vb2)
{
  if (ts1.isEmpty()) {
    ASS(ts2.isEmpty());
    return true;
  }
  
  return unify(ts1.head(),ts2.head(),vb1,vb2) &&
         unify(ts1.tail(),ts2.tail(),vb1,vb2);
} 



bool DoubleSubstitution::unify (const Term& t1,
				const Term& t2,
				VariableBank* vb1,
				VariableBank* vb2)
{
  if (t1.tag() == VAR) {
    Var v1 = t1.var();
    Binding* b = vb1->get(v1);
    if (b) {
      return unify(b->term,t2,b->bank,vb2);
    }
    
    switch (t2.tag()) {
    case VAR:
      {
	Var v2 = t2.var();
	b = vb2->get(v2);
	if (b) {
	  return unify(b->term,t1,b->bank,vb1);
	}
	
	if (v1 != v2 || &vb1 != &vb2) {
	  vb1->set(v1,t2,vb2);
	}
	return true;
      }

    case COMPOUND:
      
      return vb1->bind(v1,t2,vb2);
    
    case NUMERIC:
      vb1->set(v1,t2,vb2);
      return true;
    }
  }

  switch (t2.tag()) {
  case VAR: 
    {
      Var v2 = t2.var();
      Binding* b = vb2->get(v2);
      if (b) {
	return unify(b->term,t1,b->bank,vb1);
      }
      
      return vb2->bind(v2,t1,vb1);
    }

    
  case NUMERIC:
    return t1.tag() == NUMERIC && 
           t1.number() == t2.number();

  case COMPOUND:
    return t2.tag() == COMPOUND && 
           t1.functor() == t2.functor() &&
           unify(t1.args(),t2.args(),vb1,vb2);

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



DoubleSubstitution::VariableBank::VariableBank ()
  : _capacity(0),
    _bindings(0)
{
} 



DoubleSubstitution::VariableBank::~VariableBank ()
{
  for (int i = _capacity-1;i >= 0; i--) {
    delete _bindings[i];
  }

  delete [] _bindings;
} 



DoubleSubstitution::Binding* 
DoubleSubstitution::VariableBank::get (Var v) const
{
  int w = v.toInt();
  if (w >= _capacity) {
    return 0;
  }
  return _bindings[w];
} 



void DoubleSubstitution::VariableBank::set (Var v, 
					    const Term& t,
					    VariableBank* vb)
{
  int w = v.toInt();
  if (w >= _capacity) {
    extend(w);
  }
  _bindings[w] = new Binding(t,vb);
} 


bool DoubleSubstitution::VariableBank::bind (Var v1, 
					     const Term& t2,
					     VariableBank* b2)
{
  if (occurs(v1,t2,b2)) {
    return false;
  }
  set(v1,t2,b2);
  return true;
} 



void DoubleSubstitution::VariableBank::extend (int ind)
{
  ASS(ind >= _capacity);

  int newCapacity = Int::max(_capacity+8,ind+1);
  Binding** newBindings = new Binding*[newCapacity];
  if (! newBindings) {
    ALLOCATION_FAILURE;
  }
  for (int i = _capacity-1; i >= 0; i--) {
    newBindings[i] = _bindings[i];
  }
  for (int j = _capacity; j < newCapacity; j++) {
    newBindings[j] = 0;
  }
  delete [] _bindings;
  _bindings = newBindings;
  _capacity = newCapacity;
} 



bool DoubleSubstitution::VariableBank::occurs (Var v1,
					       const Term& t2,
					       const VariableBank* b2) const
{
  switch (t2.tag()) {
  case VAR: 
    {
      Var v2 = t2.var();
      Binding* b = b2->get(v2);
      if (b) {
	return occurs(v1,b->term,b->bank);
      }
      
      return v1 == v2 && this == b2;
    }

  case COMPOUND:
    return occurs(v1,t2.args(),b2);

  case NUMERIC:
    return false;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



bool DoubleSubstitution::VariableBank::occurs (Var v1,
					       const TermList& ts2,
					       const VariableBank* b2) const
{
  Iterator<Term> args(ts2);
  while (args.hasNext()) {
    if (occurs(v1,args.next(),b2)) {
      return true;
    }
  }
  return false;
} 



void DoubleSubstitution::apply (LiteralList& lits,int index)
{
  ASS(index >= 0 && index <= 1);

  apply(lits,_bank+index);
} 



void DoubleSubstitution::apply (LiteralList& lits,VariableBank* bank)
{
  if (lits.isEmpty()) {
    return;
  }

  Literal l(lits.head());
  apply(l,bank);
  LiteralList ls(lits.tail());
  apply(ls,bank);
  if (l == lits.head() && ls == lits.tail()) {
    return;
  }
  lits = LiteralList(l,ls);
} 



void DoubleSubstitution::apply (Literal& lit,VariableBank* bank)
{
  Atom atm(lit.atom());
  apply(atm,bank);
  if (atm == lit.atom()) {
    return;
  }
  lit = Literal(lit.sign(),atm);
} 



void DoubleSubstitution::apply (Atom& atm,VariableBank* bank)
{
  TermList ts(atm.args());
  apply(ts,bank);
  if (ts == atm.args()) {
    return;
  }
  atm = Atom(atm.functor(),ts);
} 



void DoubleSubstitution::apply (TermList& terms,VariableBank* bank)
{
  if (terms.isEmpty()) {
    return;
  }

  Term t(terms.head());
  apply(t,bank);
  TermList ts(terms.tail());
  apply(ts,bank);
  if (t == terms.head() && ts == terms.tail()) {
    return;
  }
  terms = TermList(t,ts);

} 



void DoubleSubstitution::apply (Term& term,VariableBank* bank)
{
  switch (term.tag()) {
  case VAR:
    {
      Var v = term.var();
      Binding* b = bank->get(v);
      if (b) {
	if (b->bank == 0) {
	  term = b->term;
	  return;
	}
	Term t(b->term);
	apply(t,b->bank);
	b->term = t;
	b->bank = 0;
	term = t;
	return;
      }
      
      Term t(Var(_nextVariable++));
      bank->set(v,t,0);
      term = t;
    }
    return;

  case COMPOUND: 
    {
      TermList ts(term.args());
      apply(ts,bank);
      if (ts == term.args()) {
	return;
      }
      term = Term(term.functor(),ts);
    }
    return;

  case NUMERIC:
    return;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 



string DoubleSubstitution::toString () const
{
  return _bank[0].toString(_bank) + 
         _bank[1].toString(_bank); 
} 



string DoubleSubstitution::VariableBank::toString (const VariableBank* firstBank) const
{
  string result;
  if (! _bindings) {
    return result;
  }

  for (int i=0;i < _capacity;i++) {
    Binding* b = _bindings[i];
    if (b) {
      result += (string)"(" + 
                Int::toString((this-firstBank) + 1) + ',' + 
                Var(i).toString() + ") -> ";
      if (b->bank) {
	result += (string)"(" + 
	          Int::toString(((b->bank)-firstBank) + 1) + 
	          ',' + b->term.toString() + ")\n";
      }
      else {
	result += b->term.toString() + '\n';
      }
    }
  }

  return result;
} 


