
/*
 * File EqualityAxioms.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#include "Tracer.hpp"
#include "EqualityAxioms.hpp"
#include "Unit.hpp"



bool EqualityAxioms::isEqualityAxiom (const Unit& u)
{
  switch (u.unitType()) {
  case CLAUSE:  
    return isEqualityAxiom (u.clause());

  case FORMULA:
    return isEqualityAxiom (u.formula());

#if DEBUG_PREPRO
  default:
    ASS (false);
#endif
  }
} 



bool EqualityAxioms::isReflexivityAxiom (const Formula& f)
{
  TRACER("EqualityAxioms::isReflexivityAxiom");

  Formula g(f);
  while (g.connective() == FORALL) {
    g = g.qarg ();
  }

  
  return g.connective() == ATOM &&
         isTautology(g.atom());
} 



bool EqualityAxioms::isSymmetryAxiom (const Formula& g)
{
  TRACER ("EqualityAxioms::isSymmetryAxiom (const Formula& g)");

  Formula f(g);
  while (f.connective() == FORALL) {
    f = f.qarg();
  }

  
  if (f.connective() != IMP) {
    return false;
  }

  
  Formula f1 (f.left());
  if (f1.connective() != ATOM) {
    return false;
  }
  
  Formula f2 (f.right());
  if (f2.connective() != ATOM) {
    return false;
  }

  return isSwapOf(f1.atom(),f2.atom());
} 



bool EqualityAxioms::isTransitivityAxiom (const Formula& g)
{
  TRACER ("EqualityAxioms::isTransitivityAxiom");

  Formula f (g);
  while (f.connective() == FORALL) {
    f = f.qarg ();
  }

  
  if (f.connective() != IMP) {
    return false;
  }

  
  Formula c (f.left());
  if (c.connective() != AND || ! c.args().hasLength(2)) {
    return false;
  }

  
  Formula c1 (c.args().head ());
  if (c1.connective() != ATOM) {
    return false;
  }
  Formula c2 (c.args().second());
  if (c2.connective() != ATOM) {
    return false;
  }
  
  Formula c3 (f.right());
  if (c3.connective() != ATOM) {
    return false;
  }

  return isTransitivity(c1.atom(), c2.atom(), c3.atom());
} 



bool EqualityAxioms::isFunctionReflexivityAxiom (const Formula& g)
{
  TRACER("isFunctionReflexivityAxiom");

  Formula f (g);
  while (f.connective() == FORALL) {
    f = f.qarg ();
  }

  
  if (f.connective() != IMP) {
    return false;
  }

  
  if (f.left().connective() != ATOM || 
      f.right().connective() != ATOM ) {
    return false;
  }
  
  return isFunctionMonotonicity(f.left().atom(), f.right().atom());
} 



bool EqualityAxioms::isPredicateReflexivityAxiom (const Formula& g)
{
  TRACER("isPredicateReflexivityAxiom");
  Formula f (g);
  while (f.connective() == FORALL) {
    f = f.qarg();
  }

  
  if (f.connective() != IMP) {
    return false;
  }

  
  Formula c (f.left());
  if (c.connective() != AND || ! c.args().hasLength(2)) {
    return false;
  }
  Formula c1 (c.args().head());
  if ( c1.connective() != ATOM ) {
    return false;
  }
  Formula c2 (c.args().second());
  if ( c2.connective() != ATOM ) {
    return false;
  }
  Formula c3 (f.right()); 
  if ( c3.connective() != ATOM ) {
    return false;
  }
  Atom a1 (c1.atom());
  Atom a2 (c2.atom());
  Atom a3 (c3.atom());

  return isPredicateMonotonicity(a1,a2,a3) ||
         isPredicateMonotonicity(a2,a1,a3);
} 



bool EqualityAxioms::isEqualityAxiom (const Formula& f)
{
  TRACER("EqualityAxioms::isEqualityAxiom (const Formula& f)");

  return isReflexivityAxiom(f) || 
         isSymmetryAxiom(f) ||
         isTransitivityAxiom(f) ||
         isFunctionReflexivityAxiom(f) ||
         isPredicateReflexivityAxiom(f);
} 



bool EqualityAxioms::isTautology (const Atom& a)
{
  TRACER("EqualityAxioms::isTautology");

  if ( ! a.isEquality() )
    return false;

  return a.args().head().equals(a.args().second());
} 



bool EqualityAxioms::isSwapOf (const Atom& a1, const Atom& a2)
{
  TRACER("EqualityAxioms::isSwapOf");

  if (! (a1.isEquality() && a2.isEquality())) {
    return false;
  }

  return a1.args().head().equals(a2.args().second()) &&
         a2.args().head().equals(a1.args().second());
} 



bool EqualityAxioms::isTransitivity (const Atom& a1, 
				     const Atom& a2, 
				     const Atom& a3) 
{
  TRACER("EqualityAxioms::isTransitivity");

  if ( ! (a1.isEquality() && 
          a2.isEquality() && 
          a3.isEquality()) ) {
    return false;
  }
  
  Term l1 (a1.args().head());
  Term r1 (a1.args().second());
  if ( l1.equals(r1) ) {
    return false;
  }
  Term l2 (a2.args().head());
  Term r2 (a2.args().second());
  if ( l2.equals(r2) ) {
    return false;
  }
  Term l3 (a3.args().head());
  Term r3 (a3.args().second());
  if ( l3.equals(r3) ) {
    return false;
  }

  if ( l1.equals(l2) ) { 
    return (r1.equals(l3) && r2.equals(r3)) ||
           (r2.equals(l3) && r1.equals(r3));
  }
  
  if ( l1.equals(r2) ) { 
    return (r1.equals(l3) && l2.equals(r3)) ||
           (l2.equals(l3) && r1.equals(r3));
  }
  
  if ( r1.equals(l2) ) { 
    return (l1.equals(l3) && r2.equals(r3)) ||
           (r2.equals(l3) && l1.equals(r3));
  }
  
  if ( r1.equals(r2) ) { 
    return (l1.equals(l3) && l2.equals(r3)) ||
           (l2.equals(l3) && l1.equals(r3));
  }

  return false;
} 



bool EqualityAxioms::isFunctionMonotonicity (const Atom& a1, const Atom& a2) 
{
  TRACER("EqualityAxioms::isFunctionMonotonicity");

  if (! (a1.isEquality() && a2.isEquality())) {
    return false;
  }

  Term x (a1.args().head());
  Term y (a1.args().second());
  if (x.tag() != VAR || y.tag() != VAR) {
    return false;
  }

  Var vx = x.var();
  Var vy = y.var();
  
  return vx != vy &&
         equalsUpTo(a2.args().head(),a2.args().second(),vx,vy);
} 



bool EqualityAxioms::isPredicateMonotonicity (const Atom& a1, 
					      const Atom& a2, 
					      const Atom& a3)
{
  TRACER("EqualityAxioms::isPredicateMonotonicity");

  if ( ! a1.isEquality() )
    return false;
  if ( a2.isEquality() )
    return false;
  if ( a3.isEquality() )
    return false;

  Term x (a1.args().head());
  Term y (a1.args().second());
  if ( x.tag() != VAR || y.tag() != VAR ) {
    return false;
  }

  Var vx = x.var();
  Var vy = y.var();
  
  if ( vx == vy ) 
    return false;

  if ( a2.functor() != a3.functor() )
    return false;
  
  return equalsUpTo(a2.args(),a3.args(),vx,vy);
} 



bool EqualityAxioms::isEqualityAxiom (const Clause& c)
{
  TRACER("EqualityAxioms::isEqualityAxiom (const Clause& c)");

  return isReflexivityAxiom (c) || 
         isSymmetryAxiom (c) ||
         isTransitivityAxiom (c) ||
         isFunctionReflexivityAxiom (c) ||
         isPredicateReflexivityAxiom (c);
} 



bool EqualityAxioms::isReflexivityAxiom (const Clause& c)
{
  TRACER("EqualityAxioms::isReflexivityAxiom");

  LiteralList ls(c.literals());
  if (! ls.hasLength(1)) {
    return false;
  }
  Literal l(ls.head());
  return l.isPositive() && 
         isTautology(l.atom());
} 



bool EqualityAxioms::isSymmetryAxiom (const Clause& c)
{
  TRACER("EqualityAxioms::isSymmetryAxiom");

  LiteralList ls (c.literals());
  if (! ls.hasLength(2)) {
    return false;
  }
  Literal l1 (ls.head());
  Literal l2 (ls.second());
  if (l1.sign() == l2.sign()) {
    return false;
  }
  return isSwapOf(l1.atom(),l2.atom());
} 



bool EqualityAxioms::isTransitivityAxiom (const Clause& c)
{
  TRACER("EqualityAxioms::isTransitivityAxiom");

  LiteralList ls (c.literals());
  if (! ls.hasLength(3)) { 
    return false;
  }
  Literal l1 (ls.head());
  Literal l2 (ls.second());
  Literal l3 (ls.third());
  Atom a1;
  Atom a2;
  Atom a3;
  if (l1.isPositive()) {
    if (l2.isPositive() || l3.isPositive()) {
      return false;
    }
    a3 = l1.atom ();
    a2 = l2.atom ();
    a1 = l3.atom ();
  }
  
  else if (l2.isPositive()) {
    if (l3.isPositive()) {
      return false;
    }
    a3 = l2.atom ();
    a2 = l1.atom ();
    a1 = l3.atom ();
  }
  
  else if (l3.isPositive()) {
    a1 = l1.atom ();
    a2 = l2.atom ();
    a3 = l3.atom ();
  }
  
  else {
    return false;
  }

  return isTransitivity(a1, a2, a3);
} 



bool EqualityAxioms::isFunctionReflexivityAxiom (const Clause& c)
{
  TRACER("EqualityAxioms::isFunctionReflexivityAxiom");

  LiteralList ls (c.literals());
  if (! ls.hasLength(2)) { 
    return false;
  }
  Literal l1 (ls.head());
  Literal l2 (ls.second());
  ls = ls.tail();
  if (ls.isNonEmpty()) { 
    return false;
  }
  if (l1.sign() == l2.sign()) {
    return false;
  }
  Atom a1;
  Atom a2;
  if (l1.isPositive()) {
    a1 = l2.atom ();
    a2 = l1.atom ();
  }
  else { 
    a1 = l1.atom ();
    a2 = l2.atom ();
  }
  return isFunctionMonotonicity (a1,a2);
} 



bool EqualityAxioms::isPredicateReflexivityAxiom (const Clause& c)
{
  TRACER("EqualityAxioms::isPredicateReflexivityAxiom");

  LiteralList ls (c.literals());
  if (! ls.hasLength(3)) { 
    return false;
  }
  Literal l1 (ls.head());
  Literal l2 (ls.second());
  Literal l3 (ls.third());

  Atom a1;
  Atom a2;
  Atom a3;
  if (l1.isPositive()) {
    if (l2.isPositive() || l3.isPositive()) {
      return false;
    }
    a3 = l1.atom ();
    a2 = l2.atom ();
    a1 = l3.atom ();
  }
  
  else if (l2.isPositive()) {
    if (l3.isPositive()) {
      return false;
    }
    a3 = l2.atom ();
    a2 = l1.atom ();
    a1 = l3.atom ();
  }
  
  else if (l3.isPositive()) {
    a1 = l1.atom ();
    a2 = l2.atom ();
    a3 = l3.atom ();
  }
  
  else {
    return false;
  }

  return isPredicateMonotonicity (a1,a2,a3) ||
         isPredicateMonotonicity (a2,a1,a3);
} 



bool EqualityAxioms::equalsUpTo (const TermList& ls, const TermList& rs, 
				 Var x, Var y)
{
  ASS (ls.length() == rs.length());

  Iterator<Term> ts1 (ls);
  Iterator<Term> ts2 (rs);

  while (ts1.hasNext()) {
    if (! equalsUpTo(ts1.next(),ts2.next(),x,y)) {
      return false;
    }
  }
  return true;
} 



bool EqualityAxioms::equalsUpTo (const Term& l, const Term& r, Var x, Var y)
{
  if (l.tag() != r.tag()) {
    return false;
  }

  switch (l.tag()) {
  case VAR:
    if (l.var() == x) {
      return r.var() == y;
    }
    if (l.var() == y) {
      return r.var() == x;
    }
    return l.var() == r.var();
    
  case NUMERIC:
    return l.number() == r.number();

  case COMPOUND:
    return l.functor() == r.functor() &&
           equalsUpTo(l.args(),r.args(),x,y);

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 
