
/*
 * File FormulaVarIterator.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Tracer.hpp"
#include "FormulaVarIterator.hpp"
#include "Formula.hpp"


class FormulaVarIterator::Element
{
public:
  enum Tag {
    TERMS = 0,
    FORMULA = 1,
    FORMULAS = 2
  };

  Tag _tag;
  union { 
    FormulaList* _formulas;
    Formula* _formula;
    TermList* _terms;
  };
  Element* _rest;

  Element (const FormulaList& fs, Element* rest);
  Element (const TermList& ts, Element* rest);
  Element (const Formula& f, Element* rest);
}; 



FormulaVarIterator::FormulaVarIterator (const Formula& f)
  : _current(0),
    _reserve(new Element(f,0))
{
} 



FormulaVarIterator::Element::Element (const FormulaList& fs, Element* rest)
  : _tag(FORMULAS),
    _formulas(const_cast<FormulaList*>(&fs)), 
    _rest(rest)
{
} 



FormulaVarIterator::Element::Element (const Formula& f, Element* rest)
  : _tag(FORMULA),
    _formula(const_cast<Formula*>(&f)), 
    _rest(rest)
{
} 



FormulaVarIterator::Element::Element (const TermList& ts, Element* rest)
  : _tag(TERMS),
    _terms(const_cast<TermList*>(&ts)), 
    _rest(rest)
{
} 



bool FormulaVarIterator::hasNext ()
{
  TRACER("FormulaVarIterator::hasNext");

  for (;;) {
    if (_current) { 
      switch (_current->tag()) {
      case VAR:
	if (_bound.getCount(_current->var()) == 0) { 
	  return true;
	}
	_current = 0;
	break;

      case COMPOUND:
	_reserve = new Element(_current->args(),_reserve);
	_current = 0;
	break;

      case NUMERIC:
	_current = 0;
	break;

#if DEBUG_PREPRO
      default:
	ASS(false);
#endif
      }
      continue;
    }
    
    if (! _reserve) {
      return false;
    }
    
    switch (_reserve->_tag) {
    case Element::TERMS:
      {
	TermList& ts = *_reserve->_terms;
	if (ts.isEmpty()) {
	  popReserve();
	  break;
	}
	
	_current = const_cast<Term*>(&ts.head());
	_reserve->_terms = const_cast<TermList*>(&ts.tail());
	break;
      }

    case Element::FORMULAS:
      {
	FormulaList& fs = *_reserve->_formulas;
	if (fs.isEmpty()) {
	  popReserve();
	  break;
	}
	
	const Formula& first = fs.head();
	_reserve->_formulas = const_cast<FormulaList*>(&fs.tail());
	_reserve = new Element(first,_reserve);
	break;
      }

    case Element::FORMULA:
      {
	Formula& f = *(_reserve->_formula);
	switch (f.connective()) {
	case ATOM:
	  _reserve->_tag = Element::TERMS;
	  _reserve->_terms = const_cast<TermList*>(&f.atom().args());
	  break;

	case AND:
	case OR:
	  _reserve->_tag = Element::FORMULAS;
	  _reserve->_formulas = const_cast<FormulaList*>(&f.args());
	  break;

	case IMP:
	case IFF:
	case XOR:
	  _reserve->_formula = const_cast<Formula*>(&f.right());
	  _reserve = new Element(f.left(),_reserve);
	  break;

	case NOT:
	  _reserve->_formula = const_cast<Formula*>(&f.uarg());
	  break;

	case FORALL:
	case EXISTS:
	  {
	    
	    Iterator<Var> vs(f.vars());
	    while (vs.hasNext()) {
	      _bound.incCount(vs.next());
	    }
	    _reserve->_formula = const_cast<Formula*>(&f.qarg());
	    break;
	  }
	  
#if DEBUG_PREPRO
	default:
	  ASS(false);
#endif
	}
	break;
      }
#if DEBUG_PREPRO
    default:
      ASS(false);
#endif
    }
  }
} 



void FormulaVarIterator::popReserve ()
{
  TRACER("FormulaVarIterator::popReserve()");
  ASS(_reserve);

  Element* next = _reserve->_rest;
  delete _reserve;
  _reserve = next;
} 



Var FormulaVarIterator::next ()
{
  TRACER("FormulaVarIterator::next");

  Var result = _current->var();
  _current = 0;
  return result;
} 



FormulaVarIterator::~FormulaVarIterator ()
{
  TRACER("FormulaVarIterator::~FormulaVarIterator");

  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
