
/*
 * File GapFiller.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Refutation.hpp"
#include "GapFiller.hpp"
#include "ProofChecker.hpp"
#include "Unit.hpp"
#include "XML.hpp"
#include "Tracer.hpp"
#include "Substitution.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"
#include "VarCounter.hpp"
#include "FormulaVarIterator.hpp"


void GapFiller::fill (Unit& unit)
{
  TRACER("GapFiller::fill");

  Inference* inference = unit.inference();

  switch (inference->rule()) {
  case IR_INPUT: 
  case IR_RECTIFY: 
  case IR_CLAUSIFY: 
  case IR_HALF_EQUIV:
  case IR_KERNEL: 
    
    return;

  case IR_SKOLEMIZE: 
    fillSkolem(unit);
    return;

  case IR_NNF:
  case IR_ENNF:
    fillNNF(unit);
    return;

  case IR_FLATTEN:
    fillFlatten(unit);
    return;

  case IR_MINISCOPE:
    fillMiniscope(unit);
    return;

  default:
    cout << Inference::toString(inference->rule()) << "\n";
    inference->toXML()->write(cout);
    cout << unit.toStringAsInference() << "\n";
    ASSERTION_VIOLATION;
    return;
  }
} 



void GapFiller::fillSkolem (Unit& unit)
{
  TRACER("GapFiller::fillSkolem");

  Unit premise;
  InferenceInfo* info;
  onePremiseFormulaInference(unit,premise,info);

  InferenceInfoTermList* tinfo = static_cast<InferenceInfoTermList*>(info);
  Unit result1;
  buildChoiceAxioms(premise,result1,tinfo);
  Unit result2;
  eliminateAQuantifiers(result1,result2,tinfo);
  if (unit == result2) { 
    return;
  }
  if (! unit.formula().equals(result2.formula())) {
    PROOF_CHECKING_ERROR(unit,false,0);
  }
  
  delete unit.inference();
  unit.setInference(result2.inference());
  result2.setInference(0);
} 



void GapFiller::buildChoiceAxioms (const Unit& premise,
				   Unit& conclusion,
				   const InferenceInfoTermList* info)
{
  TRACER("GapFiller::buildChoiceAxioms");

  if (! info) { 
    conclusion = premise;
    return;
  }
  const TermList& ts = info->terms();
  ASS(ts.isNonEmpty());
  const Term& t = ts.head();
  if (t.tag() != COMPOUND) { 
    buildChoiceAxioms(premise,conclusion,
		      static_cast<InferenceInfoTermList*>(info->next()));
    return;
  }

  
  try {
    const Formula& F = premise.formula();
    const Formula& f = info->position().subformula(F);
    if (f.connective() != EXISTS) {
      PROOF_CHECKING_ERROR(premise,true,info);
    }

    
    Substitution subst;
    
    const VarList& vs = f.vars();
    if (vs.length() != ts.length()) {
      PROOF_CHECKING_ERROR(premise,true,info);
    }

    Iterator<Var> ws(vs);
    Iterator<Term> ss(ts);
    while (ws.hasNext()) {
      subst.bind(ws.next(),ss.next());
    }

    
    Formula g(f.qarg());
    subst.apply(g);
    Formula choiceAxiom(IMP,f,g);

    
    
    const TermList& args = t.args();
    if (! args.isEmpty()) {
      VarList xs;
      termListToVarList(args,xs,premise,info);
      choiceAxiom = Formula(FORALL,xs,choiceAxiom);
    }

    
    Inference* newInference = new Inference(IR_CHOICE_AXIOM);
    InferenceInfo* newInfo =  new InferenceInfoTermList(IR_CHOICE_AXIOM,
							info->position(),
							0,
							info->terms());
    newInference->setInfo(newInfo);
    Unit choiceAxiomUnit(choiceAxiom,newInference);
    
    Formula G(F);
    info->position().replace(G,g);
    
    Inference* secondInference = new Inference(IR_MONOTONE_REPLACEMENT,
					       UnitList(choiceAxiomUnit,
							UnitList(premise)));
    InferenceInfo* secondInfo = new InferenceInfo(IR_MONOTONE_REPLACEMENT, 
						  info->position(),
						  0);
    secondInference->setInfo(secondInfo);
    Unit secondUnit(G,secondInference);
    buildChoiceAxioms(secondUnit,conclusion,
		      static_cast<InferenceInfoTermList*>(info->next()));
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }
} 



void GapFiller::eliminateAQuantifiers (const Unit& premise,
				       Unit& conclusion,
				       const InferenceInfoTermList* info)
{
  TRACER("GapFiller::eliminateAQuantifiers");

  if (! info) { 
    conclusion = premise;
    return;
  }
  
  
  const TermList& ts = info->terms();
  ASS(ts.isNonEmpty());
  const Term& t = ts.head();
  if (t.tag() != VAR) { 
    eliminateAQuantifiers(premise, 
			  conclusion, 
		          static_cast<InferenceInfoTermList*>(info->next()));
    return;
  }
  try {
    Formula F(premise.formula());
    
    const Formula& f = info->position().subformula(F);

    if(f.connective() != FORALL) {
      PROOF_CHECKING_ERROR(premise,true,info);
    }
    VarList xs(f.vars());
    
    
    if (ts.length() != xs.length()) {
      PROOF_CHECKING_ERROR(premise,true,info);
    }
    Iterator<Var> vs(xs);
    Iterator<Term> ss(ts);
    Substitution subst;
    while (vs.hasNext()) {
      subst.bind(vs.next(),ss.next());
    }
    Formula g(f.qarg());
    subst.apply(g);
    info->position().replace(F,g);

    Unit mid;
    
    Inference* newInference = new Inference(IR_FORALL_ELIMINATION,
					    UnitList(premise));
    InferenceInfo* newInfo = new InferenceInfoTermList(IR_FORALL_ELIMINATION,
						       info->position(),
						       0,
						       ts);
    newInference->setInfo(newInfo);
    Unit newUnit(F,newInference);
    eliminateAQuantifiers(newUnit, 
			  conclusion,
			  static_cast<InferenceInfoTermList*>(info->next()));
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }
} 



void GapFiller::termListToVarList (const TermList& 
				   ts,VarList& vs, 
				   const Unit& unit,
				   const InferenceInfo* info)
{
  TRACER("GapFiller::termListToVarList");

  if (ts.isEmpty()) {
    return;
  }
  const Term& t = ts.head();
  if (t.tag() != VAR) {
    PROOF_CHECKING_ERROR(unit,true,info);
  }
  termListToVarList(ts.tail(),vs,unit,info);
  vs.push(t.var());
} 



void GapFiller::fillNNF (Unit& unit)
{
  TRACER("GapFiller::fillNNF/1");

  Unit premise;
  InferenceInfo* info;
  onePremiseFormulaInference(unit,premise,info);

  Unit result;
  fillNNF(premise,result,info);
  if (! unit.formula().equals(result.formula())) {
    PROOF_CHECKING_ERROR(unit,false,0);
  }
  
  delete unit.inference();
  unit.setInference(result.inference());
  result.setInference(0);
} 



void GapFiller::fillNNF (const Unit& premise,
			 Unit& conclusion,
			 const InferenceInfo* info)
{
  TRACER("GapFiller::fillNNF/3");

  if (! info) { 
    conclusion = premise;
    return;
  }

  try {
    Formula F(premise.formula());
    
    const Formula& f = info->position().subformula(F);

    switch (info->rule()) {
    case IR_NOT_AND:
      {
	if (f.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& f1 = f.uarg();
	if (f1.connective() != AND) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	FormulaList fs;
	negateAll(f1.args(),fs);
	info->position().replace(F,Formula(OR,fs));
      }
      break;

    case IR_NOT_OR:
      {
	if (f.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& f1 = f.uarg();
	if (f1.connective() != OR) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	FormulaList fs;
	negateAll(f1.args(),fs);
	info->position().replace(F,Formula(AND,fs));
      }
      break;

    case IR_IMP_TO_OR:
      {
	if (f.connective() != IMP) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	info->position().replace(F,Formula(OR,
					   FormulaList(Formula(NOT,f.left()),
					   FormulaList(f.right()))));
      }
      break;

    case IR_NOT_IMP:
      {
	if (f.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& f1 = f.uarg();
	if (f1.connective() != IMP) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& l = f1.left();
	const Formula& r = f1.right();
	info->position().replace(F,Formula(AND,
					   FormulaList(l,
					   FormulaList(Formula(NOT,r)))));
      }
      break;

    case IR_NOT_IFF:
      {
	if (f.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& f1 = f.uarg();
	if (f1.connective() != IFF) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& l = f1.left();
	const Formula& r = f1.right();
	info->position().replace(F,Formula(XOR,l,r));
      }
      break;

    case IR_NOT_XOR:
      {
	if (f.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& f1 = f.uarg();
	if (f1.connective() != XOR) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& l = f1.left();
	const Formula& r = f1.right();
	info->position().replace(F,Formula(IFF,l,r));
      }
      break;

    case IR_NOT_NOT:
      {
	if (f.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& f1 = f.uarg();
	if (f1.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	info->position().replace(F,f1.uarg());
      }
      break;

    case IR_NOT_FORALL:
      {
	if (f.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& f1 = f.uarg();
	if (f1.connective() != FORALL) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	info->position().replace(F,Formula(EXISTS,f1.vars(),
					   Formula(NOT,f1.qarg())));
      }
      break;

    case IR_NOT_EXISTS:
      {
	if (f.connective() != NOT) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& f1 = f.uarg();
	if (f1.connective() != EXISTS) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	info->position().replace(F,Formula(FORALL,f1.vars(),
					   Formula(NOT,f1.qarg())));
      }
      break;

    case IR_IFF_TO_AND:
      {
	if (f.connective() != IFF) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& l = f.left();
	const Formula& r = f.right();
	info->position().replace(F,Formula(AND,
					   FormulaList(Formula(IMP,l,r),
					   FormulaList(Formula(IMP,r,l)))));
      }
      break;

    case IR_XOR_TO_AND:
      {
	if (f.connective() != XOR) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const Formula& l = f.left();
	const Formula& r = f.right();
	info->position().replace(F,
				 Formula(AND,
					 FormulaList(Formula(OR,
							     FormulaList(l,
							     FormulaList(r))),
					 FormulaList(Formula(OR,
							     FormulaList(Formula(NOT,l),
							     FormulaList(Formula(NOT,r))))))));
      }
      break;

    default:
      PROOF_CHECKING_ERROR(premise,true,info);
    }

    
    Unit mid;
    
    Inference* newInference = new Inference(info->rule(),
					    UnitList(premise));
    InferenceInfo* newInfo = new InferenceInfo(info->rule(),
					       info->position(),
					       0);
    newInference->setInfo(newInfo);
    Unit newUnit(F,newInference);
    fillNNF(newUnit, conclusion, info->next());
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }
} 



void GapFiller::negateAll (const FormulaList& in, FormulaList& out)
{
  if (in.isEmpty()) {
    return;
  }
  negateAll(in.tail(),out);
  out.push(Formula(NOT,in.head()));
} 



void GapFiller::fillFlatten (Unit& unit)
{
  TRACER("GapFiller::fillFlatten/1");

  Unit premise;
  InferenceInfo* info;
  onePremiseFormulaInference(unit,premise,info);

  Unit result;
  fillFlatten(premise,result,info);
  if (! unit.formula().equals(result.formula())) {
    PROOF_CHECKING_ERROR(unit,false,0);
  }
  
  delete unit.inference();
  unit.setInference(result.inference());
  result.setInference(0);
} 



void GapFiller::fillFlatten (const Unit& premise,
			     Unit& conclusion,
			     const InferenceInfo* info)
{
  TRACER("GapFiller::fillFlatten/3");

  if (! info) { 
    conclusion = premise;
    return;
  }
  Unit newUnit;

  try {
    
    fillFlatten(premise, newUnit, info->next());

    Formula F(newUnit.formula());
    if (info->rule() != IR_FLATTEN) {
      PROOF_CHECKING_ERROR(newUnit,true,info);
    }
    
    const Formula& f = info->position().subformula(F);
    switch (f.connective()) {
    case AND:
    case OR:
      {
	FormulaList fs(f.args());
	flatten(fs,f.connective());
	info->position().replace(F,Formula(f.connective(),fs));
      }
      break;

    case FORALL:
    case EXISTS:
      {
	Formula g(f.qarg());
	if (g.connective() != f.connective()) {
	  PROOF_CHECKING_ERROR(newUnit,true,info);
	}
	VarList vs (f.vars());
	vs.append(g.vars());
	info->position().replace(F,Formula(f.connective(),vs,g.qarg()));
      }
      break;

    default:
      
      PROOF_CHECKING_ERROR(newUnit,true,info);
    }

    
    Inference* newInference = new Inference(info->rule(),
					    UnitList(newUnit));
    InferenceInfo* newInfo = new InferenceInfo(info->rule(),
					       info->position(),
					       0);
    newInference->setInfo(newInfo);
    conclusion = Unit(F,newInference);
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(newUnit,true,info);
  }
} 



void GapFiller::flatten (FormulaList& fs,Connective con)
{
  if (fs.isEmpty()) {
    return;
  }
  Formula f(fs.head());
  FormulaList gs(fs.tail());
  flatten(gs,con);
  if (f.connective() == con) {
    FormulaList hs(f.args());
    hs.append(gs);
    fs = hs;
    return;
  }
  if (f == fs.head() && gs == fs.tail()) {
    return;
  }
  fs = FormulaList(f,gs);
} 



void GapFiller::onePremiseFormulaInference (const Unit& conclusion,
					    Unit& premise,
					    InferenceInfo*& info)
{
  TRACER("GapFiller::onePremiseFormulaInference");

  Inference* inference = conclusion.inference();

  const UnitList& premises = inference->premises();
  if (! premises.hasLength(1)) {
    PROOF_CHECKING_ERROR(conclusion,false,0);
  }
  premise = premises.head();
  if (conclusion.unitType() != FORMULA) {
    PROOF_CHECKING_ERROR(conclusion,false,0);    
  }
  if (premise.unitType() != FORMULA) {
    PROOF_CHECKING_ERROR(conclusion,false,0);    
  }

  info = inference->info();
} 



void GapFiller::fillMiniscope (Unit& unit)
{
  TRACER("GapFiller::fillMiniscope/1");

  Unit premise;
  InferenceInfo* info;
  onePremiseFormulaInference(unit,premise,info);

  Unit result;
  fillMiniscope(premise,result,info);
  if (! unit.formula().equals(result.formula())) {
    PROOF_CHECKING_ERROR(unit,false,0);
  }
  
  delete unit.inference();
  unit.setInference(result.inference());
  result.setInference(0);
} 



void GapFiller::fillMiniscope (const Unit& premise,
			       Unit& conclusion,
			       const InferenceInfo* info)
{
  TRACER("GapFiller::fillMiniscope/3");

  if (! info) { 
    conclusion = premise;
    return;
  }

  Unit mid;
  fillMiniscope(premise, mid, info->next());

  try {
    Formula F(mid.formula());
    
    const Formula& f = info->position().subformula(F);

    switch (info->rule()) {
    case IR_FORALL_AND:
      fillForallAnd(mid,true,info,F,f);
      break;

    case IR_EXISTS_OR:
      fillForallAnd(mid,false,info,F,f);
      break;

    case IR_FORALL_OR:
      fillForallOr(mid,true,
		   static_cast<const InferenceInfoTermList*>(info),F,f);
      break;

    case IR_EXISTS_AND:
      fillForallOr(mid,false,
		   static_cast<const InferenceInfoTermList*>(info),F,f);
      break;

    case IR_DUMMY_QUANTIFIER_REMOVAL:
      fillDummyQuantifierRemoval(mid,
				 static_cast<const InferenceInfoTermList*>(info),F,f);
      break;

    case IR_FLATTEN:
      fillFlatten(mid,info,F,f);
      break;

    case IR_QUANTIFIER_SWAP:
      fillQuantifierSwap(mid,info,F,f);
      break;

    default:
      PROOF_CHECKING_ERROR(mid,true,info);
    }

    
    
    Inference* newInference = new Inference(info->rule(),UnitList(mid));
    InferenceInfo* newInfo;
    switch (info->rule()) {
    case IR_FORALL_AND:
    case IR_EXISTS_OR:
    case IR_FLATTEN:
    case IR_QUANTIFIER_SWAP:
      newInfo = new InferenceInfo(info->rule(),info->position(),0);
      break;

    case IR_FORALL_OR:
    case IR_EXISTS_AND:
    case IR_DUMMY_QUANTIFIER_REMOVAL:
      newInfo = new InferenceInfoTermList(info->rule(),
					  info->position(),
					  0,
					  static_cast<const InferenceInfoTermList*>(info)->terms());
      break;

#if DEBUG_PREPRO
    default:
      ASS(false);
#endif
    }

    newInference->setInfo(newInfo);
    conclusion = Unit (F,newInference);
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(mid,true,info);
  }
} 



void GapFiller::markOccurrences (VarCounter& vc, const Formula& f)
{
  FormulaVarIterator it(f);
  while (it.hasNext()) {
    vc.setCount(it.next(),1);
  }
} 



void GapFiller::fillForallAnd (const Unit& premise, 
			       bool isForall,
			       const InferenceInfo* info,
			       Formula& F,
			       const Formula& f)
{
  Connective Q = isForall ? FORALL : EXISTS;
  Connective c = isForall ? AND : OR;

  if (f.connective() != Q) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }
  const Formula& f1 = f.qarg();
  if (f1.connective() != c) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }
  
  FormulaList fs;
  Stack<Formula> stack(fs);
  Iterator<Formula> gs(f1.args());
  while (gs.hasNext()) {
    stack.push(Formula(Q,f.vars(),gs.next()));
  }
  info->position().replace(F,Formula(c,fs));
} 



void GapFiller::fillForallOr (const Unit& premise, 
			      bool isForall,
			      const InferenceInfoTermList* info,
			      Formula& F,
			      const Formula& f)
{
  Connective Q = isForall ? FORALL : EXISTS;
  Connective c = isForall ? OR : AND;

  if (f.connective() != Q) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }
  const Formula& g = f.qarg();
  if (g.connective() != c) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }

  VarList vs;
  termListToVarList(info->terms(),vs,premise,info);
  if (vs.isEmpty()) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }

  const FormulaList& gs = g.args();
  
  
  
  int length = gs.length();
  Formula* G = new Formula [length];
  Iterator<Formula> Gs(gs);
  int i = 0;
  
  FormulaList occurring;
  Stack<Formula> stack(occurring);
  int firstOccPosition = -1;

  while (Gs.hasNext()) {
    const Formula& h = Gs.next();
    
    
    VarCounter vc;
    markOccurrences(vc,h);
    int occurs = vc.getCount(vs.head());
    Iterator<Var> ws(vs.tail());
    while (ws.hasNext()) {
      if (vc.getCount(ws.next()) != occurs) {
	PROOF_CHECKING_ERROR(premise,true,info);
      }
    }
    if (occurs) {
      if (firstOccPosition == -1) {
	firstOccPosition = i;
      }
      stack.push(h);
    }
    else {
      G[i] = h;
    }
    i++;
  }
  if (firstOccPosition == -1) { 
    PROOF_CHECKING_ERROR(premise,true,info);
  }
  
  
  VarList remaining;
  Stack<Var> rem(remaining);
  VarList qs(f.vars());
  VarList ys(vs);
  do {
    if (ys.isEmpty() || !(qs.head() == ys.head())) {
      rem.push(qs.head());
    }
    else {
      ys.pop();
    }
    qs.pop();
  }
  while (qs.isNonEmpty());

  if (ys.isNonEmpty()) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }

  if (occurring.hasLength(1)) {
    G[firstOccPosition] = Formula(Q,vs,occurring.head());
  }
  else {
    G[firstOccPosition] = Formula(Q,vs,Formula(c,occurring));
  }

  
  FormulaList fs;
  for (int j = length-1; j >= 0; j--) {
    const Formula& g = G[j];
    if (! g.isNull()) {
      fs.push(g);
    }
  }

  delete [] G;
  Formula result(c,fs);
  if (remaining.isNonEmpty()) {
    result = Formula(Q,remaining,result);
  } 
  info->position().replace(F,result);
} 



void GapFiller::fillDummyQuantifierRemoval (const Unit& premise, 
					    const InferenceInfoTermList* info,
					    Formula& F,
					    const Formula& f)
{
  Connective Q = f.connective();
  switch (Q) 
    {
    case FORALL:
    case EXISTS:
      break;

    default:
    PROOF_CHECKING_ERROR(premise,true,info);
    }

  
  VarCounter removed;
  Iterator<Term> ts(info->terms());
  while (ts.hasNext()) {
    Term t(ts.next());
    if (t.tag() != VAR) {
      PROOF_CHECKING_ERROR(premise,true,info);
    }
    removed.incCount(t.var());
  }

  VarList remains;
  Stack<Var> ws(remains); 
  Iterator<Var> vs(f.vars());
  while (vs.hasNext()) {
    Var v(vs.next());
    if (removed.getCount(v) == 0) {
      ws.push(v);
    }
  }

  const Formula& g = f.qarg();
  info->position().replace(F,remains.isEmpty() ? 
			     g : 
			     Formula(Q,remains,g));
} 



void GapFiller::fillFlatten (const Unit& premise, 
			     const InferenceInfo* info,
			     Formula& F,
			     const Formula& f)
{
  try {
    switch (f.connective()) {
    case AND:
    case OR:
      {
	FormulaList fs(f.args());
	flatten(fs,f.connective());
	info->position().replace(F,Formula(f.connective(),fs));
      }
      break;
      
    case FORALL:
    case EXISTS:
      {
	Formula g(f.qarg());
	if (g.connective() != f.connective()) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	VarList vs (f.vars());
	vs.append(g.vars());
	info->position().replace(F,Formula(f.connective(),vs,g.qarg()));
      }
      break;

    default:
      
      PROOF_CHECKING_ERROR(premise,true,info);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }
} 



void GapFiller::fillQuantifierSwap (const Unit& premise, 
				    const InferenceInfo* info,
				    Formula& F,
				    const Formula& f)
{
  try {
    Connective Q = f.connective();
    switch (Q) {
    case FORALL:
    case EXISTS:
      {
	Formula g(f.qarg());
	if (g.connective() != Q) {
	  PROOF_CHECKING_ERROR(premise,true,info);
	}
	const VarList& fvs = f.vars();
	const VarList& gvs = g.vars();
	info->position().replace(F,Formula(Q,gvs,Formula(Q,fvs,g.qarg())));
      }
      break;

    default:
      
      PROOF_CHECKING_ERROR(premise,true,info);
    }
  }
  catch (NonExistingPositionException e) {
    PROOF_CHECKING_ERROR(premise,true,info);
  }
} 
