
/*
 * File Rectify.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Tracer.hpp"
#include "Rectify.hpp"
#include "Unit.hpp"
#include "Substitution.hpp"
#include "Signature.hpp"


void Rectify::rectify (Unit& unit, VarList& answerVariables)
{
  TRACER("Rectify::rectify (Unit&...)");

  Formula f (unit.formula());
  Rectify rect;
  rect.rectify(f);
  VarList freeVars(rect._free);

  if (unit.inputType() == CONJECTURE) {
    
    TermList answerAtomArgs;
    rect._subst.getDomain(answerVariables);
    rect._subst.getRange(answerAtomArgs);
    if (freeVars.isNonEmpty()) {
      
      Signature::sig.createAnswerAtom (answerAtomArgs);
    }
  }

  if (!f.equals(unit.formula())) {
    unit = Unit(IR_RECTIFY, f, unit);
  }

  if (freeVars.isNonEmpty()) {
    Formula f1 (FORALL,freeVars,f);
    unit = Unit(IR_CLOSURE, f1,unit);
  }
} 



void Rectify::rectify (Formula& f)
{
  TRACER("Rectify::rectify (Formula&)");

  switch (f.connective()) {
  case ATOM: 
    {
      Atom a(f.atom());
      rectify(a);
      if (a == f.atom()) {
	return;
      }
      f = Formula(a);
      return;
    }

  case AND: 
  case OR: 
    {
      FormulaList newArgs (f.args());
      rectify(newArgs);
      if (newArgs == f.args()) {
	return;
      }
      f = Formula(f.connective(), newArgs);
      return;
    }

  case IMP: 
  case IFF: 
  case XOR:
    {
      Formula l (f.left());
      rectify(l);
      Formula r (f.right());
      rectify (r);
      if (l == f.left() && r == f.right()) {
	return;
      }
      f = Formula(f.connective(), l, r);
      return;
    }

  case NOT:
    {
      Formula arg(f.uarg());
      rectify(arg);
      if (f.uarg() == arg) {
	return;
      }
      f = Formula(NOT, arg);
      return;
    }

  case FORALL: 
  case EXISTS:
    {
      VarList vs;
      rectify(f.vars(), vs);
      Formula arg(f.qarg());
      rectify(arg);
      Iterator<Var> ws (f.vars());
      while (ws.hasNext()) {
	Var w(ws.next());
	_subst.undoBinding (w);
      }
      if (vs == f.vars() && arg == f.qarg()) {
	return;
      }
      f = Formula(f.connective(), vs, arg);
      return;
    }

  default:
    ASS( false );
  }
} 



void Rectify::rectify (FormulaList& fs)
{
  TRACER("Rectify::rectify (FormulaList&)");

  if (fs.isEmpty()) {
    return;
  }
  Formula h (fs.head());
  rectify(h);
  FormulaList t (fs.tail());
  rectify(t);
  if (h == fs.head() && t == fs.tail()) {
    return;
  }

  fs = FormulaList (h,t);
} 



void Rectify::rectify (Atom& a)
{
  TRACER("Rectify::rectify (Atom& a)");

  TermList ts (a.args());
  rectify(ts);
  if (ts == a.args()) { 
    return;
  }

  a = Atom(a.functor(),ts);
} 



void Rectify::rectify (Term& t) 
{
  TRACER("Rectify::rectify (Term& t)");

  switch (t.tag()) {
  case VAR:
    {
      Var v = t.var();
      Term t1;
      if (_subst.isBound(v,t1)) {
	if (! t.equals(t1)) {
	  t = t1;
	}
	return;
      }
      
      Var newV(_subst.freshVariable());
      _free.push (newV);
      if (t.var() != newV) { 
	Term s (newV);
	t = s;
      }
      _subst.bind(v,t);
      return;
    }

  case NUMERIC:
    return;

  case COMPOUND:
    {
      TermList ts(t.args());
      rectify(ts);

      if (ts == t.args()) { 
	return;
      }

      t = Term(t.functor(), ts);
      return;
    }
  }
} 



void Rectify::rectify (TermList& ts)
{
  TRACER("Rectify::rectify (TermList& ts)");

  if (ts.isEmpty()) {
    return;
  }
  
  Term hd (ts.head());
  rectify(hd);
  TermList tl (ts.tail());
  rectify(tl);

  if (hd == ts.head() && tl == ts.tail()) {
    return;
  }

  ts = TermList(hd,tl);
} 



void Rectify::rectify (const VarList& vs, VarList& result)
{
  TRACER ("Rectify::rectify (const VarList& vs...)");

  if (vs.isEmpty()) {
    return;
  }

  Var v (vs.head());
  Var newV(_subst.freshVariable());
  _subst.bind(v, Term(newV));
  rectify(vs.tail(), result);
  result = VarList(newV, result);
} 
