
/*
 * File Refutation.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "GapFiller.hpp"
#include "Inference.hpp"
#include "IntNameTable.hpp"
#include "Kernel.hpp"
#include "KIFMode.hpp"
#include "LaTeX.hpp"
#include "Map.hpp"
#include "Options.hpp"
#include "Problem.hpp"
#include "ProofChecker.hpp"
#include "Refutation.hpp"
#include "Renaming.hpp"
#include "TPTP.hpp"
#include "XML.hpp"


class Refutation::Map 
  : public ::Map <int, Unit>
{
}; 



Refutation::Refutation (const Kernel& kernel, 
			const Problem& problem,
			const Options& options)
  : _kernel (kernel),
    _problem(problem),
    _options(options)
{
  TRACER ("Refutation::Refutation");
 
  Unit::setLastNumber(kernel.refutation()->number());
  Map map;
  collectKernelClause(kernel.refutation(), map, options.proofChecking());

  
  
  UnitList refutation;
  int lastNumber = 0; 
  Iterator<Unit> units (_refutation);
  while (units.hasNext()) {
    Unit u (units.next());
    u.setNumberInRefutation(++ lastNumber);
    refutation.push(u);
  }
  _refutation = refutation;
  if (options.proofChecking()) {
    ProofChecker::check (*this);
  }
} 



Unit Refutation::collectKernelClause (const void* kernelClause, 
				      Map& map,
				      bool fillGaps)
{
  TRACER ("Refutation::collect (VampireKernel::Clause ...)");

  const VampireKernel::Clause* clause = reinterpret_cast<const VampireKernel::Clause*>(kernelClause);

  int number = clause->number();
  Unit unit;
  if (map.find(number, unit)) { 
    return unit;
  }

  
  if (clause->isInputClause()) { 
    
    Unit parent (clause->inputClauseOrigin());
    collect(parent,map,fillGaps);
    return parent;
  }

  
  
  
  UnitList parents = collectKernelClauseParents(clause->firstAncestor(),
						map,
						fillGaps);

  
  Clause cls;
  _kernel.getClause(clause,cls);
  
  if (parents.hasLength(1)) {
    Unit parent(parents.head());
    if (parent.unitType() == CLAUSE && 
	Renaming::isVariant(parent.clause(),cls)) {
      return parent;
    }
  }
  VampireKernel::Clause::BackgroundFlagIterator flags(clause-> backgroundFlagIterator());
  Unit result(cls,parents,clause->number(),flags);

  
  collect(result,map,fillGaps);
  return result;
} 



void Refutation::collect (Unit& unit, Map& map, bool fillGaps)
{
  TRACER ("Refutation::collect (const Unit& ...)");

  int number = unit.number();

  Unit unit1;
  if (map.find(number,unit1)) { 
    return;
  }

  
  
  Iterator<Unit> parents (unit.inference()->premises());
  while (parents.hasNext()) {
    Unit par(parents.next());
    collect(par, map, fillGaps);
  }

  if (fillGaps) {
    GapFiller::fill(unit);
    
    
    parents.reset(unit.inference()->premises());
    while (parents.hasNext()) {
      Unit par(parents.next());
      collect(par, map, false);
    }
  }

  
  map.insert (number, unit);
  _refutation.push(unit);
} 



UnitList Refutation::collectKernelClauseParents (const void* ps,
						 Map& map,
						 bool fillGaps)
{
  TRACER ("Refutation::collect (VampireKernel::Clause::Ancestor* ...)");

  const VampireKernel::Clause::Ancestor* parent =
    reinterpret_cast<const VampireKernel::Clause::Ancestor*>(ps);

  if (! parent) {
    return UnitList ();
  }

  
  const VampireKernel::Clause* clause = parent->clause();
  const VampireKernel::Clause::Ancestor* rest = parent->next ();
  if (clause->isBuiltInTheoryFact()) {
    return collectKernelClauseParents (rest, map, fillGaps);
  }

  
  Unit u(collectKernelClause(clause, map, fillGaps));
  return UnitList(u, collectKernelClauseParents(rest, map, fillGaps));
} 



XML::Element* Refutation::toXML () const
{
  TRACER("Refutation::toXML");

  XML::DeepElement* ref = new XML::DeepElement("refutation");
  Iterator<Unit> units (_refutation);
  while (units.hasNext()) {
    ref->addChild(units.next().toXML());
  }
  return ref;
} 



void Refutation::prettyPrint (ostream& str, 
			      int answerNumber)
{
  LaTeX::output (*this,_options);

  switch (_options.outputSyntax()) {
  case Options::OS_KIF:
    {
      XML::Element* xml = KIFMode::toXML(*this,answerNumber);
      xml->write(str);
      xml->destroy();
      return;
    }

  case Options::OS_NATIVE:
    {
      str << "=========== Refutation ==========\n";

      Iterator<Unit> units(refutation());
      while (units.hasNext()) {
	str << units.next().toStringAsInference();
      }

      str << "======= End of refutation =======\n";
      return;
    }

  case Options::OS_TPTP:
    str << TPTP::toString(*this);
    return;

  case Options::OS_XML:
    {
      XML::Element* xml = toXML();
      xml->write(str);
      xml->destroy();
      return;
    }
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 

