
/*
 * File Substitution.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Tracer.hpp"
#include "Int.hpp"
#include "XML.hpp"
#include "Atom.hpp"
#include "Formula.hpp"
#include "Substitution.hpp"



Substitution::Substitution ()
  : _top(0),
    _bindings(0),
    _lastNewVar(-1)
{
  TRACER("Substitution::Substitution");
} 



Substitution::~Substitution ()
{
  TRACER("Substitution::~Substitution");

  for (int i = 0; i < _top; i++) {
    _bindings[i]->destroy();
  }
  delete [] _bindings;
} 



void Substitution::bind (Var w, const Term& t)
{
  TRACER("Substitution::bind");

  int v = w.toInt();
  if (v >= _top) { 
    extend(v);
  }
  
  _bindings[v] = new Binding(t,_bindings[v]);
} 



bool Substitution::isBound (Var w, Term& t) const
{
  TRACER( "Substitution::isBound/2" );

  int v = w.toInt();
  if (v >= _top) { 
    return false;
  }
  Binding* b = _bindings[v];
  if (b->isEmpty()) {
    return false;
  }
  t = b->head();
  return true;
} 



bool Substitution::isBound (Var w) const
{
  TRACER( "Substitution::isBound/1" );

  int v = w.toInt();
  if (v >= _top) { 
    return false;
  }
  return _bindings[v]->isNonEmpty();
} 



void Substitution::getDomain (VarList& vs) const
{
  TRACER("Substitution::getDomain");

  VarList result;

  for (int i = 0; i < _top; i++) {
    if (_bindings[i]->isNonEmpty()) {
      result.push(Var(i));
    }
  }
  vs = result;
} 



void Substitution::getRange (TermList& ts) const
{
  TRACER("Substitution::getRange");

  TermList result;

  for (int i = 0; i < _top; i++) {
    Binding* b = _bindings[i];
    if (b->isNonEmpty()) {
      result = TermList(b->head(),result);
    }
  }
  ts = result;
} 



void Substitution::undoBinding (Var w)
{
  TRACER("Substitution::undoBinding");

  int v = w.toInt();

  ASS(v < _top);
  ASS(_bindings[v]->isNonEmpty());

  Binding* b = _bindings[v];
  _bindings[v] = b->tail();
  delete b;
} 



bool Substitution::isInRange (const Term& t) const
{
  TRACER("Substitution::isInRange");

  for (int i = 0; i < _top; i++) {
    Binding* b = _bindings[i];
    if (b->isNonEmpty() &&
	t.equals(b->head())) {
      return true;
    }
  }

  return false;
} 



void Substitution::extend (int v)
{
  TRACER("Substitution::extend");

  
  int newTop = Int::max(_top+16,v+1);

  
  Binding** newBindings = new Binding*[newTop];
  if (! newBindings) {
    ALLOCATION_FAILURE;
  }
  
  for (int i = 0;i < _top;i++) {
    newBindings[i] = _bindings[i];
  }
  
  for (int k = _top;k < newTop;k++) {
    newBindings[k] = 0;
  }

  delete [] _bindings;

  _bindings = newBindings;
  _top = newTop;
} 



XML::Element* Substitution::toXML () const
{
  XML::DeepElement* subst = new XML::DeepElement("substitution");
  subst->addChild(new XML::IntegerElement("top",_top));
  XML::DeepElement* bindings = new XML::DeepElement("bindings");
  subst->addChild(bindings);
  for (int i = 0;i < _top;i++) {
    if (_bindings[i]->isNonEmpty()) {
      XML::EmptyElement* binding = new XML::EmptyElement("binding");
      bindings->addChild(binding);
      binding->addAttribute(new XML::IntegerAttribute("var",i));
      binding->addAttribute(
	new XML::TextAttribute("term",_bindings[i]->head().toString()));
    }
  }
  return subst;
} 




























void Substitution::apply (Term& t) const
{
  TRACER ("Substitution::apply (Term& t)");

  switch (t.tag()) {
  case VAR:
    isBound(t.var(),t);
    return;
    
  case NUMERIC:
    return;

  case COMPOUND:
    {
      TermList ts (t.args());
      apply(ts);
      if (ts == t.args()) { 
	return;
      }
      t = Term (t.functor(),ts);
      return;
    }
  }
} 



void Substitution::apply (TermList& ts) const
{
  TRACER ("Substitution::apply (TermList& ts)");

  if (ts.isEmpty()) {
    return;
  }
  
  Term t (ts.head());
  apply(t);
  TermList rs (ts.tail());
  apply(rs);

  if (t == ts.head() && rs == ts.tail()) {
    return;
  }

  ts = TermList(t,rs);
} 



void Substitution::apply (Atom& a) const
{
  TRACER ("Substitution::apply(Atom&)");

  TermList ts (a.args());
  apply(ts);
  if (ts == a.args()) { 
    return;
  }

  a = Atom(a.functor(),ts);
} 



void Substitution::apply (FormulaList& fs)
{
  TRACER ("Substitution::apply (FormulaList& fs)");

  if (fs.isEmpty()) {
    return;
  }
  
  Formula f (fs.head());
  apply(f);
  FormulaList rs (fs.tail());
  apply(rs);

  if (f == fs.head() && rs == fs.tail()) {
    return;
  }

  fs = FormulaList(f,rs);
} 



void Substitution::apply (Formula& f)
{
  TRACER("Substitution::apply (Formula& f)");

  switch (f.connective()) {
  case ATOM:
    {
      Atom a(f.atom());
      apply(a);
      if (! (a == f.atom())) {
	f = Formula(a);
      }
      return;
    }

  case AND:
  case OR:
    {
      FormulaList fs(f.args());
      apply(fs);
      if (! (fs == f.args())) {
	f = Formula(f.connective(),fs);
      }
      return;
    }

  case IMP:
  case IFF:
  case XOR:
    {
      Formula l(f.left());
      apply(l);
      Formula r(f.right());
      apply(r);
      if (! (l == f.left() && r == f.right())) {
	f = Formula(f.connective(),l,r);
      }
      return;
    }

  case NOT:
    {
      Formula arg(f.uarg());
      apply(arg);
      if (! (arg == f.uarg())) {
	f = Formula(NOT,arg);
      }
      return;
    }
  case FORALL:
  case EXISTS:
    {
      
      VarList xs (f.vars());
      Iterator<Var> vs (xs);
      while (vs.hasNext()) {
	Var v(vs.next());
	bind(v,Term(v));
      }
      Formula arg(f.qarg());
      apply(arg);
      if (! (arg == f.qarg())) {
	f = Formula(f.connective(),xs,arg);
      }
      
      vs.reset (xs);
      while (vs.hasNext()) {
	undoBinding(vs.next());
      }
      return;
    }
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 
