
/*
 * File SymCounter.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#include "Tracer.hpp"
#include "SymCounter.hpp"
#include "Problem.hpp"
#include "Unit.hpp"
#include "Signature.hpp"



SymCounter::SymCounter (Signature& sig)
  :
  _noOfPreds (sig.noOfPreds()),
  _noOfFuns (sig.noOfFuns()),
  _preds ( new Pred[_noOfPreds] ),
  _funs ( new Fun[_noOfFuns] )
{
  TRACER("SymCounter::SymCounter");

  if (! _preds || ! _funs) {
    ALLOCATION_FAILURE;
  }
  Signature::Iterator ps (sig._pbuckets);

  while (ps.hasNext()) {
    PredicateSymbol& p  = static_cast<PredicateSymbol&>(ps.next());
    _preds[p.number()].pred(p);
  }

  Signature::Iterator fs (sig._fbuckets);
  while ( fs.hasNext() ) {
    FunctionSymbol& f = static_cast<FunctionSymbol&>(fs.next());
    _funs[f.number()].fun(f);
  }
} 



SymCounter::~SymCounter ()
{
  TRACER("SymCounter::~SymCounter");

  delete [] _preds;
  delete [] _funs;
} 



void SymCounter::count (const Problem& p,int c)
{
  UnitChain::Iterator us ( p.units() );
  while ( us.hasNext() ) {
    count (us.next(),c);
  }
} 


void SymCounter::count (const Unit& u,int c)
{
  switch (u.unitType()) {
    case FORMULA:
      count (u.formula(), 1, c);
      return;

    case CLAUSE:
      count (u.clause(), c);
      return;
  }
} 


void SymCounter::count (const Clause& u,int add)
{
  Iterator<Literal> ls (u.literals());
  while ( ls.hasNext() ) {
    count (ls.next(), add);
  }
} 



void SymCounter::count (const Formula& f,int polarity,int add)
{
  switch (f.connective()) {
    case ATOM:
      count (f.atom(), polarity, add);
      return;

    case AND:
    case OR: {
      Iterator<Formula> fs (f.args());
      while (fs.hasNext()) {
        count (fs.next(), polarity, add);
      }
      return;
    }

    case IMP:
      count (f.left(), -polarity, add);
      count (f.right(), polarity, add);
      return;

    case NOT:
      count (f.uarg(), -polarity, add);
      return;

    case IFF:
    case XOR:
      count (f.left(), 0, add);
      count (f.right(), 0, add);
      return;

    case FORALL:
    case EXISTS:
      count (f.qarg(), polarity, add);
      return;

    default:
      ASS(false);
      return;
  }
} 



void SymCounter::count (const Literal& l,int add)
{
  count (l.atom(), l.isPositive() ? 1 : -1, add);
} 



void SymCounter::count (const Atom& a,int polarity,int add)
{
  get(a.functor())->add (polarity,add);

  Iterator<Term> ts (a.args());
  while ( ts.hasNext() ) {
    count (ts.next(), add);
  }
} 



void SymCounter::count (const Term& t,int add)
{
  switch (t.tag())
    {
    case VAR:
    case NUMERIC:
      return;

    case COMPOUND:
      {
	get(t.functor())->add (add);

	Iterator<Term> ts (t.args());
	while ( ts.hasNext() ) {
	  count (ts.next(), add);
	}
      }
    }
} 



void SymCounter::Pred::add (int polarity, int add)
{
  switch (polarity) {
    case -1:
      _nocc += add;
      return;

    case 0:
      _docc += add;
      return;

    case 1:
      _pocc += add;
      return;

    default:
      ASS(false);
      return;
  }
} 


