
/*
 * File TPTP.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "TPTP.hpp"
#include "Unit.hpp"
#include "Int.hpp"
#include "Problem.hpp"
#include "Refutation.hpp"


string TPTP::toString (const Atom& a)
{
  return a.functor().name() + a.args().toString(",");
} 



string TPTP::toString (const Clause& c)
{
  if (c.isEmpty()) {
    return "[]";
  }

  string result = string("[") + toString(c.literals().head());

  Iterator<Literal> lits (c.literals().tail());
  while (lits.hasNext()) {
    result += string(",\n  ") + toString(lits.next());
  }
  return result + "]";;
} 



string TPTP::toString (const Formula& f)
{
  static string names [] =
    { "", " & ", " | ", " => ", " <=> ", " <~> ", "~", "!", "?"};

  Connective c = f.connective();
  string con = names[(int)c];
  switch (c) {
  case ATOM:
    return toString(f.atom());

  case AND:
  case OR: 
    {
      string result = toString(f.args().head(),c);
      Iterator<Formula> arg (f.args().tail());
      while (arg.hasNext()) {
	result += con + toString(arg.next(),c);
      }
      return result;
    }

  case IMP:
  case IFF:
  case XOR:
    return toString(f.left(),c) + con + toString(f.right(),c);

  case NOT:
    return con + toString(f.uarg(),c);

  case FORALL:
  case EXISTS:
    {
      string result = con + "[" + f.vars().head().toString();
      Iterator<Var> vs(f.vars().tail());
      while (vs.hasNext()) {
	result += string(",") + vs.next().toString();
      }
      return result + "] : " + toString(f.qarg(),c);
    }

#if DEBUG_PREPRO
  default:
    ASS (false);
#endif
  }
} 


string TPTP::toString (const Literal& l)
{
  return string(l.isPositive() ? "++" : "--") + toString(l.atom());
} 


string TPTP::toString (const Unit& u)
{
  static string inputTypes[] = {"axiom","conjecture","hypothesis","derived"};

  string nm = u.name() != "" ? 
              u.name() : 
              string("u_") + Int::toString(u.number());

  switch (u.unitType()) {
  case CLAUSE:
    return string("input_clause(") + nm + "," + 
           inputTypes[u.inputType()] + ",\n  " +
           toString(u.clause()) + ").";
  case FORMULA:
    return string("input_formula(") + nm + "," + 
           inputTypes[u.inputType()] + ",\n  " +
           toString(u.formula()) + ").";
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 


string TPTP::toStringAsInference (const Unit& u)
{
  return toString(u);
} 



string TPTP::toString (const Formula& f, Connective outer)
{
  return f.parenthesesRequired(outer) ?
         string("(") + toString(f) + ")" :
         toString(f);
} 



string TPTP::toString (const Refutation& ref)
{
  string result = "%=========== Refutation ==========\n";

  Iterator<Unit> units(ref.refutation());
  while (units.hasNext()) {
    result += toStringAsInference(units.next());
  }

  return result + "%======= End of refutation =======\n";
} 



string TPTP::toString (const Problem& problem)
{
  string result("");

  UnitChain::Iterator us (problem.units());
  while (us.hasNext()) {
    result += toString(us.next()) + "\n\n";
  }

  return result;
} 
