
/*
 * File Tautology.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Tautology.hpp"
#include "Unit.hpp"
#include "Sort.hpp"
#include "Int.hpp"
#include "Tracer.hpp"
#include "Inference.hpp"



int Tautology::isTautology (const Unit& in, Unit& out)
{
  TRACER("Tautology::isTautology(const Unit&...");
  LiteralList newLiterals;

  int result = isTautology(in.clause().literals(),newLiterals);
  if (result != -1) {
    return result;
  }
  
  out = Unit(Clause(newLiterals),
	     new Inference(IR_REMOVE_DUPLICATE_LITERALS,in));
  return -1;
} 



int Tautology::isTautology (const LiteralList& in, LiteralList& out)
{
  TRACER("Tautology::isTautology(const LiteralList&...");

  int len = in.length();
  if (len == 0) {
    return 0;
  }

  
  bool removed = false;
  LiteralList result;

  
  
  Tautology comparator;
  Sort<Literal,Tautology> sorter(len,comparator);
  Iterator<Literal> lits(in);
  while (lits.hasNext()) {
    const Literal& lit = lits.next();
    const Atom& atm = lit.atom();
    if (atm.isEquality()) {
      const TermList& ts = atm.args();
      const Term& t1 = ts.head();
      const Term& t2 = ts.second();
      switch (compare(t1,t2)) {
      case -1:
	sorter.add(lit);
	break;

      case 0:
	if (lit.isPositive()) { 
	  return 1;
	}
	
	removed = true;
	len--;
	break;

      case 1: 
	sorter.add(Literal(lit.sign(),
			   Atom(atm.functor(),
				TermList(t2,TermList(t1)))));
	break;
      }
    }
    else {
      sorter.add(lit);
    }
  }

  sorter.sort();

  for (int i=len-2;i >= 0;i--) {
    const Literal& l1 = sorter[i];
    const Literal& l2 = sorter[i+1];
    if ( l1.atom().equals(l2.atom()) ) {
      if (l1.sign() != l2.sign()) {
	return 1;
      }
      removed = true;
    }
    else { 
      result.push(l2);
    }
  }
  if (removed) {
    if (len > 0) {
      out = LiteralList(sorter[0],result);
    }
    return -1;
  }
  return 0;
} 



bool Tautology::lessThan (const Literal& lit1,const Literal& lit2)
{
  TRACER("Tautology::lessThan");

  return compare(lit1.atom(),lit2.atom()) == -1;
} 



int Tautology::compare (const Atom& atom1,const Atom& atom2)
{
  TRACER("Tautology::compare (const Atom&...");

  switch (Int::compare(atom1.functor().number(),atom2.functor().number())) {
  case -1:
    return -1;
  case 0:
    break;
  case 1:
    return 1;
  }

  return compare(atom1.args(),atom2.args());
} 



int Tautology::compare (const Term& term1,const Term& term2)
{
  TRACER("Tautology::compare (const Term&...");

  switch (Int::compare(term1.tag(),term2.tag())) {
  case -1:
    return -1;
  case 0:
    break;
  case 1:
    return 1;
  }
  

  switch (term1.tag()) {
  case VAR:
    return Int::compare(term1.var().toInt(),term2.var().toInt());

  case NUMERIC:
    return Int::compare(term1.number(),term2.number());

  case COMPOUND:
    switch (Int::compare(term1.functor().number(),term2.functor().number())) {
    case -1:
      return -1;
    case 0:
      break;
    case 1:
      return 1;
    }
  }
  return compare(term1.args(),term2.args());
} 



int Tautology::compare (const TermList& terms1,const TermList& terms2)
{
  TRACER("Tautology::compare (const TermList&...");

  Iterator<Term> ts1(terms1);
  Iterator<Term> ts2(terms2);
  while (ts1.hasNext()) {
    ASS(ts2.hasNext());
    switch (compare(ts1.next(),ts2.next())) {
    case -1:
      return -1;
    case 0:
      break;
    case 1:
      return 1;
    }
  }

  ASS(! ts2.hasNext());
  return 0;
} 
