
/*
 * File Term.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include <string>
#include <cstdio>

#include "Tracer.hpp"
#include "Int.hpp"
#include "Term.hpp"
#include "XML.hpp"
#include "Signature.hpp"
#include "VarCounter.hpp"



TermList::TermList (const Term &hd)
  :
  Lst<Term> (hd)
{
} 


TermList::TermList (const Term &hd, const TermList& tl)
  :
  Lst<Term> (hd,tl)
{
} 


void Term::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;

  if (_counter == 0) {
    destroy ();
  }
} 


Term::~Term () 
{
  if (_data) {
    _data->deref ();
  }
} 



Term::Term (Var v)
  :
  _data (new VarData (v))
{
} 



Term::Term (double d)
  :
  _data (new NumericData (d))
{
} 



Term::VarData::VarData (Var v)
  : 
  Data (VAR), 
  _var (v) 
{
} 


Term::NumericData::NumericData (double value)
  : 
  Data (NUMERIC), 
  _number (value) 
{
} 


Term::CompoundData::CompoundData (const FunctionSymbol& f, 
				  const TermList& args) 
  : 
  Data (COMPOUND), 
  _functor (f), 
  _args (args) 
{
} 



Term::CompoundData::CompoundData (const FunctionSymbol& f) 
  : 
  Data (COMPOUND), 
  _functor (f) 
{
} 


void Term::operator = (const Term& t)
{
  if (t._data) {
    t._data->ref ();
  }

  if (_data) {
    _data->deref ();
  }

  _data = t._data;
} 



inline
void Term::Data::destroy () 
{ 
  ASS (this);
  ASS (_counter == 0);

  switch (tag()) 
    {
    case VAR:
      delete static_cast<VarData*>(this);
      return;

    case COMPOUND:
      delete static_cast<CompoundData*>(this);
      return;

    case NUMERIC:
      delete static_cast<NumericData*>(this);
      return;
    }
} 



bool TermList::isLinearSequenceOfVariables () const
{
  Iterator<Term> ts(*this);
  while (ts.hasNext()) {
    if (ts.next().tag() != VAR) {
      return false;
    }
  }
  
  VarCounter vs;
  ts.reset(*this);
  while (ts.hasNext()) {
    Var v(ts.next().var());
    if (vs.getCount(v) != 0) { 
      return false;
    }
    vs.incCount(v);
  }
  return true;
} 





bool Term::occurs (Var v) const
{
  TRACER ("Term::occurs (Var v)");

  switch (tag()) 
    {
    case COMPOUND:
      return args().occurs(v);

    case VAR:
      return var() == v;

    case NUMERIC:
      return false;

    #if DEBUG_PREPRO
    default:
      ASS(false);
    #endif
    }
} 




bool TermList::occurs (Var v) const
{
  TRACER ("TermList::occurs (Var v)");

  Iterator<Term> ss (*this);
  while ( ss.hasNext() ) {
    if ( ss.next().occurs(v) ) {
      return true;
    }
  }

  return false;
} 






bool Term::equals (const Term& t) const
{
  TRACER("Term::equal");

  if (tag() != t.tag()) {
    return false;
  }

  switch (tag()) 
    {
    case VAR:
      return var() == t.var();

    case COMPOUND:
      return functor() == t.functor() &&
             args().equals (t.args());

    case NUMERIC:
      return number() == t.number();

    #ifdef DEBUG_PREPRO
    default:
      ASS(false);
    #endif
    }
} 




bool TermList::equals (const TermList& ts) const
{
  TRACER ("TermList::equals");

  Iterator<Term> ss1(*this);
  Iterator<Term> ss2 (ts);
  while ( ss1.hasNext() ) {
    ASS ( ss2.hasNext() );
    
    if ( ! ss1.next().equals(ss2.next()) ) {
      return false;
    }
  }
  ASS ( ! ss2.hasNext() );

  return true;
} 






bool TermList::hasVarsNotIn (const VarListList& vs) const
{
  Iterator<Term> terms (*this);
  while ( terms.hasNext() ) {
    if (terms.next().hasVarsNotIn (vs)) {
      return true;
    }
  }

  return false;
} 



bool Term::hasVarsNotIn (const VarListList& vs) const
{
  switch (tag()) 
    {
    case VAR: {
      
      Var v = var ();

      Iterator<VarList> vss (vs);
      while ( vss.hasNext() ) {
	if ( vss.next().member(v) ) { 
	  return false;
	}
      }
      
      return true;
    }

    case NUMERIC:
      return false;

    case COMPOUND:
      return args().hasVarsNotIn (vs);

    #ifdef DEBUG_PREPRO
    default:
      ASS (false);
    #endif
    }
} 



XML::Element* Term::toXML() const
{
  switch (tag()) {
  case VAR:
    return var().toXML();

  case COMPOUND: 
    {
      if (args().isEmpty()) {
	XML::Element* t = new XML::EmptyElement("constant");
	t->addAttribute(new XML::TextAttribute("name",functor().name()));
	return t;
      }
      
      XML::DeepElement* t = new XML::DeepElement("term");
      t->addAttribute(new XML::TextAttribute("functor",functor().name()));
      Iterator<Term> ts(args());
      while ( ts.hasNext() ) {
	t->addChild(ts.next().toXML());
      }
      return t;
    }

  case NUMERIC:
    {
      XML::Element* t = new XML::EmptyElement("numeric_term");
      t->addAttribute(new XML::DoubleAttribute("value",number()));
      return t;
    }
#ifdef DEBUG_PREPRO
  default:
    ASS(false);
    return 0;
#endif
  }
} 



string Term::toString () const
{
  switch (tag()) {
  case VAR:
    return var().toString();

  case COMPOUND:
    return functor().name() + args().toString(",");

  case NUMERIC:
    return Int::toString(number());

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
}



string TermList::toString (string separator) const
{
  if (isEmpty()) {
    return "";
  }
  string result = string("(") + head().toString();
  Iterator<Term> ts (tail());
  while (ts.hasNext()) {
    result += separator + ts.next().toString();
  }
  return result + ")";
} 



TermList::TermList (const VarList& vs)
  : Lst<Term> () 
{
  if (vs.isEmpty()) {
    return;
  }
  TermList tl (vs.tail());

  *this = TermList(Term(vs.head()),tl);
} 




