
/*
 * File Term.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __Term__
#define __Term__


class TermList;
class FunctionSymbol;
class Substitution;

namespace XML {
  class Element;
};


#include <string>
#include "Var.hpp"




enum VarType {
  VAR = 0,        
  NUMERIC = 1,
  COMPOUND = 2 
};



class Term 
{
 private:
  class Data;
  class VarData;
  class CompoundData;
  class NumericData;

 public:
  
  Term () : _data(0) {}
  Term (const FunctionSymbol& f, const TermList& args); 
  Term (const FunctionSymbol& f);                       
  Term (const Term& t);
  ~Term ();
  explicit Term (Var v);
  explicit Term (double number);
  void operator= (const Term& rhs);

  
  VarType tag () const;
  const Var& var () const;
  double number () const;
  const FunctionSymbol& functor () const;
  const TermList& args () const;
  bool operator == (const Term& rhs) const;
  bool operator != (const Term& rhs) const;

  
  void* operator new (size_t);

  
  XML::Element* toXML () const;
  bool occurs (Var v) const;
  bool equals (const Term& t) const;
  bool hasVarsNotIn (const VarListList&) const;

  
  string toString () const;
 private:
  
  Data* _data;
}; 



class TermList 
  : public Lst<Term>
{
 public:
  
  TermList () : Lst<Term> () {}
  
  TermList (const TermList& ts) : Lst<Term> (ts) {}
  explicit TermList (const Term& t);
  TermList (const Term& head, const TermList& tail);
  
  explicit TermList (LstData<Term>* d) : Lst<Term> (d) {}
  explicit TermList (const VarList&);

  
  const TermList& tail () const
    { return static_cast<const TermList&>(Lst<Term>::tail()); }

  
  bool occurs (Var v) const;
  bool equals (const TermList& ts) const;
  bool hasVarsNotIn (const VarListList& ) const;
  bool isLinearSequenceOfVariables () const;

  
  string toString (string separator) const;
}; 


class Term::Data 
{
private:
  Data ();
public:
  explicit Data (VarType tag);
  ~Data () {}

  
  VarType tag () const { return _tag; }
  void ref () { _counter++; }

  void deref ();
  void destroy ();

  
private:
  void* operator new (size_t);
 protected:
  
  int _counter;
  VarType _tag;
}; 


class Term::VarData 
: public Term::Data
{
 public:
  explicit VarData (Var v);
  
  const Var& var () const {ASS(_tag == VAR); return _var;}

  
  CLASS_NAME("Term::VarData");
  USE_ALLOCATOR(VarData);
 private:
  
  Var _var;
}; 


class Term::CompoundData 
: public Term::Data
{
 public:
  CompoundData (const FunctionSymbol& f, const TermList& args);
  explicit CompoundData (const FunctionSymbol& f);
  
  const FunctionSymbol& functor () const {return _functor;}
  
  const TermList& args () const {return _args;}

  
  CLASS_NAME("Term::CompoundData");
  USE_ALLOCATOR(CompoundData);
 private:
  
  const FunctionSymbol& _functor;
  TermList _args;
}; 


class Term::NumericData 
: public Term::Data
{
 public:
  explicit NumericData (double f);
  
  double number () const { return _number; }

  
  CLASS_NAME("Term::NumericData");
  USE_ALLOCATOR(NumericData);
 private:
  
  double _number;
}; 





inline
Term::Term (const Term& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 



inline
Term::Term (const FunctionSymbol& f, const TermList& args)
  :
  _data (new CompoundData (f,args))
{
} 



inline
Term::Term (const FunctionSymbol& f)
  :
  _data (new CompoundData (f))
{
} 


inline
VarType Term::tag () const 
{ 
  return _data->tag(); 
} 



inline
bool Term::operator == (const Term& rhs) const
{ 
  ASS (_data && rhs._data);

  return _data == rhs._data; 
} 



inline
bool Term::operator != (const Term& rhs) const
{ 
  ASS (_data && rhs._data);

  return _data != rhs._data; 
} 



inline
const Var& Term::var () const
{ 
  ASS (this && tag() == VAR);

  return (static_cast<Term::VarData*>(_data))->var();
} 



inline
const FunctionSymbol& Term::functor () const
{ 
  ASS (this && tag() == COMPOUND);

  return (static_cast<Term::CompoundData*>(_data))->functor();
} 



inline
const TermList& Term::args () const
{ 
  ASS (this && tag() == COMPOUND);

  return (static_cast<Term::CompoundData*>(_data))->args();
} 



inline
double Term::number () const
{ 
  ASS (this && tag() == NUMERIC);

  return (static_cast<Term::NumericData*>(_data))->number();
} 





inline
Term::Data::Data (VarType tag)
  : 
  _counter (1),
  _tag (tag)
{
} 


#endif 

