
/*
 * File TheoryFinder.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Problem.hpp"
#include "TheoryFinder.hpp"
#include "Unit.hpp"


class TheoryFinder::MatcherState 
{
public:
  MatcherState (char* code,int* vars,const Symbol** sym);
  int match (int cp,const Term&);
  int match (int cp,const TermList&);
  bool addNewVar (int index,int var);
  bool addNewSymbol (int index,const Symbol& symb);

  char* code;
  int* vars;
  const Symbol** sym;
};


class TheoryFinder::Matcher 
{
public:
  
  Matcher (MatcherState& state,int cp) 
    : _state(state), _cp(cp)
  {}
  virtual int nextMatch() = 0;
  virtual ~Matcher () {}
protected:
  MatcherState& _state;
  int _cp;
};


class TheoryFinder::AtomMatcher 
  : public Matcher
{
public:
  AtomMatcher (MatcherState& state,int cp,const Atom& atom);
  int nextMatch();
  ~AtomMatcher() {}
private:
  const Atom& _atom;
  int _attempts;
};


class TheoryFinder::FormulaMatcher 
  : public Matcher
{
public:
  FormulaMatcher (MatcherState& state,int cp,const Formula& formula);
  int nextMatch();
  ~FormulaMatcher();
private:
  Matcher* _delegate;
};


class TheoryFinder::FormulaPairMatcher 
  : public Matcher
{
public:
  FormulaPairMatcher (MatcherState& state,
		      int cp,
		      const Formula& formula1,
		      const Formula& formula2);
  int nextMatch();
  ~FormulaPairMatcher();
private:
  const Formula& _f1;
  const Formula& _f2;
  Matcher* _fm1;
  Matcher* _fm2;
}; 


class TheoryFinder::QuantifiedFormulaMatcher 
  : public Matcher
{
public:
  QuantifiedFormulaMatcher (MatcherState& state,
			    int cp,
			    const Formula& formula);
  int nextMatch();
  ~QuantifiedFormulaMatcher();
private:
  const Formula& _formula;
  Matcher* _delegate;
}; 


class TheoryFinder::LiteralListMatcher 
  : public Matcher
{
public:
  LiteralListMatcher (MatcherState& state,int cp,const LiteralList& lits);
  int nextMatch();
  ~LiteralListMatcher();
private:
  LiteralList _reserve;
  LiteralList _tried;
  AtomMatcher* _headMatcher;
  LiteralListMatcher* _tailMatcher;
};


class TheoryFinder::FormulaListMatcher 
  : public Matcher
{
public:
  FormulaListMatcher (MatcherState& state,int cp,const FormulaList& lits);
  int nextMatch();
  ~FormulaListMatcher();
private:
  FormulaList _reserve;
  FormulaList _tried;
  FormulaMatcher* _headMatcher;
  FormulaListMatcher* _tailMatcher;
}; 


class TheoryFinder::ClauseMatcher 
  : public Matcher
{
public:
  ClauseMatcher (MatcherState& state,const Clause& clause);
  int nextMatch();
  ~ClauseMatcher();
private:
  LiteralListMatcher* _delegate;
}; 

char TheoryFinder::_commutativityAtom[] = 
  {CODE_EQL, 
    CODE_NEWSYM,0,2,CODE_NEWVAR,0,CODE_NEWVAR,1,  
    CODE_OLDSYM,0,CODE_OLDVAR,1,CODE_OLDVAR,0};   


TheoryFinder::TheoryFinder (const Problem& problem)
  : _problem(problem),
    _records (0)
{
} 


TheoryFinder::~TheoryFinder ()
{
  while (_records) {
    Record* next = _records->next;
    delete _records;
    _records = next;
  }
} 



int TheoryFinder::search ()
{
  int found = 0;
  UnitChain::Iterator units(_problem.units());
  while (units.hasNext()) {
    if (matchAll(units.next())) {
      found++;
    }
  }
  return found;
} 



bool TheoryFinder::matchAll (const Unit& u)
{
  
  Unit unit(u);
  _currentUnit = &unit;

  switch (unit.unitType()) {
  case CLAUSE:
    return matchAll(unit.clause());

  case FORMULA:
    return matchAll(unit.formula());

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 


bool TheoryFinder::matchAll (const Clause& clause)
{
  return 
          matchExtensionality(clause) ||
          matchSubset1(clause);
} 


bool TheoryFinder::matchAll (const Formula& formula)
{
  Formula f(formula);
  stripForall(f);

  return 
         matchExtensionality(f) ||
         matchSubset(f);
} 


bool TheoryFinder::matchC (const Formula& f)
{
  return f.connective() == ATOM &&
         matchC(f.atom());
} 


bool TheoryFinder::matchC (const Clause& clause)
{
  LiteralList ls(clause.literals());
  if (! ls.hasLength(1)) {
    return false;
  }
  const Literal& l = ls.head();
         
  return l.isPositive() &&
         matchC(l.atom());
} 



void TheoryFinder::stripForall (Formula& f)
{
  while (f.connective() == FORALL) {
    f = f.qarg();
  }
} 



bool TheoryFinder::matchC (const Atom& atom)
{
  int vars[2];
  const Symbol* sym[1];

  MatcherState state(_commutativityAtom,vars,sym);
  AtomMatcher am(state,0,atom);
  if (am.nextMatch()) {  
    Record* r = new Record(AX_C,*_currentUnit,sym[0],_records);
    _records = r;
    return true;
  }
  return false;
} 



TheoryFinder::Record::Record (Axiom knd,Unit& u,const Symbol* s,Record* rec)
  : kind(knd),
    unit(u),
    symbol1(s),
    next(rec)
{
} 


TheoryFinder::Record::Record (Axiom knd,Unit& u,
			      const Symbol* s1,
			      const Symbol* s2,
			      Record* rec)
  : kind(knd),
    unit(u),
    symbol1(s1),
    symbol2(s2),
    next(rec)
{
} 


TheoryFinder::Record::Record (Axiom knd,Unit& u,
			      const Symbol* s1,
			      const Symbol* s2,
			      const Symbol* s3,
			      Record* rec)
  : kind(knd),
    unit(u),
    symbol1(s1),
    symbol2(s2),
    symbol3(s3),
    next(rec)
{
} 


TheoryFinder::MatcherState::MatcherState (char* c,int* vs,const Symbol** ss)
  : code(c),
    vars(vs),
    sym(ss)
{
}


int TheoryFinder::AtomMatcher::nextMatch ()
{
  char* code = _state.code+_cp;
  for (;;) {
    switch (_attempts) {
    case 0:
      switch (code[0]) {
      case CODE_EQL:
	if (_atom.isEquality()) {
	  _attempts ++;
	  int newCp = _state.match(_cp+1,_atom.args());
	  if (newCp) {
	    return newCp;
	  }
	}
	else {
	  return 0;
	}
	break;

      case CODE_NEWSYM:
	if (_atom.isEquality()) {
	  return 0;
	}
	if (_atom.functor().arity() != code[2]) {
	  return 0;
	}
	_attempts++;
	if (!_state.addNewSymbol(code[1],_atom.functor())) {
	  return 0;
	}
	return _state.match(_cp+3,_atom.args());
    
      case CODE_OLDSYM:
	if (_state.sym[code[1]] != &_atom.functor()) {
	  return 0;
	}
	_attempts ++;
	return _state.match(_cp+2,_atom.args());

      default:
	return false;
      }

    case 1:
      if (code[0] == CODE_EQL) {
	int newCp = _state.match(_cp+1,_atom.args().second());
	if (! newCp) {
	  return 0;
	}
	_attempts++;
	return _state.match(newCp,_atom.args().head());
      }
      
      return 0;

    default:
      
      ASS(_attempts == 2);
      return 0;
    }
  }
} 


int TheoryFinder::MatcherState::match (int cp,const Term& t)
{
  char* cd = code + cp;
  switch (cd[0]) {
  case CODE_NEWVAR:
    if (t.tag() != VAR) {
      return 0;
    }
    if (! addNewVar(cd[1],t.var().toInt())) {
      return 0;
    }
    return cp+2;

  case CODE_OLDVAR:
    if (t.tag() != VAR) {
      return 0;
    }
    if (vars[cd[1]] != t.var().toInt()) {
      return 0;
    }
    return cp+2;

  case CODE_NEWSYM:
    if (t.tag() != COMPOUND) {
      return 0;
    }
    if (t.args().length() != cd[2]) {
      return 0;
    }
    if (!addNewSymbol(cd[1],t.functor())) {
      return 0;
    }
    return match(cp+3,t.args());

  case CODE_OLDSYM:
    if (t.tag() != COMPOUND) {
      return 0;
    }
    if (sym[cd[1]] != &t.functor()) {
      return 0;
    }
    return match(cp+2,t.args());

  default:
    return false;
  }
} 


int TheoryFinder::MatcherState::match (int cp,const TermList& ts)
{
  TermList terms(ts);
  while (terms.isNonEmpty()) {
    cp = match(cp,terms.head());
    if (! cp) {
      return 0;
    }
    terms.pop();
  }
  
  return cp;
} 



TheoryFinder::AtomMatcher::AtomMatcher (MatcherState& state,
					int cp,
					const Atom& atom)
  : Matcher(state,cp),
    _atom(atom),
    _attempts(0)
{
} 


bool TheoryFinder::MatcherState::addNewSymbol (int index,const Symbol& s)
{
  for (int i = index-1;i >=0;i--) {
    if (sym[i] == &s) {
      return false;
    }
  }
  sym[index] = &s;
  return true;
} 



bool TheoryFinder::MatcherState::addNewVar (int index,int var)
{
  for (int i = index-1;i >=0;i--) {
    if (vars[i] == var) {
      return false;
    }
  }
  vars[index] = var;
  return true;
} 


bool TheoryFinder::matchExtensionality (const Clause& c)
{
  static char code[] =
  {3,                                
    CODE_NEWSYM,0,2,                 
     CODE_NEWSYM,1,2,CODE_NEWVAR,0,CODE_NEWVAR,1,CODE_OLDVAR,0,
    CODE_OLDSYM,0,                   
     CODE_OLDSYM,1,CODE_OLDVAR,0,CODE_OLDVAR,1,CODE_OLDVAR,1,
    CODE_EQL,CODE_OLDVAR,0,CODE_OLDVAR,1}; 

  int vars[2];
  const Symbol* sym[2];

  MatcherState state(code,vars,sym);
  ClauseMatcher matcher(state,c);
  if (matcher.nextMatch()) {
    Record* r = new Record(AX_EXTENSIONALITY,*_currentUnit,sym[0],sym[1],_records);
    _records = r;
    return true;
  }
  return false;
} 


bool TheoryFinder::matchSubset1 (const Clause& c)
{
  static char code[] =
  {2,                                
    CODE_NEWSYM,0,2,                 
     CODE_NEWSYM,1,2,CODE_NEWVAR,0,CODE_NEWVAR,1,CODE_OLDVAR,0,
    CODE_NEWSYM,2,2,CODE_OLDVAR,0,CODE_OLDVAR,1}; 

  int vars[2];
  const Symbol* sym[3];

  MatcherState state(code,vars,sym);
  ClauseMatcher matcher(state,c);
  if (matcher.nextMatch()) {
    Record* r = new Record(AX_SUBSET1,*_currentUnit,sym[0],sym[1],sym[2],_records);
    _records = r;
    return true;
  }
  return false;
} 


TheoryFinder::ClauseMatcher::ClauseMatcher (MatcherState& state,
					    const Clause& clause)
  : Matcher(state,0)
{
  _delegate = clause.literals().hasLength(state.code[0]) ?
              new LiteralListMatcher(state,1,clause.literals()) : 
              0;
} 


int TheoryFinder::ClauseMatcher::nextMatch ()
{
  if (_delegate) {
    return _delegate->nextMatch();
  }
  return 0;
} 


TheoryFinder::ClauseMatcher::~ClauseMatcher ()
{
  delete _delegate;
} 


TheoryFinder::LiteralListMatcher::LiteralListMatcher (MatcherState& state,
						      int cp,
						      const LiteralList& lits)
  : Matcher(state,cp),
    _reserve(lits),
    _headMatcher(0),
    _tailMatcher(0)
{
} 


TheoryFinder::FormulaListMatcher::FormulaListMatcher (MatcherState& state,
						      int cp,
						      const FormulaList& fs)
  : Matcher(state,cp),
    _reserve(fs),
    _headMatcher(0),
    _tailMatcher(0)
{
} 


TheoryFinder::LiteralListMatcher::~LiteralListMatcher ()
{
  delete _headMatcher;
  delete _tailMatcher;
} 


TheoryFinder::FormulaListMatcher::~FormulaListMatcher ()
{
  delete _headMatcher;
  delete _tailMatcher;
} 


int TheoryFinder::LiteralListMatcher::nextMatch ()
{
  for (;;) {
    if (_tailMatcher) {
      int cp = _tailMatcher->nextMatch();
      if (cp) {
	return cp;
      }
      delete _tailMatcher;
      _tailMatcher = 0;
    }
    
    else if (_headMatcher) {
      int cp = _headMatcher->nextMatch();
      if (cp) {
	if (_tried.isEmpty() && _reserve.tail().isEmpty()) {
	  return cp;
	}
	
	LiteralList remaining(_tried);
	remaining.append(_reserve.tail());
	_tailMatcher = new LiteralListMatcher(_state,cp,remaining);
      }
      else { 
	delete _headMatcher;
	_headMatcher = 0;
	_tried.push(_reserve.head());
	_reserve.pop();
      }
    }
    
    else if (_reserve.isEmpty()) {
      return 0;
    }
    
    else if (_state.code[_cp] == CODE_NOT) { 
      const Literal& hd = _reserve.head();
      if (hd.isPositive()) { 
	_tried.push(_reserve.head());
	_reserve.pop();
      }
      else {
	_headMatcher = new AtomMatcher(_state,_cp+1,hd.atom());
      }
    }
    else { 
      const Literal& hd = _reserve.head();
      if (hd.isNegative()) { 
	_tried.push(_reserve.head());
	_reserve.pop();
      }
      else {
	_headMatcher = new AtomMatcher(_state,_cp,hd.atom());
      }
    }
  }
} 


int TheoryFinder::FormulaListMatcher::nextMatch ()
{
  for (;;) {
    if (_tailMatcher) {
      int cp = _tailMatcher->nextMatch();
      if (cp) {
	return cp;
      }
      delete _tailMatcher;
      _tailMatcher = 0;
    }
    
    else if (_headMatcher) {
      int cp = _headMatcher->nextMatch();
      if (cp) {
	if (_tried.isEmpty() && _reserve.tail().isEmpty()) {
	  return cp;
	}
	
	FormulaList remaining(_tried);
	remaining.append(_reserve.tail());
	_tailMatcher = new FormulaListMatcher(_state,cp,remaining);
      }
      else { 
	delete _headMatcher;
	_headMatcher = 0;
	_tried.push(_reserve.head());
	_reserve.pop();
      }
    }
    
    else if (_reserve.isEmpty()) {
      return 0;
    }
    
    else { 
      _headMatcher = new FormulaMatcher(_state,_cp,_reserve.head());
    }
  }
} 


bool TheoryFinder::matchSubset (const Formula& f)
{
  static char code[] =
    {CODE_IFF,                                     
      CODE_NEWSYM,0,2,CODE_NEWVAR,0,CODE_NEWVAR,1, 
      CODE_FORALL,2,1,CODE_IMP,                    
       CODE_NEWSYM,1,2,CODE_NEWVAR,2,CODE_OLDVAR,0,
       CODE_OLDSYM,1,CODE_OLDVAR,2,CODE_OLDVAR,1}; 

  int vars[3];
  const Symbol* sym[2];

  MatcherState state(code,vars,sym);
  FormulaMatcher matcher(state,0,f);
  if (matcher.nextMatch()) {
    Record* r = new Record(AX_SUBSET,*_currentUnit,sym[0],sym[1],_records);
    _records = r;
    return true;
  }
  return false;
} 



TheoryFinder::FormulaMatcher::FormulaMatcher (MatcherState& state,
					      int cp,
					      const Formula& f)
  : Matcher(state,cp)
{
  Formula formula(f);

  for (;;) {
    
    switch (state.code[cp]) {
    case CODE_AND:
      if (formula.connective() == AND) {
	_delegate = new FormulaListMatcher(state,cp+1,formula.args());
      }
      else {
	_delegate = 0;
      }
      return;

    case CODE_OR:
      if (formula.connective() == OR) {
	_delegate = new FormulaListMatcher(state,cp+1,formula.args());
      }
      else {
	_delegate = 0;
      }
      return;

    case CODE_IMP:
      if (formula.connective() == IMP) {
	_delegate = new FormulaPairMatcher(state,
					   cp+1,
					   formula.left(),formula.right());
      }
      else {
	_delegate = 0;
      }
      return;

    case CODE_NOT:
      if (formula.connective() != NOT) {
	_delegate = 0;
	return;
      }
      formula = formula.uarg();
      cp++;
      break;

    case CODE_IFF:
      if (formula.connective() == IFF) {
	FormulaList args(formula.right());
	args.push(formula.left());
	_delegate = new FormulaListMatcher(state,cp+1,args);
      }
      else {
	_delegate = 0;
      }
      return;

    case CODE_XOR:
      if (formula.connective() == XOR) {
	FormulaList args(formula.right());
	args.push(formula.left());
	_delegate = new FormulaListMatcher(state,cp+1,args);
      }
      else {
	_delegate = 0;
      }
      return;

    case CODE_FORALL:
      if (formula.connective() == FORALL &&
	  formula.vars().hasLength(state.code[cp+2])) {
	_delegate = new QuantifiedFormulaMatcher(state,cp+1,formula);
      }
      else {
	_delegate = 0;
      }
      return;

    case CODE_EXISTS:
      if (formula.connective() == EXISTS &&
	  formula.vars().hasLength(state.code[cp+2])) {
	_delegate = new QuantifiedFormulaMatcher(state,cp+1,formula);
      }
      else {
	_delegate = 0;
      }
      return;

    default: 
      if (formula.connective() == ATOM) {
	_delegate = new AtomMatcher(state,cp,formula.atom());
      }
      else {
	_delegate = 0;
      }
      return;
    }
  }
} 



TheoryFinder::FormulaMatcher::~FormulaMatcher ()
{
  delete _delegate;
} 


int TheoryFinder::FormulaMatcher::nextMatch ()
{
  if (_delegate) {
    return _delegate->nextMatch();
  }
  return 0;
} 


TheoryFinder::QuantifiedFormulaMatcher::~QuantifiedFormulaMatcher ()
{
  delete _delegate;
} 



TheoryFinder::QuantifiedFormulaMatcher::QuantifiedFormulaMatcher (MatcherState& state,
								  int cp,
								  const Formula& formula)
  : Matcher(state,cp),
    _formula(formula)
{
  _delegate = new FormulaMatcher(state,cp+2,formula.qarg());
} 



int TheoryFinder::QuantifiedFormulaMatcher::nextMatch ()
{
  int cp = _delegate->nextMatch();
  if (! cp) {
    return 0;
  }
  
  
  
  int firstVar = _state.code[_cp];
  int lastVar = firstVar + _state.code[_cp+1];
  for (int v = firstVar;v < lastVar;v++) {
    Var w(_state.vars[v]);
    if (! _formula.vars().member(w)) {
      return 0;
    }
  }
  return cp;
} 



TheoryFinder::FormulaPairMatcher::FormulaPairMatcher (MatcherState& state,
						      int cp,
						      const Formula& f1,
						      const Formula& f2)
  : Matcher(state,cp),
    _f1(f1),
    _f2(f2),
    _fm1(0),
    _fm2(0)
{
} 



TheoryFinder::FormulaPairMatcher::~FormulaPairMatcher ()
{
  delete _fm1;
  delete _fm2;
} 



int TheoryFinder::FormulaPairMatcher::nextMatch ()
{
  for (;;) {
    if (_fm2) {
      int cp = _fm2->nextMatch();
      if (cp) {
	return cp;
      }
      delete _fm2;
      _fm2 = 0;
    }
    
    else if (_fm1) {
      int cp = _fm1->nextMatch();
      if (! cp) {
	return 0;
      }
      _fm2 = new FormulaMatcher(_state,cp,_f2);
    }
    
    else {
      _fm1 = new FormulaMatcher(_state,_cp,_f1);
    }
  }
} 




bool TheoryFinder::matchExtensionality (const Formula& f)
{
  static char code[] =
    {CODE_IMP,                                     
      CODE_FORALL,0,1,CODE_IFF,                    
      CODE_NEWSYM,0,2,CODE_NEWVAR,0,CODE_NEWVAR,1, 
       CODE_NEWSYM,1,2,CODE_NEWVAR,2,CODE_OLDVAR,0,
       CODE_OLDSYM,1,CODE_OLDVAR,2,CODE_OLDVAR,1}; 

  int vars[3];
  const Symbol* sym[1];

  MatcherState state(code,vars,sym);
  for (int i = 1;i >= 0;i--) {
    code[0] = i ? CODE_IMP : CODE_IFF;
    FormulaMatcher matcher(state,0,f);
    if (matcher.nextMatch()) {
      Record* r = new Record(AX_EXTENSIONALITY,*_currentUnit,sym[0],_records);
      _records = r;
      return true;
    }
  }
  return false;
} 


