
/*
 * File Unit.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __unit__
#define __unit__


#include "Chain.hpp"
#include "Clause.hpp"
#include "Formula.hpp"
#include "InferenceRule.hpp"
#include "VampireKernel.hpp"

class Options;
class Clause;
class Problem;
class UnitList;
class Inference;



enum InputType {
  
  AXIOM = 0,
  
  ASSUMPTION = 1,
  
  CONJECTURE = 2
};


enum UnitType {
  CLAUSE,
  FORMULA
};


enum DefType {
  DT_NONE,
  DT_PRED,
  DT_FUN
};





class Unit {
 private:
  class Data;

 public:
  
  Unit ();
  Unit (const Unit& a);
  Unit (const string& name, InputType untype, const Formula&);
  Unit (const string& name, InputType untype, const Clause&);
  Unit (InferenceRule, const Formula&, const Unit& parent);
  Unit (const Formula&, Inference*);
  Unit (const Clause&, Inference*);
  Unit (InferenceRule, const Clause&, const Unit& parent);
  Unit (const Clause&, 
	const UnitList& parents, 
	long number,
	VampireKernel::Clause::BackgroundFlagIterator&);
  Unit (const Formula&, const UnitList& parents);
  explicit Unit (void*); 
  ~Unit ();
  void operator = (const Unit& rhs);

  
  bool operator == (const Unit& rhs) const;
  bool operator != (const Unit& rhs) const;
  UnitType unitType () const;
  InputType inputType () const;
  DefType defType () const;
  void setDefType (DefType);
  const Formula& formula () const;
  const Clause& clause () const;
  void setFormula (const Formula& f);
  const string& name () const;
  Inference* inference () const;
  void setInference (Inference* inf);
  long number () const;
  int numberInRefutation () const;
  void setNumberInRefutation (int newNumber);
  int mark () const;
  void setMark (int mark);
  
  void* giveAway () { return _data; }

  
  void* operator new (size_t);

  
  XML::Element* toXML () const;
  static string toString (InputType t);
  static string toString (UnitType t);
  static string toString (DefType dt);

  
  string toString () const;
  string toStringAsInference () const;

  
  static long nextNumber () { return ++_lastNumber; }
  
  static void setLastNumber (long n) { _lastNumber = n; }
  static InputType max (InputType it1,InputType it2);
 private:
  
  Data* _data;
  static long _lastNumber;
  friend class Data;
}; 


class UnitList 
: public Lst<Unit>
{
 public:
  
  UnitList ();
  UnitList (const UnitList&);
  explicit UnitList (const Unit &hd);
  UnitList (const Unit& head, const UnitList& tail);
  explicit UnitList (LstData<Unit>*);
  UnitList (InputType tp, const FormulaList& formulas);

  
  const UnitList& tail () const
    { return static_cast<const UnitList&>(Lst<Unit>::tail()); }
  InputType maxInputType() const;
}; 


class Unit::Data
{
 public:
  Data (const string& name, InputType untype, const Formula&);
  Data (const string& name, InputType untype, const Clause&);
  Data (InferenceRule, const Formula&, const Unit& parent);
  Data (InferenceRule, const Clause&, const Unit& parent);
  Data (const Clause&, 
	const UnitList& parents, 
	long number,
	VampireKernel::Clause::BackgroundFlagIterator&);
  Data (const Formula&, const UnitList& parents);
  Data (const Formula&, Inference*);
  Data (const Clause&, Inference*);
  ~Data ();

  void ref () { ASS (this); _counter++; }
  void deref ();

  
  UnitType unitType () const { return _untype; }
  
  InputType inputType () const { return _intype; }
  
  DefType defType () const { return _defType; }
  
  void setDefType (DefType dt) { _defType = dt; }
  
  const Formula& formula () const { return _formula; }
  
  const Clause& clause () const { return _clause; }
  
  void setFormula (const Formula& f) { _formula = f; }
  
  const string& name () const { return _name; }
  
  Inference* inference () const { return _inference; }
  
  void setInference (Inference* inf) { _inference = inf; }
  
  long number () const { return _number; }
  
  int numberInRefutation () const { return _numberInRefutation; }
  
  int mark () const { return _mark; }
  
  void setMark (int mark) { _mark = mark; }
  
  void setNumberInRefutation (int newNumber)
  {
    ASS(newNumber > 0);
    _numberInRefutation = newNumber;
  } 

  
  CLASS_NAME("Unit::Data");
  USE_ALLOCATOR(Data);
 private:
  
  int _counter;
  UnitType _untype;
  InputType _intype;
  
  long _number;
  
  int _numberInRefutation;
  
  string _name;
  
  Clause _clause;
  
  Formula _formula;
  
  DefType _defType;
  
  Inference* _inference;
  
  int _mark;
}; 


typedef Chain<Unit> UnitChain;




inline
Unit::Unit () 
  : 
  _data (0) 
{
} 


inline
Unit::~Unit () 
{
  if (_data) {
    _data->deref ();
  }
} 



inline
Unit::Unit (const Unit& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 



inline
Unit::Unit (void* d)
  :
  _data (reinterpret_cast<Data*>(d))
{
  _data->ref();
} 



inline
Unit::Unit (const Formula& f, Inference* inf)
  : 
  _data (new Data(f, inf))
{
} 



inline
Unit::Unit (const Clause& c, Inference* inf)
  : 
  _data (new Data(c, inf))
{
} 



inline
Unit::Unit (const string& name, InputType untype, const Formula& formula) 
  : 
  _data (new Data(name, untype, formula))
{
} 



inline
Unit::Unit (const string& name, InputType untype, const Clause& clause) 
  : 
  _data (new Data(name, untype, clause)) 
{
} 



inline
Unit::Unit (InferenceRule ir, const Formula& f, const Unit& parent) 
  : 
  _data (new Data(ir, f, parent))
{
} 



inline
Unit::Unit (InferenceRule ir, const Clause& clause, const Unit& parent) 
  : 
  _data (new Data(ir, clause, parent)) 
{
} 


inline
Unit::Unit (const Clause& clause,
	    const UnitList& parents,
	    long number,
	    VampireKernel::Clause::BackgroundFlagIterator& flags)
  : 
  _data (new Data(clause, parents, number,flags)) 
{
} 


inline
Unit::Unit (const Formula& formula, const UnitList& parents) 
  : 
  _data (new Data(formula, parents)) 
{
} 


inline
bool Unit::operator == (const Unit& rhs) const
{ 
  return _data == rhs._data; 
} 


inline
bool Unit::operator != (const Unit& rhs) const
{ 
  return _data != rhs._data; 
} 



inline
UnitType Unit::unitType () const
{
  return _data->unitType ();
} 


inline
const Formula& Unit::formula () const
{
  ASS (unitType() == FORMULA);

  return _data->formula();
} 


inline
void Unit::setFormula (const Formula& f)
{
  _data->setFormula(f);
} 



inline
DefType Unit::defType () const
{
  return _data->defType ();
} 



inline
void Unit::setDefType (DefType dt)
{
  _data->setDefType (dt);
} 


inline
const Clause& Unit::clause () const
{
  ASS (unitType() == CLAUSE);

  return _data->clause();
} 



inline
InputType Unit::inputType () const
{
  return _data->inputType();
} 


inline
long Unit::number () const
{
  return _data->number ();
} 



inline
int Unit::numberInRefutation () const
{
  return _data->numberInRefutation ();
} 


inline
void Unit::setNumberInRefutation (int newNumber)
{
  _data->setNumberInRefutation (newNumber);
} 


inline
Inference* Unit::inference () const
{
  return _data->inference();
} 


inline
const string& Unit::name () const
{
  return _data->name ();
} 


inline
void Unit::setInference (Inference* inf)
{
  _data->setInference(inf);
} 


inline 
int Unit::mark () const 
{ 
  return _data->mark(); 
} 


inline
void Unit::setMark (int mark) 
{ 
  _data->setMark(mark);
} 


#endif 
