
/*
 * File VampireException.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "VampireException.hpp"
#include "XML.hpp"
#include "Run.hpp"


VampireException::VampireException (const string& file, int line)
  : _file(file),
    _line(line)
{
}


void VampireException::outputFileAndLine (ostream& str) const
{
  str << "(file: '" << _file
      << "', line: " << _line << ")";
} 



void VampireException::addFileAndLine (XML::DeepElement& xml) const
{
  xml.addChild(new XML::TextElement("file",_file));
  xml.addChild(new XML::IntegerElement("line",_line));
} 


UserErrorException::UserErrorException (const string& file, 
					int line,
					string message)
  : VampireException(file,line),
    _message(message)
{
}


void UserErrorException::cry (ostream& str)
{
  str << "Error: " <<_message << "\n";
} 



XML::Element* UserErrorException::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","user error"));
  xml->addChild(new XML::TextElement("message",_message));
  addFileAndLine(*xml);

  return xml;
} 



AllocationFailureException::AllocationFailureException (const string& file, 
							int line)
  : VampireException(file,line)
{
}


void AllocationFailureException::cry (ostream& str)
{
  str << "Allocation failure ";
  outputFileAndLine(str);
  str << '\n';
} 



XML::Element* AllocationFailureException::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","allocation failure"));
  addFileAndLine(*xml);

  return xml;
} 



AssertionViolationException::AssertionViolationException (const string& file, 
							  int line)
  : VampireException(file,line)
{
#if DEBUG_TRACE_ALL
  _stack = MyTracer::stackToXML();
#endif
  cry(cout);
}


void AssertionViolationException::cry (ostream& str)
{
  str << "Assertion violation ";
  outputFileAndLine(str);
  str << '\n';
#if DEBUG_TRACE_ALL
  MyTracer::printStack(cout);
#endif
} 



XML::Element* AssertionViolationException::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","assertion violation"));
  addFileAndLine(*xml);
#if DEBUG_TRACE_ALL
  xml->addChild(_stack);
#endif
  return xml;
} 



VampireTerminationException::VampireTerminationException (const string& file, 
							  int line,
							  string reason,
							  RunResult result)
  : VampireException(file,line),
    _reason(reason),
    _result(result)
{
}


void VampireTerminationException::cry (ostream& str)
{
  str << "Vampire termination exception ";
  outputFileAndLine(str);
  str << ", reason: " << _reason 
      << ", result: " << Run::toString(_result) << '\n';
} 



XML::Element* VampireTerminationException::toXML () const
{
  XML::DeepElement* xml = new XML::DeepElement("exception");
  xml->addAttribute(new XML::TextAttribute("kind","vampire termination"));
  addFileAndLine(*xml);
  xml->addChild(new XML::TextElement("reason",_reason));
  xml->addChild(new XML::TextElement("result",Run::toString(_result)));

  return xml;
} 


