//
// File:         BackDemodulation.cpp
// Description:  Back simplification by ordered unit equalities.
// Created:      Apr 16, 2000.
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Revised:      Dec 23, 2001.
//               typedef Multiset<ulong,HellConst::MaxNumOfDifferentHeaders> 
//                 SetOfSplittingBranches
//                
//               replaced by 
// 
//               typedef ExpandingMultiset<GlobAlloc,ulong,1UL,BackDemodulation> 
//                 SetOfSplittingBranches
// Revised:      May 13, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               1) Indexing for backward matching is now implemented 
//               with parh indexing and joins.
//               2) Ordering constraints reimplemented. 
//====================================================
#include "BackDemodulation.hpp"
#include "Clause.inl"
#include "DestructionMode.hpp"
#include "TemporaryClauseCollector.hpp"
#include "VampireKernelInterceptMacros.hpp"
using namespace BK;
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACK_DEMODULATION
#define DEBUG_NAMESPACE "BackDemodulation"
#endif
#include "debugMacros.hpp"        
//====================================================



VK::BackDemodulation::BackDemodulation() : 
  _bDOnSplittingBranchesFlag(false),
  _clauseCollector(0),
  _splittingBranches(DOP("BackDemodulation::_splittingBranches")),
  _numOfBDOnSplittingBranches(0UL)
{
  CALL("constructor BackDemodulation()");
  DOP(_splittingBranches.freeze());
  _compiledConstraint = 
    SimplificationOrdering::current()->reserveCompiledConstraintForInstanceRetrieval();
  unlock();
}; // VK::BackDemodulation::BackDemodulation()

VK::BackDemodulation::~BackDemodulation() 
{ 
  CALL("destructor ~BackDemodulation()");  
  if (DestructionMode::isThorough()) 
    {
      ASSERT(_compiledConstraint);
      SimplificationOrdering::current()->releaseCompiledConstraintForInstanceRetrieval(_compiledConstraint);
    };
}; // VK::BackDemodulation::~BackDemodulation() 
  

void VK::BackDemodulation::reset()
{
  CALL("reset()");
  BK::destroy(*this);
  BK::init(*this);
}; // void VK::BackDemodulation::reset()


void VK::BackDemodulation::integrate(Clause* cl)
{
  CALL("integrate(Clause* cl)");   
  ASSERT(!isLocked());
  lock();
  cl->protect();
  TERM* lit;
  ulong ln = 0;
  for(LiteralList::Iterator litIter(cl->LitList());
      litIter.notEnd();
      litIter.next())
    {
      lit = litIter.currentLiteral();
      if (lit->arity())
	{
	  integrateLiteral(lit,ln,cl);
	};         
      ln++;
    };
  cl->unprotect();
  unlock();
}; // void VK::BackDemodulation::integrate(Clause* cl)

inline
void VK::BackDemodulation::integrateLiteral(TERM* lit,ulong litNum,Clause* cl)
{
  CALL("integrateLiteral(TERM* lit,ulong litNum,Clause* cl)"); 
  ASSERT(isLocked());
  TERM* arg = lit->Args();
  const bool* nonBasicArgs;
  if (_prohibitDemodulationIntoInputNonBasic &&
      ((nonBasicArgs = lit->nonBasicArguments())))
    {
      ulong argNum = 0UL;
      while (!arg->IsNull())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    {
	      integrateComplexTerm(arg->First(),litNum,cl);
	    };
	  arg = arg->Next();
	  ++argNum;
	};
    }
  else
    {
      while (!arg->IsNull())
	{
	  if (!arg->isVariable())
	    {
	      integrateComplexTerm(arg->First(),litNum,cl);
	    };
	  arg = arg->Next();
	};
    };
}; // void VK::integrateLiteral(TERM* lit,ulong litNum,Clause* cl)



void VK::BackDemodulation::integrateComplexTerm(TERM* term,ulong litNum,Clause* cl)
{
  CALL("integrateComplexTerm(TERM* term,ulong litNum,Clause* cl)");
  ASSERT(isLocked());
  ASSERT(term->isComplex());

#ifdef INTERCEPT_BACKWARD_MATCHING_INDEXING
  if (!term->Occurences())
    {
      INTERCEPT_BACKWARD_MATCHING_INDEXING_INTEGRATE(term);
    };
#endif
  term->RegisterOccurence(cl,litNum);
  _index.integrate(term);
  


  TERM* arg = term->Args();
  const bool* nonBasicArgs;
  if (_prohibitDemodulationIntoInputNonBasic &&
      ((nonBasicArgs = term->nonBasicArguments())))
    {
      ulong argNum = 0UL;
      while (!arg->IsNull())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    {
	      integrateComplexTerm(arg->First(),litNum,cl);
	    };
	  arg = arg->Next();
	  ++argNum;
	};
    }
  else
    {
      while (!arg->IsNull())
	{
	  if (!arg->isVariable())
	    {
	      integrateComplexTerm(arg->First(),litNum,cl);
	    };
	  arg = arg->Next();
	};
    };
}; // void VK::integrateComplexTerm(TERM* term,ulong litNum,Clause* cl)




void VK::BackDemodulation::remove(Clause* cl) 
{
  CALL("remove(Clause* cl)");
  ASSERT(!isLocked());
  lock();
  cl->protect();
  TERM* lit;

  for(LiteralList::Iterator litIter(cl->LitList());litIter.notEnd();litIter.next())
    {
      lit = litIter.currentLiteral();
      if (lit->arity())
	{
	  removeLiteral(lit,cl);
	};   
    };  
  cl->unprotect();
  unlock();
}; // void VK::BackDemodulation::remove(Clause* cl)  


inline
void VK::BackDemodulation::removeLiteral(TERM* lit,Clause* cl)
{
  CALL("removeLiteral(TERM* lit,Clause* cl)");
  ASSERT(isLocked());
  TERM* arg = lit->Args();
  const bool* nonBasicArgs;
  if (_prohibitDemodulationIntoInputNonBasic &&
      ((nonBasicArgs = lit->nonBasicArguments())))
    {
      ulong argNum = 0UL;
      while (!arg->IsNull())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    {
	      removeComplexTerm(arg->First(),cl);
	    };
	  arg = arg->Next();
	  ++argNum;
	};
    }
  else
    {
      while (!arg->IsNull())
	{
	  if (!arg->isVariable())
	    {
	      removeComplexTerm(arg->First(),cl);
	    };
	  arg = arg->Next();
	};
    };
}; // void VK::BackDemodulation::removeLiteral(TERM* lit,Clause* cl)

void VK::BackDemodulation::removeComplexTerm(TERM* term,Clause* cl)
{
  CALL("removeComplexTerm(TERM* term,Clause* cl)");
  ASSERT(isLocked());
  ASSERT(term->isComplex());
  
#ifdef INTERCEPT_BACKWARD_MATCHING_INDEXING
  bool nonemptyOcc = term->Occurences() && 
    (*term->Occurences());       
#endif        

  term->RemoveOccurences(cl);

#ifdef INTERCEPT_BACKWARD_MATCHING_INDEXING
  if ((!term->Occurences()) && nonemptyOcc)
    {
      INTERCEPT_BACKWARD_MATCHING_INDEXING_REMOVE(term);
    };
#endif
  
  if (!term->Occurences()) 
    _index.remove(term);

  TERM* arg = term->Args();
  const bool* nonBasicArgs;
  if (_prohibitDemodulationIntoInputNonBasic &&
      ((nonBasicArgs = term->nonBasicArguments())))
    {
      ulong argNum = 0UL;
      while (!arg->IsNull())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    {
	      removeComplexTerm(arg->First(),cl);
	    };
	  arg = arg->Next();
	  ++argNum;
	};
    }
  else
    {
      while (!arg->IsNull())
	{
	  if (!arg->isVariable())
	    {
	      removeComplexTerm(arg->First(),cl);
	    };
	  arg = arg->Next();
	};
    };
}; // void VK::BackDemodulation::removeComplexTerm(TERM* term,Clause* cl)


bool VK::BackDemodulation::eligibleAsQuery(const Clause* cl) const
{
  CALL("eligibleAsQuery(const Clause* cl) const");

  if (cl->Unit())
    {
      const TERM* lit = cl->LitList()->firstLit();
      if (lit->Negative() || !lit->IsEquality()) return false;
      if (!lit->IsOrderedEq() && _degree <= 1) return false; 
    }
  else // !cl->Unit()
    {
      if (!_bDOnSplittingBranchesFlag) return false;
      if (cl->LitList()->numOfSelectedLits() != 1) return false;
      const TERM* lit = cl->LitList()->firstLit(); // the selected one
      if (lit->Negative() || !lit->IsEquality()) return false;
      if (!lit->IsOrderedEq() && _degree <= 1) return false;       
      LiteralList::Iterator iter(cl->LitList());
      iter.next(); // skip the selected literal
      while (iter.notEnd())
	{
	  if (!iter.currentLiteral()->IsSplitting()) return false;
	  iter.next();
	};
    };
  return true;
}; // bool VK::BackDemodulation::eligibleAsQuery(const Clause* cl) const


bool VK::BackDemodulation::query(Clause* cl) 
{ 
  CALL("query(Clause* cl)");  

  //DF; cout << "TRY " << cl << "\n";

  ASSERT(!isLocked());
  lock();
  cl->protect();

  // At the moment no answer literals are allowed in rewriting rules.
  if (cl->flag(Clause::FlagContainsAnswerLiterals)) goto return_false;
  if (cl->isSupportClause() && (!_demodulateFromSupport)) goto return_false;
  if (!cl->Active()) goto return_false;;
  if (cl->Unit() || (_bDOnSplittingBranchesFlag && (cl->LitList()->numOfSelectedLits() == 1)))
    {   
      const TERM* lit = cl->LitList()->firstLit(); // the selected one
      if (lit->Positive() && lit->IsEquality())
	{
	  if (_bDOnSplittingBranchesFlag)
	    {
	      _splittingBranches.reset();
	      LiteralList::Iterator iter(cl->LitList());
	      iter.next();
	      DOP(_splittingBranches.unfreeze());
	      while (iter.notEnd())
		{
		  if (iter.currentLiteral()->IsSplitting()) 
		    {
		      ASSERT(!iter.currentLiteral()->arity());
		      _splittingBranches.addSafe(iter.currentLiteral()->HeaderNum()); 
		    }
		  else goto return_false;;
		  iter.next();
		};
	      DOP(_splittingBranches.freeze());
	    }
	  else 
	    if (!cl->Unit()) goto return_false;;

	  _query = cl;
	  _candidates = 0;
	  if (lit->IsOrderedEq())
	    {
	      if (_checkSubstSafetyWRTInputNonBasicness &&
		  !safeRuleWRTInputNonBasicness(*(lit->Arg1()),*(lit->Arg2())))
		goto return_false;;

	      _queryLHS = lit->Args()->First();
	      ASSERT(_queryLHS->isComplex());
	      INTERCEPT_BACKWARD_MATCHING_INDEXING_RETRIEVAL_QUERY(_queryLHS);
	      if (!_index.newQuery(_queryLHS))
               {
		 INTERCEPT_BACKWARD_MATCHING_INDEXING_RETRIEVAL_TOTAL;
                 goto return_false;;
               };
	      _queryRHS = *(lit->Args()->Next());
	      _swapQuerySides = false;

	      _currentConstraint = 0;
	      ASSERT(isLocked());
	      ASSERT(cl->isProtected());
	      return true;
	    }
	  else // not preordered
	    {
	      if (_degree <= 1) goto return_false;;
	      _swapQuerySides = true;
	      if (query(lit->Arg1(),lit->Arg2())) 
		{
		  ASSERT(isLocked());
		  ASSERT(cl->isProtected());
		  return true;
		};
	      _swapQuerySides = false;
	      if (!query(lit->Arg2(),lit->Arg1()))
		goto return_false;
	      ASSERT(isLocked());
	      ASSERT(cl->isProtected());
	      return true;
	    };
	};
    };
  
 return_false:
  unlock();
  cl->unprotect();
  return false; 
}; // bool VK::BackDemodulation::query(Clause* cl) 

void VK::BackDemodulation::abandonCurrentQuery()
{
  CALL("abandonCurrentQuery()");
  ASSERT(isLocked());
  unlock();
  _query->unprotect();
}; // void VK::BackDemodulation::abandonCurrentQuery()
  

bool VK::BackDemodulation::query(const TERM* lhsref,const TERM* rhsref) 
{
  CALL("query(const TERM* lhsref,const TERM* rhsref)");
  ASSERT(isLocked());
  if (lhsref->isVariable() || rhsref->isVariable()) return false;
  
  if (_checkSubstSafetyWRTInputNonBasicness &&
      !safeRuleWRTInputNonBasicness(*lhsref,*rhsref))
    return false;


  ASSERT(lhsref->IsReference());
  ASSERT(rhsref->IsReference());
  _queryLHS = lhsref->First();
  ASSERT(_queryLHS->isComplex());
  _queryRHS = *rhsref;

  _compiledConstraint->loadGreater(_queryLHS,_queryRHS.First());
  if ((!_compiledConstraint->lhsVarInstMayOrder()) ||
      (!appropriateDegree(_compiledConstraint->category())))
    {
      return false;
    };

  //DF; cout << "QUERY " << _queryLHS << "\n";

  INTERCEPT_BACKWARD_MATCHING_INDEXING_RETRIEVAL_QUERY(_queryLHS);

  if (_index.newQuery(_queryLHS))
    {
      _compiledConstraint->compile(_index.substitution());
      _currentConstraint = _compiledConstraint;
      return true;
    }
  else
    {

      //DF; cout << "CANCELLED\n";
      INTERCEPT_BACKWARD_MATCHING_INDEXING_RETRIEVAL_TOTAL;
      return false;
    };
}; // bool VK::BackDemodulation::query(const TERM* lhsref,const TERM* rhsref) 


bool VK::BackDemodulation::safeRuleWRTInputNonBasicness(const TERM& lhs,const TERM& rhs)
{
  CALL("safeRuleWRTInputNonBasicness(const TERM& lhs,const TERM& rhs)");
  _inputNonBasicVarsInRHS.reset();
  collectInputNonBasicVars(rhs,false);
  if (_inputNonBasicVarsInRHS.length() == 0UL) return true;
  cancelInputNonBasicVars(lhs,false);
  return _inputNonBasicVarsInRHS.length() == 0UL;
}; // bool VK::BackDemodulation::safeRuleWRTInputNonBasicness(..)


void VK::BackDemodulation::collectInputNonBasicVars(const TERM& term,bool alreadyInputNonBasic)
{
  CALL("collectInputNonBasicVars(const TERM& term,bool alreadyInputNonBasic)");
  if (term.isVariable())
    {
      if (alreadyInputNonBasic) _inputNonBasicVarsInRHS.add1(term.var());
    }
  else
    {
      const TERM* complexTerm = term.First();
      const bool* nonBasicArgs;
      if (alreadyInputNonBasic ||
	  (!((nonBasicArgs = complexTerm->nonBasicArguments()))))
	{
	  for (const TERM* arg = complexTerm->Args(); 
	       !arg->IsNull();
	       arg = arg->Next())
	    collectInputNonBasicVars(*arg,alreadyInputNonBasic);
	}
      else
	{
	  ASSERT(!alreadyInputNonBasic && nonBasicArgs);
	  ulong argNum = 0UL;
	  for (const TERM* arg = complexTerm->Args(); 
	       !arg->IsNull();
	       arg = arg->Next())
	    {
	      collectInputNonBasicVars(*arg,nonBasicArgs[argNum]);
	      ++argNum;
	    };
	};
    };
}; // void VK::BackDemodulation::collectInputNonBasicVars(const TERM& term,bool alreadyInputNonBasic)

void VK::BackDemodulation::cancelInputNonBasicVars(const TERM& term,bool alreadyInputNonBasic)
{
  CALL("cancelInputNonBasicVars(const TERM& term,bool alreadyInputNonBasic)");
  if (term.isVariable())
    {
      if (alreadyInputNonBasic) _inputNonBasicVarsInRHS.remove(term.var());
    }
  else
    {
      const TERM* complexTerm = term.First();
      const bool* nonBasicArgs;
      if (alreadyInputNonBasic ||
	  (!((nonBasicArgs = complexTerm->nonBasicArguments()))))
	{
	  for (const TERM* arg = complexTerm->Args(); 
	       !arg->IsNull();
	       arg = arg->Next())
	    cancelInputNonBasicVars(*arg,alreadyInputNonBasic);
	}
      else
	{
	  ASSERT(!alreadyInputNonBasic && nonBasicArgs);
	  ulong argNum = 0UL;
	  for (const TERM* arg = complexTerm->Args(); 
	       !arg->IsNull();
	       arg = arg->Next())
	    {
	      cancelInputNonBasicVars(*arg,nonBasicArgs[argNum]);
	      ++argNum;
	    };
	};
    };
}; // void VK::BackDemodulation::cancelInputNonBasicVars(const TERM& term,bool alreadyInputNonBasic)


bool VK::BackDemodulation::nextSimplified(bool& onBranch) 
{ 
  CALL("nextSimplified(bool& onBranch)");
  ASSERT(isLocked());
  
  if (!_query->Active()) 
    { 
      INTERCEPT_BACKWARD_MATCHING_INDEXING_RETRIEVAL_ABORT;
      return false;
    };  
  onBranch = false;
  
 check_candidates:
  if (_candidates) 
    {   
      _recentlySimplified = _candidates->key();
      if (_recentlySimplified == _query) 
	{ 
	  _candidates = _candidates->next();
	  goto check_candidates; 
	}; 
 

      if (!_recentlySimplified)
	{
	  _candidates = _candidates->next();
	  goto check_candidates; 
	};

      ASSERT(_recentlySimplified);
      if (!_recentlySimplified->Active())
	{ 
	  _candidates = _candidates->next();
	  goto check_candidates; 
	}; 
      if (_bDOnSplittingBranchesFlag)
	{
	  ulong numOfBranchesToCover = _splittingBranches.numOfDiffElements();
	  if (!numOfBranchesToCover) return true;
	  if (_recentlySimplified->numOfAllLiterals() <= numOfBranchesToCover)
	    {
	      // Definitely can not be rewritten: does not have 
	      // enough splitting branches.
	      _candidates = _candidates->next();
	      goto check_candidates; 
	    };     
	  LiteralList::Iterator iter(_recentlySimplified->LitList());
	  do
	    {
	      if ((iter.currentLiteral()->IsSplitting()) 
		  && (_splittingBranches.safeScoreOf(iter.currentLiteral()->HeaderNum())))
		{
		  ASSERT(!iter.currentLiteral()->arity());          
		  ASSERT(numOfBranchesToCover);
		  numOfBranchesToCover--;
		  if (!numOfBranchesToCover) 
		    { 
		      _numOfBDOnSplittingBranches++;
		      onBranch = true;
		      return true; 
		    };
		};          
	      iter.next();
	    }
	  while (iter.notEnd());
	  ASSERT(numOfBranchesToCover);
	  _candidates = _candidates->next();
	  goto check_candidates; 
	};

      return true;
    };


 try_next_match:
  
  if (_index.nextMatch(_redex))
    {
      INTERCEPT_BACKWARD_MATCHING_INDEXING_RETRIEVAL_RESULT(_redex);
      if (_currentConstraint && 
	  (!_currentConstraint->holds()))
	{
	  goto try_next_match;
	};
    }
  else
    {
      INTERCEPT_BACKWARD_MATCHING_INDEXING_RETRIEVAL_TOTAL;
      if (_swapQuerySides)
	{
	  _swapQuerySides = false;
	  const TERM* lit = _query->LitList()->firstLit();
	  if (!query(lit->Arg2(),lit->Arg1())) return false;
	  goto try_next_match;  
	}
      else 
        return false;
    };


  ASSERT(_redex && _redex->isComplex());

  const TERM::OccList* occ = _redex->Occurences();   

  if (occ)
    {
      _candidates = occ->first();
      goto check_candidates;
    }
  else goto try_next_match; 
}; // nextSimplified(bool& onBranch) 
 
bool VK::BackDemodulation::collectInference() 
{
  CALL("collectInference()");
  ASSERT(isLocked());
  ASSERT(_candidates);
  if (rewrite(_candidates)) { _candidates = _candidates->next(); return true; };
  _candidates = _candidates->next();   
  return false; 
}; // bool collectInference() 





bool VK::BackDemodulation::rewrite(const TERM::OccList::Node* occ)
{
  CALL("rewrite(const TERM::OccList::Node* occ)");

  _clauseCollector->openNewClause(ClauseBackground::BackDemod,
				  (Clause*)_query,
				  (Clause*)_recentlySimplified);  
  _clauseCollector->announceNewClause();

  TERM::LitNumMask* litMask = occ->value();
  ASSERT(litMask);
   
  ulong litNum = 0;
  for (LiteralList::Iterator iter(_recentlySimplified->LitList());
       iter.notEnd();
       iter.next())
    {
      if (litMask->contains(litNum)) // this literal contains rewritten redex  
	{
	  if (!rewriteLit(iter.currentLiteral(),
			  iter.currentElement()->properties()))
	    goto rejected;
	}
      else // literal remains unchanhed 
	if (!_clauseCollector->pushLiteral(iter.currentLiteral(),
					   iter.currentElement()->properties())) 
	  goto rejected;  
      litNum++;
    };
                  
  
  _clauseCollector->makeVIP();
  if (!_clauseCollector->finaliseClause())
    {
    rejected:
      return _clauseCollector->rejectedAsRedundant();
    };
  return true; 
}; // bool VK::BackDemodulation::rewrite(TERM::OccList::Node* occ)

bool VK::BackDemodulation::rewriteLit(TERM* lit,const LiteralProperties& prop)             
{
  CALL("rewriteLit(TERM* lit,const LiteralProperties& prop)");
  ASSERT(_redex->isComplex());
  TERM header = *lit;
  if (header.IsEquality() && header.IsOrderedEq()) header.MakeUnordered(); 
 
  if (!_clauseCollector->openNewNonpropLiteral(header,prop)) 
    return false; 
  _rewrittenLitIter.Reset(lit);
  ALWAYS(_rewrittenLitIter.Next());
 check_position:    
  if (_rewrittenLitIter.CurrPos() == _redex)
    {
      if (!collectRHSInstance()) return false;      
      if (_rewrittenLitIter.SkipTerm()) goto check_position;
    }
  else
    {
      if (!_clauseCollector->pushSymbol(_rewrittenLitIter.symbolRef(),false)) 
	return false;
      if (_rewrittenLitIter.Next()) goto check_position;
    };
  return _clauseCollector->closeNonpropLiteral();
}; // bool VK::BackDemodulation::rewriteLit(TERM* lit,const LiteralProperties& prop)  



inline bool VK::BackDemodulation::collectRHSInstance()
{
  CALL("collectRHSInstance()");
  _rhsInstance.reset(_queryRHS,_index.substitution());
  do
    {
      if (!_clauseCollector->pushSymbol(_rhsInstance.symbol(),false)) 
	return false;
    }
  while (_rhsInstance.next());
  return true;
}; // bool VK::BackDemodulation::collectRHSInstance()

//====================================================
