//
// File:         BackDemodulation.hpp
// Description:  Back simplification by ordered unit equalities.
// Created:      Apr 16, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 23, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               typedef Multiset<ulong,HellConst::MaxNumOfDifferentHeaders> 
//                 SetOfSplittingBranches
//                
//               replaced by 
// 
//               typedef ExpandingMultiset<GlobAlloc,ulong,1UL,BackDemodulation> 
//                 SetOfSplittingBranches
// Revised:      May 13, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               1) Indexing for backward matching is now implemented 
//               with parh indexing and joins.
//               2) Ordering constraints reimplemented. 
//====================================================
#ifndef BACK_DEMODULATION_H
//====================================================
#define BACK_DEMODULATION_H
#include "jargon.hpp"
#include "GlobAlloc.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "ExpandingMultiset.hpp"
#include "Term.hpp"
#include "InstanceRetrievalForDemodulation.hpp"
#include "LiteralList.hpp"
#include "RuntimeError.hpp"
#include "SimplificationOrdering.hpp"
namespace VK
{
  class TemporaryClauseCollector;
}; // namespace VK
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACK_DEMODULATION
#define DEBUG_NAMESPACE "BackDemodulation"
#endif
#include "debugMacros.hpp"
//====================================================

namespace VK
{
class Clause;
class SHARING_INDEX;
class BackDemodulation
{
 public:
  typedef BK::ExpandingMultiset<BK::GlobAlloc,ulong,32UL,BackDemodulation> SetOfSplittingBranches;

 public:
  BackDemodulation();

  ~BackDemodulation();
  
  void reset();

  bool isLocked() const { return _isLocked; };

  void setClauseCollector(TemporaryClauseCollector* cc)
  {
    _clauseCollector = cc;
  };

  void setDegree(long d) { _degree = d; };
  void setBDOnSplittingBranchesFlag(bool fl) 
    {
      if (_bDOnSplittingBranchesFlag && (!fl)) BK::RuntimeError::report("Switching _bDOnSplittingBranchesFlag off.");
      _bDOnSplittingBranchesFlag = fl; 
    };
  void setDemodulateFromSupport(bool fl)
  {
    _demodulateFromSupport = fl;
  };
  void setProhibitDemodulationIntoInputNonBasic(bool fl)
  {
    _prohibitDemodulationIntoInputNonBasic = fl;
  };
  
  void setCheckSubstSafetyWRTInputNonBasicness(bool fl)
  {
    _checkSubstSafetyWRTInputNonBasicness = fl;
  };

  bool eligibleAsQuery(const Clause* cl) const;
  bool query(Clause* cl);
  void abandonCurrentQuery();
  bool nextSimplified(bool& onBranch);
  bool collectInference();
  
  Clause* simplified() { return _recentlySimplified; }; 
  void integrate(Clause* cl);
  void remove(Clause* cl);
  ulong statBDOnSplittingBranches() const { return _numOfBDOnSplittingBranches; };
 private:
  void integrateLiteral(TERM* lit,ulong litNum,Clause* cl);
  void integrateComplexTerm(TERM* term,ulong litNum,Clause* cl);
  void removeLiteral(TERM* lit,Clause* cl);
  void removeComplexTerm(TERM* term,Clause* cl);

  bool query(const TERM* lhsref,const TERM* rhsref);

  bool safeRuleWRTInputNonBasicness(const TERM& lhs,const TERM& rhs);
  void collectInputNonBasicVars(const TERM& term,bool alreadyInputNonBasic);
  void cancelInputNonBasicVars(const TERM& term,bool alreadyInputNonBasic); 

  bool appropriateDegree(const SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Category& cat)
    {
      switch (cat)
	{	  
	case SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyTautology:
	  return  (_degree >= 2);
	case SimplificationOrdering::CompiledConstraintForInstanceRetrieval::Volatile:
	  return (_degree >= 3);
	case SimplificationOrdering::CompiledConstraintForInstanceRetrieval::NearlyContradiction:
	  return (_degree >= 4);
	default: return false;
	};
    };

  bool rewrite(const TERM::OccList::Node* occ);
  bool rewriteLit(TERM* lit,const LiteralProperties& prop);  

  bool collectRHSInstance();
  
private:
  void lock() { _isLocked = true; };
  void unlock() { _isLocked = false; };


 private:
  bool _isLocked;
  bool _swapQuerySides;
  long _degree;
  bool _bDOnSplittingBranchesFlag;
  bool _demodulateFromSupport;
  bool _prohibitDemodulationIntoInputNonBasic;
  bool _checkSubstSafetyWRTInputNonBasicness;
  TemporaryClauseCollector* _clauseCollector;

  InstRet _index;
  InstRet::Substitution::Instance _rhsInstance;

  SimplificationOrdering::CompiledConstraintForInstanceRetrieval* _compiledConstraint;

  SimplificationOrdering::CompiledConstraintForInstanceRetrieval* _currentConstraint;



  Clause* _query;
  const TERM* _queryLHS;
  TERM _queryRHS;
  TERM::Iterator _termIter;
  const TERM::OccList::Node* _candidates; 
  Clause* _recentlySimplified;
  const TERM* _redex;
  TERM::Iterator _rewrittenLitIter;
  TERM::Iterator _RHSIter;
  SetOfSplittingBranches _splittingBranches;
  ulong _numOfBDOnSplittingBranches;

  BK::LinearCombination<long,VampireKernelConst::MaxNumOfVariables> _inputNonBasicVarsInRHS;
}; // class BackDemodulation


}; // namespace VK
//====================================================
#endif
