//
// File:         BasicTempLitList.hpp
// Description:  Structure for temporary lists of temporary literals.
// Created:      Oct 24, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef BASIC_TEMP_LIT_LIST_H
#define BASIC_TEMP_LIT_LIST_H
//===================================================================
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "GlobAlloc.hpp"
#include "ObjectPool.hpp"
#include "ObjectPoolWithMemory.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BASIC_TEMP_LIT_LIST
#  define DEBUG_NAMESPACE "BasicTempLitList"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class BasicTempLitList
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    class Element
    {
    public:
      Element() {};
      ~Element() {};
      void hibernate() { _literal.clear(); };
      const LitHandle& literal() const { return _literal; };
      LitHandle& literal() { return _literal; };
      const Element* next() const { return _next; }; 
      Element* next() { return _next; }; 
      const Element* previous() const { return _previous; }; 
      Element* previous() { return _previous; }; 
      void remove() { _list->remove(this); };
      void moveTo(BasicTempLitList& lst) { _list->passTo(this,lst); };
    private:
      Element* _next;
      Element* _previous;
      LitHandle _literal;
      BasicTempLitList* _list;
      friend class BasicTempLitList;
    }; // class BasicTempLitList::Element

    typedef BK::ObjectPool<Element> ElementPool;
    class Context;

  public:
    BasicTempLitList();
    BasicTempLitList(ElementPool* elemPool);
    BasicTempLitList(const BasicTempLitList& lst);
    ~BasicTempLitList();
    void relocateTo(void* target);
    void reset();
    ElementPool* elementPool() const { return _elementPool; }; 
    bool isEmpty() const { return !_length; };
    ulong length() const { return _length; };
    void push(const LitHandle& lit);
    LitHandle pop();
    void popTo(LitHandle& lit);
    void popTo(BasicTempLitList& lst);
    void skipTop();
    void collectAllFrom(BasicTempLitList& lst);
    void passAllTo(BasicTempLitList& lst) 
    {
      CALL("passAllTo(BasicTempLitList& lst)");
      lst.collectAllFrom(*this);
    };
    void collect(LitHandle& lit);
    void collect(Element* el);
    void passTo(Element* el,BasicTempLitList& lst);
    void remove(Element* el);
    const Element* first() const { return _first; };
    Element* first() { return _first; };
    const Element* last() const { return _last; };
    Element* last() { return _last; };
    ostream& output(ostream& str) const;
#ifndef NO_DEBUG
    bool contains(const LitHandle& lit) const;
    bool checkIntegrity() const;
#endif
  private:
    BasicTempLitList& operator=(const BasicTempLitList&) 
    { 
      return *this;
    };

#ifndef NO_DEBUG
    ulong computeLengthLowerBound() const;
    bool checkLinks() const;
#endif

  private:
    Element* _first;
    Element* _last;
    ulong _length;
    ElementPool* _elementPool;
    friend class Element;
  }; // class BasicTempLitList
}; // namespace VK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BASIC_TEMP_LIT_LIST
#  define DEBUG_NAMESPACE "<global>"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace BK
{

  // specialisation of the hibernation template for BasicTempLitList
  template <>
  inline
  static
  void hibernate(VK::BasicTempLitList& obj)
  {
    CALL("hibernate(VK::BasicTempLitList& obj)");
    obj.reset();
  };

}; // namespace BK



//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BASIC_TEMP_LIT_LIST
#  define DEBUG_NAMESPACE "<global>"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace BK
{
  // specialisation of the hibernation template for BasicTempLitList::Element
  template <>
  inline
  static
  void hibernate(VK::BasicTempLitList::Element& obj)
  {
    obj.hibernate();
  };
}; // namespace BK


//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BASIC_TEMP_LIT_LIST
#  define DEBUG_NAMESPACE "BasicTempLitList::Context"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  class BasicTempLitList::Context
  {
  public:
    class AutoActivator
    {
    public:
      AutoActivator() : _previous(Context::getCurrent())
      {
	Context::setCurrent(static_cast<Context*>(0));
      };
      AutoActivator(Context* c) : _previous(Context::getCurrent())
      {
	Context::setCurrent(c);
      };
      ~AutoActivator()
      {
	Context::setCurrent(_previous);
      };
    private:
      Context* _previous;
    }; // class AutoActivator

  public:
    Context() {};
    ~Context() {};
    BK::ObjectPool<BasicTempLitList::Element>* defaultElementPool() 
    { 
      return &_defaultElementPool; 
    };
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
  private:
    BK::ObjectPoolWithMemory<BK::GlobAlloc,BasicTempLitList::Element> _defaultElementPool;
    static Context* _current;
  }; // class BasicTempLitList::Context
    

}; // namespace VK



//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BASIC_TEMP_LIT_LIST
#  define DEBUG_NAMESPACE "BasicTempLitList"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  inline
  void BasicTempLitList::reset()
  {
    CALL("reset()");
    ASSERT(checkIntegrity());
    while (_first)
      {
	Element* tmp = _first;
	_first = _first->next();
	ElementPool::returnObject(tmp);
      };
    _last = 0;
    _length = 0UL;
    ASSERT(checkIntegrity());
  }; // void BasicTempLitList::reset()

  inline
  void BasicTempLitList::push(const LitHandle& lit)
  {
    CALL("push(const LitHandle& lit)");
    ASSERT(checkIntegrity());
    Element* newElem = _elementPool->lendObject();
    newElem->literal() = lit;
    newElem->_list = this;
    newElem->_next = 0;
    if (_last)
      {
	_last->_next = newElem;
	newElem->_previous = _last;
	++_length;
      }
    else
      {
	_first = newElem;
	newElem->_previous = 0;
	_length = 1UL;
      };
    _last = newElem;
    ASSERT(checkIntegrity());
    ASSERT(contains(lit));
  }; // void BasicTempLitList::push(const LitHandle& lit)

  
  inline
  BasicTempLitList::LitHandle BasicTempLitList::pop()
  {
    CALL("BasicTempLitList::pop()");
    ASSERT(checkIntegrity());
    ASSERT(!isEmpty());
    LitHandle res(_last->literal(),0);
    Element* obsoleteElem = _last;
    _last = _last->previous();
    if (_last)
      {
	_last->_next = 0;
	--_length;
      }
    else
      {
	_first = 0;
	_length = 0UL;
      };
    ElementPool::returnObject(obsoleteElem);
    ASSERT(checkIntegrity());
    return res;
  }; // LitHandle BasicTempLitList::pop()

  inline
  void BasicTempLitList::popTo(LitHandle& lit)
  {
    CALL("popTo(LitHandle& lit)");
    ASSERT(checkIntegrity());
    ASSERT(!isEmpty());
    lit.collectFrom(_last->literal());
    Element* obsoleteElem = _last;
    _last = _last->previous();
    if (_last)
      {
	_last->_next = 0;
	--_length;
      }
    else
      {
	_first = 0;
	_length = 0UL;
      };
    ElementPool::returnObject(obsoleteElem);
    ASSERT(checkIntegrity());
  }; // void BasicTempLitList::popTo(LitHandle& lit)

  inline
  void BasicTempLitList::popTo(BasicTempLitList& lst)
  {
    CALL("popTo(BasicTempLitList& lst)");
    ASSERT(this != &lst);
    ASSERT(checkIntegrity() && lst.checkIntegrity());
    ASSERT(!isEmpty());
    if (lst._elementPool == _elementPool)
      {
	// no need to return the element object to the pool, 
	// it can just be passed to lst

	Element* elem = _last;
	_last = _last->previous();
	if (_last)
	  {
	    _last->_next = 0;
	    --_length;
	  }
	else
	  {
	    _first = 0;
	    _length = 0UL;
	  };
	
	// integrate elem into lst
	elem->_list = &lst;
	elem->_next = 0;

	if (lst._last)
	  {
	    lst._last->_next = elem;
	    elem->_previous = lst._last;
	    ++(lst._length);
	  }
	else
	  {
	    lst._first = elem;
	    elem->_previous = 0;
	    lst._length = 1UL;
	  };
	lst._last = elem;
      }
    else // the lists are using different element pools
      {
	// the element object has to be returned to the pool,
        // and a new one has to be allocated in lst
	Element* newElem = lst._elementPool->lendObject();
	newElem->literal().collectFrom(_last->literal());
	newElem->_list = &lst;
	newElem->_next = 0;
	
	if (lst._last)
	  {
	    lst._last->_next = newElem;
	    newElem->_previous = lst._last;
	    ++(lst._length);
	  }
	else
	  {
	    lst._first = newElem;
	    newElem->_previous = 0;
	    lst._length = 1UL;
	  };
	lst._last = newElem;
	
	// now return the obsolete element object
	
	Element* obsoleteElem = _last;
	_last = _last->previous();
	if (_last)
	  {
	    _last->_next = 0;
	    --_length;
	  }
	else
	  {
	    _first = 0;
	    _length = 0UL;
	  };
	ElementPool::returnObject(obsoleteElem);
      };
    ASSERT(checkIntegrity() && lst.checkIntegrity());
  }; // void BasicTempLitList::popTo(BasicTempLitList& lst) 


  inline
  void BasicTempLitList::skipTop()
  {
    CALL("skipTop()");
    ASSERT(checkIntegrity());
    ASSERT(!isEmpty());
    Element* obsoleteElem = _last;
    _last = _last->previous();
    if (_last)
      {
	_last->_next = 0;
	--_length;
      }
    else
      {
	_first = 0;
	_length = 0UL;
      };
    ElementPool::returnObject(obsoleteElem);
    ASSERT(checkIntegrity());
  }; // void BasicTempLitList::skipTop()



  inline
  void BasicTempLitList::collect(LitHandle& lit)
  {
    CALL("collect(LitHandle& lit)");
    ASSERT(checkIntegrity());
    Element* newElem = _elementPool->lendObject();
    newElem->literal().collectFrom(lit);
    newElem->_list = this;
    newElem->_next = 0;
    if (_last)
      {
	_last->_next = newElem;
	newElem->_previous = _last;
	++_length;
      }
    else
      {
	_first = newElem;
	newElem->_previous = 0;
	_length = 1UL;
      };
    _last = newElem;
    ASSERT(checkIntegrity());
    ASSERT(lit.isNull());
  }; // void BasicTempLitList::collect(LitHandle& lit)

  inline
  void BasicTempLitList::collect(Element* el)
  {
    CALL("collect(Element* el)");
    el->_list->passTo(el,*this);
  };
  
  inline
  void BasicTempLitList::passTo(Element* el,BasicTempLitList& lst)
  {
    CALL("passTo(Element* el,BasicTempLitList& lst)");
    ASSERT(this != &lst);
    ASSERT(el->_list == this);
    ASSERT(checkIntegrity() && lst.checkIntegrity());
    
    if (el->_previous)
      {
	el->_previous->_next = el->_next;
      }
    else
      {
	ASSERT(_first == el);
	_first = el->_next;
      };
    if (el->_next)
      {
	el->_next->_previous = el->_previous;
      }
    else
      {
	ASSERT(_last == el);
	_last = el->_previous;
      };
    --(_length);
    

    if (_elementPool == lst._elementPool)
      {
	// no need to return the element object to its pool, 
	// it can just be passed to lst

	el->_list = &lst;
	el->_next = 0;
	if (lst._last)
	  {
	    lst._last->_next = el;
	    el->_previous = lst._last;
	  }
	else
	  {
	    lst._first = el;
	    el->_previous = 0;
	  };
	lst._last = el;
      }
    else // lst and *this are using different element pools
      {
	// the element object has to be returned to its pool,
        // and a new one has to be allocated in lst

	Element* newElem = lst._elementPool->lendObject();
	newElem->literal().collectFrom(el->literal());
	newElem->_list = &lst;
	newElem->_next = 0;
	
	if (lst._last)
	  {
	    lst._last->_next = newElem;
	    newElem->_previous = lst._last;
	  }
	else
	  {
	    lst._first = newElem;
	    newElem->_previous = 0;
	  };
	lst._last = newElem;
	ElementPool::returnObject(el);
      };
    ++lst._length;
    ASSERT(checkIntegrity() && lst.checkIntegrity());
  }; // void BasicTempLitList::passTo(Element* el,BasicTempLitList& lst)


  inline
  void BasicTempLitList::remove(Element* el)
  {
    CALL("remove(Element* el)");
    ASSERT(checkIntegrity());
    ASSERT(el->_list == this);
    
    if (el->_previous)
      {
	el->_previous->_next = el->_next;
      }
    else
      {
	ASSERT(_first == el);
	_first = el->_next;
      };
    if (el->_next)
      {
	el->_next->_previous = el->_previous;
      }
    else
      {
	ASSERT(_last == el);
	_last = el->_previous;
      };
    --_length;
    ElementPool::returnObject(el);
    ASSERT(checkIntegrity());
  }; // void BasicTempLitList::remove(Element* el)

#ifndef NO_DEBUG
  inline
  bool BasicTempLitList::contains(const LitHandle& lit) const
  {
    for (const Element* el = first(); el; el = el->next())
      if (el->literal() == lit)
	return true;
    return false;
  }; // bool BasicTempLitList::contains(const LitHandle& lit) const

  inline
  bool BasicTempLitList::checkIntegrity() const
  {
    if (_length)
      {
	if (!_first)
	  {
	    cout << "% BasicTempLitList::checkIntegrity()> "
		 << " _length && (!_first) "
		 << "\n";
	    return false;
	  };
	if (!_last)
	  {
	    cout << "% BasicTempLitList::checkIntegrity()> "
		 << " _length && (!_last) "
		 << "\n";
	    return false;
	  };
	if (computeLengthLowerBound() != _length)
	  {
	    cout << "% BasicTempLitList::checkIntegrity()> "
		 << " (computeLengthLowerBound() != _length) "
		 << "\n";
	    return false;
	  };
	if (!checkLinks())
	  {
	      
	    cout << "% BasicTempLitList::checkIntegrity()> "
		 << " !checkLinks() "
		 << "\n";
	    return false;
	  };
      }
    else
      {
	if (_first)
	  {
	    cout << "% BasicTempLitList::checkIntegrity()> "
		 << " (!_length) && _first "
		 << "\n";
	    return false;
	  };
	if (_last)
	  {
	    cout << "% BasicTempLitList::checkIntegrity()> "
		 << " (!_length) && _last "
		 << "\n";
	    return false;
	  };
      };
    return true;
  }; // bool BasicTempLitList::checkIntegrity() const


  inline
  ulong BasicTempLitList::computeLengthLowerBound() const
  {
    ulong res = 0UL;
    for (const Element* el = first(); (el && (res <= _length)); el = el->next())
      ++res;
    return res;
  }; // ulong BasicTempLitList::computeLengthLowerBound() const
  
  inline
  bool BasicTempLitList::checkLinks() const
  {
    if (!_first) return true;
    if (!_last) 
      {
	cout << "% BasicTempLitList::checkLinks()> "
	     << " !_last"
	     << "\n";
	return false;
      };
    if (_first->_previous) 
      {
	cout << "% BasicTempLitList::checkLinks()> "
	     << " _first->_previous"
	     << "\n";
	return false;
      };
    if (_last->_next) 
      {
	cout << "% BasicTempLitList::checkLinks()> "
	     << " _last->_next"
	     << "\n";
	return false;
      };
    for (const Element* el = _first; el->next();  el = el->next())
      if (el->next()->_previous != el)
	{
	  cout << "% BasicTempLitList::checkLinks()> "
	       << " el->next()->_previous != el"
	       << "\n";
	  return false;
	};
    return true;
  }; // bool BasicTempLitList::checkLinks() const


#endif

}; // namespace VK

//===================================================================

#endif
