//
// File:         Chaining.hpp
// Description:  Ordered chaining for transitive predicates.  
// Created:      Jun 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//==================================================================
#ifndef CHAINING_H
#define CHAINING_H
//==================================================================
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "premise_factoring.hpp"
#include "weight_formula.hpp"
#include "SimplificationOrdering.hpp"
#include "iu_machine.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
namespace VK
{
  class TERM;
  class Clause;
  class SymbolInfo;
  class OR_INDEX;
  class RES_PAIR_LIST;
  class ForwardProcessing;
  class Statistics;
}; // namespace VK
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CHAINING
#  define DEBUG_NAMESPACE "Chaining"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{
  class Chaining
  {
  public:
    Chaining(OR_INDEX* unifIndex);
    ~Chaining();

    void setStatistics(Statistics* stat)
    {
      _dummyClauseCollector.setStatistics(stat);
      _statistics = stat;
    };

    void setClauseCollector(TemporaryClauseCollector* cc) 
    {
      _clauseCollector = cc;
    };
    
    void setForwardProcessing(ForwardProcessing* fp)
    {
      _forwardProcessing = fp;
    };


    void setTransitivity(const SymbolInfo* table) 
    {
      _transitivity = table;
    };
    void setChainingThroughVariablesFlag(bool fl)
    {
      _chainingThroughVariables = fl;
    };
    void setMainPremiseSubsumption(bool fl)
    {
      _mainPremiseSubsumption = fl;
    };
    void setSidePremiseSubsumption(bool fl)
    {
      _sidePremiseSubsumption = fl;
    };


    void setMainSelectedLitDemodulation(bool fl)
    {
      _mainSelectedLitDemodulation = fl;
    };
    void setSideSelectedLitDemodulation(bool fl)
    {
      _sideSelectedLitDemodulation = fl;
    };

    
    void setClauseSizeLimit(long lim)
    {
      _clauseSizeLimit = lim;
    };


    bool loadQuery(Clause* cl);
    bool nextInference();
  private:
    enum OuterContinuation
    {
      LoadForwardPositive,
      NextForwardPositive,
      NextBackwardPositive,
      NextForwardNegative1,
      NextBackwardNegative1,
      NextForwardNegative2,
      NextBackwardNegative2
    }; // enum OuterContinuation

    enum InnerContinuation
    {
      FirstLit,
      FirstFactoring,
      NextFactoring
    }; // enum InnerContinuation
  private:
    
    bool nextForwardPositive();
    bool nextBackwardPositive();
    bool nextForwardNegative1();
    bool nextBackwardNegative1();
    bool nextForwardNegative2();
    bool nextBackwardNegative2();



    bool posChainingLitSelectedIn(Clause* cl);
    bool negChainingLitSelectedIn(Clause* cl);

    bool buildResolvent(PrefixSym* newLitArg1,
			const TERM& newLitArg2,
			bool inheritFromQuery);
    
    bool buildResolvent(const TERM& newLitArg1,
			PrefixSym* newLitArg2,
			bool inheritFromQuery);


    bool nextFactoring();

    bool subsumeMainPremise();
    bool collectMainPremiseInstance();
    bool subsumeSidePremise();
    bool collectSidePremiseInstance();


    bool demodulateSelectedLits();
    bool collectMainPremiseSubtermsAvailableForParamodulationInto();
    bool collectSidePremiseSubtermsAvailableForParamodulationInto();


  private:  
    OPEN_BR_DECL(1);
    Statistics* _statistics;
    const SymbolInfo* _transitivity;
    bool _chainingThroughVariables;
    Clause* _query;
    TemporaryClauseCollector* _clauseCollector;
    ForwardProcessing* _forwardProcessing;
    OR_INDEX* _unifIndex;
    BK::Comparison _headTailComparison;
    OuterContinuation _outerContinuation;
    InnerContinuation _innerContinuation;
    OR_PREMISE_WITH_FACTORING _queryPremise;
    OR_PREMISE_WITH_FACTORING _storedPremise;
    ulong _queryVarIndex;
    ulong _storedVarIndex;
    WEIGHT_FORMULA _residualQuerySizeFormula;
    long _residualQuerySize;
    long _clauseSizeLimit;
    ulong _currentQueryLit;
    SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
    SimplificationOrdering::CompiledConstraintForUnifiers* _constraintOnQuery;
    SimplificationOrdering::ConstraintCheckOnUnifier* _storedConstraintCheck;
    SimplificationOrdering::StoredConstraint* _storedConstraint;
    TERM _pseudoHeader;
    IU_MACH _unificationMachine;
    const Clause* _currentTransitivityLaw;
    PrefixSym* _queryHead;
    PrefixSym* _queryTail;
    RES_PAIR_LIST* _currentCandidate;
    TERM _newLitHeader;
    TemporaryClause::Extras::Handle _dummyClause;
    TemporaryClauseCollector _dummyClauseCollector;
    bool _forwardMode;
    bool _mainPremiseSubsumption;
    bool _sidePremiseSubsumption;
    bool _mainSelectedLitDemodulation;
    bool _sideSelectedLitDemodulation;   
    bool _mainSelectedLitDemodulationAttempted;
    bool _mainSelectedLitDemodulated;
    CLOSE_BR_DECL(1);
  }; // class Chaining

}; // namespace VK

//=================================================================
/*
  Expalanatory scheme for notation.

  Forward positive chaining:
     Query: C \/ r(queryTail,head) : head > queryTail
     Stored: D \/ r(head,storedTail) : head > storedTail
     Conclusion: C \/ D \/ r(queryTail,storedTail)
      
     The literal r(queryTail,storedTail) in the conclusion 
     inherits properties of r(queryTail,head) in the query.

  Backward positive chaining:
     Query: C \/ r(head,queryTail) : head > queryTail
     Stored: D \/ r(storedTail,head) : head > storedTail
     Conclusion: C \/ D \/ r(storedTail,queryTail)
     
     The literal r(storedTail,queryTail) in the conclusion 
     inherits properties of r(storedTail,head) in
     the stored clause.

  Forward negative chaining 1: 
     Query: C \/ r(head,queryTail) : head > queryTail
     Stored: D \/ ~r(head,storedTail) : head > storedTail
     Conclusion: C \/ D \/ ~r(queryTail,storedTail)

     The literal ~r(queryTail,storedTail) in the conclusion 
     inherits properties of ~r(head,storedTail) in
     the stored clause.

  Backward negative chaining 1:
    Query: C \/  ~r(head,queryTail) : head > queryTail
    Stored: D \/ r(head,storedTail) : head > storedTail
    Conclusion: C \/ D \/ ~r(storedTail,queryTail)
    
    The literal ~r(storedTail,queryTail) in the conclusion 
    inherits properties of ~r(head,queryTail) in
    the query.
     
  Forward negative chaining 2: 
    Query: C \/ r(queryTail,head) : head > queryTail
    Stored: D \/ ~r(storedTail,head) : head >= storedTail
    Conclusion: C \/ D \/ ~r(storedTail,queryTail)
 
    The literal ~r(storedTail,queryTail) in the conclusion 
    inherits properties of ~r(storedTail,head) in
    the stored clause.

  Backward negative chaining 2: 
    Query: C \/ ~r(queryTail,head) : head >= queryTail
    Stored: D \/ r(storedTail,head) : head > storedTail
    Conclusion: C \/ D \/ ~r(queryTail,storedTail)

    The literal ~r(queryTail,storedTail) in the conclusion 
    inherits properties of ~r(queryTail,head) in
    the query.

 */

//=================================================================
#endif
