//
// File:         Clause.cpp
// Description:  Representation of stored clauses.
// Created:      Feb 13, 2000
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Revised:      Dec 11, 2001. 
//               Freed from VampireKernelConst::MaxNumOfLiterals.
//====================================================
#include "LinearCombination.hpp"
#include "Clause.inl"
#include "VampireKernelInterceptMacros.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE
#define DEBUG_NAMESPACE "Clause"
#endif 
#include "debugMacros.hpp"
//================================================= 
using namespace std;
using namespace BK;
using namespace VK;

#ifdef DEBUG_ALLOC_OBJ_TYPE
ClassDesc Clause::List::_classDesc("Clause::List",Clause::List::minSize(),Clause::List::maxSize());
#endif

Clause::Context* 
Clause::Context::_current; // implicitely initialised to 0


Clause* Clause::unsharedCopy() const
{
  CALL("unsharedCopy() const");
  ASSERT(!isBuiltInTheoryFact());

  LiteralList::Iterator iter(LitList());
  LiteralList* targetLitList = 
    new (LitList()->length()) 
    LiteralList(LitList()->length(),LitList()->numOfSelectedLits());
  LiteralList::Iterator targetIter(targetLitList);

  while (iter.notEnd()) 
    {
      ASSERT(targetIter.notEnd());
      targetIter.currentElement()->setLiteral(iter.currentLiteral()->unsharedLitCopy());
      targetIter.currentElement()->properties() = 
	iter.currentElement()->properties();

      iter.next();
      targetIter.next();
    };
  ASSERT(iter.end());
  ASSERT(targetIter.end());
 
  Clause* res = new Clause(targetLitList);
  _background.copyTo(res->background());
  res->number = number;


  res->_size = _size;
  res->_practicallyMinResidualSize = _practicallyMinResidualSize;

  //res->_flags = _flags; 
  ASSERT(!res->_flags);
  res->setMainSet(MainSetFree);
  if (flag(FlagHasChildren)) res->setFlag(FlagHasChildren);

  res->num_of_pos_eq = num_of_pos_eq;

  return res;
}; // Clause* Clause::unsharedCopy() const


void Clause::Destroy()         
{
  CALL("Destroy()");
  ASSERT(checkObj());
  INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_UNBIND_CLAUSE_ID(this);
  delete litList;
  //delete ancestors;
  delete this;
};  // Clause::Destroy()

void Clause::DestroyUnshared()         
{
  CALL("DestroyUnshared()");

  ASSERT(checkObj());
  INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_UNBIND_CLAUSE_ID(this);
  for (LiteralList::Iterator iter(LitList());iter.notEnd();iter.next()) 
    iter.currentLiteral()->destroyUnsharedLit();
   
  delete litList;
  //delete ancestors;
  delete this;
};  // Clause::DestroyUnshared()


 
bool Clause::MakeDefinitionOf(const TERM& definedHeader)
{
  CALL("MakeDefinitionOf(const TERM& definedHeader)");
  // true iff there is at least one literal with this header
  ulong headSize = 0UL;
  LiteralList::Iterator nextHeadLit(LitList());
  LiteralList::Iterator nextTailLit(LitList());
 
 find_next_head_lit:
  if (nextHeadLit.end()) 
    {
      if (headSize) 
	{
	  LitList()->setNumOfSelected(headSize); 
	  return true;
	};
      return false;
    };
  if (*(nextHeadLit.currentLiteral()) != definedHeader)
    {
      nextHeadLit.next();
      goto find_next_head_lit;   
    };
 find_next_tail_lit: 
  if (nextTailLit.litNum() < nextHeadLit.litNum())
    { 
      if (*(nextTailLit.currentLiteral()) != definedHeader)    
	{
	  // swap this literal with the current head literal
	  TERM* headLit = nextHeadLit.currentLiteral();
	  nextHeadLit.currentElement()->setLiteral(nextTailLit.currentLiteral());      
	  nextTailLit.currentElement()->setLiteral(headLit);
	  nextHeadLit.next();
	  nextTailLit.next();
	  goto find_next_head_lit;
	}
      else
	{
	  nextTailLit.next();
	  goto find_next_tail_lit;
	};
    };
  nextHeadLit.next();
  goto find_next_head_lit;

}; // bool Clause::MakeDefinitionOf(const TERM& definedHeader)

bool Clause::Definition(const DefinedHeaderTable* definedHeaderTable,TERM& definedHeader) const
{
  CALL("Definition(const DefinedHeaderTable* definedHeaderTable) const");
  for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next())
    if (definedHeaderTable->isDefined(iter.currentLiteral()->HeaderNum())) 
      {
	definedHeader = *(iter.currentLiteral());
	return true;
      };
  return false; 
}; // bool Clause::Definition(const DefinedHeaderTable* definedHeaderTable,TERM& definedHeader) const

bool Clause::HornClause() const
{
  CALL("HornClause() const");
  if (Unit()) { return true; };
  ulong num_of_pos = 0;
  for(LiteralList::Iterator iter(LitList());iter.notEnd();iter.next())
    {
      if (!iter.currentLiteral()->isAnswerLiteral())
	{
	  if (iter.currentLiteral()->Positive()) 
	    {
	      num_of_pos++;
	      if (num_of_pos > 1) { return false; };
	    };
	};
    };
  return true;
}; // bool Clause::HornClause() const

bool Clause::isSymmetryLaw() const 
{
  if (numOfAllLiterals() != 2) return false;
  const TERM* lit1 = LitList()->firstLit();
  const TERM* lit2 = LitList()->nth(1)->literal();
  if (!TERM::Symmetry(lit1,lit2)) return false;
  return lit1->polarity() != lit2->polarity();
}; // bool Clause::isSymmetryLaw() const


bool Clause::isTotalityLaw() const 
{
  if (numOfAllLiterals() != 2) return false;
  const TERM* lit1 = LitList()->firstLit();
  const TERM* lit2 = LitList()->nth(1)->literal();
  if (!TERM::Symmetry(lit1,lit2)) return false;
  return lit1->polarity() == lit2->polarity();
}; // bool Clause::isTotalityLaw() const


bool Clause::isCommutativityLaw(ulong& func) const
{
  CALL("isCommutativityLaw(ulong& func) const");
  if (Unit())
    {
      const TERM* lit = LitList()->firstLit();
      if  (lit->Positive() 
	   && lit->IsEquality()
	   && lit->Arg1()->IsReference()
	   && lit->Arg2()->IsReference()
	   && (TERM::Symmetry(lit->Arg1()->First(),lit->Arg2()->First())))
	{
	  func = lit->Arg1()->First()->functor();
	  return true;
	};
    }; 
  return false;     
}; // bool Clause::isCommutativityLaw(ulong& func) const


bool Clause::isTransitivityLaw(ulong& pred) const
{
  CALL("isTransitivityLaw(ulong& pred) const");
  if ((numOfAllLiterals() != 3) ||
      (flag(FlagContainsAnswerLiterals))) 
    return false;
  const TERM* lit1 = LitList()->firstLit();
  const TERM* lit2 = LitList()->nth(1)->literal();
  const TERM* lit3 = LitList()->nth(2)->literal();

  pred = lit1->functor();
  if ((lit2->functor() != pred) ||
      (lit3->functor() != pred) ||
      (lit1->arity() != 2UL))
    return false;
  
  if ((!lit1->Arg1()->isVariable()) ||
      (!lit1->Arg2()->isVariable()) ||
      (*(lit1->Arg1()) == *(lit1->Arg2())))
    return false;

  if ((!lit2->Arg1()->isVariable()) ||
      (!lit2->Arg2()->isVariable()) ||
      (*(lit2->Arg1()) == *(lit2->Arg2())))
    return false;

  if ((!lit2->Arg1()->isVariable()) ||
      (!lit2->Arg2()->isVariable()) ||
      (*(lit2->Arg1()) == *(lit2->Arg2())))
    return false;


  const TERM* negLit1 = 0;
  const TERM* negLit2 = 0;
  const TERM* posLit = 0;

  if (lit1->Negative()) 
    {
      negLit1 = lit1;
      if (lit2->Negative())
	{
	  negLit2 = lit2;
	  // lit3 should be positive
	  if (lit3->Negative()) return false;
	  posLit = lit3;
	}
      else
	{
	  posLit = lit2;
	  // lit3 should be negative
	  if (lit3->Positive()) return false;
	  negLit2 = lit3;
	};
    }
  else
    {
      posLit = lit1;
      if (lit2->Positive() || lit3->Positive()) return false;
      negLit1 = lit2;
      negLit2 = lit3;
    };

  if (*(posLit->Arg1()) == *(negLit1->Arg1()))
    {
      if ((*(negLit1->Arg2()) != *(negLit2->Arg1())) || 
	  (*(negLit2->Arg2()) != *(posLit->Arg2())))
	return false;
    }
  else
    if (*(posLit->Arg1()) == *(negLit2->Arg1()))
      {
	if ((*(negLit2->Arg2()) != *(negLit1->Arg1())) || 
	    (*(negLit1->Arg2()) != *(posLit->Arg2())))
	  return false;
      }
    else
      return false;
  return true;
}; // bool Clause::isTransitivityLaw(ulong& pred) const


ostream& Clause::outputAsPrologTerm(ostream& str) const
{
  CALL("outputAsPrologTerm(ostream& str) const");
  str << '[';

  if (isNameIntro())
    {
      // name introduction, output it as a pair (<literal>,<literal list>)
      str << '(';
      LiteralList::Iterator iter(LitList());
      str << (iter.currentLiteral()->Positive() ? "++" : "--");
      iter.currentLiteral()->outputAsPrologTerm(str);
      str << " <=> ";
      iter.next();
      str << '[';
      while (iter.notEnd())
	{
	  str << (iter.currentLiteral()->Positive() ? "++" : "--");
	  iter.currentLiteral()->outputAsPrologTerm(str);
	  if (!iter.last()) str << ',';
	  iter.next();
	};
      str << ']';
      str << ')';
    }
  else // !isNameIntro()
    {
      // ordinary clause body
      str << '[';
      
      ulong maxLits = LitList()->numOfSelectedLits();
      for (LiteralList::Iterator iter(LitList());iter.notEnd();iter.next())
	{
	  if (maxLits) 
	    {
	      str << (iter.currentLiteral()->Positive() ? "+++" : "---");
	      maxLits--;
	    }
	  else 
	    str << (iter.currentLiteral()->Positive() ? "++" : "--");
	  iter.currentLiteral()->outputAsPrologTerm(str);
	  if (!iter.last()) str << ','; 
	};
      
      str << ']';
    }; // if (isNameIntro())


  // number
  str << ',' << Number();
 
  // ancestors
  str << ",[";

  if (!isInputClause())
    {
      for (const ClauseBackground::ClauseList* cl = _background.ancestors(); 
	   cl; 
	   cl = cl->tl())
	{
	  str << cl->hd()->Number();
	  if (cl->tl()) str << ',';
	};
    };
  str << ']';
 
  // rules
  str << ",[";
  ClauseBackground::Rules rules = _background.properties();
  size_t numOfRules = rules.count();
  if (rules.bit(ClauseBackground::VIP)) numOfRules--; 
  if (rules.bit(ClauseBackground::Input))
    { 
      str << "in";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::Prepro)) 
    {
      str << "pp";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::Subgoal)) 
    {
      str << "sg";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::BuiltInTheoryFact)) 
    {
      str << "bitf";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::Support)) 
    {
      str << "sprt";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::NonSubsuming)) 
    {
      str << "nsmng";
      numOfRules--;
      if (numOfRules) str << ',';
    }; 
  
  if (rules.bit(ClauseBackground::NonSubsumable)) 
    {
      str << "nsmbl";
      numOfRules--;
      if (numOfRules) str << ',';
    }; 
  
  if (rules.bit(ClauseBackground::BinRes))
    {
      str << "br";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::ForwTransChain))
    {
      str << "ftc";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::BackTransChain))
    {
      str << "btc";
      numOfRules--;
      if (numOfRules) str << ',';
    };


  if (rules.bit(ClauseBackground::ForwSup)) 
    {
      str << "fs";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::BackSup)) 
    {
      str << "bs";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::EqRes)) 
    {
      str << "er";
      numOfRules--;
      if (numOfRules) str << ',';
    };  

  if (rules.bit(ClauseBackground::EqFact)) 
    {
      str << "ef";
      numOfRules--;
      if (numOfRules) str << ',';
    };     

  if (rules.bit(ClauseBackground::ForwDemod)) 
    {
      str << "fd";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::BackDemod)) 
    {
      str << "bd";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::EqResSimp)) 
    {
      str << "ers";
      numOfRules--;
      if (numOfRules) str << ',';
    };  

  if (rules.bit(ClauseBackground::ForwSubsRes)) 
    {
      str << "fsr";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::BackSubsRes)) 
    {
      str << "bsr";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::Split)) 
    {
      str << "sp"; 
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::NegEqSplit)) 
    {
      str << "nesp"; 
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::Reanimation)) 
    {
      str << "rea"; 
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::Name)) 
    {
      str << "nm"; 
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::SimplificationByBuiltInTheories)) 
    {
      str << "sbit";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::NameIntro)) 
    {
      str << "ni";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  
  if (rules.bit(ClauseBackground::Folding)) 
    {
      str << "fld";
      numOfRules--;
      if (numOfRules) str << ',';
    };

  if (rules.bit(ClauseBackground::Clausification)) 
    {
      str << "cl";
    };
  

  str << ']'; 


  return str << ']';
}; // ostream& Clause::outputAsPrologTerm(ostream& str) const

ostream& Clause::outputBackground(ostream& str) const
{
  if (IsVIP()) str << "vip ";
  str << '['; 
  BitWord rules = _background.properties();

  ClauseBackground::outputRules(str,rules);
  if (!isInputClause())
    {
      for (const ClauseBackground::ClauseList* cl = _background.ancestors(); 
	   cl; 
	   cl = cl->tl())
	{
	  str << cl->hd()->Number();
	  if (cl->tl()) str << ',';
	};
    };

  return str << ']';
}; // ostream& Clause::outputBackground(ostream& str) const


ostream& Clause::output(ostream& str) const  
{
  if (mainSet() == MainSetActive) str << "* ";
  if (mainSet() == MainSetDefinitions) str << "DEF ";
  if (isNameIntro()) str << "NAME INTRO ";
  str << Number() << ". ";
  if (isBuiltInTheoryFact())
    {
      str << "built_in_theory_fact ";
    }
  else
    {
      if (isNameIntro())
	{	  
	  LiteralList::Iterator iter(LitList());
	  const TERM* head = iter.currentLiteral();
	  iter.next();
	  ASSERT(iter.notEnd());
	  
	  // write the top-level quantifier if necessary

	  TERM::Iterator litIter(head);
	  static LinearCombination<ulong,VampireKernelConst::MaxNumOfVariables> varRegister;
	  varRegister.reset();

	  while (litIter.NextVar())
	    {
	      ulong varNum = litIter.symbolRef().var();
	      if (varRegister.coefficient(varNum) == 0UL)
		{
		  if (varRegister.isZero())
		    {
		      str << "! [";
		    }
		  else
		    str << ",";
		  str << litIter.symbolRef();		  
		  varRegister.add1(varNum);
		};
	    };

	  if (!varRegister.isZero()) str << "] : ";
	  
	  str << head << " <=> (";
	  
	  // write the inner quantifier if necessary
	  bool innerVar = false;
	  while (iter.notEnd())
	    {
	      const TERM* lit = iter.currentLiteral();
	      iter.next();
	      litIter.Reset(lit);
	      while (litIter.NextVar())
		{
		  ulong varNum = litIter.symbolRef().var();
		  if (varRegister.coefficient(varNum) == 0UL)
		    {
		      if (innerVar)
			{
			  str << ","; 
			}
		      else
			{
			  str << "! [";
			  innerVar = true;
			};
		      str << litIter.symbolRef();
		      varRegister.add1(varNum);
		    };
		};
	    };
	  if (innerVar) str << "] : (";
	  
	  iter.reset(LitList());
	  iter.next();
	  while (iter.notEnd())
	    {
	      str << iter.currentLiteral();
	      if (!iter.last()) str << " | ";
	      iter.next();
	    };
	  
	  if (innerVar) str << ')';
	  str << ')';
	}
      else // !isNameIntro();
	{
	  if (Empty()) { str << '#'; }
	  else 
	    LitList()->output(str,Context::getCurrent()->_showLiteralProperties);
	};

      str << " /" << weight() << '/' << size() << '/' << practicallyMinResidualSize() << "/"
	  << inferenceDepth() << "/ " 
	  <<  NumOfPosEq() << "pe ";
    };
  return outputBackground(str);   
}; // ostream& Clause::output(ostream& str) const


//=================================================  
