 
//
// File:         ClauseAssembler.cpp
// Description:  Implements clause sharing.
// Created:      Dec 13, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Note:         Was CLAUSE_INDEX in clause_index.h.
// Revised:      Feb 17, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bug fix in assembleClause(..): wrong use of variable weighting.
//============================================================================
#include "ClauseAssembler.hpp"
#include "LiteralList.hpp"
#include "BitWord.hpp"
#include "sharing.hpp"
#include "dtree.hpp"
#include "LiteralSelection.hpp"
#include "TemporaryLiteral.inl"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "TemporaryClause.hpp"
#include "Clause.inl"
#include "PagedPassiveClauseStorage.inl"
#include "LiteralProperties.inl"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_ASSEMBLER
#define DEBUG_NAMESPACE "ClauseAssembler"
#endif
#include "debugMacros.hpp"
//============================================================================

using namespace BK;
using namespace Gem;
using namespace VK;


ClauseAssembler::ClauseAssembler(LiteralSharing* litSharing,const LiteralSelection* litSelection) 
  : 
  _litSharing(litSharing), 
  _litSelection(litSelection),
  _inheritNegativeSelection(false),
  _equalityIsPossible(true),
  _eliminatedPredicateTable(0),
  _maxMaxWeight(VampireKernelConst::MaxClauseSize),
  _maxWeight(VampireKernelConst::MaxClauseSize)
{
};


void ClauseAssembler::reset()
{
  CALL("reset()");
  LiteralSharing* litSharing = _litSharing;
  const LiteralSelection* litSelection = _litSelection;
  
  BK::destroy(*this);
  BK::init(*this,litSharing,litSelection);
}; 


Clause* ClauseAssembler::assembleShared(TemporaryClause& cl)
{
  CALL("assembleShared(TemporaryClause& cl)");
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();
  ulong length = cl.literals()->length();
  Clause* res;
  if (length == 0UL)
    {
     res = assembleEmptyClause(cl.background());
     _litSharing->unlock();
     return res;
    };
  if (length == 1UL)
    {
      res = assembleUnitClause(*(cl.literals()->first()->literal()),
			       cl.background());
     _litSharing->unlock();
      return res;
    };

  // general case
  res = assembleClause(*(cl.literals()),cl.background()); 
  _litSharing->unlock(); 
  return res;
}; // Clause* ClauseAssembler::assembleShared(TemporaryClause& cl)




Clause* ClauseAssembler::assembleSharedWithoutSelection(TemporaryClause& cl)
{
  CALL("assembleSharedWithoutSelection(TemporaryClause& cl)");
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();
  ulong length = cl.literals()->length();
  Clause* res;
  if (length == 0UL) 
    {
      res = assembleEmptyClause(cl.background());
     _litSharing->unlock();
     return res;
    };
  if (length == 1UL)
    {
      res =  assembleUnitClause(*(cl.literals()->first()->literal()),
				cl.background());
     _litSharing->unlock();
      return res;
    };
  
  // general case
  res = assembleClauseWithoutSelection(*(cl.literals()),cl.background());
  _litSharing->unlock();
  return res;
}; // Clause* ClauseAssembler::assembleSharedWithoutSelection(TemporaryClause& cl)


Clause* ClauseAssembler::assembleUnshared(TemporaryClause& cl)
{
  CALL("assembleUnshared(TemporaryClause& cl)");
  return 
    assembleUnsharedClause(*(cl.literals()),cl.background());
}; // Clause* ClauseAssembler::assembleUnshared(TemporaryClause& cl)

bool 
ClauseAssembler::assembleInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
						     TemporaryClause& cl,
						     long clauseNum,
						     ostream* showNewClause,
						     bool markAsName)
{
  CALL("assembleInPagedPassiveClauseStorage(..)");

  return
    assembleClauseInPagedPassiveClauseStorage(storage,
					      *(cl.literals()),
					      cl.background(),
					      clauseNum,
					      showNewClause,
					      markAsName);
}; // bool ClauseAssembler::assembleInPagedPassiveClauseStorage(..)



Clause* ClauseAssembler::assembleSharedCopy(Clause* cl)
{
  CALL("assembleSharedCopy(Clause* cl)");
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();
  LiteralList* oldLitList = cl->LitList();
  LiteralList* newLitList = 
    new (oldLitList->length())
    LiteralList(oldLitList->length(),oldLitList->numOfSelectedLits());
  LiteralList::Iterator oldIter(oldLitList);
  LiteralList::Iterator newIter(newLitList); 
  while (oldIter.notEnd()) 
    {
      ASSERT(newIter.notEnd());
      TERM* oldLit = oldIter.currentLiteral();
      newIter.currentElement()->setLiteral(shareLiteralCopy(oldLit));
      newIter.currentElement()->properties() = 
	oldIter.currentElement()->properties();
      oldIter.next();
      newIter.next();
    };
  ASSERT(newIter.end());

  Clause* res = new Clause(newLitList);
  cl->background().copyTo(res->background());
  res->SetNumber(cl->Number());
  res->setSize(cl->size());
  res->setWeightBase(cl->weightBase());
  res->SetNumOfPosEq(cl->numOfPosEq());
  res->setPracticallyMinResidualSize(cl->practicallyMinResidualSize());
  res->setMainSet(Clause::MainSetFree);
  res->setFlag(Clause::FlagLiteralsShared);
  if (cl->flag(Clause::FlagHasChildren)) res->setFlag(Clause::FlagHasChildren);
  _litSharing->unlock();
  return res;
}; // Clause* ClauseAssembler::assembleSharedCopy(Clause* cl)



Clause* ClauseAssembler::assembleNameIntro(const TemporaryLiteral& definedLit,
					   const BasicTempLitList& body)
{
  CALL("assembleNameIntro(..)");

  ASSERT(!body.isEmpty());  
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();

  LiteralList* lit_list = 
    new(body.length() + 1) 
    LiteralList(body.length() + 1,1UL);

  Clause* result = new Clause(lit_list);
  LiteralList::Iterator iter(lit_list);
  TERM* lit;

  lit = shareTmpLiteral(definedLit,result);
  iter.currentElement()->setLiteral(lit);
  iter.currentElement()->properties() = definedLit.properties();
  iter.next();
  
  for (const BasicTempLitList::Element* el = body.first();
       el;
       el = el->next())
    {
      lit = shareTmpLiteral(*(el->literal()),result);
      ASSERT(iter.notEnd());  
      iter.currentElement()->setLiteral(lit);
      iter.currentElement()->properties() = el->literal()->properties();
      iter.next();
    };

  result->clearFlags();
  result->markAsNameIntro();
  ClauseBackground background;
  background.markAsNameIntro();
  background.copyTo(result->background());

  result->setSize(0);
  result->setWeightBase(0);
  result->SetNumOfPosEq(0);
  result->setPracticallyMinResidualSize(0);
  result->setFlag(Clause::FlagLiteralsShared);
  _litSharing->unlock();
  return result; 
}; // Clause* ClauseAssembler::assembleNameIntro(..)


void ClauseAssembler::removeFromSharing(Clause* cl)
{  
  CALL("removeFromSharing(Clause* cl)");
  ASSERT(!_litSharing->isLocked());
  _litSharing->lock();
  for (LiteralList::Iterator iter(cl->LitList());iter.notEnd();iter.next())
    removeFromSharing(iter.currentLiteral()); 
  _litSharing->unlock();
}; // void ClauseAssembler::removeFromSharing(Clause* cl)


inline 
TERM* 
ClauseAssembler::shareTmpLiteral(const TemporaryLiteral& lit,
				 Clause* cl) 
{
  CALL("shareTmpLiteral(const TemporaryLiteral& lit,const Clause* cl)");
  if (lit.isEquality())
    {
      return IntegrateEq(lit.header(),lit.arg1(),lit.arg2(),cl);
    }
  else 
    {
      return Integrate(lit.flatterm(),cl); 
    };
}; // TERM* ClauseAssembler::shareTmpLiteral(const TemporaryLiteral& lit,Clause* cl)

inline TERM* ClauseAssembler::Integrate(const Flatterm* lit,Clause* cl)
{
  CALL("Integrate(const Flatterm* lit,Clause* cl)");
#ifdef _SUPPRESS_WARNINGS_
  cl = cl;
#endif
    
  DTREE_NODE* leaf = _litSharing->Integrate(lit);
  leaf->NumOfClauses()++;
  return leaf->Term();
}; // TERM* ClauseAssembler::Integrate(const Flatterm* lit,Clause* cl)

inline TERM* ClauseAssembler::IntegrateEq(const TERM& header,const Flatterm* arg1,const Flatterm* arg2,Clause* cl)
{
#ifdef _SUPPRESS_WARNINGS_
  cl = cl;
#endif
    
  DTREE_NODE* leaf = _litSharing->IntegrateEq(header,arg1,arg2);
  leaf->NumOfClauses()++;
  return leaf->Term();
}; // TERM* ClauseAssembler::IntegrateEq(const TERM& header,const Flatterm* arg1,const Flatterm* arg2,Clause* cl) 


inline void ClauseAssembler::removeFromSharing(TERM* lit)
{
  CALL("removeFromSharing(TERM* lit,Clause* cl)");
  if (lit->arity())
    {
      DTREE_NODE** branch_to_destroy;
      DTREE_NODE* leaf;
      DTREE_NODE* last_fork;
      _litSharing->InfoForDeletion(lit,branch_to_destroy,leaf,last_fork);

      ASSERT(leaf->IsLeaf());
    
      leaf->NumOfClauses()--;
    
      if (!(leaf->NumOfClauses())) // no more clauses with this literal
	{
	  TERM* lit_copy = leaf->Term();

	  DTREE::DestroyBranchWithoutForks(*branch_to_destroy);
	  if (last_fork)
	    {
	      DTREE::ReduceFork(last_fork);
	    }
	  else
	    {
	      *branch_to_destroy = 0;
	    }; 

     
	  _litSharing->destroyLiteralCopy(lit_copy);
        
	}; 
     
    }
  else // propositional
    {
      ((*(_litSharing->Tree(*lit)))->NumOfClauses())--;
      if (!((*(_litSharing->Tree(*lit)))->NumOfClauses())) // no more clauses with this literal
	{
	  delete *(_litSharing->Tree(*lit));
      
	  *(_litSharing->Tree(*lit)) = 0;
	};
    };
}; // void ClauseAssembler::removeFromSharing(TERM* lit,Clause* cl,bool remove_from_sharing)




Clause* ClauseAssembler::assembleEmptyClause(OpenClauseBackground& ancestors) const
{
  CALL("assembleEmptyClause(OpenClauseBackground& ancestors) const");
  Clause* res = Clause::MakeEmptyClause();
  res->setSize(0);
  res->SetNumOfPosEq(0);
  res->setPracticallyMinResidualSize(0);
  ancestors.writeNewBackgroundTo(res->background());
  return res;
}; // Clause* ClauseAssembler::assembleEmptyClause(OpenClauseBackground& ancestors) const


Clause* ClauseAssembler::assembleUnitClause(TemporaryLiteral& lit,OpenClauseBackground& ancestors) 
{
  CALL("assembleUnitClause(TemporaryLiteral& lit,OpenClauseBackground& ancestors)");
  lit.markAsSelected();
  lit.properties().setSelectionLayerModifier(LiteralProperties::DefaultSLModifier);
  lit.properties().setSelectionLayer(0UL);
  _intVarRenaming.Reset();
  lit.renameVariables(_intVarRenaming);
  LiteralList* lit_list;
  bool isAnswer = lit.isAnswerLiteral();
  long weightBase;
  bool posEqSelected;
  if (isAnswer)
    {
      weightBase = 0L;
      posEqSelected = false;
    }
  else
    {
      weightBase = lit.size();
      posEqSelected =
	lit.isEquality() &&
	lit.isPositive() && 
	lit.properties().useEqForParamodulationFrom();      
    };
  
  long weight = 
    Clause::Context::getCurrent()->computeWeight(weightBase,
						 ancestors.isSubgoal(),
						 posEqSelected);
  if (weight > _maxWeight) return 0;
  


  if (isAnswer)
    {
      lit_list = new (1) LiteralList(1,0);
    }
  else
    {
      lit_list = new (1) LiteralList(1,1);
    };
  

  Clause* result = new Clause(lit_list);
  lit_list->firstElem()->setLiteral(shareTmpLiteral(lit,result));
 

  LiteralProperties prop = lit.properties();
  
  if (_inheritNegativeSelection && 
      lit.isNegativeForSelection() &&
      !lit.isAnswerLiteral())
    {
      prop.setSelectabilityCategory(LiteralProperties::HighSelectability);
      prop.setSelectabilityPriority(LiteralProperties::MaximalSelectabilityPriority);
      prop.setInfiniteSelectabilityInheritanceLimit();
    };
  lit_list->firstElem()->properties() = prop;
  ancestors.writeNewBackgroundTo(result->background()); 
  result->setSize(lit.size());
  result->setWeightBase(weightBase);
  if (lit.isEquality() && lit.isPositive()) 
    { 
      result->SetNumOfPosEq(1); 
    }
  else 
    result->SetNumOfPosEq(0); 

  if (lit.isEquality() && lit.isPositive())
    { 
      result->setPracticallyMinResidualSize(lit.minPosEqLHSSize() + 1); 
    }
  else 
    result->setPracticallyMinResidualSize(0);

  result->setFlag(Clause::FlagLiteralsShared);
  if (isAnswer) result->setFlag(Clause::FlagContainsAnswerLiterals);
  return result;
}; // Clause* ClauseAssembler::assembleUnitClause(TemporaryLiteral& lit,OpenClauseBackground& ancestors)




Clause* ClauseAssembler::assembleClause(BasicTempLitList& lits,OpenClauseBackground& ancestors)
{
  CALL("assembleClause(BasicTempLitList& lits,OpenClauseBackground& ancestors)");
  ASSERT(lits.checkIntegrity());
  
  if (lits.isEmpty()) return assembleEmptyClause(ancestors);
  long size;
  long weightBase;
  long num_of_pos_eq;
  if (_equalityIsPossible)
    {
      TempLitListExtras::computeSizeAndWeightBaseAndNumOfPosEq(lits,size,weightBase,num_of_pos_eq);
    }
  else
    { 
      TempLitListExtras::computeSizeAndWeightBase(lits,size,weightBase);
      num_of_pos_eq = 0;
    }; 
   
  BasicTempLitList max_lits;
  BasicTempLitList nonmax_lits; 
  BasicTempLitList tmp_lits;
  BasicTempLitList* ordered_max_lits;
  BasicTempLitList* ordered_nonmax_lits; 

  DOP(ulong debugLitsLength = lits.length());
  bool allMaximalSelected = false;
  if (_eliminatedPredicateTable && 
      _litSelection->selectForElimination(lits,max_lits,nonmax_lits,_eliminatedPredicateTable))
    {
      // nothing to do here 
    }
  else
    if (ancestors.isSupport())
      {
	_litSelection->selectForSupport(lits,max_lits,nonmax_lits);
      }
    else 
      {
	// order literals as usual
	_litSelection->select(lits,max_lits,nonmax_lits,allMaximalSelected);
      };
  // don't use lits anymore! it must be empty here!

  ASSERT(lits.isEmpty());
  ASSERT((!max_lits.isEmpty()) || 
	 TempLitListExtras::containsOnlyAnswerLiterals(nonmax_lits));  


  bool posEqSelected = false;
  if (num_of_pos_eq)
    {
      for (const BasicTempLitList::Element* el = max_lits.first();
	   el && !posEqSelected;
	   el = el->next())
	{
	  posEqSelected = 
	    el->literal()->isEquality() &&
	    el->literal()->isPositive() &&
	    el->literal()->properties().useEqForParamodulationFrom();
	};
    };

  long weight = 
    Clause::Context::getCurrent()->computeWeight(weightBase,
						 ancestors.isSubgoal(),
						 posEqSelected);

  if (weight > _maxWeight) return 0;


  
  ulong num_of_max = max_lits.length();
  ulong num_of_nonmax = nonmax_lits.length();

  ASSERT(debugLitsLength == num_of_max + num_of_nonmax);
  if (num_of_max > 1)
    {
      _varWeighting.Reset();
      const BasicTempLitList::Element* el;
      for (el = max_lits.first(); el; el = el->next())
	_varWeighting.AnotherTermVariables(el->literal()->variableOccurences());  
      for (el = nonmax_lits.first(); el; el = el->next())
	_varWeighting.AnotherTermVariables(el->literal()->variableOccurences());  

      TempLitListExtras::passInOrder(max_lits,tmp_lits,_varWeighting);
      ordered_max_lits = &tmp_lits;
      if (num_of_nonmax > 1) 
	{
	  ASSERT(max_lits.isEmpty());
	  TempLitListExtras::passInOrder(nonmax_lits,max_lits,_varWeighting);
	  ordered_nonmax_lits = &max_lits; // this is a hack, not an error!
	}
      else 
	ordered_nonmax_lits = &nonmax_lits;  
    }
  else
    {
      ordered_max_lits = &max_lits;
      if (num_of_nonmax > 1)
	{
	  _varWeighting.Reset();
	  const BasicTempLitList::Element* el;
	  for (el = max_lits.first(); el; el = el->next())
	    _varWeighting.AnotherTermVariables(el->literal()->variableOccurences());  
	  for (el = nonmax_lits.first(); el; el = el->next())
	    _varWeighting.AnotherTermVariables(el->literal()->variableOccurences());  
	  
	  TempLitListExtras::passInOrder(nonmax_lits,tmp_lits,_varWeighting);
	  ordered_nonmax_lits = &tmp_lits;
	}
      else 
	{
	  ordered_nonmax_lits = &nonmax_lits;
	};
    };  
 
  _intVarRenaming.Reset();
  // order of renaming is significant!
  TempLitListExtras::renameVariables(*ordered_max_lits,_intVarRenaming);
  TempLitListExtras::renameVariables(*ordered_nonmax_lits,_intVarRenaming);
  
  long weight_base_of_nonmax = 
    TempLitListExtras::computeWeightBase(*ordered_nonmax_lits);
  
  long min_size_of_max_pos_eq_lhs = 0;
  if (_equalityIsPossible)
    {
      min_size_of_max_pos_eq_lhs = 
	TempLitListExtras::minimalPosEqLHSSize(*ordered_max_lits);
      if (min_size_of_max_pos_eq_lhs) min_size_of_max_pos_eq_lhs++;
    };
  
  LiteralList* lit_list = 
    new(num_of_max + num_of_nonmax) 
    LiteralList(num_of_max + num_of_nonmax,num_of_max);

  Clause* result = new Clause(lit_list);

  LiteralList::Iterator iter(lit_list);
  TERM* lit;
  LitHandle tmpLit;
  bool containsAnswerLiterals = false;
  while (num_of_max)
    {
      ordered_max_lits->popTo(tmpLit);
      lit = shareTmpLiteral(*tmpLit,result);
      ASSERT(iter.notEnd());  
      iter.currentElement()->setLiteral(lit);
      ASSERT(!lit->isAnswerLiteral());
      LiteralProperties prop = tmpLit->properties();

      if (_inheritNegativeSelection &&
	  tmpLit->isNegativeForSelection())
	{	  
	  prop.setSelectabilityCategory(LiteralProperties::HighSelectability);
	  prop.setSelectabilityPriority(LiteralProperties::MaximalSelectabilityPriority);
	  prop.setInfiniteSelectabilityInheritanceLimit();
	};
 
      iter.currentElement()->properties() = prop;
      iter.next();
      num_of_max--;
    };
           
  
  while (num_of_nonmax)
    {
      ordered_nonmax_lits->popTo(tmpLit);
      lit = shareTmpLiteral(*tmpLit,result);
      ASSERT(iter.notEnd());  
      iter.currentElement()->setLiteral(lit);   
      containsAnswerLiterals = containsAnswerLiterals || lit->isAnswerLiteral(); 
      iter.currentElement()->properties() = tmpLit->properties();
      iter.next();
      num_of_nonmax--;
    };

  ASSERT(iter.end());  
    
  ancestors.writeNewBackgroundTo(result->background());
  if (allMaximalSelected)
    result->setFlag(Clause::FlagAllMaximalLiteralsSelected);
  result->setSize(size);
  result->setWeightBase(weightBase);
  result->SetNumOfPosEq(num_of_pos_eq);
  result->setPracticallyMinResidualSize(weight_base_of_nonmax + min_size_of_max_pos_eq_lhs);
  result->setFlag(Clause::FlagLiteralsShared);
  if (containsAnswerLiterals) result->setFlag(Clause::FlagContainsAnswerLiterals);

  return result; 
}; // Clause* ClauseAssembler::assembleClause(BasicTempLitList& lits,OpenClauseBackground& ancestors)




Clause* ClauseAssembler::assembleClauseWithoutSelection(BasicTempLitList& lits,
							OpenClauseBackground& ancestors)
{
  CALL("assembleClauseWithoutSelection(BasicTempLitList& lits,OpenClauseBackground& ancestors)");
  ASSERT(lits.checkIntegrity());
  
  if (lits.isEmpty()) return assembleEmptyClause(ancestors);
  long size;
  long weightBase;
  long num_of_pos_eq;
  if (_equalityIsPossible)
    {
      TempLitListExtras::computeSizeAndWeightBaseAndNumOfPosEq(lits,size,weightBase,num_of_pos_eq);
    }
  else
    { 
      TempLitListExtras::computeSizeAndWeightBase(lits,size,weightBase);
      num_of_pos_eq = 0;
    };    

  bool allMaximalSelected = true;

  bool posEqSelected = false;
  if (num_of_pos_eq)
    {
      for (const BasicTempLitList::Element* el = lits.first();
	   el && !posEqSelected;
	   el = el->next())
	{
	  posEqSelected = 
	    el->literal()->isEquality() &&
	    el->literal()->isPositive() &&
	    el->literal()->properties().useEqForParamodulationFrom();
	};
    };

  _intVarRenaming.Reset();
  TempLitListExtras::renameVariables(lits,_intVarRenaming);

  long min_size_of_max_pos_eq_lhs = 0;
  if (_equalityIsPossible)
    {
      min_size_of_max_pos_eq_lhs = 
	TempLitListExtras::minimalPosEqLHSSize(lits);
      if (min_size_of_max_pos_eq_lhs) min_size_of_max_pos_eq_lhs++;
    };
  
  LiteralList* lit_list = 
    new(lits.length()) 
    LiteralList(lits.length(),lits.length());

  Clause* result = new Clause(lit_list);

  LiteralList::Iterator iter(lit_list);
  TERM* lit;
  LitHandle tmpLit;
  bool containsAnswerLiterals = false;
  while (!lits.isEmpty())
    {
      lits.popTo(tmpLit);

      lit = shareTmpLiteral(*tmpLit,result);
      ASSERT(iter.notEnd());  
      iter.currentElement()->setLiteral(lit);
      iter.currentElement()->properties() = tmpLit->properties();
      iter.next();
    };
           
  ASSERT(iter.end());  
    
  ancestors.writeNewBackgroundTo(result->background());
  if (allMaximalSelected)
    result->setFlag(Clause::FlagAllMaximalLiteralsSelected);
  result->setSize(size);
  result->setWeightBase(weightBase);
  result->SetNumOfPosEq(num_of_pos_eq);
  result->setPracticallyMinResidualSize(min_size_of_max_pos_eq_lhs);
  result->setFlag(Clause::FlagLiteralsShared);
  if (containsAnswerLiterals) result->setFlag(Clause::FlagContainsAnswerLiterals);
  return result; 
}; // Clause* ClauseAssembler::assembleClauseWithoutSelection(BasicTempLitList& lits,OpenClauseBackground& ancestors)



Clause* ClauseAssembler::assembleUnsharedClause(BasicTempLitList& lits,OpenClauseBackground& ancestors)
{
  CALL("assembleUnsharedClause(BasicTempLitList& lits,OpenClauseBackground& ancestors)");
  ulong num_of_lits = lits.length();
  long size;
  long weightBase;
  TempLitListExtras::computeSizeAndWeightBase(lits,size,weightBase);
  LiteralList* lit_list = new (num_of_lits) LiteralList(num_of_lits,num_of_lits);
  Clause* result = new Clause(lit_list);
          
  bool containsAnswerLiterals = false;
  LitHandle tmpLit;
  for(LiteralList::Iterator iter(lit_list); iter.notEnd(); iter.next())
    { 
      lits.popTo(tmpLit);
      tmpLit->markAsSelected();
      TERM* lit = assembleUnsharedLit(*tmpLit);
      iter.currentElement()->setLiteral(lit);
      containsAnswerLiterals = containsAnswerLiterals || lit->isAnswerLiteral();  
      iter.currentElement()->properties() = tmpLit->properties();
    };


  ancestors.writeNewBackgroundTo(result->background()); 
  result->setSize(size);
  result->setWeightBase(weightBase);
  result->SetNumOfPosEq(0);
  result->setPracticallyMinResidualSize(0);  
  if (containsAnswerLiterals) result->setFlag(Clause::FlagContainsAnswerLiterals);
  return result;  
}; // Clause* ClauseAssembler::assembleUnsharedClause(BasicTempLitList& lits,OpenClauseBackground& ancestors) 




TERM* ClauseAssembler::assembleUnsharedLit(const TemporaryLiteral& lit) 
{
  CALL("assembleUnsharedLit(const TemporaryLiteral& lit)");

  _openUnsharedTerm.reset();
  TERM header(lit.header());
  _openUnsharedTerm.litHeader(header);
  if (header.IsEquality())
    {
      Flatterm::ConstIterator iter(lit.arg1());
      while (iter.NotEndOfTerm())
	{
	  _openUnsharedTerm.symbol(iter.CurrPos()->symbolRef()); 
	  iter.NextSym();
	};     
      iter.Reset(lit.arg2());
      while (iter.NotEndOfTerm())
	{
	  _openUnsharedTerm.symbol(iter.CurrPos()->symbolRef()); 
	  iter.NextSym();
	};  
    }
  else
    {
      Flatterm::ConstIterator iter(lit.flatterm());
      iter.NextSym();
      while (iter.NotEndOfTerm())
	{
	  _openUnsharedTerm.symbol(iter.CurrPos()->symbolRef()); 
	  iter.NextSym();
	};
    };
  return _openUnsharedTerm.result().First();
}; // TERM* ClauseAssembler::assembleUnsharedLit(const TemporaryLiteral& lit) 





bool 
ClauseAssembler::assembleClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
							   BasicTempLitList& lits,
							   OpenClauseBackground& ancestors,
							   long clauseNum,
							   ostream* showNewClause,
							   bool markAsName)
{
  CALL("assembleClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,BasicTempLitList& lits,OpenClauseBackground& ancestors,long clauseNum,ostream* showNewClause,bool markAsName");
  ASSERT(!lits.isEmpty());

  long size;
  long weightBase; 
  TempLitListExtras::computeSizeAndWeightBase(lits,size,weightBase);
  
  bool posEqHasToBeSelected = true;
  for (const BasicTempLitList::Element* el = lits.first();
       el && posEqHasToBeSelected;
       el = el->next())
    {
      posEqHasToBeSelected = 
	el->literal()->isEquality() &&
	el->literal()->isPositive() &&
	el->literal()->properties().useEqForParamodulationFrom();
    };
  long weight = 
    Clause::Context::getCurrent()->computeWeight(weightBase,
						 ancestors.isSubgoal(),
						 posEqHasToBeSelected);
  if (weight > _maxWeight) return false;

  BitWord rules = ancestors.Rules();
  

  if (markAsName) rules.set(ClauseBackground::Name);
  ulong numOfLits = lits.length();
  ulong numOfAncestors = ancestors.numberOfAncestors();
  PagedPassiveClauseStorage::StoredClauseHandle newClauseHandle;
 try_to_open: 
  storage.openClause(newClauseHandle,clauseNum,size,weight,rules,numOfLits,numOfAncestors);
  if (!newClauseHandle)
    {
      // try to cancel some bad clauses in storage
      if (storage.cancelWorstClause())
	goto try_to_open;
      return false;
    };
 

  // now write the literals 
  writeIntoPagedPassiveClauseStorage(storage,lits);
 
  // now write the ancestors


  if (ancestors.numberOfAncestors())
    for (Clause* const * anc = ancestors.last(); anc >= ancestors.begin(); anc--)
      {
	storage.writeAncestor(*anc); 
      };

  storage.endOfClause();

  if (showNewClause)
    {
      PagedPassiveClauseStorage::StoredClause newClause;
      PagedPassiveClauseStorage::loadClause(newClauseHandle,newClause);
      newClause.output(*showNewClause << "% ") << "\n";
    };
  return true;
}; // bool ClauseAssembler::assembleClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,BasicTempLitList& lits,OpenClauseBackground& ancestors,long clauseNum,ostream* showNewClause,bool markAsName)



void 
ClauseAssembler::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
						    BasicTempLitList& lits)
{
  CALL("writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,BasicTempLitList& lits)"); 

  for (const BasicTempLitList::Element* el = lits.first(); 
       el;
       el = el->next())
    writeIntoPagedPassiveClauseStorage(storage,el->literal());   

}; // void ClauseAssembler::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,BasicTempLitList& lits)

void 
ClauseAssembler::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
						    const LitHandle& lit)
{
  CALL("writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,const LitHandle& lit)");
  LiteralProperties prop = lit->properties();
  prop.makeSelected();
  storage.openLiteral(prop);
  TERM header = lit->header();
  Flatterm::Iterator iter;
  if (header.IsEquality()) 
    {
      storage.writeSymbol(header);
      iter.Reset(lit->arg1());
      do
	{
	  storage.writeSymbol(iter.CurrPos()->symbolRef());          
	  iter.NextSym(); 
	}
      while (iter.NotEndOfTerm());
      iter.Reset(lit->arg2());
      do
	{
	  storage.writeSymbol(iter.CurrPos()->symbolRef());          
	  iter.NextSym();
	}
      while (iter.NotEndOfTerm());
    }
  else 
    {
      iter.Reset(lit->flatterm());
      do
	{
	  storage.writeSymbol(iter.CurrPos()->symbolRef());          
	  iter.NextSym();
	}
      while (iter.NotEndOfTerm());
    };
  storage.closeLiteral(); 
};  // void ClauseAssembler::writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,const LitHandle& lit)

TERM* ClauseAssembler::shareLiteralCopy(TERM* lit)
{
  CALL("shareLiteralCopy(TERM* lit)");
  DTREE_NODE* leaf = _litSharing->Integrate(lit);
  leaf->NumOfClauses()++;
  return leaf->Term();
}; // TERM* ClauseAssembler::shareLiteralCopy(TERM* lit)




//============================================================================
