//
// File:         CustomLiteralSelection.hpp
// Description:  Flexible schemes for literal selection.
// Created:      Jun 03, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//============================================================================
#ifndef CUSTOM_LITERAL_SELECTION_H
//============================================================================
#define CUSTOM_LITERAL_SELECTION_H 
#include "jargon.hpp"
#include "RuntimeError.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "StandardLiteralOrdering.hpp"
#include "ExpandingStack.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    typedef BasicTempLitList LitList;
    typedef BasicTempLitList::Element LitListElement;
    typedef TempLitListExtras::Characteristics LitListStatistics;
  public:
    typedef long (*NumericLiteralCharacteristic)(TemporaryLiteral&);
    typedef bool (*LiteralProperty)(TemporaryLiteral&);
    


    class ComplexLiteralCharacteristic;
    class InverseVarDepthProfile;
    // More shallow occurences of variables
    // lead to bigger instances of InverseVarDepthProfile 
    // in the corresponding ordering.
    
    


    class Filter;
    class MaxValFilter;
    class MinValFilter;
    class MaxComplexValFilter;
    class AllPOrAllFilter;
    class AllPThroughF1OrAllThroughF2;
    class CompositionFilter;
    class OneOfEachPolarityFilter;
    
    class SelectionAcceptability;
    class ConjOfPrimitiveSelectionAcceptability;

  public:
    CustomLiteralSelection() {};
    virtual ~CustomLiteralSelection();
    void setStandardLiteralOrdering(StandardLiteralOrdering* ord) { _standardLiteralOrdering = ord; };
    void select(LitList& lits,
		LitList& selectedLits,
		LitList& unselectedLits,
		bool& allMaximalSelected,
		const LitListStatistics& stat); 
   
    // numerical literal characteristics 

    static long literalPriority(TemporaryLiteral& lit);

    static long literalSize(TemporaryLiteral& lit);

    static long numOfDifferentVariables(TemporaryLiteral& lit);

    // literal properties 

    static bool isGround(TemporaryLiteral& lit);

    static bool isNegativeEquality(TemporaryLiteral& lit);

    
    static bool isNotPositiveEquality(TemporaryLiteral& lit);

    static bool isNotAvailableForParamodFrom(TemporaryLiteral& lit);


    // selection properties

    static bool oneLiteralSelected(LitListElement* firstSelected,
				   LitListElement* firstUnselected); 

    static 
    bool noNegativeWithHighPriorityUnselected(LitListElement* firstSelected,
					      LitListElement* firstUnselected); 
    

    static bool noBigNegativeUnselected(LitListElement* firstSelected,
					LitListElement* firstUnselected); 



    static bool noGroundNegativeUnselected(LitListElement* firstSelected,
					   LitListElement* firstUnselected); 

    static bool noNegativeEqualitiesUnselected(LitListElement* firstSelected,
					       LitListElement* firstUnselected); 

    static bool noNegativeWithManyDiffVarsUnselected(LitListElement* firstSelected,
						      LitListElement* firstUnselected);

    

  protected:
    void selectMaximal(LitList& lits,
		       LitList& selectedLits,
		       LitList& unselectedLits);

    void selectOneMaximal(LiteralOrdering* litOrd,
		       LitList& lits,
		       LitList& selectedLits,
		       LitList& unselectedLits);


    void unselectPositive(LitList& lits,LitList& unselectedLits,ulong numOfPositive);

  protected:
    StandardLiteralOrdering* _standardLiteralOrdering;

  private:
    virtual
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat) = 0;
    
  }; // class CustomLiteralSelection 
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ComplexLiteralCharacteristic"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{
  class CustomLiteralSelection::ComplexLiteralCharacteristic
  {
  public:
    ComplexLiteralCharacteristic() {};
    virtual ~ComplexLiteralCharacteristic();
    virtual void computeFor(TemporaryLiteral& lit) = 0;
    virtual BK::Comparison compareWith(ComplexLiteralCharacteristic* c) = 0;
  }; // class CustomLiteralSelection::ComplexLiteralCharacteristic
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::InverseVarDepthProfile"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{
  class CustomLiteralSelection::InverseVarDepthProfile : public ComplexLiteralCharacteristic
  {
  public:
    InverseVarDepthProfile() {};
    ~InverseVarDepthProfile();
    void computeFor(TemporaryLiteral& lit);
    BK::Comparison compareWith(ComplexLiteralCharacteristic* c);
  private:
    TemporaryLiteral::VarDepthProfile _data;
  }; // class CustomLiteralSelection::ComplexLiteralCharacteristic
}; // namespace VK





//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::Filter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::Filter
  {
  public:
    Filter() : _alwaysSelectOne(false) {};
    virtual ~Filter();
    void setSelectOne(bool fl) { _alwaysSelectOne = fl; };
    void select(LitList& lits,
		LitList& selectedLits,
		LitList& unselectedLits)
    {
      CALL("select(..)");
      ASSERT(!lits.isEmpty());
      DOP(ulong debugNumOfLits = lits.length());
      DOP(ulong debugNumOfSelectedLits = selectedLits.length());
      DOP(ulong debugNumOfUnselected = unselectedLits.length());

      if (_alwaysSelectOne)
	{
	  selectOne(lits,selectedLits,unselectedLits);
	  ASSERT(selectedLits.length() - debugNumOfSelectedLits == 1UL);
	}
      else
	{
	  selectAll(lits,selectedLits,unselectedLits);
	  ASSERT(selectedLits.length() > debugNumOfSelectedLits);
	};
      ASSERT(lits.isEmpty());
      ASSERT(selectedLits.length() + 
	     unselectedLits.length() 
	     ==
	     debugNumOfLits + 
	     debugNumOfSelectedLits + 
	     debugNumOfUnselected);
    };
  private:
    virtual void selectOne(LitList& lits,
			   LitList& selectedLits,
			   LitList& unselectedLits) = 0;
    virtual void selectAll(LitList& lits,
			   LitList& selectedLits,
			   LitList& unselectedLits) = 0;
  private:
    bool _alwaysSelectOne;
    friend class CustomLiteralSelection::CompositionFilter;
    friend class CustomLiteralSelection::AllPThroughF1OrAllThroughF2;
  }; // class CustomLiteralSelection::Filter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::MaxValFilter : 
    public Filter
  {
  public:
    MaxValFilter(NumericLiteralCharacteristic val) : _val(val)
    {
    };
    ~MaxValFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    NumericLiteralCharacteristic _val;
  }; // class CustomLiteralSelection::MaxValFilter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MinValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::MinValFilter : 
    public Filter
  {
  public:
    MinValFilter(NumericLiteralCharacteristic val) : _val(val)
    {
    };
    ~MinValFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    NumericLiteralCharacteristic _val;
  }; // class CustomLiteralSelection::MinValFilter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxComplexValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::MaxComplexValFilter : 
    public Filter
  {
  public:
    MaxComplexValFilter(ComplexLiteralCharacteristic* tmpVal1,
			ComplexLiteralCharacteristic* tmpVal2) : 
      _tmpVal1(tmpVal1),
      _tmpVal2(tmpVal2)
    {
    };
    ~MaxComplexValFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    ComplexLiteralCharacteristic* _tmpVal1;
    ComplexLiteralCharacteristic* _tmpVal2;
  }; // class CustomLiteralSelection::MaxComplexValFilter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::AllPOrAllFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::AllPOrAllFilter : 
    public Filter
  {
  public:
    AllPOrAllFilter(LiteralProperty prop) : _prop(prop)
    {
    };
    ~AllPOrAllFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    LiteralProperty _prop;
  }; // class CustomLiteralSelection::AllPOrAllFilter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::AllPThroughF1OrAllThroughF2"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::AllPThroughF1OrAllThroughF2 : 
    public Filter
  {
  public:
    AllPThroughF1OrAllThroughF2(LiteralProperty prop,
				Filter* filter1,
				Filter* filter2) : 
      _prop(prop),
      _filter1(filter1),
      _filter2(filter2)
    {
    };
    ~AllPThroughF1OrAllThroughF2();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    LiteralProperty _prop;
    Filter* _filter1;
    Filter* _filter2;
  }; // class CustomLiteralSelection::AllPThroughF1OrAllThroughF2
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::CompositionFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::CompositionFilter : 
    public Filter
  {
  public:
    CompositionFilter() 
    {
    };
    ~CompositionFilter();
    void resetFilters() { _filters.reset(); };
    void addFilter(Filter* f) { _filters.pushSafe(f); };
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  private:
    BK::ExpandingStack<BK::GlobAlloc,Filter*,BK::FastRelocation<Filter*>,10UL> _filters;
  }; // class CustomLiteralSelection::CompositionFilter
}; // namespace VK



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::OneOfEachPolarityFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::OneOfEachPolarityFilter : 
    public Filter
  {
  public:
    OneOfEachPolarityFilter() 
    {
    };
    ~OneOfEachPolarityFilter();
  private:    
    void selectOne(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
    void selectAll(LitList& lits,
		   LitList& selectedLits,
		   LitList& unselectedLits);
  }; // class CustomLiteralSelection::OneOfEachPolarityFilter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::SelectionAcceptability"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::SelectionAcceptability
  {
  public:
    SelectionAcceptability() {};
    virtual ~SelectionAcceptability();
    virtual bool accepts(LitListElement* firstSelected,
			 LitListElement* firstUnselected) = 0;
  }; // class CustomLiteralSelection::SelectionAcceptability
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability :
    public SelectionAcceptability
  {
  public:
    ConjOfPrimitiveSelectionAcceptability() 
    {
    };
    ~ConjOfPrimitiveSelectionAcceptability();
    bool accepts(LitListElement* firstSelected,
		 LitListElement* firstUnselected);
    void resetPrimitives() { _primitives.reset(); };
    void addPrimitive(bool (*fun)(LitListElement*,LitListElement*)) 
    {
      _primitives.pushSafe(fun);
    };
  private:
    typedef bool (*Primitive)(LitListElement*,LitListElement*);
  private:
    BK::ExpandingStack<BK::GlobAlloc,Primitive,BK::FastRelocation<Primitive>,10UL> _primitives;
  }; // class CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "TotalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class TotalLiteralSelection : public CustomLiteralSelection
  {
  public:
    TotalLiteralSelection() {};
    ~TotalLiteralSelection() {};
  private:    
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  }; // class TotalLiteralSelection
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaximalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class MaximalLiteralSelection : public CustomLiteralSelection
  {
  public:
    MaximalLiteralSelection() {};
    ~MaximalLiteralSelection() {};
  private:    

    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  }; // class MaximalLiteralSelection
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "HyperLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class HyperLiteralSelection : public CustomLiteralSelection
  {
  public:
    HyperLiteralSelection() {};
    ~HyperLiteralSelection() {};
    void setFilterForNegative(CustomLiteralSelection::Filter* f)
    {
      _filterForNegative = f;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filterForNegative;    
  }; // class HyperLiteralSelection
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaxWithCorrectionLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class MaxWithCorrectionLiteralSelection : public CustomLiteralSelection
  {
  public:
    MaxWithCorrectionLiteralSelection() 
    {
      CALL("constructor MaxWithCorrectionLiteralSelection()");
      DOP(_filterForNegative = 0);
      DOP(_maxSelectionAcceptability = 0);
    };
    ~MaxWithCorrectionLiteralSelection() {};
    void setFilterForNegative(CustomLiteralSelection::Filter* f)
    {
      _filterForNegative = f;
    };
    void setMaxSelectionAcceptability(CustomLiteralSelection::SelectionAcceptability* test)
    {
      _maxSelectionAcceptability = test;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::SelectionAcceptability* _maxSelectionAcceptability;
    CustomLiteralSelection::Filter* _filterForNegative;
  }; // class MaxWithCorrectionLiteralSelection
}; // namespace VK



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "BestOnlyLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class BestOnlyLiteralSelection : public CustomLiteralSelection
  {
  public:
    BestOnlyLiteralSelection() {};
    ~BestOnlyLiteralSelection() {};
    void setLiteralQualityOrdering(LiteralOrdering* o)
    {
      _litOrdering = o;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat); 
  private:
    LiteralOrdering* _litOrdering;
  }; // class BestOnlyLiteralSelection
}; // namespace VK



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "ArbitraryLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class ArbitraryLiteralSelection : public CustomLiteralSelection
  {
  public:
    ArbitraryLiteralSelection() {};
    ~ArbitraryLiteralSelection() {};
    void setFilter(CustomLiteralSelection::Filter* f)
    {
      _filter = f;
    };
  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filter;    
  }; // class ArbitraryLiteralSelection
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CompleteBidirectionalLiteralSelection : public CustomLiteralSelection
  {
  public:
    CompleteBidirectionalLiteralSelection() {};
    ~CompleteBidirectionalLiteralSelection() {};
    void setFilterForPositiveInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForPositiveInRules = f;
    };
    void setFilterForNegativeInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeInRules = f;
    };
    void setFilterForNegativeFacts(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeFacts = f;
    };

  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filterForPositiveInRules; 
    CustomLiteralSelection::Filter* _filterForNegativeInRules;    
    CustomLiteralSelection::Filter* _filterForNegativeFacts;    
    LitList _buffer;
  }; // class CompleteBidirectionalLiteralSelection
}; // namespace VK





//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "IncompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class IncompleteBidirectionalLiteralSelection : public CustomLiteralSelection
  {
  public:
    IncompleteBidirectionalLiteralSelection() {};
    ~IncompleteBidirectionalLiteralSelection() {};
    void setFilterForPositiveInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForPositiveInRules = f;
    };
    void setFilterForNegativeInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeInRules = f;
    };
    void setFilterForPositiveFacts(CustomLiteralSelection::Filter* f)
    {
      _filterForPositiveFacts = f;
    };
    void setFilterForNegativeFacts(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeFacts = f;
    };

  private:  
    void 
    _select(LitList& lits,
	    LitList& selectedLits,
	    LitList& unselectedLits,
	    bool& allMaximalSelected,
	    const LitListStatistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filterForPositiveInRules; 
    CustomLiteralSelection::Filter* _filterForNegativeInRules;    
    CustomLiteralSelection::Filter* _filterForPositiveFacts;      
    CustomLiteralSelection::Filter* _filterForNegativeFacts;    
    LitList _buffer;
  }; // class IncompleteBidirectionalLiteralSelection
}; // namespace VK






//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{
  inline
  void 
  CustomLiteralSelection::selectMaximal(LitList& lits,
					LitList& selectedLits,
					LitList& unselectedLits)
  {
    CALL("selectMaximal(..)");
    LitListElement* el;
    LitListElement* tmp;

    if (_standardLiteralOrdering->isStratified())
      {
	// filter out literals from nonmaximal layers
	LitListElement* maxLayerRepresentative = lits.last();  
      
	for (el = lits.last()->previous(); el; /* nothing here */)
	  {
	    switch (_standardLiteralOrdering->compareLayers(*(el->literal()),
							    *(maxLayerRepresentative->literal())))
	      {
	      case BK::Greater:
		maxLayerRepresentative = el;	      
		do 
		  { 
		    lits.popTo(unselectedLits);
		  } 
		while (lits.last() != el);
		el = el->previous();
		break;

	      case BK::Less:
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
		break;

	      case BK::Equal:
		el = el->previous();
		break;
	      
#ifdef DEBUG_NAMESPACE
	      default: ICP("ICP0"); break;
#else
#  ifdef _SUPPRESS_WARNINGS_
	      default: 
		BK::RuntimeError::report("Wrong BK::Comparison value in CustomLiteralSelection::selectMaximal(..)");
		break;
#  endif
#endif	     
	      }; // switch ..
	  }; //  for ..    
      };

    // all remaining literals belong to the maximal layer
  
    LitHandle cand; 

  next_candidate:
    lits.popTo(cand);
    if (lits.isEmpty()) 
      { 
	selectedLits.collect(cand); 
	return; 
      };
    for (el = lits.last(); el; /* nothing here */)
      {      
	switch (_standardLiteralOrdering->compare(*(el->literal()),*cand))
	  {
	  case BK::Greater:
	    unselectedLits.collect(cand); 
	    goto next_candidate;

	  case BK::Less:
	    tmp = el->previous();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	    break;
 
	  case BK::Equal: // same as below

	  case BK::Incomparable: 
	    el = el->previous();
	    break;

#ifdef DEBUG_NAMESPACE
	  default: ICP("ICP1"); break;
#else
#  ifdef _SUPPRESS_WARNINGS_
	  default: 
	    BK::RuntimeError::report("Wrong BK::Comparison value in CustomLiteralSelection::selectMaximal(..)");
	    break;
#  endif
#endif	     
	  }; // switch ..
      }; // for (el = lits.last();...

    selectedLits.collect(cand);
    if (lits.isEmpty()) return;
    goto next_candidate;
  }; // void CustomLiteralSelection::selectMaximal(..)



  inline
  void 
  CustomLiteralSelection::selectOneMaximal(LiteralOrdering* litOrd,
					   LitList& lits,
					   LitList& selectedLits,
					   LitList& unselectedLits)
  {
    CALL("selectOneMaximal(LiteralOrdering* litOrd,..)");

    LitListElement* el;
    LitListElement* tmp;
    if (litOrd->isStratified())
      {
	// filter out literals from nonmaximal layers
	LitListElement* maxLayerRepresentative = lits.last();  
      
	for (el = lits.last()->previous(); el; /* nothing here */)
	  {
	    switch (litOrd->compareLayers(*(el->literal()),
					  *(maxLayerRepresentative->literal())))
	      {
	      case BK::Greater:
		maxLayerRepresentative = el;	      
		do 
		  { 
		    lits.popTo(unselectedLits);
		  } 
		while (lits.last() != el);
		el = el->previous();
		break;

	      case BK::Less:
		tmp = el->previous();
		lits.passTo(el,unselectedLits);
		el = tmp;
		break;

	      case BK::Equal:
		el = el->previous();
		break;
	      
#ifdef DEBUG_NAMESPACE
	      default: ICP("ICP0"); break;
#else
#  ifdef _SUPPRESS_WARNINGS_
	      default: 
		BK::RuntimeError::report("Wrong BK::Comparison value in CustomLiteralSelection::selectMaximal(..)");
		break;
#  endif
#endif	     
	      }; // switch ..
	  }; //  for ..    
      };

    // all remaining literals belong to the maximal layer
  
    LitHandle cand;

  next_candidate:
    lits.popTo(cand);
    if (lits.isEmpty()) 
      { 
	selectedLits.collect(cand); 
	return; 
      };
    
    for (el = lits.last(); el; /* nothing here */)
      {      
	switch (litOrd->compare(*(el->literal()),*cand))
	  {
	  case BK::Greater:
	    unselectedLits.collect(cand); 
	    goto next_candidate;

	  case BK::Less: // same as below
	  case BK::Equal: // same as below
	  case BK::Incomparable:
	    tmp = el->previous();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	    break;
	   

#ifdef DEBUG_NAMESPACE
	  default: ICP("ICP1"); break;
#else
#  ifdef _SUPPRESS_WARNINGS_
	  default: 
	    BK::RuntimeError::report("Wrong BK::Comparison value in CustomLiteralSelection::selectMaximal(..)");
	    break;
#  endif
#endif	     
	  }; // switch ..
      }; // for (el = lits.last(); ...

    selectedLits.collect(cand);
    if (lits.isEmpty()) return;
    goto next_candidate;
  }; // void CustomLiteralSelection::selectOneMaximal(LiteralOrdering* litOrd,..)




  
  inline
  void CustomLiteralSelection::unselectPositive(LitList& lits,LitList& unselectedLits,ulong numOfPositive)
  {
    CALL("unselectPositive(..)");    
    for (LitListElement* el = lits.first(); numOfPositive; /* nothing here */)
      {
	ASSERT(el);
	if (el->literal()->isPositiveForSelection())
	  {
	    LitListElement* tmp = el->next();
	    lits.passTo(el,unselectedLits);
	    el = tmp;
	    numOfPositive--;
	  }
	else
	  el = el->next();
      };
  }; // void CustomLiteralSelection::unselectPositive(..)


}; // namespace VK


//============================================================================
#endif
