//
// File:         ForwardSubsumptionFrontEnd.hpp
// Description:  An interface to the forward subsumption core.
// Created:      Dec 15, 2001 
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef FORWARD_SUBSUMPTION_FRONT_END_H
//============================================================================
#define FORWARD_SUBSUMPTION_FRONT_END_H 
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "forward_subsumption.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "ClauseBackground.hpp"
#include "LiteralList.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_FRONT_END
#define DEBUG_NAMESPACE "ForwardSubsumptionFrontEnd"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
class ForwardSubsumptionFrontEnd : public ForwardSubsumption
{
 public: 
  ForwardSubsumptionFrontEnd() 
    : _subsumptionResolutionMinLength(0UL) 
    {
      unlock();
    };
  ~ForwardSubsumptionFrontEnd() {};
  
  void reset()
  {
    CALL("reset()");
    ForwardSubsumption::reset();
    _subsumptionResolutionMinLength = 0UL; 
    unlock();
  };
  
  bool isLocked() const { return _isLocked; };

  void setSubsumptionResolutionMinLength(ulong ln) 
    {
      _subsumptionResolutionMinLength = ln;
    };
  void integrate(Clause* cl)
  {
    CALL("integrate(Clause* cl)");
    ASSERT(!isLocked());
    lock();
    cl->protect();
    ForwardSubsumption::Integrate(cl);
    cl->unprotect();
    unlock();
  };
  void remove(Clause* cl)
  {
    CALL("remove(Clause* cl)");
    ASSERT(!isLocked());
    lock();
    cl->protect();
    ForwardSubsumption::Remove(cl);
    cl->unprotect();
    unlock();
  };
  Clause* subsume(BasicTempLitList& lits); 
  ulong simplifyBySubsumptionResolution(BasicTempLitList& lits,OpenClauseBackground& ancestors);


 private: 
  void lock() { _isLocked = true; };
  void unlock() { _isLocked = false; };
  bool pushLiteral(const TemporaryLiteral::Extras::Handle& lit);
 private:
  ulong _subsumptionResolutionMinLength;
  bool _isLocked;
}; // class ForwardSubsumptionFrontEnd

}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION_FRONT_END
#define DEBUG_NAMESPACE "ForwardSubsumptionFrontEnd"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
inline Clause* ForwardSubsumptionFrontEnd::subsume(BasicTempLitList& lits) 
{
  CALL("subsume(BasicTempLitList& lits)");
  ASSERT(!isLocked());
  lock();
  
  ForwardSubsumption::ResetQuery(); 
  
  for (BasicTempLitList::Element* el = lits.first(); el; el = el->next())
    {  
      if (!pushLiteral(el->literal()))
	{
	  unlock();
	  return static_cast<Clause*>(0);
	};
    };
  Clause* subsumingClause = ForwardSubsumption::Subsume();
  if (subsumingClause) subsumingClause->MakeVIP();
  unlock();
  return subsumingClause;
}; //Clause* ForwardSubsumptionFrontEnd::subsume(BasicTempLitList& lits)  

inline bool ForwardSubsumptionFrontEnd::pushLiteral(const TemporaryLiteral::Extras::Handle& lit)
{
  CALL("pushLiteral(const TemporaryLiteral::Extras::Handle& lit)");
  if (lit->isEquality())
    {       
      lit->linkEqArgs();

      if (lit->isOrderedEquality())
	{
	  return ForwardSubsumption::PushOrdEq(lit->header(),lit->arg1(),lit->arg2());
	}
      else 
	return ForwardSubsumption::PushUnordEq(lit->header(),lit->arg1(),lit->arg2()); 
    }
  else 
    return ForwardSubsumption::PushLit(lit->flatterm()); 
}; // bool ForwardSubsumptionFrontEnd::pushLiteral(const TemporaryLiteral::Extras::Handle& lit)




inline ulong ForwardSubsumptionFrontEnd::simplifyBySubsumptionResolution(BasicTempLitList& lits,OpenClauseBackground& ancestors)
{
  CALL("simplifyBySubsumptionResolution(BasicTempLitList& lits,OpenClauseBackground& ancestors)");
  ASSERT(!isLocked());
  lock();
  if (lits.length() < _subsumptionResolutionMinLength) 
    {
      unlock();
      return 0UL;
    };
  ulong res = 0UL;
  

  BasicTempLitList::Element* candEl = lits.last();

 try_lit: 
  if (candEl->literal()->isEquality() ||
      candEl->literal()->isAnswerLiteral())
    {
      if (candEl == lits.first()) 
	{
	  unlock();
	  return res;
	};
      candEl = candEl->previous();
      goto try_lit;
    };

  // try to simplify by resolving cand_lit
  candEl->literal()->flatterm()->symbolRef().Inversepolarity();
   
  ResetQuery(); 

  for (BasicTempLitList::Element* el = lits.first(); el; el = el->next())
    if (!pushLiteral(el->literal())) 
      {
	unlock();
	return 0UL;
      };
            
  Clause* simplifier = Subsume();
  if (simplifier)
    {
      simplifier->MakeVIP();
      res++;
      ancestors.pushAncestor(simplifier,simplifier->inferenceDepth());
      if (simplifier->isSubgoal()) ancestors.markAsSubgoal();
      simplifier->setFlag(Clause::FlagHasChildren);
      
      
      BasicTempLitList::Element* tmp = candEl;
      candEl = candEl->next();
      lits.remove(tmp);
      
      if ((lits.length() < _subsumptionResolutionMinLength) 
	  || (candEl == lits.first()))
	{
	  unlock();
	  return res;
	};
      if (candEl) { candEl = candEl->previous(); }
      else
	candEl = lits.last();
      goto try_lit;  
    }
  else 
    {
      candEl->literal()->flatterm()->symbolRef().Inversepolarity();
      if (candEl == lits.first()) 
	{
	  unlock();
	  return res;
	};
      candEl = candEl->previous(); 
      goto try_lit;
    }; 
}; // ulong ForwardSubsumptionFrontEnd::simplifyBySubsumptionResolution(BasicTempLitList& lits,OpenClauseBackground& ancestors)


}; // namespace VK



//======================================================================
#endif
