//
// File:         LiteralProperties.cpp
// Description:  Various literal labels.
// Created:      Feb 13, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#include "LiteralProperties.inl"
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_PROPERTIES
#  define DEBUG_NAMESPACE "LiteralProperties"
#endif
#include "debugMacros.hpp"
//===================================================================


LiteralProperties::LiteralProperties(const VampireKernel::LiteralAnnotation& annot) :
  _data(0UL)
{
  CALL("constructor LiteralProperties(const VampireKernel::LiteralAnnotation& annot)");
  if (annot.deploymentMode() == VampireKernel::LiteralAnnotation::RuleHead)
    {
      setDeploymentMode(RuleHead);
    }
  else
    if (annot.deploymentMode() == VampireKernel::LiteralAnnotation::Subquery)
      {
	setDeploymentMode(Subquery);
      };
  if (annot.equalityMode() == VampireKernel::LiteralAnnotation::AddResolution)
    {
      setEqualityMode(UseEqForResolution);
    }
  else
    if (annot.equalityMode() == VampireKernel::LiteralAnnotation::ResolutionOnly)
      {
	setEqualityMode(UseEqForResolutionOnly);
      };
  setSelectabilityPriority(annot.selectabilityPriority());
  ulong inhLim;
  if (annot.selectabilityInheritanceLimit().isFinite(inhLim))
    setFiniteSelectabilityInheritanceLimit(inhLim);
  if (annot.selectabilityCategory() == VampireKernel::LiteralAnnotation::LowSelectability)
    {
      setSelectabilityCategory(LowSelectability);
    }
  else
    if (annot.selectabilityCategory() == VampireKernel::LiteralAnnotation::HighSelectability)
      {
	setSelectabilityCategory(HighSelectability);
      };
}; // LiteralProperties::LiteralProperties(const VampireKernel::LiteralAnnotation& annot)

void LiteralProperties::inheritFrom(const LiteralProperties& prop)
{
  CALL("inheritFrom(const LiteralProperties& prop)");
  make(prop.content());
  if (prop.isDefault()) return;

  ulong inhLim;
  if (prop.selectabilityInheritanceLimitIsFinite(inhLim))
    {
      if (inhLim)
	{
	  --inhLim;
	  setFiniteSelectabilityInheritanceLimit(inhLim);    
	}
      else // selectability is not inherited
	{
	  setFiniteSelectabilityInheritanceLimit(0UL);
	  setSelectabilityCategory(DefaultSelectability);
	  if (prop.selectabilityCategory() == HighSelectability)
	    {
	      setSelectabilityPriority(MaximalSelectabilityPriority);
	    }
	  else 
	    if (prop.selectabilityCategory() == LowSelectability)
	      {
		setSelectabilityPriority(MinimalSelectabilityPriority);
	      };
	};
    };
}; // void LiteralProperties::inheritFrom(const LiteralProperties& prop)


void LiteralProperties::inheritFrom(const LiteralProperties& prop,
				    const SLModifier& slmod)
{
  CALL("inheritFrom(const LiteralProperties& prop,const SLModifier& slmod)");  
  make(prop.content());
  if (prop.isDefault()) return;
  setSelectionLayerModifier(slmod);
  ulong inhLim;
  if (prop.selectabilityInheritanceLimitIsFinite(inhLim))
    {
      if (inhLim)
	{
	  --inhLim;
	  setFiniteSelectabilityInheritanceLimit(inhLim);    
	}
      else // selectability is not inherited
	{
	  setFiniteSelectabilityInheritanceLimit(0UL);
	  setSelectabilityCategory(DefaultSelectability);
	  if (prop.selectabilityCategory() == HighSelectability)
	    {
	      setSelectabilityPriority(MaximalSelectabilityPriority);
	    }
	  else 
	    if (prop.selectabilityCategory() == LowSelectability)
	      {
		setSelectabilityPriority(MinimalSelectabilityPriority);
	      };
	};
    };
}; // void LiteralProperties::inheritFrom(const LiteralProperties& prop,const SLModifier& slmod)


void LiteralProperties::factoriseWith(const LiteralProperties& prop)
{
  CALL("factoriseWith(const LiteralProperties& prop)");
  if (isDefault()) 
    {
      bool selected = isSelected() || prop.isSelected();
      _data = prop._data;
      setFlagSelected(selected);
      return;
    };
  if (prop.isDefault())
    {
      setFlagSelected(isSelected() || prop.isSelected());
      return;
    };
    
    
  if (selectionLayer() > prop.selectionLayer())
    setSelectionLayer(prop.selectionLayer());

  if (deploymentMode() < prop.deploymentMode())
    setDeploymentMode(prop.deploymentMode());

  if (equalityMode() < prop.equalityMode())
    setEqualityMode(prop.equalityMode());

  if (selectionLayerModifier() < prop.selectionLayerModifier())
    setSelectionLayerModifier(prop.selectionLayerModifier());
    
  if (prop.isSelected())
    setFlagSelected(true);
   
  if (compareSelectabilityForFactoring(prop) == BK::Less)
    {
      setSelectabilityPriority(prop.selectabilityPriority());
      ulong inhLim;
      if (prop.selectabilityInheritanceLimitIsFinite(inhLim))
	{
	  setFiniteSelectabilityInheritanceLimit(inhLim);
	}
      else
	setInfiniteSelectabilityInheritanceLimit();
      setSelectabilityCategory(prop.selectabilityCategory());
    };
}; // void LiteralProperties::factoriseWith(const LiteralProperties& prop)
  

ostream& LiteralProperties::output(ostream& str) const
{
  CALL("output(ostream& str) const");

  if (isDefault()) return str;

  
  ulong numOfNondefaultEntries = 0UL;
  if (selectionLayer()) ++numOfNondefaultEntries;
  if (deploymentMode() != DefaultDeploymentMode) 
    ++numOfNondefaultEntries;
  if (equalityMode() != DefaultEqualityMode) 
    ++numOfNondefaultEntries;
  if (selectionLayerModifier() != DefaultSLModifier)
    ++numOfNondefaultEntries;
  if (selectabilityPriority()) ++numOfNondefaultEntries;
  if (selectabilityInheritanceLimitIsFinite()) 
    ++numOfNondefaultEntries;
  if (selectabilityCategory() != 
      DefaultSelectability) 
    ++numOfNondefaultEntries;
      
  if (selectionLayer())
    {
      str << "sl(" << selectionLayer() << ")";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
      
  if (deploymentMode() != DefaultDeploymentMode) 
    {
      if (deploymentMode() == RuleHead)
	{
	  str << "rh";
	}
      else
	str << "sq";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
      
  if (equalityMode() != DefaultEqualityMode) 
    {
      if (equalityMode() == UseEqForResolution)
	{
	  str << "addres";
	}
      else
	str << "resonly";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
      

  if (selectionLayerModifier() != DefaultSLModifier)
    {
      if (selectionLayerModifier() == AccelerateSelection)
	{
	  str << "accsel";
	}
      else
	str << "postp";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };


  if (selectabilityCategory() != DefaultSelectability) 
    {
      if (selectabilityCategory() == LowSelectability)
	{
	  str << "lo";
	}
      else
	str << "hi";
	  
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };

  if (selectabilityPriority())
    {
      str << "pri(" << selectabilityPriority() << ")";
      --numOfNondefaultEntries;
      if (numOfNondefaultEntries) str << ",";
    };
  ulong inhLim;
  if (selectabilityInheritanceLimitIsFinite(inhLim))
    {
      ASSERT(numOfNondefaultEntries == 1UL);
      str << "inh(" << inhLim << ")";
    };

  
  return str;
}; // ostream& LiteralProperties::output(ostream& str) const

//=================================================================

