//
// File:         LiteralProperties.hpp
// Description:  Various literal labels.
// Created:      Feb 13, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef LITERAL_PROPERTIES_H
#  define LITERAL_PROPERTIES_H
//===================================================================
#include <iostream>
#include <climits>                          // for CHAR_BIT
#include "jargon.hpp"
#include "Comparison.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernel.hpp"
using namespace std;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_PROPERTIES
#  define DEBUG_NAMESPACE "LiteralProperties"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class LiteralProperties
  {
  private:

    enum 
    {
      // all sizes are in bits
      _SizeOfULong = sizeof(ulong)*CHAR_BIT,


      _SizeOfSelectionLayer = 8,
      _SizeOfDeploymentMode = 2,
      _SizeOfSLModifier = 2,
      _SizeOfEqualityMode = 2,
      _SizeOfFlagSelected = 1,
      _SizeOfSelectabilityPrioritySignBit = 1,
      _SizeOfSelectabilityPriority = 8,
      _SizeOfSelectabilityInheritanceLimit = 5,
      _SizeOfSelectabilityCategory = 2,
      _SizeOfReserved = 
      _SizeOfULong -
      (_SizeOfSelectionLayer + 
       _SizeOfDeploymentMode + 
       _SizeOfSLModifier +
       _SizeOfFlagSelected +
       _SizeOfSelectabilityPrioritySignBit +
       _SizeOfSelectabilityPriority +
       _SizeOfSelectabilityInheritanceLimit +
       _SizeOfSelectabilityCategory)
    };

    enum
    {
      // layout
      _BeforeSelectionLayer = 0,
      _AfterSelectionLayer = _SizeOfULong - _SizeOfSelectionLayer,
      _WithoutSelectionLayer = _BeforeSelectionLayer + _AfterSelectionLayer,

      _BeforeDeploymentMode = _SizeOfSelectionLayer,
      _AfterDeploymentMode = _SizeOfULong - (_BeforeDeploymentMode + _SizeOfDeploymentMode),
      _WithoutDeploymentMode = _BeforeDeploymentMode + _AfterDeploymentMode,


      _BeforeSLModifier = _BeforeDeploymentMode + _SizeOfDeploymentMode,
      _AfterSLModifier = _SizeOfULong - (_BeforeSLModifier + _SizeOfSLModifier),
      _WithoutSLModifier = _BeforeSLModifier + _AfterSLModifier,



      _BeforeEqualityMode = _BeforeSLModifier + _SizeOfSLModifier,
      _AfterEqualityMode = _SizeOfULong - (_BeforeEqualityMode + _SizeOfEqualityMode),
      _WithoutEqualityMode = _BeforeEqualityMode + _AfterEqualityMode,


      _BeforeFlagSelected = _BeforeEqualityMode + _SizeOfEqualityMode,
      _AfterFlagSelected = _SizeOfULong - (_BeforeFlagSelected + _SizeOfFlagSelected),
      _WithoutFlagSelected = _BeforeFlagSelected + _AfterFlagSelected,

      _BeforeSelectabilityPrioritySignBit = _BeforeFlagSelected + _SizeOfFlagSelected,
      _AfterSelectabilityPrioritySignBit = _SizeOfULong - (_BeforeSelectabilityPrioritySignBit + _SizeOfSelectabilityPrioritySignBit),
      _WithoutSelectabilityPrioritySignBit =  _BeforeSelectabilityPrioritySignBit + _AfterSelectabilityPrioritySignBit,

      _BeforeSelectabilityPriority = _BeforeSelectabilityPrioritySignBit + _SizeOfSelectabilityPrioritySignBit,
      _AfterSelectabilityPriority = _SizeOfULong - (_BeforeSelectabilityPriority + _SizeOfSelectabilityPriority),
      _WithoutSelectabilityPriority = _BeforeSelectabilityPriority + _AfterSelectabilityPriority,

      _BeforeSelectabilityInheritanceLimit = _BeforeSelectabilityPriority + _SizeOfSelectabilityPriority,
      _AfterSelectabilityInheritanceLimit = _SizeOfULong - (_BeforeSelectabilityInheritanceLimit + _SizeOfSelectabilityInheritanceLimit),
      _WithoutSelectabilityInheritanceLimit = _BeforeSelectabilityInheritanceLimit + _AfterSelectabilityInheritanceLimit,
    
      _BeforeSelectabilityCategory = _SizeOfULong - _SizeOfSelectabilityCategory,
      _AfterSelectabilityCategory = 0,
      _WithoutSelectabilityCategory = _BeforeSelectabilityCategory + _AfterSelectabilityCategory
    };


  public:
    enum DeploymentMode
    {
      DefaultDeploymentMode = 0UL,
      RuleHead = 1UL,
      Subquery = 2UL
    };

    enum SLModifier // selection layer modifier
    {
      DefaultSLModifier = 0UL,
      PostponeSelection = 1UL,
      AccelerateSelection = 2UL
    };
  
    enum EqualityMode
    {
      DefaultEqualityMode = 0UL,
      UseEqForResolution = 1UL,
      UseEqForResolutionOnly = 2UL
    };

    enum SelectabilityCategory
    {
      DefaultSelectability = 0UL,
      LowSelectability = 1UL,
      HighSelectability = 2UL
    };

    enum 
    {
      MaximalSelectionLayer = 255UL,
      MinimalSelectabilityPriority = -255L,
      MaximalSelectabilityPriority = 255L,
      MaximalFiniteSelectabilityInheritanceLimit = 254UL    
    };
  

  public:
    LiteralProperties() : _data(0UL) {};
    explicit LiteralProperties(const VampireKernel::LiteralAnnotation& annot);
    explicit LiteralProperties(ulong data) : _data(data) {};
    void makeDefault() { _data = 0UL; };
    void make(ulong data) { _data = data; };
    ulong content() const { return _data; };

    void setSelectionLayer(ulong l);
    void setDeploymentMode(const DeploymentMode& m);
    void resetDeploymentMode() { setDeploymentMode(DefaultDeploymentMode); };
    void makeRuleHead() { setDeploymentMode(RuleHead); };
    void makeSubquery() { setDeploymentMode(Subquery); };
    void setEqualityMode(const EqualityMode& m);
    void setSelectionLayerModifier(const SLModifier& m);
    void setFlagSelected(bool fl);
    void makeSelected();
    void makeUnselected();
    void setSelectabilityPriority(long p);
    void setFiniteSelectabilityInheritanceLimit(ulong l);
    void setInfiniteSelectabilityInheritanceLimit();
    void setSelectabilityCategory(const SelectabilityCategory& c);

    bool isDefault() const 
    { 
      return 
	(_data == 0UL) || (_data == (1UL << _AfterFlagSelected)); 
    };
    ulong selectionLayer() const;
    DeploymentMode deploymentMode() const;
    EqualityMode equalityMode() const;
    bool useEqForResolution() const;
    bool useEqForParamodulationFrom() const;
    SLModifier selectionLayerModifier() const;
    bool isSelected() const;
    long selectabilityPriority() const;
    bool selectabilityInheritanceLimitIsInfinite() const;
    bool selectabilityInheritanceLimitIsFinite() const;
    bool selectabilityInheritanceLimitIsFinite(ulong& limit) const;
    SelectabilityCategory selectabilityCategory() const;

    void inheritFrom(const LiteralProperties& prop);
    void inheritFrom(const LiteralProperties& prop,
		     const SLModifier& slmod);
    void factoriseWith(const LiteralProperties& prop);

    ostream& output(ostream& str) const;

  private:
    BK::Comparison compareSelectabilityForFactoring(const LiteralProperties& prop) const;
    void setSelectabilityPrioritySignBit(bool fl);
    bool selectabilityPrioritySignBit() const;
  private: 
    // layout: 
    // [ _SizeOfSelectionLayer ]                            // left-most bits     
    // [ _SizeOfDeploymentMode ]  
    // [ _SizeOfSLModifier ] 
    // [ _SizeOfEqualityMode ] 
    // [ _SizeOfFlagSelected ]   
    // [ _SizeOfSelectabilityPrioritySignBit ] 
    // [ _SizeOfSelectabilityPriority ] 
    // [ _SizeOfSelectabilityInheritanceLimit ] 
    // [ _SizeOfReserved ] 
    // [ _SizeOfSelectabilityCategory ]                    // right-most bits
  
    ulong _data;
  }; // class LiteralProperties

}; // namespace VK

namespace std
{
  inline ostream& operator<<(ostream& str,const VK::LiteralProperties& prop)
  {
    return prop.output(str);
  }; 
};

//===================================================================

#endif
