//
// File:         NegativeEqualitySplitting.cpp
// Description:  Core module for negative equality splitting.
// Created:      Nov 07, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#include "NegativeEqualitySplitting.hpp"
#include "BasicTempLitList.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryLiteral.inl"
#include "TemporaryClauseCollector.hpp"
#include "ComponentName.hpp"
#include "Clause.inl"
using namespace BK;
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEGATIVE_EQUALITY_SPLITTING
#  define DEBUG_NAMESPACE "NegativeEqualitySplitting"
#endif
#include "debugMacros.hpp"
//===================================================================


NegativeEqualitySplitting::NegativeEqualitySplitting(ComponentName* componentNaming) :
  _minimalEligibleArgSize(0L),
  _maximalEligibleVarCount(0L),
  _componentNaming(componentNaming)   
{    
  CALL("constructor NegativeEqualitySplitting(ComponentName* componentNaming)");
  TemporaryLiteral::Extras::freshLiteral(_definedLiteral);
  TemporaryLiteral::Extras::freshLiteral(_definitionBodyLiteral);
  _definitionBody.push(_definitionBodyLiteral);
};


NegativeEqualitySplitting::~NegativeEqualitySplitting()
{
  CALL("destructor ~NegativeEqualitySplitting()");
}; // NegativeEqualitySplitting::~NegativeEqualitySplitting()
  


void NegativeEqualitySplitting::reset()
{
  CALL("reset()");
  _minimalEligibleArgSize = 0L;
}; // void NegativeEqualitySplitting::reset()


bool
NegativeEqualitySplitting::splitOffAppropriateNegEqArguments(TemporaryClause& cl,
							     TemporaryClauseCollector* newClauseCollector,
							     ulong& numOfSplitOffArgs,
							     ulong& numOfNewNames)
{
  CALL("splitOffAppropriateNegEqArguments(..)");
  numOfSplitOffArgs = 0UL;
  numOfNewNames = 0UL;
  bool newName;
  Clause* nameIntroClause;
  for (BasicTempLitList::Element* el = cl.literals()->first();
       el;
       el = el->next())
    {

      if (el->literal()->isEquality() &&
	  el->literal()->isNegative() &&
	  !el->literal()->properties().useEqForResolution() &&
	  splitOffAppropriateArgument(el->literal(),newClauseCollector,newName,nameIntroClause))
	{
	  ++numOfSplitOffArgs;
	  if (newName)
	    ++numOfNewNames;
	  cl.background().UsedRule(ClauseBackground::Folding);
	  cl.background().pushAncestor(nameIntroClause,0L);
	};
      
    };
  return (numOfSplitOffArgs > 0UL);
}; // bool NegativeEqualitySplitting::splitOffAppropriateNegEqArguments(..)



bool 
NegativeEqualitySplitting::splitOffAppropriateArgument(LitHandle& lit,
						       TemporaryClauseCollector* newClauseCollector,
						       bool& newNameWasIntroduced,
						       Clause*& nameIntroClause)
{
  CALL("splitOffAppropriateArgument(..)");  

  ASSERT(lit->isEquality());
  ASSERT(lit->isNegative());
  LiteralProperties prop;
  LitHandle replacementLiteral;
  bool arg1Eligible;
  bool arg2Eligible;
  bool name1Exists;
  bool name2Exists;
  ulong name1;
  ulong name2;
  Clause* nameIntroClause1;
  Clause* nameIntroClause2;

  long numOfDiffVarsInArg1 = lit->numOfDifferentVariablesInArg1();
  long numOfDiffVarsInArg2 = lit->numOfDifferentVariablesInArg2();
  TERM::Arity name1Arity = 1 + numOfDiffVarsInArg1;
  TERM::Arity name2Arity = 1 + numOfDiffVarsInArg2;


  arg1Eligible = 
    numOfDiffVarsInArg1 <= _maximalEligibleVarCount &&
    (lit->arg1()->size() >= _minimalEligibleArgSize);
  arg2Eligible = 
    numOfDiffVarsInArg2 <= _maximalEligibleVarCount &&
    (lit->arg2()->size() >= _minimalEligibleArgSize);
  

  if (arg1Eligible)
    {
      if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(replacementLiteral))
	return false;

      if (arg2Eligible)
        {
	  // both eligible        
	  name1Exists = _componentNaming->findName(lit->arg1(),nameIntroClause1);
	  name2Exists = _componentNaming->findName(lit->arg2(),nameIntroClause2);
	  if (name1Exists)
	    {
	      name1 = nameIntroClause1->firstLiteral()->functor();
	      if (name2Exists)
		{ 
		  // both arguments have names
		  name2 = nameIntroClause2->firstLiteral()->functor();
		  if (lit->arg1()->size() > lit->arg2()->size())
		    {  
		      replacementLiteral->pushPseudoLit(TERM(name1Arity,(TERM::Functor)name1,Signature::NegativePolarity),
							lit->arg2(),
							lit->arg1Weight().monomials(),
							lit->properties());
		      nameIntroClause = nameIntroClause1;
		    }          
		  else // lit->arg1()->size() <= lit->arg2()->size()
		    {
		      replacementLiteral->pushPseudoLit(TERM(name2Arity,(TERM::Functor)name2,Signature::NegativePolarity),
						  
							lit->arg1(),
							lit->arg2Weight().monomials(),
							lit->properties());
		      nameIntroClause = nameIntroClause2;
		    };
		}
	      else // name1Exists && !name2Exits
		{
		  replacementLiteral->pushPseudoLit(TERM(name1Arity,(TERM::Functor)name1,Signature::NegativePolarity),
						    lit->arg2(),
						    lit->arg1Weight().monomials(),
						    lit->properties());
		  nameIntroClause = nameIntroClause1;
		};
	   
	      lit.collectFrom(replacementLiteral);
	      newNameWasIntroduced = false;
	      return true;
	    }
	  else // !name1Exists 
	    if (name2Exists)
	      { 
		name2 = nameIntroClause2->firstLiteral()->functor();
		replacementLiteral->pushPseudoLit(TERM(name2Arity,(TERM::Functor)name2,Signature::NegativePolarity),
						  lit->arg1(),
						  lit->arg2Weight().monomials(),
						  lit->properties());
		nameIntroClause = nameIntroClause2;

		lit.collectFrom(replacementLiteral);
		newNameWasIntroduced = false;
		return true;
	      }
	    else // !name1Exists && !name2Exits
	      {
		// a new name must be created            
		if (lit->arg1()->size() > lit->arg2()->size())
		  {
		    if (!Signature::current()->createNewSkolemPredicateNumber(name1Arity,name1))
		      return false; 
		    if (name1 > TERM::maxFunctor()) return false;


		    TERM posHeader(name1Arity,(TERM::Functor)name1,Signature::PositivePolarity); 
		    TERM negHeader(name1Arity,(TERM::Functor)name1,Signature::NegativePolarity);


		    Signature::current()->assignMaximalSymbolPrecedence(name1);
		    Signature::current()->assignMaximalHeaderPrecedence(posHeader.polarity(),posHeader.functor());

		    Signature::current()->assignMaximalHeaderPrecedence(negHeader.polarity(),negHeader.functor());
			
		    for (ulong argNum = 1UL; // Not 0UL!
			 argNum < name1Arity;
			 ++argNum)
		      {
			Signature::current()->markNonBasicArgument(true,name1,argNum);
			Signature::current()->markNonBasicArgument(false,name1,argNum);
		      };

		    _newVariable->SetVar(variableOutside(lit->arg1Weight().monomials()));

		    
		    if (!_definedLiteral->pushPseudoLit(posHeader,
							_newVariable,
							lit->arg1Weight().monomials(),
							prop))
		      {
			return false;
		      };
		    if (!_definitionBodyLiteral->pushBinaryLit(TERM::positiveUnordEqHeader(),
							       _newVariable,
							       lit->arg1(),
							       prop))
		      {
			return false;
		      };
		    
		    nameIntroClause1 = 
		      _componentNaming->registerNameIntro(lit->arg1(),
							  *_definedLiteral,
							  _definitionBody);
		    
		    newClauseCollector->openNewClause(ClauseBackground::Prepro,nameIntroClause1);
			
		    newClauseCollector->announceNewClause();
		    if (!newClauseCollector->pushPseudoLiteral(posHeader,
							       lit->arg1(),
							       lit->arg1Weight().monomials(),
							       prop))
		      return false;
		    if (!newClauseCollector->finaliseClause())
		      return false;
			
		    replacementLiteral->pushPseudoLit(negHeader,
						      lit->arg2(),
						      lit->arg1Weight().monomials(),
						      lit->properties());

		    nameIntroClause = nameIntroClause1;

		    lit.collectFrom(replacementLiteral);
		    newNameWasIntroduced = true;
		    return true;
		  } 
		else // lit->arg1()->size() <= lit->arg2()->size() 
		  {
		    if (!Signature::current()->createNewSkolemPredicateNumber(name2Arity,name2))
		      return false; 
		    if (name2 > TERM::maxFunctor()) return false;
		    
		    TERM posHeader(name2Arity,(TERM::Functor)name2,Signature::PositivePolarity); 
		    TERM negHeader(name2Arity,(TERM::Functor)name2,Signature::NegativePolarity);

		    Signature::current()->assignMaximalSymbolPrecedence(name2);
		    Signature::current()->assignMaximalHeaderPrecedence(posHeader.polarity(),posHeader.functor());

		    Signature::current()->assignMaximalHeaderPrecedence(negHeader.polarity(),negHeader.functor());

		    for (ulong argNum = 1UL; // Not 0UL!
			 argNum < name2Arity;
			 ++argNum)
		      {
			Signature::current()->markNonBasicArgument(true,name2,argNum);
			Signature::current()->markNonBasicArgument(false,name2,argNum);
		      };
				
		    
		    
		    _newVariable->SetVar(variableOutside(lit->arg2Weight().monomials()));

		    if (!_definedLiteral->pushPseudoLit(posHeader,
							_newVariable,
							lit->arg2Weight().monomials(),
							prop))
		      {
			return false;
		      };
		    if (!_definitionBodyLiteral->pushBinaryLit(TERM::positiveUnordEqHeader(),
							       _newVariable,
							       lit->arg2(),
							       prop))
		      {
			return false;
		      };

		    nameIntroClause2 = 
		      _componentNaming->registerNameIntro(lit->arg2(),
							  *_definedLiteral,
							  _definitionBody);

		    newClauseCollector->openNewClause(ClauseBackground::Prepro,nameIntroClause2);

		    newClauseCollector->announceNewClause();
		    if (!newClauseCollector->pushPseudoLiteral(posHeader,
							       lit->arg2(),
							       lit->arg2Weight().monomials(),
							       prop))
		      return false;
		    if (!newClauseCollector->finaliseClause())
		      return false;

		    replacementLiteral->pushPseudoLit(negHeader,
						      lit->arg1(),
						      lit->arg2Weight().monomials(),
						      lit->properties());

		    
		    nameIntroClause = nameIntroClause2;

		    lit.collectFrom(replacementLiteral);
		    newNameWasIntroduced = true;
		    return true;
		  };
	      };
        }
      else // only arg1Eligible
        {
	  name1Exists = _componentNaming->findName(lit->arg1(),nameIntroClause1);
	  if (name1Exists) // the old name can be used
	    {
	      name1 = nameIntroClause1->firstLiteral()->functor();
	      replacementLiteral->pushPseudoLit(TERM(name1Arity,(TERM::Functor)name1,Signature::NegativePolarity),
						lit->arg2(),
						lit->arg1Weight().monomials(),
						lit->properties());
	      
	      
	      nameIntroClause = nameIntroClause1;

	      lit.collectFrom(replacementLiteral);
	      newNameWasIntroduced = false;
	      return true;
	    }
	  else // a new name must be introduced
	    {
	      if (lit->arg2()->size() == 1) return false;
	      
	      if (!Signature::current()->createNewSkolemPredicateNumber(name1Arity,name1))
		return false; 
	      if (name1 > TERM::maxFunctor()) return false;

	      TERM posHeader(name1Arity,(TERM::Functor)name1,Signature::PositivePolarity); 
	      TERM negHeader(name1Arity,(TERM::Functor)name1,Signature::NegativePolarity);

	      Signature::current()->assignMaximalSymbolPrecedence(name1);
	      Signature::current()->assignMaximalHeaderPrecedence(posHeader.polarity(),posHeader.functor());

	      Signature::current()->assignMaximalHeaderPrecedence(negHeader.polarity(),negHeader.functor());

	      for (ulong argNum = 1UL; // Not 0UL!
		   argNum < name1Arity;
		   ++argNum)
		{
		  Signature::current()->markNonBasicArgument(true,name1,argNum);
		  Signature::current()->markNonBasicArgument(false,name1,argNum);
		};
		  
	      
	      _newVariable->SetVar(variableOutside(lit->arg1Weight().monomials()));

		    
	      if (!_definedLiteral->pushPseudoLit(posHeader,
						  _newVariable,
						  lit->arg1Weight().monomials(),
						  prop))
		{
		  return false;
		};
	      if (!_definitionBodyLiteral->pushBinaryLit(TERM::positiveUnordEqHeader(),
							 _newVariable,
							 lit->arg1(),
							 prop))
		{
		  return false;
		};
		    

	      nameIntroClause1 = 
		_componentNaming->registerNameIntro(lit->arg1(),
						    *_definedLiteral,
						    _definitionBody);

	      newClauseCollector->openNewClause(ClauseBackground::Prepro,nameIntroClause1);

	      newClauseCollector->announceNewClause();
	      if (!newClauseCollector->pushPseudoLiteral(posHeader,
							 lit->arg1(),
							 lit->arg1Weight().monomials(),
							 prop))
		return false;
	      if (!newClauseCollector->finaliseClause())
		return false;

	      replacementLiteral->pushPseudoLit(negHeader,
						lit->arg2(),
						lit->arg1Weight().monomials(),
						lit->properties());

	      
	      nameIntroClause = nameIntroClause1;

	      lit.collectFrom(replacementLiteral);
	      newNameWasIntroduced = true;
	      return true;
	    };
        };
    }
  else // !arg1Eligible
    if (arg2Eligible)
      {
	// only arg2Eligible 
	if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(replacementLiteral))
	  return false;   

        name2Exists = _componentNaming->findName(lit->arg2(),nameIntroClause2);     
        if (name2Exists) // the old name can be used
	  {
	    name2 = nameIntroClause2->firstLiteral()->functor();
	    replacementLiteral->pushPseudoLit(TERM(name2Arity,(TERM::Functor)name2,Signature::NegativePolarity),
					      lit->arg1(),
					      lit->arg2Weight().monomials(),
					      lit->properties());

	    nameIntroClause = nameIntroClause2;

	    lit.collectFrom(replacementLiteral);
	    newNameWasIntroduced = false;
	    return true;
	  }
        else // a new name must be introduced
	  {
	    if (lit->arg1()->size() == 1) return false;
	    
	    if (!Signature::current()->createNewSkolemPredicateNumber(name2Arity,name2))
	      return false; 
	    if (name2 > TERM::maxFunctor()) return false;
		    

	    TERM posHeader(name2Arity,(TERM::Functor)name2,Signature::PositivePolarity); 
	    TERM negHeader(name2Arity,(TERM::Functor)name2,Signature::NegativePolarity);

	    Signature::current()->assignMaximalSymbolPrecedence(name2);
	    Signature::current()->assignMaximalHeaderPrecedence(posHeader.polarity(),posHeader.functor());
            
	    Signature::current()->assignMaximalHeaderPrecedence(negHeader.polarity(),negHeader.functor());

	    for (ulong argNum = 1UL; // Not 0UL!
		 argNum < name2Arity;
		 ++argNum)
	      {
		Signature::current()->markNonBasicArgument(true,name2,argNum);
		Signature::current()->markNonBasicArgument(false,name2,argNum);
	      };


	    _newVariable->SetVar(variableOutside(lit->arg2Weight().monomials()));

	    if (!_definedLiteral->pushPseudoLit(posHeader,
						_newVariable,
						lit->arg2Weight().monomials(),
						prop))
	      {
		return false;
	      };
	    if (!_definitionBodyLiteral->pushBinaryLit(TERM::positiveUnordEqHeader(),
						       _newVariable,
						       lit->arg2(),
						       prop))
	      {
		return false;
	      };

	    nameIntroClause2 = 
	      _componentNaming->registerNameIntro(lit->arg2(),
						  *_definedLiteral,
						  _definitionBody);

	    newClauseCollector->openNewClause(ClauseBackground::Prepro,nameIntroClause2);

	    newClauseCollector->announceNewClause();
	    if (!newClauseCollector->pushPseudoLiteral(posHeader,
						       lit->arg2(),
						       lit->arg2Weight().monomials(),
						       prop))
	      return false;
	    if (!newClauseCollector->finaliseClause())
	      return false;
		
	    replacementLiteral->pushPseudoLit(negHeader,
					      lit->arg1(),
					      lit->arg2Weight().monomials(),
					      lit->properties());

	    nameIntroClause = nameIntroClause2;

	    lit.collectFrom(replacementLiteral);
	    newNameWasIntroduced = true;
	    return true;
	  };
      }
    else // no argument is eligible
      return false;

  ICP("END");
  return false;
}; // bool NegativeEqualitySplitting::splitOffAppropriateArgument(..)

inline
ulong 
NegativeEqualitySplitting::variableOutside(const WeightPolynomial::Monomials& vars)
{
  CALL("variableOutside(const WeightPolynomial::Monomials& vars)");
  if (vars.isZero()) return 0UL;
  ulong cand = 0UL;
  while (vars.coefficient(cand) != 0UL)
    {
      ++cand;
    };
  return cand;
}; // ulong NegativeEqualitySplitting::variableOutside(const WeightPolynomial::Monomials& vars)


//=================================================================
