//
// File:         PagedPassiveClauseStorage.cpp
// Description:  Various signature related operations.  
// Created:      May 02, 2001, 13:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Math.hpp"
#include "PagedPassiveClauseStorage.inl"
#include "Clause.inl"
using namespace BK;
using namespace VK;
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PAGED_PASSIVE_CLAUSE_STORAGE
#define DEBUG_NAMESPACE "PagedPassiveClauseStorage"
#endif 
#include "debugMacros.hpp"
//=================================================  


PagedPassiveClauseStorage::PagedPassiveClauseStorage() 
{
  CALL("constructor PagedPassiveClauseStorage()");
  _ageWeightRatioNumerator = 1;
  _ageWeightRatioDenominator = 0;
  _bestGoodWeightNumerator = 1;
  _bestGoodWeightDenominator = 0;
  _maximalPriorityFlag = false;
  _minNonemptyCatNum = MaximalCategoryNumber + 1; 
  _maxNonemptyCatNum = -1;
  _totalSelectedByAgeOrWeight = 0;
  _totalSelectedByWeight = 0;
  _totalNumberOfClauses = 0;
  _numberOfSelectedClauses = 0;  
  _numberOfCancelledClauses = 0; 
  for (ulong cat = 0; cat <= MaximalCategoryNumber; cat++)
    {
      _category[cat].reset(&_freePages);
      _numberOfCurrentlyKept[cat] = 0L;
    };
  
#ifdef DEBUG_NAMESPACE
  _debugClauseOpenForWriting = false; 
#endif  
};  


PagedPassiveClauseStorage::~PagedPassiveClauseStorage() 
{
  CALL("destructor ~PagedPassiveClauseStorage()");
}; // PagedPassiveClauseStorage::~PagedPassiveClauseStorage()
  


void PagedPassiveClauseStorage::setAgeWeightRatio(long numerator,long denominator) 
{ 
  CALL("setAgeWeightRatio(long numerator,long denominator)");
  ASSERT(denominator >= 0);
  _ageWeightRatioNumerator = numerator;
  _ageWeightRatioDenominator = denominator;
  if (_ageWeightRatioNumerator < 0) _ageWeightRatioNumerator = 0;
  if (_ageWeightRatioDenominator < 0) _ageWeightRatioDenominator = 0;
  if (_ageWeightRatioNumerator == 0 && _ageWeightRatioDenominator == 0) 
    _ageWeightRatioNumerator = 1;
  if (_ageWeightRatioNumerator != 0 && _ageWeightRatioDenominator != 0) 
    {
      long gcd = 
	BK::Math::greatestCommonDivisor(_ageWeightRatioNumerator,_ageWeightRatioDenominator);
      _ageWeightRatioNumerator = _ageWeightRatioNumerator/gcd;
      _ageWeightRatioDenominator = _ageWeightRatioDenominator/gcd;
    };
  
};


void PagedPassiveClauseStorage::setBestGoodWeightRatio(long numerator,long denominator)
{
  CALL("setBestGoodWeightRatio(long numerator,long denominator)");
  _bestGoodWeightNumerator = numerator;
  _bestGoodWeightDenominator = denominator;
  if (_bestGoodWeightNumerator <= 0L) _bestGoodWeightNumerator = 1L;
  if (_bestGoodWeightDenominator < 0L) _bestGoodWeightDenominator = 0L;
};


void  
PagedPassiveClauseStorage::openClause(StoredClauseHandle& handle,
				      long clNum,
				      long clSize,
				      long clWeight,
				      const BitWord rules,
				      ulong numOfLits,
				      ulong numOfAncestors) 
{
  CALL("openClause(StoredClauseHandle& handle,long clNum,long clSize,long clWeight,const BitWord rules,ulong numOfLits,ulong numOfAncestors)");
  ASSERT(empty() || (_category[_minNonemptyCatNum].nonempty()));
  ASSERT(empty() || (_category[_maxNonemptyCatNum].nonempty()));

  long catNum = 
    (_maximalPriorityFlag || (!_ageWeightRatioDenominator)) ? 0L : categoryNum(clWeight);
  ASSERT(catNum >= 0L);
  ASSERT(catNum <= (long)MaximalCategoryNumber);
  _currentCategory = _category + catNum;
  ulong totalSize = sizeInWords(sizeof(ClauseProfile)) 
    + sizeInWords((clSize + 2*numOfLits)*sizeof(TERM)) 
    + numOfAncestors;

  _cursor = static_cast<void**>(_currentCategory->tryToAllocate(totalSize));
  if (!_cursor) 
    {
      handle.clear();
      return;
    };
  _cursor = static_cast<void**>(Storage::objectContent(static_cast<void*>(_cursor)));;   
  handle.set(static_cast<void*>(_cursor)); 
  DOP(_debugEndOfCurrentClause = _cursor + totalSize);
  DOP(_debugNumberOfLiteralsToWrite = numOfLits);
  DOP(_debugNumberOfAncestorsToWrite = numOfAncestors);
  DOP(_debugClauseOpenForWriting = true);
   
  (static_cast<ClauseProfile*>(static_cast<void*>(_cursor)))->_clauseNumber = clNum;
  (static_cast<ClauseProfile*>(static_cast<void*>(_cursor)))->_clauseSize = clSize;
  (static_cast<ClauseProfile*>(static_cast<void*>(_cursor)))->_clauseWeight = clWeight;
  (static_cast<ClauseProfile*>(static_cast<void*>(_cursor)))->_rules = rules;
  (static_cast<ClauseProfile*>(static_cast<void*>(_cursor)))->_numOfLiterals = numOfLits;
  (static_cast<ClauseProfile*>(static_cast<void*>(_cursor)))->_numOfAncestors = numOfAncestors;
  _cursor += sizeInWords(sizeof(ClauseProfile));
   
  if (empty()) 
    {
      _minNonemptyCatNum = catNum;
      _maxNonemptyCatNum = catNum; 
    }
  else
    {
      if (_minNonemptyCatNum > catNum) _minNonemptyCatNum = catNum;
      if (_maxNonemptyCatNum < catNum) _maxNonemptyCatNum = catNum;
    };

  ASSERT(!empty());
  ASSERT(_category[_minNonemptyCatNum].nonempty());
  ASSERT(_category[_maxNonemptyCatNum].nonempty());

  ++_totalNumberOfClauses;
  ++_numberOfCurrentlyKept[catNum];
}; //void  PagedPassiveClauseStorage::openClause(StoredClauseHandle& handle,long clNum,long clSize,long clWeight,const BitWord rules,ulong numOfLits,ulong numOfAncestors)  


bool PagedPassiveClauseStorage::select(StoredClauseHandle& selectedClauseHandle) 
{
  CALL("select(ClauseHandle& selectedClauseHandle)");
  if (empty()) return false;

  if (_minNonemptyCatNum == 0L) 
    {
      _selectedCategoryNumber = 0L;
      selectFrom(_category + _selectedCategoryNumber,selectedClauseHandle);
      _lastSelectedByWeight = false;
      _lastSelectedByAge = false;
      return true;
    };

  if ((_totalSelectedByAgeOrWeight % 
       (_ageWeightRatioNumerator + _ageWeightRatioDenominator)) >=
      _ageWeightRatioDenominator)
    {      
      // select by age
      long oldestCatNum = _minNonemptyCatNum;
      ASSERT(_category[_minNonemptyCatNum].nonempty());
      long oldestClauseNum = (static_cast<ClauseProfile*>(Storage::objectContent(_category[_minNonemptyCatNum].head())))->clauseNumber();
      for (long catNum = _minNonemptyCatNum + 1; catNum <= _maxNonemptyCatNum; catNum++)
	{
	  if (_category[catNum].nonempty())
	    {
	      if ((static_cast<ClauseProfile*>(Storage::objectContent(_category[catNum].head())))->clauseNumber() < oldestClauseNum)
		{
		  oldestCatNum = catNum;
		  oldestClauseNum = (static_cast<ClauseProfile*>(Storage::objectContent(_category[catNum].head())))->clauseNumber();     
		};
	    };
	};
      // now we know the oldest category 
      _selectedCategoryNumber =  oldestCatNum;   
      selectFrom(_category + _selectedCategoryNumber,selectedClauseHandle);
      _lastSelectedByWeight = false;
      _lastSelectedByAge = true;
    }   
  else // select by weight 
    {
      // determing the interval of eligible good weights
      long concession = 0UL;
      long tmp = _totalSelectedByWeight;
      while ((tmp % (_bestGoodWeightNumerator + _bestGoodWeightDenominator) >=
	      _bestGoodWeightNumerator) &&
	     _minNonemptyCatNum + concession < MaximalCategoryNumber)
	{
	  ++concession;
	  tmp = tmp / (_bestGoodWeightNumerator + _bestGoodWeightDenominator);
	};


      _selectedCategoryNumber = _minNonemptyCatNum + concession;
      while (_category[_selectedCategoryNumber].empty())
	{
	  ASSERT(_selectedCategoryNumber > _minNonemptyCatNum);
	  --_selectedCategoryNumber;
	};
      ASSERT(_category[_selectedCategoryNumber].nonempty());

      selectFrom(_category + _selectedCategoryNumber,selectedClauseHandle);
      _lastSelectedByWeight = true;
      _lastSelectedByAge = false;
    };
  ASSERT(selectedClauseHandle);
  return true;  
}; // bool PagedPassiveClauseStorage::select(ClauseHandle& selectedClauseHandle)   

bool PagedPassiveClauseStorage::select(StoredClause& selectedClause)
{
  CALL("select(StoredClause& selectedClause)");
  if (select(_selectedClauseHandle))
    {
      ASSERT((bool)_selectedClauseHandle);
      loadClause(_selectedClauseHandle,selectedClause); 
      return true;
    };
  return false;
}; // bool PagedPassiveClauseStorage::select(StoredClause& selectedClause)

void PagedPassiveClauseStorage::popSelected()
{
  CALL("popSelected()");
  ASSERT(_category[_selectedCategoryNumber].nonempty()); 
  ASSERT(_selectedClauseHandle.memory() == Storage::objectContent(_category[_selectedCategoryNumber].head()));
  _category[_selectedCategoryNumber].pop();
  _selectedClauseHandle.clear();

  // this category might be empty now 
  
  if (_category[_selectedCategoryNumber].empty())
    {
      if (_selectedCategoryNumber == _minNonemptyCatNum)
	{
	  do 
	    {
	      _minNonemptyCatNum++;     
	    }
	  while ((_minNonemptyCatNum <= _maxNonemptyCatNum) 
		 && (_category[_minNonemptyCatNum].empty()));
	}
      else
	if (_selectedCategoryNumber == _maxNonemptyCatNum)
	  {
	    do
	      {
		_maxNonemptyCatNum--;
	      }
	    while ((_maxNonemptyCatNum >= _minNonemptyCatNum) 
		   && (_category[_maxNonemptyCatNum].empty()));
	  };
    };

  ASSERT(empty() || (_category[_minNonemptyCatNum].nonempty()));
  ASSERT(empty() || (_category[_maxNonemptyCatNum].nonempty()));

  // statistics
  ++_numberOfSelectedClauses;
  --_numberOfCurrentlyKept[_selectedCategoryNumber];
  if (_lastSelectedByWeight || _lastSelectedByAge)
    ++_totalSelectedByAgeOrWeight;
  if (_lastSelectedByWeight) 
    {
      ++_totalSelectedByWeight; 
    };


}; // void PagedPassiveClauseStorage::popSelected()
       
// Cancelling clauses at the end of the queue and recycling pages. 



bool PagedPassiveClauseStorage::cancelWorstClause()
{
  CALL("cancelWorstClause()");

  if (empty()) return false;
  Storage* cat = _category + _maxNonemptyCatNum; 
  ASSERT(cat->nonempty()); 
  if (_selectedClauseHandle.memory() != Storage::objectContent(cat->endOfQueue()))
    {
      cat->cancelLastObject();
      ++_numberOfCancelledClauses;
      --_numberOfCurrentlyKept[_maxNonemptyCatNum];
      if (cat->empty())
	{
	  do
	    {
	      _maxNonemptyCatNum--;
	    }
	  while ((_maxNonemptyCatNum >= _minNonemptyCatNum) 
		 && (_category[_maxNonemptyCatNum].empty()));
	};
      return true;
    }
  else // the worst clause is locked as selected
    {
      --cat;
      while (cat >= _category + _minNonemptyCatNum)
	{
	  if (cat->nonempty())
	    {
	      cat->cancelLastObject();
	      ++_numberOfCancelledClauses;
	      --_numberOfCurrentlyKept[cat - _category];
	      // adjust _minNonemptyCatNum
	      while (_category[_minNonemptyCatNum].empty() &&
		     (_minNonemptyCatNum <= _maxNonemptyCatNum))
		++_minNonemptyCatNum;
	      return true;
	    }
	  else
	    --cat;
	};
      return false;
    };
}; // bool PagedPassiveClauseStorage::cancelWorstClause()


ostream& PagedPassiveClauseStorage::outputProfile(ostream& str) const
{
  CALL("outputProfile(ostream& str) const");
  str << "%============ Profile of Passive (weight:stored:oldest):";
  ulong columns = 0UL;
  for (long cat = _minNonemptyCatNum; cat <= _maxNonemptyCatNum; ++cat)
    {
      if (_category[cat].nonempty())
	{
	  if (columns == 0UL) 
	    {
	      str << "\n% ";
	      columns = 5UL;
	    };
	  str << cat << ":" <<  _numberOfCurrentlyKept[cat] 
	      << ":" <<  (static_cast<const ClauseProfile*>(Storage::objectContent(_category[cat].head())))->clauseNumber()
	      << "   ";
	  --columns;
	};
    };

  return
    str << "\n%======================================================\n";
}; // ostream& PagedPassiveClauseStorage::outputProfile(ostream& str) const



void PagedPassiveClauseStorage::readFromMemory(const void* memory,
					       ClauseProfile& prof,
					       ClauseLiterals& lits,
					       ClauseAncestors& ancestors)
{
  CALL("readFromMemory(const void* memory,ClauseProfile& prof,ClauseLiterals& lits,ClauseAncestors& ancestors)"); 
  ASSERT(memory);
  const ClauseProfile* profile = static_cast<const ClauseProfile*>(memory);
  prof._clauseNumber = profile->_clauseNumber;
  prof._clauseSize = profile->_clauseSize;
  prof._rules = profile->_rules;
  prof._numOfLiterals = profile->_numOfLiterals;
  prof._numOfAncestors = profile->_numOfAncestors;
  lits._numberOfLiterals = profile->_numOfLiterals;
  ancestors._numberOfAncestors = profile->_numOfAncestors;
  const TERM* storedLiterals = 
    static_cast<const TERM*>(static_cast<const void*>(profile + 1));
  if (profile->_numOfLiterals)
    {
      lits._beginning = storedLiterals;
      // skip all the literals to get to the ancestors
      ulong numOfLitsToSkip = profile->_numOfLiterals; 
      while (numOfLitsToSkip)
	{
	  storedLiterals += (storedLiterals->Content() + 2);
	  numOfLitsToSkip--;
	};
    };
  ancestors._beginning = 
    static_cast<Clause**>
    (const_cast<void*>
     (static_cast<const void*>
      (storedLiterals)));
  lits.reset();
  ancestors.reset();
}; // void PagedPassiveClauseStorage::readFromMemory(const void* memory,ClauseProfile& prof,ClauseLiterals& lits,ClauseAncestors& ancestors)

ostream& PagedPassiveClauseStorage::StoredClause::output(ostream& str)
{
  CALL("StoredClause::output(ostream& str)"); 
  str << "@ " << clauseNumber() << ". "; 

  if (!numOfLiterals()) { str << '#'; } 
  else // output literals
    {
      resetLiterals();
      TERM symbol;
      static TERM literal[2*VampireKernelConst::MaxTermSize];
      ALWAYS(nextLiteral());
      TERM::OpenCompact openLiteral(literal);
 
      while (nextSymbol(symbol)) 
	{ 
	  openLiteral.push(symbol);
	}; 
  
      str << literal;
      if (!literals().currentLiteralProperties().isDefault())
	{
	  str << " <@[" 
	      << literals().currentLiteralProperties()
	      << "]";
	};
      while (nextLiteral())
	{ 
	  openLiteral.reset(literal);
     
	  while (nextSymbol(symbol)) 
	    {
	      openLiteral.push(symbol);
	    }; 

	  str << " \\/ " << literal;
	  
	  if (!literals().currentLiteralProperties().isDefault())
	    {
	      str << " <@[" 
		  << literals().currentLiteralProperties()
		  << "]";
	    };
	};  
    }; 

  str << ' ' << clauseSize() << "(0) ?pe";

  // now output clause background
  str << ((rules().bit(ClauseBackground::VIP)) ? "vip " : "    ") << '[';
  ClauseBackground::outputRules(str,rules());
  resetAncestors();
  Clause* ancestor;
  if (nextAncestor(ancestor))
    {
      str << ancestor->Number();
      while (nextAncestor(ancestor)) str << ',' << ancestor->Number(); 
    };
  str << "]";
  return str;
}; // ostream& PagedPassiveClauseStorage::StoredClause::output(ostream& str)  

ostream& PagedPassiveClauseStorage::ClauseProfile::output(ostream& str) const
{
  CALL("ClauseProfile::output(ostream& str) const");
  str << clauseNumber() << ". "; 
  str << ' ' << clauseSize() << "(0) ?pe";
  // now output clause background
  str << ((rules().bit(ClauseBackground::VIP)) ? "vip " : "    ") << '[';
  ClauseBackground::outputRules(str,rules());
  str << ']';
  return str;  
}; // ostream& PagedPassiveClauseStorage::ClauseProfile::output(ostream& str) const


//====================================================




