//
// File:         PropositionalSplitting.cpp
// Description:  Core module for propositional splitting without backtracking.
// Created:      Nov 07, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#include "PropositionalSplitting.hpp"
#include "TemporaryLiteral.inl"
#include "Clause.inl"
using namespace BK;
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PROPOSITIONAL_SPLITTING
#  define DEBUG_NAMESPACE "PropositionalSplitting"
#endif
#include "debugMacros.hpp"
//===================================================================


PropositionalSplitting::PropositionalSplitting(ComponentName* componentNaming) :
  _style(HyperParallel),
  _smallBasicComponentInHyper(false),
  _componentNaming(componentNaming),
  _numLitMap(DOP("Splitting::_numLitMap")),
  _components(DOP("Splitting::_components")),
  _registeredCompNums(DOP("Splitting::_registeredCompNums")),
  _componentName(DOP("Splitting::_componentName")),
  _newCompName(DOP("Splitting::_newCompName")),
  _maximalNumOfLiteralsSoFar(0UL)
{    
};


PropositionalSplitting::~PropositionalSplitting()
{
  CALL("destructor ~PropositionalSplitting()");
}; // PropositionalSplitting::~PropositionalSplitting()
  


void PropositionalSplitting::reset()
{
  CALL("reset()");
  _style = HyperParallel;
  _smallBasicComponentInHyper = false;
  for (ulong i = 0UL; i < _components.size(); ++i)
    _components[i].reset();
}; // void PropositionalSplitting::reset()

bool 
PropositionalSplitting::splitWithBinaryBlocking(TemporaryClause& cl,
						TemporaryClauseCollector* newClauseCollector,
						ulong& numOfComponents,
						ulong& numOfNewNames)
{
  CALL("splitWithBinaryBlocking(..)");
  ASSERT(!cl.literals()->isEmpty());
  DOP(ulong debugNumOfLits = cl.literals()->length());
  ulong n;
  TERM::Polarity basicCompPol = Signature::NegativePolarity;
  TERM::Polarity nonBasicCompPol = Signature::PositivePolarity;
  LiteralProperties prop; // default value
  

  // load the infrmation about variable occurences in the literals
  // into _divisionIntoComponents
 
  _divisionIntoComponents.setGroundCompOptimalWeight(ULONG_MAX);
  _divisionIntoComponents.reset();
  for (LitListElement* el = cl.literals()->first(); el; el = el->next())
    {
      if (el->literal()->header().IsSplitting())
	{
	  // this literal represents a branch for previously made splitting 
	  ASSERT(Signature::current()->isSkolem(el->literal()->header().functor()));
	  ASSERT(!(el->literal()->numOfDifferentVariables()));
	  ASSERT(!el->literal()->header().arity());
	  if (el->literal()->header().polarity() == basicCompPol) 
	    return false; // the clause is blocked
	  // the literal remains in cl.literals()
	}
      else
	{
	  bool groundLit = !(el->literal()->numOfDifferentVariables());
	  _numLitMap[_divisionIntoComponents.nextLit(groundLit,0UL)] = el;
	  const ulong* beginVar;
	  const ulong* endVar;
	  el->literal()->variables(beginVar,endVar);
	  for (const ulong* v = beginVar; v != endVar; ++v)
	    _divisionIntoComponents.nextVar(*v);
	};  
    }; // for (LitListElement* el...
  
  if (_divisionIntoComponents.numOfComponents() <= 1)
    return false;
  

  // more than one component => splitting is applicable
  

  _registeredCompNums.reset();      
  _divisionIntoComponents.startResultCollection();
  ulong litNum;
  ulong compNum;
  while (_divisionIntoComponents.litCompPair(litNum,compNum))
    {      
      if (!_registeredCompNums.scoreOf(compNum))
	{
	  _registeredCompNums.add(compNum);
	  _components[compNum].reset();    
	};
      cl.literals()->passTo(_numLitMap[litNum],_components[compNum]);
      DOP(_numLitMap[litNum] = 0);
    };
  
  TERM header;
  LitHandle newLiteral;
  ulong numOfNewComponents = 0UL;
  ulong numOfNamedComponents = 0UL;

  // check which components can be folded

  for (n = 0UL; n < _registeredCompNums.numOfDiffElements(); ++n)
    {
      Clause* nameIntroClause;
      compNum = _registeredCompNums.nth(n);
      _newCompName[compNum] = 
	!_componentNaming->findName(_components[compNum],
				    nameIntroClause);
      if (_newCompName[compNum])
	{
	  ++numOfNewComponents;
	}
      else // this component has already been given name   
	{
	  _componentName[compNum] = nameIntroClause->firstLiteral()->functor();
	  if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(newLiteral))
	    {
	      // abort splitting
	      for (ulong m = 0; 
		   m < _registeredCompNums.numOfDiffElements(); 
		   ++m)
		cl.literals()->collectAllFrom(_components[_registeredCompNums.nth(m)]);
	      if (numOfNamedComponents)
		{
		  numOfComponents = numOfNamedComponents + 1;
		  numOfNewNames = 0UL;
		  ASSERT(debugNumOfLits >= cl.literals()->length());
		  return true;
		}
	      else
		{
		  ASSERT(debugNumOfLits == cl.literals()->length());
		  return false;
		};
	    }
	  else
	    {
	      header.MakeComplex((TERM::Arity)0,
				 (TERM::Functor)(_componentName[compNum]),
				 basicCompPol);
	      header.MakeSplitting();
	      newLiteral->pushPropLit(header,prop);

	      cl.background().pushAncestor(nameIntroClause,0L);
	      cl.background().UsedRule(ClauseBackground::Folding);

	      cl.literals()->collect(newLiteral);	      
	      // it makes sense to check cl for tautology here
	      numOfNamedComponents++;
	      _components[compNum].reset();
	    };
	};      
    }; // for (ulong n = 0UL;...

  if (numOfNamedComponents) 
    {
      // collect all other components back to cl.literals()
      for (n = 0UL; n < _registeredCompNums.numOfDiffElements(); ++n) 
	{
	  compNum = _registeredCompNums.nth(n);  
	  if (_newCompName[compNum])
	    _components[compNum].passAllTo(*(cl.literals()));
	};
      numOfComponents = numOfNamedComponents + 1;
      numOfNewNames = 0UL;
      return true;
    }; // if (numOfNamedComponents) 


  // no components have been given names yet. 
  // Give name to one of them (e.g. _components[_registeredCompNums.nth(0)])

  if (!Signature::current()->canCreateNSkolemSymbols(1) || 
      Signature::current()->nextSymbolNumber() > TERM::maxFunctor() ||
      !TemporaryLiteral::Extras::tryToGetFreshLiteral(newLiteral))
    {
      // roll back
      for (n = 0; n < _registeredCompNums.numOfDiffElements(); ++n)
	cl.literals()->collectAllFrom(_components[_registeredCompNums.nth(n)]);
      ASSERT(debugNumOfLits == cl.literals()->length());
      return false;
    };
  
  ulong openComponentNum = _registeredCompNums.nth(0);
  ALWAYS(Signature::current()->createNewSkolemPredicateNumber(0UL,_componentName[openComponentNum]));
 ASSERT(_componentName[openComponentNum] <= TERM::maxFunctor());
 
 header.MakeComplex((TERM::Arity)0,(TERM::Functor)(_componentName[openComponentNum]),basicCompPol);
 header.MakeSplitting();
 Signature::current()->assignMinimalSymbolPrecedence(header.functor());
 Signature::current()->assignMinimalHeaderPrecedence(header.polarity(),
						     header.functor());
 
 
 newLiteral->pushPropLit(header,prop);
 Clause* nameIntroClause = 
   _componentNaming->registerNameIntro(_components[openComponentNum],
				       *newLiteral);

 cl.background().pushAncestor(nameIntroClause,0L);
 cl.background().UsedRule(ClauseBackground::Folding);
 cl.literals()->collect(newLiteral);
  
 // collect the remaining (blocked) components in cl.literals()
 
 for (n = 0; n < _registeredCompNums.numOfDiffElements(); ++n) 
  {
   compNum = _registeredCompNums.nth(n);  
   ASSERT(_newCompName[compNum]);
   if (compNum != openComponentNum)
     _components[compNum].passAllTo(*(cl.literals()));
  };

 // pass the open component to newClauseCollector
 
 collectNameIntroduction((TERM::Functor)(_componentName[openComponentNum]),
			 nonBasicCompPol,
			 _components[openComponentNum],
			 newClauseCollector,
			 nameIntroClause);
 numOfComponents = 2UL;
 numOfNewNames = 1UL; 
 return true;
}; // bool PropositionalSplitting::splitWithBinaryBlocking(..)






bool 
PropositionalSplitting::splitWithHyperParallel(TemporaryClause& cl,
					       TemporaryClauseCollector* newClauseCollector,
					       ulong& numOfComponents,
					       ulong& numOfNewNames)
{
  CALL("splitWithHyperParallel(..)");


  ASSERT(!cl.literals()->isEmpty());
  DOP(ulong debugNumOfLits = cl.literals()->length());
  ulong n;
  TERM::Polarity basicCompPol = Signature::NegativePolarity;
  TERM::Polarity nonBasicCompPol = Signature::PositivePolarity;
  LiteralProperties prop; // default value

  // load the infrmation about variable occurences in the literals
  // into _divisionIntoComponents
 
  _divisionIntoComponents.setGroundCompOptimalWeight(ULONG_MAX);
  _divisionIntoComponents.reset();
  
  for (LitListElement* el = cl.literals()->first(); el; el = el->next())
    {
      if (el->literal()->header().IsSplitting())
	{
	  // this literal represents a branch for previously made splitting 
	  ASSERT(Signature::current()->isSkolem(el->literal()->header().functor()));
	  ASSERT(!(el->literal()->numOfDifferentVariables()));
	  ASSERT(!el->literal()->header().arity());
	  // the literal remains in cl.literals()
	}
      else
	{
	  bool groundLit = !(el->literal()->numOfDifferentVariables());
	  _numLitMap[_divisionIntoComponents.nextLit(groundLit,0UL)] = el;
	  const ulong* beginVar;
	  const ulong* endVar;
	  el->literal()->variables(beginVar,endVar);
	  for (const ulong* v = beginVar; v != endVar; ++v)
	    _divisionIntoComponents.nextVar(*v);
	};  
    }; // for (LitListElement* el...


  if (_divisionIntoComponents.numOfComponents() <= 1)
    return false;
  
  // more than one component => splitting is applicable

  
  numOfComponents = _divisionIntoComponents.numOfComponents();
  numOfNewNames = 0UL;

  _registeredCompNums.reset();      
  _divisionIntoComponents.startResultCollection();
  ulong litNum;
  ulong compNum;
  while (_divisionIntoComponents.litCompPair(litNum,compNum))
    {      
      if (!_registeredCompNums.scoreOf(compNum))
	{
	  _registeredCompNums.add(compNum);
	  _components[compNum].reset();    
	};

      cl.literals()->passTo(_numLitMap[litNum],_components[compNum]);
      DOP(_numLitMap[litNum] = 0);
    };
  

  TERM header;
  LitHandle newLiteral;
  ulong numOfNewComponents = 0UL;
  ulong numOfNamedComponents = 0UL;

  // check which components can be folded

  for (n = 0UL; n < _registeredCompNums.numOfDiffElements(); ++n)
    {
      Clause* nameIntroClause;
      compNum = _registeredCompNums.nth(n);
      _newCompName[compNum] = 
	!_componentNaming->findName(_components[compNum],
				    nameIntroClause);
      if (_newCompName[compNum])
	{
	  ++numOfNewComponents;
	}
      else // this component has already been given name   
	{
	  
	  _componentName[compNum] = nameIntroClause->firstLiteral()->functor();
	  if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(newLiteral))
	    {
	      // abort splitting
	      for (ulong m = 0; 
		   m < _registeredCompNums.numOfDiffElements(); 
		   ++m)
		cl.literals()->collectAllFrom(_components[_registeredCompNums.nth(m)]);
	      if (numOfNamedComponents)
		{
		  numOfComponents = numOfNamedComponents + 1;
		  numOfNewNames = 0UL;
		  ASSERT(debugNumOfLits >= cl.literals()->length());
		  return true;
		}
	      else
		{
		  ASSERT(debugNumOfLits == cl.literals()->length());
		  return false;
		};
	    }
	  else
	    {
	      header.MakeComplex((TERM::Arity)0,
				 (TERM::Functor)(_componentName[compNum]),
				 basicCompPol);
	      header.MakeSplitting();
	      newLiteral->pushPropLit(header,prop);
	      

	      
	      cl.background().pushAncestor(nameIntroClause,0L);
	      cl.background().UsedRule(ClauseBackground::Folding);


	      cl.literals()->collect(newLiteral);
	      // it makes sense to check cl for tautology here
	      numOfNamedComponents++;
	      _components[compNum].reset();
	    };
	};      
    }; // for (ulong n = 0UL;...


  if (!Signature::current()->canCreateNSkolemSymbols(numOfNewComponents) ||
      Signature::current()->nextSymbolNumber() + numOfNewComponents - 1 > TERM::maxFunctor())
    {
      for (n = 0; n < _registeredCompNums.numOfDiffElements(); ++n)
	if (_newCompName[_registeredCompNums.nth(n)])
	  cl.literals()->collectAllFrom(_components[_registeredCompNums.nth(n)]);

      ASSERT(debugNumOfLits >= cl.literals()->length());
      if (numOfNamedComponents > 0UL)
	{
	  numOfComponents = numOfNamedComponents + 1;
	  return true;
	}
      else
	return false;
    };

  // give names to all the remaining components
  
  bool basicComponentChosen = false;
  for (n = 0UL; n < _registeredCompNums.numOfDiffElements(); ++n)
    {
      Clause* nameIntroClause;
      compNum = _registeredCompNums.nth(n);
      
      if (_newCompName[compNum]) 
	{
	  // this could change since we could have several components that are variants of each other, 
	  // simply check this condition again to avoid duplicate names 
	  _newCompName[compNum] = 
	    !_componentNaming->findName(_components[compNum],
					nameIntroClause);
	  if (!_newCompName[compNum])
	    _componentName[compNum] = nameIntroClause->firstLiteral()->functor();
	};
      
      if (_newCompName[compNum]) 
	{
	  if (basicComponentChosen || _smallBasicComponentInHyper)
	    {
	      // _components[compNum] will be given a name

	      if (!TemporaryLiteral::Extras::tryToGetFreshLiteral(newLiteral))
		{
		  // abort splitting

		  for (n = 0; n < _registeredCompNums.numOfDiffElements(); ++n)
		    if (_newCompName[_registeredCompNums.nth(n)])
		      cl.literals()->collectAllFrom(_components[_registeredCompNums.nth(n)]);
		  if (numOfNamedComponents > 0UL)
		    {
		      numOfComponents = numOfNamedComponents + 1;
		      ASSERT(debugNumOfLits >= cl.literals()->length());
		      return true;
		    }
		  else
		    {
		      ASSERT(debugNumOfLits == cl.literals()->length());
		      return false;
		    };
		};

	      ALWAYS(Signature::current()->createNewSkolemPredicateNumber(0UL,_componentName[compNum]));
 
	      ASSERT(_componentName[compNum] <= TERM::maxFunctor());
	      
	      ++numOfNewNames;
	      ++numOfNamedComponents;

	      header.MakeComplex((TERM::Arity)0,
				 (TERM::Functor)(_componentName[compNum]),
				 basicCompPol);
	      header.MakeSplitting();
	      Signature::current()->assignMinimalSymbolPrecedence(header.functor());
	      Signature::current()->assignMinimalHeaderPrecedence(header.polarity(),
								  header.functor()); 
	      newLiteral->pushPropLit(header,prop);

	      nameIntroClause = 
		_componentNaming->registerNameIntro(_components[compNum],
						    *newLiteral);
	      cl.literals()->collect(newLiteral);
	      

	      cl.background().pushAncestor(nameIntroClause,0L);
	      cl.background().UsedRule(ClauseBackground::Folding);

	      if (collectNameIntroduction((TERM::Functor)(_componentName[compNum]),
					  nonBasicCompPol,
					  _components[compNum],
					  newClauseCollector,
					  nameIntroClause))
		{
		  ASSERT(_components[compNum].isEmpty());
		}
	      else
		{
		  _components[compNum].reset();
		};
	    }
	  else // !basicComponentChosen
	    {
	      // _components[compNum] becomes basic component	      
	      basicComponentChosen = true;
	      _components[compNum].passAllTo(*(cl.literals()));
	    };

	}; // if (_newCompName[compNum])
 
    }; // for (ulong n = 0UL;..

  return true;
}; // bool PropositionalSplitting::splitWithHyperParallel(..)



bool 
PropositionalSplitting::collectNameIntroduction(TERM::Functor branchNum,
						TERM::Polarity branchPolarity,
						BasicTempLitList& component,
						TemporaryClauseCollector* newClauseCollector,
						Clause* nameIntroClause)
{
  CALL("collectNameIntroduction(..)");
  LiteralProperties prop; // default value
  TERM header((TERM::Arity)0,branchNum,branchPolarity);
  header.MakeSplitting();  
  Signature::current()->assignMinimalHeaderPrecedence(branchPolarity,
						      branchNum);

  newClauseCollector->openNewClause(ClauseBackground::Prepro,nameIntroClause);

  newClauseCollector->announceNewClause();
  if (!newClauseCollector->pushPropLiteral(header,prop)) 
    return false;
  if (!newClauseCollector->collectLiteralList(component))
    return false;
  return newClauseCollector->finaliseClause();
}; // bool PropositionalSplitting::collectNameIntroduction(..)

bool PropositionalSplitting::adjustSizes(ulong numOfLiterals)
{
  CALL("adjustSizes(ulong numOfLiterals)");

  if (numOfLiterals > _maximalNumOfLiteralsSoFar)
    {
      if (!_divisionIntoComponents.adjustSizes(numOfLiterals))
	return false;
   
      DOP(_numLitMap.unfreeze());
      if (!_numLitMap.tryToExpand(numOfLiterals)) return false;
      DOP(_numLitMap.freeze());

      DOP(_components.unfreeze());  
      if (!_components.tryToExpand(numOfLiterals)) return false;
      DOP(_components.freeze());


      DOP(_registeredCompNums.unfreeze());
      if (!_registeredCompNums.tryToExpand(numOfLiterals)) return false;
      DOP(_registeredCompNums.freeze());


      DOP(_componentName.unfreeze());
      if (!_componentName.tryToExpand(numOfLiterals)) return false;
      DOP(_componentName.freeze());   


      DOP(_newCompName.unfreeze());
      if (!_newCompName.tryToExpand(numOfLiterals)) return false;
      DOP(_newCompName.freeze());

      _maximalNumOfLiteralsSoFar = numOfLiterals;
    }; 
  return true;
}; // bool PropositionalSplitting::adjustSizes(ulong numOfLiterals)

//=================================================================
