//
// File:         StandardLiteralOrdering.hpp
// Description:  Extension of the simplification ordering on terms
//               to literals.  
// Created:      Nov 07, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//==================================================================
#include "StandardLiteralOrdering.hpp"
#include "TemporaryLiteral.inl"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_LITERAL_ORDERING
 #define DEBUG_NAMESPACE "StandardLiteralOrdering"
#endif
#include "debugMacros.hpp"
//===================================================================

using namespace BK;
using namespace VK;


StandardLiteralOrdering::StandardLiteralOrdering() :
  _mode(0L)
{
};
   
StandardLiteralOrdering::~StandardLiteralOrdering()
{
};


Comparison StandardLiteralOrdering::compare(TemporaryLiteral& lit1,
					    TemporaryLiteral& lit2) const
{
  CALL("compare(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");

  Comparison cmp;
  switch (_mode)
    {
    case 0L:
      if (lit1.isEquality())
	{
	  if (lit2.isEquality())
	    {
	      return compareEq(lit1,lit2);
	    }
	  else
	    return Less;
	}
      else
	if (lit2.isEquality())
	  return Greater;
 
      // both are nonequalities

      // compare atoms with the simplification ordering
    compare_atoms:
      cmp =  
	SimplificationOrdering::current()->compareAtoms(lit1.flatterm(),
							lit1.weight(),
							lit2.flatterm(),
							lit2.weight()); 
      if (cmp == Equal)
	{
	  if (lit1.polarity() == lit2.polarity()) return Equal;
	  return (lit1.isPositive()) ? Less : Equal;
	};
      return cmp;

    case 1L:
      if (lit1.isEquality())
	{
	  if (lit2.isEquality())
	    {
	      if (lit1.isNegative())
		{
		  if (lit2.isNegative())
		    {
		      // both are negative
		      return compareEq(lit1,lit2);
		    }
		  else
		    return Greater;
		}
	      else
		if (lit2.isNegative())
		  {
		    return Less;
		  }
		else // both are positive
		  return compareEq(lit1,lit2);
	    }
	  else
	    return Less;
	}
      else
	if (lit2.isEquality())
	  return Greater;
 
      // both are nonequalities

      // compare headers first
      cmp = 
	Signature::current()->compareHeaderPrecedences(lit1.polarity(),
						       lit1.header().functor(),
						       lit2.polarity(),
						       lit2.header().functor());
      ASSERT(cmp != Incomparable);
      if (cmp == Equal)
	{
	  ASSERT(lit1.header() == lit2.header());

	  // compare atoms with the simplification ordering
	  return SimplificationOrdering::current()->compareAtoms(lit1.flatterm(),
								 lit1.weight(),
								 lit2.flatterm(),
								 lit2.weight()); 
	}
      else
	return cmp;	

    case 2L:
      if (lit1.isEquality())
	{
	  if (lit2.isEquality())
	    {
	      return compareEq(lit1,lit2);
	    }
	  else
	    return Less;
	}
      else
	if (lit2.isEquality())
	  return Greater;
 
      // both are nonequalities
      // compare predicates first
      if (lit1.header().functor() == lit2.header().functor())
	{
	  goto compare_atoms;
	}
      else
	return 
	  Signature::current()->compareHeaderPrecedences(lit1.polarity(),
							 lit1.header().functor(),
							 lit2.polarity(),
							 lit2.header().functor());
    case 3L:      
      if (lit1.isEquality())
	{
	  if (lit2.isEquality())
	    {
	      return compareEq(lit1,lit2);
	    }
	  else
	    {
	      return inverse(compareNeqWithEq(lit2,lit1));
	    };
	}
      else
	if (lit2.isEquality())
	  {
	    return compareNeqWithEq(lit1,lit2);
	  }
	else // neither is equality, compare atoms
	  { 
	    goto compare_atoms;
	  };


#ifdef DEBUG_NAMESPACE
    default:
      ICP("ICP0");
      return Incomparable;
#else
#  ifdef _SUPPRESS_WARNINGS_      
    default:
      RuntimeError::report("Wrong value of _mode in StandardLiteralOrdering::compare(..)");
      return Incomparable;
#  endif
#endif

    }; // switch (_mode)
	ICP("END");
	return Incomparable;
}; // Comparison StandardLiteralOrdering::compare(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const




Comparison StandardLiteralOrdering::compareLayers(const TemporaryLiteral& lit1,
						      const TemporaryLiteral& lit2) const
{
  CALL("compareLayers(const TemporaryLiteral& lit1,const TemporaryLiteral& lit2) const");
  if (_mode == 3L) return Equal;
 
  if (_mode == 0L)
    {      
      if (lit1.isEquality())
	{
	  if (lit2.isEquality())
	    {
	      return Equal;
	    }
	  else
	    return Less;
	}
      else
	if (lit2.isEquality())
	  return Greater;
 
      // both are nonequalities
      return Equal;
    };


  if (_mode == 1L)
    {      
      if (lit1.isEquality())
	{
	  if (lit2.isEquality())
	    {
	      if (lit1.isNegative())
		{
		  if (lit2.isNegative())
		    {
		      // both are negative
		      return Equal;
		    }
		  else
		    return Greater;
		}
	      else
		if (lit2.isNegative())
		  {
		    return Less;
		  }
		else // both are positive
		  return Equal;
	    }
	  else
	    return Less;
	}
      else
	if (lit2.isEquality())
	  return Greater;
 
      // both are nonequalities

      return 
	Signature::current()->compareHeaderPrecedences(lit1.polarity(),
						       lit1.header().functor(),
						       lit2.polarity(),
						       lit2.header().functor());
    };

  ASSERT(_mode == 2L);
    
  if (lit1.isEquality())
    {
      if (lit2.isEquality())
	{
	  return Equal;
	}
      else
	return Less;
    }
  else
    if (lit2.isEquality())
      return Greater;
 
  // both are nonequalities

  if (lit1.header().functor() == lit2.header().functor())
    {
      return Equal;
    }
  else
    return 
      Signature::current()->compareHeaderPrecedences(lit1.polarity(),
						     lit1.header().functor(),
						     lit2.polarity(),
						     lit2.header().functor());
}; // Comparison StandardLiteralOrdering::compareLayers(const TemporaryLiteral& lit1,const TemporaryLiteral& lit2) const






inline
Comparison StandardLiteralOrdering::compareEq(TemporaryLiteral& lit1,
				      TemporaryLiteral& lit2) const
{
  CALL("compareEq(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isEquality());
  ASSERT(lit2.isEquality());
  
  // compare atoms first
  Comparison acmp;
  if (lit1.isOrderedEquality())
    {
      if (lit2.isOrderedEquality())
	{
	  // both are ordered
	  acmp = compareEqAtomsOO(lit1,lit2);
	}
      else
	{
	  acmp = compareEqAtomsOU(lit1,lit2);
	};
    }
  else
    if (lit2.isOrderedEquality())
      {
	acmp = inverse(compareEqAtomsOU(lit2,lit1));
      }
    else // both are unordered
      {
	acmp = compareEqAtomsUU(lit1,lit2);
      };
  
  if (acmp == Equal)
    {     
      if (lit1.polarity() == lit2.polarity()) return Equal;
      return (lit1.isPositive()) ? Less : Equal;
    };
  return acmp;
}; // Comparison StandardLiteralOrdering::compareEq(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const




Comparison StandardLiteralOrdering::compareEqAtomsOO(TemporaryLiteral& lit1,
					     TemporaryLiteral& lit2) const

{
  CALL("compareEqAtomsOO(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isOrderedEquality());
  ASSERT(lit2.isOrderedEquality());

  Comparison cmp11 = 
    SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
					       lit2.arg1(),lit2.arg1Weight()); 
  if (cmp11 != Equal) return cmp11;
  return SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
						    lit2.arg2(),lit2.arg2Weight()); 

}; // Comparison StandardLiteralOrdering::compareEqAtomsOO(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const

Comparison StandardLiteralOrdering::compareEqAtomsOU(TemporaryLiteral& lit1,
					     TemporaryLiteral& lit2) const

{
  CALL("compareEqAtomsOU(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(lit1.isOrderedEquality());
  ASSERT(!lit2.isOrderedEquality());
  
  // lit1.arg1() > lit1.arg2()
  switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
						     lit2.arg1(),lit2.arg1Weight()))
    {
    case Less:
      // lit1.arg2() < lit1.arg1() < lit2.arg1()
      return Less;
	  
    case Equal: 
      // lit1.arg1() = lit2.arg1() * lit2.arg2(), compare lit1.arg2() with lit2.arg2()
      return SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							lit2.arg2(),lit2.arg2Weight());	
    case Greater: 
      // lit1.arg1() > lit2.arg1(), compare lit1.arg1() with lit2.arg2()
      return SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							lit2.arg2(),lit2.arg2Weight());
    case Incomparable:         
      // compare lit1.arg2() with lit2.arg1() and lit2.arg2()
      switch (SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),
							 lit2.arg1(),lit2.arg1Weight()))
	{
	case Less: 
	  if (SimplificationOrdering::current()->lessOrEqual(lit1.arg1(),lit1.arg1Weight(),
							     lit2.arg2(),lit2.arg2Weight()))
	    {
	      return Less;
	    }
	  else
	    return Incomparable;

	case Equal:
	  return SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							    lit2.arg2(),lit2.arg2Weight());

	case Greater:
	  // lit1.arg1() * lit2.arg1(), lit1.arg2() > lit2.arg1()
	  switch (SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),
							     lit2.arg2(),lit2.arg2Weight()))
	    {
	    case Less: return Less;
	    case Equal: return Greater;
	    case Greater: return Greater;
	    case Incomparable: 
	      if (SimplificationOrdering::current()->greaterOrEqual(lit1.arg2(),lit1.arg2Weight(),
								    lit2.arg2(),lit2.arg2Weight()))
		{
		  return Greater;
		}
	      else
		return Incomparable;
	    };

	case Incomparable:
	  // lit1.arg1() * lit2.arg1(), lit1.arg2() * lit2.arg1()
	  if (SimplificationOrdering::current()->less(lit1.arg1(),lit1.arg1Weight(),
						      lit2.arg2(),lit2.arg2Weight()))
	    {
	      return Less;
	    }
	  else
	    return Incomparable;
	};

    }; // switch

  ICP("ICP0");
  return Incomparable;

}; // Comparison StandardLiteralOrdering::compareEqAtomsOU(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const


Comparison StandardLiteralOrdering::compareEqAtomsUU(TemporaryLiteral& lit1,
					     TemporaryLiteral& lit2) const

{
  CALL("compareEqAtomsUU(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const");
  ASSERT(!lit1.isOrderedEquality());
  ASSERT(!lit2.isOrderedEquality());

  Comparison cmp = SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),lit2.arg1(),lit2.arg1Weight());
  switch (cmp)
    {
    case Greater:
      // lit1.arg1() > lit2.arg1()
      cmp = SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),lit2.arg2(),lit2.arg2Weight()); 
      switch (cmp)
	{
	case Greater:
	  // lit1.arg1() > lit2.arg1(), lit1.arg1() > lit2.arg2()
	  return Greater;
        
	case Less:
	  // lit1.arg1() > lit2.arg1(), lit1.arg1() < lit2.arg2(). Not an error since v2.73
	  return SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
       
	case Incomparable:
	  // lit1.arg1() > lit2.arg1(), lit1.arg1() * lit2.arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1.arg1() > lit2.arg1(), lit1.arg1() * lit2.arg2(), lit1.arg2() > lit2.arg2()
	      return Greater;
            
	    case Less:
	      // lit1.arg1() > lit2.arg1(), lit1.arg1() * lit2.arg2(), lit1.arg2() < lit2.arg2()
	      return Incomparable;
           
	    case Incomparable:
	      // lit1.arg1() > lit2.arg1(), lit1.arg1() * lit2.arg2(), lit1.arg2() * lit2.arg2()
	      return Incomparable;
           
	    case Equal:
	      // lit1.arg1() > lit2.arg1(), lit1.arg1() * lit2.arg2(), lit1.arg2() = lit2.arg2()
	      return Greater;
	    };
        
	case Equal:
	  // lit1.arg1() > lit2.arg1(), lit1.arg1() = lit2.arg2() : error because lit2.arg1() * lit2.arg2()
#ifdef DEBUG_NAMESPACE 
	  ICP("DE20"); 
#endif
	  return Incomparable;
	};

    case Less:
      // lit1.arg1() < lit2.arg1()
      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
      switch (cmp)
	{
	case Greater:
	  // lit1.arg1() < lit2.arg1(), lit1.arg2() > lit2.arg1(). Not an error since v2.73
	  return SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
        
	case Less:
	  // lit1.arg1() < lit2.arg1(), lit1.arg2() < lit2.arg1()
	  return Less; // was Greater in the CASC-16 version (obvious incompleteness)
        
	case Incomparable:
	  // lit1.arg1() < lit2.arg1(), lit1.arg2() * lit2.arg1()
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1.arg1() < lit2.arg1(), lit1.arg2() * lit2.arg1(), lit1.arg2() > lit2.arg2()
	      return Incomparable;
           
	    case Less:
	      // lit1.arg1() < lit2.arg1(), lit1.arg2() * lit2.arg1(), lit1.arg2() < lit2.arg2()
	      return Less;
           
	    case Incomparable:
	      // lit1.arg1() < lit2.arg1(), lit1.arg2() * lit2.arg1(), lit1.arg2() * lit2.arg2()
	      return Incomparable;
           
	    case Equal:
	      // lit1.arg1() < lit2.arg1(), lit1.arg2() * lit2.arg1(), lit1.arg2() = lit2.arg2()
	      return Less; 
	    }; 
        
	case Equal:
	  // lit1.arg1() < lit2.arg1(), lit1.arg2() = lit2.arg1() : error because lit1.arg1() * lit1.arg2()
#ifdef DEBUG_NAMESPACE 
	  ICP("DE40");
#endif
	  return Incomparable;
	};
      
    case Incomparable:
      // lit1.arg1() * lit2.arg1()
      cmp = SimplificationOrdering::current()->compare(lit1.arg1(),lit1.arg1Weight(),lit2.arg2(),lit2.arg2Weight()); 
      switch (cmp)
	{
	case Greater:
	  // lit1.arg1() * lit2.arg1(), lit1.arg1() > lit2.arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() > lit2.arg2(), lit1.arg2() > lit2.arg1()
	      return Greater; 
           
	    case Less:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() > lit2.arg2(), lit1.arg2() < lit2.arg1()
	      return Incomparable;
           
	    case Incomparable:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() > lit2.arg2(), lit1.arg2() * lit2.arg1()
	      return Incomparable;
           
	    case Equal:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() > lit2.arg2(), lit1.arg2() = lit2.arg1()
	      return Greater;
	    };
          
	case Less:
	  // lit1.arg1() * lit2.arg1(), lit1.arg1() < lit2.arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() < lit2.arg2(), lit1.arg2() > lit2.arg1()
	      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight());
	      ASSERT(cmp != Equal);
	      return cmp;
          
	    case Less:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() < lit2.arg2(), lit1.arg2() < lit2.arg1()
	      return Less;
            
	    case Incomparable:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() < lit2.arg2(), lit1.arg2() * lit2.arg1()
	      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	      if (cmp == Less) return Less;
	      return Incomparable;
            
	    case Equal:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() < lit2.arg2(), lit1.arg2() = lit2.arg1()
	      return Less; 
	    }; 
         

	case Incomparable:
	  // lit1.arg1() * lit2.arg1(), lit1.arg1() * lit2.arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() * lit2.arg2(), lit1.arg2() > lit2.arg1()
	      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
	      if (cmp == Greater) return Greater;
	      ASSERT(cmp != Equal);
	      return Incomparable;
           
	    case Less:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() * lit2.arg2(), lit1.arg2() < lit2.arg1()
	      return Incomparable;
           
	    case Incomparable:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() * lit2.arg2(), lit1.arg2() * lit2.arg1()
	      return Incomparable;
           
	    case Equal:
	      // lit1.arg1() * lit2.arg1(), lit1.arg1() * lit2.arg2(), lit1.arg2() = lit2.arg1()
	      return Incomparable;
	    }; 
        
	case Equal:
	  // lit1.arg1() * lit2.arg1(), lit1.arg1() = lit2.arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg1(),lit2.arg1Weight()); 
	  return cmp;
	};
      
    case Equal:
      // lit1.arg1() = lit2.arg1()
      cmp = SimplificationOrdering::current()->compare(lit1.arg2(),lit1.arg2Weight(),lit2.arg2(),lit2.arg2Weight()); 
      return cmp;
    };
  ICP("DE100");
  return Incomparable; 

}; // Comparison StandardLiteralOrdering::compareEqAtomsUU(TemporaryLiteral& lit1,TemporaryLiteral& lit2) const


Comparison StandardLiteralOrdering::compareNeqWithEq(TemporaryLiteral& neq,
						     TemporaryLiteral& eq) const
{
  CALL("compareNeqWithEq(TemporaryLiteral& neq,TemporaryLiteral& eq) const");
  ASSERT(!neq.isEquality());
  ASSERT(eq.isEquality());
  Comparison cmp1 = 
    SimplificationOrdering::current()->compare(neq.flatterm(),neq.weight(),
					       eq.arg1(),eq.arg1Weight()); 
    
  if (eq.isOrderedEquality()) return cmp1;
  
  if (cmp1 == Less) return Less;
  
  Comparison cmp2 = 
    SimplificationOrdering::current()->compare(neq.flatterm(),neq.weight(),
					       eq.arg2(),eq.arg2Weight()); 
  
  if (cmp2 == Incomparable) return Incomparable;
  if (cmp2 == Less) return Less;
  ASSERT(cmp2 == Greater);
  return cmp1;
}; // Comparison StandardLiteralOrdering::compareNeqWithEq(..)



//=================================================================
