//
// File:         Statistics.cpp
// Description:  Statistics on inferences and clauses.
// Created:      Nov 17, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#include "Statistics.hpp"
using namespace BK;
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics"
#endif
#include "debugMacros.hpp"
//===================================================================

void Statistics::reset()
{
  inferences.reset();
  clauses.reset();
}; // void Statistics::reset()

void Statistics::absorb(const Statistics& stat)
{
  inferences.absorb(stat.inferences);
  clauses.absorb(stat.clauses);
}; // void Statistics::absorb(const Statistics& stat)


bool Statistics::completenessViolated() const
{
  return 
    inferences.someNonredundantDiscarded() ||
    clauses.someNonredundantDiscarded();
}; // bool Statistics::completenessViolated() const

ostream& Statistics::outputDetailed(ostream& str) const
{
  inferences.outputDetailed(str);
  clauses.outputDetailed(str);
  return str;
}; // ostream& Statistics::outputDetailed(ostream& str) const


ostream& Statistics::outputSummary(ostream& str) const
{
  inferences.outputSummary(str);
  clauses.outputSummary(str);
  return str;
}; // ostream& Statistics::outputSummary(ostream& str) const




//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Inferences"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Inferences::reset()
{
  deduction.reset();
  simplifying.reset();
  propositionalSplitting.reset();
  negativeEqualitySplitting.reset();
}; // void Statistics::Inferences::reset()

void Statistics::Inferences::absorb(const Inferences& stat)
{
  deduction.absorb(stat.deduction);
  simplifying.absorb(stat.simplifying);
  propositionalSplitting.absorb(stat.propositionalSplitting);
  negativeEqualitySplitting.absorb(stat.negativeEqualitySplitting);
}; // void Statistics::Inferences::absorb(const Inferences& stat)

ostream& Statistics::Inferences::outputDetailed(ostream& str) const
{
  str << "%== Inferences\n";
  
  str << "%     deduction\n";
  str << "%       accomplished " << deduction.totalAccomplished() << "\n";
  str << "%       redundant discarded " << deduction.totalRedundantDiscarded() 
      << " + " << deduction.totalRedundantBranchesDiscarded() << " branches\n";
  str << "%       nonredundant discarded " << deduction.totalNonredundantDiscarded() 
      << " + " << deduction.totalNonredundantBranchesDiscarded() << " branches\n";
  if (deduction.somethingRegistered())
    str << "%       rules\n";
  if (deduction.resolution.somethingRegistered())
    deduction.resolution.outputDetailed(str << "%         resolution ") << "\n";
  if (deduction.forwardSuperposition.somethingRegistered())
    deduction.forwardSuperposition.outputDetailed(str << "%         forw superposition ") << "\n";
  if (deduction.backwardSuperposition.somethingRegistered())
    deduction.backwardSuperposition.outputDetailed(str << "%         back superposition ") << "\n";
  if (deduction.equalityFactoring.somethingRegistered())
    deduction.equalityFactoring.outputDetailed(str << "%         eq factoring") << "\n";
  if (deduction.equalityResolution.somethingRegistered())
    deduction.equalityResolution.outputDetailed(str << "%         eq resolution ") << "\n";
  if (deduction.forwardChaining.somethingRegistered())
    deduction.forwardChaining.outputDetailed(str << "%         forw chaining ") << "\n";
  if (deduction.backwardChaining.somethingRegistered())
    deduction.backwardChaining.outputDetailed(str << "%         back chaining ") << "\n";

  str << "%     simplifying\n";
  if (simplifying.propositionalTautologyDeletion)
    str << "%       prop tautology del " 
	<< simplifying.propositionalTautologyDeletion 
	<< "\n";
  if (simplifying.equationalTautologyDeletion)
    str << "%       eq tautology del " 
	<< simplifying.equationalTautologyDeletion
	<< "\n";
  
  if (simplifying.reductionInNonBasicPosition)
    str << "%       red in non-basic pos " 
	<< simplifying.reductionInNonBasicPosition
	<< "\n";
  
  if (simplifying.forwardSubsumption)
    str << "%       forw subsumption " 
	<< simplifying.forwardSubsumption 
	<< "\n";
  
  if (simplifying.backwardSubsumption)
    str << "%       backward subsumption " 
	<< simplifying.backwardSubsumption
	<< "\n";
  
  if (simplifying.forwardSubsumptionResolution)
    str << "%       forw subs resolution " 
	<< simplifying.forwardSubsumptionResolution
	<< "\n";
  
  if (simplifying.forwardDemodulation)
    str << "%       forw demodulation " 
	<<  simplifying.forwardDemodulation
	<< "\n";
  
  if (simplifying.onBranchForwardDemodulation)
    str << "%       on-branch forw demod " 
	<< simplifying.onBranchForwardDemodulation 
	<< "\n";
  
  if (simplifying.backwardDemodulation)
    str << "%       back demodulation " 
	<< simplifying.backwardDemodulation 
	<< "\n";
  
  if (simplifying.onBranchBackwardDemodulation)
    str << "%       on-branch back demod " 
	<< simplifying.onBranchBackwardDemodulation
	<< "\n";
  
  if (simplifying.destructiveEqualityResolution)
    str << "%       destr eq resolution " 
	<< simplifying.destructiveEqualityResolution 
	<< "\n";
  
  if (simplifying.builtInTheory)
    str << "%       built-in theory simpl " 
	<< simplifying.builtInTheory
	<< "\n";

  if (propositionalSplitting.somethingRegistered())
    {
      str << "%     propositional splitting " 
	  << propositionalSplitting.accomplished 
	  << "  avg comp " 
	  << ((propositionalSplitting.accomplished) ? 
	      (propositionalSplitting.components / propositionalSplitting.accomplished)
	      :
	      0UL)
          << " names " << propositionalSplitting.introducedComponentNames << "\n";
    };


  if (negativeEqualitySplitting.somethingRegistered())
    {
      str << "%     negative equality splitting "
	  << negativeEqualitySplitting.accomplished 
	  << "  avg arg split off " 
	  << ((negativeEqualitySplitting.accomplished) ? 
	      (negativeEqualitySplitting.argumentsSplitOff / negativeEqualitySplitting.accomplished)
	      :
	      0UL)
          << " names " << negativeEqualitySplitting.introducedNames << "\n";
    };
  return str;
}; // ostream& Statistics::Inferences::outputDetailed(ostream& str) const


ostream& Statistics::Inferences::outputSummary(ostream& str) const
{
  str << "%== Deduction inferences\n";
  str << "%     accomplished " 
      << deduction.totalAccomplished() << " = " 
      << deduction.resolution.accomplished << " res + "
      << (deduction.forwardSuperposition.accomplished +
	  deduction.backwardSuperposition.accomplished +
	  deduction.equalityFactoring.accomplished +
	  deduction.equalityResolution.accomplished) << " eq + "
      << (deduction.forwardChaining.accomplished +
	  deduction.backwardChaining.accomplished) << " ch\n";
  
  str << "%     redundant " << deduction.totalRedundantDiscarded() 
      << " + " << deduction.totalRedundantBranchesDiscarded() << " branches\n";
  str << "%     nonredundant discarded " << deduction.totalNonredundantDiscarded() 
      << " + " << deduction.totalNonredundantBranchesDiscarded() << " branches\n";

  str << "%== Simplifying inferences " << simplifying.total() << "\n";
  str << "%     forward  " << simplifying.totalForward() << "\n";

  
  if (simplifying.propositionalTautologyDeletion || 
      simplifying.equationalTautologyDeletion)
    str << "%       tautology del " 
	<< simplifying.propositionalTautologyDeletion << " prop + "
	<< simplifying.equationalTautologyDeletion << " eq\n";
   
  if (simplifying.reductionInNonBasicPosition)
    str << "%       red in non-basic pos " 
	<< simplifying.reductionInNonBasicPosition
	<< "\n";
  
  if (simplifying.forwardSubsumption)
    str << "%       subsumption " 
	<< simplifying.forwardSubsumption 
	<< "\n";
  
  if (simplifying.forwardSubsumptionResolution)
    str << "%       subs resolution " 
	<< simplifying.forwardSubsumptionResolution
	<< "\n";
  
  if (simplifying.forwardDemodulation || 
      simplifying.onBranchForwardDemodulation)
    str << "%       demodulation " << simplifying.forwardDemodulation 
	<< " + " << simplifying.onBranchForwardDemodulation << " on-branch\n";
  
  if (simplifying.destructiveEqualityResolution)
    str << "%       destr eq resolution " 
	<< simplifying.destructiveEqualityResolution 
	<< "\n";
  
  if (simplifying.builtInTheory)
    str << "%       built-in theory simpl " 
	<< simplifying.builtInTheory
	<< "\n";
 
  str << "%     backward " << simplifying.totalBackward() << "\n";
 
  if (simplifying.backwardSubsumption)
    str << "%       subsumption " 
	<< simplifying.backwardSubsumption
	<< "\n";
  

  if (simplifying.backwardDemodulation || 
      simplifying.onBranchBackwardDemodulation)
    str << "%       demodulation " << simplifying.backwardDemodulation 
	<< " + " << simplifying.onBranchBackwardDemodulation << " on-branch\n";


  if (propositionalSplitting.somethingRegistered())
    {
      str << "%     propositional splitting " 
	  << propositionalSplitting.accomplished 
	  << "  avg comp " 
	  << ((propositionalSplitting.accomplished) ? 
	      (propositionalSplitting.components / propositionalSplitting.accomplished)
	      :
	      0UL)
          << " names " << propositionalSplitting.introducedComponentNames << "\n";
    };


  if (negativeEqualitySplitting.somethingRegistered())
    {
      str << "%     negative equality splitting "
	  << negativeEqualitySplitting.accomplished 
	  << "  avg arg split off " 
	  << ((negativeEqualitySplitting.accomplished) ? 
	      (negativeEqualitySplitting.argumentsSplitOff / negativeEqualitySplitting.accomplished)
	      :
	      0UL)
          << " names " << negativeEqualitySplitting.introducedNames << "\n";
    };


  return str;
}; // ostream& Statistics::Inferences::outputSummary(ostream& str) const



//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Inferences::Deduction"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Inferences::Deduction::reset()
{
  resolution.reset();
  forwardSuperposition.reset();
  backwardSuperposition.reset();
  equalityFactoring.reset();
  equalityResolution.reset();
  forwardChaining.reset();
  backwardChaining.reset();
}; // void Statistics::Inferences::Deduction::reset()

void Statistics::Inferences::Deduction::absorb(const Deduction& stat)
{
  resolution.absorb(stat.resolution);
  forwardSuperposition.absorb(stat.forwardSuperposition);
  backwardSuperposition.absorb(stat.backwardSuperposition);
  equalityFactoring.absorb(stat.equalityFactoring);
  equalityResolution.absorb(stat.equalityResolution);
  forwardChaining.absorb(stat.forwardChaining);
  backwardChaining.absorb(stat.backwardChaining);
}; // void Statistics::Inferences::Deduction::absorb(const Deduction& stat)


ulong Statistics::Inferences::Deduction::totalAccomplished() const
{
  return
    resolution.accomplished +
    forwardSuperposition.accomplished +
    backwardSuperposition.accomplished +
    equalityFactoring.accomplished +
    equalityResolution.accomplished +
    forwardChaining.accomplished +
    backwardChaining.accomplished;
}; // ulong Statistics::Inferences::Deduction::totalAccomplished() const

ulong Statistics::Inferences::Deduction::totalRedundantDiscarded() const
{
  return 
    resolution.redundantDiscarded +
    forwardSuperposition.redundantDiscarded +
    backwardSuperposition.redundantDiscarded +
    equalityFactoring.redundantDiscarded +
    equalityResolution.redundantDiscarded +
    forwardChaining.redundantDiscarded +
    backwardChaining.redundantDiscarded;
}; // ulong Statistics::Inferences::Deduction::totalRedundantDiscarded() const

ulong Statistics::Inferences::Deduction::totalRedundantBranchesDiscarded() const
{
  return
    resolution.redundantBranchesDiscarded +
    forwardSuperposition.redundantBranchesDiscarded +
    backwardSuperposition.redundantBranchesDiscarded +
    equalityFactoring.redundantBranchesDiscarded +
    equalityResolution.redundantBranchesDiscarded +
    forwardChaining.redundantBranchesDiscarded +
    backwardChaining.redundantBranchesDiscarded;
}; // ulong Statistics::Inferences::Deduction::totalRedundantBranchesDiscarded() const

ulong Statistics::Inferences::Deduction::totalNonredundantDiscarded() const
{
  return
    resolution.totalNonredundantDiscarded() +
    forwardSuperposition.totalNonredundantDiscarded() +
    backwardSuperposition.totalNonredundantDiscarded() +
    equalityFactoring.totalNonredundantDiscarded() +
    equalityResolution.totalNonredundantDiscarded() +
    forwardChaining.totalNonredundantDiscarded() +
    backwardChaining.totalNonredundantDiscarded();
}; // ulong Statistics::Inferences::Deduction::totalNonredundantDiscarded() const

ulong Statistics::Inferences::Deduction::totalNonredundantBranchesDiscarded() const
{
  return
    resolution.totalNonredundantBranchesDiscarded() +
    forwardSuperposition.totalNonredundantBranchesDiscarded() +
    backwardSuperposition.totalNonredundantBranchesDiscarded() +
    equalityFactoring.totalNonredundantBranchesDiscarded() +
    equalityResolution.totalNonredundantBranchesDiscarded() +
    forwardChaining.totalNonredundantBranchesDiscarded() +
    backwardChaining.totalNonredundantBranchesDiscarded();
}; // ulong Statistics::Inferences::Deduction::totalNonredundantBranchesDiscarded() const



//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Inferences::Deduction::RuleProfile"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Inferences::Deduction::RuleProfile::reset()
{  
  accomplished = 0UL; 
  redundantDiscarded = 0UL; 
  redundantBranchesDiscarded = 0UL;
  discardedBySize = 0UL;
  branchesDiscardedBySize = 0UL;
  discardedParamodulationsFromVariables = 0UL;
  branchesDiscardedWithParamodulationsFromVariables = 0UL;
  discardedChainingsThroughVariables = 0UL;
  branchesDiscardedWithChainingsThroughVariables = 0UL;
  mainPremisesSubsumed = 0UL;
  sidePremisesSubsumed = 0UL;
  mainPremisesDemodulated = 0UL;
  sidePremisesDemodulated = 0UL; 
  discardedForOtherReasons = 0UL;
  branchesDiscardedForOtherReasons = 0UL;
}; // void Statistics::Inferences::Deduction::RuleProfile::reset()



void Statistics::Inferences::Deduction::RuleProfile::absorb(const RuleProfile& stat)
{  
  accomplished += stat.accomplished; 
  redundantDiscarded += stat.redundantDiscarded; 
  redundantBranchesDiscarded += stat.redundantBranchesDiscarded;
  discardedBySize += stat.discardedBySize;
  branchesDiscardedBySize += stat.branchesDiscardedBySize;
  discardedParamodulationsFromVariables += stat.discardedParamodulationsFromVariables;
  branchesDiscardedWithParamodulationsFromVariables += stat.branchesDiscardedWithParamodulationsFromVariables;
  discardedChainingsThroughVariables += stat.discardedChainingsThroughVariables;
  branchesDiscardedWithChainingsThroughVariables += stat.branchesDiscardedWithChainingsThroughVariables;
  mainPremisesSubsumed += stat.mainPremisesSubsumed;
  sidePremisesSubsumed += stat.sidePremisesSubsumed;
  mainPremisesDemodulated += stat.mainPremisesDemodulated;
  sidePremisesDemodulated += stat.sidePremisesDemodulated; 
  discardedForOtherReasons += stat.discardedForOtherReasons;
  branchesDiscardedForOtherReasons += stat.branchesDiscardedForOtherReasons;
}; // void Statistics::Inferences::Deduction::RuleProfile::absorb(const RuleProfile& stat)


bool Statistics::Inferences::Deduction::RuleProfile::somethingRegistered() const
{
  return
    accomplished || 
    redundantDiscarded || 
    redundantBranchesDiscarded ||
    discardedBySize ||
    branchesDiscardedBySize ||
    discardedParamodulationsFromVariables ||
    branchesDiscardedWithParamodulationsFromVariables ||
    discardedChainingsThroughVariables ||
    branchesDiscardedWithChainingsThroughVariables ||
    mainPremisesSubsumed ||
    sidePremisesSubsumed ||
    mainPremisesDemodulated ||
    sidePremisesDemodulated ||
    discardedForOtherReasons ||
    branchesDiscardedForOtherReasons;
}; // bool Statistics::Inferences::Deduction::RuleProfile::somethingRegistered() cons

bool Statistics::Inferences::Deduction::RuleProfile::someNonredundantDiscarded() const
{
  return
    discardedBySize ||
    branchesDiscardedBySize ||
    discardedParamodulationsFromVariables ||
    branchesDiscardedWithParamodulationsFromVariables ||
    discardedChainingsThroughVariables ||
    branchesDiscardedWithChainingsThroughVariables ||
    mainPremisesSubsumed ||
    sidePremisesSubsumed ||
    mainPremisesDemodulated ||
    sidePremisesDemodulated ||
    discardedForOtherReasons ||
    branchesDiscardedForOtherReasons;
}; // bool Statistics::Inferences::Deduction::RuleProfile::someNonredundantDiscarded() const


ulong Statistics::Inferences::Deduction::RuleProfile::totalNonredundantDiscarded() const
{
  return
    discardedBySize +
    discardedParamodulationsFromVariables +
    discardedChainingsThroughVariables +
    discardedForOtherReasons;
}; // ulong Statistics::Inferences::Deduction::RuleProfile::totalNonredundantDiscarded() const

ulong Statistics::Inferences::Deduction::RuleProfile::totalNonredundantBranchesDiscarded() const
{
  return
    branchesDiscardedBySize +
    branchesDiscardedWithParamodulationsFromVariables +
    branchesDiscardedWithChainingsThroughVariables +
    mainPremisesSubsumed +
    mainPremisesDemodulated + 
    sidePremisesSubsumed + 
    sidePremisesDemodulated +
    branchesDiscardedForOtherReasons;
}; // ulong Statistics::Inferences::Deduction::RuleProfile::totalNonredundantBranchesDiscarded() const



ostream& Statistics::Inferences::Deduction::RuleProfile::outputDetailed(ostream& str) const
{
  str << accomplished;
  if (somethingRegistered())
    {
      if (someNonredundantDiscarded())
	str << " *";
      str << " [ ";
      if (redundantDiscarded)
	str << "red " << redundantDiscarded << " ";
      if (redundantBranchesDiscarded)
	str << "rbr " << redundantBranchesDiscarded << " ";
      if (discardedBySize)
	str << "sz " << discardedBySize << " ";
      if (branchesDiscardedBySize)
	str << "brsz " << branchesDiscardedBySize << " ";
      if (discardedParamodulationsFromVariables)
	str << "pfv " << discardedParamodulationsFromVariables << " ";
      if (branchesDiscardedWithParamodulationsFromVariables)
	str << "brpfv " << branchesDiscardedWithParamodulationsFromVariables << " ";
      if (discardedChainingsThroughVariables)
	str << "ctv " << discardedChainingsThroughVariables << " ";
      if (branchesDiscardedWithChainingsThroughVariables)
	str << "brctv " << branchesDiscardedWithChainingsThroughVariables << " ";
      if (mainPremisesSubsumed)
	str << "mps " << mainPremisesSubsumed << " ";
      if (sidePremisesSubsumed)
	str << "sps " << sidePremisesSubsumed << " ";
      if (mainPremisesDemodulated)
	str << "mpd " << mainPremisesDemodulated << " ";
      if (sidePremisesDemodulated)
	str << "spd " << sidePremisesDemodulated << " ";
      if (discardedForOtherReasons)
	str << "othd " << discardedForOtherReasons << " ";
      if (branchesDiscardedForOtherReasons)
	str << "othbrd " << branchesDiscardedForOtherReasons << " ";
      str << "]";
    };
  return str;
}; // ostream& Statistics::Inferences::Deduction::RuleProfile::outputDetailed(ostream& str) const
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Inferences::Simplifying"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Inferences::Simplifying::reset()
{
  propositionalTautologyDeletion = 0UL;
  equationalTautologyDeletion = 0UL;
  reductionInNonBasicPosition = 0UL;
  forwardSubsumption = 0UL;
  backwardSubsumption = 0UL;
  forwardSubsumptionResolution = 0UL;
  forwardDemodulation = 0UL;
  onBranchForwardDemodulation = 0UL;
  backwardDemodulation = 0UL;
  onBranchBackwardDemodulation = 0UL;
  destructiveEqualityResolution = 0UL;
  builtInTheory = 0UL;
}; // void Statistics::Inferences::Simplifying::reset()

void Statistics::Inferences::Simplifying::absorb(const Simplifying& stat)
{
  propositionalTautologyDeletion += stat.propositionalTautologyDeletion;
  equationalTautologyDeletion += stat.equationalTautologyDeletion;
  reductionInNonBasicPosition += stat.reductionInNonBasicPosition;
  forwardSubsumption += stat.forwardSubsumption;
  backwardSubsumption += stat.backwardSubsumption;
  forwardSubsumptionResolution += stat.forwardSubsumptionResolution;
  forwardDemodulation += stat.forwardDemodulation;
  onBranchForwardDemodulation += stat.onBranchForwardDemodulation;
  backwardDemodulation += stat.backwardDemodulation;
  onBranchBackwardDemodulation += stat.onBranchBackwardDemodulation;
  destructiveEqualityResolution += stat.destructiveEqualityResolution;
  builtInTheory += stat.builtInTheory;
}; // void Statistics::Inferences::Simplifying::absorb(const Simplifying& stat)


ulong Statistics::Inferences::Simplifying::total() const
{
  return 
    propositionalTautologyDeletion +
    equationalTautologyDeletion +
    reductionInNonBasicPosition +
    forwardSubsumption +
    backwardSubsumption +
    forwardSubsumptionResolution +
    forwardDemodulation +
    onBranchForwardDemodulation +
    backwardDemodulation +
    onBranchBackwardDemodulation +
    destructiveEqualityResolution +
    builtInTheory;
}; // ulong Statistics::Inferences::Simplifying::total() const

ulong Statistics::Inferences::Simplifying::totalForward() const
{
  return 
    propositionalTautologyDeletion +
    equationalTautologyDeletion +
    reductionInNonBasicPosition +
    forwardSubsumption +
    forwardSubsumptionResolution +
    forwardDemodulation +
    onBranchForwardDemodulation +
    destructiveEqualityResolution +
    builtInTheory;
}; // ulong Statistics::Inferences::Simplifying::totalForward() const

ulong Statistics::Inferences::Simplifying::totalBackward() const
{
  return 
    backwardSubsumption +
    backwardDemodulation +
    onBranchBackwardDemodulation;
}; // ulong Statistics::Inferences::Simplifying::totalBackward() const

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Inferences::PropositionalSplitting"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Inferences::PropositionalSplitting::reset()
{  
  accomplished = 0UL;
  components = 0UL;
  introducedComponentNames = 0UL;
}; // void Statistics::Inferences::PropositionalSplitting::reset()


void Statistics::Inferences::PropositionalSplitting::absorb(const PropositionalSplitting& stat)
{  
  accomplished += stat.accomplished;
  components += stat.components;
  introducedComponentNames += stat.introducedComponentNames;
}; // void Statistics::Inferences::PropositionalSplitting::absorb(const PropositionalSplitting& stat)



bool Statistics::Inferences::PropositionalSplitting::somethingRegistered() const
{
  return accomplished || components || introducedComponentNames;
}; 

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Inferences::NegativeEqualitySplitting"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Inferences::NegativeEqualitySplitting::reset()
{  
  accomplished = 0UL;
  argumentsSplitOff = 0UL;
  introducedNames = 0UL;
}; // void Statistics::Inferences::NegativeEqualitySplitting::reset()


void Statistics::Inferences::NegativeEqualitySplitting::absorb(const NegativeEqualitySplitting& stat)
{  
  accomplished += stat.accomplished;
  argumentsSplitOff += stat.argumentsSplitOff;
  introducedNames += stat.introducedNames;
}; // void Statistics::Inferences::NegativeEqualitySplitting::absorb(const NegativeEqualitySplitting& stat)



bool Statistics::Inferences::NegativeEqualitySplitting::somethingRegistered() const
{
  return accomplished || argumentsSplitOff || introducedNames;
}; // bool Statistics::Inferences::NegativeEqualitySplitting::somethingRegistered() const

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Clauses"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Clauses::reset()
{
  generated.reset();
  retained.reset();
}; // void Statistics::Clauses::reset()

void Statistics::Clauses::absorb(const Clauses& stat)
{
  generated.absorb(stat.generated);
  retained.absorb(stat.retained);
}; // void Statistics::Clauses::absorb(const Clauses& stat)



ostream& Statistics::Clauses::outputDetailed(ostream& str) const
{
  str << "%== Clauses\n";

  str << "%     generated " << generated.total() << "\n";
  if (generated.byResolution)
    str << "%         by resolution " << generated.byResolution << "\n"; 
  if (generated.byForwardSuperposition)
    str << "%         by forw superposition " << generated.byForwardSuperposition << "\n"; 
  if (generated.byBackwardSuperposition)
    str << "%         by back superposition " << generated.byBackwardSuperposition << "\n"; 
  if (generated.byEqualityFactoring)
    str << "%         by eq factoring " << generated.byEqualityFactoring << "\n"; 
  if (generated.byEqualityResolution)
    str << "%         by eq resolution " << generated.byEqualityResolution << "\n"; 
  if (generated.byForwardChaining)
    str << "%         by forw chaining " << generated.byForwardChaining << "\n"; 
  if (generated.byBackwardChaining)
    str << "%         by back chaining " << generated.byBackwardChaining << "\n"; 
  if (generated.byBackwardDemodulation)
    str << "%         by back demodulation " << generated.byBackwardDemodulation << "\n"; 
  if (generated.byOnBranchBackwardDemodulation)
    str << "%         by forw demodulation " << generated.byOnBranchBackwardDemodulation << "\n"; 
  if (generated.byPropositionalSplitting)
    str << "%         by prop splitting " << generated.byPropositionalSplitting << "\n"; 
  if (generated.byNegativeEqualitySplitting)
    str << "%         by neg eq splitting " << generated.byNegativeEqualitySplitting << "\n"; 
  if (generated.byOther)
    str << "%         by other " << generated.byOther << "\n"; 
  
  str << "%       reduntant " << generated.totalRedundant() << "\n";
  if (generated.discardedPropositionalTautologies)
    str << "%         prop tautologies " << generated.discardedPropositionalTautologies << "\n";
  if (generated.discardedEquationalTautologies)
    str << "%         eq tautologies " << generated.discardedEquationalTautologies << "\n";
  if (generated.reducedInNonBasicPosition)
    str << "%         red in non-basic pos " << generated.reducedInNonBasicPosition << "\n";
  if (generated.subsumed)
    str << "%         subsumed " << generated.subsumed << "\n";


  str << "%       nonreduntant discarded " << generated.totalNonredundantDiscarded() << "\n";
  if (generated.discardedWithTooDeepInferences)
    str << "%         too deep inference " << generated.discardedWithTooDeepInferences << "\n";
  if (generated.discardedTooLarge)
    str << "%         too large " << generated.discardedTooLarge << "\n";
  if (generated.discardedWithTooLargeLiterals)
    str << "%         too large literals " << generated.discardedWithTooLargeLiterals << "\n";
  if (generated.discardedWithTooManyLiterals)
    str << "%         too many literals " << generated.discardedWithTooManyLiterals << "\n";
  if (generated.discardedTooDeep)
    str << "%         too deep " << generated.discardedTooDeep << "\n";
  if (generated.discardedWithTooManyVariables)
    str << "%         too many variables " << generated.discardedWithTooManyVariables << "\n";
  if (generated.discardedForOtherReasons)
    str << "%         other reasons " << generated.discardedForOtherReasons << "\n";
  
  str << "%       simplified " << generated.totalSimplified() << "\n";
  
  if (generated.simplifiedBySubsumptionResolution)
    str << "%         by subs resolution " << generated.simplifiedBySubsumptionResolution << "\n";
  if (generated.simplifiedByDemodulation)
    str << "%         by demodulation " << generated.simplifiedByDemodulation << "\n";
  if (generated.simplifiedByOnBranchDemodulation)
    str << "%         by on-branch demod " << generated.simplifiedByOnBranchDemodulation << "\n";
  if (generated.simplifiedByDestructiveEqualityResolution)
    str << "%         by destr eq res " << generated.simplifiedByDestructiveEqualityResolution << "\n";
  if (generated.simplifiedByBuiltInTheories)
    str << "%         by built-in theories " << generated.simplifiedByBuiltInTheories << "\n";
   
  if (generated.transformedByPropositionalSplitting)
    str << "%       transformed by prop splitting " << generated.transformedByPropositionalSplitting << "\n";
  if (generated.transformedByNegativeEqualitySplitting)
    str << "%       transformed by neg eq splitting " << generated.transformedByNegativeEqualitySplitting << "\n";
  

  str << "%     retained " << retained.total << "\n";
  str << "%       selected " << retained.selected << "\n";
  str << "%       subsumed " 
      << retained.passiveSubsumed + retained.activeSubsumed 
      << " = " << retained.passiveSubsumed << " passive + " 
      << retained.activeSubsumed << " active\n";
  str << "%       simplified " << retained.totalSimplified() << "\n";
  if (retained.passiveSimplifiedByDemodulation +
      retained.activeSimplifiedByDemodulation)
    str << "%         by demodulation " 
	<< (retained.passiveSimplifiedByDemodulation +
	    retained.activeSimplifiedByDemodulation)
	<< " = " << retained.passiveSimplifiedByDemodulation << " passive + "
        << retained.activeSimplifiedByDemodulation << " active \n";
  if (retained.passiveSimplifiedByOnBranchDemodulation +
      retained.activeSimplifiedByOnBranchDemodulation)
    str << "%         by on-branch demodulation " 
	<< (retained.passiveSimplifiedByOnBranchDemodulation +
	    retained.activeSimplifiedByOnBranchDemodulation)
	<< " = " << retained.passiveSimplifiedByOnBranchDemodulation << " passive + "
        << retained.activeSimplifiedByOnBranchDemodulation << " active\n";
  str << "%       orphans murdered " 
      << retained.passiveOrphansMurdered + retained.activeOrphansMurdered 
      << " = " << retained.passiveOrphansMurdered << " passive + " 
      << retained.activeOrphansMurdered << " active\n";
  str << "%       nonredundant recycled " 
      << retained.passiveRecycled + retained.activeRecycled 
      << " = " << retained.passiveRecycled << " passive + " 
      << retained.activeRecycled << " active\n";
  return str;
}; // ostream& Statistics::Clauses::outputDetailed(ostream& str) const


ostream& Statistics::Clauses::outputSummary(ostream& str) const
{
  str << "%== Generated clauses " << generated.total() << " = " 
      << generated.totalGeneratedByDeduction() << " ded + "
      << generated.totalGeneratedBySimplification() << " simpl";
  if (generated.byPropositionalSplitting)
    str << " + " << generated.byPropositionalSplitting << " prop spl";
  if (generated.byNegativeEqualitySplitting)
    str << " + " << generated.byNegativeEqualitySplitting << " neg eq spl";
  if (generated.byOther)
    str << " + " << generated.byOther << " other";
  str << "\n";

  str << "%     redundant " << generated.totalRedundant() << " = "
      << generated.discardedPropositionalTautologies << " prop taut + "
      << generated.discardedEquationalTautologies << " eq taut + "
      << generated.reducedInNonBasicPosition << " red nb + "
      << generated.subsumed << " subs\n";
    
  str << "%     nonredundant discarded " << generated.totalNonredundantDiscarded() << "\n";
  str << "%     simplified " << generated.totalSimplified() << "\n";
  if (generated.transformedByPropositionalSplitting)
    str << "%     transformed by prop splitting " << generated.transformedByPropositionalSplitting << "\n";
  if (generated.transformedByNegativeEqualitySplitting)
    str << "%     transformed by neg eq splitting " << generated.transformedByNegativeEqualitySplitting << "\n";
  

  str << "%== Retained clauses " << retained.total << "\n";
  str << "%     selected " << retained.selected << "\n";
  str << "%     subsumed " 
      << retained.passiveSubsumed + retained.activeSubsumed 
      << " = " << retained.passiveSubsumed << " passive + " 
      << retained.activeSubsumed << " active\n";
  str << "%     simplified " << retained.totalSimplified() << "\n";
  if (retained.passiveSimplifiedByDemodulation +
      retained.activeSimplifiedByDemodulation)
    str << "%       by demodulation " 
	<< (retained.passiveSimplifiedByDemodulation +
	    retained.activeSimplifiedByDemodulation)
	<< " = " << retained.passiveSimplifiedByDemodulation << " passive + "
        << retained.activeSimplifiedByDemodulation << " active \n";
  if (retained.passiveSimplifiedByOnBranchDemodulation +
      retained.activeSimplifiedByOnBranchDemodulation)
    str << "%       by on-branch demodulation " 
	<< (retained.passiveSimplifiedByOnBranchDemodulation +
	    retained.activeSimplifiedByOnBranchDemodulation)
	<< " = " << retained.passiveSimplifiedByOnBranchDemodulation << " passive + "
        << retained.activeSimplifiedByOnBranchDemodulation << " active\n";
  str << "%     orphans murdered " 
      << retained.passiveOrphansMurdered + retained.activeOrphansMurdered 
      << " = " << retained.passiveOrphansMurdered << " passive + " 
      << retained.activeOrphansMurdered << " active\n";
  str << "%     nonredundant recycled " 
      << retained.passiveRecycled + retained.activeRecycled 
      << " = " << retained.passiveRecycled << " passive + " 
      << retained.activeRecycled << " active\n";

  return str;
}; // ostream& Statistics::Clauses::outputSummary(ostream& str) const

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Clauses::Generated"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Clauses::Generated::reset()
{
  byResolution = 0UL;
  byForwardSuperposition = 0UL;
  byBackwardSuperposition = 0UL;
  byEqualityFactoring = 0UL;
  byEqualityResolution = 0UL;
  byForwardChaining = 0UL;
  byBackwardChaining = 0UL;
  byBackwardDemodulation = 0UL;
  byOnBranchBackwardDemodulation = 0UL;
  byPropositionalSplitting = 0UL;
  byNegativeEqualitySplitting = 0UL;
  byOther = 0UL;

  discardedPropositionalTautologies = 0UL;
  discardedEquationalTautologies = 0UL;
  reducedInNonBasicPosition = 0UL;
  subsumed = 0UL;
	
  discardedWithTooDeepInferences = 0UL;
  discardedTooLarge = 0UL;
  discardedWithTooLargeLiterals = 0UL;
  discardedWithTooManyLiterals = 0UL;
  discardedTooDeep = 0UL;
  discardedWithTooManyVariables = 0UL;
  discardedForOtherReasons = 0UL;
	
  simplifiedBySubsumptionResolution = 0UL;
  simplifiedByDemodulation = 0UL;
  simplifiedByOnBranchDemodulation = 0UL;
  simplifiedByDestructiveEqualityResolution = 0UL;
  simplifiedByBuiltInTheories = 0UL;
	
  transformedByPropositionalSplitting = 0UL;
  transformedByNegativeEqualitySplitting = 0UL;
}; // void Statistics::Clauses::Generated::reset()

void Statistics::Clauses::Generated::absorb(const Generated& stat)
{
  byResolution += stat.byResolution;
  byForwardSuperposition += stat.byForwardSuperposition;
  byBackwardSuperposition += stat.byBackwardSuperposition;
  byEqualityFactoring += stat.byEqualityFactoring;
  byEqualityResolution += stat.byEqualityResolution;
  byForwardChaining += stat.byForwardChaining;
  byBackwardChaining += stat.byBackwardChaining;
  byBackwardDemodulation += stat.byBackwardDemodulation;
  byOnBranchBackwardDemodulation += stat.byOnBranchBackwardDemodulation;
  byPropositionalSplitting += stat.byPropositionalSplitting;
  byNegativeEqualitySplitting += stat.byNegativeEqualitySplitting;
  byOther += stat.byOther;

  discardedPropositionalTautologies += stat.discardedPropositionalTautologies;
  discardedEquationalTautologies += stat.discardedEquationalTautologies;
  reducedInNonBasicPosition += stat.reducedInNonBasicPosition;
  subsumed += stat.subsumed;
	
  discardedWithTooDeepInferences += stat.discardedWithTooDeepInferences;
  discardedTooLarge += stat.discardedTooLarge;
  discardedWithTooLargeLiterals += stat.discardedWithTooLargeLiterals;
  discardedWithTooManyLiterals += stat.discardedWithTooManyLiterals;
  discardedTooDeep += stat.discardedTooDeep;
  discardedWithTooManyVariables += stat.discardedWithTooManyVariables;
  discardedForOtherReasons += stat.discardedForOtherReasons;
	
  simplifiedBySubsumptionResolution += stat.simplifiedBySubsumptionResolution;
  simplifiedByDemodulation += stat.simplifiedByDemodulation;
  simplifiedByOnBranchDemodulation += stat.simplifiedByOnBranchDemodulation;
  simplifiedByDestructiveEqualityResolution += stat.simplifiedByDestructiveEqualityResolution;
  simplifiedByBuiltInTheories += stat.simplifiedByBuiltInTheories;
	
  transformedByPropositionalSplitting += stat.transformedByPropositionalSplitting;
  transformedByNegativeEqualitySplitting += stat.transformedByNegativeEqualitySplitting;
}; // void Statistics::Clauses::Generated::absorb(const Generated& stat)





ulong Statistics::Clauses::Generated::total() const
{
  return
    byResolution +
    byForwardSuperposition +
    byBackwardSuperposition +
    byEqualityFactoring +
    byEqualityResolution +
    byForwardChaining +
    byBackwardChaining +
    byBackwardDemodulation +
    byOnBranchBackwardDemodulation +
    byPropositionalSplitting +
    byNegativeEqualitySplitting +
    byOther;
}; // ulong Statistics::Clauses::Generated::total() const

ulong Statistics::Clauses::Generated::totalGeneratedByDeduction() const
{
  return
    byResolution +
    byForwardSuperposition +
    byBackwardSuperposition +
    byEqualityFactoring +
    byEqualityResolution +
    byForwardChaining +
    byBackwardChaining;
}; // ulong Statistics::Clauses::Generated::totalGeneratedByDeduction() const

ulong Statistics::Clauses::Generated::totalGeneratedBySimplification() const
{
  return
    byBackwardDemodulation +
    byOnBranchBackwardDemodulation;
}; // ulong Statistics::Clauses::Generated::totalGeneratedBySimplification() const

ulong Statistics::Clauses::Generated::totalRedundant() const
{
  return 
    discardedPropositionalTautologies + 
    discardedEquationalTautologies +
    reducedInNonBasicPosition +
    subsumed;
}; // ulong Statistics::Clauses::Generated::totalRedundant() const 


ulong Statistics::Clauses::Generated::totalNonredundantDiscarded() const
{
  return
    discardedWithTooDeepInferences +
    discardedTooLarge +
    discardedWithTooLargeLiterals +
    discardedWithTooManyLiterals +
    discardedTooDeep +
    discardedWithTooManyVariables +
    discardedForOtherReasons;
}; // ulong Statistics::Clauses::Generated::totalNonredundantDiscarded() const

ulong Statistics::Clauses::Generated::totalSimplified() const
{
  return
    simplifiedBySubsumptionResolution +
    simplifiedByDemodulation +
    simplifiedByOnBranchDemodulation +
    simplifiedByDestructiveEqualityResolution +
    simplifiedByBuiltInTheories;
}; // ulong Statistics::Clauses::Generated::totalSimplified() const

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STATISTICS
#  define DEBUG_NAMESPACE "Statistics::Clauses::Retained"
#endif
#include "debugMacros.hpp"
//===================================================================


void Statistics::Clauses::Retained::reset()
{
  total = 0UL;
  selected = 0UL;
  passiveSubsumed = 0UL;
  activeSubsumed = 0UL;
  passiveSimplifiedByDemodulation = 0UL;
  activeSimplifiedByDemodulation = 0UL;
  passiveSimplifiedByOnBranchDemodulation = 0UL;
  activeSimplifiedByOnBranchDemodulation = 0UL;
  passiveOrphansMurdered = 0UL;
  activeOrphansMurdered = 0UL;
  passiveRecycled = 0UL;
  activeRecycled = 0UL;
}; // void Statistics::Clauses::Retained::reset()

void Statistics::Clauses::Retained::absorb(const Retained& stat)
{
  total += stat.total;
  selected += stat.selected;
  passiveSubsumed += stat.passiveSubsumed;
  activeSubsumed += stat.activeSubsumed;
  passiveSimplifiedByDemodulation += stat.passiveSimplifiedByDemodulation;
  activeSimplifiedByDemodulation += stat.activeSimplifiedByDemodulation;
  passiveSimplifiedByOnBranchDemodulation += stat.passiveSimplifiedByOnBranchDemodulation;
  activeSimplifiedByOnBranchDemodulation += stat.activeSimplifiedByOnBranchDemodulation;
  passiveOrphansMurdered += stat.passiveOrphansMurdered;
  activeOrphansMurdered += stat.activeOrphansMurdered;
  passiveRecycled += stat.passiveRecycled;
  activeRecycled += stat.activeRecycled;
}; // void Statistics::Clauses::Retained::absorb(const Retained& stat)



ulong Statistics::Clauses::Retained::totalSimplified() const
{
  return
    passiveSimplifiedByDemodulation +
    activeSimplifiedByDemodulation +
    passiveSimplifiedByOnBranchDemodulation +
    activeSimplifiedByOnBranchDemodulation;
}; // ulong Statistics::Clauses::Retained::totalSimplified() const


//=================================================================
