//
// File:         TempLitListExtras.hpp
// Description:  Utilities for processing temporary lists
//               of temporary literals.
// Created:      Oct 24, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef TEMP_LIT_LIST_EXTRAS_H
#define TEMP_LIT_LIST_EXTRAS_H
//===================================================================
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "ObjectPool.hpp"
#include "Receiver.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.inl"
#include "BasicTempLitList.hpp"
#include "Term.hpp"

namespace VK
{  
  class SymbolInfo;
  class Clause;
}; // namespace VK
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class TempLitListExtras
  {
  public:
    typedef BasicTempLitList::Element Element;
    typedef TemporaryLiteral::Extras::Handle LitHandle;
    typedef BK::ObjectPool<BasicTempLitList> Pool;
    typedef BK::ObjectPool<BasicTempLitList>::EntryHandle Handle;
    class Characteristics;
    class Context;
  public:

    static long computeWeightBase(const BasicTempLitList& lst);

    static 
    void
    computeSizeAndWeightBase(const BasicTempLitList& lst,
			     long& sz,
			     long& weightBase);
    static 
    void
    computeSizeAndWeightBaseAndNumOfPosEq(const BasicTempLitList& lst,
					  long& sz,
					  long& weightBase,
					  long& numOfPosEq);
    static
    long minimalPosEqLHSSize(const BasicTempLitList& lst);


    static 
    bool 
    instantiationIsPossible(const BasicTempLitList& lst,
			    ulong var,
			    ulong instanceDepth,
			    long instanceSize,
			    bool checkInputNonBasicness);
    
    static
    void
    passInOrder(BasicTempLitList& src,
		BasicTempLitList& dest,
		const VAR_WEIGHTING& varWeighting);

    static 
    void 
    renameVariables(const BasicTempLitList& lst,
		    INT_VAR_RENAMING& varRenaming);
    
    static
    void
    weed(BasicTempLitList& lst,
	 bool& isTautology,
	 bool& isEqTautology);

    static
    void
    weed(BasicTempLitList& lst,
	 bool& isTautology,
	 bool& isEqTautology,
	 const SymbolInfo* commutativityTableForEqTaut,
	 const SymbolInfo* commutativityTableForPropTaut,
	 const SymbolInfo* commutativityTableForTrivialEqRes,
	 const SymbolInfo* commutativityTableForIdenticalLitFactoring,
	 BK::Receiver<const Clause*>* usedCommutativityLaws);

    static 
    bool 
    containsOnlyAnswerLiterals(const BasicTempLitList& lst);

    static 
    bool 
    containsAnswerLiterals(const BasicTempLitList& lst);

    static 
    bool 
    containsSplittingBranches(const BasicTempLitList& lst);
    
    static 
    ulong
    countDifferentVariables(const BasicTempLitList& lst);
    
    static
    void 
    collectVariableOccurences(const BasicTempLitList& lst,
			      WeightPolynomial::Monomials& vars);

    static
    void 
    preorientEqualities(const BasicTempLitList& lst);


    static 
    void adjustSelectionLayers(BasicTempLitList& lst);

    static 
    void freshList(Handle& h,Pool* pool);

    static 
    void freshList(Handle& h);

  private:
    static
    bool complementary(const LitHandle& lit1,
		       const LitHandle& lit2,
		       const SymbolInfo* commutativityTable,
		       BK::Receiver<const Clause*>* usedCommutativityLaws);

    static
    bool equal(const LitHandle& lit1,
	       const LitHandle& lit2,
	       const SymbolInfo* commutativityTable,
	       BK::Receiver<const Clause*>* usedCommutativityLaws);
    

    static
    bool complementaryEq(const LitHandle& lit1,
			 const LitHandle& lit2,
			 const SymbolInfo* commutativityTable,
			 BK::Receiver<const Clause*>* usedCommutativityLaws);

    static
    bool equalEq(const LitHandle& lit1,
		 const LitHandle& lit2,
		 const SymbolInfo* commutativityTable,
		 BK::Receiver<const Clause*>* usedCommutativityLaws);
    


  }; // class TempLitListExtras
}; // namespace VK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::Context"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  class TempLitListExtras::Context
  {
  public:
    class AutoActivator
    {
    public:
      AutoActivator() : _previous(Context::getCurrent())
      {
	Context::setCurrent(static_cast<Context*>(0));
      };
      AutoActivator(Context* c) : _previous(Context::getCurrent())
      {
	Context::setCurrent(c);
      };
      ~AutoActivator()
      {
	Context::setCurrent(_previous);
      };
    private:
      Context* _previous;
    }; // class AutoActivator
  public:
    Context() {};
    ~Context() {};
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
    BK::ObjectPool<BasicTempLitList>* defaultPool()
    {
      return static_cast<BK::ObjectPool<BasicTempLitList>*>(&_defaultPool);
    };
  private:
    BK::ObjectPoolWithMemory<BK::GlobAlloc,BasicTempLitList> _defaultPool;
    static Context* _current;
  }; // class TempLitListExtras::Context
    

}; // namespace VK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::Characteristics"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  class TempLitListExtras::Characteristics
  {
  public:
    Characteristics() {};
    ~Characteristics() {};
    void reset();
    ulong numOfAllLits() const { return _numOfNonanswerLits + _numOfAnswerLits; };
    ulong numOfNonanswerLits() const { return _numOfNonanswerLits; };
    ulong numOfAnswerLits() const { return _numOfAnswerLits; };
    ulong numOfEqLits() const { return _numOfEqLits; };
    ulong numOfNoneqNonanswerLits() const { return _numOfNonanswerLits - _numOfEqLits;  };
    ulong numOfNegNonanswerLits() const { return _numOfNegNonanswerLits; };
    ulong numOfPosNonanswerLits() const { return _numOfNonanswerLits - _numOfNegNonanswerLits; };
    ulong numOfNegSplittingBranches() const { return _numOfNegSplittingBranches; };
    ulong numOfPosSplittingBranches() const { return _numOfPosSplittingBranches; };
    const TERM& maxNoneqHeader() const { return _maxNoneqHeader; };
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    ulong numOfNextStateLits() const { return _numOfNextStateLits; };
    ulong numOfCurrStateLits() const { return _numOfCurrStateLits; };
#endif
    
    void collectFor(const BasicTempLitList& lst)
    {
      CALL("collectFor(const BasicTempLitList& lst)");
      reset();
      addFor(lst);
    };
    void addFor(const BasicTempLitList& lst);
    void addFor(const LitHandle& lit);

  private:
    ulong _numOfNonanswerLits;
    ulong _numOfAnswerLits;
    ulong _numOfEqLits;
    ulong _numOfNegNonanswerLits;
    ulong _numOfNegSplittingBranches;
    ulong _numOfPosSplittingBranches;
    TERM _maxNoneqHeader;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    ulong _numOfNextStateLits;
    ulong _numOfCurrStateLits;
#endif
  }; // class TempLitListExtras::Characteristics
}; // namespace VK



//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  inline
  bool 
  TempLitListExtras::containsOnlyAnswerLiterals(const BasicTempLitList& lst)
  {
    CALL("containsOnlyAnswerLiterals(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      if (!el->literal()->isAnswerLiteral())
	return false;
    return true;
  }; // containsOnlyAnswerLiterals(const BasicTempLitList& lst)

  inline
  bool 
  TempLitListExtras::containsAnswerLiterals(const BasicTempLitList& lst)
  {
    CALL("containsAnswerLiterals(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      if (el->literal()->isAnswerLiteral())
	return true;
    return false;
  }; // containsAnswerLiterals(const BasicTempLitList& lst)

  inline
  bool 
  TempLitListExtras::containsSplittingBranches(const BasicTempLitList& lst)
  {
    CALL("containsSplittingBranches(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      if (el->literal()->isSplittingBranch())
	return true;
    return false;
  }; // containsSplittingBranches(const BasicTempLitList& lst)
  
  inline
  void 
  TempLitListExtras::preorientEqualities(const BasicTempLitList& lst)
  {
    CALL("preorientEqualities(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      if (el->literal()->isEquality())
	el->literal()->orderEqArgs();
  };


  inline
  void TempLitListExtras::adjustSelectionLayers(BasicTempLitList& lst)
  {
    CALL("adjustSelectionLayers(BasicTempLitList& lst)");
    if (lst.isEmpty()) return;
    if (lst.length() == 1UL)
      {
	LiteralProperties& prop = lst.first()->literal()->properties();
	prop.setSelectionLayerModifier(LiteralProperties::DefaultSLModifier);
	prop.setSelectionLayer(0UL);
	return;
      };
    // first pass: collect info

    bool defExists = false;
    ulong minLayerWithDef = ULONG_MAX;
    ulong maxLayerWithDef = 0UL;
    
    bool postExists = false;
    ulong minLayerWithPost = ULONG_MAX;
    ulong maxLayerWithPost = 0UL;
    
    bool accelExists = false;
    ulong minLayerWithAccel = ULONG_MAX;
    ulong maxLayerWithAccel = 0UL;

	const Element* el;
    for (el = lst.first(); el; el = el->next())
      {
	LiteralProperties& prop = el->literal()->properties();
	ulong layer = prop.selectionLayer();
	switch (prop.selectionLayerModifier())
	  {
	  case LiteralProperties::DefaultSLModifier: 
	    defExists = true;
	    if (layer < minLayerWithDef) minLayerWithDef = layer;
	    if (layer > maxLayerWithDef) maxLayerWithDef = layer;
	    break;	
	  case LiteralProperties::PostponeSelection:
	    postExists = true;
	    if (layer < minLayerWithPost) minLayerWithPost = layer;
	    if (layer > maxLayerWithPost) maxLayerWithPost = layer;
	    break;	
	    
	  case LiteralProperties::AccelerateSelection:
	    accelExists = true;
	    if (layer < minLayerWithAccel) minLayerWithAccel = layer;
	    if (layer > maxLayerWithAccel) maxLayerWithAccel = layer;
	    break;	
	  };
      };

    if (!postExists && !accelExists && minLayerWithDef == 0UL)
      return; 

    ulong incForDef = 
      (accelExists) ? ((maxLayerWithAccel - minLayerWithAccel) + 1) : 0UL;
    ulong incForPost = 
      (defExists) ? ((maxLayerWithDef - minLayerWithDef) + 1 + incForDef) : incForDef;

    // second pass: adjust the layers and reset layer modifiers

    for (el = lst.first(); el; el = el->next())
      {
	LiteralProperties& prop = el->literal()->properties();
	ulong layer = prop.selectionLayer();
	switch (prop.selectionLayerModifier())
	  {
	  case LiteralProperties::DefaultSLModifier:
	    prop.setSelectionLayer(layer - minLayerWithDef + incForDef);
	    break;	
	  case LiteralProperties::PostponeSelection:
	    prop.setSelectionLayer(layer - minLayerWithPost + incForPost);
	    break;	
	  case LiteralProperties::AccelerateSelection:
	    prop.setSelectionLayer(layer - minLayerWithAccel);
	    break;	
	  };
	prop.setSelectionLayerModifier(LiteralProperties::DefaultSLModifier);
      };
  }; // void TempLitListExtras::adjustSelectionLayers(BasicTempLitList& lst)

  inline
  void TempLitListExtras::freshList(Handle& h,Pool* pool)
  {
    CALL("freshList(Handle& h,Pool* pool)");
    ASSERT(pool);
    pool->reserveObject(h);
    h->reset();
  };
  
  inline
  void TempLitListExtras::freshList(Handle& h)
  {
    CALL("freshList(Handle& h)");
    ASSERT(Context::getCurrent());
    freshList(h,Context::getCurrent()->defaultPool());
  };
  
}; // namespace VK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMP_LIT_LIST_EXTRAS
#  define DEBUG_NAMESPACE "TempLitListExtras::Characteristics"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  inline
  void TempLitListExtras::Characteristics::reset()
  {
    CALL("reset()");
    _numOfNonanswerLits = 0UL;
    _numOfAnswerLits = 0UL;
    _numOfEqLits = 0UL;
    _numOfNegNonanswerLits = 0UL;
    _numOfNegSplittingBranches = 0UL;
    _numOfPosSplittingBranches = 0UL;
#ifdef VKERNEL_FOR_STEP_RESOLUTION
    _numOfNextStateLits = 0UL;
    _numOfCurrStateLits = 0UL;
#endif
  }; // void TempLitListExtras::Characteristics::reset()
  
  inline
  void TempLitListExtras::Characteristics::addFor(const BasicTempLitList& lst)
  {
    CALL("addFor(const BasicTempLitList& lst)");
    for (const Element* el = lst.first(); el; el = el->next())
      addFor(el->literal());
  }; // void TempLitListExtras::Characteristics::addFor(const BasicTempLitList& lst)

  inline
	  void TempLitListExtras::Characteristics::addFor(const TemporaryLiteral::Extras::Handle& lit)
  {
    CALL("addFor(const TemporaryLiteral::Handle& lit)");
    if (lit->isAnswerLiteral())
      {
	ASSERT(!lit->isEquality());      
	++_numOfAnswerLits;
      }
    else // nonanswer literal
      {
	++_numOfNonanswerLits;
	if (lit->isEquality()) 
	  {
	    ++_numOfEqLits;
	  }
	else
	  {
	    if (_numOfNonanswerLits - _numOfEqLits)
	      {
		if (Signature::current()->compareHeaderPrecedences(lit->header().polarity(),
								   lit->header().functor(),
								   _maxNoneqHeader.polarity(),
								   _maxNoneqHeader.functor()) == 
		    BK::Greater)
		    {
		      _maxNoneqHeader = lit->header();
		    };
	      }
	    else
	      _maxNoneqHeader = lit->header();
	    if (lit->isSplittingBranch())
	      if (lit->isNegativeForSelection())
		{
		  ++_numOfNegSplittingBranches;
		}
	      else 
		++_numOfPosSplittingBranches;
	  };
	if (lit->isNegativeForSelection()) 
	  ++_numOfNegNonanswerLits;
      };

  }; // void TempLitListExtras::Characteristics::addFor(const TemporaryLiteral::Handle& lit)
}; // namespace VK

//===================================================================

#endif
