//
// File:         TemporaryClauseQueue.cpp
// Description:  Queues for communicating temporary clauses
//               between different modules.
// Created:      Nov 08, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#include "TemporaryClauseQueue.hpp"
using namespace BK;
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_QUEUE
#  define DEBUG_NAMESPACE "TemporaryClauseQueue"
#endif
#include "debugMacros.hpp"
//===================================================================


TemporaryClauseQueue::TemporaryClauseQueue() :
  _nextIn(0UL),
  _nextOut(0UL)
{
#ifndef NO_DEBUG
  _name = "?";
#endif
  DOP(_store.freeze());
};

#ifndef NO_DEBUG
TemporaryClauseQueue::TemporaryClauseQueue(const char* name) :
  _nextIn(0UL),
  _nextOut(0UL)
{
  _name = name;
  DOP(_store.freeze());
};
#endif

TemporaryClauseQueue::~TemporaryClauseQueue()
{
};


void TemporaryClauseQueue::reset()
{
  CALLM(_name,"reset()");
  BK::destroy(_store);
  BK::init(_store);
  _nextIn = 0UL;
  _nextOut = 0UL;
}; // void TemporaryClauseQueue::reset()


void TemporaryClauseQueue::push(const ClauseHandle& cl)
{
  CALLM(_name,"push(const ClauseHandle& cl)");
  if (_store.size() == 0UL) 
    {
      ASSERT(_nextIn == 0UL);
      ASSERT(_nextOut == 0UL);
      DOP(_store.unfreeze());
      _store.expand();
      DOP(_store.freeze());
    };
  ASSERT(_nextIn < _store.size());
  ASSERT(_store[_nextIn].isNull());
  _store[_nextIn] = cl;
  incNextIn();
}; // void TemporaryClauseQueue::push(const ClauseHandle& cl)


void TemporaryClauseQueue::collect(ClauseHandle& cl)
{
  CALLM(_name,"collect(ClauseHandle& cl)");
  if (_store.size() == 0UL) 
    {
      ASSERT(_nextIn == 0UL);
      ASSERT(_nextOut == 0UL);
      DOP(_store.unfreeze());
      _store.expand();
      DOP(_store.freeze());
    };
  ASSERT(_nextIn < _store.size());
  ASSERT(_store[_nextIn].isNull());
  _store[_nextIn].collectFrom(cl);
  incNextIn();
}; // void TemporaryClauseQueue::collect(ClauseHandle& cl)


 
TemporaryClauseQueue::ClauseHandle TemporaryClauseQueue::pop()
{
  CALLM(_name,"pop()");
  ulong tmp = _nextOut;   
  incNextOut();
  return ClauseHandle(_store[tmp],0);
}; // ClauseHandle TemporaryClauseQueue::pop()
  
void TemporaryClauseQueue::popTo(ClauseHandle& cl)
{
  CALLM(_name,"popTo(ClauseHandle& cl)");
  ASSERT(!isEmpty());
  cl.collectFrom(_store[_nextOut]);
  incNextOut();
}; // void TemporaryClauseQueue::popTo(ClauseHandle& cl)


void TemporaryClauseQueue::popTo(TemporaryClauseQueue& receiver)
{
  CALLM(_name,"popTo(TemporaryClauseQueue& receiver)");
  ASSERT(!isEmpty());
  if (receiver._store.size() == 0UL) 
    {
      ASSERT(receiver._nextIn == 0UL);
      ASSERT(receiver._nextOut == 0UL);
      DOP(receiver._store.unfreeze());
      receiver._store.expand();
      DOP(receiver._store.freeze());
    };
  ASSERT(receiver._nextIn < receiver._store.size());
  ASSERT(receiver._store[receiver._nextIn].isNull());
  receiver._store[receiver._nextIn].collectFrom(_store[_nextOut]);
  receiver.incNextIn();
  incNextOut();
}; // void TemporaryClauseQueue::popTo(ClauseHandle& cl) 

inline void TemporaryClauseQueue::incNextIn()
{
  CALLM(_name,"incNextIn()");
  ++_nextIn;
  if (_nextIn == _store.size()) _nextIn = 0UL;
  if (_nextIn == _nextOut)
    {
      DOP(_store.unfreeze());
      ulong createdExtraCapacity;
      _store.stretch(_nextIn,createdExtraCapacity);
      DOP(_store.freeze());
      ASSERT(createdExtraCapacity > 0UL);
      _nextOut += createdExtraCapacity;
    };
}; // void TemporaryClauseQueue::incNextIn()



ostream& TemporaryClauseQueue::output(ostream& str) const
{
  CALLM(_name,"output(ostream& str) const");
  if (_nextIn >= _nextOut)
    {
      for (ulong i = _nextOut; i < _nextIn; ++i)
	_store[i]->outputComplete(str) << "\n";
    }
  else
    {
      for (ulong i = _nextOut; i < _store.size(); ++i)
	_store[i]->outputComplete(str) << "\n";
      for (ulong j = 0UL; j < _nextIn; ++j)
	_store[j]->outputComplete(str) << "\n";
    };
  return str;
}; // ostream& TemporaryClauseQueue::output(ostream& str) const


//=================================================================
