//
// File:         VampireKernel.hpp
// Description:  Vampire kernel interface. 
// Created:      May 19, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//======================================================================
#ifndef VAMPIRE_KERNEL_H
#define VAMPIRE_KERNEL_H
//======================================================================
#include <iostream>
#include <fstream>
#include <string>
#include "jargon.hpp"
//======================================================================

class VampireKernel
{
 public:
  typedef long TermWeightType;
  class Options;
  class Emergency;
  enum TerminationReason
  {
    UnknownTerminationReason,
    RefutationFound,
    Satisfiable,
    EmptyPassive,
    MaxPassive,
    MaxActive,
    TimeLimitExpired,
    AbortedOnAllocationFailure,
    AbortedOnExternalInterrupt,
    AbortedOnErrorSignal,
    AbortedOnError
  };
  class Statistics;
  class Subterm;
  class Literal;
  class Clause;
  class ClauseSetIterator;
  class LiteralAnnotation;
  class ClauseAnnotation;
 public:
  // Initialisation and destruction
  VampireKernel();
  ~VampireKernel();
  void openSession(const Options& opt);
  void closeSession();
  
  // Version information
  
  static float versionNumber();
  static string versionDescription();

  // Receiving signature information

  void beginInputSignature();  
  void endInputSignature();
  void registerInputFunction(string name,
			     ulong arity,
			     ulong& id); 

  void registerInputPredicate(string name,
			      ulong arity,
			      bool isAnswer,
			      ulong& id); 
  
  void setSymbolWeightConstPart(ulong symId,const TermWeightType& weight);
  void setSymbolPrecedence(ulong symId,long prec); 
  // ^ prec cannot be 0
 
  void setHeaderPrecedence(bool positive,ulong symId,long prec); 
  // ^ prec cannot be -1, 0 or 1
  
  void interpretFunction(string inputSymbol,
			 ulong arity,
			 string theory,
			 string nativeSymbol);
  void interpretPredicate(string inputSymbol,
			  ulong arity,
			  string theory,
			  string nativeSymbol);

  // Receiving input clauses
  
  void beginInputClauses();
  void endInputClauses();
  
  void beginInputClause(ulong clauseNumber,
			void* origin,
			bool goal,
			bool support);
  
  void beginInputClause(ulong clauseNumber,
			void* origin,
			const ClauseAnnotation& annotation);
  
  void endInputClause();

  void beginInputLiteralDefinition(ulong clauseNumber,
                                   void* origin,
				   bool goal,
				   bool support);
  
  void endInputLiteralDefinition();

  void beginInputLiteralDefinitionHead(bool positive,ulong predicateId);
  void endInputLiteralDefinitionHead();

  void beginInputLiteralDefinitionTail();
  void endInputLiteralDefinitionTail();

  void beginInputLiteralDefinitionTailComponent();
  void endInputLiteralDefinitionTailComponent();

  void beginLiteral(bool positive,ulong predicateId);
  void beginLiteral(bool positive,ulong predicateId,bool preselected);
  void beginLiteral(bool positive,
		    ulong predicateId,
		    const LiteralAnnotation& annotation);
  void endLiteral();  


  void receiveFunction(ulong symId);
  void receiveNumericConstant(double value);
  void receiveVariable(ulong varNum);
  
  // Searching for a proof, includes preprocessing
  
  void saturate(ulong nextVacantClauseId,
		long timeLimit); // deciseconds 
  void lookForAnotherAnswer();
  // throws Emergency and intercepts SIGVTALRM
  static VampireKernel* currentSession();
  void suspend();
  void resume();
  void tabulate();
  void reportInterrupt(int signalNumber);
  void reportErrorSignal(int signalNumber);
  void reportError();

  // Communicating search results

  TerminationReason terminationReason() const;
  const Clause* refutation() const;
  ClauseSetIterator* activeClauses(); // saturation if everything is kosher
  void disposeOf(ClauseSetIterator* iterator); 

  // Internal signature interface

  bool isNumericConstantId(ulong symbolId) const;
  bool isInputSymbol(ulong symbolId) const; 
  bool isAnswerPredicateId(ulong symbolId) const;
  ulong symbolArity(ulong symbolId) const;
  string symbolPrintName(ulong symbolId) const; 
  // ^ does not work for numeric constants
  double numericConstantValue(ulong symbolId) const; 

  // Getting the session statistics

  const Statistics& statistics(); 
  
  void outputPreproStat(ostream* str,bool full);
  void outputCurrentLoopStat(ostream* str,bool full);
  void outputCumulativeStat(ostream* str,bool full);


 private:
  class _Implementation; 
  
 private:
  _Implementation* _implementation;
}; // class VampireKernel

//======================================================================

class VampireKernel::Emergency
{
 public:
  enum Type
  {
    AllocationFailure,
    TimeLimit
  };
 public:
  Emergency(const Type& t) : type_(t) {}; 
  ~Emergency() {};
  const Type& type() const { return type_; };
 private:
  Emergency() {};
 private:
  Type type_;
}; // class VampireKernel::Emergency

//======================================================================

class VampireKernel::Statistics
{
public:
  class DeductionInferences
  {
  public:
    ulong totalAccomplished;
    ulong resolutionAccomplished;
    ulong forwardSuperposition;
    ulong backwardSuperposition;
    ulong equalityFactoring;
    ulong equalityResolution;
    ulong forwardChaining;
    ulong backwardChaining;
    ulong totalRedundantDiscarded;
    ulong totalRedundantBranchesDiscarded;
    ulong totalNonredundantDiscarded;
    ulong totalNonredundantBranchesDiscarded;
  }; // class DeductionInferences

  class SimplifyingInferences
  {
  public:
    ulong totalForward;
    ulong propositionalTautologyDeletion;
    ulong equationalTautologyDeletion;
    ulong forwardSubsumption;
    ulong forwardSubsumptionResolution;
    ulong forwardDemodulation;
    ulong onBranchForwardDemodulation;
    ulong destructiveEqualityResolution;
    ulong builtInTheory;
    
    ulong totalBackward;
    ulong backwardSubsumption;
    ulong backwardDemodulation;
    ulong onBranchBackwardDemodulation;
  }; // class SimplifyingInferences

  class PropositionalSplitting
  {
  public:
    ulong accomplished;
    float averageNumOfComponents;
    ulong introducedComponentNames;
  }; // class PropositionalSplitting

  class NegativeEqualitySplitting
  {
  public:
    ulong accomplished;
    float averageNumOfArgumentsSplitOff;
    ulong introducedNames;
  }; // class NegativeEqualitySplitting

  class GeneratedClauses
  {
  public:
    ulong total;
    ulong byDeductionInferences;
    ulong bySimplificationInferences;
    ulong byPropositionalSplitting;
    ulong byNegativeEqualitySplitting;
    ulong byOther;
    
    ulong totalRedundantDiscarded;
    ulong discardedPropositionalTautologies;
    ulong discardedEquationalTautologies;
    ulong discardedSubsumed;
    
    ulong totalNonredundantDiscarded;
    
    ulong totalSimplified;
    ulong transformedByPropositionalSplitting;
    ulong transformedByNegativeEqualitySplitting;
  }; // class GeneratedClauses

  class RetainedClauses
  {
  public:
    ulong total;
    ulong selected;
    ulong totalSubsumed;
    ulong totalSimplified;
    ulong totalSimplifiedByDemodulation;
    ulong totalSimplifiedByOnBranchDemodulation;
    ulong totalOrphansMurdered; // thank the Waldmeister team for the nice name
    ulong totalRecycled;
    ulong currentlyPassive;
    ulong currentlyActive;
  }; // class RetainedClauses

public:
  float time; // seconds
  ulong memory; // bytes
  DeductionInferences deductionInferences;
  SimplifyingInferences simplifyingInferences;
  PropositionalSplitting propositionalSplittingInferences;
  NegativeEqualitySplitting negativeEqualitySplittingInferences;
  
  GeneratedClauses generatedClauses;
  RetainedClauses retainedClauses;
}; // class VampireKernel::Statistics

//======================================================================

class VampireKernel::Subterm
{
 public:
  const Subterm* next() const; 
  // ^ returns 0 if this is the last subterm
  bool isVariable() const;
  ulong variableNumber() const; 
  ulong topFunction() const;
  const Subterm* firstArgument() const; 
  // ^ returns 0 if there are no arguments
 private:
  Subterm() {};
  ~Subterm() {};
  friend class VampireKernel;
}; // class VampireKernel::Subterm

class VampireKernel::Literal
{
 public:
  const Literal* next() const;
  // ^ return 0 if this is the last literal
  bool isPositive() const;
  ulong predicate() const;
  const Subterm* firstArgument() const; 
  // ^ returns 0 if there are no arguments
 private:
  Literal() {};
  ~Literal() {};
  friend class VampireKernel;
}; // class VampireKernel::Literal

//======================================================================

class VampireKernel::Clause
{
 public:
  class Ancestor
    {
    public:
      const Clause* clause() const;
      const Ancestor* next() const;
    private:
      Ancestor() {};
      ~Ancestor() {};
      friend class Clause;
    };

  enum BackgroundFlag
    { 
      IsInputClause,
      IsObtainedDuringPreprocessing,
      BinaryResolutionUsed,
      ForwardTransitiveChainingUsed,
      BackwardTransitiveChainingUsed,
      ForwardSuperpositionUsed,
      BackwardSuperpositionUsed,
      EqualityResolutionUsed,
      EqualityFactoringUsed,
      ForwardDemodulationUsed,
      BackwardDemodulationUsed,
      SimplificationByEqualityResolutionUsed,
      ForwardSubsumptionResolutionUsed,
      BackwardSubsumptionResolutionUsed,
      PropositionalSplittingUsed, 
      NegativeEqualitySplittingUsed, 
      IsReanimatedPassive,
      IsPartOfNameIntroduction,
      SimplificationByBuiltInTheoriesUsed,
      IsNameIntro,
      FoldingUsed,
      ClausificationUsed,
      IsBuiltInTheoryFact
    };

  class BackgroundFlagIterator
  {
  public:
    BackgroundFlagIterator();
    ~BackgroundFlagIterator();
    bool getCurrentFlag(BackgroundFlag& flag) const;
    void goToNextFlag(); 
  private:
    const void* _base;
    unsigned long _offset;
    friend class Clause;
  };

 public:
  ulong number() const;
  bool isLiteralDefinition(const Literal*& definedLiteral,
			   const Literal*& bodyLiterals);
  // ^ if definedLiteral = L(x) and bodyLiterals = {L1(x,y),..,Ln(x,y)},
  // then the clause should be interpreted as
  // FORALL x (L(x) <=> FORALL y (L1(x,y) \/ .. \/ Ln(x,y)))

  const Literal* literals() const; 
  // ^ should not be applied to literal definitions
  ulong numberOfSelectedLiterals() const; 
  // ^ return 0 if the clause is empty,
  //   should not be applied to literal definitions
  bool isInputClause() const;
  void* inputClauseOrigin() const;
  bool isBuiltInTheoryFact() const;
  bool isNameIntroduction() const; 
  // ^ implies backgroundFlagIsSet(IsNameIntro) and isLiteralDefinition(..)
  const Ancestor* firstAncestor() const;
  BackgroundFlagIterator backgroundFlagIterator() const;
  bool backgroundFlagIsSet(BackgroundFlag fl) const; 
 private:
  Clause() {};
  ~Clause() {};
  friend class VampireKernel;
}; // class VampireKernel::Clause

class VampireKernel::ClauseSetIterator
{
public:
  ClauseSetIterator() {};
  virtual ~ClauseSetIterator();
  operator bool() const { return !isEnd(); };
  virtual bool isEnd() const = 0;
  ClauseSetIterator& operator ++() { next(); return *this; };
  virtual void next() = 0;
  virtual const Clause* clause() const = 0;
}; // class VampireKernel::ClauseSetIterator

//======================================================================

class VampireKernel::LiteralAnnotation
{
public:
  enum SelectabilityCategory
  {
    DefaultSelectability,
    LowSelectability,
    HighSelectability
  };
  enum 
  {
    MinimalSelectabilityPriority = -255L,
    MaximalSelectabilityPriority = 255L
  };

  class SelectabilityInheritanceLimit
  {
  public:
    enum
    {
      MaximalFinite = 30UL
    };
  public:
    SelectabilityInheritanceLimit();
    void makeFinite(ulong lim);
    void makeInfinite();
    
    bool isFinite(ulong& lim) const;

  private:
    bool _finite;
    ulong _value;
  }; // class SelectabilityInheritanceLimit

  enum DeploymentMode
  {
    DefaultDeploymentMode,
    RuleHead,
    Subquery
  };

  enum EqualityMode
  {
    DefaultEqualityMode,
    AddResolution,
    ResolutionOnly
  };

public:
  LiteralAnnotation(); 
  void makeDefault();
  void makeDefaultSelectability(long selectabilityPriority,
				const DeploymentMode& deploymentMode,
				const EqualityMode& equalityMode);
  void makeHighSelectability(long selectabilityPriority,
			     const SelectabilityInheritanceLimit& selectabilityInheritanceLimit,
			     const DeploymentMode& deploymentMode,
			     const EqualityMode& equalityMode);
  void makeLowSelectability(long selectabilityPriority,
			    const SelectabilityInheritanceLimit& selectabilityInheritanceLimit,
			    const DeploymentMode& deploymentMode,
			    const EqualityMode& equalityMode);

  const SelectabilityCategory& selectabilityCategory() const;
  long selectabilityPriority() const;
  const SelectabilityInheritanceLimit& selectabilityInheritanceLimit() const;
  const DeploymentMode& deploymentMode() const;
  const EqualityMode& equalityMode() const;
private:
  SelectabilityCategory _selectabilityCategory;
  long _selectabilityPriority;
  SelectabilityInheritanceLimit _selectabilityInheritanceLimit;
  DeploymentMode _deploymentMode;
  EqualityMode _equalityMode;
}; // class VampireKernel::LiteralAnnotation

//======================================================================

class VampireKernel::ClauseAnnotation
{
public:
  ClauseAnnotation(); 
  void makeDefault();
  void make(bool isGoal,bool isSupport,bool isNonSubsuming,bool isNonSubsumable);
  bool isGoal() const;
  bool isSupport() const;
  bool isNonSubsuming() const;
  bool isNonSubsumable() const;  
  void markAsGoal();
  void markAsSupport();
  void markAsNonSubsuming();
  void markAsNonSubsumable();
private:
  bool _isGoal;
  bool _isSupport;
  bool _isNonSubsuming;
  bool _isNonSubsumable;
}; // class VampireKernel::ClauseAnnotation

//======================================================================

class VampireKernel::Options
{
 public:
  enum SimplificationOrdering
  {
    NonrecursiveKBOrdering,
    StandardKBOrdering
  };
  enum TermWeightingScheme
  {
    UniformTermWeighting,
    NonuniformConstantTermWeighting
  };
 public:
  Options() { setDefaultValues(); };
  ~Options() {};
  void setDefaultValues();
  std::ostream& output(std::ostream& str);
 public:
  
  // session attributes
  const char* job_id;
  

  // main algorithm
    
  long main_alg;

  // inference rules
  bool no_inferences_between_rules;
  bool static_splitting;
  bool dynamic_splitting;
  long splitting_style;
  bool small_basic_comp_in_hyp_splitting;
  long static_neg_eq_splitting;
  long stat_neg_eq_spl_var_count_limit;
  long dynamic_neg_eq_splitting;
  long dyn_neg_eq_spl_var_count_limit;
  bool paramodulation;
  bool paramodulate_from_variables;
  long sim_back_sup;
  bool forw_sup_no_dup_red;
  bool back_sup_no_dup_red;
  bool prohibit_non_basic_superposition;
  bool prohibit_non_basic_fd;
  bool prohibit_non_basic_bd;
  bool reduce_non_basic;
  bool input_non_basicness;
  bool run_time_non_basicness;
  long selection;
  long literal_comparison_mode;
  bool inherited_negative_selection;
     
  // strategy 
  bool forward_subsumption;
  bool fs_set_mode;
  bool fs_use_sig_filters;
  long simplify_by_fs;
  long main_premise_subsumption;
  long side_premise_subsumption;
  long main_selected_lit_demodulation;
  long side_selected_lit_demodulation;
  long forward_demodulation;
  bool fd_on_splitting_branches;
  bool normalize_by_commutativity;
  bool eq_taut_mod_C;
  bool prop_taut_mod_C;
  bool trivial_eq_res_mod_C;
  bool identical_lit_factoring_mod_C;
  bool chaining_for_transitive;
  bool chaining_through_variables;
  long backward_demodulation;
  bool bd_on_splitting_branches;
  bool simplify_by_eq_res;
  bool backward_subsumption;
  bool bs_set_mode;
  long orphan_murder;
  long lrs;
  long elim_def;
  SimplificationOrdering simplification_ordering;
  long symbol_precedence_by_arity;
  bool header_precedence_kinky;
  TermWeightingScheme term_weighting_scheme;
  long symbol_weight_by_arity;
  bool orient_pos_equalities_in_support;
  bool orient_neg_equalities_in_support;
  LiteralAnnotation::EqualityMode all_pos_eq_mode;
  LiteralAnnotation::EqualityMode all_neg_eq_mode;

  // numeric options 
  long memory_limit; // kilobytes
  long allocation_buffer_size;
  double allocation_buffer_price;
  long max_skolem; 
  long max_active;
  long max_passive;
  long max_weight;
  long max_inference_depth;
  long age_weight_ratio_numerator;
  long age_weight_ratio_denominator;
  long best_good_weight_ratio_numerator;
  long best_good_weight_ratio_denominator;
  float selected_pos_eq_penalty_coefficient;    
  float nongoal_penalty_coefficient;
  float lrs_softness;
  long first_time_check;   
  long first_mem_check;   
    
  // interface options    
  bool show_opt;
  bool show_input;
  bool show_prepro;
  bool show_gen;
  bool show_passive;
  bool show_active;
  bool show_name_intro;
  bool show_lit_prop; 
  long show_passive_profile_interval;
  bool proof;
  bool auto_stat_prepro;
  bool auto_stat_curr_loop;
  bool auto_stat_cumulative;
  bool auto_stat_full;
  bool auto_stat_summary;
  bool show_profile;
  bool show_weight_limit_changes;  
  std::ofstream* tabulation;
  std::ostream* std_output;

#ifdef VKERNEL_FOR_STEP_RESOLUTION
  bool step_resolution_literal_selection;
#endif

}; // class VampireKernel::Options

//======================================================================

namespace std
{
inline 
ostream& operator<<(ostream& str,
		    const VampireKernel::Options::SimplificationOrdering& so)
{
  switch (so)
    {
    case VampireKernel::Options::NonrecursiveKBOrdering: 
      str << "nonrecursive_KB";
      break;
    case VampireKernel::Options::StandardKBOrdering: 
      str << "standard_KB";
      break;
    };
  return str;
};

inline 
ostream& operator<<(ostream& str,
		    const VampireKernel::Options::TermWeightingScheme& tws)
{
  switch (tws)
    {
    case VampireKernel::Options::UniformTermWeighting:
      str << "uniform";
      break;
    case VampireKernel::Options::NonuniformConstantTermWeighting:
      str << "nonuniform_constant";
      break;
    };
  return str;
};
};

//======================================================================
#endif



