//
// File:         backward_superposition.cpp
// Description:  Classes dealing with backward superposition.
// Created:      May 06, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::CompiledConstraintForUnifiers
//               replaced ORD_CONSTR
// Revised:      May 16, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bug fix in RecollectMaxWeightForm().
// Revised:      Jul 15, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Major refurbishment.
// Revised:      Jul 16, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Computation of main premise residual size was optimised.
//==================================================================
#include "backward_superposition.hpp"
#include "DestructionMode.hpp"
#include "VampireKernelInterceptMacros.hpp"
#include "TemporaryClauseCollector.hpp"
#include "ForwardProcessing.hpp"
#include "Statistics.hpp"
#include "prefix.hpp"
#include "Clause.inl"
using namespace BK;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "BACKWARD_SUPERPOSITION"
#endif
#include "debugMacros.hpp"
//================================================


VK::BACKWARD_SUPERPOSITION::BACKWARD_SUPERPOSITION(OR_INDEX* unifIndex) : 
  OPEN_BR(1,"br 1") DCOMMA 
  // main parameters
  _statistics(0),
  _clauseCollector(0),
  _forwardProcessing(0),
  _unifIndex(unifIndex),
  _simultaneousness(0L),
  _noInfBetweenRulesFlag(false),
  _paramodulateFromVariables(true),
  _mainPremiseSubsumption(false),
  _sidePremiseSubsumption(false),
  _mainSelectedLitDemodulation(false),
  _sideSelectedLitDemodulation(false),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),

  
  // main premise data
  _mainPremiseVarIndex(0UL),

  // side premise data
  _sidePremiseVarIndex(1UL), 
  
  // misc
  _dummyClauseCollector(DOP("BACKWARD_SUPERPOSITION::_dummyClauseCollector"))
{
  _mainPremiseResidualSizeFormula.SetIndex(0UL);
  _mainPremiseSelectedLitResidualSizeFormula.SetIndex(0UL);

  _currentMainPosEqArg1Mem[0].Head().MkFunc(TERM((TERM::Arity)1,(TERM::Functor)0,Signature::PositivePolarity));
  _currentMainPosEqArg2Mem[0].Head().MkFunc(TERM((TERM::Arity)1,(TERM::Functor)0,Signature::PositivePolarity));
  
  _retrievalOfUnifiable.SetIndex(_sidePremiseVarIndex);
  _compiledConstraint = 
    SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
  _storedConstraintCheck = 
    SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
}; // BACKWARD_SUPERPOSITION::BACKWARD_SUPERPOSITION(OR_INDEX* unifIndex)

VK::BACKWARD_SUPERPOSITION::~BACKWARD_SUPERPOSITION() 
{ 
  CALL("destructor ~BACKWARD_SUPERPOSITION()"); 
    
  if (DestructionMode::isThorough())
    {
      ASSERT(_storedConstraintCheck);
      SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
      ASSERT(_compiledConstraint);
      SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
    };
};  // BACKWARD_SUPERPOSITION::~BACKWARD_SUPERPOSITION() 
  



bool VK::BACKWARD_SUPERPOSITION::SetQuery(Clause* cl) 
{ 
  CALL("SetQuery(Clause* cl)");
  if (cl->isSupportClause())
    {
      // just in case we decide to change 
      // the treatment of support clauses
      _orientPosEqualitiesInMainPremise = _orientPosEqualitiesInSupport;
    }
  else
    _orientPosEqualitiesInMainPremise = true;
         
  _noRulesAsSidePremises = _noInfBetweenRulesFlag && cl->IsRule();

  loadMainPremise(cl);
 try_next_pos_eq:
  _currentMainPosEqLit = findNextMainSelectedPosEq();
  if (!_currentMainPosEqLit) return false;
  if (!LoadEq(_currentMainPosEqLit,_mainPremiseVarIndex))
    goto try_next_pos_eq;
  return true; 
}; // bool VK::BACKWARD_SUPERPOSITION::SetQuery(Clause* cl) 

bool VK::BACKWARD_SUPERPOSITION::CollectResult()
{
  CALL("CollectResult()");
  if ((!_mainPremise->Active()) 
      || (!_sidePremise->Active()))
    return false;
      
  if (_mainPremiseResidualSize
      + _sidePremise->practicallyMinResidualSize()
      > _clauseSizeLimit)
    {
      ++(_statistics->inferences.deduction.backwardSuperposition.discardedBySize);
      return false;
    };            

  _clauseCollector->openNewClause(ClauseBackground::BackSup,
				  _mainPremise,
				  _sidePremise);
  _clauseCollector->announceNewClause();
  return
    collectSidePremiseLiterals() && 
    collectMainPremiseLiterals() && 
    _clauseCollector->finaliseClause();
}; // bool VK::BACKWARD_SUPERPOSITION::CollectResult()

bool VK::BACKWARD_SUPERPOSITION::NextInference()
{
 process_literal:
  while (Paramodulate())
    { 
      if (CollectResult()) return true;
    };
  // try next literal
 try_next_pos_eq:
  _currentMainPosEqLit = findNextMainSelectedPosEq();       
  if (!_currentMainPosEqLit) return false;
  if (!LoadEq(_currentMainPosEqLit,_mainPremiseVarIndex))
    goto try_next_pos_eq;
  goto process_literal;
}; // bool VK::BACKWARD_SUPERPOSITION::NextInference()


inline void VK::BACKWARD_SUPERPOSITION::compileMainPremiseResidualSizeFormula()
{
  CALL("compileMainPremiseResidualSizeFormula()");
  _mainPremiseResidualSizeFormula.Reset();

  for (LiteralList::Iterator iter(_allMainPremiseLiterals,
				  _firstNonselectedMainLit);
       iter.notEnd();
       iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      _mainPremiseResidualSizeFormula.Compile(iter.currentLiteral());
}; // void VK::BACKWARD_SUPERPOSITION::compileMainPremiseResidualSizeFormula()


inline void VK::BACKWARD_SUPERPOSITION::recompileMainPremiseSelectedLitResidualSizeFormula()
{
  CALL("recompileMainPremiseSelectedLitResidualSizeFormula()");
  _mainPremiseSelectedLitResidualSizeFormula.Reset();

  for (LiteralList::Iterator iter(_allMainPremiseLiterals); 
       iter.currentElement() != _firstNonselectedMainLit;
       iter.next())
    {
      ASSERT(!iter.currentLiteral()->isAnswerLiteral());
      if (iter.currentElement() == _currentSelectedLitListElement)
	{              
	  ASSERT(iter.currentElement()->isEquality());
	  ASSERT(iter.currentElement()->positive());
	  if (iter.currentElement()->isOrderedEq())
	    _mainPremiseSelectedLitResidualSizeFormula.Compile(*(iter.currentElement()->arg2()));
	}
      else 
	_mainPremiseSelectedLitResidualSizeFormula.Compile(iter.currentElement()->literal());
    }; 
}; // void VK::BACKWARD_SUPERPOSITION::recompileMainPremiseSelectedLitResidualSizeFormula()


inline void VK::BACKWARD_SUPERPOSITION::loadMainPremise(Clause* cl) 
{
  CALL("loadMainPremise(Clause* cl)");
  _mainPremise = cl;
  _allMainPremiseLiterals = cl->LitList();
  _firstNonselectedMainLit = _allMainPremiseLiterals->nth(_allMainPremiseLiterals->numOfSelectedLits());
  _mainSelectedLitIter.reset(_allMainPremiseLiterals);
  compileMainPremiseResidualSizeFormula();
}; // void VK::BACKWARD_SUPERPOSITION::loadMainPremise(Clause* cl) 


inline const VK::TERM* VK::BACKWARD_SUPERPOSITION::findNextMainSelectedPosEq()
{
  CALL("findNextMainSelectedPosEq()");
  const TERM* lit;
 check_lit:
  if (_mainSelectedLitIter.currentElement() == _firstNonselectedMainLit) return 0;  
  _currentSelectedLitListElement = _mainSelectedLitIter.currentElement();
  lit = _mainSelectedLitIter.currentLiteral();
  if (lit->Positive() && 
      lit->IsEquality() &&
      _mainSelectedLitIter.currentElement()->properties().useEqForParamodulationFrom()) 
    {
      _mainSelectedLitIter.next();
      recompileMainPremiseSelectedLitResidualSizeFormula();
      return lit;
    }; 
  _mainSelectedLitIter.next();
  goto check_lit;
}; //  const TERM* VK::BACKWARD_SUPERPOSITION::findNextMainSelectedPosEq()


inline bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseLiterals() 
{
  CALL("collectMainPremiseLiterals()");
  // collect max. literals except the used equality
  LiteralProperties::SLModifier slmod = LiteralProperties::DefaultSLModifier;
  
  if (_currentSelectedLitListElement->properties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (_currentSelectedLitListElement->properties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  slmod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(_currentSelectedLitListElement->properties().deploymentMode() == 
		 LiteralProperties::Subquery);
	  slmod = LiteralProperties::PostponeSelection;
	};
    };

  LiteralProperties prop;
  for (LiteralList::Iterator iter(_allMainPremiseLiterals); 
       iter.currentElement() != _firstNonselectedMainLit;
       iter.next())
    if (iter.currentElement() != _currentSelectedLitListElement)
      {  
	prop.inheritFrom(iter.currentElement()->properties(),slmod);
	if (!_clauseCollector->pushLiteralWithUnifier(_mainPremiseVarIndex,
						      iter.currentLiteral(),
						      prop)) 
	  return false;
      };      

  // collect nonmax. literals
  for (LiteralList::Iterator iter1(_allMainPremiseLiterals,_firstNonselectedMainLit);
       iter1.notEnd();
       iter1.next())
    {
      prop.inheritFrom(iter1.currentElement()->properties(),slmod);
      if (!_clauseCollector->pushLiteralWithUnifier(_mainPremiseVarIndex,
						    iter1.currentLiteral(),
						    prop)) 
	return false;   
    };  
  return true;
}; // bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseLiterals() 




inline bool VK::BACKWARD_SUPERPOSITION::LoadEq(const TERM* eq_lit,ulong eq_ind)
{
  CALL("LoadEq(const TERM* eq_lit,ulong eq_ind)");
  ASSERT(eq_lit->IsEquality());
  ASSERT(eq_lit->Positive());
  _mainPremiseVarIndex = eq_ind; 
  const TERM* arg_ref = eq_lit->Args();
  PrefixSym* mem = _currentMainPosEqArg1Mem + 1;
  Prefix::StandardTermToPrefix(_mainPremiseVarIndex,*arg_ref,mem);
  _currentMainPosEqArg1Mem[0].SetBracket(mem);
     
  // another argument
  arg_ref = arg_ref->Next();
  mem = _currentMainPosEqArg2Mem + 1;
  Prefix::StandardTermToPrefix(_mainPremiseVarIndex,*arg_ref,mem);
  _currentMainPosEqArg2Mem[0].SetBracket(mem);
     
  if (!LoadLHS(_currentMainPosEqArg1Mem))
    {
      ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedForOtherReasons);
      return false;
    };
  _currentMainPosEqRHS = _currentMainPosEqArg2Mem + 1;
  if (_orientPosEqualitiesInMainPremise)
    {
      _currentMainPosEqIsUnordered = eq_lit->IsUnorderedEq();
    }
  else
    _currentMainPosEqIsUnordered = true;

  _swapArguments = _currentMainPosEqIsUnordered;

  _currentCompiledConstraint = 0;
  if (_currentMainPosEqIsUnordered)
    {
      if (_orientPosEqualitiesInMainPremise)
	{
	  _compiledConstraint->load(_currentMainPosEqArg2Mem,_currentMainPosEqArg1Mem);
	  if (_compiledConstraint->rhsVarInstMayOrder())
	    {
	      _compiledConstraint->compileGreaterOrEqual();
	      _currentCompiledConstraint = _compiledConstraint;
	    };
	};
    };
  return true;
}; //  bool VK::BACKWARD_SUPERPOSITION::LoadEq(const TERM* eq_lit,ulong eq_ind)



inline bool VK::BACKWARD_SUPERPOSITION::Paramodulate()
{
  CALL("Paramodulate()");
 try_next_cand:
  if (NextCandidate(_currentCompiledConstraint,
		    _sidePremise,
		    _candLitRedexPair)) 
    {
      if (_noRulesAsSidePremises && _sidePremise->IsRule())
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.discardedForOtherReasons);
	  goto try_next_cand;
	};
      if (demodulateSelectedLits())
	goto try_next_cand;

      if (subsumeSidePremise()) goto try_next_cand;
      return true;
    };
  if (_swapArguments)
    {
      _swapArguments = false;
      if (!LoadLHS(_currentMainPosEqArg2Mem))
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedForOtherReasons);
	  return false;
	};
      _currentMainPosEqRHS = _currentMainPosEqArg1Mem + 1;


      _currentCompiledConstraint = 0;
      _compiledConstraint->load(_currentMainPosEqArg1Mem,_currentMainPosEqArg2Mem);
      if (_compiledConstraint->rhsVarInstMayOrder())
	{
	  _compiledConstraint->compileGreaterOrEqual();
	  _currentCompiledConstraint = _compiledConstraint;
	};
      goto try_next_cand;          
    };
  return false; 
}; // bool VK::BACKWARD_SUPERPOSITION::Paramodulate()

inline bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseLiterals() 
{
  CALL("collectSidePremiseLiterals()");

  const SimplificationOrdering::StoredConstraint* constr = 
    _candLitRedexPair->constraint();
  if (constr) 
    {
      if (_storedConstraintCheck->holds(constr,_sidePremiseVarIndex))
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.redundantDiscarded);
	  return false;
	};
    };
    
  ulong rewritten_lit_num = _candLitRedexPair->litNum();

  LiteralList::Element* rewritten_lit = 
    _sidePremise->LitList()->nth(rewritten_lit_num);
 
  LiteralProperties::SLModifier rewLitSLMod =
    LiteralProperties::DefaultSLModifier;
 
  LiteralProperties::SLModifier slmod =
    LiteralProperties::DefaultSLModifier;
  
  if (rewritten_lit->properties().deploymentMode() != 
      LiteralProperties::DefaultDeploymentMode)
    {
      if (rewritten_lit->properties().deploymentMode() == 
	  LiteralProperties::RuleHead)
	{
	  slmod = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(rewritten_lit->properties().deploymentMode() == 
		 LiteralProperties::Subquery);
	  slmod = LiteralProperties::PostponeSelection;
	};

    };

  LiteralProperties prop;
  for (LiteralList::Iterator iter(_sidePremise->LitList()); 
       iter.notEnd(); 
       iter.next())
    {
      if (iter.currentElement() == rewritten_lit)
	{        
	  prop.inheritFrom(iter.currentElement()->properties(),
			   rewLitSLMod);
	  if (!_clauseCollector->pushLiteralWithReplacement(_sidePremiseVarIndex,
							    iter.currentLiteral(),
							    _candLitRedexPair->redex(),
							    _currentMainPosEqRHS,
							    prop))
	    return false;
	}
      else
	{
	  if ((_simultaneousness > 0) && (iter.currentElement()->arity()))
	    {
	      prop.inheritFrom(iter.currentElement()->properties(),slmod);
	      if (!_clauseCollector->pushLiteralWithReplacement(_sidePremiseVarIndex,
								iter.currentLiteral(),
								_candLitRedexPair->redex(),
								_currentMainPosEqRHS,
								prop))
		return false;   
	    }
	  else
	    {
	      prop.inheritFrom(iter.currentElement()->properties(),slmod);
	      if (!_clauseCollector->pushLiteralWithUnifier(_sidePremiseVarIndex,
							    iter.currentElement()->literal(),
							    prop)) 
		return false;  
	    };
	};  
    };     

  return true;
};  // bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseLiterals()

bool VK::BACKWARD_SUPERPOSITION::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");

  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (_sidePremise->hasSubsumedPremiseInstance()) return false;
  _sidePremise->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);

  if (collectSidePremiseInstance()  && 
      _dummyClauseCollector.finaliseClause(_dummyClause)) 
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _sidePremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
  
 return_false:
  _sidePremise->unblockForSubsumption();
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.backwardSuperposition.sidePremisesSubsumed);
 discarded:
  _sidePremise->unblockForSubsumption();
  return true;
}; // bool VK::BACKWARD_SUPERPOSITION::subsumeSidePremise()

bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");

  for(LiteralList::Iterator iter(_sidePremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties())) 
      return false; 
  return true;

}; // bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseInstance()



bool VK::BACKWARD_SUPERPOSITION::LoadLHS(PrefixSym* t)
{
  CALL("LoadLHS(PrefixSym* t)");
  if (!_retrievalOfUnifiable.CompileComplexTerm(t))
    return false;
  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_QUERY(t->Arg1());
  _newRetrievalAttempt = true;
  _isFromVariable = t->Arg1()->Head().isVariable();
  return true;
}; // bool VK::BACKWARD_SUPERPOSITION::LoadLHS(PrefixSym* t)
            
            
inline 
bool 
VK::BACKWARD_SUPERPOSITION::NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr,
			  Clause*& candCl,
			  BackParaCandidates::LitRedexPair*& candLitRedexPair)
{
  CALL("NextCandidate(..)"); 

  if (_newRetrievalAttempt)
    {
      _newRetrievalAttempt = false;
      if ((!_paramodulateFromVariables) && _isFromVariable)
	{
	  ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedWithParamodulationsFromVariables);
	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_TOTAL;
	  return false;
	};

      if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->BParaTree())))
	{	       
	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_RESULT(_retrievalOfUnifiable.CurrentLeaf()->Term().First());

	  computeMainPremiseResidualWeight();
	  if (_mainPremiseResidualSize > _clauseSizeLimit)
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedBySize);
	      goto try_next_leaf;
	    };

	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) 
	    {
	       ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	      goto try_next_leaf;
	    };
	  if (subsumeMainPremise()) goto try_next_leaf;
	  _candidates.reset(*(_retrievalOfUnifiable.BPCandidatesFromLeaf()));            
	  ALWAYS(_candidates.next(candCl,candLitRedexPair));
	  return true;
	}
      else 
	{
	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_TOTAL;
	  return false;
	};
    }
  else // !_newRetrievalAttempt
    {
      if (_candidates.next(candCl,candLitRedexPair)) 
	{
	  return true;
	};
    try_next_leaf: 

      if (_retrievalOfUnifiable.FindNextLeaf())
	{	        

	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_RESULT(_retrievalOfUnifiable.CurrentLeaf()->Term().First());

	  computeMainPremiseResidualWeight();
	  if (_mainPremiseResidualSize > _clauseSizeLimit)
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.branchesDiscardedBySize);
	      goto try_next_leaf;
	    };

	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) 
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	      goto try_next_leaf;
	    };
	  if (subsumeMainPremise()) goto try_next_leaf;
	  _candidates.reset(*(_retrievalOfUnifiable.BPCandidatesFromLeaf()));            
	  ALWAYS(_candidates.next(candCl,candLitRedexPair));
	  return true;
	}
      else 
	{
	  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_RETRIEVAL_TOTAL;
	  return false;  
	};
    };     
}; // bool VK::BACKWARD_SUPERPOSITION::NextCandidate(..) 


bool VK::BACKWARD_SUPERPOSITION::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");

  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_mainPremise->hasSubsumedPremiseInstance()) return false;
  _mainPremise->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);

  if (collectMainPremiseInstance() && 
      _dummyClauseCollector.finaliseClause(_dummyClause))   
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != _mainPremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {      
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{	
	  ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
  
 return_false:
  _mainPremise->unblockForSubsumption();
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.backwardSuperposition.mainPremisesSubsumed);
 discarded:
  _mainPremise->unblockForSubsumption();
  return true;
}; // bool VK::BACKWARD_SUPERPOSITION::subsumeMainPremise()

bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");

  for(LiteralList::Iterator iter(_mainPremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
						      iter.currentLiteral(),
						      iter.currentElement()->properties())) 
      return false; 
  return true;
}; // bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseInstance()




bool VK::BACKWARD_SUPERPOSITION::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");
  if (_sidePremise->paramodulationsIntoProtected() &&
      _mainPremise->Unit())
    return false; // paramodulations into this clause are protected

  Clause* simplifyingClause;
  if (_mainSelectedLitDemodulation)
    {   
      if (_mainSelectedLitDemodulationAttempted)
	{
	  if (_mainSelectedLitDemodulated)
	    {
	      ++(_statistics->inferences.deduction.backwardSuperposition.mainPremisesDemodulated);
	      return true;
	    };
	  // proceed with the side premise
	}
      else
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  _mainPremise->blockForDemodulationFrom();
	  _sidePremise->blockForDemodulationFrom();
          _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto() && 
	      _dummyClauseCollector.finaliseClause(_dummyClause))
	    {		  
	      simplifyingClause = 
		_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);     		      
	      ASSERT(simplifyingClause != _mainPremise);
	      ASSERT(simplifyingClause != _sidePremise);
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      
	      if (simplifyingClause)
		{	
		  _mainPremise->protectParamodulationsInto();
		  _sidePremise->protectParamodulationsInto();
		  ++(_statistics->inferences.deduction.backwardSuperposition.mainPremisesDemodulated);
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	      // proceed with the side premise
	    }
	  else // analyse the result
	    {
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      bool tautology = 
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::PropositionalTautology) ||
		(_dummyClauseCollector.rejectionReason() == 
		 TemporaryClauseCollector::EquationalTautology);
	      if (tautology)
		++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	      return tautology;
	    };
	};
    };

  if (_sideSelectedLitDemodulation)
    {
      _mainPremise->blockForDemodulationFrom();
      _sidePremise->blockForDemodulationFrom();
      _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified); 

      if (collectSidePremiseSubtermsAvailableForParamodulationInto()&&
	  _dummyClauseCollector.finaliseClause(_dummyClause))
	{		    	      
	  simplifyingClause = 
	    _forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause); 
	  ASSERT(simplifyingClause != _mainPremise);
	  ASSERT(simplifyingClause != _sidePremise);
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	      
	  if (simplifyingClause)
	    {	
	      _mainPremise->protectParamodulationsInto();
	      _sidePremise->protectParamodulationsInto();
	      ++(_statistics->inferences.deduction.backwardSuperposition.sidePremisesDemodulated);
	      return true;
	    }
	  else
	    return false;
	}
      else // analyse the result
	{
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  bool tautology =
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::PropositionalTautology) ||
	    (_dummyClauseCollector.rejectionReason() == 
	     TemporaryClauseCollector::EquationalTautology);
	  if (tautology)
	    ++(_statistics->inferences.deduction.backwardSuperposition.redundantBranchesDiscarded);
	  return tautology;
	};
    }
  else
    return false;
}; // bool VK::BACKWARD_SUPERPOSITION::demodulateSelectedLits()




bool 
VK::BACKWARD_SUPERPOSITION::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()"); 
  LiteralProperties prop;
  bool atLeastOneLiteralCollected = false;
  for(LiteralList::Iterator iter(_allMainPremiseLiterals);
      iter.currentElement() != _firstNonselectedMainLit;
      iter.next())
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_mainPremiseVarIndex,
								      TERM::dummyPositiveHeader1(),
								      *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_mainPremiseVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
    };
  return atLeastOneLiteralCollected;
}; // bool VK::BACKWARD_SUPERPOSITION::collectMainPremiseSubtermsAvailableForParamodulationInto()



bool 
VK::BACKWARD_SUPERPOSITION::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  LiteralProperties prop;
  ulong numOfSelectedLits = _sidePremise->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_sidePremise->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_dummyClauseCollector.pushPseudoLiteralWithUnifier(_sidePremiseVarIndex,
					     TERM::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
							      lit,
							      prop))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_dummyClauseCollector.pushLiteralWithUnifier(_sidePremiseVarIndex,
							      lit,
							      prop))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; // bool VK::BACKWARD_SUPERPOSITION::collectSidePremiseSubtermsAvailableForParamodulationInto()





//==================================================================
