// Revised: May 19, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//          Two more ways of estimating weight reachability added:
//          definitelyReachable(..) and potentiallyReachable(..).
// Revised: May 21, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//          Bug fix in Category(long weight).
//======================================================================

#ifndef KEPT_H
//======================================================================
#define KEPT_H
#include <climits>
#include <iostream>
#include "jargon.hpp"
#include "Math.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Clause.inl"

//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_KEPT
#define DEBUG_NAMESPACE "KEPT"
#endif
#include "debugMacros.hpp"
//====================================================
namespace VK
{

const long NUMBER_OF_KEPT_CATEGORIES = VampireKernelConst::MaxClauseSize + 1;

class KEPT
{
 public:
  KEPT() : 
    _ageWeightRatioNumerator(1),
    _ageWeightRatioDenominator(0),
    _bestGoodWeightNumerator(1),
    _bestGoodWeightDenominator(0),
    _magicCoefficient(0.0)
  { 
    Reset(); 
  };
  ~KEPT() { CALL("destructor ~KEPT()"); }; 
  

  void Reset()
    {
      _maxPriorityFlag = false;
      num_of_currently_kept = 0L;
      num_of_used = 0L;
      num_of_removed = 0L;
      num_of_cancelled = 0L;
      numOfReachable = LONG_MAX;
      for (long cat = 0; cat < NUMBER_OF_KEPT_CATEGORIES; cat++) StatReset(cat);
      _minNonemptyCatNum = 0;
      _maxNonemptyCatNum = 0;
      max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
      _totalSelectedByAgeOrWeight = 0;
      _totalSelectedByWeight = 0;
    };
  void DestroyAllClauses();
  void setAgeWeightRatio(long num,long denom) 
    {
      _ageWeightRatioNumerator = num;
      _ageWeightRatioDenominator = denom;
      if (_ageWeightRatioNumerator < 0) _ageWeightRatioNumerator = 0;
      if (_ageWeightRatioDenominator < 0) _ageWeightRatioDenominator = 0;
      if (_ageWeightRatioNumerator == 0 && _ageWeightRatioDenominator == 0) 
	_ageWeightRatioNumerator = 1;
      if (_ageWeightRatioNumerator != 0 && _ageWeightRatioDenominator != 0) 
	{
	  long gcd = 
	    BK::Math::greatestCommonDivisor(_ageWeightRatioNumerator,_ageWeightRatioDenominator);
	  _ageWeightRatioNumerator = _ageWeightRatioNumerator/gcd;
	  _ageWeightRatioDenominator = _ageWeightRatioDenominator/gcd;
	};
      if (_ageWeightRatioDenominator > 0) 
	_magicCoefficient = 
	  1.0 + 
	  (((float)_ageWeightRatioNumerator)/((float)_ageWeightRatioDenominator));
    };
  void setBestGoodWeightRatio(long numerator,long denominator)
  {
    _bestGoodWeightNumerator = numerator;
    _bestGoodWeightDenominator = denominator;
    if (_bestGoodWeightNumerator <= 0L) _bestGoodWeightNumerator = 1L;
    if (_bestGoodWeightDenominator < 0L) _bestGoodWeightDenominator = 0L;
  };
  void MaxPriorityOn() { _maxPriorityFlag = true; };
  void MaxPriorityOff() { _maxPriorityFlag = false; }; 
  bool isEmpty() const { return !num_of_currently_kept; };
  long Size() const { return num_of_currently_kept; };
  long Total() const { return num_of_currently_kept + num_of_used + num_of_removed + num_of_cancelled; };
  long CurrentlyKept() const { return num_of_currently_kept; };
  long NumOfReachable() const { return numOfReachable; };
  long UsedAndRemoved() const { return (num_of_used + num_of_removed); };
  bool MaxPriorityEmpty() { return EmptyCat(0); };
  long UnreachableKept() const
    {
      return (num_of_currently_kept > numOfReachable)? (num_of_currently_kept - numOfReachable) : 0; 
    };
  void SetNumOfReachable1(long n) 
    {
      CALL("SetNumOfReachable1(const long n)");
      ASSERT(CheckStatistics());    
      numOfReachable = n; 
      CorrectMaxAcceptedCat1();
    };  
  void SetNumOfReachable2(long n) 
    {
      CALL("SetNumOfReachable2(const long n)");
      ASSERT(CheckStatistics());    
      numOfReachable = n; 
      CorrectMaxAcceptedCat2();
    };
  void SetNumOfReachable3(long n) 
    {
      CALL("SetNumOfReachable3(const long n)");
      ASSERT(CheckStatistics());    
      numOfReachable = n; 
      CorrectMaxAcceptedCat3();
    };
  long MaxAcceptedWeight() const
    {
      if (num_of_currently_kept < numOfReachable) return VampireKernelConst::MaxClauseSize; 
      if (_ageWeightRatioDenominator > 0) { return Weight(max_accepted_cat); }
      else return 0L; // ??
    };
  long MaxPresentWeight() const { return Weight(_maxNonemptyCatNum); };
  long MinPresentWeight() const { return Weight(_minNonemptyCatNum); };

  void Cancel(Clause* c) 
  { 
    CALL("Cancel(Clause* c)");
    Clause::Remove(c); 
    StatCancelled(Category(c)); 
  };
  void Remove(Clause* c) 
    { 
      CALL("Remove(Clause* c)");
      Clause::Remove(c); 
      StatRemoved(Category(c)); 
    };
  void Insert(Clause* c) 
    { 
      if (_maxPriorityFlag) 
	{ 
	  c->AssignMaxPriority();
	};
      InsertAs(c,Category(c)); 
    };
  Clause* Next();
  Clause* TheWorstClause()
    {
      CALL("TheWorstClause()");
      for (long cat = _maxNonemptyCatNum; cat >= _minNonemptyCatNum; cat--) 
	if (!(EmptyCat(cat))) return _category[cat].Previous();
      return 0;
    };

  Clause* TheWorstUnblockedClause()
    {
      CALL("TheWorstUnblockedClause()");
      for (long cat = _maxNonemptyCatNum; cat >= _minNonemptyCatNum; cat--) 
	if (!(EmptyCat(cat))) 
	  for (Clause* cand = _category[cat].Previous(); 
	       cand != &(_category[cat]); 
	       cand = cand->Previous())
	    if (!cand->isProtected()) 
	      {
		return cand;
	      };
      return 0;   
    };
  
  Clause* KEPT::theWorstUnblockedClauseAfter(Clause* cl);
  Clause* TheMostUnreachableClause()
    {
      CALL("TheMostUnreachableClause()");
      if (!UnreachableKept()) return 0;
      long cat;
      ulong youngestNumber = 0UL;
      for (cat = _maxNonemptyCatNum;
	   (cat >= _minNonemptyCatNum) && (!definitelyReachable(cat));
	   --cat)
	{
	  if (!(EmptyCat(cat))) 
	    {
	      if (_category[cat].Previous()->Number() > youngestNumber)
		youngestNumber = _category[cat].Previous()->Number();
	    };
	};
      ++cat;
      if (!youngestNumber) return 0;
      while (cat <= _maxNonemptyCatNum)
	{
	  if (!(EmptyCat(cat))) 
	    {
	      if (_category[cat].Previous()->Number() == youngestNumber)
		return _category[cat].Previous();
	    };
	  ++cat;
	};
      ICP("END");
      return 0;
    }; // Clause* TheMostUnreachableClause()



  void ResetIterator()
    {
      if (EmptyCat(_minNonemptyCatNum)) { next_clause = 0; return; };
      iter_cat = _minNonemptyCatNum;
      next_clause = _category[iter_cat].Next();
    };
  Clause* IterNext() 
    {
      Clause* res = next_clause;     
      if (res)
	{
	  next_clause = res->Next();
	  if (next_clause == &(_category[iter_cat]))
	    {
	      // try another category
	      iter_cat++;       
	      while ((iter_cat <= _maxNonemptyCatNum) && (EmptyCat(iter_cat))) { iter_cat++; };
	      if (iter_cat > _maxNonemptyCatNum) 
		{ 
		  next_clause = 0; 
		}
	      else { next_clause = _category[iter_cat].Next(); };
	    };
	};
      return res;
    }; // Clause* IterNext() 

  ostream& outputProfile(ostream& str) const;
 
 private:   
  void NormalizeMinMaxCat(long lb,long rb)
    {
      _minNonemptyCatNum = lb;
      _maxNonemptyCatNum = rb;
      while ((EmptyCat(_minNonemptyCatNum)) && (_minNonemptyCatNum < _maxNonemptyCatNum)) _minNonemptyCatNum++;
      while ((EmptyCat(_maxNonemptyCatNum)) && (_minNonemptyCatNum < _maxNonemptyCatNum)) _maxNonemptyCatNum--;
    };
  
  void StatReset(long cat) 
    {
      currently_kept[cat] = 0L;
      used[cat] = 0L;
      removed[cat] = 0L;
    };
  
  void StatInserted(long cat)
    {
      currently_kept[cat]++;
      num_of_currently_kept++;    
      if (cat > _maxNonemptyCatNum) { NormalizeMinMaxCat(_minNonemptyCatNum,cat); }
      else
	if (cat < _minNonemptyCatNum) NormalizeMinMaxCat(cat,_maxNonemptyCatNum);
    };
   
  void StatUsed(long cat)
    {
      used[cat]++;
      currently_kept[cat]--;
      num_of_used++;     
      num_of_currently_kept--;
      numOfReachable--; // experimental
      if (EmptyCat(cat) && ((cat == _minNonemptyCatNum) || (cat == _maxNonemptyCatNum)))
	NormalizeMinMaxCat(_minNonemptyCatNum,_maxNonemptyCatNum);
    };
   
  void StatRemoved(long cat)
    {
      removed[cat]++;
      currently_kept[cat]--;
      num_of_removed++;
      num_of_currently_kept--;
      numOfReachable--; // experimental
      if (EmptyCat(cat) && ((cat == _minNonemptyCatNum) || (cat == _maxNonemptyCatNum)))
	NormalizeMinMaxCat(_minNonemptyCatNum,_maxNonemptyCatNum);
    };
  
  void StatCancelled(long cat)
    {
      num_of_cancelled++;
      currently_kept[cat]--;
      num_of_currently_kept--;
      if (EmptyCat(cat) && ((cat == _minNonemptyCatNum) || (cat == _maxNonemptyCatNum)))
	NormalizeMinMaxCat(_minNonemptyCatNum,_maxNonemptyCatNum);     
    };

  long Total(long cat) const { return currently_kept[cat] + used[cat] + removed[cat]; };
 
  long NumOfLighter(long cat)
    {
      long res = 0L;
      for (long c = 0; c < cat; c++) res += currently_kept[c]; 
      return res;
    };
   
  long NumOfHeavier(long cat)
    {
      long res = 0L;
      for (long c = cat + 1; c <= _maxNonemptyCatNum; c++) res += currently_kept[c];
      return res;
    };


  // old, well tuned version:
  bool Reachable(long cat)
    {
      CALL("Reachable(long cat)");
      if (numOfReachable > num_of_currently_kept) return true;   
      long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
      if (!num_of_lighter_or_same_weight) return true; 
      if (numOfReachable <= num_of_lighter_or_same_weight) return false;
      long magic_number = (long)(num_of_lighter_or_same_weight * _magicCoefficient);
      return (numOfReachable - (num_of_currently_kept - numOfReachable) > magic_number);
    };


  bool definitelyReachable(long cat)
    {
      CALL("definitelyReachable(long cat)");   
      if (numOfReachable > num_of_currently_kept) return true; 
      long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
      if (!num_of_lighter_or_same_weight) return true; 
      if (numOfReachable <= num_of_lighter_or_same_weight) return false;
      long magic_number = (long)(num_of_lighter_or_same_weight * _magicCoefficient);
      DOP(bool debugRes = (magic_number < numOfReachable));
      ASSERT((!debugRes) || potentiallyReachable(cat));
      // Definitely rechable provided that magic_number < numOfReachable,
      // even if some reachable clauses are of greater weight.
      return (magic_number < numOfReachable); 
    }; // bool definitelyReachable(long cat)


  bool potentiallyReachable(long cat)
    {
      CALL("potentiallyReachable(long cat)");   
      if (numOfReachable > num_of_currently_kept) return true; 
      long num_of_lighter_or_same_weight = NumOfLighter(cat) + currently_kept[cat];
      if (!num_of_lighter_or_same_weight) return true; 
      if (numOfReachable <= num_of_lighter_or_same_weight) return false;

      // May be reachable if all reachable clauses are of the same
      // or smaller weight. 
      return true; 
    }; // bool potentiallyReachable(long cat)
  

 
  void CorrectMaxAcceptedCat1();
  void CorrectMaxAcceptedCat2();
  void CorrectMaxAcceptedCat3();

  long Category(long weight) 
    { 
      if (_ageWeightRatioDenominator > 0)
	{
	  if (weight < 0L) return 0L;
	  if (weight >= NUMBER_OF_KEPT_CATEGORIES) return NUMBER_OF_KEPT_CATEGORIES - 1;
          return weight;
	};
      return 0L;
    };

  long Weight(long cat) const { return cat; };
   
  long Category(Clause* c) 
    {
      if (c->MaxPriority()) return 0; 
      return Category(c->weight());
    };
  void InsertAs(Clause* c,long cat) { Clause::InsertBefore(c,&(_category[cat])); StatInserted(cat); };
  bool EmptyCat(long cat) const { return (_category[cat].Next() == &(_category[cat])); };
         
 public: // for debugging
#ifndef NO_DEBUG
  bool clean() const;
  long CountCurrentlyKept(long cat);
  bool CheckStatistics();
#endif 

 private: // main data
  bool _maxPriorityFlag;
  long _ageWeightRatioNumerator; // <=0 if no selection by age
  long _ageWeightRatioDenominator; // <=0 if no selection by weight
  long _bestGoodWeightNumerator;
  long _bestGoodWeightDenominator;
  float _magicCoefficient; // must be = 1 + (_ageWeightRatioNumerator/_ageWeightRatioDenominator);
  long _totalSelectedByAgeOrWeight;
  long _totalSelectedByWeight;
  Clause _category[NUMBER_OF_KEPT_CATEGORIES];
  long _minNonemptyCatNum;
  long _maxNonemptyCatNum;
  long max_accepted_cat;
 private: // statistics
  long currently_kept[NUMBER_OF_KEPT_CATEGORIES];
  long used[NUMBER_OF_KEPT_CATEGORIES];
  long removed[NUMBER_OF_KEPT_CATEGORIES];
  long num_of_currently_kept;
  long num_of_used;
  long num_of_removed;
  long num_of_cancelled;
  long numOfReachable;

 private: // iterator
  Clause* next_clause;
  long iter_cat; 
}; // class KEPT

}; // namespace VK
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_KEPT
#define DEBUG_NAMESPACE "KEPT"
#endif
#include "debugMacros.hpp"
//====================================================

namespace VK
{
inline void KEPT::CorrectMaxAcceptedCat1()
{
  if (num_of_currently_kept < numOfReachable) 
    {
      max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
      return;
    };
  if (definitelyReachable(max_accepted_cat))
    {
      while (definitelyReachable(max_accepted_cat + 1))
	{
	  max_accepted_cat++;
	  if (max_accepted_cat > _maxNonemptyCatNum) 
	    {
	      max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
	      return;
	    };
	};
    }
  else
    {
      do
	{
	  max_accepted_cat--;
	}
      while ((max_accepted_cat > 0) && (!(definitelyReachable(max_accepted_cat))));
    };     
}; // void KEPT::CorrectMaxAcceptedCat1()


inline void KEPT::CorrectMaxAcceptedCat2()
{
  if (num_of_currently_kept < numOfReachable) 
    {
      max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
      return;
    };
  if (potentiallyReachable(max_accepted_cat))
    {
      while (potentiallyReachable(max_accepted_cat + 1))
	{
	  max_accepted_cat++;
	  if (max_accepted_cat > _maxNonemptyCatNum) 
	    {
	      max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
	      return;
	    };
	};
    }
  else
    {
      do
	{
	  max_accepted_cat--;
	}
      while ((max_accepted_cat > 0) && (!(potentiallyReachable(max_accepted_cat))));
    };     
}; // void KEPT::CorrectMaxAcceptedCat2()

inline void KEPT::CorrectMaxAcceptedCat3()
{
  if (num_of_currently_kept < numOfReachable) 
    {
      max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
      return;
    };
  if (Reachable(max_accepted_cat))
    {
      while (Reachable(max_accepted_cat + 1))
	{
	  max_accepted_cat++;
	  if (max_accepted_cat > _maxNonemptyCatNum) 
	    {
	      max_accepted_cat = NUMBER_OF_KEPT_CATEGORIES-1;
	      return;
	    };
	};
    }
  else
    {
      do
	{
	  max_accepted_cat--;
	}
      while ((max_accepted_cat > 0) && (!(Reachable(max_accepted_cat))));
    };     
}; // void KEPT::CorrectMaxAcceptedCat3()

}; // namespace VK
//====================================================
#endif
