
/*
 * File XML.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifdef _MSC_VER
#  include <winsock2.h>
#  include <process.h>
#else 
#  include <unistd.h>
#endif

#include <ctime>

#include "Tracer.hpp"
#include "XML.hpp"
#include "VampireException.hpp"
#include "Options.hpp"
#include "VampireKernel.hpp"

namespace XML 
{



Attribute::Attribute (Type type, const string& name)
  : _type (type),
    _name (name),
    _next (0),
    _previous (0)
{
  TRACER("XML::Attribute::Attribute");
}


IntegerAttribute::IntegerAttribute (const string& name, int value)
  : Attribute (INTEGER, name),
    _value (value)
{
  TRACER("XML::IntegerAttribute::IntegerAttribute");
}


LongAttribute::LongAttribute (const string& name, long value)
  : Attribute (LONG, name),
    _value (value)
{
  TRACER("XML::LongAttribute::LongAttribute");
}


DoubleAttribute::DoubleAttribute (const string& name, double value)
  : Attribute (DOUBLE, name),
    _value (value)
{
  TRACER("XML::DoubleAttribute::DoubleAttribute");
}


TextAttribute::TextAttribute (const string& name, const string& text)
  : Attribute (TEXT, name),
    _value (text)
{
  TRACER("XML::TextAttribute::TextAttribute");
}


void DeepElement::destroyContent ()
{
  TRACER("XML::DeepElement::destroyContent");

  Content* current = _firstSubelement;
  while (current) {
    Content* next = current->_next;
    current->destroy();
    current = next;
  }
} 



void Element::write (ostream& str)
{
  TRACER("XML::Element::write");

  write(str,0);
} 


Element::Element (Type tp, const string& name)
  :
  Content (tp),
  _name (name),
  _firstAttribute (0),
  _lastAttribute (0)
{
  TRACER("XML::Element::Element");
}



Content::Content (Type ct)
  : _type (ct),
    _next (0),
    _previous (0)
{
  TRACER("XML::Content::Content");
} 



void Element::addAttribute(Attribute* att)
{
  TRACER("XML::Element::addAttribute");

  if (_lastAttribute) {
    _lastAttribute->setNext(att);
    att->setPrevious(_lastAttribute);
    _lastAttribute = att;
    return;
  }
  _firstAttribute = att;
  _lastAttribute = att;
} 



void DeepElement::addChild (Content* sub)
{
  TRACER("XML::DeepElement::addChild");

  if (_lastSubelement) {
    _lastSubelement->_next = sub;
    sub->_previous = _lastSubelement;
    _lastSubelement = sub;
    return;
  }
  _firstSubelement = sub;
  _lastSubelement = sub;
} 



TextElement::TextElement (const string& name, const string& text)
  : Element(TEXT,name),
    _text(text)
{
  TRACER("XML::TextElement::TextElement");
}


DeepElement::DeepElement (const string& name)
  : Element(DEEP, name),
    _firstSubelement(0),
    _lastSubelement(0)
{
  TRACER("XML::DeepElement::DeepElement");
}


FloatElement::FloatElement (const string& name, float value)
  : Element(FLOAT, name),
    _value(value)
{
  TRACER("XML::FloatElement::FloatElement");
}


LongElement::LongElement (const string& name, long value)
  : Element(LONG, name),
    _value(value)
{
  TRACER("XML::LongElement::LongElement");
}


IntegerElement::IntegerElement (const string& name, int value)
  : Element(INTEGER, name),
    _value(value)
{
  TRACER("XML::IntegerElement::IntegerElement");
}


EmptyElement::EmptyElement (const string& name)
  : Element(EMPTY, name)
{
  TRACER("XML::EmptyElement::EmptyElement");
}


Element* Element::generalProofAttemptInfo ()
{
  TRACER("XML::Element::generalProofAttemptInfo ()");

  DeepElement* attempt = new DeepElement("general");
  attempt->addChild(new TextElement("problem_name",
					 Options::globalOptions.problemName()));
  char hostname[256];
  gethostname(hostname,256);
  attempt->addChild(new TextElement("hostname",hostname));

  
  time_t jobTime = time(static_cast<time_t*>(0));
  tm* gmt = gmtime(&jobTime); 

  Element* date = new EmptyElement("date");
  date->addAttribute(new IntegerAttribute("day",gmt->tm_mday));
  date->addAttribute(new IntegerAttribute("month",gmt->tm_mon + 1));
  date->addAttribute(new IntegerAttribute("year",1900 + gmt->tm_year));
  attempt->addChild(date);

  Element* time = new EmptyElement("time");
  time->addAttribute(new IntegerAttribute("hour",gmt->tm_hour));
  time->addAttribute(new IntegerAttribute("min",gmt->tm_min));
  time->addAttribute(new IntegerAttribute("sec",gmt->tm_sec));
  attempt->addChild(time);

  DeepElement* version = new DeepElement("vampire_version");
  version->addChild(new FloatElement("number",
					  VampireKernel::versionNumber()));
  version->addChild(new TextElement("description",
					 VampireKernel::versionDescription()));
  version->addChild(new TextElement("date",__DATE__));
  version->addChild(new TextElement("time",__TIME__));
  attempt->addChild(version);

  return attempt;
} 



void Attribute::writeList (ostream& str)
{
  TRACER("XML::Attribute::writeList (ostream& str)");

  for (Attribute* attr = this; attr; attr = attr->_next) {
    str << ' ' << attr->_name << "=\"";
    attr->writeValue(str);
    str << '\"';
  }
} 



void TextAttribute::writeValue (ostream& str)
{
  str << _value;
} 



void IntegerAttribute::writeValue (ostream& str)
{
  str << _value;
} 



void LongAttribute::writeValue (ostream& str)
{
  str << _value;
} 



void DoubleAttribute::writeValue (ostream& str)
{
  str << _value;
} 



void DeepElement::writeContent(ostream& str, int indent)
{
  TRACER("XML::DeepElement::writeContent(ostream& str, int indent)");

  if (_firstSubelement) {
    str << '\n';
    for (Content* sub = _firstSubelement; sub; sub = sub->_next) {
      sub->write(str,indent+1);
    }
    writeSpaces(str,indent);
  }
} 


void TextElement::writeContent(ostream& str, int indent)
{
  str << _text;
} 


void FloatElement::writeContent(ostream& str, int indent)
{
  str << _value;
} 


void LongElement::writeContent(ostream& str, int indent)
{
  str << _value;
} 


void IntegerElement::writeContent(ostream& str, int indent)
{
  str << _value;
} 


void EmptyElement::writeContent(ostream& str, int indent)
{
} 



void Element::writeFooter(ostream& str)
{
  str << "</" << _name << '>';
} 


void Element::writeSpaces(ostream& str,int n)
{
  ASS(n >= 0);

  while (n > 0) {
    n--;
    str << ' ';
  }
} 



void Element::write (ostream& str, int indent)
{
  TRACER("XML::Element::write(ostream& str, int indent)");

  writeSpaces(str,indent);
  str << '<' << _name;
  _firstAttribute->writeList(str);
  writeEndOfHeader(str);
  str << '>';
  writeContent(str,indent);
  writeFooter(str);
  str << '\n';
} 


void EmptyElement::writeEndOfHeader (ostream& str)
{
  str << " /";
}



void Element::append (const string& file)
{
  TRACER("XML::Element::append");

  ofstream stream(file.c_str(),ios::app);
  CONTROL("X1");
  if (! stream) { 
    USER_ERROR("cannot open file XML file for appending");
  }
  CONTROL("X2");
  write(stream);
  CONTROL("X3");
} 


void Content::destroy ()
{
  TRACER("XML::Content::destroy");

  
  destroyAttributes();
  destroyContent();

  
  switch (_type) {
  case PCDATA:
    delete static_cast<PCData*>(this);
    return;

  case EMPTY:
    delete static_cast<EmptyElement*>(this);
    return;

  case TEXT:
    delete static_cast<TextElement*>(this);
    return;

  case FLOAT:
    delete static_cast<FloatElement*>(this);
    return;

  case LONG:
    delete static_cast<LongElement*>(this);
    return;

  case INTEGER:
    delete static_cast<IntegerElement*>(this);
    return;

  case DEEP:
    delete static_cast<DeepElement*>(this);
    return;
  }
}


void Element::destroyAttributes ()
{
  TRACER("XML::Element::destroyAttributes");

  _firstAttribute->destroy();
}



void Attribute::destroy ()
{
  TRACER("XML::Attribute::destroy");

  Attribute* current = this;
  while (current) {
    Attribute* next = current->_next;
    switch (current->_type) {
    case TEXT:
      delete static_cast<TextAttribute*>(current);
      break;

    case LONG:
      delete static_cast<LongAttribute*>(current);
      break;

    case DOUBLE:
      delete static_cast<DoubleAttribute*>(current);
      break;

    case INTEGER:
      delete static_cast<IntegerAttribute*>(current);
      break;
    }
    current = next;
  }
} 


void XML::Attribute::operator delete (void* obj)
{
  
  ASSERTION_VIOLATION;
}

void XML::Content::operator delete (void* obj)
{
  
  ASSERTION_VIOLATION;
}

void XML::Element::operator delete (void* obj)
{
  
  ASSERTION_VIOLATION;
}
}; 
